############################################################################################
## Hexaploid likelihood, LOD and recombination frequency functions
## Assumes random bivalent pairing. 
## Peter Bourke, Plant Breeding, Wageningen University and Research. January 2017
############################################################################################


#' Calculate recombination frequency, LOD and log-likelihood from frequency tables in a random pairing hexaploid
#' @description This group of functions is called by \code{\link{linkage}}.
#' @param x A frequency table of the different classes of dosages in the progeny. The column names start with \code{"n_"}. Followed by the dosage of the first marker and then of the second.
#' @return
#' A list with the following items:
#' \item{r_mat}{A matrix with recombination frequencies for the different phases}
#' \item{LOD_mat}{A matrix with LOD scores for the different phases}
#' \item{logL_mat}{A matrix with log likelihood ratios for the different phases}
#' \item{phasing_strategy}{A character string specifying the phasing strategy. \code{"MLL"} for maximum likelihood en \code{"MINR"} for minimum recombination frequency.}
#' \item{possible_phases}{The phases between markers that are possible. Same order and length as column names of output matrices.}
#' @name r6_functions
NULL

#' @rdname r6_functions 
#' @noRd
r6_1.0_1.0 <- function(x){
r_c <- (x[,"n_01"] + x[,"n_10"])/(x[,"n_00"] + x[,"n_01"] + x[,"n_10"] + x[,"n_11"])
logL_c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_10"] - x[,"n_11"])*log(2) + (x[,"n_00"] + x[,"n_11"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_10"])*log(pmax(1e-6,r_c))
LOD_c <- (x[,"n_01"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_11"])*log10(2) + (x[,"n_00"] + x[,"n_11"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_10"])*log10(pmax(1e-6,r_c))
r_r <- (3*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_10"] + 3*x[,"n_11"])/(x[,"n_00"] + x[,"n_01"] + x[,"n_10"] + x[,"n_11"])
logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_10"] - x[,"n_11"])*(log(2) + log(5)) + (x[,"n_01"] + x[,"n_10"])*log(3 - r_r) + (x[,"n_00"] + x[,"n_11"])*log(2 + r_r)
LOD_r <- -((x[,"n_01"] + x[,"n_10"])*(-log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_11"])*(-log10(2) + log10(5)) + (x[,"n_01"] + x[,"n_10"])*log10(3 - r_r) + (x[,"n_00"] + x[,"n_11"])*log10(2 + r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_1.1 <- function(x){
r_c <- (x[,"n_02"] + x[,"n_10"])/(x[,"n_00"] + x[,"n_02"] + x[,"n_10"] + x[,"n_12"])
logL_c <- 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"])*log(2) + (x[,"n_00"] + x[,"n_12"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_10"])*log(pmax(1e-6,r_c))
LOD_c <- (x[,"n_02"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_12"])*log10(2) + (x[,"n_00"] + x[,"n_12"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_10"])*log10(pmax(1e-6,r_c))
r_r <- (3*x[,"n_00"] - 2*x[,"n_02"] - 2*x[,"n_10"] + 3*x[,"n_12"])/(x[,"n_00"] + x[,"n_02"] + x[,"n_10"] + x[,"n_12"])
logL_r <- 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"])*log(2) + (-x[,"n_00"] - x[,"n_02"] - x[,"n_10"] - x[,"n_12"])*log(5) + (x[,"n_02"] + x[,"n_10"])*log(3 - r_r) + (x[,"n_00"] + x[,"n_12"])*log(2 + r_r)
LOD_r <- -((x[,"n_02"] + x[,"n_10"])*(-log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_12"])*(-log10(2) + log10(5)) + (x[,"n_02"] + x[,"n_10"])*log10(3 - r_r) + (x[,"n_00"] + x[,"n_12"])*log10(2 + r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_1.2 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n10,n11,n12,n13) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13)*(log(2) + log(5)) + (n01 + n12)*log(3 - 2*r) + (n00 + n13)*log(pmax(1e-6,1 - r)) + (n03 + n10)*log(pmax(1e-6,r)) + (n02 + n11)*log(1 + 2*r)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n10,n11,n12,n13) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"])
LOD_c <- (x[,"n_03"] + x[,"n_10"])*log10(2) - (x[,"n_02"] + x[,"n_11"])*log10(2) - (x[,"n_01"] + x[,"n_12"])*log10(2) + (x[,"n_00"] + x[,"n_13"])*log10(2) + (x[,"n_01"] + x[,"n_12"])*log10(3 - 2*r_c) + (x[,"n_00"] + x[,"n_13"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_03"] + x[,"n_10"])*log10(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_11"])*log10(1 + 2*r_c)
logL_c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"])*(log(2) + log(5)) + (x[,"n_01"] + x[,"n_12"])*log(3 - 2*r_c) + (x[,"n_00"] + x[,"n_13"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_03"] + x[,"n_10"])*log(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_11"])*log(1 + 2*r_c)
logL_r <- function(r,n00,n01,n02,n03,n10,n11,n12,n13) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13)*(log(2) + 2*log(5)) + (n02 + n11)*log(11 - 2*r) + (n03 + n10)*log(3 - r) + (n00 + n13)*log(2 + r) + (n01 + n12)*log(9 + 2*r)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n10,n11,n12,n13) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"])
LOD_r <- -((x[,"n_03"] + x[,"n_10"])*(-log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_13"])*(-log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_11"])*(log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_12"])*(log10(2) + log10(5)) + (x[,"n_02"] + x[,"n_11"])*log10(11 - 2*r_r) + (x[,"n_03"] + x[,"n_10"])*log10(3 - r_r) + (x[,"n_00"] + x[,"n_13"])*log10(2 + r_r) + (x[,"n_01"] + x[,"n_12"])*log10(9 + 2*r_r)
logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"])*(log(2) + 2*log(5)) + (x[,"n_02"] + x[,"n_11"])*log(11 - 2*r_r) + (x[,"n_03"] + x[,"n_10"])*log(3 - r_r) + (x[,"n_00"] + x[,"n_13"])*log(2 + r_r) + (x[,"n_01"] + x[,"n_12"])*log(9 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_1.3 <- function(x){
logL_c <- function(r,n02,n12,n00,n01,n03,n04,n10,n11,n13,n14) {
L <- 2*(n02 + n12)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14)*(3*log(2) + log(5)) + (n01 + n13)*log(9 - 8*r) + (n00 + n14)*log(pmax(1e-6,1 - r)) + (n04 + n10)*log(pmax(1e-6,r)) + (n03 + n11)*log(1 + 8*r)
return(L)}
interlogL_c <- function(n02,n12,n00,n01,n03,n04,n10,n11,n13,n14) {
optimize(logL_c,c(0,0.5), n02,n12,n00,n01,n03,n04,n10,n11,n13,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_02"],x[,"n_12"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_13"],x[,"n_14"])
LOD_c <- (x[,"n_04"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_14"])*log10(2) - (x[,"n_03"] + x[,"n_11"])*log10(5) - (x[,"n_01"] + x[,"n_13"])*log10(5) + (x[,"n_01"] + x[,"n_13"])*log10(9 - 8*r_c) + (x[,"n_00"] + x[,"n_14"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_10"])*log10(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_11"])*log10(1 + 8*r_c)
logL_c <- 2*(x[,"n_02"] + x[,"n_12"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*(3*log(2) + log(5)) + (x[,"n_01"] + x[,"n_13"])*log(9 - 8*r_c) + (x[,"n_00"] + x[,"n_14"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_10"])*log(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_11"])*log(1 + 8*r_c)
logL_r <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14) {
L <- 3*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14)*log(2) + 2*(n02 + n12)*log(3) + (-2*n00 - 2*n01 - n02 - 2*n03 - 2*n04 - 2*n10 - 2*n11 - n12 - 2*n13 - 2*n14)*log(5) + (n03 + n11)*log(29 - 8*r) + (n04 + n10)*log(3 - r) + (n00 + n14)*log(2 + r) + (n01 + n13)*log(21 + 8*r)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"])
LOD_r <- -2*(x[,"n_03"] + x[,"n_11"])*log10(5) - 2*(x[,"n_01"] + x[,"n_13"])*log10(5) - (x[,"n_04"] + x[,"n_10"])*(-log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_14"])*(-log10(2) + log10(5)) + (x[,"n_03"] + x[,"n_11"])*log10(29 - 8*r_r) + (x[,"n_04"] + x[,"n_10"])*log10(3 - r_r) + (x[,"n_00"] + x[,"n_14"])*log10(2 + r_r) + (x[,"n_01"] + x[,"n_13"])*log10(21 + 8*r_r)
logL_r <- 3*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*log(2) + 2*(x[,"n_02"] + x[,"n_12"])*log(3) + (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"])*log(5) + (x[,"n_03"] + x[,"n_11"])*log(29 - 8*r_r) + (x[,"n_04"] + x[,"n_10"])*log(3 - r_r) + (x[,"n_00"] + x[,"n_14"])*log(2 + r_r) + (x[,"n_01"] + x[,"n_13"])*log(21 + 8*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_1.4 <- function(x){
logL_c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14)*(log(2) + log(5)) + (n02 + n13)*log(3 - 2*r) + (n01 + n14)*log(pmax(1e-6,1 - r)) + (n04 + n11)*log(pmax(1e-6,r)) + (n03 + n12)*log(1 + 2*r)
return(L)}
interlogL_c <- function(n01,n02,n03,n04,n11,n12,n13,n14) {
optimize(logL_c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"])
LOD_c <- (x[,"n_04"] + x[,"n_11"])*log10(2) - (x[,"n_03"] + x[,"n_12"])*log10(2) - (x[,"n_02"] + x[,"n_13"])*log10(2) + (x[,"n_01"] + x[,"n_14"])*log10(2) + (x[,"n_02"] + x[,"n_13"])*log10(3 - 2*r_c) + (x[,"n_01"] + x[,"n_14"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_11"])*log10(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_12"])*log10(1 + 2*r_c)
logL_c <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*(log(2) + log(5)) + (x[,"n_02"] + x[,"n_13"])*log(3 - 2*r_c) + (x[,"n_01"] + x[,"n_14"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_11"])*log(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_12"])*log(1 + 2*r_c)
logL_r <- function(r,n01,n02,n03,n04,n11,n12,n13,n14) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14)*(log(2) + 2*log(5)) + (n03 + n12)*log(11 - 2*r) + (n04 + n11)*log(3 - r) + (n01 + n14)*log(2 + r) + (n02 + n13)*log(9 + 2*r)
return(L)}
interlogL_r <- function(n01,n02,n03,n04,n11,n12,n13,n14) {
optimize(logL_r,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"])
LOD_r <- -((x[,"n_04"] + x[,"n_11"])*(-log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_14"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_12"])*(log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_13"])*(log10(2) + log10(5)) + (x[,"n_03"] + x[,"n_12"])*log10(11 - 2*r_r) + (x[,"n_04"] + x[,"n_11"])*log10(3 - r_r) + (x[,"n_01"] + x[,"n_14"])*log10(2 + r_r) + (x[,"n_02"] + x[,"n_13"])*log10(9 + 2*r_r)
logL_r <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*(log(2) + 2*log(5)) + (x[,"n_03"] + x[,"n_12"])*log(11 - 2*r_r) + (x[,"n_04"] + x[,"n_11"])*log(3 - r_r) + (x[,"n_01"] + x[,"n_14"])*log(2 + r_r) + (x[,"n_02"] + x[,"n_13"])*log(9 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_1.5 <- function(x){
r_c <- (x[,"n_04"] + x[,"n_12"])/(x[,"n_02"] + x[,"n_04"] + x[,"n_12"] + x[,"n_14"])
logL_c <- 2*(-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*log(2) + (x[,"n_02"] + x[,"n_14"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_12"])*log(pmax(1e-6,r_c))
LOD_c <- (x[,"n_04"] + x[,"n_12"])*log10(2) + (x[,"n_02"] + x[,"n_14"])*log10(2) + (x[,"n_02"] + x[,"n_14"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_12"])*log10(pmax(1e-6,r_c))
r_r <- (3*x[,"n_02"] - 2*x[,"n_04"] - 2*x[,"n_12"] + 3*x[,"n_14"])/(x[,"n_02"] + x[,"n_04"] + x[,"n_12"] + x[,"n_14"])
logL_r <- 2*(-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*log(2) + (-x[,"n_02"] - x[,"n_04"] - x[,"n_12"] - x[,"n_14"])*log(5) + (x[,"n_04"] + x[,"n_12"])*log(3 - r_r) + (x[,"n_02"] + x[,"n_14"])*log(2 + r_r)
LOD_r <- -((x[,"n_04"] + x[,"n_12"])*(-log10(2) + log10(5))) - (x[,"n_02"] + x[,"n_14"])*(-log10(2) + log10(5)) + (x[,"n_04"] + x[,"n_12"])*log10(3 - r_r) + (x[,"n_02"] + x[,"n_14"])*log10(2 + r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_2.0 <- function(x){
r_c <- (x[,"n_02"] + x[,"n_10"])/(x[,"n_00"] + x[,"n_02"] + x[,"n_10"] + x[,"n_12"])
logL_c <- -((-x[,"n_01"] - x[,"n_11"])*(-log(2) + log(3))) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"])*log(5) + (x[,"n_00"] + x[,"n_12"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_10"])*log(pmax(1e-6,r_c))
LOD_c <- (x[,"n_02"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_12"])*log10(2) + (x[,"n_00"] + x[,"n_12"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_10"])*log10(pmax(1e-6,r_c))
r_r <- (3*x[,"n_00"] - x[,"n_02"] - x[,"n_10"] + 3*x[,"n_12"])/(2*(x[,"n_00"] + x[,"n_02"] + x[,"n_10"] + x[,"n_12"]))
logL_r <- (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_10"] - x[,"n_11"] - 2*x[,"n_12"])*log(2) + (x[,"n_01"] + x[,"n_11"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"])*log(5) + (x[,"n_02"] + x[,"n_10"])*log(3 - 2*r_r) + (x[,"n_00"] + x[,"n_12"])*log(1 + 2*r_r)
LOD_r <- -((x[,"n_02"] + x[,"n_10"])*log10(2)) - (x[,"n_00"] + x[,"n_12"])*log10(2) + (x[,"n_02"] + x[,"n_10"])*log10(3 - 2*r_r) + (x[,"n_00"] + x[,"n_12"])*log10(1 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_2.1 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n10,n11,n12,n13) {
L <- (-n00 - 2*n01 - 2*n02 - n03 - n10 - 2*n11 - 2*n12 - n13)*log(2) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13)*log(5) + (n01 + n12)*log(5 - 2*r) + (n00 + n13)*log(pmax(1e-6,1 - r)) + (n03 + n10)*log(pmax(1e-6,r)) + (n02 + n11)*log(3 + 2*r)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n10,n11,n12,n13) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"])
LOD_c <- (x[,"n_03"] + x[,"n_10"])*log10(2) - 2*(x[,"n_02"] + x[,"n_11"])*log10(2) - 2*(x[,"n_01"] + x[,"n_12"])*log10(2) + (x[,"n_00"] + x[,"n_13"])*log10(2) + (x[,"n_01"] + x[,"n_12"])*log10(5 - 2*r_c) + (x[,"n_00"] + x[,"n_13"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_03"] + x[,"n_10"])*log10(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_11"])*log10(3 + 2*r_c)
logL_c <- (-x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"])*log(5) + (x[,"n_01"] + x[,"n_12"])*log(5 - 2*r_c) + (x[,"n_00"] + x[,"n_13"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_03"] + x[,"n_10"])*log(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_11"])*log(3 + 2*r_c)
logL_r <- function(r,n00,n01,n02,n03,n10,n11,n12,n13) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13)*(3*log(2) + log(5)) + (n03 + n10)*log(3 - 2*r) + (n02 + n11)*log(9 - 2*r) + (n00 + n13)*log(1 + 2*r) + (n01 + n12)*log(7 + 2*r)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n10,n11,n12,n13) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"])
LOD_r <- -((x[,"n_03"] + x[,"n_10"])*log10(2)) - 3*(x[,"n_02"] + x[,"n_11"])*log10(2) - 3*(x[,"n_01"] + x[,"n_12"])*log10(2) - (x[,"n_00"] + x[,"n_13"])*log10(2) + (x[,"n_03"] + x[,"n_10"])*log10(3 - 2*r_r) + (x[,"n_02"] + x[,"n_11"])*log10(9 - 2*r_r) + (x[,"n_00"] + x[,"n_13"])*log10(1 + 2*r_r) + (x[,"n_01"] + x[,"n_12"])*log10(7 + 2*r_r)
logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"])*(3*log(2) + log(5)) + (x[,"n_03"] + x[,"n_10"])*log(3 - 2*r_r) + (x[,"n_02"] + x[,"n_11"])*log(9 - 2*r_r) + (x[,"n_00"] + x[,"n_13"])*log(1 + 2*r_r) + (x[,"n_01"] + x[,"n_12"])*log(7 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_2.2 <- function(x){
logL_c <- function(r,n01,n02,n03,n11,n12,n13,n00,n04,n10,n14) {
L <- (-n01 - n02 - n03 - n11 - n12 - n13)*log(2) + (n01 + n03 + n11 + n13)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14)*log(5) + (n02 + n12)*log(11) + (n01 + n13)*log(3 - 2*r) + (n00 + n14)*log(pmax(1e-6,1 - r)) + (n04 + n10)*log(pmax(1e-6,r)) + (n03 + n11)*log(1 + 2*r)
return(L)}
interlogL_c <- function(n01,n02,n03,n11,n12,n13,n00,n04,n10,n14) {
optimize(logL_c,c(0,0.5), n01,n02,n03,n11,n12,n13,n00,n04,n10,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_00"],x[,"n_04"],x[,"n_10"],x[,"n_14"])
LOD_c <- (x[,"n_04"] + x[,"n_10"])*log10(2) - (x[,"n_03"] + x[,"n_11"])*log10(2) - (x[,"n_01"] + x[,"n_13"])*log10(2) + (x[,"n_00"] + x[,"n_14"])*log10(2) + (x[,"n_01"] + x[,"n_13"])*log10(3 - 2*r_c) + (x[,"n_00"] + x[,"n_14"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_10"])*log10(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_11"])*log10(1 + 2*r_c)
logL_c <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"])*log(2) + (x[,"n_01"] + x[,"n_03"] + x[,"n_11"] + x[,"n_13"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*log(5) + (x[,"n_02"] + x[,"n_12"])*log(11) + (x[,"n_01"] + x[,"n_13"])*log(3 - 2*r_c) + (x[,"n_00"] + x[,"n_14"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_10"])*log(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_11"])*log(1 + 2*r_c)
logL_r <- function(r,n00,n01,n02,n04,n10,n12,n13,n14,n03,n11) {
L <- (-2*n00 - 2*n01 - n02 - 2*n04 - 2*n10 - n12 - 2*n13 - 2*n14)*log(2) + (n01 + n13)*log(3) + 2*(-n00 - n01 - n02 - n04 - n10 - n12 - n13 - n14)*log(5) + (n02 + n12)*log(11) + (n04 + n10)*log(3 - 2*r) + (n03 + n11)*log(3/20 - (3*r)/50) + (n00 + n14)*log(1 + 2*r) + (n01 + n13)*log(3 + 2*r)
return(L)}
interlogL_r <- function(n00,n01,n02,n04,n10,n12,n13,n14,n03,n11) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n04,n10,n12,n13,n14,n03,n11, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_04"],x[,"n_10"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_03"],x[,"n_11"])
LOD_r <- -((x[,"n_04"] + x[,"n_10"])*log10(2)) - 2*(x[,"n_01"] + x[,"n_13"])*log10(2) - (x[,"n_00"] + x[,"n_14"])*log10(2) + (x[,"n_03"] + x[,"n_11"])*(-log10(3) + 2*log10(5)) + (x[,"n_04"] + x[,"n_10"])*log10(3 - 2*r_r) + (x[,"n_03"] + x[,"n_11"])*log10(3/20 - (3*r_r)/50) + (x[,"n_00"] + x[,"n_14"])*log10(1 + 2*r_r) + (x[,"n_01"] + x[,"n_13"])*log10(3 + 2*r_r)
logL_r <- (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_04"] - 2*x[,"n_10"] - x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"])*log(2) + (x[,"n_01"] + x[,"n_13"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_04"] - x[,"n_10"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*log(5) + (x[,"n_02"] + x[,"n_12"])*log(11) + (x[,"n_04"] + x[,"n_10"])*log(3 - 2*r_r) + (x[,"n_03"] + x[,"n_11"])*log(3/20 - (3*r_r)/50) + (x[,"n_00"] + x[,"n_14"])*log(1 + 2*r_r) + (x[,"n_01"] + x[,"n_13"])*log(3 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_2.3 <- function(x){
logL_c <- function(r,n00,n01,n03,n04,n05,n10,n11,n12,n14,n15,n02,n13) {
L <- (-2*n00 - 3*n01 - 3*n03 - 3*n04 - 2*n05 - 2*n10 - 3*n11 - 3*n12 - 3*n14 - 2*n15)*log(2) + (n01 + n04 + n11 + n14)*log(3) + 2*(-n00 - n01 - n03 - n04 - n05 - n10 - n11 - n12 - n14 - n15)*log(5) + (n01 + n14)*log(7 - 6*r) + (n00 + n15)*log(pmax(1e-6,1 - r)) + (n02 + n13)*log(9/40 - (2*r)/25) + (n05 + n10)*log(pmax(1e-6,r)) + (n04 + n11)*log(1 + 6*r) + (n03 + n12)*log(29 + 16*r)
return(L)}
interlogL_c <- function(n00,n01,n03,n04,n05,n10,n11,n12,n14,n15,n02,n13) {
optimize(logL_c,c(0,0.5), n00,n01,n03,n04,n05,n10,n11,n12,n14,n15,n02,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_02"],x[,"n_13"])
LOD_c <- (x[,"n_05"] + x[,"n_10"])*log10(2) - 2*(x[,"n_04"] + x[,"n_11"])*log10(2) - 2*(x[,"n_01"] + x[,"n_14"])*log10(2) + (x[,"n_00"] + x[,"n_15"])*log10(2) + (x[,"n_02"] + x[,"n_13"])*(3*log10(2) + 2*log10(5) - log10(37)) - (x[,"n_03"] + x[,"n_12"])*log10(37) + (x[,"n_01"] + x[,"n_14"])*log10(7 - 6*r_c) + (x[,"n_00"] + x[,"n_15"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_13"])*log10(9/40 - (2*r_c)/25) + (x[,"n_05"] + x[,"n_10"])*log10(pmax(1e-6,r_c)) + (x[,"n_04"] + x[,"n_11"])*log10(1 + 6*r_c) + (x[,"n_03"] + x[,"n_12"])*log10(29 + 16*r_c)
logL_c <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_14"] - 2*x[,"n_15"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_14"] - x[,"n_15"])*log(5) + (x[,"n_01"] + x[,"n_14"])*log(7 - 6*r_c) + (x[,"n_00"] + x[,"n_15"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_13"])*log(9/40 - (2*r_c)/25) + (x[,"n_05"] + x[,"n_10"])*log(pmax(1e-6,r_c)) + (x[,"n_04"] + x[,"n_11"])*log(1 + 6*r_c) + (x[,"n_03"] + x[,"n_12"])*log(29 + 16*r_c)
logL_r <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15) {
L <- (-4*n00 - 4*n01 - 3*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 3*n12 - 3*n13 - 4*n14 - 4*n15)*log(2) + (n01 + n04 + n11 + n14)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15)*log(5) + (n03 + n12)*log(41 - 8*r) + (n04 + n11)*log(11 - 6*r) + (n05 + n10)*log(3 - 2*r) + (n00 + n15)*log(1 + 2*r) + (n01 + n14)*log(5 + 6*r) + (n02 + n13)*log(33 + 8*r)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"])
LOD_r <- -((x[,"n_05"] + x[,"n_10"])*log10(2)) - 3*(x[,"n_04"] + x[,"n_11"])*log10(2) - 3*(x[,"n_01"] + x[,"n_14"])*log10(2) - (x[,"n_00"] + x[,"n_15"])*log10(2) - (x[,"n_03"] + x[,"n_12"])*log10(37) - (x[,"n_02"] + x[,"n_13"])*log10(37) + (x[,"n_03"] + x[,"n_12"])*log10(41 - 8*r_r) + (x[,"n_04"] + x[,"n_11"])*log10(11 - 6*r_r) + (x[,"n_05"] + x[,"n_10"])*log10(3 - 2*r_r) + (x[,"n_00"] + x[,"n_15"])*log10(1 + 2*r_r) + (x[,"n_01"] + x[,"n_14"])*log10(5 + 6*r_r) + (x[,"n_02"] + x[,"n_13"])*log10(33 + 8*r_r)
logL_r <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"])*log(5) + (x[,"n_03"] + x[,"n_12"])*log(41 - 8*r_r) + (x[,"n_04"] + x[,"n_11"])*log(11 - 6*r_r) + (x[,"n_05"] + x[,"n_10"])*log(3 - 2*r_r) + (x[,"n_00"] + x[,"n_15"])*log(1 + 2*r_r) + (x[,"n_01"] + x[,"n_14"])*log(5 + 6*r_r) + (x[,"n_02"] + x[,"n_13"])*log(33 + 8*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_2.4 <- function(x){
logL_c <- function(r,n02,n03,n04,n12,n13,n14,n01,n05,n11,n15) {
L <- (-n02 - n03 - n04 - n12 - n13 - n14)*log(2) + (n02 + n04 + n12 + n14)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15)*log(5) + (n03 + n13)*log(11) + (n02 + n14)*log(3 - 2*r) + (n01 + n15)*log(pmax(1e-6,1 - r)) + (n05 + n11)*log(pmax(1e-6,r)) + (n04 + n12)*log(1 + 2*r)
return(L)}
interlogL_c <- function(n02,n03,n04,n12,n13,n14,n01,n05,n11,n15) {
optimize(logL_c,c(0,0.5), n02,n03,n04,n12,n13,n14,n01,n05,n11,n15, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_01"],x[,"n_05"],x[,"n_11"],x[,"n_15"])
LOD_c <- (x[,"n_05"] + x[,"n_11"])*log10(2) - (x[,"n_04"] + x[,"n_12"])*log10(2) - (x[,"n_02"] + x[,"n_14"])*log10(2) + (x[,"n_01"] + x[,"n_15"])*log10(2) + (x[,"n_02"] + x[,"n_14"])*log10(3 - 2*r_c) + (x[,"n_01"] + x[,"n_15"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_05"] + x[,"n_11"])*log10(pmax(1e-6,r_c)) + (x[,"n_04"] + x[,"n_12"])*log10(1 + 2*r_c)
logL_c <- (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*log(2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_12"] + x[,"n_14"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"])*log(5) + (x[,"n_03"] + x[,"n_13"])*log(11) + (x[,"n_02"] + x[,"n_14"])*log(3 - 2*r_c) + (x[,"n_01"] + x[,"n_15"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_05"] + x[,"n_11"])*log(pmax(1e-6,r_c)) + (x[,"n_04"] + x[,"n_12"])*log(1 + 2*r_c)
logL_r <- function(r,n01,n02,n03,n05,n11,n13,n14,n15,n04,n12) {
L <- (-2*n01 - 2*n02 - n03 - 2*n05 - 2*n11 - n13 - 2*n14 - 2*n15)*log(2) + (n02 + n14)*log(3) + 2*(-n01 - n02 - n03 - n05 - n11 - n13 - n14 - n15)*log(5) + (n03 + n13)*log(11) + (n05 + n11)*log(3 - 2*r) + (n04 + n12)*log(3/20 - (3*r)/50) + (n01 + n15)*log(1 + 2*r) + (n02 + n14)*log(3 + 2*r)
return(L)}
interlogL_r <- function(n01,n02,n03,n05,n11,n13,n14,n15,n04,n12) {
optimize(logL_r,c(0,0.5), n01,n02,n03,n05,n11,n13,n14,n15,n04,n12, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_05"],x[,"n_11"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_04"],x[,"n_12"])
LOD_r <- -((x[,"n_05"] + x[,"n_11"])*log10(2)) - 2*(x[,"n_02"] + x[,"n_14"])*log10(2) - (x[,"n_01"] + x[,"n_15"])*log10(2) + (x[,"n_04"] + x[,"n_12"])*(-log10(3) + 2*log10(5)) + (x[,"n_05"] + x[,"n_11"])*log10(3 - 2*r_r) + (x[,"n_04"] + x[,"n_12"])*log10(3/20 - (3*r_r)/50) + (x[,"n_01"] + x[,"n_15"])*log10(1 + 2*r_r) + (x[,"n_02"] + x[,"n_14"])*log10(3 + 2*r_r)
logL_r <- (-2*x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_05"] - 2*x[,"n_11"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"])*log(2) + (x[,"n_02"] + x[,"n_14"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_05"] - x[,"n_11"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"])*log(5) + (x[,"n_03"] + x[,"n_13"])*log(11) + (x[,"n_05"] + x[,"n_11"])*log(3 - 2*r_r) + (x[,"n_04"] + x[,"n_12"])*log(3/20 - (3*r_r)/50) + (x[,"n_01"] + x[,"n_15"])*log(1 + 2*r_r) + (x[,"n_02"] + x[,"n_14"])*log(3 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_3.0 <- function(x){
logL_c <- function(r,n01,n02,n11,n12,n00,n03,n10,n13) {
L <- (n01 + n02 + n11 + n12)*log(3) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13)*(2*log(2) + log(5)) + (n00 + n13)*log(pmax(1e-6,1 - r)) + (n01 + n12)*log(2 - r) + (n03 + n10)*log(pmax(1e-6,r)) + (n02 + n11)*log(1 + r)
return(L)}
interlogL_c <- function(n01,n02,n11,n12,n00,n03,n10,n13) {
optimize(logL_c,c(0,0.5), n01,n02,n11,n12,n00,n03,n10,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_11"],x[,"n_12"],x[,"n_00"],x[,"n_03"],x[,"n_10"],x[,"n_13"])
LOD_c <- (x[,"n_03"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_13"])*log10(2) - (x[,"n_02"] + x[,"n_11"])*(-log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_12"])*(-log10(2) + log10(3)) + (x[,"n_00"] + x[,"n_13"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_12"])*log10(2 - r_c) + (x[,"n_03"] + x[,"n_10"])*log10(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_11"])*log10(1 + r_c)
logL_c <- (x[,"n_01"] + x[,"n_02"] + x[,"n_11"] + x[,"n_12"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"])*(2*log(2) + log(5)) + (x[,"n_00"] + x[,"n_13"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_12"])*log(2 - r_c) + (x[,"n_03"] + x[,"n_10"])*log(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_11"])*log(1 + r_c)
logL_r <- function(r,n01,n02,n11,n12,n00,n03,n10,n13) {
L <- (n01 + n02 + n11 + n12)*log(3) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13)*(2*log(2) + log(5)) + (n03 + n10)*log(pmax(1e-6,1 - r)) + (n02 + n11)*log(2 - r) + (n00 + n13)*log(pmax(1e-6,r)) + (n01 + n12)*log(1 + r)
return(L)}
interlogL_r <- function(n01,n02,n11,n12,n00,n03,n10,n13) {
optimize(logL_r,c(0,0.5), n01,n02,n11,n12,n00,n03,n10,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_11"],x[,"n_12"],x[,"n_00"],x[,"n_03"],x[,"n_10"],x[,"n_13"])
LOD_r <- (x[,"n_03"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_13"])*log10(2) - (x[,"n_02"] + x[,"n_11"])*(-log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_12"])*(-log10(2) + log10(3)) + (x[,"n_03"] + x[,"n_10"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_02"] + x[,"n_11"])*log10(2 - r_r) + (x[,"n_00"] + x[,"n_13"])*log10(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_12"])*log10(1 + r_r)
logL_r <- (x[,"n_01"] + x[,"n_02"] + x[,"n_11"] + x[,"n_12"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"])*(2*log(2) + log(5)) + (x[,"n_03"] + x[,"n_10"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_02"] + x[,"n_11"])*log(2 - r_r) + (x[,"n_00"] + x[,"n_13"])*log(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_12"])*log(1 + r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_3.1 <- function(x){
logL_c <- function(r,n02,n12,n00,n01,n03,n04,n10,n11,n13,n14) {
L <- 2*(n02 + n12)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14)*(3*log(2) + log(5)) + (n01 + n13)*log(7 - 4*r) + (n00 + n14)*log(pmax(1e-6,1 - r)) + (n04 + n10)*log(pmax(1e-6,r)) + (n03 + n11)*log(3 + 4*r)
return(L)}
interlogL_c <- function(n02,n12,n00,n01,n03,n04,n10,n11,n13,n14) {
optimize(logL_c,c(0,0.5), n02,n12,n00,n01,n03,n04,n10,n11,n13,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_02"],x[,"n_12"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_13"],x[,"n_14"])
LOD_c <- (x[,"n_04"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_14"])*log10(2) - (x[,"n_03"] + x[,"n_11"])*log10(5) - (x[,"n_01"] + x[,"n_13"])*log10(5) + (x[,"n_01"] + x[,"n_13"])*log10(7 - 4*r_c) + (x[,"n_00"] + x[,"n_14"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_10"])*log10(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_11"])*log10(3 + 4*r_c)
logL_c <- 2*(x[,"n_02"] + x[,"n_12"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*(3*log(2) + log(5)) + (x[,"n_01"] + x[,"n_13"])*log(7 - 4*r_c) + (x[,"n_00"] + x[,"n_14"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_10"])*log(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_11"])*log(3 + 4*r_c)
logL_r <- function(r,n02,n12,n00,n01,n03,n04,n10,n11,n13,n14) {
L <- 2*(n02 + n12)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14)*(3*log(2) + log(5)) + (n03 + n11)*log(7 - 4*r) + (n04 + n10)*log(pmax(1e-6,1 - r)) + (n00 + n14)*log(pmax(1e-6,r)) + (n01 + n13)*log(3 + 4*r)
return(L)}
interlogL_r <- function(n02,n12,n00,n01,n03,n04,n10,n11,n13,n14) {
optimize(logL_r,c(0,0.5), n02,n12,n00,n01,n03,n04,n10,n11,n13,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_02"],x[,"n_12"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_13"],x[,"n_14"])
LOD_r <- (x[,"n_04"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_14"])*log10(2) - (x[,"n_03"] + x[,"n_11"])*log10(5) - (x[,"n_01"] + x[,"n_13"])*log10(5) + (x[,"n_03"] + x[,"n_11"])*log10(7 - 4*r_r) + (x[,"n_04"] + x[,"n_10"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_00"] + x[,"n_14"])*log10(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_13"])*log10(3 + 4*r_r)
logL_r <- 2*(x[,"n_02"] + x[,"n_12"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*(3*log(2) + log(5)) + (x[,"n_03"] + x[,"n_11"])*log(7 - 4*r_r) + (x[,"n_04"] + x[,"n_10"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_00"] + x[,"n_14"])*log(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_13"])*log(3 + 4*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_3.2 <- function(x){
logL_c <- function(r,n01,n04,n11,n14,n00,n02,n03,n05,n10,n12,n13,n15) {
L <- (n01 + n04 + n11 + n14)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15)*(2*log(2) + 2*log(5)) + (n02 + n13)*log(22 - 7*r) + (n01 + n14)*log(3 - 2*r) + (n00 + n15)*log(pmax(1e-6,1 - r)) + (n05 + n10)*log(pmax(1e-6,r)) + (n04 + n11)*log(1 + 2*r) + (n03 + n12)*log(15 + 7*r)
return(L)}
interlogL_c <- function(n01,n04,n11,n14,n00,n02,n03,n05,n10,n12,n13,n15) {
optimize(logL_c,c(0,0.5), n01,n04,n11,n14,n00,n02,n03,n05,n10,n12,n13,n15, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_04"],x[,"n_11"],x[,"n_14"],x[,"n_00"],x[,"n_02"],x[,"n_03"],x[,"n_05"],x[,"n_10"],x[,"n_12"],x[,"n_13"],x[,"n_15"])
LOD_c <- (x[,"n_05"] + x[,"n_10"])*log10(2) - (x[,"n_04"] + x[,"n_11"])*log10(2) - (x[,"n_01"] + x[,"n_14"])*log10(2) + (x[,"n_00"] + x[,"n_15"])*log10(2) - (x[,"n_03"] + x[,"n_12"])*(-log10(2) + log10(37)) - (x[,"n_02"] + x[,"n_13"])*(-log10(2) + log10(37)) + (x[,"n_02"] + x[,"n_13"])*log10(22 - 7*r_c) + (x[,"n_01"] + x[,"n_14"])*log10(3 - 2*r_c) + (x[,"n_00"] + x[,"n_15"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_05"] + x[,"n_10"])*log10(pmax(1e-6,r_c)) + (x[,"n_04"] + x[,"n_11"])*log10(1 + 2*r_c) + (x[,"n_03"] + x[,"n_12"])*log10(15 + 7*r_c)
logL_c <- (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"])*(2*log(2) + 2*log(5)) + (x[,"n_02"] + x[,"n_13"])*log(22 - 7*r_c) + (x[,"n_01"] + x[,"n_14"])*log(3 - 2*r_c) + (x[,"n_00"] + x[,"n_15"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_05"] + x[,"n_10"])*log(pmax(1e-6,r_c)) + (x[,"n_04"] + x[,"n_11"])*log(1 + 2*r_c) + (x[,"n_03"] + x[,"n_12"])*log(15 + 7*r_c)
logL_r <- function(r,n01,n04,n11,n14,n00,n02,n03,n05,n10,n12,n13,n15) {
L <- (n01 + n04 + n11 + n14)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15)*(2*log(2) + 2*log(5)) + (n03 + n12)*log(22 - 7*r) + (n04 + n11)*log(3 - 2*r) + (n05 + n10)*log(pmax(1e-6,1 - r)) + (n00 + n15)*log(pmax(1e-6,r)) + (n01 + n14)*log(1 + 2*r) + (n02 + n13)*log(15 + 7*r)
return(L)}
interlogL_r <- function(n01,n04,n11,n14,n00,n02,n03,n05,n10,n12,n13,n15) {
optimize(logL_r,c(0,0.5), n01,n04,n11,n14,n00,n02,n03,n05,n10,n12,n13,n15, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_04"],x[,"n_11"],x[,"n_14"],x[,"n_00"],x[,"n_02"],x[,"n_03"],x[,"n_05"],x[,"n_10"],x[,"n_12"],x[,"n_13"],x[,"n_15"])
LOD_r <- (x[,"n_05"] + x[,"n_10"])*log10(2) - (x[,"n_04"] + x[,"n_11"])*log10(2) - (x[,"n_01"] + x[,"n_14"])*log10(2) + (x[,"n_00"] + x[,"n_15"])*log10(2) - (x[,"n_03"] + x[,"n_12"])*(-log10(2) + log10(37)) - (x[,"n_02"] + x[,"n_13"])*(-log10(2) + log10(37)) + (x[,"n_03"] + x[,"n_12"])*log10(22 - 7*r_r) + (x[,"n_04"] + x[,"n_11"])*log10(3 - 2*r_r) + (x[,"n_05"] + x[,"n_10"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_00"] + x[,"n_15"])*log10(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_14"])*log10(1 + 2*r_r) + (x[,"n_02"] + x[,"n_13"])*log10(15 + 7*r_r)
logL_r <- (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"])*(2*log(2) + 2*log(5)) + (x[,"n_03"] + x[,"n_12"])*log(22 - 7*r_r) + (x[,"n_04"] + x[,"n_11"])*log(3 - 2*r_r) + (x[,"n_05"] + x[,"n_10"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_00"] + x[,"n_15"])*log(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_14"])*log(1 + 2*r_r) + (x[,"n_02"] + x[,"n_13"])*log(15 + 7*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_3.3 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16) {
L <- (-4*n00 - 4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 4*n11 - 4*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n16)*log(2) + (n01 + n02 + n04 + n05 + n11 + n12 + n14 + n15)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16)*log(5) + (n02 + n04 + n12 + n14)*log(11) + (n03 + n13)*log(41) + (n01 + n15)*log(5 - 4*r) + (n00 + n16)*log(pmax(1e-6,1 - r)) + (n02 + n14)*log(2 - r) + (n06 + n10)*log(pmax(1e-6,r)) + (n04 + n12)*log(1 + r) + (n05 + n11)*log(1 + 4*r)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"])
LOD_c <- (x[,"n_06"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_16"])*log10(2) - (x[,"n_05"] + x[,"n_11"])*log10(3) - (x[,"n_01"] + x[,"n_15"])*log10(3) - (x[,"n_04"] + x[,"n_12"])*(-log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_14"])*(-log10(2) + log10(3)) + (x[,"n_01"] + x[,"n_15"])*log10(5 - 4*r_c) + (x[,"n_00"] + x[,"n_16"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_14"])*log10(2 - r_c) + (x[,"n_06"] + x[,"n_10"])*log10(pmax(1e-6,r_c)) + (x[,"n_04"] + x[,"n_12"])*log10(1 + r_c) + (x[,"n_05"] + x[,"n_11"])*log10(1 + 4*r_c)
logL_c <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_11"] + x[,"n_12"] + x[,"n_14"] + x[,"n_15"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"])*log(5) + (x[,"n_02"] + x[,"n_04"] + x[,"n_12"] + x[,"n_14"])*log(11) + (x[,"n_03"] + x[,"n_13"])*log(41) + (x[,"n_01"] + x[,"n_15"])*log(5 - 4*r_c) + (x[,"n_00"] + x[,"n_16"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_14"])*log(2 - r_c) + (x[,"n_06"] + x[,"n_10"])*log(pmax(1e-6,r_c)) + (x[,"n_04"] + x[,"n_12"])*log(1 + r_c) + (x[,"n_05"] + x[,"n_11"])*log(1 + 4*r_c)
logL_r <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16) {
L <- (-4*n00 - 4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 4*n11 - 4*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n16)*log(2) + (n01 + n02 + n04 + n05 + n11 + n12 + n14 + n15)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16)*log(5) + (n02 + n04 + n12 + n14)*log(11) + (n03 + n13)*log(41) + (n05 + n11)*log(5 - 4*r) + (n06 + n10)*log(pmax(1e-6,1 - r)) + (n04 + n12)*log(2 - r) + (n00 + n16)*log(pmax(1e-6,r)) + (n02 + n14)*log(1 + r) + (n01 + n15)*log(1 + 4*r)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"])
LOD_r <- (x[,"n_06"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_16"])*log10(2) - (x[,"n_05"] + x[,"n_11"])*log10(3) - (x[,"n_01"] + x[,"n_15"])*log10(3) - (x[,"n_04"] + x[,"n_12"])*(-log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_14"])*(-log10(2) + log10(3)) + (x[,"n_05"] + x[,"n_11"])*log10(5 - 4*r_r) + (x[,"n_06"] + x[,"n_10"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_04"] + x[,"n_12"])*log10(2 - r_r) + (x[,"n_00"] + x[,"n_16"])*log10(pmax(1e-6,r_r)) + (x[,"n_02"] + x[,"n_14"])*log10(1 + r_r) + (x[,"n_01"] + x[,"n_15"])*log10(1 + 4*r_r)
logL_r <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_11"] + x[,"n_12"] + x[,"n_14"] + x[,"n_15"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"])*log(5) + (x[,"n_02"] + x[,"n_04"] + x[,"n_12"] + x[,"n_14"])*log(11) + (x[,"n_03"] + x[,"n_13"])*log(41) + (x[,"n_05"] + x[,"n_11"])*log(5 - 4*r_r) + (x[,"n_06"] + x[,"n_10"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_04"] + x[,"n_12"])*log(2 - r_r) + (x[,"n_00"] + x[,"n_16"])*log(pmax(1e-6,r_r)) + (x[,"n_02"] + x[,"n_14"])*log(1 + r_r) + (x[,"n_01"] + x[,"n_15"])*log(1 + 4*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.0_4.1 <- function(x){
logL_c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14)*(3*log(2) + log(5)) + (n01 + n14)*log(3 - 2*r) + (n02 + n13)*log(9 - 2*r) + (n04 + n11)*log(1 + 2*r) + (n03 + n12)*log(7 + 2*r)
return(L)}
interlogL_c <- function(n01,n02,n03,n04,n11,n12,n13,n14) {
optimize(logL_c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"])
LOD_c <- -((x[,"n_04"] + x[,"n_11"])*log10(2)) - 3*(x[,"n_03"] + x[,"n_12"])*log10(2) - 3*(x[,"n_02"] + x[,"n_13"])*log10(2) - (x[,"n_01"] + x[,"n_14"])*log10(2) + (x[,"n_01"] + x[,"n_14"])*log10(3 - 2*r_c) + (x[,"n_02"] + x[,"n_13"])*log10(9 - 2*r_c) + (x[,"n_04"] + x[,"n_11"])*log10(1 + 2*r_c) + (x[,"n_03"] + x[,"n_12"])*log10(7 + 2*r_c)
logL_c <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*(3*log(2) + log(5)) + (x[,"n_01"] + x[,"n_14"])*log(3 - 2*r_c) + (x[,"n_02"] + x[,"n_13"])*log(9 - 2*r_c) + (x[,"n_04"] + x[,"n_11"])*log(1 + 2*r_c) + (x[,"n_03"] + x[,"n_12"])*log(7 + 2*r_c)
logL_r <- function(r,n01,n02,n03,n04,n11,n12,n13,n14) {
L <- (-n01 - 2*n02 - 2*n03 - n04 - n11 - 2*n12 - 2*n13 - n14)*log(2) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14)*log(5) + (n03 + n12)*log(5 - 2*r) + (n04 + n11)*log(pmax(1e-6,1 - r)) + (n01 + n14)*log(pmax(1e-6,r)) + (n02 + n13)*log(3 + 2*r)
return(L)}
interlogL_r <- function(n01,n02,n03,n04,n11,n12,n13,n14) {
optimize(logL_r,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"])
LOD_r <- (x[,"n_04"] + x[,"n_11"])*log10(2) - 2*(x[,"n_03"] + x[,"n_12"])*log10(2) - 2*(x[,"n_02"] + x[,"n_13"])*log10(2) + (x[,"n_01"] + x[,"n_14"])*log10(2) + (x[,"n_03"] + x[,"n_12"])*log10(5 - 2*r_r) + (x[,"n_04"] + x[,"n_11"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_01"] + x[,"n_14"])*log10(pmax(1e-6,r_r)) + (x[,"n_02"] + x[,"n_13"])*log10(3 + 2*r_r)
logL_r <- (-x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"])*log(5) + (x[,"n_03"] + x[,"n_12"])*log(5 - 2*r_r) + (x[,"n_04"] + x[,"n_11"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_01"] + x[,"n_14"])*log(pmax(1e-6,r_r)) + (x[,"n_02"] + x[,"n_13"])*log(3 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_1.1 <- function(x){
logL_cc <- function(r,n00,n01,n02,n10,n12,n20,n21,n22,n11) {
L <- (-2*n00 - n01 - 2*n02 - n10 - n12 - 2*n20 - n21 - 2*n22)*log(2) + 2*(n00 + n22)*log(pmax(1e-6,1 - r)) + 2*(n02 + n20)*log(pmax(1e-6,r)) + (n01 + n10 + n12 + n21)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n11*log(1/2 - r + r^2)
return(L)}
interlogL_cc <- function(n00,n01,n02,n10,n12,n20,n21,n22,n11) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n10,n12,n20,n21,n22,n11, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_12"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_11"])
LOD_cc <- 2*x[,"n_11"]*log10(2) + 2*(x[,"n_02"] + x[,"n_20"])*log10(2) + 2*(x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*log10(2) + 2*(x[,"n_00"] + x[,"n_22"])*log10(2) + 2*(x[,"n_00"] + x[,"n_22"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_02"] + x[,"n_20"])*log10(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + x[,"n_11"]*log10(1/2 - r_cc + r_cc^2)
logL_cc <- (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - x[,"n_10"] - x[,"n_12"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"])*log(2) + 2*(x[,"n_00"] + x[,"n_22"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_02"] + x[,"n_20"])*log(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + x[,"n_11"]*log(1/2 - r_cc + r_cc^2)
logL_cr <- function(r,n00,n01,n02,n10,n12,n20,n21,n22,n11) {
L <- 2*(-n00 - n01 - n02 - n10 - n12 - n20 - n21 - n22)*log(2) + (-n00 - n01 - n02 - n10 - n11 - n12 - n20 - n21 - n22)*log(5) + (n02 + n20)*(log(3 - r) + log(pmax(1e-6,r))) + (n00 + n22)*(log(pmax(1e-6,1 - r)) + log(2 + r)) + n11*log(1 + r - r^2) + (n01 + n10 + n12 + n21)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_cr <- function(n00,n01,n02,n10,n12,n20,n21,n22,n11) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n10,n12,n20,n21,n22,n11, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_12"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_11"])
LOD_cr <- -(x[,"n_11"]*(-2*log10(2) + log10(5))) - (x[,"n_02"] + x[,"n_20"])*(-2*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_22"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(-log10(2) + log10(5)) + (x[,"n_02"] + x[,"n_20"])*(log10(3 - r_cr) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_22"])*(log10(pmax(1e-6,1 - r_cr)) + log10(2 + r_cr)) + x[,"n_11"]*log10(1 + r_cr - r_cr^2) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*log10(3 - 2*r_cr + 2*r_cr^2)
logL_cr <- 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"])*log(5) + (x[,"n_02"] + x[,"n_20"])*(log(3 - r_cr) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_22"])*(log(pmax(1e-6,1 - r_cr)) + log(2 + r_cr)) + x[,"n_11"]*log(1 + r_cr - r_cr^2) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*log(3 - 2*r_cr + 2*r_cr^2)
logL_rr <- function(r,n00,n01,n02,n10,n11,n12,n20,n21,n22) {
L <- (-2*n00 - n01 - 2*n02 - n10 - n11 - n12 - 2*n20 - n21 - 2*n22)*log(2) + 2*(-n00 - n01 - n02 - n10 - n11 - n12 - n20 - n21 - n22)*log(5) + 2*(n02 + n20)*log(3 - r) + 2*(n00 + n22)*log(2 + r) + (n01 + n10 + n12 + n21)*(log(3 - r) + log(2 + r)) + n11*log(13 - 2*r + 2*r^2)
return(L)}
interlogL_rr <- function(n00,n01,n02,n10,n11,n12,n20,n21,n22) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n10,n11,n12,n20,n21,n22, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_20"],x[,"n_21"],x[,"n_22"])
LOD_rr <- -2*(x[,"n_02"] + x[,"n_20"])*(-log10(2) + log10(5)) - 2*(x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(-log10(2) + log10(5)) - 2*(x[,"n_00"] + x[,"n_22"])*(-log10(2) + log10(5)) - x[,"n_11"]*(-log10(2) + 2*log10(5)) + 2*(x[,"n_02"] + x[,"n_20"])*log10(3 - r_rr) + 2*(x[,"n_00"] + x[,"n_22"])*log10(2 + r_rr) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(log10(3 - r_rr) + log10(2 + r_rr)) + x[,"n_11"]*log10(13 - 2*r_rr + 2*r_rr^2)
logL_rr <- (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"])*log(5) + 2*(x[,"n_02"] + x[,"n_20"])*log(3 - r_rr) + 2*(x[,"n_00"] + x[,"n_22"])*log(2 + r_rr) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(log(3 - r_rr) + log(2 + r_rr)) + x[,"n_11"]*log(13 - 2*r_rr + 2*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_1.2 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n11,n12,n20,n21,n22,n23,n10,n13) {
L <- (-n00 - 2*n01 - 2*n02 - n03 - 2*n11 - 2*n12 - n20 - 2*n21 - 2*n22 - n23)*log(2) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*log(5) + 2*(n00 + n23)*log(pmax(1e-6,1 - r)) + 2*(n03 + n20)*log(pmax(1e-6,r)) + (n02 + n21)*(log(5 - 2*r) + log(pmax(1e-6,r))) + (n10 + n13)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n22)*log(3 - r - 2*r^2) + (n11 + n12)*log(5 - 4*r + 4*r^2)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n11,n12,n20,n21,n22,n23,n10,n13) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n11,n12,n20,n21,n22,n23,n10,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_11"],x[,"n_12"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_10"],x[,"n_13"])
LOD_cc <- -2*(x[,"n_11"] + x[,"n_12"])*log10(2) + 2*(x[,"n_10"] + x[,"n_13"])*log10(2) + 2*(x[,"n_03"] + x[,"n_20"])*log10(2) - (x[,"n_02"] + x[,"n_21"])*log10(2) - (x[,"n_01"] + x[,"n_22"])*log10(2) + 2*(x[,"n_00"] + x[,"n_23"])*log10(2) + 2*(x[,"n_00"] + x[,"n_23"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_03"] + x[,"n_20"])*log10(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_21"])*(log10(5 - 2*r_cc) + log10(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_13"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_22"])*log10(3 - r_cc - 2*r_cc^2) + (x[,"n_11"] + x[,"n_12"])*log10(5 - 4*r_cc + 4*r_cc^2)
logL_cc <- (-x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*log(5) + 2*(x[,"n_00"] + x[,"n_23"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_03"] + x[,"n_20"])*log(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_21"])*(log(5 - 2*r_cc) + log(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_13"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_22"])*log(3 - r_cc - 2*r_cc^2) + (x[,"n_11"] + x[,"n_12"])*log(5 - 4*r_cc + 4*r_cc^2)
logL_cr <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*(3*log(2) + log(5)) + (n03 + n20)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n00 + n23)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n11 + n12)*log(7 + 4*r - 4*r^2) + (n01 + n22)*log(6 - 5*r + 2*r^2) + (n02 + n21)*log(3 + r + 2*r^2) + (n10 + n13)*log(3 - 4*r + 4*r^2)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_cr <- -3*(x[,"n_11"] + x[,"n_12"])*log10(2) - (x[,"n_10"] + x[,"n_13"])*log10(2) - 2*(x[,"n_02"] + x[,"n_21"])*log10(2) - 2*(x[,"n_01"] + x[,"n_22"])*log10(2) + (x[,"n_03"] + x[,"n_20"])*(log10(3 - 2*r_cr) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_23"])*(log10(pmax(1e-6,1 - r_cr)) + log10(1 + 2*r_cr)) + (x[,"n_11"] + x[,"n_12"])*log10(7 + 4*r_cr - 4*r_cr^2) + (x[,"n_01"] + x[,"n_22"])*log10(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_02"] + x[,"n_21"])*log10(3 + r_cr + 2*r_cr^2) + (x[,"n_10"] + x[,"n_13"])*log10(3 - 4*r_cr + 4*r_cr^2)
logL_cr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*(3*log(2) + log(5)) + (x[,"n_03"] + x[,"n_20"])*(log(3 - 2*r_cr) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_23"])*(log(pmax(1e-6,1 - r_cr)) + log(1 + 2*r_cr)) + (x[,"n_11"] + x[,"n_12"])*log(7 + 4*r_cr - 4*r_cr^2) + (x[,"n_01"] + x[,"n_22"])*log(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_02"] + x[,"n_21"])*log(3 + r_cr + 2*r_cr^2) + (x[,"n_10"] + x[,"n_13"])*log(3 - 4*r_cr + 4*r_cr^2)
logL_rc <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
L <- (-n00 - 2*n01 - 2*n02 - n03 - n10 - 2*n11 - 2*n12 - n13 - n20 - 2*n21 - 2*n22 - n23)*log(2) + 2*(-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*log(5) + (n03 + n20)*(log(3 - r) + log(pmax(1e-6,r))) + (n00 + n23)*(log(pmax(1e-6,1 - r)) + log(2 + r)) + (n11 + n12)*log(19 + 4*r - 4*r^2) + (n01 + n22)*log(12 - 5*r + 2*r^2) + (n10 + n13)*log(3 - 2*r + 2*r^2) + (n02 + n21)*log(9 + r + 2*r^2)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_rc <- -((x[,"n_03"] + x[,"n_20"])*(-2*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_23"])*(-2*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_13"])*(-log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_21"])*(log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_22"])*(log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_12"])*(2*log10(2) + log10(5)) + (x[,"n_03"] + x[,"n_20"])*(log10(3 - r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_23"])*(log10(pmax(1e-6,1 - r_rc)) + log10(2 + r_rc)) + (x[,"n_11"] + x[,"n_12"])*log10(19 + 4*r_rc - 4*r_rc^2) + (x[,"n_01"] + x[,"n_22"])*log10(12 - 5*r_rc + 2*r_rc^2) + (x[,"n_10"] + x[,"n_13"])*log10(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_02"] + x[,"n_21"])*log10(9 + r_rc + 2*r_rc^2)
logL_rc <- (-x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*log(5) + (x[,"n_03"] + x[,"n_20"])*(log(3 - r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_23"])*(log(pmax(1e-6,1 - r_rc)) + log(2 + r_rc)) + (x[,"n_11"] + x[,"n_12"])*log(19 + 4*r_rc - 4*r_rc^2) + (x[,"n_01"] + x[,"n_22"])*log(12 - 5*r_rc + 2*r_rc^2) + (x[,"n_10"] + x[,"n_13"])*log(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_02"] + x[,"n_21"])*log(9 + r_rc + 2*r_rc^2)
logL_rr <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*(3*log(2) + 2*log(5)) + (n03 + n20)*(log(3 - 2*r) + log(3 - r)) + (n00 + n23)*(log(2 + r) + log(1 + 2*r)) + (n10 + n13)*log(9 + 4*r - 4*r^2) + (n02 + n21)*log(24 - 7*r - 2*r^2) + (n01 + n22)*log(15 + 11*r - 2*r^2) + (n11 + n12)*log(41 - 4*r + 4*r^2)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_rr <- -((x[,"n_03"] + x[,"n_20"])*log10(5)) - (x[,"n_00"] + x[,"n_23"])*log10(5) - (x[,"n_10"] + x[,"n_13"])*(log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_21"])*(2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_22"])*(2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_12"])*(3*log10(2) + log10(5)) + (x[,"n_03"] + x[,"n_20"])*(log10(3 - 2*r_rr) + log10(3 - r_rr)) + (x[,"n_00"] + x[,"n_23"])*(log10(2 + r_rr) + log10(1 + 2*r_rr)) + (x[,"n_10"] + x[,"n_13"])*log10(9 + 4*r_rr - 4*r_rr^2) + (x[,"n_02"] + x[,"n_21"])*log10(24 - 7*r_rr - 2*r_rr^2) + (x[,"n_01"] + x[,"n_22"])*log10(15 + 11*r_rr - 2*r_rr^2) + (x[,"n_11"] + x[,"n_12"])*log10(41 - 4*r_rr + 4*r_rr^2)
logL_rr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*(3*log(2) + 2*log(5)) + (x[,"n_03"] + x[,"n_20"])*(log(3 - 2*r_rr) + log(3 - r_rr)) + (x[,"n_00"] + x[,"n_23"])*(log(2 + r_rr) + log(1 + 2*r_rr)) + (x[,"n_10"] + x[,"n_13"])*log(9 + 4*r_rr - 4*r_rr^2) + (x[,"n_02"] + x[,"n_21"])*log(24 - 7*r_rr - 2*r_rr^2) + (x[,"n_01"] + x[,"n_22"])*log(15 + 11*r_rr - 2*r_rr^2) + (x[,"n_11"] + x[,"n_12"])*log(41 - 4*r_rr + 4*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_1.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-3*n00 - 2*n01 - 3*n02 - 2*n03 - 3*n04 - 2*n10 - n11 - n12 - n13 - 2*n14 - 3*n20 - 2*n21 - 3*n22 - 2*n23 - 3*n24)*log(2) + (n02 + n12 + n22)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + 2*(n00 + n24)*log(pmax(1e-6,1 - r)) + 2*(n04 + n20)*log(pmax(1e-6,r)) + (n10 + n14)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n21)*(log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n22)*log(1 + 2*r - 2*r^2) + (n11 + n13)*log(1 + r - r^2) + (n01 + n23)*log(3 - 4*r + r^2) + n12*log(1 - r + r^2)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_cc <- 2*(x[,"n_10"] + x[,"n_14"])*log10(2) + 2*(x[,"n_04"] + x[,"n_20"])*log10(2) + 2*(x[,"n_00"] + x[,"n_24"])*log10(2) + x[,"n_12"]*(2*log10(2) - log10(3)) - (x[,"n_02"] + x[,"n_22"])*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_13"])*(-2*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_21"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_23"])*(-2*log10(2) + log10(5)) + 2*(x[,"n_00"] + x[,"n_24"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_20"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_14"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_02"] + x[,"n_22"])*log10(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_11"] + x[,"n_13"])*log10(1 + r_cc - r_cc^2) + (x[,"n_01"] + x[,"n_23"])*log10(3 - 4*r_cc + r_cc^2) + x[,"n_12"]*log10(1 - r_cc + r_cc^2)
logL_cc <- (-3*x[,"n_00"] - 2*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_20"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_22"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + 2*(x[,"n_00"] + x[,"n_24"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_20"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_14"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_02"] + x[,"n_22"])*log(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_11"] + x[,"n_13"])*log(1 + r_cc - r_cc^2) + (x[,"n_01"] + x[,"n_23"])*log(3 - 4*r_cc + r_cc^2) + x[,"n_12"]*log(1 - r_cc + r_cc^2)
logL_cr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-3*n00 - 3*n01 - 2*n02 - 3*n03 - 3*n04 - 3*n10 - 2*n11 - 2*n12 - 2*n13 - 3*n14 - 3*n20 - 3*n21 - 2*n22 - 3*n23 - 3*n24)*log(2) + (n02 + n12 + n22)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + (n00 + n04 + n20 + n24)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n23)*log(3 - 2*r^2) + n12*log(1 + 2*r - 2*r^2) + (n03 + n21)*log(1 + 4*r - 2*r^2) + (n02 + n22)*log(1 - r + r^2) + (n10 + n14)*log(1 - 2*r + 2*r^2) + (n11 + n13)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_cr <- (x[,"n_10"] + x[,"n_14"])*log10(2) + 2*(x[,"n_00"] + x[,"n_04"] + x[,"n_20"] + x[,"n_24"])*log10(2) + (x[,"n_02"] + x[,"n_22"])*(2*log10(2) - log10(3)) - x[,"n_12"]*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_13"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_21"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_23"])*(-log10(2) + log10(5)) + (x[,"n_00"] + x[,"n_04"] + x[,"n_20"] + x[,"n_24"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_23"])*log10(3 - 2*r_cr^2) + x[,"n_12"]*log10(1 + 2*r_cr - 2*r_cr^2) + (x[,"n_03"] + x[,"n_21"])*log10(1 + 4*r_cr - 2*r_cr^2) + (x[,"n_02"] + x[,"n_22"])*log10(1 - r_cr + r_cr^2) + (x[,"n_10"] + x[,"n_14"])*log10(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_11"] + x[,"n_13"])*log10(3 - 2*r_cr + 2*r_cr^2)
logL_cr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 2*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_22"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_00"] + x[,"n_04"] + x[,"n_20"] + x[,"n_24"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_23"])*log(3 - 2*r_cr^2) + x[,"n_12"]*log(1 + 2*r_cr - 2*r_cr^2) + (x[,"n_03"] + x[,"n_21"])*log(1 + 4*r_cr - 2*r_cr^2) + (x[,"n_02"] + x[,"n_22"])*log(1 - r_cr + r_cr^2) + (x[,"n_10"] + x[,"n_14"])*log(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_11"] + x[,"n_13"])*log(3 - 2*r_cr + 2*r_cr^2)
logL_rc <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-3*n00 - 3*n01 - 2*n02 - 3*n03 - 3*n04 - 3*n10 - 2*n11 - 2*n12 - 2*n13 - 3*n14 - 3*n20 - 3*n21 - 2*n22 - 3*n23 - 3*n24)*log(2) + (n02 + n12 + n22)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + (n04 + n20)*(log(3 - r) + log(pmax(1e-6,r))) + (n00 + n24)*(log(pmax(1e-6,1 - r)) + log(2 + r)) + (n01 + n23)*log(15 - 4*r - 2*r^2) + n12*log(7 + 2*r - 2*r^2) + (n03 + n21)*log(9 + 8*r - 2*r^2) + (n02 + n22)*log(4 - r + r^2) + (n10 + n14)*log(3 - 2*r + 2*r^2) + (n11 + n13)*log(13 - 2*r + 2*r^2)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_rc <- -((x[,"n_04"] + x[,"n_20"])*(-2*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_14"])*(-log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_22"])*(-2*log10(2) + log10(3) + log10(5)) - x[,"n_12"]*(-log10(2) + log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_13"])*(-log10(2) + 2*log10(5)) - (x[,"n_03"] + x[,"n_21"])*(-log10(2) + 2*log10(5)) - (x[,"n_01"] + x[,"n_23"])*(-log10(2) + 2*log10(5)) + (x[,"n_04"] + x[,"n_20"])*(log10(3 - r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_24"])*(log10(pmax(1e-6,1 - r_rc)) + log10(2 + r_rc)) + (x[,"n_01"] + x[,"n_23"])*log10(15 - 4*r_rc - 2*r_rc^2) + x[,"n_12"]*log10(7 + 2*r_rc - 2*r_rc^2) + (x[,"n_03"] + x[,"n_21"])*log10(9 + 8*r_rc - 2*r_rc^2) + (x[,"n_02"] + x[,"n_22"])*log10(4 - r_rc + r_rc^2) + (x[,"n_10"] + x[,"n_14"])*log10(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_11"] + x[,"n_13"])*log10(13 - 2*r_rc + 2*r_rc^2)
logL_rc <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 2*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_22"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_04"] + x[,"n_20"])*(log(3 - r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_24"])*(log(pmax(1e-6,1 - r_rc)) + log(2 + r_rc)) + (x[,"n_01"] + x[,"n_23"])*log(15 - 4*r_rc - 2*r_rc^2) + x[,"n_12"]*log(7 + 2*r_rc - 2*r_rc^2) + (x[,"n_03"] + x[,"n_21"])*log(9 + 8*r_rc - 2*r_rc^2) + (x[,"n_02"] + x[,"n_22"])*log(4 - r_rc + r_rc^2) + (x[,"n_10"] + x[,"n_14"])*log(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_11"] + x[,"n_13"])*log(13 - 2*r_rc + 2*r_rc^2)
logL_rr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-3*n00 - 2*n01 - 3*n02 - 2*n03 - 3*n04 - 2*n10 - n11 - n12 - n13 - 2*n14 - 3*n20 - 2*n21 - 3*n22 - 2*n23 - 3*n24)*log(2) + (n02 + n12 + n22)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + (n04 + n20)*(log(pmax(1e-6,1 - r)) + log(3 - r)) + (n00 + n24)*(log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n22)*log(7 + 2*r - 2*r^2) + (n10 + n14)*log(1 + r - r^2) + (n11 + n13)*log(6 + r - r^2) + (n03 + n21)*log(10 - 8*r + r^2) + n12*log(4 - r + r^2) + (n01 + n23)*log(3 + 6*r + r^2)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_rr <- -((x[,"n_10"] + x[,"n_14"])*(-2*log10(2) + log10(5))) - (x[,"n_04"] + x[,"n_20"])*(-2*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - x[,"n_12"]*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_22"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_13"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_03"] + x[,"n_21"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_01"] + x[,"n_23"])*(-2*log10(2) + 2*log10(5)) + (x[,"n_04"] + x[,"n_20"])*(log10(pmax(1e-6,1 - r_rr)) + log10(3 - r_rr)) + (x[,"n_00"] + x[,"n_24"])*(log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_02"] + x[,"n_22"])*log10(7 + 2*r_rr - 2*r_rr^2) + (x[,"n_10"] + x[,"n_14"])*log10(1 + r_rr - r_rr^2) + (x[,"n_11"] + x[,"n_13"])*log10(6 + r_rr - r_rr^2) + (x[,"n_03"] + x[,"n_21"])*log10(10 - 8*r_rr + r_rr^2) + x[,"n_12"]*log10(4 - r_rr + r_rr^2) + (x[,"n_01"] + x[,"n_23"])*log10(3 + 6*r_rr + r_rr^2)
logL_rr <- (-3*x[,"n_00"] - 2*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_20"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_22"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_04"] + x[,"n_20"])*(log(pmax(1e-6,1 - r_rr)) + log(3 - r_rr)) + (x[,"n_00"] + x[,"n_24"])*(log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_02"] + x[,"n_22"])*log(7 + 2*r_rr - 2*r_rr^2) + (x[,"n_10"] + x[,"n_14"])*log(1 + r_rr - r_rr^2) + (x[,"n_11"] + x[,"n_13"])*log(6 + r_rr - r_rr^2) + (x[,"n_03"] + x[,"n_21"])*log(10 - 8*r_rr + r_rr^2) + x[,"n_12"]*log(4 - r_rr + r_rr^2) + (x[,"n_01"] + x[,"n_23"])*log(3 + 6*r_rr + r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_1.4 <- function(x){
logL_cc <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24)*(3*log(2) + log(5)) + (n01 + n24)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n04 + n21)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n11 + n14)*log(1 + 4*r - 4*r^2) + (n02 + n23)*log(6 - 3*r - 2*r^2) + (n03 + n22)*log(1 + 7*r - 2*r^2) + (n12 + n13)*log(9 - 4*r + 4*r^2)
return(L)}
interlogL_cc <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
optimize(logL_cc,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_cc <- -3*(x[,"n_12"] + x[,"n_13"])*log10(2) - (x[,"n_11"] + x[,"n_14"])*log10(2) - 2*(x[,"n_03"] + x[,"n_22"])*log10(2) - 2*(x[,"n_02"] + x[,"n_23"])*log10(2) + (x[,"n_01"] + x[,"n_24"])*(log10(3 - 2*r_cc) + log10(pmax(1e-6,1 - r_cc))) + (x[,"n_04"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(1 + 2*r_cc)) + (x[,"n_11"] + x[,"n_14"])*log10(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_02"] + x[,"n_23"])*log10(6 - 3*r_cc - 2*r_cc^2) + (x[,"n_03"] + x[,"n_22"])*log10(1 + 7*r_cc - 2*r_cc^2) + (x[,"n_12"] + x[,"n_13"])*log10(9 - 4*r_cc + 4*r_cc^2)
logL_cc <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*(3*log(2) + log(5)) + (x[,"n_01"] + x[,"n_24"])*(log(3 - 2*r_cc) + log(pmax(1e-6,1 - r_cc))) + (x[,"n_04"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(1 + 2*r_cc)) + (x[,"n_11"] + x[,"n_14"])*log(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_02"] + x[,"n_23"])*log(6 - 3*r_cc - 2*r_cc^2) + (x[,"n_03"] + x[,"n_22"])*log(1 + 7*r_cc - 2*r_cc^2) + (x[,"n_12"] + x[,"n_13"])*log(9 - 4*r_cc + 4*r_cc^2)
logL_cr <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
L <- (-n01 - 2*n02 - 2*n03 - n04 - n11 - 2*n12 - 2*n13 - n14 - n21 - 2*n22 - 2*n23 - n24)*log(2) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24)*log(5) + (n01 + n04 + n21 + n24)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n12 + n13)*log(3 + 4*r - 4*r^2) + (n02 + n23)*log(3 - 3*r + 2*r^2) + (n11 + n14)*log(1 - 2*r + 2*r^2) + (n03 + n22)*log(2 - r + 2*r^2)
return(L)}
interlogL_cr <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
optimize(logL_cr,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_cr <- -2*(x[,"n_12"] + x[,"n_13"])*log10(2) + (x[,"n_11"] + x[,"n_14"])*log10(2) - (x[,"n_03"] + x[,"n_22"])*log10(2) - (x[,"n_02"] + x[,"n_23"])*log10(2) + 2*(x[,"n_01"] + x[,"n_04"] + x[,"n_21"] + x[,"n_24"])*log10(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_21"] + x[,"n_24"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_12"] + x[,"n_13"])*log10(3 + 4*r_cr - 4*r_cr^2) + (x[,"n_02"] + x[,"n_23"])*log10(3 - 3*r_cr + 2*r_cr^2) + (x[,"n_11"] + x[,"n_14"])*log10(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_03"] + x[,"n_22"])*log10(2 - r_cr + 2*r_cr^2)
logL_cr <- (-x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_01"] + x[,"n_04"] + x[,"n_21"] + x[,"n_24"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_12"] + x[,"n_13"])*log(3 + 4*r_cr - 4*r_cr^2) + (x[,"n_02"] + x[,"n_23"])*log(3 - 3*r_cr + 2*r_cr^2) + (x[,"n_11"] + x[,"n_14"])*log(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_03"] + x[,"n_22"])*log(2 - r_cr + 2*r_cr^2)
logL_rc <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24)*(3*log(2) + 2*log(5)) + (n01 + n24)*(log(3 - 2*r) + log(2 + r)) + (n04 + n21)*(log(3 - r) + log(1 + 2*r)) + (n12 + n13)*log(39 + 4*r - 4*r^2) + (n02 + n23)*log(21 - 3*r + 2*r^2) + (n03 + n22)*log(20 - r + 2*r^2) + (n11 + n14)*log(11 - 4*r + 4*r^2)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_rc <- -((x[,"n_04"] + x[,"n_21"])*log10(5)) - (x[,"n_01"] + x[,"n_24"])*log10(5) - (x[,"n_11"] + x[,"n_14"])*(log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_22"])*(2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_23"])*(2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_13"])*(3*log10(2) + log10(5)) + (x[,"n_01"] + x[,"n_24"])*(log10(3 - 2*r_rc) + log10(2 + r_rc)) + (x[,"n_04"] + x[,"n_21"])*(log10(3 - r_rc) + log10(1 + 2*r_rc)) + (x[,"n_12"] + x[,"n_13"])*log10(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_02"] + x[,"n_23"])*log10(21 - 3*r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_22"])*log10(20 - r_rc + 2*r_rc^2) + (x[,"n_11"] + x[,"n_14"])*log10(11 - 4*r_rc + 4*r_rc^2)
logL_rc <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*(3*log(2) + 2*log(5)) + (x[,"n_01"] + x[,"n_24"])*(log(3 - 2*r_rc) + log(2 + r_rc)) + (x[,"n_04"] + x[,"n_21"])*(log(3 - r_rc) + log(1 + 2*r_rc)) + (x[,"n_12"] + x[,"n_13"])*log(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_02"] + x[,"n_23"])*log(21 - 3*r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_22"])*log(20 - r_rc + 2*r_rc^2) + (x[,"n_11"] + x[,"n_14"])*log(11 - 4*r_rc + 4*r_rc^2)
logL_rr <- function(r,n01,n02,n03,n04,n12,n13,n21,n22,n23,n24,n11,n14) {
L <- (-n01 - 2*n02 - 2*n03 - n04 - 2*n12 - 2*n13 - n21 - 2*n22 - 2*n23 - n24)*log(2) + 2*(-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24)*log(5) + (n04 + n21)*(log(pmax(1e-6,1 - r)) + log(3 - r)) + (n01 + n24)*(log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n22)*log(13 - 5*r - 2*r^2) + (n02 + n23)*log(6 + 9*r - 2*r^2) + (n11 + n14)*log(1 + r - r^2) + (n12 + n13)*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_rr <- function(n01,n02,n03,n04,n12,n13,n21,n22,n23,n24,n11,n14) {
optimize(logL_rr,c(0,0.5), n01,n02,n03,n04,n12,n13,n21,n22,n23,n24,n11,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_11"],x[,"n_14"])
LOD_rr <- -((x[,"n_11"] + x[,"n_14"])*(-2*log10(2) + log10(5))) - (x[,"n_04"] + x[,"n_21"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_22"])*(log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_23"])*(log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_13"])*(2*log10(2) + log10(5)) + (x[,"n_04"] + x[,"n_21"])*(log10(pmax(1e-6,1 - r_rr)) + log10(3 - r_rr)) + (x[,"n_01"] + x[,"n_24"])*(log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_03"] + x[,"n_22"])*log10(13 - 5*r_rr - 2*r_rr^2) + (x[,"n_02"] + x[,"n_23"])*log10(6 + 9*r_rr - 2*r_rr^2) + (x[,"n_11"] + x[,"n_14"])*log10(1 + r_rr - r_rr^2) + (x[,"n_12"] + x[,"n_13"])*log10(21 - 4*r_rr + 4*r_rr^2)
logL_rr <- (-x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_04"] + x[,"n_21"])*(log(pmax(1e-6,1 - r_rr)) + log(3 - r_rr)) + (x[,"n_01"] + x[,"n_24"])*(log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_03"] + x[,"n_22"])*log(13 - 5*r_rr - 2*r_rr^2) + (x[,"n_02"] + x[,"n_23"])*log(6 + 9*r_rr - 2*r_rr^2) + (x[,"n_11"] + x[,"n_14"])*log(1 + r_rr - r_rr^2) + (x[,"n_12"] + x[,"n_13"])*log(21 - 4*r_rr + 4*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_1.5 <- function(x){
logL_cc <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24) {
L <- (-2*n02 - n03 - 2*n04 - n12 - n13 - n14 - 2*n22 - n23 - 2*n24)*log(2) + (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24)*log(5) + (n02 + n24)*(log(pmax(1e-6,1 - r)) + log(3 - r)) + (n04 + n22)*(log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n12 + n14 + n23)*log(1 + r - r^2) + n13*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_cc <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24) {
optimize(logL_cc,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_cc <- -((x[,"n_04"] + x[,"n_22"])*(-2*log10(2) + log10(5))) - (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - x[,"n_13"]*(-log10(2) + log10(5)) + (x[,"n_02"] + x[,"n_24"])*(log10(pmax(1e-6,1 - r_cc)) + log10(3 - r_cc)) + (x[,"n_04"] + x[,"n_22"])*(log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*log10(1 + r_cc - r_cc^2) + x[,"n_13"]*log10(3 - 2*r_cc + 2*r_cc^2)
logL_cc <- (-2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"])*log(2) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_02"] + x[,"n_24"])*(log(pmax(1e-6,1 - r_cc)) + log(3 - r_cc)) + (x[,"n_04"] + x[,"n_22"])*(log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*log(1 + r_cc - r_cc^2) + x[,"n_13"]*log(3 - 2*r_cc + 2*r_cc^2)
logL_cr <- function(r,n02,n03,n04,n12,n14,n22,n23,n24,n13) {
L <- 2*(-n02 - n03 - n04 - n12 - n14 - n22 - n23 - n24)*log(2) + (n02 + n04 + n13 + n22 + n24)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n12 + n14 + n23)*log(1 - 2*r + 2*r^2)
return(L)}
interlogL_cr <- function(n02,n03,n04,n12,n14,n22,n23,n24,n13) {
optimize(logL_cr,c(0,0.5), n02,n03,n04,n12,n14,n22,n23,n24,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_13"])
LOD_cr <- (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*log10(2) + 2*(x[,"n_02"] + x[,"n_04"] + x[,"n_13"] + x[,"n_22"] + x[,"n_24"])*log10(2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_13"] + x[,"n_22"] + x[,"n_24"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*log10(1 - 2*r_cr + 2*r_cr^2)
logL_cr <- 2*(-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_13"] + x[,"n_22"] + x[,"n_24"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*log(1 - 2*r_cr + 2*r_cr^2)
logL_rc <- function(r,n02,n03,n04,n12,n14,n22,n23,n24,n13) {
L <- 2*(-n02 - n03 - n04 - n12 - n14 - n22 - n23 - n24)*log(2) + 2*(-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24)*log(5) + (n02 + n04 + n13 + n22 + n24)*(log(3 - r) + log(2 + r)) + (n03 + n12 + n14 + n23)*log(13 - 2*r + 2*r^2)
return(L)}
interlogL_rc <- function(n02,n03,n04,n12,n14,n22,n23,n24,n13) {
optimize(logL_rc,c(0,0.5), n02,n03,n04,n12,n14,n22,n23,n24,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_13"])
LOD_rc <- -2*(x[,"n_02"] + x[,"n_04"] + x[,"n_13"] + x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*(-log10(2) + 2*log10(5)) + (x[,"n_02"] + x[,"n_04"] + x[,"n_13"] + x[,"n_22"] + x[,"n_24"])*(log10(3 - r_rc) + log10(2 + r_rc)) + (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*log10(13 - 2*r_rc + 2*r_rc^2)
logL_rc <- 2*(-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(2) + 2*(-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_02"] + x[,"n_04"] + x[,"n_13"] + x[,"n_22"] + x[,"n_24"])*(log(3 - r_rc) + log(2 + r_rc)) + (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*log(13 - 2*r_rc + 2*r_rc^2)
logL_rr <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24) {
L <- (-2*n02 - n03 - 2*n04 - n12 - n13 - n14 - 2*n22 - n23 - 2*n24)*log(2) + (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24)*log(5) + (n04 + n22)*(log(pmax(1e-6,1 - r)) + log(3 - r)) + (n02 + n24)*(log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n12 + n14 + n23)*log(1 + r - r^2) + n13*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_rr <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24) {
optimize(logL_rr,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_rr <- -((x[,"n_04"] + x[,"n_22"])*(-2*log10(2) + log10(5))) - (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - x[,"n_13"]*(-log10(2) + log10(5)) + (x[,"n_04"] + x[,"n_22"])*(log10(pmax(1e-6,1 - r_rr)) + log10(3 - r_rr)) + (x[,"n_02"] + x[,"n_24"])*(log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*log10(1 + r_rr - r_rr^2) + x[,"n_13"]*log10(3 - 2*r_rr + 2*r_rr^2)
logL_rr <- (-2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"])*log(2) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_04"] + x[,"n_22"])*(log(pmax(1e-6,1 - r_rr)) + log(3 - r_rr)) + (x[,"n_02"] + x[,"n_24"])*(log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_03"] + x[,"n_12"] + x[,"n_14"] + x[,"n_23"])*log(1 + r_rr - r_rr^2) + x[,"n_13"]*log(3 - 2*r_rr + 2*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_2.0 <- function(x){
logL_c <- function(r,n00,n01,n02,n10,n11,n12,n20,n21,n22) {
L <- (-n00 - 2*n01 - n02 - n10 - n11 - n12 - n20 - 2*n21 - n22)*log(2) + (n01 + n11 + n21)*log(3) + (-n00 - n01 - n02 - n10 - n11 - n12 - n20 - n21 - n22)*log(5) + (n00 + n22)*log(pmax(1e-6,1 - r)) + (n02 + n20)*log(pmax(1e-6,r))
return(L)}
interlogL_c <- function(n00,n01,n02,n10,n11,n12,n20,n21,n22) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n10,n11,n12,n20,n21,n22, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_20"],x[,"n_21"],x[,"n_22"])
LOD_c <- (x[,"n_02"] + x[,"n_20"])*log10(2) + (x[,"n_00"] + x[,"n_22"])*log10(2) + (x[,"n_00"] + x[,"n_22"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_20"])*log10(pmax(1e-6,r_c))
logL_c <- (-x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"])*log(2) + (x[,"n_01"] + x[,"n_11"] + x[,"n_21"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"])*log(5) + (x[,"n_00"] + x[,"n_22"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_02"] + x[,"n_20"])*log(pmax(1e-6,r_c))
logL_r <- function(r,n00,n01,n02,n10,n11,n12,n20,n21,n22) {
L <- (-3*n00 - 2*n01 - 3*n02 - n10 - n11 - n12 - 3*n20 - 2*n21 - 3*n22)*log(2) + (n01 + n11 + n21)*log(3) + (-n00 - n01 - n02 - n10 - n11 - n12 - n20 - n21 - n22)*log(5) + (n02 + n20)*log(3 - 2*r) + (n00 + n22)*log(1 + 2*r)
return(L)}
interlogL_r <- function(n00,n01,n02,n10,n11,n12,n20,n21,n22) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n10,n11,n12,n20,n21,n22, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_20"],x[,"n_21"],x[,"n_22"])
LOD_r <- -((x[,"n_02"] + x[,"n_20"])*log10(2)) - (x[,"n_00"] + x[,"n_22"])*log10(2) + (x[,"n_02"] + x[,"n_20"])*log10(3 - 2*r_r) + (x[,"n_00"] + x[,"n_22"])*log10(1 + 2*r_r)
logL_r <- (-3*x[,"n_00"] - 2*x[,"n_01"] - 3*x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - 3*x[,"n_20"] - 2*x[,"n_21"] - 3*x[,"n_22"])*log(2) + (x[,"n_01"] + x[,"n_11"] + x[,"n_21"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"])*log(5) + (x[,"n_02"] + x[,"n_20"])*log(3 - 2*r_r) + (x[,"n_00"] + x[,"n_22"])*log(1 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_2.2 <- function(x){
logL_cc <- function(r,n02,n10,n12,n14,n22,n01,n03,n11,n13,n21,n23,n00,n04,n20,n24) {
L <- (-2*n02 + n10 - n12 + n14 - 2*n22)*log(2) + (n01 + n03 + n11 + n13 + n21 + n23)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + (n01 + n23)*log(pmax(1e-6,1 - r)) + 2*(n00 + n24)*log(pmax(1e-6,1 - r)) + 2*(n04 + n20)*log(pmax(1e-6,r)) + (n03 + n21)*log(pmax(1e-6,r)) + (n10 + n14)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n22)*log(9 + 8*r - 8*r^2) + n12*log(13 - 8*r + 8*r^2)
return(L)}
interlogL_cc <- function(n02,n10,n12,n14,n22,n01,n03,n11,n13,n21,n23,n00,n04,n20,n24) {
optimize(logL_cc,c(0,0.5), n02,n10,n12,n14,n22,n01,n03,n11,n13,n21,n23,n00,n04,n20,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_02"],x[,"n_10"],x[,"n_12"],x[,"n_14"],x[,"n_22"],x[,"n_01"],x[,"n_03"],x[,"n_11"],x[,"n_13"],x[,"n_21"],x[,"n_23"],x[,"n_00"],x[,"n_04"],x[,"n_20"],x[,"n_24"])
LOD_cc <- 2*(x[,"n_10"] + x[,"n_14"])*log10(2) + 2*(x[,"n_04"] + x[,"n_20"])*log10(2) + (x[,"n_03"] + x[,"n_21"])*log10(2) + (x[,"n_01"] + x[,"n_23"])*log10(2) + 2*(x[,"n_00"] + x[,"n_24"])*log10(2) - x[,"n_12"]*log10(11) - (x[,"n_02"] + x[,"n_22"])*log10(11) + (x[,"n_01"] + x[,"n_23"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_00"] + x[,"n_24"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_20"])*log10(pmax(1e-6,r_cc)) + (x[,"n_03"] + x[,"n_21"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_14"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_22"])*log10(9 + 8*r_cc - 8*r_cc^2) + x[,"n_12"]*log10(13 - 8*r_cc + 8*r_cc^2)
logL_cc <- (-2*x[,"n_02"] + x[,"n_10"] - x[,"n_12"] + x[,"n_14"] - 2*x[,"n_22"])*log(2) + (x[,"n_01"] + x[,"n_03"] + x[,"n_11"] + x[,"n_13"] + x[,"n_21"] + x[,"n_23"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_01"] + x[,"n_23"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_00"] + x[,"n_24"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_20"])*log(pmax(1e-6,r_cc)) + (x[,"n_03"] + x[,"n_21"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_14"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_22"])*log(9 + 8*r_cc - 8*r_cc^2) + x[,"n_12"]*log(13 - 8*r_cc + 8*r_cc^2)
logL_cr <- function(r,n00,n01,n03,n04,n10,n14,n20,n21,n23,n24,n11,n13,n02,n12,n22) {
L <- (-2*n00 - 3*n01 - 3*n03 - 2*n04 - 2*n10 - 2*n14 - 2*n20 - 3*n21 - 3*n23 - 2*n24)*log(2) + (n01 + n03 + n11 + n13 + n21 + n23)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + (n01 + n23)*log(5 - 2*r) + (n04 + n20)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n00 + n24)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n03 + n21)*log(3 + 2*r) + n12*log(5 + 2*r - 2*r^2) + (n02 + n22)*log(3 - r + r^2) + (n10 + n14)*log(3 - 4*r + 4*r^2)
return(L)}
interlogL_cr <- function(n00,n01,n03,n04,n10,n14,n20,n21,n23,n24,n11,n13,n02,n12,n22) {
optimize(logL_cr,c(0,0.5), n00,n01,n03,n04,n10,n14,n20,n21,n23,n24,n11,n13,n02,n12,n22, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_23"],x[,"n_24"],x[,"n_11"],x[,"n_13"],x[,"n_02"],x[,"n_12"],x[,"n_22"])
LOD_cr <- -((x[,"n_10"] + x[,"n_14"])*log10(2)) - 2*(x[,"n_03"] + x[,"n_21"])*log10(2) - 2*(x[,"n_01"] + x[,"n_23"])*log10(2) - (x[,"n_02"] + x[,"n_22"])*(-2*log10(2) + log10(11)) - x[,"n_12"]*(-log10(2) + log10(11)) + (x[,"n_01"] + x[,"n_23"])*log10(5 - 2*r_cr) + (x[,"n_04"] + x[,"n_20"])*(log10(3 - 2*r_cr) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_24"])*(log10(pmax(1e-6,1 - r_cr)) + log10(1 + 2*r_cr)) + (x[,"n_03"] + x[,"n_21"])*log10(3 + 2*r_cr) + x[,"n_12"]*log10(5 + 2*r_cr - 2*r_cr^2) + (x[,"n_02"] + x[,"n_22"])*log10(3 - r_cr + r_cr^2) + (x[,"n_10"] + x[,"n_14"])*log10(3 - 4*r_cr + 4*r_cr^2)
logL_cr <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_23"] - 2*x[,"n_24"])*log(2) + (x[,"n_01"] + x[,"n_03"] + x[,"n_11"] + x[,"n_13"] + x[,"n_21"] + x[,"n_23"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_01"] + x[,"n_23"])*log(5 - 2*r_cr) + (x[,"n_04"] + x[,"n_20"])*(log(3 - 2*r_cr) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_24"])*(log(pmax(1e-6,1 - r_cr)) + log(1 + 2*r_cr)) + (x[,"n_03"] + x[,"n_21"])*log(3 + 2*r_cr) + x[,"n_12"]*log(5 + 2*r_cr - 2*r_cr^2) + (x[,"n_02"] + x[,"n_22"])*log(3 - r_cr + r_cr^2) + (x[,"n_10"] + x[,"n_14"])*log(3 - 4*r_cr + 4*r_cr^2)
logL_rr <- function(r,n00,n01,n02,n03,n04,n10,n12,n14,n20,n21,n22,n23,n24,n11,n13) {
L <- (-4*n00 - 2*n01 - 3*n02 - 2*n03 - 4*n04 - 3*n10 - 2*n12 - 3*n14 - 4*n20 - 2*n21 - 3*n22 - 2*n23 - 4*n24)*log(2) + (n01 + n03 + n11 + n13 + n21 + n23)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + 2*(n04 + n20)*log(3 - 2*r) + (n03 + n21)*log(3 - 2*r) + (n01 + n23)*log(1 + 2*r) + 2*(n00 + n24)*log(1 + 2*r) + (n10 + n14)*(log(3 - 2*r) + log(1 + 2*r)) + (n02 + n22)*log(21 + 4*r - 4*r^2) + n12*log(23 - 4*r + 4*r^2)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n10,n12,n14,n20,n21,n22,n23,n24,n11,n13) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n10,n12,n14,n20,n21,n22,n23,n24,n11,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_12"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_11"],x[,"n_13"])
LOD_rr <- -2*(x[,"n_10"] + x[,"n_14"])*log10(2) - 2*(x[,"n_04"] + x[,"n_20"])*log10(2) - (x[,"n_03"] + x[,"n_21"])*log10(2) - (x[,"n_01"] + x[,"n_23"])*log10(2) - 2*(x[,"n_00"] + x[,"n_24"])*log10(2) - x[,"n_12"]*(log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_22"])*(log10(2) + log10(11)) + 2*(x[,"n_04"] + x[,"n_20"])*log10(3 - 2*r_rr) + (x[,"n_03"] + x[,"n_21"])*log10(3 - 2*r_rr) + (x[,"n_01"] + x[,"n_23"])*log10(1 + 2*r_rr) + 2*(x[,"n_00"] + x[,"n_24"])*log10(1 + 2*r_rr) + (x[,"n_10"] + x[,"n_14"])*(log10(3 - 2*r_rr) + log10(1 + 2*r_rr)) + (x[,"n_02"] + x[,"n_22"])*log10(21 + 4*r_rr - 4*r_rr^2) + x[,"n_12"]*log10(23 - 4*r_rr + 4*r_rr^2)
logL_rr <- (-4*x[,"n_00"] - 2*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_10"] - 2*x[,"n_12"] - 3*x[,"n_14"] - 4*x[,"n_20"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"])*log(2) + (x[,"n_01"] + x[,"n_03"] + x[,"n_11"] + x[,"n_13"] + x[,"n_21"] + x[,"n_23"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + 2*(x[,"n_04"] + x[,"n_20"])*log(3 - 2*r_rr) + (x[,"n_03"] + x[,"n_21"])*log(3 - 2*r_rr) + (x[,"n_01"] + x[,"n_23"])*log(1 + 2*r_rr) + 2*(x[,"n_00"] + x[,"n_24"])*log(1 + 2*r_rr) + (x[,"n_10"] + x[,"n_14"])*(log(3 - 2*r_rr) + log(1 + 2*r_rr)) + (x[,"n_02"] + x[,"n_22"])*log(21 + 4*r_rr - 4*r_rr^2) + x[,"n_12"]*log(23 - 4*r_rr + 4*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_2.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25)*log(2) + (n01 + n04 + n11 + n14 + n21 + n24)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(5) + 2*(n00 + n25)*log(pmax(1e-6,1 - r)) + 2*(n05 + n20)*log(pmax(1e-6,r)) + (n10 + n15)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n21)*(log(pmax(1e-6,r)) + log(3 + 2*r)) + (n02 + n23)*log(24 - 7*r - 8*r^2) + (n03 + n22)*log(9 + 23*r - 8*r^2) + (n11 + n14)*log(3 + 4*r - 4*r^2) + (n01 + n24)*log(5 - 7*r + 2*r^2) + (n12 + n13)*log(41 - 16*r + 16*r^2)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_cc <- -2*(x[,"n_11"] + x[,"n_14"])*log10(2) + 2*(x[,"n_10"] + x[,"n_15"])*log10(2) + 2*(x[,"n_05"] + x[,"n_20"])*log10(2) - (x[,"n_04"] + x[,"n_21"])*log10(2) - (x[,"n_01"] + x[,"n_24"])*log10(2) + 2*(x[,"n_00"] + x[,"n_25"])*log10(2) - (x[,"n_12"] + x[,"n_13"])*log10(37) - (x[,"n_03"] + x[,"n_22"])*(-log10(2) + log10(37)) - (x[,"n_02"] + x[,"n_23"])*(-log10(2) + log10(37)) + 2*(x[,"n_00"] + x[,"n_25"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_05"] + x[,"n_20"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_15"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(3 + 2*r_cc)) + (x[,"n_02"] + x[,"n_23"])*log10(24 - 7*r_cc - 8*r_cc^2) + (x[,"n_03"] + x[,"n_22"])*log10(9 + 23*r_cc - 8*r_cc^2) + (x[,"n_11"] + x[,"n_14"])*log10(3 + 4*r_cc - 4*r_cc^2) + (x[,"n_01"] + x[,"n_24"])*log10(5 - 7*r_cc + 2*r_cc^2) + (x[,"n_12"] + x[,"n_13"])*log10(41 - 16*r_cc + 16*r_cc^2)
logL_cc <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + 2*(x[,"n_00"] + x[,"n_25"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_05"] + x[,"n_20"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_15"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(3 + 2*r_cc)) + (x[,"n_02"] + x[,"n_23"])*log(24 - 7*r_cc - 8*r_cc^2) + (x[,"n_03"] + x[,"n_22"])*log(9 + 23*r_cc - 8*r_cc^2) + (x[,"n_11"] + x[,"n_14"])*log(3 + 4*r_cc - 4*r_cc^2) + (x[,"n_01"] + x[,"n_24"])*log(5 - 7*r_cc + 2*r_cc^2) + (x[,"n_12"] + x[,"n_13"])*log(41 - 16*r_cc + 16*r_cc^2)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 2*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25)*log(2) + (n01 + n04 + n11 + n14 + n21 + n24)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(5) + (n00 + n05 + n20 + n25)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n12 + n13)*log(33 + 16*r - 16*r^2) + (n01 + n24)*log(2 + r - 2*r^2) + (n04 + n21)*log(1 + 3*r - 2*r^2) + (n10 + n15)*log(1 - 2*r + 2*r^2) + (n11 + n14)*log(5 - 4*r + 4*r^2) + (n02 + n23)*log(21 - 9*r + 8*r^2) + (n03 + n22)*log(20 - 7*r + 8*r^2)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_cr <- -2*(x[,"n_11"] + x[,"n_14"])*log10(2) + (x[,"n_10"] + x[,"n_15"])*log10(2) - (x[,"n_04"] + x[,"n_21"])*log10(2) - (x[,"n_01"] + x[,"n_24"])*log10(2) + 2*(x[,"n_00"] + x[,"n_05"] + x[,"n_20"] + x[,"n_25"])*log10(2) - (x[,"n_12"] + x[,"n_13"])*log10(37) - (x[,"n_03"] + x[,"n_22"])*(-log10(2) + log10(37)) - (x[,"n_02"] + x[,"n_23"])*(-log10(2) + log10(37)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_12"] + x[,"n_13"])*log10(33 + 16*r_cr - 16*r_cr^2) + (x[,"n_01"] + x[,"n_24"])*log10(2 + r_cr - 2*r_cr^2) + (x[,"n_04"] + x[,"n_21"])*log10(1 + 3*r_cr - 2*r_cr^2) + (x[,"n_10"] + x[,"n_15"])*log10(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_11"] + x[,"n_14"])*log10(5 - 4*r_cr + 4*r_cr^2) + (x[,"n_02"] + x[,"n_23"])*log10(21 - 9*r_cr + 8*r_cr^2) + (x[,"n_03"] + x[,"n_22"])*log10(20 - 7*r_cr + 8*r_cr^2)
logL_cr <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_00"] + x[,"n_05"] + x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_12"] + x[,"n_13"])*log(33 + 16*r_cr - 16*r_cr^2) + (x[,"n_01"] + x[,"n_24"])*log(2 + r_cr - 2*r_cr^2) + (x[,"n_04"] + x[,"n_21"])*log(1 + 3*r_cr - 2*r_cr^2) + (x[,"n_10"] + x[,"n_15"])*log(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_11"] + x[,"n_14"])*log(5 - 4*r_cr + 4*r_cr^2) + (x[,"n_02"] + x[,"n_23"])*log(21 - 9*r_cr + 8*r_cr^2) + (x[,"n_03"] + x[,"n_22"])*log(20 - 7*r_cr + 8*r_cr^2)
logL_rc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- (-4*n00 - 4*n01 - 3*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 3*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25)*log(2) + (n01 + n04 + n11 + n14 + n21 + n24)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(5) + (n05 + n20)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n00 + n25)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n12 + n13)*log(35 + 8*r - 8*r^2) + (n01 + n24)*log(4 + r - 2*r^2) + (n04 + n21)*log(3 + 3*r - 2*r^2) + (n02 + n23)*log(21 - 7*r + 4*r^2) + (n10 + n15)*log(3 - 4*r + 4*r^2) + (n11 + n14)*log(9 - 4*r + 4*r^2) + (n03 + n22)*log(18 - r + 4*r^2)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_rc <- -3*(x[,"n_11"] + x[,"n_14"])*log10(2) - (x[,"n_10"] + x[,"n_15"])*log10(2) - 2*(x[,"n_04"] + x[,"n_21"])*log10(2) - 2*(x[,"n_01"] + x[,"n_24"])*log10(2) - (x[,"n_12"] + x[,"n_13"])*log10(37) - (x[,"n_03"] + x[,"n_22"])*(-log10(2) + log10(37)) - (x[,"n_02"] + x[,"n_23"])*(-log10(2) + log10(37)) + (x[,"n_05"] + x[,"n_20"])*(log10(3 - 2*r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_12"] + x[,"n_13"])*log10(35 + 8*r_rc - 8*r_rc^2) + (x[,"n_01"] + x[,"n_24"])*log10(4 + r_rc - 2*r_rc^2) + (x[,"n_04"] + x[,"n_21"])*log10(3 + 3*r_rc - 2*r_rc^2) + (x[,"n_02"] + x[,"n_23"])*log10(21 - 7*r_rc + 4*r_rc^2) + (x[,"n_10"] + x[,"n_15"])*log10(3 - 4*r_rc + 4*r_rc^2) + (x[,"n_11"] + x[,"n_14"])*log10(9 - 4*r_rc + 4*r_rc^2) + (x[,"n_03"] + x[,"n_22"])*log10(18 - r_rc + 4*r_rc^2)
logL_rc <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_05"] + x[,"n_20"])*(log(3 - 2*r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_12"] + x[,"n_13"])*log(35 + 8*r_rc - 8*r_rc^2) + (x[,"n_01"] + x[,"n_24"])*log(4 + r_rc - 2*r_rc^2) + (x[,"n_04"] + x[,"n_21"])*log(3 + 3*r_rc - 2*r_rc^2) + (x[,"n_02"] + x[,"n_23"])*log(21 - 7*r_rc + 4*r_rc^2) + (x[,"n_10"] + x[,"n_15"])*log(3 - 4*r_rc + 4*r_rc^2) + (x[,"n_11"] + x[,"n_14"])*log(9 - 4*r_rc + 4*r_rc^2) + (x[,"n_03"] + x[,"n_22"])*log(18 - r_rc + 4*r_rc^2)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- (-4*n00 - 4*n01 - 3*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 3*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25)*log(2) + (n01 + n04 + n11 + n14 + n21 + n24)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(5) + (n05 + n20)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n00 + n25)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n03 + n22)*log(23 - 7*r - 4*r^2) + (n10 + n15)*log(1 + 4*r - 4*r^2) + (n11 + n14)*log(7 + 4*r - 4*r^2) + (n02 + n23)*log(12 + 15*r - 4*r^2) + (n04 + n21)*log(8 - 9*r + 2*r^2) + (n01 + n24)*log(1 + 5*r + 2*r^2) + (n12 + n13)*log(39 - 8*r + 8*r^2)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_rr <- -3*(x[,"n_11"] + x[,"n_14"])*log10(2) - (x[,"n_10"] + x[,"n_15"])*log10(2) - 2*(x[,"n_04"] + x[,"n_21"])*log10(2) - 2*(x[,"n_01"] + x[,"n_24"])*log10(2) - (x[,"n_12"] + x[,"n_13"])*log10(37) - (x[,"n_03"] + x[,"n_22"])*(-log10(2) + log10(37)) - (x[,"n_02"] + x[,"n_23"])*(-log10(2) + log10(37)) + (x[,"n_05"] + x[,"n_20"])*(log10(3 - 2*r_rr) + log10(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_03"] + x[,"n_22"])*log10(23 - 7*r_rr - 4*r_rr^2) + (x[,"n_10"] + x[,"n_15"])*log10(1 + 4*r_rr - 4*r_rr^2) + (x[,"n_11"] + x[,"n_14"])*log10(7 + 4*r_rr - 4*r_rr^2) + (x[,"n_02"] + x[,"n_23"])*log10(12 + 15*r_rr - 4*r_rr^2) + (x[,"n_04"] + x[,"n_21"])*log10(8 - 9*r_rr + 2*r_rr^2) + (x[,"n_01"] + x[,"n_24"])*log10(1 + 5*r_rr + 2*r_rr^2) + (x[,"n_12"] + x[,"n_13"])*log10(39 - 8*r_rr + 8*r_rr^2)
logL_rr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_05"] + x[,"n_20"])*(log(3 - 2*r_rr) + log(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_03"] + x[,"n_22"])*log(23 - 7*r_rr - 4*r_rr^2) + (x[,"n_10"] + x[,"n_15"])*log(1 + 4*r_rr - 4*r_rr^2) + (x[,"n_11"] + x[,"n_14"])*log(7 + 4*r_rr - 4*r_rr^2) + (x[,"n_02"] + x[,"n_23"])*log(12 + 15*r_rr - 4*r_rr^2) + (x[,"n_04"] + x[,"n_21"])*log(8 - 9*r_rr + 2*r_rr^2) + (x[,"n_01"] + x[,"n_24"])*log(1 + 5*r_rr + 2*r_rr^2) + (x[,"n_12"] + x[,"n_13"])*log(39 - 8*r_rr + 8*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_2.4 <- function(x){
logL_cc <- function(r,n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n12,n14) {
L <- (-2*n01 - 3*n02 - n03 - 3*n04 - 2*n05 - 2*n11 + n13 - 2*n15 - 2*n21 - 3*n22 - n23 - 3*n24 - 2*n25)*log(2) + (n02 + n04 + n12 + n14 + n22 + n24)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25)*log(5) + (n02 + n24)*log(7 - 6*r) + (n01 + n25)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n05 + n21)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n04 + n22)*log(1 + 6*r) + (n11 + n15)*log(1 + 4*r - 4*r^2) + (n03 + n23)*log(5 + 2*r - 2*r^2) + n13*log(3 - r + r^2)
return(L)}
interlogL_cc <- function(n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n12,n14) {
optimize(logL_cc,c(0,0.5), n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n12,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_13"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_12"],x[,"n_14"])
LOD_cc <- -((x[,"n_11"] + x[,"n_15"])*log10(2)) - 2*(x[,"n_04"] + x[,"n_22"])*log10(2) - 2*(x[,"n_02"] + x[,"n_24"])*log10(2) - x[,"n_13"]*(-2*log10(2) + log10(11)) - (x[,"n_03"] + x[,"n_23"])*(-log10(2) + log10(11)) + (x[,"n_02"] + x[,"n_24"])*log10(7 - 6*r_cc) + (x[,"n_01"] + x[,"n_25"])*(log10(3 - 2*r_cc) + log10(pmax(1e-6,1 - r_cc))) + (x[,"n_05"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(1 + 2*r_cc)) + (x[,"n_04"] + x[,"n_22"])*log10(1 + 6*r_cc) + (x[,"n_11"] + x[,"n_15"])*log10(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_03"] + x[,"n_23"])*log10(5 + 2*r_cc - 2*r_cc^2) + x[,"n_13"]*log10(3 - r_cc + r_cc^2)
logL_cc <- (-2*x[,"n_01"] - 3*x[,"n_02"] - x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] + x[,"n_13"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 3*x[,"n_22"] - x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"])*log(2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_12"] + x[,"n_14"] + x[,"n_22"] + x[,"n_24"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_02"] + x[,"n_24"])*log(7 - 6*r_cc) + (x[,"n_01"] + x[,"n_25"])*(log(3 - 2*r_cc) + log(pmax(1e-6,1 - r_cc))) + (x[,"n_05"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(1 + 2*r_cc)) + (x[,"n_04"] + x[,"n_22"])*log(1 + 6*r_cc) + (x[,"n_11"] + x[,"n_15"])*log(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_03"] + x[,"n_23"])*log(5 + 2*r_cc - 2*r_cc^2) + x[,"n_13"]*log(3 - r_cc + r_cc^2)
logL_cr <- function(r,n02,n03,n04,n13,n22,n23,n24,n12,n14,n01,n05,n11,n15,n21,n25) {
L <- (-n02 - 2*n03 - n04 - n13 - n22 - 2*n23 - n24)*log(2) + (n02 + n04 + n12 + n14 + n22 + n24)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25)*log(5) + (n01 + n05 + n21 + n25)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n13*log(9 + 8*r - 8*r^2) + (n11 + n15)*log(1 - 2*r + 2*r^2) + (n03 + n23)*log(13 - 8*r + 8*r^2)
return(L)}
interlogL_cr <- function(n02,n03,n04,n13,n22,n23,n24,n12,n14,n01,n05,n11,n15,n21,n25) {
optimize(logL_cr,c(0,0.5), n02,n03,n04,n13,n22,n23,n24,n12,n14,n01,n05,n11,n15,n21,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_13"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_12"],x[,"n_14"],x[,"n_01"],x[,"n_05"],x[,"n_11"],x[,"n_15"],x[,"n_21"],x[,"n_25"])
LOD_cr <- (x[,"n_11"] + x[,"n_15"])*log10(2) + 2*(x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*log10(2) - x[,"n_13"]*log10(11) - (x[,"n_03"] + x[,"n_23"])*log10(11) + (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + x[,"n_13"]*log10(9 + 8*r_cr - 8*r_cr^2) + (x[,"n_11"] + x[,"n_15"])*log10(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_03"] + x[,"n_23"])*log10(13 - 8*r_cr + 8*r_cr^2)
logL_cr <- (-x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - x[,"n_13"] - x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"])*log(2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_12"] + x[,"n_14"] + x[,"n_22"] + x[,"n_24"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + x[,"n_13"]*log(9 + 8*r_cr - 8*r_cr^2) + (x[,"n_11"] + x[,"n_15"])*log(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_03"] + x[,"n_23"])*log(13 - 8*r_cr + 8*r_cr^2)
logL_rc <- function(r,n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n12,n14) {
L <- (-4*n01 - n02 - 3*n03 - n04 - 4*n05 - 3*n11 - 2*n13 - 3*n15 - 4*n21 - n22 - 3*n23 - n24 - 4*n25)*log(2) + (n02 + n04 + n12 + n14 + n22 + n24)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25)*log(5) + (n01 + n05 + n21 + n25)*(log(3 - 2*r) + log(1 + 2*r)) + n13*log(21 + 4*r - 4*r^2) + (n11 + n15)*log(5 - 4*r + 4*r^2) + (n03 + n23)*log(23 - 4*r + 4*r^2)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n12,n14) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n12,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_13"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_12"],x[,"n_14"])
LOD_rc <- -2*(x[,"n_11"] + x[,"n_15"])*log10(2) - 2*(x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*log10(2) - x[,"n_13"]*(log10(2) + log10(11)) - (x[,"n_03"] + x[,"n_23"])*(log10(2) + log10(11)) + (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*(log10(3 - 2*r_rc) + log10(1 + 2*r_rc)) + x[,"n_13"]*log10(21 + 4*r_rc - 4*r_rc^2) + (x[,"n_11"] + x[,"n_15"])*log10(5 - 4*r_rc + 4*r_rc^2) + (x[,"n_03"] + x[,"n_23"])*log10(23 - 4*r_rc + 4*r_rc^2)
logL_rc <- (-4*x[,"n_01"] - x[,"n_02"] - 3*x[,"n_03"] - x[,"n_04"] - 4*x[,"n_05"] - 3*x[,"n_11"] - 2*x[,"n_13"] - 3*x[,"n_15"] - 4*x[,"n_21"] - x[,"n_22"] - 3*x[,"n_23"] - x[,"n_24"] - 4*x[,"n_25"])*log(2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_12"] + x[,"n_14"] + x[,"n_22"] + x[,"n_24"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*(log(3 - 2*r_rc) + log(1 + 2*r_rc)) + x[,"n_13"]*log(21 + 4*r_rc - 4*r_rc^2) + (x[,"n_11"] + x[,"n_15"])*log(5 - 4*r_rc + 4*r_rc^2) + (x[,"n_03"] + x[,"n_23"])*log(23 - 4*r_rc + 4*r_rc^2)
logL_rr <- function(r,n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n12,n14) {
L <- (-2*n01 - 3*n02 - n03 - 3*n04 - 2*n05 - 2*n11 + n13 - 2*n15 - 2*n21 - 3*n22 - n23 - 3*n24 - 2*n25)*log(2) + (n02 + n04 + n12 + n14 + n22 + n24)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25)*log(5) + (n04 + n22)*log(7 - 6*r) + (n05 + n21)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n01 + n25)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n02 + n24)*log(1 + 6*r) + (n11 + n15)*log(1 + 4*r - 4*r^2) + (n03 + n23)*log(5 + 2*r - 2*r^2) + n13*log(3 - r + r^2)
return(L)}
interlogL_rr <- function(n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n12,n14) {
optimize(logL_rr,c(0,0.5), n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n12,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_13"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_12"],x[,"n_14"])
LOD_rr <- -((x[,"n_11"] + x[,"n_15"])*log10(2)) - 2*(x[,"n_04"] + x[,"n_22"])*log10(2) - 2*(x[,"n_02"] + x[,"n_24"])*log10(2) - x[,"n_13"]*(-2*log10(2) + log10(11)) - (x[,"n_03"] + x[,"n_23"])*(-log10(2) + log10(11)) + (x[,"n_04"] + x[,"n_22"])*log10(7 - 6*r_rr) + (x[,"n_05"] + x[,"n_21"])*(log10(3 - 2*r_rr) + log10(pmax(1e-6,1 - r_rr))) + (x[,"n_01"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_02"] + x[,"n_24"])*log10(1 + 6*r_rr) + (x[,"n_11"] + x[,"n_15"])*log10(1 + 4*r_rr - 4*r_rr^2) + (x[,"n_03"] + x[,"n_23"])*log10(5 + 2*r_rr - 2*r_rr^2) + x[,"n_13"]*log10(3 - r_rr + r_rr^2)
logL_rr <- (-2*x[,"n_01"] - 3*x[,"n_02"] - x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] + x[,"n_13"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 3*x[,"n_22"] - x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"])*log(2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_12"] + x[,"n_14"] + x[,"n_22"] + x[,"n_24"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_04"] + x[,"n_22"])*log(7 - 6*r_rr) + (x[,"n_05"] + x[,"n_21"])*(log(3 - 2*r_rr) + log(pmax(1e-6,1 - r_rr))) + (x[,"n_01"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_02"] + x[,"n_24"])*log(1 + 6*r_rr) + (x[,"n_11"] + x[,"n_15"])*log(1 + 4*r_rr - 4*r_rr^2) + (x[,"n_03"] + x[,"n_23"])*log(5 + 2*r_rr - 2*r_rr^2) + x[,"n_13"]*log(3 - r_rr + r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_3.0 <- function(x){
logL_c <- function(r,n01,n02,n11,n12,n21,n22,n00,n03,n10,n13,n20,n23) {
L <- (n01 + n02 + 2*n11 + 2*n12 + n21 + n22)*log(3) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*(3*log(2) + log(5)) + (n00 + n23)*log(pmax(1e-6,1 - r)) + (n01 + n22)*log(2 - r) + (n03 + n20)*log(pmax(1e-6,r)) + (n02 + n21)*log(1 + r)
return(L)}
interlogL_c <- function(n01,n02,n11,n12,n21,n22,n00,n03,n10,n13,n20,n23) {
optimize(logL_c,c(0,0.5), n01,n02,n11,n12,n21,n22,n00,n03,n10,n13,n20,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_11"],x[,"n_12"],x[,"n_21"],x[,"n_22"],x[,"n_00"],x[,"n_03"],x[,"n_10"],x[,"n_13"],x[,"n_20"],x[,"n_23"])
LOD_c <- (x[,"n_03"] + x[,"n_20"])*log10(2) + (x[,"n_00"] + x[,"n_23"])*log10(2) - (x[,"n_02"] + x[,"n_21"])*(-log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_22"])*(-log10(2) + log10(3)) + (x[,"n_00"] + x[,"n_23"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_22"])*log10(2 - r_c) + (x[,"n_03"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_21"])*log10(1 + r_c)
logL_c <- (x[,"n_01"] + x[,"n_02"] + 2*x[,"n_11"] + 2*x[,"n_12"] + x[,"n_21"] + x[,"n_22"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*(3*log(2) + log(5)) + (x[,"n_00"] + x[,"n_23"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_22"])*log(2 - r_c) + (x[,"n_03"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_21"])*log(1 + r_c)
logL_r <- function(r,n01,n02,n11,n12,n21,n22,n00,n03,n10,n13,n20,n23) {
L <- (n01 + n02 + 2*n11 + 2*n12 + n21 + n22)*log(3) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*(3*log(2) + log(5)) + (n03 + n20)*log(pmax(1e-6,1 - r)) + (n02 + n21)*log(2 - r) + (n00 + n23)*log(pmax(1e-6,r)) + (n01 + n22)*log(1 + r)
return(L)}
interlogL_r <- function(n01,n02,n11,n12,n21,n22,n00,n03,n10,n13,n20,n23) {
optimize(logL_r,c(0,0.5), n01,n02,n11,n12,n21,n22,n00,n03,n10,n13,n20,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_11"],x[,"n_12"],x[,"n_21"],x[,"n_22"],x[,"n_00"],x[,"n_03"],x[,"n_10"],x[,"n_13"],x[,"n_20"],x[,"n_23"])
LOD_r <- (x[,"n_03"] + x[,"n_20"])*log10(2) + (x[,"n_00"] + x[,"n_23"])*log10(2) - (x[,"n_02"] + x[,"n_21"])*(-log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_22"])*(-log10(2) + log10(3)) + (x[,"n_03"] + x[,"n_20"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_02"] + x[,"n_21"])*log10(2 - r_r) + (x[,"n_00"] + x[,"n_23"])*log10(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_22"])*log10(1 + r_r)
logL_r <- (x[,"n_01"] + x[,"n_02"] + 2*x[,"n_11"] + 2*x[,"n_12"] + x[,"n_21"] + x[,"n_22"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*(3*log(2) + log(5)) + (x[,"n_03"] + x[,"n_20"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_02"] + x[,"n_21"])*log(2 - r_r) + (x[,"n_00"] + x[,"n_23"])*log(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_22"])*log(1 + r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.1_3.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
L <- (-4*n00 - 3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26)*log(2) + (n01 + n02 + n04 + n05 + n11 + n12 + n14 + n15 + n21 + n22 + n24 + n25)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26)*log(5) + 2*(n00 + n26)*log(pmax(1e-6,1 - r)) + (n01 + n25)*(log(pmax(1e-6,1 - r)) + log(2 - r)) + 2*(n06 + n20)*log(pmax(1e-6,r)) + (n10 + n16)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n21)*(log(pmax(1e-6,r)) + log(1 + r)) + (n03 + n23)*log(9 + 5*r - 5*r^2) + (n11 + n15)*log(1 + 2*r - 2*r^2) + (n12 + n14)*log(8 + r - r^2) + (n02 + n24)*log(14 - 12*r + r^2) + (n04 + n22)*log(3 + 10*r + r^2) + n13*log(23 - 10*r + 10*r^2)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"])
LOD_cc <- 2*(x[,"n_10"] + x[,"n_16"])*log10(2) + 2*(x[,"n_06"] + x[,"n_20"])*log10(2) + 2*(x[,"n_00"] + x[,"n_26"])*log10(2) - (x[,"n_05"] + x[,"n_21"])*(-2*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_25"])*(-2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_15"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_14"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_22"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_02"] + x[,"n_24"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_03"] + x[,"n_23"])*(-2*log10(2) + log10(41)) - x[,"n_13"]*(-log10(2) + log10(41)) + 2*(x[,"n_00"] + x[,"n_26"])*log10(pmax(1e-6,1 - r_cc)) + (x[,"n_01"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_cc)) + log10(2 - r_cc)) + 2*(x[,"n_06"] + x[,"n_20"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_16"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(1 + r_cc)) + (x[,"n_03"] + x[,"n_23"])*log10(9 + 5*r_cc - 5*r_cc^2) + (x[,"n_11"] + x[,"n_15"])*log10(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_12"] + x[,"n_14"])*log10(8 + r_cc - r_cc^2) + (x[,"n_02"] + x[,"n_24"])*log10(14 - 12*r_cc + r_cc^2) + (x[,"n_04"] + x[,"n_22"])*log10(3 + 10*r_cc + r_cc^2) + x[,"n_13"]*log10(23 - 10*r_cc + 10*r_cc^2)
logL_cc <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_11"] + x[,"n_12"] + x[,"n_14"] + x[,"n_15"] + x[,"n_21"] + x[,"n_22"] + x[,"n_24"] + x[,"n_25"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"])*log(5) + 2*(x[,"n_00"] + x[,"n_26"])*log(pmax(1e-6,1 - r_cc)) + (x[,"n_01"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_cc)) + log(2 - r_cc)) + 2*(x[,"n_06"] + x[,"n_20"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_16"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(1 + r_cc)) + (x[,"n_03"] + x[,"n_23"])*log(9 + 5*r_cc - 5*r_cc^2) + (x[,"n_11"] + x[,"n_15"])*log(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_12"] + x[,"n_14"])*log(8 + r_cc - r_cc^2) + (x[,"n_02"] + x[,"n_24"])*log(14 - 12*r_cc + r_cc^2) + (x[,"n_04"] + x[,"n_22"])*log(3 + 10*r_cc + r_cc^2) + x[,"n_13"]*log(23 - 10*r_cc + 10*r_cc^2)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
L <- (-4*n00 - 4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 2*n11 - 4*n12 - n13 - 4*n14 - 2*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26)*log(2) + (n01 + n02 + n04 + n05 + n11 + n12 + n14 + n15 + n21 + n22 + n24 + n25)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26)*log(5) + (n00 + n06 + n20 + n26)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n13*log(9 + 5*r - 5*r^2) + (n01 + n05 + n21 + n25)*log(1 + 2*r - 2*r^2) + (n02 + n04 + n22 + n24)*log(8 + r - r^2) + (n11 + n15)*log(1 - r + r^2) + (n10 + n16)*log(1 - 2*r + 2*r^2) + (n12 + n14)*log(17 - 2*r + 2*r^2) + (n03 + n23)*log(23 - 10*r + 10*r^2)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"])
LOD_cr <- (x[,"n_10"] + x[,"n_16"])*log10(2) + 2*(x[,"n_00"] + x[,"n_06"] + x[,"n_20"] + x[,"n_26"])*log10(2) + (x[,"n_11"] + x[,"n_15"])*(2*log10(2) - log10(3)) - (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*(-log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_04"] + x[,"n_22"] + x[,"n_24"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_12"] + x[,"n_14"])*(-log10(2) + log10(3) + log10(11)) - x[,"n_13"]*(-2*log10(2) + log10(41)) - (x[,"n_03"] + x[,"n_23"])*(-log10(2) + log10(41)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_20"] + x[,"n_26"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + x[,"n_13"]*log10(9 + 5*r_cr - 5*r_cr^2) + (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*log10(1 + 2*r_cr - 2*r_cr^2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_22"] + x[,"n_24"])*log10(8 + r_cr - r_cr^2) + (x[,"n_11"] + x[,"n_15"])*log10(1 - r_cr + r_cr^2) + (x[,"n_10"] + x[,"n_16"])*log10(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_12"] + x[,"n_14"])*log10(17 - 2*r_cr + 2*r_cr^2) + (x[,"n_03"] + x[,"n_23"])*log10(23 - 10*r_cr + 10*r_cr^2)
logL_cr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 2*x[,"n_11"] - 4*x[,"n_12"] - x[,"n_13"] - 4*x[,"n_14"] - 2*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_11"] + x[,"n_12"] + x[,"n_14"] + x[,"n_15"] + x[,"n_21"] + x[,"n_22"] + x[,"n_24"] + x[,"n_25"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"])*log(5) + (x[,"n_00"] + x[,"n_06"] + x[,"n_20"] + x[,"n_26"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + x[,"n_13"]*log(9 + 5*r_cr - 5*r_cr^2) + (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*log(1 + 2*r_cr - 2*r_cr^2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_22"] + x[,"n_24"])*log(8 + r_cr - r_cr^2) + (x[,"n_11"] + x[,"n_15"])*log(1 - r_cr + r_cr^2) + (x[,"n_10"] + x[,"n_16"])*log(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_12"] + x[,"n_14"])*log(17 - 2*r_cr + 2*r_cr^2) + (x[,"n_03"] + x[,"n_23"])*log(23 - 10*r_cr + 10*r_cr^2)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
L <- (-4*n00 - 3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26)*log(2) + (n01 + n02 + n04 + n05 + n11 + n12 + n14 + n15 + n21 + n22 + n24 + n25)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26)*log(5) + 2*(n06 + n20)*log(pmax(1e-6,1 - r)) + (n05 + n21)*(log(pmax(1e-6,1 - r)) + log(2 - r)) + 2*(n00 + n26)*log(pmax(1e-6,r)) + (n10 + n16)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n25)*(log(pmax(1e-6,r)) + log(1 + r)) + (n03 + n23)*log(9 + 5*r - 5*r^2) + (n11 + n15)*log(1 + 2*r - 2*r^2) + (n12 + n14)*log(8 + r - r^2) + (n04 + n22)*log(14 - 12*r + r^2) + (n02 + n24)*log(3 + 10*r + r^2) + n13*log(23 - 10*r + 10*r^2)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"])
LOD_rr <- 2*(x[,"n_10"] + x[,"n_16"])*log10(2) + 2*(x[,"n_06"] + x[,"n_20"])*log10(2) + 2*(x[,"n_00"] + x[,"n_26"])*log10(2) - (x[,"n_05"] + x[,"n_21"])*(-2*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_25"])*(-2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_15"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_14"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_22"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_02"] + x[,"n_24"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_03"] + x[,"n_23"])*(-2*log10(2) + log10(41)) - x[,"n_13"]*(-log10(2) + log10(41)) + 2*(x[,"n_06"] + x[,"n_20"])*log10(pmax(1e-6,1 - r_rr)) + (x[,"n_05"] + x[,"n_21"])*(log10(pmax(1e-6,1 - r_rr)) + log10(2 - r_rr)) + 2*(x[,"n_00"] + x[,"n_26"])*log10(pmax(1e-6,r_rr)) + (x[,"n_10"] + x[,"n_16"])*(log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(1 + r_rr)) + (x[,"n_03"] + x[,"n_23"])*log10(9 + 5*r_rr - 5*r_rr^2) + (x[,"n_11"] + x[,"n_15"])*log10(1 + 2*r_rr - 2*r_rr^2) + (x[,"n_12"] + x[,"n_14"])*log10(8 + r_rr - r_rr^2) + (x[,"n_04"] + x[,"n_22"])*log10(14 - 12*r_rr + r_rr^2) + (x[,"n_02"] + x[,"n_24"])*log10(3 + 10*r_rr + r_rr^2) + x[,"n_13"]*log10(23 - 10*r_rr + 10*r_rr^2)
logL_rr <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_11"] + x[,"n_12"] + x[,"n_14"] + x[,"n_15"] + x[,"n_21"] + x[,"n_22"] + x[,"n_24"] + x[,"n_25"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"])*log(5) + 2*(x[,"n_06"] + x[,"n_20"])*log(pmax(1e-6,1 - r_rr)) + (x[,"n_05"] + x[,"n_21"])*(log(pmax(1e-6,1 - r_rr)) + log(2 - r_rr)) + 2*(x[,"n_00"] + x[,"n_26"])*log(pmax(1e-6,r_rr)) + (x[,"n_10"] + x[,"n_16"])*(log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(1 + r_rr)) + (x[,"n_03"] + x[,"n_23"])*log(9 + 5*r_rr - 5*r_rr^2) + (x[,"n_11"] + x[,"n_15"])*log(1 + 2*r_rr - 2*r_rr^2) + (x[,"n_12"] + x[,"n_14"])*log(8 + r_rr - r_rr^2) + (x[,"n_04"] + x[,"n_22"])*log(14 - 12*r_rr + r_rr^2) + (x[,"n_02"] + x[,"n_24"])*log(3 + 10*r_rr + r_rr^2) + x[,"n_13"]*log(23 - 10*r_rr + 10*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_1.2 <- function(x){
logL_cc <- function(r,n01,n02,n10,n13,n20,n23,n31,n32,n00,n03,n11,n12,n21,n22,n30,n33) {
L <- (n01 + n02 + n10 + n13 + n20 + n23 + n31 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(log(2) + log(5)) + 3*(n00 + n33)*log(pmax(1e-6,1 - r)) + 3*(n03 + n30)*log(pmax(1e-6,r)) + (n01 + n10 + n23 + n32)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n13 + n20 + n31)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n12 + n21)*(log(pmax(1e-6,r)) + log(7 - 12*r + 9*r^2)) + (n11 + n22)*log(4 - 10*r + 15*r^2 - 9*r^3)
return(L)}
interlogL_cc <- function(n01,n02,n10,n13,n20,n23,n31,n32,n00,n03,n11,n12,n21,n22,n30,n33) {
optimize(logL_cc,c(0,0.5), n01,n02,n10,n13,n20,n23,n31,n32,n00,n03,n11,n12,n21,n22,n30,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_13"],x[,"n_20"],x[,"n_23"],x[,"n_31"],x[,"n_32"],x[,"n_00"],x[,"n_03"],x[,"n_11"],x[,"n_12"],x[,"n_21"],x[,"n_22"],x[,"n_30"],x[,"n_33"])
LOD_cc <- 3*(x[,"n_03"] + x[,"n_30"])*log10(2) + 3*(x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log10(2) + 3*(x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(2) + 3*(x[,"n_00"] + x[,"n_33"])*log10(2) - (x[,"n_12"] + x[,"n_21"])*(-3*log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_22"])*(-3*log10(2) + log10(13)) + 3*(x[,"n_00"] + x[,"n_33"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_03"] + x[,"n_30"])*log10(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_12"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(7 - 12*r_cc + 9*r_cc^2)) + (x[,"n_11"] + x[,"n_22"])*log10(4 - 10*r_cc + 15*r_cc^2 - 9*r_cc^3)
logL_cc <- (x[,"n_01"] + x[,"n_02"] + x[,"n_10"] + x[,"n_13"] + x[,"n_20"] + x[,"n_23"] + x[,"n_31"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(log(2) + log(5)) + 3*(x[,"n_00"] + x[,"n_33"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_03"] + x[,"n_30"])*log(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_12"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(7 - 12*r_cc + 9*r_cc^2)) + (x[,"n_11"] + x[,"n_22"])*log(4 - 10*r_cc + 15*r_cc^2 - 9*r_cc^3)
logL_cm <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(3*log(2) + log(5)) + (n03 + n30)*(log(2 - r) + 2*log(pmax(1e-6,r))) + (n00 + n33)*(log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n02 + n13 + n20 + n31)*(log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n12 + n21)*log(6 - 4*r + 14*r^2 - 9*r^3) + (n01 + n10 + n23 + n32)*log(3 - 4*r + 4*r^2 - 3*r^3) + (n11 + n22)*log(7 + 3*r - 13*r^2 + 9*r^3)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_cm <- -((x[,"n_03"] + x[,"n_30"])*(-3*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_33"])*(-3*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(-3*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(13)) - (x[,"n_12"] + x[,"n_21"])*(-3*log10(2) + log10(3) + log10(17)) - (x[,"n_11"] + x[,"n_22"])*(-3*log10(2) + log10(3) + log10(17)) + (x[,"n_03"] + x[,"n_30"])*(log10(2 - r_cm) + 2*log10(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_33"])*(log10(pmax(1e-6,1 - r_cm)) + log10(1 - r_cm^2)) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_12"] + x[,"n_21"])*log10(6 - 4*r_cm + 14*r_cm^2 - 9*r_cm^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(3 - 4*r_cm + 4*r_cm^2 - 3*r_cm^3) + (x[,"n_11"] + x[,"n_22"])*log10(7 + 3*r_cm - 13*r_cm^2 + 9*r_cm^3)
logL_cm <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(3*log(2) + log(5)) + (x[,"n_03"] + x[,"n_30"])*(log(2 - r_cm) + 2*log(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_33"])*(log(pmax(1e-6,1 - r_cm)) + log(1 - r_cm^2)) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_12"] + x[,"n_21"])*log(6 - 4*r_cm + 14*r_cm^2 - 9*r_cm^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log(3 - 4*r_cm + 4*r_cm^2 - 3*r_cm^3) + (x[,"n_11"] + x[,"n_22"])*log(7 + 3*r_cm - 13*r_cm^2 + 9*r_cm^3)
logL_cr <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(2*log(2) + log(3) + log(5)) + (n00 + n33)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n30)*(log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n11 + n22)*log(12 - 8*r + 11*r^2 - 9*r^3) + (n02 + n13 + n20 + n31)*log(3 - 4*r + 7*r^2 - 3*r^3) + (n01 + n10 + n23 + n32)*log(3 - r - 2*r^2 + 3*r^3) + (n12 + n21)*log(6 + 13*r - 16*r^2 + 9*r^3)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_cr <- -((x[,"n_03"] + x[,"n_30"])*(-3*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_33"])*(-3*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_21"])*(-3*log10(2) + log10(7) + log10(11)) - (x[,"n_11"] + x[,"n_22"])*(-3*log10(2) + log10(7) + log10(11)) - (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(-3*log10(2) + log10(19)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(19)) + (x[,"n_00"] + x[,"n_33"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr)) + log10(2 + r_cr)) + (x[,"n_03"] + x[,"n_30"])*(log10(pmax(1e-6,r_cr)) + log10(3 - 4*r_cr + r_cr^2)) + (x[,"n_11"] + x[,"n_22"])*log10(12 - 8*r_cr + 11*r_cr^2 - 9*r_cr^3) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log10(3 - 4*r_cr + 7*r_cr^2 - 3*r_cr^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(3 - r_cr - 2*r_cr^2 + 3*r_cr^3) + (x[,"n_12"] + x[,"n_21"])*log10(6 + 13*r_cr - 16*r_cr^2 + 9*r_cr^3)
logL_cr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(2*log(2) + log(3) + log(5)) + (x[,"n_00"] + x[,"n_33"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr)) + log(2 + r_cr)) + (x[,"n_03"] + x[,"n_30"])*(log(pmax(1e-6,r_cr)) + log(3 - 4*r_cr + r_cr^2)) + (x[,"n_11"] + x[,"n_22"])*log(12 - 8*r_cr + 11*r_cr^2 - 9*r_cr^3) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log(3 - 4*r_cr + 7*r_cr^2 - 3*r_cr^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log(3 - r_cr - 2*r_cr^2 + 3*r_cr^3) + (x[,"n_12"] + x[,"n_21"])*log(6 + 13*r_cr - 16*r_cr^2 + 9*r_cr^3)
logL_rc <- function(r,n01,n02,n10,n13,n20,n23,n31,n32,n00,n03,n11,n12,n21,n22,n30,n33) {
L <- (n01 + n02 + n10 + n13 + n20 + n23 + n31 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(log(2) + 2*log(5)) + (n03 + n30)*(log(3 - r) + 2*log(pmax(1e-6,r))) + (n00 + n33)*(2*log(pmax(1e-6,1 - r)) + log(2 + r)) + (n02 + n13 + n20 + n31)*(log(pmax(1e-6,r)) + log(2 - 2*r + r^2)) + (n12 + n21)*log(9 - 7*r + 15*r^2 - 9*r^3) + (n01 + n10 + n23 + n32)*log(1 - r + r^2 - r^3) + (n11 + n22)*log(8 + 4*r - 12*r^2 + 9*r^3)
return(L)}
interlogL_rc <- function(n01,n02,n10,n13,n20,n23,n31,n32,n00,n03,n11,n12,n21,n22,n30,n33) {
optimize(logL_rc,c(0,0.5), n01,n02,n10,n13,n20,n23,n31,n32,n00,n03,n11,n12,n21,n22,n30,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_13"],x[,"n_20"],x[,"n_23"],x[,"n_31"],x[,"n_32"],x[,"n_00"],x[,"n_03"],x[,"n_11"],x[,"n_12"],x[,"n_21"],x[,"n_22"],x[,"n_30"],x[,"n_33"])
LOD_rc <- (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(3*log10(2) - log10(5)) - (x[,"n_03"] + x[,"n_30"])*(-3*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(-3*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_33"])*(-3*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_21"])*(-3*log10(2) + log10(5) + log10(13)) - (x[,"n_11"] + x[,"n_22"])*(-3*log10(2) + log10(5) + log10(13)) + (x[,"n_03"] + x[,"n_30"])*(log10(3 - r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_33"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(2 + r_rc)) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 2*r_rc + r_rc^2)) + (x[,"n_12"] + x[,"n_21"])*log10(9 - 7*r_rc + 15*r_rc^2 - 9*r_rc^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(1 - r_rc + r_rc^2 - r_rc^3) + (x[,"n_11"] + x[,"n_22"])*log10(8 + 4*r_rc - 12*r_rc^2 + 9*r_rc^3)
logL_rc <- (x[,"n_01"] + x[,"n_02"] + x[,"n_10"] + x[,"n_13"] + x[,"n_20"] + x[,"n_23"] + x[,"n_31"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(log(2) + 2*log(5)) + (x[,"n_03"] + x[,"n_30"])*(log(3 - r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_33"])*(2*log(pmax(1e-6,1 - r_rc)) + log(2 + r_rc)) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(2 - 2*r_rc + r_rc^2)) + (x[,"n_12"] + x[,"n_21"])*log(9 - 7*r_rc + 15*r_rc^2 - 9*r_rc^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log(1 - r_rc + r_rc^2 - r_rc^3) + (x[,"n_11"] + x[,"n_22"])*log(8 + 4*r_rc - 12*r_rc^2 + 9*r_rc^3)
logL_rm <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(3*log(2) + 2*log(5)) + (n03 + n30)*(log(2 - r) + log(3 - r) + log(pmax(1e-6,r))) + (n00 + n33)*(log(2 + r) + log(1 - r^2)) + (n11 + n22)*log(32 - 3*r + 10*r^2 - 9*r^3) + (n02 + n13 + n20 + n31)*log(9 - 5*r + 8*r^2 - 3*r^3) + (n01 + n10 + n23 + n32)*log(9 - 2*r - r^2 + 3*r^3) + (n12 + n21)*log(30 + 10*r - 17*r^2 + 9*r^3)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_rm <- -((x[,"n_03"] + x[,"n_30"])*(-3*log10(2) + log10(3) + log10(5))) - (x[,"n_00"] + x[,"n_33"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(-3*log10(2) + log10(5) + log10(13)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(5) + log10(13)) - (x[,"n_12"] + x[,"n_21"])*(-3*log10(2) + log10(3) + log10(5) + log10(17)) - (x[,"n_11"] + x[,"n_22"])*(-3*log10(2) + log10(3) + log10(5) + log10(17)) + (x[,"n_03"] + x[,"n_30"])*(log10(2 - r_rm) + log10(3 - r_rm) + log10(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_33"])*(log10(2 + r_rm) + log10(1 - r_rm^2)) + (x[,"n_11"] + x[,"n_22"])*log10(32 - 3*r_rm + 10*r_rm^2 - 9*r_rm^3) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log10(9 - 5*r_rm + 8*r_rm^2 - 3*r_rm^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(9 - 2*r_rm - r_rm^2 + 3*r_rm^3) + (x[,"n_12"] + x[,"n_21"])*log10(30 + 10*r_rm - 17*r_rm^2 + 9*r_rm^3)
logL_rm <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(3*log(2) + 2*log(5)) + (x[,"n_03"] + x[,"n_30"])*(log(2 - r_rm) + log(3 - r_rm) + log(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_33"])*(log(2 + r_rm) + log(1 - r_rm^2)) + (x[,"n_11"] + x[,"n_22"])*log(32 - 3*r_rm + 10*r_rm^2 - 9*r_rm^3) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log(9 - 5*r_rm + 8*r_rm^2 - 3*r_rm^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log(9 - 2*r_rm - r_rm^2 + 3*r_rm^3) + (x[,"n_12"] + x[,"n_21"])*log(30 + 10*r_rm - 17*r_rm^2 + 9*r_rm^3)
logL_rr <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(2*log(2) + log(3) + 2*log(5)) + (n00 + n33)*(log(pmax(1e-6,r)) + 2*log(2 + r)) + (n03 + n30)*(log(3 - r) + log(3 - 4*r + r^2)) + (n12 + n21)*log(57 - 25*r + 19*r^2 - 9*r^3) + (n01 + n10 + n23 + n32)*log(6 + 13*r - r^2 - 3*r^3) + (n02 + n13 + n20 + n31)*log(15 - 2*r - 10*r^2 + 3*r^3) + (n11 + n22)*log(42 + 14*r - 8*r^2 + 9*r^3)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_rr <- -((x[,"n_03"] + x[,"n_30"])*(-3*log10(2) + 2*log10(5))) - (x[,"n_00"] + x[,"n_33"])*(-log10(2) + 2*(-log10(2) + log10(5))) - (x[,"n_12"] + x[,"n_21"])*(-3*log10(2) + log10(5) + log10(7) + log10(11)) - (x[,"n_11"] + x[,"n_22"])*(-3*log10(2) + log10(5) + log10(7) + log10(11)) - (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(-3*log10(2) + log10(5) + log10(19)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(5) + log10(19)) + (x[,"n_00"] + x[,"n_33"])*(log10(pmax(1e-6,r_rr)) + 2*log10(2 + r_rr)) + (x[,"n_03"] + x[,"n_30"])*(log10(3 - r_rr) + log10(3 - 4*r_rr + r_rr^2)) + (x[,"n_12"] + x[,"n_21"])*log10(57 - 25*r_rr + 19*r_rr^2 - 9*r_rr^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(6 + 13*r_rr - r_rr^2 - 3*r_rr^3) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log10(15 - 2*r_rr - 10*r_rr^2 + 3*r_rr^3) + (x[,"n_11"] + x[,"n_22"])*log10(42 + 14*r_rr - 8*r_rr^2 + 9*r_rr^3)
logL_rr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(2*log(2) + log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_33"])*(log(pmax(1e-6,r_rr)) + 2*log(2 + r_rr)) + (x[,"n_03"] + x[,"n_30"])*(log(3 - r_rr) + log(3 - 4*r_rr + r_rr^2)) + (x[,"n_12"] + x[,"n_21"])*log(57 - 25*r_rr + 19*r_rr^2 - 9*r_rr^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log(6 + 13*r_rr - r_rr^2 - 3*r_rr^3) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log(15 - 2*r_rr - 10*r_rr^2 + 3*r_rr^3) + (x[,"n_11"] + x[,"n_22"])*log(42 + 14*r_rr - 8*r_rr^2 + 9*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_1.3 <- function(x){
logL_cc <- function(r,n10,n14,n20,n24,n00,n01,n02,n03,n04,n11,n12,n13,n21,n22,n23,n30,n31,n32,n33,n34) {
L <- (n10 + n14 + n20 + n24)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(3*log(2) + log(5)) + (n02 + n32)*log(7) + 3*(n00 + n34)*log(pmax(1e-6,1 - r)) + 3*(n04 + n30)*log(pmax(1e-6,r)) + (n02 + n32)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n24)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n31)*(log(5 - 2*r) + 2*log(pmax(1e-6,r))) + (n14 + n20)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n33)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r)) + (n13 + n21)*(log(pmax(1e-6,r)) + log(12 - 11*r + 6*r^2)) + (n12 + n22)*log(9 - 7*r + 7*r^2) + (n11 + n23)*log(7 - 8*r + 7*r^2 - 6*r^3)
return(L)}
interlogL_cc <- function(n10,n14,n20,n24,n00,n01,n02,n03,n04,n11,n12,n13,n21,n22,n23,n30,n31,n32,n33,n34) {
optimize(logL_cc,c(0,0.5), n10,n14,n20,n24,n00,n01,n02,n03,n04,n11,n12,n13,n21,n22,n23,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_10"],x[,"n_14"],x[,"n_20"],x[,"n_24"],x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cc <- 3*(x[,"n_14"] + x[,"n_20"])*log10(2) - 2*(x[,"n_13"] + x[,"n_21"])*log10(2) - 2*(x[,"n_11"] + x[,"n_23"])*log10(2) + 3*(x[,"n_10"] + x[,"n_24"])*log10(2) + 3*(x[,"n_04"] + x[,"n_30"])*log10(2) + 2*(x[,"n_02"] + x[,"n_32"])*log10(2) + 3*(x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_12"] + x[,"n_22"])*(-2*log10(2) + log10(29)) + 3*(x[,"n_00"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_04"] + x[,"n_30"])*log10(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_32"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_24"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_31"])*(log10(5 - 2*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_20"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_33"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 2*r_cc)) + (x[,"n_13"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(12 - 11*r_cc + 6*r_cc^2)) + (x[,"n_12"] + x[,"n_22"])*log10(9 - 7*r_cc + 7*r_cc^2) + (x[,"n_11"] + x[,"n_23"])*log10(7 - 8*r_cc + 7*r_cc^2 - 6*r_cc^3)
logL_cc <- (x[,"n_10"] + x[,"n_14"] + x[,"n_20"] + x[,"n_24"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + (x[,"n_02"] + x[,"n_32"])*log(7) + 3*(x[,"n_00"] + x[,"n_34"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_04"] + x[,"n_30"])*log(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_32"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_24"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_31"])*(log(5 - 2*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_20"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_33"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 2*r_cc)) + (x[,"n_13"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(12 - 11*r_cc + 6*r_cc^2)) + (x[,"n_12"] + x[,"n_22"])*log(9 - 7*r_cc + 7*r_cc^2) + (x[,"n_11"] + x[,"n_23"])*log(7 - 8*r_cc + 7*r_cc^2 - 6*r_cc^3)
logL_cm <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n12 - n14 - n20 - n22 - n24 - n30 - n31 - n32 - n33 - n34)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*log(5) + (n00 + n34)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n12 + n22)*log(21 + 2*r - 2*r^2) + (n02 + n32)*log(3 - r + r^2) + (n03 + n31)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n14 + n20)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2)) + (n10 + n24)*log(3 - 7*r + 10*r^2 - 6*r^3) + (n13 + n21)*log(3 + 6*r^2 - 4*r^3) + (n01 + n33)*log(3 - 4*r + 3*r^2 - 2*r^3) + (n11 + n23)*log(5 - 6*r^2 + 4*r^3)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cm <- -2*(x[,"n_13"] + x[,"n_21"])*log10(2) - 2*(x[,"n_11"] + x[,"n_23"])*log10(2) + 3*(x[,"n_04"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_03"] + x[,"n_31"])*(-log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_33"])*(-log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_20"])*(-2*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_32"])*(-2*log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_22"])*(-log10(2) + log10(43)) + (x[,"n_00"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cm)) + 2*log10(pmax(1e-6,r_cm))) + (x[,"n_12"] + x[,"n_22"])*log10(21 + 2*r_cm - 2*r_cm^2) + (x[,"n_02"] + x[,"n_32"])*log10(3 - r_cm + r_cm^2) + (x[,"n_03"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(4 - 3*r_cm + 2*r_cm^2)) + (x[,"n_14"] + x[,"n_20"])*(log10(pmax(1e-6,r_cm)) + log10(5 - 8*r_cm + 6*r_cm^2)) + (x[,"n_10"] + x[,"n_24"])*log10(3 - 7*r_cm + 10*r_cm^2 - 6*r_cm^3) + (x[,"n_13"] + x[,"n_21"])*log10(3 + 6*r_cm^2 - 4*r_cm^3) + (x[,"n_01"] + x[,"n_33"])*log10(3 - 4*r_cm + 3*r_cm^2 - 2*r_cm^3) + (x[,"n_11"] + x[,"n_23"])*log10(5 - 6*r_cm^2 + 4*r_cm^3)
logL_cm <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_12"] - x[,"n_14"] - x[,"n_20"] - x[,"n_22"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_00"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cm)) + 2*log(pmax(1e-6,r_cm))) + (x[,"n_12"] + x[,"n_22"])*log(21 + 2*r_cm - 2*r_cm^2) + (x[,"n_02"] + x[,"n_32"])*log(3 - r_cm + r_cm^2) + (x[,"n_03"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(4 - 3*r_cm + 2*r_cm^2)) + (x[,"n_14"] + x[,"n_20"])*(log(pmax(1e-6,r_cm)) + log(5 - 8*r_cm + 6*r_cm^2)) + (x[,"n_10"] + x[,"n_24"])*log(3 - 7*r_cm + 10*r_cm^2 - 6*r_cm^3) + (x[,"n_13"] + x[,"n_21"])*log(3 + 6*r_cm^2 - 4*r_cm^3) + (x[,"n_01"] + x[,"n_33"])*log(3 - 4*r_cm + 3*r_cm^2 - 2*r_cm^3) + (x[,"n_11"] + x[,"n_23"])*log(5 - 6*r_cm^2 + 4*r_cm^3)
logL_cr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(3*log(2) + log(5)) + (n04 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n34)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n12 + n22)*log(6 + 5*r - 5*r^2) + (n01 + n33)*(log(pmax(1e-6,r)) + log(4 - 5*r + 2*r^2)) + (n10 + n24)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n02 + n32)*log(3 - 5*r + 5*r^2) + (n11 + n23)*log(6 - 8*r + 11*r^2 - 6*r^3) + (n14 + n20)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n03 + n31)*log(1 + r^2 - 2*r^3) + (n13 + n21)*log(3 + 4*r - 7*r^2 + 6*r^3)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cr <- -2*(x[,"n_13"] + x[,"n_21"])*log10(2) - 2*(x[,"n_11"] + x[,"n_23"])*log10(2) + 3*(x[,"n_04"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_34"])*log10(2) + (x[,"n_14"] + x[,"n_20"])*(3*log10(2) - log10(3)) - (x[,"n_10"] + x[,"n_24"])*(-3*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_32"])*(-2*log10(2) + log10(7)) - (x[,"n_12"] + x[,"n_22"])*(-2*log10(2) + log10(29)) + (x[,"n_04"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_12"] + x[,"n_22"])*log10(6 + 5*r_cr - 5*r_cr^2) + (x[,"n_01"] + x[,"n_33"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 5*r_cr + 2*r_cr^2)) + (x[,"n_10"] + x[,"n_24"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_02"] + x[,"n_32"])*log10(3 - 5*r_cr + 5*r_cr^2) + (x[,"n_11"] + x[,"n_23"])*log10(6 - 8*r_cr + 11*r_cr^2 - 6*r_cr^3) + (x[,"n_14"] + x[,"n_20"])*log10(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_03"] + x[,"n_31"])*log10(1 + r_cr^2 - 2*r_cr^3) + (x[,"n_13"] + x[,"n_21"])*log10(3 + 4*r_cr - 7*r_cr^2 + 6*r_cr^3)
logL_cr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + (x[,"n_04"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_12"] + x[,"n_22"])*log(6 + 5*r_cr - 5*r_cr^2) + (x[,"n_01"] + x[,"n_33"])*(log(pmax(1e-6,r_cr)) + log(4 - 5*r_cr + 2*r_cr^2)) + (x[,"n_10"] + x[,"n_24"])*(log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_02"] + x[,"n_32"])*log(3 - 5*r_cr + 5*r_cr^2) + (x[,"n_11"] + x[,"n_23"])*log(6 - 8*r_cr + 11*r_cr^2 - 6*r_cr^3) + (x[,"n_14"] + x[,"n_20"])*log(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_03"] + x[,"n_31"])*log(1 + r_cr^2 - 2*r_cr^3) + (x[,"n_13"] + x[,"n_21"])*log(3 + 4*r_cr - 7*r_cr^2 + 6*r_cr^3)
logL_rc <- function(r,n10,n14,n20,n24,n00,n01,n02,n03,n04,n11,n12,n13,n21,n22,n23,n30,n31,n32,n33,n34) {
L <- (n10 + n14 + n20 + n24)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(3*log(2) + 2*log(5)) + (n04 + n30)*(log(3 - r) + 2*log(pmax(1e-6,r))) + (n00 + n34)*(2*log(pmax(1e-6,1 - r)) + log(2 + r)) + (n12 + n22)*log(36 + r - r^2) + (n14 + n20)*(log(pmax(1e-6,r)) + log(2 - 2*r + r^2)) + (n02 + n32)*log(9 - r + r^2) + (n03 + n31)*(log(pmax(1e-6,r)) + log(12 - 5*r + 2*r^2)) + (n13 + n21)*log(18 + 11*r^2 - 6*r^3) + (n01 + n33)*log(9 - 8*r + r^2 - 2*r^3) + (n10 + n24)*log(1 - r + r^2 - r^3) + (n11 + n23)*log(23 - 4*r - 7*r^2 + 6*r^3)
return(L)}
interlogL_rc <- function(n10,n14,n20,n24,n00,n01,n02,n03,n04,n11,n12,n13,n21,n22,n23,n30,n31,n32,n33,n34) {
optimize(logL_rc,c(0,0.5), n10,n14,n20,n24,n00,n01,n02,n03,n04,n11,n12,n13,n21,n22,n23,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_10"],x[,"n_14"],x[,"n_20"],x[,"n_24"],x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rc <- (x[,"n_10"] + x[,"n_24"])*(3*log10(2) - log10(5)) - (x[,"n_03"] + x[,"n_31"])*log10(5) - (x[,"n_01"] + x[,"n_33"])*log10(5) - (x[,"n_14"] + x[,"n_20"])*(-3*log10(2) + log10(5)) - (x[,"n_04"] + x[,"n_30"])*(-3*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_34"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_21"])*(2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_23"])*(2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_22"])*(-2*log10(2) + log10(5) + log10(29)) + (x[,"n_04"] + x[,"n_30"])*(log10(3 - r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(2 + r_rc)) + (x[,"n_12"] + x[,"n_22"])*log10(36 + r_rc - r_rc^2) + (x[,"n_14"] + x[,"n_20"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 2*r_rc + r_rc^2)) + (x[,"n_02"] + x[,"n_32"])*log10(9 - r_rc + r_rc^2) + (x[,"n_03"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(12 - 5*r_rc + 2*r_rc^2)) + (x[,"n_13"] + x[,"n_21"])*log10(18 + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_01"] + x[,"n_33"])*log10(9 - 8*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_10"] + x[,"n_24"])*log10(1 - r_rc + r_rc^2 - r_rc^3) + (x[,"n_11"] + x[,"n_23"])*log10(23 - 4*r_rc - 7*r_rc^2 + 6*r_rc^3)
logL_rc <- (x[,"n_10"] + x[,"n_14"] + x[,"n_20"] + x[,"n_24"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + 2*log(5)) + (x[,"n_04"] + x[,"n_30"])*(log(3 - r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_rc)) + log(2 + r_rc)) + (x[,"n_12"] + x[,"n_22"])*log(36 + r_rc - r_rc^2) + (x[,"n_14"] + x[,"n_20"])*(log(pmax(1e-6,r_rc)) + log(2 - 2*r_rc + r_rc^2)) + (x[,"n_02"] + x[,"n_32"])*log(9 - r_rc + r_rc^2) + (x[,"n_03"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(12 - 5*r_rc + 2*r_rc^2)) + (x[,"n_13"] + x[,"n_21"])*log(18 + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_01"] + x[,"n_33"])*log(9 - 8*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_10"] + x[,"n_24"])*log(1 - r_rc + r_rc^2 - r_rc^3) + (x[,"n_11"] + x[,"n_23"])*log(23 - 4*r_rc - 7*r_rc^2 + 6*r_rc^3)
logL_rm <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n12 - n14 - n20 - n22 - n24 - n30 - n31 - n32 - n33 - n34)*log(3) + (-2*n00 - 2*n01 - n02 - 2*n03 - 2*n04 - 2*n10 - 2*n11 - n12 - 2*n13 - 2*n14 - 2*n20 - 2*n21 - n22 - 2*n23 - 2*n24 - 2*n30 - 2*n31 - n32 - 2*n33 - 2*n34)*log(5) + (n04 + n30)*(log(pmax(1e-6,1 - r)) + log(3 - r) + log(pmax(1e-6,r))) + (n00 + n34)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n12 + n22)*log(21 + 2*r - 2*r^2) + (n02 + n32)*log(3 - r + r^2) + (n14 + n20)*log(9 - 11*r + 14*r^2 - 6*r^3) + (n11 + n23)*log(19 + 6*r^2 - 4*r^3) + (n03 + n31)*log(9 - 4*r + 3*r^2 - 2*r^3) + (n01 + n33)*log(6 + 4*r - 3*r^2 + 2*r^3) + (n13 + n21)*log(21 - 6*r^2 + 4*r^3) + (n10 + n24)*log(6 + r - 4*r^2 + 6*r^3)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rm <- -((x[,"n_04"] + x[,"n_30"])*(-3*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_34"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_21"])*(2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_23"])*(2*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_31"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_01"] + x[,"n_33"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_14"] + x[,"n_20"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_10"] + x[,"n_24"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_02"] + x[,"n_32"])*(-2*log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_22"])*(-log10(2) + log10(43)) + (x[,"n_04"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_rm)) + log10(3 - r_rm) + log10(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm)) + log10(2 + r_rm)) + (x[,"n_12"] + x[,"n_22"])*log10(21 + 2*r_rm - 2*r_rm^2) + (x[,"n_02"] + x[,"n_32"])*log10(3 - r_rm + r_rm^2) + (x[,"n_14"] + x[,"n_20"])*log10(9 - 11*r_rm + 14*r_rm^2 - 6*r_rm^3) + (x[,"n_11"] + x[,"n_23"])*log10(19 + 6*r_rm^2 - 4*r_rm^3) + (x[,"n_03"] + x[,"n_31"])*log10(9 - 4*r_rm + 3*r_rm^2 - 2*r_rm^3) + (x[,"n_01"] + x[,"n_33"])*log10(6 + 4*r_rm - 3*r_rm^2 + 2*r_rm^3) + (x[,"n_13"] + x[,"n_21"])*log10(21 - 6*r_rm^2 + 4*r_rm^3) + (x[,"n_10"] + x[,"n_24"])*log10(6 + r_rm - 4*r_rm^2 + 6*r_rm^3)
logL_rm <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_12"] - x[,"n_14"] - x[,"n_20"] - x[,"n_22"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(3) + (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"])*log(5) + (x[,"n_04"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_rm)) + log(3 - r_rm) + log(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm)) + log(2 + r_rm)) + (x[,"n_12"] + x[,"n_22"])*log(21 + 2*r_rm - 2*r_rm^2) + (x[,"n_02"] + x[,"n_32"])*log(3 - r_rm + r_rm^2) + (x[,"n_14"] + x[,"n_20"])*log(9 - 11*r_rm + 14*r_rm^2 - 6*r_rm^3) + (x[,"n_11"] + x[,"n_23"])*log(19 + 6*r_rm^2 - 4*r_rm^3) + (x[,"n_03"] + x[,"n_31"])*log(9 - 4*r_rm + 3*r_rm^2 - 2*r_rm^3) + (x[,"n_01"] + x[,"n_33"])*log(6 + 4*r_rm - 3*r_rm^2 + 2*r_rm^3) + (x[,"n_13"] + x[,"n_21"])*log(21 - 6*r_rm^2 + 4*r_rm^3) + (x[,"n_10"] + x[,"n_24"])*log(6 + r_rm - 4*r_rm^2 + 6*r_rm^3)
logL_rr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(3*log(2) + 2*log(5)) + (n04 + n30)*(2*log(pmax(1e-6,1 - r)) + log(3 - r)) + (n00 + n34)*(2*log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n32)*log(6 + 11*r - 11*r^2) + (n10 + n24)*(log(pmax(1e-6,r)) + log(4 + r - 3*r^2)) + (n01 + n33)*(log(pmax(1e-6,r)) + log(8 + 5*r - 2*r^2)) + (n12 + n22)*log(39 - 11*r + 11*r^2) + (n13 + n21)*log(27 - 16*r + 7*r^2 - 6*r^3) + (n03 + n31)*log(11 - 12*r - r^2 + 2*r^3) + (n14 + n20)*log(2 + 3*r - 8*r^2 + 3*r^3) + (n11 + n23)*log(12 + 20*r - 11*r^2 + 6*r^3)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rr <- -((x[,"n_03"] + x[,"n_31"])*log10(5)) - (x[,"n_01"] + x[,"n_33"])*log10(5) - (x[,"n_04"] + x[,"n_30"])*(-3*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_34"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_21"])*(2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_23"])*(2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_20"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_10"] + x[,"n_24"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_22"])*(-2*log10(2) + log10(5) + log10(29)) + (x[,"n_04"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 - r_rr)) + (x[,"n_00"] + x[,"n_34"])*(2*log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_02"] + x[,"n_32"])*log10(6 + 11*r_rr - 11*r_rr^2) + (x[,"n_10"] + x[,"n_24"])*(log10(pmax(1e-6,r_rr)) + log10(4 + r_rr - 3*r_rr^2)) + (x[,"n_01"] + x[,"n_33"])*(log10(pmax(1e-6,r_rr)) + log10(8 + 5*r_rr - 2*r_rr^2)) + (x[,"n_12"] + x[,"n_22"])*log10(39 - 11*r_rr + 11*r_rr^2) + (x[,"n_13"] + x[,"n_21"])*log10(27 - 16*r_rr + 7*r_rr^2 - 6*r_rr^3) + (x[,"n_03"] + x[,"n_31"])*log10(11 - 12*r_rr - r_rr^2 + 2*r_rr^3) + (x[,"n_14"] + x[,"n_20"])*log10(2 + 3*r_rr - 8*r_rr^2 + 3*r_rr^3) + (x[,"n_11"] + x[,"n_23"])*log10(12 + 20*r_rr - 11*r_rr^2 + 6*r_rr^3)
logL_rr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + 2*log(5)) + (x[,"n_04"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 - r_rr)) + (x[,"n_00"] + x[,"n_34"])*(2*log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_02"] + x[,"n_32"])*log(6 + 11*r_rr - 11*r_rr^2) + (x[,"n_10"] + x[,"n_24"])*(log(pmax(1e-6,r_rr)) + log(4 + r_rr - 3*r_rr^2)) + (x[,"n_01"] + x[,"n_33"])*(log(pmax(1e-6,r_rr)) + log(8 + 5*r_rr - 2*r_rr^2)) + (x[,"n_12"] + x[,"n_22"])*log(39 - 11*r_rr + 11*r_rr^2) + (x[,"n_13"] + x[,"n_21"])*log(27 - 16*r_rr + 7*r_rr^2 - 6*r_rr^3) + (x[,"n_03"] + x[,"n_31"])*log(11 - 12*r_rr - r_rr^2 + 2*r_rr^3) + (x[,"n_14"] + x[,"n_20"])*log(2 + 3*r_rr - 8*r_rr^2 + 3*r_rr^3) + (x[,"n_11"] + x[,"n_23"])*log(12 + 20*r_rr - 11*r_rr^2 + 6*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_1.4 <- function(x){
logL_cc <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(2*log(2) + log(3) + log(5)) + (n04 + n31)*(2*log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n14 + n21 + n32)*(log(pmax(1e-6,r)) + log(5 + r - 3*r^2)) + (n01 + n34)*(log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n12 + n23)*log(15 - 17*r + 17*r^2 - 9*r^3) + (n02 + n11 + n24 + n33)*log(3 + 2*r - 8*r^2 + 3*r^3) + (n13 + n22)*log(6 + 10*r - 10*r^2 + 9*r^3)
return(L)}
interlogL_cc <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_cc,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cc <- -((x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + log10(7) + log10(11)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + log10(7) + log10(11)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*(-3*log10(2) + log10(19)) - (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(-3*log10(2) + log10(19)) + (x[,"n_04"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(5 + r_cc - 3*r_cc^2)) + (x[,"n_01"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cc)) + log10(3 - 4*r_cc + r_cc^2)) + (x[,"n_12"] + x[,"n_23"])*log10(15 - 17*r_cc + 17*r_cc^2 - 9*r_cc^3) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*log10(3 + 2*r_cc - 8*r_cc^2 + 3*r_cc^3) + (x[,"n_13"] + x[,"n_22"])*log10(6 + 10*r_cc - 10*r_cc^2 + 9*r_cc^3)
logL_cc <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(2*log(2) + log(3) + log(5)) + (x[,"n_04"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(5 + r_cc - 3*r_cc^2)) + (x[,"n_01"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cc)) + log(3 - 4*r_cc + r_cc^2)) + (x[,"n_12"] + x[,"n_23"])*log(15 - 17*r_cc + 17*r_cc^2 - 9*r_cc^3) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*log(3 + 2*r_cc - 8*r_cc^2 + 3*r_cc^3) + (x[,"n_13"] + x[,"n_22"])*log(6 + 10*r_cc - 10*r_cc^2 + 9*r_cc^3)
logL_cm <- function(r,n12,n13,n22,n23,n01,n02,n03,n04,n11,n14,n21,n24,n31,n32,n33,n34) {
L <- (n12 + n13 + n22 + n23)*log(3) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(3*log(2) + log(5)) + (n01 + n34)*(log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n04 + n31)*(log(pmax(1e-6,r)) + log(1 - r^2)) + (n13 + n22)*log(2 - r + 4*r^2 - 3*r^3) + (n02 + n11 + n24 + n33)*log(3 - 5*r + 6*r^2 - 3*r^3) + (n12 + n23)*log(2 + 2*r - 5*r^2 + 3*r^3) + (n03 + n14 + n21 + n32)*log(1 + 2*r - 3*r^2 + 3*r^3)
return(L)}
interlogL_cm <- function(n12,n13,n22,n23,n01,n02,n03,n04,n11,n14,n21,n24,n31,n32,n33,n34) {
optimize(logL_cm,c(0,0.5), n12,n13,n22,n23,n01,n02,n03,n04,n11,n14,n21,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_12"],x[,"n_13"],x[,"n_22"],x[,"n_23"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_14"],x[,"n_21"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cm <- -((x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(3))) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*(-3*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(-3*log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + log10(17)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + log10(17)) + (x[,"n_01"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cm)) + log10(2 - r_cm) + log10(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(1 - r_cm^2)) + (x[,"n_13"] + x[,"n_22"])*log10(2 - r_cm + 4*r_cm^2 - 3*r_cm^3) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*log10(3 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3) + (x[,"n_12"] + x[,"n_23"])*log10(2 + 2*r_cm - 5*r_cm^2 + 3*r_cm^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log10(1 + 2*r_cm - 3*r_cm^2 + 3*r_cm^3)
logL_cm <- (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log(3) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + (x[,"n_01"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cm)) + log(2 - r_cm) + log(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(1 - r_cm^2)) + (x[,"n_13"] + x[,"n_22"])*log(2 - r_cm + 4*r_cm^2 - 3*r_cm^3) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*log(3 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3) + (x[,"n_12"] + x[,"n_23"])*log(2 + 2*r_cm - 5*r_cm^2 + 3*r_cm^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log(1 + 2*r_cm - 3*r_cm^2 + 3*r_cm^3)
logL_cr <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(log(2) + log(5)) + (n04 + n31)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n34)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n02 + n11 + n24 + n33)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n13 + n22)*(log(pmax(1e-6,r)) + log(8 - 14*r + 9*r^2)) + (n12 + n23)*log(3 - 7*r + 13*r^2 - 9*r^3) + (n03 + n14 + n21 + n32)*log(1 - 3*r + 5*r^2 - 3*r^3)
return(L)}
interlogL_cr <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_cr,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cr <- 3*(x[,"n_04"] + x[,"n_31"])*log10(2) + 3*(x[,"n_01"] + x[,"n_34"])*log10(2) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*(3*log10(2) - log10(3)) - (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(-3*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + log10(13)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + log10(13)) + (x[,"n_04"] + x[,"n_31"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_13"] + x[,"n_22"])*(log10(pmax(1e-6,r_cr)) + log10(8 - 14*r_cr + 9*r_cr^2)) + (x[,"n_12"] + x[,"n_23"])*log10(3 - 7*r_cr + 13*r_cr^2 - 9*r_cr^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log10(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3)
logL_cr <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(log(2) + log(5)) + (x[,"n_04"] + x[,"n_31"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_13"] + x[,"n_22"])*(log(pmax(1e-6,r_cr)) + log(8 - 14*r_cr + 9*r_cr^2)) + (x[,"n_12"] + x[,"n_23"])*log(3 - 7*r_cr + 13*r_cr^2 - 9*r_cr^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3)
logL_rc <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(2*log(2) + log(3) + 2*log(5)) + (n04 + n31)*(log(3 - r) + log(pmax(1e-6,r)) + log(2 + r)) + (n01 + n34)*(log(2 + r) + log(3 - 4*r + r^2)) + (n13 + n22)*log(48 - 4*r + 13*r^2 - 9*r^3) + (n02 + n11 + n24 + n33)*log(15 - 8*r + 5*r^2 - 3*r^3) + (n03 + n14 + n21 + n32)*log(9 + 7*r - 4*r^2 + 3*r^3) + (n12 + n23)*log(48 + 5*r - 14*r^2 + 9*r^3)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rc <- -((x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + 2*log10(5))) - (x[,"n_04"] + x[,"n_31"])*(-log10(2) + 2*(-log10(2) + log10(5))) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + log10(5) + log10(7) + log10(11)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + log10(5) + log10(7) + log10(11)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*(-3*log10(2) + log10(5) + log10(19)) - (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(-3*log10(2) + log10(5) + log10(19)) + (x[,"n_04"] + x[,"n_31"])*(log10(3 - r_rc) + log10(pmax(1e-6,r_rc)) + log10(2 + r_rc)) + (x[,"n_01"] + x[,"n_34"])*(log10(2 + r_rc) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_13"] + x[,"n_22"])*log10(48 - 4*r_rc + 13*r_rc^2 - 9*r_rc^3) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*log10(15 - 8*r_rc + 5*r_rc^2 - 3*r_rc^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log10(9 + 7*r_rc - 4*r_rc^2 + 3*r_rc^3) + (x[,"n_12"] + x[,"n_23"])*log10(48 + 5*r_rc - 14*r_rc^2 + 9*r_rc^3)
logL_rc <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(2*log(2) + log(3) + 2*log(5)) + (x[,"n_04"] + x[,"n_31"])*(log(3 - r_rc) + log(pmax(1e-6,r_rc)) + log(2 + r_rc)) + (x[,"n_01"] + x[,"n_34"])*(log(2 + r_rc) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_13"] + x[,"n_22"])*log(48 - 4*r_rc + 13*r_rc^2 - 9*r_rc^3) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*log(15 - 8*r_rc + 5*r_rc^2 - 3*r_rc^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log(9 + 7*r_rc - 4*r_rc^2 + 3*r_rc^3) + (x[,"n_12"] + x[,"n_23"])*log(48 + 5*r_rc - 14*r_rc^2 + 9*r_rc^3)
logL_rm <- function(r,n12,n13,n22,n23,n01,n02,n03,n04,n11,n14,n21,n24,n31,n32,n33,n34) {
L <- (n12 + n13 + n22 + n23)*log(3) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(3*log(2) + 2*log(5)) + (n01 + n34)*(log(2 - r) + log(pmax(1e-6,r)) + log(2 + r)) + (n04 + n31)*(log(3 - r) + log(1 - r^2)) + (n12 + n23)*log(10 + 4*r^2 - 3*r^3) + (n03 + n14 + n21 + n32)*log(11 - 8*r + 6*r^2 - 3*r^3) + (n13 + n22)*log(11 + r - 5*r^2 + 3*r^3) + (n02 + n11 + n24 + n33)*log(6 + 5*r - 3*r^2 + 3*r^3)
return(L)}
interlogL_rm <- function(n12,n13,n22,n23,n01,n02,n03,n04,n11,n14,n21,n24,n31,n32,n33,n34) {
optimize(logL_rm,c(0,0.5), n12,n13,n22,n23,n01,n02,n03,n04,n11,n14,n21,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_12"],x[,"n_13"],x[,"n_22"],x[,"n_23"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_14"],x[,"n_21"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rm <- -((x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(3) + log10(5))) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*(-3*log10(2) + log10(5) + log10(13)) - (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(-3*log10(2) + log10(5) + log10(13)) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + log10(5) + log10(17)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + log10(5) + log10(17)) + (x[,"n_01"] + x[,"n_34"])*(log10(2 - r_rm) + log10(pmax(1e-6,r_rm)) + log10(2 + r_rm)) + (x[,"n_04"] + x[,"n_31"])*(log10(3 - r_rm) + log10(1 - r_rm^2)) + (x[,"n_12"] + x[,"n_23"])*log10(10 + 4*r_rm^2 - 3*r_rm^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log10(11 - 8*r_rm + 6*r_rm^2 - 3*r_rm^3) + (x[,"n_13"] + x[,"n_22"])*log10(11 + r_rm - 5*r_rm^2 + 3*r_rm^3) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*log10(6 + 5*r_rm - 3*r_rm^2 + 3*r_rm^3)
logL_rm <- (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log(3) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + 2*log(5)) + (x[,"n_01"] + x[,"n_34"])*(log(2 - r_rm) + log(pmax(1e-6,r_rm)) + log(2 + r_rm)) + (x[,"n_04"] + x[,"n_31"])*(log(3 - r_rm) + log(1 - r_rm^2)) + (x[,"n_12"] + x[,"n_23"])*log(10 + 4*r_rm^2 - 3*r_rm^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log(11 - 8*r_rm + 6*r_rm^2 - 3*r_rm^3) + (x[,"n_13"] + x[,"n_22"])*log(11 + r_rm - 5*r_rm^2 + 3*r_rm^3) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*log(6 + 5*r_rm - 3*r_rm^2 + 3*r_rm^3)
logL_rr <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(log(2) + 2*log(5)) + (n04 + n31)*(2*log(pmax(1e-6,1 - r)) + log(3 - r)) + (n01 + n34)*(2*log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n11 + n24 + n33)*(log(pmax(1e-6,r)) + log(4 + r - 3*r^2)) + (n13 + n22)*log(12 - 14*r + 17*r^2 - 9*r^3) + (n03 + n14 + n21 + n32)*log(2 + 3*r - 8*r^2 + 3*r^3) + (n12 + n23)*log(6 + 7*r - 10*r^2 + 9*r^3)
return(L)}
interlogL_rr <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_rr,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rr <- -((x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + log10(5) + log10(13)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + log10(5) + log10(13)) + (x[,"n_04"] + x[,"n_31"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 - r_rr)) + (x[,"n_01"] + x[,"n_34"])*(2*log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(log10(pmax(1e-6,r_rr)) + log10(4 + r_rr - 3*r_rr^2)) + (x[,"n_13"] + x[,"n_22"])*log10(12 - 14*r_rr + 17*r_rr^2 - 9*r_rr^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log10(2 + 3*r_rr - 8*r_rr^2 + 3*r_rr^3) + (x[,"n_12"] + x[,"n_23"])*log10(6 + 7*r_rr - 10*r_rr^2 + 9*r_rr^3)
logL_rr <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(log(2) + 2*log(5)) + (x[,"n_04"] + x[,"n_31"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 - r_rr)) + (x[,"n_01"] + x[,"n_34"])*(2*log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_02"] + x[,"n_11"] + x[,"n_24"] + x[,"n_33"])*(log(pmax(1e-6,r_rr)) + log(4 + r_rr - 3*r_rr^2)) + (x[,"n_13"] + x[,"n_22"])*log(12 - 14*r_rr + 17*r_rr^2 - 9*r_rr^3) + (x[,"n_03"] + x[,"n_14"] + x[,"n_21"] + x[,"n_32"])*log(2 + 3*r_rr - 8*r_rr^2 + 3*r_rr^3) + (x[,"n_12"] + x[,"n_23"])*log(6 + 7*r_rr - 10*r_rr^2 + 9*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_1.5 <- function(x){
logL_cc <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
L <- (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*(3*log(2) + log(5)) + (n02 + n34)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n04 + n32)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n03 + n33)*log(1 + 4*r - 4*r^2) + (n12 + n24)*log(6 - 3*r - 2*r^2) + (n14 + n22)*log(1 + 7*r - 2*r^2) + (n13 + n23)*log(9 - 4*r + 4*r^2)
return(L)}
interlogL_cc <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
optimize(logL_cc,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cc <- -2*(x[,"n_14"] + x[,"n_22"])*log10(2) - 3*(x[,"n_13"] + x[,"n_23"])*log10(2) - 2*(x[,"n_12"] + x[,"n_24"])*log10(2) - (x[,"n_03"] + x[,"n_33"])*log10(2) + (x[,"n_02"] + x[,"n_34"])*(log10(3 - 2*r_cc) + log10(pmax(1e-6,1 - r_cc))) + (x[,"n_04"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(1 + 2*r_cc)) + (x[,"n_03"] + x[,"n_33"])*log10(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_12"] + x[,"n_24"])*log10(6 - 3*r_cc - 2*r_cc^2) + (x[,"n_14"] + x[,"n_22"])*log10(1 + 7*r_cc - 2*r_cc^2) + (x[,"n_13"] + x[,"n_23"])*log10(9 - 4*r_cc + 4*r_cc^2)
logL_cc <- (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + (x[,"n_02"] + x[,"n_34"])*(log(3 - 2*r_cc) + log(pmax(1e-6,1 - r_cc))) + (x[,"n_04"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(1 + 2*r_cc)) + (x[,"n_03"] + x[,"n_33"])*log(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_12"] + x[,"n_24"])*log(6 - 3*r_cc - 2*r_cc^2) + (x[,"n_14"] + x[,"n_22"])*log(1 + 7*r_cc - 2*r_cc^2) + (x[,"n_13"] + x[,"n_23"])*log(9 - 4*r_cc + 4*r_cc^2)
logL_cm <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
L <- (-n02 - n03 - n04 - 2*n12 - 2*n13 - 2*n14 - 2*n22 - 2*n23 - 2*n24 - n32 - n33 - n34)*log(2) + (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*log(5) + (n02 + n04 + n32 + n34)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n13 + n23)*log(3 + 4*r - 4*r^2) + (n12 + n24)*log(3 - 3*r + 2*r^2) + (n03 + n33)*log(1 - 2*r + 2*r^2) + (n14 + n22)*log(2 - r + 2*r^2)
return(L)}
interlogL_cm <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
optimize(logL_cm,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cm <- -((x[,"n_14"] + x[,"n_22"])*log10(2)) - 2*(x[,"n_13"] + x[,"n_23"])*log10(2) - (x[,"n_12"] + x[,"n_24"])*log10(2) + (x[,"n_03"] + x[,"n_33"])*log10(2) + 2*(x[,"n_02"] + x[,"n_04"] + x[,"n_32"] + x[,"n_34"])*log10(2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_32"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_13"] + x[,"n_23"])*log10(3 + 4*r_cm - 4*r_cm^2) + (x[,"n_12"] + x[,"n_24"])*log10(3 - 3*r_cm + 2*r_cm^2) + (x[,"n_03"] + x[,"n_33"])*log10(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_14"] + x[,"n_22"])*log10(2 - r_cm + 2*r_cm^2)
logL_cm <- (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(2) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_02"] + x[,"n_04"] + x[,"n_32"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_13"] + x[,"n_23"])*log(3 + 4*r_cm - 4*r_cm^2) + (x[,"n_12"] + x[,"n_24"])*log(3 - 3*r_cm + 2*r_cm^2) + (x[,"n_03"] + x[,"n_33"])*log(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_14"] + x[,"n_22"])*log(2 - r_cm + 2*r_cm^2)
logL_rc <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
L <- (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*(3*log(2) + 2*log(5)) + (n02 + n34)*(log(3 - 2*r) + log(2 + r)) + (n04 + n32)*(log(3 - r) + log(1 + 2*r)) + (n13 + n23)*log(39 + 4*r - 4*r^2) + (n12 + n24)*log(21 - 3*r + 2*r^2) + (n14 + n22)*log(20 - r + 2*r^2) + (n03 + n33)*log(11 - 4*r + 4*r^2)
return(L)}
interlogL_rc <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
optimize(logL_rc,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rc <- -((x[,"n_04"] + x[,"n_32"])*log10(5)) - (x[,"n_02"] + x[,"n_34"])*log10(5) - (x[,"n_03"] + x[,"n_33"])*(log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_22"])*(2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_24"])*(2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_23"])*(3*log10(2) + log10(5)) + (x[,"n_02"] + x[,"n_34"])*(log10(3 - 2*r_rc) + log10(2 + r_rc)) + (x[,"n_04"] + x[,"n_32"])*(log10(3 - r_rc) + log10(1 + 2*r_rc)) + (x[,"n_13"] + x[,"n_23"])*log10(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_12"] + x[,"n_24"])*log10(21 - 3*r_rc + 2*r_rc^2) + (x[,"n_14"] + x[,"n_22"])*log10(20 - r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_33"])*log10(11 - 4*r_rc + 4*r_rc^2)
logL_rc <- (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + 2*log(5)) + (x[,"n_02"] + x[,"n_34"])*(log(3 - 2*r_rc) + log(2 + r_rc)) + (x[,"n_04"] + x[,"n_32"])*(log(3 - r_rc) + log(1 + 2*r_rc)) + (x[,"n_13"] + x[,"n_23"])*log(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_12"] + x[,"n_24"])*log(21 - 3*r_rc + 2*r_rc^2) + (x[,"n_14"] + x[,"n_22"])*log(20 - r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_33"])*log(11 - 4*r_rc + 4*r_rc^2)
logL_rm <- function(r,n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n03,n33) {
L <- (-n02 - n04 - 2*n12 - 2*n13 - 2*n14 - 2*n22 - 2*n23 - 2*n24 - n32 - n34)*log(2) + 2*(-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*log(5) + (n04 + n32)*(log(pmax(1e-6,1 - r)) + log(3 - r)) + (n02 + n34)*(log(pmax(1e-6,r)) + log(2 + r)) + (n14 + n22)*log(13 - 5*r - 2*r^2) + (n12 + n24)*log(6 + 9*r - 2*r^2) + (n03 + n33)*log(1 + r - r^2) + (n13 + n23)*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_rm <- function(n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n03,n33) {
optimize(logL_rm,c(0,0.5), n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n03,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_02"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_34"],x[,"n_03"],x[,"n_33"])
LOD_rm <- -((x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(5))) - (x[,"n_03"] + x[,"n_33"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_22"])*(log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_24"])*(log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_23"])*(2*log10(2) + log10(5)) + (x[,"n_04"] + x[,"n_32"])*(log10(pmax(1e-6,1 - r_rm)) + log10(3 - r_rm)) + (x[,"n_02"] + x[,"n_34"])*(log10(pmax(1e-6,r_rm)) + log10(2 + r_rm)) + (x[,"n_14"] + x[,"n_22"])*log10(13 - 5*r_rm - 2*r_rm^2) + (x[,"n_12"] + x[,"n_24"])*log10(6 + 9*r_rm - 2*r_rm^2) + (x[,"n_03"] + x[,"n_33"])*log10(1 + r_rm - r_rm^2) + (x[,"n_13"] + x[,"n_23"])*log10(21 - 4*r_rm + 4*r_rm^2)
logL_rm <- (-x[,"n_02"] - x[,"n_04"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_32"] - x[,"n_34"])*log(2) + 2*(-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_04"] + x[,"n_32"])*(log(pmax(1e-6,1 - r_rm)) + log(3 - r_rm)) + (x[,"n_02"] + x[,"n_34"])*(log(pmax(1e-6,r_rm)) + log(2 + r_rm)) + (x[,"n_14"] + x[,"n_22"])*log(13 - 5*r_rm - 2*r_rm^2) + (x[,"n_12"] + x[,"n_24"])*log(6 + 9*r_rm - 2*r_rm^2) + (x[,"n_03"] + x[,"n_33"])*log(1 + r_rm - r_rm^2) + (x[,"n_13"] + x[,"n_23"])*log(21 - 4*r_rm + 4*r_rm^2)

return(list(
r_mat = cbind(r_cc,r_cm,r_rc,r_rm,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_rc,LOD_rm,0),
logL_mat = cbind(logL_cc,logL_cm,logL_rc,logL_rm,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","repulsion coupling","repulsion mixed","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_2.1 <- function(x){
logL_cc <- function(r,n01,n02,n10,n11,n12,n13,n20,n21,n22,n23,n31,n32,n00,n03,n30,n33) {
L <- (-n01 - n02 - n10 - 2*n11 - 2*n12 - n13 - n20 - 2*n21 - 2*n22 - n23 - n31 - n32)*log(2) + 2*(-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*log(5) + 2*(n00 + n33)*log(pmax(1e-6,1 - r)) + 2*(n03 + n30)*log(pmax(1e-6,r)) + (n02 + n13 + n20 + n31)*(log(5 - 2*r) + log(pmax(1e-6,r))) + (n01 + n10 + n23 + n32)*log(3 - r - 2*r^2) + (n11 + n22)*log(21 - 12*r + 4*r^2) + (n12 + n21)*log(13 + 4*r + 4*r^2)
return(L)}
interlogL_cc <- function(n01,n02,n10,n11,n12,n13,n20,n21,n22,n23,n31,n32,n00,n03,n30,n33) {
optimize(logL_cc,c(0,0.5), n01,n02,n10,n11,n12,n13,n20,n21,n22,n23,n31,n32,n00,n03,n30,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_31"],x[,"n_32"],x[,"n_00"],x[,"n_03"],x[,"n_30"],x[,"n_33"])
LOD_cc <- -4*(x[,"n_12"] + x[,"n_21"])*log10(2) - 4*(x[,"n_11"] + x[,"n_22"])*log10(2) + 2*(x[,"n_03"] + x[,"n_30"])*log10(2) - (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log10(2) - (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(2) + 2*(x[,"n_00"] + x[,"n_33"])*log10(2) + 2*(x[,"n_00"] + x[,"n_33"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_03"] + x[,"n_30"])*log10(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log10(5 - 2*r_cc) + log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(3 - r_cc - 2*r_cc^2) + (x[,"n_11"] + x[,"n_22"])*log10(21 - 12*r_cc + 4*r_cc^2) + (x[,"n_12"] + x[,"n_21"])*log10(13 + 4*r_cc + 4*r_cc^2)
logL_cc <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - x[,"n_31"] - x[,"n_32"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*log(5) + 2*(x[,"n_00"] + x[,"n_33"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_03"] + x[,"n_30"])*log(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log(5 - 2*r_cc) + log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log(3 - r_cc - 2*r_cc^2) + (x[,"n_11"] + x[,"n_22"])*log(21 - 12*r_cc + 4*r_cc^2) + (x[,"n_12"] + x[,"n_21"])*log(13 + 4*r_cc + 4*r_cc^2)
logL_cr <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-2*n00 - 3*n01 - 3*n02 - 2*n03 - 2*n10 - 3*n11 - 3*n12 - 2*n13 - 2*n20 - 3*n21 - 3*n22 - 2*n23 - 2*n30 - 3*n31 - 3*n32 - 2*n33)*log(2) + 2*(-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*log(5) + (n03 + n30)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n00 + n33)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n11 + n22)*log(33 - 4*r^2) + (n12 + n21)*log(29 + 8*r - 4*r^2) + (n10 + n23)*log(6 - 5*r + 2*r^2) + (n13 + n20)*log(3 + r + 2*r^2) + (n01 + n02 + n31 + n32)*log(9 - 4*r + 4*r^2)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_cr <- -2*(x[,"n_13"] + x[,"n_20"])*log10(2) - 5*(x[,"n_12"] + x[,"n_21"])*log10(2) - 5*(x[,"n_11"] + x[,"n_22"])*log10(2) - 2*(x[,"n_10"] + x[,"n_23"])*log10(2) - 3*(x[,"n_01"] + x[,"n_02"] + x[,"n_31"] + x[,"n_32"])*log10(2) + (x[,"n_03"] + x[,"n_30"])*(log10(3 - 2*r_cr) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_33"])*(log10(pmax(1e-6,1 - r_cr)) + log10(1 + 2*r_cr)) + (x[,"n_11"] + x[,"n_22"])*log10(33 - 4*r_cr^2) + (x[,"n_12"] + x[,"n_21"])*log10(29 + 8*r_cr - 4*r_cr^2) + (x[,"n_10"] + x[,"n_23"])*log10(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_13"] + x[,"n_20"])*log10(3 + r_cr + 2*r_cr^2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_31"] + x[,"n_32"])*log10(9 - 4*r_cr + 4*r_cr^2)
logL_cr <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 2*x[,"n_33"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*log(5) + (x[,"n_03"] + x[,"n_30"])*(log(3 - 2*r_cr) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_33"])*(log(pmax(1e-6,1 - r_cr)) + log(1 + 2*r_cr)) + (x[,"n_11"] + x[,"n_22"])*log(33 - 4*r_cr^2) + (x[,"n_12"] + x[,"n_21"])*log(29 + 8*r_cr - 4*r_cr^2) + (x[,"n_10"] + x[,"n_23"])*log(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_13"] + x[,"n_20"])*log(3 + r_cr + 2*r_cr^2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_31"] + x[,"n_32"])*log(9 - 4*r_cr + 4*r_cr^2)
logL_rc <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 2*n30 - 2*n31 - 2*n32 - 2*n33)*log(2) + 2*(-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*log(5) + (n03 + n30)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n00 + n33)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n11 + n22)*log(33 - 4*r^2) + (n12 + n21)*log(29 + 8*r - 4*r^2) + (n01 + n32)*log(6 - 5*r + 2*r^2) + (n02 + n31)*log(3 + r + 2*r^2) + (n10 + n13 + n20 + n23)*log(9 - 4*r + 4*r^2)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_rc <- -5*(x[,"n_12"] + x[,"n_21"])*log10(2) - 5*(x[,"n_11"] + x[,"n_22"])*log10(2) - 3*(x[,"n_10"] + x[,"n_13"] + x[,"n_20"] + x[,"n_23"])*log10(2) - 2*(x[,"n_02"] + x[,"n_31"])*log10(2) - 2*(x[,"n_01"] + x[,"n_32"])*log10(2) + (x[,"n_03"] + x[,"n_30"])*(log10(3 - 2*r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_33"])*(log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_11"] + x[,"n_22"])*log10(33 - 4*r_rc^2) + (x[,"n_12"] + x[,"n_21"])*log10(29 + 8*r_rc - 4*r_rc^2) + (x[,"n_01"] + x[,"n_32"])*log10(6 - 5*r_rc + 2*r_rc^2) + (x[,"n_02"] + x[,"n_31"])*log10(3 + r_rc + 2*r_rc^2) + (x[,"n_10"] + x[,"n_13"] + x[,"n_20"] + x[,"n_23"])*log10(9 - 4*r_rc + 4*r_rc^2)
logL_rc <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*log(5) + (x[,"n_03"] + x[,"n_30"])*(log(3 - 2*r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_33"])*(log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_11"] + x[,"n_22"])*log(33 - 4*r_rc^2) + (x[,"n_12"] + x[,"n_21"])*log(29 + 8*r_rc - 4*r_rc^2) + (x[,"n_01"] + x[,"n_32"])*log(6 - 5*r_rc + 2*r_rc^2) + (x[,"n_02"] + x[,"n_31"])*log(3 + r_rc + 2*r_rc^2) + (x[,"n_10"] + x[,"n_13"] + x[,"n_20"] + x[,"n_23"])*log(9 - 4*r_rc + 4*r_rc^2)
logL_rr <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(4*log(2) + 2*log(5)) + 2*(n03 + n30)*log(3 - 2*r) + 2*(n00 + n33)*log(1 + 2*r) + (n02 + n13 + n20 + n31)*log(21 - 8*r - 4*r^2) + (n01 + n10 + n23 + n32)*log(9 + 16*r - 4*r^2) + (n12 + n21)*log(73 - 20*r + 4*r^2) + (n11 + n22)*log(57 + 12*r + 4*r^2)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_rr <- -6*(x[,"n_12"] + x[,"n_21"])*log10(2) - 6*(x[,"n_11"] + x[,"n_22"])*log10(2) - 2*(x[,"n_03"] + x[,"n_30"])*log10(2) - 4*(x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log10(2) - 4*(x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(2) - 2*(x[,"n_00"] + x[,"n_33"])*log10(2) + 2*(x[,"n_03"] + x[,"n_30"])*log10(3 - 2*r_rr) + 2*(x[,"n_00"] + x[,"n_33"])*log10(1 + 2*r_rr) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log10(21 - 8*r_rr - 4*r_rr^2) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(9 + 16*r_rr - 4*r_rr^2) + (x[,"n_12"] + x[,"n_21"])*log10(73 - 20*r_rr + 4*r_rr^2) + (x[,"n_11"] + x[,"n_22"])*log10(57 + 12*r_rr + 4*r_rr^2)
logL_rr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(4*log(2) + 2*log(5)) + 2*(x[,"n_03"] + x[,"n_30"])*log(3 - 2*r_rr) + 2*(x[,"n_00"] + x[,"n_33"])*log(1 + 2*r_rr) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log(21 - 8*r_rr - 4*r_rr^2) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log(9 + 16*r_rr - 4*r_rr^2) + (x[,"n_12"] + x[,"n_21"])*log(73 - 20*r_rr + 4*r_rr^2) + (x[,"n_11"] + x[,"n_22"])*log(57 + 12*r_rr + 4*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_2.2 <- function(x){
logL_cc <- function(r,n01,n02,n03,n11,n12,n13,n21,n22,n23,n31,n32,n33,n10,n14,n20,n24,n00,n04,n30,n34) {
L <- (-n01 + 2*n02 - n03 - n11 - n12 - n13 - n21 - n22 - n23 - n31 + 2*n32 - n33)*log(2) + (n10 + n14 + n20 + n24)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*log(5) + 3*(n00 + n34)*log(pmax(1e-6,1 - r)) + 3*(n04 + n30)*log(pmax(1e-6,r)) + (n02 + n32)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n24)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n31)*(log(7 - 4*r) + 2*log(pmax(1e-6,r))) + (n14 + n20)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n33)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r)) + (n12 + n22)*log(11 - 8*r + 8*r^2) + (n13 + n21)*(log(pmax(1e-6,r)) + log(16 - 19*r + 12*r^2)) + (n11 + n23)*log(9 - 14*r + 17*r^2 - 12*r^3)
return(L)}
interlogL_cc <- function(n01,n02,n03,n11,n12,n13,n21,n22,n23,n31,n32,n33,n10,n14,n20,n24,n00,n04,n30,n34) {
optimize(logL_cc,c(0,0.5), n01,n02,n03,n11,n12,n13,n21,n22,n23,n31,n32,n33,n10,n14,n20,n24,n00,n04,n30,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_10"],x[,"n_14"],x[,"n_20"],x[,"n_24"],x[,"n_00"],x[,"n_04"],x[,"n_30"],x[,"n_34"])
LOD_cc <- 3*(x[,"n_14"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_24"])*log10(2) + 3*(x[,"n_04"] + x[,"n_30"])*log10(2) + 2*(x[,"n_02"] + x[,"n_32"])*log10(2) + 3*(x[,"n_00"] + x[,"n_34"])*log10(2) - 2*(x[,"n_12"] + x[,"n_22"])*log10(3) - (x[,"n_03"] + x[,"n_31"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_33"])*(-2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_21"])*(-2*log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_23"])*(-2*log10(2) + log10(19)) + 3*(x[,"n_00"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_04"] + x[,"n_30"])*log10(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_32"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_24"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_31"])*(log10(7 - 4*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_20"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_33"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 4*r_cc)) + (x[,"n_12"] + x[,"n_22"])*log10(11 - 8*r_cc + 8*r_cc^2) + (x[,"n_13"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(16 - 19*r_cc + 12*r_cc^2)) + (x[,"n_11"] + x[,"n_23"])*log10(9 - 14*r_cc + 17*r_cc^2 - 12*r_cc^3)
logL_cc <- (-x[,"n_01"] + 2*x[,"n_02"] - x[,"n_03"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_31"] + 2*x[,"n_32"] - x[,"n_33"])*log(2) + (x[,"n_10"] + x[,"n_14"] + x[,"n_20"] + x[,"n_24"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + 3*(x[,"n_00"] + x[,"n_34"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_04"] + x[,"n_30"])*log(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_32"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_24"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_31"])*(log(7 - 4*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_20"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_33"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 4*r_cc)) + (x[,"n_12"] + x[,"n_22"])*log(11 - 8*r_cc + 8*r_cc^2) + (x[,"n_13"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(16 - 19*r_cc + 12*r_cc^2)) + (x[,"n_11"] + x[,"n_23"])*log(9 - 14*r_cc + 17*r_cc^2 - 12*r_cc^3)
logL_cm <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 2*n04 - 2*n10 - 3*n11 - 3*n12 - 3*n13 - 2*n14 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 2*n24 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 2*n34)*log(2) + 2*(n02 + n32)*log(3) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n10 - 2*n11 - n12 - 2*n13 - 2*n14 - 2*n20 - 2*n21 - n22 - 2*n23 - 2*n24 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34)*log(5) + (n12 + n22)*log(7) + (n04 + n30)*(log(2 - r) + 2*log(pmax(1e-6,r))) + (n00 + n34)*(log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n14 + n20)*(log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n03 + n31)*(log(pmax(1e-6,r)) + log(12 - 7*r + 4*r^2)) + (n13 + n21)*log(15 + 2*r + 19*r^2 - 12*r^3) + (n01 + n33)*log(9 - 10*r + 5*r^2 - 4*r^3) + (n10 + n24)*log(3 - 4*r + 4*r^2 - 3*r^3) + (n11 + n23)*log(24 - 4*r - 17*r^2 + 12*r^3)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cm <- -((x[,"n_04"] + x[,"n_30"])*(-3*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_34"])*(-3*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_21"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_11"] + x[,"n_23"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_14"] + x[,"n_20"])*(-3*log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_24"])*(-3*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_31"])*(-2*log10(2) + log10(19)) - (x[,"n_01"] + x[,"n_33"])*(-2*log10(2) + log10(19)) + (x[,"n_04"] + x[,"n_30"])*(log10(2 - r_cm) + 2*log10(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cm)) + log10(1 - r_cm^2)) + (x[,"n_14"] + x[,"n_20"])*(log10(pmax(1e-6,r_cm)) + log10(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_03"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(12 - 7*r_cm + 4*r_cm^2)) + (x[,"n_13"] + x[,"n_21"])*log10(15 + 2*r_cm + 19*r_cm^2 - 12*r_cm^3) + (x[,"n_01"] + x[,"n_33"])*log10(9 - 10*r_cm + 5*r_cm^2 - 4*r_cm^3) + (x[,"n_10"] + x[,"n_24"])*log10(3 - 4*r_cm + 4*r_cm^2 - 3*r_cm^3) + (x[,"n_11"] + x[,"n_23"])*log10(24 - 4*r_cm - 17*r_cm^2 + 12*r_cm^3)
logL_cm <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 2*x[,"n_34"])*log(2) + 2*(x[,"n_02"] + x[,"n_32"])*log(3) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"])*log(5) + (x[,"n_12"] + x[,"n_22"])*log(7) + (x[,"n_04"] + x[,"n_30"])*(log(2 - r_cm) + 2*log(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cm)) + log(1 - r_cm^2)) + (x[,"n_14"] + x[,"n_20"])*(log(pmax(1e-6,r_cm)) + log(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_03"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(12 - 7*r_cm + 4*r_cm^2)) + (x[,"n_13"] + x[,"n_21"])*log(15 + 2*r_cm + 19*r_cm^2 - 12*r_cm^3) + (x[,"n_01"] + x[,"n_33"])*log(9 - 10*r_cm + 5*r_cm^2 - 4*r_cm^3) + (x[,"n_10"] + x[,"n_24"])*log(3 - 4*r_cm + 4*r_cm^2 - 3*r_cm^3) + (x[,"n_11"] + x[,"n_23"])*log(24 - 4*r_cm - 17*r_cm^2 + 12*r_cm^3)
logL_cr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-n00 - 2*n01 - 2*n02 - 2*n03 - n04 - n10 - 2*n11 - 2*n12 - 2*n13 - n14 - n20 - 2*n21 - 2*n22 - 2*n23 - n24 - n30 - 2*n31 - 2*n32 - 2*n33 - n34)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(log(3) + 2*log(5)) + (n00 + n34)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n12 + n22)*log(51 + 8*r - 8*r^2) + (n04 + n30)*(log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n02 + n32)*log(15 - 8*r + 8*r^2) + (n11 + n23)*log(33 - 14*r + 17*r^2 - 12*r^3) + (n03 + n31)*log(9 - 6*r + 7*r^2 - 4*r^3) + (n14 + n20)*log(3 - 4*r + 7*r^2 - 3*r^3) + (n10 + n24)*log(3 - r - 2*r^2 + 3*r^3) + (n01 + n33)*log(6 + 4*r - 5*r^2 + 4*r^3) + (n13 + n21)*log(24 + 16*r - 19*r^2 + 12*r^3)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cr <- -((x[,"n_04"] + x[,"n_30"])*(-3*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_34"])*(-3*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_32"])*log10(13) - (x[,"n_14"] + x[,"n_20"])*(-3*log10(2) + log10(19)) - (x[,"n_10"] + x[,"n_24"])*(-3*log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_21"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_11"] + x[,"n_23"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_03"] + x[,"n_31"])*(-2*log10(2) + log10(29)) - (x[,"n_01"] + x[,"n_33"])*(-2*log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_22"])*log10(53) + (x[,"n_00"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr)) + log10(2 + r_cr)) + (x[,"n_12"] + x[,"n_22"])*log10(51 + 8*r_cr - 8*r_cr^2) + (x[,"n_04"] + x[,"n_30"])*(log10(pmax(1e-6,r_cr)) + log10(3 - 4*r_cr + r_cr^2)) + (x[,"n_02"] + x[,"n_32"])*log10(15 - 8*r_cr + 8*r_cr^2) + (x[,"n_11"] + x[,"n_23"])*log10(33 - 14*r_cr + 17*r_cr^2 - 12*r_cr^3) + (x[,"n_03"] + x[,"n_31"])*log10(9 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_14"] + x[,"n_20"])*log10(3 - 4*r_cr + 7*r_cr^2 - 3*r_cr^3) + (x[,"n_10"] + x[,"n_24"])*log10(3 - r_cr - 2*r_cr^2 + 3*r_cr^3) + (x[,"n_01"] + x[,"n_33"])*log10(6 + 4*r_cr - 5*r_cr^2 + 4*r_cr^3) + (x[,"n_13"] + x[,"n_21"])*log10(24 + 16*r_cr - 19*r_cr^2 + 12*r_cr^3)
logL_cr <- (-x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr)) + log(2 + r_cr)) + (x[,"n_12"] + x[,"n_22"])*log(51 + 8*r_cr - 8*r_cr^2) + (x[,"n_04"] + x[,"n_30"])*(log(pmax(1e-6,r_cr)) + log(3 - 4*r_cr + r_cr^2)) + (x[,"n_02"] + x[,"n_32"])*log(15 - 8*r_cr + 8*r_cr^2) + (x[,"n_11"] + x[,"n_23"])*log(33 - 14*r_cr + 17*r_cr^2 - 12*r_cr^3) + (x[,"n_03"] + x[,"n_31"])*log(9 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_14"] + x[,"n_20"])*log(3 - 4*r_cr + 7*r_cr^2 - 3*r_cr^3) + (x[,"n_10"] + x[,"n_24"])*log(3 - r_cr - 2*r_cr^2 + 3*r_cr^3) + (x[,"n_01"] + x[,"n_33"])*log(6 + 4*r_cr - 5*r_cr^2 + 4*r_cr^3) + (x[,"n_13"] + x[,"n_21"])*log(24 + 16*r_cr - 19*r_cr^2 + 12*r_cr^3)
logL_rc <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-2*n00 - n01 - 2*n02 - n03 - 2*n04 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n30 - n31 - 2*n32 - n33 - 2*n34)*log(2) + (n10 + n14 + n20 + n24)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*log(5) + (n04 + n30)*(log(3 - 2*r) + 2*log(pmax(1e-6,r))) + (n00 + n34)*(2*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n02 + n32)*log(3 + 4*r - 4*r^2) + (n14 + n20)*(log(pmax(1e-6,r)) + log(2 - 3*r + 2*r^2)) + (n03 + n31)*(log(pmax(1e-6,r)) + log(3 - 2*r + 2*r^2)) + (n12 + n22)*log(19 - 4*r + 4*r^2) + (n13 + n21)*log(9 - 4*r + 16*r^2 - 12*r^3) + (n10 + n24)*log(1 - 2*r + 3*r^2 - 2*r^3) + (n01 + n33)*log(3 - 5*r + 4*r^2 - 2*r^3) + (n11 + n23)*log(9 + 8*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rc <- (x[,"n_14"] + x[,"n_20"])*log10(2) + (x[,"n_10"] + x[,"n_24"])*log10(2) + (x[,"n_04"] + x[,"n_30"])*log10(2) - 2*(x[,"n_02"] + x[,"n_32"])*log10(2) + (x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_12"] + x[,"n_22"])*(log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_31"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_33"])*(-2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_21"])*(-log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_23"])*(-log10(2) + log10(19)) + (x[,"n_04"] + x[,"n_30"])*(log10(3 - 2*r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_02"] + x[,"n_32"])*log10(3 + 4*r_rc - 4*r_rc^2) + (x[,"n_14"] + x[,"n_20"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 3*r_rc + 2*r_rc^2)) + (x[,"n_03"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(3 - 2*r_rc + 2*r_rc^2)) + (x[,"n_12"] + x[,"n_22"])*log10(19 - 4*r_rc + 4*r_rc^2) + (x[,"n_13"] + x[,"n_21"])*log10(9 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_10"] + x[,"n_24"])*log10(1 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3) + (x[,"n_01"] + x[,"n_33"])*log10(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3) + (x[,"n_11"] + x[,"n_23"])*log10(9 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rc <- (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"])*log(2) + (x[,"n_10"] + x[,"n_14"] + x[,"n_20"] + x[,"n_24"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_04"] + x[,"n_30"])*(log(3 - 2*r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_02"] + x[,"n_32"])*log(3 + 4*r_rc - 4*r_rc^2) + (x[,"n_14"] + x[,"n_20"])*(log(pmax(1e-6,r_rc)) + log(2 - 3*r_rc + 2*r_rc^2)) + (x[,"n_03"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(3 - 2*r_rc + 2*r_rc^2)) + (x[,"n_12"] + x[,"n_22"])*log(19 - 4*r_rc + 4*r_rc^2) + (x[,"n_13"] + x[,"n_21"])*log(9 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_10"] + x[,"n_24"])*log(1 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3) + (x[,"n_01"] + x[,"n_33"])*log(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3) + (x[,"n_11"] + x[,"n_23"])*log(9 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rm <- function(r,n02,n32,n00,n01,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n33,n34) {
L <- (n02 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(4*log(2) + 2*log(5)) + (n04 + n30)*(log(3 - 2*r) + log(2 - r) + log(pmax(1e-6,r))) + (n12 + n22)*log(67 + 12*r - 12*r^2) + (n00 + n34)*(log(1 + 2*r) + log(1 - r^2)) + (n02 + n32)*log(7 - 4*r + 4*r^2) + (n11 + n23)*log(39 - 8*r + 20*r^2 - 12*r^3) + (n14 + n20)*log(9 - 10*r + 13*r^2 - 6*r^3) + (n03 + n31)*log(9 + 4*r^2 - 4*r^3) + (n01 + n33)*log(9 + 4*r - 8*r^2 + 4*r^3) + (n10 + n24)*log(6 + 2*r - 5*r^2 + 6*r^3) + (n13 + n21)*log(39 + 4*r - 16*r^2 + 12*r^3)
return(L)}
interlogL_rm <- function(n02,n32,n00,n01,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n33,n34) {
optimize(logL_rm,c(0,0.5), n02,n32,n00,n01,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_02"],x[,"n_32"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"])
LOD_rm <- -((x[,"n_04"] + x[,"n_30"])*(-log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_34"])*(-log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_32"])*(log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_22"])*(log10(2) + log10(5) + log10(7)) - (x[,"n_13"] + x[,"n_21"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_11"] + x[,"n_23"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_14"] + x[,"n_20"])*(-log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_24"])*(-log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_31"])*(-log10(2) + log10(19)) - (x[,"n_01"] + x[,"n_33"])*(-log10(2) + log10(19)) + (x[,"n_04"] + x[,"n_30"])*(log10(3 - 2*r_rm) + log10(2 - r_rm) + log10(pmax(1e-6,r_rm))) + (x[,"n_12"] + x[,"n_22"])*log10(67 + 12*r_rm - 12*r_rm^2) + (x[,"n_00"] + x[,"n_34"])*(log10(1 + 2*r_rm) + log10(1 - r_rm^2)) + (x[,"n_02"] + x[,"n_32"])*log10(7 - 4*r_rm + 4*r_rm^2) + (x[,"n_11"] + x[,"n_23"])*log10(39 - 8*r_rm + 20*r_rm^2 - 12*r_rm^3) + (x[,"n_14"] + x[,"n_20"])*log10(9 - 10*r_rm + 13*r_rm^2 - 6*r_rm^3) + (x[,"n_03"] + x[,"n_31"])*log10(9 + 4*r_rm^2 - 4*r_rm^3) + (x[,"n_01"] + x[,"n_33"])*log10(9 + 4*r_rm - 8*r_rm^2 + 4*r_rm^3) + (x[,"n_10"] + x[,"n_24"])*log10(6 + 2*r_rm - 5*r_rm^2 + 6*r_rm^3) + (x[,"n_13"] + x[,"n_21"])*log10(39 + 4*r_rm - 16*r_rm^2 + 12*r_rm^3)
logL_rm <- (x[,"n_02"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(4*log(2) + 2*log(5)) + (x[,"n_04"] + x[,"n_30"])*(log(3 - 2*r_rm) + log(2 - r_rm) + log(pmax(1e-6,r_rm))) + (x[,"n_12"] + x[,"n_22"])*log(67 + 12*r_rm - 12*r_rm^2) + (x[,"n_00"] + x[,"n_34"])*(log(1 + 2*r_rm) + log(1 - r_rm^2)) + (x[,"n_02"] + x[,"n_32"])*log(7 - 4*r_rm + 4*r_rm^2) + (x[,"n_11"] + x[,"n_23"])*log(39 - 8*r_rm + 20*r_rm^2 - 12*r_rm^3) + (x[,"n_14"] + x[,"n_20"])*log(9 - 10*r_rm + 13*r_rm^2 - 6*r_rm^3) + (x[,"n_03"] + x[,"n_31"])*log(9 + 4*r_rm^2 - 4*r_rm^3) + (x[,"n_01"] + x[,"n_33"])*log(9 + 4*r_rm - 8*r_rm^2 + 4*r_rm^3) + (x[,"n_10"] + x[,"n_24"])*log(6 + 2*r_rm - 5*r_rm^2 + 6*r_rm^3) + (x[,"n_13"] + x[,"n_21"])*log(39 + 4*r_rm - 16*r_rm^2 + 12*r_rm^3)
logL_rr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(3*log(2) + log(3) + 2*log(5)) + (n00 + n34)*(log(pmax(1e-6,r)) + log(2 + r) + log(1 + 2*r)) + (n02 + n32)*log(21 + 20*r - 20*r^2) + (n04 + n30)*(log(3 - 2*r) + log(3 - 4*r + r^2)) + (n12 + n22)*log(111 - 20*r + 20*r^2) + (n13 + n21)*log(75 - 40*r + 16*r^2 - 12*r^3) + (n10 + n24)*log(3 + 14*r + r^2 - 6*r^3) + (n01 + n33)*log(3 + 20*r + 8*r^2 - 4*r^3) + (n03 + n31)*log(27 - 24*r - 4*r^2 + 4*r^3) + (n14 + n20)*log(12 + 2*r - 17*r^2 + 6*r^3) + (n11 + n23)*log(39 + 44*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rr <- -((x[,"n_04"] + x[,"n_30"])*(-log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_34"])*(-log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_32"])*(log10(2) + log10(13)) - (x[,"n_14"] + x[,"n_20"])*(-log10(2) + log10(19)) - (x[,"n_10"] + x[,"n_24"])*(-log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_21"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_11"] + x[,"n_23"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_03"] + x[,"n_31"])*(-log10(2) + log10(29)) - (x[,"n_01"] + x[,"n_33"])*(-log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_22"])*(log10(2) + log10(53)) + (x[,"n_00"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(2 + r_rr) + log10(1 + 2*r_rr)) + (x[,"n_02"] + x[,"n_32"])*log10(21 + 20*r_rr - 20*r_rr^2) + (x[,"n_04"] + x[,"n_30"])*(log10(3 - 2*r_rr) + log10(3 - 4*r_rr + r_rr^2)) + (x[,"n_12"] + x[,"n_22"])*log10(111 - 20*r_rr + 20*r_rr^2) + (x[,"n_13"] + x[,"n_21"])*log10(75 - 40*r_rr + 16*r_rr^2 - 12*r_rr^3) + (x[,"n_10"] + x[,"n_24"])*log10(3 + 14*r_rr + r_rr^2 - 6*r_rr^3) + (x[,"n_01"] + x[,"n_33"])*log10(3 + 20*r_rr + 8*r_rr^2 - 4*r_rr^3) + (x[,"n_03"] + x[,"n_31"])*log10(27 - 24*r_rr - 4*r_rr^2 + 4*r_rr^3) + (x[,"n_14"] + x[,"n_20"])*log10(12 + 2*r_rr - 17*r_rr^2 + 6*r_rr^3) + (x[,"n_11"] + x[,"n_23"])*log10(39 + 44*r_rr - 20*r_rr^2 + 12*r_rr^3)
logL_rr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(2 + r_rr) + log(1 + 2*r_rr)) + (x[,"n_02"] + x[,"n_32"])*log(21 + 20*r_rr - 20*r_rr^2) + (x[,"n_04"] + x[,"n_30"])*(log(3 - 2*r_rr) + log(3 - 4*r_rr + r_rr^2)) + (x[,"n_12"] + x[,"n_22"])*log(111 - 20*r_rr + 20*r_rr^2) + (x[,"n_13"] + x[,"n_21"])*log(75 - 40*r_rr + 16*r_rr^2 - 12*r_rr^3) + (x[,"n_10"] + x[,"n_24"])*log(3 + 14*r_rr + r_rr^2 - 6*r_rr^3) + (x[,"n_01"] + x[,"n_33"])*log(3 + 20*r_rr + 8*r_rr^2 - 4*r_rr^3) + (x[,"n_03"] + x[,"n_31"])*log(27 - 24*r_rr - 4*r_rr^2 + 4*r_rr^3) + (x[,"n_14"] + x[,"n_20"])*log(12 + 2*r_rr - 17*r_rr^2 + 6*r_rr^3) + (x[,"n_11"] + x[,"n_23"])*log(39 + 44*r_rr - 20*r_rr^2 + 12*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_2.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 2*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35)*log(2) + (n10 + n15 + n20 + n25)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(5) + 3*(n00 + n35)*log(pmax(1e-6,1 - r)) + 3*(n05 + n30)*log(pmax(1e-6,r)) + (n10 + n25)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n31)*(log(11 - 2*r) + 2*log(pmax(1e-6,r))) + (n15 + n20)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n34)*(2*log(pmax(1e-6,1 - r)) + log(9 + 2*r)) + (n03 + n32)*(log(pmax(1e-6,r)) + log(18 - 5*r - 4*r^2)) + (n14 + n21)*(log(pmax(1e-6,r)) + log(26 - 17*r + 6*r^2)) + (n12 + n23)*log(41 - 34*r + 29*r^2 - 12*r^3) + (n11 + n24)*log(15 - 10*r + r^2 - 6*r^3) + (n02 + n33)*log(9 + 4*r - 17*r^2 + 4*r^3) + (n13 + n22)*log(24 + 12*r - 7*r^2 + 12*r^3)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_cc <- 3*(x[,"n_15"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_25"])*log10(2) + 3*(x[,"n_05"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_04"] + x[,"n_31"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_34"])*(-log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_22"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_12"] + x[,"n_23"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_14"] + x[,"n_21"])*(-log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_24"])*(-log10(2) + log10(19)) - (x[,"n_03"] + x[,"n_32"])*(-2*log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_33"])*(-2*log10(2) + log10(29)) + 3*(x[,"n_00"] + x[,"n_35"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_05"] + x[,"n_30"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_25"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_31"])*(log10(11 - 2*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_20"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(9 + 2*r_cc)) + (x[,"n_03"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(18 - 5*r_cc - 4*r_cc^2)) + (x[,"n_14"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(26 - 17*r_cc + 6*r_cc^2)) + (x[,"n_12"] + x[,"n_23"])*log10(41 - 34*r_cc + 29*r_cc^2 - 12*r_cc^3) + (x[,"n_11"] + x[,"n_24"])*log10(15 - 10*r_cc + r_cc^2 - 6*r_cc^3) + (x[,"n_02"] + x[,"n_33"])*log10(9 + 4*r_cc - 17*r_cc^2 + 4*r_cc^3) + (x[,"n_13"] + x[,"n_22"])*log10(24 + 12*r_cc - 7*r_cc^2 + 12*r_cc^3)
logL_cc <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"])*log(2) + (x[,"n_10"] + x[,"n_15"] + x[,"n_20"] + x[,"n_25"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + 3*(x[,"n_00"] + x[,"n_35"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_05"] + x[,"n_30"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_25"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_31"])*(log(11 - 2*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_20"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_cc)) + log(9 + 2*r_cc)) + (x[,"n_03"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(18 - 5*r_cc - 4*r_cc^2)) + (x[,"n_14"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(26 - 17*r_cc + 6*r_cc^2)) + (x[,"n_12"] + x[,"n_23"])*log(41 - 34*r_cc + 29*r_cc^2 - 12*r_cc^3) + (x[,"n_11"] + x[,"n_24"])*log(15 - 10*r_cc + r_cc^2 - 6*r_cc^3) + (x[,"n_02"] + x[,"n_33"])*log(9 + 4*r_cc - 17*r_cc^2 + 4*r_cc^3) + (x[,"n_13"] + x[,"n_22"])*log(24 + 12*r_cc - 7*r_cc^2 + 12*r_cc^3)
logL_cm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - n05 - 2*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - n35)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(5) + (n00 + n35)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n31)*(log(pmax(1e-6,r)) + log(9 - 5*r + 2*r^2)) + (n15 + n20)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2)) + (n14 + n21)*log(21 + 8*r + 22*r^2 - 12*r^3) + (n13 + n22)*log(27 + 2*r + 10*r^2 - 8*r^3) + (n10 + n25)*log(3 - 7*r + 10*r^2 - 6*r^3) + (n02 + n33)*log(15 - 11*r + 9*r^2 - 4*r^3) + (n01 + n34)*log(6 - 5*r + r^2 - 2*r^3) + (n03 + n32)*log(9 + 5*r - 3*r^2 + 4*r^3) + (n12 + n23)*log(31 + 2*r - 14*r^2 + 8*r^3) + (n11 + n24)*log(39 - 16*r - 14*r^2 + 12*r^3)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_cm <- 3*(x[,"n_05"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_15"] + x[,"n_20"])*(-2*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_25"])*(-2*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_32"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_33"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_04"] + x[,"n_31"])*(-log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_34"])*(-log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_21"])*log10(29) - (x[,"n_11"] + x[,"n_24"])*log10(29) - (x[,"n_13"] + x[,"n_22"])*(-log10(2) + log10(59)) - (x[,"n_12"] + x[,"n_23"])*(-log10(2) + log10(59)) + (x[,"n_00"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cm)) + 2*log10(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(9 - 5*r_cm + 2*r_cm^2)) + (x[,"n_15"] + x[,"n_20"])*(log10(pmax(1e-6,r_cm)) + log10(5 - 8*r_cm + 6*r_cm^2)) + (x[,"n_14"] + x[,"n_21"])*log10(21 + 8*r_cm + 22*r_cm^2 - 12*r_cm^3) + (x[,"n_13"] + x[,"n_22"])*log10(27 + 2*r_cm + 10*r_cm^2 - 8*r_cm^3) + (x[,"n_10"] + x[,"n_25"])*log10(3 - 7*r_cm + 10*r_cm^2 - 6*r_cm^3) + (x[,"n_02"] + x[,"n_33"])*log10(15 - 11*r_cm + 9*r_cm^2 - 4*r_cm^3) + (x[,"n_01"] + x[,"n_34"])*log10(6 - 5*r_cm + r_cm^2 - 2*r_cm^3) + (x[,"n_03"] + x[,"n_32"])*log10(9 + 5*r_cm - 3*r_cm^2 + 4*r_cm^3) + (x[,"n_12"] + x[,"n_23"])*log10(31 + 2*r_cm - 14*r_cm^2 + 8*r_cm^3) + (x[,"n_11"] + x[,"n_24"])*log10(39 - 16*r_cm - 14*r_cm^2 + 12*r_cm^3)
logL_cm <- (-x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - x[,"n_05"] - 2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_00"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cm)) + 2*log(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(9 - 5*r_cm + 2*r_cm^2)) + (x[,"n_15"] + x[,"n_20"])*(log(pmax(1e-6,r_cm)) + log(5 - 8*r_cm + 6*r_cm^2)) + (x[,"n_14"] + x[,"n_21"])*log(21 + 8*r_cm + 22*r_cm^2 - 12*r_cm^3) + (x[,"n_13"] + x[,"n_22"])*log(27 + 2*r_cm + 10*r_cm^2 - 8*r_cm^3) + (x[,"n_10"] + x[,"n_25"])*log(3 - 7*r_cm + 10*r_cm^2 - 6*r_cm^3) + (x[,"n_02"] + x[,"n_33"])*log(15 - 11*r_cm + 9*r_cm^2 - 4*r_cm^3) + (x[,"n_01"] + x[,"n_34"])*log(6 - 5*r_cm + r_cm^2 - 2*r_cm^3) + (x[,"n_03"] + x[,"n_32"])*log(9 + 5*r_cm - 3*r_cm^2 + 4*r_cm^3) + (x[,"n_12"] + x[,"n_23"])*log(31 + 2*r_cm - 14*r_cm^2 + 8*r_cm^3) + (x[,"n_11"] + x[,"n_24"])*log(39 - 16*r_cm - 14*r_cm^2 + 12*r_cm^3)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 2*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(5) + (n05 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n35)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n34)*(log(pmax(1e-6,r)) + log(8 - 7*r + 2*r^2)) + (n10 + n25)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n12 + n23)*log(30 - 4*r + 13*r^2 - 12*r^3) + (n11 + n24)*log(12 - 10*r + 13*r^2 - 6*r^3) + (n03 + n32)*log(11 - 12*r + 11*r^2 - 4*r^3) + (n15 + n20)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n04 + n31)*log(3 - r^2 - 2*r^3) + (n02 + n33)*log(6 + 2*r - r^2 + 4*r^3) + (n14 + n21)*log(9 + 2*r - 5*r^2 + 6*r^3) + (n13 + n22)*log(27 + 14*r - 23*r^2 + 12*r^3)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_cr <- 3*(x[,"n_05"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_35"])*log10(2) + (x[,"n_15"] + x[,"n_20"])*(3*log10(2) - log10(3)) - (x[,"n_10"] + x[,"n_25"])*(-3*log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_31"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_34"])*(-log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_22"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_12"] + x[,"n_23"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_14"] + x[,"n_21"])*(-log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_24"])*(-log10(2) + log10(19)) - (x[,"n_03"] + x[,"n_32"])*(-2*log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_33"])*(-2*log10(2) + log10(29)) + (x[,"n_05"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_34"])*(log10(pmax(1e-6,r_cr)) + log10(8 - 7*r_cr + 2*r_cr^2)) + (x[,"n_10"] + x[,"n_25"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_12"] + x[,"n_23"])*log10(30 - 4*r_cr + 13*r_cr^2 - 12*r_cr^3) + (x[,"n_11"] + x[,"n_24"])*log10(12 - 10*r_cr + 13*r_cr^2 - 6*r_cr^3) + (x[,"n_03"] + x[,"n_32"])*log10(11 - 12*r_cr + 11*r_cr^2 - 4*r_cr^3) + (x[,"n_15"] + x[,"n_20"])*log10(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_04"] + x[,"n_31"])*log10(3 - r_cr^2 - 2*r_cr^3) + (x[,"n_02"] + x[,"n_33"])*log10(6 + 2*r_cr - r_cr^2 + 4*r_cr^3) + (x[,"n_14"] + x[,"n_21"])*log10(9 + 2*r_cr - 5*r_cr^2 + 6*r_cr^3) + (x[,"n_13"] + x[,"n_22"])*log10(27 + 14*r_cr - 23*r_cr^2 + 12*r_cr^3)
logL_cr <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_05"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_34"])*(log(pmax(1e-6,r_cr)) + log(8 - 7*r_cr + 2*r_cr^2)) + (x[,"n_10"] + x[,"n_25"])*(log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_12"] + x[,"n_23"])*log(30 - 4*r_cr + 13*r_cr^2 - 12*r_cr^3) + (x[,"n_11"] + x[,"n_24"])*log(12 - 10*r_cr + 13*r_cr^2 - 6*r_cr^3) + (x[,"n_03"] + x[,"n_32"])*log(11 - 12*r_cr + 11*r_cr^2 - 4*r_cr^3) + (x[,"n_15"] + x[,"n_20"])*log(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_04"] + x[,"n_31"])*log(3 - r_cr^2 - 2*r_cr^3) + (x[,"n_02"] + x[,"n_33"])*log(6 + 2*r_cr - r_cr^2 + 4*r_cr^3) + (x[,"n_14"] + x[,"n_21"])*log(9 + 2*r_cr - 5*r_cr^2 + 6*r_cr^3) + (x[,"n_13"] + x[,"n_22"])*log(27 + 14*r_cr - 23*r_cr^2 + 12*r_cr^3)
logL_rc <- function(r,n10,n15,n20,n25,n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n21,n22,n23,n24,n30,n31,n32,n33,n34,n35) {
L <- (n10 + n15 + n20 + n25)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(4*log(2) + 2*log(5)) + (n05 + n30)*(log(3 - 2*r) + 2*log(pmax(1e-6,r))) + (n00 + n35)*(2*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n04 + n31)*(log(pmax(1e-6,r)) + log(12 - 5*r + 2*r^2)) + (n15 + n20)*(log(pmax(1e-6,r)) + log(2 - 3*r + 2*r^2)) + (n13 + n22)*log(57 + 16*r^2 - 12*r^3) + (n14 + n21)*log(18 - 2*r + 11*r^2 - 6*r^3) + (n02 + n33)*log(21 - 16*r + 8*r^2 - 4*r^3) + (n01 + n34)*log(9 - 8*r + r^2 - 2*r^3) + (n10 + n25)*log(1 - 2*r + 3*r^2 - 2*r^3) + (n03 + n32)*log(9 + 12*r - 4*r^2 + 4*r^3) + (n11 + n24)*log(21 - 2*r - 7*r^2 + 6*r^3) + (n12 + n23)*log(61 + 4*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rc <- function(n10,n15,n20,n25,n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n21,n22,n23,n24,n30,n31,n32,n33,n34,n35) {
optimize(logL_rc,c(0,0.5), n10,n15,n20,n25,n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n21,n22,n23,n24,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_10"],x[,"n_15"],x[,"n_20"],x[,"n_25"],x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rc <- (x[,"n_15"] + x[,"n_20"])*log10(2) + (x[,"n_10"] + x[,"n_25"])*log10(2) + (x[,"n_05"] + x[,"n_30"])*log10(2) + (x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_04"] + x[,"n_31"])*log10(5) - (x[,"n_01"] + x[,"n_34"])*log10(5) - (x[,"n_13"] + x[,"n_22"])*(-log10(2) + log10(7) + log10(17)) - (x[,"n_12"] + x[,"n_23"])*(-log10(2) + log10(7) + log10(17)) - (x[,"n_14"] + x[,"n_21"])*log10(19) - (x[,"n_11"] + x[,"n_24"])*log10(19) - (x[,"n_03"] + x[,"n_32"])*(-log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_33"])*(-log10(2) + log10(29)) + (x[,"n_05"] + x[,"n_30"])*(log10(3 - 2*r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_04"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(12 - 5*r_rc + 2*r_rc^2)) + (x[,"n_15"] + x[,"n_20"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 3*r_rc + 2*r_rc^2)) + (x[,"n_13"] + x[,"n_22"])*log10(57 + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_14"] + x[,"n_21"])*log10(18 - 2*r_rc + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_02"] + x[,"n_33"])*log10(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_01"] + x[,"n_34"])*log10(9 - 8*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_10"] + x[,"n_25"])*log10(1 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3) + (x[,"n_03"] + x[,"n_32"])*log10(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_11"] + x[,"n_24"])*log10(21 - 2*r_rc - 7*r_rc^2 + 6*r_rc^3) + (x[,"n_12"] + x[,"n_23"])*log10(61 + 4*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rc <- (x[,"n_10"] + x[,"n_15"] + x[,"n_20"] + x[,"n_25"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_30"])*(log(3 - 2*r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_04"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(12 - 5*r_rc + 2*r_rc^2)) + (x[,"n_15"] + x[,"n_20"])*(log(pmax(1e-6,r_rc)) + log(2 - 3*r_rc + 2*r_rc^2)) + (x[,"n_13"] + x[,"n_22"])*log(57 + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_14"] + x[,"n_21"])*log(18 - 2*r_rc + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_02"] + x[,"n_33"])*log(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_01"] + x[,"n_34"])*log(9 - 8*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_10"] + x[,"n_25"])*log(1 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3) + (x[,"n_03"] + x[,"n_32"])*log(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_11"] + x[,"n_24"])*log(21 - 2*r_rc - 7*r_rc^2 + 6*r_rc^3) + (x[,"n_12"] + x[,"n_23"])*log(61 + 4*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 4*n10 - 4*n11 - 2*n12 - 2*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 2*n22 - 2*n23 - 4*n24 - 4*n25 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(5) + (n05 + n30)*(log(3 - 2*r) + log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n35)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 + 2*r)) + (n15 + n20)*log(9 - 16*r + 22*r^2 - 12*r^3) + (n11 + n24)*log(51 + 4*r + 26*r^2 - 12*r^3) + (n03 + n32)*log(27 - 14*r + 12*r^2 - 4*r^3) + (n04 + n31)*log(9 - 3*r - r^2 - 2*r^3) + (n12 + n23)*log(14 + r + 2*r^2 - 2*r^3) + (n01 + n34)*log(3 + 11*r - 7*r^2 + 2*r^3) + (n13 + n22)*log(15 + r - 4*r^2 + 2*r^3) + (n02 + n33)*log(21 + 2*r + 4*r^3) + (n10 + n25)*log(3 + 8*r - 14*r^2 + 12*r^3) + (n14 + n21)*log(69 - 20*r - 10*r^2 + 12*r^3)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rm <- (x[,"n_05"] + x[,"n_30"])*log10(2) + (x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_15"] + x[,"n_20"])*log10(5) - (x[,"n_10"] + x[,"n_25"])*log10(5) - (x[,"n_03"] + x[,"n_32"])*(-log10(2) + 2*log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_33"])*(-log10(2) + 2*log10(3) + log10(5)) - (x[,"n_04"] + x[,"n_31"])*log10(7) - (x[,"n_01"] + x[,"n_34"])*log10(7) - (x[,"n_14"] + x[,"n_21"])*(log10(2) + log10(29)) - (x[,"n_11"] + x[,"n_24"])*(log10(2) + log10(29)) - (x[,"n_13"] + x[,"n_22"])*(-2*log10(2) + log10(59)) - (x[,"n_12"] + x[,"n_23"])*(-2*log10(2) + log10(59)) + (x[,"n_05"] + x[,"n_30"])*(log10(3 - 2*r_rm) + log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm)) + log10(1 + 2*r_rm)) + (x[,"n_15"] + x[,"n_20"])*log10(9 - 16*r_rm + 22*r_rm^2 - 12*r_rm^3) + (x[,"n_11"] + x[,"n_24"])*log10(51 + 4*r_rm + 26*r_rm^2 - 12*r_rm^3) + (x[,"n_03"] + x[,"n_32"])*log10(27 - 14*r_rm + 12*r_rm^2 - 4*r_rm^3) + (x[,"n_04"] + x[,"n_31"])*log10(9 - 3*r_rm - r_rm^2 - 2*r_rm^3) + (x[,"n_12"] + x[,"n_23"])*log10(14 + r_rm + 2*r_rm^2 - 2*r_rm^3) + (x[,"n_01"] + x[,"n_34"])*log10(3 + 11*r_rm - 7*r_rm^2 + 2*r_rm^3) + (x[,"n_13"] + x[,"n_22"])*log10(15 + r_rm - 4*r_rm^2 + 2*r_rm^3) + (x[,"n_02"] + x[,"n_33"])*log10(21 + 2*r_rm + 4*r_rm^3) + (x[,"n_10"] + x[,"n_25"])*log10(3 + 8*r_rm - 14*r_rm^2 + 12*r_rm^3) + (x[,"n_14"] + x[,"n_21"])*log10(69 - 20*r_rm - 10*r_rm^2 + 12*r_rm^3)
logL_rm <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_05"] + x[,"n_30"])*(log(3 - 2*r_rm) + log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm)) + log(1 + 2*r_rm)) + (x[,"n_15"] + x[,"n_20"])*log(9 - 16*r_rm + 22*r_rm^2 - 12*r_rm^3) + (x[,"n_11"] + x[,"n_24"])*log(51 + 4*r_rm + 26*r_rm^2 - 12*r_rm^3) + (x[,"n_03"] + x[,"n_32"])*log(27 - 14*r_rm + 12*r_rm^2 - 4*r_rm^3) + (x[,"n_04"] + x[,"n_31"])*log(9 - 3*r_rm - r_rm^2 - 2*r_rm^3) + (x[,"n_12"] + x[,"n_23"])*log(14 + r_rm + 2*r_rm^2 - 2*r_rm^3) + (x[,"n_01"] + x[,"n_34"])*log(3 + 11*r_rm - 7*r_rm^2 + 2*r_rm^3) + (x[,"n_13"] + x[,"n_22"])*log(15 + r_rm - 4*r_rm^2 + 2*r_rm^3) + (x[,"n_02"] + x[,"n_33"])*log(21 + 2*r_rm + 4*r_rm^3) + (x[,"n_10"] + x[,"n_25"])*log(3 + 8*r_rm - 14*r_rm^2 + 12*r_rm^3) + (x[,"n_14"] + x[,"n_21"])*log(69 - 20*r_rm - 10*r_rm^2 + 12*r_rm^3)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(4*log(2) + 2*log(5)) + (n05 + n30)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r))) + (n00 + n35)*(2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n10 + n25)*(log(pmax(1e-6,r)) + log(2 + 5*r - 6*r^2)) + (n01 + n34)*(log(pmax(1e-6,r)) + log(4 + 13*r - 2*r^2)) + (n13 + n22)*log(75 - 44*r + 32*r^2 - 12*r^3) + (n14 + n21)*log(27 - 14*r - r^2 - 6*r^3) + (n02 + n33)*log(3 + 28*r - 8*r^2 - 4*r^3) + (n04 + n31)*log(15 - 24*r + 7*r^2 + 2*r^3) + (n03 + n32)*log(19 - 20*r^2 + 4*r^3) + (n11 + n24)*log(6 + 34*r - 19*r^2 + 6*r^3) + (n15 + n20)*log(1 + 6*r - 13*r^2 + 6*r^3) + (n12 + n23)*log(51 + 16*r - 4*r^2 + 12*r^3)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rr <- (x[,"n_05"] + x[,"n_30"])*log10(2) + (x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_15"] + x[,"n_20"])*(-log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_25"])*(-log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_31"])*log10(5) - (x[,"n_01"] + x[,"n_34"])*log10(5) - (x[,"n_13"] + x[,"n_22"])*(-log10(2) + log10(7) + log10(17)) - (x[,"n_12"] + x[,"n_23"])*(-log10(2) + log10(7) + log10(17)) - (x[,"n_14"] + x[,"n_21"])*log10(19) - (x[,"n_11"] + x[,"n_24"])*log10(19) - (x[,"n_03"] + x[,"n_32"])*(-log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_33"])*(-log10(2) + log10(29)) + (x[,"n_05"] + x[,"n_30"])*(log10(3 - 2*r_rr) + 2*log10(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_10"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(2 + 5*r_rr - 6*r_rr^2)) + (x[,"n_01"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(4 + 13*r_rr - 2*r_rr^2)) + (x[,"n_13"] + x[,"n_22"])*log10(75 - 44*r_rr + 32*r_rr^2 - 12*r_rr^3) + (x[,"n_14"] + x[,"n_21"])*log10(27 - 14*r_rr - r_rr^2 - 6*r_rr^3) + (x[,"n_02"] + x[,"n_33"])*log10(3 + 28*r_rr - 8*r_rr^2 - 4*r_rr^3) + (x[,"n_04"] + x[,"n_31"])*log10(15 - 24*r_rr + 7*r_rr^2 + 2*r_rr^3) + (x[,"n_03"] + x[,"n_32"])*log10(19 - 20*r_rr^2 + 4*r_rr^3) + (x[,"n_11"] + x[,"n_24"])*log10(6 + 34*r_rr - 19*r_rr^2 + 6*r_rr^3) + (x[,"n_15"] + x[,"n_20"])*log10(1 + 6*r_rr - 13*r_rr^2 + 6*r_rr^3) + (x[,"n_12"] + x[,"n_23"])*log10(51 + 16*r_rr - 4*r_rr^2 + 12*r_rr^3)
logL_rr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_30"])*(log(3 - 2*r_rr) + 2*log(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_35"])*(2*log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_10"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(2 + 5*r_rr - 6*r_rr^2)) + (x[,"n_01"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(4 + 13*r_rr - 2*r_rr^2)) + (x[,"n_13"] + x[,"n_22"])*log(75 - 44*r_rr + 32*r_rr^2 - 12*r_rr^3) + (x[,"n_14"] + x[,"n_21"])*log(27 - 14*r_rr - r_rr^2 - 6*r_rr^3) + (x[,"n_02"] + x[,"n_33"])*log(3 + 28*r_rr - 8*r_rr^2 - 4*r_rr^3) + (x[,"n_04"] + x[,"n_31"])*log(15 - 24*r_rr + 7*r_rr^2 + 2*r_rr^3) + (x[,"n_03"] + x[,"n_32"])*log(19 - 20*r_rr^2 + 4*r_rr^3) + (x[,"n_11"] + x[,"n_24"])*log(6 + 34*r_rr - 19*r_rr^2 + 6*r_rr^3) + (x[,"n_15"] + x[,"n_20"])*log(1 + 6*r_rr - 13*r_rr^2 + 6*r_rr^3) + (x[,"n_12"] + x[,"n_23"])*log(51 + 16*r_rr - 4*r_rr^2 + 12*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_2.4 <- function(x){
logL_cc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-n01 - 2*n02 - 2*n03 - 2*n04 - n05 - n11 - 2*n12 - 2*n13 - 2*n14 - n15 - n21 - 2*n22 - 2*n23 - 2*n24 - n25 - n31 - 2*n32 - 2*n33 - 2*n34 - n35)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*(log(3) + 2*log(5)) + (n05 + n31)*(2*log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n33)*log(9 + 16*r - 16*r^2) + (n04 + n32)*(log(pmax(1e-6,r)) + log(12 + 7*r - 4*r^2)) + (n15 + n21)*(log(pmax(1e-6,r)) + log(5 + r - 3*r^2)) + (n01 + n35)*(log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n13 + n23)*log(57 - 16*r + 16*r^2) + (n12 + n24)*log(42 - 32*r + 17*r^2 - 12*r^3) + (n11 + n25)*log(3 + 2*r - 8*r^2 + 3*r^3) + (n02 + n34)*log(15 - 14*r - 5*r^2 + 4*r^3) + (n14 + n22)*log(15 + 34*r - 19*r^2 + 12*r^3)
return(L)}
interlogL_cc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_cc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_cc <- -((x[,"n_05"] + x[,"n_31"])*(-3*log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_35"])*(-3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_33"])*log10(13) - (x[,"n_15"] + x[,"n_21"])*(-3*log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_25"])*(-3*log10(2) + log10(19)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(29)) - (x[,"n_13"] + x[,"n_23"])*log10(53) + (x[,"n_05"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_03"] + x[,"n_33"])*log10(9 + 16*r_cc - 16*r_cc^2) + (x[,"n_04"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(12 + 7*r_cc - 4*r_cc^2)) + (x[,"n_15"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(5 + r_cc - 3*r_cc^2)) + (x[,"n_01"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_cc)) + log10(3 - 4*r_cc + r_cc^2)) + (x[,"n_13"] + x[,"n_23"])*log10(57 - 16*r_cc + 16*r_cc^2) + (x[,"n_12"] + x[,"n_24"])*log10(42 - 32*r_cc + 17*r_cc^2 - 12*r_cc^3) + (x[,"n_11"] + x[,"n_25"])*log10(3 + 2*r_cc - 8*r_cc^2 + 3*r_cc^3) + (x[,"n_02"] + x[,"n_34"])*log10(15 - 14*r_cc - 5*r_cc^2 + 4*r_cc^3) + (x[,"n_14"] + x[,"n_22"])*log10(15 + 34*r_cc - 19*r_cc^2 + 12*r_cc^3)
logL_cc <- (-x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_03"] + x[,"n_33"])*log(9 + 16*r_cc - 16*r_cc^2) + (x[,"n_04"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(12 + 7*r_cc - 4*r_cc^2)) + (x[,"n_15"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(5 + r_cc - 3*r_cc^2)) + (x[,"n_01"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_cc)) + log(3 - 4*r_cc + r_cc^2)) + (x[,"n_13"] + x[,"n_23"])*log(57 - 16*r_cc + 16*r_cc^2) + (x[,"n_12"] + x[,"n_24"])*log(42 - 32*r_cc + 17*r_cc^2 - 12*r_cc^3) + (x[,"n_11"] + x[,"n_25"])*log(3 + 2*r_cc - 8*r_cc^2 + 3*r_cc^3) + (x[,"n_02"] + x[,"n_34"])*log(15 - 14*r_cc - 5*r_cc^2 + 4*r_cc^3) + (x[,"n_14"] + x[,"n_22"])*log(15 + 34*r_cc - 19*r_cc^2 + 12*r_cc^3)
logL_cm <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-2*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 2*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35)*log(2) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*log(5) + (n01 + n35)*(log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n13 + n23)*log(33 + 8*r - 8*r^2) + (n05 + n31)*(log(pmax(1e-6,r)) + log(1 - r^2)) + (n03 + n33)*log(11 - 8*r + 8*r^2) + (n14 + n22)*log(18 - 4*r + 19*r^2 - 12*r^3) + (n02 + n34)*log(6 - 4*r + 5*r^2 - 4*r^3) + (n11 + n25)*log(3 - 5*r + 6*r^2 - 3*r^3) + (n15 + n21)*log(1 + 2*r - 3*r^2 + 3*r^3) + (n04 + n32)*log(3 + 6*r - 7*r^2 + 4*r^3) + (n12 + n24)*log(21 + 2*r - 17*r^2 + 12*r^3)
return(L)}
interlogL_cm <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_cm,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_cm <- -2*(x[,"n_03"] + x[,"n_33"])*log10(3) - (x[,"n_05"] + x[,"n_31"])*(-3*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_35"])*(-3*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_23"])*(log10(5) + log10(7)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_15"] + x[,"n_21"])*(-3*log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_25"])*(-3*log10(2) + log10(13)) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(19)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(19)) + (x[,"n_01"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_cm)) + log10(2 - r_cm) + log10(pmax(1e-6,r_cm))) + (x[,"n_13"] + x[,"n_23"])*log10(33 + 8*r_cm - 8*r_cm^2) + (x[,"n_05"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(1 - r_cm^2)) + (x[,"n_03"] + x[,"n_33"])*log10(11 - 8*r_cm + 8*r_cm^2) + (x[,"n_14"] + x[,"n_22"])*log10(18 - 4*r_cm + 19*r_cm^2 - 12*r_cm^3) + (x[,"n_02"] + x[,"n_34"])*log10(6 - 4*r_cm + 5*r_cm^2 - 4*r_cm^3) + (x[,"n_11"] + x[,"n_25"])*log10(3 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3) + (x[,"n_15"] + x[,"n_21"])*log10(1 + 2*r_cm - 3*r_cm^2 + 3*r_cm^3) + (x[,"n_04"] + x[,"n_32"])*log10(3 + 6*r_cm - 7*r_cm^2 + 4*r_cm^3) + (x[,"n_12"] + x[,"n_24"])*log10(21 + 2*r_cm - 17*r_cm^2 + 12*r_cm^3)
logL_cm <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_01"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_cm)) + log(2 - r_cm) + log(pmax(1e-6,r_cm))) + (x[,"n_13"] + x[,"n_23"])*log(33 + 8*r_cm - 8*r_cm^2) + (x[,"n_05"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(1 - r_cm^2)) + (x[,"n_03"] + x[,"n_33"])*log(11 - 8*r_cm + 8*r_cm^2) + (x[,"n_14"] + x[,"n_22"])*log(18 - 4*r_cm + 19*r_cm^2 - 12*r_cm^3) + (x[,"n_02"] + x[,"n_34"])*log(6 - 4*r_cm + 5*r_cm^2 - 4*r_cm^3) + (x[,"n_11"] + x[,"n_25"])*log(3 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3) + (x[,"n_15"] + x[,"n_21"])*log(1 + 2*r_cm - 3*r_cm^2 + 3*r_cm^3) + (x[,"n_04"] + x[,"n_32"])*log(3 + 6*r_cm - 7*r_cm^2 + 4*r_cm^3) + (x[,"n_12"] + x[,"n_24"])*log(21 + 2*r_cm - 17*r_cm^2 + 12*r_cm^3)
logL_cr <- function(r,n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n01,n03,n05,n11,n15,n21,n25,n31,n33,n35) {
L <- (-n02 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n34)*log(2) + 2*(n13 + n23)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*log(5) + (n05 + n31)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n35)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n25)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n02 + n34)*(log(pmax(1e-6,r)) + log(4 - 5*r + 4*r^2)) + (n12 + n24)*log(6 - 8*r + 17*r^2 - 12*r^3) + (n04 + n32)*log(3 - 6*r + 7*r^2 - 4*r^3) + (n15 + n21)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n14 + n22)*log(3 + 10*r - 19*r^2 + 12*r^3)
return(L)}
interlogL_cr <- function(n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n01,n03,n05,n11,n15,n21,n25,n31,n33,n35) {
optimize(logL_cr,c(0,0.5), n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n01,n03,n05,n11,n15,n21,n25,n31,n33,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_02"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_34"],x[,"n_01"],x[,"n_03"],x[,"n_05"],x[,"n_11"],x[,"n_15"],x[,"n_21"],x[,"n_25"],x[,"n_31"],x[,"n_33"],x[,"n_35"])
LOD_cr <- 3*(x[,"n_05"] + x[,"n_31"])*log10(2) + 3*(x[,"n_01"] + x[,"n_35"])*log10(2) + (x[,"n_15"] + x[,"n_21"])*(3*log10(2) - log10(3)) - (x[,"n_11"] + x[,"n_25"])*(-3*log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + log10(19)) + (x[,"n_05"] + x[,"n_31"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_11"] + x[,"n_25"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_02"] + x[,"n_34"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_12"] + x[,"n_24"])*log10(6 - 8*r_cr + 17*r_cr^2 - 12*r_cr^3) + (x[,"n_04"] + x[,"n_32"])*log10(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_15"] + x[,"n_21"])*log10(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_14"] + x[,"n_22"])*log10(3 + 10*r_cr - 19*r_cr^2 + 12*r_cr^3)
logL_cr <- (-x[,"n_02"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_34"])*log(2) + 2*(x[,"n_13"] + x[,"n_23"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_05"] + x[,"n_31"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_11"] + x[,"n_25"])*(log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_02"] + x[,"n_34"])*(log(pmax(1e-6,r_cr)) + log(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_12"] + x[,"n_24"])*log(6 - 8*r_cr + 17*r_cr^2 - 12*r_cr^3) + (x[,"n_04"] + x[,"n_32"])*log(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_15"] + x[,"n_21"])*log(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_14"] + x[,"n_22"])*log(3 + 10*r_cr - 19*r_cr^2 + 12*r_cr^3)
logL_rc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*(3*log(2) + log(3) + 2*log(5)) + (n05 + n31)*(log(3 - 2*r) + log(pmax(1e-6,r)) + log(2 + r)) + (n13 + n23)*log(105 + 4*r - 4*r^2) + (n01 + n35)*(log(1 + 2*r) + log(3 - 4*r + r^2)) + (n03 + n33)*log(27 - 4*r + 4*r^2) + (n14 + n22)*log(57 - 4*r + 16*r^2 - 12*r^3) + (n11 + n25)*log(12 - 10*r + 13*r^2 - 6*r^3) + (n02 + n34)*log(21 - 16*r + 8*r^2 - 4*r^3) + (n04 + n32)*log(9 + 12*r - 4*r^2 + 4*r^3) + (n15 + n21)*log(9 + 2*r - 5*r^2 + 6*r^3) + (n12 + n24)*log(57 + 8*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rc <- -((x[,"n_05"] + x[,"n_31"])*(-log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_35"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_33"])*(log10(2) + log10(13)) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + log10(19)) - (x[,"n_14"] + x[,"n_22"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_12"] + x[,"n_24"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_04"] + x[,"n_32"])*(-log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_34"])*(-log10(2) + log10(29)) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(53)) + (x[,"n_05"] + x[,"n_31"])*(log10(3 - 2*r_rc) + log10(pmax(1e-6,r_rc)) + log10(2 + r_rc)) + (x[,"n_13"] + x[,"n_23"])*log10(105 + 4*r_rc - 4*r_rc^2) + (x[,"n_01"] + x[,"n_35"])*(log10(1 + 2*r_rc) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_03"] + x[,"n_33"])*log10(27 - 4*r_rc + 4*r_rc^2) + (x[,"n_14"] + x[,"n_22"])*log10(57 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_11"] + x[,"n_25"])*log10(12 - 10*r_rc + 13*r_rc^2 - 6*r_rc^3) + (x[,"n_02"] + x[,"n_34"])*log10(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_04"] + x[,"n_32"])*log10(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_15"] + x[,"n_21"])*log10(9 + 2*r_rc - 5*r_rc^2 + 6*r_rc^3) + (x[,"n_12"] + x[,"n_24"])*log10(57 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rc <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(3*log(2) + log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_31"])*(log(3 - 2*r_rc) + log(pmax(1e-6,r_rc)) + log(2 + r_rc)) + (x[,"n_13"] + x[,"n_23"])*log(105 + 4*r_rc - 4*r_rc^2) + (x[,"n_01"] + x[,"n_35"])*(log(1 + 2*r_rc) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_03"] + x[,"n_33"])*log(27 - 4*r_rc + 4*r_rc^2) + (x[,"n_14"] + x[,"n_22"])*log(57 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_11"] + x[,"n_25"])*log(12 - 10*r_rc + 13*r_rc^2 - 6*r_rc^3) + (x[,"n_02"] + x[,"n_34"])*log(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_04"] + x[,"n_32"])*log(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_15"] + x[,"n_21"])*log(9 + 2*r_rc - 5*r_rc^2 + 6*r_rc^3) + (x[,"n_12"] + x[,"n_24"])*log(57 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rm <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*(4*log(2) + 2*log(5)) + (n01 + n35)*(log(2 - r) + log(pmax(1e-6,r)) + log(1 + 2*r)) + (n13 + n23)*log(69 + 4*r - 4*r^2) + (n05 + n31)*(log(3 - 2*r) + log(1 - r^2)) + (n03 + n33)*log(19 - 4*r + 4*r^2) + (n12 + n24)*log(33 + 4*r + 20*r^2 - 12*r^3) + (n15 + n21)*log(10 - 10*r + 9*r^2 - 6*r^3) + (n04 + n32)*log(15 - 12*r + 4*r^2 - 4*r^3) + (n02 + n34)*log(3 + 16*r - 8*r^2 + 4*r^3) + (n11 + n25)*log(3 + 10*r - 9*r^2 + 6*r^3) + (n14 + n22)*log(45 - 8*r - 16*r^2 + 12*r^3)
return(L)}
interlogL_rm <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_rm,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rm <- -((x[,"n_05"] + x[,"n_31"])*(-log10(2) + log10(3))) - (x[,"n_01"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_33"])*(log10(2) + 2*log10(3)) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(5) + log10(7)) - (x[,"n_14"] + x[,"n_22"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_12"] + x[,"n_24"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + log10(13)) - (x[,"n_04"] + x[,"n_32"])*(-log10(2) + log10(19)) - (x[,"n_02"] + x[,"n_34"])*(-log10(2) + log10(19)) + (x[,"n_01"] + x[,"n_35"])*(log10(2 - r_rm) + log10(pmax(1e-6,r_rm)) + log10(1 + 2*r_rm)) + (x[,"n_13"] + x[,"n_23"])*log10(69 + 4*r_rm - 4*r_rm^2) + (x[,"n_05"] + x[,"n_31"])*(log10(3 - 2*r_rm) + log10(1 - r_rm^2)) + (x[,"n_03"] + x[,"n_33"])*log10(19 - 4*r_rm + 4*r_rm^2) + (x[,"n_12"] + x[,"n_24"])*log10(33 + 4*r_rm + 20*r_rm^2 - 12*r_rm^3) + (x[,"n_15"] + x[,"n_21"])*log10(10 - 10*r_rm + 9*r_rm^2 - 6*r_rm^3) + (x[,"n_04"] + x[,"n_32"])*log10(15 - 12*r_rm + 4*r_rm^2 - 4*r_rm^3) + (x[,"n_02"] + x[,"n_34"])*log10(3 + 16*r_rm - 8*r_rm^2 + 4*r_rm^3) + (x[,"n_11"] + x[,"n_25"])*log10(3 + 10*r_rm - 9*r_rm^2 + 6*r_rm^3) + (x[,"n_14"] + x[,"n_22"])*log10(45 - 8*r_rm - 16*r_rm^2 + 12*r_rm^3)
logL_rm <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(4*log(2) + 2*log(5)) + (x[,"n_01"] + x[,"n_35"])*(log(2 - r_rm) + log(pmax(1e-6,r_rm)) + log(1 + 2*r_rm)) + (x[,"n_13"] + x[,"n_23"])*log(69 + 4*r_rm - 4*r_rm^2) + (x[,"n_05"] + x[,"n_31"])*(log(3 - 2*r_rm) + log(1 - r_rm^2)) + (x[,"n_03"] + x[,"n_33"])*log(19 - 4*r_rm + 4*r_rm^2) + (x[,"n_12"] + x[,"n_24"])*log(33 + 4*r_rm + 20*r_rm^2 - 12*r_rm^3) + (x[,"n_15"] + x[,"n_21"])*log(10 - 10*r_rm + 9*r_rm^2 - 6*r_rm^3) + (x[,"n_04"] + x[,"n_32"])*log(15 - 12*r_rm + 4*r_rm^2 - 4*r_rm^3) + (x[,"n_02"] + x[,"n_34"])*log(3 + 16*r_rm - 8*r_rm^2 + 4*r_rm^3) + (x[,"n_11"] + x[,"n_25"])*log(3 + 10*r_rm - 9*r_rm^2 + 6*r_rm^3) + (x[,"n_14"] + x[,"n_22"])*log(45 - 8*r_rm - 16*r_rm^2 + 12*r_rm^3)
logL_rr <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-2*n01 - n02 - 2*n03 - n04 - 2*n05 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - 2*n31 - n32 - 2*n33 - n34 - 2*n35)*log(2) + (n13 + n23)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*log(5) + (n05 + n31)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r))) + (n01 + n35)*(2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n03 + n33)*log(1 + 12*r - 12*r^2) + (n11 + n25)*(log(pmax(1e-6,r)) + log(2 + 5*r - 6*r^2)) + (n13 + n23)*log(7 - 4*r + 4*r^2) + (n14 + n22)*log(15 - 16*r + 16*r^2 - 12*r^3) + (n02 + n34)*log(r + 4*r^2 - 2*r^3) + (n04 + n32)*log(3 - 3*r - 2*r^2 + 2*r^3) + (n15 + n21)*log(1 + 6*r - 13*r^2 + 6*r^3) + (n12 + n24)*log(3 + 20*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rr <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_rr,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rr <- (x[,"n_05"] + x[,"n_31"])*log10(2) - 2*(x[,"n_03"] + x[,"n_33"])*log10(2) + (x[,"n_01"] + x[,"n_35"])*log10(2) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_22"])*(-log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_24"])*(-log10(2) + log10(19)) + (x[,"n_05"] + x[,"n_31"])*(log10(3 - 2*r_rr) + 2*log10(pmax(1e-6,1 - r_rr))) + (x[,"n_01"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_03"] + x[,"n_33"])*log10(1 + 12*r_rr - 12*r_rr^2) + (x[,"n_11"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(2 + 5*r_rr - 6*r_rr^2)) + (x[,"n_13"] + x[,"n_23"])*log10(7 - 4*r_rr + 4*r_rr^2) + (x[,"n_14"] + x[,"n_22"])*log10(15 - 16*r_rr + 16*r_rr^2 - 12*r_rr^3) + (x[,"n_02"] + x[,"n_34"])*log10(r_rr + 4*r_rr^2 - 2*r_rr^3) + (x[,"n_04"] + x[,"n_32"])*log10(3 - 3*r_rr - 2*r_rr^2 + 2*r_rr^3) + (x[,"n_15"] + x[,"n_21"])*log10(1 + 6*r_rr - 13*r_rr^2 + 6*r_rr^3) + (x[,"n_12"] + x[,"n_24"])*log10(3 + 20*r_rr - 20*r_rr^2 + 12*r_rr^3)
logL_rr <- (-2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"] - 2*x[,"n_35"])*log(2) + (x[,"n_13"] + x[,"n_23"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_05"] + x[,"n_31"])*(log(3 - 2*r_rr) + 2*log(pmax(1e-6,1 - r_rr))) + (x[,"n_01"] + x[,"n_35"])*(2*log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_03"] + x[,"n_33"])*log(1 + 12*r_rr - 12*r_rr^2) + (x[,"n_11"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(2 + 5*r_rr - 6*r_rr^2)) + (x[,"n_13"] + x[,"n_23"])*log(7 - 4*r_rr + 4*r_rr^2) + (x[,"n_14"] + x[,"n_22"])*log(15 - 16*r_rr + 16*r_rr^2 - 12*r_rr^3) + (x[,"n_02"] + x[,"n_34"])*log(r_rr + 4*r_rr^2 - 2*r_rr^3) + (x[,"n_04"] + x[,"n_32"])*log(3 - 3*r_rr - 2*r_rr^2 + 2*r_rr^3) + (x[,"n_15"] + x[,"n_21"])*log(1 + 6*r_rr - 13*r_rr^2 + 6*r_rr^3) + (x[,"n_12"] + x[,"n_24"])*log(3 + 20*r_rr - 20*r_rr^2 + 12*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_3.0 <- function(x){
logL_c <- function(r,n01,n02,n11,n12,n21,n22,n31,n32,n00,n03,n10,n13,n20,n23,n30,n33) {
L <- (n01 + n02 + n11 + n12 + n21 + n22 + n31 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(2*log(2) + 2*log(5)) + (n10 + n23)*log(3 - 2*r) + (n11 + n22)*log(7 - 2*r) + (n00 + n33)*log(pmax(1e-6,1 - r)) + (n01 + n32)*log(2 - r) + (n03 + n30)*log(pmax(1e-6,r)) + (n02 + n31)*log(1 + r) + (n13 + n20)*log(1 + 2*r) + (n12 + n21)*log(5 + 2*r)
return(L)}
interlogL_c <- function(n01,n02,n11,n12,n21,n22,n31,n32,n00,n03,n10,n13,n20,n23,n30,n33) {
optimize(logL_c,c(0,0.5), n01,n02,n11,n12,n21,n22,n31,n32,n00,n03,n10,n13,n20,n23,n30,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_11"],x[,"n_12"],x[,"n_21"],x[,"n_22"],x[,"n_31"],x[,"n_32"],x[,"n_00"],x[,"n_03"],x[,"n_10"],x[,"n_13"],x[,"n_20"],x[,"n_23"],x[,"n_30"],x[,"n_33"])
LOD_c <- -((x[,"n_13"] + x[,"n_20"])*log10(2)) - (x[,"n_10"] + x[,"n_23"])*log10(2) + (x[,"n_03"] + x[,"n_30"])*log10(2) + (x[,"n_00"] + x[,"n_33"])*log10(2) - (x[,"n_02"] + x[,"n_31"])*(-log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_32"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_21"])*(log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_22"])*(log10(2) + log10(3)) + (x[,"n_10"] + x[,"n_23"])*log10(3 - 2*r_c) + (x[,"n_11"] + x[,"n_22"])*log10(7 - 2*r_c) + (x[,"n_00"] + x[,"n_33"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_32"])*log10(2 - r_c) + (x[,"n_03"] + x[,"n_30"])*log10(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_31"])*log10(1 + r_c) + (x[,"n_13"] + x[,"n_20"])*log10(1 + 2*r_c) + (x[,"n_12"] + x[,"n_21"])*log10(5 + 2*r_c)
logL_c <- (x[,"n_01"] + x[,"n_02"] + x[,"n_11"] + x[,"n_12"] + x[,"n_21"] + x[,"n_22"] + x[,"n_31"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(2*log(2) + 2*log(5)) + (x[,"n_10"] + x[,"n_23"])*log(3 - 2*r_c) + (x[,"n_11"] + x[,"n_22"])*log(7 - 2*r_c) + (x[,"n_00"] + x[,"n_33"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_32"])*log(2 - r_c) + (x[,"n_03"] + x[,"n_30"])*log(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_31"])*log(1 + r_c) + (x[,"n_13"] + x[,"n_20"])*log(1 + 2*r_c) + (x[,"n_12"] + x[,"n_21"])*log(5 + 2*r_c)
logL_r <- function(r,n01,n02,n11,n12,n21,n22,n31,n32,n00,n03,n10,n13,n20,n23,n30,n33) {
L <- (n01 + n02 + n11 + n12 + n21 + n22 + n31 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(2*log(2) + 2*log(5)) + (n13 + n20)*log(3 - 2*r) + (n12 + n21)*log(7 - 2*r) + (n03 + n30)*log(pmax(1e-6,1 - r)) + (n02 + n31)*log(2 - r) + (n00 + n33)*log(pmax(1e-6,r)) + (n01 + n32)*log(1 + r) + (n10 + n23)*log(1 + 2*r) + (n11 + n22)*log(5 + 2*r)
return(L)}
interlogL_r <- function(n01,n02,n11,n12,n21,n22,n31,n32,n00,n03,n10,n13,n20,n23,n30,n33) {
optimize(logL_r,c(0,0.5), n01,n02,n11,n12,n21,n22,n31,n32,n00,n03,n10,n13,n20,n23,n30,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_11"],x[,"n_12"],x[,"n_21"],x[,"n_22"],x[,"n_31"],x[,"n_32"],x[,"n_00"],x[,"n_03"],x[,"n_10"],x[,"n_13"],x[,"n_20"],x[,"n_23"],x[,"n_30"],x[,"n_33"])
LOD_r <- -((x[,"n_13"] + x[,"n_20"])*log10(2)) - (x[,"n_10"] + x[,"n_23"])*log10(2) + (x[,"n_03"] + x[,"n_30"])*log10(2) + (x[,"n_00"] + x[,"n_33"])*log10(2) - (x[,"n_02"] + x[,"n_31"])*(-log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_32"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_21"])*(log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_22"])*(log10(2) + log10(3)) + (x[,"n_13"] + x[,"n_20"])*log10(3 - 2*r_r) + (x[,"n_12"] + x[,"n_21"])*log10(7 - 2*r_r) + (x[,"n_03"] + x[,"n_30"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_02"] + x[,"n_31"])*log10(2 - r_r) + (x[,"n_00"] + x[,"n_33"])*log10(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_32"])*log10(1 + r_r) + (x[,"n_10"] + x[,"n_23"])*log10(1 + 2*r_r) + (x[,"n_11"] + x[,"n_22"])*log10(5 + 2*r_r)
logL_r <- (x[,"n_01"] + x[,"n_02"] + x[,"n_11"] + x[,"n_12"] + x[,"n_21"] + x[,"n_22"] + x[,"n_31"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(2*log(2) + 2*log(5)) + (x[,"n_13"] + x[,"n_20"])*log(3 - 2*r_r) + (x[,"n_12"] + x[,"n_21"])*log(7 - 2*r_r) + (x[,"n_03"] + x[,"n_30"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_02"] + x[,"n_31"])*log(2 - r_r) + (x[,"n_00"] + x[,"n_33"])*log(pmax(1e-6,r_r)) + (x[,"n_01"] + x[,"n_32"])*log(1 + r_r) + (x[,"n_10"] + x[,"n_23"])*log(1 + 2*r_r) + (x[,"n_11"] + x[,"n_22"])*log(5 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_3.1 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-2*n00 - n01 - 2*n02 - n03 - 2*n04 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n30 - n31 - 2*n32 - n33 - 2*n34)*log(2) + (n02 + n12 + n22 + n32)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*log(5) + 2*(n00 + n34)*log(pmax(1e-6,1 - r)) + 2*(n04 + n30)*log(pmax(1e-6,r)) + (n14 + n20)*(log(5 - 2*r) + log(pmax(1e-6,r))) + (n03 + n31)*(log(pmax(1e-6,r)) + log(2 + r)) + (n11 + n23)*log(27 - 12*r - 4*r^2) + (n13 + n21)*log(11 + 20*r - 4*r^2) + (n10 + n24)*log(3 - r - 2*r^2) + (n02 + n32)*log(1 + 2*r - 2*r^2) + (n01 + n33)*log(3 - 4*r + r^2) + (n12 + n22)*log(13 - 4*r + 4*r^2)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cc <- -((x[,"n_14"] + x[,"n_20"])*log10(2)) - (x[,"n_10"] + x[,"n_24"])*log10(2) + 2*(x[,"n_04"] + x[,"n_30"])*log10(2) + 2*(x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_02"] + x[,"n_32"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_22"])*(2*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_31"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_33"])*(-2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_21"])*(2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_23"])*(2*log10(2) + log10(5)) + 2*(x[,"n_00"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_30"])*log10(pmax(1e-6,r_cc)) + (x[,"n_14"] + x[,"n_20"])*(log10(5 - 2*r_cc) + log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_31"])*(log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_11"] + x[,"n_23"])*log10(27 - 12*r_cc - 4*r_cc^2) + (x[,"n_13"] + x[,"n_21"])*log10(11 + 20*r_cc - 4*r_cc^2) + (x[,"n_10"] + x[,"n_24"])*log10(3 - r_cc - 2*r_cc^2) + (x[,"n_02"] + x[,"n_32"])*log10(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_01"] + x[,"n_33"])*log10(3 - 4*r_cc + r_cc^2) + (x[,"n_12"] + x[,"n_22"])*log10(13 - 4*r_cc + 4*r_cc^2)
logL_cc <- (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_30"] - x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_22"] + x[,"n_32"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + 2*(x[,"n_00"] + x[,"n_34"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_30"])*log(pmax(1e-6,r_cc)) + (x[,"n_14"] + x[,"n_20"])*(log(5 - 2*r_cc) + log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_31"])*(log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_11"] + x[,"n_23"])*log(27 - 12*r_cc - 4*r_cc^2) + (x[,"n_13"] + x[,"n_21"])*log(11 + 20*r_cc - 4*r_cc^2) + (x[,"n_10"] + x[,"n_24"])*log(3 - r_cc - 2*r_cc^2) + (x[,"n_02"] + x[,"n_32"])*log(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_01"] + x[,"n_33"])*log(3 - 4*r_cc + r_cc^2) + (x[,"n_12"] + x[,"n_22"])*log(13 - 4*r_cc + 4*r_cc^2)
logL_cr <- function(r,n02,n12,n22,n32,n00,n01,n03,n04,n10,n11,n13,n14,n20,n21,n23,n24,n30,n31,n33,n34) {
L <- (n02 + n12 + n22 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(4*log(2) + 2*log(5)) + (n04 + n30)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n00 + n34)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n01 + n03 + n31 + n33)*log(9 + 4*r - 4*r^2) + (n12 + n22)*log(23 + 4*r - 4*r^2) + (n10 + n24)*log(6 - 5*r + 2*r^2) + (n14 + n20)*log(3 + r + 2*r^2) + (n11 + n23)*log(45 - 12*r + 4*r^2) + (n02 + n32)*log(7 - 4*r + 4*r^2) + (n13 + n21)*log(37 + 4*r + 4*r^2)
return(L)}
interlogL_cr <- function(n02,n12,n22,n32,n00,n01,n03,n04,n10,n11,n13,n14,n20,n21,n23,n24,n30,n31,n33,n34) {
optimize(logL_cr,c(0,0.5), n02,n12,n22,n32,n00,n01,n03,n04,n10,n11,n13,n14,n20,n21,n23,n24,n30,n31,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_02"],x[,"n_12"],x[,"n_22"],x[,"n_32"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"])
LOD_cr <- -2*(x[,"n_14"] + x[,"n_20"])*log10(2) - 2*(x[,"n_10"] + x[,"n_24"])*log10(2) - (x[,"n_02"] + x[,"n_32"])*(log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_22"])*(3*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_03"] + x[,"n_31"] + x[,"n_33"])*(log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_21"])*(3*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_23"])*(3*log10(2) + log10(5)) + (x[,"n_04"] + x[,"n_30"])*(log10(3 - 2*r_cr) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cr)) + log10(1 + 2*r_cr)) + (x[,"n_01"] + x[,"n_03"] + x[,"n_31"] + x[,"n_33"])*log10(9 + 4*r_cr - 4*r_cr^2) + (x[,"n_12"] + x[,"n_22"])*log10(23 + 4*r_cr - 4*r_cr^2) + (x[,"n_10"] + x[,"n_24"])*log10(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_14"] + x[,"n_20"])*log10(3 + r_cr + 2*r_cr^2) + (x[,"n_11"] + x[,"n_23"])*log10(45 - 12*r_cr + 4*r_cr^2) + (x[,"n_02"] + x[,"n_32"])*log10(7 - 4*r_cr + 4*r_cr^2) + (x[,"n_13"] + x[,"n_21"])*log10(37 + 4*r_cr + 4*r_cr^2)
logL_cr <- (x[,"n_02"] + x[,"n_12"] + x[,"n_22"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(4*log(2) + 2*log(5)) + (x[,"n_04"] + x[,"n_30"])*(log(3 - 2*r_cr) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cr)) + log(1 + 2*r_cr)) + (x[,"n_01"] + x[,"n_03"] + x[,"n_31"] + x[,"n_33"])*log(9 + 4*r_cr - 4*r_cr^2) + (x[,"n_12"] + x[,"n_22"])*log(23 + 4*r_cr - 4*r_cr^2) + (x[,"n_10"] + x[,"n_24"])*log(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_14"] + x[,"n_20"])*log(3 + r_cr + 2*r_cr^2) + (x[,"n_11"] + x[,"n_23"])*log(45 - 12*r_cr + 4*r_cr^2) + (x[,"n_02"] + x[,"n_32"])*log(7 - 4*r_cr + 4*r_cr^2) + (x[,"n_13"] + x[,"n_21"])*log(37 + 4*r_cr + 4*r_cr^2)
logL_rc <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-2*n00 - 2*n01 - n02 - 2*n03 - 2*n04 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n30 - 2*n31 - n32 - 2*n33 - 2*n34)*log(2) + (n02 + n12 + n22 + n32)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*log(5) + (n00 + n04 + n30 + n34)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n12 + n22)*log(11 + 4*r - 4*r^2) + (n01 + n33)*log(3 - 2*r^2) + (n03 + n31)*log(1 + 4*r - 2*r^2) + (n02 + n32)*log(1 - r + r^2) + (n14 + n20)*log(3 - 3*r + 2*r^2) + (n10 + n24)*log(2 - r + 2*r^2) + (n11 + n13 + n21 + n23)*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rc <- -((x[,"n_14"] + x[,"n_20"])*log10(2)) - (x[,"n_10"] + x[,"n_24"])*log10(2) + 2*(x[,"n_00"] + x[,"n_04"] + x[,"n_30"] + x[,"n_34"])*log10(2) + (x[,"n_02"] + x[,"n_32"])*(2*log10(2) - log10(3)) - (x[,"n_12"] + x[,"n_22"])*(2*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_31"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_33"])*(-log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_13"] + x[,"n_21"] + x[,"n_23"])*(2*log10(2) + log10(5)) + (x[,"n_00"] + x[,"n_04"] + x[,"n_30"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc))) + (x[,"n_12"] + x[,"n_22"])*log10(11 + 4*r_rc - 4*r_rc^2) + (x[,"n_01"] + x[,"n_33"])*log10(3 - 2*r_rc^2) + (x[,"n_03"] + x[,"n_31"])*log10(1 + 4*r_rc - 2*r_rc^2) + (x[,"n_02"] + x[,"n_32"])*log10(1 - r_rc + r_rc^2) + (x[,"n_14"] + x[,"n_20"])*log10(3 - 3*r_rc + 2*r_rc^2) + (x[,"n_10"] + x[,"n_24"])*log10(2 - r_rc + 2*r_rc^2) + (x[,"n_11"] + x[,"n_13"] + x[,"n_21"] + x[,"n_23"])*log10(21 - 4*r_rc + 4*r_rc^2)
logL_rc <- (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_22"] + x[,"n_32"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_00"] + x[,"n_04"] + x[,"n_30"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc))) + (x[,"n_12"] + x[,"n_22"])*log(11 + 4*r_rc - 4*r_rc^2) + (x[,"n_01"] + x[,"n_33"])*log(3 - 2*r_rc^2) + (x[,"n_03"] + x[,"n_31"])*log(1 + 4*r_rc - 2*r_rc^2) + (x[,"n_02"] + x[,"n_32"])*log(1 - r_rc + r_rc^2) + (x[,"n_14"] + x[,"n_20"])*log(3 - 3*r_rc + 2*r_rc^2) + (x[,"n_10"] + x[,"n_24"])*log(2 - r_rc + 2*r_rc^2) + (x[,"n_11"] + x[,"n_13"] + x[,"n_21"] + x[,"n_23"])*log(21 - 4*r_rc + 4*r_rc^2)
logL_rr <- function(r,n02,n12,n22,n32,n00,n01,n03,n04,n10,n11,n13,n14,n20,n21,n23,n24,n30,n31,n33,n34) {
L <- (n02 + n12 + n22 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(4*log(2) + 2*log(5)) + (n04 + n30)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n00 + n34)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n13 + n21)*log(51 - 20*r - 4*r^2) + (n02 + n32)*log(5 + 4*r - 4*r^2) + (n11 + n23)*log(27 + 28*r - 4*r^2) + (n14 + n20)*log(6 - 3*r - 2*r^2) + (n10 + n24)*log(1 + 7*r - 2*r^2) + (n03 + n31)*log(19 - 20*r + 4*r^2) + (n12 + n22)*log(25 - 4*r + 4*r^2) + (n01 + n33)*log(3 + 12*r + 4*r^2)
return(L)}
interlogL_rr <- function(n02,n12,n22,n32,n00,n01,n03,n04,n10,n11,n13,n14,n20,n21,n23,n24,n30,n31,n33,n34) {
optimize(logL_rr,c(0,0.5), n02,n12,n22,n32,n00,n01,n03,n04,n10,n11,n13,n14,n20,n21,n23,n24,n30,n31,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_02"],x[,"n_12"],x[,"n_22"],x[,"n_32"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"])
LOD_rr <- -2*(x[,"n_14"] + x[,"n_20"])*log10(2) - 2*(x[,"n_10"] + x[,"n_24"])*log10(2) - (x[,"n_02"] + x[,"n_32"])*(log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_22"])*(3*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_31"])*(log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_33"])*(log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_21"])*(3*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_23"])*(3*log10(2) + log10(5)) + (x[,"n_04"] + x[,"n_30"])*(log10(3 - 2*r_rr) + log10(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_13"] + x[,"n_21"])*log10(51 - 20*r_rr - 4*r_rr^2) + (x[,"n_02"] + x[,"n_32"])*log10(5 + 4*r_rr - 4*r_rr^2) + (x[,"n_11"] + x[,"n_23"])*log10(27 + 28*r_rr - 4*r_rr^2) + (x[,"n_14"] + x[,"n_20"])*log10(6 - 3*r_rr - 2*r_rr^2) + (x[,"n_10"] + x[,"n_24"])*log10(1 + 7*r_rr - 2*r_rr^2) + (x[,"n_03"] + x[,"n_31"])*log10(19 - 20*r_rr + 4*r_rr^2) + (x[,"n_12"] + x[,"n_22"])*log10(25 - 4*r_rr + 4*r_rr^2) + (x[,"n_01"] + x[,"n_33"])*log10(3 + 12*r_rr + 4*r_rr^2)
logL_rr <- (x[,"n_02"] + x[,"n_12"] + x[,"n_22"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(4*log(2) + 2*log(5)) + (x[,"n_04"] + x[,"n_30"])*(log(3 - 2*r_rr) + log(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_13"] + x[,"n_21"])*log(51 - 20*r_rr - 4*r_rr^2) + (x[,"n_02"] + x[,"n_32"])*log(5 + 4*r_rr - 4*r_rr^2) + (x[,"n_11"] + x[,"n_23"])*log(27 + 28*r_rr - 4*r_rr^2) + (x[,"n_14"] + x[,"n_20"])*log(6 - 3*r_rr - 2*r_rr^2) + (x[,"n_10"] + x[,"n_24"])*log(1 + 7*r_rr - 2*r_rr^2) + (x[,"n_03"] + x[,"n_31"])*log(19 - 20*r_rr + 4*r_rr^2) + (x[,"n_12"] + x[,"n_22"])*log(25 - 4*r_rr + 4*r_rr^2) + (x[,"n_01"] + x[,"n_33"])*log(3 + 12*r_rr + 4*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_3.2 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n10 - 2*n11 - n12 - n13 - 2*n14 - 2*n15 - 2*n20 - 2*n21 - n22 - n23 - 2*n24 - 2*n25 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35)*log(2) + (n10 + n15 + n20 + n25)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(5) + 3*(n00 + n35)*log(pmax(1e-6,1 - r)) + (n01 + n34)*(2*log(pmax(1e-6,1 - r)) + log(6 - r)) + 3*(n05 + n30)*log(pmax(1e-6,r)) + (n10 + n25)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n15 + n20)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n31)*(2*log(pmax(1e-6,r)) + log(5 + r)) + (n03 + n32)*(log(pmax(1e-6,r)) + log(7 + 4*r - 8*r^2)) + (n14 + n21)*(log(pmax(1e-6,r)) + log(11 - 2*r - 3*r^2)) + (n12 + n23)*log(12 - 17*r + 22*r^2 - 12*r^3) + (n11 + n24)*log(6 + 2*r - 11*r^2 + 3*r^3) + (n02 + n33)*log(3 + 9*r - 20*r^2 + 8*r^3) + (n13 + n22)*log(5 + 9*r - 14*r^2 + 12*r^3)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_cc <- 3*(x[,"n_15"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_25"])*log10(2) + 3*(x[,"n_05"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_13"] + x[,"n_22"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_23"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_03"] + x[,"n_32"])*(-log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_33"])*(-log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_21"])*(-3*log10(2) + log10(37)) - (x[,"n_11"] + x[,"n_24"])*(-3*log10(2) + log10(37)) + 3*(x[,"n_00"] + x[,"n_35"])*log10(pmax(1e-6,1 - r_cc)) + (x[,"n_01"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(6 - r_cc)) + 3*(x[,"n_05"] + x[,"n_30"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_25"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_20"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(5 + r_cc)) + (x[,"n_03"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(7 + 4*r_cc - 8*r_cc^2)) + (x[,"n_14"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(11 - 2*r_cc - 3*r_cc^2)) + (x[,"n_12"] + x[,"n_23"])*log10(12 - 17*r_cc + 22*r_cc^2 - 12*r_cc^3) + (x[,"n_11"] + x[,"n_24"])*log10(6 + 2*r_cc - 11*r_cc^2 + 3*r_cc^3) + (x[,"n_02"] + x[,"n_33"])*log10(3 + 9*r_cc - 20*r_cc^2 + 8*r_cc^3) + (x[,"n_13"] + x[,"n_22"])*log10(5 + 9*r_cc - 14*r_cc^2 + 12*r_cc^3)
logL_cc <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"])*log(2) + (x[,"n_10"] + x[,"n_15"] + x[,"n_20"] + x[,"n_25"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + 3*(x[,"n_00"] + x[,"n_35"])*log(pmax(1e-6,1 - r_cc)) + (x[,"n_01"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_cc)) + log(6 - r_cc)) + 3*(x[,"n_05"] + x[,"n_30"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_25"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_20"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(5 + r_cc)) + (x[,"n_03"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(7 + 4*r_cc - 8*r_cc^2)) + (x[,"n_14"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(11 - 2*r_cc - 3*r_cc^2)) + (x[,"n_12"] + x[,"n_23"])*log(12 - 17*r_cc + 22*r_cc^2 - 12*r_cc^3) + (x[,"n_11"] + x[,"n_24"])*log(6 + 2*r_cc - 11*r_cc^2 + 3*r_cc^3) + (x[,"n_02"] + x[,"n_33"])*log(3 + 9*r_cc - 20*r_cc^2 + 8*r_cc^3) + (x[,"n_13"] + x[,"n_22"])*log(5 + 9*r_cc - 14*r_cc^2 + 12*r_cc^3)
logL_cm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(4*log(2) + 2*log(5)) + (n05 + n30)*(log(2 - r) + 2*log(pmax(1e-6,r))) + (n00 + n35)*(log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n04 + n31)*(log(pmax(1e-6,r)) + log(9 + r - r^2)) + (n15 + n20)*(log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n13 + n22)*log(55 - 4*r + 36*r^2 - 24*r^3) + (n02 + n33)*log(21 - 16*r + 12*r^2 - 8*r^3) + (n10 + n25)*log(3 - 4*r + 4*r^2 - 3*r^3) + (n11 + n24)*log(27 - 17*r + 5*r^2 - 3*r^3) + (n01 + n34)*log(9 - 8*r - 2*r^2 + r^3) + (n14 + n21)*log(12 + 16*r - 4*r^2 + 3*r^3) + (n03 + n32)*log(9 + 16*r - 12*r^2 + 8*r^3) + (n12 + n23)*log(63 + 4*r - 36*r^2 + 24*r^3)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_cm <- -((x[,"n_05"] + x[,"n_30"])*(-3*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_35"])*(-3*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_32"])*(log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_33"])*(log10(3) + log10(5)) - (x[,"n_15"] + x[,"n_20"])*(-3*log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_25"])*(-3*log10(2) + log10(13)) - (x[,"n_14"] + x[,"n_21"])*(-3*log10(2) + log10(5) + log10(31)) - (x[,"n_11"] + x[,"n_24"])*(-3*log10(2) + log10(5) + log10(31)) - (x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(37)) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(37)) - (x[,"n_13"] + x[,"n_22"])*log10(59) - (x[,"n_12"] + x[,"n_23"])*log10(59) + (x[,"n_05"] + x[,"n_30"])*(log10(2 - r_cm) + 2*log10(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_cm)) + log10(1 - r_cm^2)) + (x[,"n_04"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(9 + r_cm - r_cm^2)) + (x[,"n_15"] + x[,"n_20"])*(log10(pmax(1e-6,r_cm)) + log10(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_13"] + x[,"n_22"])*log10(55 - 4*r_cm + 36*r_cm^2 - 24*r_cm^3) + (x[,"n_02"] + x[,"n_33"])*log10(21 - 16*r_cm + 12*r_cm^2 - 8*r_cm^3) + (x[,"n_10"] + x[,"n_25"])*log10(3 - 4*r_cm + 4*r_cm^2 - 3*r_cm^3) + (x[,"n_11"] + x[,"n_24"])*log10(27 - 17*r_cm + 5*r_cm^2 - 3*r_cm^3) + (x[,"n_01"] + x[,"n_34"])*log10(9 - 8*r_cm - 2*r_cm^2 + r_cm^3) + (x[,"n_14"] + x[,"n_21"])*log10(12 + 16*r_cm - 4*r_cm^2 + 3*r_cm^3) + (x[,"n_03"] + x[,"n_32"])*log10(9 + 16*r_cm - 12*r_cm^2 + 8*r_cm^3) + (x[,"n_12"] + x[,"n_23"])*log10(63 + 4*r_cm - 36*r_cm^2 + 24*r_cm^3)
logL_cm <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_30"])*(log(2 - r_cm) + 2*log(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_cm)) + log(1 - r_cm^2)) + (x[,"n_04"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(9 + r_cm - r_cm^2)) + (x[,"n_15"] + x[,"n_20"])*(log(pmax(1e-6,r_cm)) + log(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_13"] + x[,"n_22"])*log(55 - 4*r_cm + 36*r_cm^2 - 24*r_cm^3) + (x[,"n_02"] + x[,"n_33"])*log(21 - 16*r_cm + 12*r_cm^2 - 8*r_cm^3) + (x[,"n_10"] + x[,"n_25"])*log(3 - 4*r_cm + 4*r_cm^2 - 3*r_cm^3) + (x[,"n_11"] + x[,"n_24"])*log(27 - 17*r_cm + 5*r_cm^2 - 3*r_cm^3) + (x[,"n_01"] + x[,"n_34"])*log(9 - 8*r_cm - 2*r_cm^2 + r_cm^3) + (x[,"n_14"] + x[,"n_21"])*log(12 + 16*r_cm - 4*r_cm^2 + 3*r_cm^3) + (x[,"n_03"] + x[,"n_32"])*log(9 + 16*r_cm - 12*r_cm^2 + 8*r_cm^3) + (x[,"n_12"] + x[,"n_23"])*log(63 + 4*r_cm - 36*r_cm^2 + 24*r_cm^3)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(3*log(2) + log(3) + 2*log(5)) + (n00 + n35)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n30)*(log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n12 + n23)*log(93 - 16*r + 28*r^2 - 24*r^3) + (n03 + n32)*log(27 - 18*r + 20*r^2 - 8*r^3) + (n15 + n20)*log(3 - 4*r + 7*r^2 - 3*r^3) + (n14 + n21)*log(30 - 7*r + 10*r^2 - 3*r^3) + (n01 + n34)*log(3 + 11*r - 4*r^2 - r^3) + (n04 + n31)*log(9 - 7*r^2 + r^3) + (n10 + n25)*log(3 - r - 2*r^2 + 3*r^3) + (n11 + n24)*log(30 - 4*r + r^2 + 3*r^3) + (n02 + n33)*log(21 + 2*r - 4*r^2 + 8*r^3) + (n13 + n22)*log(81 + 32*r - 44*r^2 + 24*r^3)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_cr <- -((x[,"n_05"] + x[,"n_30"])*(-3*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_35"])*(-3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_32"])*(log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_33"])*(log10(2) + log10(11)) - (x[,"n_15"] + x[,"n_20"])*(-3*log10(2) + log10(19)) - (x[,"n_10"] + x[,"n_25"])*(-3*log10(2) + log10(19)) - (x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(59)) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(59)) - (x[,"n_13"] + x[,"n_22"])*log10(89) - (x[,"n_12"] + x[,"n_23"])*log10(89) - (x[,"n_14"] + x[,"n_21"])*(-3*log10(2) + log10(229)) - (x[,"n_11"] + x[,"n_24"])*(-3*log10(2) + log10(229)) + (x[,"n_00"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr)) + log10(2 + r_cr)) + (x[,"n_05"] + x[,"n_30"])*(log10(pmax(1e-6,r_cr)) + log10(3 - 4*r_cr + r_cr^2)) + (x[,"n_12"] + x[,"n_23"])*log10(93 - 16*r_cr + 28*r_cr^2 - 24*r_cr^3) + (x[,"n_03"] + x[,"n_32"])*log10(27 - 18*r_cr + 20*r_cr^2 - 8*r_cr^3) + (x[,"n_15"] + x[,"n_20"])*log10(3 - 4*r_cr + 7*r_cr^2 - 3*r_cr^3) + (x[,"n_14"] + x[,"n_21"])*log10(30 - 7*r_cr + 10*r_cr^2 - 3*r_cr^3) + (x[,"n_01"] + x[,"n_34"])*log10(3 + 11*r_cr - 4*r_cr^2 - r_cr^3) + (x[,"n_04"] + x[,"n_31"])*log10(9 - 7*r_cr^2 + r_cr^3) + (x[,"n_10"] + x[,"n_25"])*log10(3 - r_cr - 2*r_cr^2 + 3*r_cr^3) + (x[,"n_11"] + x[,"n_24"])*log10(30 - 4*r_cr + r_cr^2 + 3*r_cr^3) + (x[,"n_02"] + x[,"n_33"])*log10(21 + 2*r_cr - 4*r_cr^2 + 8*r_cr^3) + (x[,"n_13"] + x[,"n_22"])*log10(81 + 32*r_cr - 44*r_cr^2 + 24*r_cr^3)
logL_cr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(3*log(2) + log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr)) + log(2 + r_cr)) + (x[,"n_05"] + x[,"n_30"])*(log(pmax(1e-6,r_cr)) + log(3 - 4*r_cr + r_cr^2)) + (x[,"n_12"] + x[,"n_23"])*log(93 - 16*r_cr + 28*r_cr^2 - 24*r_cr^3) + (x[,"n_03"] + x[,"n_32"])*log(27 - 18*r_cr + 20*r_cr^2 - 8*r_cr^3) + (x[,"n_15"] + x[,"n_20"])*log(3 - 4*r_cr + 7*r_cr^2 - 3*r_cr^3) + (x[,"n_14"] + x[,"n_21"])*log(30 - 7*r_cr + 10*r_cr^2 - 3*r_cr^3) + (x[,"n_01"] + x[,"n_34"])*log(3 + 11*r_cr - 4*r_cr^2 - r_cr^3) + (x[,"n_04"] + x[,"n_31"])*log(9 - 7*r_cr^2 + r_cr^3) + (x[,"n_10"] + x[,"n_25"])*log(3 - r_cr - 2*r_cr^2 + 3*r_cr^3) + (x[,"n_11"] + x[,"n_24"])*log(30 - 4*r_cr + r_cr^2 + 3*r_cr^3) + (x[,"n_02"] + x[,"n_33"])*log(21 + 2*r_cr - 4*r_cr^2 + 8*r_cr^3) + (x[,"n_13"] + x[,"n_22"])*log(81 + 32*r_cr - 44*r_cr^2 + 24*r_cr^3)
logL_rc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n10 - 2*n11 - n12 - n13 - 2*n14 - 2*n15 - 2*n20 - 2*n21 - n22 - n23 - 2*n24 - 2*n25 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35)*log(2) + (n12 + n13 + n22 + n23)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(5) + (n00 + n35)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n31)*(log(pmax(1e-6,r)) + log(2 + 2*r - r^2)) + (n15 + n20)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n02 + n33)*log(6 - 9*r + 12*r^2 - 8*r^3) + (n13 + n22)*log(3 - 3*r + 6*r^2 - 4*r^3) + (n11 + n24)*log(6 - 4*r + 4*r^2 - 3*r^3) + (n10 + n25)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n01 + n34)*log(3 - 3*r - r^2 + r^3) + (n14 + n21)*log(3 + 5*r - 5*r^2 + 3*r^3) + (n12 + n23)*log(2 + 3*r - 6*r^2 + 4*r^3) + (n03 + n32)*log(1 + 9*r - 12*r^2 + 8*r^3)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rc <- 3*(x[,"n_05"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_35"])*log10(2) + (x[,"n_10"] + x[,"n_25"])*(3*log10(2) - log10(3)) - (x[,"n_15"] + x[,"n_20"])*(-3*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_22"])*(-log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_23"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_32"])*(-log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_33"])*(-log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_21"])*(-3*log10(2) + log10(37)) - (x[,"n_11"] + x[,"n_24"])*(-3*log10(2) + log10(37)) + (x[,"n_00"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc))) + (x[,"n_05"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_rc)) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_04"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(2 + 2*r_rc - r_rc^2)) + (x[,"n_15"] + x[,"n_20"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 4*r_rc + 3*r_rc^2)) + (x[,"n_02"] + x[,"n_33"])*log10(6 - 9*r_rc + 12*r_rc^2 - 8*r_rc^3) + (x[,"n_13"] + x[,"n_22"])*log10(3 - 3*r_rc + 6*r_rc^2 - 4*r_rc^3) + (x[,"n_11"] + x[,"n_24"])*log10(6 - 4*r_rc + 4*r_rc^2 - 3*r_rc^3) + (x[,"n_10"] + x[,"n_25"])*log10(1 - 3*r_rc + 5*r_rc^2 - 3*r_rc^3) + (x[,"n_01"] + x[,"n_34"])*log10(3 - 3*r_rc - r_rc^2 + r_rc^3) + (x[,"n_14"] + x[,"n_21"])*log10(3 + 5*r_rc - 5*r_rc^2 + 3*r_rc^3) + (x[,"n_12"] + x[,"n_23"])*log10(2 + 3*r_rc - 6*r_rc^2 + 4*r_rc^3) + (x[,"n_03"] + x[,"n_32"])*log10(1 + 9*r_rc - 12*r_rc^2 + 8*r_rc^3)
logL_rc <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"])*log(2) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_00"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc))) + (x[,"n_05"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_rc)) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_04"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(2 + 2*r_rc - r_rc^2)) + (x[,"n_15"] + x[,"n_20"])*(log(pmax(1e-6,r_rc)) + log(2 - 4*r_rc + 3*r_rc^2)) + (x[,"n_02"] + x[,"n_33"])*log(6 - 9*r_rc + 12*r_rc^2 - 8*r_rc^3) + (x[,"n_13"] + x[,"n_22"])*log(3 - 3*r_rc + 6*r_rc^2 - 4*r_rc^3) + (x[,"n_11"] + x[,"n_24"])*log(6 - 4*r_rc + 4*r_rc^2 - 3*r_rc^3) + (x[,"n_10"] + x[,"n_25"])*log(1 - 3*r_rc + 5*r_rc^2 - 3*r_rc^3) + (x[,"n_01"] + x[,"n_34"])*log(3 - 3*r_rc - r_rc^2 + r_rc^3) + (x[,"n_14"] + x[,"n_21"])*log(3 + 5*r_rc - 5*r_rc^2 + 3*r_rc^3) + (x[,"n_12"] + x[,"n_23"])*log(2 + 3*r_rc - 6*r_rc^2 + 4*r_rc^3) + (x[,"n_03"] + x[,"n_32"])*log(1 + 9*r_rc - 12*r_rc^2 + 8*r_rc^3)
logL_rm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(4*log(2) + 2*log(5)) + (n05 + n30)*(log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n00 + n35)*(log(pmax(1e-6,r)) + log(1 - r^2)) + (n12 + n23)*log(57 - 4*r + 28*r^2 - 24*r^3) + (n03 + n32)*log(19 - 16*r + 20*r^2 - 8*r^3) + (n15 + n20)*log(3 - 5*r + 6*r^2 - 3*r^3) + (n14 + n21)*log(24 - 14*r + 11*r^2 - 3*r^3) + (n01 + n34)*log(3 + 6*r - 5*r^2 - r^3) + (n04 + n31)*log(3 + 7*r - 8*r^2 + r^3) + (n10 + n25)*log(1 + 2*r - 3*r^2 + 3*r^3) + (n11 + n24)*log(18 + r + 2*r^2 + 3*r^3) + (n02 + n33)*log(15 - 4*r^2 + 8*r^3) + (n13 + n22)*log(57 + 20*r - 44*r^2 + 24*r^3)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rm <- -((x[,"n_05"] + x[,"n_30"])*(-3*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_35"])*(-3*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_32"])*(log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_33"])*(log10(3) + log10(5)) - (x[,"n_15"] + x[,"n_20"])*(-3*log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_25"])*(-3*log10(2) + log10(13)) - (x[,"n_14"] + x[,"n_21"])*(-3*log10(2) + log10(5) + log10(31)) - (x[,"n_11"] + x[,"n_24"])*(-3*log10(2) + log10(5) + log10(31)) - (x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(37)) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(37)) - (x[,"n_13"] + x[,"n_22"])*log10(59) - (x[,"n_12"] + x[,"n_23"])*log10(59) + (x[,"n_05"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_rm)) + log10(2 - r_rm) + log10(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm)) + log10(1 - r_rm^2)) + (x[,"n_12"] + x[,"n_23"])*log10(57 - 4*r_rm + 28*r_rm^2 - 24*r_rm^3) + (x[,"n_03"] + x[,"n_32"])*log10(19 - 16*r_rm + 20*r_rm^2 - 8*r_rm^3) + (x[,"n_15"] + x[,"n_20"])*log10(3 - 5*r_rm + 6*r_rm^2 - 3*r_rm^3) + (x[,"n_14"] + x[,"n_21"])*log10(24 - 14*r_rm + 11*r_rm^2 - 3*r_rm^3) + (x[,"n_01"] + x[,"n_34"])*log10(3 + 6*r_rm - 5*r_rm^2 - r_rm^3) + (x[,"n_04"] + x[,"n_31"])*log10(3 + 7*r_rm - 8*r_rm^2 + r_rm^3) + (x[,"n_10"] + x[,"n_25"])*log10(1 + 2*r_rm - 3*r_rm^2 + 3*r_rm^3) + (x[,"n_11"] + x[,"n_24"])*log10(18 + r_rm + 2*r_rm^2 + 3*r_rm^3) + (x[,"n_02"] + x[,"n_33"])*log10(15 - 4*r_rm^2 + 8*r_rm^3) + (x[,"n_13"] + x[,"n_22"])*log10(57 + 20*r_rm - 44*r_rm^2 + 24*r_rm^3)
logL_rm <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_rm)) + log(2 - r_rm) + log(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_35"])*(log(pmax(1e-6,r_rm)) + log(1 - r_rm^2)) + (x[,"n_12"] + x[,"n_23"])*log(57 - 4*r_rm + 28*r_rm^2 - 24*r_rm^3) + (x[,"n_03"] + x[,"n_32"])*log(19 - 16*r_rm + 20*r_rm^2 - 8*r_rm^3) + (x[,"n_15"] + x[,"n_20"])*log(3 - 5*r_rm + 6*r_rm^2 - 3*r_rm^3) + (x[,"n_14"] + x[,"n_21"])*log(24 - 14*r_rm + 11*r_rm^2 - 3*r_rm^3) + (x[,"n_01"] + x[,"n_34"])*log(3 + 6*r_rm - 5*r_rm^2 - r_rm^3) + (x[,"n_04"] + x[,"n_31"])*log(3 + 7*r_rm - 8*r_rm^2 + r_rm^3) + (x[,"n_10"] + x[,"n_25"])*log(1 + 2*r_rm - 3*r_rm^2 + 3*r_rm^3) + (x[,"n_11"] + x[,"n_24"])*log(18 + r_rm + 2*r_rm^2 + 3*r_rm^3) + (x[,"n_02"] + x[,"n_33"])*log(15 - 4*r_rm^2 + 8*r_rm^3) + (x[,"n_13"] + x[,"n_22"])*log(57 + 20*r_rm - 44*r_rm^2 + 24*r_rm^3)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(3*log(2) + log(3) + 2*log(5)) + (n00 + n35)*(2*log(pmax(1e-6,r)) + log(2 + r)) + (n10 + n25)*(log(pmax(1e-6,r)) + log(5 + r - 3*r^2)) + (n05 + n30)*(log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n01 + n34)*(log(pmax(1e-6,r)) + log(9 + 11*r + r^2)) + (n13 + n22)*log(111 - 64*r + 52*r^2 - 24*r^3) + (n02 + n33)*log(9 + 30*r - 4*r^2 - 8*r^3) + (n11 + n24)*log(12 + 38*r - 8*r^2 - 3*r^3) + (n04 + n31)*log(21 - 34*r + 14*r^2 - r^3) + (n14 + n21)*log(39 - 13*r - 17*r^2 + 3*r^3) + (n15 + n20)*log(3 + 2*r - 8*r^2 + 3*r^3) + (n03 + n32)*log(27 + 2*r - 28*r^2 + 8*r^3) + (n12 + n23)*log(75 + 32*r - 20*r^2 + 24*r^3)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rr <- -((x[,"n_05"] + x[,"n_30"])*(-3*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_35"])*(-3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_32"])*(log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_33"])*(log10(2) + log10(11)) - (x[,"n_15"] + x[,"n_20"])*(-3*log10(2) + log10(19)) - (x[,"n_10"] + x[,"n_25"])*(-3*log10(2) + log10(19)) - (x[,"n_04"] + x[,"n_31"])*(-3*log10(2) + log10(59)) - (x[,"n_01"] + x[,"n_34"])*(-3*log10(2) + log10(59)) - (x[,"n_13"] + x[,"n_22"])*log10(89) - (x[,"n_12"] + x[,"n_23"])*log10(89) - (x[,"n_14"] + x[,"n_21"])*(-3*log10(2) + log10(229)) - (x[,"n_11"] + x[,"n_24"])*(-3*log10(2) + log10(229)) + (x[,"n_00"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_10"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(5 + r_rr - 3*r_rr^2)) + (x[,"n_05"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_rr)) + log10(3 - 4*r_rr + r_rr^2)) + (x[,"n_01"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(9 + 11*r_rr + r_rr^2)) + (x[,"n_13"] + x[,"n_22"])*log10(111 - 64*r_rr + 52*r_rr^2 - 24*r_rr^3) + (x[,"n_02"] + x[,"n_33"])*log10(9 + 30*r_rr - 4*r_rr^2 - 8*r_rr^3) + (x[,"n_11"] + x[,"n_24"])*log10(12 + 38*r_rr - 8*r_rr^2 - 3*r_rr^3) + (x[,"n_04"] + x[,"n_31"])*log10(21 - 34*r_rr + 14*r_rr^2 - r_rr^3) + (x[,"n_14"] + x[,"n_21"])*log10(39 - 13*r_rr - 17*r_rr^2 + 3*r_rr^3) + (x[,"n_15"] + x[,"n_20"])*log10(3 + 2*r_rr - 8*r_rr^2 + 3*r_rr^3) + (x[,"n_03"] + x[,"n_32"])*log10(27 + 2*r_rr - 28*r_rr^2 + 8*r_rr^3) + (x[,"n_12"] + x[,"n_23"])*log10(75 + 32*r_rr - 20*r_rr^2 + 24*r_rr^3)
logL_rr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(3*log(2) + log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_35"])*(2*log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_10"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(5 + r_rr - 3*r_rr^2)) + (x[,"n_05"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_rr)) + log(3 - 4*r_rr + r_rr^2)) + (x[,"n_01"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(9 + 11*r_rr + r_rr^2)) + (x[,"n_13"] + x[,"n_22"])*log(111 - 64*r_rr + 52*r_rr^2 - 24*r_rr^3) + (x[,"n_02"] + x[,"n_33"])*log(9 + 30*r_rr - 4*r_rr^2 - 8*r_rr^3) + (x[,"n_11"] + x[,"n_24"])*log(12 + 38*r_rr - 8*r_rr^2 - 3*r_rr^3) + (x[,"n_04"] + x[,"n_31"])*log(21 - 34*r_rr + 14*r_rr^2 - r_rr^3) + (x[,"n_14"] + x[,"n_21"])*log(39 - 13*r_rr - 17*r_rr^2 + 3*r_rr^3) + (x[,"n_15"] + x[,"n_20"])*log(3 + 2*r_rr - 8*r_rr^2 + 3*r_rr^3) + (x[,"n_03"] + x[,"n_32"])*log(27 + 2*r_rr - 28*r_rr^2 + 8*r_rr^3) + (x[,"n_12"] + x[,"n_23"])*log(75 + 32*r_rr - 20*r_rr^2 + 24*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_3.3 <- function(x){
logL_cc <- function(r,n10,n16,n20,n26,n00,n01,n02,n03,n04,n05,n06,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n36) {
L <- (n10 + n16 + n20 + n26)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(4*log(2) + 2*log(5)) + 3*(n00 + n36)*log(pmax(1e-6,1 - r)) + (n01 + n35)*(log(9 - 2*r) + 2*log(pmax(1e-6,1 - r))) + 3*(n06 + n30)*log(pmax(1e-6,r)) + (n10 + n26)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n16 + n20)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n05 + n31)*(2*log(pmax(1e-6,r)) + log(7 + 2*r)) + (n03 + n33)*log(9 + 28*r - 28*r^2) + (n04 + n32)*(log(pmax(1e-6,r)) + log(15 + 13*r - 7*r^2)) + (n15 + n21)*(log(pmax(1e-6,r)) + log(16 - r - 6*r^2)) + (n13 + n23)*log(73 - 28*r + 28*r^2) + (n12 + n24)*log(57 - 44*r + 29*r^2 - 21*r^3) + (n11 + n25)*log(9 + 4*r - 19*r^2 + 6*r^3) + (n02 + n34)*log(21 - 20*r - 8*r^2 + 7*r^3) + (n14 + n22)*log(21 + 49*r - 34*r^2 + 21*r^3)
return(L)}
interlogL_cc <- function(n10,n16,n20,n26,n00,n01,n02,n03,n04,n05,n06,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n36) {
optimize(logL_cc,c(0,0.5), n10,n16,n20,n26,n00,n01,n02,n03,n04,n05,n06,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n36, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_10"],x[,"n_16"],x[,"n_20"],x[,"n_26"],x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"])
LOD_cc <- 3*(x[,"n_16"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_26"])*log10(2) + 3*(x[,"n_06"] + x[,"n_30"])*log10(2) - (x[,"n_05"] + x[,"n_31"])*log10(2) - 4*(x[,"n_03"] + x[,"n_33"])*log10(2) - (x[,"n_01"] + x[,"n_35"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) - (x[,"n_15"] + x[,"n_21"])*log10(7) - (x[,"n_11"] + x[,"n_25"])*log10(7) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(79)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(79)) - (x[,"n_14"] + x[,"n_22"])*(-3*log10(2) + log10(317)) - (x[,"n_12"] + x[,"n_24"])*(-3*log10(2) + log10(317)) + 3*(x[,"n_00"] + x[,"n_36"])*log10(pmax(1e-6,1 - r_cc)) + (x[,"n_01"] + x[,"n_35"])*(log10(9 - 2*r_cc) + 2*log10(pmax(1e-6,1 - r_cc))) + 3*(x[,"n_06"] + x[,"n_30"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_26"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_16"] + x[,"n_20"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(7 + 2*r_cc)) + (x[,"n_03"] + x[,"n_33"])*log10(9 + 28*r_cc - 28*r_cc^2) + (x[,"n_04"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(15 + 13*r_cc - 7*r_cc^2)) + (x[,"n_15"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(16 - r_cc - 6*r_cc^2)) + (x[,"n_13"] + x[,"n_23"])*log10(73 - 28*r_cc + 28*r_cc^2) + (x[,"n_12"] + x[,"n_24"])*log10(57 - 44*r_cc + 29*r_cc^2 - 21*r_cc^3) + (x[,"n_11"] + x[,"n_25"])*log10(9 + 4*r_cc - 19*r_cc^2 + 6*r_cc^3) + (x[,"n_02"] + x[,"n_34"])*log10(21 - 20*r_cc - 8*r_cc^2 + 7*r_cc^3) + (x[,"n_14"] + x[,"n_22"])*log10(21 + 49*r_cc - 34*r_cc^2 + 21*r_cc^3)
logL_cc <- (x[,"n_10"] + x[,"n_16"] + x[,"n_20"] + x[,"n_26"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(4*log(2) + 2*log(5)) + 3*(x[,"n_00"] + x[,"n_36"])*log(pmax(1e-6,1 - r_cc)) + (x[,"n_01"] + x[,"n_35"])*(log(9 - 2*r_cc) + 2*log(pmax(1e-6,1 - r_cc))) + 3*(x[,"n_06"] + x[,"n_30"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_26"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_16"] + x[,"n_20"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(7 + 2*r_cc)) + (x[,"n_03"] + x[,"n_33"])*log(9 + 28*r_cc - 28*r_cc^2) + (x[,"n_04"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(15 + 13*r_cc - 7*r_cc^2)) + (x[,"n_15"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(16 - r_cc - 6*r_cc^2)) + (x[,"n_13"] + x[,"n_23"])*log(73 - 28*r_cc + 28*r_cc^2) + (x[,"n_12"] + x[,"n_24"])*log(57 - 44*r_cc + 29*r_cc^2 - 21*r_cc^3) + (x[,"n_11"] + x[,"n_25"])*log(9 + 4*r_cc - 19*r_cc^2 + 6*r_cc^3) + (x[,"n_02"] + x[,"n_34"])*log(21 - 20*r_cc - 8*r_cc^2 + 7*r_cc^3) + (x[,"n_14"] + x[,"n_22"])*log(21 + 49*r_cc - 34*r_cc^2 + 21*r_cc^3)
logL_cm <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n13,n23) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n06 - 4*n10 - 4*n11 - 4*n12 - 4*n14 - 4*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 4*n24 - 4*n25 - 4*n26 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n36)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(log(3) + 2*log(5)) + (n00 + n36)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n06 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n05 + n31)*(log(pmax(1e-6,r)) + log(6 - r - 2*r^2)) + (n13 + n23)*log(12 + r - r^2) + (n16 + n20)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2)) + (n03 + n33)*log(27 - 8*r + 8*r^2) + (n14 + n22)*log(105 + 7*r + 62*r^2 - 42*r^3) + (n11 + n25)*log(33 - 32*r + 26*r^2 - 12*r^3) + (n02 + n34)*log(21 - 16*r + 11*r^2 - 7*r^3) + (n10 + n26)*log(3 - 7*r + 10*r^2 - 6*r^3) + (n01 + n35)*log(3 + 2*r - 7*r^2 + 2*r^3) + (n04 + n32)*log(9 + 15*r - 10*r^2 + 7*r^3) + (n15 + n21)*log(15 + 16*r - 10*r^2 + 12*r^3) + (n12 + n24)*log(132 - 5*r - 64*r^2 + 42*r^3)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n13,n23) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n13,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_13"],x[,"n_23"])
LOD_cm <- 3*(x[,"n_06"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) - 2*(x[,"n_03"] + x[,"n_33"])*log10(5) - (x[,"n_16"] + x[,"n_20"])*(-2*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_26"])*(-2*log10(2) + log10(5)) - (x[,"n_05"] + x[,"n_31"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_35"])*(-log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_23"])*(-2*log10(2) + 2*log10(7)) - (x[,"n_15"] + x[,"n_21"])*(log10(2) + log10(11)) - (x[,"n_11"] + x[,"n_25"])*(log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + 2*log10(5) + log10(19)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + 2*log10(5) + log10(19)) + (x[,"n_00"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_06"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cm)) + 2*log10(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(6 - r_cm - 2*r_cm^2)) + (x[,"n_13"] + x[,"n_23"])*log10(12 + r_cm - r_cm^2) + (x[,"n_16"] + x[,"n_20"])*(log10(pmax(1e-6,r_cm)) + log10(5 - 8*r_cm + 6*r_cm^2)) + (x[,"n_03"] + x[,"n_33"])*log10(27 - 8*r_cm + 8*r_cm^2) + (x[,"n_14"] + x[,"n_22"])*log10(105 + 7*r_cm + 62*r_cm^2 - 42*r_cm^3) + (x[,"n_11"] + x[,"n_25"])*log10(33 - 32*r_cm + 26*r_cm^2 - 12*r_cm^3) + (x[,"n_02"] + x[,"n_34"])*log10(21 - 16*r_cm + 11*r_cm^2 - 7*r_cm^3) + (x[,"n_10"] + x[,"n_26"])*log10(3 - 7*r_cm + 10*r_cm^2 - 6*r_cm^3) + (x[,"n_01"] + x[,"n_35"])*log10(3 + 2*r_cm - 7*r_cm^2 + 2*r_cm^3) + (x[,"n_04"] + x[,"n_32"])*log10(9 + 15*r_cm - 10*r_cm^2 + 7*r_cm^3) + (x[,"n_15"] + x[,"n_21"])*log10(15 + 16*r_cm - 10*r_cm^2 + 12*r_cm^3) + (x[,"n_12"] + x[,"n_24"])*log10(132 - 5*r_cm - 64*r_cm^2 + 42*r_cm^3)
logL_cm <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_06"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cm)) + 2*log(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(6 - r_cm - 2*r_cm^2)) + (x[,"n_13"] + x[,"n_23"])*log(12 + r_cm - r_cm^2) + (x[,"n_16"] + x[,"n_20"])*(log(pmax(1e-6,r_cm)) + log(5 - 8*r_cm + 6*r_cm^2)) + (x[,"n_03"] + x[,"n_33"])*log(27 - 8*r_cm + 8*r_cm^2) + (x[,"n_14"] + x[,"n_22"])*log(105 + 7*r_cm + 62*r_cm^2 - 42*r_cm^3) + (x[,"n_11"] + x[,"n_25"])*log(33 - 32*r_cm + 26*r_cm^2 - 12*r_cm^3) + (x[,"n_02"] + x[,"n_34"])*log(21 - 16*r_cm + 11*r_cm^2 - 7*r_cm^3) + (x[,"n_10"] + x[,"n_26"])*log(3 - 7*r_cm + 10*r_cm^2 - 6*r_cm^3) + (x[,"n_01"] + x[,"n_35"])*log(3 + 2*r_cm - 7*r_cm^2 + 2*r_cm^3) + (x[,"n_04"] + x[,"n_32"])*log(9 + 15*r_cm - 10*r_cm^2 + 7*r_cm^3) + (x[,"n_15"] + x[,"n_21"])*log(15 + 16*r_cm - 10*r_cm^2 + 12*r_cm^3) + (x[,"n_12"] + x[,"n_24"])*log(132 - 5*r_cm - 64*r_cm^2 + 42*r_cm^3)
logL_cr <- function(r,n13,n23,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36) {
L <- (n13 + n23)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(4*log(2) + 2*log(5)) + (n06 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n36)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n13 + n23)*log(21 + 4*r - 4*r^2) + (n01 + n35)*(log(pmax(1e-6,r)) + log(4 + r - 2*r^2)) + (n10 + n26)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n03 + n33)*log(19 - 12*r + 12*r^2) + (n12 + n24)*log(42 - 17*r + 35*r^2 - 21*r^3) + (n04 + n32)*log(15 - 12*r + 7*r^2 - 7*r^3) + (n15 + n21)*log(9 - 8*r + 11*r^2 - 6*r^3) + (n16 + n20)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n05 + n31)*log(3 - 5*r^2 + 2*r^3) + (n11 + n25)*log(6 + 4*r - 7*r^2 + 6*r^3) + (n02 + n34)*log(3 + 19*r - 14*r^2 + 7*r^3) + (n14 + n22)*log(39 + 10*r - 28*r^2 + 21*r^3)
return(L)}
interlogL_cr <- function(n13,n23,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36) {
optimize(logL_cr,c(0,0.5), n13,n23,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_13"],x[,"n_23"],x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"])
LOD_cr <- 3*(x[,"n_06"] + x[,"n_30"])*log10(2) - (x[,"n_05"] + x[,"n_31"])*log10(2) - 4*(x[,"n_03"] + x[,"n_33"])*log10(2) - (x[,"n_01"] + x[,"n_35"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) + (x[,"n_16"] + x[,"n_20"])*(3*log10(2) - log10(3)) - (x[,"n_10"] + x[,"n_26"])*(-3*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_21"])*log10(7) - (x[,"n_11"] + x[,"n_25"])*log10(7) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(79)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(79)) - (x[,"n_14"] + x[,"n_22"])*(-3*log10(2) + log10(317)) - (x[,"n_12"] + x[,"n_24"])*(-3*log10(2) + log10(317)) + (x[,"n_06"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_36"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_13"] + x[,"n_23"])*log10(21 + 4*r_cr - 4*r_cr^2) + (x[,"n_01"] + x[,"n_35"])*(log10(pmax(1e-6,r_cr)) + log10(4 + r_cr - 2*r_cr^2)) + (x[,"n_10"] + x[,"n_26"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_03"] + x[,"n_33"])*log10(19 - 12*r_cr + 12*r_cr^2) + (x[,"n_12"] + x[,"n_24"])*log10(42 - 17*r_cr + 35*r_cr^2 - 21*r_cr^3) + (x[,"n_04"] + x[,"n_32"])*log10(15 - 12*r_cr + 7*r_cr^2 - 7*r_cr^3) + (x[,"n_15"] + x[,"n_21"])*log10(9 - 8*r_cr + 11*r_cr^2 - 6*r_cr^3) + (x[,"n_16"] + x[,"n_20"])*log10(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_05"] + x[,"n_31"])*log10(3 - 5*r_cr^2 + 2*r_cr^3) + (x[,"n_11"] + x[,"n_25"])*log10(6 + 4*r_cr - 7*r_cr^2 + 6*r_cr^3) + (x[,"n_02"] + x[,"n_34"])*log10(3 + 19*r_cr - 14*r_cr^2 + 7*r_cr^3) + (x[,"n_14"] + x[,"n_22"])*log10(39 + 10*r_cr - 28*r_cr^2 + 21*r_cr^3)
logL_cr <- (x[,"n_13"] + x[,"n_23"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(4*log(2) + 2*log(5)) + (x[,"n_06"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_36"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_13"] + x[,"n_23"])*log(21 + 4*r_cr - 4*r_cr^2) + (x[,"n_01"] + x[,"n_35"])*(log(pmax(1e-6,r_cr)) + log(4 + r_cr - 2*r_cr^2)) + (x[,"n_10"] + x[,"n_26"])*(log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_03"] + x[,"n_33"])*log(19 - 12*r_cr + 12*r_cr^2) + (x[,"n_12"] + x[,"n_24"])*log(42 - 17*r_cr + 35*r_cr^2 - 21*r_cr^3) + (x[,"n_04"] + x[,"n_32"])*log(15 - 12*r_cr + 7*r_cr^2 - 7*r_cr^3) + (x[,"n_15"] + x[,"n_21"])*log(9 - 8*r_cr + 11*r_cr^2 - 6*r_cr^3) + (x[,"n_16"] + x[,"n_20"])*log(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_05"] + x[,"n_31"])*log(3 - 5*r_cr^2 + 2*r_cr^3) + (x[,"n_11"] + x[,"n_25"])*log(6 + 4*r_cr - 7*r_cr^2 + 6*r_cr^3) + (x[,"n_02"] + x[,"n_34"])*log(3 + 19*r_cr - 14*r_cr^2 + 7*r_cr^3) + (x[,"n_14"] + x[,"n_22"])*log(39 + 10*r_cr - 28*r_cr^2 + 21*r_cr^3)
logL_rc <- function(r,n13,n23,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36) {
L <- (n13 + n23)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(4*log(2) + 2*log(5)) + (n00 + n36)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n06 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n13 + n23)*log(21 + 4*r - 4*r^2) + (n05 + n31)*(log(pmax(1e-6,r)) + log(4 + r - 2*r^2)) + (n16 + n20)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n03 + n33)*log(19 - 12*r + 12*r^2) + (n14 + n22)*log(42 - 17*r + 35*r^2 - 21*r^3) + (n02 + n34)*log(15 - 12*r + 7*r^2 - 7*r^3) + (n11 + n25)*log(9 - 8*r + 11*r^2 - 6*r^3) + (n10 + n26)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n01 + n35)*log(3 - 5*r^2 + 2*r^3) + (n15 + n21)*log(6 + 4*r - 7*r^2 + 6*r^3) + (n04 + n32)*log(3 + 19*r - 14*r^2 + 7*r^3) + (n12 + n24)*log(39 + 10*r - 28*r^2 + 21*r^3)
return(L)}
interlogL_rc <- function(n13,n23,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36) {
optimize(logL_rc,c(0,0.5), n13,n23,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_13"],x[,"n_23"],x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"])
LOD_rc <- 3*(x[,"n_06"] + x[,"n_30"])*log10(2) - (x[,"n_05"] + x[,"n_31"])*log10(2) - 4*(x[,"n_03"] + x[,"n_33"])*log10(2) - (x[,"n_01"] + x[,"n_35"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) + (x[,"n_10"] + x[,"n_26"])*(3*log10(2) - log10(3)) - (x[,"n_16"] + x[,"n_20"])*(-3*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_21"])*log10(7) - (x[,"n_11"] + x[,"n_25"])*log10(7) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(79)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(79)) - (x[,"n_14"] + x[,"n_22"])*(-3*log10(2) + log10(317)) - (x[,"n_12"] + x[,"n_24"])*(-3*log10(2) + log10(317)) + (x[,"n_00"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc))) + (x[,"n_06"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_rc)) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_13"] + x[,"n_23"])*log10(21 + 4*r_rc - 4*r_rc^2) + (x[,"n_05"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(4 + r_rc - 2*r_rc^2)) + (x[,"n_16"] + x[,"n_20"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 4*r_rc + 3*r_rc^2)) + (x[,"n_03"] + x[,"n_33"])*log10(19 - 12*r_rc + 12*r_rc^2) + (x[,"n_14"] + x[,"n_22"])*log10(42 - 17*r_rc + 35*r_rc^2 - 21*r_rc^3) + (x[,"n_02"] + x[,"n_34"])*log10(15 - 12*r_rc + 7*r_rc^2 - 7*r_rc^3) + (x[,"n_11"] + x[,"n_25"])*log10(9 - 8*r_rc + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_10"] + x[,"n_26"])*log10(1 - 3*r_rc + 5*r_rc^2 - 3*r_rc^3) + (x[,"n_01"] + x[,"n_35"])*log10(3 - 5*r_rc^2 + 2*r_rc^3) + (x[,"n_15"] + x[,"n_21"])*log10(6 + 4*r_rc - 7*r_rc^2 + 6*r_rc^3) + (x[,"n_04"] + x[,"n_32"])*log10(3 + 19*r_rc - 14*r_rc^2 + 7*r_rc^3) + (x[,"n_12"] + x[,"n_24"])*log10(39 + 10*r_rc - 28*r_rc^2 + 21*r_rc^3)
logL_rc <- (x[,"n_13"] + x[,"n_23"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(4*log(2) + 2*log(5)) + (x[,"n_00"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc))) + (x[,"n_06"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_rc)) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_13"] + x[,"n_23"])*log(21 + 4*r_rc - 4*r_rc^2) + (x[,"n_05"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(4 + r_rc - 2*r_rc^2)) + (x[,"n_16"] + x[,"n_20"])*(log(pmax(1e-6,r_rc)) + log(2 - 4*r_rc + 3*r_rc^2)) + (x[,"n_03"] + x[,"n_33"])*log(19 - 12*r_rc + 12*r_rc^2) + (x[,"n_14"] + x[,"n_22"])*log(42 - 17*r_rc + 35*r_rc^2 - 21*r_rc^3) + (x[,"n_02"] + x[,"n_34"])*log(15 - 12*r_rc + 7*r_rc^2 - 7*r_rc^3) + (x[,"n_11"] + x[,"n_25"])*log(9 - 8*r_rc + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_10"] + x[,"n_26"])*log(1 - 3*r_rc + 5*r_rc^2 - 3*r_rc^3) + (x[,"n_01"] + x[,"n_35"])*log(3 - 5*r_rc^2 + 2*r_rc^3) + (x[,"n_15"] + x[,"n_21"])*log(6 + 4*r_rc - 7*r_rc^2 + 6*r_rc^3) + (x[,"n_04"] + x[,"n_32"])*log(3 + 19*r_rc - 14*r_rc^2 + 7*r_rc^3) + (x[,"n_12"] + x[,"n_24"])*log(39 + 10*r_rc - 28*r_rc^2 + 21*r_rc^3)
logL_rm <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n13,n23) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n06 - 4*n10 - 4*n11 - 4*n12 - 4*n14 - 4*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 4*n24 - 4*n25 - 4*n26 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n36)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(log(3) + 2*log(5)) + (n06 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n36)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n35)*(log(pmax(1e-6,r)) + log(6 - r - 2*r^2)) + (n13 + n23)*log(12 + r - r^2) + (n10 + n26)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2)) + (n03 + n33)*log(27 - 8*r + 8*r^2) + (n12 + n24)*log(105 + 7*r + 62*r^2 - 42*r^3) + (n15 + n21)*log(33 - 32*r + 26*r^2 - 12*r^3) + (n04 + n32)*log(21 - 16*r + 11*r^2 - 7*r^3) + (n16 + n20)*log(3 - 7*r + 10*r^2 - 6*r^3) + (n05 + n31)*log(3 + 2*r - 7*r^2 + 2*r^3) + (n02 + n34)*log(9 + 15*r - 10*r^2 + 7*r^3) + (n11 + n25)*log(15 + 16*r - 10*r^2 + 12*r^3) + (n14 + n22)*log(132 - 5*r - 64*r^2 + 42*r^3)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n13,n23) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n13,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_13"],x[,"n_23"])
LOD_rm <- 3*(x[,"n_06"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) - 2*(x[,"n_03"] + x[,"n_33"])*log10(5) - (x[,"n_16"] + x[,"n_20"])*(-2*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_26"])*(-2*log10(2) + log10(5)) - (x[,"n_05"] + x[,"n_31"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_35"])*(-log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_23"])*(-2*log10(2) + 2*log10(7)) - (x[,"n_15"] + x[,"n_21"])*(log10(2) + log10(11)) - (x[,"n_11"] + x[,"n_25"])*(log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + 2*log10(5) + log10(19)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + 2*log10(5) + log10(19)) + (x[,"n_06"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_36"])*(log10(pmax(1e-6,1 - r_rm)) + 2*log10(pmax(1e-6,r_rm))) + (x[,"n_01"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm)) + log10(6 - r_rm - 2*r_rm^2)) + (x[,"n_13"] + x[,"n_23"])*log10(12 + r_rm - r_rm^2) + (x[,"n_10"] + x[,"n_26"])*(log10(pmax(1e-6,r_rm)) + log10(5 - 8*r_rm + 6*r_rm^2)) + (x[,"n_03"] + x[,"n_33"])*log10(27 - 8*r_rm + 8*r_rm^2) + (x[,"n_12"] + x[,"n_24"])*log10(105 + 7*r_rm + 62*r_rm^2 - 42*r_rm^3) + (x[,"n_15"] + x[,"n_21"])*log10(33 - 32*r_rm + 26*r_rm^2 - 12*r_rm^3) + (x[,"n_04"] + x[,"n_32"])*log10(21 - 16*r_rm + 11*r_rm^2 - 7*r_rm^3) + (x[,"n_16"] + x[,"n_20"])*log10(3 - 7*r_rm + 10*r_rm^2 - 6*r_rm^3) + (x[,"n_05"] + x[,"n_31"])*log10(3 + 2*r_rm - 7*r_rm^2 + 2*r_rm^3) + (x[,"n_02"] + x[,"n_34"])*log10(9 + 15*r_rm - 10*r_rm^2 + 7*r_rm^3) + (x[,"n_11"] + x[,"n_25"])*log10(15 + 16*r_rm - 10*r_rm^2 + 12*r_rm^3) + (x[,"n_14"] + x[,"n_22"])*log10(132 - 5*r_rm - 64*r_rm^2 + 42*r_rm^3)
logL_rm <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(log(3) + 2*log(5)) + (x[,"n_06"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_36"])*(log(pmax(1e-6,1 - r_rm)) + 2*log(pmax(1e-6,r_rm))) + (x[,"n_01"] + x[,"n_35"])*(log(pmax(1e-6,r_rm)) + log(6 - r_rm - 2*r_rm^2)) + (x[,"n_13"] + x[,"n_23"])*log(12 + r_rm - r_rm^2) + (x[,"n_10"] + x[,"n_26"])*(log(pmax(1e-6,r_rm)) + log(5 - 8*r_rm + 6*r_rm^2)) + (x[,"n_03"] + x[,"n_33"])*log(27 - 8*r_rm + 8*r_rm^2) + (x[,"n_12"] + x[,"n_24"])*log(105 + 7*r_rm + 62*r_rm^2 - 42*r_rm^3) + (x[,"n_15"] + x[,"n_21"])*log(33 - 32*r_rm + 26*r_rm^2 - 12*r_rm^3) + (x[,"n_04"] + x[,"n_32"])*log(21 - 16*r_rm + 11*r_rm^2 - 7*r_rm^3) + (x[,"n_16"] + x[,"n_20"])*log(3 - 7*r_rm + 10*r_rm^2 - 6*r_rm^3) + (x[,"n_05"] + x[,"n_31"])*log(3 + 2*r_rm - 7*r_rm^2 + 2*r_rm^3) + (x[,"n_02"] + x[,"n_34"])*log(9 + 15*r_rm - 10*r_rm^2 + 7*r_rm^3) + (x[,"n_11"] + x[,"n_25"])*log(15 + 16*r_rm - 10*r_rm^2 + 12*r_rm^3) + (x[,"n_14"] + x[,"n_22"])*log(132 - 5*r_rm - 64*r_rm^2 + 42*r_rm^3)
logL_rr <- function(r,n10,n16,n20,n26,n00,n01,n02,n03,n04,n05,n06,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n36) {
L <- (n10 + n16 + n20 + n26)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(4*log(2) + 2*log(5)) + 3*(n06 + n30)*log(pmax(1e-6,1 - r)) + (n05 + n31)*(log(9 - 2*r) + 2*log(pmax(1e-6,1 - r))) + 3*(n00 + n36)*log(pmax(1e-6,r)) + (n16 + n20)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n26)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n35)*(2*log(pmax(1e-6,r)) + log(7 + 2*r)) + (n03 + n33)*log(9 + 28*r - 28*r^2) + (n02 + n34)*(log(pmax(1e-6,r)) + log(15 + 13*r - 7*r^2)) + (n11 + n25)*(log(pmax(1e-6,r)) + log(16 - r - 6*r^2)) + (n13 + n23)*log(73 - 28*r + 28*r^2) + (n14 + n22)*log(57 - 44*r + 29*r^2 - 21*r^3) + (n15 + n21)*log(9 + 4*r - 19*r^2 + 6*r^3) + (n04 + n32)*log(21 - 20*r - 8*r^2 + 7*r^3) + (n12 + n24)*log(21 + 49*r - 34*r^2 + 21*r^3)
return(L)}
interlogL_rr <- function(n10,n16,n20,n26,n00,n01,n02,n03,n04,n05,n06,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n36) {
optimize(logL_rr,c(0,0.5), n10,n16,n20,n26,n00,n01,n02,n03,n04,n05,n06,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n36, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_10"],x[,"n_16"],x[,"n_20"],x[,"n_26"],x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"])
LOD_rr <- 3*(x[,"n_16"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_26"])*log10(2) + 3*(x[,"n_06"] + x[,"n_30"])*log10(2) - (x[,"n_05"] + x[,"n_31"])*log10(2) - 4*(x[,"n_03"] + x[,"n_33"])*log10(2) - (x[,"n_01"] + x[,"n_35"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) - (x[,"n_15"] + x[,"n_21"])*log10(7) - (x[,"n_11"] + x[,"n_25"])*log10(7) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(79)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(79)) - (x[,"n_14"] + x[,"n_22"])*(-3*log10(2) + log10(317)) - (x[,"n_12"] + x[,"n_24"])*(-3*log10(2) + log10(317)) + 3*(x[,"n_06"] + x[,"n_30"])*log10(pmax(1e-6,1 - r_rr)) + (x[,"n_05"] + x[,"n_31"])*(log10(9 - 2*r_rr) + 2*log10(pmax(1e-6,1 - r_rr))) + 3*(x[,"n_00"] + x[,"n_36"])*log10(pmax(1e-6,r_rr)) + (x[,"n_16"] + x[,"n_20"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_26"])*(log10(pmax(1e-6,1 - r_rr)) + 2*log10(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_rr)) + log10(7 + 2*r_rr)) + (x[,"n_03"] + x[,"n_33"])*log10(9 + 28*r_rr - 28*r_rr^2) + (x[,"n_02"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(15 + 13*r_rr - 7*r_rr^2)) + (x[,"n_11"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(16 - r_rr - 6*r_rr^2)) + (x[,"n_13"] + x[,"n_23"])*log10(73 - 28*r_rr + 28*r_rr^2) + (x[,"n_14"] + x[,"n_22"])*log10(57 - 44*r_rr + 29*r_rr^2 - 21*r_rr^3) + (x[,"n_15"] + x[,"n_21"])*log10(9 + 4*r_rr - 19*r_rr^2 + 6*r_rr^3) + (x[,"n_04"] + x[,"n_32"])*log10(21 - 20*r_rr - 8*r_rr^2 + 7*r_rr^3) + (x[,"n_12"] + x[,"n_24"])*log10(21 + 49*r_rr - 34*r_rr^2 + 21*r_rr^3)
logL_rr <- (x[,"n_10"] + x[,"n_16"] + x[,"n_20"] + x[,"n_26"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(4*log(2) + 2*log(5)) + 3*(x[,"n_06"] + x[,"n_30"])*log(pmax(1e-6,1 - r_rr)) + (x[,"n_05"] + x[,"n_31"])*(log(9 - 2*r_rr) + 2*log(pmax(1e-6,1 - r_rr))) + 3*(x[,"n_00"] + x[,"n_36"])*log(pmax(1e-6,r_rr)) + (x[,"n_16"] + x[,"n_20"])*(2*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_26"])*(log(pmax(1e-6,1 - r_rr)) + 2*log(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_35"])*(2*log(pmax(1e-6,r_rr)) + log(7 + 2*r_rr)) + (x[,"n_03"] + x[,"n_33"])*log(9 + 28*r_rr - 28*r_rr^2) + (x[,"n_02"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(15 + 13*r_rr - 7*r_rr^2)) + (x[,"n_11"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(16 - r_rr - 6*r_rr^2)) + (x[,"n_13"] + x[,"n_23"])*log(73 - 28*r_rr + 28*r_rr^2) + (x[,"n_14"] + x[,"n_22"])*log(57 - 44*r_rr + 29*r_rr^2 - 21*r_rr^3) + (x[,"n_15"] + x[,"n_21"])*log(9 + 4*r_rr - 19*r_rr^2 + 6*r_rr^3) + (x[,"n_04"] + x[,"n_32"])*log(21 - 20*r_rr - 8*r_rr^2 + 7*r_rr^3) + (x[,"n_12"] + x[,"n_24"])*log(21 + 49*r_rr - 34*r_rr^2 + 21*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.2_4.1 <- function(x){
logL_cc <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-2*n01 - 2*n02 - 2*n03 - 2*n04 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n31 - 2*n32 - 2*n33 - 2*n34)*log(2) + 2*(-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*log(5) + (n01 + n34)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n04 + n31)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n11 + n24)*log(11 - 4*r - 4*r^2) + (n14 + n21)*log(3 + 12*r - 4*r^2) + (n02 + n33)*log(6 - 3*r - 2*r^2) + (n03 + n32)*log(1 + 7*r - 2*r^2) + (n12 + n23)*log(39 - 16*r + 4*r^2) + (n13 + n22)*log(27 + 8*r + 4*r^2)
return(L)}
interlogL_cc <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_cc,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cc <- -3*(x[,"n_14"] + x[,"n_21"])*log10(2) - 5*(x[,"n_13"] + x[,"n_22"])*log10(2) - 5*(x[,"n_12"] + x[,"n_23"])*log10(2) - 3*(x[,"n_11"] + x[,"n_24"])*log10(2) - 2*(x[,"n_03"] + x[,"n_32"])*log10(2) - 2*(x[,"n_02"] + x[,"n_33"])*log10(2) + (x[,"n_01"] + x[,"n_34"])*(log10(3 - 2*r_cc) + log10(pmax(1e-6,1 - r_cc))) + (x[,"n_04"] + x[,"n_31"])*(log10(pmax(1e-6,r_cc)) + log10(1 + 2*r_cc)) + (x[,"n_11"] + x[,"n_24"])*log10(11 - 4*r_cc - 4*r_cc^2) + (x[,"n_14"] + x[,"n_21"])*log10(3 + 12*r_cc - 4*r_cc^2) + (x[,"n_02"] + x[,"n_33"])*log10(6 - 3*r_cc - 2*r_cc^2) + (x[,"n_03"] + x[,"n_32"])*log10(1 + 7*r_cc - 2*r_cc^2) + (x[,"n_12"] + x[,"n_23"])*log10(39 - 16*r_cc + 4*r_cc^2) + (x[,"n_13"] + x[,"n_22"])*log10(27 + 8*r_cc + 4*r_cc^2)
logL_cc <- (-2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_01"] + x[,"n_34"])*(log(3 - 2*r_cc) + log(pmax(1e-6,1 - r_cc))) + (x[,"n_04"] + x[,"n_31"])*(log(pmax(1e-6,r_cc)) + log(1 + 2*r_cc)) + (x[,"n_11"] + x[,"n_24"])*log(11 - 4*r_cc - 4*r_cc^2) + (x[,"n_14"] + x[,"n_21"])*log(3 + 12*r_cc - 4*r_cc^2) + (x[,"n_02"] + x[,"n_33"])*log(6 - 3*r_cc - 2*r_cc^2) + (x[,"n_03"] + x[,"n_32"])*log(1 + 7*r_cc - 2*r_cc^2) + (x[,"n_12"] + x[,"n_23"])*log(39 - 16*r_cc + 4*r_cc^2) + (x[,"n_13"] + x[,"n_22"])*log(27 + 8*r_cc + 4*r_cc^2)
logL_cr <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(4*log(2) + 2*log(5)) + (n01 + n04 + n31 + n34)*(log(3 - 2*r) + log(1 + 2*r)) + (n12 + n13 + n22 + n23)*log(63 + 4*r - 4*r^2) + (n02 + n14 + n21 + n33)*log(15 + 4*r^2) + (n03 + n11 + n24 + n32)*log(19 - 8*r + 4*r^2)
return(L)}
interlogL_cr <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_cr,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cr <- -6*(x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log10(2) - 4*(x[,"n_03"] + x[,"n_11"] + x[,"n_24"] + x[,"n_32"])*log10(2) - 4*(x[,"n_02"] + x[,"n_14"] + x[,"n_21"] + x[,"n_33"])*log10(2) - 2*(x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*log10(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*(log10(3 - 2*r_cr) + log10(1 + 2*r_cr)) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log10(63 + 4*r_cr - 4*r_cr^2) + (x[,"n_02"] + x[,"n_14"] + x[,"n_21"] + x[,"n_33"])*log10(15 + 4*r_cr^2) + (x[,"n_03"] + x[,"n_11"] + x[,"n_24"] + x[,"n_32"])*log10(19 - 8*r_cr + 4*r_cr^2)
logL_cr <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(4*log(2) + 2*log(5)) + (x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*(log(3 - 2*r_cr) + log(1 + 2*r_cr)) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log(63 + 4*r_cr - 4*r_cr^2) + (x[,"n_02"] + x[,"n_14"] + x[,"n_21"] + x[,"n_33"])*log(15 + 4*r_cr^2) + (x[,"n_03"] + x[,"n_11"] + x[,"n_24"] + x[,"n_32"])*log(19 - 8*r_cr + 4*r_cr^2)
logL_rc <- function(r,n02,n03,n11,n12,n13,n14,n21,n22,n23,n24,n32,n33,n01,n04,n31,n34) {
L <- (-n02 - n03 - n11 - 2*n12 - 2*n13 - n14 - n21 - 2*n22 - 2*n23 - n24 - n32 - n33)*log(2) + 2*(-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*log(5) + (n01 + n04 + n31 + n34)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n12 + n13 + n22 + n23)*log(15 + 4*r - 4*r^2) + (n02 + n14 + n21 + n33)*log(3 - 3*r + 2*r^2) + (n03 + n11 + n24 + n32)*log(2 - r + 2*r^2)
return(L)}
interlogL_rc <- function(n02,n03,n11,n12,n13,n14,n21,n22,n23,n24,n32,n33,n01,n04,n31,n34) {
optimize(logL_rc,c(0,0.5), n02,n03,n11,n12,n13,n14,n21,n22,n23,n24,n32,n33,n01,n04,n31,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_02"],x[,"n_03"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_01"],x[,"n_04"],x[,"n_31"],x[,"n_34"])
LOD_rc <- -4*(x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log10(2) - (x[,"n_03"] + x[,"n_11"] + x[,"n_24"] + x[,"n_32"])*log10(2) - (x[,"n_02"] + x[,"n_14"] + x[,"n_21"] + x[,"n_33"])*log10(2) + 2*(x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*log10(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc))) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log10(15 + 4*r_rc - 4*r_rc^2) + (x[,"n_02"] + x[,"n_14"] + x[,"n_21"] + x[,"n_33"])*log10(3 - 3*r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_11"] + x[,"n_24"] + x[,"n_32"])*log10(2 - r_rc + 2*r_rc^2)
logL_rc <- (-x[,"n_02"] - x[,"n_03"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc))) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log(15 + 4*r_rc - 4*r_rc^2) + (x[,"n_02"] + x[,"n_14"] + x[,"n_21"] + x[,"n_33"])*log(3 - 3*r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_11"] + x[,"n_24"] + x[,"n_32"])*log(2 - r_rc + 2*r_rc^2)
logL_rr <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-2*n01 - 3*n02 - 3*n03 - 2*n04 - 2*n11 - 3*n12 - 3*n13 - 2*n14 - 2*n21 - 3*n22 - 3*n23 - 2*n24 - 2*n31 - 3*n32 - 3*n33 - 2*n34)*log(2) + 2*(-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*log(5) + (n04 + n31)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n01 + n34)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n03 + n32)*log(11 - 4*r - 4*r^2) + (n02 + n33)*log(3 + 12*r - 4*r^2) + (n14 + n21)*log(6 - 3*r - 2*r^2) + (n11 + n24)*log(1 + 7*r - 2*r^2) + (n13 + n22)*log(39 - 16*r + 4*r^2) + (n12 + n23)*log(27 + 8*r + 4*r^2)
return(L)}
interlogL_rr <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_rr,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rr <- -2*(x[,"n_14"] + x[,"n_21"])*log10(2) - 5*(x[,"n_13"] + x[,"n_22"])*log10(2) - 5*(x[,"n_12"] + x[,"n_23"])*log10(2) - 2*(x[,"n_11"] + x[,"n_24"])*log10(2) - 3*(x[,"n_03"] + x[,"n_32"])*log10(2) - 3*(x[,"n_02"] + x[,"n_33"])*log10(2) + (x[,"n_04"] + x[,"n_31"])*(log10(3 - 2*r_rr) + log10(pmax(1e-6,1 - r_rr))) + (x[,"n_01"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_03"] + x[,"n_32"])*log10(11 - 4*r_rr - 4*r_rr^2) + (x[,"n_02"] + x[,"n_33"])*log10(3 + 12*r_rr - 4*r_rr^2) + (x[,"n_14"] + x[,"n_21"])*log10(6 - 3*r_rr - 2*r_rr^2) + (x[,"n_11"] + x[,"n_24"])*log10(1 + 7*r_rr - 2*r_rr^2) + (x[,"n_13"] + x[,"n_22"])*log10(39 - 16*r_rr + 4*r_rr^2) + (x[,"n_12"] + x[,"n_23"])*log10(27 + 8*r_rr + 4*r_rr^2)
logL_rr <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 2*x[,"n_34"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_04"] + x[,"n_31"])*(log(3 - 2*r_rr) + log(pmax(1e-6,1 - r_rr))) + (x[,"n_01"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_03"] + x[,"n_32"])*log(11 - 4*r_rr - 4*r_rr^2) + (x[,"n_02"] + x[,"n_33"])*log(3 + 12*r_rr - 4*r_rr^2) + (x[,"n_14"] + x[,"n_21"])*log(6 - 3*r_rr - 2*r_rr^2) + (x[,"n_11"] + x[,"n_24"])*log(1 + 7*r_rr - 2*r_rr^2) + (x[,"n_13"] + x[,"n_22"])*log(39 - 16*r_rr + 4*r_rr^2) + (x[,"n_12"] + x[,"n_23"])*log(27 + 8*r_rr + 4*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_1.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n22,n24,n30,n31,n33,n34,n40,n41,n42,n43,n44,n12,n21,n23,n32) {
L <- (-3*n00 - n01 - 2*n02 - n03 - 3*n04 - n10 - 2*n11 - 2*n13 - n14 - 2*n20 - 2*n22 - 2*n24 - n30 - 2*n31 - 2*n33 - n34 - 3*n40 - n41 - 2*n42 - n43 - 3*n44)*log(2) + (n02 + n12 + n20 + n21 + n22 + n23 + n24 + n32 + n42)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + 4*(n00 + n44)*log(pmax(1e-6,1 - r)) + 4*(n04 + n40)*log(pmax(1e-6,r)) + (n01 + n10 + n34 + n43)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n20 + n24 + n42)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n03 + n14 + n30 + n41)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n13 + n31)*(2*log(pmax(1e-6,r)) + log(9 - 12*r + 8*r^2)) + (n11 + n33)*(2*log(pmax(1e-6,1 - r)) + log(5 - 4*r + 8*r^2)) + (n12 + n21 + n23 + n32)*(log(pmax(1e-6,r)) + log(1 - 2*r + 2*r^2 - r^3)) + n22*log(3 - 8*r + 14*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n22,n24,n30,n31,n33,n34,n40,n41,n42,n43,n44,n12,n21,n23,n32) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n22,n24,n30,n31,n33,n34,n40,n41,n42,n43,n44,n12,n21,n23,n32, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_22"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_12"],x[,"n_21"],x[,"n_23"],x[,"n_32"])
LOD_cc <- 4*(x[,"n_04"] + x[,"n_40"])*log10(2) + 4*(x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*log10(2) + 4*(x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(2) + 4*(x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log10(2) + 4*(x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_31"])*(-2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-2*log10(2) + log10(5)) - x[,"n_22"]*(-3*log10(2) + log10(11)) + 4*(x[,"n_00"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_04"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(9 - 12*r_cc + 8*r_cc^2)) + (x[,"n_11"] + x[,"n_33"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(5 - 4*r_cc + 8*r_cc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(1 - 2*r_cc + 2*r_cc^2 - r_cc^3)) + x[,"n_22"]*log10(3 - 8*r_cc + 14*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cc <- (-3*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 3*x[,"n_04"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_13"] - x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_22"] - 2*x[,"n_24"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_33"] - x[,"n_34"] - 3*x[,"n_40"] - x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 3*x[,"n_44"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"] + x[,"n_42"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + 4*(x[,"n_00"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_04"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(9 - 12*r_cc + 8*r_cc^2)) + (x[,"n_11"] + x[,"n_33"])*(2*log(pmax(1e-6,1 - r_cc)) + log(5 - 4*r_cc + 8*r_cc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(1 - 2*r_cc + 2*r_cc^2 - r_cc^3)) + x[,"n_22"]*log(3 - 8*r_cc + 14*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cm2c <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n10 - n11 - 3*n12 - n13 - 3*n14 - 3*n20 - 3*n21 - 2*n22 - 3*n23 - 3*n24 - 3*n30 - n31 - 3*n32 - n33 - 3*n34 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n13 - n14 - n20 - n22 - n24 - n30 - n31 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n00 + n44)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n40)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n03 + n14 + n30 + n41)*(2*log(pmax(1e-6,r)) + log(5 - 6*r + 4*r^2)) + (n01 + n10 + n34 + n43)*(2*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 4*r^2)) + (n02 + n20 + n24 + n42)*(log(pmax(1e-6,r)) + log(7 - 13*r + 12*r^2 - 6*r^3)) + (n13 + n31)*(log(pmax(1e-6,r)) + log(5 - 5*r + 7*r^2 - 4*r^3)) + n22*log(12 + 5*r - 23*r^2 + 36*r^3 - 18*r^4) + (n11 + n33)*log(3 - 8*r^2 + 9*r^3 - 4*r^4) + (n12 + n21 + n23 + n32)*log(5 - 4*r + 12*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_cm2c <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_cm2c,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cm2c <- 4*(x[,"n_04"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-2*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-log10(2) + 2*log10(3)) - (x[,"n_13"] + x[,"n_31"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-3*log10(2) + log10(11)) - x[,"n_22"]*(-3*log10(2) + log10(97)) + (x[,"n_00"] + x[,"n_44"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_04"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cm2c)) + 3*log10(pmax(1e-6,r_cm2c))) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(5 - 6*r_cm2c + 4*r_cm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 - 2*r_cm2c + 4*r_cm2c^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm2c)) + log10(7 - 13*r_cm2c + 12*r_cm2c^2 - 6*r_cm2c^3)) + (x[,"n_13"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm2c)) + log10(5 - 5*r_cm2c + 7*r_cm2c^2 - 4*r_cm2c^3)) + x[,"n_22"]*log10(12 + 5*r_cm2c - 23*r_cm2c^2 + 36*r_cm2c^3 - 18*r_cm2c^4) + (x[,"n_11"] + x[,"n_33"])*log10(3 - 8*r_cm2c^2 + 9*r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(5 - 4*r_cm2c + 12*r_cm2c^2 - 16*r_cm2c^3 + 8*r_cm2c^4)
logL_cm2c <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_10"] - x[,"n_11"] - 3*x[,"n_12"] - x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_30"] - x[,"n_31"] - 3*x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_22"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_00"] + x[,"n_44"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_04"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cm2c)) + 3*log(pmax(1e-6,r_cm2c))) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm2c)) + log(5 - 6*r_cm2c + 4*r_cm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(3 - 2*r_cm2c + 4*r_cm2c^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log(pmax(1e-6,r_cm2c)) + log(7 - 13*r_cm2c + 12*r_cm2c^2 - 6*r_cm2c^3)) + (x[,"n_13"] + x[,"n_31"])*(log(pmax(1e-6,r_cm2c)) + log(5 - 5*r_cm2c + 7*r_cm2c^2 - 4*r_cm2c^3)) + x[,"n_22"]*log(12 + 5*r_cm2c - 23*r_cm2c^2 + 36*r_cm2c^3 - 18*r_cm2c^4) + (x[,"n_11"] + x[,"n_33"])*log(3 - 8*r_cm2c^2 + 9*r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(5 - 4*r_cm2c + 12*r_cm2c^2 - 16*r_cm2c^3 + 8*r_cm2c^4)
logL_cm1c <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-3*n00 - n01 - 3*n02 - n03 - 3*n04 - n10 - 2*n11 - n12 - 2*n13 - n14 - 3*n20 - n21 - n22 - n23 - 3*n24 - n30 - 2*n31 - n32 - 2*n33 - n34 - 3*n40 - n41 - 3*n42 - n43 - 3*n44)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n13 - n14 - n20 - n22 - n24 - n30 - n31 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n00 + n04 + n40 + n44)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n03 + n10 + n14 + n30 + n34 + n41 + n43)*(log(pmax(1e-6,r)) + log(1 - 2*r + 2*r^2 - r^3)) + (n12 + n21 + n23 + n32)*log(1 + r - 3*r^2 + 4*r^3 - 2*r^4) + (n02 + n20 + n24 + n42)*log(3 - 8*r + 14*r^2 - 12*r^3 + 6*r^4) + (n11 + n33)*log(6 - 8*r + 13*r^2 - 16*r^3 + 8*r^4) + (n13 + n31)*log(3 - 2*r + 13*r^2 - 16*r^3 + 8*r^4) + n22*log(6 - 2*r + 11*r^2 - 18*r^3 + 9*r^4)
return(L)}
interlogL_cm1c <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_cm1c,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cm1c <- 4*(x[,"n_00"] + x[,"n_04"] + x[,"n_40"] + x[,"n_44"])*log10(2) - (x[,"n_01"] + x[,"n_03"] + x[,"n_10"] + x[,"n_14"] + x[,"n_30"] + x[,"n_34"] + x[,"n_41"] + x[,"n_43"])*(-4*log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_13"] + x[,"n_31"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-3*log10(2) + log10(11)) - x[,"n_22"]*(-4*log10(2) + log10(97)) + (x[,"n_00"] + x[,"n_04"] + x[,"n_40"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_03"] + x[,"n_10"] + x[,"n_14"] + x[,"n_30"] + x[,"n_34"] + x[,"n_41"] + x[,"n_43"])*(log10(pmax(1e-6,r_cm1c)) + log10(1 - 2*r_cm1c + 2*r_cm1c^2 - r_cm1c^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(1 + r_cm1c - 3*r_cm1c^2 + 4*r_cm1c^3 - 2*r_cm1c^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(3 - 8*r_cm1c + 14*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_11"] + x[,"n_33"])*log10(6 - 8*r_cm1c + 13*r_cm1c^2 - 16*r_cm1c^3 + 8*r_cm1c^4) + (x[,"n_13"] + x[,"n_31"])*log10(3 - 2*r_cm1c + 13*r_cm1c^2 - 16*r_cm1c^3 + 8*r_cm1c^4) + x[,"n_22"]*log10(6 - 2*r_cm1c + 11*r_cm1c^2 - 18*r_cm1c^3 + 9*r_cm1c^4)
logL_cm1c <- (-3*x[,"n_00"] - x[,"n_01"] - 3*x[,"n_02"] - x[,"n_03"] - 3*x[,"n_04"] - x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - 3*x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 3*x[,"n_24"] - x[,"n_30"] - 2*x[,"n_31"] - x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"] - 3*x[,"n_40"] - x[,"n_41"] - 3*x[,"n_42"] - x[,"n_43"] - 3*x[,"n_44"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_22"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_00"] + x[,"n_04"] + x[,"n_40"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_03"] + x[,"n_10"] + x[,"n_14"] + x[,"n_30"] + x[,"n_34"] + x[,"n_41"] + x[,"n_43"])*(log(pmax(1e-6,r_cm1c)) + log(1 - 2*r_cm1c + 2*r_cm1c^2 - r_cm1c^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(1 + r_cm1c - 3*r_cm1c^2 + 4*r_cm1c^3 - 2*r_cm1c^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(3 - 8*r_cm1c + 14*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_11"] + x[,"n_33"])*log(6 - 8*r_cm1c + 13*r_cm1c^2 - 16*r_cm1c^3 + 8*r_cm1c^4) + (x[,"n_13"] + x[,"n_31"])*log(3 - 2*r_cm1c + 13*r_cm1c^2 - 16*r_cm1c^3 + 8*r_cm1c^4) + x[,"n_22"]*log(6 - 2*r_cm1c + 11*r_cm1c^2 - 18*r_cm1c^3 + 9*r_cm1c^4)
logL_cr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n10 - n11 - 3*n12 - n13 - 3*n14 - 3*n20 - 3*n21 - 2*n22 - 3*n23 - 3*n24 - 3*n30 - n31 - 3*n32 - n33 - 3*n34 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44)*log(2) + (n02 + n12 + n20 + n21 + n22 + n23 + n24 + n32 + n42)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n04 + n40)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n44)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n10 + n34 + n43)*(2*log(pmax(1e-6,r)) + log(3 - 6*r + 4*r^2)) + (n03 + n14 + n30 + n41)*(2*log(pmax(1e-6,1 - r)) + log(1 - 2*r + 4*r^2)) + n22*(log(pmax(1e-6,r)) + log(7 - 13*r + 12*r^2 - 6*r^3)) + (n11 + n33)*(log(pmax(1e-6,r)) + log(3 - 6*r + 7*r^2 - 4*r^3)) + (n13 + n31)*(log(pmax(1e-6,r)) + log(4 - 9*r + 9*r^2 - 4*r^3)) + (n02 + n20 + n24 + n42)*(log(pmax(1e-6,r)) + log(1 - 3*r + 4*r^2 - 2*r^3)) + (n12 + n21 + n23 + n32)*log(3 - 8*r + 16*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cr <- 4*(x[,"n_04"] + x[,"n_40"])*log10(2) + 2*(x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*log10(2) + 3*(x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(2) + 2*(x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log10(2) + 4*(x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_31"])*(-3*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-3*log10(2) + log10(5)) - x[,"n_22"]*(-3*log10(2) + log10(11)) + (x[,"n_04"] + x[,"n_40"])*(3*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cr)) + 3*log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(2*log10(pmax(1e-6,r_cr)) + log10(3 - 6*r_cr + 4*r_cr^2)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(1 - 2*r_cr + 4*r_cr^2)) + x[,"n_22"]*(log10(pmax(1e-6,r_cr)) + log10(7 - 13*r_cr + 12*r_cr^2 - 6*r_cr^3)) + (x[,"n_11"] + x[,"n_33"])*(log10(pmax(1e-6,r_cr)) + log10(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3)) + (x[,"n_13"] + x[,"n_31"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 9*r_cr + 9*r_cr^2 - 4*r_cr^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log10(pmax(1e-6,r_cr)) + log10(1 - 3*r_cr + 4*r_cr^2 - 2*r_cr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(3 - 8*r_cr + 16*r_cr^2 - 16*r_cr^3 + 8*r_cr^4)
logL_cr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_10"] - x[,"n_11"] - 3*x[,"n_12"] - x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_30"] - x[,"n_31"] - 3*x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"] + x[,"n_42"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(3*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cr)) + 3*log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(2*log(pmax(1e-6,r_cr)) + log(3 - 6*r_cr + 4*r_cr^2)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_cr)) + log(1 - 2*r_cr + 4*r_cr^2)) + x[,"n_22"]*(log(pmax(1e-6,r_cr)) + log(7 - 13*r_cr + 12*r_cr^2 - 6*r_cr^3)) + (x[,"n_11"] + x[,"n_33"])*(log(pmax(1e-6,r_cr)) + log(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3)) + (x[,"n_13"] + x[,"n_31"])*(log(pmax(1e-6,r_cr)) + log(4 - 9*r_cr + 9*r_cr^2 - 4*r_cr^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log(pmax(1e-6,r_cr)) + log(1 - 3*r_cr + 4*r_cr^2 - 2*r_cr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(3 - 8*r_cr + 16*r_cr^2 - 16*r_cr^3 + 8*r_cr^4)
logL_rc <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n10 - n11 - 3*n12 - n13 - 3*n14 - 3*n20 - 3*n21 - 2*n22 - 3*n23 - 3*n24 - 3*n30 - n31 - 3*n32 - n33 - 3*n34 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44)*log(2) + (n02 + n12 + n20 + n21 + n22 + n23 + n24 + n32 + n42)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n04 + n40)*(log(3 - r) + 3*log(pmax(1e-6,r))) + (n00 + n44)*(3*log(pmax(1e-6,1 - r)) + log(2 + r)) + (n03 + n14 + n30 + n41)*(2*log(pmax(1e-6,r)) + log(9 - 10*r + 4*r^2)) + (n01 + n10 + n34 + n43)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 4*r^2)) + (n13 + n31)*(log(pmax(1e-6,r)) + log(9 - 9*r + 9*r^2 - 4*r^3)) + (n02 + n20 + n24 + n42)*(log(pmax(1e-6,r)) + log(3 - 5*r + 4*r^2 - 2*r^3)) + n22*log(6 + 5*r - 11*r^2 + 12*r^3 - 6*r^4) + (n11 + n33)*log(5 - 2*r - 6*r^2 + 7*r^3 - 4*r^4) + (n12 + n21 + n23 + n32)*log(9 - 8*r + 16*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rc <- -((x[,"n_04"] + x[,"n_40"])*(-4*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_44"])*(-4*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_31"])*(-3*log10(2) + 2*log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-3*log10(2) + 2*log10(5)) - x[,"n_22"]*(-3*log10(2) + log10(5) + log10(11)) + (x[,"n_04"] + x[,"n_40"])*(log10(3 - r_rc) + 3*log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_44"])*(3*log10(pmax(1e-6,1 - r_rc)) + log10(2 + r_rc)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_rc)) + log10(9 - 10*r_rc + 4*r_rc^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(3 + 2*r_rc + 4*r_rc^2)) + (x[,"n_13"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(9 - 9*r_rc + 9*r_rc^2 - 4*r_rc^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log10(pmax(1e-6,r_rc)) + log10(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3)) + x[,"n_22"]*log10(6 + 5*r_rc - 11*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_11"] + x[,"n_33"])*log10(5 - 2*r_rc - 6*r_rc^2 + 7*r_rc^3 - 4*r_rc^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(9 - 8*r_rc + 16*r_rc^2 - 16*r_rc^3 + 8*r_rc^4)
logL_rc <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_10"] - x[,"n_11"] - 3*x[,"n_12"] - x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_30"] - x[,"n_31"] - 3*x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"] + x[,"n_42"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(log(3 - r_rc) + 3*log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_44"])*(3*log(pmax(1e-6,1 - r_rc)) + log(2 + r_rc)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log(pmax(1e-6,r_rc)) + log(9 - 10*r_rc + 4*r_rc^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_rc)) + log(3 + 2*r_rc + 4*r_rc^2)) + (x[,"n_13"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(9 - 9*r_rc + 9*r_rc^2 - 4*r_rc^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log(pmax(1e-6,r_rc)) + log(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3)) + x[,"n_22"]*log(6 + 5*r_rc - 11*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_11"] + x[,"n_33"])*log(5 - 2*r_rc - 6*r_rc^2 + 7*r_rc^3 - 4*r_rc^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(9 - 8*r_rc + 16*r_rc^2 - 16*r_rc^3 + 8*r_rc^4)
logL_rm2c <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-3*n00 - 2*n01 - 3*n02 - 2*n03 - 3*n04 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 3*n20 - 2*n21 - n22 - 2*n23 - 3*n24 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 3*n40 - 2*n41 - 3*n42 - 2*n43 - 3*n44)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n13 - n14 - n20 - n22 - n24 - n30 - n31 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n04 + n40)*(log(pmax(1e-6,1 - r)) + log(3 - r) + 2*log(pmax(1e-6,r))) + (n00 + n44)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n14 + n30 + n41)*(log(pmax(1e-6,r)) + log(6 - 7*r + 6*r^2 - 2*r^3)) + (n12 + n21 + n23 + n32)*log(11 + 2*r - 6*r^2 + 8*r^3 - 4*r^4) + (n01 + n10 + n34 + n43)*log(3 - 2*r - r^2 + 2*r^3 - 2*r^4) + (n02 + n20 + n24 + n42)*log(9 - 10*r + 16*r^2 - 12*r^3 + 6*r^4) + (n13 + n31)*log(18 - 4*r + 19*r^2 - 20*r^3 + 8*r^4) + (n11 + n33)*log(21 - 6*r + 7*r^2 - 12*r^3 + 8*r^4) + n22*log(30 - r + 10*r^2 - 18*r^3 + 9*r^4)
return(L)}
interlogL_rm2c <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rm2c,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm2c <- -((x[,"n_04"] + x[,"n_40"])*(-4*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_44"])*(-4*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_31"])*(-2*log10(2) + log10(3) + 2*log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-2*log10(2) + log10(3) + 2*log10(5)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-3*log10(2) + log10(5) + log10(11)) - x[,"n_22"]*(-4*log10(2) + log10(5) + log10(97)) + (x[,"n_04"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_rm2c)) + log10(3 - r_rm2c) + 2*log10(pmax(1e-6,r_rm2c))) + (x[,"n_00"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c)) + log10(2 + r_rm2c)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log10(pmax(1e-6,r_rm2c)) + log10(6 - 7*r_rm2c + 6*r_rm2c^2 - 2*r_rm2c^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(11 + 2*r_rm2c - 6*r_rm2c^2 + 8*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log10(3 - 2*r_rm2c - r_rm2c^2 + 2*r_rm2c^3 - 2*r_rm2c^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(9 - 10*r_rm2c + 16*r_rm2c^2 - 12*r_rm2c^3 + 6*r_rm2c^4) + (x[,"n_13"] + x[,"n_31"])*log10(18 - 4*r_rm2c + 19*r_rm2c^2 - 20*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_11"] + x[,"n_33"])*log10(21 - 6*r_rm2c + 7*r_rm2c^2 - 12*r_rm2c^3 + 8*r_rm2c^4) + x[,"n_22"]*log10(30 - r_rm2c + 10*r_rm2c^2 - 18*r_rm2c^3 + 9*r_rm2c^4)
logL_rm2c <- (-3*x[,"n_00"] - 2*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 3*x[,"n_40"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 2*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_22"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_rm2c)) + log(3 - r_rm2c) + 2*log(pmax(1e-6,r_rm2c))) + (x[,"n_00"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c)) + log(2 + r_rm2c)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log(pmax(1e-6,r_rm2c)) + log(6 - 7*r_rm2c + 6*r_rm2c^2 - 2*r_rm2c^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(11 + 2*r_rm2c - 6*r_rm2c^2 + 8*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log(3 - 2*r_rm2c - r_rm2c^2 + 2*r_rm2c^3 - 2*r_rm2c^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(9 - 10*r_rm2c + 16*r_rm2c^2 - 12*r_rm2c^3 + 6*r_rm2c^4) + (x[,"n_13"] + x[,"n_31"])*log(18 - 4*r_rm2c + 19*r_rm2c^2 - 20*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_11"] + x[,"n_33"])*log(21 - 6*r_rm2c + 7*r_rm2c^2 - 12*r_rm2c^3 + 8*r_rm2c^4) + x[,"n_22"]*log(30 - r_rm2c + 10*r_rm2c^2 - 18*r_rm2c^3 + 9*r_rm2c^4)
logL_rm1c <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n10 - n11 - 3*n12 - n13 - 3*n14 - 3*n20 - 3*n21 - 2*n22 - 3*n23 - 3*n24 - 3*n30 - n31 - 3*n32 - n33 - 3*n34 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n13 - n14 - n20 - n22 - n24 - n30 - n31 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n04 + n40)*(2*log(pmax(1e-6,1 - r)) + log(3 - r) + log(pmax(1e-6,r))) + (n00 + n44)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(2 + r)) + (n01 + n10 + n34 + n43)*(log(pmax(1e-6,r)) + log(8 - r - 2*r^2 + 4*r^3)) + n22*log(60 + 7*r - 25*r^2 + 36*r^3 - 18*r^4) + (n02 + n20 + n24 + n42)*log(6 + 5*r - 11*r^2 + 12*r^3 - 6*r^4) + (n11 + n33)*log(6 + 7*r - 2*r^2 + 5*r^3 - 4*r^4) + (n13 + n31)*log(12 - 2*r - 11*r^2 + 11*r^3 - 4*r^4) + (n03 + n14 + n30 + n41)*log(9 - 16*r + 17*r^2 - 14*r^3 + 4*r^4) + (n12 + n21 + n23 + n32)*log(23 - 4*r + 12*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_rm1c <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rm1c,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm1c <- -((x[,"n_04"] + x[,"n_40"])*(-4*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_44"])*(-4*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-log10(2) + 2*log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_31"])*(-3*log10(2) + log10(3) + 2*log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-3*log10(2) + log10(3) + 2*log10(5)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-3*log10(2) + log10(5) + log10(11)) - x[,"n_22"]*(-3*log10(2) + log10(5) + log10(97)) + (x[,"n_04"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 - r_rm1c) + log10(pmax(1e-6,r_rm1c))) + (x[,"n_00"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_rm1c)) + 2*log10(pmax(1e-6,r_rm1c)) + log10(2 + r_rm1c)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(log10(pmax(1e-6,r_rm1c)) + log10(8 - r_rm1c - 2*r_rm1c^2 + 4*r_rm1c^3)) + x[,"n_22"]*log10(60 + 7*r_rm1c - 25*r_rm1c^2 + 36*r_rm1c^3 - 18*r_rm1c^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(6 + 5*r_rm1c - 11*r_rm1c^2 + 12*r_rm1c^3 - 6*r_rm1c^4) + (x[,"n_11"] + x[,"n_33"])*log10(6 + 7*r_rm1c - 2*r_rm1c^2 + 5*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_13"] + x[,"n_31"])*log10(12 - 2*r_rm1c - 11*r_rm1c^2 + 11*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*log10(9 - 16*r_rm1c + 17*r_rm1c^2 - 14*r_rm1c^3 + 4*r_rm1c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(23 - 4*r_rm1c + 12*r_rm1c^2 - 16*r_rm1c^3 + 8*r_rm1c^4)
logL_rm1c <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_10"] - x[,"n_11"] - 3*x[,"n_12"] - x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_30"] - x[,"n_31"] - 3*x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_22"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(3 - r_rm1c) + log(pmax(1e-6,r_rm1c))) + (x[,"n_00"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_rm1c)) + 2*log(pmax(1e-6,r_rm1c)) + log(2 + r_rm1c)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(log(pmax(1e-6,r_rm1c)) + log(8 - r_rm1c - 2*r_rm1c^2 + 4*r_rm1c^3)) + x[,"n_22"]*log(60 + 7*r_rm1c - 25*r_rm1c^2 + 36*r_rm1c^3 - 18*r_rm1c^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(6 + 5*r_rm1c - 11*r_rm1c^2 + 12*r_rm1c^3 - 6*r_rm1c^4) + (x[,"n_11"] + x[,"n_33"])*log(6 + 7*r_rm1c - 2*r_rm1c^2 + 5*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_13"] + x[,"n_31"])*log(12 - 2*r_rm1c - 11*r_rm1c^2 + 11*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*log(9 - 16*r_rm1c + 17*r_rm1c^2 - 14*r_rm1c^3 + 4*r_rm1c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(23 - 4*r_rm1c + 12*r_rm1c^2 - 16*r_rm1c^3 + 8*r_rm1c^4)
logL_rr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-3*n00 - 2*n01 - 2*n02 - 2*n03 - 3*n04 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 3*n40 - 2*n41 - 2*n42 - 2*n43 - 3*n44)*log(2) + (n02 + n12 + n20 + n21 + n22 + n23 + n24 + n32 + n42)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n04 + n40)*(3*log(pmax(1e-6,1 - r)) + log(3 - r)) + (n00 + n44)*(3*log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n14 + n30 + n41)*(2*log(pmax(1e-6,1 - r)) + log(1 + 4*r - 2*r^2)) + (n11 + n33)*(log(pmax(1e-6,r)) + log(12 + 3*r - 8*r^2 + 8*r^3)) + (n12 + n21 + n23 + n32)*log(3 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n01 + n10 + n34 + n43)*log(3*r^2 - 2*r^4) + (n02 + n20 + n24 + n42)*log(r - 2*r^3 + r^4) + n22*log(9 - 10*r + 16*r^2 - 12*r^3 + 6*r^4) + (n13 + n31)*log(15 - 26*r + 27*r^2 - 24*r^3 + 8*r^4)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rr <- (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(3*log10(2) - log10(5)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(4*log10(2) - log10(5)) - (x[,"n_04"] + x[,"n_40"])*(-4*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_44"])*(-4*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-3*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_31"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-2*log10(2) + 2*log10(5)) - x[,"n_22"]*(-3*log10(2) + log10(5) + log10(11)) + (x[,"n_04"] + x[,"n_40"])*(3*log10(pmax(1e-6,1 - r_rr)) + log10(3 - r_rr)) + (x[,"n_00"] + x[,"n_44"])*(3*log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(1 + 4*r_rr - 2*r_rr^2)) + (x[,"n_11"] + x[,"n_33"])*(log10(pmax(1e-6,r_rr)) + log10(12 + 3*r_rr - 8*r_rr^2 + 8*r_rr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(3 + 4*r_rr - 8*r_rr^2 + 8*r_rr^3 - 4*r_rr^4) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log10(3*r_rr^2 - 2*r_rr^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(r_rr - 2*r_rr^3 + r_rr^4) + x[,"n_22"]*log10(9 - 10*r_rr + 16*r_rr^2 - 12*r_rr^3 + 6*r_rr^4) + (x[,"n_13"] + x[,"n_31"])*log10(15 - 26*r_rr + 27*r_rr^2 - 24*r_rr^3 + 8*r_rr^4)
logL_rr <- (-3*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 3*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"] + x[,"n_42"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(3*log(pmax(1e-6,1 - r_rr)) + log(3 - r_rr)) + (x[,"n_00"] + x[,"n_44"])*(3*log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(1 + 4*r_rr - 2*r_rr^2)) + (x[,"n_11"] + x[,"n_33"])*(log(pmax(1e-6,r_rr)) + log(12 + 3*r_rr - 8*r_rr^2 + 8*r_rr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(3 + 4*r_rr - 8*r_rr^2 + 8*r_rr^3 - 4*r_rr^4) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log(3*r_rr^2 - 2*r_rr^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(r_rr - 2*r_rr^3 + r_rr^4) + x[,"n_22"]*log(9 - 10*r_rr + 16*r_rr^2 - 12*r_rr^3 + 6*r_rr^4) + (x[,"n_13"] + x[,"n_31"])*log(15 - 26*r_rr + 27*r_rr^2 - 24*r_rr^3 + 8*r_rr^4)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_cr,r_rc,r_rm2c,r_rm1c,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_cr,LOD_rc,LOD_rm2c,LOD_rm1c,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_cr,logL_rc,logL_rm2c,logL_rm1c,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","coupling repulsion","repulsion coupling","repulsion mixed2c","repulsion mixed1c","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_1.4 <- function(x){
logL_cc <- function(r,n02,n03,n42,n43,n01,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n44) {
L <- (n02 + n03 + n42 + n43)*log(3) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(3*log(2) + log(5)) + (n21 + n24)*log(7) + 3*(n01 + n44)*log(pmax(1e-6,1 - r)) + 3*(n04 + n41)*log(pmax(1e-6,r)) + (n21 + n24)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n43)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n14 + n31)*(log(5 - 2*r) + 2*log(pmax(1e-6,r))) + (n03 + n42)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n34)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r)) + (n13 + n32)*(log(pmax(1e-6,r)) + log(12 - 11*r + 6*r^2)) + (n22 + n23)*log(9 - 7*r + 7*r^2) + (n12 + n33)*log(7 - 8*r + 7*r^2 - 6*r^3)
return(L)}
interlogL_cc <- function(n02,n03,n42,n43,n01,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n44) {
optimize(logL_cc,c(0,0.5), n02,n03,n42,n43,n01,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_02"],x[,"n_03"],x[,"n_42"],x[,"n_43"],x[,"n_01"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_44"])
LOD_cc <- 2*(x[,"n_21"] + x[,"n_24"])*log10(2) - 2*(x[,"n_13"] + x[,"n_32"])*log10(2) - 2*(x[,"n_12"] + x[,"n_33"])*log10(2) + 3*(x[,"n_04"] + x[,"n_41"])*log10(2) + 3*(x[,"n_03"] + x[,"n_42"])*log10(2) + 3*(x[,"n_02"] + x[,"n_43"])*log10(2) + 3*(x[,"n_01"] + x[,"n_44"])*log10(2) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(29)) + 3*(x[,"n_01"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_04"] + x[,"n_41"])*log10(pmax(1e-6,r_cc)) + (x[,"n_21"] + x[,"n_24"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_31"])*(log10(5 - 2*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 2*r_cc)) + (x[,"n_13"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(12 - 11*r_cc + 6*r_cc^2)) + (x[,"n_22"] + x[,"n_23"])*log10(9 - 7*r_cc + 7*r_cc^2) + (x[,"n_12"] + x[,"n_33"])*log10(7 - 8*r_cc + 7*r_cc^2 - 6*r_cc^3)
logL_cc <- (x[,"n_02"] + x[,"n_03"] + x[,"n_42"] + x[,"n_43"])*log(3) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(3*log(2) + log(5)) + (x[,"n_21"] + x[,"n_24"])*log(7) + 3*(x[,"n_01"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_04"] + x[,"n_41"])*log(pmax(1e-6,r_cc)) + (x[,"n_21"] + x[,"n_24"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_31"])*(log(5 - 2*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 2*r_cc)) + (x[,"n_13"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(12 - 11*r_cc + 6*r_cc^2)) + (x[,"n_22"] + x[,"n_23"])*log(9 - 7*r_cc + 7*r_cc^2) + (x[,"n_12"] + x[,"n_33"])*log(7 - 8*r_cc + 7*r_cc^2 - 6*r_cc^3)
logL_cm2c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-2*n01 - 3*n02 - 3*n03 - 2*n04 - 2*n11 - 3*n12 - 3*n13 - 2*n14 - 2*n21 - 3*n22 - 3*n23 - 2*n24 - 2*n31 - 3*n32 - 3*n33 - 2*n34 - 2*n41 - 3*n42 - 3*n43 - 2*n44)*log(2) + (-n01 - n02 - n03 - n04 - n11 - n14 - n21 - n22 - n23 - n24 - n31 - n34 - n41 - n42 - n43 - n44)*log(3) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*log(5) + (n01 + n44)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n41)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n22 + n23)*log(21 + 2*r - 2*r^2) + (n21 + n24)*log(3 - r + r^2) + (n14 + n31)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n03 + n42)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2)) + (n02 + n43)*log(3 - 7*r + 10*r^2 - 6*r^3) + (n13 + n32)*log(3 + 6*r^2 - 4*r^3) + (n11 + n34)*log(3 - 4*r + 3*r^2 - 2*r^3) + (n12 + n33)*log(5 - 6*r^2 + 4*r^3)
return(L)}
interlogL_cm2c <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_cm2c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cm2c <- -2*(x[,"n_13"] + x[,"n_32"])*log10(2) - 2*(x[,"n_12"] + x[,"n_33"])*log10(2) + 3*(x[,"n_04"] + x[,"n_41"])*log10(2) + 3*(x[,"n_01"] + x[,"n_44"])*log10(2) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_42"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_23"])*(-log10(2) + log10(43)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,1 - r_cm2c)) + 2*log10(pmax(1e-6,r_cm2c))) + (x[,"n_22"] + x[,"n_23"])*log10(21 + 2*r_cm2c - 2*r_cm2c^2) + (x[,"n_21"] + x[,"n_24"])*log10(3 - r_cm2c + r_cm2c^2) + (x[,"n_14"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm2c)) + log10(4 - 3*r_cm2c + 2*r_cm2c^2)) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm2c)) + log10(5 - 8*r_cm2c + 6*r_cm2c^2)) + (x[,"n_02"] + x[,"n_43"])*log10(3 - 7*r_cm2c + 10*r_cm2c^2 - 6*r_cm2c^3) + (x[,"n_13"] + x[,"n_32"])*log10(3 + 6*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_11"] + x[,"n_34"])*log10(3 - 4*r_cm2c + 3*r_cm2c^2 - 2*r_cm2c^3) + (x[,"n_12"] + x[,"n_33"])*log10(5 - 6*r_cm2c^2 + 4*r_cm2c^3)
logL_cm2c <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(3) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,1 - r_cm2c)) + 2*log(pmax(1e-6,r_cm2c))) + (x[,"n_22"] + x[,"n_23"])*log(21 + 2*r_cm2c - 2*r_cm2c^2) + (x[,"n_21"] + x[,"n_24"])*log(3 - r_cm2c + r_cm2c^2) + (x[,"n_14"] + x[,"n_31"])*(log(pmax(1e-6,r_cm2c)) + log(4 - 3*r_cm2c + 2*r_cm2c^2)) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,r_cm2c)) + log(5 - 8*r_cm2c + 6*r_cm2c^2)) + (x[,"n_02"] + x[,"n_43"])*log(3 - 7*r_cm2c + 10*r_cm2c^2 - 6*r_cm2c^3) + (x[,"n_13"] + x[,"n_32"])*log(3 + 6*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_11"] + x[,"n_34"])*log(3 - 4*r_cm2c + 3*r_cm2c^2 - 2*r_cm2c^3) + (x[,"n_12"] + x[,"n_33"])*log(5 - 6*r_cm2c^2 + 4*r_cm2c^3)
logL_cm1c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(3*log(2) + log(5)) + (n04 + n41)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n44)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n22 + n23)*log(6 + 5*r - 5*r^2) + (n11 + n34)*(log(pmax(1e-6,r)) + log(4 - 5*r + 2*r^2)) + (n02 + n43)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n21 + n24)*log(3 - 5*r + 5*r^2) + (n12 + n33)*log(6 - 8*r + 11*r^2 - 6*r^3) + (n03 + n42)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n14 + n31)*log(1 + r^2 - 2*r^3) + (n13 + n32)*log(3 + 4*r - 7*r^2 + 6*r^3)
return(L)}
interlogL_cm1c <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_cm1c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cm1c <- -2*(x[,"n_13"] + x[,"n_32"])*log10(2) - 2*(x[,"n_12"] + x[,"n_33"])*log10(2) + 3*(x[,"n_04"] + x[,"n_41"])*log10(2) + 3*(x[,"n_01"] + x[,"n_44"])*log10(2) + (x[,"n_03"] + x[,"n_42"])*(3*log10(2) - log10(3)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(3)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(7)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(29)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + log10(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_22"] + x[,"n_23"])*log10(6 + 5*r_cm1c - 5*r_cm1c^2) + (x[,"n_11"] + x[,"n_34"])*(log10(pmax(1e-6,r_cm1c)) + log10(4 - 5*r_cm1c + 2*r_cm1c^2)) + (x[,"n_02"] + x[,"n_43"])*(log10(pmax(1e-6,r_cm1c)) + log10(2 - 4*r_cm1c + 3*r_cm1c^2)) + (x[,"n_21"] + x[,"n_24"])*log10(3 - 5*r_cm1c + 5*r_cm1c^2) + (x[,"n_12"] + x[,"n_33"])*log10(6 - 8*r_cm1c + 11*r_cm1c^2 - 6*r_cm1c^3) + (x[,"n_03"] + x[,"n_42"])*log10(1 - 3*r_cm1c + 5*r_cm1c^2 - 3*r_cm1c^3) + (x[,"n_14"] + x[,"n_31"])*log10(1 + r_cm1c^2 - 2*r_cm1c^3) + (x[,"n_13"] + x[,"n_32"])*log10(3 + 4*r_cm1c - 7*r_cm1c^2 + 6*r_cm1c^3)
logL_cm1c <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(3*log(2) + log(5)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_cm1c)) + log(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_22"] + x[,"n_23"])*log(6 + 5*r_cm1c - 5*r_cm1c^2) + (x[,"n_11"] + x[,"n_34"])*(log(pmax(1e-6,r_cm1c)) + log(4 - 5*r_cm1c + 2*r_cm1c^2)) + (x[,"n_02"] + x[,"n_43"])*(log(pmax(1e-6,r_cm1c)) + log(2 - 4*r_cm1c + 3*r_cm1c^2)) + (x[,"n_21"] + x[,"n_24"])*log(3 - 5*r_cm1c + 5*r_cm1c^2) + (x[,"n_12"] + x[,"n_33"])*log(6 - 8*r_cm1c + 11*r_cm1c^2 - 6*r_cm1c^3) + (x[,"n_03"] + x[,"n_42"])*log(1 - 3*r_cm1c + 5*r_cm1c^2 - 3*r_cm1c^3) + (x[,"n_14"] + x[,"n_31"])*log(1 + r_cm1c^2 - 2*r_cm1c^3) + (x[,"n_13"] + x[,"n_32"])*log(3 + 4*r_cm1c - 7*r_cm1c^2 + 6*r_cm1c^3)
logL_rc <- function(r,n02,n03,n42,n43,n01,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n44) {
L <- (n02 + n03 + n42 + n43)*log(3) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(3*log(2) + 2*log(5)) + (n04 + n41)*(log(3 - r) + 2*log(pmax(1e-6,r))) + (n01 + n44)*(2*log(pmax(1e-6,1 - r)) + log(2 + r)) + (n22 + n23)*log(36 + r - r^2) + (n03 + n42)*(log(pmax(1e-6,r)) + log(2 - 2*r + r^2)) + (n21 + n24)*log(9 - r + r^2) + (n14 + n31)*(log(pmax(1e-6,r)) + log(12 - 5*r + 2*r^2)) + (n13 + n32)*log(18 + 11*r^2 - 6*r^3) + (n11 + n34)*log(9 - 8*r + r^2 - 2*r^3) + (n02 + n43)*log(1 - r + r^2 - r^3) + (n12 + n33)*log(23 - 4*r - 7*r^2 + 6*r^3)
return(L)}
interlogL_rc <- function(n02,n03,n42,n43,n01,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n44) {
optimize(logL_rc,c(0,0.5), n02,n03,n42,n43,n01,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_02"],x[,"n_03"],x[,"n_42"],x[,"n_43"],x[,"n_01"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_44"])
LOD_rc <- (x[,"n_02"] + x[,"n_43"])*(3*log10(2) - log10(5)) - (x[,"n_14"] + x[,"n_31"])*log10(5) - (x[,"n_11"] + x[,"n_34"])*log10(5) - (x[,"n_04"] + x[,"n_41"])*(-3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_44"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_32"])*(2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_33"])*(2*log10(2) + log10(5)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(5) + log10(29)) + (x[,"n_04"] + x[,"n_41"])*(log10(3 - r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(2 + r_rc)) + (x[,"n_22"] + x[,"n_23"])*log10(36 + r_rc - r_rc^2) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 2*r_rc + r_rc^2)) + (x[,"n_21"] + x[,"n_24"])*log10(9 - r_rc + r_rc^2) + (x[,"n_14"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(12 - 5*r_rc + 2*r_rc^2)) + (x[,"n_13"] + x[,"n_32"])*log10(18 + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_11"] + x[,"n_34"])*log10(9 - 8*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_02"] + x[,"n_43"])*log10(1 - r_rc + r_rc^2 - r_rc^3) + (x[,"n_12"] + x[,"n_33"])*log10(23 - 4*r_rc - 7*r_rc^2 + 6*r_rc^3)
logL_rc <- (x[,"n_02"] + x[,"n_03"] + x[,"n_42"] + x[,"n_43"])*log(3) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(3*log(2) + 2*log(5)) + (x[,"n_04"] + x[,"n_41"])*(log(3 - r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_rc)) + log(2 + r_rc)) + (x[,"n_22"] + x[,"n_23"])*log(36 + r_rc - r_rc^2) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,r_rc)) + log(2 - 2*r_rc + r_rc^2)) + (x[,"n_21"] + x[,"n_24"])*log(9 - r_rc + r_rc^2) + (x[,"n_14"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(12 - 5*r_rc + 2*r_rc^2)) + (x[,"n_13"] + x[,"n_32"])*log(18 + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_11"] + x[,"n_34"])*log(9 - 8*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_02"] + x[,"n_43"])*log(1 - r_rc + r_rc^2 - r_rc^3) + (x[,"n_12"] + x[,"n_33"])*log(23 - 4*r_rc - 7*r_rc^2 + 6*r_rc^3)
logL_rm2c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-2*n01 - 3*n02 - 3*n03 - 2*n04 - 2*n11 - 3*n12 - 3*n13 - 2*n14 - 2*n21 - 3*n22 - 3*n23 - 2*n24 - 2*n31 - 3*n32 - 3*n33 - 2*n34 - 2*n41 - 3*n42 - 3*n43 - 2*n44)*log(2) + (-n01 - n02 - n03 - n04 - n11 - n14 - n21 - n22 - n23 - n24 - n31 - n34 - n41 - n42 - n43 - n44)*log(3) + (-2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - n21 - n22 - n23 - n24 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n41 - 2*n42 - 2*n43 - 2*n44)*log(5) + (n04 + n41)*(log(pmax(1e-6,1 - r)) + log(3 - r) + log(pmax(1e-6,r))) + (n01 + n44)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n22 + n23)*log(21 + 2*r - 2*r^2) + (n21 + n24)*log(3 - r + r^2) + (n03 + n42)*log(9 - 11*r + 14*r^2 - 6*r^3) + (n12 + n33)*log(19 + 6*r^2 - 4*r^3) + (n14 + n31)*log(9 - 4*r + 3*r^2 - 2*r^3) + (n11 + n34)*log(6 + 4*r - 3*r^2 + 2*r^3) + (n13 + n32)*log(21 - 6*r^2 + 4*r^3) + (n02 + n43)*log(6 + r - 4*r^2 + 6*r^3)
return(L)}
interlogL_rm2c <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_rm2c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm2c <- -((x[,"n_04"] + x[,"n_41"])*(-3*log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_44"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_32"])*(2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_33"])*(2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_03"] + x[,"n_42"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_02"] + x[,"n_43"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_23"])*(-log10(2) + log10(43)) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,1 - r_rm2c)) + log10(3 - r_rm2c) + log10(pmax(1e-6,r_rm2c))) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c)) + log10(2 + r_rm2c)) + (x[,"n_22"] + x[,"n_23"])*log10(21 + 2*r_rm2c - 2*r_rm2c^2) + (x[,"n_21"] + x[,"n_24"])*log10(3 - r_rm2c + r_rm2c^2) + (x[,"n_03"] + x[,"n_42"])*log10(9 - 11*r_rm2c + 14*r_rm2c^2 - 6*r_rm2c^3) + (x[,"n_12"] + x[,"n_33"])*log10(19 + 6*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_14"] + x[,"n_31"])*log10(9 - 4*r_rm2c + 3*r_rm2c^2 - 2*r_rm2c^3) + (x[,"n_11"] + x[,"n_34"])*log10(6 + 4*r_rm2c - 3*r_rm2c^2 + 2*r_rm2c^3) + (x[,"n_13"] + x[,"n_32"])*log10(21 - 6*r_rm2c^2 + 4*r_rm2c^3) + (x[,"n_02"] + x[,"n_43"])*log10(6 + r_rm2c - 4*r_rm2c^2 + 6*r_rm2c^3)
logL_rm2c <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(3) + (-2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,1 - r_rm2c)) + log(3 - r_rm2c) + log(pmax(1e-6,r_rm2c))) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c)) + log(2 + r_rm2c)) + (x[,"n_22"] + x[,"n_23"])*log(21 + 2*r_rm2c - 2*r_rm2c^2) + (x[,"n_21"] + x[,"n_24"])*log(3 - r_rm2c + r_rm2c^2) + (x[,"n_03"] + x[,"n_42"])*log(9 - 11*r_rm2c + 14*r_rm2c^2 - 6*r_rm2c^3) + (x[,"n_12"] + x[,"n_33"])*log(19 + 6*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_14"] + x[,"n_31"])*log(9 - 4*r_rm2c + 3*r_rm2c^2 - 2*r_rm2c^3) + (x[,"n_11"] + x[,"n_34"])*log(6 + 4*r_rm2c - 3*r_rm2c^2 + 2*r_rm2c^3) + (x[,"n_13"] + x[,"n_32"])*log(21 - 6*r_rm2c^2 + 4*r_rm2c^3) + (x[,"n_02"] + x[,"n_43"])*log(6 + r_rm2c - 4*r_rm2c^2 + 6*r_rm2c^3)
logL_rm1c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(3*log(2) + 2*log(5)) + (n04 + n41)*(2*log(pmax(1e-6,1 - r)) + log(3 - r)) + (n01 + n44)*(2*log(pmax(1e-6,r)) + log(2 + r)) + (n21 + n24)*log(6 + 11*r - 11*r^2) + (n02 + n43)*(log(pmax(1e-6,r)) + log(4 + r - 3*r^2)) + (n11 + n34)*(log(pmax(1e-6,r)) + log(8 + 5*r - 2*r^2)) + (n22 + n23)*log(39 - 11*r + 11*r^2) + (n13 + n32)*log(27 - 16*r + 7*r^2 - 6*r^3) + (n14 + n31)*log(11 - 12*r - r^2 + 2*r^3) + (n03 + n42)*log(2 + 3*r - 8*r^2 + 3*r^3) + (n12 + n33)*log(12 + 20*r - 11*r^2 + 6*r^3)
return(L)}
interlogL_rm1c <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_rm1c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm1c <- -((x[,"n_14"] + x[,"n_31"])*log10(5)) - (x[,"n_11"] + x[,"n_34"])*log10(5) - (x[,"n_04"] + x[,"n_41"])*(-3*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_44"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_32"])*(2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_33"])*(2*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(5) + log10(29)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 - r_rm1c)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(2 + r_rm1c)) + (x[,"n_21"] + x[,"n_24"])*log10(6 + 11*r_rm1c - 11*r_rm1c^2) + (x[,"n_02"] + x[,"n_43"])*(log10(pmax(1e-6,r_rm1c)) + log10(4 + r_rm1c - 3*r_rm1c^2)) + (x[,"n_11"] + x[,"n_34"])*(log10(pmax(1e-6,r_rm1c)) + log10(8 + 5*r_rm1c - 2*r_rm1c^2)) + (x[,"n_22"] + x[,"n_23"])*log10(39 - 11*r_rm1c + 11*r_rm1c^2) + (x[,"n_13"] + x[,"n_32"])*log10(27 - 16*r_rm1c + 7*r_rm1c^2 - 6*r_rm1c^3) + (x[,"n_14"] + x[,"n_31"])*log10(11 - 12*r_rm1c - r_rm1c^2 + 2*r_rm1c^3) + (x[,"n_03"] + x[,"n_42"])*log10(2 + 3*r_rm1c - 8*r_rm1c^2 + 3*r_rm1c^3) + (x[,"n_12"] + x[,"n_33"])*log10(12 + 20*r_rm1c - 11*r_rm1c^2 + 6*r_rm1c^3)
logL_rm1c <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(3*log(2) + 2*log(5)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(3 - r_rm1c)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rm1c)) + log(2 + r_rm1c)) + (x[,"n_21"] + x[,"n_24"])*log(6 + 11*r_rm1c - 11*r_rm1c^2) + (x[,"n_02"] + x[,"n_43"])*(log(pmax(1e-6,r_rm1c)) + log(4 + r_rm1c - 3*r_rm1c^2)) + (x[,"n_11"] + x[,"n_34"])*(log(pmax(1e-6,r_rm1c)) + log(8 + 5*r_rm1c - 2*r_rm1c^2)) + (x[,"n_22"] + x[,"n_23"])*log(39 - 11*r_rm1c + 11*r_rm1c^2) + (x[,"n_13"] + x[,"n_32"])*log(27 - 16*r_rm1c + 7*r_rm1c^2 - 6*r_rm1c^3) + (x[,"n_14"] + x[,"n_31"])*log(11 - 12*r_rm1c - r_rm1c^2 + 2*r_rm1c^3) + (x[,"n_03"] + x[,"n_42"])*log(2 + 3*r_rm1c - 8*r_rm1c^2 + 3*r_rm1c^3) + (x[,"n_12"] + x[,"n_33"])*log(12 + 20*r_rm1c - 11*r_rm1c^2 + 6*r_rm1c^3)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_rc,r_rm2c,r_rm1c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_rc,LOD_rm2c,LOD_rm1c,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_rc,logL_rm2c,logL_rm1c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","repulsion coupling","repulsion mixed2c","repulsion mixed1c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_1.5 <- function(x){
logL_cc <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
L <- (-3*n02 - 2*n03 - 3*n04 - 2*n12 - n13 - 2*n14 - 3*n22 - n23 - 3*n24 - 2*n32 - n33 - 2*n34 - 3*n42 - 2*n43 - 3*n44)*log(2) + (n22 + n23 + n24)*log(3) + (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*log(5) + 2*(n02 + n44)*log(pmax(1e-6,1 - r)) + 2*(n04 + n42)*log(pmax(1e-6,r)) + (n03 + n43)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n14 + n32)*(log(pmax(1e-6,r)) + log(2 + r)) + (n22 + n24)*log(1 + 2*r - 2*r^2) + (n13 + n33)*log(1 + r - r^2) + (n12 + n34)*log(3 - 4*r + r^2) + n23*log(1 - r + r^2)
return(L)}
interlogL_cc <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
optimize(logL_cc,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cc <- 2*(x[,"n_04"] + x[,"n_42"])*log10(2) + 2*(x[,"n_03"] + x[,"n_43"])*log10(2) + 2*(x[,"n_02"] + x[,"n_44"])*log10(2) + x[,"n_23"]*(2*log10(2) - log10(3)) - (x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(5)) + 2*(x[,"n_02"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_42"])*log10(pmax(1e-6,r_cc)) + (x[,"n_03"] + x[,"n_43"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_22"] + x[,"n_24"])*log10(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_13"] + x[,"n_33"])*log10(1 + r_cc - r_cc^2) + (x[,"n_12"] + x[,"n_34"])*log10(3 - 4*r_cc + r_cc^2) + x[,"n_23"]*log10(1 - r_cc + r_cc^2)
logL_cc <- (-3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_22"] - x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 3*x[,"n_42"] - 2*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + 2*(x[,"n_02"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_42"])*log(pmax(1e-6,r_cc)) + (x[,"n_03"] + x[,"n_43"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_22"] + x[,"n_24"])*log(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_13"] + x[,"n_33"])*log(1 + r_cc - r_cc^2) + (x[,"n_12"] + x[,"n_34"])*log(3 - 4*r_cc + r_cc^2) + x[,"n_23"]*log(1 - r_cc + r_cc^2)
logL_cm2c <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
L <- (-3*n02 - 3*n03 - 3*n04 - 3*n12 - 2*n13 - 3*n14 - 2*n22 - 2*n23 - 2*n24 - 3*n32 - 2*n33 - 3*n34 - 3*n42 - 3*n43 - 3*n44)*log(2) + (n22 + n23 + n24)*log(3) + (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*log(5) + (n02 + n04 + n42 + n44)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n12 + n34)*log(3 - 2*r^2) + n23*log(1 + 2*r - 2*r^2) + (n14 + n32)*log(1 + 4*r - 2*r^2) + (n22 + n24)*log(1 - r + r^2) + (n03 + n43)*log(1 - 2*r + 2*r^2) + (n13 + n33)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_cm2c <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
optimize(logL_cm2c,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cm2c <- (x[,"n_03"] + x[,"n_43"])*log10(2) + 2*(x[,"n_02"] + x[,"n_04"] + x[,"n_42"] + x[,"n_44"])*log10(2) + (x[,"n_22"] + x[,"n_24"])*(2*log10(2) - log10(3)) - x[,"n_23"]*(-log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_32"])*(-log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_33"])*(-log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_34"])*(-log10(2) + log10(5)) + (x[,"n_02"] + x[,"n_04"] + x[,"n_42"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_12"] + x[,"n_34"])*log10(3 - 2*r_cm2c^2) + x[,"n_23"]*log10(1 + 2*r_cm2c - 2*r_cm2c^2) + (x[,"n_14"] + x[,"n_32"])*log10(1 + 4*r_cm2c - 2*r_cm2c^2) + (x[,"n_22"] + x[,"n_24"])*log10(1 - r_cm2c + r_cm2c^2) + (x[,"n_03"] + x[,"n_43"])*log10(1 - 2*r_cm2c + 2*r_cm2c^2) + (x[,"n_13"] + x[,"n_33"])*log10(3 - 2*r_cm2c + 2*r_cm2c^2)
logL_cm2c <- (-3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 3*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_02"] + x[,"n_04"] + x[,"n_42"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_12"] + x[,"n_34"])*log(3 - 2*r_cm2c^2) + x[,"n_23"]*log(1 + 2*r_cm2c - 2*r_cm2c^2) + (x[,"n_14"] + x[,"n_32"])*log(1 + 4*r_cm2c - 2*r_cm2c^2) + (x[,"n_22"] + x[,"n_24"])*log(1 - r_cm2c + r_cm2c^2) + (x[,"n_03"] + x[,"n_43"])*log(1 - 2*r_cm2c + 2*r_cm2c^2) + (x[,"n_13"] + x[,"n_33"])*log(3 - 2*r_cm2c + 2*r_cm2c^2)
logL_rc <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
L <- (-3*n02 - 3*n03 - 3*n04 - 3*n12 - 2*n13 - 3*n14 - 2*n22 - 2*n23 - 2*n24 - 3*n32 - 2*n33 - 3*n34 - 3*n42 - 3*n43 - 3*n44)*log(2) + (n22 + n23 + n24)*log(3) + 2*(-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*log(5) + (n04 + n42)*(log(3 - r) + log(pmax(1e-6,r))) + (n02 + n44)*(log(pmax(1e-6,1 - r)) + log(2 + r)) + (n12 + n34)*log(15 - 4*r - 2*r^2) + n23*log(7 + 2*r - 2*r^2) + (n14 + n32)*log(9 + 8*r - 2*r^2) + (n22 + n24)*log(4 - r + r^2) + (n03 + n43)*log(3 - 2*r + 2*r^2) + (n13 + n33)*log(13 - 2*r + 2*r^2)
return(L)}
interlogL_rc <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
optimize(logL_rc,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rc <- -((x[,"n_04"] + x[,"n_42"])*(-2*log10(2) + log10(5))) - (x[,"n_02"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_43"])*(-log10(2) + log10(5)) - (x[,"n_22"] + x[,"n_24"])*(-2*log10(2) + log10(3) + log10(5)) - x[,"n_23"]*(-log10(2) + log10(3) + log10(5)) - (x[,"n_14"] + x[,"n_32"])*(-log10(2) + 2*log10(5)) - (x[,"n_13"] + x[,"n_33"])*(-log10(2) + 2*log10(5)) - (x[,"n_12"] + x[,"n_34"])*(-log10(2) + 2*log10(5)) + (x[,"n_04"] + x[,"n_42"])*(log10(3 - r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_02"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_rc)) + log10(2 + r_rc)) + (x[,"n_12"] + x[,"n_34"])*log10(15 - 4*r_rc - 2*r_rc^2) + x[,"n_23"]*log10(7 + 2*r_rc - 2*r_rc^2) + (x[,"n_14"] + x[,"n_32"])*log10(9 + 8*r_rc - 2*r_rc^2) + (x[,"n_22"] + x[,"n_24"])*log10(4 - r_rc + r_rc^2) + (x[,"n_03"] + x[,"n_43"])*log10(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_13"] + x[,"n_33"])*log10(13 - 2*r_rc + 2*r_rc^2)
logL_rc <- (-3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 3*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + 2*(-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_42"])*(log(3 - r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_02"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_rc)) + log(2 + r_rc)) + (x[,"n_12"] + x[,"n_34"])*log(15 - 4*r_rc - 2*r_rc^2) + x[,"n_23"]*log(7 + 2*r_rc - 2*r_rc^2) + (x[,"n_14"] + x[,"n_32"])*log(9 + 8*r_rc - 2*r_rc^2) + (x[,"n_22"] + x[,"n_24"])*log(4 - r_rc + r_rc^2) + (x[,"n_03"] + x[,"n_43"])*log(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_13"] + x[,"n_33"])*log(13 - 2*r_rc + 2*r_rc^2)
logL_rm2c <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
L <- (-3*n02 - 2*n03 - 3*n04 - 2*n12 - n13 - 2*n14 - 3*n22 - n23 - 3*n24 - 2*n32 - n33 - 2*n34 - 3*n42 - 2*n43 - 3*n44)*log(2) + (n22 + n23 + n24)*log(3) + 2*(-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*log(5) + (n04 + n42)*(log(pmax(1e-6,1 - r)) + log(3 - r)) + (n02 + n44)*(log(pmax(1e-6,r)) + log(2 + r)) + (n22 + n24)*log(7 + 2*r - 2*r^2) + (n03 + n43)*log(1 + r - r^2) + (n13 + n33)*log(6 + r - r^2) + (n14 + n32)*log(10 - 8*r + r^2) + n23*log(4 - r + r^2) + (n12 + n34)*log(3 + 6*r + r^2)
return(L)}
interlogL_rm2c <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
optimize(logL_rm2c,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm2c <- -((x[,"n_04"] + x[,"n_42"])*(-2*log10(2) + log10(5))) - (x[,"n_03"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - x[,"n_23"]*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + 2*log10(5)) + (x[,"n_04"] + x[,"n_42"])*(log10(pmax(1e-6,1 - r_rm2c)) + log10(3 - r_rm2c)) + (x[,"n_02"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm2c)) + log10(2 + r_rm2c)) + (x[,"n_22"] + x[,"n_24"])*log10(7 + 2*r_rm2c - 2*r_rm2c^2) + (x[,"n_03"] + x[,"n_43"])*log10(1 + r_rm2c - r_rm2c^2) + (x[,"n_13"] + x[,"n_33"])*log10(6 + r_rm2c - r_rm2c^2) + (x[,"n_14"] + x[,"n_32"])*log10(10 - 8*r_rm2c + r_rm2c^2) + x[,"n_23"]*log10(4 - r_rm2c + r_rm2c^2) + (x[,"n_12"] + x[,"n_34"])*log10(3 + 6*r_rm2c + r_rm2c^2)
logL_rm2c <- (-3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_22"] - x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 3*x[,"n_42"] - 2*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + 2*(-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_42"])*(log(pmax(1e-6,1 - r_rm2c)) + log(3 - r_rm2c)) + (x[,"n_02"] + x[,"n_44"])*(log(pmax(1e-6,r_rm2c)) + log(2 + r_rm2c)) + (x[,"n_22"] + x[,"n_24"])*log(7 + 2*r_rm2c - 2*r_rm2c^2) + (x[,"n_03"] + x[,"n_43"])*log(1 + r_rm2c - r_rm2c^2) + (x[,"n_13"] + x[,"n_33"])*log(6 + r_rm2c - r_rm2c^2) + (x[,"n_14"] + x[,"n_32"])*log(10 - 8*r_rm2c + r_rm2c^2) + x[,"n_23"]*log(4 - r_rm2c + r_rm2c^2) + (x[,"n_12"] + x[,"n_34"])*log(3 + 6*r_rm2c + r_rm2c^2)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_rc,r_rm2c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_rc,LOD_rm2c,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_rc,logL_rm2c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","repulsion coupling","repulsion mixed2c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_2.2 <- function(x){
logL_cc <- function(r,n00,n01,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n43,n44,n02,n42) {
L <- (-2*n00 - 3*n01 - 3*n03 - 2*n04 - 2*n10 - n11 - n12 - n13 - 2*n14 - 2*n20 - 3*n21 + n22 - 3*n23 - 2*n24 - 2*n30 - n31 - n32 - n33 - 2*n34 - 2*n40 - 3*n41 - 3*n43 - 2*n44)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n20 + n24)*log(7) + 3*(n00 + n44)*log(pmax(1e-6,1 - r)) + 3*(n04 + n40)*log(pmax(1e-6,r)) + (n02 + n20 + n24 + n42)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n41)*(log(7 - 4*r) + 2*log(pmax(1e-6,r))) + (n14 + n30)*(log(5 - 2*r) + 2*log(pmax(1e-6,r))) + (n10 + n34)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r)) + (n01 + n43)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r)) + (n12 + n32)*log(5 + 2*r - 2*r^2) + (n21 + n23)*log(21 + 2*r - 2*r^2) + (n13 + n31)*(log(pmax(1e-6,r)) + log(7 - 3*r + 2*r^2)) + n22*log(3 - 2*r + 2*r^2) + (n11 + n33)*log(6 - 7*r + 3*r^2 - 2*r^3)
return(L)}
interlogL_cc <- function(n00,n01,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n43,n44,n02,n42) {
optimize(logL_cc,c(0,0.5), n00,n01,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n43,n44,n02,n42, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_43"],x[,"n_44"],x[,"n_02"],x[,"n_42"])
LOD_cc <- 3*(x[,"n_04"] + x[,"n_40"])*log10(2) + 2*(x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(2) + 3*(x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_13"] + x[,"n_31"])*log10(3) - (x[,"n_11"] + x[,"n_33"])*log10(3) - (x[,"n_03"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - x[,"n_22"]*(-log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_32"])*(-log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_23"])*(-log10(2) + log10(43)) + 3*(x[,"n_00"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_04"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_41"])*(log10(7 - 4*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_30"])*(log10(5 - 2*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 2*r_cc)) + (x[,"n_01"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 4*r_cc)) + (x[,"n_12"] + x[,"n_32"])*log10(5 + 2*r_cc - 2*r_cc^2) + (x[,"n_21"] + x[,"n_23"])*log10(21 + 2*r_cc - 2*r_cc^2) + (x[,"n_13"] + x[,"n_31"])*(log10(pmax(1e-6,r_cc)) + log10(7 - 3*r_cc + 2*r_cc^2)) + x[,"n_22"]*log10(3 - 2*r_cc + 2*r_cc^2) + (x[,"n_11"] + x[,"n_33"])*log10(6 - 7*r_cc + 3*r_cc^2 - 2*r_cc^3)
logL_cc <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 3*x[,"n_21"] + x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_20"] + x[,"n_24"])*log(7) + 3*(x[,"n_00"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_04"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_41"])*(log(7 - 4*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_30"])*(log(5 - 2*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 2*r_cc)) + (x[,"n_01"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 4*r_cc)) + (x[,"n_12"] + x[,"n_32"])*log(5 + 2*r_cc - 2*r_cc^2) + (x[,"n_21"] + x[,"n_23"])*log(21 + 2*r_cc - 2*r_cc^2) + (x[,"n_13"] + x[,"n_31"])*(log(pmax(1e-6,r_cc)) + log(7 - 3*r_cc + 2*r_cc^2)) + x[,"n_22"]*log(3 - 2*r_cc + 2*r_cc^2) + (x[,"n_11"] + x[,"n_33"])*log(6 - 7*r_cc + 3*r_cc^2 - 2*r_cc^3)
logL_cm <- function(r,n00,n01,n02,n03,n04,n10,n12,n14,n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n11,n13,n31,n33) {
L <- (-n00 - 2*n01 - 3*n02 - 2*n03 - n04 - n10 - 2*n12 - n14 - n20 - 2*n21 - 2*n22 - 2*n23 - n24 - n30 - 2*n32 - n34 - n40 - 2*n41 - 3*n42 - 2*n43 - n44)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n12 + n32)*log(11) + (n00 + n44)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n40)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + n22*log(57 + 8*r - 8*r^2) + (n20 + n24)*log(3 - r + r^2) + (n14 + n30)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n21 + n23)*log(33 - 2*r + 2*r^2) + (n03 + n41)*(log(pmax(1e-6,r)) + log(6 - 7*r + 4*r^2)) + (n02 + n42)*log(9 - 8*r + 8*r^2) + (n01 + n43)*log(3 - 4*r + 5*r^2 - 4*r^3) + (n10 + n34)*log(3 - 4*r + 3*r^2 - 2*r^3) + (n13 + n31)*log(3 + 2*r + 3*r^2 - 2*r^3) + (n11 + n33)*log(6 - 2*r - 3*r^2 + 2*r^3)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n10,n12,n14,n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n11,n13,n31,n33) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n10,n12,n14,n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n11,n13,n31,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_12"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_32"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_11"],x[,"n_13"],x[,"n_31"],x[,"n_33"])
LOD_cm <- 3*(x[,"n_04"] + x[,"n_40"])*log10(2) + 3*(x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_14"] + x[,"n_30"])*(-log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_34"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_31"])*(-log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_33"])*(-log10(2) + 2*log10(3)) - (x[,"n_02"] + x[,"n_42"])*log10(7) - (x[,"n_03"] + x[,"n_41"])*(-2*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_43"])*(-2*log10(2) + log10(7)) - (x[,"n_20"] + x[,"n_24"])*(-2*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_23"])*(-log10(2) + log10(5) + log10(13)) - x[,"n_22"]*log10(59) + (x[,"n_00"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cm)) + 2*log10(pmax(1e-6,r_cm))) + x[,"n_22"]*log10(57 + 8*r_cm - 8*r_cm^2) + (x[,"n_20"] + x[,"n_24"])*log10(3 - r_cm + r_cm^2) + (x[,"n_14"] + x[,"n_30"])*(log10(pmax(1e-6,r_cm)) + log10(4 - 3*r_cm + 2*r_cm^2)) + (x[,"n_21"] + x[,"n_23"])*log10(33 - 2*r_cm + 2*r_cm^2) + (x[,"n_03"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm)) + log10(6 - 7*r_cm + 4*r_cm^2)) + (x[,"n_02"] + x[,"n_42"])*log10(9 - 8*r_cm + 8*r_cm^2) + (x[,"n_01"] + x[,"n_43"])*log10(3 - 4*r_cm + 5*r_cm^2 - 4*r_cm^3) + (x[,"n_10"] + x[,"n_34"])*log10(3 - 4*r_cm + 3*r_cm^2 - 2*r_cm^3) + (x[,"n_13"] + x[,"n_31"])*log10(3 + 2*r_cm + 3*r_cm^2 - 2*r_cm^3) + (x[,"n_11"] + x[,"n_33"])*log10(6 - 2*r_cm - 3*r_cm^2 + 2*r_cm^3)
logL_cm <- (-x[,"n_00"] - 2*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - 2*x[,"n_12"] - x[,"n_14"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - 2*x[,"n_32"] - x[,"n_34"] - x[,"n_40"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 2*x[,"n_43"] - x[,"n_44"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_12"] + x[,"n_32"])*log(11) + (x[,"n_00"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cm)) + 2*log(pmax(1e-6,r_cm))) + x[,"n_22"]*log(57 + 8*r_cm - 8*r_cm^2) + (x[,"n_20"] + x[,"n_24"])*log(3 - r_cm + r_cm^2) + (x[,"n_14"] + x[,"n_30"])*(log(pmax(1e-6,r_cm)) + log(4 - 3*r_cm + 2*r_cm^2)) + (x[,"n_21"] + x[,"n_23"])*log(33 - 2*r_cm + 2*r_cm^2) + (x[,"n_03"] + x[,"n_41"])*(log(pmax(1e-6,r_cm)) + log(6 - 7*r_cm + 4*r_cm^2)) + (x[,"n_02"] + x[,"n_42"])*log(9 - 8*r_cm + 8*r_cm^2) + (x[,"n_01"] + x[,"n_43"])*log(3 - 4*r_cm + 5*r_cm^2 - 4*r_cm^3) + (x[,"n_10"] + x[,"n_34"])*log(3 - 4*r_cm + 3*r_cm^2 - 2*r_cm^3) + (x[,"n_13"] + x[,"n_31"])*log(3 + 2*r_cm + 3*r_cm^2 - 2*r_cm^3) + (x[,"n_11"] + x[,"n_33"])*log(6 - 2*r_cm - 3*r_cm^2 + 2*r_cm^3)
logL_cr <- function(r,n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n21,n22,n23,n24,n30,n31,n33,n34,n40,n41,n42,n43,n44,n12,n32) {
L <- (-2*n00 - 3*n01 - 2*n02 - 3*n03 - 2*n04 - 2*n10 - n11 - n13 - 2*n14 - 2*n20 - 3*n21 - n22 - 3*n23 - 2*n24 - 2*n30 - n31 - n33 - 2*n34 - 2*n40 - 3*n41 - 2*n42 - 3*n43 - 2*n44)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n04 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n44)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + n22*log(9 + 4*r - 4*r^2) + (n21 + n23)*log(21 + 2*r - 2*r^2) + (n12 + n32)*log(3 - r + r^2) + (n10 + n34)*(log(pmax(1e-6,r)) + log(4 - 5*r + 2*r^2)) + (n01 + n43)*(log(pmax(1e-6,r)) + log(4 - 5*r + 4*r^2)) + (n20 + n24)*log(3 - 5*r + 5*r^2) + (n03 + n41)*log(3 - 6*r + 7*r^2 - 4*r^3) + (n14 + n30)*log(1 + r^2 - 2*r^3) + (n11 + n33)*log(3 - r + 3*r^2 - 2*r^3) + (n13 + n31)*log(3 + r - 3*r^2 + 2*r^3)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n21,n22,n23,n24,n30,n31,n33,n34,n40,n41,n42,n43,n44,n12,n32) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n21,n22,n23,n24,n30,n31,n33,n34,n40,n41,n42,n43,n44,n12,n32, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_12"],x[,"n_32"])
LOD_cr <- 3*(x[,"n_04"] + x[,"n_40"])*log10(2) + 3*(x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_13"] + x[,"n_31"])*log10(3) - (x[,"n_11"] + x[,"n_33"])*log10(3) - (x[,"n_03"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - x[,"n_22"]*(log10(2) + log10(5)) - (x[,"n_20"] + x[,"n_24"])*(-2*log10(2) + log10(7)) - (x[,"n_12"] + x[,"n_32"])*(-2*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_23"])*(-log10(2) + log10(43)) + (x[,"n_04"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + x[,"n_22"]*log10(9 + 4*r_cr - 4*r_cr^2) + (x[,"n_21"] + x[,"n_23"])*log10(21 + 2*r_cr - 2*r_cr^2) + (x[,"n_12"] + x[,"n_32"])*log10(3 - r_cr + r_cr^2) + (x[,"n_10"] + x[,"n_34"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 5*r_cr + 2*r_cr^2)) + (x[,"n_01"] + x[,"n_43"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_20"] + x[,"n_24"])*log10(3 - 5*r_cr + 5*r_cr^2) + (x[,"n_03"] + x[,"n_41"])*log10(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_14"] + x[,"n_30"])*log10(1 + r_cr^2 - 2*r_cr^3) + (x[,"n_11"] + x[,"n_33"])*log10(3 - r_cr + 3*r_cr^2 - 2*r_cr^3) + (x[,"n_13"] + x[,"n_31"])*log10(3 + r_cr - 3*r_cr^2 + 2*r_cr^3)
logL_cr <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 2*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - x[,"n_11"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 3*x[,"n_21"] - x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - x[,"n_31"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 2*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + x[,"n_22"]*log(9 + 4*r_cr - 4*r_cr^2) + (x[,"n_21"] + x[,"n_23"])*log(21 + 2*r_cr - 2*r_cr^2) + (x[,"n_12"] + x[,"n_32"])*log(3 - r_cr + r_cr^2) + (x[,"n_10"] + x[,"n_34"])*(log(pmax(1e-6,r_cr)) + log(4 - 5*r_cr + 2*r_cr^2)) + (x[,"n_01"] + x[,"n_43"])*(log(pmax(1e-6,r_cr)) + log(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_20"] + x[,"n_24"])*log(3 - 5*r_cr + 5*r_cr^2) + (x[,"n_03"] + x[,"n_41"])*log(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_14"] + x[,"n_30"])*log(1 + r_cr^2 - 2*r_cr^3) + (x[,"n_11"] + x[,"n_33"])*log(3 - r_cr + 3*r_cr^2 - 2*r_cr^3) + (x[,"n_13"] + x[,"n_31"])*log(3 + r_cr - 3*r_cr^2 + 2*r_cr^3)
logL_rc <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 4*n20 - 3*n21 - 3*n22 - 3*n23 - 4*n24 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n04 + n40)*(log(3 - 2*r) + 2*log(pmax(1e-6,r))) + (n00 + n44)*(2*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n02 + n42)*log(3 + 4*r - 4*r^2) + n22*log(39 + 4*r - 4*r^2) + (n21 + n23)*log(21 + 2*r - 2*r^2) + (n14 + n30)*(log(pmax(1e-6,r)) + log(6 - 5*r + 2*r^2)) + (n03 + n41)*(log(pmax(1e-6,r)) + log(3 - 2*r + 2*r^2)) + (n12 + n32)*log(23 - 4*r + 4*r^2) + (n20 + n24)*log(9 - 8*r + 8*r^2) + (n13 + n31)*log(9 + 4*r + 6*r^2 - 4*r^3) + (n10 + n34)*log(3 - 2*r + r^2 - 2*r^3) + (n01 + n43)*log(3 - 5*r + 4*r^2 - 2*r^3) + (n11 + n33)*log(15 - 4*r - 6*r^2 + 4*r^3)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rc <- -((x[,"n_14"] + x[,"n_30"])*log10(2)) - (x[,"n_10"] + x[,"n_34"])*log10(2) + (x[,"n_04"] + x[,"n_40"])*log10(2) - 2*(x[,"n_02"] + x[,"n_42"])*log10(2) + (x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_13"] + x[,"n_31"])*(2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_33"])*(2*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - x[,"n_22"]*(3*log10(2) + log10(5)) - (x[,"n_20"] + x[,"n_24"])*log10(7) - (x[,"n_12"] + x[,"n_32"])*(log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_23"])*(-log10(2) + log10(43)) + (x[,"n_04"] + x[,"n_40"])*(log10(3 - 2*r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_02"] + x[,"n_42"])*log10(3 + 4*r_rc - 4*r_rc^2) + x[,"n_22"]*log10(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_21"] + x[,"n_23"])*log10(21 + 2*r_rc - 2*r_rc^2) + (x[,"n_14"] + x[,"n_30"])*(log10(pmax(1e-6,r_rc)) + log10(6 - 5*r_rc + 2*r_rc^2)) + (x[,"n_03"] + x[,"n_41"])*(log10(pmax(1e-6,r_rc)) + log10(3 - 2*r_rc + 2*r_rc^2)) + (x[,"n_12"] + x[,"n_32"])*log10(23 - 4*r_rc + 4*r_rc^2) + (x[,"n_20"] + x[,"n_24"])*log10(9 - 8*r_rc + 8*r_rc^2) + (x[,"n_13"] + x[,"n_31"])*log10(9 + 4*r_rc + 6*r_rc^2 - 4*r_rc^3) + (x[,"n_10"] + x[,"n_34"])*log10(3 - 2*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_01"] + x[,"n_43"])*log10(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3) + (x[,"n_11"] + x[,"n_33"])*log10(15 - 4*r_rc - 6*r_rc^2 + 4*r_rc^3)
logL_rc <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(log(3 - 2*r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_02"] + x[,"n_42"])*log(3 + 4*r_rc - 4*r_rc^2) + x[,"n_22"]*log(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_21"] + x[,"n_23"])*log(21 + 2*r_rc - 2*r_rc^2) + (x[,"n_14"] + x[,"n_30"])*(log(pmax(1e-6,r_rc)) + log(6 - 5*r_rc + 2*r_rc^2)) + (x[,"n_03"] + x[,"n_41"])*(log(pmax(1e-6,r_rc)) + log(3 - 2*r_rc + 2*r_rc^2)) + (x[,"n_12"] + x[,"n_32"])*log(23 - 4*r_rc + 4*r_rc^2) + (x[,"n_20"] + x[,"n_24"])*log(9 - 8*r_rc + 8*r_rc^2) + (x[,"n_13"] + x[,"n_31"])*log(9 + 4*r_rc + 6*r_rc^2 - 4*r_rc^3) + (x[,"n_10"] + x[,"n_34"])*log(3 - 2*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_01"] + x[,"n_43"])*log(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3) + (x[,"n_11"] + x[,"n_33"])*log(15 - 4*r_rc - 6*r_rc^2 + 4*r_rc^3)
logL_rm <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-3*n00 - 4*n01 - 3*n02 - 4*n03 - 3*n04 - 3*n10 - 2*n11 - 2*n12 - 2*n13 - 3*n14 - n20 - 2*n21 - 2*n22 - 2*n23 - n24 - 3*n30 - 2*n31 - 2*n32 - 2*n33 - 3*n34 - 3*n40 - 4*n41 - 3*n42 - 4*n43 - 3*n44)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n12 + n32)*log(11) + (n04 + n40)*(log(3 - 2*r) + log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n44)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 + 2*r)) + n22*log(57 + 8*r - 8*r^2) + (n20 + n24)*log(3 - r + r^2) + (n21 + n23)*log(33 - 2*r + 2*r^2) + (n02 + n42)*log(9 - 8*r + 8*r^2) + (n03 + n41)*log(9 - 6*r + 8*r^2 - 8*r^3) + (n14 + n30)*log(9 - 8*r + 6*r^2 - 4*r^3) + (n11 + n33)*log(15 + 4*r + 6*r^2 - 4*r^3) + (n13 + n31)*log(21 - 4*r - 6*r^2 + 4*r^3) + (n10 + n34)*log(3 + 8*r - 6*r^2 + 4*r^3) + (n01 + n43)*log(3 + 14*r - 16*r^2 + 8*r^3)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm <- (x[,"n_04"] + x[,"n_40"])*log10(2) + (x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_14"] + x[,"n_30"])*(log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_34"])*(log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_31"])*(log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_33"])*(log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_41"])*log10(7) - (x[,"n_02"] + x[,"n_42"])*log10(7) - (x[,"n_01"] + x[,"n_43"])*log10(7) - (x[,"n_20"] + x[,"n_24"])*(-2*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_23"])*(-log10(2) + log10(5) + log10(13)) - x[,"n_22"]*log10(59) + (x[,"n_04"] + x[,"n_40"])*(log10(3 - 2*r_rm) + log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm)) + log10(1 + 2*r_rm)) + x[,"n_22"]*log10(57 + 8*r_rm - 8*r_rm^2) + (x[,"n_20"] + x[,"n_24"])*log10(3 - r_rm + r_rm^2) + (x[,"n_21"] + x[,"n_23"])*log10(33 - 2*r_rm + 2*r_rm^2) + (x[,"n_02"] + x[,"n_42"])*log10(9 - 8*r_rm + 8*r_rm^2) + (x[,"n_03"] + x[,"n_41"])*log10(9 - 6*r_rm + 8*r_rm^2 - 8*r_rm^3) + (x[,"n_14"] + x[,"n_30"])*log10(9 - 8*r_rm + 6*r_rm^2 - 4*r_rm^3) + (x[,"n_11"] + x[,"n_33"])*log10(15 + 4*r_rm + 6*r_rm^2 - 4*r_rm^3) + (x[,"n_13"] + x[,"n_31"])*log10(21 - 4*r_rm - 6*r_rm^2 + 4*r_rm^3) + (x[,"n_10"] + x[,"n_34"])*log10(3 + 8*r_rm - 6*r_rm^2 + 4*r_rm^3) + (x[,"n_01"] + x[,"n_43"])*log10(3 + 14*r_rm - 16*r_rm^2 + 8*r_rm^3)
logL_rm <- (-3*x[,"n_00"] - 4*x[,"n_01"] - 3*x[,"n_02"] - 4*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - 3*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_40"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 4*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_12"] + x[,"n_32"])*log(11) + (x[,"n_04"] + x[,"n_40"])*(log(3 - 2*r_rm) + log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm)) + log(1 + 2*r_rm)) + x[,"n_22"]*log(57 + 8*r_rm - 8*r_rm^2) + (x[,"n_20"] + x[,"n_24"])*log(3 - r_rm + r_rm^2) + (x[,"n_21"] + x[,"n_23"])*log(33 - 2*r_rm + 2*r_rm^2) + (x[,"n_02"] + x[,"n_42"])*log(9 - 8*r_rm + 8*r_rm^2) + (x[,"n_03"] + x[,"n_41"])*log(9 - 6*r_rm + 8*r_rm^2 - 8*r_rm^3) + (x[,"n_14"] + x[,"n_30"])*log(9 - 8*r_rm + 6*r_rm^2 - 4*r_rm^3) + (x[,"n_11"] + x[,"n_33"])*log(15 + 4*r_rm + 6*r_rm^2 - 4*r_rm^3) + (x[,"n_13"] + x[,"n_31"])*log(21 - 4*r_rm - 6*r_rm^2 + 4*r_rm^3) + (x[,"n_10"] + x[,"n_34"])*log(3 + 8*r_rm - 6*r_rm^2 + 4*r_rm^3) + (x[,"n_01"] + x[,"n_43"])*log(3 + 14*r_rm - 16*r_rm^2 + 8*r_rm^3)
logL_rr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 4*n20 - 3*n21 - 3*n22 - 3*n23 - 4*n24 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44)*log(2) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n20 - 2*n21 - n22 - 2*n23 - 2*n24 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44)*log(5) + (n04 + n40)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r))) + (n00 + n44)*(2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n20 + n24)*log(3 + 16*r - 16*r^2) + (n02 + n42)*log(1 + 12*r - 12*r^2) + (n12 + n32)*log(21 + 4*r - 4*r^2) + (n21 + n23)*log(21 + 2*r - 2*r^2) + (n10 + n34)*(log(pmax(1e-6,r)) + log(2 + 5*r - 2*r^2)) + n22*log(9 - 4*r + 4*r^2) + (n13 + n31)*log(21 - 20*r + 6*r^2 - 4*r^3) + (n01 + n43)*log(r + 4*r^2 - 2*r^3) + (n03 + n41)*log(3 - 3*r - 2*r^2 + 2*r^3) + (n14 + n30)*log(5 - 6*r - r^2 + 2*r^3) + (n11 + n33)*log(3 + 20*r - 6*r^2 + 4*r^3)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rr <- -3*x[,"n_22"]*log10(2) - (x[,"n_14"] + x[,"n_30"])*log10(2) - (x[,"n_10"] + x[,"n_34"])*log10(2) + (x[,"n_04"] + x[,"n_40"])*log10(2) - 2*(x[,"n_02"] + x[,"n_42"])*log10(2) + (x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_13"] + x[,"n_31"])*(2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_33"])*(2*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - (x[,"n_20"] + x[,"n_24"])*log10(7) - (x[,"n_12"] + x[,"n_32"])*(log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_23"])*(-log10(2) + log10(43)) + (x[,"n_04"] + x[,"n_40"])*(log10(3 - 2*r_rr) + 2*log10(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_20"] + x[,"n_24"])*log10(3 + 16*r_rr - 16*r_rr^2) + (x[,"n_02"] + x[,"n_42"])*log10(1 + 12*r_rr - 12*r_rr^2) + (x[,"n_12"] + x[,"n_32"])*log10(21 + 4*r_rr - 4*r_rr^2) + (x[,"n_21"] + x[,"n_23"])*log10(21 + 2*r_rr - 2*r_rr^2) + (x[,"n_10"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(2 + 5*r_rr - 2*r_rr^2)) + x[,"n_22"]*log10(9 - 4*r_rr + 4*r_rr^2) + (x[,"n_13"] + x[,"n_31"])*log10(21 - 20*r_rr + 6*r_rr^2 - 4*r_rr^3) + (x[,"n_01"] + x[,"n_43"])*log10(r_rr + 4*r_rr^2 - 2*r_rr^3) + (x[,"n_03"] + x[,"n_41"])*log10(3 - 3*r_rr - 2*r_rr^2 + 2*r_rr^3) + (x[,"n_14"] + x[,"n_30"])*log10(5 - 6*r_rr - r_rr^2 + 2*r_rr^3) + (x[,"n_11"] + x[,"n_33"])*log10(3 + 20*r_rr - 6*r_rr^2 + 4*r_rr^3)
logL_rr <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"])*log(2) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(log(3 - 2*r_rr) + 2*log(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_20"] + x[,"n_24"])*log(3 + 16*r_rr - 16*r_rr^2) + (x[,"n_02"] + x[,"n_42"])*log(1 + 12*r_rr - 12*r_rr^2) + (x[,"n_12"] + x[,"n_32"])*log(21 + 4*r_rr - 4*r_rr^2) + (x[,"n_21"] + x[,"n_23"])*log(21 + 2*r_rr - 2*r_rr^2) + (x[,"n_10"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(2 + 5*r_rr - 2*r_rr^2)) + x[,"n_22"]*log(9 - 4*r_rr + 4*r_rr^2) + (x[,"n_13"] + x[,"n_31"])*log(21 - 20*r_rr + 6*r_rr^2 - 4*r_rr^3) + (x[,"n_01"] + x[,"n_43"])*log(r_rr + 4*r_rr^2 - 2*r_rr^3) + (x[,"n_03"] + x[,"n_41"])*log(3 - 3*r_rr - 2*r_rr^2 + 2*r_rr^3) + (x[,"n_14"] + x[,"n_30"])*log(5 - 6*r_rr - r_rr^2 + 2*r_rr^3) + (x[,"n_11"] + x[,"n_33"])*log(3 + 20*r_rr - 6*r_rr^2 + 4*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_2.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - n25 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + (n01 + n04 + n11 + n14 + n20 + n21 + n22 + n23 + n24 + n25 + n31 + n34 + n41 + n44)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + 4*(n00 + n45)*log(pmax(1e-6,1 - r)) + 4*(n05 + n40)*log(pmax(1e-6,r)) + (n10 + n35)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n20 + n25)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n41)*(log(3 - 2*r) + 3*log(pmax(1e-6,r))) + (n15 + n30)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n44)*(3*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n02 + n43)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(11 + 4*r)) + (n03 + n42)*(2*log(pmax(1e-6,r)) + log(15 - 19*r + 4*r^2)) + (n14 + n31)*(2*log(pmax(1e-6,r)) + log(13 - 14*r + 8*r^2)) + (n11 + n34)*(2*log(pmax(1e-6,1 - r)) + log(7 - 2*r + 8*r^2)) + (n13 + n32)*(log(pmax(1e-6,r)) + log(60 - 69*r + 54*r^2 - 16*r^3)) + (n21 + n24)*(log(pmax(1e-6,r)) + log(17 - 29*r + 24*r^2 - 12*r^3)) + (n12 + n33)*log(29 - 20*r - 3*r^2 + 10*r^3 - 16*r^4) + (n22 + n23)*log(15 - 17*r + 25*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_10"],x[,"n_15"],x[,"n_30"],x[,"n_35"])
LOD_cc <- 4*(x[,"n_20"] + x[,"n_25"])*log10(2) + 4*(x[,"n_15"] + x[,"n_30"])*log10(2) - (x[,"n_14"] + x[,"n_31"])*log10(2) - (x[,"n_11"] + x[,"n_34"])*log10(2) + 4*(x[,"n_10"] + x[,"n_35"])*log10(2) + 4*(x[,"n_05"] + x[,"n_40"])*log10(2) + 2*(x[,"n_04"] + x[,"n_41"])*log10(2) + 2*(x[,"n_01"] + x[,"n_44"])*log10(2) + 4*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_21"] + x[,"n_24"])*(-log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(37)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(37)) + 4*(x[,"n_00"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_35"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_25"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_41"])*(log10(3 - 2*r_cc) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_44"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(1 + 2*r_cc)) + (x[,"n_02"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc)) + log10(11 + 4*r_cc)) + (x[,"n_03"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(15 - 19*r_cc + 4*r_cc^2)) + (x[,"n_14"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(13 - 14*r_cc + 8*r_cc^2)) + (x[,"n_11"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(7 - 2*r_cc + 8*r_cc^2)) + (x[,"n_13"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(60 - 69*r_cc + 54*r_cc^2 - 16*r_cc^3)) + (x[,"n_21"] + x[,"n_24"])*(log10(pmax(1e-6,r_cc)) + log10(17 - 29*r_cc + 24*r_cc^2 - 12*r_cc^3)) + (x[,"n_12"] + x[,"n_33"])*log10(29 - 20*r_cc - 3*r_cc^2 + 10*r_cc^3 - 16*r_cc^4) + (x[,"n_22"] + x[,"n_23"])*log10(15 - 17*r_cc + 25*r_cc^2 - 16*r_cc^3 + 8*r_cc^4)
logL_cc <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - x[,"n_25"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + 4*(x[,"n_00"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_35"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_25"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_41"])*(log(3 - 2*r_cc) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_44"])*(3*log(pmax(1e-6,1 - r_cc)) + log(1 + 2*r_cc)) + (x[,"n_02"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc)) + log(11 + 4*r_cc)) + (x[,"n_03"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(15 - 19*r_cc + 4*r_cc^2)) + (x[,"n_14"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(13 - 14*r_cc + 8*r_cc^2)) + (x[,"n_11"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_cc)) + log(7 - 2*r_cc + 8*r_cc^2)) + (x[,"n_13"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(60 - 69*r_cc + 54*r_cc^2 - 16*r_cc^3)) + (x[,"n_21"] + x[,"n_24"])*(log(pmax(1e-6,r_cc)) + log(17 - 29*r_cc + 24*r_cc^2 - 12*r_cc^3)) + (x[,"n_12"] + x[,"n_33"])*log(29 - 20*r_cc - 3*r_cc^2 + 10*r_cc^3 - 16*r_cc^4) + (x[,"n_22"] + x[,"n_23"])*log(15 - 17*r_cc + 25*r_cc^2 - 16*r_cc^3 + 8*r_cc^4)
logL_cm2c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 2*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n00 + n45)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n40)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n15 + n30)*(2*log(pmax(1e-6,r)) + log(5 - 6*r + 4*r^2)) + (n10 + n35)*(2*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 4*r^2)) + (n04 + n41)*(2*log(pmax(1e-6,r)) + log(11 - 11*r + 6*r^2)) + (n01 + n44)*(2*log(pmax(1e-6,1 - r)) + log(6 - r + 6*r^2)) + (n14 + n31)*(log(pmax(1e-6,r)) + log(44 - 33*r + 46*r^2 - 24*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(7 - 13*r + 12*r^2 - 6*r^3)) + (n03 + n42)*(log(pmax(1e-6,r)) + log(18 - 20*r + 15*r^2 - 4*r^3)) + (n11 + n34)*log(33 - 20*r - 39*r^2 + 50*r^3 - 24*r^4) + (n22 + n23)*log(34 - r - 7*r^2 + 16*r^3 - 8*r^4) + (n02 + n43)*log(9 - 7*r + r^2 + r^3 - 4*r^4) + (n13 + n32)*log(42 + 16*r + 41*r^2 - 46*r^3 + 16*r^4) + (n12 + n33)*log(69 - 24*r - r^2 - 18*r^3 + 16*r^4) + (n21 + n24)*log(33 - 11*r + 47*r^2 - 72*r^3 + 36*r^4)
return(L)}
interlogL_cm2c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cm2c <- 4*(x[,"n_05"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_15"] + x[,"n_30"])*(-2*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_35"])*(-2*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_31"])*(log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_34"])*(log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_04"] + x[,"n_41"])*(-2*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_44"])*(-2*log10(2) + log10(7)) - (x[,"n_20"] + x[,"n_25"])*(-3*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-log10(2) + log10(5) + log10(13)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(7) + log10(19)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(3) + log10(37)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(3) + log10(37)) + (x[,"n_00"] + x[,"n_45"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cm2c)) + 3*log10(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(5 - 6*r_cm2c + 4*r_cm2c^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 - 2*r_cm2c + 4*r_cm2c^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(11 - 11*r_cm2c + 6*r_cm2c^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(6 - r_cm2c + 6*r_cm2c^2)) + (x[,"n_14"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm2c)) + log10(44 - 33*r_cm2c + 46*r_cm2c^2 - 24*r_cm2c^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_cm2c)) + log10(7 - 13*r_cm2c + 12*r_cm2c^2 - 6*r_cm2c^3)) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm2c)) + log10(18 - 20*r_cm2c + 15*r_cm2c^2 - 4*r_cm2c^3)) + (x[,"n_11"] + x[,"n_34"])*log10(33 - 20*r_cm2c - 39*r_cm2c^2 + 50*r_cm2c^3 - 24*r_cm2c^4) + (x[,"n_22"] + x[,"n_23"])*log10(34 - r_cm2c - 7*r_cm2c^2 + 16*r_cm2c^3 - 8*r_cm2c^4) + (x[,"n_02"] + x[,"n_43"])*log10(9 - 7*r_cm2c + r_cm2c^2 + r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_13"] + x[,"n_32"])*log10(42 + 16*r_cm2c + 41*r_cm2c^2 - 46*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_12"] + x[,"n_33"])*log10(69 - 24*r_cm2c - r_cm2c^2 - 18*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_21"] + x[,"n_24"])*log10(33 - 11*r_cm2c + 47*r_cm2c^2 - 72*r_cm2c^3 + 36*r_cm2c^4)
logL_cm2c <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_00"] + x[,"n_45"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cm2c)) + 3*log(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,r_cm2c)) + log(5 - 6*r_cm2c + 4*r_cm2c^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(3 - 2*r_cm2c + 4*r_cm2c^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm2c)) + log(11 - 11*r_cm2c + 6*r_cm2c^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(6 - r_cm2c + 6*r_cm2c^2)) + (x[,"n_14"] + x[,"n_31"])*(log(pmax(1e-6,r_cm2c)) + log(44 - 33*r_cm2c + 46*r_cm2c^2 - 24*r_cm2c^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_cm2c)) + log(7 - 13*r_cm2c + 12*r_cm2c^2 - 6*r_cm2c^3)) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,r_cm2c)) + log(18 - 20*r_cm2c + 15*r_cm2c^2 - 4*r_cm2c^3)) + (x[,"n_11"] + x[,"n_34"])*log(33 - 20*r_cm2c - 39*r_cm2c^2 + 50*r_cm2c^3 - 24*r_cm2c^4) + (x[,"n_22"] + x[,"n_23"])*log(34 - r_cm2c - 7*r_cm2c^2 + 16*r_cm2c^3 - 8*r_cm2c^4) + (x[,"n_02"] + x[,"n_43"])*log(9 - 7*r_cm2c + r_cm2c^2 + r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_13"] + x[,"n_32"])*log(42 + 16*r_cm2c + 41*r_cm2c^2 - 46*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_12"] + x[,"n_33"])*log(69 - 24*r_cm2c - r_cm2c^2 - 18*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_21"] + x[,"n_24"])*log(33 - 11*r_cm2c + 47*r_cm2c^2 - 72*r_cm2c^3 + 36*r_cm2c^4)
logL_cm1c <- function(r,n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n00 + n05 + n40 + n45)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n44)*(log(pmax(1e-6,r)) + log(8 - 13*r + 11*r^2 - 6*r^3)) + (n04 + n41)*(log(pmax(1e-6,r)) + log(9 - 16*r + 13*r^2 - 6*r^3)) + (n10 + n15 + n30 + n35)*(log(pmax(1e-6,r)) + log(1 - 2*r + 2*r^2 - r^3)) + (n21 + n24)*log(33 + 7*r - 43*r^2 + 72*r^3 - 36*r^4) + (n12 + n33)*log(60 - 8*r - 11*r^2 + 26*r^3 - 16*r^4) + (n13 + n32)*log(51 + 16*r - 29*r^2 + 38*r^3 - 16*r^4) + (n03 + n42)*log(9 - 13*r + 17*r^2 - 11*r^3 + 4*r^4) + (n02 + n43)*log(6 - 4*r + 8*r^2 - 5*r^3 + 4*r^4) + (n20 + n25)*log(3 - 8*r + 14*r^2 - 12*r^3 + 6*r^4) + (n22 + n23)*log(32 + 3*r + 5*r^2 - 16*r^3 + 8*r^4) + (n14 + n31)*log(15 - 4*r + 39*r^2 - 50*r^3 + 24*r^4) + (n11 + n34)*log(24 - 20*r + 33*r^2 - 46*r^3 + 24*r^4)
return(L)}
interlogL_cm1c <- function(n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35) {
optimize(logL_cm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_10"],x[,"n_15"],x[,"n_30"],x[,"n_35"])
LOD_cm1c <- 4*(x[,"n_00"] + x[,"n_05"] + x[,"n_40"] + x[,"n_45"])*log10(2) - (x[,"n_10"] + x[,"n_15"] + x[,"n_30"] + x[,"n_35"])*(-4*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_31"])*(log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_34"])*(log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_04"] + x[,"n_41"])*(-2*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_44"])*(-2*log10(2) + log10(7)) - (x[,"n_20"] + x[,"n_25"])*(-3*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-log10(2) + log10(5) + log10(13)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(7) + log10(19)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(3) + log10(37)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(3) + log10(37)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_40"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_cm1c)) + log10(8 - 13*r_cm1c + 11*r_cm1c^2 - 6*r_cm1c^3)) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm1c)) + log10(9 - 16*r_cm1c + 13*r_cm1c^2 - 6*r_cm1c^3)) + (x[,"n_10"] + x[,"n_15"] + x[,"n_30"] + x[,"n_35"])*(log10(pmax(1e-6,r_cm1c)) + log10(1 - 2*r_cm1c + 2*r_cm1c^2 - r_cm1c^3)) + (x[,"n_21"] + x[,"n_24"])*log10(33 + 7*r_cm1c - 43*r_cm1c^2 + 72*r_cm1c^3 - 36*r_cm1c^4) + (x[,"n_12"] + x[,"n_33"])*log10(60 - 8*r_cm1c - 11*r_cm1c^2 + 26*r_cm1c^3 - 16*r_cm1c^4) + (x[,"n_13"] + x[,"n_32"])*log10(51 + 16*r_cm1c - 29*r_cm1c^2 + 38*r_cm1c^3 - 16*r_cm1c^4) + (x[,"n_03"] + x[,"n_42"])*log10(9 - 13*r_cm1c + 17*r_cm1c^2 - 11*r_cm1c^3 + 4*r_cm1c^4) + (x[,"n_02"] + x[,"n_43"])*log10(6 - 4*r_cm1c + 8*r_cm1c^2 - 5*r_cm1c^3 + 4*r_cm1c^4) + (x[,"n_20"] + x[,"n_25"])*log10(3 - 8*r_cm1c + 14*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_22"] + x[,"n_23"])*log10(32 + 3*r_cm1c + 5*r_cm1c^2 - 16*r_cm1c^3 + 8*r_cm1c^4) + (x[,"n_14"] + x[,"n_31"])*log10(15 - 4*r_cm1c + 39*r_cm1c^2 - 50*r_cm1c^3 + 24*r_cm1c^4) + (x[,"n_11"] + x[,"n_34"])*log10(24 - 20*r_cm1c + 33*r_cm1c^2 - 46*r_cm1c^3 + 24*r_cm1c^4)
logL_cm1c <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_00"] + x[,"n_05"] + x[,"n_40"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_cm1c)) + log(8 - 13*r_cm1c + 11*r_cm1c^2 - 6*r_cm1c^3)) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_cm1c)) + log(9 - 16*r_cm1c + 13*r_cm1c^2 - 6*r_cm1c^3)) + (x[,"n_10"] + x[,"n_15"] + x[,"n_30"] + x[,"n_35"])*(log(pmax(1e-6,r_cm1c)) + log(1 - 2*r_cm1c + 2*r_cm1c^2 - r_cm1c^3)) + (x[,"n_21"] + x[,"n_24"])*log(33 + 7*r_cm1c - 43*r_cm1c^2 + 72*r_cm1c^3 - 36*r_cm1c^4) + (x[,"n_12"] + x[,"n_33"])*log(60 - 8*r_cm1c - 11*r_cm1c^2 + 26*r_cm1c^3 - 16*r_cm1c^4) + (x[,"n_13"] + x[,"n_32"])*log(51 + 16*r_cm1c - 29*r_cm1c^2 + 38*r_cm1c^3 - 16*r_cm1c^4) + (x[,"n_03"] + x[,"n_42"])*log(9 - 13*r_cm1c + 17*r_cm1c^2 - 11*r_cm1c^3 + 4*r_cm1c^4) + (x[,"n_02"] + x[,"n_43"])*log(6 - 4*r_cm1c + 8*r_cm1c^2 - 5*r_cm1c^3 + 4*r_cm1c^4) + (x[,"n_20"] + x[,"n_25"])*log(3 - 8*r_cm1c + 14*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_22"] + x[,"n_23"])*log(32 + 3*r_cm1c + 5*r_cm1c^2 - 16*r_cm1c^3 + 8*r_cm1c^4) + (x[,"n_14"] + x[,"n_31"])*log(15 - 4*r_cm1c + 39*r_cm1c^2 - 50*r_cm1c^3 + 24*r_cm1c^4) + (x[,"n_11"] + x[,"n_34"])*log(24 - 20*r_cm1c + 33*r_cm1c^2 - 46*r_cm1c^3 + 24*r_cm1c^4)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 2*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + (n01 + n04 + n11 + n14 + n20 + n21 + n22 + n23 + n24 + n25 + n31 + n34 + n41 + n44)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n40)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n45)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n44)*(2*log(pmax(1e-6,r)) + log(2 - 3*r + 2*r^2)) + (n04 + n41)*(2*log(pmax(1e-6,1 - r)) + log(1 - r + 2*r^2)) + (n10 + n35)*(2*log(pmax(1e-6,r)) + log(3 - 6*r + 4*r^2)) + (n15 + n30)*(2*log(pmax(1e-6,1 - r)) + log(1 - 2*r + 4*r^2)) + (n11 + n34)*(log(pmax(1e-6,r)) + log(8 - 13*r + 14*r^2 - 8*r^3)) + (n02 + n43)*(log(pmax(1e-6,r)) + log(6 - 9*r + 9*r^2 - 4*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(1 - 3*r + 4*r^2 - 2*r^3)) + (n22 + n23)*log(9 + 11*r - 19*r^2 + 16*r^3 - 8*r^4) + (n14 + n31)*log(1 + 8*r - 19*r^2 + 18*r^3 - 8*r^4) + (n03 + n42)*log(2 + r - 6*r^2 + 7*r^3 - 4*r^4) + (n21 + n24)*log(6 - 13*r + 25*r^2 - 24*r^3 + 12*r^4) + (n12 + n33)*log(18 - 12*r + 39*r^2 - 34*r^3 + 16*r^4) + (n13 + n32)*log(27 - 28*r + 33*r^2 - 30*r^3 + 16*r^4)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cr <- 3*(x[,"n_20"] + x[,"n_25"])*log10(2) + 2*(x[,"n_15"] + x[,"n_30"])*log10(2) - (x[,"n_14"] + x[,"n_31"])*log10(2) - (x[,"n_11"] + x[,"n_34"])*log10(2) + 2*(x[,"n_10"] + x[,"n_35"])*log10(2) + 4*(x[,"n_05"] + x[,"n_40"])*log10(2) + 2*(x[,"n_04"] + x[,"n_41"])*log10(2) + 2*(x[,"n_01"] + x[,"n_44"])*log10(2) + 4*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_21"] + x[,"n_24"])*(-log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(37)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(37)) + (x[,"n_05"] + x[,"n_40"])*(3*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cr)) + 3*log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_cr)) + log10(2 - 3*r_cr + 2*r_cr^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(1 - r_cr + 2*r_cr^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_cr)) + log10(3 - 6*r_cr + 4*r_cr^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(1 - 2*r_cr + 4*r_cr^2)) + (x[,"n_11"] + x[,"n_34"])*(log10(pmax(1e-6,r_cr)) + log10(8 - 13*r_cr + 14*r_cr^2 - 8*r_cr^3)) + (x[,"n_02"] + x[,"n_43"])*(log10(pmax(1e-6,r_cr)) + log10(6 - 9*r_cr + 9*r_cr^2 - 4*r_cr^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_cr)) + log10(1 - 3*r_cr + 4*r_cr^2 - 2*r_cr^3)) + (x[,"n_22"] + x[,"n_23"])*log10(9 + 11*r_cr - 19*r_cr^2 + 16*r_cr^3 - 8*r_cr^4) + (x[,"n_14"] + x[,"n_31"])*log10(1 + 8*r_cr - 19*r_cr^2 + 18*r_cr^3 - 8*r_cr^4) + (x[,"n_03"] + x[,"n_42"])*log10(2 + r_cr - 6*r_cr^2 + 7*r_cr^3 - 4*r_cr^4) + (x[,"n_21"] + x[,"n_24"])*log10(6 - 13*r_cr + 25*r_cr^2 - 24*r_cr^3 + 12*r_cr^4) + (x[,"n_12"] + x[,"n_33"])*log10(18 - 12*r_cr + 39*r_cr^2 - 34*r_cr^3 + 16*r_cr^4) + (x[,"n_13"] + x[,"n_32"])*log10(27 - 28*r_cr + 33*r_cr^2 - 30*r_cr^3 + 16*r_cr^4)
logL_cr <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_40"])*(3*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cr)) + 3*log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,r_cr)) + log(2 - 3*r_cr + 2*r_cr^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_cr)) + log(1 - r_cr + 2*r_cr^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log(pmax(1e-6,r_cr)) + log(3 - 6*r_cr + 4*r_cr^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_cr)) + log(1 - 2*r_cr + 4*r_cr^2)) + (x[,"n_11"] + x[,"n_34"])*(log(pmax(1e-6,r_cr)) + log(8 - 13*r_cr + 14*r_cr^2 - 8*r_cr^3)) + (x[,"n_02"] + x[,"n_43"])*(log(pmax(1e-6,r_cr)) + log(6 - 9*r_cr + 9*r_cr^2 - 4*r_cr^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_cr)) + log(1 - 3*r_cr + 4*r_cr^2 - 2*r_cr^3)) + (x[,"n_22"] + x[,"n_23"])*log(9 + 11*r_cr - 19*r_cr^2 + 16*r_cr^3 - 8*r_cr^4) + (x[,"n_14"] + x[,"n_31"])*log(1 + 8*r_cr - 19*r_cr^2 + 18*r_cr^3 - 8*r_cr^4) + (x[,"n_03"] + x[,"n_42"])*log(2 + r_cr - 6*r_cr^2 + 7*r_cr^3 - 4*r_cr^4) + (x[,"n_21"] + x[,"n_24"])*log(6 - 13*r_cr + 25*r_cr^2 - 24*r_cr^3 + 12*r_cr^4) + (x[,"n_12"] + x[,"n_33"])*log(18 - 12*r_cr + 39*r_cr^2 - 34*r_cr^3 + 16*r_cr^4) + (x[,"n_13"] + x[,"n_32"])*log(27 - 28*r_cr + 33*r_cr^2 - 30*r_cr^3 + 16*r_cr^4)
logL_rc <- function(r,n01,n04,n11,n14,n20,n21,n22,n23,n24,n25,n31,n34,n41,n44,n00,n02,n03,n05,n10,n12,n13,n15,n30,n32,n33,n35,n40,n42,n43,n45) {
L <- (n01 + n04 + n11 + n14 + n20 + n21 + n22 + n23 + n24 + n25 + n31 + n34 + n41 + n44)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n05 + n40)*(log(3 - 2*r) + 3*log(pmax(1e-6,r))) + (n00 + n45)*(3*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n04 + n41)*(2*log(pmax(1e-6,r)) + log(3 - 3*r + 2*r^2)) + (n01 + n44)*(2*log(pmax(1e-6,1 - r)) + log(2 - r + 2*r^2)) + (n15 + n30)*(2*log(pmax(1e-6,r)) + log(9 - 14*r + 8*r^2)) + (n10 + n35)*(2*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 8*r^2)) + (n14 + n31)*(log(pmax(1e-6,r)) + log(12 - 13*r + 14*r^2 - 8*r^3)) + (n03 + n42)*(log(pmax(1e-6,r)) + log(9 - 6*r + 4*r^2 - 4*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(3 - 7*r + 8*r^2 - 4*r^3)) + (n22 + n23)*log(21 + 8*r - 16*r^2 + 16*r^3 - 8*r^4) + (n11 + n34)*log(5 + 4*r - 19*r^2 + 18*r^3 - 8*r^4) + (n02 + n43)*log(3 + 7*r - 18*r^2 + 12*r^3 - 4*r^4) + (n21 + n24)*log(9 - 11*r + 23*r^2 - 24*r^3 + 12*r^4) + (n12 + n33)*log(55 - 52*r + 48*r^2 - 40*r^3 + 16*r^4) + (n13 + n32)*log(27 + 12*r + 24*r^2 - 24*r^3 + 16*r^4)
return(L)}
interlogL_rc <- function(n01,n04,n11,n14,n20,n21,n22,n23,n24,n25,n31,n34,n41,n44,n00,n02,n03,n05,n10,n12,n13,n15,n30,n32,n33,n35,n40,n42,n43,n45) {
optimize(logL_rc,c(0,0.5), n01,n04,n11,n14,n20,n21,n22,n23,n24,n25,n31,n34,n41,n44,n00,n02,n03,n05,n10,n12,n13,n15,n30,n32,n33,n35,n40,n42,n43,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_04"],x[,"n_11"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_34"],x[,"n_41"],x[,"n_44"],x[,"n_00"],x[,"n_02"],x[,"n_03"],x[,"n_05"],x[,"n_10"],x[,"n_12"],x[,"n_13"],x[,"n_15"],x[,"n_30"],x[,"n_32"],x[,"n_33"],x[,"n_35"],x[,"n_40"],x[,"n_42"],x[,"n_43"],x[,"n_45"])
LOD_rc <- (x[,"n_20"] + x[,"n_25"])*log10(2) - 2*(x[,"n_14"] + x[,"n_31"])*log10(2) - 2*(x[,"n_11"] + x[,"n_34"])*log10(2) + 2*(x[,"n_05"] + x[,"n_40"])*log10(2) + (x[,"n_04"] + x[,"n_41"])*log10(2) + (x[,"n_01"] + x[,"n_44"])*log10(2) + 2*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_22"] + x[,"n_23"])*(-log10(2) + 2*log10(3) + log10(5)) - (x[,"n_21"] + x[,"n_24"])*log10(7) - (x[,"n_03"] + x[,"n_42"])*(-2*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_43"])*(-2*log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_32"])*log10(37) - (x[,"n_12"] + x[,"n_33"])*log10(37) + (x[,"n_05"] + x[,"n_40"])*(log10(3 - 2*r_rc) + 3*log10(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_45"])*(3*log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_rc)) + log10(3 - 3*r_rc + 2*r_rc^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(2 - r_rc + 2*r_rc^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_rc)) + log10(9 - 14*r_rc + 8*r_rc^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(3 - 2*r_rc + 8*r_rc^2)) + (x[,"n_14"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(12 - 13*r_rc + 14*r_rc^2 - 8*r_rc^3)) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,r_rc)) + log10(9 - 6*r_rc + 4*r_rc^2 - 4*r_rc^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_rc)) + log10(3 - 7*r_rc + 8*r_rc^2 - 4*r_rc^3)) + (x[,"n_22"] + x[,"n_23"])*log10(21 + 8*r_rc - 16*r_rc^2 + 16*r_rc^3 - 8*r_rc^4) + (x[,"n_11"] + x[,"n_34"])*log10(5 + 4*r_rc - 19*r_rc^2 + 18*r_rc^3 - 8*r_rc^4) + (x[,"n_02"] + x[,"n_43"])*log10(3 + 7*r_rc - 18*r_rc^2 + 12*r_rc^3 - 4*r_rc^4) + (x[,"n_21"] + x[,"n_24"])*log10(9 - 11*r_rc + 23*r_rc^2 - 24*r_rc^3 + 12*r_rc^4) + (x[,"n_12"] + x[,"n_33"])*log10(55 - 52*r_rc + 48*r_rc^2 - 40*r_rc^3 + 16*r_rc^4) + (x[,"n_13"] + x[,"n_32"])*log10(27 + 12*r_rc + 24*r_rc^2 - 24*r_rc^3 + 16*r_rc^4)
logL_rc <- (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_40"])*(log(3 - 2*r_rc) + 3*log(pmax(1e-6,r_rc))) + (x[,"n_00"] + x[,"n_45"])*(3*log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,r_rc)) + log(3 - 3*r_rc + 2*r_rc^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_rc)) + log(2 - r_rc + 2*r_rc^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,r_rc)) + log(9 - 14*r_rc + 8*r_rc^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_rc)) + log(3 - 2*r_rc + 8*r_rc^2)) + (x[,"n_14"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(12 - 13*r_rc + 14*r_rc^2 - 8*r_rc^3)) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,r_rc)) + log(9 - 6*r_rc + 4*r_rc^2 - 4*r_rc^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_rc)) + log(3 - 7*r_rc + 8*r_rc^2 - 4*r_rc^3)) + (x[,"n_22"] + x[,"n_23"])*log(21 + 8*r_rc - 16*r_rc^2 + 16*r_rc^3 - 8*r_rc^4) + (x[,"n_11"] + x[,"n_34"])*log(5 + 4*r_rc - 19*r_rc^2 + 18*r_rc^3 - 8*r_rc^4) + (x[,"n_02"] + x[,"n_43"])*log(3 + 7*r_rc - 18*r_rc^2 + 12*r_rc^3 - 4*r_rc^4) + (x[,"n_21"] + x[,"n_24"])*log(9 - 11*r_rc + 23*r_rc^2 - 24*r_rc^3 + 12*r_rc^4) + (x[,"n_12"] + x[,"n_33"])*log(55 - 52*r_rc + 48*r_rc^2 - 40*r_rc^3 + 16*r_rc^4) + (x[,"n_13"] + x[,"n_32"])*log(27 + 12*r_rc + 24*r_rc^2 - 24*r_rc^3 + 16*r_rc^4)
logL_rm2c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n05 + n40)*(log(3 - 2*r) + log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n00 + n45)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 + 2*r)) + (n15 + n30)*(log(pmax(1e-6,r)) + log(12 - 19*r + 18*r^2 - 8*r^3)) + (n04 + n41)*(log(pmax(1e-6,r)) + log(12 - 14*r + 11*r^2 - 6*r^3)) + (n21 + n24)*log(66 + 5*r - 41*r^2 + 72*r^3 - 36*r^4) + (n13 + n32)*log(111 - 4*r + 4*r^2 + 16*r^3 - 16*r^4) + (n12 + n33)*log(111 + 12*r - 44*r^2 + 48*r^3 - 16*r^4) + (n10 + n35)*log(3 + 4*r - 13*r^2 + 14*r^3 - 8*r^4) + (n01 + n44)*log(3 + 7*r - 17*r^2 + 13*r^3 - 6*r^4) + (n03 + n42)*log(9 + 3*r + 2*r^2 + 4*r^4) + (n02 + n43)*log(18 - 23*r + 26*r^2 - 16*r^3 + 4*r^4) + (n22 + n23)*log(65 + 4*r + 4*r^2 - 16*r^3 + 8*r^4) + (n20 + n25)*log(9 - 17*r + 29*r^2 - 24*r^3 + 12*r^4) + (n11 + n34)*log(39 - 16*r + 39*r^2 - 50*r^3 + 24*r^4) + (n14 + n31)*log(36 - 8*r + 33*r^2 - 46*r^3 + 24*r^4)
return(L)}
interlogL_rm2c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
optimize(logL_rm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_22"],x[,"n_23"])
LOD_rm2c <- 2*(x[,"n_05"] + x[,"n_40"])*log10(2) + 2*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_15"] + x[,"n_30"])*log10(3) - (x[,"n_10"] + x[,"n_35"])*log10(3) - (x[,"n_14"] + x[,"n_31"])*(2*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_34"])*(2*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_42"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_43"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_04"] + x[,"n_41"])*(-log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_44"])*(-log10(2) + log10(7)) - (x[,"n_20"] + x[,"n_25"])*(-log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(log10(5) + log10(13)) - (x[,"n_22"] + x[,"n_23"])*(-log10(2) + log10(7) + log10(19)) - (x[,"n_13"] + x[,"n_32"])*(log10(3) + log10(37)) - (x[,"n_12"] + x[,"n_33"])*(log10(3) + log10(37)) + (x[,"n_05"] + x[,"n_40"])*(log10(3 - 2*r_rm2c) + log10(pmax(1e-6,1 - r_rm2c)) + 2*log10(pmax(1e-6,r_rm2c))) + (x[,"n_00"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c)) + log10(1 + 2*r_rm2c)) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,r_rm2c)) + log10(12 - 19*r_rm2c + 18*r_rm2c^2 - 8*r_rm2c^3)) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_rm2c)) + log10(12 - 14*r_rm2c + 11*r_rm2c^2 - 6*r_rm2c^3)) + (x[,"n_21"] + x[,"n_24"])*log10(66 + 5*r_rm2c - 41*r_rm2c^2 + 72*r_rm2c^3 - 36*r_rm2c^4) + (x[,"n_13"] + x[,"n_32"])*log10(111 - 4*r_rm2c + 4*r_rm2c^2 + 16*r_rm2c^3 - 16*r_rm2c^4) + (x[,"n_12"] + x[,"n_33"])*log10(111 + 12*r_rm2c - 44*r_rm2c^2 + 48*r_rm2c^3 - 16*r_rm2c^4) + (x[,"n_10"] + x[,"n_35"])*log10(3 + 4*r_rm2c - 13*r_rm2c^2 + 14*r_rm2c^3 - 8*r_rm2c^4) + (x[,"n_01"] + x[,"n_44"])*log10(3 + 7*r_rm2c - 17*r_rm2c^2 + 13*r_rm2c^3 - 6*r_rm2c^4) + (x[,"n_03"] + x[,"n_42"])*log10(9 + 3*r_rm2c + 2*r_rm2c^2 + 4*r_rm2c^4) + (x[,"n_02"] + x[,"n_43"])*log10(18 - 23*r_rm2c + 26*r_rm2c^2 - 16*r_rm2c^3 + 4*r_rm2c^4) + (x[,"n_22"] + x[,"n_23"])*log10(65 + 4*r_rm2c + 4*r_rm2c^2 - 16*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_20"] + x[,"n_25"])*log10(9 - 17*r_rm2c + 29*r_rm2c^2 - 24*r_rm2c^3 + 12*r_rm2c^4) + (x[,"n_11"] + x[,"n_34"])*log10(39 - 16*r_rm2c + 39*r_rm2c^2 - 50*r_rm2c^3 + 24*r_rm2c^4) + (x[,"n_14"] + x[,"n_31"])*log10(36 - 8*r_rm2c + 33*r_rm2c^2 - 46*r_rm2c^3 + 24*r_rm2c^4)
logL_rm2c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_40"])*(log(3 - 2*r_rm2c) + log(pmax(1e-6,1 - r_rm2c)) + 2*log(pmax(1e-6,r_rm2c))) + (x[,"n_00"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c)) + log(1 + 2*r_rm2c)) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,r_rm2c)) + log(12 - 19*r_rm2c + 18*r_rm2c^2 - 8*r_rm2c^3)) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_rm2c)) + log(12 - 14*r_rm2c + 11*r_rm2c^2 - 6*r_rm2c^3)) + (x[,"n_21"] + x[,"n_24"])*log(66 + 5*r_rm2c - 41*r_rm2c^2 + 72*r_rm2c^3 - 36*r_rm2c^4) + (x[,"n_13"] + x[,"n_32"])*log(111 - 4*r_rm2c + 4*r_rm2c^2 + 16*r_rm2c^3 - 16*r_rm2c^4) + (x[,"n_12"] + x[,"n_33"])*log(111 + 12*r_rm2c - 44*r_rm2c^2 + 48*r_rm2c^3 - 16*r_rm2c^4) + (x[,"n_10"] + x[,"n_35"])*log(3 + 4*r_rm2c - 13*r_rm2c^2 + 14*r_rm2c^3 - 8*r_rm2c^4) + (x[,"n_01"] + x[,"n_44"])*log(3 + 7*r_rm2c - 17*r_rm2c^2 + 13*r_rm2c^3 - 6*r_rm2c^4) + (x[,"n_03"] + x[,"n_42"])*log(9 + 3*r_rm2c + 2*r_rm2c^2 + 4*r_rm2c^4) + (x[,"n_02"] + x[,"n_43"])*log(18 - 23*r_rm2c + 26*r_rm2c^2 - 16*r_rm2c^3 + 4*r_rm2c^4) + (x[,"n_22"] + x[,"n_23"])*log(65 + 4*r_rm2c + 4*r_rm2c^2 - 16*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_20"] + x[,"n_25"])*log(9 - 17*r_rm2c + 29*r_rm2c^2 - 24*r_rm2c^3 + 12*r_rm2c^4) + (x[,"n_11"] + x[,"n_34"])*log(39 - 16*r_rm2c + 39*r_rm2c^2 - 50*r_rm2c^3 + 24*r_rm2c^4) + (x[,"n_14"] + x[,"n_31"])*log(36 - 8*r_rm2c + 33*r_rm2c^2 - 46*r_rm2c^3 + 24*r_rm2c^4)
logL_rm1c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n05 + n40)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n45)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n01 + n44)*(log(pmax(1e-6,r)) + log(4 + 10*r - 11*r^2 + 6*r^3)) + (n10 + n35)*(log(pmax(1e-6,r)) + log(4 + 7*r - 10*r^2 + 8*r^3)) + (n11 + n34)*log(12 + 56*r - 33*r^2 + 46*r^3 - 24*r^4) + (n14 + n31)*log(57 - 32*r - 39*r^2 + 50*r^3 - 24*r^4) + (n20 + n25)*log(3 + 13*r - 25*r^2 + 24*r^3 - 12*r^4) + (n22 + n23)*log(67 - 8*r^2 + 16*r^3 - 8*r^4) + (n03 + n42)*log(18 - 17*r + 10*r^2 - 4*r^3 - 4*r^4) + (n02 + n43)*log(3 + 25*r - 26*r^2 + 20*r^3 - 4*r^4) + (n04 + n41)*log(9 - 15*r + 13*r^2 - 13*r^3 + 6*r^4) + (n15 + n30)*log(9 - 20*r + 25*r^2 - 22*r^3 + 8*r^4) + (n12 + n33)*log(93 + 20*r + 56*r^2 - 56*r^3 + 16*r^4) + (n13 + n32)*log(129 - 28*r - 16*r^2 - 8*r^3 + 16*r^4) + (n21 + n24)*log(66 - 13*r + 49*r^2 - 72*r^3 + 36*r^4)
return(L)}
interlogL_rm1c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
optimize(logL_rm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_22"],x[,"n_23"])
LOD_rm1c <- 2*(x[,"n_05"] + x[,"n_40"])*log10(2) + 2*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_15"] + x[,"n_30"])*log10(3) - (x[,"n_10"] + x[,"n_35"])*log10(3) - (x[,"n_14"] + x[,"n_31"])*(2*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_34"])*(2*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_42"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_43"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_04"] + x[,"n_41"])*(-log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_44"])*(-log10(2) + log10(7)) - (x[,"n_20"] + x[,"n_25"])*(-log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(log10(5) + log10(13)) - (x[,"n_22"] + x[,"n_23"])*(-log10(2) + log10(7) + log10(19)) - (x[,"n_13"] + x[,"n_32"])*(log10(3) + log10(37)) - (x[,"n_12"] + x[,"n_33"])*(log10(3) + log10(37)) + (x[,"n_05"] + x[,"n_40"])*(log10(3 - 2*r_rm1c) + 2*log10(pmax(1e-6,1 - r_rm1c)) + log10(pmax(1e-6,r_rm1c))) + (x[,"n_00"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_rm1c)) + 2*log10(pmax(1e-6,r_rm1c)) + log10(1 + 2*r_rm1c)) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm1c)) + log10(4 + 10*r_rm1c - 11*r_rm1c^2 + 6*r_rm1c^3)) + (x[,"n_10"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm1c)) + log10(4 + 7*r_rm1c - 10*r_rm1c^2 + 8*r_rm1c^3)) + (x[,"n_11"] + x[,"n_34"])*log10(12 + 56*r_rm1c - 33*r_rm1c^2 + 46*r_rm1c^3 - 24*r_rm1c^4) + (x[,"n_14"] + x[,"n_31"])*log10(57 - 32*r_rm1c - 39*r_rm1c^2 + 50*r_rm1c^3 - 24*r_rm1c^4) + (x[,"n_20"] + x[,"n_25"])*log10(3 + 13*r_rm1c - 25*r_rm1c^2 + 24*r_rm1c^3 - 12*r_rm1c^4) + (x[,"n_22"] + x[,"n_23"])*log10(67 - 8*r_rm1c^2 + 16*r_rm1c^3 - 8*r_rm1c^4) + (x[,"n_03"] + x[,"n_42"])*log10(18 - 17*r_rm1c + 10*r_rm1c^2 - 4*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_02"] + x[,"n_43"])*log10(3 + 25*r_rm1c - 26*r_rm1c^2 + 20*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_04"] + x[,"n_41"])*log10(9 - 15*r_rm1c + 13*r_rm1c^2 - 13*r_rm1c^3 + 6*r_rm1c^4) + (x[,"n_15"] + x[,"n_30"])*log10(9 - 20*r_rm1c + 25*r_rm1c^2 - 22*r_rm1c^3 + 8*r_rm1c^4) + (x[,"n_12"] + x[,"n_33"])*log10(93 + 20*r_rm1c + 56*r_rm1c^2 - 56*r_rm1c^3 + 16*r_rm1c^4) + (x[,"n_13"] + x[,"n_32"])*log10(129 - 28*r_rm1c - 16*r_rm1c^2 - 8*r_rm1c^3 + 16*r_rm1c^4) + (x[,"n_21"] + x[,"n_24"])*log10(66 - 13*r_rm1c + 49*r_rm1c^2 - 72*r_rm1c^3 + 36*r_rm1c^4)
logL_rm1c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_40"])*(log(3 - 2*r_rm1c) + 2*log(pmax(1e-6,1 - r_rm1c)) + log(pmax(1e-6,r_rm1c))) + (x[,"n_00"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_rm1c)) + 2*log(pmax(1e-6,r_rm1c)) + log(1 + 2*r_rm1c)) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_rm1c)) + log(4 + 10*r_rm1c - 11*r_rm1c^2 + 6*r_rm1c^3)) + (x[,"n_10"] + x[,"n_35"])*(log(pmax(1e-6,r_rm1c)) + log(4 + 7*r_rm1c - 10*r_rm1c^2 + 8*r_rm1c^3)) + (x[,"n_11"] + x[,"n_34"])*log(12 + 56*r_rm1c - 33*r_rm1c^2 + 46*r_rm1c^3 - 24*r_rm1c^4) + (x[,"n_14"] + x[,"n_31"])*log(57 - 32*r_rm1c - 39*r_rm1c^2 + 50*r_rm1c^3 - 24*r_rm1c^4) + (x[,"n_20"] + x[,"n_25"])*log(3 + 13*r_rm1c - 25*r_rm1c^2 + 24*r_rm1c^3 - 12*r_rm1c^4) + (x[,"n_22"] + x[,"n_23"])*log(67 - 8*r_rm1c^2 + 16*r_rm1c^3 - 8*r_rm1c^4) + (x[,"n_03"] + x[,"n_42"])*log(18 - 17*r_rm1c + 10*r_rm1c^2 - 4*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_02"] + x[,"n_43"])*log(3 + 25*r_rm1c - 26*r_rm1c^2 + 20*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_04"] + x[,"n_41"])*log(9 - 15*r_rm1c + 13*r_rm1c^2 - 13*r_rm1c^3 + 6*r_rm1c^4) + (x[,"n_15"] + x[,"n_30"])*log(9 - 20*r_rm1c + 25*r_rm1c^2 - 22*r_rm1c^3 + 8*r_rm1c^4) + (x[,"n_12"] + x[,"n_33"])*log(93 + 20*r_rm1c + 56*r_rm1c^2 - 56*r_rm1c^3 + 16*r_rm1c^4) + (x[,"n_13"] + x[,"n_32"])*log(129 - 28*r_rm1c - 16*r_rm1c^2 - 8*r_rm1c^3 + 16*r_rm1c^4) + (x[,"n_21"] + x[,"n_24"])*log(66 - 13*r_rm1c + 49*r_rm1c^2 - 72*r_rm1c^3 + 36*r_rm1c^4)
logL_rr <- function(r,n01,n04,n11,n14,n20,n21,n22,n23,n24,n25,n31,n34,n41,n44,n00,n02,n03,n05,n10,n12,n13,n15,n30,n32,n33,n35,n40,n42,n43,n45) {
L <- (n01 + n04 + n11 + n14 + n20 + n21 + n22 + n23 + n24 + n25 + n31 + n34 + n41 + n44)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n05 + n40)*(log(3 - 2*r) + 3*log(pmax(1e-6,1 - r))) + (n00 + n45)*(3*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n10 + n35)*(2*log(pmax(1e-6,r)) + log(3 + 6*r - 8*r^2)) + (n15 + n30)*(2*log(pmax(1e-6,1 - r)) + log(1 + 10*r - 8*r^2)) + (n04 + n41)*(2*log(pmax(1e-6,1 - r)) + log(2 + r - 2*r^2)) + (n01 + n44)*(2*log(pmax(1e-6,r)) + log(1 + 3*r - 2*r^2)) + (n02 + n43)*(log(pmax(1e-6,r)) + log(3 + 18*r - 24*r^2 + 4*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(1 + 3*r - 8*r^2 + 4*r^3)) + (n11 + n34)*(log(pmax(1e-6,r)) + log(4 + 13*r - 14*r^2 + 8*r^3)) + (n13 + n32)*log(57 - 44*r + 12*r^2 - 16*r^4) + (n12 + n33)*log(9 + 84*r - 84*r^2 + 64*r^3 - 16*r^4) + (n21 + n24)*log(3 + 19*r - 31*r^2 + 24*r^3 - 12*r^4) + (n03 + n42)*log(1 + 17*r - 30*r^2 + 8*r^3 + 4*r^4) + (n14 + n31)*log(11 - 20*r + 19*r^2 - 18*r^3 + 8*r^4) + (n22 + n23)*log(27 - 20*r + 28*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_rr <- function(n01,n04,n11,n14,n20,n21,n22,n23,n24,n25,n31,n34,n41,n44,n00,n02,n03,n05,n10,n12,n13,n15,n30,n32,n33,n35,n40,n42,n43,n45) {
optimize(logL_rr,c(0,0.5), n01,n04,n11,n14,n20,n21,n22,n23,n24,n25,n31,n34,n41,n44,n00,n02,n03,n05,n10,n12,n13,n15,n30,n32,n33,n35,n40,n42,n43,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_01"],x[,"n_04"],x[,"n_11"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_34"],x[,"n_41"],x[,"n_44"],x[,"n_00"],x[,"n_02"],x[,"n_03"],x[,"n_05"],x[,"n_10"],x[,"n_12"],x[,"n_13"],x[,"n_15"],x[,"n_30"],x[,"n_32"],x[,"n_33"],x[,"n_35"],x[,"n_40"],x[,"n_42"],x[,"n_43"],x[,"n_45"])
LOD_rr <- (x[,"n_20"] + x[,"n_25"])*log10(2) - 2*(x[,"n_14"] + x[,"n_31"])*log10(2) - 2*(x[,"n_11"] + x[,"n_34"])*log10(2) + 2*(x[,"n_05"] + x[,"n_40"])*log10(2) + (x[,"n_04"] + x[,"n_41"])*log10(2) + (x[,"n_01"] + x[,"n_44"])*log10(2) + 2*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_22"] + x[,"n_23"])*(-log10(2) + 2*log10(3) + log10(5)) - (x[,"n_21"] + x[,"n_24"])*log10(7) - (x[,"n_03"] + x[,"n_42"])*(-2*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_43"])*(-2*log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_32"])*log10(37) - (x[,"n_12"] + x[,"n_33"])*log10(37) + (x[,"n_05"] + x[,"n_40"])*(log10(3 - 2*r_rr) + 3*log10(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_45"])*(3*log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_10"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_rr)) + log10(3 + 6*r_rr - 8*r_rr^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(1 + 10*r_rr - 8*r_rr^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(2 + r_rr - 2*r_rr^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(1 + 3*r_rr - 2*r_rr^2)) + (x[,"n_02"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(3 + 18*r_rr - 24*r_rr^2 + 4*r_rr^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(1 + 3*r_rr - 8*r_rr^2 + 4*r_rr^3)) + (x[,"n_11"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(4 + 13*r_rr - 14*r_rr^2 + 8*r_rr^3)) + (x[,"n_13"] + x[,"n_32"])*log10(57 - 44*r_rr + 12*r_rr^2 - 16*r_rr^4) + (x[,"n_12"] + x[,"n_33"])*log10(9 + 84*r_rr - 84*r_rr^2 + 64*r_rr^3 - 16*r_rr^4) + (x[,"n_21"] + x[,"n_24"])*log10(3 + 19*r_rr - 31*r_rr^2 + 24*r_rr^3 - 12*r_rr^4) + (x[,"n_03"] + x[,"n_42"])*log10(1 + 17*r_rr - 30*r_rr^2 + 8*r_rr^3 + 4*r_rr^4) + (x[,"n_14"] + x[,"n_31"])*log10(11 - 20*r_rr + 19*r_rr^2 - 18*r_rr^3 + 8*r_rr^4) + (x[,"n_22"] + x[,"n_23"])*log10(27 - 20*r_rr + 28*r_rr^2 - 16*r_rr^3 + 8*r_rr^4)
logL_rr <- (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_40"])*(log(3 - 2*r_rr) + 3*log(pmax(1e-6,1 - r_rr))) + (x[,"n_00"] + x[,"n_45"])*(3*log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_10"] + x[,"n_35"])*(2*log(pmax(1e-6,r_rr)) + log(3 + 6*r_rr - 8*r_rr^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_rr)) + log(1 + 10*r_rr - 8*r_rr^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(2 + r_rr - 2*r_rr^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(1 + 3*r_rr - 2*r_rr^2)) + (x[,"n_02"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(3 + 18*r_rr - 24*r_rr^2 + 4*r_rr^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(1 + 3*r_rr - 8*r_rr^2 + 4*r_rr^3)) + (x[,"n_11"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(4 + 13*r_rr - 14*r_rr^2 + 8*r_rr^3)) + (x[,"n_13"] + x[,"n_32"])*log(57 - 44*r_rr + 12*r_rr^2 - 16*r_rr^4) + (x[,"n_12"] + x[,"n_33"])*log(9 + 84*r_rr - 84*r_rr^2 + 64*r_rr^3 - 16*r_rr^4) + (x[,"n_21"] + x[,"n_24"])*log(3 + 19*r_rr - 31*r_rr^2 + 24*r_rr^3 - 12*r_rr^4) + (x[,"n_03"] + x[,"n_42"])*log(1 + 17*r_rr - 30*r_rr^2 + 8*r_rr^3 + 4*r_rr^4) + (x[,"n_14"] + x[,"n_31"])*log(11 - 20*r_rr + 19*r_rr^2 - 18*r_rr^3 + 8*r_rr^4) + (x[,"n_22"] + x[,"n_23"])*log(27 - 20*r_rr + 28*r_rr^2 - 16*r_rr^3 + 8*r_rr^4)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_cr,r_rc,r_rm2c,r_rm1c,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_cr,LOD_rc,LOD_rm2c,LOD_rm1c,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_cr,logL_rc,logL_rm2c,logL_rm1c,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","coupling repulsion","repulsion coupling","repulsion mixed2c","repulsion mixed1c","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_2.4 <- function(x){
logL_cc <- function(r,n01,n02,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n44,n45,n03,n43) {
L <- (-2*n01 - 3*n02 - 3*n04 - 2*n05 - 2*n11 - n12 - n13 - n14 - 2*n15 - 2*n21 - 3*n22 + n23 - 3*n24 - 2*n25 - 2*n31 - n32 - n33 - n34 - 2*n35 - 2*n41 - 3*n42 - 3*n44 - 2*n45)*log(2) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n21 + n25)*log(7) + 3*(n01 + n45)*log(pmax(1e-6,1 - r)) + 3*(n05 + n41)*log(pmax(1e-6,r)) + (n03 + n21 + n25 + n43)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n42)*(log(7 - 4*r) + 2*log(pmax(1e-6,r))) + (n15 + n31)*(log(5 - 2*r) + 2*log(pmax(1e-6,r))) + (n11 + n35)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r)) + (n02 + n44)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r)) + (n13 + n33)*log(5 + 2*r - 2*r^2) + (n22 + n24)*log(21 + 2*r - 2*r^2) + (n14 + n32)*(log(pmax(1e-6,r)) + log(7 - 3*r + 2*r^2)) + n23*log(3 - 2*r + 2*r^2) + (n12 + n34)*log(6 - 7*r + 3*r^2 - 2*r^3)
return(L)}
interlogL_cc <- function(n01,n02,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n44,n45,n03,n43) {
optimize(logL_cc,c(0,0.5), n01,n02,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n44,n45,n03,n43, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_44"],x[,"n_45"],x[,"n_03"],x[,"n_43"])
LOD_cc <- 3*(x[,"n_05"] + x[,"n_41"])*log10(2) + 2*(x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log10(2) + 3*(x[,"n_01"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_32"])*log10(3) - (x[,"n_12"] + x[,"n_34"])*log10(3) - (x[,"n_04"] + x[,"n_42"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - x[,"n_23"]*(-log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_33"])*(-log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(43)) + 3*(x[,"n_01"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_05"] + x[,"n_41"])*log10(pmax(1e-6,r_cc)) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_42"])*(log10(7 - 4*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_31"])*(log10(5 - 2*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 2*r_cc)) + (x[,"n_02"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 4*r_cc)) + (x[,"n_13"] + x[,"n_33"])*log10(5 + 2*r_cc - 2*r_cc^2) + (x[,"n_22"] + x[,"n_24"])*log10(21 + 2*r_cc - 2*r_cc^2) + (x[,"n_14"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(7 - 3*r_cc + 2*r_cc^2)) + x[,"n_23"]*log10(3 - 2*r_cc + 2*r_cc^2) + (x[,"n_12"] + x[,"n_34"])*log10(6 - 7*r_cc + 3*r_cc^2 - 2*r_cc^3)
logL_cc <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 3*x[,"n_22"] + x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_21"] + x[,"n_25"])*log(7) + 3*(x[,"n_01"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_05"] + x[,"n_41"])*log(pmax(1e-6,r_cc)) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_42"])*(log(7 - 4*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_31"])*(log(5 - 2*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 2*r_cc)) + (x[,"n_02"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 4*r_cc)) + (x[,"n_13"] + x[,"n_33"])*log(5 + 2*r_cc - 2*r_cc^2) + (x[,"n_22"] + x[,"n_24"])*log(21 + 2*r_cc - 2*r_cc^2) + (x[,"n_14"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(7 - 3*r_cc + 2*r_cc^2)) + x[,"n_23"]*log(3 - 2*r_cc + 2*r_cc^2) + (x[,"n_12"] + x[,"n_34"])*log(6 - 7*r_cc + 3*r_cc^2 - 2*r_cc^3)
logL_cm2c <- function(r,n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n31,n33,n35,n41,n42,n43,n44,n45,n12,n14,n32,n34) {
L <- (-n01 - 2*n02 - 3*n03 - 2*n04 - n05 - n11 - 2*n13 - n15 - n21 - 2*n22 - 2*n23 - 2*n24 - n25 - n31 - 2*n33 - n35 - n41 - 2*n42 - 3*n43 - 2*n44 - n45)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n13 + n33)*log(11) + (n01 + n45)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n41)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + n23*log(57 + 8*r - 8*r^2) + (n21 + n25)*log(3 - r + r^2) + (n15 + n31)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n22 + n24)*log(33 - 2*r + 2*r^2) + (n04 + n42)*(log(pmax(1e-6,r)) + log(6 - 7*r + 4*r^2)) + (n03 + n43)*log(9 - 8*r + 8*r^2) + (n02 + n44)*log(3 - 4*r + 5*r^2 - 4*r^3) + (n11 + n35)*log(3 - 4*r + 3*r^2 - 2*r^3) + (n14 + n32)*log(3 + 2*r + 3*r^2 - 2*r^3) + (n12 + n34)*log(6 - 2*r - 3*r^2 + 2*r^3)
return(L)}
interlogL_cm2c <- function(n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n31,n33,n35,n41,n42,n43,n44,n45,n12,n14,n32,n34) {
optimize(logL_cm2c,c(0,0.5), n01,n02,n03,n04,n05,n11,n13,n15,n21,n22,n23,n24,n25,n31,n33,n35,n41,n42,n43,n44,n45,n12,n14,n32,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_13"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_33"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_12"],x[,"n_14"],x[,"n_32"],x[,"n_34"])
LOD_cm2c <- 3*(x[,"n_05"] + x[,"n_41"])*log10(2) + 3*(x[,"n_01"] + x[,"n_45"])*log10(2) - (x[,"n_15"] + x[,"n_31"])*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_32"])*(-log10(2) + 2*log10(3)) - (x[,"n_12"] + x[,"n_34"])*(-log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_43"])*log10(7) - (x[,"n_04"] + x[,"n_42"])*(-2*log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_44"])*(-2*log10(2) + log10(7)) - (x[,"n_21"] + x[,"n_25"])*(-2*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(5) + log10(13)) - x[,"n_23"]*log10(59) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_41"])*(log10(pmax(1e-6,1 - r_cm2c)) + 2*log10(pmax(1e-6,r_cm2c))) + x[,"n_23"]*log10(57 + 8*r_cm2c - 8*r_cm2c^2) + (x[,"n_21"] + x[,"n_25"])*log10(3 - r_cm2c + r_cm2c^2) + (x[,"n_15"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm2c)) + log10(4 - 3*r_cm2c + 2*r_cm2c^2)) + (x[,"n_22"] + x[,"n_24"])*log10(33 - 2*r_cm2c + 2*r_cm2c^2) + (x[,"n_04"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm2c)) + log10(6 - 7*r_cm2c + 4*r_cm2c^2)) + (x[,"n_03"] + x[,"n_43"])*log10(9 - 8*r_cm2c + 8*r_cm2c^2) + (x[,"n_02"] + x[,"n_44"])*log10(3 - 4*r_cm2c + 5*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_11"] + x[,"n_35"])*log10(3 - 4*r_cm2c + 3*r_cm2c^2 - 2*r_cm2c^3) + (x[,"n_14"] + x[,"n_32"])*log10(3 + 2*r_cm2c + 3*r_cm2c^2 - 2*r_cm2c^3) + (x[,"n_12"] + x[,"n_34"])*log10(6 - 2*r_cm2c - 3*r_cm2c^2 + 2*r_cm2c^3)
logL_cm2c <- (-x[,"n_01"] - 2*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - 2*x[,"n_13"] - x[,"n_15"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - 2*x[,"n_33"] - x[,"n_35"] - x[,"n_41"] - 2*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"] - x[,"n_45"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_13"] + x[,"n_33"])*log(11) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_41"])*(log(pmax(1e-6,1 - r_cm2c)) + 2*log(pmax(1e-6,r_cm2c))) + x[,"n_23"]*log(57 + 8*r_cm2c - 8*r_cm2c^2) + (x[,"n_21"] + x[,"n_25"])*log(3 - r_cm2c + r_cm2c^2) + (x[,"n_15"] + x[,"n_31"])*(log(pmax(1e-6,r_cm2c)) + log(4 - 3*r_cm2c + 2*r_cm2c^2)) + (x[,"n_22"] + x[,"n_24"])*log(33 - 2*r_cm2c + 2*r_cm2c^2) + (x[,"n_04"] + x[,"n_42"])*(log(pmax(1e-6,r_cm2c)) + log(6 - 7*r_cm2c + 4*r_cm2c^2)) + (x[,"n_03"] + x[,"n_43"])*log(9 - 8*r_cm2c + 8*r_cm2c^2) + (x[,"n_02"] + x[,"n_44"])*log(3 - 4*r_cm2c + 5*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_11"] + x[,"n_35"])*log(3 - 4*r_cm2c + 3*r_cm2c^2 - 2*r_cm2c^3) + (x[,"n_14"] + x[,"n_32"])*log(3 + 2*r_cm2c + 3*r_cm2c^2 - 2*r_cm2c^3) + (x[,"n_12"] + x[,"n_34"])*log(6 - 2*r_cm2c - 3*r_cm2c^2 + 2*r_cm2c^3)
logL_cm1c <- function(r,n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n22,n23,n24,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n13,n33) {
L <- (-2*n01 - 3*n02 - 2*n03 - 3*n04 - 2*n05 - 2*n11 - n12 - n14 - 2*n15 - 2*n21 - 3*n22 - n23 - 3*n24 - 2*n25 - 2*n31 - n32 - n34 - 2*n35 - 2*n41 - 3*n42 - 2*n43 - 3*n44 - 2*n45)*log(2) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n41)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n45)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + n23*log(9 + 4*r - 4*r^2) + (n22 + n24)*log(21 + 2*r - 2*r^2) + (n13 + n33)*log(3 - r + r^2) + (n11 + n35)*(log(pmax(1e-6,r)) + log(4 - 5*r + 2*r^2)) + (n02 + n44)*(log(pmax(1e-6,r)) + log(4 - 5*r + 4*r^2)) + (n21 + n25)*log(3 - 5*r + 5*r^2) + (n04 + n42)*log(3 - 6*r + 7*r^2 - 4*r^3) + (n15 + n31)*log(1 + r^2 - 2*r^3) + (n12 + n34)*log(3 - r + 3*r^2 - 2*r^3) + (n14 + n32)*log(3 + r - 3*r^2 + 2*r^3)
return(L)}
interlogL_cm1c <- function(n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n22,n23,n24,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n13,n33) {
optimize(logL_cm1c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n22,n23,n24,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n13,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_13"],x[,"n_33"])
LOD_cm1c <- 3*(x[,"n_05"] + x[,"n_41"])*log10(2) + 3*(x[,"n_01"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_32"])*log10(3) - (x[,"n_12"] + x[,"n_34"])*log10(3) - (x[,"n_04"] + x[,"n_42"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - x[,"n_23"]*(log10(2) + log10(5)) - (x[,"n_21"] + x[,"n_25"])*(-2*log10(2) + log10(7)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(43)) + (x[,"n_05"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + log10(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + x[,"n_23"]*log10(9 + 4*r_cm1c - 4*r_cm1c^2) + (x[,"n_22"] + x[,"n_24"])*log10(21 + 2*r_cm1c - 2*r_cm1c^2) + (x[,"n_13"] + x[,"n_33"])*log10(3 - r_cm1c + r_cm1c^2) + (x[,"n_11"] + x[,"n_35"])*(log10(pmax(1e-6,r_cm1c)) + log10(4 - 5*r_cm1c + 2*r_cm1c^2)) + (x[,"n_02"] + x[,"n_44"])*(log10(pmax(1e-6,r_cm1c)) + log10(4 - 5*r_cm1c + 4*r_cm1c^2)) + (x[,"n_21"] + x[,"n_25"])*log10(3 - 5*r_cm1c + 5*r_cm1c^2) + (x[,"n_04"] + x[,"n_42"])*log10(3 - 6*r_cm1c + 7*r_cm1c^2 - 4*r_cm1c^3) + (x[,"n_15"] + x[,"n_31"])*log10(1 + r_cm1c^2 - 2*r_cm1c^3) + (x[,"n_12"] + x[,"n_34"])*log10(3 - r_cm1c + 3*r_cm1c^2 - 2*r_cm1c^3) + (x[,"n_14"] + x[,"n_32"])*log10(3 + r_cm1c - 3*r_cm1c^2 + 2*r_cm1c^3)
logL_cm1c <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - x[,"n_12"] - x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 3*x[,"n_22"] - x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - x[,"n_32"] - x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 2*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_cm1c)) + log(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + x[,"n_23"]*log(9 + 4*r_cm1c - 4*r_cm1c^2) + (x[,"n_22"] + x[,"n_24"])*log(21 + 2*r_cm1c - 2*r_cm1c^2) + (x[,"n_13"] + x[,"n_33"])*log(3 - r_cm1c + r_cm1c^2) + (x[,"n_11"] + x[,"n_35"])*(log(pmax(1e-6,r_cm1c)) + log(4 - 5*r_cm1c + 2*r_cm1c^2)) + (x[,"n_02"] + x[,"n_44"])*(log(pmax(1e-6,r_cm1c)) + log(4 - 5*r_cm1c + 4*r_cm1c^2)) + (x[,"n_21"] + x[,"n_25"])*log(3 - 5*r_cm1c + 5*r_cm1c^2) + (x[,"n_04"] + x[,"n_42"])*log(3 - 6*r_cm1c + 7*r_cm1c^2 - 4*r_cm1c^3) + (x[,"n_15"] + x[,"n_31"])*log(1 + r_cm1c^2 - 2*r_cm1c^3) + (x[,"n_12"] + x[,"n_34"])*log(3 - r_cm1c + 3*r_cm1c^2 - 2*r_cm1c^3) + (x[,"n_14"] + x[,"n_32"])*log(3 + r_cm1c - 3*r_cm1c^2 + 2*r_cm1c^3)
logL_rc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-4*n01 - 3*n02 - 4*n03 - 3*n04 - 4*n05 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 4*n21 - 3*n22 - 3*n23 - 3*n24 - 4*n25 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 4*n41 - 3*n42 - 4*n43 - 3*n44 - 4*n45)*log(2) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n41)*(log(3 - 2*r) + 2*log(pmax(1e-6,r))) + (n01 + n45)*(2*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n03 + n43)*log(3 + 4*r - 4*r^2) + n23*log(39 + 4*r - 4*r^2) + (n22 + n24)*log(21 + 2*r - 2*r^2) + (n15 + n31)*(log(pmax(1e-6,r)) + log(6 - 5*r + 2*r^2)) + (n04 + n42)*(log(pmax(1e-6,r)) + log(3 - 2*r + 2*r^2)) + (n13 + n33)*log(23 - 4*r + 4*r^2) + (n21 + n25)*log(9 - 8*r + 8*r^2) + (n14 + n32)*log(9 + 4*r + 6*r^2 - 4*r^3) + (n11 + n35)*log(3 - 2*r + r^2 - 2*r^3) + (n02 + n44)*log(3 - 5*r + 4*r^2 - 2*r^3) + (n12 + n34)*log(15 - 4*r - 6*r^2 + 4*r^3)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rc <- -((x[,"n_15"] + x[,"n_31"])*log10(2)) - (x[,"n_11"] + x[,"n_35"])*log10(2) + (x[,"n_05"] + x[,"n_41"])*log10(2) - 2*(x[,"n_03"] + x[,"n_43"])*log10(2) + (x[,"n_01"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_32"])*(2*log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_34"])*(2*log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_42"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - x[,"n_23"]*(3*log10(2) + log10(5)) - (x[,"n_21"] + x[,"n_25"])*log10(7) - (x[,"n_13"] + x[,"n_33"])*(log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(43)) + (x[,"n_05"] + x[,"n_41"])*(log10(3 - 2*r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_03"] + x[,"n_43"])*log10(3 + 4*r_rc - 4*r_rc^2) + x[,"n_23"]*log10(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_22"] + x[,"n_24"])*log10(21 + 2*r_rc - 2*r_rc^2) + (x[,"n_15"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(6 - 5*r_rc + 2*r_rc^2)) + (x[,"n_04"] + x[,"n_42"])*(log10(pmax(1e-6,r_rc)) + log10(3 - 2*r_rc + 2*r_rc^2)) + (x[,"n_13"] + x[,"n_33"])*log10(23 - 4*r_rc + 4*r_rc^2) + (x[,"n_21"] + x[,"n_25"])*log10(9 - 8*r_rc + 8*r_rc^2) + (x[,"n_14"] + x[,"n_32"])*log10(9 + 4*r_rc + 6*r_rc^2 - 4*r_rc^3) + (x[,"n_11"] + x[,"n_35"])*log10(3 - 2*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_02"] + x[,"n_44"])*log10(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3) + (x[,"n_12"] + x[,"n_34"])*log10(15 - 4*r_rc - 6*r_rc^2 + 4*r_rc^3)
logL_rc <- (-4*x[,"n_01"] - 3*x[,"n_02"] - 4*x[,"n_03"] - 3*x[,"n_04"] - 4*x[,"n_05"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 4*x[,"n_25"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 4*x[,"n_43"] - 3*x[,"n_44"] - 4*x[,"n_45"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_41"])*(log(3 - 2*r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_03"] + x[,"n_43"])*log(3 + 4*r_rc - 4*r_rc^2) + x[,"n_23"]*log(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_22"] + x[,"n_24"])*log(21 + 2*r_rc - 2*r_rc^2) + (x[,"n_15"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(6 - 5*r_rc + 2*r_rc^2)) + (x[,"n_04"] + x[,"n_42"])*(log(pmax(1e-6,r_rc)) + log(3 - 2*r_rc + 2*r_rc^2)) + (x[,"n_13"] + x[,"n_33"])*log(23 - 4*r_rc + 4*r_rc^2) + (x[,"n_21"] + x[,"n_25"])*log(9 - 8*r_rc + 8*r_rc^2) + (x[,"n_14"] + x[,"n_32"])*log(9 + 4*r_rc + 6*r_rc^2 - 4*r_rc^3) + (x[,"n_11"] + x[,"n_35"])*log(3 - 2*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_02"] + x[,"n_44"])*log(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3) + (x[,"n_12"] + x[,"n_34"])*log(15 - 4*r_rc - 6*r_rc^2 + 4*r_rc^3)
logL_rm2c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n05 - 3*n11 - 2*n12 - 2*n13 - 2*n14 - 3*n15 - n21 - 2*n22 - 2*n23 - 2*n24 - n25 - 3*n31 - 2*n32 - 2*n33 - 2*n34 - 3*n35 - 3*n41 - 4*n42 - 3*n43 - 4*n44 - 3*n45)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n13 + n33)*log(11) + (n05 + n41)*(log(3 - 2*r) + log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n45)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 + 2*r)) + n23*log(57 + 8*r - 8*r^2) + (n21 + n25)*log(3 - r + r^2) + (n22 + n24)*log(33 - 2*r + 2*r^2) + (n03 + n43)*log(9 - 8*r + 8*r^2) + (n04 + n42)*log(9 - 6*r + 8*r^2 - 8*r^3) + (n15 + n31)*log(9 - 8*r + 6*r^2 - 4*r^3) + (n12 + n34)*log(15 + 4*r + 6*r^2 - 4*r^3) + (n14 + n32)*log(21 - 4*r - 6*r^2 + 4*r^3) + (n11 + n35)*log(3 + 8*r - 6*r^2 + 4*r^3) + (n02 + n44)*log(3 + 14*r - 16*r^2 + 8*r^3)
return(L)}
interlogL_rm2c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_rm2c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm2c <- (x[,"n_05"] + x[,"n_41"])*log10(2) + (x[,"n_01"] + x[,"n_45"])*log10(2) - (x[,"n_15"] + x[,"n_31"])*(log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_35"])*(log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_32"])*(log10(2) + 2*log10(3)) - (x[,"n_12"] + x[,"n_34"])*(log10(2) + 2*log10(3)) - (x[,"n_04"] + x[,"n_42"])*log10(7) - (x[,"n_03"] + x[,"n_43"])*log10(7) - (x[,"n_02"] + x[,"n_44"])*log10(7) - (x[,"n_21"] + x[,"n_25"])*(-2*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(5) + log10(13)) - x[,"n_23"]*log10(59) + (x[,"n_05"] + x[,"n_41"])*(log10(3 - 2*r_rm2c) + log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c))) + (x[,"n_01"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c)) + log10(1 + 2*r_rm2c)) + x[,"n_23"]*log10(57 + 8*r_rm2c - 8*r_rm2c^2) + (x[,"n_21"] + x[,"n_25"])*log10(3 - r_rm2c + r_rm2c^2) + (x[,"n_22"] + x[,"n_24"])*log10(33 - 2*r_rm2c + 2*r_rm2c^2) + (x[,"n_03"] + x[,"n_43"])*log10(9 - 8*r_rm2c + 8*r_rm2c^2) + (x[,"n_04"] + x[,"n_42"])*log10(9 - 6*r_rm2c + 8*r_rm2c^2 - 8*r_rm2c^3) + (x[,"n_15"] + x[,"n_31"])*log10(9 - 8*r_rm2c + 6*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_12"] + x[,"n_34"])*log10(15 + 4*r_rm2c + 6*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_14"] + x[,"n_32"])*log10(21 - 4*r_rm2c - 6*r_rm2c^2 + 4*r_rm2c^3) + (x[,"n_11"] + x[,"n_35"])*log10(3 + 8*r_rm2c - 6*r_rm2c^2 + 4*r_rm2c^3) + (x[,"n_02"] + x[,"n_44"])*log10(3 + 14*r_rm2c - 16*r_rm2c^2 + 8*r_rm2c^3)
logL_rm2c <- (-3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_15"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - 3*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_13"] + x[,"n_33"])*log(11) + (x[,"n_05"] + x[,"n_41"])*(log(3 - 2*r_rm2c) + log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c))) + (x[,"n_01"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c)) + log(1 + 2*r_rm2c)) + x[,"n_23"]*log(57 + 8*r_rm2c - 8*r_rm2c^2) + (x[,"n_21"] + x[,"n_25"])*log(3 - r_rm2c + r_rm2c^2) + (x[,"n_22"] + x[,"n_24"])*log(33 - 2*r_rm2c + 2*r_rm2c^2) + (x[,"n_03"] + x[,"n_43"])*log(9 - 8*r_rm2c + 8*r_rm2c^2) + (x[,"n_04"] + x[,"n_42"])*log(9 - 6*r_rm2c + 8*r_rm2c^2 - 8*r_rm2c^3) + (x[,"n_15"] + x[,"n_31"])*log(9 - 8*r_rm2c + 6*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_12"] + x[,"n_34"])*log(15 + 4*r_rm2c + 6*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_14"] + x[,"n_32"])*log(21 - 4*r_rm2c - 6*r_rm2c^2 + 4*r_rm2c^3) + (x[,"n_11"] + x[,"n_35"])*log(3 + 8*r_rm2c - 6*r_rm2c^2 + 4*r_rm2c^3) + (x[,"n_02"] + x[,"n_44"])*log(3 + 14*r_rm2c - 16*r_rm2c^2 + 8*r_rm2c^3)
logL_rm1c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-4*n01 - 3*n02 - 4*n03 - 3*n04 - 4*n05 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 4*n21 - 3*n22 - 3*n23 - 3*n24 - 4*n25 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 4*n41 - 3*n42 - 4*n43 - 3*n44 - 4*n45)*log(2) + (-2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - n23 - 2*n24 - 2*n25 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(5) + (n05 + n41)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r))) + (n01 + n45)*(2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n21 + n25)*log(3 + 16*r - 16*r^2) + (n03 + n43)*log(1 + 12*r - 12*r^2) + (n13 + n33)*log(21 + 4*r - 4*r^2) + (n22 + n24)*log(21 + 2*r - 2*r^2) + (n11 + n35)*(log(pmax(1e-6,r)) + log(2 + 5*r - 2*r^2)) + n23*log(9 - 4*r + 4*r^2) + (n14 + n32)*log(21 - 20*r + 6*r^2 - 4*r^3) + (n02 + n44)*log(r + 4*r^2 - 2*r^3) + (n04 + n42)*log(3 - 3*r - 2*r^2 + 2*r^3) + (n15 + n31)*log(5 - 6*r - r^2 + 2*r^3) + (n12 + n34)*log(3 + 20*r - 6*r^2 + 4*r^3)
return(L)}
interlogL_rm1c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_rm1c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm1c <- -3*x[,"n_23"]*log10(2) - (x[,"n_15"] + x[,"n_31"])*log10(2) - (x[,"n_11"] + x[,"n_35"])*log10(2) + (x[,"n_05"] + x[,"n_41"])*log10(2) - 2*(x[,"n_03"] + x[,"n_43"])*log10(2) + (x[,"n_01"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_32"])*(2*log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_34"])*(2*log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_42"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - (x[,"n_21"] + x[,"n_25"])*log10(7) - (x[,"n_13"] + x[,"n_33"])*(log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(43)) + (x[,"n_05"] + x[,"n_41"])*(log10(3 - 2*r_rm1c) + 2*log10(pmax(1e-6,1 - r_rm1c))) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(1 + 2*r_rm1c)) + (x[,"n_21"] + x[,"n_25"])*log10(3 + 16*r_rm1c - 16*r_rm1c^2) + (x[,"n_03"] + x[,"n_43"])*log10(1 + 12*r_rm1c - 12*r_rm1c^2) + (x[,"n_13"] + x[,"n_33"])*log10(21 + 4*r_rm1c - 4*r_rm1c^2) + (x[,"n_22"] + x[,"n_24"])*log10(21 + 2*r_rm1c - 2*r_rm1c^2) + (x[,"n_11"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm1c)) + log10(2 + 5*r_rm1c - 2*r_rm1c^2)) + x[,"n_23"]*log10(9 - 4*r_rm1c + 4*r_rm1c^2) + (x[,"n_14"] + x[,"n_32"])*log10(21 - 20*r_rm1c + 6*r_rm1c^2 - 4*r_rm1c^3) + (x[,"n_02"] + x[,"n_44"])*log10(r_rm1c + 4*r_rm1c^2 - 2*r_rm1c^3) + (x[,"n_04"] + x[,"n_42"])*log10(3 - 3*r_rm1c - 2*r_rm1c^2 + 2*r_rm1c^3) + (x[,"n_15"] + x[,"n_31"])*log10(5 - 6*r_rm1c - r_rm1c^2 + 2*r_rm1c^3) + (x[,"n_12"] + x[,"n_34"])*log10(3 + 20*r_rm1c - 6*r_rm1c^2 + 4*r_rm1c^3)
logL_rm1c <- (-4*x[,"n_01"] - 3*x[,"n_02"] - 4*x[,"n_03"] - 3*x[,"n_04"] - 4*x[,"n_05"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 4*x[,"n_25"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 4*x[,"n_43"] - 3*x[,"n_44"] - 4*x[,"n_45"])*log(2) + (-2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_41"])*(log(3 - 2*r_rm1c) + 2*log(pmax(1e-6,1 - r_rm1c))) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,r_rm1c)) + log(1 + 2*r_rm1c)) + (x[,"n_21"] + x[,"n_25"])*log(3 + 16*r_rm1c - 16*r_rm1c^2) + (x[,"n_03"] + x[,"n_43"])*log(1 + 12*r_rm1c - 12*r_rm1c^2) + (x[,"n_13"] + x[,"n_33"])*log(21 + 4*r_rm1c - 4*r_rm1c^2) + (x[,"n_22"] + x[,"n_24"])*log(21 + 2*r_rm1c - 2*r_rm1c^2) + (x[,"n_11"] + x[,"n_35"])*(log(pmax(1e-6,r_rm1c)) + log(2 + 5*r_rm1c - 2*r_rm1c^2)) + x[,"n_23"]*log(9 - 4*r_rm1c + 4*r_rm1c^2) + (x[,"n_14"] + x[,"n_32"])*log(21 - 20*r_rm1c + 6*r_rm1c^2 - 4*r_rm1c^3) + (x[,"n_02"] + x[,"n_44"])*log(r_rm1c + 4*r_rm1c^2 - 2*r_rm1c^3) + (x[,"n_04"] + x[,"n_42"])*log(3 - 3*r_rm1c - 2*r_rm1c^2 + 2*r_rm1c^3) + (x[,"n_15"] + x[,"n_31"])*log(5 - 6*r_rm1c - r_rm1c^2 + 2*r_rm1c^3) + (x[,"n_12"] + x[,"n_34"])*log(3 + 20*r_rm1c - 6*r_rm1c^2 + 4*r_rm1c^3)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_rc,r_rm2c,r_rm1c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_rc,LOD_rm2c,LOD_rm1c,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_rc,logL_rm2c,logL_rm1c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","repulsion coupling","repulsion mixed2c","repulsion mixed1c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_3.1 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 4*n20 - 3*n21 - 3*n22 - 3*n23 - 4*n24 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44)*log(2) + (n02 + n12 + n20 + n21 + 2*n22 + n23 + n24 + n32 + n42)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + 2*(n00 + n44)*log(pmax(1e-6,1 - r)) + 2*(n04 + n40)*log(pmax(1e-6,r)) + (n03 + n14 + n30 + n41)*(log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n20 + n24 + n42)*log(1 + 2*r - 2*r^2) + (n12 + n21 + n23 + n32)*log(7 + 2*r - 2*r^2) + (n01 + n10 + n34 + n43)*log(3 - 4*r + r^2) + (n11 + n33)*log(21 - 18*r + 2*r^2) + n22*log(5 - 2*r + 2*r^2) + (n13 + n31)*log(5 + 14*r + 2*r^2)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cc <- 2*(x[,"n_04"] + x[,"n_40"])*log10(2) + 2*(x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-log10(2) + log10(3)) - x[,"n_22"]*(-log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_31"])*(-log10(2) + 2*log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-log10(2) + 2*log10(5)) + 2*(x[,"n_00"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(7 + 2*r_cc - 2*r_cc^2) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log10(3 - 4*r_cc + r_cc^2) + (x[,"n_11"] + x[,"n_33"])*log10(21 - 18*r_cc + 2*r_cc^2) + x[,"n_22"]*log10(5 - 2*r_cc + 2*r_cc^2) + (x[,"n_13"] + x[,"n_31"])*log10(5 + 14*r_cc + 2*r_cc^2)
logL_cc <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_20"] + x[,"n_21"] + 2*x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"] + x[,"n_42"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + 2*(x[,"n_00"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_04"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(7 + 2*r_cc - 2*r_cc^2) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log(3 - 4*r_cc + r_cc^2) + (x[,"n_11"] + x[,"n_33"])*log(21 - 18*r_cc + 2*r_cc^2) + x[,"n_22"]*log(5 - 2*r_cc + 2*r_cc^2) + (x[,"n_13"] + x[,"n_31"])*log(5 + 14*r_cc + 2*r_cc^2)
logL_cr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-4*n00 - 4*n01 - 3*n02 - 4*n03 - 4*n04 - 4*n10 - 2*n11 - 2*n12 - 2*n13 - 4*n14 - 3*n20 - 2*n21 - 2*n22 - 2*n23 - 3*n24 - 4*n30 - 2*n31 - 2*n32 - 2*n33 - 4*n34 - 4*n40 - 4*n41 - 3*n42 - 4*n43 - 4*n44)*log(2) + (n02 + n12 + n20 + n21 + 2*n22 + n23 + n24 + n32 + n42)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n00 + n04 + n40 + n44)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n22*(log(2 - r) + log(1 + r)) + (n03 + n10 + n34 + n41)*log(3 - 2*r^2) + (n01 + n14 + n30 + n43)*log(1 + 4*r - 2*r^2) + (n11 + n13 + n31 + n33)*log(6 + r - r^2) + (n02 + n20 + n24 + n42)*log(1 - r + r^2) + (n12 + n21 + n23 + n32)*log(4 - r + r^2)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cr <- 2*(x[,"n_00"] + x[,"n_04"] + x[,"n_40"] + x[,"n_44"])*log10(2) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(2*log10(2) - log10(3)) - 2*x[,"n_22"]*(-log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_10"] + x[,"n_34"] + x[,"n_41"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_14"] + x[,"n_30"] + x[,"n_43"])*(-log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_13"] + x[,"n_31"] + x[,"n_33"])*(-2*log10(2) + 2*log10(5)) + (x[,"n_00"] + x[,"n_04"] + x[,"n_40"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + x[,"n_22"]*(log10(2 - r_cr) + log10(1 + r_cr)) + (x[,"n_03"] + x[,"n_10"] + x[,"n_34"] + x[,"n_41"])*log10(3 - 2*r_cr^2) + (x[,"n_01"] + x[,"n_14"] + x[,"n_30"] + x[,"n_43"])*log10(1 + 4*r_cr - 2*r_cr^2) + (x[,"n_11"] + x[,"n_13"] + x[,"n_31"] + x[,"n_33"])*log10(6 + r_cr - r_cr^2) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(1 - r_cr + r_cr^2) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(4 - r_cr + r_cr^2)
logL_cr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 3*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 4*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_20"] + x[,"n_21"] + 2*x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"] + x[,"n_42"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_00"] + x[,"n_04"] + x[,"n_40"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + x[,"n_22"]*(log(2 - r_cr) + log(1 + r_cr)) + (x[,"n_03"] + x[,"n_10"] + x[,"n_34"] + x[,"n_41"])*log(3 - 2*r_cr^2) + (x[,"n_01"] + x[,"n_14"] + x[,"n_30"] + x[,"n_43"])*log(1 + 4*r_cr - 2*r_cr^2) + (x[,"n_11"] + x[,"n_13"] + x[,"n_31"] + x[,"n_33"])*log(6 + r_cr - r_cr^2) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(1 - r_cr + r_cr^2) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(4 - r_cr + r_cr^2)
logL_rc <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-4*n00 - 4*n01 - 3*n02 - 4*n03 - 4*n04 - 4*n10 - 2*n11 - 2*n12 - 2*n13 - 4*n14 - 3*n20 - 2*n21 - 2*n22 - 2*n23 - 3*n24 - 4*n30 - 2*n31 - 2*n32 - 2*n33 - 4*n34 - 4*n40 - 4*n41 - 3*n42 - 4*n43 - 4*n44)*log(2) + (n02 + n12 + n20 + n21 + 2*n22 + n23 + n24 + n32 + n42)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n00 + n04 + n40 + n44)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n22*(log(2 - r) + log(1 + r)) + (n01 + n14 + n30 + n43)*log(3 - 2*r^2) + (n03 + n10 + n34 + n41)*log(1 + 4*r - 2*r^2) + (n11 + n13 + n31 + n33)*log(6 + r - r^2) + (n02 + n20 + n24 + n42)*log(1 - r + r^2) + (n12 + n21 + n23 + n32)*log(4 - r + r^2)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rc <- 2*(x[,"n_00"] + x[,"n_04"] + x[,"n_40"] + x[,"n_44"])*log10(2) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(2*log10(2) - log10(3)) - 2*x[,"n_22"]*(-log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_10"] + x[,"n_34"] + x[,"n_41"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_14"] + x[,"n_30"] + x[,"n_43"])*(-log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_13"] + x[,"n_31"] + x[,"n_33"])*(-2*log10(2) + 2*log10(5)) + (x[,"n_00"] + x[,"n_04"] + x[,"n_40"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc))) + x[,"n_22"]*(log10(2 - r_rc) + log10(1 + r_rc)) + (x[,"n_01"] + x[,"n_14"] + x[,"n_30"] + x[,"n_43"])*log10(3 - 2*r_rc^2) + (x[,"n_03"] + x[,"n_10"] + x[,"n_34"] + x[,"n_41"])*log10(1 + 4*r_rc - 2*r_rc^2) + (x[,"n_11"] + x[,"n_13"] + x[,"n_31"] + x[,"n_33"])*log10(6 + r_rc - r_rc^2) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(1 - r_rc + r_rc^2) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(4 - r_rc + r_rc^2)
logL_rc <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 3*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 4*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_20"] + x[,"n_21"] + 2*x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"] + x[,"n_42"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_00"] + x[,"n_04"] + x[,"n_40"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc))) + x[,"n_22"]*(log(2 - r_rc) + log(1 + r_rc)) + (x[,"n_01"] + x[,"n_14"] + x[,"n_30"] + x[,"n_43"])*log(3 - 2*r_rc^2) + (x[,"n_03"] + x[,"n_10"] + x[,"n_34"] + x[,"n_41"])*log(1 + 4*r_rc - 2*r_rc^2) + (x[,"n_11"] + x[,"n_13"] + x[,"n_31"] + x[,"n_33"])*log(6 + r_rc - r_rc^2) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(1 - r_rc + r_rc^2) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(4 - r_rc + r_rc^2)
logL_rr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 4*n20 - 3*n21 - 3*n22 - 3*n23 - 4*n24 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44)*log(2) + (n02 + n12 + n20 + n21 + 2*n22 + n23 + n24 + n32 + n42)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + 2*(n04 + n40)*log(pmax(1e-6,1 - r)) + 2*(n00 + n44)*log(pmax(1e-6,r)) + (n01 + n10 + n34 + n43)*(log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n20 + n24 + n42)*log(1 + 2*r - 2*r^2) + (n12 + n21 + n23 + n32)*log(7 + 2*r - 2*r^2) + (n03 + n14 + n30 + n41)*log(3 - 4*r + r^2) + (n13 + n31)*log(21 - 18*r + 2*r^2) + n22*log(5 - 2*r + 2*r^2) + (n11 + n33)*log(5 + 14*r + 2*r^2)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rr <- 2*(x[,"n_04"] + x[,"n_40"])*log10(2) + 2*(x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-log10(2) + log10(3)) - x[,"n_22"]*(-log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_31"])*(-log10(2) + 2*log10(5)) - (x[,"n_11"] + x[,"n_33"])*(-log10(2) + 2*log10(5)) + 2*(x[,"n_04"] + x[,"n_40"])*log10(pmax(1e-6,1 - r_rr)) + 2*(x[,"n_00"] + x[,"n_44"])*log10(pmax(1e-6,r_rr)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(1 + 2*r_rr - 2*r_rr^2) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(7 + 2*r_rr - 2*r_rr^2) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*log10(3 - 4*r_rr + r_rr^2) + (x[,"n_13"] + x[,"n_31"])*log10(21 - 18*r_rr + 2*r_rr^2) + x[,"n_22"]*log10(5 - 2*r_rr + 2*r_rr^2) + (x[,"n_11"] + x[,"n_33"])*log10(5 + 14*r_rr + 2*r_rr^2)
logL_rr <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"])*log(2) + (x[,"n_02"] + x[,"n_12"] + x[,"n_20"] + x[,"n_21"] + 2*x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"] + x[,"n_42"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + 2*(x[,"n_04"] + x[,"n_40"])*log(pmax(1e-6,1 - r_rr)) + 2*(x[,"n_00"] + x[,"n_44"])*log(pmax(1e-6,r_rr)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(1 + 2*r_rr - 2*r_rr^2) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(7 + 2*r_rr - 2*r_rr^2) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*log(3 - 4*r_rr + r_rr^2) + (x[,"n_13"] + x[,"n_31"])*log(21 - 18*r_rr + 2*r_rr^2) + x[,"n_22"]*log(5 - 2*r_rr + 2*r_rr^2) + (x[,"n_11"] + x[,"n_33"])*log(5 + 14*r_rr + 2*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_3.2 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- 4*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(2) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n20 - n21 - 2*n22 - 2*n23 - n24 - 2*n25 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(5) + (n20 + n25)*log(7) + 3*(n00 + n45)*log(pmax(1e-6,1 - r)) + (n01 + n44)*(2*log(pmax(1e-6,1 - r)) + log(6 - r)) + 3*(n05 + n40)*log(pmax(1e-6,r)) + (n20 + n25)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n15 + n30)*(log(5 - 2*r) + 2*log(pmax(1e-6,r))) + (n04 + n41)*(2*log(pmax(1e-6,r)) + log(5 + r)) + (n10 + n35)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r)) + (n03 + n42)*(log(pmax(1e-6,r)) + log(7 + 4*r - 8*r^2)) + (n21 + n24)*log(3 + 5*r - 5*r^2) + (n14 + n31)*(log(pmax(1e-6,r)) + log(20 + 9*r - 2*r^2)) + (n22 + n23)*log(75 - 32*r + 32*r^2) + (n12 + n33)*log(51 - 32*r + 16*r^2 - 16*r^3) + (n11 + n34)*log(27 - 32*r + 3*r^2 + 2*r^3) + (n02 + n43)*log(3 + 9*r - 20*r^2 + 8*r^3) + (n13 + n32)*log(19 + 48*r - 32*r^2 + 16*r^3)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cc <- 2*(x[,"n_20"] + x[,"n_25"])*log10(2) + 3*(x[,"n_05"] + x[,"n_40"])*log10(2) + 3*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_31"])*(2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_34"])*(2*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_42"])*(-log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_43"])*(-log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_41"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_44"])*(-3*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(17)) - (x[,"n_13"] + x[,"n_32"])*log10(37) - (x[,"n_12"] + x[,"n_33"])*log10(37) - (x[,"n_22"] + x[,"n_23"])*log10(67) + 3*(x[,"n_00"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(6 - r_cc)) + 3*(x[,"n_05"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_30"])*(log10(5 - 2*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cc)) + log10(5 + r_cc)) + (x[,"n_10"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 2*r_cc)) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,r_cc)) + log10(7 + 4*r_cc - 8*r_cc^2)) + (x[,"n_21"] + x[,"n_24"])*log10(3 + 5*r_cc - 5*r_cc^2) + (x[,"n_14"] + x[,"n_31"])*(log10(pmax(1e-6,r_cc)) + log10(20 + 9*r_cc - 2*r_cc^2)) + (x[,"n_22"] + x[,"n_23"])*log10(75 - 32*r_cc + 32*r_cc^2) + (x[,"n_12"] + x[,"n_33"])*log10(51 - 32*r_cc + 16*r_cc^2 - 16*r_cc^3) + (x[,"n_11"] + x[,"n_34"])*log10(27 - 32*r_cc + 3*r_cc^2 + 2*r_cc^3) + (x[,"n_02"] + x[,"n_43"])*log10(3 + 9*r_cc - 20*r_cc^2 + 8*r_cc^3) + (x[,"n_13"] + x[,"n_32"])*log10(19 + 48*r_cc - 32*r_cc^2 + 16*r_cc^3)
logL_cc <- 4*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(2) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(5) + (x[,"n_20"] + x[,"n_25"])*log(7) + 3*(x[,"n_00"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(6 - r_cc)) + 3*(x[,"n_05"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_30"])*(log(5 - 2*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cc)) + log(5 + r_cc)) + (x[,"n_10"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 2*r_cc)) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,r_cc)) + log(7 + 4*r_cc - 8*r_cc^2)) + (x[,"n_21"] + x[,"n_24"])*log(3 + 5*r_cc - 5*r_cc^2) + (x[,"n_14"] + x[,"n_31"])*(log(pmax(1e-6,r_cc)) + log(20 + 9*r_cc - 2*r_cc^2)) + (x[,"n_22"] + x[,"n_23"])*log(75 - 32*r_cc + 32*r_cc^2) + (x[,"n_12"] + x[,"n_33"])*log(51 - 32*r_cc + 16*r_cc^2 - 16*r_cc^3) + (x[,"n_11"] + x[,"n_34"])*log(27 - 32*r_cc + 3*r_cc^2 + 2*r_cc^3) + (x[,"n_02"] + x[,"n_43"])*log(3 + 9*r_cc - 20*r_cc^2 + 8*r_cc^3) + (x[,"n_13"] + x[,"n_32"])*log(19 + 48*r_cc - 32*r_cc^2 + 16*r_cc^3)
logL_cm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
L <- (-3*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 3*n05 - 3*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 3*n15 - 3*n20 - 4*n21 - 4*n24 - 3*n25 - 3*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 3*n35 - 3*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 3*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n00 + n45)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n40)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n41)*(log(pmax(1e-6,r)) + log(9 - 4*r - 2*r^2)) + (n22 + n23)*log(13/80 + r/75 - r^2/75) + (n20 + n25)*log(3 - r + r^2) + (n15 + n30)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n21 + n24)*log(69 - 14*r + 14*r^2) + (n13 + n32)*log(99 + 8*r + 48*r^2 - 32*r^3) + (n02 + n43)*log(18 - 25*r + 32*r^2 - 16*r^3) + (n11 + n34)*log(51 - 32*r + 6*r^2 - 4*r^3) + (n10 + n35)*log(3 - 4*r + 3*r^2 - 2*r^3) + (n01 + n44)*log(3 + 5*r - 10*r^2 + 2*r^3) + (n14 + n31)*log(21 + 32*r - 6*r^2 + 4*r^3) + (n03 + n42)*log(9 + 9*r - 16*r^2 + 16*r^3) + (n12 + n33)*log(123 - 8*r - 48*r^2 + 32*r^3)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_22"],x[,"n_23"])
LOD_cm <- 3*(x[,"n_05"] + x[,"n_40"])*log10(2) + 3*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_15"] + x[,"n_30"])*(-log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_31"])*(2*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_34"])*(2*log10(2) + 2*log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-2*log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_41"])*(-2*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_44"])*(-2*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_42"])*(-log10(2) + log10(23)) - (x[,"n_02"] + x[,"n_43"])*(-log10(2) + log10(23)) - (x[,"n_13"] + x[,"n_32"])*(log10(3) + log10(37)) - (x[,"n_12"] + x[,"n_33"])*(log10(3) + log10(37)) - (x[,"n_21"] + x[,"n_24"])*(-log10(2) + log10(131)) + (x[,"n_22"] + x[,"n_23"])*(4*log10(2) + log10(3) + 2*log10(5) - log10(199)) + (x[,"n_00"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cm)) + 2*log10(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm)) + log10(9 - 4*r_cm - 2*r_cm^2)) + (x[,"n_22"] + x[,"n_23"])*log10(13/80 + r_cm/75 - r_cm^2/75) + (x[,"n_20"] + x[,"n_25"])*log10(3 - r_cm + r_cm^2) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,r_cm)) + log10(4 - 3*r_cm + 2*r_cm^2)) + (x[,"n_21"] + x[,"n_24"])*log10(69 - 14*r_cm + 14*r_cm^2) + (x[,"n_13"] + x[,"n_32"])*log10(99 + 8*r_cm + 48*r_cm^2 - 32*r_cm^3) + (x[,"n_02"] + x[,"n_43"])*log10(18 - 25*r_cm + 32*r_cm^2 - 16*r_cm^3) + (x[,"n_11"] + x[,"n_34"])*log10(51 - 32*r_cm + 6*r_cm^2 - 4*r_cm^3) + (x[,"n_10"] + x[,"n_35"])*log10(3 - 4*r_cm + 3*r_cm^2 - 2*r_cm^3) + (x[,"n_01"] + x[,"n_44"])*log10(3 + 5*r_cm - 10*r_cm^2 + 2*r_cm^3) + (x[,"n_14"] + x[,"n_31"])*log10(21 + 32*r_cm - 6*r_cm^2 + 4*r_cm^3) + (x[,"n_03"] + x[,"n_42"])*log10(9 + 9*r_cm - 16*r_cm^2 + 16*r_cm^3) + (x[,"n_12"] + x[,"n_33"])*log10(123 - 8*r_cm - 48*r_cm^2 + 32*r_cm^3)
logL_cm <- (-3*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cm)) + 2*log(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_cm)) + log(9 - 4*r_cm - 2*r_cm^2)) + (x[,"n_22"] + x[,"n_23"])*log(13/80 + r_cm/75 - r_cm^2/75) + (x[,"n_20"] + x[,"n_25"])*log(3 - r_cm + r_cm^2) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,r_cm)) + log(4 - 3*r_cm + 2*r_cm^2)) + (x[,"n_21"] + x[,"n_24"])*log(69 - 14*r_cm + 14*r_cm^2) + (x[,"n_13"] + x[,"n_32"])*log(99 + 8*r_cm + 48*r_cm^2 - 32*r_cm^3) + (x[,"n_02"] + x[,"n_43"])*log(18 - 25*r_cm + 32*r_cm^2 - 16*r_cm^3) + (x[,"n_11"] + x[,"n_34"])*log(51 - 32*r_cm + 6*r_cm^2 - 4*r_cm^3) + (x[,"n_10"] + x[,"n_35"])*log(3 - 4*r_cm + 3*r_cm^2 - 2*r_cm^3) + (x[,"n_01"] + x[,"n_44"])*log(3 + 5*r_cm - 10*r_cm^2 + 2*r_cm^3) + (x[,"n_14"] + x[,"n_31"])*log(21 + 32*r_cm - 6*r_cm^2 + 4*r_cm^3) + (x[,"n_03"] + x[,"n_42"])*log(9 + 9*r_cm - 16*r_cm^2 + 16*r_cm^3) + (x[,"n_12"] + x[,"n_33"])*log(123 - 8*r_cm - 48*r_cm^2 + 32*r_cm^3)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n05 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n45)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n22 + n23)*log(63 + 16*r - 16*r^2) + (n01 + n44)*(log(pmax(1e-6,r)) + log(2 + 2*r - r^2)) + (n10 + n35)*(log(pmax(1e-6,r)) + log(4 - 5*r + 2*r^2)) + (n20 + n25)*log(3 - 5*r + 5*r^2) + (n21 + n24)*log(24 - 11*r + 11*r^2) + (n12 + n33)*log(39 - 16*r + 32*r^2 - 16*r^3) + (n03 + n42)*log(6 - 9*r + 12*r^2 - 8*r^3) + (n14 + n31)*log(15 - 4*r - 3*r^2 - 2*r^3) + (n15 + n30)*log(1 + r^2 - 2*r^3) + (n04 + n41)*log(3 - 3*r - r^2 + r^3) + (n11 + n34)*log(6 + 16*r - 9*r^2 + 2*r^3) + (n02 + n43)*log(1 + 9*r - 12*r^2 + 8*r^3) + (n13 + n32)*log(39 - 16*r^2 + 16*r^3)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cr <- 3*(x[,"n_05"] + x[,"n_40"])*log10(2) + 3*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_31"])*(2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_34"])*(2*log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-2*log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_42"])*(-log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_43"])*(-log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_41"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_44"])*(-3*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(17)) - (x[,"n_13"] + x[,"n_32"])*log10(37) - (x[,"n_12"] + x[,"n_33"])*log10(37) - (x[,"n_22"] + x[,"n_23"])*log10(67) + (x[,"n_05"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_22"] + x[,"n_23"])*log10(63 + 16*r_cr - 16*r_cr^2) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_cr)) + log10(2 + 2*r_cr - r_cr^2)) + (x[,"n_10"] + x[,"n_35"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 5*r_cr + 2*r_cr^2)) + (x[,"n_20"] + x[,"n_25"])*log10(3 - 5*r_cr + 5*r_cr^2) + (x[,"n_21"] + x[,"n_24"])*log10(24 - 11*r_cr + 11*r_cr^2) + (x[,"n_12"] + x[,"n_33"])*log10(39 - 16*r_cr + 32*r_cr^2 - 16*r_cr^3) + (x[,"n_03"] + x[,"n_42"])*log10(6 - 9*r_cr + 12*r_cr^2 - 8*r_cr^3) + (x[,"n_14"] + x[,"n_31"])*log10(15 - 4*r_cr - 3*r_cr^2 - 2*r_cr^3) + (x[,"n_15"] + x[,"n_30"])*log10(1 + r_cr^2 - 2*r_cr^3) + (x[,"n_04"] + x[,"n_41"])*log10(3 - 3*r_cr - r_cr^2 + r_cr^3) + (x[,"n_11"] + x[,"n_34"])*log10(6 + 16*r_cr - 9*r_cr^2 + 2*r_cr^3) + (x[,"n_02"] + x[,"n_43"])*log10(1 + 9*r_cr - 12*r_cr^2 + 8*r_cr^3) + (x[,"n_13"] + x[,"n_32"])*log10(39 - 16*r_cr^2 + 16*r_cr^3)
logL_cr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_22"] + x[,"n_23"])*log(63 + 16*r_cr - 16*r_cr^2) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_cr)) + log(2 + 2*r_cr - r_cr^2)) + (x[,"n_10"] + x[,"n_35"])*(log(pmax(1e-6,r_cr)) + log(4 - 5*r_cr + 2*r_cr^2)) + (x[,"n_20"] + x[,"n_25"])*log(3 - 5*r_cr + 5*r_cr^2) + (x[,"n_21"] + x[,"n_24"])*log(24 - 11*r_cr + 11*r_cr^2) + (x[,"n_12"] + x[,"n_33"])*log(39 - 16*r_cr + 32*r_cr^2 - 16*r_cr^3) + (x[,"n_03"] + x[,"n_42"])*log(6 - 9*r_cr + 12*r_cr^2 - 8*r_cr^3) + (x[,"n_14"] + x[,"n_31"])*log(15 - 4*r_cr - 3*r_cr^2 - 2*r_cr^3) + (x[,"n_15"] + x[,"n_30"])*log(1 + r_cr^2 - 2*r_cr^3) + (x[,"n_04"] + x[,"n_41"])*log(3 - 3*r_cr - r_cr^2 + r_cr^3) + (x[,"n_11"] + x[,"n_34"])*log(6 + 16*r_cr - 9*r_cr^2 + 2*r_cr^3) + (x[,"n_02"] + x[,"n_43"])*log(1 + 9*r_cr - 12*r_cr^2 + 8*r_cr^3) + (x[,"n_13"] + x[,"n_32"])*log(39 - 16*r_cr^2 + 16*r_cr^3)
logL_rc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n00 + n45)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n40)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n22 + n23)*log(63 + 16*r - 16*r^2) + (n04 + n41)*(log(pmax(1e-6,r)) + log(2 + 2*r - r^2)) + (n15 + n30)*(log(pmax(1e-6,r)) + log(4 - 5*r + 2*r^2)) + (n20 + n25)*log(3 - 5*r + 5*r^2) + (n21 + n24)*log(24 - 11*r + 11*r^2) + (n13 + n32)*log(39 - 16*r + 32*r^2 - 16*r^3) + (n02 + n43)*log(6 - 9*r + 12*r^2 - 8*r^3) + (n11 + n34)*log(15 - 4*r - 3*r^2 - 2*r^3) + (n10 + n35)*log(1 + r^2 - 2*r^3) + (n01 + n44)*log(3 - 3*r - r^2 + r^3) + (n14 + n31)*log(6 + 16*r - 9*r^2 + 2*r^3) + (n03 + n42)*log(1 + 9*r - 12*r^2 + 8*r^3) + (n12 + n33)*log(39 - 16*r^2 + 16*r^3)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rc <- 3*(x[,"n_05"] + x[,"n_40"])*log10(2) + 3*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_31"])*(2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_34"])*(2*log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-2*log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_42"])*(-log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_43"])*(-log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_41"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_44"])*(-3*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(17)) - (x[,"n_13"] + x[,"n_32"])*log10(37) - (x[,"n_12"] + x[,"n_33"])*log10(37) - (x[,"n_22"] + x[,"n_23"])*log10(67) + (x[,"n_00"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc))) + (x[,"n_05"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_rc)) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_22"] + x[,"n_23"])*log10(63 + 16*r_rc - 16*r_rc^2) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_rc)) + log10(2 + 2*r_rc - r_rc^2)) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,r_rc)) + log10(4 - 5*r_rc + 2*r_rc^2)) + (x[,"n_20"] + x[,"n_25"])*log10(3 - 5*r_rc + 5*r_rc^2) + (x[,"n_21"] + x[,"n_24"])*log10(24 - 11*r_rc + 11*r_rc^2) + (x[,"n_13"] + x[,"n_32"])*log10(39 - 16*r_rc + 32*r_rc^2 - 16*r_rc^3) + (x[,"n_02"] + x[,"n_43"])*log10(6 - 9*r_rc + 12*r_rc^2 - 8*r_rc^3) + (x[,"n_11"] + x[,"n_34"])*log10(15 - 4*r_rc - 3*r_rc^2 - 2*r_rc^3) + (x[,"n_10"] + x[,"n_35"])*log10(1 + r_rc^2 - 2*r_rc^3) + (x[,"n_01"] + x[,"n_44"])*log10(3 - 3*r_rc - r_rc^2 + r_rc^3) + (x[,"n_14"] + x[,"n_31"])*log10(6 + 16*r_rc - 9*r_rc^2 + 2*r_rc^3) + (x[,"n_03"] + x[,"n_42"])*log10(1 + 9*r_rc - 12*r_rc^2 + 8*r_rc^3) + (x[,"n_12"] + x[,"n_33"])*log10(39 - 16*r_rc^2 + 16*r_rc^3)
logL_rc <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_00"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc))) + (x[,"n_05"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_rc)) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_22"] + x[,"n_23"])*log(63 + 16*r_rc - 16*r_rc^2) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_rc)) + log(2 + 2*r_rc - r_rc^2)) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,r_rc)) + log(4 - 5*r_rc + 2*r_rc^2)) + (x[,"n_20"] + x[,"n_25"])*log(3 - 5*r_rc + 5*r_rc^2) + (x[,"n_21"] + x[,"n_24"])*log(24 - 11*r_rc + 11*r_rc^2) + (x[,"n_13"] + x[,"n_32"])*log(39 - 16*r_rc + 32*r_rc^2 - 16*r_rc^3) + (x[,"n_02"] + x[,"n_43"])*log(6 - 9*r_rc + 12*r_rc^2 - 8*r_rc^3) + (x[,"n_11"] + x[,"n_34"])*log(15 - 4*r_rc - 3*r_rc^2 - 2*r_rc^3) + (x[,"n_10"] + x[,"n_35"])*log(1 + r_rc^2 - 2*r_rc^3) + (x[,"n_01"] + x[,"n_44"])*log(3 - 3*r_rc - r_rc^2 + r_rc^3) + (x[,"n_14"] + x[,"n_31"])*log(6 + 16*r_rc - 9*r_rc^2 + 2*r_rc^3) + (x[,"n_03"] + x[,"n_42"])*log(1 + 9*r_rc - 12*r_rc^2 + 8*r_rc^3) + (x[,"n_12"] + x[,"n_33"])*log(39 - 16*r_rc^2 + 16*r_rc^3)
logL_rm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
L <- (-3*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 3*n05 - 3*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 3*n15 - 3*n20 - 4*n21 - 4*n24 - 3*n25 - 3*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 3*n35 - 3*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 3*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n05 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n45)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n44)*(log(pmax(1e-6,r)) + log(9 - 4*r - 2*r^2)) + (n22 + n23)*log(13/80 + r/75 - r^2/75) + (n20 + n25)*log(3 - r + r^2) + (n10 + n35)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n21 + n24)*log(69 - 14*r + 14*r^2) + (n12 + n33)*log(99 + 8*r + 48*r^2 - 32*r^3) + (n03 + n42)*log(18 - 25*r + 32*r^2 - 16*r^3) + (n14 + n31)*log(51 - 32*r + 6*r^2 - 4*r^3) + (n15 + n30)*log(3 - 4*r + 3*r^2 - 2*r^3) + (n04 + n41)*log(3 + 5*r - 10*r^2 + 2*r^3) + (n11 + n34)*log(21 + 32*r - 6*r^2 + 4*r^3) + (n02 + n43)*log(9 + 9*r - 16*r^2 + 16*r^3) + (n13 + n32)*log(123 - 8*r - 48*r^2 + 32*r^3)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_22"],x[,"n_23"])
LOD_rm <- 3*(x[,"n_05"] + x[,"n_40"])*log10(2) + 3*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_15"] + x[,"n_30"])*(-log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_31"])*(2*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_34"])*(2*log10(2) + 2*log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-2*log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_41"])*(-2*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_44"])*(-2*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_42"])*(-log10(2) + log10(23)) - (x[,"n_02"] + x[,"n_43"])*(-log10(2) + log10(23)) - (x[,"n_13"] + x[,"n_32"])*(log10(3) + log10(37)) - (x[,"n_12"] + x[,"n_33"])*(log10(3) + log10(37)) - (x[,"n_21"] + x[,"n_24"])*(-log10(2) + log10(131)) + (x[,"n_22"] + x[,"n_23"])*(4*log10(2) + log10(3) + 2*log10(5) - log10(199)) + (x[,"n_05"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_rm)) + 2*log10(pmax(1e-6,r_rm))) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm)) + log10(9 - 4*r_rm - 2*r_rm^2)) + (x[,"n_22"] + x[,"n_23"])*log10(13/80 + r_rm/75 - r_rm^2/75) + (x[,"n_20"] + x[,"n_25"])*log10(3 - r_rm + r_rm^2) + (x[,"n_10"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm)) + log10(4 - 3*r_rm + 2*r_rm^2)) + (x[,"n_21"] + x[,"n_24"])*log10(69 - 14*r_rm + 14*r_rm^2) + (x[,"n_12"] + x[,"n_33"])*log10(99 + 8*r_rm + 48*r_rm^2 - 32*r_rm^3) + (x[,"n_03"] + x[,"n_42"])*log10(18 - 25*r_rm + 32*r_rm^2 - 16*r_rm^3) + (x[,"n_14"] + x[,"n_31"])*log10(51 - 32*r_rm + 6*r_rm^2 - 4*r_rm^3) + (x[,"n_15"] + x[,"n_30"])*log10(3 - 4*r_rm + 3*r_rm^2 - 2*r_rm^3) + (x[,"n_04"] + x[,"n_41"])*log10(3 + 5*r_rm - 10*r_rm^2 + 2*r_rm^3) + (x[,"n_11"] + x[,"n_34"])*log10(21 + 32*r_rm - 6*r_rm^2 + 4*r_rm^3) + (x[,"n_02"] + x[,"n_43"])*log10(9 + 9*r_rm - 16*r_rm^2 + 16*r_rm^3) + (x[,"n_13"] + x[,"n_32"])*log10(123 - 8*r_rm - 48*r_rm^2 + 32*r_rm^3)
logL_rm <- (-3*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_rm)) + 2*log(pmax(1e-6,r_rm))) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_rm)) + log(9 - 4*r_rm - 2*r_rm^2)) + (x[,"n_22"] + x[,"n_23"])*log(13/80 + r_rm/75 - r_rm^2/75) + (x[,"n_20"] + x[,"n_25"])*log(3 - r_rm + r_rm^2) + (x[,"n_10"] + x[,"n_35"])*(log(pmax(1e-6,r_rm)) + log(4 - 3*r_rm + 2*r_rm^2)) + (x[,"n_21"] + x[,"n_24"])*log(69 - 14*r_rm + 14*r_rm^2) + (x[,"n_12"] + x[,"n_33"])*log(99 + 8*r_rm + 48*r_rm^2 - 32*r_rm^3) + (x[,"n_03"] + x[,"n_42"])*log(18 - 25*r_rm + 32*r_rm^2 - 16*r_rm^3) + (x[,"n_14"] + x[,"n_31"])*log(51 - 32*r_rm + 6*r_rm^2 - 4*r_rm^3) + (x[,"n_15"] + x[,"n_30"])*log(3 - 4*r_rm + 3*r_rm^2 - 2*r_rm^3) + (x[,"n_04"] + x[,"n_41"])*log(3 + 5*r_rm - 10*r_rm^2 + 2*r_rm^3) + (x[,"n_11"] + x[,"n_34"])*log(21 + 32*r_rm - 6*r_rm^2 + 4*r_rm^3) + (x[,"n_02"] + x[,"n_43"])*log(9 + 9*r_rm - 16*r_rm^2 + 16*r_rm^3) + (x[,"n_13"] + x[,"n_32"])*log(123 - 8*r_rm - 48*r_rm^2 + 32*r_rm^3)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- 4*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(2) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n20 - n21 - 2*n22 - 2*n23 - n24 - 2*n25 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(5) + (n20 + n25)*log(7) + 3*(n05 + n40)*log(pmax(1e-6,1 - r)) + (n04 + n41)*(2*log(pmax(1e-6,1 - r)) + log(6 - r)) + 3*(n00 + n45)*log(pmax(1e-6,r)) + (n20 + n25)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n35)*(log(5 - 2*r) + 2*log(pmax(1e-6,r))) + (n01 + n44)*(2*log(pmax(1e-6,r)) + log(5 + r)) + (n15 + n30)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r)) + (n02 + n43)*(log(pmax(1e-6,r)) + log(7 + 4*r - 8*r^2)) + (n21 + n24)*log(3 + 5*r - 5*r^2) + (n11 + n34)*(log(pmax(1e-6,r)) + log(20 + 9*r - 2*r^2)) + (n22 + n23)*log(75 - 32*r + 32*r^2) + (n13 + n32)*log(51 - 32*r + 16*r^2 - 16*r^3) + (n14 + n31)*log(27 - 32*r + 3*r^2 + 2*r^3) + (n03 + n42)*log(3 + 9*r - 20*r^2 + 8*r^3) + (n12 + n33)*log(19 + 48*r - 32*r^2 + 16*r^3)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rr <- 2*(x[,"n_20"] + x[,"n_25"])*log10(2) + 3*(x[,"n_05"] + x[,"n_40"])*log10(2) + 3*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_31"])*(2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_34"])*(2*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_42"])*(-log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_43"])*(-log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_41"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_44"])*(-3*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(17)) - (x[,"n_13"] + x[,"n_32"])*log10(37) - (x[,"n_12"] + x[,"n_33"])*log10(37) - (x[,"n_22"] + x[,"n_23"])*log10(67) + 3*(x[,"n_05"] + x[,"n_40"])*log10(pmax(1e-6,1 - r_rr)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(6 - r_rr)) + 3*(x[,"n_00"] + x[,"n_45"])*log10(pmax(1e-6,r_rr)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_35"])*(log10(5 - 2*r_rr) + 2*log10(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(5 + r_rr)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 + 2*r_rr)) + (x[,"n_02"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(7 + 4*r_rr - 8*r_rr^2)) + (x[,"n_21"] + x[,"n_24"])*log10(3 + 5*r_rr - 5*r_rr^2) + (x[,"n_11"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(20 + 9*r_rr - 2*r_rr^2)) + (x[,"n_22"] + x[,"n_23"])*log10(75 - 32*r_rr + 32*r_rr^2) + (x[,"n_13"] + x[,"n_32"])*log10(51 - 32*r_rr + 16*r_rr^2 - 16*r_rr^3) + (x[,"n_14"] + x[,"n_31"])*log10(27 - 32*r_rr + 3*r_rr^2 + 2*r_rr^3) + (x[,"n_03"] + x[,"n_42"])*log10(3 + 9*r_rr - 20*r_rr^2 + 8*r_rr^3) + (x[,"n_12"] + x[,"n_33"])*log10(19 + 48*r_rr - 32*r_rr^2 + 16*r_rr^3)
logL_rr <- 4*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(2) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(5) + (x[,"n_20"] + x[,"n_25"])*log(7) + 3*(x[,"n_05"] + x[,"n_40"])*log(pmax(1e-6,1 - r_rr)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(6 - r_rr)) + 3*(x[,"n_00"] + x[,"n_45"])*log(pmax(1e-6,r_rr)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_35"])*(log(5 - 2*r_rr) + 2*log(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(5 + r_rr)) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 + 2*r_rr)) + (x[,"n_02"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(7 + 4*r_rr - 8*r_rr^2)) + (x[,"n_21"] + x[,"n_24"])*log(3 + 5*r_rr - 5*r_rr^2) + (x[,"n_11"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(20 + 9*r_rr - 2*r_rr^2)) + (x[,"n_22"] + x[,"n_23"])*log(75 - 32*r_rr + 32*r_rr^2) + (x[,"n_13"] + x[,"n_32"])*log(51 - 32*r_rr + 16*r_rr^2 - 16*r_rr^3) + (x[,"n_14"] + x[,"n_31"])*log(27 - 32*r_rr + 3*r_rr^2 + 2*r_rr^3) + (x[,"n_03"] + x[,"n_42"])*log(3 + 9*r_rr - 20*r_rr^2 + 8*r_rr^3) + (x[,"n_12"] + x[,"n_33"])*log(19 + 48*r_rr - 32*r_rr^2 + 16*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_3.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n05 - 4*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 3*n20 - 3*n21 - 3*n22 - 2*n23 - 3*n24 - 3*n25 - 3*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44 - 3*n45 - 4*n46)*log(2) + (n01 + n02 + n04 + n05 + n11 + n12 + n14 + n15 + n20 + n21 + n22 + n23 + n24 + n25 + n26 + n31 + n32 + n34 + n35 + n41 + n42 + n44 + n45)*log(3) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 2*n20 - n21 - 2*n22 - 2*n23 - 2*n24 - n25 - 2*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n46)*log(5) + 3*(n01 + n45)*log(pmax(1e-6,1 - r)) + 4*(n00 + n46)*log(pmax(1e-6,1 - r)) + 4*(n06 + n40)*log(pmax(1e-6,r)) + 3*(n05 + n41)*log(pmax(1e-6,r)) + (n21 + n25)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n36)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n20 + n26)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n15 + n31)*(log(2 - r) + 2*log(pmax(1e-6,r))) + (n16 + n30)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n11 + n35)*(2*log(pmax(1e-6,1 - r)) + log(1 + r)) + (n04 + n42)*(2*log(pmax(1e-6,r)) + log(4 - 3*r^2)) + (n02 + n44)*(2*log(pmax(1e-6,1 - r)) + log(1 + 6*r - 3*r^2)) + (n14 + n32)*(log(pmax(1e-6,r)) + log(4 + r - 3*r^2 + 3*r^3)) + (n03 + n43)*(log(pmax(1e-6,r)) + log(5 + 3*r - 16*r^2 + 8*r^3)) + (n13 + n33)*log(7 + 17*r - 33*r^2 + 32*r^3 - 16*r^4) + (n22 + n24)*log(6 + 8*r - 17*r^2 + 18*r^3 - 9*r^4) + (n12 + n34)*log(5 - 9*r + 10*r^2 - 9*r^3 + 3*r^4) + n23*log(9 - 13*r + 21*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cc <- 2*(x[,"n_21"] + x[,"n_25"])*log10(2) + 4*(x[,"n_20"] + x[,"n_26"])*log10(2) + 4*(x[,"n_16"] + x[,"n_30"])*log10(2) + 4*(x[,"n_10"] + x[,"n_36"])*log10(2) + 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 3*(x[,"n_05"] + x[,"n_41"])*log10(2) + 3*(x[,"n_01"] + x[,"n_45"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) - (x[,"n_15"] + x[,"n_31"])*(-3*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-3*log10(2) + log10(3)) - x[,"n_23"]*(-2*log10(2) + 2*log10(5)) - (x[,"n_03"] + x[,"n_43"])*(-2*log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_32"])*(-4*log10(2) + log10(3) + log10(11)) - (x[,"n_12"] + x[,"n_34"])*(-4*log10(2) + log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_24"])*(-4*log10(2) + log10(7) + log10(17)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(41)) + 3*(x[,"n_01"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_00"] + x[,"n_46"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_06"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + 3*(x[,"n_05"] + x[,"n_41"])*log10(pmax(1e-6,r_cc)) + (x[,"n_21"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_36"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_26"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_31"])*(log10(2 - r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_16"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(1 + r_cc)) + (x[,"n_04"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(4 - 3*r_cc^2)) + (x[,"n_02"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(1 + 6*r_cc - 3*r_cc^2)) + (x[,"n_14"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(4 + r_cc - 3*r_cc^2 + 3*r_cc^3)) + (x[,"n_03"] + x[,"n_43"])*(log10(pmax(1e-6,r_cc)) + log10(5 + 3*r_cc - 16*r_cc^2 + 8*r_cc^3)) + (x[,"n_13"] + x[,"n_33"])*log10(7 + 17*r_cc - 33*r_cc^2 + 32*r_cc^3 - 16*r_cc^4) + (x[,"n_22"] + x[,"n_24"])*log10(6 + 8*r_cc - 17*r_cc^2 + 18*r_cc^3 - 9*r_cc^4) + (x[,"n_12"] + x[,"n_34"])*log10(5 - 9*r_cc + 10*r_cc^2 - 9*r_cc^3 + 3*r_cc^4) + x[,"n_23"]*log10(9 - 13*r_cc + 21*r_cc^2 - 16*r_cc^3 + 8*r_cc^4)
logL_cc <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_11"] + x[,"n_12"] + x[,"n_14"] + x[,"n_15"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_26"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_41"] + x[,"n_42"] + x[,"n_44"] + x[,"n_45"])*log(3) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - 2*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_46"])*log(5) + 3*(x[,"n_01"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_00"] + x[,"n_46"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_06"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + 3*(x[,"n_05"] + x[,"n_41"])*log(pmax(1e-6,r_cc)) + (x[,"n_21"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_10"] + x[,"n_36"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_26"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_31"])*(log(2 - r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_16"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cc)) + log(1 + r_cc)) + (x[,"n_04"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(4 - 3*r_cc^2)) + (x[,"n_02"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(1 + 6*r_cc - 3*r_cc^2)) + (x[,"n_14"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(4 + r_cc - 3*r_cc^2 + 3*r_cc^3)) + (x[,"n_03"] + x[,"n_43"])*(log(pmax(1e-6,r_cc)) + log(5 + 3*r_cc - 16*r_cc^2 + 8*r_cc^3)) + (x[,"n_13"] + x[,"n_33"])*log(7 + 17*r_cc - 33*r_cc^2 + 32*r_cc^3 - 16*r_cc^4) + (x[,"n_22"] + x[,"n_24"])*log(6 + 8*r_cc - 17*r_cc^2 + 18*r_cc^3 - 9*r_cc^4) + (x[,"n_12"] + x[,"n_34"])*log(5 - 9*r_cc + 10*r_cc^2 - 9*r_cc^3 + 3*r_cc^4) + x[,"n_23"]*log(9 - 13*r_cc + 21*r_cc^2 - 16*r_cc^3 + 8*r_cc^4)
logL_cm2c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 3*n11 - 4*n12 - 3*n13 - 4*n14 - 3*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 3*n31 - 4*n32 - 3*n33 - 4*n34 - 3*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(log(3) + 2*log(5)) + (n21 + n25)*log(7) + (n00 + n46)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n41)*(log(7 - 4*r) + 2*log(pmax(1e-6,r))) + (n06 + n40)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n45)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r)) + (n21 + n25)*log(3 + 2*r - 2*r^2) + (n16 + n30)*(2*log(pmax(1e-6,r)) + log(5 - 6*r + 4*r^2)) + (n15 + n31)*(log(pmax(1e-6,r)) + log(14 - 3*r + 4*r^2)) + (n10 + n36)*(2*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 4*r^2)) + (n20 + n26)*(log(pmax(1e-6,r)) + log(7 - 13*r + 12*r^2 - 6*r^3)) + (n11 + n35)*log(15 - 20*r + 9*r^2 - 4*r^3) + (n04 + n42)*(log(pmax(1e-6,r)) + log(15 + r - 7*r^2 + 9*r^3)) + n23*log(111 + 10*r - 58*r^2 + 96*r^3 - 48*r^4) + (n14 + n32)*log(39 + 76*r - 27*r^2 + 50*r^3 - 36*r^4) + (n12 + n34)*log(102 - 28*r - 93*r^2 + 94*r^3 - 36*r^4) + (n03 + n43)*log(9 + 18*r - 34*r^2 + 32*r^3 - 16*r^4) + (n02 + n44)*log(18 - 32*r + 34*r^2 - 29*r^3 + 9*r^4) + (n13 + n33)*log(63 - 14*r + 46*r^2 - 64*r^3 + 32*r^4) + (n22 + n24)*log(138 - 31*r + 85*r^2 - 108*r^3 + 54*r^4)
return(L)}
interlogL_cm2c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cm2c <- 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) - (x[,"n_16"] + x[,"n_30"])*(-2*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_36"])*(-2*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_31"])*(-2*log10(2) + 3*log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-2*log10(2) + 3*log10(3)) - (x[,"n_05"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_45"])*(-2*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_43"])*(-log10(2) + 2*log10(5)) - (x[,"n_21"] + x[,"n_25"])*(-log10(2) + log10(7)) - (x[,"n_20"] + x[,"n_26"])*(-3*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + 3*log10(3) + log10(11)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + 3*log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + log10(7) + log10(17)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + log10(7) + log10(17)) - x[,"n_23"]*(-log10(2) + log10(13) + log10(17)) - (x[,"n_13"] + x[,"n_33"])*(-log10(2) + log10(3) + log10(41)) - (x[,"n_22"] + x[,"n_24"])*(-3*log10(2) + log10(1069)) + (x[,"n_00"] + x[,"n_46"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_41"])*(log10(7 - 4*r_cm2c) + 2*log10(pmax(1e-6,r_cm2c))) + (x[,"n_06"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cm2c)) + 3*log10(pmax(1e-6,r_cm2c))) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 + 4*r_cm2c)) + (x[,"n_21"] + x[,"n_25"])*log10(3 + 2*r_cm2c - 2*r_cm2c^2) + (x[,"n_16"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(5 - 6*r_cm2c + 4*r_cm2c^2)) + (x[,"n_15"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm2c)) + log10(14 - 3*r_cm2c + 4*r_cm2c^2)) + (x[,"n_10"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 - 2*r_cm2c + 4*r_cm2c^2)) + (x[,"n_20"] + x[,"n_26"])*(log10(pmax(1e-6,r_cm2c)) + log10(7 - 13*r_cm2c + 12*r_cm2c^2 - 6*r_cm2c^3)) + (x[,"n_11"] + x[,"n_35"])*log10(15 - 20*r_cm2c + 9*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_04"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm2c)) + log10(15 + r_cm2c - 7*r_cm2c^2 + 9*r_cm2c^3)) + x[,"n_23"]*log10(111 + 10*r_cm2c - 58*r_cm2c^2 + 96*r_cm2c^3 - 48*r_cm2c^4) + (x[,"n_14"] + x[,"n_32"])*log10(39 + 76*r_cm2c - 27*r_cm2c^2 + 50*r_cm2c^3 - 36*r_cm2c^4) + (x[,"n_12"] + x[,"n_34"])*log10(102 - 28*r_cm2c - 93*r_cm2c^2 + 94*r_cm2c^3 - 36*r_cm2c^4) + (x[,"n_03"] + x[,"n_43"])*log10(9 + 18*r_cm2c - 34*r_cm2c^2 + 32*r_cm2c^3 - 16*r_cm2c^4) + (x[,"n_02"] + x[,"n_44"])*log10(18 - 32*r_cm2c + 34*r_cm2c^2 - 29*r_cm2c^3 + 9*r_cm2c^4) + (x[,"n_13"] + x[,"n_33"])*log10(63 - 14*r_cm2c + 46*r_cm2c^2 - 64*r_cm2c^3 + 32*r_cm2c^4) + (x[,"n_22"] + x[,"n_24"])*log10(138 - 31*r_cm2c + 85*r_cm2c^2 - 108*r_cm2c^3 + 54*r_cm2c^4)
logL_cm2c <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 3*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 3*x[,"n_31"] - 4*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(log(3) + 2*log(5)) + (x[,"n_21"] + x[,"n_25"])*log(7) + (x[,"n_00"] + x[,"n_46"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_41"])*(log(7 - 4*r_cm2c) + 2*log(pmax(1e-6,r_cm2c))) + (x[,"n_06"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cm2c)) + 3*log(pmax(1e-6,r_cm2c))) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(3 + 4*r_cm2c)) + (x[,"n_21"] + x[,"n_25"])*log(3 + 2*r_cm2c - 2*r_cm2c^2) + (x[,"n_16"] + x[,"n_30"])*(2*log(pmax(1e-6,r_cm2c)) + log(5 - 6*r_cm2c + 4*r_cm2c^2)) + (x[,"n_15"] + x[,"n_31"])*(log(pmax(1e-6,r_cm2c)) + log(14 - 3*r_cm2c + 4*r_cm2c^2)) + (x[,"n_10"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(3 - 2*r_cm2c + 4*r_cm2c^2)) + (x[,"n_20"] + x[,"n_26"])*(log(pmax(1e-6,r_cm2c)) + log(7 - 13*r_cm2c + 12*r_cm2c^2 - 6*r_cm2c^3)) + (x[,"n_11"] + x[,"n_35"])*log(15 - 20*r_cm2c + 9*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_04"] + x[,"n_42"])*(log(pmax(1e-6,r_cm2c)) + log(15 + r_cm2c - 7*r_cm2c^2 + 9*r_cm2c^3)) + x[,"n_23"]*log(111 + 10*r_cm2c - 58*r_cm2c^2 + 96*r_cm2c^3 - 48*r_cm2c^4) + (x[,"n_14"] + x[,"n_32"])*log(39 + 76*r_cm2c - 27*r_cm2c^2 + 50*r_cm2c^3 - 36*r_cm2c^4) + (x[,"n_12"] + x[,"n_34"])*log(102 - 28*r_cm2c - 93*r_cm2c^2 + 94*r_cm2c^3 - 36*r_cm2c^4) + (x[,"n_03"] + x[,"n_43"])*log(9 + 18*r_cm2c - 34*r_cm2c^2 + 32*r_cm2c^3 - 16*r_cm2c^4) + (x[,"n_02"] + x[,"n_44"])*log(18 - 32*r_cm2c + 34*r_cm2c^2 - 29*r_cm2c^3 + 9*r_cm2c^4) + (x[,"n_13"] + x[,"n_33"])*log(63 - 14*r_cm2c + 46*r_cm2c^2 - 64*r_cm2c^3 + 32*r_cm2c^4) + (x[,"n_22"] + x[,"n_24"])*log(138 - 31*r_cm2c + 85*r_cm2c^2 - 108*r_cm2c^3 + 54*r_cm2c^4)
logL_cm1c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n05 - 4*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44 - 3*n45 - 4*n46)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(log(3) + 2*log(5)) + (n00 + n06 + n40 + n46)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n45)*(log(pmax(1e-6,r)) + log(2 - r - r^2)) + (n05 + n41)*(log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n21 + n25)*log(15 - 11*r + 11*r^2) + (n15 + n31)*log(3 + r - r^3) + (n10 + n16 + n30 + n36)*(log(pmax(1e-6,r)) + log(1 - 2*r + 2*r^2 - r^3)) + (n11 + n35)*log(3 + 2*r - 3*r^2 + r^3) + (n22 + n24)*log(129 + 32*r - 86*r^2 + 108*r^3 - 54*r^4) + (n13 + n33)*log(30 + 7*r - 23*r^2 + 32*r^3 - 16*r^4) + (n04 + n42)*log(9 - 11*r^2 + 14*r^3 - 9*r^4) + (n02 + n44)*log(3 + 16*r - 23*r^2 + 22*r^3 - 9*r^4) + (n20 + n26)*log(3 - 8*r + 14*r^2 - 12*r^3 + 6*r^4) + (n03 + n43)*log(9 - 13*r + 21*r^2 - 16*r^3 + 8*r^4) + (n12 + n34)*log(21 - 10*r + 18*r^2 - 20*r^3 + 9*r^4) + (n14 + n32)*log(18 - 2*r + 12*r^2 - 16*r^3 + 9*r^4) + n23*log(54 - r + 25*r^2 - 48*r^3 + 24*r^4)
return(L)}
interlogL_cm1c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cm1c <- 4*(x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*log10(2) - (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*(-4*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_31"])*(-3*log10(2) + 3*log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-3*log10(2) + 3*log10(3)) - (x[,"n_05"] + x[,"n_41"])*(-3*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_45"])*(-3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_43"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_21"] + x[,"n_25"])*(-2*log10(2) + 2*log10(7)) - (x[,"n_20"] + x[,"n_26"])*(-3*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_32"])*(-4*log10(2) + 3*log10(3) + log10(11)) - (x[,"n_12"] + x[,"n_34"])*(-4*log10(2) + 3*log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + log10(7) + log10(17)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + log10(7) + log10(17)) - x[,"n_23"]*(-2*log10(2) + log10(13) + log10(17)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(3) + log10(41)) - (x[,"n_22"] + x[,"n_24"])*(-3*log10(2) + log10(1069)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_45"])*(log10(pmax(1e-6,r_cm1c)) + log10(2 - r_cm1c - r_cm1c^2)) + (x[,"n_05"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm1c)) + log10(3 - 4*r_cm1c + r_cm1c^2)) + (x[,"n_21"] + x[,"n_25"])*log10(15 - 11*r_cm1c + 11*r_cm1c^2) + (x[,"n_15"] + x[,"n_31"])*log10(3 + r_cm1c - r_cm1c^3) + (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*(log10(pmax(1e-6,r_cm1c)) + log10(1 - 2*r_cm1c + 2*r_cm1c^2 - r_cm1c^3)) + (x[,"n_11"] + x[,"n_35"])*log10(3 + 2*r_cm1c - 3*r_cm1c^2 + r_cm1c^3) + (x[,"n_22"] + x[,"n_24"])*log10(129 + 32*r_cm1c - 86*r_cm1c^2 + 108*r_cm1c^3 - 54*r_cm1c^4) + (x[,"n_13"] + x[,"n_33"])*log10(30 + 7*r_cm1c - 23*r_cm1c^2 + 32*r_cm1c^3 - 16*r_cm1c^4) + (x[,"n_04"] + x[,"n_42"])*log10(9 - 11*r_cm1c^2 + 14*r_cm1c^3 - 9*r_cm1c^4) + (x[,"n_02"] + x[,"n_44"])*log10(3 + 16*r_cm1c - 23*r_cm1c^2 + 22*r_cm1c^3 - 9*r_cm1c^4) + (x[,"n_20"] + x[,"n_26"])*log10(3 - 8*r_cm1c + 14*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_03"] + x[,"n_43"])*log10(9 - 13*r_cm1c + 21*r_cm1c^2 - 16*r_cm1c^3 + 8*r_cm1c^4) + (x[,"n_12"] + x[,"n_34"])*log10(21 - 10*r_cm1c + 18*r_cm1c^2 - 20*r_cm1c^3 + 9*r_cm1c^4) + (x[,"n_14"] + x[,"n_32"])*log10(18 - 2*r_cm1c + 12*r_cm1c^2 - 16*r_cm1c^3 + 9*r_cm1c^4) + x[,"n_23"]*log10(54 - r_cm1c + 25*r_cm1c^2 - 48*r_cm1c^3 + 24*r_cm1c^4)
logL_cm1c <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*(2*log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_45"])*(log(pmax(1e-6,r_cm1c)) + log(2 - r_cm1c - r_cm1c^2)) + (x[,"n_05"] + x[,"n_41"])*(log(pmax(1e-6,r_cm1c)) + log(3 - 4*r_cm1c + r_cm1c^2)) + (x[,"n_21"] + x[,"n_25"])*log(15 - 11*r_cm1c + 11*r_cm1c^2) + (x[,"n_15"] + x[,"n_31"])*log(3 + r_cm1c - r_cm1c^3) + (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*(log(pmax(1e-6,r_cm1c)) + log(1 - 2*r_cm1c + 2*r_cm1c^2 - r_cm1c^3)) + (x[,"n_11"] + x[,"n_35"])*log(3 + 2*r_cm1c - 3*r_cm1c^2 + r_cm1c^3) + (x[,"n_22"] + x[,"n_24"])*log(129 + 32*r_cm1c - 86*r_cm1c^2 + 108*r_cm1c^3 - 54*r_cm1c^4) + (x[,"n_13"] + x[,"n_33"])*log(30 + 7*r_cm1c - 23*r_cm1c^2 + 32*r_cm1c^3 - 16*r_cm1c^4) + (x[,"n_04"] + x[,"n_42"])*log(9 - 11*r_cm1c^2 + 14*r_cm1c^3 - 9*r_cm1c^4) + (x[,"n_02"] + x[,"n_44"])*log(3 + 16*r_cm1c - 23*r_cm1c^2 + 22*r_cm1c^3 - 9*r_cm1c^4) + (x[,"n_20"] + x[,"n_26"])*log(3 - 8*r_cm1c + 14*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_03"] + x[,"n_43"])*log(9 - 13*r_cm1c + 21*r_cm1c^2 - 16*r_cm1c^3 + 8*r_cm1c^4) + (x[,"n_12"] + x[,"n_34"])*log(21 - 10*r_cm1c + 18*r_cm1c^2 - 20*r_cm1c^3 + 9*r_cm1c^4) + (x[,"n_14"] + x[,"n_32"])*log(18 - 2*r_cm1c + 12*r_cm1c^2 - 16*r_cm1c^3 + 9*r_cm1c^4) + x[,"n_23"]*log(54 - r_cm1c + 25*r_cm1c^2 - 48*r_cm1c^3 + 24*r_cm1c^4)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 3*n11 - 4*n12 - 3*n13 - 4*n14 - 3*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 3*n31 - 4*n32 - 3*n33 - 4*n34 - 3*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46)*log(2) + (n01 + n02 + n04 + n05 + n11 + n12 + n14 + n15 + n20 + n21 + n22 + n23 + n24 + n25 + n26 + n31 + n32 + n34 + n35 + n41 + n42 + n44 + n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*log(5) + 2*(n05 + n41)*log(pmax(1e-6,1 - r)) + 2*(n01 + n45)*log(pmax(1e-6,r)) + (n06 + n40)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n11 + n35)*(log(2 - r) + log(pmax(1e-6,r))) + (n00 + n46)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n15 + n31)*log(1 - r^2) + (n21 + n25)*log(3 - 2*r + 2*r^2) + (n10 + n36)*(2*log(pmax(1e-6,r)) + log(3 - 6*r + 4*r^2)) + (n16 + n30)*(2*log(pmax(1e-6,1 - r)) + log(1 - 2*r + 4*r^2)) + (n20 + n26)*(log(pmax(1e-6,r)) + log(1 - 3*r + 4*r^2 - 2*r^3)) + (n02 + n44)*(log(pmax(1e-6,r)) + log(1 + 3*r - 5*r^2 + 3*r^3)) + n23*log(9 + 18*r - 34*r^2 + 32*r^3 - 16*r^4) + (n03 + n43)*log(1 + 14*r - 30*r^2 + 32*r^3 - 16*r^4) + (n12 + n34)*log(3 + 16*r - 19*r^2 + 22*r^3 - 12*r^4) + (n14 + n32)*log(10 + 4*r - 25*r^2 + 26*r^3 - 12*r^4) + (n04 + n42)*log(2 - 4*r + 6*r^2 - 7*r^3 + 3*r^4) + (n22 + n24)*log(18 - 17*r + 35*r^2 - 36*r^3 + 18*r^4) + (n13 + n33)*log(27 - 34*r + 66*r^2 - 64*r^3 + 32*r^4)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cr <- 3*(x[,"n_20"] + x[,"n_26"])*log10(2) + 2*(x[,"n_16"] + x[,"n_30"])*log10(2) + 2*(x[,"n_10"] + x[,"n_36"])*log10(2) + 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 2*(x[,"n_05"] + x[,"n_41"])*log10(2) + 2*(x[,"n_01"] + x[,"n_45"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) + (x[,"n_15"] + x[,"n_31"])*(2*log10(2) - log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-2*log10(2) + log10(3)) - (x[,"n_21"] + x[,"n_25"])*(-log10(2) + log10(5)) - x[,"n_23"]*(-log10(2) + 2*log10(5)) - (x[,"n_03"] + x[,"n_43"])*(-log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(11)) + (x[,"n_04"] + x[,"n_42"])*(4*log10(2) - log10(13)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_24"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_13"] + x[,"n_33"])*(-log10(2) + log10(41)) + 2*(x[,"n_05"] + x[,"n_41"])*log10(pmax(1e-6,1 - r_cr)) + 2*(x[,"n_01"] + x[,"n_45"])*log10(pmax(1e-6,r_cr)) + (x[,"n_06"] + x[,"n_40"])*(3*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_11"] + x[,"n_35"])*(log10(2 - r_cr) + log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_46"])*(log10(pmax(1e-6,1 - r_cr)) + 3*log10(pmax(1e-6,r_cr))) + (x[,"n_15"] + x[,"n_31"])*log10(1 - r_cr^2) + (x[,"n_21"] + x[,"n_25"])*log10(3 - 2*r_cr + 2*r_cr^2) + (x[,"n_10"] + x[,"n_36"])*(2*log10(pmax(1e-6,r_cr)) + log10(3 - 6*r_cr + 4*r_cr^2)) + (x[,"n_16"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(1 - 2*r_cr + 4*r_cr^2)) + (x[,"n_20"] + x[,"n_26"])*(log10(pmax(1e-6,r_cr)) + log10(1 - 3*r_cr + 4*r_cr^2 - 2*r_cr^3)) + (x[,"n_02"] + x[,"n_44"])*(log10(pmax(1e-6,r_cr)) + log10(1 + 3*r_cr - 5*r_cr^2 + 3*r_cr^3)) + x[,"n_23"]*log10(9 + 18*r_cr - 34*r_cr^2 + 32*r_cr^3 - 16*r_cr^4) + (x[,"n_03"] + x[,"n_43"])*log10(1 + 14*r_cr - 30*r_cr^2 + 32*r_cr^3 - 16*r_cr^4) + (x[,"n_12"] + x[,"n_34"])*log10(3 + 16*r_cr - 19*r_cr^2 + 22*r_cr^3 - 12*r_cr^4) + (x[,"n_14"] + x[,"n_32"])*log10(10 + 4*r_cr - 25*r_cr^2 + 26*r_cr^3 - 12*r_cr^4) + (x[,"n_04"] + x[,"n_42"])*log10(2 - 4*r_cr + 6*r_cr^2 - 7*r_cr^3 + 3*r_cr^4) + (x[,"n_22"] + x[,"n_24"])*log10(18 - 17*r_cr + 35*r_cr^2 - 36*r_cr^3 + 18*r_cr^4) + (x[,"n_13"] + x[,"n_33"])*log10(27 - 34*r_cr + 66*r_cr^2 - 64*r_cr^3 + 32*r_cr^4)
logL_cr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 3*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 3*x[,"n_31"] - 4*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_11"] + x[,"n_12"] + x[,"n_14"] + x[,"n_15"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_26"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_41"] + x[,"n_42"] + x[,"n_44"] + x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*log(5) + 2*(x[,"n_05"] + x[,"n_41"])*log(pmax(1e-6,1 - r_cr)) + 2*(x[,"n_01"] + x[,"n_45"])*log(pmax(1e-6,r_cr)) + (x[,"n_06"] + x[,"n_40"])*(3*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_11"] + x[,"n_35"])*(log(2 - r_cr) + log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_46"])*(log(pmax(1e-6,1 - r_cr)) + 3*log(pmax(1e-6,r_cr))) + (x[,"n_15"] + x[,"n_31"])*log(1 - r_cr^2) + (x[,"n_21"] + x[,"n_25"])*log(3 - 2*r_cr + 2*r_cr^2) + (x[,"n_10"] + x[,"n_36"])*(2*log(pmax(1e-6,r_cr)) + log(3 - 6*r_cr + 4*r_cr^2)) + (x[,"n_16"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_cr)) + log(1 - 2*r_cr + 4*r_cr^2)) + (x[,"n_20"] + x[,"n_26"])*(log(pmax(1e-6,r_cr)) + log(1 - 3*r_cr + 4*r_cr^2 - 2*r_cr^3)) + (x[,"n_02"] + x[,"n_44"])*(log(pmax(1e-6,r_cr)) + log(1 + 3*r_cr - 5*r_cr^2 + 3*r_cr^3)) + x[,"n_23"]*log(9 + 18*r_cr - 34*r_cr^2 + 32*r_cr^3 - 16*r_cr^4) + (x[,"n_03"] + x[,"n_43"])*log(1 + 14*r_cr - 30*r_cr^2 + 32*r_cr^3 - 16*r_cr^4) + (x[,"n_12"] + x[,"n_34"])*log(3 + 16*r_cr - 19*r_cr^2 + 22*r_cr^3 - 12*r_cr^4) + (x[,"n_14"] + x[,"n_32"])*log(10 + 4*r_cr - 25*r_cr^2 + 26*r_cr^3 - 12*r_cr^4) + (x[,"n_04"] + x[,"n_42"])*log(2 - 4*r_cr + 6*r_cr^2 - 7*r_cr^3 + 3*r_cr^4) + (x[,"n_22"] + x[,"n_24"])*log(18 - 17*r_cr + 35*r_cr^2 - 36*r_cr^3 + 18*r_cr^4) + (x[,"n_13"] + x[,"n_33"])*log(27 - 34*r_cr + 66*r_cr^2 - 64*r_cr^3 + 32*r_cr^4)
logL_rc <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 3*n11 - 4*n12 - 3*n13 - 4*n14 - 3*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 3*n31 - 4*n32 - 3*n33 - 4*n34 - 3*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46)*log(2) + (n01 + n02 + n04 + n05 + n11 + n12 + n14 + n15 + n20 + n21 + n22 + n23 + n24 + n25 + n26 + n31 + n32 + n34 + n35 + n41 + n42 + n44 + n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*log(5) + 2*(n01 + n45)*log(pmax(1e-6,1 - r)) + 2*(n05 + n41)*log(pmax(1e-6,r)) + (n00 + n46)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n15 + n31)*(log(2 - r) + log(pmax(1e-6,r))) + (n06 + n40)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n11 + n35)*log(1 - r^2) + (n21 + n25)*log(3 - 2*r + 2*r^2) + (n16 + n30)*(2*log(pmax(1e-6,r)) + log(3 - 6*r + 4*r^2)) + (n10 + n36)*(2*log(pmax(1e-6,1 - r)) + log(1 - 2*r + 4*r^2)) + (n20 + n26)*(log(pmax(1e-6,r)) + log(1 - 3*r + 4*r^2 - 2*r^3)) + (n04 + n42)*(log(pmax(1e-6,r)) + log(1 + 3*r - 5*r^2 + 3*r^3)) + n23*log(9 + 18*r - 34*r^2 + 32*r^3 - 16*r^4) + (n03 + n43)*log(1 + 14*r - 30*r^2 + 32*r^3 - 16*r^4) + (n14 + n32)*log(3 + 16*r - 19*r^2 + 22*r^3 - 12*r^4) + (n12 + n34)*log(10 + 4*r - 25*r^2 + 26*r^3 - 12*r^4) + (n02 + n44)*log(2 - 4*r + 6*r^2 - 7*r^3 + 3*r^4) + (n22 + n24)*log(18 - 17*r + 35*r^2 - 36*r^3 + 18*r^4) + (n13 + n33)*log(27 - 34*r + 66*r^2 - 64*r^3 + 32*r^4)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_rc <- 3*(x[,"n_20"] + x[,"n_26"])*log10(2) + 2*(x[,"n_16"] + x[,"n_30"])*log10(2) + 2*(x[,"n_10"] + x[,"n_36"])*log10(2) + 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 2*(x[,"n_05"] + x[,"n_41"])*log10(2) + 2*(x[,"n_01"] + x[,"n_45"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) + (x[,"n_11"] + x[,"n_35"])*(2*log10(2) - log10(3)) - (x[,"n_15"] + x[,"n_31"])*(-2*log10(2) + log10(3)) - (x[,"n_21"] + x[,"n_25"])*(-log10(2) + log10(5)) - x[,"n_23"]*(-log10(2) + 2*log10(5)) - (x[,"n_03"] + x[,"n_43"])*(-log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(11)) + (x[,"n_02"] + x[,"n_44"])*(4*log10(2) - log10(13)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_24"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_13"] + x[,"n_33"])*(-log10(2) + log10(41)) + 2*(x[,"n_01"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_rc)) + 2*(x[,"n_05"] + x[,"n_41"])*log10(pmax(1e-6,r_rc)) + (x[,"n_00"] + x[,"n_46"])*(3*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc))) + (x[,"n_15"] + x[,"n_31"])*(log10(2 - r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_06"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_rc)) + 3*log10(pmax(1e-6,r_rc))) + (x[,"n_11"] + x[,"n_35"])*log10(1 - r_rc^2) + (x[,"n_21"] + x[,"n_25"])*log10(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_16"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_rc)) + log10(3 - 6*r_rc + 4*r_rc^2)) + (x[,"n_10"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(1 - 2*r_rc + 4*r_rc^2)) + (x[,"n_20"] + x[,"n_26"])*(log10(pmax(1e-6,r_rc)) + log10(1 - 3*r_rc + 4*r_rc^2 - 2*r_rc^3)) + (x[,"n_04"] + x[,"n_42"])*(log10(pmax(1e-6,r_rc)) + log10(1 + 3*r_rc - 5*r_rc^2 + 3*r_rc^3)) + x[,"n_23"]*log10(9 + 18*r_rc - 34*r_rc^2 + 32*r_rc^3 - 16*r_rc^4) + (x[,"n_03"] + x[,"n_43"])*log10(1 + 14*r_rc - 30*r_rc^2 + 32*r_rc^3 - 16*r_rc^4) + (x[,"n_14"] + x[,"n_32"])*log10(3 + 16*r_rc - 19*r_rc^2 + 22*r_rc^3 - 12*r_rc^4) + (x[,"n_12"] + x[,"n_34"])*log10(10 + 4*r_rc - 25*r_rc^2 + 26*r_rc^3 - 12*r_rc^4) + (x[,"n_02"] + x[,"n_44"])*log10(2 - 4*r_rc + 6*r_rc^2 - 7*r_rc^3 + 3*r_rc^4) + (x[,"n_22"] + x[,"n_24"])*log10(18 - 17*r_rc + 35*r_rc^2 - 36*r_rc^3 + 18*r_rc^4) + (x[,"n_13"] + x[,"n_33"])*log10(27 - 34*r_rc + 66*r_rc^2 - 64*r_rc^3 + 32*r_rc^4)
logL_rc <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 3*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 3*x[,"n_31"] - 4*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_11"] + x[,"n_12"] + x[,"n_14"] + x[,"n_15"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_26"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_41"] + x[,"n_42"] + x[,"n_44"] + x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*log(5) + 2*(x[,"n_01"] + x[,"n_45"])*log(pmax(1e-6,1 - r_rc)) + 2*(x[,"n_05"] + x[,"n_41"])*log(pmax(1e-6,r_rc)) + (x[,"n_00"] + x[,"n_46"])*(3*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc))) + (x[,"n_15"] + x[,"n_31"])*(log(2 - r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_06"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_rc)) + 3*log(pmax(1e-6,r_rc))) + (x[,"n_11"] + x[,"n_35"])*log(1 - r_rc^2) + (x[,"n_21"] + x[,"n_25"])*log(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_16"] + x[,"n_30"])*(2*log(pmax(1e-6,r_rc)) + log(3 - 6*r_rc + 4*r_rc^2)) + (x[,"n_10"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_rc)) + log(1 - 2*r_rc + 4*r_rc^2)) + (x[,"n_20"] + x[,"n_26"])*(log(pmax(1e-6,r_rc)) + log(1 - 3*r_rc + 4*r_rc^2 - 2*r_rc^3)) + (x[,"n_04"] + x[,"n_42"])*(log(pmax(1e-6,r_rc)) + log(1 + 3*r_rc - 5*r_rc^2 + 3*r_rc^3)) + x[,"n_23"]*log(9 + 18*r_rc - 34*r_rc^2 + 32*r_rc^3 - 16*r_rc^4) + (x[,"n_03"] + x[,"n_43"])*log(1 + 14*r_rc - 30*r_rc^2 + 32*r_rc^3 - 16*r_rc^4) + (x[,"n_14"] + x[,"n_32"])*log(3 + 16*r_rc - 19*r_rc^2 + 22*r_rc^3 - 12*r_rc^4) + (x[,"n_12"] + x[,"n_34"])*log(10 + 4*r_rc - 25*r_rc^2 + 26*r_rc^3 - 12*r_rc^4) + (x[,"n_02"] + x[,"n_44"])*log(2 - 4*r_rc + 6*r_rc^2 - 7*r_rc^3 + 3*r_rc^4) + (x[,"n_22"] + x[,"n_24"])*log(18 - 17*r_rc + 35*r_rc^2 - 36*r_rc^3 + 18*r_rc^4) + (x[,"n_13"] + x[,"n_33"])*log(27 - 34*r_rc + 66*r_rc^2 - 64*r_rc^3 + 32*r_rc^4)
logL_rm2c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n05 - 4*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44 - 3*n45 - 4*n46)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(log(3) + 2*log(5)) + (n00 + n06 + n40 + n46)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n05 + n41)*(log(pmax(1e-6,r)) + log(2 - r - r^2)) + (n01 + n45)*(log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n21 + n25)*log(15 - 11*r + 11*r^2) + (n11 + n35)*log(3 + r - r^3) + (n10 + n16 + n30 + n36)*(log(pmax(1e-6,r)) + log(1 - 2*r + 2*r^2 - r^3)) + (n15 + n31)*log(3 + 2*r - 3*r^2 + r^3) + (n22 + n24)*log(129 + 32*r - 86*r^2 + 108*r^3 - 54*r^4) + (n13 + n33)*log(30 + 7*r - 23*r^2 + 32*r^3 - 16*r^4) + (n02 + n44)*log(9 - 11*r^2 + 14*r^3 - 9*r^4) + (n04 + n42)*log(3 + 16*r - 23*r^2 + 22*r^3 - 9*r^4) + (n20 + n26)*log(3 - 8*r + 14*r^2 - 12*r^3 + 6*r^4) + (n03 + n43)*log(9 - 13*r + 21*r^2 - 16*r^3 + 8*r^4) + (n14 + n32)*log(21 - 10*r + 18*r^2 - 20*r^3 + 9*r^4) + (n12 + n34)*log(18 - 2*r + 12*r^2 - 16*r^3 + 9*r^4) + n23*log(54 - r + 25*r^2 - 48*r^3 + 24*r^4)
return(L)}
interlogL_rm2c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_rm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_rm2c <- 4*(x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*log10(2) - (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*(-4*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_31"])*(-3*log10(2) + 3*log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-3*log10(2) + 3*log10(3)) - (x[,"n_05"] + x[,"n_41"])*(-3*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_45"])*(-3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_43"])*(-2*log10(2) + 2*log10(5)) - (x[,"n_21"] + x[,"n_25"])*(-2*log10(2) + 2*log10(7)) - (x[,"n_20"] + x[,"n_26"])*(-3*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_32"])*(-4*log10(2) + 3*log10(3) + log10(11)) - (x[,"n_12"] + x[,"n_34"])*(-4*log10(2) + 3*log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + log10(7) + log10(17)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + log10(7) + log10(17)) - x[,"n_23"]*(-2*log10(2) + log10(13) + log10(17)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(3) + log10(41)) - (x[,"n_22"] + x[,"n_24"])*(-3*log10(2) + log10(1069)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + 2*log10(pmax(1e-6,r_rm2c))) + (x[,"n_05"] + x[,"n_41"])*(log10(pmax(1e-6,r_rm2c)) + log10(2 - r_rm2c - r_rm2c^2)) + (x[,"n_01"] + x[,"n_45"])*(log10(pmax(1e-6,r_rm2c)) + log10(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_21"] + x[,"n_25"])*log10(15 - 11*r_rm2c + 11*r_rm2c^2) + (x[,"n_11"] + x[,"n_35"])*log10(3 + r_rm2c - r_rm2c^3) + (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*(log10(pmax(1e-6,r_rm2c)) + log10(1 - 2*r_rm2c + 2*r_rm2c^2 - r_rm2c^3)) + (x[,"n_15"] + x[,"n_31"])*log10(3 + 2*r_rm2c - 3*r_rm2c^2 + r_rm2c^3) + (x[,"n_22"] + x[,"n_24"])*log10(129 + 32*r_rm2c - 86*r_rm2c^2 + 108*r_rm2c^3 - 54*r_rm2c^4) + (x[,"n_13"] + x[,"n_33"])*log10(30 + 7*r_rm2c - 23*r_rm2c^2 + 32*r_rm2c^3 - 16*r_rm2c^4) + (x[,"n_02"] + x[,"n_44"])*log10(9 - 11*r_rm2c^2 + 14*r_rm2c^3 - 9*r_rm2c^4) + (x[,"n_04"] + x[,"n_42"])*log10(3 + 16*r_rm2c - 23*r_rm2c^2 + 22*r_rm2c^3 - 9*r_rm2c^4) + (x[,"n_20"] + x[,"n_26"])*log10(3 - 8*r_rm2c + 14*r_rm2c^2 - 12*r_rm2c^3 + 6*r_rm2c^4) + (x[,"n_03"] + x[,"n_43"])*log10(9 - 13*r_rm2c + 21*r_rm2c^2 - 16*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_14"] + x[,"n_32"])*log10(21 - 10*r_rm2c + 18*r_rm2c^2 - 20*r_rm2c^3 + 9*r_rm2c^4) + (x[,"n_12"] + x[,"n_34"])*log10(18 - 2*r_rm2c + 12*r_rm2c^2 - 16*r_rm2c^3 + 9*r_rm2c^4) + x[,"n_23"]*log10(54 - r_rm2c + 25*r_rm2c^2 - 48*r_rm2c^3 + 24*r_rm2c^4)
logL_rm2c <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*(2*log(pmax(1e-6,1 - r_rm2c)) + 2*log(pmax(1e-6,r_rm2c))) + (x[,"n_05"] + x[,"n_41"])*(log(pmax(1e-6,r_rm2c)) + log(2 - r_rm2c - r_rm2c^2)) + (x[,"n_01"] + x[,"n_45"])*(log(pmax(1e-6,r_rm2c)) + log(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_21"] + x[,"n_25"])*log(15 - 11*r_rm2c + 11*r_rm2c^2) + (x[,"n_11"] + x[,"n_35"])*log(3 + r_rm2c - r_rm2c^3) + (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*(log(pmax(1e-6,r_rm2c)) + log(1 - 2*r_rm2c + 2*r_rm2c^2 - r_rm2c^3)) + (x[,"n_15"] + x[,"n_31"])*log(3 + 2*r_rm2c - 3*r_rm2c^2 + r_rm2c^3) + (x[,"n_22"] + x[,"n_24"])*log(129 + 32*r_rm2c - 86*r_rm2c^2 + 108*r_rm2c^3 - 54*r_rm2c^4) + (x[,"n_13"] + x[,"n_33"])*log(30 + 7*r_rm2c - 23*r_rm2c^2 + 32*r_rm2c^3 - 16*r_rm2c^4) + (x[,"n_02"] + x[,"n_44"])*log(9 - 11*r_rm2c^2 + 14*r_rm2c^3 - 9*r_rm2c^4) + (x[,"n_04"] + x[,"n_42"])*log(3 + 16*r_rm2c - 23*r_rm2c^2 + 22*r_rm2c^3 - 9*r_rm2c^4) + (x[,"n_20"] + x[,"n_26"])*log(3 - 8*r_rm2c + 14*r_rm2c^2 - 12*r_rm2c^3 + 6*r_rm2c^4) + (x[,"n_03"] + x[,"n_43"])*log(9 - 13*r_rm2c + 21*r_rm2c^2 - 16*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_14"] + x[,"n_32"])*log(21 - 10*r_rm2c + 18*r_rm2c^2 - 20*r_rm2c^3 + 9*r_rm2c^4) + (x[,"n_12"] + x[,"n_34"])*log(18 - 2*r_rm2c + 12*r_rm2c^2 - 16*r_rm2c^3 + 9*r_rm2c^4) + x[,"n_23"]*log(54 - r_rm2c + 25*r_rm2c^2 - 48*r_rm2c^3 + 24*r_rm2c^4)
logL_rm1c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 3*n11 - 4*n12 - 3*n13 - 4*n14 - 3*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 3*n31 - 4*n32 - 3*n33 - 4*n34 - 3*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(log(3) + 2*log(5)) + (n21 + n25)*log(7) + (n06 + n40)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n45)*(log(7 - 4*r) + 2*log(pmax(1e-6,r))) + (n00 + n46)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n05 + n41)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r)) + (n21 + n25)*log(3 + 2*r - 2*r^2) + (n10 + n36)*(2*log(pmax(1e-6,r)) + log(5 - 6*r + 4*r^2)) + (n11 + n35)*(log(pmax(1e-6,r)) + log(14 - 3*r + 4*r^2)) + (n16 + n30)*(2*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 4*r^2)) + (n20 + n26)*(log(pmax(1e-6,r)) + log(7 - 13*r + 12*r^2 - 6*r^3)) + (n15 + n31)*log(15 - 20*r + 9*r^2 - 4*r^3) + (n02 + n44)*(log(pmax(1e-6,r)) + log(15 + r - 7*r^2 + 9*r^3)) + n23*log(111 + 10*r - 58*r^2 + 96*r^3 - 48*r^4) + (n12 + n34)*log(39 + 76*r - 27*r^2 + 50*r^3 - 36*r^4) + (n14 + n32)*log(102 - 28*r - 93*r^2 + 94*r^3 - 36*r^4) + (n03 + n43)*log(9 + 18*r - 34*r^2 + 32*r^3 - 16*r^4) + (n04 + n42)*log(18 - 32*r + 34*r^2 - 29*r^3 + 9*r^4) + (n13 + n33)*log(63 - 14*r + 46*r^2 - 64*r^3 + 32*r^4) + (n22 + n24)*log(138 - 31*r + 85*r^2 - 108*r^3 + 54*r^4)
return(L)}
interlogL_rm1c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_rm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_rm1c <- 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) - (x[,"n_16"] + x[,"n_30"])*(-2*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_36"])*(-2*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_31"])*(-2*log10(2) + 3*log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-2*log10(2) + 3*log10(3)) - (x[,"n_05"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_45"])*(-2*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_43"])*(-log10(2) + 2*log10(5)) - (x[,"n_21"] + x[,"n_25"])*(-log10(2) + log10(7)) - (x[,"n_20"] + x[,"n_26"])*(-3*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + 3*log10(3) + log10(11)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + 3*log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + log10(7) + log10(17)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + log10(7) + log10(17)) - x[,"n_23"]*(-log10(2) + log10(13) + log10(17)) - (x[,"n_13"] + x[,"n_33"])*(-log10(2) + log10(3) + log10(41)) - (x[,"n_22"] + x[,"n_24"])*(-3*log10(2) + log10(1069)) + (x[,"n_06"] + x[,"n_40"])*(3*log10(pmax(1e-6,1 - r_rm1c)) + log10(pmax(1e-6,r_rm1c))) + (x[,"n_01"] + x[,"n_45"])*(log10(7 - 4*r_rm1c) + 2*log10(pmax(1e-6,r_rm1c))) + (x[,"n_00"] + x[,"n_46"])*(log10(pmax(1e-6,1 - r_rm1c)) + 3*log10(pmax(1e-6,r_rm1c))) + (x[,"n_05"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 + 4*r_rm1c)) + (x[,"n_21"] + x[,"n_25"])*log10(3 + 2*r_rm1c - 2*r_rm1c^2) + (x[,"n_10"] + x[,"n_36"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(5 - 6*r_rm1c + 4*r_rm1c^2)) + (x[,"n_11"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm1c)) + log10(14 - 3*r_rm1c + 4*r_rm1c^2)) + (x[,"n_16"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 - 2*r_rm1c + 4*r_rm1c^2)) + (x[,"n_20"] + x[,"n_26"])*(log10(pmax(1e-6,r_rm1c)) + log10(7 - 13*r_rm1c + 12*r_rm1c^2 - 6*r_rm1c^3)) + (x[,"n_15"] + x[,"n_31"])*log10(15 - 20*r_rm1c + 9*r_rm1c^2 - 4*r_rm1c^3) + (x[,"n_02"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm1c)) + log10(15 + r_rm1c - 7*r_rm1c^2 + 9*r_rm1c^3)) + x[,"n_23"]*log10(111 + 10*r_rm1c - 58*r_rm1c^2 + 96*r_rm1c^3 - 48*r_rm1c^4) + (x[,"n_12"] + x[,"n_34"])*log10(39 + 76*r_rm1c - 27*r_rm1c^2 + 50*r_rm1c^3 - 36*r_rm1c^4) + (x[,"n_14"] + x[,"n_32"])*log10(102 - 28*r_rm1c - 93*r_rm1c^2 + 94*r_rm1c^3 - 36*r_rm1c^4) + (x[,"n_03"] + x[,"n_43"])*log10(9 + 18*r_rm1c - 34*r_rm1c^2 + 32*r_rm1c^3 - 16*r_rm1c^4) + (x[,"n_04"] + x[,"n_42"])*log10(18 - 32*r_rm1c + 34*r_rm1c^2 - 29*r_rm1c^3 + 9*r_rm1c^4) + (x[,"n_13"] + x[,"n_33"])*log10(63 - 14*r_rm1c + 46*r_rm1c^2 - 64*r_rm1c^3 + 32*r_rm1c^4) + (x[,"n_22"] + x[,"n_24"])*log10(138 - 31*r_rm1c + 85*r_rm1c^2 - 108*r_rm1c^3 + 54*r_rm1c^4)
logL_rm1c <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 3*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 3*x[,"n_31"] - 4*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(log(3) + 2*log(5)) + (x[,"n_21"] + x[,"n_25"])*log(7) + (x[,"n_06"] + x[,"n_40"])*(3*log(pmax(1e-6,1 - r_rm1c)) + log(pmax(1e-6,r_rm1c))) + (x[,"n_01"] + x[,"n_45"])*(log(7 - 4*r_rm1c) + 2*log(pmax(1e-6,r_rm1c))) + (x[,"n_00"] + x[,"n_46"])*(log(pmax(1e-6,1 - r_rm1c)) + 3*log(pmax(1e-6,r_rm1c))) + (x[,"n_05"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(3 + 4*r_rm1c)) + (x[,"n_21"] + x[,"n_25"])*log(3 + 2*r_rm1c - 2*r_rm1c^2) + (x[,"n_10"] + x[,"n_36"])*(2*log(pmax(1e-6,r_rm1c)) + log(5 - 6*r_rm1c + 4*r_rm1c^2)) + (x[,"n_11"] + x[,"n_35"])*(log(pmax(1e-6,r_rm1c)) + log(14 - 3*r_rm1c + 4*r_rm1c^2)) + (x[,"n_16"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(3 - 2*r_rm1c + 4*r_rm1c^2)) + (x[,"n_20"] + x[,"n_26"])*(log(pmax(1e-6,r_rm1c)) + log(7 - 13*r_rm1c + 12*r_rm1c^2 - 6*r_rm1c^3)) + (x[,"n_15"] + x[,"n_31"])*log(15 - 20*r_rm1c + 9*r_rm1c^2 - 4*r_rm1c^3) + (x[,"n_02"] + x[,"n_44"])*(log(pmax(1e-6,r_rm1c)) + log(15 + r_rm1c - 7*r_rm1c^2 + 9*r_rm1c^3)) + x[,"n_23"]*log(111 + 10*r_rm1c - 58*r_rm1c^2 + 96*r_rm1c^3 - 48*r_rm1c^4) + (x[,"n_12"] + x[,"n_34"])*log(39 + 76*r_rm1c - 27*r_rm1c^2 + 50*r_rm1c^3 - 36*r_rm1c^4) + (x[,"n_14"] + x[,"n_32"])*log(102 - 28*r_rm1c - 93*r_rm1c^2 + 94*r_rm1c^3 - 36*r_rm1c^4) + (x[,"n_03"] + x[,"n_43"])*log(9 + 18*r_rm1c - 34*r_rm1c^2 + 32*r_rm1c^3 - 16*r_rm1c^4) + (x[,"n_04"] + x[,"n_42"])*log(18 - 32*r_rm1c + 34*r_rm1c^2 - 29*r_rm1c^3 + 9*r_rm1c^4) + (x[,"n_13"] + x[,"n_33"])*log(63 - 14*r_rm1c + 46*r_rm1c^2 - 64*r_rm1c^3 + 32*r_rm1c^4) + (x[,"n_22"] + x[,"n_24"])*log(138 - 31*r_rm1c + 85*r_rm1c^2 - 108*r_rm1c^3 + 54*r_rm1c^4)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n05 - 4*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 3*n20 - 3*n21 - 3*n22 - 2*n23 - 3*n24 - 3*n25 - 3*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44 - 3*n45 - 4*n46)*log(2) + (n01 + n02 + n04 + n05 + n11 + n12 + n14 + n15 + n20 + n21 + n22 + n23 + n24 + n25 + n26 + n31 + n32 + n34 + n35 + n41 + n42 + n44 + n45)*log(3) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 2*n20 - n21 - 2*n22 - 2*n23 - 2*n24 - n25 - 2*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n46)*log(5) + 4*(n06 + n40)*log(pmax(1e-6,1 - r)) + 3*(n05 + n41)*log(pmax(1e-6,1 - r)) + 3*(n01 + n45)*log(pmax(1e-6,r)) + 4*(n00 + n46)*log(pmax(1e-6,r)) + (n21 + n25)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n16 + n30)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n20 + n26)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n35)*(log(2 - r) + 2*log(pmax(1e-6,r))) + (n10 + n36)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n15 + n31)*(2*log(pmax(1e-6,1 - r)) + log(1 + r)) + (n02 + n44)*(2*log(pmax(1e-6,r)) + log(4 - 3*r^2)) + (n04 + n42)*(2*log(pmax(1e-6,1 - r)) + log(1 + 6*r - 3*r^2)) + (n12 + n34)*(log(pmax(1e-6,r)) + log(4 + r - 3*r^2 + 3*r^3)) + (n03 + n43)*(log(pmax(1e-6,r)) + log(5 + 3*r - 16*r^2 + 8*r^3)) + (n13 + n33)*log(7 + 17*r - 33*r^2 + 32*r^3 - 16*r^4) + (n22 + n24)*log(6 + 8*r - 17*r^2 + 18*r^3 - 9*r^4) + (n14 + n32)*log(5 - 9*r + 10*r^2 - 9*r^3 + 3*r^4) + n23*log(9 - 13*r + 21*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_rr <- 2*(x[,"n_21"] + x[,"n_25"])*log10(2) + 4*(x[,"n_20"] + x[,"n_26"])*log10(2) + 4*(x[,"n_16"] + x[,"n_30"])*log10(2) + 4*(x[,"n_10"] + x[,"n_36"])*log10(2) + 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 3*(x[,"n_05"] + x[,"n_41"])*log10(2) + 3*(x[,"n_01"] + x[,"n_45"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) - (x[,"n_15"] + x[,"n_31"])*(-3*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-3*log10(2) + log10(3)) - x[,"n_23"]*(-2*log10(2) + 2*log10(5)) - (x[,"n_03"] + x[,"n_43"])*(-2*log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_32"])*(-4*log10(2) + log10(3) + log10(11)) - (x[,"n_12"] + x[,"n_34"])*(-4*log10(2) + log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_24"])*(-4*log10(2) + log10(7) + log10(17)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(41)) + 4*(x[,"n_06"] + x[,"n_40"])*log10(pmax(1e-6,1 - r_rr)) + 3*(x[,"n_05"] + x[,"n_41"])*log10(pmax(1e-6,1 - r_rr)) + 3*(x[,"n_01"] + x[,"n_45"])*log10(pmax(1e-6,r_rr)) + 4*(x[,"n_00"] + x[,"n_46"])*log10(pmax(1e-6,r_rr)) + (x[,"n_21"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr))) + (x[,"n_16"] + x[,"n_30"])*(3*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr))) + (x[,"n_20"] + x[,"n_26"])*(2*log10(pmax(1e-6,1 - r_rr)) + 2*log10(pmax(1e-6,r_rr))) + (x[,"n_11"] + x[,"n_35"])*(log10(2 - r_rr) + 2*log10(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_36"])*(log10(pmax(1e-6,1 - r_rr)) + 3*log10(pmax(1e-6,r_rr))) + (x[,"n_15"] + x[,"n_31"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(1 + r_rr)) + (x[,"n_02"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(4 - 3*r_rr^2)) + (x[,"n_04"] + x[,"n_42"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(1 + 6*r_rr - 3*r_rr^2)) + (x[,"n_12"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(4 + r_rr - 3*r_rr^2 + 3*r_rr^3)) + (x[,"n_03"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(5 + 3*r_rr - 16*r_rr^2 + 8*r_rr^3)) + (x[,"n_13"] + x[,"n_33"])*log10(7 + 17*r_rr - 33*r_rr^2 + 32*r_rr^3 - 16*r_rr^4) + (x[,"n_22"] + x[,"n_24"])*log10(6 + 8*r_rr - 17*r_rr^2 + 18*r_rr^3 - 9*r_rr^4) + (x[,"n_14"] + x[,"n_32"])*log10(5 - 9*r_rr + 10*r_rr^2 - 9*r_rr^3 + 3*r_rr^4) + x[,"n_23"]*log10(9 - 13*r_rr + 21*r_rr^2 - 16*r_rr^3 + 8*r_rr^4)
logL_rr <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_11"] + x[,"n_12"] + x[,"n_14"] + x[,"n_15"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_26"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_41"] + x[,"n_42"] + x[,"n_44"] + x[,"n_45"])*log(3) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - 2*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_46"])*log(5) + 4*(x[,"n_06"] + x[,"n_40"])*log(pmax(1e-6,1 - r_rr)) + 3*(x[,"n_05"] + x[,"n_41"])*log(pmax(1e-6,1 - r_rr)) + 3*(x[,"n_01"] + x[,"n_45"])*log(pmax(1e-6,r_rr)) + 4*(x[,"n_00"] + x[,"n_46"])*log(pmax(1e-6,r_rr)) + (x[,"n_21"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr))) + (x[,"n_16"] + x[,"n_30"])*(3*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr))) + (x[,"n_20"] + x[,"n_26"])*(2*log(pmax(1e-6,1 - r_rr)) + 2*log(pmax(1e-6,r_rr))) + (x[,"n_11"] + x[,"n_35"])*(log(2 - r_rr) + 2*log(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_36"])*(log(pmax(1e-6,1 - r_rr)) + 3*log(pmax(1e-6,r_rr))) + (x[,"n_15"] + x[,"n_31"])*(2*log(pmax(1e-6,1 - r_rr)) + log(1 + r_rr)) + (x[,"n_02"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(4 - 3*r_rr^2)) + (x[,"n_04"] + x[,"n_42"])*(2*log(pmax(1e-6,1 - r_rr)) + log(1 + 6*r_rr - 3*r_rr^2)) + (x[,"n_12"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(4 + r_rr - 3*r_rr^2 + 3*r_rr^3)) + (x[,"n_03"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(5 + 3*r_rr - 16*r_rr^2 + 8*r_rr^3)) + (x[,"n_13"] + x[,"n_33"])*log(7 + 17*r_rr - 33*r_rr^2 + 32*r_rr^3 - 16*r_rr^4) + (x[,"n_22"] + x[,"n_24"])*log(6 + 8*r_rr - 17*r_rr^2 + 18*r_rr^3 - 9*r_rr^4) + (x[,"n_14"] + x[,"n_32"])*log(5 - 9*r_rr + 10*r_rr^2 - 9*r_rr^3 + 3*r_rr^4) + x[,"n_23"]*log(9 - 13*r_rr + 21*r_rr^2 - 16*r_rr^3 + 8*r_rr^4)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_cr,r_rc,r_rm2c,r_rm1c,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_cr,LOD_rc,LOD_rm2c,LOD_rm1c,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_cr,logL_rc,logL_rm2c,logL_rm1c,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","coupling repulsion","repulsion coupling","repulsion mixed2c","repulsion mixed1c","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.3_4.1 <- function(x){
logL_cc <- function(r,n21,n22,n23,n24,n01,n02,n03,n04,n11,n12,n13,n14,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (n21 + n22 + n23 + n24)*log(3) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(4*log(2) + 2*log(5)) + (n01 + n44)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n04 + n41)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n12 + n33)*log(51 - 20*r - 4*r^2) + (n21 + n24)*log(5 + 4*r - 4*r^2) + (n13 + n32)*log(27 + 28*r - 4*r^2) + (n02 + n43)*log(6 - 3*r - 2*r^2) + (n03 + n42)*log(1 + 7*r - 2*r^2) + (n11 + n34)*log(19 - 20*r + 4*r^2) + (n22 + n23)*log(25 - 4*r + 4*r^2) + (n14 + n31)*log(3 + 12*r + 4*r^2)
return(L)}
interlogL_cc <- function(n21,n22,n23,n24,n01,n02,n03,n04,n11,n12,n13,n14,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_cc,c(0,0.5), n21,n22,n23,n24,n01,n02,n03,n04,n11,n12,n13,n14,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cc <- -2*(x[,"n_03"] + x[,"n_42"])*log10(2) - 2*(x[,"n_02"] + x[,"n_43"])*log10(2) - (x[,"n_21"] + x[,"n_24"])*(log10(2) + log10(3)) - (x[,"n_22"] + x[,"n_23"])*(3*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_31"])*(log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_34"])*(log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_32"])*(3*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_33"])*(3*log10(2) + log10(5)) + (x[,"n_01"] + x[,"n_44"])*(log10(3 - 2*r_cc) + log10(pmax(1e-6,1 - r_cc))) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_cc)) + log10(1 + 2*r_cc)) + (x[,"n_12"] + x[,"n_33"])*log10(51 - 20*r_cc - 4*r_cc^2) + (x[,"n_21"] + x[,"n_24"])*log10(5 + 4*r_cc - 4*r_cc^2) + (x[,"n_13"] + x[,"n_32"])*log10(27 + 28*r_cc - 4*r_cc^2) + (x[,"n_02"] + x[,"n_43"])*log10(6 - 3*r_cc - 2*r_cc^2) + (x[,"n_03"] + x[,"n_42"])*log10(1 + 7*r_cc - 2*r_cc^2) + (x[,"n_11"] + x[,"n_34"])*log10(19 - 20*r_cc + 4*r_cc^2) + (x[,"n_22"] + x[,"n_23"])*log10(25 - 4*r_cc + 4*r_cc^2) + (x[,"n_14"] + x[,"n_31"])*log10(3 + 12*r_cc + 4*r_cc^2)
logL_cc <- (x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(4*log(2) + 2*log(5)) + (x[,"n_01"] + x[,"n_44"])*(log(3 - 2*r_cc) + log(pmax(1e-6,1 - r_cc))) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_cc)) + log(1 + 2*r_cc)) + (x[,"n_12"] + x[,"n_33"])*log(51 - 20*r_cc - 4*r_cc^2) + (x[,"n_21"] + x[,"n_24"])*log(5 + 4*r_cc - 4*r_cc^2) + (x[,"n_13"] + x[,"n_32"])*log(27 + 28*r_cc - 4*r_cc^2) + (x[,"n_02"] + x[,"n_43"])*log(6 - 3*r_cc - 2*r_cc^2) + (x[,"n_03"] + x[,"n_42"])*log(1 + 7*r_cc - 2*r_cc^2) + (x[,"n_11"] + x[,"n_34"])*log(19 - 20*r_cc + 4*r_cc^2) + (x[,"n_22"] + x[,"n_23"])*log(25 - 4*r_cc + 4*r_cc^2) + (x[,"n_14"] + x[,"n_31"])*log(3 + 12*r_cc + 4*r_cc^2)
logL_cr <- function(r,n21,n22,n23,n24,n01,n02,n03,n04,n11,n12,n13,n14,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (n21 + n22 + n23 + n24)*log(3) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(4*log(2) + 2*log(5)) + (n01 + n44)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n04 + n41)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n11 + n14 + n31 + n34)*log(9 + 4*r - 4*r^2) + (n22 + n23)*log(23 + 4*r - 4*r^2) + (n03 + n42)*log(6 - 5*r + 2*r^2) + (n02 + n43)*log(3 + r + 2*r^2) + (n13 + n32)*log(45 - 12*r + 4*r^2) + (n21 + n24)*log(7 - 4*r + 4*r^2) + (n12 + n33)*log(37 + 4*r + 4*r^2)
return(L)}
interlogL_cr <- function(n21,n22,n23,n24,n01,n02,n03,n04,n11,n12,n13,n14,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_cr,c(0,0.5), n21,n22,n23,n24,n01,n02,n03,n04,n11,n12,n13,n14,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cr <- -2*(x[,"n_03"] + x[,"n_42"])*log10(2) - 2*(x[,"n_02"] + x[,"n_43"])*log10(2) - (x[,"n_21"] + x[,"n_24"])*(log10(2) + log10(3)) - (x[,"n_22"] + x[,"n_23"])*(3*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_14"] + x[,"n_31"] + x[,"n_34"])*(log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_32"])*(3*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_33"])*(3*log10(2) + log10(5)) + (x[,"n_01"] + x[,"n_44"])*(log10(3 - 2*r_cr) + log10(pmax(1e-6,r_cr))) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,1 - r_cr)) + log10(1 + 2*r_cr)) + (x[,"n_11"] + x[,"n_14"] + x[,"n_31"] + x[,"n_34"])*log10(9 + 4*r_cr - 4*r_cr^2) + (x[,"n_22"] + x[,"n_23"])*log10(23 + 4*r_cr - 4*r_cr^2) + (x[,"n_03"] + x[,"n_42"])*log10(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_02"] + x[,"n_43"])*log10(3 + r_cr + 2*r_cr^2) + (x[,"n_13"] + x[,"n_32"])*log10(45 - 12*r_cr + 4*r_cr^2) + (x[,"n_21"] + x[,"n_24"])*log10(7 - 4*r_cr + 4*r_cr^2) + (x[,"n_12"] + x[,"n_33"])*log10(37 + 4*r_cr + 4*r_cr^2)
logL_cr <- (x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(4*log(2) + 2*log(5)) + (x[,"n_01"] + x[,"n_44"])*(log(3 - 2*r_cr) + log(pmax(1e-6,r_cr))) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,1 - r_cr)) + log(1 + 2*r_cr)) + (x[,"n_11"] + x[,"n_14"] + x[,"n_31"] + x[,"n_34"])*log(9 + 4*r_cr - 4*r_cr^2) + (x[,"n_22"] + x[,"n_23"])*log(23 + 4*r_cr - 4*r_cr^2) + (x[,"n_03"] + x[,"n_42"])*log(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_02"] + x[,"n_43"])*log(3 + r_cr + 2*r_cr^2) + (x[,"n_13"] + x[,"n_32"])*log(45 - 12*r_cr + 4*r_cr^2) + (x[,"n_21"] + x[,"n_24"])*log(7 - 4*r_cr + 4*r_cr^2) + (x[,"n_12"] + x[,"n_33"])*log(37 + 4*r_cr + 4*r_cr^2)
logL_rc <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-2*n01 - 3*n02 - 3*n03 - 2*n04 - 2*n11 - 3*n12 - 3*n13 - 2*n14 - n21 - 3*n22 - 3*n23 - n24 - 2*n31 - 3*n32 - 3*n33 - 2*n34 - 2*n41 - 3*n42 - 3*n43 - 2*n44)*log(2) + (n21 + n22 + n23 + n24)*log(3) + 2*(-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*log(5) + (n01 + n04 + n41 + n44)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n22 + n23)*log(11 + 4*r - 4*r^2) + (n14 + n31)*log(3 - 2*r^2) + (n11 + n34)*log(1 + 4*r - 2*r^2) + (n21 + n24)*log(1 - r + r^2) + (n02 + n43)*log(3 - 3*r + 2*r^2) + (n03 + n42)*log(2 - r + 2*r^2) + (n12 + n13 + n32 + n33)*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rc <- -((x[,"n_03"] + x[,"n_42"])*log10(2)) - (x[,"n_02"] + x[,"n_43"])*log10(2) + 2*(x[,"n_01"] + x[,"n_04"] + x[,"n_41"] + x[,"n_44"])*log10(2) + (x[,"n_21"] + x[,"n_24"])*(2*log10(2) - log10(3)) - (x[,"n_22"] + x[,"n_23"])*(2*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_13"] + x[,"n_32"] + x[,"n_33"])*(2*log10(2) + log10(5)) + (x[,"n_01"] + x[,"n_04"] + x[,"n_41"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc))) + (x[,"n_22"] + x[,"n_23"])*log10(11 + 4*r_rc - 4*r_rc^2) + (x[,"n_14"] + x[,"n_31"])*log10(3 - 2*r_rc^2) + (x[,"n_11"] + x[,"n_34"])*log10(1 + 4*r_rc - 2*r_rc^2) + (x[,"n_21"] + x[,"n_24"])*log10(1 - r_rc + r_rc^2) + (x[,"n_02"] + x[,"n_43"])*log10(3 - 3*r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_42"])*log10(2 - r_rc + 2*r_rc^2) + (x[,"n_12"] + x[,"n_13"] + x[,"n_32"] + x[,"n_33"])*log10(21 - 4*r_rc + 4*r_rc^2)
logL_rc <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - x[,"n_24"] - 2*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + (x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_01"] + x[,"n_04"] + x[,"n_41"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc))) + (x[,"n_22"] + x[,"n_23"])*log(11 + 4*r_rc - 4*r_rc^2) + (x[,"n_14"] + x[,"n_31"])*log(3 - 2*r_rc^2) + (x[,"n_11"] + x[,"n_34"])*log(1 + 4*r_rc - 2*r_rc^2) + (x[,"n_21"] + x[,"n_24"])*log(1 - r_rc + r_rc^2) + (x[,"n_02"] + x[,"n_43"])*log(3 - 3*r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_42"])*log(2 - r_rc + 2*r_rc^2) + (x[,"n_12"] + x[,"n_13"] + x[,"n_32"] + x[,"n_33"])*log(21 - 4*r_rc + 4*r_rc^2)
logL_rr <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-2*n01 - 3*n02 - 3*n03 - 2*n04 - n11 - 3*n12 - 3*n13 - n14 - 2*n21 - 3*n22 - 3*n23 - 2*n24 - n31 - 3*n32 - 3*n33 - n34 - 2*n41 - 3*n42 - 3*n43 - 2*n44)*log(2) + (n21 + n22 + n23 + n24)*log(3) + 2*(-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*log(5) + 2*(n04 + n41)*log(pmax(1e-6,1 - r)) + 2*(n01 + n44)*log(pmax(1e-6,r)) + (n02 + n43)*(log(5 - 2*r) + log(pmax(1e-6,r))) + (n11 + n34)*(log(pmax(1e-6,r)) + log(2 + r)) + (n13 + n32)*log(27 - 12*r - 4*r^2) + (n12 + n33)*log(11 + 20*r - 4*r^2) + (n03 + n42)*log(3 - r - 2*r^2) + (n21 + n24)*log(1 + 2*r - 2*r^2) + (n14 + n31)*log(3 - 4*r + r^2) + (n22 + n23)*log(13 - 4*r + 4*r^2)
return(L)}
interlogL_rr <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_rr,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rr <- 2*(x[,"n_04"] + x[,"n_41"])*log10(2) - (x[,"n_03"] + x[,"n_42"])*log10(2) - (x[,"n_02"] + x[,"n_43"])*log10(2) + 2*(x[,"n_01"] + x[,"n_44"])*log10(2) - (x[,"n_21"] + x[,"n_24"])*(-log10(2) + log10(3)) - (x[,"n_22"] + x[,"n_23"])*(2*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_31"])*(-2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_34"])*(-2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_32"])*(2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_33"])*(2*log10(2) + log10(5)) + 2*(x[,"n_04"] + x[,"n_41"])*log10(pmax(1e-6,1 - r_rr)) + 2*(x[,"n_01"] + x[,"n_44"])*log10(pmax(1e-6,r_rr)) + (x[,"n_02"] + x[,"n_43"])*(log10(5 - 2*r_rr) + log10(pmax(1e-6,r_rr))) + (x[,"n_11"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_13"] + x[,"n_32"])*log10(27 - 12*r_rr - 4*r_rr^2) + (x[,"n_12"] + x[,"n_33"])*log10(11 + 20*r_rr - 4*r_rr^2) + (x[,"n_03"] + x[,"n_42"])*log10(3 - r_rr - 2*r_rr^2) + (x[,"n_21"] + x[,"n_24"])*log10(1 + 2*r_rr - 2*r_rr^2) + (x[,"n_14"] + x[,"n_31"])*log10(3 - 4*r_rr + r_rr^2) + (x[,"n_22"] + x[,"n_23"])*log10(13 - 4*r_rr + 4*r_rr^2)
logL_rr <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - x[,"n_14"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - x[,"n_34"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + (x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + 2*(x[,"n_04"] + x[,"n_41"])*log(pmax(1e-6,1 - r_rr)) + 2*(x[,"n_01"] + x[,"n_44"])*log(pmax(1e-6,r_rr)) + (x[,"n_02"] + x[,"n_43"])*(log(5 - 2*r_rr) + log(pmax(1e-6,r_rr))) + (x[,"n_11"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_13"] + x[,"n_32"])*log(27 - 12*r_rr - 4*r_rr^2) + (x[,"n_12"] + x[,"n_33"])*log(11 + 20*r_rr - 4*r_rr^2) + (x[,"n_03"] + x[,"n_42"])*log(3 - r_rr - 2*r_rr^2) + (x[,"n_21"] + x[,"n_24"])*log(1 + 2*r_rr - 2*r_rr^2) + (x[,"n_14"] + x[,"n_31"])*log(3 - 4*r_rr + r_rr^2) + (x[,"n_22"] + x[,"n_23"])*log(13 - 4*r_rr + 4*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.4_1.4 <- function(x){
logL_cc <- function(r,n12,n13,n21,n24,n31,n34,n42,n43,n11,n14,n22,n23,n32,n33,n41,n44) {
L <- (n12 + n13 + n21 + n24 + n31 + n34 + n42 + n43)*log(3) + (-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(log(2) + log(5)) + 3*(n11 + n44)*log(pmax(1e-6,1 - r)) + 3*(n14 + n41)*log(pmax(1e-6,r)) + (n12 + n21 + n34 + n43)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n13 + n24 + n31 + n42)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n23 + n32)*(log(pmax(1e-6,r)) + log(7 - 12*r + 9*r^2)) + (n22 + n33)*log(4 - 10*r + 15*r^2 - 9*r^3)
return(L)}
interlogL_cc <- function(n12,n13,n21,n24,n31,n34,n42,n43,n11,n14,n22,n23,n32,n33,n41,n44) {
optimize(logL_cc,c(0,0.5), n12,n13,n21,n24,n31,n34,n42,n43,n11,n14,n22,n23,n32,n33,n41,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_12"],x[,"n_13"],x[,"n_21"],x[,"n_24"],x[,"n_31"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_11"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_32"],x[,"n_33"],x[,"n_41"],x[,"n_44"])
LOD_cc <- 3*(x[,"n_14"] + x[,"n_41"])*log10(2) + 3*(x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(2) + 3*(x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(2) + 3*(x[,"n_11"] + x[,"n_44"])*log10(2) - (x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_33"])*(-3*log10(2) + log10(13)) + 3*(x[,"n_11"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_14"] + x[,"n_41"])*log10(pmax(1e-6,r_cc)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_23"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(7 - 12*r_cc + 9*r_cc^2)) + (x[,"n_22"] + x[,"n_33"])*log10(4 - 10*r_cc + 15*r_cc^2 - 9*r_cc^3)
logL_cc <- (x[,"n_12"] + x[,"n_13"] + x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_42"] + x[,"n_43"])*log(3) + (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(log(2) + log(5)) + 3*(x[,"n_11"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_14"] + x[,"n_41"])*log(pmax(1e-6,r_cc)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_23"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(7 - 12*r_cc + 9*r_cc^2)) + (x[,"n_22"] + x[,"n_33"])*log(4 - 10*r_cc + 15*r_cc^2 - 9*r_cc^3)
logL_cm3c <- function(r,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(3*log(2) + log(5)) + (n14 + n41)*(log(2 - r) + 2*log(pmax(1e-6,r))) + (n11 + n44)*(log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n13 + n24 + n31 + n42)*(log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n23 + n32)*log(6 - 4*r + 14*r^2 - 9*r^3) + (n12 + n21 + n34 + n43)*log(3 - 4*r + 4*r^2 - 3*r^3) + (n22 + n33)*log(7 + 3*r - 13*r^2 + 9*r^3)
return(L)}
interlogL_cm3c <- function(n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_cm3c,c(0,0.5), n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm3c <- mapply(interlogL_cm3c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cm3c <- -((x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(3))) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-3*log10(2) + log10(13)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-3*log10(2) + log10(13)) - (x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(3) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-3*log10(2) + log10(3) + log10(17)) + (x[,"n_14"] + x[,"n_41"])*(log10(2 - r_cm3c) + 2*log10(pmax(1e-6,r_cm3c))) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cm3c)) + log10(1 - r_cm3c^2)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm3c)) + log10(5 - 5*r_cm3c + 3*r_cm3c^2)) + (x[,"n_23"] + x[,"n_32"])*log10(6 - 4*r_cm3c + 14*r_cm3c^2 - 9*r_cm3c^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(3 - 4*r_cm3c + 4*r_cm3c^2 - 3*r_cm3c^3) + (x[,"n_22"] + x[,"n_33"])*log10(7 + 3*r_cm3c - 13*r_cm3c^2 + 9*r_cm3c^3)
logL_cm3c <- (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(3*log(2) + log(5)) + (x[,"n_14"] + x[,"n_41"])*(log(2 - r_cm3c) + 2*log(pmax(1e-6,r_cm3c))) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cm3c)) + log(1 - r_cm3c^2)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log(pmax(1e-6,r_cm3c)) + log(5 - 5*r_cm3c + 3*r_cm3c^2)) + (x[,"n_23"] + x[,"n_32"])*log(6 - 4*r_cm3c + 14*r_cm3c^2 - 9*r_cm3c^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(3 - 4*r_cm3c + 4*r_cm3c^2 - 3*r_cm3c^3) + (x[,"n_22"] + x[,"n_33"])*log(7 + 3*r_cm3c - 13*r_cm3c^2 + 9*r_cm3c^3)
logL_cm2c <- function(r,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(2*log(2) + log(3) + log(5)) + (n11 + n44)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n14 + n41)*(log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n22 + n33)*log(12 - 8*r + 11*r^2 - 9*r^3) + (n13 + n24 + n31 + n42)*log(3 - 4*r + 7*r^2 - 3*r^3) + (n12 + n21 + n34 + n43)*log(3 - r - 2*r^2 + 3*r^3) + (n23 + n32)*log(6 + 13*r - 16*r^2 + 9*r^3)
return(L)}
interlogL_cm2c <- function(n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_cm2c,c(0,0.5), n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cm2c <- -((x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(5))) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(7) + log10(11)) - (x[,"n_22"] + x[,"n_33"])*(-3*log10(2) + log10(7) + log10(11)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-3*log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-3*log10(2) + log10(19)) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c)) + log10(2 + r_cm2c)) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm2c)) + log10(3 - 4*r_cm2c + r_cm2c^2)) + (x[,"n_22"] + x[,"n_33"])*log10(12 - 8*r_cm2c + 11*r_cm2c^2 - 9*r_cm2c^3) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(3 - 4*r_cm2c + 7*r_cm2c^2 - 3*r_cm2c^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(3 - r_cm2c - 2*r_cm2c^2 + 3*r_cm2c^3) + (x[,"n_23"] + x[,"n_32"])*log10(6 + 13*r_cm2c - 16*r_cm2c^2 + 9*r_cm2c^3)
logL_cm2c <- (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(2*log(2) + log(3) + log(5)) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c)) + log(2 + r_cm2c)) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,r_cm2c)) + log(3 - 4*r_cm2c + r_cm2c^2)) + (x[,"n_22"] + x[,"n_33"])*log(12 - 8*r_cm2c + 11*r_cm2c^2 - 9*r_cm2c^3) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(3 - 4*r_cm2c + 7*r_cm2c^2 - 3*r_cm2c^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(3 - r_cm2c - 2*r_cm2c^2 + 3*r_cm2c^3) + (x[,"n_23"] + x[,"n_32"])*log(6 + 13*r_cm2c - 16*r_cm2c^2 + 9*r_cm2c^3)
logL_rc <- function(r,n12,n13,n21,n24,n31,n34,n42,n43,n11,n14,n22,n23,n32,n33,n41,n44) {
L <- (n12 + n13 + n21 + n24 + n31 + n34 + n42 + n43)*log(3) + (-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(log(2) + 2*log(5)) + (n14 + n41)*(log(3 - r) + 2*log(pmax(1e-6,r))) + (n11 + n44)*(2*log(pmax(1e-6,1 - r)) + log(2 + r)) + (n13 + n24 + n31 + n42)*(log(pmax(1e-6,r)) + log(2 - 2*r + r^2)) + (n23 + n32)*log(9 - 7*r + 15*r^2 - 9*r^3) + (n12 + n21 + n34 + n43)*log(1 - r + r^2 - r^3) + (n22 + n33)*log(8 + 4*r - 12*r^2 + 9*r^3)
return(L)}
interlogL_rc <- function(n12,n13,n21,n24,n31,n34,n42,n43,n11,n14,n22,n23,n32,n33,n41,n44) {
optimize(logL_rc,c(0,0.5), n12,n13,n21,n24,n31,n34,n42,n43,n11,n14,n22,n23,n32,n33,n41,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_12"],x[,"n_13"],x[,"n_21"],x[,"n_24"],x[,"n_31"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_11"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_32"],x[,"n_33"],x[,"n_41"],x[,"n_44"])
LOD_rc <- (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(3*log10(2) - log10(5)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-3*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(5) + log10(13)) - (x[,"n_22"] + x[,"n_33"])*(-3*log10(2) + log10(5) + log10(13)) + (x[,"n_14"] + x[,"n_41"])*(log10(3 - r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_11"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(2 + r_rc)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 2*r_rc + r_rc^2)) + (x[,"n_23"] + x[,"n_32"])*log10(9 - 7*r_rc + 15*r_rc^2 - 9*r_rc^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(1 - r_rc + r_rc^2 - r_rc^3) + (x[,"n_22"] + x[,"n_33"])*log10(8 + 4*r_rc - 12*r_rc^2 + 9*r_rc^3)
logL_rc <- (x[,"n_12"] + x[,"n_13"] + x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_42"] + x[,"n_43"])*log(3) + (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(log(2) + 2*log(5)) + (x[,"n_14"] + x[,"n_41"])*(log(3 - r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_11"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_rc)) + log(2 + r_rc)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log(pmax(1e-6,r_rc)) + log(2 - 2*r_rc + r_rc^2)) + (x[,"n_23"] + x[,"n_32"])*log(9 - 7*r_rc + 15*r_rc^2 - 9*r_rc^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(1 - r_rc + r_rc^2 - r_rc^3) + (x[,"n_22"] + x[,"n_33"])*log(8 + 4*r_rc - 12*r_rc^2 + 9*r_rc^3)
logL_rm3c <- function(r,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(3*log(2) + 2*log(5)) + (n14 + n41)*(log(2 - r) + log(3 - r) + log(pmax(1e-6,r))) + (n11 + n44)*(log(2 + r) + log(1 - r^2)) + (n22 + n33)*log(32 - 3*r + 10*r^2 - 9*r^3) + (n13 + n24 + n31 + n42)*log(9 - 5*r + 8*r^2 - 3*r^3) + (n12 + n21 + n34 + n43)*log(9 - 2*r - r^2 + 3*r^3) + (n23 + n32)*log(30 + 10*r - 17*r^2 + 9*r^3)
return(L)}
interlogL_rm3c <- function(n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_rm3c,c(0,0.5), n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm3c <- mapply(interlogL_rm3c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm3c <- -((x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(3) + log10(5))) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-3*log10(2) + log10(5) + log10(13)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-3*log10(2) + log10(5) + log10(13)) - (x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(3) + log10(5) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-3*log10(2) + log10(3) + log10(5) + log10(17)) + (x[,"n_14"] + x[,"n_41"])*(log10(2 - r_rm3c) + log10(3 - r_rm3c) + log10(pmax(1e-6,r_rm3c))) + (x[,"n_11"] + x[,"n_44"])*(log10(2 + r_rm3c) + log10(1 - r_rm3c^2)) + (x[,"n_22"] + x[,"n_33"])*log10(32 - 3*r_rm3c + 10*r_rm3c^2 - 9*r_rm3c^3) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(9 - 5*r_rm3c + 8*r_rm3c^2 - 3*r_rm3c^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(9 - 2*r_rm3c - r_rm3c^2 + 3*r_rm3c^3) + (x[,"n_23"] + x[,"n_32"])*log10(30 + 10*r_rm3c - 17*r_rm3c^2 + 9*r_rm3c^3)
logL_rm3c <- (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(3*log(2) + 2*log(5)) + (x[,"n_14"] + x[,"n_41"])*(log(2 - r_rm3c) + log(3 - r_rm3c) + log(pmax(1e-6,r_rm3c))) + (x[,"n_11"] + x[,"n_44"])*(log(2 + r_rm3c) + log(1 - r_rm3c^2)) + (x[,"n_22"] + x[,"n_33"])*log(32 - 3*r_rm3c + 10*r_rm3c^2 - 9*r_rm3c^3) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(9 - 5*r_rm3c + 8*r_rm3c^2 - 3*r_rm3c^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(9 - 2*r_rm3c - r_rm3c^2 + 3*r_rm3c^3) + (x[,"n_23"] + x[,"n_32"])*log(30 + 10*r_rm3c - 17*r_rm3c^2 + 9*r_rm3c^3)
logL_rm2c <- function(r,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(2*log(2) + log(3) + 2*log(5)) + (n11 + n44)*(log(pmax(1e-6,r)) + 2*log(2 + r)) + (n14 + n41)*(log(3 - r) + log(3 - 4*r + r^2)) + (n23 + n32)*log(57 - 25*r + 19*r^2 - 9*r^3) + (n12 + n21 + n34 + n43)*log(6 + 13*r - r^2 - 3*r^3) + (n13 + n24 + n31 + n42)*log(15 - 2*r - 10*r^2 + 3*r^3) + (n22 + n33)*log(42 + 14*r - 8*r^2 + 9*r^3)
return(L)}
interlogL_rm2c <- function(n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_rm2c,c(0,0.5), n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm2c <- -((x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + 2*log10(5))) - (x[,"n_11"] + x[,"n_44"])*(-log10(2) + 2*(-log10(2) + log10(5))) - (x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(5) + log10(7) + log10(11)) - (x[,"n_22"] + x[,"n_33"])*(-3*log10(2) + log10(5) + log10(7) + log10(11)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-3*log10(2) + log10(5) + log10(19)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-3*log10(2) + log10(5) + log10(19)) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm2c)) + 2*log10(2 + r_rm2c)) + (x[,"n_14"] + x[,"n_41"])*(log10(3 - r_rm2c) + log10(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_23"] + x[,"n_32"])*log10(57 - 25*r_rm2c + 19*r_rm2c^2 - 9*r_rm2c^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(6 + 13*r_rm2c - r_rm2c^2 - 3*r_rm2c^3) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(15 - 2*r_rm2c - 10*r_rm2c^2 + 3*r_rm2c^3) + (x[,"n_22"] + x[,"n_33"])*log10(42 + 14*r_rm2c - 8*r_rm2c^2 + 9*r_rm2c^3)
logL_rm2c <- (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(2*log(2) + log(3) + 2*log(5)) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_rm2c)) + 2*log(2 + r_rm2c)) + (x[,"n_14"] + x[,"n_41"])*(log(3 - r_rm2c) + log(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_23"] + x[,"n_32"])*log(57 - 25*r_rm2c + 19*r_rm2c^2 - 9*r_rm2c^3) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(6 + 13*r_rm2c - r_rm2c^2 - 3*r_rm2c^3) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(15 - 2*r_rm2c - 10*r_rm2c^2 + 3*r_rm2c^3) + (x[,"n_22"] + x[,"n_33"])*log(42 + 14*r_rm2c - 8*r_rm2c^2 + 9*r_rm2c^3)

return(list(
r_mat = cbind(r_cc,r_cm3c,r_cm2c,r_rc,r_rm3c,r_rm2c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm3c,LOD_cm2c,LOD_rc,LOD_rm3c,LOD_rm2c,0),
logL_mat = cbind(logL_cc,logL_cm3c,logL_cm2c,logL_rc,logL_rm3c,logL_rm2c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed3c","coupling mixed2c","repulsion coupling","repulsion mixed3c","repulsion mixed2c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.4_1.5 <- function(x){
logL_cc <- function(r,n12,n14,n22,n23,n24,n32,n33,n34,n42,n44,n13,n43) {
L <- (-n12 - n14 - 2*n22 - 2*n23 - 2*n24 - 2*n32 - 2*n33 - 2*n34 - n42 - n44)*log(2) + (-n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*log(5) + 2*(n12 + n44)*log(pmax(1e-6,1 - r)) + 2*(n14 + n42)*log(pmax(1e-6,r)) + (n24 + n32)*(log(5 - 2*r) + log(pmax(1e-6,r))) + (n13 + n43)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n22 + n34)*log(3 - r - 2*r^2) + (n23 + n33)*log(5 - 4*r + 4*r^2)
return(L)}
interlogL_cc <- function(n12,n14,n22,n23,n24,n32,n33,n34,n42,n44,n13,n43) {
optimize(logL_cc,c(0,0.5), n12,n14,n22,n23,n24,n32,n33,n34,n42,n44,n13,n43, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_12"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_44"],x[,"n_13"],x[,"n_43"])
LOD_cc <- -((x[,"n_24"] + x[,"n_32"])*log10(2)) - 2*(x[,"n_23"] + x[,"n_33"])*log10(2) - (x[,"n_22"] + x[,"n_34"])*log10(2) + 2*(x[,"n_14"] + x[,"n_42"])*log10(2) + 2*(x[,"n_13"] + x[,"n_43"])*log10(2) + 2*(x[,"n_12"] + x[,"n_44"])*log10(2) + 2*(x[,"n_12"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_14"] + x[,"n_42"])*log10(pmax(1e-6,r_cc)) + (x[,"n_24"] + x[,"n_32"])*(log10(5 - 2*r_cc) + log10(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_43"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_22"] + x[,"n_34"])*log10(3 - r_cc - 2*r_cc^2) + (x[,"n_23"] + x[,"n_33"])*log10(5 - 4*r_cc + 4*r_cc^2)
logL_cc <- (-x[,"n_12"] - x[,"n_14"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_42"] - x[,"n_44"])*log(2) + (-x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + 2*(x[,"n_12"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_14"] + x[,"n_42"])*log(pmax(1e-6,r_cc)) + (x[,"n_24"] + x[,"n_32"])*(log(5 - 2*r_cc) + log(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_43"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_22"] + x[,"n_34"])*log(3 - r_cc - 2*r_cc^2) + (x[,"n_23"] + x[,"n_33"])*log(5 - 4*r_cc + 4*r_cc^2)
logL_cm3c <- function(r,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
L <- (-n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*(3*log(2) + log(5)) + (n14 + n42)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n12 + n44)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n23 + n33)*log(7 + 4*r - 4*r^2) + (n22 + n34)*log(6 - 5*r + 2*r^2) + (n24 + n32)*log(3 + r + 2*r^2) + (n13 + n43)*log(3 - 4*r + 4*r^2)
return(L)}
interlogL_cm3c <- function(n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
optimize(logL_cm3c,c(0,0.5), n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm3c <- mapply(interlogL_cm3c,x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cm3c <- -2*(x[,"n_24"] + x[,"n_32"])*log10(2) - 3*(x[,"n_23"] + x[,"n_33"])*log10(2) - 2*(x[,"n_22"] + x[,"n_34"])*log10(2) - (x[,"n_13"] + x[,"n_43"])*log10(2) + (x[,"n_14"] + x[,"n_42"])*(log10(3 - 2*r_cm3c) + log10(pmax(1e-6,r_cm3c))) + (x[,"n_12"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cm3c)) + log10(1 + 2*r_cm3c)) + (x[,"n_23"] + x[,"n_33"])*log10(7 + 4*r_cm3c - 4*r_cm3c^2) + (x[,"n_22"] + x[,"n_34"])*log10(6 - 5*r_cm3c + 2*r_cm3c^2) + (x[,"n_24"] + x[,"n_32"])*log10(3 + r_cm3c + 2*r_cm3c^2) + (x[,"n_13"] + x[,"n_43"])*log10(3 - 4*r_cm3c + 4*r_cm3c^2)
logL_cm3c <- (-x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(3*log(2) + log(5)) + (x[,"n_14"] + x[,"n_42"])*(log(3 - 2*r_cm3c) + log(pmax(1e-6,r_cm3c))) + (x[,"n_12"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cm3c)) + log(1 + 2*r_cm3c)) + (x[,"n_23"] + x[,"n_33"])*log(7 + 4*r_cm3c - 4*r_cm3c^2) + (x[,"n_22"] + x[,"n_34"])*log(6 - 5*r_cm3c + 2*r_cm3c^2) + (x[,"n_24"] + x[,"n_32"])*log(3 + r_cm3c + 2*r_cm3c^2) + (x[,"n_13"] + x[,"n_43"])*log(3 - 4*r_cm3c + 4*r_cm3c^2)
logL_rc <- function(r,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
L <- (-n12 - n13 - n14 - 2*n22 - 2*n23 - 2*n24 - 2*n32 - 2*n33 - 2*n34 - n42 - n43 - n44)*log(2) + 2*(-n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*log(5) + (n14 + n42)*(log(3 - r) + log(pmax(1e-6,r))) + (n12 + n44)*(log(pmax(1e-6,1 - r)) + log(2 + r)) + (n23 + n33)*log(19 + 4*r - 4*r^2) + (n22 + n34)*log(12 - 5*r + 2*r^2) + (n13 + n43)*log(3 - 2*r + 2*r^2) + (n24 + n32)*log(9 + r + 2*r^2)
return(L)}
interlogL_rc <- function(n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
optimize(logL_rc,c(0,0.5), n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rc <- -((x[,"n_14"] + x[,"n_42"])*(-2*log10(2) + log10(5))) - (x[,"n_12"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_43"])*(-log10(2) + log10(5)) - (x[,"n_24"] + x[,"n_32"])*(log10(2) + log10(5)) - (x[,"n_22"] + x[,"n_34"])*(log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_33"])*(2*log10(2) + log10(5)) + (x[,"n_14"] + x[,"n_42"])*(log10(3 - r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_12"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_rc)) + log10(2 + r_rc)) + (x[,"n_23"] + x[,"n_33"])*log10(19 + 4*r_rc - 4*r_rc^2) + (x[,"n_22"] + x[,"n_34"])*log10(12 - 5*r_rc + 2*r_rc^2) + (x[,"n_13"] + x[,"n_43"])*log10(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_24"] + x[,"n_32"])*log10(9 + r_rc + 2*r_rc^2)
logL_rc <- (-x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(2) + 2*(-x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_14"] + x[,"n_42"])*(log(3 - r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_12"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_rc)) + log(2 + r_rc)) + (x[,"n_23"] + x[,"n_33"])*log(19 + 4*r_rc - 4*r_rc^2) + (x[,"n_22"] + x[,"n_34"])*log(12 - 5*r_rc + 2*r_rc^2) + (x[,"n_13"] + x[,"n_43"])*log(3 - 2*r_rc + 2*r_rc^2) + (x[,"n_24"] + x[,"n_32"])*log(9 + r_rc + 2*r_rc^2)
logL_rm3c <- function(r,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
L <- (-n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*(3*log(2) + 2*log(5)) + (n14 + n42)*(log(3 - 2*r) + log(3 - r)) + (n12 + n44)*(log(2 + r) + log(1 + 2*r)) + (n13 + n43)*log(9 + 4*r - 4*r^2) + (n24 + n32)*log(24 - 7*r - 2*r^2) + (n22 + n34)*log(15 + 11*r - 2*r^2) + (n23 + n33)*log(41 - 4*r + 4*r^2)
return(L)}
interlogL_rm3c <- function(n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
optimize(logL_rm3c,c(0,0.5), n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm3c <- mapply(interlogL_rm3c,x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm3c <- -((x[,"n_14"] + x[,"n_42"])*log10(5)) - (x[,"n_12"] + x[,"n_44"])*log10(5) - (x[,"n_13"] + x[,"n_43"])*(log10(2) + log10(5)) - (x[,"n_24"] + x[,"n_32"])*(2*log10(2) + log10(5)) - (x[,"n_22"] + x[,"n_34"])*(2*log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_33"])*(3*log10(2) + log10(5)) + (x[,"n_14"] + x[,"n_42"])*(log10(3 - 2*r_rm3c) + log10(3 - r_rm3c)) + (x[,"n_12"] + x[,"n_44"])*(log10(2 + r_rm3c) + log10(1 + 2*r_rm3c)) + (x[,"n_13"] + x[,"n_43"])*log10(9 + 4*r_rm3c - 4*r_rm3c^2) + (x[,"n_24"] + x[,"n_32"])*log10(24 - 7*r_rm3c - 2*r_rm3c^2) + (x[,"n_22"] + x[,"n_34"])*log10(15 + 11*r_rm3c - 2*r_rm3c^2) + (x[,"n_23"] + x[,"n_33"])*log10(41 - 4*r_rm3c + 4*r_rm3c^2)
logL_rm3c <- (-x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(3*log(2) + 2*log(5)) + (x[,"n_14"] + x[,"n_42"])*(log(3 - 2*r_rm3c) + log(3 - r_rm3c)) + (x[,"n_12"] + x[,"n_44"])*(log(2 + r_rm3c) + log(1 + 2*r_rm3c)) + (x[,"n_13"] + x[,"n_43"])*log(9 + 4*r_rm3c - 4*r_rm3c^2) + (x[,"n_24"] + x[,"n_32"])*log(24 - 7*r_rm3c - 2*r_rm3c^2) + (x[,"n_22"] + x[,"n_34"])*log(15 + 11*r_rm3c - 2*r_rm3c^2) + (x[,"n_23"] + x[,"n_33"])*log(41 - 4*r_rm3c + 4*r_rm3c^2)

return(list(
r_mat = cbind(r_cc,r_cm3c,r_rc,r_rm3c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm3c,LOD_rc,LOD_rm3c,0),
logL_mat = cbind(logL_cc,logL_cm3c,logL_rc,logL_rm3c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed3c","repulsion coupling","repulsion mixed3c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.4_2.2 <- function(x){
logL_cc <- function(r,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-n10 - 2*n11 - 2*n12 - 2*n13 - n14 - n20 - 2*n21 - 2*n22 - 2*n23 - n24 - n30 - 2*n31 - 2*n32 - 2*n33 - n34 - n40 - 2*n41 - 2*n42 - 2*n43 - n44)*log(2) + (-n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*(log(3) + 2*log(5)) + (n14 + n40)*(2*log(pmax(1e-6,r)) + log(2 + r)) + (n12 + n42)*log(9 + 16*r - 16*r^2) + (n13 + n41)*(log(pmax(1e-6,r)) + log(12 + 7*r - 4*r^2)) + (n24 + n30)*(log(pmax(1e-6,r)) + log(5 + r - 3*r^2)) + (n10 + n44)*(log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n22 + n32)*log(57 - 16*r + 16*r^2) + (n21 + n33)*log(42 - 32*r + 17*r^2 - 12*r^3) + (n20 + n34)*log(3 + 2*r - 8*r^2 + 3*r^3) + (n11 + n43)*log(15 - 14*r - 5*r^2 + 4*r^3) + (n23 + n31)*log(15 + 34*r - 19*r^2 + 12*r^3)
return(L)}
interlogL_cc <- function(n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_cc,c(0,0.5), n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cc <- -((x[,"n_14"] + x[,"n_40"])*(-3*log10(2) + log10(5))) - (x[,"n_10"] + x[,"n_44"])*(-3*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_42"])*log10(13) - (x[,"n_24"] + x[,"n_30"])*(-3*log10(2) + log10(19)) - (x[,"n_20"] + x[,"n_34"])*(-3*log10(2) + log10(19)) - (x[,"n_23"] + x[,"n_31"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_21"] + x[,"n_33"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_13"] + x[,"n_41"])*(-2*log10(2) + log10(29)) - (x[,"n_11"] + x[,"n_43"])*(-2*log10(2) + log10(29)) - (x[,"n_22"] + x[,"n_32"])*log10(53) + (x[,"n_14"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_12"] + x[,"n_42"])*log10(9 + 16*r_cc - 16*r_cc^2) + (x[,"n_13"] + x[,"n_41"])*(log10(pmax(1e-6,r_cc)) + log10(12 + 7*r_cc - 4*r_cc^2)) + (x[,"n_24"] + x[,"n_30"])*(log10(pmax(1e-6,r_cc)) + log10(5 + r_cc - 3*r_cc^2)) + (x[,"n_10"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cc)) + log10(3 - 4*r_cc + r_cc^2)) + (x[,"n_22"] + x[,"n_32"])*log10(57 - 16*r_cc + 16*r_cc^2) + (x[,"n_21"] + x[,"n_33"])*log10(42 - 32*r_cc + 17*r_cc^2 - 12*r_cc^3) + (x[,"n_20"] + x[,"n_34"])*log10(3 + 2*r_cc - 8*r_cc^2 + 3*r_cc^3) + (x[,"n_11"] + x[,"n_43"])*log10(15 - 14*r_cc - 5*r_cc^2 + 4*r_cc^3) + (x[,"n_23"] + x[,"n_31"])*log10(15 + 34*r_cc - 19*r_cc^2 + 12*r_cc^3)
logL_cc <- (-x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - x[,"n_44"])*log(2) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(log(3) + 2*log(5)) + (x[,"n_14"] + x[,"n_40"])*(2*log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_12"] + x[,"n_42"])*log(9 + 16*r_cc - 16*r_cc^2) + (x[,"n_13"] + x[,"n_41"])*(log(pmax(1e-6,r_cc)) + log(12 + 7*r_cc - 4*r_cc^2)) + (x[,"n_24"] + x[,"n_30"])*(log(pmax(1e-6,r_cc)) + log(5 + r_cc - 3*r_cc^2)) + (x[,"n_10"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cc)) + log(3 - 4*r_cc + r_cc^2)) + (x[,"n_22"] + x[,"n_32"])*log(57 - 16*r_cc + 16*r_cc^2) + (x[,"n_21"] + x[,"n_33"])*log(42 - 32*r_cc + 17*r_cc^2 - 12*r_cc^3) + (x[,"n_20"] + x[,"n_34"])*log(3 + 2*r_cc - 8*r_cc^2 + 3*r_cc^3) + (x[,"n_11"] + x[,"n_43"])*log(15 - 14*r_cc - 5*r_cc^2 + 4*r_cc^3) + (x[,"n_23"] + x[,"n_31"])*log(15 + 34*r_cc - 19*r_cc^2 + 12*r_cc^3)
logL_cm <- function(r,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-2*n10 - 3*n11 - 3*n12 - 3*n13 - 2*n14 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 2*n24 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 2*n34 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 2*n44)*log(2) + 2*(-n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n10 + n44)*(log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n22 + n32)*log(33 + 8*r - 8*r^2) + (n14 + n40)*(log(pmax(1e-6,r)) + log(1 - r^2)) + (n12 + n42)*log(11 - 8*r + 8*r^2) + (n23 + n31)*log(18 - 4*r + 19*r^2 - 12*r^3) + (n11 + n43)*log(6 - 4*r + 5*r^2 - 4*r^3) + (n20 + n34)*log(3 - 5*r + 6*r^2 - 3*r^3) + (n24 + n30)*log(1 + 2*r - 3*r^2 + 3*r^3) + (n13 + n41)*log(3 + 6*r - 7*r^2 + 4*r^3) + (n21 + n33)*log(21 + 2*r - 17*r^2 + 12*r^3)
return(L)}
interlogL_cm <- function(n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_cm,c(0,0.5), n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cm <- -2*(x[,"n_12"] + x[,"n_42"])*log10(3) - (x[,"n_14"] + x[,"n_40"])*(-3*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_44"])*(-3*log10(2) + log10(3)) - (x[,"n_22"] + x[,"n_32"])*(log10(5) + log10(7)) - (x[,"n_23"] + x[,"n_31"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_21"] + x[,"n_33"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_24"] + x[,"n_30"])*(-3*log10(2) + log10(13)) - (x[,"n_20"] + x[,"n_34"])*(-3*log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_41"])*(-2*log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_43"])*(-2*log10(2) + log10(19)) + (x[,"n_10"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cm)) + log10(2 - r_cm) + log10(pmax(1e-6,r_cm))) + (x[,"n_22"] + x[,"n_32"])*log10(33 + 8*r_cm - 8*r_cm^2) + (x[,"n_14"] + x[,"n_40"])*(log10(pmax(1e-6,r_cm)) + log10(1 - r_cm^2)) + (x[,"n_12"] + x[,"n_42"])*log10(11 - 8*r_cm + 8*r_cm^2) + (x[,"n_23"] + x[,"n_31"])*log10(18 - 4*r_cm + 19*r_cm^2 - 12*r_cm^3) + (x[,"n_11"] + x[,"n_43"])*log10(6 - 4*r_cm + 5*r_cm^2 - 4*r_cm^3) + (x[,"n_20"] + x[,"n_34"])*log10(3 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3) + (x[,"n_24"] + x[,"n_30"])*log10(1 + 2*r_cm - 3*r_cm^2 + 3*r_cm^3) + (x[,"n_13"] + x[,"n_41"])*log10(3 + 6*r_cm - 7*r_cm^2 + 4*r_cm^3) + (x[,"n_21"] + x[,"n_33"])*log10(21 + 2*r_cm - 17*r_cm^2 + 12*r_cm^3)
logL_cm <- (-2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + 2*(-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_10"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cm)) + log(2 - r_cm) + log(pmax(1e-6,r_cm))) + (x[,"n_22"] + x[,"n_32"])*log(33 + 8*r_cm - 8*r_cm^2) + (x[,"n_14"] + x[,"n_40"])*(log(pmax(1e-6,r_cm)) + log(1 - r_cm^2)) + (x[,"n_12"] + x[,"n_42"])*log(11 - 8*r_cm + 8*r_cm^2) + (x[,"n_23"] + x[,"n_31"])*log(18 - 4*r_cm + 19*r_cm^2 - 12*r_cm^3) + (x[,"n_11"] + x[,"n_43"])*log(6 - 4*r_cm + 5*r_cm^2 - 4*r_cm^3) + (x[,"n_20"] + x[,"n_34"])*log(3 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3) + (x[,"n_24"] + x[,"n_30"])*log(1 + 2*r_cm - 3*r_cm^2 + 3*r_cm^3) + (x[,"n_13"] + x[,"n_41"])*log(3 + 6*r_cm - 7*r_cm^2 + 4*r_cm^3) + (x[,"n_21"] + x[,"n_33"])*log(21 + 2*r_cm - 17*r_cm^2 + 12*r_cm^3)
logL_cr <- function(r,n11,n13,n21,n22,n23,n31,n32,n33,n41,n43,n10,n12,n14,n20,n24,n30,n34,n40,n42,n44) {
L <- (-n11 - n13 - n21 - n22 - n23 - n31 - n32 - n33 - n41 - n43)*log(2) + 2*(n22 + n32)*log(3) + 2*(-n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n14 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n44)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n20 + n34)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n11 + n43)*(log(pmax(1e-6,r)) + log(4 - 5*r + 4*r^2)) + (n21 + n33)*log(6 - 8*r + 17*r^2 - 12*r^3) + (n13 + n41)*log(3 - 6*r + 7*r^2 - 4*r^3) + (n24 + n30)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n23 + n31)*log(3 + 10*r - 19*r^2 + 12*r^3)
return(L)}
interlogL_cr <- function(n11,n13,n21,n22,n23,n31,n32,n33,n41,n43,n10,n12,n14,n20,n24,n30,n34,n40,n42,n44) {
optimize(logL_cr,c(0,0.5), n11,n13,n21,n22,n23,n31,n32,n33,n41,n43,n10,n12,n14,n20,n24,n30,n34,n40,n42,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_11"],x[,"n_13"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_41"],x[,"n_43"],x[,"n_10"],x[,"n_12"],x[,"n_14"],x[,"n_20"],x[,"n_24"],x[,"n_30"],x[,"n_34"],x[,"n_40"],x[,"n_42"],x[,"n_44"])
LOD_cr <- 3*(x[,"n_14"] + x[,"n_40"])*log10(2) + 3*(x[,"n_10"] + x[,"n_44"])*log10(2) + (x[,"n_24"] + x[,"n_30"])*(3*log10(2) - log10(3)) - (x[,"n_20"] + x[,"n_34"])*(-3*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_31"])*(-2*log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_33"])*(-2*log10(2) + log10(19)) + (x[,"n_14"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_10"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_20"] + x[,"n_34"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_11"] + x[,"n_43"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_21"] + x[,"n_33"])*log10(6 - 8*r_cr + 17*r_cr^2 - 12*r_cr^3) + (x[,"n_13"] + x[,"n_41"])*log10(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_24"] + x[,"n_30"])*log10(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_23"] + x[,"n_31"])*log10(3 + 10*r_cr - 19*r_cr^2 + 12*r_cr^3)
logL_cr <- (-x[,"n_11"] - x[,"n_13"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_41"] - x[,"n_43"])*log(2) + 2*(x[,"n_22"] + x[,"n_32"])*log(3) + 2*(-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_14"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_10"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_20"] + x[,"n_34"])*(log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_11"] + x[,"n_43"])*(log(pmax(1e-6,r_cr)) + log(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_21"] + x[,"n_33"])*log(6 - 8*r_cr + 17*r_cr^2 - 12*r_cr^3) + (x[,"n_13"] + x[,"n_41"])*log(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_24"] + x[,"n_30"])*log(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_23"] + x[,"n_31"])*log(3 + 10*r_cr - 19*r_cr^2 + 12*r_cr^3)
logL_rc <- function(r,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*(3*log(2) + log(3) + 2*log(5)) + (n14 + n40)*(log(3 - 2*r) + log(pmax(1e-6,r)) + log(2 + r)) + (n22 + n32)*log(105 + 4*r - 4*r^2) + (n10 + n44)*(log(1 + 2*r) + log(3 - 4*r + r^2)) + (n12 + n42)*log(27 - 4*r + 4*r^2) + (n23 + n31)*log(57 - 4*r + 16*r^2 - 12*r^3) + (n20 + n34)*log(12 - 10*r + 13*r^2 - 6*r^3) + (n11 + n43)*log(21 - 16*r + 8*r^2 - 4*r^3) + (n13 + n41)*log(9 + 12*r - 4*r^2 + 4*r^3) + (n24 + n30)*log(9 + 2*r - 5*r^2 + 6*r^3) + (n21 + n33)*log(57 + 8*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rc <- function(n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rc,c(0,0.5), n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rc <- -((x[,"n_14"] + x[,"n_40"])*(-log10(2) + log10(5))) - (x[,"n_10"] + x[,"n_44"])*(-log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_42"])*(log10(2) + log10(13)) - (x[,"n_24"] + x[,"n_30"])*(-log10(2) + log10(19)) - (x[,"n_20"] + x[,"n_34"])*(-log10(2) + log10(19)) - (x[,"n_23"] + x[,"n_31"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_21"] + x[,"n_33"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_13"] + x[,"n_41"])*(-log10(2) + log10(29)) - (x[,"n_11"] + x[,"n_43"])*(-log10(2) + log10(29)) - (x[,"n_22"] + x[,"n_32"])*(log10(2) + log10(53)) + (x[,"n_14"] + x[,"n_40"])*(log10(3 - 2*r_rc) + log10(pmax(1e-6,r_rc)) + log10(2 + r_rc)) + (x[,"n_22"] + x[,"n_32"])*log10(105 + 4*r_rc - 4*r_rc^2) + (x[,"n_10"] + x[,"n_44"])*(log10(1 + 2*r_rc) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_12"] + x[,"n_42"])*log10(27 - 4*r_rc + 4*r_rc^2) + (x[,"n_23"] + x[,"n_31"])*log10(57 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_20"] + x[,"n_34"])*log10(12 - 10*r_rc + 13*r_rc^2 - 6*r_rc^3) + (x[,"n_11"] + x[,"n_43"])*log10(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_13"] + x[,"n_41"])*log10(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_24"] + x[,"n_30"])*log10(9 + 2*r_rc - 5*r_rc^2 + 6*r_rc^3) + (x[,"n_21"] + x[,"n_33"])*log10(57 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rc <- (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(3*log(2) + log(3) + 2*log(5)) + (x[,"n_14"] + x[,"n_40"])*(log(3 - 2*r_rc) + log(pmax(1e-6,r_rc)) + log(2 + r_rc)) + (x[,"n_22"] + x[,"n_32"])*log(105 + 4*r_rc - 4*r_rc^2) + (x[,"n_10"] + x[,"n_44"])*(log(1 + 2*r_rc) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_12"] + x[,"n_42"])*log(27 - 4*r_rc + 4*r_rc^2) + (x[,"n_23"] + x[,"n_31"])*log(57 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_20"] + x[,"n_34"])*log(12 - 10*r_rc + 13*r_rc^2 - 6*r_rc^3) + (x[,"n_11"] + x[,"n_43"])*log(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_13"] + x[,"n_41"])*log(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_24"] + x[,"n_30"])*log(9 + 2*r_rc - 5*r_rc^2 + 6*r_rc^3) + (x[,"n_21"] + x[,"n_33"])*log(57 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rm <- function(r,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*(4*log(2) + 2*log(5)) + (n10 + n44)*(log(2 - r) + log(pmax(1e-6,r)) + log(1 + 2*r)) + (n22 + n32)*log(69 + 4*r - 4*r^2) + (n14 + n40)*(log(3 - 2*r) + log(1 - r^2)) + (n12 + n42)*log(19 - 4*r + 4*r^2) + (n21 + n33)*log(33 + 4*r + 20*r^2 - 12*r^3) + (n24 + n30)*log(10 - 10*r + 9*r^2 - 6*r^3) + (n13 + n41)*log(15 - 12*r + 4*r^2 - 4*r^3) + (n11 + n43)*log(3 + 16*r - 8*r^2 + 4*r^3) + (n20 + n34)*log(3 + 10*r - 9*r^2 + 6*r^3) + (n23 + n31)*log(45 - 8*r - 16*r^2 + 12*r^3)
return(L)}
interlogL_rm <- function(n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rm,c(0,0.5), n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm <- -((x[,"n_14"] + x[,"n_40"])*(-log10(2) + log10(3))) - (x[,"n_10"] + x[,"n_44"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_42"])*(log10(2) + 2*log10(3)) - (x[,"n_22"] + x[,"n_32"])*(log10(2) + log10(5) + log10(7)) - (x[,"n_23"] + x[,"n_31"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_21"] + x[,"n_33"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_24"] + x[,"n_30"])*(-log10(2) + log10(13)) - (x[,"n_20"] + x[,"n_34"])*(-log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_41"])*(-log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_43"])*(-log10(2) + log10(19)) + (x[,"n_10"] + x[,"n_44"])*(log10(2 - r_rm) + log10(pmax(1e-6,r_rm)) + log10(1 + 2*r_rm)) + (x[,"n_22"] + x[,"n_32"])*log10(69 + 4*r_rm - 4*r_rm^2) + (x[,"n_14"] + x[,"n_40"])*(log10(3 - 2*r_rm) + log10(1 - r_rm^2)) + (x[,"n_12"] + x[,"n_42"])*log10(19 - 4*r_rm + 4*r_rm^2) + (x[,"n_21"] + x[,"n_33"])*log10(33 + 4*r_rm + 20*r_rm^2 - 12*r_rm^3) + (x[,"n_24"] + x[,"n_30"])*log10(10 - 10*r_rm + 9*r_rm^2 - 6*r_rm^3) + (x[,"n_13"] + x[,"n_41"])*log10(15 - 12*r_rm + 4*r_rm^2 - 4*r_rm^3) + (x[,"n_11"] + x[,"n_43"])*log10(3 + 16*r_rm - 8*r_rm^2 + 4*r_rm^3) + (x[,"n_20"] + x[,"n_34"])*log10(3 + 10*r_rm - 9*r_rm^2 + 6*r_rm^3) + (x[,"n_23"] + x[,"n_31"])*log10(45 - 8*r_rm - 16*r_rm^2 + 12*r_rm^3)
logL_rm <- (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(4*log(2) + 2*log(5)) + (x[,"n_10"] + x[,"n_44"])*(log(2 - r_rm) + log(pmax(1e-6,r_rm)) + log(1 + 2*r_rm)) + (x[,"n_22"] + x[,"n_32"])*log(69 + 4*r_rm - 4*r_rm^2) + (x[,"n_14"] + x[,"n_40"])*(log(3 - 2*r_rm) + log(1 - r_rm^2)) + (x[,"n_12"] + x[,"n_42"])*log(19 - 4*r_rm + 4*r_rm^2) + (x[,"n_21"] + x[,"n_33"])*log(33 + 4*r_rm + 20*r_rm^2 - 12*r_rm^3) + (x[,"n_24"] + x[,"n_30"])*log(10 - 10*r_rm + 9*r_rm^2 - 6*r_rm^3) + (x[,"n_13"] + x[,"n_41"])*log(15 - 12*r_rm + 4*r_rm^2 - 4*r_rm^3) + (x[,"n_11"] + x[,"n_43"])*log(3 + 16*r_rm - 8*r_rm^2 + 4*r_rm^3) + (x[,"n_20"] + x[,"n_34"])*log(3 + 10*r_rm - 9*r_rm^2 + 6*r_rm^3) + (x[,"n_23"] + x[,"n_31"])*log(45 - 8*r_rm - 16*r_rm^2 + 12*r_rm^3)
logL_rr <- function(r,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-2*n10 - n11 - 2*n12 - n13 - 2*n14 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n40 - n41 - 2*n42 - n43 - 2*n44)*log(2) + (n22 + n32)*log(3) + 2*(-n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n14 + n40)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r))) + (n10 + n44)*(2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n12 + n42)*log(1 + 12*r - 12*r^2) + (n20 + n34)*(log(pmax(1e-6,r)) + log(2 + 5*r - 6*r^2)) + (n22 + n32)*log(7 - 4*r + 4*r^2) + (n23 + n31)*log(15 - 16*r + 16*r^2 - 12*r^3) + (n11 + n43)*log(r + 4*r^2 - 2*r^3) + (n13 + n41)*log(3 - 3*r - 2*r^2 + 2*r^3) + (n24 + n30)*log(1 + 6*r - 13*r^2 + 6*r^3) + (n21 + n33)*log(3 + 20*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rr <- function(n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rr,c(0,0.5), n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rr <- (x[,"n_14"] + x[,"n_40"])*log10(2) - 2*(x[,"n_12"] + x[,"n_42"])*log10(2) + (x[,"n_10"] + x[,"n_44"])*log10(2) - (x[,"n_24"] + x[,"n_30"])*(-log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_34"])*(-log10(2) + log10(3)) - (x[,"n_22"] + x[,"n_32"])*(log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_41"])*(-2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_43"])*(-2*log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_31"])*(-log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_33"])*(-log10(2) + log10(19)) + (x[,"n_14"] + x[,"n_40"])*(log10(3 - 2*r_rr) + 2*log10(pmax(1e-6,1 - r_rr))) + (x[,"n_10"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_12"] + x[,"n_42"])*log10(1 + 12*r_rr - 12*r_rr^2) + (x[,"n_20"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(2 + 5*r_rr - 6*r_rr^2)) + (x[,"n_22"] + x[,"n_32"])*log10(7 - 4*r_rr + 4*r_rr^2) + (x[,"n_23"] + x[,"n_31"])*log10(15 - 16*r_rr + 16*r_rr^2 - 12*r_rr^3) + (x[,"n_11"] + x[,"n_43"])*log10(r_rr + 4*r_rr^2 - 2*r_rr^3) + (x[,"n_13"] + x[,"n_41"])*log10(3 - 3*r_rr - 2*r_rr^2 + 2*r_rr^3) + (x[,"n_24"] + x[,"n_30"])*log10(1 + 6*r_rr - 13*r_rr^2 + 6*r_rr^3) + (x[,"n_21"] + x[,"n_33"])*log10(3 + 20*r_rr - 20*r_rr^2 + 12*r_rr^3)
logL_rr <- (-2*x[,"n_10"] - x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_40"] - x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"])*log(2) + (x[,"n_22"] + x[,"n_32"])*log(3) + 2*(-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_14"] + x[,"n_40"])*(log(3 - 2*r_rr) + 2*log(pmax(1e-6,1 - r_rr))) + (x[,"n_10"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_12"] + x[,"n_42"])*log(1 + 12*r_rr - 12*r_rr^2) + (x[,"n_20"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(2 + 5*r_rr - 6*r_rr^2)) + (x[,"n_22"] + x[,"n_32"])*log(7 - 4*r_rr + 4*r_rr^2) + (x[,"n_23"] + x[,"n_31"])*log(15 - 16*r_rr + 16*r_rr^2 - 12*r_rr^3) + (x[,"n_11"] + x[,"n_43"])*log(r_rr + 4*r_rr^2 - 2*r_rr^3) + (x[,"n_13"] + x[,"n_41"])*log(3 - 3*r_rr - 2*r_rr^2 + 2*r_rr^3) + (x[,"n_24"] + x[,"n_30"])*log(1 + 6*r_rr - 13*r_rr^2 + 6*r_rr^3) + (x[,"n_21"] + x[,"n_33"])*log(3 + 20*r_rr - 20*r_rr^2 + 12*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.4_2.3 <- function(x){
logL_cc <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + (n20 + n25 + n30 + n35)*log(3) + 2*(-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + 3*(n10 + n45)*log(pmax(1e-6,1 - r)) + 3*(n15 + n40)*log(pmax(1e-6,r)) + (n20 + n35)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n14 + n41)*(log(11 - 2*r) + 2*log(pmax(1e-6,r))) + (n25 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n44)*(2*log(pmax(1e-6,1 - r)) + log(9 + 2*r)) + (n13 + n42)*(log(pmax(1e-6,r)) + log(18 - 5*r - 4*r^2)) + (n24 + n31)*(log(pmax(1e-6,r)) + log(26 - 17*r + 6*r^2)) + (n22 + n33)*log(41 - 34*r + 29*r^2 - 12*r^3) + (n21 + n34)*log(15 - 10*r + r^2 - 6*r^3) + (n12 + n43)*log(9 + 4*r - 17*r^2 + 4*r^3) + (n23 + n32)*log(24 + 12*r - 7*r^2 + 12*r^3)
return(L)}
interlogL_cc <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cc,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cc <- 3*(x[,"n_25"] + x[,"n_30"])*log10(2) + 3*(x[,"n_20"] + x[,"n_35"])*log10(2) + 3*(x[,"n_15"] + x[,"n_40"])*log10(2) + 3*(x[,"n_10"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_41"])*(-log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_44"])*(-log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_24"] + x[,"n_31"])*(-log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_34"])*(-log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_42"])*(-2*log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_43"])*(-2*log10(2) + log10(29)) + 3*(x[,"n_10"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_15"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_20"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_41"])*(log10(11 - 2*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_25"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(9 + 2*r_cc)) + (x[,"n_13"] + x[,"n_42"])*(log10(pmax(1e-6,r_cc)) + log10(18 - 5*r_cc - 4*r_cc^2)) + (x[,"n_24"] + x[,"n_31"])*(log10(pmax(1e-6,r_cc)) + log10(26 - 17*r_cc + 6*r_cc^2)) + (x[,"n_22"] + x[,"n_33"])*log10(41 - 34*r_cc + 29*r_cc^2 - 12*r_cc^3) + (x[,"n_21"] + x[,"n_34"])*log10(15 - 10*r_cc + r_cc^2 - 6*r_cc^3) + (x[,"n_12"] + x[,"n_43"])*log10(9 + 4*r_cc - 17*r_cc^2 + 4*r_cc^3) + (x[,"n_23"] + x[,"n_32"])*log10(24 + 12*r_cc - 7*r_cc^2 + 12*r_cc^3)
logL_cc <- (-2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_20"] + x[,"n_25"] + x[,"n_30"] + x[,"n_35"])*log(3) + 2*(-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + 3*(x[,"n_10"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_15"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_20"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_41"])*(log(11 - 2*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_25"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(9 + 2*r_cc)) + (x[,"n_13"] + x[,"n_42"])*(log(pmax(1e-6,r_cc)) + log(18 - 5*r_cc - 4*r_cc^2)) + (x[,"n_24"] + x[,"n_31"])*(log(pmax(1e-6,r_cc)) + log(26 - 17*r_cc + 6*r_cc^2)) + (x[,"n_22"] + x[,"n_33"])*log(41 - 34*r_cc + 29*r_cc^2 - 12*r_cc^3) + (x[,"n_21"] + x[,"n_34"])*log(15 - 10*r_cc + r_cc^2 - 6*r_cc^3) + (x[,"n_12"] + x[,"n_43"])*log(9 + 4*r_cc - 17*r_cc^2 + 4*r_cc^3) + (x[,"n_23"] + x[,"n_32"])*log(24 + 12*r_cc - 7*r_cc^2 + 12*r_cc^3)
logL_cm2c <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35 - n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - n45)*log(2) + (-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n10 + n45)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n15 + n40)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n14 + n41)*(log(pmax(1e-6,r)) + log(9 - 5*r + 2*r^2)) + (n25 + n30)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2)) + (n24 + n31)*log(21 + 8*r + 22*r^2 - 12*r^3) + (n23 + n32)*log(27 + 2*r + 10*r^2 - 8*r^3) + (n20 + n35)*log(3 - 7*r + 10*r^2 - 6*r^3) + (n12 + n43)*log(15 - 11*r + 9*r^2 - 4*r^3) + (n11 + n44)*log(6 - 5*r + r^2 - 2*r^3) + (n13 + n42)*log(9 + 5*r - 3*r^2 + 4*r^3) + (n22 + n33)*log(31 + 2*r - 14*r^2 + 8*r^3) + (n21 + n34)*log(39 - 16*r - 14*r^2 + 12*r^3)
return(L)}
interlogL_cm2c <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cm2c,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cm2c <- 3*(x[,"n_15"] + x[,"n_40"])*log10(2) + 3*(x[,"n_10"] + x[,"n_45"])*log10(2) - (x[,"n_25"] + x[,"n_30"])*(-2*log10(2) + log10(5)) - (x[,"n_20"] + x[,"n_35"])*(-2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_42"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_43"])*(-2*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_14"] + x[,"n_41"])*(-log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_44"])*(-log10(2) + log10(7)) - (x[,"n_24"] + x[,"n_31"])*log10(29) - (x[,"n_21"] + x[,"n_34"])*log10(29) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(59)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(59)) + (x[,"n_10"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cm2c)) + 2*log10(pmax(1e-6,r_cm2c))) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm2c)) + log10(9 - 5*r_cm2c + 2*r_cm2c^2)) + (x[,"n_25"] + x[,"n_30"])*(log10(pmax(1e-6,r_cm2c)) + log10(5 - 8*r_cm2c + 6*r_cm2c^2)) + (x[,"n_24"] + x[,"n_31"])*log10(21 + 8*r_cm2c + 22*r_cm2c^2 - 12*r_cm2c^3) + (x[,"n_23"] + x[,"n_32"])*log10(27 + 2*r_cm2c + 10*r_cm2c^2 - 8*r_cm2c^3) + (x[,"n_20"] + x[,"n_35"])*log10(3 - 7*r_cm2c + 10*r_cm2c^2 - 6*r_cm2c^3) + (x[,"n_12"] + x[,"n_43"])*log10(15 - 11*r_cm2c + 9*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_11"] + x[,"n_44"])*log10(6 - 5*r_cm2c + r_cm2c^2 - 2*r_cm2c^3) + (x[,"n_13"] + x[,"n_42"])*log10(9 + 5*r_cm2c - 3*r_cm2c^2 + 4*r_cm2c^3) + (x[,"n_22"] + x[,"n_33"])*log10(31 + 2*r_cm2c - 14*r_cm2c^2 + 8*r_cm2c^3) + (x[,"n_21"] + x[,"n_34"])*log10(39 - 16*r_cm2c - 14*r_cm2c^2 + 12*r_cm2c^3)
logL_cm2c <- (-x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"] - x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - x[,"n_45"])*log(2) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_10"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cm2c)) + 2*log(pmax(1e-6,r_cm2c))) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,r_cm2c)) + log(9 - 5*r_cm2c + 2*r_cm2c^2)) + (x[,"n_25"] + x[,"n_30"])*(log(pmax(1e-6,r_cm2c)) + log(5 - 8*r_cm2c + 6*r_cm2c^2)) + (x[,"n_24"] + x[,"n_31"])*log(21 + 8*r_cm2c + 22*r_cm2c^2 - 12*r_cm2c^3) + (x[,"n_23"] + x[,"n_32"])*log(27 + 2*r_cm2c + 10*r_cm2c^2 - 8*r_cm2c^3) + (x[,"n_20"] + x[,"n_35"])*log(3 - 7*r_cm2c + 10*r_cm2c^2 - 6*r_cm2c^3) + (x[,"n_12"] + x[,"n_43"])*log(15 - 11*r_cm2c + 9*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_11"] + x[,"n_44"])*log(6 - 5*r_cm2c + r_cm2c^2 - 2*r_cm2c^3) + (x[,"n_13"] + x[,"n_42"])*log(9 + 5*r_cm2c - 3*r_cm2c^2 + 4*r_cm2c^3) + (x[,"n_22"] + x[,"n_33"])*log(31 + 2*r_cm2c - 14*r_cm2c^2 + 8*r_cm2c^3) + (x[,"n_21"] + x[,"n_34"])*log(39 - 16*r_cm2c - 14*r_cm2c^2 + 12*r_cm2c^3)
logL_cm1c <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + 2*(-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n15 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n45)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n44)*(log(pmax(1e-6,r)) + log(8 - 7*r + 2*r^2)) + (n20 + n35)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n22 + n33)*log(30 - 4*r + 13*r^2 - 12*r^3) + (n21 + n34)*log(12 - 10*r + 13*r^2 - 6*r^3) + (n13 + n42)*log(11 - 12*r + 11*r^2 - 4*r^3) + (n25 + n30)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n14 + n41)*log(3 - r^2 - 2*r^3) + (n12 + n43)*log(6 + 2*r - r^2 + 4*r^3) + (n24 + n31)*log(9 + 2*r - 5*r^2 + 6*r^3) + (n23 + n32)*log(27 + 14*r - 23*r^2 + 12*r^3)
return(L)}
interlogL_cm1c <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cm1c,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cm1c <- 3*(x[,"n_15"] + x[,"n_40"])*log10(2) + 3*(x[,"n_10"] + x[,"n_45"])*log10(2) + (x[,"n_25"] + x[,"n_30"])*(3*log10(2) - log10(3)) - (x[,"n_20"] + x[,"n_35"])*(-3*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_41"])*(-log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_44"])*(-log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_24"] + x[,"n_31"])*(-log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_34"])*(-log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_42"])*(-2*log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_43"])*(-2*log10(2) + log10(29)) + (x[,"n_15"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + log10(pmax(1e-6,r_cm1c))) + (x[,"n_10"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_cm1c)) + log10(8 - 7*r_cm1c + 2*r_cm1c^2)) + (x[,"n_20"] + x[,"n_35"])*(log10(pmax(1e-6,r_cm1c)) + log10(2 - 4*r_cm1c + 3*r_cm1c^2)) + (x[,"n_22"] + x[,"n_33"])*log10(30 - 4*r_cm1c + 13*r_cm1c^2 - 12*r_cm1c^3) + (x[,"n_21"] + x[,"n_34"])*log10(12 - 10*r_cm1c + 13*r_cm1c^2 - 6*r_cm1c^3) + (x[,"n_13"] + x[,"n_42"])*log10(11 - 12*r_cm1c + 11*r_cm1c^2 - 4*r_cm1c^3) + (x[,"n_25"] + x[,"n_30"])*log10(1 - 3*r_cm1c + 5*r_cm1c^2 - 3*r_cm1c^3) + (x[,"n_14"] + x[,"n_41"])*log10(3 - r_cm1c^2 - 2*r_cm1c^3) + (x[,"n_12"] + x[,"n_43"])*log10(6 + 2*r_cm1c - r_cm1c^2 + 4*r_cm1c^3) + (x[,"n_24"] + x[,"n_31"])*log10(9 + 2*r_cm1c - 5*r_cm1c^2 + 6*r_cm1c^3) + (x[,"n_23"] + x[,"n_32"])*log10(27 + 14*r_cm1c - 23*r_cm1c^2 + 12*r_cm1c^3)
logL_cm1c <- (-2*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + 2*(-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_15"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_cm1c)) + log(pmax(1e-6,r_cm1c))) + (x[,"n_10"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_cm1c)) + log(8 - 7*r_cm1c + 2*r_cm1c^2)) + (x[,"n_20"] + x[,"n_35"])*(log(pmax(1e-6,r_cm1c)) + log(2 - 4*r_cm1c + 3*r_cm1c^2)) + (x[,"n_22"] + x[,"n_33"])*log(30 - 4*r_cm1c + 13*r_cm1c^2 - 12*r_cm1c^3) + (x[,"n_21"] + x[,"n_34"])*log(12 - 10*r_cm1c + 13*r_cm1c^2 - 6*r_cm1c^3) + (x[,"n_13"] + x[,"n_42"])*log(11 - 12*r_cm1c + 11*r_cm1c^2 - 4*r_cm1c^3) + (x[,"n_25"] + x[,"n_30"])*log(1 - 3*r_cm1c + 5*r_cm1c^2 - 3*r_cm1c^3) + (x[,"n_14"] + x[,"n_41"])*log(3 - r_cm1c^2 - 2*r_cm1c^3) + (x[,"n_12"] + x[,"n_43"])*log(6 + 2*r_cm1c - r_cm1c^2 + 4*r_cm1c^3) + (x[,"n_24"] + x[,"n_31"])*log(9 + 2*r_cm1c - 5*r_cm1c^2 + 6*r_cm1c^3) + (x[,"n_23"] + x[,"n_32"])*log(27 + 14*r_cm1c - 23*r_cm1c^2 + 12*r_cm1c^3)
logL_rc <- function(r,n20,n25,n30,n35,n10,n11,n12,n13,n14,n15,n21,n22,n23,n24,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45) {
L <- (n20 + n25 + n30 + n35)*log(3) + (-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n15 + n40)*(log(3 - 2*r) + 2*log(pmax(1e-6,r))) + (n10 + n45)*(2*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n14 + n41)*(log(pmax(1e-6,r)) + log(12 - 5*r + 2*r^2)) + (n25 + n30)*(log(pmax(1e-6,r)) + log(2 - 3*r + 2*r^2)) + (n23 + n32)*log(57 + 16*r^2 - 12*r^3) + (n24 + n31)*log(18 - 2*r + 11*r^2 - 6*r^3) + (n12 + n43)*log(21 - 16*r + 8*r^2 - 4*r^3) + (n11 + n44)*log(9 - 8*r + r^2 - 2*r^3) + (n20 + n35)*log(1 - 2*r + 3*r^2 - 2*r^3) + (n13 + n42)*log(9 + 12*r - 4*r^2 + 4*r^3) + (n21 + n34)*log(21 - 2*r - 7*r^2 + 6*r^3) + (n22 + n33)*log(61 + 4*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rc <- function(n20,n25,n30,n35,n10,n11,n12,n13,n14,n15,n21,n22,n23,n24,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45) {
optimize(logL_rc,c(0,0.5), n20,n25,n30,n35,n10,n11,n12,n13,n14,n15,n21,n22,n23,n24,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_20"],x[,"n_25"],x[,"n_30"],x[,"n_35"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rc <- (x[,"n_25"] + x[,"n_30"])*log10(2) + (x[,"n_20"] + x[,"n_35"])*log10(2) + (x[,"n_15"] + x[,"n_40"])*log10(2) + (x[,"n_10"] + x[,"n_45"])*log10(2) - (x[,"n_14"] + x[,"n_41"])*log10(5) - (x[,"n_11"] + x[,"n_44"])*log10(5) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(7) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(7) + log10(17)) - (x[,"n_24"] + x[,"n_31"])*log10(19) - (x[,"n_21"] + x[,"n_34"])*log10(19) - (x[,"n_13"] + x[,"n_42"])*(-log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_43"])*(-log10(2) + log10(29)) + (x[,"n_15"] + x[,"n_40"])*(log10(3 - 2*r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_10"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,r_rc)) + log10(12 - 5*r_rc + 2*r_rc^2)) + (x[,"n_25"] + x[,"n_30"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 3*r_rc + 2*r_rc^2)) + (x[,"n_23"] + x[,"n_32"])*log10(57 + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_24"] + x[,"n_31"])*log10(18 - 2*r_rc + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_12"] + x[,"n_43"])*log10(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_11"] + x[,"n_44"])*log10(9 - 8*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_20"] + x[,"n_35"])*log10(1 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3) + (x[,"n_13"] + x[,"n_42"])*log10(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_21"] + x[,"n_34"])*log10(21 - 2*r_rc - 7*r_rc^2 + 6*r_rc^3) + (x[,"n_22"] + x[,"n_33"])*log10(61 + 4*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rc <- (x[,"n_20"] + x[,"n_25"] + x[,"n_30"] + x[,"n_35"])*log(3) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_15"] + x[,"n_40"])*(log(3 - 2*r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_10"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,r_rc)) + log(12 - 5*r_rc + 2*r_rc^2)) + (x[,"n_25"] + x[,"n_30"])*(log(pmax(1e-6,r_rc)) + log(2 - 3*r_rc + 2*r_rc^2)) + (x[,"n_23"] + x[,"n_32"])*log(57 + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_24"] + x[,"n_31"])*log(18 - 2*r_rc + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_12"] + x[,"n_43"])*log(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_11"] + x[,"n_44"])*log(9 - 8*r_rc + r_rc^2 - 2*r_rc^3) + (x[,"n_20"] + x[,"n_35"])*log(1 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3) + (x[,"n_13"] + x[,"n_42"])*log(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_21"] + x[,"n_34"])*log(21 - 2*r_rc - 7*r_rc^2 + 6*r_rc^3) + (x[,"n_22"] + x[,"n_33"])*log(61 + 4*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rm2c <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 4*n20 - 4*n21 - 2*n22 - 2*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 2*n32 - 2*n33 - 4*n34 - 4*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45)*log(2) + (-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n15 + n40)*(log(3 - 2*r) + log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n45)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 + 2*r)) + (n25 + n30)*log(9 - 16*r + 22*r^2 - 12*r^3) + (n21 + n34)*log(51 + 4*r + 26*r^2 - 12*r^3) + (n13 + n42)*log(27 - 14*r + 12*r^2 - 4*r^3) + (n14 + n41)*log(9 - 3*r - r^2 - 2*r^3) + (n22 + n33)*log(14 + r + 2*r^2 - 2*r^3) + (n11 + n44)*log(3 + 11*r - 7*r^2 + 2*r^3) + (n23 + n32)*log(15 + r - 4*r^2 + 2*r^3) + (n12 + n43)*log(21 + 2*r + 4*r^3) + (n20 + n35)*log(3 + 8*r - 14*r^2 + 12*r^3) + (n24 + n31)*log(69 - 20*r - 10*r^2 + 12*r^3)
return(L)}
interlogL_rm2c <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rm2c,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm2c <- (x[,"n_15"] + x[,"n_40"])*log10(2) + (x[,"n_10"] + x[,"n_45"])*log10(2) - (x[,"n_25"] + x[,"n_30"])*log10(5) - (x[,"n_20"] + x[,"n_35"])*log10(5) - (x[,"n_13"] + x[,"n_42"])*(-log10(2) + 2*log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_43"])*(-log10(2) + 2*log10(3) + log10(5)) - (x[,"n_14"] + x[,"n_41"])*log10(7) - (x[,"n_11"] + x[,"n_44"])*log10(7) - (x[,"n_24"] + x[,"n_31"])*(log10(2) + log10(29)) - (x[,"n_21"] + x[,"n_34"])*(log10(2) + log10(29)) - (x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + log10(59)) - (x[,"n_22"] + x[,"n_33"])*(-2*log10(2) + log10(59)) + (x[,"n_15"] + x[,"n_40"])*(log10(3 - 2*r_rm2c) + log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c))) + (x[,"n_10"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c)) + log10(1 + 2*r_rm2c)) + (x[,"n_25"] + x[,"n_30"])*log10(9 - 16*r_rm2c + 22*r_rm2c^2 - 12*r_rm2c^3) + (x[,"n_21"] + x[,"n_34"])*log10(51 + 4*r_rm2c + 26*r_rm2c^2 - 12*r_rm2c^3) + (x[,"n_13"] + x[,"n_42"])*log10(27 - 14*r_rm2c + 12*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_14"] + x[,"n_41"])*log10(9 - 3*r_rm2c - r_rm2c^2 - 2*r_rm2c^3) + (x[,"n_22"] + x[,"n_33"])*log10(14 + r_rm2c + 2*r_rm2c^2 - 2*r_rm2c^3) + (x[,"n_11"] + x[,"n_44"])*log10(3 + 11*r_rm2c - 7*r_rm2c^2 + 2*r_rm2c^3) + (x[,"n_23"] + x[,"n_32"])*log10(15 + r_rm2c - 4*r_rm2c^2 + 2*r_rm2c^3) + (x[,"n_12"] + x[,"n_43"])*log10(21 + 2*r_rm2c + 4*r_rm2c^3) + (x[,"n_20"] + x[,"n_35"])*log10(3 + 8*r_rm2c - 14*r_rm2c^2 + 12*r_rm2c^3) + (x[,"n_24"] + x[,"n_31"])*log10(69 - 20*r_rm2c - 10*r_rm2c^2 + 12*r_rm2c^3)
logL_rm2c <- (-3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_15"] + x[,"n_40"])*(log(3 - 2*r_rm2c) + log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c))) + (x[,"n_10"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c)) + log(1 + 2*r_rm2c)) + (x[,"n_25"] + x[,"n_30"])*log(9 - 16*r_rm2c + 22*r_rm2c^2 - 12*r_rm2c^3) + (x[,"n_21"] + x[,"n_34"])*log(51 + 4*r_rm2c + 26*r_rm2c^2 - 12*r_rm2c^3) + (x[,"n_13"] + x[,"n_42"])*log(27 - 14*r_rm2c + 12*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_14"] + x[,"n_41"])*log(9 - 3*r_rm2c - r_rm2c^2 - 2*r_rm2c^3) + (x[,"n_22"] + x[,"n_33"])*log(14 + r_rm2c + 2*r_rm2c^2 - 2*r_rm2c^3) + (x[,"n_11"] + x[,"n_44"])*log(3 + 11*r_rm2c - 7*r_rm2c^2 + 2*r_rm2c^3) + (x[,"n_23"] + x[,"n_32"])*log(15 + r_rm2c - 4*r_rm2c^2 + 2*r_rm2c^3) + (x[,"n_12"] + x[,"n_43"])*log(21 + 2*r_rm2c + 4*r_rm2c^3) + (x[,"n_20"] + x[,"n_35"])*log(3 + 8*r_rm2c - 14*r_rm2c^2 + 12*r_rm2c^3) + (x[,"n_24"] + x[,"n_31"])*log(69 - 20*r_rm2c - 10*r_rm2c^2 + 12*r_rm2c^3)
logL_rm1c <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n15 + n40)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r))) + (n10 + n45)*(2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n20 + n35)*(log(pmax(1e-6,r)) + log(2 + 5*r - 6*r^2)) + (n11 + n44)*(log(pmax(1e-6,r)) + log(4 + 13*r - 2*r^2)) + (n23 + n32)*log(75 - 44*r + 32*r^2 - 12*r^3) + (n24 + n31)*log(27 - 14*r - r^2 - 6*r^3) + (n12 + n43)*log(3 + 28*r - 8*r^2 - 4*r^3) + (n14 + n41)*log(15 - 24*r + 7*r^2 + 2*r^3) + (n13 + n42)*log(19 - 20*r^2 + 4*r^3) + (n21 + n34)*log(6 + 34*r - 19*r^2 + 6*r^3) + (n25 + n30)*log(1 + 6*r - 13*r^2 + 6*r^3) + (n22 + n33)*log(51 + 16*r - 4*r^2 + 12*r^3)
return(L)}
interlogL_rm1c <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rm1c,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm1c <- (x[,"n_15"] + x[,"n_40"])*log10(2) + (x[,"n_10"] + x[,"n_45"])*log10(2) - (x[,"n_25"] + x[,"n_30"])*(-log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_41"])*log10(5) - (x[,"n_11"] + x[,"n_44"])*log10(5) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(7) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(7) + log10(17)) - (x[,"n_24"] + x[,"n_31"])*log10(19) - (x[,"n_21"] + x[,"n_34"])*log10(19) - (x[,"n_13"] + x[,"n_42"])*(-log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_43"])*(-log10(2) + log10(29)) + (x[,"n_15"] + x[,"n_40"])*(log10(3 - 2*r_rm1c) + 2*log10(pmax(1e-6,1 - r_rm1c))) + (x[,"n_10"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(1 + 2*r_rm1c)) + (x[,"n_20"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm1c)) + log10(2 + 5*r_rm1c - 6*r_rm1c^2)) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm1c)) + log10(4 + 13*r_rm1c - 2*r_rm1c^2)) + (x[,"n_23"] + x[,"n_32"])*log10(75 - 44*r_rm1c + 32*r_rm1c^2 - 12*r_rm1c^3) + (x[,"n_24"] + x[,"n_31"])*log10(27 - 14*r_rm1c - r_rm1c^2 - 6*r_rm1c^3) + (x[,"n_12"] + x[,"n_43"])*log10(3 + 28*r_rm1c - 8*r_rm1c^2 - 4*r_rm1c^3) + (x[,"n_14"] + x[,"n_41"])*log10(15 - 24*r_rm1c + 7*r_rm1c^2 + 2*r_rm1c^3) + (x[,"n_13"] + x[,"n_42"])*log10(19 - 20*r_rm1c^2 + 4*r_rm1c^3) + (x[,"n_21"] + x[,"n_34"])*log10(6 + 34*r_rm1c - 19*r_rm1c^2 + 6*r_rm1c^3) + (x[,"n_25"] + x[,"n_30"])*log10(1 + 6*r_rm1c - 13*r_rm1c^2 + 6*r_rm1c^3) + (x[,"n_22"] + x[,"n_33"])*log10(51 + 16*r_rm1c - 4*r_rm1c^2 + 12*r_rm1c^3)
logL_rm1c <- (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_15"] + x[,"n_40"])*(log(3 - 2*r_rm1c) + 2*log(pmax(1e-6,1 - r_rm1c))) + (x[,"n_10"] + x[,"n_45"])*(2*log(pmax(1e-6,r_rm1c)) + log(1 + 2*r_rm1c)) + (x[,"n_20"] + x[,"n_35"])*(log(pmax(1e-6,r_rm1c)) + log(2 + 5*r_rm1c - 6*r_rm1c^2)) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_rm1c)) + log(4 + 13*r_rm1c - 2*r_rm1c^2)) + (x[,"n_23"] + x[,"n_32"])*log(75 - 44*r_rm1c + 32*r_rm1c^2 - 12*r_rm1c^3) + (x[,"n_24"] + x[,"n_31"])*log(27 - 14*r_rm1c - r_rm1c^2 - 6*r_rm1c^3) + (x[,"n_12"] + x[,"n_43"])*log(3 + 28*r_rm1c - 8*r_rm1c^2 - 4*r_rm1c^3) + (x[,"n_14"] + x[,"n_41"])*log(15 - 24*r_rm1c + 7*r_rm1c^2 + 2*r_rm1c^3) + (x[,"n_13"] + x[,"n_42"])*log(19 - 20*r_rm1c^2 + 4*r_rm1c^3) + (x[,"n_21"] + x[,"n_34"])*log(6 + 34*r_rm1c - 19*r_rm1c^2 + 6*r_rm1c^3) + (x[,"n_25"] + x[,"n_30"])*log(1 + 6*r_rm1c - 13*r_rm1c^2 + 6*r_rm1c^3) + (x[,"n_22"] + x[,"n_33"])*log(51 + 16*r_rm1c - 4*r_rm1c^2 + 12*r_rm1c^3)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_rc,r_rm2c,r_rm1c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_rc,LOD_rm2c,LOD_rm1c,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_rc,logL_rm2c,logL_rm1c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","repulsion coupling","repulsion mixed2c","repulsion mixed1c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.4_2.4 <- function(x){
logL_cc <- function(r,n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44,n21,n25,n31,n35,n11,n15,n41,n45) {
L <- (-n12 + 2*n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34 - n42 + 2*n43 - n44)*log(2) + (n21 + n25 + n31 + n35)*log(3) + 2*(-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + 3*(n11 + n45)*log(pmax(1e-6,1 - r)) + 3*(n15 + n41)*log(pmax(1e-6,r)) + (n13 + n43)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n21 + n35)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n14 + n42)*(log(7 - 4*r) + 2*log(pmax(1e-6,r))) + (n25 + n31)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n12 + n44)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r)) + (n23 + n33)*log(11 - 8*r + 8*r^2) + (n24 + n32)*(log(pmax(1e-6,r)) + log(16 - 19*r + 12*r^2)) + (n22 + n34)*log(9 - 14*r + 17*r^2 - 12*r^3)
return(L)}
interlogL_cc <- function(n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44,n21,n25,n31,n35,n11,n15,n41,n45) {
optimize(logL_cc,c(0,0.5), n12,n13,n14,n22,n23,n24,n32,n33,n34,n42,n43,n44,n21,n25,n31,n35,n11,n15,n41,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_21"],x[,"n_25"],x[,"n_31"],x[,"n_35"],x[,"n_11"],x[,"n_15"],x[,"n_41"],x[,"n_45"])
LOD_cc <- 3*(x[,"n_25"] + x[,"n_31"])*log10(2) + 3*(x[,"n_21"] + x[,"n_35"])*log10(2) + 3*(x[,"n_15"] + x[,"n_41"])*log10(2) + 2*(x[,"n_13"] + x[,"n_43"])*log10(2) + 3*(x[,"n_11"] + x[,"n_45"])*log10(2) - 2*(x[,"n_23"] + x[,"n_33"])*log10(3) - (x[,"n_14"] + x[,"n_42"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - (x[,"n_24"] + x[,"n_32"])*(-2*log10(2) + log10(19)) - (x[,"n_22"] + x[,"n_34"])*(-2*log10(2) + log10(19)) + 3*(x[,"n_11"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_15"] + x[,"n_41"])*log10(pmax(1e-6,r_cc)) + (x[,"n_13"] + x[,"n_43"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_21"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_42"])*(log10(7 - 4*r_cc) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_25"] + x[,"n_31"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_12"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 4*r_cc)) + (x[,"n_23"] + x[,"n_33"])*log10(11 - 8*r_cc + 8*r_cc^2) + (x[,"n_24"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(16 - 19*r_cc + 12*r_cc^2)) + (x[,"n_22"] + x[,"n_34"])*log10(9 - 14*r_cc + 17*r_cc^2 - 12*r_cc^3)
logL_cc <- (-x[,"n_12"] + 2*x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] + 2*x[,"n_43"] - x[,"n_44"])*log(2) + (x[,"n_21"] + x[,"n_25"] + x[,"n_31"] + x[,"n_35"])*log(3) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + 3*(x[,"n_11"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + 3*(x[,"n_15"] + x[,"n_41"])*log(pmax(1e-6,r_cc)) + (x[,"n_13"] + x[,"n_43"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_21"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_42"])*(log(7 - 4*r_cc) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_25"] + x[,"n_31"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_12"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 4*r_cc)) + (x[,"n_23"] + x[,"n_33"])*log(11 - 8*r_cc + 8*r_cc^2) + (x[,"n_24"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(16 - 19*r_cc + 12*r_cc^2)) + (x[,"n_22"] + x[,"n_34"])*log(9 - 14*r_cc + 17*r_cc^2 - 12*r_cc^3)
logL_cm3c <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-2*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35 - 2*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + 2*(n13 + n43)*log(3) + (-2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - n23 - 2*n24 - 2*n25 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(5) + (n23 + n33)*log(7) + (n15 + n41)*(log(2 - r) + 2*log(pmax(1e-6,r))) + (n11 + n45)*(log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n25 + n31)*(log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n14 + n42)*(log(pmax(1e-6,r)) + log(12 - 7*r + 4*r^2)) + (n24 + n32)*log(15 + 2*r + 19*r^2 - 12*r^3) + (n12 + n44)*log(9 - 10*r + 5*r^2 - 4*r^3) + (n21 + n35)*log(3 - 4*r + 4*r^2 - 3*r^3) + (n22 + n34)*log(24 - 4*r - 17*r^2 + 12*r^3)
return(L)}
interlogL_cm3c <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_cm3c,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm3c <- mapply(interlogL_cm3c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cm3c <- -((x[,"n_15"] + x[,"n_41"])*(-3*log10(2) + log10(3))) - (x[,"n_11"] + x[,"n_45"])*(-3*log10(2) + log10(3)) - (x[,"n_24"] + x[,"n_32"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_22"] + x[,"n_34"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(13)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(13)) - (x[,"n_14"] + x[,"n_42"])*(-2*log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_44"])*(-2*log10(2) + log10(19)) + (x[,"n_15"] + x[,"n_41"])*(log10(2 - r_cm3c) + 2*log10(pmax(1e-6,r_cm3c))) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cm3c)) + log10(1 - r_cm3c^2)) + (x[,"n_25"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm3c)) + log10(5 - 5*r_cm3c + 3*r_cm3c^2)) + (x[,"n_14"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm3c)) + log10(12 - 7*r_cm3c + 4*r_cm3c^2)) + (x[,"n_24"] + x[,"n_32"])*log10(15 + 2*r_cm3c + 19*r_cm3c^2 - 12*r_cm3c^3) + (x[,"n_12"] + x[,"n_44"])*log10(9 - 10*r_cm3c + 5*r_cm3c^2 - 4*r_cm3c^3) + (x[,"n_21"] + x[,"n_35"])*log10(3 - 4*r_cm3c + 4*r_cm3c^2 - 3*r_cm3c^3) + (x[,"n_22"] + x[,"n_34"])*log10(24 - 4*r_cm3c - 17*r_cm3c^2 + 12*r_cm3c^3)
logL_cm3c <- (-2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + 2*(x[,"n_13"] + x[,"n_43"])*log(3) + (-2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(5) + (x[,"n_23"] + x[,"n_33"])*log(7) + (x[,"n_15"] + x[,"n_41"])*(log(2 - r_cm3c) + 2*log(pmax(1e-6,r_cm3c))) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cm3c)) + log(1 - r_cm3c^2)) + (x[,"n_25"] + x[,"n_31"])*(log(pmax(1e-6,r_cm3c)) + log(5 - 5*r_cm3c + 3*r_cm3c^2)) + (x[,"n_14"] + x[,"n_42"])*(log(pmax(1e-6,r_cm3c)) + log(12 - 7*r_cm3c + 4*r_cm3c^2)) + (x[,"n_24"] + x[,"n_32"])*log(15 + 2*r_cm3c + 19*r_cm3c^2 - 12*r_cm3c^3) + (x[,"n_12"] + x[,"n_44"])*log(9 - 10*r_cm3c + 5*r_cm3c^2 - 4*r_cm3c^3) + (x[,"n_21"] + x[,"n_35"])*log(3 - 4*r_cm3c + 4*r_cm3c^2 - 3*r_cm3c^3) + (x[,"n_22"] + x[,"n_34"])*log(24 - 4*r_cm3c - 17*r_cm3c^2 + 12*r_cm3c^3)
logL_cm2c <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-n11 - 2*n12 - 2*n13 - 2*n14 - n15 - n21 - 2*n22 - 2*n23 - 2*n24 - n25 - n31 - 2*n32 - 2*n33 - 2*n34 - n35 - n41 - 2*n42 - 2*n43 - 2*n44 - n45)*log(2) + (-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n11 + n45)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n23 + n33)*log(51 + 8*r - 8*r^2) + (n15 + n41)*(log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n13 + n43)*log(15 - 8*r + 8*r^2) + (n22 + n34)*log(33 - 14*r + 17*r^2 - 12*r^3) + (n14 + n42)*log(9 - 6*r + 7*r^2 - 4*r^3) + (n25 + n31)*log(3 - 4*r + 7*r^2 - 3*r^3) + (n21 + n35)*log(3 - r - 2*r^2 + 3*r^3) + (n12 + n44)*log(6 + 4*r - 5*r^2 + 4*r^3) + (n24 + n32)*log(24 + 16*r - 19*r^2 + 12*r^3)
return(L)}
interlogL_cm2c <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_cm2c,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cm2c <- -((x[,"n_15"] + x[,"n_41"])*(-3*log10(2) + log10(5))) - (x[,"n_11"] + x[,"n_45"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_43"])*log10(13) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(19)) - (x[,"n_24"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_22"] + x[,"n_34"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_14"] + x[,"n_42"])*(-2*log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_44"])*(-2*log10(2) + log10(29)) - (x[,"n_23"] + x[,"n_33"])*log10(53) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c)) + log10(2 + r_cm2c)) + (x[,"n_23"] + x[,"n_33"])*log10(51 + 8*r_cm2c - 8*r_cm2c^2) + (x[,"n_15"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm2c)) + log10(3 - 4*r_cm2c + r_cm2c^2)) + (x[,"n_13"] + x[,"n_43"])*log10(15 - 8*r_cm2c + 8*r_cm2c^2) + (x[,"n_22"] + x[,"n_34"])*log10(33 - 14*r_cm2c + 17*r_cm2c^2 - 12*r_cm2c^3) + (x[,"n_14"] + x[,"n_42"])*log10(9 - 6*r_cm2c + 7*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_25"] + x[,"n_31"])*log10(3 - 4*r_cm2c + 7*r_cm2c^2 - 3*r_cm2c^3) + (x[,"n_21"] + x[,"n_35"])*log10(3 - r_cm2c - 2*r_cm2c^2 + 3*r_cm2c^3) + (x[,"n_12"] + x[,"n_44"])*log10(6 + 4*r_cm2c - 5*r_cm2c^2 + 4*r_cm2c^3) + (x[,"n_24"] + x[,"n_32"])*log10(24 + 16*r_cm2c - 19*r_cm2c^2 + 12*r_cm2c^3)
logL_cm2c <- (-x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - x[,"n_45"])*log(2) + (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c)) + log(2 + r_cm2c)) + (x[,"n_23"] + x[,"n_33"])*log(51 + 8*r_cm2c - 8*r_cm2c^2) + (x[,"n_15"] + x[,"n_41"])*(log(pmax(1e-6,r_cm2c)) + log(3 - 4*r_cm2c + r_cm2c^2)) + (x[,"n_13"] + x[,"n_43"])*log(15 - 8*r_cm2c + 8*r_cm2c^2) + (x[,"n_22"] + x[,"n_34"])*log(33 - 14*r_cm2c + 17*r_cm2c^2 - 12*r_cm2c^3) + (x[,"n_14"] + x[,"n_42"])*log(9 - 6*r_cm2c + 7*r_cm2c^2 - 4*r_cm2c^3) + (x[,"n_25"] + x[,"n_31"])*log(3 - 4*r_cm2c + 7*r_cm2c^2 - 3*r_cm2c^3) + (x[,"n_21"] + x[,"n_35"])*log(3 - r_cm2c - 2*r_cm2c^2 + 3*r_cm2c^3) + (x[,"n_12"] + x[,"n_44"])*log(6 + 4*r_cm2c - 5*r_cm2c^2 + 4*r_cm2c^3) + (x[,"n_24"] + x[,"n_32"])*log(24 + 16*r_cm2c - 19*r_cm2c^2 + 12*r_cm2c^3)
logL_rc <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-2*n11 - n12 - 2*n13 - n14 - 2*n15 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n41 - n42 - 2*n43 - n44 - 2*n45)*log(2) + (n21 + n25 + n31 + n35)*log(3) + 2*(-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n15 + n41)*(log(3 - 2*r) + 2*log(pmax(1e-6,r))) + (n11 + n45)*(2*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n13 + n43)*log(3 + 4*r - 4*r^2) + (n25 + n31)*(log(pmax(1e-6,r)) + log(2 - 3*r + 2*r^2)) + (n14 + n42)*(log(pmax(1e-6,r)) + log(3 - 2*r + 2*r^2)) + (n23 + n33)*log(19 - 4*r + 4*r^2) + (n24 + n32)*log(9 - 4*r + 16*r^2 - 12*r^3) + (n21 + n35)*log(1 - 2*r + 3*r^2 - 2*r^3) + (n12 + n44)*log(3 - 5*r + 4*r^2 - 2*r^3) + (n22 + n34)*log(9 + 8*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rc <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_rc,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rc <- (x[,"n_25"] + x[,"n_31"])*log10(2) + (x[,"n_21"] + x[,"n_35"])*log10(2) + (x[,"n_15"] + x[,"n_41"])*log10(2) - 2*(x[,"n_13"] + x[,"n_43"])*log10(2) + (x[,"n_11"] + x[,"n_45"])*log10(2) - (x[,"n_23"] + x[,"n_33"])*(log10(2) + 2*log10(3)) - (x[,"n_14"] + x[,"n_42"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - (x[,"n_24"] + x[,"n_32"])*(-log10(2) + log10(19)) - (x[,"n_22"] + x[,"n_34"])*(-log10(2) + log10(19)) + (x[,"n_15"] + x[,"n_41"])*(log10(3 - 2*r_rc) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_11"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_13"] + x[,"n_43"])*log10(3 + 4*r_rc - 4*r_rc^2) + (x[,"n_25"] + x[,"n_31"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 3*r_rc + 2*r_rc^2)) + (x[,"n_14"] + x[,"n_42"])*(log10(pmax(1e-6,r_rc)) + log10(3 - 2*r_rc + 2*r_rc^2)) + (x[,"n_23"] + x[,"n_33"])*log10(19 - 4*r_rc + 4*r_rc^2) + (x[,"n_24"] + x[,"n_32"])*log10(9 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_21"] + x[,"n_35"])*log10(1 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3) + (x[,"n_12"] + x[,"n_44"])*log10(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3) + (x[,"n_22"] + x[,"n_34"])*log10(9 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rc <- (-2*x[,"n_11"] - x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - x[,"n_42"] - 2*x[,"n_43"] - x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_25"] + x[,"n_31"] + x[,"n_35"])*log(3) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_15"] + x[,"n_41"])*(log(3 - 2*r_rc) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_11"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_13"] + x[,"n_43"])*log(3 + 4*r_rc - 4*r_rc^2) + (x[,"n_25"] + x[,"n_31"])*(log(pmax(1e-6,r_rc)) + log(2 - 3*r_rc + 2*r_rc^2)) + (x[,"n_14"] + x[,"n_42"])*(log(pmax(1e-6,r_rc)) + log(3 - 2*r_rc + 2*r_rc^2)) + (x[,"n_23"] + x[,"n_33"])*log(19 - 4*r_rc + 4*r_rc^2) + (x[,"n_24"] + x[,"n_32"])*log(9 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_21"] + x[,"n_35"])*log(1 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3) + (x[,"n_12"] + x[,"n_44"])*log(3 - 5*r_rc + 4*r_rc^2 - 2*r_rc^3) + (x[,"n_22"] + x[,"n_34"])*log(9 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rm3c <- function(r,n13,n43,n11,n12,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n44,n45) {
L <- (n13 + n43)*log(3) + (-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n15 + n41)*(log(3 - 2*r) + log(2 - r) + log(pmax(1e-6,r))) + (n23 + n33)*log(67 + 12*r - 12*r^2) + (n11 + n45)*(log(1 + 2*r) + log(1 - r^2)) + (n13 + n43)*log(7 - 4*r + 4*r^2) + (n22 + n34)*log(39 - 8*r + 20*r^2 - 12*r^3) + (n25 + n31)*log(9 - 10*r + 13*r^2 - 6*r^3) + (n14 + n42)*log(9 + 4*r^2 - 4*r^3) + (n12 + n44)*log(9 + 4*r - 8*r^2 + 4*r^3) + (n21 + n35)*log(6 + 2*r - 5*r^2 + 6*r^3) + (n24 + n32)*log(39 + 4*r - 16*r^2 + 12*r^3)
return(L)}
interlogL_rm3c <- function(n13,n43,n11,n12,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n44,n45) {
optimize(logL_rm3c,c(0,0.5), n13,n43,n11,n12,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm3c <- mapply(interlogL_rm3c,x[,"n_13"],x[,"n_43"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_44"],x[,"n_45"])
LOD_rm3c <- -((x[,"n_15"] + x[,"n_41"])*(-log10(2) + log10(3))) - (x[,"n_11"] + x[,"n_45"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_43"])*(log10(2) + log10(3)) - (x[,"n_23"] + x[,"n_33"])*(log10(2) + log10(5) + log10(7)) - (x[,"n_24"] + x[,"n_32"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_22"] + x[,"n_34"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_25"] + x[,"n_31"])*(-log10(2) + log10(13)) - (x[,"n_21"] + x[,"n_35"])*(-log10(2) + log10(13)) - (x[,"n_14"] + x[,"n_42"])*(-log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_44"])*(-log10(2) + log10(19)) + (x[,"n_15"] + x[,"n_41"])*(log10(3 - 2*r_rm3c) + log10(2 - r_rm3c) + log10(pmax(1e-6,r_rm3c))) + (x[,"n_23"] + x[,"n_33"])*log10(67 + 12*r_rm3c - 12*r_rm3c^2) + (x[,"n_11"] + x[,"n_45"])*(log10(1 + 2*r_rm3c) + log10(1 - r_rm3c^2)) + (x[,"n_13"] + x[,"n_43"])*log10(7 - 4*r_rm3c + 4*r_rm3c^2) + (x[,"n_22"] + x[,"n_34"])*log10(39 - 8*r_rm3c + 20*r_rm3c^2 - 12*r_rm3c^3) + (x[,"n_25"] + x[,"n_31"])*log10(9 - 10*r_rm3c + 13*r_rm3c^2 - 6*r_rm3c^3) + (x[,"n_14"] + x[,"n_42"])*log10(9 + 4*r_rm3c^2 - 4*r_rm3c^3) + (x[,"n_12"] + x[,"n_44"])*log10(9 + 4*r_rm3c - 8*r_rm3c^2 + 4*r_rm3c^3) + (x[,"n_21"] + x[,"n_35"])*log10(6 + 2*r_rm3c - 5*r_rm3c^2 + 6*r_rm3c^3) + (x[,"n_24"] + x[,"n_32"])*log10(39 + 4*r_rm3c - 16*r_rm3c^2 + 12*r_rm3c^3)
logL_rm3c <- (x[,"n_13"] + x[,"n_43"])*log(3) + (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_15"] + x[,"n_41"])*(log(3 - 2*r_rm3c) + log(2 - r_rm3c) + log(pmax(1e-6,r_rm3c))) + (x[,"n_23"] + x[,"n_33"])*log(67 + 12*r_rm3c - 12*r_rm3c^2) + (x[,"n_11"] + x[,"n_45"])*(log(1 + 2*r_rm3c) + log(1 - r_rm3c^2)) + (x[,"n_13"] + x[,"n_43"])*log(7 - 4*r_rm3c + 4*r_rm3c^2) + (x[,"n_22"] + x[,"n_34"])*log(39 - 8*r_rm3c + 20*r_rm3c^2 - 12*r_rm3c^3) + (x[,"n_25"] + x[,"n_31"])*log(9 - 10*r_rm3c + 13*r_rm3c^2 - 6*r_rm3c^3) + (x[,"n_14"] + x[,"n_42"])*log(9 + 4*r_rm3c^2 - 4*r_rm3c^3) + (x[,"n_12"] + x[,"n_44"])*log(9 + 4*r_rm3c - 8*r_rm3c^2 + 4*r_rm3c^3) + (x[,"n_21"] + x[,"n_35"])*log(6 + 2*r_rm3c - 5*r_rm3c^2 + 6*r_rm3c^3) + (x[,"n_24"] + x[,"n_32"])*log(39 + 4*r_rm3c - 16*r_rm3c^2 + 12*r_rm3c^3)
logL_rm2c <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*(3*log(2) + log(3) + 2*log(5)) + (n11 + n45)*(log(pmax(1e-6,r)) + log(2 + r) + log(1 + 2*r)) + (n13 + n43)*log(21 + 20*r - 20*r^2) + (n15 + n41)*(log(3 - 2*r) + log(3 - 4*r + r^2)) + (n23 + n33)*log(111 - 20*r + 20*r^2) + (n24 + n32)*log(75 - 40*r + 16*r^2 - 12*r^3) + (n21 + n35)*log(3 + 14*r + r^2 - 6*r^3) + (n12 + n44)*log(3 + 20*r + 8*r^2 - 4*r^3) + (n14 + n42)*log(27 - 24*r - 4*r^2 + 4*r^3) + (n25 + n31)*log(12 + 2*r - 17*r^2 + 6*r^3) + (n22 + n34)*log(39 + 44*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rm2c <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_rm2c,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm2c <- -((x[,"n_15"] + x[,"n_41"])*(-log10(2) + log10(5))) - (x[,"n_11"] + x[,"n_45"])*(-log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_43"])*(log10(2) + log10(13)) - (x[,"n_25"] + x[,"n_31"])*(-log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_35"])*(-log10(2) + log10(19)) - (x[,"n_24"] + x[,"n_32"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_22"] + x[,"n_34"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_14"] + x[,"n_42"])*(-log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_44"])*(-log10(2) + log10(29)) - (x[,"n_23"] + x[,"n_33"])*(log10(2) + log10(53)) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,r_rm2c)) + log10(2 + r_rm2c) + log10(1 + 2*r_rm2c)) + (x[,"n_13"] + x[,"n_43"])*log10(21 + 20*r_rm2c - 20*r_rm2c^2) + (x[,"n_15"] + x[,"n_41"])*(log10(3 - 2*r_rm2c) + log10(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_23"] + x[,"n_33"])*log10(111 - 20*r_rm2c + 20*r_rm2c^2) + (x[,"n_24"] + x[,"n_32"])*log10(75 - 40*r_rm2c + 16*r_rm2c^2 - 12*r_rm2c^3) + (x[,"n_21"] + x[,"n_35"])*log10(3 + 14*r_rm2c + r_rm2c^2 - 6*r_rm2c^3) + (x[,"n_12"] + x[,"n_44"])*log10(3 + 20*r_rm2c + 8*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_14"] + x[,"n_42"])*log10(27 - 24*r_rm2c - 4*r_rm2c^2 + 4*r_rm2c^3) + (x[,"n_25"] + x[,"n_31"])*log10(12 + 2*r_rm2c - 17*r_rm2c^2 + 6*r_rm2c^3) + (x[,"n_22"] + x[,"n_34"])*log10(39 + 44*r_rm2c - 20*r_rm2c^2 + 12*r_rm2c^3)
logL_rm2c <- (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(3*log(2) + log(3) + 2*log(5)) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,r_rm2c)) + log(2 + r_rm2c) + log(1 + 2*r_rm2c)) + (x[,"n_13"] + x[,"n_43"])*log(21 + 20*r_rm2c - 20*r_rm2c^2) + (x[,"n_15"] + x[,"n_41"])*(log(3 - 2*r_rm2c) + log(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_23"] + x[,"n_33"])*log(111 - 20*r_rm2c + 20*r_rm2c^2) + (x[,"n_24"] + x[,"n_32"])*log(75 - 40*r_rm2c + 16*r_rm2c^2 - 12*r_rm2c^3) + (x[,"n_21"] + x[,"n_35"])*log(3 + 14*r_rm2c + r_rm2c^2 - 6*r_rm2c^3) + (x[,"n_12"] + x[,"n_44"])*log(3 + 20*r_rm2c + 8*r_rm2c^2 - 4*r_rm2c^3) + (x[,"n_14"] + x[,"n_42"])*log(27 - 24*r_rm2c - 4*r_rm2c^2 + 4*r_rm2c^3) + (x[,"n_25"] + x[,"n_31"])*log(12 + 2*r_rm2c - 17*r_rm2c^2 + 6*r_rm2c^3) + (x[,"n_22"] + x[,"n_34"])*log(39 + 44*r_rm2c - 20*r_rm2c^2 + 12*r_rm2c^3)

return(list(
r_mat = cbind(r_cc,r_cm3c,r_cm2c,r_rc,r_rm3c,r_rm2c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm3c,LOD_cm2c,LOD_rc,LOD_rm3c,LOD_rm2c,0),
logL_mat = cbind(logL_cc,logL_cm3c,logL_cm2c,logL_rc,logL_rm3c,logL_rm2c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed3c","coupling mixed2c","repulsion coupling","repulsion mixed3c","repulsion mixed2c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.4_3.2 <- function(x){
logL_cc <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(3*log(2) + log(3) + 2*log(5)) + (n15 + n40)*(2*log(pmax(1e-6,r)) + log(2 + r)) + (n25 + n30)*(log(pmax(1e-6,r)) + log(5 + r - 3*r^2)) + (n10 + n45)*(log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n14 + n41)*(log(pmax(1e-6,r)) + log(9 + 11*r + r^2)) + (n22 + n33)*log(111 - 64*r + 52*r^2 - 24*r^3) + (n13 + n42)*log(9 + 30*r - 4*r^2 - 8*r^3) + (n24 + n31)*log(12 + 38*r - 8*r^2 - 3*r^3) + (n11 + n44)*log(21 - 34*r + 14*r^2 - r^3) + (n21 + n34)*log(39 - 13*r - 17*r^2 + 3*r^3) + (n20 + n35)*log(3 + 2*r - 8*r^2 + 3*r^3) + (n12 + n43)*log(27 + 2*r - 28*r^2 + 8*r^3) + (n23 + n32)*log(75 + 32*r - 20*r^2 + 24*r^3)
return(L)}
interlogL_cc <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cc,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cc <- -((x[,"n_15"] + x[,"n_40"])*(-3*log10(2) + log10(5))) - (x[,"n_10"] + x[,"n_45"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_42"])*(log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_43"])*(log10(2) + log10(11)) - (x[,"n_25"] + x[,"n_30"])*(-3*log10(2) + log10(19)) - (x[,"n_20"] + x[,"n_35"])*(-3*log10(2) + log10(19)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(59)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(59)) - (x[,"n_23"] + x[,"n_32"])*log10(89) - (x[,"n_22"] + x[,"n_33"])*log10(89) - (x[,"n_24"] + x[,"n_31"])*(-3*log10(2) + log10(229)) - (x[,"n_21"] + x[,"n_34"])*(-3*log10(2) + log10(229)) + (x[,"n_15"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_25"] + x[,"n_30"])*(log10(pmax(1e-6,r_cc)) + log10(5 + r_cc - 3*r_cc^2)) + (x[,"n_10"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cc)) + log10(3 - 4*r_cc + r_cc^2)) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,r_cc)) + log10(9 + 11*r_cc + r_cc^2)) + (x[,"n_22"] + x[,"n_33"])*log10(111 - 64*r_cc + 52*r_cc^2 - 24*r_cc^3) + (x[,"n_13"] + x[,"n_42"])*log10(9 + 30*r_cc - 4*r_cc^2 - 8*r_cc^3) + (x[,"n_24"] + x[,"n_31"])*log10(12 + 38*r_cc - 8*r_cc^2 - 3*r_cc^3) + (x[,"n_11"] + x[,"n_44"])*log10(21 - 34*r_cc + 14*r_cc^2 - r_cc^3) + (x[,"n_21"] + x[,"n_34"])*log10(39 - 13*r_cc - 17*r_cc^2 + 3*r_cc^3) + (x[,"n_20"] + x[,"n_35"])*log10(3 + 2*r_cc - 8*r_cc^2 + 3*r_cc^3) + (x[,"n_12"] + x[,"n_43"])*log10(27 + 2*r_cc - 28*r_cc^2 + 8*r_cc^3) + (x[,"n_23"] + x[,"n_32"])*log10(75 + 32*r_cc - 20*r_cc^2 + 24*r_cc^3)
logL_cc <- (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(3*log(2) + log(3) + 2*log(5)) + (x[,"n_15"] + x[,"n_40"])*(2*log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_25"] + x[,"n_30"])*(log(pmax(1e-6,r_cc)) + log(5 + r_cc - 3*r_cc^2)) + (x[,"n_10"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cc)) + log(3 - 4*r_cc + r_cc^2)) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,r_cc)) + log(9 + 11*r_cc + r_cc^2)) + (x[,"n_22"] + x[,"n_33"])*log(111 - 64*r_cc + 52*r_cc^2 - 24*r_cc^3) + (x[,"n_13"] + x[,"n_42"])*log(9 + 30*r_cc - 4*r_cc^2 - 8*r_cc^3) + (x[,"n_24"] + x[,"n_31"])*log(12 + 38*r_cc - 8*r_cc^2 - 3*r_cc^3) + (x[,"n_11"] + x[,"n_44"])*log(21 - 34*r_cc + 14*r_cc^2 - r_cc^3) + (x[,"n_21"] + x[,"n_34"])*log(39 - 13*r_cc - 17*r_cc^2 + 3*r_cc^3) + (x[,"n_20"] + x[,"n_35"])*log(3 + 2*r_cc - 8*r_cc^2 + 3*r_cc^3) + (x[,"n_12"] + x[,"n_43"])*log(27 + 2*r_cc - 28*r_cc^2 + 8*r_cc^3) + (x[,"n_23"] + x[,"n_32"])*log(75 + 32*r_cc - 20*r_cc^2 + 24*r_cc^3)
logL_cm <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n10 + n45)*(log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n15 + n40)*(log(pmax(1e-6,r)) + log(1 - r^2)) + (n23 + n32)*log(57 - 4*r + 28*r^2 - 24*r^3) + (n12 + n43)*log(19 - 16*r + 20*r^2 - 8*r^3) + (n20 + n35)*log(3 - 5*r + 6*r^2 - 3*r^3) + (n21 + n34)*log(24 - 14*r + 11*r^2 - 3*r^3) + (n14 + n41)*log(3 + 6*r - 5*r^2 - r^3) + (n11 + n44)*log(3 + 7*r - 8*r^2 + r^3) + (n25 + n30)*log(1 + 2*r - 3*r^2 + 3*r^3) + (n24 + n31)*log(18 + r + 2*r^2 + 3*r^3) + (n13 + n42)*log(15 - 4*r^2 + 8*r^3) + (n22 + n33)*log(57 + 20*r - 44*r^2 + 24*r^3)
return(L)}
interlogL_cm <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cm,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cm <- -((x[,"n_15"] + x[,"n_40"])*(-3*log10(2) + log10(3))) - (x[,"n_10"] + x[,"n_45"])*(-3*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_42"])*(log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_43"])*(log10(3) + log10(5)) - (x[,"n_25"] + x[,"n_30"])*(-3*log10(2) + log10(13)) - (x[,"n_20"] + x[,"n_35"])*(-3*log10(2) + log10(13)) - (x[,"n_24"] + x[,"n_31"])*(-3*log10(2) + log10(5) + log10(31)) - (x[,"n_21"] + x[,"n_34"])*(-3*log10(2) + log10(5) + log10(31)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(37)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(37)) - (x[,"n_23"] + x[,"n_32"])*log10(59) - (x[,"n_22"] + x[,"n_33"])*log10(59) + (x[,"n_10"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cm)) + log10(2 - r_cm) + log10(pmax(1e-6,r_cm))) + (x[,"n_15"] + x[,"n_40"])*(log10(pmax(1e-6,r_cm)) + log10(1 - r_cm^2)) + (x[,"n_23"] + x[,"n_32"])*log10(57 - 4*r_cm + 28*r_cm^2 - 24*r_cm^3) + (x[,"n_12"] + x[,"n_43"])*log10(19 - 16*r_cm + 20*r_cm^2 - 8*r_cm^3) + (x[,"n_20"] + x[,"n_35"])*log10(3 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3) + (x[,"n_21"] + x[,"n_34"])*log10(24 - 14*r_cm + 11*r_cm^2 - 3*r_cm^3) + (x[,"n_14"] + x[,"n_41"])*log10(3 + 6*r_cm - 5*r_cm^2 - r_cm^3) + (x[,"n_11"] + x[,"n_44"])*log10(3 + 7*r_cm - 8*r_cm^2 + r_cm^3) + (x[,"n_25"] + x[,"n_30"])*log10(1 + 2*r_cm - 3*r_cm^2 + 3*r_cm^3) + (x[,"n_24"] + x[,"n_31"])*log10(18 + r_cm + 2*r_cm^2 + 3*r_cm^3) + (x[,"n_13"] + x[,"n_42"])*log10(15 - 4*r_cm^2 + 8*r_cm^3) + (x[,"n_22"] + x[,"n_33"])*log10(57 + 20*r_cm - 44*r_cm^2 + 24*r_cm^3)
logL_cm <- (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_10"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cm)) + log(2 - r_cm) + log(pmax(1e-6,r_cm))) + (x[,"n_15"] + x[,"n_40"])*(log(pmax(1e-6,r_cm)) + log(1 - r_cm^2)) + (x[,"n_23"] + x[,"n_32"])*log(57 - 4*r_cm + 28*r_cm^2 - 24*r_cm^3) + (x[,"n_12"] + x[,"n_43"])*log(19 - 16*r_cm + 20*r_cm^2 - 8*r_cm^3) + (x[,"n_20"] + x[,"n_35"])*log(3 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3) + (x[,"n_21"] + x[,"n_34"])*log(24 - 14*r_cm + 11*r_cm^2 - 3*r_cm^3) + (x[,"n_14"] + x[,"n_41"])*log(3 + 6*r_cm - 5*r_cm^2 - r_cm^3) + (x[,"n_11"] + x[,"n_44"])*log(3 + 7*r_cm - 8*r_cm^2 + r_cm^3) + (x[,"n_25"] + x[,"n_30"])*log(1 + 2*r_cm - 3*r_cm^2 + 3*r_cm^3) + (x[,"n_24"] + x[,"n_31"])*log(18 + r_cm + 2*r_cm^2 + 3*r_cm^3) + (x[,"n_13"] + x[,"n_42"])*log(15 - 4*r_cm^2 + 8*r_cm^3) + (x[,"n_22"] + x[,"n_33"])*log(57 + 20*r_cm - 44*r_cm^2 + 24*r_cm^3)
logL_cr <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n20 - 2*n21 - n22 - n23 - 2*n24 - 2*n25 - 2*n30 - 2*n31 - n32 - n33 - 2*n34 - 2*n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(2) + (n22 + n23 + n32 + n33)*log(3) + 2*(-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n15 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n45)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n44)*(log(pmax(1e-6,r)) + log(2 + 2*r - r^2)) + (n20 + n35)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n13 + n42)*log(6 - 9*r + 12*r^2 - 8*r^3) + (n22 + n33)*log(3 - 3*r + 6*r^2 - 4*r^3) + (n24 + n31)*log(6 - 4*r + 4*r^2 - 3*r^3) + (n25 + n30)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n14 + n41)*log(3 - 3*r - r^2 + r^3) + (n21 + n34)*log(3 + 5*r - 5*r^2 + 3*r^3) + (n23 + n32)*log(2 + 3*r - 6*r^2 + 4*r^3) + (n12 + n43)*log(1 + 9*r - 12*r^2 + 8*r^3)
return(L)}
interlogL_cr <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cr,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cr <- 3*(x[,"n_15"] + x[,"n_40"])*log10(2) + 3*(x[,"n_10"] + x[,"n_45"])*log10(2) + (x[,"n_25"] + x[,"n_30"])*(3*log10(2) - log10(3)) - (x[,"n_20"] + x[,"n_35"])*(-3*log10(2) + log10(3)) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(5)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_42"])*(-log10(2) + log10(7)) - (x[,"n_12"] + x[,"n_43"])*(-log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(11)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(11)) - (x[,"n_24"] + x[,"n_31"])*(-3*log10(2) + log10(37)) - (x[,"n_21"] + x[,"n_34"])*(-3*log10(2) + log10(37)) + (x[,"n_15"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_10"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_cr)) + log10(2 + 2*r_cr - r_cr^2)) + (x[,"n_20"] + x[,"n_35"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_13"] + x[,"n_42"])*log10(6 - 9*r_cr + 12*r_cr^2 - 8*r_cr^3) + (x[,"n_22"] + x[,"n_33"])*log10(3 - 3*r_cr + 6*r_cr^2 - 4*r_cr^3) + (x[,"n_24"] + x[,"n_31"])*log10(6 - 4*r_cr + 4*r_cr^2 - 3*r_cr^3) + (x[,"n_25"] + x[,"n_30"])*log10(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_14"] + x[,"n_41"])*log10(3 - 3*r_cr - r_cr^2 + r_cr^3) + (x[,"n_21"] + x[,"n_34"])*log10(3 + 5*r_cr - 5*r_cr^2 + 3*r_cr^3) + (x[,"n_23"] + x[,"n_32"])*log10(2 + 3*r_cr - 6*r_cr^2 + 4*r_cr^3) + (x[,"n_12"] + x[,"n_43"])*log10(1 + 9*r_cr - 12*r_cr^2 + 8*r_cr^3)
logL_cr <- (-2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 2*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*log(3) + 2*(-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_15"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_10"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_cr)) + log(2 + 2*r_cr - r_cr^2)) + (x[,"n_20"] + x[,"n_35"])*(log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_13"] + x[,"n_42"])*log(6 - 9*r_cr + 12*r_cr^2 - 8*r_cr^3) + (x[,"n_22"] + x[,"n_33"])*log(3 - 3*r_cr + 6*r_cr^2 - 4*r_cr^3) + (x[,"n_24"] + x[,"n_31"])*log(6 - 4*r_cr + 4*r_cr^2 - 3*r_cr^3) + (x[,"n_25"] + x[,"n_30"])*log(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_14"] + x[,"n_41"])*log(3 - 3*r_cr - r_cr^2 + r_cr^3) + (x[,"n_21"] + x[,"n_34"])*log(3 + 5*r_cr - 5*r_cr^2 + 3*r_cr^3) + (x[,"n_23"] + x[,"n_32"])*log(2 + 3*r_cr - 6*r_cr^2 + 4*r_cr^3) + (x[,"n_12"] + x[,"n_43"])*log(1 + 9*r_cr - 12*r_cr^2 + 8*r_cr^3)
logL_rc <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(3*log(2) + log(3) + 2*log(5)) + (n15 + n40)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n10 + n45)*(log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n23 + n32)*log(93 - 16*r + 28*r^2 - 24*r^3) + (n12 + n43)*log(27 - 18*r + 20*r^2 - 8*r^3) + (n20 + n35)*log(3 - 4*r + 7*r^2 - 3*r^3) + (n21 + n34)*log(30 - 7*r + 10*r^2 - 3*r^3) + (n14 + n41)*log(3 + 11*r - 4*r^2 - r^3) + (n11 + n44)*log(9 - 7*r^2 + r^3) + (n25 + n30)*log(3 - r - 2*r^2 + 3*r^3) + (n24 + n31)*log(30 - 4*r + r^2 + 3*r^3) + (n13 + n42)*log(21 + 2*r - 4*r^2 + 8*r^3) + (n22 + n33)*log(81 + 32*r - 44*r^2 + 24*r^3)
return(L)}
interlogL_rc <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rc,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rc <- -((x[,"n_15"] + x[,"n_40"])*(-3*log10(2) + log10(5))) - (x[,"n_10"] + x[,"n_45"])*(-3*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_42"])*(log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_43"])*(log10(2) + log10(11)) - (x[,"n_25"] + x[,"n_30"])*(-3*log10(2) + log10(19)) - (x[,"n_20"] + x[,"n_35"])*(-3*log10(2) + log10(19)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(59)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(59)) - (x[,"n_23"] + x[,"n_32"])*log10(89) - (x[,"n_22"] + x[,"n_33"])*log10(89) - (x[,"n_24"] + x[,"n_31"])*(-3*log10(2) + log10(229)) - (x[,"n_21"] + x[,"n_34"])*(-3*log10(2) + log10(229)) + (x[,"n_15"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc)) + log10(2 + r_rc)) + (x[,"n_10"] + x[,"n_45"])*(log10(pmax(1e-6,r_rc)) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_23"] + x[,"n_32"])*log10(93 - 16*r_rc + 28*r_rc^2 - 24*r_rc^3) + (x[,"n_12"] + x[,"n_43"])*log10(27 - 18*r_rc + 20*r_rc^2 - 8*r_rc^3) + (x[,"n_20"] + x[,"n_35"])*log10(3 - 4*r_rc + 7*r_rc^2 - 3*r_rc^3) + (x[,"n_21"] + x[,"n_34"])*log10(30 - 7*r_rc + 10*r_rc^2 - 3*r_rc^3) + (x[,"n_14"] + x[,"n_41"])*log10(3 + 11*r_rc - 4*r_rc^2 - r_rc^3) + (x[,"n_11"] + x[,"n_44"])*log10(9 - 7*r_rc^2 + r_rc^3) + (x[,"n_25"] + x[,"n_30"])*log10(3 - r_rc - 2*r_rc^2 + 3*r_rc^3) + (x[,"n_24"] + x[,"n_31"])*log10(30 - 4*r_rc + r_rc^2 + 3*r_rc^3) + (x[,"n_13"] + x[,"n_42"])*log10(21 + 2*r_rc - 4*r_rc^2 + 8*r_rc^3) + (x[,"n_22"] + x[,"n_33"])*log10(81 + 32*r_rc - 44*r_rc^2 + 24*r_rc^3)
logL_rc <- (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(3*log(2) + log(3) + 2*log(5)) + (x[,"n_15"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc)) + log(2 + r_rc)) + (x[,"n_10"] + x[,"n_45"])*(log(pmax(1e-6,r_rc)) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_23"] + x[,"n_32"])*log(93 - 16*r_rc + 28*r_rc^2 - 24*r_rc^3) + (x[,"n_12"] + x[,"n_43"])*log(27 - 18*r_rc + 20*r_rc^2 - 8*r_rc^3) + (x[,"n_20"] + x[,"n_35"])*log(3 - 4*r_rc + 7*r_rc^2 - 3*r_rc^3) + (x[,"n_21"] + x[,"n_34"])*log(30 - 7*r_rc + 10*r_rc^2 - 3*r_rc^3) + (x[,"n_14"] + x[,"n_41"])*log(3 + 11*r_rc - 4*r_rc^2 - r_rc^3) + (x[,"n_11"] + x[,"n_44"])*log(9 - 7*r_rc^2 + r_rc^3) + (x[,"n_25"] + x[,"n_30"])*log(3 - r_rc - 2*r_rc^2 + 3*r_rc^3) + (x[,"n_24"] + x[,"n_31"])*log(30 - 4*r_rc + r_rc^2 + 3*r_rc^3) + (x[,"n_13"] + x[,"n_42"])*log(21 + 2*r_rc - 4*r_rc^2 + 8*r_rc^3) + (x[,"n_22"] + x[,"n_33"])*log(81 + 32*r_rc - 44*r_rc^2 + 24*r_rc^3)
logL_rm <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(4*log(2) + 2*log(5)) + (n10 + n45)*(log(2 - r) + 2*log(pmax(1e-6,r))) + (n15 + n40)*(log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n11 + n44)*(log(pmax(1e-6,r)) + log(9 + r - r^2)) + (n20 + n35)*(log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n22 + n33)*log(55 - 4*r + 36*r^2 - 24*r^3) + (n13 + n42)*log(21 - 16*r + 12*r^2 - 8*r^3) + (n25 + n30)*log(3 - 4*r + 4*r^2 - 3*r^3) + (n24 + n31)*log(27 - 17*r + 5*r^2 - 3*r^3) + (n14 + n41)*log(9 - 8*r - 2*r^2 + r^3) + (n21 + n34)*log(12 + 16*r - 4*r^2 + 3*r^3) + (n12 + n43)*log(9 + 16*r - 12*r^2 + 8*r^3) + (n23 + n32)*log(63 + 4*r - 36*r^2 + 24*r^3)
return(L)}
interlogL_rm <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rm,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm <- -((x[,"n_15"] + x[,"n_40"])*(-3*log10(2) + log10(3))) - (x[,"n_10"] + x[,"n_45"])*(-3*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_42"])*(log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_43"])*(log10(3) + log10(5)) - (x[,"n_25"] + x[,"n_30"])*(-3*log10(2) + log10(13)) - (x[,"n_20"] + x[,"n_35"])*(-3*log10(2) + log10(13)) - (x[,"n_24"] + x[,"n_31"])*(-3*log10(2) + log10(5) + log10(31)) - (x[,"n_21"] + x[,"n_34"])*(-3*log10(2) + log10(5) + log10(31)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(37)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(37)) - (x[,"n_23"] + x[,"n_32"])*log10(59) - (x[,"n_22"] + x[,"n_33"])*log10(59) + (x[,"n_10"] + x[,"n_45"])*(log10(2 - r_rm) + 2*log10(pmax(1e-6,r_rm))) + (x[,"n_15"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_rm)) + log10(1 - r_rm^2)) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm)) + log10(9 + r_rm - r_rm^2)) + (x[,"n_20"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm)) + log10(5 - 5*r_rm + 3*r_rm^2)) + (x[,"n_22"] + x[,"n_33"])*log10(55 - 4*r_rm + 36*r_rm^2 - 24*r_rm^3) + (x[,"n_13"] + x[,"n_42"])*log10(21 - 16*r_rm + 12*r_rm^2 - 8*r_rm^3) + (x[,"n_25"] + x[,"n_30"])*log10(3 - 4*r_rm + 4*r_rm^2 - 3*r_rm^3) + (x[,"n_24"] + x[,"n_31"])*log10(27 - 17*r_rm + 5*r_rm^2 - 3*r_rm^3) + (x[,"n_14"] + x[,"n_41"])*log10(9 - 8*r_rm - 2*r_rm^2 + r_rm^3) + (x[,"n_21"] + x[,"n_34"])*log10(12 + 16*r_rm - 4*r_rm^2 + 3*r_rm^3) + (x[,"n_12"] + x[,"n_43"])*log10(9 + 16*r_rm - 12*r_rm^2 + 8*r_rm^3) + (x[,"n_23"] + x[,"n_32"])*log10(63 + 4*r_rm - 36*r_rm^2 + 24*r_rm^3)
logL_rm <- (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(4*log(2) + 2*log(5)) + (x[,"n_10"] + x[,"n_45"])*(log(2 - r_rm) + 2*log(pmax(1e-6,r_rm))) + (x[,"n_15"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_rm)) + log(1 - r_rm^2)) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_rm)) + log(9 + r_rm - r_rm^2)) + (x[,"n_20"] + x[,"n_35"])*(log(pmax(1e-6,r_rm)) + log(5 - 5*r_rm + 3*r_rm^2)) + (x[,"n_22"] + x[,"n_33"])*log(55 - 4*r_rm + 36*r_rm^2 - 24*r_rm^3) + (x[,"n_13"] + x[,"n_42"])*log(21 - 16*r_rm + 12*r_rm^2 - 8*r_rm^3) + (x[,"n_25"] + x[,"n_30"])*log(3 - 4*r_rm + 4*r_rm^2 - 3*r_rm^3) + (x[,"n_24"] + x[,"n_31"])*log(27 - 17*r_rm + 5*r_rm^2 - 3*r_rm^3) + (x[,"n_14"] + x[,"n_41"])*log(9 - 8*r_rm - 2*r_rm^2 + r_rm^3) + (x[,"n_21"] + x[,"n_34"])*log(12 + 16*r_rm - 4*r_rm^2 + 3*r_rm^3) + (x[,"n_12"] + x[,"n_43"])*log(9 + 16*r_rm - 12*r_rm^2 + 8*r_rm^3) + (x[,"n_23"] + x[,"n_32"])*log(63 + 4*r_rm - 36*r_rm^2 + 24*r_rm^3)
logL_rr <- function(r,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n20 - 2*n21 - n22 - n23 - 2*n24 - 2*n25 - 2*n30 - 2*n31 - n32 - n33 - 2*n34 - 2*n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(2) + (n20 + n25 + n30 + n35)*log(3) + 2*(-n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + 3*(n15 + n40)*log(pmax(1e-6,1 - r)) + (n14 + n41)*(2*log(pmax(1e-6,1 - r)) + log(6 - r)) + 3*(n10 + n45)*log(pmax(1e-6,r)) + (n25 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n20 + n35)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n44)*(2*log(pmax(1e-6,r)) + log(5 + r)) + (n12 + n43)*(log(pmax(1e-6,r)) + log(7 + 4*r - 8*r^2)) + (n21 + n34)*(log(pmax(1e-6,r)) + log(11 - 2*r - 3*r^2)) + (n23 + n32)*log(12 - 17*r + 22*r^2 - 12*r^3) + (n24 + n31)*log(6 + 2*r - 11*r^2 + 3*r^3) + (n13 + n42)*log(3 + 9*r - 20*r^2 + 8*r^3) + (n22 + n33)*log(5 + 9*r - 14*r^2 + 12*r^3)
return(L)}
interlogL_rr <- function(n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rr,c(0,0.5), n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rr <- 3*(x[,"n_25"] + x[,"n_30"])*log10(2) + 3*(x[,"n_20"] + x[,"n_35"])*log10(2) + 3*(x[,"n_15"] + x[,"n_40"])*log10(2) + 3*(x[,"n_10"] + x[,"n_45"])*log10(2) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_42"])*(-log10(2) + log10(7)) - (x[,"n_12"] + x[,"n_43"])*(-log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(11)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(11)) - (x[,"n_24"] + x[,"n_31"])*(-3*log10(2) + log10(37)) - (x[,"n_21"] + x[,"n_34"])*(-3*log10(2) + log10(37)) + 3*(x[,"n_15"] + x[,"n_40"])*log10(pmax(1e-6,1 - r_rr)) + (x[,"n_14"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(6 - r_rr)) + 3*(x[,"n_10"] + x[,"n_45"])*log10(pmax(1e-6,r_rr)) + (x[,"n_25"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr))) + (x[,"n_20"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_rr)) + 2*log10(pmax(1e-6,r_rr))) + (x[,"n_11"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(5 + r_rr)) + (x[,"n_12"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(7 + 4*r_rr - 8*r_rr^2)) + (x[,"n_21"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(11 - 2*r_rr - 3*r_rr^2)) + (x[,"n_23"] + x[,"n_32"])*log10(12 - 17*r_rr + 22*r_rr^2 - 12*r_rr^3) + (x[,"n_24"] + x[,"n_31"])*log10(6 + 2*r_rr - 11*r_rr^2 + 3*r_rr^3) + (x[,"n_13"] + x[,"n_42"])*log10(3 + 9*r_rr - 20*r_rr^2 + 8*r_rr^3) + (x[,"n_22"] + x[,"n_33"])*log10(5 + 9*r_rr - 14*r_rr^2 + 12*r_rr^3)
logL_rr <- (-2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 2*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_20"] + x[,"n_25"] + x[,"n_30"] + x[,"n_35"])*log(3) + 2*(-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + 3*(x[,"n_15"] + x[,"n_40"])*log(pmax(1e-6,1 - r_rr)) + (x[,"n_14"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(6 - r_rr)) + 3*(x[,"n_10"] + x[,"n_45"])*log(pmax(1e-6,r_rr)) + (x[,"n_25"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr))) + (x[,"n_20"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_rr)) + 2*log(pmax(1e-6,r_rr))) + (x[,"n_11"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(5 + r_rr)) + (x[,"n_12"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(7 + 4*r_rr - 8*r_rr^2)) + (x[,"n_21"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(11 - 2*r_rr - 3*r_rr^2)) + (x[,"n_23"] + x[,"n_32"])*log(12 - 17*r_rr + 22*r_rr^2 - 12*r_rr^3) + (x[,"n_24"] + x[,"n_31"])*log(6 + 2*r_rr - 11*r_rr^2 + 3*r_rr^3) + (x[,"n_13"] + x[,"n_42"])*log(3 + 9*r_rr - 20*r_rr^2 + 8*r_rr^3) + (x[,"n_22"] + x[,"n_33"])*log(5 + 9*r_rr - 14*r_rr^2 + 12*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.4_3.3 <- function(x){
logL_cc <- function(r,n20,n26,n30,n36,n10,n11,n12,n13,n14,n15,n16,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n46) {
L <- (n20 + n26 + n30 + n36)*log(3) + (-n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(4*log(2) + 2*log(5)) + 3*(n10 + n46)*log(pmax(1e-6,1 - r)) + (n11 + n45)*(log(9 - 2*r) + 2*log(pmax(1e-6,1 - r))) + 3*(n16 + n40)*log(pmax(1e-6,r)) + (n20 + n36)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n26 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n15 + n41)*(2*log(pmax(1e-6,r)) + log(7 + 2*r)) + (n13 + n43)*log(9 + 28*r - 28*r^2) + (n14 + n42)*(log(pmax(1e-6,r)) + log(15 + 13*r - 7*r^2)) + (n25 + n31)*(log(pmax(1e-6,r)) + log(16 - r - 6*r^2)) + (n23 + n33)*log(73 - 28*r + 28*r^2) + (n22 + n34)*log(57 - 44*r + 29*r^2 - 21*r^3) + (n21 + n35)*log(9 + 4*r - 19*r^2 + 6*r^3) + (n12 + n44)*log(21 - 20*r - 8*r^2 + 7*r^3) + (n24 + n32)*log(21 + 49*r - 34*r^2 + 21*r^3)
return(L)}
interlogL_cc <- function(n20,n26,n30,n36,n10,n11,n12,n13,n14,n15,n16,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cc,c(0,0.5), n20,n26,n30,n36,n10,n11,n12,n13,n14,n15,n16,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_20"],x[,"n_26"],x[,"n_30"],x[,"n_36"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cc <- 3*(x[,"n_26"] + x[,"n_30"])*log10(2) + 3*(x[,"n_20"] + x[,"n_36"])*log10(2) + 3*(x[,"n_16"] + x[,"n_40"])*log10(2) - (x[,"n_15"] + x[,"n_41"])*log10(2) - 4*(x[,"n_13"] + x[,"n_43"])*log10(2) - (x[,"n_11"] + x[,"n_45"])*log10(2) + 3*(x[,"n_10"] + x[,"n_46"])*log10(2) - (x[,"n_25"] + x[,"n_31"])*log10(7) - (x[,"n_21"] + x[,"n_35"])*log10(7) - (x[,"n_23"] + x[,"n_33"])*(log10(2) + log10(3) + log10(11)) - (x[,"n_14"] + x[,"n_42"])*(-3*log10(2) + log10(79)) - (x[,"n_12"] + x[,"n_44"])*(-3*log10(2) + log10(79)) - (x[,"n_24"] + x[,"n_32"])*(-3*log10(2) + log10(317)) - (x[,"n_22"] + x[,"n_34"])*(-3*log10(2) + log10(317)) + 3*(x[,"n_10"] + x[,"n_46"])*log10(pmax(1e-6,1 - r_cc)) + (x[,"n_11"] + x[,"n_45"])*(log10(9 - 2*r_cc) + 2*log10(pmax(1e-6,1 - r_cc))) + 3*(x[,"n_16"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_20"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_26"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cc)) + log10(7 + 2*r_cc)) + (x[,"n_13"] + x[,"n_43"])*log10(9 + 28*r_cc - 28*r_cc^2) + (x[,"n_14"] + x[,"n_42"])*(log10(pmax(1e-6,r_cc)) + log10(15 + 13*r_cc - 7*r_cc^2)) + (x[,"n_25"] + x[,"n_31"])*(log10(pmax(1e-6,r_cc)) + log10(16 - r_cc - 6*r_cc^2)) + (x[,"n_23"] + x[,"n_33"])*log10(73 - 28*r_cc + 28*r_cc^2) + (x[,"n_22"] + x[,"n_34"])*log10(57 - 44*r_cc + 29*r_cc^2 - 21*r_cc^3) + (x[,"n_21"] + x[,"n_35"])*log10(9 + 4*r_cc - 19*r_cc^2 + 6*r_cc^3) + (x[,"n_12"] + x[,"n_44"])*log10(21 - 20*r_cc - 8*r_cc^2 + 7*r_cc^3) + (x[,"n_24"] + x[,"n_32"])*log10(21 + 49*r_cc - 34*r_cc^2 + 21*r_cc^3)
logL_cc <- (x[,"n_20"] + x[,"n_26"] + x[,"n_30"] + x[,"n_36"])*log(3) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(4*log(2) + 2*log(5)) + 3*(x[,"n_10"] + x[,"n_46"])*log(pmax(1e-6,1 - r_cc)) + (x[,"n_11"] + x[,"n_45"])*(log(9 - 2*r_cc) + 2*log(pmax(1e-6,1 - r_cc))) + 3*(x[,"n_16"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_20"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_26"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cc)) + log(7 + 2*r_cc)) + (x[,"n_13"] + x[,"n_43"])*log(9 + 28*r_cc - 28*r_cc^2) + (x[,"n_14"] + x[,"n_42"])*(log(pmax(1e-6,r_cc)) + log(15 + 13*r_cc - 7*r_cc^2)) + (x[,"n_25"] + x[,"n_31"])*(log(pmax(1e-6,r_cc)) + log(16 - r_cc - 6*r_cc^2)) + (x[,"n_23"] + x[,"n_33"])*log(73 - 28*r_cc + 28*r_cc^2) + (x[,"n_22"] + x[,"n_34"])*log(57 - 44*r_cc + 29*r_cc^2 - 21*r_cc^3) + (x[,"n_21"] + x[,"n_35"])*log(9 + 4*r_cc - 19*r_cc^2 + 6*r_cc^3) + (x[,"n_12"] + x[,"n_44"])*log(21 - 20*r_cc - 8*r_cc^2 + 7*r_cc^3) + (x[,"n_24"] + x[,"n_32"])*log(21 + 49*r_cc - 34*r_cc^2 + 21*r_cc^3)
logL_cm2c <- function(r,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n23,n33) {
L <- (-3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 4*n21 - 4*n22 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 4*n31 - 4*n32 - 4*n34 - 4*n35 - 4*n36 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 3*n46)*log(2) + (-n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(log(3) + 2*log(5)) + (n10 + n46)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n16 + n40)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n15 + n41)*(log(pmax(1e-6,r)) + log(6 - r - 2*r^2)) + (n23 + n33)*log(12 + r - r^2) + (n26 + n30)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2)) + (n13 + n43)*log(27 - 8*r + 8*r^2) + (n24 + n32)*log(105 + 7*r + 62*r^2 - 42*r^3) + (n21 + n35)*log(33 - 32*r + 26*r^2 - 12*r^3) + (n12 + n44)*log(21 - 16*r + 11*r^2 - 7*r^3) + (n20 + n36)*log(3 - 7*r + 10*r^2 - 6*r^3) + (n11 + n45)*log(3 + 2*r - 7*r^2 + 2*r^3) + (n14 + n42)*log(9 + 15*r - 10*r^2 + 7*r^3) + (n25 + n31)*log(15 + 16*r - 10*r^2 + 12*r^3) + (n22 + n34)*log(132 - 5*r - 64*r^2 + 42*r^3)
return(L)}
interlogL_cm2c <- function(n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n23,n33) {
optimize(logL_cm2c,c(0,0.5), n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n23,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_23"],x[,"n_33"])
LOD_cm2c <- 3*(x[,"n_16"] + x[,"n_40"])*log10(2) + 3*(x[,"n_10"] + x[,"n_46"])*log10(2) - 2*(x[,"n_13"] + x[,"n_43"])*log10(5) - (x[,"n_26"] + x[,"n_30"])*(-2*log10(2) + log10(5)) - (x[,"n_20"] + x[,"n_36"])*(-2*log10(2) + log10(5)) - (x[,"n_15"] + x[,"n_41"])*(-log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_45"])*(-log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_33"])*(-2*log10(2) + 2*log10(7)) - (x[,"n_25"] + x[,"n_31"])*(log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_35"])*(log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_42"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_12"] + x[,"n_44"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_24"] + x[,"n_32"])*(-2*log10(2) + 2*log10(5) + log10(19)) - (x[,"n_22"] + x[,"n_34"])*(-2*log10(2) + 2*log10(5) + log10(19)) + (x[,"n_10"] + x[,"n_46"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_16"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cm2c)) + 2*log10(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm2c)) + log10(6 - r_cm2c - 2*r_cm2c^2)) + (x[,"n_23"] + x[,"n_33"])*log10(12 + r_cm2c - r_cm2c^2) + (x[,"n_26"] + x[,"n_30"])*(log10(pmax(1e-6,r_cm2c)) + log10(5 - 8*r_cm2c + 6*r_cm2c^2)) + (x[,"n_13"] + x[,"n_43"])*log10(27 - 8*r_cm2c + 8*r_cm2c^2) + (x[,"n_24"] + x[,"n_32"])*log10(105 + 7*r_cm2c + 62*r_cm2c^2 - 42*r_cm2c^3) + (x[,"n_21"] + x[,"n_35"])*log10(33 - 32*r_cm2c + 26*r_cm2c^2 - 12*r_cm2c^3) + (x[,"n_12"] + x[,"n_44"])*log10(21 - 16*r_cm2c + 11*r_cm2c^2 - 7*r_cm2c^3) + (x[,"n_20"] + x[,"n_36"])*log10(3 - 7*r_cm2c + 10*r_cm2c^2 - 6*r_cm2c^3) + (x[,"n_11"] + x[,"n_45"])*log10(3 + 2*r_cm2c - 7*r_cm2c^2 + 2*r_cm2c^3) + (x[,"n_14"] + x[,"n_42"])*log10(9 + 15*r_cm2c - 10*r_cm2c^2 + 7*r_cm2c^3) + (x[,"n_25"] + x[,"n_31"])*log10(15 + 16*r_cm2c - 10*r_cm2c^2 + 12*r_cm2c^3) + (x[,"n_22"] + x[,"n_34"])*log10(132 - 5*r_cm2c - 64*r_cm2c^2 + 42*r_cm2c^3)
logL_cm2c <- (-3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_36"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_46"])*log(2) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(log(3) + 2*log(5)) + (x[,"n_10"] + x[,"n_46"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_16"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cm2c)) + 2*log(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_41"])*(log(pmax(1e-6,r_cm2c)) + log(6 - r_cm2c - 2*r_cm2c^2)) + (x[,"n_23"] + x[,"n_33"])*log(12 + r_cm2c - r_cm2c^2) + (x[,"n_26"] + x[,"n_30"])*(log(pmax(1e-6,r_cm2c)) + log(5 - 8*r_cm2c + 6*r_cm2c^2)) + (x[,"n_13"] + x[,"n_43"])*log(27 - 8*r_cm2c + 8*r_cm2c^2) + (x[,"n_24"] + x[,"n_32"])*log(105 + 7*r_cm2c + 62*r_cm2c^2 - 42*r_cm2c^3) + (x[,"n_21"] + x[,"n_35"])*log(33 - 32*r_cm2c + 26*r_cm2c^2 - 12*r_cm2c^3) + (x[,"n_12"] + x[,"n_44"])*log(21 - 16*r_cm2c + 11*r_cm2c^2 - 7*r_cm2c^3) + (x[,"n_20"] + x[,"n_36"])*log(3 - 7*r_cm2c + 10*r_cm2c^2 - 6*r_cm2c^3) + (x[,"n_11"] + x[,"n_45"])*log(3 + 2*r_cm2c - 7*r_cm2c^2 + 2*r_cm2c^3) + (x[,"n_14"] + x[,"n_42"])*log(9 + 15*r_cm2c - 10*r_cm2c^2 + 7*r_cm2c^3) + (x[,"n_25"] + x[,"n_31"])*log(15 + 16*r_cm2c - 10*r_cm2c^2 + 12*r_cm2c^3) + (x[,"n_22"] + x[,"n_34"])*log(132 - 5*r_cm2c - 64*r_cm2c^2 + 42*r_cm2c^3)
logL_cm1c <- function(r,n23,n33,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (n23 + n33)*log(3) + (-n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(4*log(2) + 2*log(5)) + (n16 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n46)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n23 + n33)*log(21 + 4*r - 4*r^2) + (n11 + n45)*(log(pmax(1e-6,r)) + log(4 + r - 2*r^2)) + (n20 + n36)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n13 + n43)*log(19 - 12*r + 12*r^2) + (n22 + n34)*log(42 - 17*r + 35*r^2 - 21*r^3) + (n14 + n42)*log(15 - 12*r + 7*r^2 - 7*r^3) + (n25 + n31)*log(9 - 8*r + 11*r^2 - 6*r^3) + (n26 + n30)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n15 + n41)*log(3 - 5*r^2 + 2*r^3) + (n21 + n35)*log(6 + 4*r - 7*r^2 + 6*r^3) + (n12 + n44)*log(3 + 19*r - 14*r^2 + 7*r^3) + (n24 + n32)*log(39 + 10*r - 28*r^2 + 21*r^3)
return(L)}
interlogL_cm1c <- function(n23,n33,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cm1c,c(0,0.5), n23,n33,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_23"],x[,"n_33"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cm1c <- 3*(x[,"n_16"] + x[,"n_40"])*log10(2) - (x[,"n_15"] + x[,"n_41"])*log10(2) - 4*(x[,"n_13"] + x[,"n_43"])*log10(2) - (x[,"n_11"] + x[,"n_45"])*log10(2) + 3*(x[,"n_10"] + x[,"n_46"])*log10(2) + (x[,"n_26"] + x[,"n_30"])*(3*log10(2) - log10(3)) - (x[,"n_20"] + x[,"n_36"])*(-3*log10(2) + log10(3)) - (x[,"n_25"] + x[,"n_31"])*log10(7) - (x[,"n_21"] + x[,"n_35"])*log10(7) - (x[,"n_23"] + x[,"n_33"])*(log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_42"])*(-3*log10(2) + log10(79)) - (x[,"n_12"] + x[,"n_44"])*(-3*log10(2) + log10(79)) - (x[,"n_24"] + x[,"n_32"])*(-3*log10(2) + log10(317)) - (x[,"n_22"] + x[,"n_34"])*(-3*log10(2) + log10(317)) + (x[,"n_16"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + log10(pmax(1e-6,r_cm1c))) + (x[,"n_10"] + x[,"n_46"])*(log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_23"] + x[,"n_33"])*log10(21 + 4*r_cm1c - 4*r_cm1c^2) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,r_cm1c)) + log10(4 + r_cm1c - 2*r_cm1c^2)) + (x[,"n_20"] + x[,"n_36"])*(log10(pmax(1e-6,r_cm1c)) + log10(2 - 4*r_cm1c + 3*r_cm1c^2)) + (x[,"n_13"] + x[,"n_43"])*log10(19 - 12*r_cm1c + 12*r_cm1c^2) + (x[,"n_22"] + x[,"n_34"])*log10(42 - 17*r_cm1c + 35*r_cm1c^2 - 21*r_cm1c^3) + (x[,"n_14"] + x[,"n_42"])*log10(15 - 12*r_cm1c + 7*r_cm1c^2 - 7*r_cm1c^3) + (x[,"n_25"] + x[,"n_31"])*log10(9 - 8*r_cm1c + 11*r_cm1c^2 - 6*r_cm1c^3) + (x[,"n_26"] + x[,"n_30"])*log10(1 - 3*r_cm1c + 5*r_cm1c^2 - 3*r_cm1c^3) + (x[,"n_15"] + x[,"n_41"])*log10(3 - 5*r_cm1c^2 + 2*r_cm1c^3) + (x[,"n_21"] + x[,"n_35"])*log10(6 + 4*r_cm1c - 7*r_cm1c^2 + 6*r_cm1c^3) + (x[,"n_12"] + x[,"n_44"])*log10(3 + 19*r_cm1c - 14*r_cm1c^2 + 7*r_cm1c^3) + (x[,"n_24"] + x[,"n_32"])*log10(39 + 10*r_cm1c - 28*r_cm1c^2 + 21*r_cm1c^3)
logL_cm1c <- (x[,"n_23"] + x[,"n_33"])*log(3) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(4*log(2) + 2*log(5)) + (x[,"n_16"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_cm1c)) + log(pmax(1e-6,r_cm1c))) + (x[,"n_10"] + x[,"n_46"])*(log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_23"] + x[,"n_33"])*log(21 + 4*r_cm1c - 4*r_cm1c^2) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,r_cm1c)) + log(4 + r_cm1c - 2*r_cm1c^2)) + (x[,"n_20"] + x[,"n_36"])*(log(pmax(1e-6,r_cm1c)) + log(2 - 4*r_cm1c + 3*r_cm1c^2)) + (x[,"n_13"] + x[,"n_43"])*log(19 - 12*r_cm1c + 12*r_cm1c^2) + (x[,"n_22"] + x[,"n_34"])*log(42 - 17*r_cm1c + 35*r_cm1c^2 - 21*r_cm1c^3) + (x[,"n_14"] + x[,"n_42"])*log(15 - 12*r_cm1c + 7*r_cm1c^2 - 7*r_cm1c^3) + (x[,"n_25"] + x[,"n_31"])*log(9 - 8*r_cm1c + 11*r_cm1c^2 - 6*r_cm1c^3) + (x[,"n_26"] + x[,"n_30"])*log(1 - 3*r_cm1c + 5*r_cm1c^2 - 3*r_cm1c^3) + (x[,"n_15"] + x[,"n_41"])*log(3 - 5*r_cm1c^2 + 2*r_cm1c^3) + (x[,"n_21"] + x[,"n_35"])*log(6 + 4*r_cm1c - 7*r_cm1c^2 + 6*r_cm1c^3) + (x[,"n_12"] + x[,"n_44"])*log(3 + 19*r_cm1c - 14*r_cm1c^2 + 7*r_cm1c^3) + (x[,"n_24"] + x[,"n_32"])*log(39 + 10*r_cm1c - 28*r_cm1c^2 + 21*r_cm1c^3)
logL_rc <- function(r,n23,n33,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (n23 + n33)*log(3) + (-n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(4*log(2) + 2*log(5)) + (n10 + n46)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n16 + n40)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n23 + n33)*log(21 + 4*r - 4*r^2) + (n15 + n41)*(log(pmax(1e-6,r)) + log(4 + r - 2*r^2)) + (n26 + n30)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n13 + n43)*log(19 - 12*r + 12*r^2) + (n24 + n32)*log(42 - 17*r + 35*r^2 - 21*r^3) + (n12 + n44)*log(15 - 12*r + 7*r^2 - 7*r^3) + (n21 + n35)*log(9 - 8*r + 11*r^2 - 6*r^3) + (n20 + n36)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n11 + n45)*log(3 - 5*r^2 + 2*r^3) + (n25 + n31)*log(6 + 4*r - 7*r^2 + 6*r^3) + (n14 + n42)*log(3 + 19*r - 14*r^2 + 7*r^3) + (n22 + n34)*log(39 + 10*r - 28*r^2 + 21*r^3)
return(L)}
interlogL_rc <- function(n23,n33,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_rc,c(0,0.5), n23,n33,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_23"],x[,"n_33"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_rc <- 3*(x[,"n_16"] + x[,"n_40"])*log10(2) - (x[,"n_15"] + x[,"n_41"])*log10(2) - 4*(x[,"n_13"] + x[,"n_43"])*log10(2) - (x[,"n_11"] + x[,"n_45"])*log10(2) + 3*(x[,"n_10"] + x[,"n_46"])*log10(2) + (x[,"n_20"] + x[,"n_36"])*(3*log10(2) - log10(3)) - (x[,"n_26"] + x[,"n_30"])*(-3*log10(2) + log10(3)) - (x[,"n_25"] + x[,"n_31"])*log10(7) - (x[,"n_21"] + x[,"n_35"])*log10(7) - (x[,"n_23"] + x[,"n_33"])*(log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_42"])*(-3*log10(2) + log10(79)) - (x[,"n_12"] + x[,"n_44"])*(-3*log10(2) + log10(79)) - (x[,"n_24"] + x[,"n_32"])*(-3*log10(2) + log10(317)) - (x[,"n_22"] + x[,"n_34"])*(-3*log10(2) + log10(317)) + (x[,"n_10"] + x[,"n_46"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc))) + (x[,"n_16"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_rc)) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_23"] + x[,"n_33"])*log10(21 + 4*r_rc - 4*r_rc^2) + (x[,"n_15"] + x[,"n_41"])*(log10(pmax(1e-6,r_rc)) + log10(4 + r_rc - 2*r_rc^2)) + (x[,"n_26"] + x[,"n_30"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 4*r_rc + 3*r_rc^2)) + (x[,"n_13"] + x[,"n_43"])*log10(19 - 12*r_rc + 12*r_rc^2) + (x[,"n_24"] + x[,"n_32"])*log10(42 - 17*r_rc + 35*r_rc^2 - 21*r_rc^3) + (x[,"n_12"] + x[,"n_44"])*log10(15 - 12*r_rc + 7*r_rc^2 - 7*r_rc^3) + (x[,"n_21"] + x[,"n_35"])*log10(9 - 8*r_rc + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_20"] + x[,"n_36"])*log10(1 - 3*r_rc + 5*r_rc^2 - 3*r_rc^3) + (x[,"n_11"] + x[,"n_45"])*log10(3 - 5*r_rc^2 + 2*r_rc^3) + (x[,"n_25"] + x[,"n_31"])*log10(6 + 4*r_rc - 7*r_rc^2 + 6*r_rc^3) + (x[,"n_14"] + x[,"n_42"])*log10(3 + 19*r_rc - 14*r_rc^2 + 7*r_rc^3) + (x[,"n_22"] + x[,"n_34"])*log10(39 + 10*r_rc - 28*r_rc^2 + 21*r_rc^3)
logL_rc <- (x[,"n_23"] + x[,"n_33"])*log(3) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(4*log(2) + 2*log(5)) + (x[,"n_10"] + x[,"n_46"])*(2*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc))) + (x[,"n_16"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_rc)) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_23"] + x[,"n_33"])*log(21 + 4*r_rc - 4*r_rc^2) + (x[,"n_15"] + x[,"n_41"])*(log(pmax(1e-6,r_rc)) + log(4 + r_rc - 2*r_rc^2)) + (x[,"n_26"] + x[,"n_30"])*(log(pmax(1e-6,r_rc)) + log(2 - 4*r_rc + 3*r_rc^2)) + (x[,"n_13"] + x[,"n_43"])*log(19 - 12*r_rc + 12*r_rc^2) + (x[,"n_24"] + x[,"n_32"])*log(42 - 17*r_rc + 35*r_rc^2 - 21*r_rc^3) + (x[,"n_12"] + x[,"n_44"])*log(15 - 12*r_rc + 7*r_rc^2 - 7*r_rc^3) + (x[,"n_21"] + x[,"n_35"])*log(9 - 8*r_rc + 11*r_rc^2 - 6*r_rc^3) + (x[,"n_20"] + x[,"n_36"])*log(1 - 3*r_rc + 5*r_rc^2 - 3*r_rc^3) + (x[,"n_11"] + x[,"n_45"])*log(3 - 5*r_rc^2 + 2*r_rc^3) + (x[,"n_25"] + x[,"n_31"])*log(6 + 4*r_rc - 7*r_rc^2 + 6*r_rc^3) + (x[,"n_14"] + x[,"n_42"])*log(3 + 19*r_rc - 14*r_rc^2 + 7*r_rc^3) + (x[,"n_22"] + x[,"n_34"])*log(39 + 10*r_rc - 28*r_rc^2 + 21*r_rc^3)
logL_rm2c <- function(r,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n23,n33) {
L <- (-3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 4*n21 - 4*n22 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 4*n31 - 4*n32 - 4*n34 - 4*n35 - 4*n36 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 3*n46)*log(2) + (-n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(log(3) + 2*log(5)) + (n16 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n46)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n45)*(log(pmax(1e-6,r)) + log(6 - r - 2*r^2)) + (n23 + n33)*log(12 + r - r^2) + (n20 + n36)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2)) + (n13 + n43)*log(27 - 8*r + 8*r^2) + (n22 + n34)*log(105 + 7*r + 62*r^2 - 42*r^3) + (n25 + n31)*log(33 - 32*r + 26*r^2 - 12*r^3) + (n14 + n42)*log(21 - 16*r + 11*r^2 - 7*r^3) + (n26 + n30)*log(3 - 7*r + 10*r^2 - 6*r^3) + (n15 + n41)*log(3 + 2*r - 7*r^2 + 2*r^3) + (n12 + n44)*log(9 + 15*r - 10*r^2 + 7*r^3) + (n21 + n35)*log(15 + 16*r - 10*r^2 + 12*r^3) + (n24 + n32)*log(132 - 5*r - 64*r^2 + 42*r^3)
return(L)}
interlogL_rm2c <- function(n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n23,n33) {
optimize(logL_rm2c,c(0,0.5), n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n23,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_23"],x[,"n_33"])
LOD_rm2c <- 3*(x[,"n_16"] + x[,"n_40"])*log10(2) + 3*(x[,"n_10"] + x[,"n_46"])*log10(2) - 2*(x[,"n_13"] + x[,"n_43"])*log10(5) - (x[,"n_26"] + x[,"n_30"])*(-2*log10(2) + log10(5)) - (x[,"n_20"] + x[,"n_36"])*(-2*log10(2) + log10(5)) - (x[,"n_15"] + x[,"n_41"])*(-log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_45"])*(-log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_33"])*(-2*log10(2) + 2*log10(7)) - (x[,"n_25"] + x[,"n_31"])*(log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_35"])*(log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_42"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_12"] + x[,"n_44"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_24"] + x[,"n_32"])*(-2*log10(2) + 2*log10(5) + log10(19)) - (x[,"n_22"] + x[,"n_34"])*(-2*log10(2) + 2*log10(5) + log10(19)) + (x[,"n_16"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c))) + (x[,"n_10"] + x[,"n_46"])*(log10(pmax(1e-6,1 - r_rm2c)) + 2*log10(pmax(1e-6,r_rm2c))) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,r_rm2c)) + log10(6 - r_rm2c - 2*r_rm2c^2)) + (x[,"n_23"] + x[,"n_33"])*log10(12 + r_rm2c - r_rm2c^2) + (x[,"n_20"] + x[,"n_36"])*(log10(pmax(1e-6,r_rm2c)) + log10(5 - 8*r_rm2c + 6*r_rm2c^2)) + (x[,"n_13"] + x[,"n_43"])*log10(27 - 8*r_rm2c + 8*r_rm2c^2) + (x[,"n_22"] + x[,"n_34"])*log10(105 + 7*r_rm2c + 62*r_rm2c^2 - 42*r_rm2c^3) + (x[,"n_25"] + x[,"n_31"])*log10(33 - 32*r_rm2c + 26*r_rm2c^2 - 12*r_rm2c^3) + (x[,"n_14"] + x[,"n_42"])*log10(21 - 16*r_rm2c + 11*r_rm2c^2 - 7*r_rm2c^3) + (x[,"n_26"] + x[,"n_30"])*log10(3 - 7*r_rm2c + 10*r_rm2c^2 - 6*r_rm2c^3) + (x[,"n_15"] + x[,"n_41"])*log10(3 + 2*r_rm2c - 7*r_rm2c^2 + 2*r_rm2c^3) + (x[,"n_12"] + x[,"n_44"])*log10(9 + 15*r_rm2c - 10*r_rm2c^2 + 7*r_rm2c^3) + (x[,"n_21"] + x[,"n_35"])*log10(15 + 16*r_rm2c - 10*r_rm2c^2 + 12*r_rm2c^3) + (x[,"n_24"] + x[,"n_32"])*log10(132 - 5*r_rm2c - 64*r_rm2c^2 + 42*r_rm2c^3)
logL_rm2c <- (-3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_36"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_46"])*log(2) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(log(3) + 2*log(5)) + (x[,"n_16"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c))) + (x[,"n_10"] + x[,"n_46"])*(log(pmax(1e-6,1 - r_rm2c)) + 2*log(pmax(1e-6,r_rm2c))) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,r_rm2c)) + log(6 - r_rm2c - 2*r_rm2c^2)) + (x[,"n_23"] + x[,"n_33"])*log(12 + r_rm2c - r_rm2c^2) + (x[,"n_20"] + x[,"n_36"])*(log(pmax(1e-6,r_rm2c)) + log(5 - 8*r_rm2c + 6*r_rm2c^2)) + (x[,"n_13"] + x[,"n_43"])*log(27 - 8*r_rm2c + 8*r_rm2c^2) + (x[,"n_22"] + x[,"n_34"])*log(105 + 7*r_rm2c + 62*r_rm2c^2 - 42*r_rm2c^3) + (x[,"n_25"] + x[,"n_31"])*log(33 - 32*r_rm2c + 26*r_rm2c^2 - 12*r_rm2c^3) + (x[,"n_14"] + x[,"n_42"])*log(21 - 16*r_rm2c + 11*r_rm2c^2 - 7*r_rm2c^3) + (x[,"n_26"] + x[,"n_30"])*log(3 - 7*r_rm2c + 10*r_rm2c^2 - 6*r_rm2c^3) + (x[,"n_15"] + x[,"n_41"])*log(3 + 2*r_rm2c - 7*r_rm2c^2 + 2*r_rm2c^3) + (x[,"n_12"] + x[,"n_44"])*log(9 + 15*r_rm2c - 10*r_rm2c^2 + 7*r_rm2c^3) + (x[,"n_21"] + x[,"n_35"])*log(15 + 16*r_rm2c - 10*r_rm2c^2 + 12*r_rm2c^3) + (x[,"n_24"] + x[,"n_32"])*log(132 - 5*r_rm2c - 64*r_rm2c^2 + 42*r_rm2c^3)
logL_rm1c <- function(r,n20,n26,n30,n36,n10,n11,n12,n13,n14,n15,n16,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n46) {
L <- (n20 + n26 + n30 + n36)*log(3) + (-n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(4*log(2) + 2*log(5)) + 3*(n16 + n40)*log(pmax(1e-6,1 - r)) + (n15 + n41)*(log(9 - 2*r) + 2*log(pmax(1e-6,1 - r))) + 3*(n10 + n46)*log(pmax(1e-6,r)) + (n26 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n20 + n36)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n45)*(2*log(pmax(1e-6,r)) + log(7 + 2*r)) + (n13 + n43)*log(9 + 28*r - 28*r^2) + (n12 + n44)*(log(pmax(1e-6,r)) + log(15 + 13*r - 7*r^2)) + (n21 + n35)*(log(pmax(1e-6,r)) + log(16 - r - 6*r^2)) + (n23 + n33)*log(73 - 28*r + 28*r^2) + (n24 + n32)*log(57 - 44*r + 29*r^2 - 21*r^3) + (n25 + n31)*log(9 + 4*r - 19*r^2 + 6*r^3) + (n14 + n42)*log(21 - 20*r - 8*r^2 + 7*r^3) + (n22 + n34)*log(21 + 49*r - 34*r^2 + 21*r^3)
return(L)}
interlogL_rm1c <- function(n20,n26,n30,n36,n10,n11,n12,n13,n14,n15,n16,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_rm1c,c(0,0.5), n20,n26,n30,n36,n10,n11,n12,n13,n14,n15,n16,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_20"],x[,"n_26"],x[,"n_30"],x[,"n_36"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_rm1c <- 3*(x[,"n_26"] + x[,"n_30"])*log10(2) + 3*(x[,"n_20"] + x[,"n_36"])*log10(2) + 3*(x[,"n_16"] + x[,"n_40"])*log10(2) - (x[,"n_15"] + x[,"n_41"])*log10(2) - 4*(x[,"n_13"] + x[,"n_43"])*log10(2) - (x[,"n_11"] + x[,"n_45"])*log10(2) + 3*(x[,"n_10"] + x[,"n_46"])*log10(2) - (x[,"n_25"] + x[,"n_31"])*log10(7) - (x[,"n_21"] + x[,"n_35"])*log10(7) - (x[,"n_23"] + x[,"n_33"])*(log10(2) + log10(3) + log10(11)) - (x[,"n_14"] + x[,"n_42"])*(-3*log10(2) + log10(79)) - (x[,"n_12"] + x[,"n_44"])*(-3*log10(2) + log10(79)) - (x[,"n_24"] + x[,"n_32"])*(-3*log10(2) + log10(317)) - (x[,"n_22"] + x[,"n_34"])*(-3*log10(2) + log10(317)) + 3*(x[,"n_16"] + x[,"n_40"])*log10(pmax(1e-6,1 - r_rm1c)) + (x[,"n_15"] + x[,"n_41"])*(log10(9 - 2*r_rm1c) + 2*log10(pmax(1e-6,1 - r_rm1c))) + 3*(x[,"n_10"] + x[,"n_46"])*log10(pmax(1e-6,r_rm1c)) + (x[,"n_26"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(pmax(1e-6,r_rm1c))) + (x[,"n_20"] + x[,"n_36"])*(log10(pmax(1e-6,1 - r_rm1c)) + 2*log10(pmax(1e-6,r_rm1c))) + (x[,"n_11"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(7 + 2*r_rm1c)) + (x[,"n_13"] + x[,"n_43"])*log10(9 + 28*r_rm1c - 28*r_rm1c^2) + (x[,"n_12"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm1c)) + log10(15 + 13*r_rm1c - 7*r_rm1c^2)) + (x[,"n_21"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm1c)) + log10(16 - r_rm1c - 6*r_rm1c^2)) + (x[,"n_23"] + x[,"n_33"])*log10(73 - 28*r_rm1c + 28*r_rm1c^2) + (x[,"n_24"] + x[,"n_32"])*log10(57 - 44*r_rm1c + 29*r_rm1c^2 - 21*r_rm1c^3) + (x[,"n_25"] + x[,"n_31"])*log10(9 + 4*r_rm1c - 19*r_rm1c^2 + 6*r_rm1c^3) + (x[,"n_14"] + x[,"n_42"])*log10(21 - 20*r_rm1c - 8*r_rm1c^2 + 7*r_rm1c^3) + (x[,"n_22"] + x[,"n_34"])*log10(21 + 49*r_rm1c - 34*r_rm1c^2 + 21*r_rm1c^3)
logL_rm1c <- (x[,"n_20"] + x[,"n_26"] + x[,"n_30"] + x[,"n_36"])*log(3) + (-x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(4*log(2) + 2*log(5)) + 3*(x[,"n_16"] + x[,"n_40"])*log(pmax(1e-6,1 - r_rm1c)) + (x[,"n_15"] + x[,"n_41"])*(log(9 - 2*r_rm1c) + 2*log(pmax(1e-6,1 - r_rm1c))) + 3*(x[,"n_10"] + x[,"n_46"])*log(pmax(1e-6,r_rm1c)) + (x[,"n_26"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(pmax(1e-6,r_rm1c))) + (x[,"n_20"] + x[,"n_36"])*(log(pmax(1e-6,1 - r_rm1c)) + 2*log(pmax(1e-6,r_rm1c))) + (x[,"n_11"] + x[,"n_45"])*(2*log(pmax(1e-6,r_rm1c)) + log(7 + 2*r_rm1c)) + (x[,"n_13"] + x[,"n_43"])*log(9 + 28*r_rm1c - 28*r_rm1c^2) + (x[,"n_12"] + x[,"n_44"])*(log(pmax(1e-6,r_rm1c)) + log(15 + 13*r_rm1c - 7*r_rm1c^2)) + (x[,"n_21"] + x[,"n_35"])*(log(pmax(1e-6,r_rm1c)) + log(16 - r_rm1c - 6*r_rm1c^2)) + (x[,"n_23"] + x[,"n_33"])*log(73 - 28*r_rm1c + 28*r_rm1c^2) + (x[,"n_24"] + x[,"n_32"])*log(57 - 44*r_rm1c + 29*r_rm1c^2 - 21*r_rm1c^3) + (x[,"n_25"] + x[,"n_31"])*log(9 + 4*r_rm1c - 19*r_rm1c^2 + 6*r_rm1c^3) + (x[,"n_14"] + x[,"n_42"])*log(21 - 20*r_rm1c - 8*r_rm1c^2 + 7*r_rm1c^3) + (x[,"n_22"] + x[,"n_34"])*log(21 + 49*r_rm1c - 34*r_rm1c^2 + 21*r_rm1c^3)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_rc,r_rm2c,r_rm1c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_rc,LOD_rm2c,LOD_rm1c,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_rc,logL_rm2c,logL_rm1c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","repulsion coupling","repulsion mixed2c","repulsion mixed1c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.4_4.1 <- function(x){
logL_cc <- function(r,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*(4*log(2) + 2*log(5)) + 2*(n11 + n44)*log(3 - 2*r) + 2*(n14 + n41)*log(1 + 2*r) + (n12 + n21 + n34 + n43)*log(21 - 8*r - 4*r^2) + (n13 + n24 + n31 + n42)*log(9 + 16*r - 4*r^2) + (n22 + n33)*log(73 - 20*r + 4*r^2) + (n23 + n32)*log(57 + 12*r + 4*r^2)
return(L)}
interlogL_cc <- function(n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_cc,c(0,0.5), n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cc <- -6*(x[,"n_23"] + x[,"n_32"])*log10(2) - 6*(x[,"n_22"] + x[,"n_33"])*log10(2) - 2*(x[,"n_14"] + x[,"n_41"])*log10(2) - 4*(x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(2) - 4*(x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(2) - 2*(x[,"n_11"] + x[,"n_44"])*log10(2) + 2*(x[,"n_11"] + x[,"n_44"])*log10(3 - 2*r_cc) + 2*(x[,"n_14"] + x[,"n_41"])*log10(1 + 2*r_cc) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(21 - 8*r_cc - 4*r_cc^2) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(9 + 16*r_cc - 4*r_cc^2) + (x[,"n_22"] + x[,"n_33"])*log10(73 - 20*r_cc + 4*r_cc^2) + (x[,"n_23"] + x[,"n_32"])*log10(57 + 12*r_cc + 4*r_cc^2)
logL_cc <- (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(4*log(2) + 2*log(5)) + 2*(x[,"n_11"] + x[,"n_44"])*log(3 - 2*r_cc) + 2*(x[,"n_14"] + x[,"n_41"])*log(1 + 2*r_cc) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(21 - 8*r_cc - 4*r_cc^2) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(9 + 16*r_cc - 4*r_cc^2) + (x[,"n_22"] + x[,"n_33"])*log(73 - 20*r_cc + 4*r_cc^2) + (x[,"n_23"] + x[,"n_32"])*log(57 + 12*r_cc + 4*r_cc^2)
logL_cr <- function(r,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-2*n11 - 2*n12 - 2*n13 - 2*n14 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n41 - 2*n42 - 2*n43 - 2*n44)*log(2) + 2*(-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*log(5) + (n11 + n44)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n14 + n41)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n23 + n32)*log(33 - 4*r^2) + (n22 + n33)*log(29 + 8*r - 4*r^2) + (n13 + n42)*log(6 - 5*r + 2*r^2) + (n12 + n43)*log(3 + r + 2*r^2) + (n21 + n24 + n31 + n34)*log(9 - 4*r + 4*r^2)
return(L)}
interlogL_cr <- function(n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_cr,c(0,0.5), n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cr <- -5*(x[,"n_23"] + x[,"n_32"])*log10(2) - 5*(x[,"n_22"] + x[,"n_33"])*log10(2) - 3*(x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"])*log10(2) - 2*(x[,"n_13"] + x[,"n_42"])*log10(2) - 2*(x[,"n_12"] + x[,"n_43"])*log10(2) + (x[,"n_11"] + x[,"n_44"])*(log10(3 - 2*r_cr) + log10(pmax(1e-6,r_cr))) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,1 - r_cr)) + log10(1 + 2*r_cr)) + (x[,"n_23"] + x[,"n_32"])*log10(33 - 4*r_cr^2) + (x[,"n_22"] + x[,"n_33"])*log10(29 + 8*r_cr - 4*r_cr^2) + (x[,"n_13"] + x[,"n_42"])*log10(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_12"] + x[,"n_43"])*log10(3 + r_cr + 2*r_cr^2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"])*log10(9 - 4*r_cr + 4*r_cr^2)
logL_cr <- (-2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"])*log(2) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_11"] + x[,"n_44"])*(log(3 - 2*r_cr) + log(pmax(1e-6,r_cr))) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,1 - r_cr)) + log(1 + 2*r_cr)) + (x[,"n_23"] + x[,"n_32"])*log(33 - 4*r_cr^2) + (x[,"n_22"] + x[,"n_33"])*log(29 + 8*r_cr - 4*r_cr^2) + (x[,"n_13"] + x[,"n_42"])*log(6 - 5*r_cr + 2*r_cr^2) + (x[,"n_12"] + x[,"n_43"])*log(3 + r_cr + 2*r_cr^2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"])*log(9 - 4*r_cr + 4*r_cr^2)
logL_rc <- function(r,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
L <- (-2*n11 - 3*n12 - 3*n13 - 2*n14 - 2*n21 - 3*n22 - 3*n23 - 2*n24 - 2*n31 - 3*n32 - 3*n33 - 2*n34 - 2*n41 - 3*n42 - 3*n43 - 2*n44)*log(2) + 2*(-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*log(5) + (n11 + n44)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n14 + n41)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n23 + n32)*log(33 - 4*r^2) + (n22 + n33)*log(29 + 8*r - 4*r^2) + (n24 + n31)*log(6 - 5*r + 2*r^2) + (n21 + n34)*log(3 + r + 2*r^2) + (n12 + n13 + n42 + n43)*log(9 - 4*r + 4*r^2)
return(L)}
interlogL_rc <- function(n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44) {
optimize(logL_rc,c(0,0.5), n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rc <- -2*(x[,"n_24"] + x[,"n_31"])*log10(2) - 5*(x[,"n_23"] + x[,"n_32"])*log10(2) - 5*(x[,"n_22"] + x[,"n_33"])*log10(2) - 2*(x[,"n_21"] + x[,"n_34"])*log10(2) - 3*(x[,"n_12"] + x[,"n_13"] + x[,"n_42"] + x[,"n_43"])*log10(2) + (x[,"n_11"] + x[,"n_44"])*(log10(3 - 2*r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_23"] + x[,"n_32"])*log10(33 - 4*r_rc^2) + (x[,"n_22"] + x[,"n_33"])*log10(29 + 8*r_rc - 4*r_rc^2) + (x[,"n_24"] + x[,"n_31"])*log10(6 - 5*r_rc + 2*r_rc^2) + (x[,"n_21"] + x[,"n_34"])*log10(3 + r_rc + 2*r_rc^2) + (x[,"n_12"] + x[,"n_13"] + x[,"n_42"] + x[,"n_43"])*log10(9 - 4*r_rc + 4*r_rc^2)
logL_rc <- (-2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_11"] + x[,"n_44"])*(log(3 - 2*r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_23"] + x[,"n_32"])*log(33 - 4*r_rc^2) + (x[,"n_22"] + x[,"n_33"])*log(29 + 8*r_rc - 4*r_rc^2) + (x[,"n_24"] + x[,"n_31"])*log(6 - 5*r_rc + 2*r_rc^2) + (x[,"n_21"] + x[,"n_34"])*log(3 + r_rc + 2*r_rc^2) + (x[,"n_12"] + x[,"n_13"] + x[,"n_42"] + x[,"n_43"])*log(9 - 4*r_rc + 4*r_rc^2)
logL_rr <- function(r,n12,n13,n21,n22,n23,n24,n31,n32,n33,n34,n42,n43,n11,n14,n41,n44) {
L <- (-n12 - n13 - n21 - 2*n22 - 2*n23 - n24 - n31 - 2*n32 - 2*n33 - n34 - n42 - n43)*log(2) + 2*(-n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34 - n41 - n42 - n43 - n44)*log(5) + 2*(n14 + n41)*log(pmax(1e-6,1 - r)) + 2*(n11 + n44)*log(pmax(1e-6,r)) + (n12 + n21 + n34 + n43)*(log(5 - 2*r) + log(pmax(1e-6,r))) + (n13 + n24 + n31 + n42)*log(3 - r - 2*r^2) + (n23 + n32)*log(21 - 12*r + 4*r^2) + (n22 + n33)*log(13 + 4*r + 4*r^2)
return(L)}
interlogL_rr <- function(n12,n13,n21,n22,n23,n24,n31,n32,n33,n34,n42,n43,n11,n14,n41,n44) {
optimize(logL_rr,c(0,0.5), n12,n13,n21,n22,n23,n24,n31,n32,n33,n34,n42,n43,n11,n14,n41,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_12"],x[,"n_13"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_11"],x[,"n_14"],x[,"n_41"],x[,"n_44"])
LOD_rr <- -4*(x[,"n_23"] + x[,"n_32"])*log10(2) - 4*(x[,"n_22"] + x[,"n_33"])*log10(2) + 2*(x[,"n_14"] + x[,"n_41"])*log10(2) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(2) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(2) + 2*(x[,"n_11"] + x[,"n_44"])*log10(2) + 2*(x[,"n_14"] + x[,"n_41"])*log10(pmax(1e-6,1 - r_rr)) + 2*(x[,"n_11"] + x[,"n_44"])*log10(pmax(1e-6,r_rr)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(log10(5 - 2*r_rr) + log10(pmax(1e-6,r_rr))) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(3 - r_rr - 2*r_rr^2) + (x[,"n_23"] + x[,"n_32"])*log10(21 - 12*r_rr + 4*r_rr^2) + (x[,"n_22"] + x[,"n_33"])*log10(13 + 4*r_rr + 4*r_rr^2)
logL_rr <- (-x[,"n_12"] - x[,"n_13"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"])*log(2) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + 2*(x[,"n_14"] + x[,"n_41"])*log(pmax(1e-6,1 - r_rr)) + 2*(x[,"n_11"] + x[,"n_44"])*log(pmax(1e-6,r_rr)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(log(5 - 2*r_rr) + log(pmax(1e-6,r_rr))) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(3 - r_rr - 2*r_rr^2) + (x[,"n_23"] + x[,"n_32"])*log(21 - 12*r_rr + 4*r_rr^2) + (x[,"n_22"] + x[,"n_33"])*log(13 + 4*r_rr + 4*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.5_1.5 <- function(x){
logL_cc <- function(r,n22,n23,n24,n32,n34,n42,n43,n44,n33) {
L <- (-2*n22 - n23 - 2*n24 - n32 - n34 - 2*n42 - n43 - 2*n44)*log(2) + 2*(n22 + n44)*log(pmax(1e-6,1 - r)) + 2*(n24 + n42)*log(pmax(1e-6,r)) + (n23 + n32 + n34 + n43)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n33*log(1/2 - r + r^2)
return(L)}
interlogL_cc <- function(n22,n23,n24,n32,n34,n42,n43,n44,n33) {
optimize(logL_cc,c(0,0.5), n22,n23,n24,n32,n34,n42,n43,n44,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_33"])
LOD_cc <- 2*x[,"n_33"]*log10(2) + 2*(x[,"n_24"] + x[,"n_42"])*log10(2) + 2*(x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(2) + 2*(x[,"n_22"] + x[,"n_44"])*log10(2) + 2*(x[,"n_22"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_24"] + x[,"n_42"])*log10(pmax(1e-6,r_cc)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + x[,"n_33"]*log10(1/2 - r_cc + r_cc^2)
logL_cc <- (-2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - x[,"n_32"] - x[,"n_34"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"])*log(2) + 2*(x[,"n_22"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_24"] + x[,"n_42"])*log(pmax(1e-6,r_cc)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + x[,"n_33"]*log(1/2 - r_cc + r_cc^2)
logL_cm <- function(r,n22,n23,n24,n32,n34,n42,n43,n44,n33) {
L <- 2*(-n22 - n23 - n24 - n32 - n34 - n42 - n43 - n44)*log(2) + (-n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*log(5) + (n24 + n42)*(log(3 - r) + log(pmax(1e-6,r))) + (n22 + n44)*(log(pmax(1e-6,1 - r)) + log(2 + r)) + n33*log(1 + r - r^2) + (n23 + n32 + n34 + n43)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_cm <- function(n22,n23,n24,n32,n34,n42,n43,n44,n33) {
optimize(logL_cm,c(0,0.5), n22,n23,n24,n32,n34,n42,n43,n44,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_33"])
LOD_cm <- -(x[,"n_33"]*(-2*log10(2) + log10(5))) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(5)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-log10(2) + log10(5)) + (x[,"n_24"] + x[,"n_42"])*(log10(3 - r_cm) + log10(pmax(1e-6,r_cm))) + (x[,"n_22"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cm)) + log10(2 + r_cm)) + x[,"n_33"]*log10(1 + r_cm - r_cm^2) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(3 - 2*r_cm + 2*r_cm^2)
logL_cm <- 2*(-x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(2) + (-x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_24"] + x[,"n_42"])*(log(3 - r_cm) + log(pmax(1e-6,r_cm))) + (x[,"n_22"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cm)) + log(2 + r_cm)) + x[,"n_33"]*log(1 + r_cm - r_cm^2) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(3 - 2*r_cm + 2*r_cm^2)
logL_rm <- function(r,n22,n23,n24,n32,n33,n34,n42,n43,n44) {
L <- (-2*n22 - n23 - 2*n24 - n32 - n33 - n34 - 2*n42 - n43 - 2*n44)*log(2) + 2*(-n22 - n23 - n24 - n32 - n33 - n34 - n42 - n43 - n44)*log(5) + 2*(n24 + n42)*log(3 - r) + 2*(n22 + n44)*log(2 + r) + (n23 + n32 + n34 + n43)*(log(3 - r) + log(2 + r)) + n33*log(13 - 2*r + 2*r^2)
return(L)}
interlogL_rm <- function(n22,n23,n24,n32,n33,n34,n42,n43,n44) {
optimize(logL_rm,c(0,0.5), n22,n23,n24,n32,n33,n34,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rm <- -2*(x[,"n_24"] + x[,"n_42"])*(-log10(2) + log10(5)) - 2*(x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-log10(2) + log10(5)) - 2*(x[,"n_22"] + x[,"n_44"])*(-log10(2) + log10(5)) - x[,"n_33"]*(-log10(2) + 2*log10(5)) + 2*(x[,"n_24"] + x[,"n_42"])*log10(3 - r_rm) + 2*(x[,"n_22"] + x[,"n_44"])*log10(2 + r_rm) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log10(3 - r_rm) + log10(2 + r_rm)) + x[,"n_33"]*log10(13 - 2*r_rm + 2*r_rm^2)
logL_rm <- (-2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"])*log(2) + 2*(-x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + 2*(x[,"n_24"] + x[,"n_42"])*log(3 - r_rm) + 2*(x[,"n_22"] + x[,"n_44"])*log(2 + r_rm) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log(3 - r_rm) + log(2 + r_rm)) + x[,"n_33"]*log(13 - 2*r_rm + 2*r_rm^2)

return(list(
r_mat = cbind(r_cc,r_cm,r_rm,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_rm,0),
logL_mat = cbind(logL_cc,logL_cm,logL_rm,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","repulsion mixed","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.5_2.2 <- function(x){
logL_cc <- function(r,n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n31,n33) {
L <- (-2*n20 - 3*n21 - n22 - 3*n23 - 2*n24 - 2*n30 + n32 - 2*n34 - 2*n40 - 3*n41 - n42 - 3*n43 - 2*n44)*log(2) + (n21 + n23 + n31 + n33 + n41 + n43)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n21 + n43)*log(7 - 6*r) + (n20 + n44)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n24 + n40)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n23 + n41)*log(1 + 6*r) + (n30 + n34)*log(1 + 4*r - 4*r^2) + (n22 + n42)*log(5 + 2*r - 2*r^2) + n32*log(3 - r + r^2)
return(L)}
interlogL_cc <- function(n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n31,n33) {
optimize(logL_cc,c(0,0.5), n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n31,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_32"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_31"],x[,"n_33"])
LOD_cc <- -((x[,"n_30"] + x[,"n_34"])*log10(2)) - 2*(x[,"n_23"] + x[,"n_41"])*log10(2) - 2*(x[,"n_21"] + x[,"n_43"])*log10(2) - x[,"n_32"]*(-2*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_42"])*(-log10(2) + log10(11)) + (x[,"n_21"] + x[,"n_43"])*log10(7 - 6*r_cc) + (x[,"n_20"] + x[,"n_44"])*(log10(3 - 2*r_cc) + log10(pmax(1e-6,1 - r_cc))) + (x[,"n_24"] + x[,"n_40"])*(log10(pmax(1e-6,r_cc)) + log10(1 + 2*r_cc)) + (x[,"n_23"] + x[,"n_41"])*log10(1 + 6*r_cc) + (x[,"n_30"] + x[,"n_34"])*log10(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_22"] + x[,"n_42"])*log10(5 + 2*r_cc - 2*r_cc^2) + x[,"n_32"]*log10(3 - r_cc + r_cc^2)
logL_cc <- (-2*x[,"n_20"] - 3*x[,"n_21"] - x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] + x[,"n_32"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 3*x[,"n_41"] - x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + (x[,"n_21"] + x[,"n_23"] + x[,"n_31"] + x[,"n_33"] + x[,"n_41"] + x[,"n_43"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_21"] + x[,"n_43"])*log(7 - 6*r_cc) + (x[,"n_20"] + x[,"n_44"])*(log(3 - 2*r_cc) + log(pmax(1e-6,1 - r_cc))) + (x[,"n_24"] + x[,"n_40"])*(log(pmax(1e-6,r_cc)) + log(1 + 2*r_cc)) + (x[,"n_23"] + x[,"n_41"])*log(1 + 6*r_cc) + (x[,"n_30"] + x[,"n_34"])*log(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_22"] + x[,"n_42"])*log(5 + 2*r_cc - 2*r_cc^2) + x[,"n_32"]*log(3 - r_cc + r_cc^2)
logL_cm <- function(r,n21,n22,n23,n32,n41,n42,n43,n31,n33,n20,n24,n30,n34,n40,n44) {
L <- (-n21 - 2*n22 - n23 - n32 - n41 - 2*n42 - n43)*log(2) + (n21 + n23 + n31 + n33 + n41 + n43)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n20 + n24 + n40 + n44)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n32*log(9 + 8*r - 8*r^2) + (n30 + n34)*log(1 - 2*r + 2*r^2) + (n22 + n42)*log(13 - 8*r + 8*r^2)
return(L)}
interlogL_cm <- function(n21,n22,n23,n32,n41,n42,n43,n31,n33,n20,n24,n30,n34,n40,n44) {
optimize(logL_cm,c(0,0.5), n21,n22,n23,n32,n41,n42,n43,n31,n33,n20,n24,n30,n34,n40,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_32"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_31"],x[,"n_33"],x[,"n_20"],x[,"n_24"],x[,"n_30"],x[,"n_34"],x[,"n_40"],x[,"n_44"])
LOD_cm <- (x[,"n_30"] + x[,"n_34"])*log10(2) + 2*(x[,"n_20"] + x[,"n_24"] + x[,"n_40"] + x[,"n_44"])*log10(2) - x[,"n_32"]*log10(11) - (x[,"n_22"] + x[,"n_42"])*log10(11) + (x[,"n_20"] + x[,"n_24"] + x[,"n_40"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + x[,"n_32"]*log10(9 + 8*r_cm - 8*r_cm^2) + (x[,"n_30"] + x[,"n_34"])*log10(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_22"] + x[,"n_42"])*log10(13 - 8*r_cm + 8*r_cm^2)
logL_cm <- (-x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - x[,"n_32"] - x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"])*log(2) + (x[,"n_21"] + x[,"n_23"] + x[,"n_31"] + x[,"n_33"] + x[,"n_41"] + x[,"n_43"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_20"] + x[,"n_24"] + x[,"n_40"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + x[,"n_32"]*log(9 + 8*r_cm - 8*r_cm^2) + (x[,"n_30"] + x[,"n_34"])*log(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_22"] + x[,"n_42"])*log(13 - 8*r_cm + 8*r_cm^2)
logL_rc <- function(r,n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n31,n33) {
L <- (-4*n20 - n21 - 3*n22 - n23 - 4*n24 - 3*n30 - 2*n32 - 3*n34 - 4*n40 - n41 - 3*n42 - n43 - 4*n44)*log(2) + (n21 + n23 + n31 + n33 + n41 + n43)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n20 + n24 + n40 + n44)*(log(3 - 2*r) + log(1 + 2*r)) + n32*log(21 + 4*r - 4*r^2) + (n30 + n34)*log(5 - 4*r + 4*r^2) + (n22 + n42)*log(23 - 4*r + 4*r^2)
return(L)}
interlogL_rc <- function(n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n31,n33) {
optimize(logL_rc,c(0,0.5), n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n31,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_32"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_31"],x[,"n_33"])
LOD_rc <- -2*(x[,"n_30"] + x[,"n_34"])*log10(2) - 2*(x[,"n_20"] + x[,"n_24"] + x[,"n_40"] + x[,"n_44"])*log10(2) - x[,"n_32"]*(log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_42"])*(log10(2) + log10(11)) + (x[,"n_20"] + x[,"n_24"] + x[,"n_40"] + x[,"n_44"])*(log10(3 - 2*r_rc) + log10(1 + 2*r_rc)) + x[,"n_32"]*log10(21 + 4*r_rc - 4*r_rc^2) + (x[,"n_30"] + x[,"n_34"])*log10(5 - 4*r_rc + 4*r_rc^2) + (x[,"n_22"] + x[,"n_42"])*log10(23 - 4*r_rc + 4*r_rc^2)
logL_rc <- (-4*x[,"n_20"] - x[,"n_21"] - 3*x[,"n_22"] - x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_30"] - 2*x[,"n_32"] - 3*x[,"n_34"] - 4*x[,"n_40"] - x[,"n_41"] - 3*x[,"n_42"] - x[,"n_43"] - 4*x[,"n_44"])*log(2) + (x[,"n_21"] + x[,"n_23"] + x[,"n_31"] + x[,"n_33"] + x[,"n_41"] + x[,"n_43"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_20"] + x[,"n_24"] + x[,"n_40"] + x[,"n_44"])*(log(3 - 2*r_rc) + log(1 + 2*r_rc)) + x[,"n_32"]*log(21 + 4*r_rc - 4*r_rc^2) + (x[,"n_30"] + x[,"n_34"])*log(5 - 4*r_rc + 4*r_rc^2) + (x[,"n_22"] + x[,"n_42"])*log(23 - 4*r_rc + 4*r_rc^2)
logL_rm <- function(r,n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n31,n33) {
L <- (-2*n20 - 3*n21 - n22 - 3*n23 - 2*n24 - 2*n30 + n32 - 2*n34 - 2*n40 - 3*n41 - n42 - 3*n43 - 2*n44)*log(2) + (n21 + n23 + n31 + n33 + n41 + n43)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n23 + n41)*log(7 - 6*r) + (n24 + n40)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n20 + n44)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n21 + n43)*log(1 + 6*r) + (n30 + n34)*log(1 + 4*r - 4*r^2) + (n22 + n42)*log(5 + 2*r - 2*r^2) + n32*log(3 - r + r^2)
return(L)}
interlogL_rm <- function(n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n31,n33) {
optimize(logL_rm,c(0,0.5), n20,n21,n22,n23,n24,n30,n32,n34,n40,n41,n42,n43,n44,n31,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_32"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_31"],x[,"n_33"])
LOD_rm <- -((x[,"n_30"] + x[,"n_34"])*log10(2)) - 2*(x[,"n_23"] + x[,"n_41"])*log10(2) - 2*(x[,"n_21"] + x[,"n_43"])*log10(2) - x[,"n_32"]*(-2*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_42"])*(-log10(2) + log10(11)) + (x[,"n_23"] + x[,"n_41"])*log10(7 - 6*r_rm) + (x[,"n_24"] + x[,"n_40"])*(log10(3 - 2*r_rm) + log10(pmax(1e-6,1 - r_rm))) + (x[,"n_20"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm)) + log10(1 + 2*r_rm)) + (x[,"n_21"] + x[,"n_43"])*log10(1 + 6*r_rm) + (x[,"n_30"] + x[,"n_34"])*log10(1 + 4*r_rm - 4*r_rm^2) + (x[,"n_22"] + x[,"n_42"])*log10(5 + 2*r_rm - 2*r_rm^2) + x[,"n_32"]*log10(3 - r_rm + r_rm^2)
logL_rm <- (-2*x[,"n_20"] - 3*x[,"n_21"] - x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] + x[,"n_32"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 3*x[,"n_41"] - x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"])*log(2) + (x[,"n_21"] + x[,"n_23"] + x[,"n_31"] + x[,"n_33"] + x[,"n_41"] + x[,"n_43"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_23"] + x[,"n_41"])*log(7 - 6*r_rm) + (x[,"n_24"] + x[,"n_40"])*(log(3 - 2*r_rm) + log(pmax(1e-6,1 - r_rm))) + (x[,"n_20"] + x[,"n_44"])*(log(pmax(1e-6,r_rm)) + log(1 + 2*r_rm)) + (x[,"n_21"] + x[,"n_43"])*log(1 + 6*r_rm) + (x[,"n_30"] + x[,"n_34"])*log(1 + 4*r_rm - 4*r_rm^2) + (x[,"n_22"] + x[,"n_42"])*log(5 + 2*r_rm - 2*r_rm^2) + x[,"n_32"]*log(3 - r_rm + r_rm^2)

return(list(
r_mat = cbind(r_cc,r_cm,r_rc,r_rm,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_rc,LOD_rm,0),
logL_mat = cbind(logL_cc,logL_cm,logL_rc,logL_rm,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","repulsion coupling","repulsion mixed","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.5_2.3 <- function(x){
logL_cc <- function(r,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - n35 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + (n21 + n24 + n31 + n34 + n41 + n44)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + 2*(n20 + n45)*log(pmax(1e-6,1 - r)) + 2*(n25 + n40)*log(pmax(1e-6,r)) + (n30 + n35)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n24 + n41)*(log(pmax(1e-6,r)) + log(3 + 2*r)) + (n22 + n43)*log(24 - 7*r - 8*r^2) + (n23 + n42)*log(9 + 23*r - 8*r^2) + (n31 + n34)*log(3 + 4*r - 4*r^2) + (n21 + n44)*log(5 - 7*r + 2*r^2) + (n32 + n33)*log(41 - 16*r + 16*r^2)
return(L)}
interlogL_cc <- function(n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cc,c(0,0.5), n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cc <- -2*(x[,"n_31"] + x[,"n_34"])*log10(2) + 2*(x[,"n_30"] + x[,"n_35"])*log10(2) + 2*(x[,"n_25"] + x[,"n_40"])*log10(2) - (x[,"n_24"] + x[,"n_41"])*log10(2) - (x[,"n_21"] + x[,"n_44"])*log10(2) + 2*(x[,"n_20"] + x[,"n_45"])*log10(2) - (x[,"n_32"] + x[,"n_33"])*log10(37) - (x[,"n_23"] + x[,"n_42"])*(-log10(2) + log10(37)) - (x[,"n_22"] + x[,"n_43"])*(-log10(2) + log10(37)) + 2*(x[,"n_20"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_25"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_30"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_24"] + x[,"n_41"])*(log10(pmax(1e-6,r_cc)) + log10(3 + 2*r_cc)) + (x[,"n_22"] + x[,"n_43"])*log10(24 - 7*r_cc - 8*r_cc^2) + (x[,"n_23"] + x[,"n_42"])*log10(9 + 23*r_cc - 8*r_cc^2) + (x[,"n_31"] + x[,"n_34"])*log10(3 + 4*r_cc - 4*r_cc^2) + (x[,"n_21"] + x[,"n_44"])*log10(5 - 7*r_cc + 2*r_cc^2) + (x[,"n_32"] + x[,"n_33"])*log10(41 - 16*r_cc + 16*r_cc^2)
logL_cc <- (-2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - x[,"n_35"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + 2*(x[,"n_20"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_25"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_30"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_24"] + x[,"n_41"])*(log(pmax(1e-6,r_cc)) + log(3 + 2*r_cc)) + (x[,"n_22"] + x[,"n_43"])*log(24 - 7*r_cc - 8*r_cc^2) + (x[,"n_23"] + x[,"n_42"])*log(9 + 23*r_cc - 8*r_cc^2) + (x[,"n_31"] + x[,"n_34"])*log(3 + 4*r_cc - 4*r_cc^2) + (x[,"n_21"] + x[,"n_44"])*log(5 - 7*r_cc + 2*r_cc^2) + (x[,"n_32"] + x[,"n_33"])*log(41 - 16*r_cc + 16*r_cc^2)
logL_cm <- function(r,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + (n21 + n24 + n31 + n34 + n41 + n44)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n20 + n25 + n40 + n45)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n32 + n33)*log(33 + 16*r - 16*r^2) + (n21 + n44)*log(2 + r - 2*r^2) + (n24 + n41)*log(1 + 3*r - 2*r^2) + (n30 + n35)*log(1 - 2*r + 2*r^2) + (n31 + n34)*log(5 - 4*r + 4*r^2) + (n22 + n43)*log(21 - 9*r + 8*r^2) + (n23 + n42)*log(20 - 7*r + 8*r^2)
return(L)}
interlogL_cm <- function(n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cm,c(0,0.5), n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cm <- -2*(x[,"n_31"] + x[,"n_34"])*log10(2) + (x[,"n_30"] + x[,"n_35"])*log10(2) - (x[,"n_24"] + x[,"n_41"])*log10(2) - (x[,"n_21"] + x[,"n_44"])*log10(2) + 2*(x[,"n_20"] + x[,"n_25"] + x[,"n_40"] + x[,"n_45"])*log10(2) - (x[,"n_32"] + x[,"n_33"])*log10(37) - (x[,"n_23"] + x[,"n_42"])*(-log10(2) + log10(37)) - (x[,"n_22"] + x[,"n_43"])*(-log10(2) + log10(37)) + (x[,"n_20"] + x[,"n_25"] + x[,"n_40"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_32"] + x[,"n_33"])*log10(33 + 16*r_cm - 16*r_cm^2) + (x[,"n_21"] + x[,"n_44"])*log10(2 + r_cm - 2*r_cm^2) + (x[,"n_24"] + x[,"n_41"])*log10(1 + 3*r_cm - 2*r_cm^2) + (x[,"n_30"] + x[,"n_35"])*log10(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_31"] + x[,"n_34"])*log10(5 - 4*r_cm + 4*r_cm^2) + (x[,"n_22"] + x[,"n_43"])*log10(21 - 9*r_cm + 8*r_cm^2) + (x[,"n_23"] + x[,"n_42"])*log10(20 - 7*r_cm + 8*r_cm^2)
logL_cm <- (-2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_20"] + x[,"n_25"] + x[,"n_40"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_32"] + x[,"n_33"])*log(33 + 16*r_cm - 16*r_cm^2) + (x[,"n_21"] + x[,"n_44"])*log(2 + r_cm - 2*r_cm^2) + (x[,"n_24"] + x[,"n_41"])*log(1 + 3*r_cm - 2*r_cm^2) + (x[,"n_30"] + x[,"n_35"])*log(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_31"] + x[,"n_34"])*log(5 - 4*r_cm + 4*r_cm^2) + (x[,"n_22"] + x[,"n_43"])*log(21 - 9*r_cm + 8*r_cm^2) + (x[,"n_23"] + x[,"n_42"])*log(20 - 7*r_cm + 8*r_cm^2)
logL_rc <- function(r,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 3*n42 - 3*n43 - 4*n44 - 4*n45)*log(2) + (n21 + n24 + n31 + n34 + n41 + n44)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n25 + n40)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n20 + n45)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n32 + n33)*log(35 + 8*r - 8*r^2) + (n21 + n44)*log(4 + r - 2*r^2) + (n24 + n41)*log(3 + 3*r - 2*r^2) + (n22 + n43)*log(21 - 7*r + 4*r^2) + (n30 + n35)*log(3 - 4*r + 4*r^2) + (n31 + n34)*log(9 - 4*r + 4*r^2) + (n23 + n42)*log(18 - r + 4*r^2)
return(L)}
interlogL_rc <- function(n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rc,c(0,0.5), n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rc <- -3*(x[,"n_31"] + x[,"n_34"])*log10(2) - (x[,"n_30"] + x[,"n_35"])*log10(2) - 2*(x[,"n_24"] + x[,"n_41"])*log10(2) - 2*(x[,"n_21"] + x[,"n_44"])*log10(2) - (x[,"n_32"] + x[,"n_33"])*log10(37) - (x[,"n_23"] + x[,"n_42"])*(-log10(2) + log10(37)) - (x[,"n_22"] + x[,"n_43"])*(-log10(2) + log10(37)) + (x[,"n_25"] + x[,"n_40"])*(log10(3 - 2*r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_20"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_32"] + x[,"n_33"])*log10(35 + 8*r_rc - 8*r_rc^2) + (x[,"n_21"] + x[,"n_44"])*log10(4 + r_rc - 2*r_rc^2) + (x[,"n_24"] + x[,"n_41"])*log10(3 + 3*r_rc - 2*r_rc^2) + (x[,"n_22"] + x[,"n_43"])*log10(21 - 7*r_rc + 4*r_rc^2) + (x[,"n_30"] + x[,"n_35"])*log10(3 - 4*r_rc + 4*r_rc^2) + (x[,"n_31"] + x[,"n_34"])*log10(9 - 4*r_rc + 4*r_rc^2) + (x[,"n_23"] + x[,"n_42"])*log10(18 - r_rc + 4*r_rc^2)
logL_rc <- (-4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_25"] + x[,"n_40"])*(log(3 - 2*r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_20"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_32"] + x[,"n_33"])*log(35 + 8*r_rc - 8*r_rc^2) + (x[,"n_21"] + x[,"n_44"])*log(4 + r_rc - 2*r_rc^2) + (x[,"n_24"] + x[,"n_41"])*log(3 + 3*r_rc - 2*r_rc^2) + (x[,"n_22"] + x[,"n_43"])*log(21 - 7*r_rc + 4*r_rc^2) + (x[,"n_30"] + x[,"n_35"])*log(3 - 4*r_rc + 4*r_rc^2) + (x[,"n_31"] + x[,"n_34"])*log(9 - 4*r_rc + 4*r_rc^2) + (x[,"n_23"] + x[,"n_42"])*log(18 - r_rc + 4*r_rc^2)
logL_rm <- function(r,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 3*n42 - 3*n43 - 4*n44 - 4*n45)*log(2) + (n21 + n24 + n31 + n34 + n41 + n44)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n25 + n40)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n20 + n45)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n23 + n42)*log(23 - 7*r - 4*r^2) + (n30 + n35)*log(1 + 4*r - 4*r^2) + (n31 + n34)*log(7 + 4*r - 4*r^2) + (n22 + n43)*log(12 + 15*r - 4*r^2) + (n24 + n41)*log(8 - 9*r + 2*r^2) + (n21 + n44)*log(1 + 5*r + 2*r^2) + (n32 + n33)*log(39 - 8*r + 8*r^2)
return(L)}
interlogL_rm <- function(n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rm,c(0,0.5), n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm <- -3*(x[,"n_31"] + x[,"n_34"])*log10(2) - (x[,"n_30"] + x[,"n_35"])*log10(2) - 2*(x[,"n_24"] + x[,"n_41"])*log10(2) - 2*(x[,"n_21"] + x[,"n_44"])*log10(2) - (x[,"n_32"] + x[,"n_33"])*log10(37) - (x[,"n_23"] + x[,"n_42"])*(-log10(2) + log10(37)) - (x[,"n_22"] + x[,"n_43"])*(-log10(2) + log10(37)) + (x[,"n_25"] + x[,"n_40"])*(log10(3 - 2*r_rm) + log10(pmax(1e-6,1 - r_rm))) + (x[,"n_20"] + x[,"n_45"])*(log10(pmax(1e-6,r_rm)) + log10(1 + 2*r_rm)) + (x[,"n_23"] + x[,"n_42"])*log10(23 - 7*r_rm - 4*r_rm^2) + (x[,"n_30"] + x[,"n_35"])*log10(1 + 4*r_rm - 4*r_rm^2) + (x[,"n_31"] + x[,"n_34"])*log10(7 + 4*r_rm - 4*r_rm^2) + (x[,"n_22"] + x[,"n_43"])*log10(12 + 15*r_rm - 4*r_rm^2) + (x[,"n_24"] + x[,"n_41"])*log10(8 - 9*r_rm + 2*r_rm^2) + (x[,"n_21"] + x[,"n_44"])*log10(1 + 5*r_rm + 2*r_rm^2) + (x[,"n_32"] + x[,"n_33"])*log10(39 - 8*r_rm + 8*r_rm^2)
logL_rm <- (-4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_25"] + x[,"n_40"])*(log(3 - 2*r_rm) + log(pmax(1e-6,1 - r_rm))) + (x[,"n_20"] + x[,"n_45"])*(log(pmax(1e-6,r_rm)) + log(1 + 2*r_rm)) + (x[,"n_23"] + x[,"n_42"])*log(23 - 7*r_rm - 4*r_rm^2) + (x[,"n_30"] + x[,"n_35"])*log(1 + 4*r_rm - 4*r_rm^2) + (x[,"n_31"] + x[,"n_34"])*log(7 + 4*r_rm - 4*r_rm^2) + (x[,"n_22"] + x[,"n_43"])*log(12 + 15*r_rm - 4*r_rm^2) + (x[,"n_24"] + x[,"n_41"])*log(8 - 9*r_rm + 2*r_rm^2) + (x[,"n_21"] + x[,"n_44"])*log(1 + 5*r_rm + 2*r_rm^2) + (x[,"n_32"] + x[,"n_33"])*log(39 - 8*r_rm + 8*r_rm^2)

return(list(
r_mat = cbind(r_cc,r_cm,r_rc,r_rm,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_rc,LOD_rm,0),
logL_mat = cbind(logL_cc,logL_cm,logL_rc,logL_rm,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","repulsion coupling","repulsion mixed","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.5_2.4 <- function(x){
logL_cc <- function(r,n23,n31,n33,n35,n43,n22,n24,n32,n34,n42,n44,n21,n25,n41,n45) {
L <- (-2*n23 + n31 - n33 + n35 - 2*n43)*log(2) + (n22 + n24 + n32 + n34 + n42 + n44)*log(3) + 2*(-n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n22 + n44)*log(pmax(1e-6,1 - r)) + 2*(n21 + n45)*log(pmax(1e-6,1 - r)) + 2*(n25 + n41)*log(pmax(1e-6,r)) + (n24 + n42)*log(pmax(1e-6,r)) + (n31 + n35)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n23 + n43)*log(9 + 8*r - 8*r^2) + n33*log(13 - 8*r + 8*r^2)
return(L)}
interlogL_cc <- function(n23,n31,n33,n35,n43,n22,n24,n32,n34,n42,n44,n21,n25,n41,n45) {
optimize(logL_cc,c(0,0.5), n23,n31,n33,n35,n43,n22,n24,n32,n34,n42,n44,n21,n25,n41,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_23"],x[,"n_31"],x[,"n_33"],x[,"n_35"],x[,"n_43"],x[,"n_22"],x[,"n_24"],x[,"n_32"],x[,"n_34"],x[,"n_42"],x[,"n_44"],x[,"n_21"],x[,"n_25"],x[,"n_41"],x[,"n_45"])
LOD_cc <- 2*(x[,"n_31"] + x[,"n_35"])*log10(2) + 2*(x[,"n_25"] + x[,"n_41"])*log10(2) + (x[,"n_24"] + x[,"n_42"])*log10(2) + (x[,"n_22"] + x[,"n_44"])*log10(2) + 2*(x[,"n_21"] + x[,"n_45"])*log10(2) - x[,"n_33"]*log10(11) - (x[,"n_23"] + x[,"n_43"])*log10(11) + (x[,"n_22"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_21"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_25"] + x[,"n_41"])*log10(pmax(1e-6,r_cc)) + (x[,"n_24"] + x[,"n_42"])*log10(pmax(1e-6,r_cc)) + (x[,"n_31"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_23"] + x[,"n_43"])*log10(9 + 8*r_cc - 8*r_cc^2) + x[,"n_33"]*log10(13 - 8*r_cc + 8*r_cc^2)
logL_cc <- (-2*x[,"n_23"] + x[,"n_31"] - x[,"n_33"] + x[,"n_35"] - 2*x[,"n_43"])*log(2) + (x[,"n_22"] + x[,"n_24"] + x[,"n_32"] + x[,"n_34"] + x[,"n_42"] + x[,"n_44"])*log(3) + 2*(-x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_22"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_21"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + 2*(x[,"n_25"] + x[,"n_41"])*log(pmax(1e-6,r_cc)) + (x[,"n_24"] + x[,"n_42"])*log(pmax(1e-6,r_cc)) + (x[,"n_31"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_23"] + x[,"n_43"])*log(9 + 8*r_cc - 8*r_cc^2) + x[,"n_33"]*log(13 - 8*r_cc + 8*r_cc^2)
logL_cm <- function(r,n21,n22,n24,n25,n31,n35,n41,n42,n44,n45,n32,n34,n23,n33,n43) {
L <- (-2*n21 - 3*n22 - 3*n24 - 2*n25 - 2*n31 - 2*n35 - 2*n41 - 3*n42 - 3*n44 - 2*n45)*log(2) + (n22 + n24 + n32 + n34 + n42 + n44)*log(3) + 2*(-n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n22 + n44)*log(5 - 2*r) + (n25 + n41)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n21 + n45)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n24 + n42)*log(3 + 2*r) + n33*log(5 + 2*r - 2*r^2) + (n23 + n43)*log(3 - r + r^2) + (n31 + n35)*log(3 - 4*r + 4*r^2)
return(L)}
interlogL_cm <- function(n21,n22,n24,n25,n31,n35,n41,n42,n44,n45,n32,n34,n23,n33,n43) {
optimize(logL_cm,c(0,0.5), n21,n22,n24,n25,n31,n35,n41,n42,n44,n45,n32,n34,n23,n33,n43, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_44"],x[,"n_45"],x[,"n_32"],x[,"n_34"],x[,"n_23"],x[,"n_33"],x[,"n_43"])
LOD_cm <- -((x[,"n_31"] + x[,"n_35"])*log10(2)) - 2*(x[,"n_24"] + x[,"n_42"])*log10(2) - 2*(x[,"n_22"] + x[,"n_44"])*log10(2) - (x[,"n_23"] + x[,"n_43"])*(-2*log10(2) + log10(11)) - x[,"n_33"]*(-log10(2) + log10(11)) + (x[,"n_22"] + x[,"n_44"])*log10(5 - 2*r_cm) + (x[,"n_25"] + x[,"n_41"])*(log10(3 - 2*r_cm) + log10(pmax(1e-6,r_cm))) + (x[,"n_21"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cm)) + log10(1 + 2*r_cm)) + (x[,"n_24"] + x[,"n_42"])*log10(3 + 2*r_cm) + x[,"n_33"]*log10(5 + 2*r_cm - 2*r_cm^2) + (x[,"n_23"] + x[,"n_43"])*log10(3 - r_cm + r_cm^2) + (x[,"n_31"] + x[,"n_35"])*log10(3 - 4*r_cm + 4*r_cm^2)
logL_cm <- (-2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_22"] + x[,"n_24"] + x[,"n_32"] + x[,"n_34"] + x[,"n_42"] + x[,"n_44"])*log(3) + 2*(-x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_22"] + x[,"n_44"])*log(5 - 2*r_cm) + (x[,"n_25"] + x[,"n_41"])*(log(3 - 2*r_cm) + log(pmax(1e-6,r_cm))) + (x[,"n_21"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cm)) + log(1 + 2*r_cm)) + (x[,"n_24"] + x[,"n_42"])*log(3 + 2*r_cm) + x[,"n_33"]*log(5 + 2*r_cm - 2*r_cm^2) + (x[,"n_23"] + x[,"n_43"])*log(3 - r_cm + r_cm^2) + (x[,"n_31"] + x[,"n_35"])*log(3 - 4*r_cm + 4*r_cm^2)
logL_rm <- function(r,n21,n22,n23,n24,n25,n31,n33,n35,n41,n42,n43,n44,n45,n32,n34) {
L <- (-4*n21 - 2*n22 - 3*n23 - 2*n24 - 4*n25 - 3*n31 - 2*n33 - 3*n35 - 4*n41 - 2*n42 - 3*n43 - 2*n44 - 4*n45)*log(2) + (n22 + n24 + n32 + n34 + n42 + n44)*log(3) + 2*(-n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + 2*(n25 + n41)*log(3 - 2*r) + (n24 + n42)*log(3 - 2*r) + (n22 + n44)*log(1 + 2*r) + 2*(n21 + n45)*log(1 + 2*r) + (n31 + n35)*(log(3 - 2*r) + log(1 + 2*r)) + (n23 + n43)*log(21 + 4*r - 4*r^2) + n33*log(23 - 4*r + 4*r^2)
return(L)}
interlogL_rm <- function(n21,n22,n23,n24,n25,n31,n33,n35,n41,n42,n43,n44,n45,n32,n34) {
optimize(logL_rm,c(0,0.5), n21,n22,n23,n24,n25,n31,n33,n35,n41,n42,n43,n44,n45,n32,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_33"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_32"],x[,"n_34"])
LOD_rm <- -2*(x[,"n_31"] + x[,"n_35"])*log10(2) - 2*(x[,"n_25"] + x[,"n_41"])*log10(2) - (x[,"n_24"] + x[,"n_42"])*log10(2) - (x[,"n_22"] + x[,"n_44"])*log10(2) - 2*(x[,"n_21"] + x[,"n_45"])*log10(2) - x[,"n_33"]*(log10(2) + log10(11)) - (x[,"n_23"] + x[,"n_43"])*(log10(2) + log10(11)) + 2*(x[,"n_25"] + x[,"n_41"])*log10(3 - 2*r_rm) + (x[,"n_24"] + x[,"n_42"])*log10(3 - 2*r_rm) + (x[,"n_22"] + x[,"n_44"])*log10(1 + 2*r_rm) + 2*(x[,"n_21"] + x[,"n_45"])*log10(1 + 2*r_rm) + (x[,"n_31"] + x[,"n_35"])*(log10(3 - 2*r_rm) + log10(1 + 2*r_rm)) + (x[,"n_23"] + x[,"n_43"])*log10(21 + 4*r_rm - 4*r_rm^2) + x[,"n_33"]*log10(23 - 4*r_rm + 4*r_rm^2)
logL_rm <- (-4*x[,"n_21"] - 2*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 4*x[,"n_25"] - 3*x[,"n_31"] - 2*x[,"n_33"] - 3*x[,"n_35"] - 4*x[,"n_41"] - 2*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"] - 4*x[,"n_45"])*log(2) + (x[,"n_22"] + x[,"n_24"] + x[,"n_32"] + x[,"n_34"] + x[,"n_42"] + x[,"n_44"])*log(3) + 2*(-x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + 2*(x[,"n_25"] + x[,"n_41"])*log(3 - 2*r_rm) + (x[,"n_24"] + x[,"n_42"])*log(3 - 2*r_rm) + (x[,"n_22"] + x[,"n_44"])*log(1 + 2*r_rm) + 2*(x[,"n_21"] + x[,"n_45"])*log(1 + 2*r_rm) + (x[,"n_31"] + x[,"n_35"])*(log(3 - 2*r_rm) + log(1 + 2*r_rm)) + (x[,"n_23"] + x[,"n_43"])*log(21 + 4*r_rm - 4*r_rm^2) + x[,"n_33"]*log(23 - 4*r_rm + 4*r_rm^2)

return(list(
r_mat = cbind(r_cc,r_cm,r_rm,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_rm,0),
logL_mat = cbind(logL_cc,logL_cm,logL_rm,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","repulsion mixed","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.5_3.2 <- function(x){
logL_cc <- function(r,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 3*n42 - 3*n43 - 4*n44 - 4*n45)*log(2) + (n21 + n24 + n31 + n34 + n41 + n44)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n20 + n45)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n25 + n40)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n22 + n43)*log(23 - 7*r - 4*r^2) + (n30 + n35)*log(1 + 4*r - 4*r^2) + (n31 + n34)*log(7 + 4*r - 4*r^2) + (n23 + n42)*log(12 + 15*r - 4*r^2) + (n21 + n44)*log(8 - 9*r + 2*r^2) + (n24 + n41)*log(1 + 5*r + 2*r^2) + (n32 + n33)*log(39 - 8*r + 8*r^2)
return(L)}
interlogL_cc <- function(n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cc,c(0,0.5), n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cc <- -3*(x[,"n_31"] + x[,"n_34"])*log10(2) - (x[,"n_30"] + x[,"n_35"])*log10(2) - 2*(x[,"n_24"] + x[,"n_41"])*log10(2) - 2*(x[,"n_21"] + x[,"n_44"])*log10(2) - (x[,"n_32"] + x[,"n_33"])*log10(37) - (x[,"n_23"] + x[,"n_42"])*(-log10(2) + log10(37)) - (x[,"n_22"] + x[,"n_43"])*(-log10(2) + log10(37)) + (x[,"n_20"] + x[,"n_45"])*(log10(3 - 2*r_cc) + log10(pmax(1e-6,1 - r_cc))) + (x[,"n_25"] + x[,"n_40"])*(log10(pmax(1e-6,r_cc)) + log10(1 + 2*r_cc)) + (x[,"n_22"] + x[,"n_43"])*log10(23 - 7*r_cc - 4*r_cc^2) + (x[,"n_30"] + x[,"n_35"])*log10(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_31"] + x[,"n_34"])*log10(7 + 4*r_cc - 4*r_cc^2) + (x[,"n_23"] + x[,"n_42"])*log10(12 + 15*r_cc - 4*r_cc^2) + (x[,"n_21"] + x[,"n_44"])*log10(8 - 9*r_cc + 2*r_cc^2) + (x[,"n_24"] + x[,"n_41"])*log10(1 + 5*r_cc + 2*r_cc^2) + (x[,"n_32"] + x[,"n_33"])*log10(39 - 8*r_cc + 8*r_cc^2)
logL_cc <- (-4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_20"] + x[,"n_45"])*(log(3 - 2*r_cc) + log(pmax(1e-6,1 - r_cc))) + (x[,"n_25"] + x[,"n_40"])*(log(pmax(1e-6,r_cc)) + log(1 + 2*r_cc)) + (x[,"n_22"] + x[,"n_43"])*log(23 - 7*r_cc - 4*r_cc^2) + (x[,"n_30"] + x[,"n_35"])*log(1 + 4*r_cc - 4*r_cc^2) + (x[,"n_31"] + x[,"n_34"])*log(7 + 4*r_cc - 4*r_cc^2) + (x[,"n_23"] + x[,"n_42"])*log(12 + 15*r_cc - 4*r_cc^2) + (x[,"n_21"] + x[,"n_44"])*log(8 - 9*r_cc + 2*r_cc^2) + (x[,"n_24"] + x[,"n_41"])*log(1 + 5*r_cc + 2*r_cc^2) + (x[,"n_32"] + x[,"n_33"])*log(39 - 8*r_cc + 8*r_cc^2)
logL_cm <- function(r,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + (n21 + n24 + n31 + n34 + n41 + n44)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n20 + n25 + n40 + n45)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n32 + n33)*log(33 + 16*r - 16*r^2) + (n24 + n41)*log(2 + r - 2*r^2) + (n21 + n44)*log(1 + 3*r - 2*r^2) + (n30 + n35)*log(1 - 2*r + 2*r^2) + (n31 + n34)*log(5 - 4*r + 4*r^2) + (n23 + n42)*log(21 - 9*r + 8*r^2) + (n22 + n43)*log(20 - 7*r + 8*r^2)
return(L)}
interlogL_cm <- function(n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cm,c(0,0.5), n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cm <- -2*(x[,"n_31"] + x[,"n_34"])*log10(2) + (x[,"n_30"] + x[,"n_35"])*log10(2) - (x[,"n_24"] + x[,"n_41"])*log10(2) - (x[,"n_21"] + x[,"n_44"])*log10(2) + 2*(x[,"n_20"] + x[,"n_25"] + x[,"n_40"] + x[,"n_45"])*log10(2) - (x[,"n_32"] + x[,"n_33"])*log10(37) - (x[,"n_23"] + x[,"n_42"])*(-log10(2) + log10(37)) - (x[,"n_22"] + x[,"n_43"])*(-log10(2) + log10(37)) + (x[,"n_20"] + x[,"n_25"] + x[,"n_40"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_32"] + x[,"n_33"])*log10(33 + 16*r_cm - 16*r_cm^2) + (x[,"n_24"] + x[,"n_41"])*log10(2 + r_cm - 2*r_cm^2) + (x[,"n_21"] + x[,"n_44"])*log10(1 + 3*r_cm - 2*r_cm^2) + (x[,"n_30"] + x[,"n_35"])*log10(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_31"] + x[,"n_34"])*log10(5 - 4*r_cm + 4*r_cm^2) + (x[,"n_23"] + x[,"n_42"])*log10(21 - 9*r_cm + 8*r_cm^2) + (x[,"n_22"] + x[,"n_43"])*log10(20 - 7*r_cm + 8*r_cm^2)
logL_cm <- (-2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_20"] + x[,"n_25"] + x[,"n_40"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_32"] + x[,"n_33"])*log(33 + 16*r_cm - 16*r_cm^2) + (x[,"n_24"] + x[,"n_41"])*log(2 + r_cm - 2*r_cm^2) + (x[,"n_21"] + x[,"n_44"])*log(1 + 3*r_cm - 2*r_cm^2) + (x[,"n_30"] + x[,"n_35"])*log(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_31"] + x[,"n_34"])*log(5 - 4*r_cm + 4*r_cm^2) + (x[,"n_23"] + x[,"n_42"])*log(21 - 9*r_cm + 8*r_cm^2) + (x[,"n_22"] + x[,"n_43"])*log(20 - 7*r_cm + 8*r_cm^2)
logL_rc <- function(r,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 3*n42 - 3*n43 - 4*n44 - 4*n45)*log(2) + (n21 + n24 + n31 + n34 + n41 + n44)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n20 + n45)*(log(3 - 2*r) + log(pmax(1e-6,r))) + (n25 + n40)*(log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n32 + n33)*log(35 + 8*r - 8*r^2) + (n24 + n41)*log(4 + r - 2*r^2) + (n21 + n44)*log(3 + 3*r - 2*r^2) + (n23 + n42)*log(21 - 7*r + 4*r^2) + (n30 + n35)*log(3 - 4*r + 4*r^2) + (n31 + n34)*log(9 - 4*r + 4*r^2) + (n22 + n43)*log(18 - r + 4*r^2)
return(L)}
interlogL_rc <- function(n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rc,c(0,0.5), n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rc <- -3*(x[,"n_31"] + x[,"n_34"])*log10(2) - (x[,"n_30"] + x[,"n_35"])*log10(2) - 2*(x[,"n_24"] + x[,"n_41"])*log10(2) - 2*(x[,"n_21"] + x[,"n_44"])*log10(2) - (x[,"n_32"] + x[,"n_33"])*log10(37) - (x[,"n_23"] + x[,"n_42"])*(-log10(2) + log10(37)) - (x[,"n_22"] + x[,"n_43"])*(-log10(2) + log10(37)) + (x[,"n_20"] + x[,"n_45"])*(log10(3 - 2*r_rc) + log10(pmax(1e-6,r_rc))) + (x[,"n_25"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_rc)) + log10(1 + 2*r_rc)) + (x[,"n_32"] + x[,"n_33"])*log10(35 + 8*r_rc - 8*r_rc^2) + (x[,"n_24"] + x[,"n_41"])*log10(4 + r_rc - 2*r_rc^2) + (x[,"n_21"] + x[,"n_44"])*log10(3 + 3*r_rc - 2*r_rc^2) + (x[,"n_23"] + x[,"n_42"])*log10(21 - 7*r_rc + 4*r_rc^2) + (x[,"n_30"] + x[,"n_35"])*log10(3 - 4*r_rc + 4*r_rc^2) + (x[,"n_31"] + x[,"n_34"])*log10(9 - 4*r_rc + 4*r_rc^2) + (x[,"n_22"] + x[,"n_43"])*log10(18 - r_rc + 4*r_rc^2)
logL_rc <- (-4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_20"] + x[,"n_45"])*(log(3 - 2*r_rc) + log(pmax(1e-6,r_rc))) + (x[,"n_25"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_rc)) + log(1 + 2*r_rc)) + (x[,"n_32"] + x[,"n_33"])*log(35 + 8*r_rc - 8*r_rc^2) + (x[,"n_24"] + x[,"n_41"])*log(4 + r_rc - 2*r_rc^2) + (x[,"n_21"] + x[,"n_44"])*log(3 + 3*r_rc - 2*r_rc^2) + (x[,"n_23"] + x[,"n_42"])*log(21 - 7*r_rc + 4*r_rc^2) + (x[,"n_30"] + x[,"n_35"])*log(3 - 4*r_rc + 4*r_rc^2) + (x[,"n_31"] + x[,"n_34"])*log(9 - 4*r_rc + 4*r_rc^2) + (x[,"n_22"] + x[,"n_43"])*log(18 - r_rc + 4*r_rc^2)
logL_rm <- function(r,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - n35 - 2*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45)*log(2) + (n21 + n24 + n31 + n34 + n41 + n44)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + 2*(n25 + n40)*log(pmax(1e-6,1 - r)) + 2*(n20 + n45)*log(pmax(1e-6,r)) + (n30 + n35)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n21 + n44)*(log(pmax(1e-6,r)) + log(3 + 2*r)) + (n23 + n42)*log(24 - 7*r - 8*r^2) + (n22 + n43)*log(9 + 23*r - 8*r^2) + (n31 + n34)*log(3 + 4*r - 4*r^2) + (n24 + n41)*log(5 - 7*r + 2*r^2) + (n32 + n33)*log(41 - 16*r + 16*r^2)
return(L)}
interlogL_rm <- function(n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rm,c(0,0.5), n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm <- -2*(x[,"n_31"] + x[,"n_34"])*log10(2) + 2*(x[,"n_30"] + x[,"n_35"])*log10(2) + 2*(x[,"n_25"] + x[,"n_40"])*log10(2) - (x[,"n_24"] + x[,"n_41"])*log10(2) - (x[,"n_21"] + x[,"n_44"])*log10(2) + 2*(x[,"n_20"] + x[,"n_45"])*log10(2) - (x[,"n_32"] + x[,"n_33"])*log10(37) - (x[,"n_23"] + x[,"n_42"])*(-log10(2) + log10(37)) - (x[,"n_22"] + x[,"n_43"])*(-log10(2) + log10(37)) + 2*(x[,"n_25"] + x[,"n_40"])*log10(pmax(1e-6,1 - r_rm)) + 2*(x[,"n_20"] + x[,"n_45"])*log10(pmax(1e-6,r_rm)) + (x[,"n_30"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm))) + (x[,"n_21"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm)) + log10(3 + 2*r_rm)) + (x[,"n_23"] + x[,"n_42"])*log10(24 - 7*r_rm - 8*r_rm^2) + (x[,"n_22"] + x[,"n_43"])*log10(9 + 23*r_rm - 8*r_rm^2) + (x[,"n_31"] + x[,"n_34"])*log10(3 + 4*r_rm - 4*r_rm^2) + (x[,"n_24"] + x[,"n_41"])*log10(5 - 7*r_rm + 2*r_rm^2) + (x[,"n_32"] + x[,"n_33"])*log10(41 - 16*r_rm + 16*r_rm^2)
logL_rm <- (-2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - x[,"n_35"] - 2*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + 2*(x[,"n_25"] + x[,"n_40"])*log(pmax(1e-6,1 - r_rm)) + 2*(x[,"n_20"] + x[,"n_45"])*log(pmax(1e-6,r_rm)) + (x[,"n_30"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm))) + (x[,"n_21"] + x[,"n_44"])*(log(pmax(1e-6,r_rm)) + log(3 + 2*r_rm)) + (x[,"n_23"] + x[,"n_42"])*log(24 - 7*r_rm - 8*r_rm^2) + (x[,"n_22"] + x[,"n_43"])*log(9 + 23*r_rm - 8*r_rm^2) + (x[,"n_31"] + x[,"n_34"])*log(3 + 4*r_rm - 4*r_rm^2) + (x[,"n_24"] + x[,"n_41"])*log(5 - 7*r_rm + 2*r_rm^2) + (x[,"n_32"] + x[,"n_33"])*log(41 - 16*r_rm + 16*r_rm^2)

return(list(
r_mat = cbind(r_cc,r_cm,r_rc,r_rm,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_rc,LOD_rm,0),
logL_mat = cbind(logL_cc,logL_cm,logL_rc,logL_rm,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","repulsion coupling","repulsion mixed","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_1.5_3.3 <- function(x){
logL_cc <- function(r,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26 - 3*n30 - 3*n31 - 3*n32 - 2*n33 - 3*n34 - 3*n35 - 3*n36 - 4*n40 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 4*n46)*log(2) + (n21 + n22 + n24 + n25 + n31 + n32 + n34 + n35 + n41 + n42 + n44 + n45)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*log(5) + 2*(n20 + n46)*log(pmax(1e-6,1 - r)) + (n21 + n45)*(log(pmax(1e-6,1 - r)) + log(2 - r)) + 2*(n26 + n40)*log(pmax(1e-6,r)) + (n30 + n36)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n25 + n41)*(log(pmax(1e-6,r)) + log(1 + r)) + (n23 + n43)*log(9 + 5*r - 5*r^2) + (n31 + n35)*log(1 + 2*r - 2*r^2) + (n32 + n34)*log(8 + r - r^2) + (n22 + n44)*log(14 - 12*r + r^2) + (n24 + n42)*log(3 + 10*r + r^2) + n33*log(23 - 10*r + 10*r^2)
return(L)}
interlogL_cc <- function(n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cc,c(0,0.5), n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cc <- 2*(x[,"n_30"] + x[,"n_36"])*log10(2) + 2*(x[,"n_26"] + x[,"n_40"])*log10(2) + 2*(x[,"n_20"] + x[,"n_46"])*log10(2) - (x[,"n_25"] + x[,"n_41"])*(-2*log10(2) + log10(3)) - (x[,"n_21"] + x[,"n_45"])*(-2*log10(2) + log10(3)) - (x[,"n_31"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_23"] + x[,"n_43"])*(-2*log10(2) + log10(41)) - x[,"n_33"]*(-log10(2) + log10(41)) + 2*(x[,"n_20"] + x[,"n_46"])*log10(pmax(1e-6,1 - r_cc)) + (x[,"n_21"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_cc)) + log10(2 - r_cc)) + 2*(x[,"n_26"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_30"] + x[,"n_36"])*(log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_25"] + x[,"n_41"])*(log10(pmax(1e-6,r_cc)) + log10(1 + r_cc)) + (x[,"n_23"] + x[,"n_43"])*log10(9 + 5*r_cc - 5*r_cc^2) + (x[,"n_31"] + x[,"n_35"])*log10(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_32"] + x[,"n_34"])*log10(8 + r_cc - r_cc^2) + (x[,"n_22"] + x[,"n_44"])*log10(14 - 12*r_cc + r_cc^2) + (x[,"n_24"] + x[,"n_42"])*log10(3 + 10*r_cc + r_cc^2) + x[,"n_33"]*log10(23 - 10*r_cc + 10*r_cc^2)
logL_cc <- (-4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (x[,"n_21"] + x[,"n_22"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_41"] + x[,"n_42"] + x[,"n_44"] + x[,"n_45"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*log(5) + 2*(x[,"n_20"] + x[,"n_46"])*log(pmax(1e-6,1 - r_cc)) + (x[,"n_21"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_cc)) + log(2 - r_cc)) + 2*(x[,"n_26"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_30"] + x[,"n_36"])*(log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_25"] + x[,"n_41"])*(log(pmax(1e-6,r_cc)) + log(1 + r_cc)) + (x[,"n_23"] + x[,"n_43"])*log(9 + 5*r_cc - 5*r_cc^2) + (x[,"n_31"] + x[,"n_35"])*log(1 + 2*r_cc - 2*r_cc^2) + (x[,"n_32"] + x[,"n_34"])*log(8 + r_cc - r_cc^2) + (x[,"n_22"] + x[,"n_44"])*log(14 - 12*r_cc + r_cc^2) + (x[,"n_24"] + x[,"n_42"])*log(3 + 10*r_cc + r_cc^2) + x[,"n_33"]*log(23 - 10*r_cc + 10*r_cc^2)
logL_cm <- function(r,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 2*n31 - 4*n32 - n33 - 4*n34 - 2*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n46)*log(2) + (n21 + n22 + n24 + n25 + n31 + n32 + n34 + n35 + n41 + n42 + n44 + n45)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*log(5) + (n20 + n26 + n40 + n46)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n33*log(9 + 5*r - 5*r^2) + (n21 + n25 + n41 + n45)*log(1 + 2*r - 2*r^2) + (n22 + n24 + n42 + n44)*log(8 + r - r^2) + (n31 + n35)*log(1 - r + r^2) + (n30 + n36)*log(1 - 2*r + 2*r^2) + (n32 + n34)*log(17 - 2*r + 2*r^2) + (n23 + n43)*log(23 - 10*r + 10*r^2)
return(L)}
interlogL_cm <- function(n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cm,c(0,0.5), n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cm <- (x[,"n_30"] + x[,"n_36"])*log10(2) + 2*(x[,"n_20"] + x[,"n_26"] + x[,"n_40"] + x[,"n_46"])*log10(2) + (x[,"n_31"] + x[,"n_35"])*(2*log10(2) - log10(3)) - (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*(-log10(2) + log10(3)) - (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_32"] + x[,"n_34"])*(-log10(2) + log10(3) + log10(11)) - x[,"n_33"]*(-2*log10(2) + log10(41)) - (x[,"n_23"] + x[,"n_43"])*(-log10(2) + log10(41)) + (x[,"n_20"] + x[,"n_26"] + x[,"n_40"] + x[,"n_46"])*(log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + x[,"n_33"]*log10(9 + 5*r_cm - 5*r_cm^2) + (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*log10(1 + 2*r_cm - 2*r_cm^2) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log10(8 + r_cm - r_cm^2) + (x[,"n_31"] + x[,"n_35"])*log10(1 - r_cm + r_cm^2) + (x[,"n_30"] + x[,"n_36"])*log10(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_32"] + x[,"n_34"])*log10(17 - 2*r_cm + 2*r_cm^2) + (x[,"n_23"] + x[,"n_43"])*log10(23 - 10*r_cm + 10*r_cm^2)
logL_cm <- (-4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 2*x[,"n_31"] - 4*x[,"n_32"] - x[,"n_33"] - 4*x[,"n_34"] - 2*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (x[,"n_21"] + x[,"n_22"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_41"] + x[,"n_42"] + x[,"n_44"] + x[,"n_45"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*log(5) + (x[,"n_20"] + x[,"n_26"] + x[,"n_40"] + x[,"n_46"])*(log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + x[,"n_33"]*log(9 + 5*r_cm - 5*r_cm^2) + (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*log(1 + 2*r_cm - 2*r_cm^2) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log(8 + r_cm - r_cm^2) + (x[,"n_31"] + x[,"n_35"])*log(1 - r_cm + r_cm^2) + (x[,"n_30"] + x[,"n_36"])*log(1 - 2*r_cm + 2*r_cm^2) + (x[,"n_32"] + x[,"n_34"])*log(17 - 2*r_cm + 2*r_cm^2) + (x[,"n_23"] + x[,"n_43"])*log(23 - 10*r_cm + 10*r_cm^2)
logL_rm <- function(r,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26 - 3*n30 - 3*n31 - 3*n32 - 2*n33 - 3*n34 - 3*n35 - 3*n36 - 4*n40 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 4*n46)*log(2) + (n21 + n22 + n24 + n25 + n31 + n32 + n34 + n35 + n41 + n42 + n44 + n45)*log(3) + 2*(-n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*log(5) + 2*(n26 + n40)*log(pmax(1e-6,1 - r)) + (n25 + n41)*(log(pmax(1e-6,1 - r)) + log(2 - r)) + 2*(n20 + n46)*log(pmax(1e-6,r)) + (n30 + n36)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n21 + n45)*(log(pmax(1e-6,r)) + log(1 + r)) + (n23 + n43)*log(9 + 5*r - 5*r^2) + (n31 + n35)*log(1 + 2*r - 2*r^2) + (n32 + n34)*log(8 + r - r^2) + (n24 + n42)*log(14 - 12*r + r^2) + (n22 + n44)*log(3 + 10*r + r^2) + n33*log(23 - 10*r + 10*r^2)
return(L)}
interlogL_rm <- function(n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_rm,c(0,0.5), n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_rm <- 2*(x[,"n_30"] + x[,"n_36"])*log10(2) + 2*(x[,"n_26"] + x[,"n_40"])*log10(2) + 2*(x[,"n_20"] + x[,"n_46"])*log10(2) - (x[,"n_25"] + x[,"n_41"])*(-2*log10(2) + log10(3)) - (x[,"n_21"] + x[,"n_45"])*(-2*log10(2) + log10(3)) - (x[,"n_31"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(11)) - (x[,"n_23"] + x[,"n_43"])*(-2*log10(2) + log10(41)) - x[,"n_33"]*(-log10(2) + log10(41)) + 2*(x[,"n_26"] + x[,"n_40"])*log10(pmax(1e-6,1 - r_rm)) + (x[,"n_25"] + x[,"n_41"])*(log10(pmax(1e-6,1 - r_rm)) + log10(2 - r_rm)) + 2*(x[,"n_20"] + x[,"n_46"])*log10(pmax(1e-6,r_rm)) + (x[,"n_30"] + x[,"n_36"])*(log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm))) + (x[,"n_21"] + x[,"n_45"])*(log10(pmax(1e-6,r_rm)) + log10(1 + r_rm)) + (x[,"n_23"] + x[,"n_43"])*log10(9 + 5*r_rm - 5*r_rm^2) + (x[,"n_31"] + x[,"n_35"])*log10(1 + 2*r_rm - 2*r_rm^2) + (x[,"n_32"] + x[,"n_34"])*log10(8 + r_rm - r_rm^2) + (x[,"n_24"] + x[,"n_42"])*log10(14 - 12*r_rm + r_rm^2) + (x[,"n_22"] + x[,"n_44"])*log10(3 + 10*r_rm + r_rm^2) + x[,"n_33"]*log10(23 - 10*r_rm + 10*r_rm^2)
logL_rm <- (-4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (x[,"n_21"] + x[,"n_22"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_41"] + x[,"n_42"] + x[,"n_44"] + x[,"n_45"])*log(3) + 2*(-x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*log(5) + 2*(x[,"n_26"] + x[,"n_40"])*log(pmax(1e-6,1 - r_rm)) + (x[,"n_25"] + x[,"n_41"])*(log(pmax(1e-6,1 - r_rm)) + log(2 - r_rm)) + 2*(x[,"n_20"] + x[,"n_46"])*log(pmax(1e-6,r_rm)) + (x[,"n_30"] + x[,"n_36"])*(log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm))) + (x[,"n_21"] + x[,"n_45"])*(log(pmax(1e-6,r_rm)) + log(1 + r_rm)) + (x[,"n_23"] + x[,"n_43"])*log(9 + 5*r_rm - 5*r_rm^2) + (x[,"n_31"] + x[,"n_35"])*log(1 + 2*r_rm - 2*r_rm^2) + (x[,"n_32"] + x[,"n_34"])*log(8 + r_rm - r_rm^2) + (x[,"n_24"] + x[,"n_42"])*log(14 - 12*r_rm + r_rm^2) + (x[,"n_22"] + x[,"n_44"])*log(3 + 10*r_rm + r_rm^2) + x[,"n_33"]*log(23 - 10*r_rm + 10*r_rm^2)

return(list(
r_mat = cbind(r_cc,r_cm,r_rm,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_rm,0),
logL_mat = cbind(logL_cc,logL_cm,logL_rm,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","repulsion mixed","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_1.2 <- function(x){
logL_c <- function(r,n10,n11,n12,n13,n00,n01,n02,n03,n20,n21,n22,n23) {
L <- (-n10 + n11 + n12 - n13)*log(2) + (n10 + n11 + n12 + n13)*log(3) + 2*(-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*log(5) + (n01 + n22)*log(3 - 2*r) + (n00 + n23)*log(pmax(1e-6,1 - r)) + (n03 + n20)*log(pmax(1e-6,r)) + (n02 + n21)*log(1 + 2*r)
return(L)}
interlogL_c <- function(n10,n11,n12,n13,n00,n01,n02,n03,n20,n21,n22,n23) {
optimize(logL_c,c(0,0.5), n10,n11,n12,n13,n00,n01,n02,n03,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_c <- (x[,"n_03"] + x[,"n_20"])*log10(2) - (x[,"n_02"] + x[,"n_21"])*log10(2) - (x[,"n_01"] + x[,"n_22"])*log10(2) + (x[,"n_00"] + x[,"n_23"])*log10(2) + (x[,"n_01"] + x[,"n_22"])*log10(3 - 2*r_c) + (x[,"n_00"] + x[,"n_23"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_03"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_21"])*log10(1 + 2*r_c)
logL_c <- (-x[,"n_10"] + x[,"n_11"] + x[,"n_12"] - x[,"n_13"])*log(2) + (x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*log(5) + (x[,"n_01"] + x[,"n_22"])*log(3 - 2*r_c) + (x[,"n_00"] + x[,"n_23"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_03"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_21"])*log(1 + 2*r_c)
logL_r <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
L <- (-2*n00 - n01 - n02 - 2*n03 - n10 + n11 + n12 - n13 - 2*n20 - n21 - n22 - 2*n23)*log(2) + (n10 + n11 + n12 + n13)*log(3) + 2*(-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*log(5) + (n03 + n20)*log(3 - 2*r) + (n02 + n21)*log(5 - 2*r) + (n00 + n23)*log(1 + 2*r) + (n01 + n22)*log(3 + 2*r)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_r <- -((x[,"n_03"] + x[,"n_20"])*log10(2)) - 2*(x[,"n_02"] + x[,"n_21"])*log10(2) - 2*(x[,"n_01"] + x[,"n_22"])*log10(2) - (x[,"n_00"] + x[,"n_23"])*log10(2) + (x[,"n_03"] + x[,"n_20"])*log10(3 - 2*r_r) + (x[,"n_02"] + x[,"n_21"])*log10(5 - 2*r_r) + (x[,"n_00"] + x[,"n_23"])*log10(1 + 2*r_r) + (x[,"n_01"] + x[,"n_22"])*log10(3 + 2*r_r)
logL_r <- (-2*x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - x[,"n_10"] + x[,"n_11"] + x[,"n_12"] - x[,"n_13"] - 2*x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - 2*x[,"n_23"])*log(2) + (x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*log(5) + (x[,"n_03"] + x[,"n_20"])*log(3 - 2*r_r) + (x[,"n_02"] + x[,"n_21"])*log(5 - 2*r_r) + (x[,"n_00"] + x[,"n_23"])*log(1 + 2*r_r) + (x[,"n_01"] + x[,"n_22"])*log(3 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_1.3 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 3*n10 - 2*n11 - 2*n12 - 2*n13 - 3*n14 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24)*log(2) + (2*n02 + n10 + n11 + 3*n12 + n13 + n14 + 2*n22)*log(3) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n10 - n11 - 2*n12 - n13 - 2*n14 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24)*log(5) + (n01 + n23)*log(9 - 8*r) + (n00 + n24)*log(pmax(1e-6,1 - r)) + (n04 + n20)*log(pmax(1e-6,r)) + (n03 + n21)*log(1 + 8*r)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_c <- (x[,"n_04"] + x[,"n_20"])*log10(2) + (x[,"n_00"] + x[,"n_24"])*log10(2) - (x[,"n_03"] + x[,"n_21"])*log10(5) - (x[,"n_01"] + x[,"n_23"])*log10(5) + (x[,"n_01"] + x[,"n_23"])*log10(9 - 8*r_c) + (x[,"n_00"] + x[,"n_24"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_21"])*log10(1 + 8*r_c)
logL_c <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 3*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"])*log(2) + (2*x[,"n_02"] + x[,"n_10"] + x[,"n_11"] + 3*x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + 2*x[,"n_22"])*log(3) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"])*log(5) + (x[,"n_01"] + x[,"n_23"])*log(9 - 8*r_c) + (x[,"n_00"] + x[,"n_24"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_21"])*log(1 + 8*r_c)
logL_r <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-4*n00 - 2*n01 - 2*n02 - 2*n03 - 4*n04 - 3*n10 - 2*n11 - 2*n12 - 2*n13 - 3*n14 - 4*n20 - 2*n21 - 2*n22 - 2*n23 - 4*n24)*log(2) + (2*n02 + n10 + n11 + 3*n12 + n13 + n14 + 2*n22)*log(3) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n10 - n11 - 2*n12 - n13 - 2*n14 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24)*log(5) + (n03 + n21)*log(7 - 4*r) + (n04 + n20)*log(3 - 2*r) + (n00 + n24)*log(1 + 2*r) + (n01 + n23)*log(3 + 4*r)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_r <- -((x[,"n_04"] + x[,"n_20"])*log10(2)) - (x[,"n_00"] + x[,"n_24"])*log10(2) - (x[,"n_03"] + x[,"n_21"])*log10(5) - (x[,"n_01"] + x[,"n_23"])*log10(5) + (x[,"n_03"] + x[,"n_21"])*log10(7 - 4*r_r) + (x[,"n_04"] + x[,"n_20"])*log10(3 - 2*r_r) + (x[,"n_00"] + x[,"n_24"])*log10(1 + 2*r_r) + (x[,"n_01"] + x[,"n_23"])*log10(3 + 4*r_r)
logL_r <- (-4*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 4*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"])*log(2) + (2*x[,"n_02"] + x[,"n_10"] + x[,"n_11"] + 3*x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + 2*x[,"n_22"])*log(3) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"])*log(5) + (x[,"n_03"] + x[,"n_21"])*log(7 - 4*r_r) + (x[,"n_04"] + x[,"n_20"])*log(3 - 2*r_r) + (x[,"n_00"] + x[,"n_24"])*log(1 + 2*r_r) + (x[,"n_01"] + x[,"n_23"])*log(3 + 4*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_1.4 <- function(x){
logL_c <- function(r,n11,n12,n13,n14,n01,n02,n03,n04,n21,n22,n23,n24) {
L <- (-n11 + n12 + n13 - n14)*log(2) + (n11 + n12 + n13 + n14)*log(3) + 2*(-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24)*log(5) + (n02 + n23)*log(3 - 2*r) + (n01 + n24)*log(pmax(1e-6,1 - r)) + (n04 + n21)*log(pmax(1e-6,r)) + (n03 + n22)*log(1 + 2*r)
return(L)}
interlogL_c <- function(n11,n12,n13,n14,n01,n02,n03,n04,n21,n22,n23,n24) {
optimize(logL_c,c(0,0.5), n11,n12,n13,n14,n01,n02,n03,n04,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_c <- (x[,"n_04"] + x[,"n_21"])*log10(2) - (x[,"n_03"] + x[,"n_22"])*log10(2) - (x[,"n_02"] + x[,"n_23"])*log10(2) + (x[,"n_01"] + x[,"n_24"])*log10(2) + (x[,"n_02"] + x[,"n_23"])*log10(3 - 2*r_c) + (x[,"n_01"] + x[,"n_24"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_21"])*log10(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_22"])*log10(1 + 2*r_c)
logL_c <- (-x[,"n_11"] + x[,"n_12"] + x[,"n_13"] - x[,"n_14"])*log(2) + (x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_02"] + x[,"n_23"])*log(3 - 2*r_c) + (x[,"n_01"] + x[,"n_24"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_21"])*log(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_22"])*log(1 + 2*r_c)
logL_r <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
L <- (-2*n01 - n02 - n03 - 2*n04 - n11 + n12 + n13 - n14 - 2*n21 - n22 - n23 - 2*n24)*log(2) + (n11 + n12 + n13 + n14)*log(3) + 2*(-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24)*log(5) + (n04 + n21)*log(3 - 2*r) + (n03 + n22)*log(5 - 2*r) + (n01 + n24)*log(1 + 2*r) + (n02 + n23)*log(3 + 2*r)
return(L)}
interlogL_r <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
optimize(logL_r,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_r <- -((x[,"n_04"] + x[,"n_21"])*log10(2)) - 2*(x[,"n_03"] + x[,"n_22"])*log10(2) - 2*(x[,"n_02"] + x[,"n_23"])*log10(2) - (x[,"n_01"] + x[,"n_24"])*log10(2) + (x[,"n_04"] + x[,"n_21"])*log10(3 - 2*r_r) + (x[,"n_03"] + x[,"n_22"])*log10(5 - 2*r_r) + (x[,"n_01"] + x[,"n_24"])*log10(1 + 2*r_r) + (x[,"n_02"] + x[,"n_23"])*log10(3 + 2*r_r)
logL_r <- (-2*x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - x[,"n_11"] + x[,"n_12"] + x[,"n_13"] - x[,"n_14"] - 2*x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"])*log(2) + (x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_04"] + x[,"n_21"])*log(3 - 2*r_r) + (x[,"n_03"] + x[,"n_22"])*log(5 - 2*r_r) + (x[,"n_01"] + x[,"n_24"])*log(1 + 2*r_r) + (x[,"n_02"] + x[,"n_23"])*log(3 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_1.5 <- function(x){
r_c <- (x[,"n_04"] + x[,"n_22"])/(x[,"n_02"] + x[,"n_04"] + x[,"n_22"] + x[,"n_24"])
logL_c <- (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(2) + (x[,"n_12"] + x[,"n_13"] + x[,"n_14"])*log(3) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_02"] + x[,"n_24"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_22"])*log(pmax(1e-6,r_c))
LOD_c <- (x[,"n_04"] + x[,"n_22"])*log10(2) + (x[,"n_02"] + x[,"n_24"])*log10(2) + (x[,"n_02"] + x[,"n_24"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_04"] + x[,"n_22"])*log10(pmax(1e-6,r_c))
r_r <- (3*x[,"n_02"] - x[,"n_04"] - x[,"n_22"] + 3*x[,"n_24"])/(2*(x[,"n_02"] + x[,"n_04"] + x[,"n_22"] + x[,"n_24"]))
logL_r <- (-3*x[,"n_02"] - x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_22"] - x[,"n_23"] - 3*x[,"n_24"])*log(2) + (x[,"n_12"] + x[,"n_13"] + x[,"n_14"])*log(3) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_04"] + x[,"n_22"])*log(3 - 2*r_r) + (x[,"n_02"] + x[,"n_24"])*log(1 + 2*r_r)
LOD_r <- -((x[,"n_04"] + x[,"n_22"])*log10(2)) - (x[,"n_02"] + x[,"n_24"])*log10(2) + (x[,"n_04"] + x[,"n_22"])*log10(3 - 2*r_r) + (x[,"n_02"] + x[,"n_24"])*log10(1 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_2.0 <- function(x){
logL_c <- function(r,n01,n10,n12,n21,n00,n02,n11,n20,n22) {
L <- (n01 + n10 + n12 + n21)*log(2) + (-n00 - n01 - n02 - n10 - n11 - n12 - n20 - n21 - n22)*log(5) + 2*(n00 + n22)*log(pmax(1e-6,1 - r)) + 2*(n02 + n20)*log(pmax(1e-6,r)) + (n01 + n10 + n12 + n21)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n11*log(3 - 4*r + 4*r^2)
return(L)}
interlogL_c <- function(n01,n10,n12,n21,n00,n02,n11,n20,n22) {
optimize(logL_c,c(0,0.5), n01,n10,n12,n21,n00,n02,n11,n20,n22, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_10"],x[,"n_12"],x[,"n_21"],x[,"n_00"],x[,"n_02"],x[,"n_11"],x[,"n_20"],x[,"n_22"])
LOD_c <- -(x[,"n_11"]*log10(2)) + 2*(x[,"n_02"] + x[,"n_20"])*log10(2) + 2*(x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*log10(2) + 2*(x[,"n_00"] + x[,"n_22"])*log10(2) + 2*(x[,"n_00"] + x[,"n_22"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_02"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + x[,"n_11"]*log10(3 - 4*r_c + 4*r_c^2)
logL_c <- (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"])*log(5) + 2*(x[,"n_00"] + x[,"n_22"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_02"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + x[,"n_11"]*log(3 - 4*r_c + 4*r_c^2)
logL_m <- function(r,n00,n01,n02,n10,n11,n12,n20,n21,n22) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n10 - n11 - 2*n12 - 2*n20 - 2*n21 - 2*n22)*log(2) + (-n00 - n01 - n02 - n10 - n11 - n12 - n20 - n21 - n22)*log(5) + (n02 + n20)*(log(2 - r) + log(pmax(1e-6,r))) + n11*log(3 + 2*r - 2*r^2) + (n00 + n22)*log(1 - r^2) + (n01 + n10 + n12 + n21)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n10,n11,n12,n20,n21,n22) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n10,n11,n12,n20,n21,n22, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_20"],x[,"n_21"],x[,"n_22"])
LOD_m <- (x[,"n_00"] + x[,"n_22"])*(2*log10(2) - log10(3)) - (x[,"n_02"] + x[,"n_20"])*(-2*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(-log10(2) + log10(5)) - x[,"n_11"]*(-log10(2) + log10(7)) + (x[,"n_02"] + x[,"n_20"])*(log10(2 - r_m) + log10(pmax(1e-6,r_m))) + x[,"n_11"]*log10(3 + 2*r_m - 2*r_m^2) + (x[,"n_00"] + x[,"n_22"])*log10(1 - r_m^2) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*log10(3 - 2*r_m + 2*r_m^2)
logL_m <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_10"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"])*log(5) + (x[,"n_02"] + x[,"n_20"])*(log(2 - r_m) + log(pmax(1e-6,r_m))) + x[,"n_11"]*log(3 + 2*r_m - 2*r_m^2) + (x[,"n_00"] + x[,"n_22"])*log(1 - r_m^2) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*log(3 - 2*r_m + 2*r_m^2)
logL_r <- function(r,n00,n01,n02,n10,n11,n12,n20,n21,n22) {
L <- (-n00 - n01 - n02 - n10 + n11 - n12 - n20 - n21 - n22)*log(2) + (-n00 - n01 - n02 - n10 - n11 - n12 - n20 - n21 - n22)*(log(3) + log(5)) + (n00 + n22)*(log(pmax(1e-6,r)) + log(2 + r)) + (n01 + n10 + n12 + n21)*log(3 + 2*r - 2*r^2) + (n02 + n20)*log(3 - 4*r + r^2) + n11*log(3 - r + r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n10,n11,n12,n20,n21,n22) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n10,n11,n12,n20,n21,n22, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_20"],x[,"n_21"],x[,"n_22"])
LOD_r <- -((x[,"n_02"] + x[,"n_20"])*(-2*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_22"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*(-log10(2) + log10(7)) - x[,"n_11"]*(-2*log10(2) + log10(11)) + (x[,"n_00"] + x[,"n_22"])*(log10(pmax(1e-6,r_r)) + log10(2 + r_r)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*log10(3 + 2*r_r - 2*r_r^2) + (x[,"n_02"] + x[,"n_20"])*log10(3 - 4*r_r + r_r^2) + x[,"n_11"]*log10(3 - r_r + r_r^2)
logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] + x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"])*(log(3) + log(5)) + (x[,"n_00"] + x[,"n_22"])*(log(pmax(1e-6,r_r)) + log(2 + r_r)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_12"] + x[,"n_21"])*log(3 + 2*r_r - 2*r_r^2) + (x[,"n_02"] + x[,"n_20"])*log(3 - 4*r_r + r_r^2) + x[,"n_11"]*log(3 - r_r + r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_2.1 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n11,n12,n20,n21,n22,n23,n10,n13) {
L <- (-n00 - n01 - n02 - n03 - n11 - n12 - n20 - n21 - n22 - n23)*log(2) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*log(5) + 2*(n00 + n23)*log(pmax(1e-6,1 - r)) + 2*(n03 + n20)*log(pmax(1e-6,r)) + (n10 + n13)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n21)*(log(2 - r) + log(pmax(1e-6,r))) + (n01 + n22)*log(1 - r^2) + (n11 + n12)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n11,n12,n20,n21,n22,n23,n10,n13) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n11,n12,n20,n21,n22,n23,n10,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_11"],x[,"n_12"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_10"],x[,"n_13"])
LOD_c <- 2*(x[,"n_10"] + x[,"n_13"])*log10(2) + 2*(x[,"n_03"] + x[,"n_20"])*log10(2) + 2*(x[,"n_00"] + x[,"n_23"])*log10(2) + (x[,"n_01"] + x[,"n_22"])*(2*log10(2) - log10(3)) - (x[,"n_02"] + x[,"n_21"])*(-2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_12"])*(-log10(2) + log10(5)) + 2*(x[,"n_00"] + x[,"n_23"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_03"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_13"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_21"])*(log10(2 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_01"] + x[,"n_22"])*log10(1 - r_c^2) + (x[,"n_11"] + x[,"n_12"])*log10(3 - 2*r_c + 2*r_c^2)
logL_c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_11"] - x[,"n_12"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*log(5) + 2*(x[,"n_00"] + x[,"n_23"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_03"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_13"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_21"])*(log(2 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_01"] + x[,"n_22"])*log(1 - r_c^2) + (x[,"n_11"] + x[,"n_12"])*log(3 - 2*r_c + 2*r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*(3*log(2) + log(5)) + (n03 + n20)*(log(2 - r) + log(pmax(1e-6,r))) + (n11 + n12)*log(9 + 2*r - 2*r^2) + (n00 + n23)*log(1 - r^2) + (n02 + n21)*log(3 + r^2) + (n01 + n22)*log(4 - 2*r + r^2) + (n10 + n13)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_m <- (x[,"n_00"] + x[,"n_23"])*(2*log10(2) - log10(3)) - (x[,"n_03"] + x[,"n_20"])*(-2*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_13"])*(-log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_21"])*(-2*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_22"])*(-2*log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_12"])*(-log10(2) + log10(19)) + (x[,"n_03"] + x[,"n_20"])*(log10(2 - r_m) + log10(pmax(1e-6,r_m))) + (x[,"n_11"] + x[,"n_12"])*log10(9 + 2*r_m - 2*r_m^2) + (x[,"n_00"] + x[,"n_23"])*log10(1 - r_m^2) + (x[,"n_02"] + x[,"n_21"])*log10(3 + r_m^2) + (x[,"n_01"] + x[,"n_22"])*log10(4 - 2*r_m + r_m^2) + (x[,"n_10"] + x[,"n_13"])*log10(3 - 2*r_m + 2*r_m^2)
logL_m <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*(3*log(2) + log(5)) + (x[,"n_03"] + x[,"n_20"])*(log(2 - r_m) + log(pmax(1e-6,r_m))) + (x[,"n_11"] + x[,"n_12"])*log(9 + 2*r_m - 2*r_m^2) + (x[,"n_00"] + x[,"n_23"])*log(1 - r_m^2) + (x[,"n_02"] + x[,"n_21"])*log(3 + r_m^2) + (x[,"n_01"] + x[,"n_22"])*log(4 - 2*r_m + r_m^2) + (x[,"n_10"] + x[,"n_13"])*log(3 - 2*r_m + 2*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*(2*log(2) + log(3) + log(5)) + (n00 + n23)*(log(pmax(1e-6,r)) + log(2 + r)) + (n10 + n13)*log(3 + 2*r - 2*r^2) + (n02 + n21)*log(6 - 2*r - r^2) + (n01 + n22)*log(3 + 4*r - r^2) + (n03 + n20)*log(3 - 4*r + r^2) + (n11 + n12)*log(15 - 2*r + 2*r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_r <- -((x[,"n_03"] + x[,"n_20"])*(-2*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_23"])*(-2*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_13"])*(-log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_21"])*(-2*log10(2) + log10(19)) - (x[,"n_01"] + x[,"n_22"])*(-2*log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_12"])*(-log10(2) + log10(29)) + (x[,"n_00"] + x[,"n_23"])*(log10(pmax(1e-6,r_r)) + log10(2 + r_r)) + (x[,"n_10"] + x[,"n_13"])*log10(3 + 2*r_r - 2*r_r^2) + (x[,"n_02"] + x[,"n_21"])*log10(6 - 2*r_r - r_r^2) + (x[,"n_01"] + x[,"n_22"])*log10(3 + 4*r_r - r_r^2) + (x[,"n_03"] + x[,"n_20"])*log10(3 - 4*r_r + r_r^2) + (x[,"n_11"] + x[,"n_12"])*log10(15 - 2*r_r + 2*r_r^2)
logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*(2*log(2) + log(3) + log(5)) + (x[,"n_00"] + x[,"n_23"])*(log(pmax(1e-6,r_r)) + log(2 + r_r)) + (x[,"n_10"] + x[,"n_13"])*log(3 + 2*r_r - 2*r_r^2) + (x[,"n_02"] + x[,"n_21"])*log(6 - 2*r_r - r_r^2) + (x[,"n_01"] + x[,"n_22"])*log(3 + 4*r_r - r_r^2) + (x[,"n_03"] + x[,"n_20"])*log(3 - 4*r_r + r_r^2) + (x[,"n_11"] + x[,"n_12"])*log(15 - 2*r_r + 2*r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_2.2 <- function(x){
logL_c <- function(r,n10,n14,n00,n01,n02,n03,n04,n11,n12,n13,n20,n21,n22,n23,n24) {
L <- (n10 + n14)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + 2*(n00 + n24)*log(pmax(1e-6,1 - r)) + 2*(n04 + n20)*log(pmax(1e-6,r)) + (n10 + n14)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n21)*(log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n22)*log(1 + 4*r - 4*r^2) + (n11 + n13)*log(3 + 2*r - 2*r^2) + (n01 + n23)*log(3 - 4*r + r^2) + n12*log(9 - 8*r + 8*r^2)
return(L)}
interlogL_c <- function(n10,n14,n00,n01,n02,n03,n04,n11,n12,n13,n20,n21,n22,n23,n24) {
optimize(logL_c,c(0,0.5), n10,n14,n00,n01,n02,n03,n04,n11,n12,n13,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_10"],x[,"n_14"],x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_c <- 2*(x[,"n_10"] + x[,"n_14"])*log10(2) + 2*(x[,"n_04"] + x[,"n_20"])*log10(2) - (x[,"n_02"] + x[,"n_22"])*log10(2) + 2*(x[,"n_00"] + x[,"n_24"])*log10(2) - (x[,"n_03"] + x[,"n_21"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_23"])*(-2*log10(2) + log10(5)) - x[,"n_12"]*log10(7) - (x[,"n_11"] + x[,"n_13"])*(-log10(2) + log10(7)) + 2*(x[,"n_00"] + x[,"n_24"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_04"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_14"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_03"] + x[,"n_21"])*(log10(pmax(1e-6,r_c)) + log10(2 + r_c)) + (x[,"n_02"] + x[,"n_22"])*log10(1 + 4*r_c - 4*r_c^2) + (x[,"n_11"] + x[,"n_13"])*log10(3 + 2*r_c - 2*r_c^2) + (x[,"n_01"] + x[,"n_23"])*log10(3 - 4*r_c + r_c^2) + x[,"n_12"]*log10(9 - 8*r_c + 8*r_c^2)
logL_c <- (x[,"n_10"] + x[,"n_14"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + 2*(x[,"n_00"] + x[,"n_24"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_04"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_14"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_03"] + x[,"n_21"])*(log(pmax(1e-6,r_c)) + log(2 + r_c)) + (x[,"n_02"] + x[,"n_22"])*log(1 + 4*r_c - 4*r_c^2) + (x[,"n_11"] + x[,"n_13"])*log(3 + 2*r_c - 2*r_c^2) + (x[,"n_01"] + x[,"n_23"])*log(3 - 4*r_c + r_c^2) + x[,"n_12"]*log(9 - 8*r_c + 8*r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-2*n00 - 2*n01 - n02 - 2*n03 - 2*n04 - 2*n10 - 2*n11 + n12 - 2*n13 - 2*n14 - 2*n20 - 2*n21 - n22 - 2*n23 - 2*n24)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + (n04 + n20)*(log(2 - r) + log(pmax(1e-6,r))) + (n00 + n24)*log(1 - r^2) + (n01 + n23)*log(6 - 2*r - r^2) + n12*log(3 + r - r^2) + (n03 + n21)*log(3 + 4*r - r^2) + (n10 + n14)*log(3 - 2*r + 2*r^2) + (n02 + n22)*log(5 - 2*r + 2*r^2) + (n11 + n13)*log(15 - 2*r + 2*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_m <- (x[,"n_00"] + x[,"n_24"])*(2*log10(2) - log10(3)) - (x[,"n_04"] + x[,"n_20"])*(-2*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_22"])*(-log10(2) + 2*log10(3)) - (x[,"n_10"] + x[,"n_14"])*(-log10(2) + log10(5)) - x[,"n_12"]*(-2*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_21"])*(-2*log10(2) + log10(19)) - (x[,"n_01"] + x[,"n_23"])*(-2*log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_13"])*(-log10(2) + log10(29)) + (x[,"n_04"] + x[,"n_20"])*(log10(2 - r_m) + log10(pmax(1e-6,r_m))) + (x[,"n_00"] + x[,"n_24"])*log10(1 - r_m^2) + (x[,"n_01"] + x[,"n_23"])*log10(6 - 2*r_m - r_m^2) + x[,"n_12"]*log10(3 + r_m - r_m^2) + (x[,"n_03"] + x[,"n_21"])*log10(3 + 4*r_m - r_m^2) + (x[,"n_10"] + x[,"n_14"])*log10(3 - 2*r_m + 2*r_m^2) + (x[,"n_02"] + x[,"n_22"])*log10(5 - 2*r_m + 2*r_m^2) + (x[,"n_11"] + x[,"n_13"])*log10(15 - 2*r_m + 2*r_m^2)
logL_m <- (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] + x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_04"] + x[,"n_20"])*(log(2 - r_m) + log(pmax(1e-6,r_m))) + (x[,"n_00"] + x[,"n_24"])*log(1 - r_m^2) + (x[,"n_01"] + x[,"n_23"])*log(6 - 2*r_m - r_m^2) + x[,"n_12"]*log(3 + r_m - r_m^2) + (x[,"n_03"] + x[,"n_21"])*log(3 + 4*r_m - r_m^2) + (x[,"n_10"] + x[,"n_14"])*log(3 - 2*r_m + 2*r_m^2) + (x[,"n_02"] + x[,"n_22"])*log(5 - 2*r_m + 2*r_m^2) + (x[,"n_11"] + x[,"n_13"])*log(15 - 2*r_m + 2*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n21,n22,n23,n24,n12) {
L <- (-n00 - n01 + n02 - n03 - n04 - n10 - n11 - n13 - n14 - n20 - n21 + n22 - n23 - n24)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*(log(3) + 2*log(5)) + (n00 + n24)*(log(pmax(1e-6,r)) + log(2 + r)) + (n10 + n14)*log(3 + 2*r - 2*r^2) + (n11 + n13)*log(21 + 2*r - 2*r^2) + (n02 + n22)*log(3 + r - r^2) + (n03 + n21)*log(12 - 10*r + r^2) + (n04 + n20)*log(3 - 4*r + r^2) + (n01 + n23)*log(3 + 8*r + r^2) + n12*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n21,n22,n23,n24,n12) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n21,n22,n23,n24,n12, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_12"])
LOD_r <- -((x[,"n_04"] + x[,"n_20"])*(-2*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - x[,"n_12"]*(2*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_14"])*(-log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_22"])*(-2*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_21"])*(-2*log10(2) + log10(29)) - (x[,"n_01"] + x[,"n_23"])*(-2*log10(2) + log10(29)) - (x[,"n_11"] + x[,"n_13"])*(-log10(2) + log10(43)) + (x[,"n_00"] + x[,"n_24"])*(log10(pmax(1e-6,r_r)) + log10(2 + r_r)) + (x[,"n_10"] + x[,"n_14"])*log10(3 + 2*r_r - 2*r_r^2) + (x[,"n_11"] + x[,"n_13"])*log10(21 + 2*r_r - 2*r_r^2) + (x[,"n_02"] + x[,"n_22"])*log10(3 + r_r - r_r^2) + (x[,"n_03"] + x[,"n_21"])*log10(12 - 10*r_r + r_r^2) + (x[,"n_04"] + x[,"n_20"])*log10(3 - 4*r_r + r_r^2) + (x[,"n_01"] + x[,"n_23"])*log10(3 + 8*r_r + r_r^2) + x[,"n_12"]*log10(21 - 4*r_r + 4*r_r^2)
logL_r <- (-x[,"n_00"] - x[,"n_01"] + x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] + x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_24"])*(log(pmax(1e-6,r_r)) + log(2 + r_r)) + (x[,"n_10"] + x[,"n_14"])*log(3 + 2*r_r - 2*r_r^2) + (x[,"n_11"] + x[,"n_13"])*log(21 + 2*r_r - 2*r_r^2) + (x[,"n_02"] + x[,"n_22"])*log(3 + r_r - r_r^2) + (x[,"n_03"] + x[,"n_21"])*log(12 - 10*r_r + r_r^2) + (x[,"n_04"] + x[,"n_20"])*log(3 - 4*r_r + r_r^2) + (x[,"n_01"] + x[,"n_23"])*log(3 + 8*r_r + r_r^2) + x[,"n_12"]*log(21 - 4*r_r + 4*r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_2.3 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - n15 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(5) + 2*(n00 + n25)*log(pmax(1e-6,1 - r)) + 2*(n05 + n20)*log(pmax(1e-6,r)) + (n10 + n15)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n21)*(log(pmax(1e-6,r)) + log(2 + 7*r)) + (n11 + n14)*log(3 + 14*r - 14*r^2) + (n02 + n23)*log(9 - 8*r^2) + (n03 + n22)*log(1 + 16*r - 8*r^2) + (n01 + n24)*log(9 - 16*r + 7*r^2) + (n12 + n13)*log(27 - 16*r + 16*r^2)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_c <- 2*(x[,"n_10"] + x[,"n_15"])*log10(2) + 2*(x[,"n_05"] + x[,"n_20"])*log10(2) + 2*(x[,"n_00"] + x[,"n_25"])*log10(2) - (x[,"n_03"] + x[,"n_22"])*log10(7) - (x[,"n_02"] + x[,"n_23"])*log10(7) - (x[,"n_04"] + x[,"n_21"])*(-2*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_24"])*(-2*log10(2) + log10(11)) - (x[,"n_11"] + x[,"n_14"])*(-log10(2) + log10(13)) - (x[,"n_12"] + x[,"n_13"])*log10(23) + 2*(x[,"n_00"] + x[,"n_25"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_15"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_21"])*(log10(pmax(1e-6,r_c)) + log10(2 + 7*r_c)) + (x[,"n_11"] + x[,"n_14"])*log10(3 + 14*r_c - 14*r_c^2) + (x[,"n_02"] + x[,"n_23"])*log10(9 - 8*r_c^2) + (x[,"n_03"] + x[,"n_22"])*log10(1 + 16*r_c - 8*r_c^2) + (x[,"n_01"] + x[,"n_24"])*log10(9 - 16*r_c + 7*r_c^2) + (x[,"n_12"] + x[,"n_13"])*log10(27 - 16*r_c + 16*r_c^2)
logL_c <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + 2*(x[,"n_00"] + x[,"n_25"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_15"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_21"])*(log(pmax(1e-6,r_c)) + log(2 + 7*r_c)) + (x[,"n_11"] + x[,"n_14"])*log(3 + 14*r_c - 14*r_c^2) + (x[,"n_02"] + x[,"n_23"])*log(9 - 8*r_c^2) + (x[,"n_03"] + x[,"n_22"])*log(1 + 16*r_c - 8*r_c^2) + (x[,"n_01"] + x[,"n_24"])*log(9 - 16*r_c + 7*r_c^2) + (x[,"n_12"] + x[,"n_13"])*log(27 - 16*r_c + 16*r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- 2*(-2*n00 - 2*n01 - n02 - n03 - 2*n04 - 2*n05 - 2*n10 - 2*n11 - n12 - n13 - 2*n14 - 2*n15 - 2*n20 - 2*n21 - n22 - n23 - 2*n24 - 2*n25)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(5) + (n05 + n20)*(log(2 - r) + log(pmax(1e-6,r))) + (n01 + n24)*log(12 - 2*r - 7*r^2) + (n04 + n21)*log(3 + 16*r - 7*r^2) + (n12 + n13)*log(21 + 4*r - 4*r^2) + (n00 + n25)*log(1 - r^2) + (n03 + n22)*log(7 + 2*r^2) + (n02 + n23)*log(9 - 4*r + 2*r^2) + (n10 + n15)*log(3 - 2*r + 2*r^2) + (n11 + n14)*log(33 - 14*r + 14*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_m <- (x[,"n_00"] + x[,"n_25"])*(2*log10(2) - log10(3)) - (x[,"n_05"] + x[,"n_20"])*(-2*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_15"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_22"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_23"])*(-log10(2) + log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_13"])*(log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_21"])*(-2*log10(2) + log10(37)) - (x[,"n_01"] + x[,"n_24"])*(-2*log10(2) + log10(37)) - (x[,"n_11"] + x[,"n_14"])*(-log10(2) + log10(59)) + (x[,"n_05"] + x[,"n_20"])*(log10(2 - r_m) + log10(pmax(1e-6,r_m))) + (x[,"n_01"] + x[,"n_24"])*log10(12 - 2*r_m - 7*r_m^2) + (x[,"n_04"] + x[,"n_21"])*log10(3 + 16*r_m - 7*r_m^2) + (x[,"n_12"] + x[,"n_13"])*log10(21 + 4*r_m - 4*r_m^2) + (x[,"n_00"] + x[,"n_25"])*log10(1 - r_m^2) + (x[,"n_03"] + x[,"n_22"])*log10(7 + 2*r_m^2) + (x[,"n_02"] + x[,"n_23"])*log10(9 - 4*r_m + 2*r_m^2) + (x[,"n_10"] + x[,"n_15"])*log10(3 - 2*r_m + 2*r_m^2) + (x[,"n_11"] + x[,"n_14"])*log10(33 - 14*r_m + 14*r_m^2)
logL_m <- 2*(-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_05"] + x[,"n_20"])*(log(2 - r_m) + log(pmax(1e-6,r_m))) + (x[,"n_01"] + x[,"n_24"])*log(12 - 2*r_m - 7*r_m^2) + (x[,"n_04"] + x[,"n_21"])*log(3 + 16*r_m - 7*r_m^2) + (x[,"n_12"] + x[,"n_13"])*log(21 + 4*r_m - 4*r_m^2) + (x[,"n_00"] + x[,"n_25"])*log(1 - r_m^2) + (x[,"n_03"] + x[,"n_22"])*log(7 + 2*r_m^2) + (x[,"n_02"] + x[,"n_23"])*log(9 - 4*r_m + 2*r_m^2) + (x[,"n_10"] + x[,"n_15"])*log(3 - 2*r_m + 2*r_m^2) + (x[,"n_11"] + x[,"n_14"])*log(33 - 14*r_m + 14*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- (-3*n00 - 3*n01 - 2*n02 - 2*n03 - 3*n04 - 3*n05 - 3*n10 - 3*n11 - 2*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n20 - 3*n21 - 2*n22 - 2*n23 - 3*n24 - 3*n25)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*(log(3) + 2*log(5)) + (n00 + n25)*(log(pmax(1e-6,r)) + log(2 + r)) + (n11 + n14)*log(39 + 14*r - 14*r^2) + (n03 + n22)*log(27 - 8*r - 4*r^2) + (n02 + n23)*log(15 + 16*r - 4*r^2) + (n10 + n15)*log(3 + 2*r - 2*r^2) + (n05 + n20)*log(3 - 4*r + r^2) + (n04 + n21)*log(30 - 34*r + 7*r^2) + (n01 + n24)*log(3 + 20*r + 7*r^2) + (n12 + n13)*log(69 - 8*r + 8*r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_r <- -((x[,"n_05"] + x[,"n_20"])*(-2*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_25"])*(-2*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_15"])*(-log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_22"])*(log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_23"])*(log10(2) + log10(11)) - (x[,"n_11"] + x[,"n_14"])*(-log10(2) + log10(5) + log10(17)) - (x[,"n_04"] + x[,"n_21"])*(-2*log10(2) + log10(59)) - (x[,"n_01"] + x[,"n_24"])*(-2*log10(2) + log10(59)) - (x[,"n_12"] + x[,"n_13"])*log10(67) + (x[,"n_00"] + x[,"n_25"])*(log10(pmax(1e-6,r_r)) + log10(2 + r_r)) + (x[,"n_11"] + x[,"n_14"])*log10(39 + 14*r_r - 14*r_r^2) + (x[,"n_03"] + x[,"n_22"])*log10(27 - 8*r_r - 4*r_r^2) + (x[,"n_02"] + x[,"n_23"])*log10(15 + 16*r_r - 4*r_r^2) + (x[,"n_10"] + x[,"n_15"])*log10(3 + 2*r_r - 2*r_r^2) + (x[,"n_05"] + x[,"n_20"])*log10(3 - 4*r_r + r_r^2) + (x[,"n_04"] + x[,"n_21"])*log10(30 - 34*r_r + 7*r_r^2) + (x[,"n_01"] + x[,"n_24"])*log10(3 + 20*r_r + 7*r_r^2) + (x[,"n_12"] + x[,"n_13"])*log10(69 - 8*r_r + 8*r_r^2)
logL_r <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_25"])*(log(pmax(1e-6,r_r)) + log(2 + r_r)) + (x[,"n_11"] + x[,"n_14"])*log(39 + 14*r_r - 14*r_r^2) + (x[,"n_03"] + x[,"n_22"])*log(27 - 8*r_r - 4*r_r^2) + (x[,"n_02"] + x[,"n_23"])*log(15 + 16*r_r - 4*r_r^2) + (x[,"n_10"] + x[,"n_15"])*log(3 + 2*r_r - 2*r_r^2) + (x[,"n_05"] + x[,"n_20"])*log(3 - 4*r_r + r_r^2) + (x[,"n_04"] + x[,"n_21"])*log(30 - 34*r_r + 7*r_r^2) + (x[,"n_01"] + x[,"n_24"])*log(3 + 20*r_r + 7*r_r^2) + (x[,"n_12"] + x[,"n_13"])*log(69 - 8*r_r + 8*r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_2.4 <- function(x){
logL_c <- function(r,n11,n15,n01,n02,n03,n04,n05,n12,n13,n14,n21,n22,n23,n24,n25) {
L <- (n11 + n15)*log(2) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25)*log(5) + 2*(n01 + n25)*log(pmax(1e-6,1 - r)) + 2*(n05 + n21)*log(pmax(1e-6,r)) + (n11 + n15)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n22)*(log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n23)*log(1 + 4*r - 4*r^2) + (n12 + n14)*log(3 + 2*r - 2*r^2) + (n02 + n24)*log(3 - 4*r + r^2) + n13*log(9 - 8*r + 8*r^2)
return(L)}
interlogL_c <- function(n11,n15,n01,n02,n03,n04,n05,n12,n13,n14,n21,n22,n23,n24,n25) {
optimize(logL_c,c(0,0.5), n11,n15,n01,n02,n03,n04,n05,n12,n13,n14,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_11"],x[,"n_15"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_c <- 2*(x[,"n_11"] + x[,"n_15"])*log10(2) + 2*(x[,"n_05"] + x[,"n_21"])*log10(2) - (x[,"n_03"] + x[,"n_23"])*log10(2) + 2*(x[,"n_01"] + x[,"n_25"])*log10(2) - (x[,"n_04"] + x[,"n_22"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - x[,"n_13"]*log10(7) - (x[,"n_12"] + x[,"n_14"])*(-log10(2) + log10(7)) + 2*(x[,"n_01"] + x[,"n_25"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_21"])*log10(pmax(1e-6,r_c)) + (x[,"n_11"] + x[,"n_15"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_22"])*(log10(pmax(1e-6,r_c)) + log10(2 + r_c)) + (x[,"n_03"] + x[,"n_23"])*log10(1 + 4*r_c - 4*r_c^2) + (x[,"n_12"] + x[,"n_14"])*log10(3 + 2*r_c - 2*r_c^2) + (x[,"n_02"] + x[,"n_24"])*log10(3 - 4*r_c + r_c^2) + x[,"n_13"]*log10(9 - 8*r_c + 8*r_c^2)
logL_c <- (x[,"n_11"] + x[,"n_15"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + 2*(x[,"n_01"] + x[,"n_25"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_21"])*log(pmax(1e-6,r_c)) + (x[,"n_11"] + x[,"n_15"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_22"])*(log(pmax(1e-6,r_c)) + log(2 + r_c)) + (x[,"n_03"] + x[,"n_23"])*log(1 + 4*r_c - 4*r_c^2) + (x[,"n_12"] + x[,"n_14"])*log(3 + 2*r_c - 2*r_c^2) + (x[,"n_02"] + x[,"n_24"])*log(3 - 4*r_c + r_c^2) + x[,"n_13"]*log(9 - 8*r_c + 8*r_c^2)
logL_m <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25) {
L <- (-2*n01 - 2*n02 - n03 - 2*n04 - 2*n05 - 2*n11 - 2*n12 + n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - n23 - 2*n24 - 2*n25)*log(2) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25)*log(5) + (n05 + n21)*(log(2 - r) + log(pmax(1e-6,r))) + (n01 + n25)*log(1 - r^2) + (n02 + n24)*log(6 - 2*r - r^2) + n13*log(3 + r - r^2) + (n04 + n22)*log(3 + 4*r - r^2) + (n11 + n15)*log(3 - 2*r + 2*r^2) + (n03 + n23)*log(5 - 2*r + 2*r^2) + (n12 + n14)*log(15 - 2*r + 2*r^2)
return(L)}
interlogL_m <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25) {
optimize(logL_m,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_m <- (x[,"n_01"] + x[,"n_25"])*(2*log10(2) - log10(3)) - (x[,"n_05"] + x[,"n_21"])*(-2*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_23"])*(-log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_15"])*(-log10(2) + log10(5)) - x[,"n_13"]*(-2*log10(2) + log10(13)) - (x[,"n_04"] + x[,"n_22"])*(-2*log10(2) + log10(19)) - (x[,"n_02"] + x[,"n_24"])*(-2*log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_14"])*(-log10(2) + log10(29)) + (x[,"n_05"] + x[,"n_21"])*(log10(2 - r_m) + log10(pmax(1e-6,r_m))) + (x[,"n_01"] + x[,"n_25"])*log10(1 - r_m^2) + (x[,"n_02"] + x[,"n_24"])*log10(6 - 2*r_m - r_m^2) + x[,"n_13"]*log10(3 + r_m - r_m^2) + (x[,"n_04"] + x[,"n_22"])*log10(3 + 4*r_m - r_m^2) + (x[,"n_11"] + x[,"n_15"])*log10(3 - 2*r_m + 2*r_m^2) + (x[,"n_03"] + x[,"n_23"])*log10(5 - 2*r_m + 2*r_m^2) + (x[,"n_12"] + x[,"n_14"])*log10(15 - 2*r_m + 2*r_m^2)
logL_m <- (-2*x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] + x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_05"] + x[,"n_21"])*(log(2 - r_m) + log(pmax(1e-6,r_m))) + (x[,"n_01"] + x[,"n_25"])*log(1 - r_m^2) + (x[,"n_02"] + x[,"n_24"])*log(6 - 2*r_m - r_m^2) + x[,"n_13"]*log(3 + r_m - r_m^2) + (x[,"n_04"] + x[,"n_22"])*log(3 + 4*r_m - r_m^2) + (x[,"n_11"] + x[,"n_15"])*log(3 - 2*r_m + 2*r_m^2) + (x[,"n_03"] + x[,"n_23"])*log(5 - 2*r_m + 2*r_m^2) + (x[,"n_12"] + x[,"n_14"])*log(15 - 2*r_m + 2*r_m^2)
logL_r <- function(r,n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n22,n23,n24,n25,n13) {
L <- (-n01 - n02 + n03 - n04 - n05 - n11 - n12 - n14 - n15 - n21 - n22 + n23 - n24 - n25)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25)*(log(3) + 2*log(5)) + (n01 + n25)*(log(pmax(1e-6,r)) + log(2 + r)) + (n11 + n15)*log(3 + 2*r - 2*r^2) + (n12 + n14)*log(21 + 2*r - 2*r^2) + (n03 + n23)*log(3 + r - r^2) + (n04 + n22)*log(12 - 10*r + r^2) + (n05 + n21)*log(3 - 4*r + r^2) + (n02 + n24)*log(3 + 8*r + r^2) + n13*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_r <- function(n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n22,n23,n24,n25,n13) {
optimize(logL_r,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n22,n23,n24,n25,n13, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_13"])
LOD_r <- -((x[,"n_05"] + x[,"n_21"])*(-2*log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_25"])*(-2*log10(2) + log10(5)) - x[,"n_13"]*(2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_15"])*(-log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_23"])*(-2*log10(2) + log10(13)) - (x[,"n_04"] + x[,"n_22"])*(-2*log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_24"])*(-2*log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_14"])*(-log10(2) + log10(43)) + (x[,"n_01"] + x[,"n_25"])*(log10(pmax(1e-6,r_r)) + log10(2 + r_r)) + (x[,"n_11"] + x[,"n_15"])*log10(3 + 2*r_r - 2*r_r^2) + (x[,"n_12"] + x[,"n_14"])*log10(21 + 2*r_r - 2*r_r^2) + (x[,"n_03"] + x[,"n_23"])*log10(3 + r_r - r_r^2) + (x[,"n_04"] + x[,"n_22"])*log10(12 - 10*r_r + r_r^2) + (x[,"n_05"] + x[,"n_21"])*log10(3 - 4*r_r + r_r^2) + (x[,"n_02"] + x[,"n_24"])*log10(3 + 8*r_r + r_r^2) + x[,"n_13"]*log10(21 - 4*r_r + 4*r_r^2)
logL_r <- (-x[,"n_01"] - x[,"n_02"] + x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] + x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*(log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_25"])*(log(pmax(1e-6,r_r)) + log(2 + r_r)) + (x[,"n_11"] + x[,"n_15"])*log(3 + 2*r_r - 2*r_r^2) + (x[,"n_12"] + x[,"n_14"])*log(21 + 2*r_r - 2*r_r^2) + (x[,"n_03"] + x[,"n_23"])*log(3 + r_r - r_r^2) + (x[,"n_04"] + x[,"n_22"])*log(12 - 10*r_r + r_r^2) + (x[,"n_05"] + x[,"n_21"])*log(3 - 4*r_r + r_r^2) + (x[,"n_02"] + x[,"n_24"])*log(3 + 8*r_r + r_r^2) + x[,"n_13"]*log(21 - 4*r_r + 4*r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_3.0 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - n10 - n11 - n12 - n13 - 2*n20 - 2*n21 - 2*n22 - 2*n23)*log(2) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*log(5) + 2*(n00 + n23)*log(pmax(1e-6,1 - r)) + 2*(n03 + n20)*log(pmax(1e-6,r)) + (n10 + n13)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n21)*(log(4 - r) + log(pmax(1e-6,r))) + (n01 + n22)*log(3 - 2*r - r^2) + (n11 + n12)*log(3 - r + r^2)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_c <- 2*(x[,"n_10"] + x[,"n_13"])*log10(2) + 2*(x[,"n_03"] + x[,"n_20"])*log10(2) + 2*(x[,"n_00"] + x[,"n_23"])*log10(2) - (x[,"n_02"] + x[,"n_21"])*(-2*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_22"])*(-2*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_12"])*(-2*log10(2) + log10(11)) + 2*(x[,"n_00"] + x[,"n_23"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_03"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_13"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_21"])*(log10(4 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_01"] + x[,"n_22"])*log10(3 - 2*r_c - r_c^2) + (x[,"n_11"] + x[,"n_12"])*log10(3 - r_c + r_c^2)
logL_c <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*log(5) + 2*(x[,"n_00"] + x[,"n_23"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_03"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_13"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_21"])*(log(4 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_01"] + x[,"n_22"])*log(3 - 2*r_c - r_c^2) + (x[,"n_11"] + x[,"n_12"])*log(3 - r_c + r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
L <- (-n00 - n01 - n02 - n03 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - n20 - n21 - n22 - n23)*log(2) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*(log(3) + log(5)) + (n00 + n03 + n20 + n23)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n11 + n12)*log(15 + 4*r - 4*r^2) + (n01 + n02 + n21 + n22)*log(3 - r + r^2) + (n10 + n13)*log(3 - 4*r + 4*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_m <- -4*(x[,"n_11"] + x[,"n_12"])*log10(2) - (x[,"n_10"] + x[,"n_13"])*log10(2) + 2*(x[,"n_00"] + x[,"n_03"] + x[,"n_20"] + x[,"n_23"])*log10(2) - (x[,"n_01"] + x[,"n_02"] + x[,"n_21"] + x[,"n_22"])*(-2*log10(2) + log10(11)) + (x[,"n_00"] + x[,"n_03"] + x[,"n_20"] + x[,"n_23"])*(log10(pmax(1e-6,1 - r_m)) + log10(pmax(1e-6,r_m))) + (x[,"n_11"] + x[,"n_12"])*log10(15 + 4*r_m - 4*r_m^2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_21"] + x[,"n_22"])*log10(3 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_13"])*log10(3 - 4*r_m + 4*r_m^2)
logL_m <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*(log(3) + log(5)) + (x[,"n_00"] + x[,"n_03"] + x[,"n_20"] + x[,"n_23"])*(log(pmax(1e-6,1 - r_m)) + log(pmax(1e-6,r_m))) + (x[,"n_11"] + x[,"n_12"])*log(15 + 4*r_m - 4*r_m^2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_21"] + x[,"n_22"])*log(3 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_13"])*log(3 - 4*r_m + 4*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - n10 - n11 - n12 - n13 - 2*n20 - 2*n21 - 2*n22 - 2*n23)*log(2) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23)*log(5) + 2*(n03 + n20)*log(pmax(1e-6,1 - r)) + 2*(n00 + n23)*log(pmax(1e-6,r)) + (n10 + n13)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n22)*(log(4 - r) + log(pmax(1e-6,r))) + (n02 + n21)*log(3 - 2*r - r^2) + (n11 + n12)*log(3 - r + r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"])
LOD_r <- 2*(x[,"n_10"] + x[,"n_13"])*log10(2) + 2*(x[,"n_03"] + x[,"n_20"])*log10(2) + 2*(x[,"n_00"] + x[,"n_23"])*log10(2) - (x[,"n_02"] + x[,"n_21"])*(-2*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_22"])*(-2*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_12"])*(-2*log10(2) + log10(11)) + 2*(x[,"n_03"] + x[,"n_20"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_23"])*log10(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_13"])*(log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_22"])*(log10(4 - r_r) + log10(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_21"])*log10(3 - 2*r_r - r_r^2) + (x[,"n_11"] + x[,"n_12"])*log10(3 - r_r + r_r^2)
logL_r <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"])*log(5) + 2*(x[,"n_03"] + x[,"n_20"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_23"])*log(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_13"])*(log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_22"])*(log(4 - r_r) + log(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_21"])*log(3 - 2*r_r - r_r^2) + (x[,"n_11"] + x[,"n_12"])*log(3 - r_r + r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_3.1 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-3*n00 - n01 - 3*n02 - n03 - 3*n04 - 2*n10 - 2*n11 - n12 - 2*n13 - 2*n14 - 3*n20 - n21 - 3*n22 - n23 - 3*n24)*log(2) + (n11 + n13)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + (n01 + n23)*log(pmax(1e-6,1 - r)) + 2*(n00 + n24)*log(pmax(1e-6,1 - r)) + 2*(n04 + n20)*log(pmax(1e-6,r)) + (n03 + n21)*log(pmax(1e-6,r)) + (n10 + n14)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n22)*log(3 + 2*r - 2*r^2) + n12*log(3 - r + r^2)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_c <- 2*(x[,"n_10"] + x[,"n_14"])*log10(2) + 2*(x[,"n_04"] + x[,"n_20"])*log10(2) + (x[,"n_03"] + x[,"n_21"])*log10(2) + (x[,"n_01"] + x[,"n_23"])*log10(2) + 2*(x[,"n_00"] + x[,"n_24"])*log10(2) - (x[,"n_02"] + x[,"n_22"])*(-log10(2) + log10(7)) - x[,"n_12"]*(-2*log10(2) + log10(11)) + (x[,"n_01"] + x[,"n_23"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_00"] + x[,"n_24"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_04"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_21"])*log10(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_14"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_22"])*log10(3 + 2*r_c - 2*r_c^2) + x[,"n_12"]*log10(3 - r_c + r_c^2)
logL_c <- (-3*x[,"n_00"] - x[,"n_01"] - 3*x[,"n_02"] - x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_20"] - x[,"n_21"] - 3*x[,"n_22"] - x[,"n_23"] - 3*x[,"n_24"])*log(2) + (x[,"n_11"] + x[,"n_13"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_01"] + x[,"n_23"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_00"] + x[,"n_24"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_04"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_21"])*log(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_14"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_22"])*log(3 + 2*r_c - 2*r_c^2) + x[,"n_12"]*log(3 - r_c + r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-2*n00 - 2*n01 - n02 - 2*n03 - 2*n04 - 3*n10 - 2*n11 - 2*n12 - 2*n13 - 3*n14 - 2*n20 - 2*n21 - n22 - 2*n23 - 2*n24)*log(2) + (-n00 - n02 - n04 - n10 + n11 - n12 + n13 - n14 - n20 - n22 - n24)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + (n00 + n04 + n20 + n24)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n12*log(15 + 4*r - 4*r^2) + (n02 + n22)*log(3 - r + r^2) + (n10 + n14)*log(3 - 4*r + 4*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_m <- -4*x[,"n_12"]*log10(2) - (x[,"n_10"] + x[,"n_14"])*log10(2) + 2*(x[,"n_00"] + x[,"n_04"] + x[,"n_20"] + x[,"n_24"])*log10(2) - (x[,"n_02"] + x[,"n_22"])*(-2*log10(2) + log10(11)) + (x[,"n_00"] + x[,"n_04"] + x[,"n_20"] + x[,"n_24"])*(log10(pmax(1e-6,1 - r_m)) + log10(pmax(1e-6,r_m))) + x[,"n_12"]*log10(15 + 4*r_m - 4*r_m^2) + (x[,"n_02"] + x[,"n_22"])*log10(3 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_14"])*log10(3 - 4*r_m + 4*r_m^2)
logL_m <- (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 3*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"])*log(2) + (-x[,"n_00"] - x[,"n_02"] - x[,"n_04"] - x[,"n_10"] + x[,"n_11"] - x[,"n_12"] + x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_22"] - x[,"n_24"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + (x[,"n_00"] + x[,"n_04"] + x[,"n_20"] + x[,"n_24"])*(log(pmax(1e-6,1 - r_m)) + log(pmax(1e-6,r_m))) + x[,"n_12"]*log(15 + 4*r_m - 4*r_m^2) + (x[,"n_02"] + x[,"n_22"])*log(3 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_14"])*log(3 - 4*r_m + 4*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
L <- (-3*n00 - n01 - 3*n02 - n03 - 3*n04 - 2*n10 - 2*n11 - n12 - 2*n13 - 2*n14 - 3*n20 - n21 - 3*n22 - n23 - 3*n24)*log(2) + (n11 + n13)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24)*log(5) + 2*(n04 + n20)*log(pmax(1e-6,1 - r)) + (n03 + n21)*log(pmax(1e-6,1 - r)) + (n01 + n23)*log(pmax(1e-6,r)) + 2*(n00 + n24)*log(pmax(1e-6,r)) + (n10 + n14)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n22)*log(3 + 2*r - 2*r^2) + n12*log(3 - r + r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_r <- 2*(x[,"n_10"] + x[,"n_14"])*log10(2) + 2*(x[,"n_04"] + x[,"n_20"])*log10(2) + (x[,"n_03"] + x[,"n_21"])*log10(2) + (x[,"n_01"] + x[,"n_23"])*log10(2) + 2*(x[,"n_00"] + x[,"n_24"])*log10(2) - (x[,"n_02"] + x[,"n_22"])*(-log10(2) + log10(7)) - x[,"n_12"]*(-2*log10(2) + log10(11)) + 2*(x[,"n_04"] + x[,"n_20"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_03"] + x[,"n_21"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_01"] + x[,"n_23"])*log10(pmax(1e-6,r_r)) + 2*(x[,"n_00"] + x[,"n_24"])*log10(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_14"])*(log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_22"])*log10(3 + 2*r_r - 2*r_r^2) + x[,"n_12"]*log10(3 - r_r + r_r^2)
logL_r <- (-3*x[,"n_00"] - x[,"n_01"] - 3*x[,"n_02"] - x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_20"] - x[,"n_21"] - 3*x[,"n_22"] - x[,"n_23"] - 3*x[,"n_24"])*log(2) + (x[,"n_11"] + x[,"n_13"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + 2*(x[,"n_04"] + x[,"n_20"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_03"] + x[,"n_21"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_01"] + x[,"n_23"])*log(pmax(1e-6,r_r)) + 2*(x[,"n_00"] + x[,"n_24"])*log(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_14"])*(log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_22"])*log(3 + 2*r_r - 2*r_r^2) + x[,"n_12"]*log(3 - r_r + r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_3.2 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- (-2*n00 - n01 - 2*n02 - 2*n03 - n04 - 2*n05 - n10 - n11 - n12 - n13 - n14 - n15 - 2*n20 - n21 - 2*n22 - 2*n23 - n24 - 2*n25)*log(2) + (n12 + n13)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(5) + 2*(n00 + n25)*log(pmax(1e-6,1 - r)) + 2*(n05 + n20)*log(pmax(1e-6,r)) + (n10 + n15)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n21)*(log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n23)*log(10 - 4*r - 3*r^2) + (n03 + n22)*log(3 + 10*r - 3*r^2) + (n11 + n14)*log(3 + 2*r - 2*r^2) + (n01 + n24)*log(3 - 4*r + r^2) + (n12 + n13)*log(4 - r + r^2)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_c <- 2*(x[,"n_10"] + x[,"n_15"])*log10(2) + 2*(x[,"n_05"] + x[,"n_20"])*log10(2) + 2*(x[,"n_00"] + x[,"n_25"])*log10(2) - (x[,"n_04"] + x[,"n_21"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_13"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_14"])*(-log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_22"])*(-2*log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_23"])*(-2*log10(2) + log10(29)) + 2*(x[,"n_00"] + x[,"n_25"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_15"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_21"])*(log10(pmax(1e-6,r_c)) + log10(2 + r_c)) + (x[,"n_02"] + x[,"n_23"])*log10(10 - 4*r_c - 3*r_c^2) + (x[,"n_03"] + x[,"n_22"])*log10(3 + 10*r_c - 3*r_c^2) + (x[,"n_11"] + x[,"n_14"])*log10(3 + 2*r_c - 2*r_c^2) + (x[,"n_01"] + x[,"n_24"])*log10(3 - 4*r_c + r_c^2) + (x[,"n_12"] + x[,"n_13"])*log10(4 - r_c + r_c^2)
logL_c <- (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - 2*x[,"n_25"])*log(2) + (x[,"n_12"] + x[,"n_13"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + 2*(x[,"n_00"] + x[,"n_25"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_15"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_21"])*(log(pmax(1e-6,r_c)) + log(2 + r_c)) + (x[,"n_02"] + x[,"n_23"])*log(10 - 4*r_c - 3*r_c^2) + (x[,"n_03"] + x[,"n_22"])*log(3 + 10*r_c - 3*r_c^2) + (x[,"n_11"] + x[,"n_14"])*log(3 + 2*r_c - 2*r_c^2) + (x[,"n_01"] + x[,"n_24"])*log(3 - 4*r_c + r_c^2) + (x[,"n_12"] + x[,"n_13"])*log(4 - r_c + r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - 2*n10 + n11 - 2*n12 - 2*n13 + n14 - 2*n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(2) + (-n00 - n01 - n04 - n05 - n10 - n11 - n14 - n15 - n20 - n21 - n24 - n25)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(5) + (n00 + n05 + n20 + n25)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n12 + n13)*log(21 + 4*r - 4*r^2) + (n01 + n04 + n21 + n24)*log(3 + 2*r - 2*r^2) + (n11 + n14)*log(3 - r + r^2) + (n02 + n03 + n22 + n23)*log(4 - r + r^2) + (n10 + n15)*log(3 - 4*r + 4*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_m <- -((x[,"n_10"] + x[,"n_15"])*log10(2)) + 2*(x[,"n_00"] + x[,"n_05"] + x[,"n_20"] + x[,"n_25"])*log10(2) - (x[,"n_02"] + x[,"n_03"] + x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_01"] + x[,"n_04"] + x[,"n_21"] + x[,"n_24"])*(-log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_14"])*(-2*log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_13"])*(log10(2) + log10(11)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_m)) + log10(pmax(1e-6,r_m))) + (x[,"n_12"] + x[,"n_13"])*log10(21 + 4*r_m - 4*r_m^2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_21"] + x[,"n_24"])*log10(3 + 2*r_m - 2*r_m^2) + (x[,"n_11"] + x[,"n_14"])*log10(3 - r_m + r_m^2) + (x[,"n_02"] + x[,"n_03"] + x[,"n_22"] + x[,"n_23"])*log10(4 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_15"])*log10(3 - 4*r_m + 4*r_m^2)
logL_m <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - 2*x[,"n_10"] + x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] + x[,"n_14"] - 2*x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + (x[,"n_00"] + x[,"n_05"] + x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_m)) + log(pmax(1e-6,r_m))) + (x[,"n_12"] + x[,"n_13"])*log(21 + 4*r_m - 4*r_m^2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_21"] + x[,"n_24"])*log(3 + 2*r_m - 2*r_m^2) + (x[,"n_11"] + x[,"n_14"])*log(3 - r_m + r_m^2) + (x[,"n_02"] + x[,"n_03"] + x[,"n_22"] + x[,"n_23"])*log(4 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_15"])*log(3 - 4*r_m + 4*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
L <- (-2*n00 - n01 - 2*n02 - 2*n03 - n04 - 2*n05 - n10 - n11 - n12 - n13 - n14 - n15 - 2*n20 - n21 - 2*n22 - 2*n23 - n24 - 2*n25)*log(2) + (n12 + n13)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25)*log(5) + 2*(n05 + n20)*log(pmax(1e-6,1 - r)) + 2*(n00 + n25)*log(pmax(1e-6,r)) + (n10 + n15)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n24)*(log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n22)*log(10 - 4*r - 3*r^2) + (n02 + n23)*log(3 + 10*r - 3*r^2) + (n11 + n14)*log(3 + 2*r - 2*r^2) + (n04 + n21)*log(3 - 4*r + r^2) + (n12 + n13)*log(4 - r + r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"])
LOD_r <- 2*(x[,"n_10"] + x[,"n_15"])*log10(2) + 2*(x[,"n_05"] + x[,"n_20"])*log10(2) + 2*(x[,"n_00"] + x[,"n_25"])*log10(2) - (x[,"n_04"] + x[,"n_21"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_13"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_14"])*(-log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_22"])*(-2*log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_23"])*(-2*log10(2) + log10(29)) + 2*(x[,"n_05"] + x[,"n_20"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_25"])*log10(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_15"])*(log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_24"])*(log10(pmax(1e-6,r_r)) + log10(2 + r_r)) + (x[,"n_03"] + x[,"n_22"])*log10(10 - 4*r_r - 3*r_r^2) + (x[,"n_02"] + x[,"n_23"])*log10(3 + 10*r_r - 3*r_r^2) + (x[,"n_11"] + x[,"n_14"])*log10(3 + 2*r_r - 2*r_r^2) + (x[,"n_04"] + x[,"n_21"])*log10(3 - 4*r_r + r_r^2) + (x[,"n_12"] + x[,"n_13"])*log10(4 - r_r + r_r^2)
logL_r <- (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - 2*x[,"n_25"])*log(2) + (x[,"n_12"] + x[,"n_13"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"])*log(5) + 2*(x[,"n_05"] + x[,"n_20"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_25"])*log(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_15"])*(log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_24"])*(log(pmax(1e-6,r_r)) + log(2 + r_r)) + (x[,"n_03"] + x[,"n_22"])*log(10 - 4*r_r - 3*r_r^2) + (x[,"n_02"] + x[,"n_23"])*log(3 + 10*r_r - 3*r_r^2) + (x[,"n_11"] + x[,"n_14"])*log(3 + 2*r_r - 2*r_r^2) + (x[,"n_04"] + x[,"n_21"])*log(3 - 4*r_r + r_r^2) + (x[,"n_12"] + x[,"n_13"])*log(4 - r_r + r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_3.3 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
L <- (-4*n00 - 2*n01 - 4*n02 - 2*n03 - 4*n04 - 2*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 2*n21 - 4*n22 - 2*n23 - 4*n24 - 2*n25 - 4*n26)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26)*log(5) + 2*(n00 + n26)*log(pmax(1e-6,1 - r)) + 2*(n06 + n20)*log(pmax(1e-6,r)) + (n10 + n16)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n21)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n11 + n15)*log(3 + 8*r - 8*r^2) + (n03 + n23)*log(7 + 4*r - 4*r^2) + (n02 + n24)*log(36 - 32*r - r^2) + (n04 + n22)*log(3 + 34*r - r^2) + (n12 + n14)*log(30 - r + r^2) + (n01 + n25)*log(3 - 5*r + 2*r^2) + n13*log(27 - 8*r + 8*r^2)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"])
LOD_c <- 2*(x[,"n_10"] + x[,"n_16"])*log10(2) + 2*(x[,"n_06"] + x[,"n_20"])*log10(2) - 3*(x[,"n_03"] + x[,"n_23"])*log10(2) + 2*(x[,"n_00"] + x[,"n_26"])*log10(2) - 2*x[,"n_13"]*log10(5) - (x[,"n_11"] + x[,"n_15"])*log10(5) - (x[,"n_12"] + x[,"n_14"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_04"] + x[,"n_22"])*(-2*log10(2) + log10(79)) - (x[,"n_02"] + x[,"n_24"])*(-2*log10(2) + log10(79)) + 2*(x[,"n_00"] + x[,"n_26"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_06"] + x[,"n_20"])*log10(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_16"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_05"] + x[,"n_21"])*(log10(pmax(1e-6,r_c)) + log10(1 + 2*r_c)) + (x[,"n_11"] + x[,"n_15"])*log10(3 + 8*r_c - 8*r_c^2) + (x[,"n_03"] + x[,"n_23"])*log10(7 + 4*r_c - 4*r_c^2) + (x[,"n_02"] + x[,"n_24"])*log10(36 - 32*r_c - r_c^2) + (x[,"n_04"] + x[,"n_22"])*log10(3 + 34*r_c - r_c^2) + (x[,"n_12"] + x[,"n_14"])*log10(30 - r_c + r_c^2) + (x[,"n_01"] + x[,"n_25"])*log10(3 - 5*r_c + 2*r_c^2) + x[,"n_13"]*log10(27 - 8*r_c + 8*r_c^2)
logL_c <- (-4*x[,"n_00"] - 2*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 2*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 2*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 2*x[,"n_25"] - 4*x[,"n_26"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"])*log(5) + 2*(x[,"n_00"] + x[,"n_26"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_06"] + x[,"n_20"])*log(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_16"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_05"] + x[,"n_21"])*(log(pmax(1e-6,r_c)) + log(1 + 2*r_c)) + (x[,"n_11"] + x[,"n_15"])*log(3 + 8*r_c - 8*r_c^2) + (x[,"n_03"] + x[,"n_23"])*log(7 + 4*r_c - 4*r_c^2) + (x[,"n_02"] + x[,"n_24"])*log(36 - 32*r_c - r_c^2) + (x[,"n_04"] + x[,"n_22"])*log(3 + 34*r_c - r_c^2) + (x[,"n_12"] + x[,"n_14"])*log(30 - r_c + r_c^2) + (x[,"n_01"] + x[,"n_25"])*log(3 - 5*r_c + 2*r_c^2) + x[,"n_13"]*log(27 - 8*r_c + 8*r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
L <- (-3*n00 - 3*n01 - 3*n02 - 2*n03 - 3*n04 - 3*n05 - 3*n06 - 4*n10 - 3*n11 - 4*n12 - 2*n13 - 4*n14 - 3*n15 - 4*n16 - 3*n20 - 3*n21 - 3*n22 - 2*n23 - 3*n24 - 3*n25 - 3*n26)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26)*(log(3) + 2*log(5)) + (n00 + n06 + n20 + n26)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + n13*log(69 + 16*r - 16*r^2) + (n01 + n05 + n21 + n25)*log(3 + 8*r - 8*r^2) + (n12 + n14)*log(177 + 4*r - 4*r^2) + (n02 + n04 + n22 + n24)*log(30 - r + r^2) + (n10 + n16)*log(3 - 4*r + 4*r^2) + (n03 + n23)*log(27 - 8*r + 8*r^2) + (n11 + n15)*log(21 - 16*r + 16*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"])
LOD_m <- -((x[,"n_10"] + x[,"n_16"])*log10(2)) + 2*(x[,"n_00"] + x[,"n_06"] + x[,"n_20"] + x[,"n_26"])*log10(2) - 2*(x[,"n_03"] + x[,"n_23"])*log10(5) - (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*log10(5) - (x[,"n_11"] + x[,"n_15"])*log10(17) - (x[,"n_02"] + x[,"n_04"] + x[,"n_22"] + x[,"n_24"])*(-2*log10(2) + log10(7) + log10(17)) - x[,"n_13"]*log10(73) - (x[,"n_12"] + x[,"n_14"])*(log10(2) + log10(89)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_20"] + x[,"n_26"])*(log10(pmax(1e-6,1 - r_m)) + log10(pmax(1e-6,r_m))) + x[,"n_13"]*log10(69 + 16*r_m - 16*r_m^2) + (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*log10(3 + 8*r_m - 8*r_m^2) + (x[,"n_12"] + x[,"n_14"])*log10(177 + 4*r_m - 4*r_m^2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_22"] + x[,"n_24"])*log10(30 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_16"])*log10(3 - 4*r_m + 4*r_m^2) + (x[,"n_03"] + x[,"n_23"])*log10(27 - 8*r_m + 8*r_m^2) + (x[,"n_11"] + x[,"n_15"])*log10(21 - 16*r_m + 16*r_m^2)
logL_m <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_06"] - 4*x[,"n_10"] - 3*x[,"n_11"] - 4*x[,"n_12"] - 2*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_16"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_26"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_20"] + x[,"n_26"])*(log(pmax(1e-6,1 - r_m)) + log(pmax(1e-6,r_m))) + x[,"n_13"]*log(69 + 16*r_m - 16*r_m^2) + (x[,"n_01"] + x[,"n_05"] + x[,"n_21"] + x[,"n_25"])*log(3 + 8*r_m - 8*r_m^2) + (x[,"n_12"] + x[,"n_14"])*log(177 + 4*r_m - 4*r_m^2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_22"] + x[,"n_24"])*log(30 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_16"])*log(3 - 4*r_m + 4*r_m^2) + (x[,"n_03"] + x[,"n_23"])*log(27 - 8*r_m + 8*r_m^2) + (x[,"n_11"] + x[,"n_15"])*log(21 - 16*r_m + 16*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
L <- (-4*n00 - 2*n01 - 4*n02 - 2*n03 - 4*n04 - 2*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 2*n21 - 4*n22 - 2*n23 - 4*n24 - 2*n25 - 4*n26)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26)*log(5) + 2*(n06 + n20)*log(pmax(1e-6,1 - r)) + 2*(n00 + n26)*log(pmax(1e-6,r)) + (n10 + n16)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n25)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n11 + n15)*log(3 + 8*r - 8*r^2) + (n03 + n23)*log(7 + 4*r - 4*r^2) + (n04 + n22)*log(36 - 32*r - r^2) + (n02 + n24)*log(3 + 34*r - r^2) + (n12 + n14)*log(30 - r + r^2) + (n05 + n21)*log(3 - 5*r + 2*r^2) + n13*log(27 - 8*r + 8*r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"])
LOD_r <- 2*(x[,"n_10"] + x[,"n_16"])*log10(2) + 2*(x[,"n_06"] + x[,"n_20"])*log10(2) - 3*(x[,"n_03"] + x[,"n_23"])*log10(2) + 2*(x[,"n_00"] + x[,"n_26"])*log10(2) - 2*x[,"n_13"]*log10(5) - (x[,"n_11"] + x[,"n_15"])*log10(5) - (x[,"n_12"] + x[,"n_14"])*(-2*log10(2) + log10(7) + log10(17)) - (x[,"n_04"] + x[,"n_22"])*(-2*log10(2) + log10(79)) - (x[,"n_02"] + x[,"n_24"])*(-2*log10(2) + log10(79)) + 2*(x[,"n_06"] + x[,"n_20"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_26"])*log10(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_16"])*(log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_25"])*(log10(pmax(1e-6,r_r)) + log10(1 + 2*r_r)) + (x[,"n_11"] + x[,"n_15"])*log10(3 + 8*r_r - 8*r_r^2) + (x[,"n_03"] + x[,"n_23"])*log10(7 + 4*r_r - 4*r_r^2) + (x[,"n_04"] + x[,"n_22"])*log10(36 - 32*r_r - r_r^2) + (x[,"n_02"] + x[,"n_24"])*log10(3 + 34*r_r - r_r^2) + (x[,"n_12"] + x[,"n_14"])*log10(30 - r_r + r_r^2) + (x[,"n_05"] + x[,"n_21"])*log10(3 - 5*r_r + 2*r_r^2) + x[,"n_13"]*log10(27 - 8*r_r + 8*r_r^2)
logL_r <- (-4*x[,"n_00"] - 2*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 2*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 2*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 2*x[,"n_25"] - 4*x[,"n_26"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"])*log(5) + 2*(x[,"n_06"] + x[,"n_20"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_26"])*log(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_16"])*(log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_25"])*(log(pmax(1e-6,r_r)) + log(1 + 2*r_r)) + (x[,"n_11"] + x[,"n_15"])*log(3 + 8*r_r - 8*r_r^2) + (x[,"n_03"] + x[,"n_23"])*log(7 + 4*r_r - 4*r_r^2) + (x[,"n_04"] + x[,"n_22"])*log(36 - 32*r_r - r_r^2) + (x[,"n_02"] + x[,"n_24"])*log(3 + 34*r_r - r_r^2) + (x[,"n_12"] + x[,"n_14"])*log(30 - r_r + r_r^2) + (x[,"n_05"] + x[,"n_21"])*log(3 - 5*r_r + 2*r_r^2) + x[,"n_13"]*log(27 - 8*r_r + 8*r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.0_4.1 <- function(x){
logL_c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24)*(2*log(2) + log(3) + log(5)) + (n04 + n21)*(log(pmax(1e-6,r)) + log(2 + r)) + (n11 + n14)*log(3 + 2*r - 2*r^2) + (n02 + n23)*log(6 - 2*r - r^2) + (n03 + n22)*log(3 + 4*r - r^2) + (n01 + n24)*log(3 - 4*r + r^2) + (n12 + n13)*log(15 - 2*r + 2*r^2)
return(L)}
interlogL_c <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
optimize(logL_c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_c <- -((x[,"n_04"] + x[,"n_21"])*(-2*log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_24"])*(-2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_14"])*(-log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_22"])*(-2*log10(2) + log10(19)) - (x[,"n_02"] + x[,"n_23"])*(-2*log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_13"])*(-log10(2) + log10(29)) + (x[,"n_04"] + x[,"n_21"])*(log10(pmax(1e-6,r_c)) + log10(2 + r_c)) + (x[,"n_11"] + x[,"n_14"])*log10(3 + 2*r_c - 2*r_c^2) + (x[,"n_02"] + x[,"n_23"])*log10(6 - 2*r_c - r_c^2) + (x[,"n_03"] + x[,"n_22"])*log10(3 + 4*r_c - r_c^2) + (x[,"n_01"] + x[,"n_24"])*log10(3 - 4*r_c + r_c^2) + (x[,"n_12"] + x[,"n_13"])*log10(15 - 2*r_c + 2*r_c^2)
logL_c <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*(2*log(2) + log(3) + log(5)) + (x[,"n_04"] + x[,"n_21"])*(log(pmax(1e-6,r_c)) + log(2 + r_c)) + (x[,"n_11"] + x[,"n_14"])*log(3 + 2*r_c - 2*r_c^2) + (x[,"n_02"] + x[,"n_23"])*log(6 - 2*r_c - r_c^2) + (x[,"n_03"] + x[,"n_22"])*log(3 + 4*r_c - r_c^2) + (x[,"n_01"] + x[,"n_24"])*log(3 - 4*r_c + r_c^2) + (x[,"n_12"] + x[,"n_13"])*log(15 - 2*r_c + 2*r_c^2)
logL_m <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24)*(3*log(2) + log(5)) + (n01 + n24)*(log(2 - r) + log(pmax(1e-6,r))) + (n12 + n13)*log(9 + 2*r - 2*r^2) + (n04 + n21)*log(1 - r^2) + (n02 + n23)*log(3 + r^2) + (n03 + n22)*log(4 - 2*r + r^2) + (n11 + n14)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_m <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24) {
optimize(logL_m,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"])
LOD_m <- (x[,"n_04"] + x[,"n_21"])*(2*log10(2) - log10(3)) - (x[,"n_01"] + x[,"n_24"])*(-2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_14"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_22"])*(-2*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_23"])*(-2*log10(2) + log10(13)) - (x[,"n_12"] + x[,"n_13"])*(-log10(2) + log10(19)) + (x[,"n_01"] + x[,"n_24"])*(log10(2 - r_m) + log10(pmax(1e-6,r_m))) + (x[,"n_12"] + x[,"n_13"])*log10(9 + 2*r_m - 2*r_m^2) + (x[,"n_04"] + x[,"n_21"])*log10(1 - r_m^2) + (x[,"n_02"] + x[,"n_23"])*log10(3 + r_m^2) + (x[,"n_03"] + x[,"n_22"])*log10(4 - 2*r_m + r_m^2) + (x[,"n_11"] + x[,"n_14"])*log10(3 - 2*r_m + 2*r_m^2)
logL_m <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*(3*log(2) + log(5)) + (x[,"n_01"] + x[,"n_24"])*(log(2 - r_m) + log(pmax(1e-6,r_m))) + (x[,"n_12"] + x[,"n_13"])*log(9 + 2*r_m - 2*r_m^2) + (x[,"n_04"] + x[,"n_21"])*log(1 - r_m^2) + (x[,"n_02"] + x[,"n_23"])*log(3 + r_m^2) + (x[,"n_03"] + x[,"n_22"])*log(4 - 2*r_m + r_m^2) + (x[,"n_11"] + x[,"n_14"])*log(3 - 2*r_m + 2*r_m^2)
logL_r <- function(r,n01,n02,n03,n04,n12,n13,n21,n22,n23,n24,n11,n14) {
L <- (-n01 - n02 - n03 - n04 - n12 - n13 - n21 - n22 - n23 - n24)*log(2) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24)*log(5) + 2*(n04 + n21)*log(pmax(1e-6,1 - r)) + 2*(n01 + n24)*log(pmax(1e-6,r)) + (n11 + n14)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n23)*(log(2 - r) + log(pmax(1e-6,r))) + (n03 + n22)*log(1 - r^2) + (n12 + n13)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_r <- function(n01,n02,n03,n04,n12,n13,n21,n22,n23,n24,n11,n14) {
optimize(logL_r,c(0,0.5), n01,n02,n03,n04,n12,n13,n21,n22,n23,n24,n11,n14, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_11"],x[,"n_14"])
LOD_r <- 2*(x[,"n_11"] + x[,"n_14"])*log10(2) + 2*(x[,"n_04"] + x[,"n_21"])*log10(2) + 2*(x[,"n_01"] + x[,"n_24"])*log10(2) + (x[,"n_03"] + x[,"n_22"])*(2*log10(2) - log10(3)) - (x[,"n_02"] + x[,"n_23"])*(-2*log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_13"])*(-log10(2) + log10(5)) + 2*(x[,"n_04"] + x[,"n_21"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_01"] + x[,"n_24"])*log10(pmax(1e-6,r_r)) + (x[,"n_11"] + x[,"n_14"])*(log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_23"])*(log10(2 - r_r) + log10(pmax(1e-6,r_r))) + (x[,"n_03"] + x[,"n_22"])*log10(1 - r_r^2) + (x[,"n_12"] + x[,"n_13"])*log10(3 - 2*r_r + 2*r_r^2)
logL_r <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"])*log(5) + 2*(x[,"n_04"] + x[,"n_21"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_01"] + x[,"n_24"])*log(pmax(1e-6,r_r)) + (x[,"n_11"] + x[,"n_14"])*(log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_23"])*(log(2 - r_r) + log(pmax(1e-6,r_r))) + (x[,"n_03"] + x[,"n_22"])*log(1 - r_r^2) + (x[,"n_12"] + x[,"n_13"])*log(3 - 2*r_r + 2*r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.1_1.5 <- function(x){
logL_cc <- function(r,n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n03,n33) {
L <- (-n02 - n04 - 2*n12 - 2*n13 - 2*n14 - 2*n22 - 2*n23 - 2*n24 - n32 - n34)*log(2) + 2*(-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*log(5) + (n02 + n34)*(log(pmax(1e-6,1 - r)) + log(3 - r)) + (n04 + n32)*(log(pmax(1e-6,r)) + log(2 + r)) + (n12 + n24)*log(13 - 5*r - 2*r^2) + (n14 + n22)*log(6 + 9*r - 2*r^2) + (n03 + n33)*log(1 + r - r^2) + (n13 + n23)*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_cc <- function(n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n03,n33) {
optimize(logL_cc,c(0,0.5), n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n03,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_02"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_34"],x[,"n_03"],x[,"n_33"])
LOD_cc <- -((x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(5))) - (x[,"n_03"] + x[,"n_33"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_22"])*(log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_24"])*(log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_23"])*(2*log10(2) + log10(5)) + (x[,"n_02"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cc)) + log10(3 - r_cc)) + (x[,"n_04"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_12"] + x[,"n_24"])*log10(13 - 5*r_cc - 2*r_cc^2) + (x[,"n_14"] + x[,"n_22"])*log10(6 + 9*r_cc - 2*r_cc^2) + (x[,"n_03"] + x[,"n_33"])*log10(1 + r_cc - r_cc^2) + (x[,"n_13"] + x[,"n_23"])*log10(21 - 4*r_cc + 4*r_cc^2)
logL_cc <- (-x[,"n_02"] - x[,"n_04"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_32"] - x[,"n_34"])*log(2) + 2*(-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_02"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cc)) + log(3 - r_cc)) + (x[,"n_04"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_12"] + x[,"n_24"])*log(13 - 5*r_cc - 2*r_cc^2) + (x[,"n_14"] + x[,"n_22"])*log(6 + 9*r_cc - 2*r_cc^2) + (x[,"n_03"] + x[,"n_33"])*log(1 + r_cc - r_cc^2) + (x[,"n_13"] + x[,"n_23"])*log(21 - 4*r_cc + 4*r_cc^2)
logL_cr <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
L <- (-n02 - n03 - n04 - 2*n12 - 2*n13 - 2*n14 - 2*n22 - 2*n23 - 2*n24 - n32 - n33 - n34)*log(2) + (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*log(5) + (n02 + n04 + n32 + n34)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n13 + n23)*log(3 + 4*r - 4*r^2) + (n14 + n22)*log(3 - 3*r + 2*r^2) + (n03 + n33)*log(1 - 2*r + 2*r^2) + (n12 + n24)*log(2 - r + 2*r^2)
return(L)}
interlogL_cr <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
optimize(logL_cr,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_cr <- -((x[,"n_14"] + x[,"n_22"])*log10(2)) - 2*(x[,"n_13"] + x[,"n_23"])*log10(2) - (x[,"n_12"] + x[,"n_24"])*log10(2) + (x[,"n_03"] + x[,"n_33"])*log10(2) + 2*(x[,"n_02"] + x[,"n_04"] + x[,"n_32"] + x[,"n_34"])*log10(2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_32"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_13"] + x[,"n_23"])*log10(3 + 4*r_cr - 4*r_cr^2) + (x[,"n_14"] + x[,"n_22"])*log10(3 - 3*r_cr + 2*r_cr^2) + (x[,"n_03"] + x[,"n_33"])*log10(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_12"] + x[,"n_24"])*log10(2 - r_cr + 2*r_cr^2)
logL_cr <- (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(2) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(5) + (x[,"n_02"] + x[,"n_04"] + x[,"n_32"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_13"] + x[,"n_23"])*log(3 + 4*r_cr - 4*r_cr^2) + (x[,"n_14"] + x[,"n_22"])*log(3 - 3*r_cr + 2*r_cr^2) + (x[,"n_03"] + x[,"n_33"])*log(1 - 2*r_cr + 2*r_cr^2) + (x[,"n_12"] + x[,"n_24"])*log(2 - r_cr + 2*r_cr^2)
logL_rc <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
L <- (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*(3*log(2) + 2*log(5)) + (n04 + n32)*(log(3 - 2*r) + log(2 + r)) + (n02 + n34)*(log(3 - r) + log(1 + 2*r)) + (n13 + n23)*log(39 + 4*r - 4*r^2) + (n14 + n22)*log(21 - 3*r + 2*r^2) + (n12 + n24)*log(20 - r + 2*r^2) + (n03 + n33)*log(11 - 4*r + 4*r^2)
return(L)}
interlogL_rc <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
optimize(logL_rc,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rc <- -((x[,"n_04"] + x[,"n_32"])*log10(5)) - (x[,"n_02"] + x[,"n_34"])*log10(5) - (x[,"n_03"] + x[,"n_33"])*(log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_22"])*(2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_24"])*(2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_23"])*(3*log10(2) + log10(5)) + (x[,"n_04"] + x[,"n_32"])*(log10(3 - 2*r_rc) + log10(2 + r_rc)) + (x[,"n_02"] + x[,"n_34"])*(log10(3 - r_rc) + log10(1 + 2*r_rc)) + (x[,"n_13"] + x[,"n_23"])*log10(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_14"] + x[,"n_22"])*log10(21 - 3*r_rc + 2*r_rc^2) + (x[,"n_12"] + x[,"n_24"])*log10(20 - r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_33"])*log10(11 - 4*r_rc + 4*r_rc^2)
logL_rc <- (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + 2*log(5)) + (x[,"n_04"] + x[,"n_32"])*(log(3 - 2*r_rc) + log(2 + r_rc)) + (x[,"n_02"] + x[,"n_34"])*(log(3 - r_rc) + log(1 + 2*r_rc)) + (x[,"n_13"] + x[,"n_23"])*log(39 + 4*r_rc - 4*r_rc^2) + (x[,"n_14"] + x[,"n_22"])*log(21 - 3*r_rc + 2*r_rc^2) + (x[,"n_12"] + x[,"n_24"])*log(20 - r_rc + 2*r_rc^2) + (x[,"n_03"] + x[,"n_33"])*log(11 - 4*r_rc + 4*r_rc^2)
logL_rr <- function(r,n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
L <- (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*(3*log(2) + log(5)) + (n04 + n32)*(log(3 - 2*r) + log(pmax(1e-6,1 - r))) + (n02 + n34)*(log(pmax(1e-6,r)) + log(1 + 2*r)) + (n03 + n33)*log(1 + 4*r - 4*r^2) + (n14 + n22)*log(6 - 3*r - 2*r^2) + (n12 + n24)*log(1 + 7*r - 2*r^2) + (n13 + n23)*log(9 - 4*r + 4*r^2)
return(L)}
interlogL_rr <- function(n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34) {
optimize(logL_rr,c(0,0.5), n02,n03,n04,n12,n13,n14,n22,n23,n24,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_rr <- -2*(x[,"n_14"] + x[,"n_22"])*log10(2) - 3*(x[,"n_13"] + x[,"n_23"])*log10(2) - 2*(x[,"n_12"] + x[,"n_24"])*log10(2) - (x[,"n_03"] + x[,"n_33"])*log10(2) + (x[,"n_04"] + x[,"n_32"])*(log10(3 - 2*r_rr) + log10(pmax(1e-6,1 - r_rr))) + (x[,"n_02"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(1 + 2*r_rr)) + (x[,"n_03"] + x[,"n_33"])*log10(1 + 4*r_rr - 4*r_rr^2) + (x[,"n_14"] + x[,"n_22"])*log10(6 - 3*r_rr - 2*r_rr^2) + (x[,"n_12"] + x[,"n_24"])*log10(1 + 7*r_rr - 2*r_rr^2) + (x[,"n_13"] + x[,"n_23"])*log10(9 - 4*r_rr + 4*r_rr^2)
logL_rr <- (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + (x[,"n_04"] + x[,"n_32"])*(log(3 - 2*r_rr) + log(pmax(1e-6,1 - r_rr))) + (x[,"n_02"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(1 + 2*r_rr)) + (x[,"n_03"] + x[,"n_33"])*log(1 + 4*r_rr - 4*r_rr^2) + (x[,"n_14"] + x[,"n_22"])*log(6 - 3*r_rr - 2*r_rr^2) + (x[,"n_12"] + x[,"n_24"])*log(1 + 7*r_rr - 2*r_rr^2) + (x[,"n_13"] + x[,"n_23"])*log(9 - 4*r_rr + 4*r_rr^2)

return(list(
r_mat = cbind(r_cc,r_cr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.1_2.4 <- function(x){
logL_cc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-2*n01 - n02 - 2*n03 - n04 - 2*n05 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - 2*n31 - n32 - 2*n33 - n34 - 2*n35)*log(2) + (n13 + n23)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*log(5) + (n01 + n35)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r))) + (n05 + n31)*(2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n03 + n33)*log(1 + 12*r - 12*r^2) + (n15 + n21)*(log(pmax(1e-6,r)) + log(2 + 5*r - 6*r^2)) + (n13 + n23)*log(7 - 4*r + 4*r^2) + (n12 + n24)*log(15 - 16*r + 16*r^2 - 12*r^3) + (n04 + n32)*log(r + 4*r^2 - 2*r^3) + (n02 + n34)*log(3 - 3*r - 2*r^2 + 2*r^3) + (n11 + n25)*log(1 + 6*r - 13*r^2 + 6*r^3) + (n14 + n22)*log(3 + 20*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_cc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_cc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_cc <- (x[,"n_05"] + x[,"n_31"])*log10(2) - 2*(x[,"n_03"] + x[,"n_33"])*log10(2) + (x[,"n_01"] + x[,"n_35"])*log10(2) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_22"])*(-log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_24"])*(-log10(2) + log10(19)) + (x[,"n_01"] + x[,"n_35"])*(log10(3 - 2*r_cc) + 2*log10(pmax(1e-6,1 - r_cc))) + (x[,"n_05"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(1 + 2*r_cc)) + (x[,"n_03"] + x[,"n_33"])*log10(1 + 12*r_cc - 12*r_cc^2) + (x[,"n_15"] + x[,"n_21"])*(log10(pmax(1e-6,r_cc)) + log10(2 + 5*r_cc - 6*r_cc^2)) + (x[,"n_13"] + x[,"n_23"])*log10(7 - 4*r_cc + 4*r_cc^2) + (x[,"n_12"] + x[,"n_24"])*log10(15 - 16*r_cc + 16*r_cc^2 - 12*r_cc^3) + (x[,"n_04"] + x[,"n_32"])*log10(r_cc + 4*r_cc^2 - 2*r_cc^3) + (x[,"n_02"] + x[,"n_34"])*log10(3 - 3*r_cc - 2*r_cc^2 + 2*r_cc^3) + (x[,"n_11"] + x[,"n_25"])*log10(1 + 6*r_cc - 13*r_cc^2 + 6*r_cc^3) + (x[,"n_14"] + x[,"n_22"])*log10(3 + 20*r_cc - 20*r_cc^2 + 12*r_cc^3)
logL_cc <- (-2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"] - 2*x[,"n_35"])*log(2) + (x[,"n_13"] + x[,"n_23"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_01"] + x[,"n_35"])*(log(3 - 2*r_cc) + 2*log(pmax(1e-6,1 - r_cc))) + (x[,"n_05"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(1 + 2*r_cc)) + (x[,"n_03"] + x[,"n_33"])*log(1 + 12*r_cc - 12*r_cc^2) + (x[,"n_15"] + x[,"n_21"])*(log(pmax(1e-6,r_cc)) + log(2 + 5*r_cc - 6*r_cc^2)) + (x[,"n_13"] + x[,"n_23"])*log(7 - 4*r_cc + 4*r_cc^2) + (x[,"n_12"] + x[,"n_24"])*log(15 - 16*r_cc + 16*r_cc^2 - 12*r_cc^3) + (x[,"n_04"] + x[,"n_32"])*log(r_cc + 4*r_cc^2 - 2*r_cc^3) + (x[,"n_02"] + x[,"n_34"])*log(3 - 3*r_cc - 2*r_cc^2 + 2*r_cc^3) + (x[,"n_11"] + x[,"n_25"])*log(1 + 6*r_cc - 13*r_cc^2 + 6*r_cc^3) + (x[,"n_14"] + x[,"n_22"])*log(3 + 20*r_cc - 20*r_cc^2 + 12*r_cc^3)
logL_cr <- function(r,n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n01,n03,n05,n11,n15,n21,n25,n31,n33,n35) {
L <- (-n02 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n34)*log(2) + 2*(n13 + n23)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*log(5) + (n01 + n35)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n31)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n15 + n21)*(log(pmax(1e-6,r)) + log(2 - 4*r + 3*r^2)) + (n04 + n32)*(log(pmax(1e-6,r)) + log(4 - 5*r + 4*r^2)) + (n14 + n22)*log(6 - 8*r + 17*r^2 - 12*r^3) + (n02 + n34)*log(3 - 6*r + 7*r^2 - 4*r^3) + (n11 + n25)*log(1 - 3*r + 5*r^2 - 3*r^3) + (n12 + n24)*log(3 + 10*r - 19*r^2 + 12*r^3)
return(L)}
interlogL_cr <- function(n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n01,n03,n05,n11,n15,n21,n25,n31,n33,n35) {
optimize(logL_cr,c(0,0.5), n02,n04,n12,n13,n14,n22,n23,n24,n32,n34,n01,n03,n05,n11,n15,n21,n25,n31,n33,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_02"],x[,"n_04"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_34"],x[,"n_01"],x[,"n_03"],x[,"n_05"],x[,"n_11"],x[,"n_15"],x[,"n_21"],x[,"n_25"],x[,"n_31"],x[,"n_33"],x[,"n_35"])
LOD_cr <- 3*(x[,"n_05"] + x[,"n_31"])*log10(2) + 3*(x[,"n_01"] + x[,"n_35"])*log10(2) + (x[,"n_11"] + x[,"n_25"])*(3*log10(2) - log10(3)) - (x[,"n_15"] + x[,"n_21"])*(-3*log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + log10(19)) + (x[,"n_01"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr))) + (x[,"n_05"] + x[,"n_31"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_15"] + x[,"n_21"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_04"] + x[,"n_32"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_14"] + x[,"n_22"])*log10(6 - 8*r_cr + 17*r_cr^2 - 12*r_cr^3) + (x[,"n_02"] + x[,"n_34"])*log10(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_11"] + x[,"n_25"])*log10(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_12"] + x[,"n_24"])*log10(3 + 10*r_cr - 19*r_cr^2 + 12*r_cr^3)
logL_cr <- (-x[,"n_02"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_34"])*log(2) + 2*(x[,"n_13"] + x[,"n_23"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_01"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr))) + (x[,"n_05"] + x[,"n_31"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_15"] + x[,"n_21"])*(log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 3*r_cr^2)) + (x[,"n_04"] + x[,"n_32"])*(log(pmax(1e-6,r_cr)) + log(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_14"] + x[,"n_22"])*log(6 - 8*r_cr + 17*r_cr^2 - 12*r_cr^3) + (x[,"n_02"] + x[,"n_34"])*log(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3) + (x[,"n_11"] + x[,"n_25"])*log(1 - 3*r_cr + 5*r_cr^2 - 3*r_cr^3) + (x[,"n_12"] + x[,"n_24"])*log(3 + 10*r_cr - 19*r_cr^2 + 12*r_cr^3)
logL_mc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*(4*log(2) + 2*log(5)) + (n05 + n31)*(log(2 - r) + log(pmax(1e-6,r)) + log(1 + 2*r)) + (n13 + n23)*log(69 + 4*r - 4*r^2) + (n01 + n35)*(log(3 - 2*r) + log(1 - r^2)) + (n03 + n33)*log(19 - 4*r + 4*r^2) + (n14 + n22)*log(33 + 4*r + 20*r^2 - 12*r^3) + (n11 + n25)*log(10 - 10*r + 9*r^2 - 6*r^3) + (n02 + n34)*log(15 - 12*r + 4*r^2 - 4*r^3) + (n04 + n32)*log(3 + 16*r - 8*r^2 + 4*r^3) + (n15 + n21)*log(3 + 10*r - 9*r^2 + 6*r^3) + (n12 + n24)*log(45 - 8*r - 16*r^2 + 12*r^3)
return(L)}
interlogL_mc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_mc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_mc <- -((x[,"n_05"] + x[,"n_31"])*(-log10(2) + log10(3))) - (x[,"n_01"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_33"])*(log10(2) + 2*log10(3)) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(5) + log10(7)) - (x[,"n_14"] + x[,"n_22"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_12"] + x[,"n_24"])*(-log10(2) + log10(7) + log10(11)) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + log10(13)) - (x[,"n_04"] + x[,"n_32"])*(-log10(2) + log10(19)) - (x[,"n_02"] + x[,"n_34"])*(-log10(2) + log10(19)) + (x[,"n_05"] + x[,"n_31"])*(log10(2 - r_mc) + log10(pmax(1e-6,r_mc)) + log10(1 + 2*r_mc)) + (x[,"n_13"] + x[,"n_23"])*log10(69 + 4*r_mc - 4*r_mc^2) + (x[,"n_01"] + x[,"n_35"])*(log10(3 - 2*r_mc) + log10(1 - r_mc^2)) + (x[,"n_03"] + x[,"n_33"])*log10(19 - 4*r_mc + 4*r_mc^2) + (x[,"n_14"] + x[,"n_22"])*log10(33 + 4*r_mc + 20*r_mc^2 - 12*r_mc^3) + (x[,"n_11"] + x[,"n_25"])*log10(10 - 10*r_mc + 9*r_mc^2 - 6*r_mc^3) + (x[,"n_02"] + x[,"n_34"])*log10(15 - 12*r_mc + 4*r_mc^2 - 4*r_mc^3) + (x[,"n_04"] + x[,"n_32"])*log10(3 + 16*r_mc - 8*r_mc^2 + 4*r_mc^3) + (x[,"n_15"] + x[,"n_21"])*log10(3 + 10*r_mc - 9*r_mc^2 + 6*r_mc^3) + (x[,"n_12"] + x[,"n_24"])*log10(45 - 8*r_mc - 16*r_mc^2 + 12*r_mc^3)
logL_mc <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(4*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_31"])*(log(2 - r_mc) + log(pmax(1e-6,r_mc)) + log(1 + 2*r_mc)) + (x[,"n_13"] + x[,"n_23"])*log(69 + 4*r_mc - 4*r_mc^2) + (x[,"n_01"] + x[,"n_35"])*(log(3 - 2*r_mc) + log(1 - r_mc^2)) + (x[,"n_03"] + x[,"n_33"])*log(19 - 4*r_mc + 4*r_mc^2) + (x[,"n_14"] + x[,"n_22"])*log(33 + 4*r_mc + 20*r_mc^2 - 12*r_mc^3) + (x[,"n_11"] + x[,"n_25"])*log(10 - 10*r_mc + 9*r_mc^2 - 6*r_mc^3) + (x[,"n_02"] + x[,"n_34"])*log(15 - 12*r_mc + 4*r_mc^2 - 4*r_mc^3) + (x[,"n_04"] + x[,"n_32"])*log(3 + 16*r_mc - 8*r_mc^2 + 4*r_mc^3) + (x[,"n_15"] + x[,"n_21"])*log(3 + 10*r_mc - 9*r_mc^2 + 6*r_mc^3) + (x[,"n_12"] + x[,"n_24"])*log(45 - 8*r_mc - 16*r_mc^2 + 12*r_mc^3)
logL_mr <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-2*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 2*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 2*n31 - 3*n32 - 3*n33 - 3*n34 - 2*n35)*log(2) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*log(5) + (n05 + n31)*(log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n13 + n23)*log(33 + 8*r - 8*r^2) + (n01 + n35)*(log(pmax(1e-6,r)) + log(1 - r^2)) + (n03 + n33)*log(11 - 8*r + 8*r^2) + (n12 + n24)*log(18 - 4*r + 19*r^2 - 12*r^3) + (n04 + n32)*log(6 - 4*r + 5*r^2 - 4*r^3) + (n15 + n21)*log(3 - 5*r + 6*r^2 - 3*r^3) + (n11 + n25)*log(1 + 2*r - 3*r^2 + 3*r^3) + (n02 + n34)*log(3 + 6*r - 7*r^2 + 4*r^3) + (n14 + n22)*log(21 + 2*r - 17*r^2 + 12*r^3)
return(L)}
interlogL_mr <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_mr,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mr <- mapply(interlogL_mr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_mr <- -2*(x[,"n_03"] + x[,"n_33"])*log10(3) - (x[,"n_05"] + x[,"n_31"])*(-3*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_35"])*(-3*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_23"])*(log10(5) + log10(7)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_15"] + x[,"n_21"])*(-3*log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_25"])*(-3*log10(2) + log10(13)) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(19)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(19)) + (x[,"n_05"] + x[,"n_31"])*(log10(pmax(1e-6,1 - r_mr)) + log10(2 - r_mr) + log10(pmax(1e-6,r_mr))) + (x[,"n_13"] + x[,"n_23"])*log10(33 + 8*r_mr - 8*r_mr^2) + (x[,"n_01"] + x[,"n_35"])*(log10(pmax(1e-6,r_mr)) + log10(1 - r_mr^2)) + (x[,"n_03"] + x[,"n_33"])*log10(11 - 8*r_mr + 8*r_mr^2) + (x[,"n_12"] + x[,"n_24"])*log10(18 - 4*r_mr + 19*r_mr^2 - 12*r_mr^3) + (x[,"n_04"] + x[,"n_32"])*log10(6 - 4*r_mr + 5*r_mr^2 - 4*r_mr^3) + (x[,"n_15"] + x[,"n_21"])*log10(3 - 5*r_mr + 6*r_mr^2 - 3*r_mr^3) + (x[,"n_11"] + x[,"n_25"])*log10(1 + 2*r_mr - 3*r_mr^2 + 3*r_mr^3) + (x[,"n_02"] + x[,"n_34"])*log10(3 + 6*r_mr - 7*r_mr^2 + 4*r_mr^3) + (x[,"n_14"] + x[,"n_22"])*log10(21 + 2*r_mr - 17*r_mr^2 + 12*r_mr^3)
logL_mr <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 2*x[,"n_35"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(5) + (x[,"n_05"] + x[,"n_31"])*(log(pmax(1e-6,1 - r_mr)) + log(2 - r_mr) + log(pmax(1e-6,r_mr))) + (x[,"n_13"] + x[,"n_23"])*log(33 + 8*r_mr - 8*r_mr^2) + (x[,"n_01"] + x[,"n_35"])*(log(pmax(1e-6,r_mr)) + log(1 - r_mr^2)) + (x[,"n_03"] + x[,"n_33"])*log(11 - 8*r_mr + 8*r_mr^2) + (x[,"n_12"] + x[,"n_24"])*log(18 - 4*r_mr + 19*r_mr^2 - 12*r_mr^3) + (x[,"n_04"] + x[,"n_32"])*log(6 - 4*r_mr + 5*r_mr^2 - 4*r_mr^3) + (x[,"n_15"] + x[,"n_21"])*log(3 - 5*r_mr + 6*r_mr^2 - 3*r_mr^3) + (x[,"n_11"] + x[,"n_25"])*log(1 + 2*r_mr - 3*r_mr^2 + 3*r_mr^3) + (x[,"n_02"] + x[,"n_34"])*log(3 + 6*r_mr - 7*r_mr^2 + 4*r_mr^3) + (x[,"n_14"] + x[,"n_22"])*log(21 + 2*r_mr - 17*r_mr^2 + 12*r_mr^3)
logL_rc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*(3*log(2) + log(3) + 2*log(5)) + (n01 + n35)*(log(3 - 2*r) + log(pmax(1e-6,r)) + log(2 + r)) + (n13 + n23)*log(105 + 4*r - 4*r^2) + (n05 + n31)*(log(1 + 2*r) + log(3 - 4*r + r^2)) + (n03 + n33)*log(27 - 4*r + 4*r^2) + (n12 + n24)*log(57 - 4*r + 16*r^2 - 12*r^3) + (n15 + n21)*log(12 - 10*r + 13*r^2 - 6*r^3) + (n04 + n32)*log(21 - 16*r + 8*r^2 - 4*r^3) + (n02 + n34)*log(9 + 12*r - 4*r^2 + 4*r^3) + (n11 + n25)*log(9 + 2*r - 5*r^2 + 6*r^3) + (n14 + n22)*log(57 + 8*r - 20*r^2 + 12*r^3)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rc <- -((x[,"n_05"] + x[,"n_31"])*(-log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_35"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_33"])*(log10(2) + log10(13)) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + log10(19)) - (x[,"n_14"] + x[,"n_22"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_12"] + x[,"n_24"])*(-log10(2) + log10(5) + log10(23)) - (x[,"n_04"] + x[,"n_32"])*(-log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_34"])*(-log10(2) + log10(29)) - (x[,"n_13"] + x[,"n_23"])*(log10(2) + log10(53)) + (x[,"n_01"] + x[,"n_35"])*(log10(3 - 2*r_rc) + log10(pmax(1e-6,r_rc)) + log10(2 + r_rc)) + (x[,"n_13"] + x[,"n_23"])*log10(105 + 4*r_rc - 4*r_rc^2) + (x[,"n_05"] + x[,"n_31"])*(log10(1 + 2*r_rc) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_03"] + x[,"n_33"])*log10(27 - 4*r_rc + 4*r_rc^2) + (x[,"n_12"] + x[,"n_24"])*log10(57 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_15"] + x[,"n_21"])*log10(12 - 10*r_rc + 13*r_rc^2 - 6*r_rc^3) + (x[,"n_04"] + x[,"n_32"])*log10(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_02"] + x[,"n_34"])*log10(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_11"] + x[,"n_25"])*log10(9 + 2*r_rc - 5*r_rc^2 + 6*r_rc^3) + (x[,"n_14"] + x[,"n_22"])*log10(57 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rc <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(3*log(2) + log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_35"])*(log(3 - 2*r_rc) + log(pmax(1e-6,r_rc)) + log(2 + r_rc)) + (x[,"n_13"] + x[,"n_23"])*log(105 + 4*r_rc - 4*r_rc^2) + (x[,"n_05"] + x[,"n_31"])*(log(1 + 2*r_rc) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_03"] + x[,"n_33"])*log(27 - 4*r_rc + 4*r_rc^2) + (x[,"n_12"] + x[,"n_24"])*log(57 - 4*r_rc + 16*r_rc^2 - 12*r_rc^3) + (x[,"n_15"] + x[,"n_21"])*log(12 - 10*r_rc + 13*r_rc^2 - 6*r_rc^3) + (x[,"n_04"] + x[,"n_32"])*log(21 - 16*r_rc + 8*r_rc^2 - 4*r_rc^3) + (x[,"n_02"] + x[,"n_34"])*log(9 + 12*r_rc - 4*r_rc^2 + 4*r_rc^3) + (x[,"n_11"] + x[,"n_25"])*log(9 + 2*r_rc - 5*r_rc^2 + 6*r_rc^3) + (x[,"n_14"] + x[,"n_22"])*log(57 + 8*r_rc - 20*r_rc^2 + 12*r_rc^3)
logL_rr <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-n01 - 2*n02 - 2*n03 - 2*n04 - n05 - n11 - 2*n12 - 2*n13 - 2*n14 - n15 - n21 - 2*n22 - 2*n23 - 2*n24 - n25 - n31 - 2*n32 - 2*n33 - 2*n34 - n35)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*(log(3) + 2*log(5)) + (n01 + n35)*(2*log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n33)*log(9 + 16*r - 16*r^2) + (n02 + n34)*(log(pmax(1e-6,r)) + log(12 + 7*r - 4*r^2)) + (n11 + n25)*(log(pmax(1e-6,r)) + log(5 + r - 3*r^2)) + (n05 + n31)*(log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n13 + n23)*log(57 - 16*r + 16*r^2) + (n14 + n22)*log(42 - 32*r + 17*r^2 - 12*r^3) + (n15 + n21)*log(3 + 2*r - 8*r^2 + 3*r^3) + (n04 + n32)*log(15 - 14*r - 5*r^2 + 4*r^3) + (n12 + n24)*log(15 + 34*r - 19*r^2 + 12*r^3)
return(L)}
interlogL_rr <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_rr,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_rr <- -((x[,"n_05"] + x[,"n_31"])*(-3*log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_35"])*(-3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_33"])*log10(13) - (x[,"n_15"] + x[,"n_21"])*(-3*log10(2) + log10(19)) - (x[,"n_11"] + x[,"n_25"])*(-3*log10(2) + log10(19)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(23)) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(29)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(29)) - (x[,"n_13"] + x[,"n_23"])*log10(53) + (x[,"n_01"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_03"] + x[,"n_33"])*log10(9 + 16*r_rr - 16*r_rr^2) + (x[,"n_02"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(12 + 7*r_rr - 4*r_rr^2)) + (x[,"n_11"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(5 + r_rr - 3*r_rr^2)) + (x[,"n_05"] + x[,"n_31"])*(log10(pmax(1e-6,1 - r_rr)) + log10(3 - 4*r_rr + r_rr^2)) + (x[,"n_13"] + x[,"n_23"])*log10(57 - 16*r_rr + 16*r_rr^2) + (x[,"n_14"] + x[,"n_22"])*log10(42 - 32*r_rr + 17*r_rr^2 - 12*r_rr^3) + (x[,"n_15"] + x[,"n_21"])*log10(3 + 2*r_rr - 8*r_rr^2 + 3*r_rr^3) + (x[,"n_04"] + x[,"n_32"])*log10(15 - 14*r_rr - 5*r_rr^2 + 4*r_rr^3) + (x[,"n_12"] + x[,"n_24"])*log10(15 + 34*r_rr - 19*r_rr^2 + 12*r_rr^3)
logL_rr <- (-x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_35"])*(2*log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_03"] + x[,"n_33"])*log(9 + 16*r_rr - 16*r_rr^2) + (x[,"n_02"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(12 + 7*r_rr - 4*r_rr^2)) + (x[,"n_11"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(5 + r_rr - 3*r_rr^2)) + (x[,"n_05"] + x[,"n_31"])*(log(pmax(1e-6,1 - r_rr)) + log(3 - 4*r_rr + r_rr^2)) + (x[,"n_13"] + x[,"n_23"])*log(57 - 16*r_rr + 16*r_rr^2) + (x[,"n_14"] + x[,"n_22"])*log(42 - 32*r_rr + 17*r_rr^2 - 12*r_rr^3) + (x[,"n_15"] + x[,"n_21"])*log(3 + 2*r_rr - 8*r_rr^2 + 3*r_rr^3) + (x[,"n_04"] + x[,"n_32"])*log(15 - 14*r_rr - 5*r_rr^2 + 4*r_rr^3) + (x[,"n_12"] + x[,"n_24"])*log(15 + 34*r_rr - 19*r_rr^2 + 12*r_rr^3)

return(list(
r_mat = cbind(r_cc,r_cr,r_mc,r_mr,r_rc,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cr,LOD_mc,LOD_mr,LOD_rc,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cr,logL_mc,logL_mr,logL_rc,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling repulsion","mixed coupling","mixed repulsion","repulsion coupling","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.1_3.0 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(3*log(2) + log(5)) + 2*(n00 + n33)*log(pmax(1e-6,1 - r)) + 2*(n03 + n30)*log(pmax(1e-6,r)) + (n13 + n20)*(log(2 - r) + log(pmax(1e-6,r))) + (n02 + n31)*(log(4 - r) + log(pmax(1e-6,r))) + (n10 + n23)*log(1 - r^2) + (n01 + n32)*log(3 - 2*r - r^2) + (n11 + n22)*log(9 - 4*r + r^2) + (n12 + n21)*log(6 + 2*r + r^2)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_c <- 2*(x[,"n_03"] + x[,"n_30"])*log10(2) + 2*(x[,"n_00"] + x[,"n_33"])*log10(2) + (x[,"n_10"] + x[,"n_23"])*(2*log10(2) - log10(3)) - (x[,"n_13"] + x[,"n_20"])*(-2*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_31"])*(-2*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_32"])*(-2*log10(2) + log10(7)) - (x[,"n_12"] + x[,"n_21"])*(-2*log10(2) + log10(29)) - (x[,"n_11"] + x[,"n_22"])*(-2*log10(2) + log10(29)) + 2*(x[,"n_00"] + x[,"n_33"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_03"] + x[,"n_30"])*log10(pmax(1e-6,r_c)) + (x[,"n_13"] + x[,"n_20"])*(log10(2 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_31"])*(log10(4 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_10"] + x[,"n_23"])*log10(1 - r_c^2) + (x[,"n_01"] + x[,"n_32"])*log10(3 - 2*r_c - r_c^2) + (x[,"n_11"] + x[,"n_22"])*log10(9 - 4*r_c + r_c^2) + (x[,"n_12"] + x[,"n_21"])*log10(6 + 2*r_c + r_c^2)
logL_c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(3*log(2) + log(5)) + 2*(x[,"n_00"] + x[,"n_33"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_03"] + x[,"n_30"])*log(pmax(1e-6,r_c)) + (x[,"n_13"] + x[,"n_20"])*(log(2 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_31"])*(log(4 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_10"] + x[,"n_23"])*log(1 - r_c^2) + (x[,"n_01"] + x[,"n_32"])*log(3 - 2*r_c - r_c^2) + (x[,"n_11"] + x[,"n_22"])*log(9 - 4*r_c + r_c^2) + (x[,"n_12"] + x[,"n_21"])*log(6 + 2*r_c + r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 2*n30 - 2*n31 - 2*n32 - 2*n33)*log(2) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(log(3) + log(5)) + (n00 + n03 + n30 + n33)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n11 + n12 + n21 + n22)*log(21 + 2*r - 2*r^2) + (n01 + n02 + n31 + n32)*log(3 - r + r^2) + (n10 + n13 + n20 + n23)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_m <- 2*(x[,"n_00"] + x[,"n_03"] + x[,"n_30"] + x[,"n_33"])*log10(2) - (x[,"n_10"] + x[,"n_13"] + x[,"n_20"] + x[,"n_23"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_02"] + x[,"n_31"] + x[,"n_32"])*(-2*log10(2) + log10(11)) - (x[,"n_11"] + x[,"n_12"] + x[,"n_21"] + x[,"n_22"])*(-log10(2) + log10(43)) + (x[,"n_00"] + x[,"n_03"] + x[,"n_30"] + x[,"n_33"])*(log10(pmax(1e-6,1 - r_m)) + log10(pmax(1e-6,r_m))) + (x[,"n_11"] + x[,"n_12"] + x[,"n_21"] + x[,"n_22"])*log10(21 + 2*r_m - 2*r_m^2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_31"] + x[,"n_32"])*log10(3 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_13"] + x[,"n_20"] + x[,"n_23"])*log10(3 - 2*r_m + 2*r_m^2)
logL_m <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(log(3) + log(5)) + (x[,"n_00"] + x[,"n_03"] + x[,"n_30"] + x[,"n_33"])*(log(pmax(1e-6,1 - r_m)) + log(pmax(1e-6,r_m))) + (x[,"n_11"] + x[,"n_12"] + x[,"n_21"] + x[,"n_22"])*log(21 + 2*r_m - 2*r_m^2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_31"] + x[,"n_32"])*log(3 - r_m + r_m^2) + (x[,"n_10"] + x[,"n_13"] + x[,"n_20"] + x[,"n_23"])*log(3 - 2*r_m + 2*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(3*log(2) + log(5)) + 2*(n03 + n30)*log(pmax(1e-6,1 - r)) + 2*(n00 + n33)*log(pmax(1e-6,r)) + (n10 + n23)*(log(2 - r) + log(pmax(1e-6,r))) + (n01 + n32)*(log(4 - r) + log(pmax(1e-6,r))) + (n13 + n20)*log(1 - r^2) + (n02 + n31)*log(3 - 2*r - r^2) + (n12 + n21)*log(9 - 4*r + r^2) + (n11 + n22)*log(6 + 2*r + r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_r <- 2*(x[,"n_03"] + x[,"n_30"])*log10(2) + 2*(x[,"n_00"] + x[,"n_33"])*log10(2) + (x[,"n_13"] + x[,"n_20"])*(2*log10(2) - log10(3)) - (x[,"n_10"] + x[,"n_23"])*(-2*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_31"])*(-2*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_32"])*(-2*log10(2) + log10(7)) - (x[,"n_12"] + x[,"n_21"])*(-2*log10(2) + log10(29)) - (x[,"n_11"] + x[,"n_22"])*(-2*log10(2) + log10(29)) + 2*(x[,"n_03"] + x[,"n_30"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_33"])*log10(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_23"])*(log10(2 - r_r) + log10(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_32"])*(log10(4 - r_r) + log10(pmax(1e-6,r_r))) + (x[,"n_13"] + x[,"n_20"])*log10(1 - r_r^2) + (x[,"n_02"] + x[,"n_31"])*log10(3 - 2*r_r - r_r^2) + (x[,"n_12"] + x[,"n_21"])*log10(9 - 4*r_r + r_r^2) + (x[,"n_11"] + x[,"n_22"])*log10(6 + 2*r_r + r_r^2)
logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(3*log(2) + log(5)) + 2*(x[,"n_03"] + x[,"n_30"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_33"])*log(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_23"])*(log(2 - r_r) + log(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_32"])*(log(4 - r_r) + log(pmax(1e-6,r_r))) + (x[,"n_13"] + x[,"n_20"])*log(1 - r_r^2) + (x[,"n_02"] + x[,"n_31"])*log(3 - 2*r_r - r_r^2) + (x[,"n_12"] + x[,"n_21"])*log(9 - 4*r_r + r_r^2) + (x[,"n_11"] + x[,"n_22"])*log(6 + 2*r_r + r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.2_2.2 <- function(x){
logL_cc <- function(r,n01,n02,n03,n10,n11,n12,n13,n14,n20,n21,n23,n24,n30,n31,n32,n33,n34,n41,n42,n43,n00,n04,n22,n40,n44) {
L <- (2*n01 + n02 + 2*n03 + 2*n10 + n11 + 3*n12 + n13 + 2*n14 + n20 + 3*n21 + 3*n23 + n24 + 2*n30 + n31 + 3*n32 + n33 + 2*n34 + 2*n41 + n42 + 2*n43)*log(2) + (n02 + n20 + n24 + n42)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + 4*(n00 + n44)*log(pmax(1e-6,1 - r)) + 4*(n04 + n40)*log(pmax(1e-6,r)) + (n01 + n10 + n34 + n43)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n20 + n24 + n42)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n03 + n14 + n30 + n41)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n13 + n31)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n11 + n33)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n12 + n21 + n23 + n32)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + n22*log(11 - 32*r + 68*r^2 - 72*r^3 + 36*r^4)
return(L)}
interlogL_cc <- function(n01,n02,n03,n10,n11,n12,n13,n14,n20,n21,n23,n24,n30,n31,n32,n33,n34,n41,n42,n43,n00,n04,n22,n40,n44) {
optimize(logL_cc,c(0,0.5), n01,n02,n03,n10,n11,n12,n13,n14,n20,n21,n23,n24,n30,n31,n32,n33,n34,n41,n42,n43,n00,n04,n22,n40,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_00"],x[,"n_04"],x[,"n_22"],x[,"n_40"],x[,"n_44"])
LOD_cc <- 4*(x[,"n_04"] + x[,"n_40"])*log10(2) + 4*(x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*log10(2) + 4*(x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(2) + 4*(x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log10(2) + 4*(x[,"n_00"] + x[,"n_44"])*log10(2) - (x[,"n_13"] + x[,"n_31"])*(-2*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_33"])*(-2*log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + log10(5)) - x[,"n_22"]*(-2*log10(2) + log10(3) + log10(7)) + 4*(x[,"n_00"] + x[,"n_44"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_04"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(7 - 12*r_cc + 8*r_cc^2)) + (x[,"n_11"] + x[,"n_33"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 - 4*r_cc + 8*r_cc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(2 - 5*r_cc + 6*r_cc^2 - 3*r_cc^3)) + x[,"n_22"]*log10(11 - 32*r_cc + 68*r_cc^2 - 72*r_cc^3 + 36*r_cc^4)
logL_cc <- (2*x[,"n_01"] + x[,"n_02"] + 2*x[,"n_03"] + 2*x[,"n_10"] + x[,"n_11"] + 3*x[,"n_12"] + x[,"n_13"] + 2*x[,"n_14"] + x[,"n_20"] + 3*x[,"n_21"] + 3*x[,"n_23"] + x[,"n_24"] + 2*x[,"n_30"] + x[,"n_31"] + 3*x[,"n_32"] + x[,"n_33"] + 2*x[,"n_34"] + 2*x[,"n_41"] + x[,"n_42"] + 2*x[,"n_43"])*log(2) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + 4*(x[,"n_00"] + x[,"n_44"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_04"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(7 - 12*r_cc + 8*r_cc^2)) + (x[,"n_11"] + x[,"n_33"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 - 4*r_cc + 8*r_cc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(2 - 5*r_cc + 6*r_cc^2 - 3*r_cc^3)) + x[,"n_22"]*log(11 - 32*r_cc + 68*r_cc^2 - 72*r_cc^3 + 36*r_cc^4)
logL_cm <- function(r,n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n24,n30,n31,n33,n34,n40,n41,n42,n43,n44,n12,n21,n23,n32,n22) {
L <- (-2*n00 - 2*n01 - n02 - 2*n03 - 2*n04 - 2*n10 - 2*n11 - 2*n13 - 2*n14 - n20 - n24 - 2*n30 - 2*n31 - 2*n33 - 2*n34 - 2*n40 - 2*n41 - n42 - 2*n43 - 2*n44)*log(2) + (n12 + n21 + n23 + n32)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n04 + n40)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n00 + n44)*(2*log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n01 + n10 + n34 + n43)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r^2)) + (n03 + n14 + n30 + n41)*(2*log(pmax(1e-6,r)) + log(7 - 8*r + 4*r^2)) + (n13 + n31)*(log(pmax(1e-6,r)) + log(18 - 25*r + 32*r^2 - 16*r^3)) + (n02 + n20 + n24 + n42)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n11 + n33)*log(9 - 25*r^2 + 32*r^3 - 16*r^4) + n22*log(5 + 2*r - 11*r^2 + 18*r^3 - 9*r^4) + (n12 + n21 + n23 + n32)*log(1 - r + 3*r^2 - 4*r^3 + 2*r^4)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n24,n30,n31,n33,n34,n40,n41,n42,n43,n44,n12,n21,n23,n32,n22) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n13,n14,n20,n24,n30,n31,n33,n34,n40,n41,n42,n43,n44,n12,n21,n23,n32,n22, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_12"],x[,"n_21"],x[,"n_23"],x[,"n_32"],x[,"n_22"])
LOD_cm <- -((x[,"n_04"] + x[,"n_40"])*(-4*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_44"])*(-4*log10(2) + log10(3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(3*log10(2) - log10(7)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-4*log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_31"])*(-2*log10(2) + log10(23)) - (x[,"n_11"] + x[,"n_33"])*(-2*log10(2) + log10(23)) - x[,"n_22"]*(-4*log10(2) + log10(79)) + (x[,"n_04"] + x[,"n_40"])*(log10(2 - r_cm) + 3*log10(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(1 - r_cm^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(3 + 4*r_cm^2)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm)) + log10(7 - 8*r_cm + 4*r_cm^2)) + (x[,"n_13"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(18 - 25*r_cm + 32*r_cm^2 - 16*r_cm^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm)) + log10(4 - 7*r_cm + 6*r_cm^2 - 3*r_cm^3)) + (x[,"n_11"] + x[,"n_33"])*log10(9 - 25*r_cm^2 + 32*r_cm^3 - 16*r_cm^4) + x[,"n_22"]*log10(5 + 2*r_cm - 11*r_cm^2 + 18*r_cm^3 - 9*r_cm^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(1 - r_cm + 3*r_cm^2 - 4*r_cm^3 + 2*r_cm^4)
logL_cm <- (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_20"] - x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 2*x[,"n_41"] - x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"])*log(2) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(log(2 - r_cm) + 3*log(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cm)) + log(1 - r_cm^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_cm)) + log(3 + 4*r_cm^2)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm)) + log(7 - 8*r_cm + 4*r_cm^2)) + (x[,"n_13"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(18 - 25*r_cm + 32*r_cm^2 - 16*r_cm^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(log(pmax(1e-6,r_cm)) + log(4 - 7*r_cm + 6*r_cm^2 - 3*r_cm^3)) + (x[,"n_11"] + x[,"n_33"])*log(9 - 25*r_cm^2 + 32*r_cm^3 - 16*r_cm^4) + x[,"n_22"]*log(5 + 2*r_cm - 11*r_cm^2 + 18*r_cm^3 - 9*r_cm^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(1 - r_cm + 3*r_cm^2 - 4*r_cm^3 + 2*r_cm^4)
logL_cr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-n00 - n01 - n02 - n03 - n04 - n10 - n11 + n12 - n13 - n14 - n20 + n21 + n22 + n23 - n24 - n30 - n31 + n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*(log(3) + 2*log(5)) + (n00 + n44)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n04 + n40)*(2*log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n03 + n14 + n30 + n41)*(log(pmax(1e-6,r)) + log(6 - 11*r + 12*r^2 - 4*r^3)) + (n12 + n21 + n23 + n32)*log(3 + 5*r - 11*r^2 + 12*r^3 - 6*r^4) + (n01 + n10 + n34 + n43)*log(3 - 4*r + r^2 + 4*r^3 - 4*r^4) + (n02 + n20 + n24 + n42)*log(3 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + n22*log(9 - 8*r + 17*r^2 - 18*r^3 + 9*r^4) + (n13 + n31)*log(9 - 12*r + 39*r^2 - 40*r^3 + 16*r^4) + (n11 + n33)*log(12 - 10*r + 15*r^2 - 24*r^3 + 16*r^4)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_cr <- -((x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-log10(2) + log10(3))) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_40"])*(-4*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_44"])*(-4*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_31"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_11"] + x[,"n_33"])*(-2*log10(2) + log10(5) + log10(7)) - x[,"n_22"]*(-4*log10(2) + 2*log10(11)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-3*log10(2) + log10(19)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(31)) + (x[,"n_00"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr)) + log10(2 + r_cr)) + (x[,"n_04"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_cr)) + log10(3 - 4*r_cr + r_cr^2)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log10(pmax(1e-6,r_cr)) + log10(6 - 11*r_cr + 12*r_cr^2 - 4*r_cr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(3 + 5*r_cr - 11*r_cr^2 + 12*r_cr^3 - 6*r_cr^4) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log10(3 - 4*r_cr + r_cr^2 + 4*r_cr^3 - 4*r_cr^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(3 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + x[,"n_22"]*log10(9 - 8*r_cr + 17*r_cr^2 - 18*r_cr^3 + 9*r_cr^4) + (x[,"n_13"] + x[,"n_31"])*log10(9 - 12*r_cr + 39*r_cr^2 - 40*r_cr^3 + 16*r_cr^4) + (x[,"n_11"] + x[,"n_33"])*log10(12 - 10*r_cr + 15*r_cr^2 - 24*r_cr^3 + 16*r_cr^4)
logL_cr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] + x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] + x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr)) + log(2 + r_cr)) + (x[,"n_04"] + x[,"n_40"])*(2*log(pmax(1e-6,r_cr)) + log(3 - 4*r_cr + r_cr^2)) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log(pmax(1e-6,r_cr)) + log(6 - 11*r_cr + 12*r_cr^2 - 4*r_cr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(3 + 5*r_cr - 11*r_cr^2 + 12*r_cr^3 - 6*r_cr^4) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log(3 - 4*r_cr + r_cr^2 + 4*r_cr^3 - 4*r_cr^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(3 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + x[,"n_22"]*log(9 - 8*r_cr + 17*r_cr^2 - 18*r_cr^3 + 9*r_cr^4) + (x[,"n_13"] + x[,"n_31"])*log(9 - 12*r_cr + 39*r_cr^2 - 40*r_cr^3 + 16*r_cr^4) + (x[,"n_11"] + x[,"n_33"])*log(12 - 10*r_cr + 15*r_cr^2 - 24*r_cr^3 + 16*r_cr^4)

logL_mm <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 4*n20 - 3*n21 - 3*n22 - 3*n23 - 4*n24 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n04 + n40)*(2*log(2 - r) + 2*log(pmax(1e-6,r))) + 2*(n00 + n44)*log(1 - r^2) + (n03 + n14 + n30 + n41)*(log(2 - r) + log(pmax(1e-6,r)) + log(3 - 2*r + 2*r^2)) + (n01 + n10 + n34 + n43)*(log(1 - r^2) + log(3 - 2*r + 2*r^2)) + (n12 + n21 + n23 + n32)*log(21 + 4*r - 16*r^2 + 24*r^3 - 12*r^4) + (n02 + n20 + n24 + n42)*log(9 - 8*r + 14*r^2 - 12*r^3 + 6*r^4) + (n13 + n31)*log(9 + 18*r^2 - 20*r^3 + 8*r^4) + (n11 + n33)*log(15 - 8*r + 6*r^2 - 12*r^3 + 8*r^4) + n22*log(37 + 18*r^2 - 36*r^3 + 18*r^4)
return(L)}
interlogL_mm <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_mm,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm <- mapply(interlogL_mm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_mm <- 2*(x[,"n_00"] + x[,"n_44"])*(2*log10(2) - log10(3)) - (x[,"n_04"] + x[,"n_40"])*(-2*log10(2) + 2*(-log10(2) + log10(3))) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(17)) - (x[,"n_13"] + x[,"n_31"])*(-log10(2) + log10(23)) - (x[,"n_11"] + x[,"n_33"])*(-log10(2) + log10(23)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-3*log10(2) + log10(59)) - x[,"n_22"]*(-3*log10(2) + log10(5) + log10(61)) + (x[,"n_04"] + x[,"n_40"])*(2*log10(2 - r_mm) + 2*log10(pmax(1e-6,r_mm))) + 2*(x[,"n_00"] + x[,"n_44"])*log10(1 - r_mm^2) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log10(2 - r_mm) + log10(pmax(1e-6,r_mm)) + log10(3 - 2*r_mm + 2*r_mm^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(log10(1 - r_mm^2) + log10(3 - 2*r_mm + 2*r_mm^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(21 + 4*r_mm - 16*r_mm^2 + 24*r_mm^3 - 12*r_mm^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(9 - 8*r_mm + 14*r_mm^2 - 12*r_mm^3 + 6*r_mm^4) + (x[,"n_13"] + x[,"n_31"])*log10(9 + 18*r_mm^2 - 20*r_mm^3 + 8*r_mm^4) + (x[,"n_11"] + x[,"n_33"])*log10(15 - 8*r_mm + 6*r_mm^2 - 12*r_mm^3 + 8*r_mm^4) + x[,"n_22"]*log10(37 + 18*r_mm^2 - 36*r_mm^3 + 18*r_mm^4)
logL_mm <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_04"] + x[,"n_40"])*(2*log(2 - r_mm) + 2*log(pmax(1e-6,r_mm))) + 2*(x[,"n_00"] + x[,"n_44"])*log(1 - r_mm^2) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log(2 - r_mm) + log(pmax(1e-6,r_mm)) + log(3 - 2*r_mm + 2*r_mm^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(log(1 - r_mm^2) + log(3 - 2*r_mm + 2*r_mm^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(21 + 4*r_mm - 16*r_mm^2 + 24*r_mm^3 - 12*r_mm^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(9 - 8*r_mm + 14*r_mm^2 - 12*r_mm^3 + 6*r_mm^4) + (x[,"n_13"] + x[,"n_31"])*log(9 + 18*r_mm^2 - 20*r_mm^3 + 8*r_mm^4) + (x[,"n_11"] + x[,"n_33"])*log(15 - 8*r_mm + 6*r_mm^2 - 12*r_mm^3 + 8*r_mm^4) + x[,"n_22"]*log(37 + 18*r_mm^2 - 36*r_mm^3 + 18*r_mm^4)
logL_mr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-3*n00 - 3*n01 - 2*n02 - 3*n03 - 3*n04 - 3*n10 - 2*n11 - 2*n12 - 2*n13 - 3*n14 - 2*n20 - 2*n21 - n22 - 2*n23 - 2*n24 - 3*n30 - 2*n31 - 2*n32 - 2*n33 - 3*n34 - 3*n40 - 3*n41 - 2*n42 - 3*n43 - 3*n44)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*(log(3) + 2*log(5)) + (n00 + n44)*(log(pmax(1e-6,r)) + log(2 + r) + log(1 - r^2)) + (n04 + n40)*(log(2 - r) + log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + n22*log(27 + 10*r - 19*r^2 + 18*r^3 - 9*r^4) + (n11 + n33)*log(15 + 4*r - r^2 + 8*r^3 - 8*r^4) + (n13 + n31)*log(18 + 6*r - 25*r^2 + 24*r^3 - 8*r^4) + (n02 + n20 + n24 + n42)*log(6 - 2*r - r^2 + 6*r^3 - 3*r^4) + (n01 + n10 + n34 + n43)*log(3 + 8*r - 6*r^2 + 4*r^4) + (n03 + n14 + n30 + n41)*log(9 - 12*r + 18*r^2 - 16*r^3 + 4*r^4) + (n12 + n21 + n23 + n32)*log(33 - 8*r + 20*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_mr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_mr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mr <- mapply(interlogL_mr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_mr <- -((x[,"n_04"] + x[,"n_40"])*(-4*log10(2) + log10(3) + log10(5))) - (x[,"n_00"] + x[,"n_44"])*(-4*log10(2) + log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-4*log10(2) + log10(5) + log10(17)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-2*log10(2) + log10(23)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + log10(23)) - (x[,"n_13"] + x[,"n_31"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_11"] + x[,"n_33"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + log10(127)) - x[,"n_22"]*(-4*log10(2) + log10(463)) + (x[,"n_00"] + x[,"n_44"])*(log10(pmax(1e-6,r_mr)) + log10(2 + r_mr) + log10(1 - r_mr^2)) + (x[,"n_04"] + x[,"n_40"])*(log10(2 - r_mr) + log10(pmax(1e-6,r_mr)) + log10(3 - 4*r_mr + r_mr^2)) + x[,"n_22"]*log10(27 + 10*r_mr - 19*r_mr^2 + 18*r_mr^3 - 9*r_mr^4) + (x[,"n_11"] + x[,"n_33"])*log10(15 + 4*r_mr - r_mr^2 + 8*r_mr^3 - 8*r_mr^4) + (x[,"n_13"] + x[,"n_31"])*log10(18 + 6*r_mr - 25*r_mr^2 + 24*r_mr^3 - 8*r_mr^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(6 - 2*r_mr - r_mr^2 + 6*r_mr^3 - 3*r_mr^4) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log10(3 + 8*r_mr - 6*r_mr^2 + 4*r_mr^4) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*log10(9 - 12*r_mr + 18*r_mr^2 - 16*r_mr^3 + 4*r_mr^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(33 - 8*r_mr + 20*r_mr^2 - 24*r_mr^3 + 12*r_mr^4)
logL_mr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 2*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 3*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 2*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_44"])*(log(pmax(1e-6,r_mr)) + log(2 + r_mr) + log(1 - r_mr^2)) + (x[,"n_04"] + x[,"n_40"])*(log(2 - r_mr) + log(pmax(1e-6,r_mr)) + log(3 - 4*r_mr + r_mr^2)) + x[,"n_22"]*log(27 + 10*r_mr - 19*r_mr^2 + 18*r_mr^3 - 9*r_mr^4) + (x[,"n_11"] + x[,"n_33"])*log(15 + 4*r_mr - r_mr^2 + 8*r_mr^3 - 8*r_mr^4) + (x[,"n_13"] + x[,"n_31"])*log(18 + 6*r_mr - 25*r_mr^2 + 24*r_mr^3 - 8*r_mr^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(6 - 2*r_mr - r_mr^2 + 6*r_mr^3 - 3*r_mr^4) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*log(3 + 8*r_mr - 6*r_mr^2 + 4*r_mr^4) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*log(9 - 12*r_mr + 18*r_mr^2 - 16*r_mr^3 + 4*r_mr^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(33 - 8*r_mr + 20*r_mr^2 - 24*r_mr^3 + 12*r_mr^4)

logL_rr <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
L <- (-2*n00 - n01 - 2*n02 - n03 - 2*n04 - n10 - n11 - n12 - n13 - n14 - 2*n20 - n21 - n22 - n23 - 2*n24 - n30 - n31 - n32 - n33 - n34 - 2*n40 - n41 - 2*n42 - n43 - 2*n44)*log(2) + (-2*n00 - 2*n01 - n02 - 2*n03 - 2*n04 - 2*n10 - 2*n11 - n12 - 2*n13 - 2*n14 - n20 - n21 - n22 - n23 - n24 - 2*n30 - 2*n31 - n32 - 2*n33 - 2*n34 - 2*n40 - 2*n41 - n42 - 2*n43 - 2*n44)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34 - n40 - n41 - n42 - n43 - n44)*log(5) + (n00 + n44)*(2*log(pmax(1e-6,r)) + 2*log(2 + r)) + (n01 + n10 + n34 + n43)*(log(pmax(1e-6,r)) + log(2 + r) + log(3 + 2*r - 2*r^2)) + 2*(n04 + n40)*log(3 - 4*r + r^2) + (n03 + n14 + n30 + n41)*(log(3 + 2*r - 2*r^2) + log(3 - 4*r + r^2)) + (n12 + n21 + n23 + n32)*log(15 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n02 + n20 + n24 + n42)*log(3 + 8*r - 6*r^2 - 4*r^3 + 2*r^4) + n22*log(33 - 16*r + 22*r^2 - 12*r^3 + 6*r^4) + (n13 + n31)*log(45 - 48*r + 32*r^2 - 28*r^3 + 8*r^4) + (n11 + n33)*log(9 + 36*r - 4*r^2 - 4*r^3 + 8*r^4)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34,n40,n41,n42,n43,n44, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"])
LOD_rr <- -2*(x[,"n_04"] + x[,"n_40"])*(-2*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_44"])*(-2*log10(2) + 2*(-log10(2) + log10(5))) - (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*(-2*log10(2) + 2*log10(3) + log10(7)) - (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(-3*log10(2) + log10(5) + log10(7)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(-3*log10(2) + log10(5) + log10(7)) - (x[,"n_13"] + x[,"n_31"])*(log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_33"])*(log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*(-3*log10(2) + log10(41)) - x[,"n_22"]*(-3*log10(2) + log10(5) + log10(47)) + (x[,"n_00"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + 2*log10(2 + r_rr)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(2 + r_rr) + log10(3 + 2*r_rr - 2*r_rr^2)) + 2*(x[,"n_04"] + x[,"n_40"])*log10(3 - 4*r_rr + r_rr^2) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log10(3 + 2*r_rr - 2*r_rr^2) + log10(3 - 4*r_rr + r_rr^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log10(15 + 4*r_rr - 8*r_rr^2 + 8*r_rr^3 - 4*r_rr^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log10(3 + 8*r_rr - 6*r_rr^2 - 4*r_rr^3 + 2*r_rr^4) + x[,"n_22"]*log10(33 - 16*r_rr + 22*r_rr^2 - 12*r_rr^3 + 6*r_rr^4) + (x[,"n_13"] + x[,"n_31"])*log10(45 - 48*r_rr + 32*r_rr^2 - 28*r_rr^3 + 8*r_rr^4) + (x[,"n_11"] + x[,"n_33"])*log10(9 + 36*r_rr - 4*r_rr^2 - 4*r_rr^3 + 8*r_rr^4)
logL_rr <- (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - 2*x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - 2*x[,"n_40"] - x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"])*log(2) + (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_10"] - 2*x[,"n_11"] - x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 2*x[,"n_41"] - x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"])*log(5) + (x[,"n_00"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + 2*log(2 + r_rr)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_34"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(2 + r_rr) + log(3 + 2*r_rr - 2*r_rr^2)) + 2*(x[,"n_04"] + x[,"n_40"])*log(3 - 4*r_rr + r_rr^2) + (x[,"n_03"] + x[,"n_14"] + x[,"n_30"] + x[,"n_41"])*(log(3 + 2*r_rr - 2*r_rr^2) + log(3 - 4*r_rr + r_rr^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_23"] + x[,"n_32"])*log(15 + 4*r_rr - 8*r_rr^2 + 8*r_rr^3 - 4*r_rr^4) + (x[,"n_02"] + x[,"n_20"] + x[,"n_24"] + x[,"n_42"])*log(3 + 8*r_rr - 6*r_rr^2 - 4*r_rr^3 + 2*r_rr^4) + x[,"n_22"]*log(33 - 16*r_rr + 22*r_rr^2 - 12*r_rr^3 + 6*r_rr^4) + (x[,"n_13"] + x[,"n_31"])*log(45 - 48*r_rr + 32*r_rr^2 - 28*r_rr^3 + 8*r_rr^4) + (x[,"n_11"] + x[,"n_33"])*log(9 + 36*r_rr - 4*r_rr^2 - 4*r_rr^3 + 8*r_rr^4)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_mm,r_mr,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_mm,LOD_mr,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_mm,logL_mr,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","mixed mixed","mixed repulsion","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.2_2.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35) {
L <- (-2*n00 - 2*n01 - n02 - n03 - 2*n04 - 2*n05 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - n20 - n21 - n22 - n23 - n24 - n25 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n40 - 2*n41 - n42 - n43 - 2*n44 - 2*n45)*log(2) + (n01 + n04 + n11 + n14 + n20 + n21 + n24 + n25 + n31 + n34 + n41 + n44)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + 4*(n00 + n45)*log(pmax(1e-6,1 - r)) + 4*(n05 + n40)*log(pmax(1e-6,r)) + (n10 + n35)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n20 + n25)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n15 + n30)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n04 + n41)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n01 + n44)*(3*log(pmax(1e-6,1 - r)) + log(1 + r)) + (n02 + n43)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(5 + r)) + (n03 + n42)*(2*log(pmax(1e-6,r)) + log(6 - 7*r + r^2)) + (n11 + n34)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r^2)) + (n14 + n31)*(2*log(pmax(1e-6,r)) + log(7 - 8*r + 4*r^2)) + (n13 + n32)*(log(pmax(1e-6,r)) + log(32 - 45*r + 36*r^2 - 8*r^3)) + (n21 + n24)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n12 + n33)*log(15 - 18*r + 15*r^2 - 4*r^3 - 8*r^4) + (n22 + n23)*log(11 - 12*r + 18*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_10"],x[,"n_15"],x[,"n_30"],x[,"n_35"])
LOD_cc <- 4*(x[,"n_20"] + x[,"n_25"])*log10(2) + 4*(x[,"n_15"] + x[,"n_30"])*log10(2) + 4*(x[,"n_10"] + x[,"n_35"])*log10(2) + 4*(x[,"n_05"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_33"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_03"] + x[,"n_42"])*(-4*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_43"])*(-4*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-4*log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_23"])*(-3*log10(2) + log10(67)) + 4*(x[,"n_00"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_35"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_25"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_41"])*(log10(2 - r_cc) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_44"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(1 + r_cc)) + (x[,"n_02"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc)) + log10(5 + r_cc)) + (x[,"n_03"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(6 - 7*r_cc + r_cc^2)) + (x[,"n_11"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 4*r_cc^2)) + (x[,"n_14"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(7 - 8*r_cc + 4*r_cc^2)) + (x[,"n_13"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(32 - 45*r_cc + 36*r_cc^2 - 8*r_cc^3)) + (x[,"n_21"] + x[,"n_24"])*(log10(pmax(1e-6,r_cc)) + log10(4 - 7*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_12"] + x[,"n_33"])*log10(15 - 18*r_cc + 15*r_cc^2 - 4*r_cc^3 - 8*r_cc^4) + (x[,"n_22"] + x[,"n_23"])*log10(11 - 12*r_cc + 18*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cc <- (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 2*x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + 4*(x[,"n_00"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_35"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_25"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_41"])*(log(2 - r_cc) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_44"])*(3*log(pmax(1e-6,1 - r_cc)) + log(1 + r_cc)) + (x[,"n_02"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc)) + log(5 + r_cc)) + (x[,"n_03"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(6 - 7*r_cc + r_cc^2)) + (x[,"n_11"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 4*r_cc^2)) + (x[,"n_14"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(7 - 8*r_cc + 4*r_cc^2)) + (x[,"n_13"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(32 - 45*r_cc + 36*r_cc^2 - 8*r_cc^3)) + (x[,"n_21"] + x[,"n_24"])*(log(pmax(1e-6,r_cc)) + log(4 - 7*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_12"] + x[,"n_33"])*log(15 - 18*r_cc + 15*r_cc^2 - 4*r_cc^3 - 8*r_cc^4) + (x[,"n_22"] + x[,"n_23"])*log(11 - 12*r_cc + 18*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n21,n22,n23,n24,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n20,n25) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n00 + n45)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n40)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n04 + n41)*(2*log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n01 + n44)*(2*log(pmax(1e-6,1 - r)) + log(3 - r + 3*r^2)) + (n15 + n30)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n10 + n35)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n14 + n31)*(log(pmax(1e-6,r)) + log(24 - 29*r + 44*r^2 - 24*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n03 + n42)*(log(pmax(1e-6,r)) + log(7 - 8*r + 6*r^2 - 2*r^3)) + (n11 + n34)*log(15 - 2*r - 41*r^2 + 52*r^3 - 24*r^4) + (n22 + n23)*log(17 - 8*r^2 + 16*r^3 - 8*r^4) + (n02 + n43)*log(3 - r - 2*r^2 + 2*r^3 - 2*r^4) + (n13 + n32)*log(21 + 2*r + 34*r^2 - 40*r^3 + 16*r^4) + (n12 + n33)*log(33 - 14*r + 10*r^2 - 24*r^3 + 16*r^4) + (n21 + n24)*log(15 - 8*r + 44*r^2 - 72*r^3 + 36*r^4)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n21,n22,n23,n24,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n20,n25) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n21,n22,n23,n24,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n20,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_20"],x[,"n_25"])
LOD_cm <- 4*(x[,"n_05"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_15"] + x[,"n_30"])*(-2*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_35"])*(-2*log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_31"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_11"] + x[,"n_34"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_22"] + x[,"n_23"])*(-log10(2) + log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(17)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(17)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(53)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(53)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(61)) + (x[,"n_00"] + x[,"n_45"])*(3*log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cm)) + 3*log10(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm)) + log10(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(3 - r_cm + 3*r_cm^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_cm)) + log10(7 - 12*r_cm + 8*r_cm^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(3 - 4*r_cm + 8*r_cm^2)) + (x[,"n_14"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(24 - 29*r_cm + 44*r_cm^2 - 24*r_cm^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_cm)) + log10(2 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3)) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm)) + log10(7 - 8*r_cm + 6*r_cm^2 - 2*r_cm^3)) + (x[,"n_11"] + x[,"n_34"])*log10(15 - 2*r_cm - 41*r_cm^2 + 52*r_cm^3 - 24*r_cm^4) + (x[,"n_22"] + x[,"n_23"])*log10(17 - 8*r_cm^2 + 16*r_cm^3 - 8*r_cm^4) + (x[,"n_02"] + x[,"n_43"])*log10(3 - r_cm - 2*r_cm^2 + 2*r_cm^3 - 2*r_cm^4) + (x[,"n_13"] + x[,"n_32"])*log10(21 + 2*r_cm + 34*r_cm^2 - 40*r_cm^3 + 16*r_cm^4) + (x[,"n_12"] + x[,"n_33"])*log10(33 - 14*r_cm + 10*r_cm^2 - 24*r_cm^3 + 16*r_cm^4) + (x[,"n_21"] + x[,"n_24"])*log10(15 - 8*r_cm + 44*r_cm^2 - 72*r_cm^3 + 36*r_cm^4)
logL_cm <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_00"] + x[,"n_45"])*(3*log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cm)) + 3*log(pmax(1e-6,r_cm))) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm)) + log(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cm)) + log(3 - r_cm + 3*r_cm^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,r_cm)) + log(7 - 12*r_cm + 8*r_cm^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cm)) + log(3 - 4*r_cm + 8*r_cm^2)) + (x[,"n_14"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(24 - 29*r_cm + 44*r_cm^2 - 24*r_cm^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_cm)) + log(2 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3)) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,r_cm)) + log(7 - 8*r_cm + 6*r_cm^2 - 2*r_cm^3)) + (x[,"n_11"] + x[,"n_34"])*log(15 - 2*r_cm - 41*r_cm^2 + 52*r_cm^3 - 24*r_cm^4) + (x[,"n_22"] + x[,"n_23"])*log(17 - 8*r_cm^2 + 16*r_cm^3 - 8*r_cm^4) + (x[,"n_02"] + x[,"n_43"])*log(3 - r_cm - 2*r_cm^2 + 2*r_cm^3 - 2*r_cm^4) + (x[,"n_13"] + x[,"n_32"])*log(21 + 2*r_cm + 34*r_cm^2 - 40*r_cm^3 + 16*r_cm^4) + (x[,"n_12"] + x[,"n_33"])*log(33 - 14*r_cm + 10*r_cm^2 - 24*r_cm^3 + 16*r_cm^4) + (x[,"n_21"] + x[,"n_24"])*log(15 - 8*r_cm + 44*r_cm^2 - 72*r_cm^3 + 36*r_cm^4)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - n15 - 2*n20 - 2*n21 - n22 - n23 - 2*n24 - 2*n25 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(2) + (n21 + n24)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n00 + n05 + n40 + n45)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n41)*(log(pmax(1e-6,r)) + log(2 - 3*r + 4*r^2 - 3*r^3)) + (n01 + n44)*(log(pmax(1e-6,r)) + log(4 - 9*r + 8*r^2 - 3*r^3)) + (n13 + n32)*log(9 - 2*r + 3*r^2 + 4*r^3 - 8*r^4) + (n12 + n33)*log(6 + 16*r - 33*r^2 + 28*r^3 - 8*r^4) + (n21 + n24)*log(1 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n10 + n15 + n30 + n35)*log(r - 3*r^2 + 4*r^3 - 2*r^4) + (n02 + n43)*log(3 - 8*r + 14*r^2 - 10*r^3 + 2*r^4) + (n03 + n42)*log(1 + 2*r - 4*r^2 + 2*r^3 + 2*r^4) + (n20 + n25)*log(1 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + (n22 + n23)*log(9 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + (n11 + n34)*log(6 - 14*r + 27*r^2 - 28*r^3 + 12*r^4) + (n14 + n31)*log(3 - 4*r + 15*r^2 - 20*r^3 + 12*r^4)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cr <- 3*(x[,"n_10"] + x[,"n_15"] + x[,"n_30"] + x[,"n_35"])*log10(2) + 4*(x[,"n_00"] + x[,"n_05"] + x[,"n_40"] + x[,"n_45"])*log10(2) + (x[,"n_20"] + x[,"n_25"])*(3*log10(2) - log10(3)) - (x[,"n_14"] + x[,"n_31"])*log10(3) - (x[,"n_11"] + x[,"n_34"])*log10(3) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_13"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_33"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-3*log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_23"])*(-3*log10(2) + log10(67)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_40"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 3*r_cr + 4*r_cr^2 - 3*r_cr^3)) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 9*r_cr + 8*r_cr^2 - 3*r_cr^3)) + (x[,"n_13"] + x[,"n_32"])*log10(9 - 2*r_cr + 3*r_cr^2 + 4*r_cr^3 - 8*r_cr^4) + (x[,"n_12"] + x[,"n_33"])*log10(6 + 16*r_cr - 33*r_cr^2 + 28*r_cr^3 - 8*r_cr^4) + (x[,"n_21"] + x[,"n_24"])*log10(1 + 4*r_cr - 10*r_cr^2 + 12*r_cr^3 - 6*r_cr^4) + (x[,"n_10"] + x[,"n_15"] + x[,"n_30"] + x[,"n_35"])*log10(r_cr - 3*r_cr^2 + 4*r_cr^3 - 2*r_cr^4) + (x[,"n_02"] + x[,"n_43"])*log10(3 - 8*r_cr + 14*r_cr^2 - 10*r_cr^3 + 2*r_cr^4) + (x[,"n_03"] + x[,"n_42"])*log10(1 + 2*r_cr - 4*r_cr^2 + 2*r_cr^3 + 2*r_cr^4) + (x[,"n_20"] + x[,"n_25"])*log10(1 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + (x[,"n_22"] + x[,"n_23"])*log10(9 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + (x[,"n_11"] + x[,"n_34"])*log10(6 - 14*r_cr + 27*r_cr^2 - 28*r_cr^3 + 12*r_cr^4) + (x[,"n_14"] + x[,"n_31"])*log10(3 - 4*r_cr + 15*r_cr^2 - 20*r_cr^3 + 12*r_cr^4)
logL_cr <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_00"] + x[,"n_05"] + x[,"n_40"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_cr)) + log(2 - 3*r_cr + 4*r_cr^2 - 3*r_cr^3)) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_cr)) + log(4 - 9*r_cr + 8*r_cr^2 - 3*r_cr^3)) + (x[,"n_13"] + x[,"n_32"])*log(9 - 2*r_cr + 3*r_cr^2 + 4*r_cr^3 - 8*r_cr^4) + (x[,"n_12"] + x[,"n_33"])*log(6 + 16*r_cr - 33*r_cr^2 + 28*r_cr^3 - 8*r_cr^4) + (x[,"n_21"] + x[,"n_24"])*log(1 + 4*r_cr - 10*r_cr^2 + 12*r_cr^3 - 6*r_cr^4) + (x[,"n_10"] + x[,"n_15"] + x[,"n_30"] + x[,"n_35"])*log(r_cr - 3*r_cr^2 + 4*r_cr^3 - 2*r_cr^4) + (x[,"n_02"] + x[,"n_43"])*log(3 - 8*r_cr + 14*r_cr^2 - 10*r_cr^3 + 2*r_cr^4) + (x[,"n_03"] + x[,"n_42"])*log(1 + 2*r_cr - 4*r_cr^2 + 2*r_cr^3 + 2*r_cr^4) + (x[,"n_20"] + x[,"n_25"])*log(1 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + (x[,"n_22"] + x[,"n_23"])*log(9 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + (x[,"n_11"] + x[,"n_34"])*log(6 - 14*r_cr + 27*r_cr^2 - 28*r_cr^3 + 12*r_cr^4) + (x[,"n_14"] + x[,"n_31"])*log(3 - 4*r_cr + 15*r_cr^2 - 20*r_cr^3 + 12*r_cr^4)
logL_mc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 3*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 3*n25 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n40)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n00 + n45)*(2*log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n04 + n41)*(2*log(pmax(1e-6,r)) + log(11 - 8*r + 3*r^2)) + (n01 + n44)*(2*log(pmax(1e-6,1 - r)) + log(6 + 2*r + 3*r^2)) + (n10 + n35)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r^2)) + (n15 + n30)*(2*log(pmax(1e-6,r)) + log(7 - 8*r + 4*r^2)) + (n14 + n31)*(log(pmax(1e-6,r)) + log(30 - 25*r + 28*r^2 - 12*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n03 + n42)*(log(pmax(1e-6,r)) + log(18 - 17*r + 10*r^2 - 2*r^3)) + (n11 + n34)*log(21 - 16*r - 13*r^2 + 20*r^3 - 12*r^4) + (n22 + n23)*log(67 - 4*r - 8*r^2 + 24*r^3 - 12*r^4) + (n02 + n43)*log(9 - 6*r + r^2 - 2*r^3 - 2*r^4) + (n13 + n32)*log(27 + 8*r + 30*r^2 - 28*r^3 + 8*r^4) + (n12 + n33)*log(45 - 16*r - 6*r^2 - 4*r^3 + 8*r^4) + (n21 + n24)*log(21 - 4*r + 22*r^2 - 36*r^3 + 18*r^4)
return(L)}
interlogL_mc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_mc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_mc <- -((x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-4*log10(2) + log10(13)) - (x[,"n_21"] + x[,"n_24"])*(-3*log10(2) + 2*log10(13)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(23)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(23)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + 2*log10(3) + log10(29)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(31)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(31)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(71)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(71)) + (x[,"n_05"] + x[,"n_40"])*(log10(2 - r_mc) + 3*log10(pmax(1e-6,r_mc))) + (x[,"n_00"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(1 - r_mc^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_mc)) + log10(11 - 8*r_mc + 3*r_mc^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(6 + 2*r_mc + 3*r_mc^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 + 4*r_mc^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_mc)) + log10(7 - 8*r_mc + 4*r_mc^2)) + (x[,"n_14"] + x[,"n_31"])*(log10(pmax(1e-6,r_mc)) + log10(30 - 25*r_mc + 28*r_mc^2 - 12*r_mc^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_mc)) + log10(4 - 7*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,r_mc)) + log10(18 - 17*r_mc + 10*r_mc^2 - 2*r_mc^3)) + (x[,"n_11"] + x[,"n_34"])*log10(21 - 16*r_mc - 13*r_mc^2 + 20*r_mc^3 - 12*r_mc^4) + (x[,"n_22"] + x[,"n_23"])*log10(67 - 4*r_mc - 8*r_mc^2 + 24*r_mc^3 - 12*r_mc^4) + (x[,"n_02"] + x[,"n_43"])*log10(9 - 6*r_mc + r_mc^2 - 2*r_mc^3 - 2*r_mc^4) + (x[,"n_13"] + x[,"n_32"])*log10(27 + 8*r_mc + 30*r_mc^2 - 28*r_mc^3 + 8*r_mc^4) + (x[,"n_12"] + x[,"n_33"])*log10(45 - 16*r_mc - 6*r_mc^2 - 4*r_mc^3 + 8*r_mc^4) + (x[,"n_21"] + x[,"n_24"])*log10(21 - 4*r_mc + 22*r_mc^2 - 36*r_mc^3 + 18*r_mc^4)
logL_mc <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 3*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_40"])*(log(2 - r_mc) + 3*log(pmax(1e-6,r_mc))) + (x[,"n_00"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_mc)) + log(1 - r_mc^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,r_mc)) + log(11 - 8*r_mc + 3*r_mc^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_mc)) + log(6 + 2*r_mc + 3*r_mc^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 + 4*r_mc^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,r_mc)) + log(7 - 8*r_mc + 4*r_mc^2)) + (x[,"n_14"] + x[,"n_31"])*(log(pmax(1e-6,r_mc)) + log(30 - 25*r_mc + 28*r_mc^2 - 12*r_mc^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_mc)) + log(4 - 7*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,r_mc)) + log(18 - 17*r_mc + 10*r_mc^2 - 2*r_mc^3)) + (x[,"n_11"] + x[,"n_34"])*log(21 - 16*r_mc - 13*r_mc^2 + 20*r_mc^3 - 12*r_mc^4) + (x[,"n_22"] + x[,"n_23"])*log(67 - 4*r_mc - 8*r_mc^2 + 24*r_mc^3 - 12*r_mc^4) + (x[,"n_02"] + x[,"n_43"])*log(9 - 6*r_mc + r_mc^2 - 2*r_mc^3 - 2*r_mc^4) + (x[,"n_13"] + x[,"n_32"])*log(27 + 8*r_mc + 30*r_mc^2 - 28*r_mc^3 + 8*r_mc^4) + (x[,"n_12"] + x[,"n_33"])*log(45 - 16*r_mc - 6*r_mc^2 - 4*r_mc^3 + 8*r_mc^4) + (x[,"n_21"] + x[,"n_24"])*log(21 - 4*r_mc + 22*r_mc^2 - 36*r_mc^3 + 18*r_mc^4)
logL_mm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 4*n10 - 4*n11 - 3*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n40)*(log(pmax(1e-6,1 - r)) + log(2 - r) + 2*log(pmax(1e-6,r))) + (n00 + n45)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 - r^2)) + (n15 + n30)*(log(pmax(1e-6,r)) + log(12 - 21*r + 20*r^2 - 8*r^3)) + (n04 + n41)*(log(pmax(1e-6,r)) + log(9 - 10*r + 7*r^2 - 3*r^3)) + (n21 + n24)*log(69 - 8*r - 28*r^2 + 72*r^3 - 36*r^4) + (n10 + n35)*log(3 + 2*r - 9*r^2 + 12*r^3 - 8*r^4) + (n13 + n32)*log(51 + 4*r - 2*r^2 + 12*r^3 - 8*r^4) + (n12 + n33)*log(57 - 4*r - 14*r^2 + 20*r^3 - 8*r^4) + (n01 + n44)*log(3 + 2*r - 7*r^2 + 5*r^3 - 3*r^4) + (n02 + n43)*log(12 - 10*r + 11*r^2 - 6*r^3 + 2*r^4) + (n03 + n42)*log(9 - 2*r + 5*r^2 - 2*r^3 + 2*r^4) + (n22 + n23)*log(31 + 4*r - 8*r^3 + 4*r^4) + (n20 + n25)*log(9 - 16*r + 28*r^2 - 24*r^3 + 12*r^4) + (n14 + n31)*log(27 + 6*r + 37*r^2 - 52*r^3 + 24*r^4) + (n11 + n34)*log(42 - 20*r + 25*r^2 - 44*r^3 + 24*r^4)
return(L)}
interlogL_mm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_mm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm <- mapply(interlogL_mm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_mm <- -((x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_30"])*(-2*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_35"])*(-2*log10(2) + log10(11)) - (x[,"n_20"] + x[,"n_25"])*(-2*log10(2) + log10(23)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(7) + log10(37)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(43)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(43)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(3) + log10(43)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(73)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(73)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(107)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(107)) - (x[,"n_14"] + x[,"n_31"])*(-2*log10(2) + log10(137)) - (x[,"n_11"] + x[,"n_34"])*(-2*log10(2) + log10(137)) + (x[,"n_05"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_mm)) + log10(2 - r_mm) + 2*log10(pmax(1e-6,r_mm))) + (x[,"n_00"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_mm)) + log10(pmax(1e-6,r_mm)) + log10(1 - r_mm^2)) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,r_mm)) + log10(12 - 21*r_mm + 20*r_mm^2 - 8*r_mm^3)) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_mm)) + log10(9 - 10*r_mm + 7*r_mm^2 - 3*r_mm^3)) + (x[,"n_21"] + x[,"n_24"])*log10(69 - 8*r_mm - 28*r_mm^2 + 72*r_mm^3 - 36*r_mm^4) + (x[,"n_10"] + x[,"n_35"])*log10(3 + 2*r_mm - 9*r_mm^2 + 12*r_mm^3 - 8*r_mm^4) + (x[,"n_13"] + x[,"n_32"])*log10(51 + 4*r_mm - 2*r_mm^2 + 12*r_mm^3 - 8*r_mm^4) + (x[,"n_12"] + x[,"n_33"])*log10(57 - 4*r_mm - 14*r_mm^2 + 20*r_mm^3 - 8*r_mm^4) + (x[,"n_01"] + x[,"n_44"])*log10(3 + 2*r_mm - 7*r_mm^2 + 5*r_mm^3 - 3*r_mm^4) + (x[,"n_02"] + x[,"n_43"])*log10(12 - 10*r_mm + 11*r_mm^2 - 6*r_mm^3 + 2*r_mm^4) + (x[,"n_03"] + x[,"n_42"])*log10(9 - 2*r_mm + 5*r_mm^2 - 2*r_mm^3 + 2*r_mm^4) + (x[,"n_22"] + x[,"n_23"])*log10(31 + 4*r_mm - 8*r_mm^3 + 4*r_mm^4) + (x[,"n_20"] + x[,"n_25"])*log10(9 - 16*r_mm + 28*r_mm^2 - 24*r_mm^3 + 12*r_mm^4) + (x[,"n_14"] + x[,"n_31"])*log10(27 + 6*r_mm + 37*r_mm^2 - 52*r_mm^3 + 24*r_mm^4) + (x[,"n_11"] + x[,"n_34"])*log10(42 - 20*r_mm + 25*r_mm^2 - 44*r_mm^3 + 24*r_mm^4)
logL_mm <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_mm)) + log(2 - r_mm) + 2*log(pmax(1e-6,r_mm))) + (x[,"n_00"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_mm)) + log(pmax(1e-6,r_mm)) + log(1 - r_mm^2)) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,r_mm)) + log(12 - 21*r_mm + 20*r_mm^2 - 8*r_mm^3)) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_mm)) + log(9 - 10*r_mm + 7*r_mm^2 - 3*r_mm^3)) + (x[,"n_21"] + x[,"n_24"])*log(69 - 8*r_mm - 28*r_mm^2 + 72*r_mm^3 - 36*r_mm^4) + (x[,"n_10"] + x[,"n_35"])*log(3 + 2*r_mm - 9*r_mm^2 + 12*r_mm^3 - 8*r_mm^4) + (x[,"n_13"] + x[,"n_32"])*log(51 + 4*r_mm - 2*r_mm^2 + 12*r_mm^3 - 8*r_mm^4) + (x[,"n_12"] + x[,"n_33"])*log(57 - 4*r_mm - 14*r_mm^2 + 20*r_mm^3 - 8*r_mm^4) + (x[,"n_01"] + x[,"n_44"])*log(3 + 2*r_mm - 7*r_mm^2 + 5*r_mm^3 - 3*r_mm^4) + (x[,"n_02"] + x[,"n_43"])*log(12 - 10*r_mm + 11*r_mm^2 - 6*r_mm^3 + 2*r_mm^4) + (x[,"n_03"] + x[,"n_42"])*log(9 - 2*r_mm + 5*r_mm^2 - 2*r_mm^3 + 2*r_mm^4) + (x[,"n_22"] + x[,"n_23"])*log(31 + 4*r_mm - 8*r_mm^3 + 4*r_mm^4) + (x[,"n_20"] + x[,"n_25"])*log(9 - 16*r_mm + 28*r_mm^2 - 24*r_mm^3 + 12*r_mm^4) + (x[,"n_14"] + x[,"n_31"])*log(27 + 6*r_mm + 37*r_mm^2 - 52*r_mm^3 + 24*r_mm^4) + (x[,"n_11"] + x[,"n_34"])*log(42 - 20*r_mm + 25*r_mm^2 - 44*r_mm^3 + 24*r_mm^4)
logL_mr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 3*n21 - 4*n22 - 4*n23 - 3*n24 - 4*n25 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45)*log(2) + (n22 + n23)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n40)*(2*log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n00 + n45)*(2*log(pmax(1e-6,r)) + log(1 - r^2)) + (n01 + n44)*(log(pmax(1e-6,r)) + log(4 + 2*r - 6*r^2 + 3*r^3)) + (n10 + n35)*(log(pmax(1e-6,r)) + log(2 + r - 4*r^2 + 4*r^3)) + (n14 + n31)*log(15 - 2*r - 19*r^2 + 24*r^3 - 12*r^4) + (n11 + n34)*log(6 + 16*r - 19*r^2 + 24*r^3 - 12*r^4) + (n20 + n25)*log(1 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n22 + n23)*log(21 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n03 + n42)*log(10 - 14*r + 15*r^2 - 6*r^3 - 2*r^4) + (n02 + n43)*log(3 + 10*r - 15*r^2 + 14*r^3 - 2*r^4) + (n04 + n41)*log(3 - 2*r + 2*r^2 - 6*r^3 + 3*r^4) + (n15 + n30)*log(3 - 8*r + 13*r^2 - 12*r^3 + 4*r^4) + (n12 + n33)*log(33 - 8*r + 42*r^2 - 36*r^3 + 8*r^4) + (n13 + n32)*log(39 - 18*r^2 + 4*r^3 + 8*r^4) + (n21 + n24)*log(12 - 8*r + 17*r^2 - 18*r^3 + 9*r^4)
return(L)}
interlogL_mr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_mr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mr <- mapply(interlogL_mr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_mr <- -((x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-3*log10(2) + log10(13)) - (x[,"n_21"] + x[,"n_24"])*(-4*log10(2) + 2*log10(13)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(23)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(23)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(3) + log10(29)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(31)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(31)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(71)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(71)) + (x[,"n_05"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(2 - r_mr) + log10(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_mr)) + log10(1 - r_mr^2)) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_mr)) + log10(4 + 2*r_mr - 6*r_mr^2 + 3*r_mr^3)) + (x[,"n_10"] + x[,"n_35"])*(log10(pmax(1e-6,r_mr)) + log10(2 + r_mr - 4*r_mr^2 + 4*r_mr^3)) + (x[,"n_14"] + x[,"n_31"])*log10(15 - 2*r_mr - 19*r_mr^2 + 24*r_mr^3 - 12*r_mr^4) + (x[,"n_11"] + x[,"n_34"])*log10(6 + 16*r_mr - 19*r_mr^2 + 24*r_mr^3 - 12*r_mr^4) + (x[,"n_20"] + x[,"n_25"])*log10(1 + 4*r_mr - 10*r_mr^2 + 12*r_mr^3 - 6*r_mr^4) + (x[,"n_22"] + x[,"n_23"])*log10(21 + 4*r_mr - 8*r_mr^2 + 8*r_mr^3 - 4*r_mr^4) + (x[,"n_03"] + x[,"n_42"])*log10(10 - 14*r_mr + 15*r_mr^2 - 6*r_mr^3 - 2*r_mr^4) + (x[,"n_02"] + x[,"n_43"])*log10(3 + 10*r_mr - 15*r_mr^2 + 14*r_mr^3 - 2*r_mr^4) + (x[,"n_04"] + x[,"n_41"])*log10(3 - 2*r_mr + 2*r_mr^2 - 6*r_mr^3 + 3*r_mr^4) + (x[,"n_15"] + x[,"n_30"])*log10(3 - 8*r_mr + 13*r_mr^2 - 12*r_mr^3 + 4*r_mr^4) + (x[,"n_12"] + x[,"n_33"])*log10(33 - 8*r_mr + 42*r_mr^2 - 36*r_mr^3 + 8*r_mr^4) + (x[,"n_13"] + x[,"n_32"])*log10(39 - 18*r_mr^2 + 4*r_mr^3 + 8*r_mr^4) + (x[,"n_21"] + x[,"n_24"])*log10(12 - 8*r_mr + 17*r_mr^2 - 18*r_mr^3 + 9*r_mr^4)
logL_mr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 3*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"])*log(2) + (x[,"n_22"] + x[,"n_23"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_mr)) + log(2 - r_mr) + log(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_45"])*(2*log(pmax(1e-6,r_mr)) + log(1 - r_mr^2)) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_mr)) + log(4 + 2*r_mr - 6*r_mr^2 + 3*r_mr^3)) + (x[,"n_10"] + x[,"n_35"])*(log(pmax(1e-6,r_mr)) + log(2 + r_mr - 4*r_mr^2 + 4*r_mr^3)) + (x[,"n_14"] + x[,"n_31"])*log(15 - 2*r_mr - 19*r_mr^2 + 24*r_mr^3 - 12*r_mr^4) + (x[,"n_11"] + x[,"n_34"])*log(6 + 16*r_mr - 19*r_mr^2 + 24*r_mr^3 - 12*r_mr^4) + (x[,"n_20"] + x[,"n_25"])*log(1 + 4*r_mr - 10*r_mr^2 + 12*r_mr^3 - 6*r_mr^4) + (x[,"n_22"] + x[,"n_23"])*log(21 + 4*r_mr - 8*r_mr^2 + 8*r_mr^3 - 4*r_mr^4) + (x[,"n_03"] + x[,"n_42"])*log(10 - 14*r_mr + 15*r_mr^2 - 6*r_mr^3 - 2*r_mr^4) + (x[,"n_02"] + x[,"n_43"])*log(3 + 10*r_mr - 15*r_mr^2 + 14*r_mr^3 - 2*r_mr^4) + (x[,"n_04"] + x[,"n_41"])*log(3 - 2*r_mr + 2*r_mr^2 - 6*r_mr^3 + 3*r_mr^4) + (x[,"n_15"] + x[,"n_30"])*log(3 - 8*r_mr + 13*r_mr^2 - 12*r_mr^3 + 4*r_mr^4) + (x[,"n_12"] + x[,"n_33"])*log(33 - 8*r_mr + 42*r_mr^2 - 36*r_mr^3 + 8*r_mr^4) + (x[,"n_13"] + x[,"n_32"])*log(39 - 18*r_mr^2 + 4*r_mr^3 + 8*r_mr^4) + (x[,"n_21"] + x[,"n_24"])*log(12 - 8*r_mr + 17*r_mr^2 - 18*r_mr^3 + 9*r_mr^4)
logL_rc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n20 - 2*n21 - 3*n22 - 3*n23 - 2*n24 - 3*n25 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n00 + n45)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n40)*(2*log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n15 + n30)*(log(pmax(1e-6,r)) + log(6 - 11*r + 12*r^2 - 4*r^3)) + (n04 + n41)*(log(pmax(1e-6,r)) + log(12 - 16*r + 10*r^2 - 3*r^3)) + (n21 + n24)*log(15 + 4*r - 13*r^2 + 18*r^3 - 9*r^4) + (n12 + n33)*log(57 - 4*r - 12*r^2 + 12*r^3 - 8*r^4) + (n13 + n32)*log(45 + 24*r - 24*r^2 + 20*r^3 - 8*r^4) + (n10 + n35)*log(3 - 4*r + r^2 + 4*r^3 - 4*r^4) + (n01 + n44)*log(3 + 2*r - 4*r^2 + 2*r^3 - 3*r^4) + (n03 + n42)*log(9 - 6*r + 13*r^2 - 6*r^3 + 2*r^4) + (n02 + n43)*log(12 - 10*r + 7*r^2 - 2*r^3 + 2*r^4) + (n20 + n25)*log(3 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + (n14 + n31)*log(18 - 12*r + 35*r^2 - 32*r^3 + 12*r^4) + (n22 + n23)*log(99 - 4*r + 16*r^2 - 24*r^3 + 12*r^4) + (n11 + n34)*log(21 - 10*r + 11*r^2 - 16*r^3 + 12*r^4)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rc <- -((x[,"n_15"] + x[,"n_30"])*(-log10(2) + log10(3))) - (x[,"n_10"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_20"] + x[,"n_25"])*(-3*log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_24"])*(-4*log10(2) + log10(13) + log10(19)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_13"] + x[,"n_32"])*log10(53) - (x[,"n_12"] + x[,"n_33"])*log10(53) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(5) + log10(79)) + (x[,"n_00"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc)) + log10(2 + r_rc)) + (x[,"n_05"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_rc)) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,r_rc)) + log10(6 - 11*r_rc + 12*r_rc^2 - 4*r_rc^3)) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_rc)) + log10(12 - 16*r_rc + 10*r_rc^2 - 3*r_rc^3)) + (x[,"n_21"] + x[,"n_24"])*log10(15 + 4*r_rc - 13*r_rc^2 + 18*r_rc^3 - 9*r_rc^4) + (x[,"n_12"] + x[,"n_33"])*log10(57 - 4*r_rc - 12*r_rc^2 + 12*r_rc^3 - 8*r_rc^4) + (x[,"n_13"] + x[,"n_32"])*log10(45 + 24*r_rc - 24*r_rc^2 + 20*r_rc^3 - 8*r_rc^4) + (x[,"n_10"] + x[,"n_35"])*log10(3 - 4*r_rc + r_rc^2 + 4*r_rc^3 - 4*r_rc^4) + (x[,"n_01"] + x[,"n_44"])*log10(3 + 2*r_rc - 4*r_rc^2 + 2*r_rc^3 - 3*r_rc^4) + (x[,"n_03"] + x[,"n_42"])*log10(9 - 6*r_rc + 13*r_rc^2 - 6*r_rc^3 + 2*r_rc^4) + (x[,"n_02"] + x[,"n_43"])*log10(12 - 10*r_rc + 7*r_rc^2 - 2*r_rc^3 + 2*r_rc^4) + (x[,"n_20"] + x[,"n_25"])*log10(3 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + (x[,"n_14"] + x[,"n_31"])*log10(18 - 12*r_rc + 35*r_rc^2 - 32*r_rc^3 + 12*r_rc^4) + (x[,"n_22"] + x[,"n_23"])*log10(99 - 4*r_rc + 16*r_rc^2 - 24*r_rc^3 + 12*r_rc^4) + (x[,"n_11"] + x[,"n_34"])*log10(21 - 10*r_rc + 11*r_rc^2 - 16*r_rc^3 + 12*r_rc^4)
logL_rc <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc)) + log(2 + r_rc)) + (x[,"n_05"] + x[,"n_40"])*(2*log(pmax(1e-6,r_rc)) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,r_rc)) + log(6 - 11*r_rc + 12*r_rc^2 - 4*r_rc^3)) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_rc)) + log(12 - 16*r_rc + 10*r_rc^2 - 3*r_rc^3)) + (x[,"n_21"] + x[,"n_24"])*log(15 + 4*r_rc - 13*r_rc^2 + 18*r_rc^3 - 9*r_rc^4) + (x[,"n_12"] + x[,"n_33"])*log(57 - 4*r_rc - 12*r_rc^2 + 12*r_rc^3 - 8*r_rc^4) + (x[,"n_13"] + x[,"n_32"])*log(45 + 24*r_rc - 24*r_rc^2 + 20*r_rc^3 - 8*r_rc^4) + (x[,"n_10"] + x[,"n_35"])*log(3 - 4*r_rc + r_rc^2 + 4*r_rc^3 - 4*r_rc^4) + (x[,"n_01"] + x[,"n_44"])*log(3 + 2*r_rc - 4*r_rc^2 + 2*r_rc^3 - 3*r_rc^4) + (x[,"n_03"] + x[,"n_42"])*log(9 - 6*r_rc + 13*r_rc^2 - 6*r_rc^3 + 2*r_rc^4) + (x[,"n_02"] + x[,"n_43"])*log(12 - 10*r_rc + 7*r_rc^2 - 2*r_rc^3 + 2*r_rc^4) + (x[,"n_20"] + x[,"n_25"])*log(3 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + (x[,"n_14"] + x[,"n_31"])*log(18 - 12*r_rc + 35*r_rc^2 - 32*r_rc^3 + 12*r_rc^4) + (x[,"n_22"] + x[,"n_23"])*log(99 - 4*r_rc + 16*r_rc^2 - 24*r_rc^3 + 12*r_rc^4) + (x[,"n_11"] + x[,"n_34"])*log(21 - 10*r_rc + 11*r_rc^2 - 16*r_rc^3 + 12*r_rc^4)
logL_rm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 3*n10 - 3*n11 - n12 - n13 - 3*n14 - 3*n15 - 3*n20 - 3*n21 - 3*n24 - 3*n25 - 3*n30 - 3*n31 - n32 - n33 - 3*n34 - 3*n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(2) + (-2*n00 - n01 - 2*n02 - 2*n03 - n04 - 2*n05 - 2*n10 - n11 - 2*n12 - 2*n13 - n14 - 2*n15 - n20 - n21 - n22 - n23 - n24 - n25 - 2*n30 - n31 - 2*n32 - 2*n33 - n34 - 2*n35 - 2*n40 - n41 - 2*n42 - 2*n43 - n44 - 2*n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n00 + n45)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n40)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n01 + n44)*(log(pmax(1e-6,r)) + log(3 - r^2 + r^3)) + (n10 + n35)*(log(pmax(1e-6,r)) + log(12 - 7*r - 4*r^2 + 8*r^3)) + (n11 + n34)*log(9 + 18*r - 7*r^2 + 12*r^3 - 8*r^4) + (n14 + n31)*log(24 - 8*r - 19*r^2 + 20*r^3 - 8*r^4) + (n20 + n25)*log(3 - 4*r^2 + 8*r^3 - 4*r^4) + (n03 + n42)*log(18 - 12*r + 7*r^2 - 2*r^3 - 2*r^4) + (n02 + n43)*log(9 + 12*r - 11*r^2 + 10*r^3 - 2*r^4) + (n22 + n23)*log(12 + r - 2*r^2 + 2*r^3 - r^4) + (n04 + n41)*log(3 - 4*r + 3*r^2 - 3*r^3 + r^4) + (n12 + n33)*log(36 + 3*r + 16*r^2 - 14*r^3 + 4*r^4) + (n13 + n32)*log(45 - 9*r - 2*r^2 - 2*r^3 + 4*r^4) + (n15 + n30)*log(9 - 18*r + 29*r^2 - 28*r^3 + 8*r^4) + (n21 + n24)*log(33 - 8*r + 20*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_22"],x[,"n_23"])
LOD_rm <- -((x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_32"])*(3*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_33"])*(3*log10(2) + log10(5)) - (x[,"n_20"] + x[,"n_25"])*(-2*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_23"])*(-4*log10(2) + log10(3) + log10(5) + log10(13)) - (x[,"n_15"] + x[,"n_30"])*(-2*log10(2) + log10(17)) - (x[,"n_10"] + x[,"n_35"])*(-2*log10(2) + log10(17)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(23)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(23)) - (x[,"n_14"] + x[,"n_31"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_11"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(107)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(107)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(127)) + (x[,"n_00"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_rm)) + 2*log10(pmax(1e-6,r_rm)) + log10(2 + r_rm)) + (x[,"n_05"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm)) + log10(3 - 4*r_rm + r_rm^2)) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm)) + log10(3 - r_rm^2 + r_rm^3)) + (x[,"n_10"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm)) + log10(12 - 7*r_rm - 4*r_rm^2 + 8*r_rm^3)) + (x[,"n_11"] + x[,"n_34"])*log10(9 + 18*r_rm - 7*r_rm^2 + 12*r_rm^3 - 8*r_rm^4) + (x[,"n_14"] + x[,"n_31"])*log10(24 - 8*r_rm - 19*r_rm^2 + 20*r_rm^3 - 8*r_rm^4) + (x[,"n_20"] + x[,"n_25"])*log10(3 - 4*r_rm^2 + 8*r_rm^3 - 4*r_rm^4) + (x[,"n_03"] + x[,"n_42"])*log10(18 - 12*r_rm + 7*r_rm^2 - 2*r_rm^3 - 2*r_rm^4) + (x[,"n_02"] + x[,"n_43"])*log10(9 + 12*r_rm - 11*r_rm^2 + 10*r_rm^3 - 2*r_rm^4) + (x[,"n_22"] + x[,"n_23"])*log10(12 + r_rm - 2*r_rm^2 + 2*r_rm^3 - r_rm^4) + (x[,"n_04"] + x[,"n_41"])*log10(3 - 4*r_rm + 3*r_rm^2 - 3*r_rm^3 + r_rm^4) + (x[,"n_12"] + x[,"n_33"])*log10(36 + 3*r_rm + 16*r_rm^2 - 14*r_rm^3 + 4*r_rm^4) + (x[,"n_13"] + x[,"n_32"])*log10(45 - 9*r_rm - 2*r_rm^2 - 2*r_rm^3 + 4*r_rm^4) + (x[,"n_15"] + x[,"n_30"])*log10(9 - 18*r_rm + 29*r_rm^2 - 28*r_rm^3 + 8*r_rm^4) + (x[,"n_21"] + x[,"n_24"])*log10(33 - 8*r_rm + 20*r_rm^2 - 24*r_rm^3 + 12*r_rm^4)
logL_rm <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - 2*x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - 2*x[,"n_30"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - x[,"n_44"] - 2*x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_00"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_rm)) + 2*log(pmax(1e-6,r_rm)) + log(2 + r_rm)) + (x[,"n_05"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm)) + log(3 - 4*r_rm + r_rm^2)) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_rm)) + log(3 - r_rm^2 + r_rm^3)) + (x[,"n_10"] + x[,"n_35"])*(log(pmax(1e-6,r_rm)) + log(12 - 7*r_rm - 4*r_rm^2 + 8*r_rm^3)) + (x[,"n_11"] + x[,"n_34"])*log(9 + 18*r_rm - 7*r_rm^2 + 12*r_rm^3 - 8*r_rm^4) + (x[,"n_14"] + x[,"n_31"])*log(24 - 8*r_rm - 19*r_rm^2 + 20*r_rm^3 - 8*r_rm^4) + (x[,"n_20"] + x[,"n_25"])*log(3 - 4*r_rm^2 + 8*r_rm^3 - 4*r_rm^4) + (x[,"n_03"] + x[,"n_42"])*log(18 - 12*r_rm + 7*r_rm^2 - 2*r_rm^3 - 2*r_rm^4) + (x[,"n_02"] + x[,"n_43"])*log(9 + 12*r_rm - 11*r_rm^2 + 10*r_rm^3 - 2*r_rm^4) + (x[,"n_22"] + x[,"n_23"])*log(12 + r_rm - 2*r_rm^2 + 2*r_rm^3 - r_rm^4) + (x[,"n_04"] + x[,"n_41"])*log(3 - 4*r_rm + 3*r_rm^2 - 3*r_rm^3 + r_rm^4) + (x[,"n_12"] + x[,"n_33"])*log(36 + 3*r_rm + 16*r_rm^2 - 14*r_rm^3 + 4*r_rm^4) + (x[,"n_13"] + x[,"n_32"])*log(45 - 9*r_rm - 2*r_rm^2 - 2*r_rm^3 + 4*r_rm^4) + (x[,"n_15"] + x[,"n_30"])*log(9 - 18*r_rm + 29*r_rm^2 - 28*r_rm^3 + 8*r_rm^4) + (x[,"n_21"] + x[,"n_24"])*log(33 - 8*r_rm + 20*r_rm^2 - 24*r_rm^3 + 12*r_rm^4)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n00 + n45)*(3*log(pmax(1e-6,r)) + log(2 + r)) + (n15 + n30)*(2*log(pmax(1e-6,1 - r)) + log(3 + 8*r - 4*r^2)) + (n04 + n41)*(2*log(pmax(1e-6,1 - r)) + log(12 + 2*r - 3*r^2)) + (n01 + n44)*(2*log(pmax(1e-6,r)) + log(11 + 4*r - 3*r^2)) + (n05 + n40)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n02 + n43)*(log(pmax(1e-6,r)) + log(18 + 7*r - 18*r^2 + 2*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(4 - r - 6*r^2 + 3*r^3)) + (n11 + n34)*(log(pmax(1e-6,r)) + log(30 + 17*r - 20*r^2 + 12*r^3)) + (n21 + n24)*log(21 + 44*r - 62*r^2 + 36*r^3 - 18*r^4) + (n13 + n32)*log(75 - 52*r + 24*r^2 - 12*r^3 - 8*r^4) + (n12 + n33)*log(27 + 72*r - 60*r^2 + 44*r^3 - 8*r^4) + (n10 + n35)*log(7*r^2 - 4*r^4) + (n03 + n42)*log(9 + 14*r - 35*r^2 + 10*r^3 + 2*r^4) + (n14 + n31)*log(39 - 52*r + 29*r^2 - 28*r^3 + 12*r^4) + (n22 + n23)*log(111 - 52*r + 64*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rr <- -((x[,"n_15"] + x[,"n_30"])*(-log10(2) + log10(3))) - (x[,"n_10"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_20"] + x[,"n_25"])*(-4*log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_24"])*(-3*log10(2) + log10(13) + log10(19)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_13"] + x[,"n_32"])*log10(53) - (x[,"n_12"] + x[,"n_33"])*log10(53) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(5) + log10(79)) + (x[,"n_00"] + x[,"n_45"])*(3*log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 + 8*r_rr - 4*r_rr^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(12 + 2*r_rr - 3*r_rr^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(11 + 4*r_rr - 3*r_rr^2)) + (x[,"n_05"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 - 4*r_rr + r_rr^2)) + (x[,"n_02"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(18 + 7*r_rr - 18*r_rr^2 + 2*r_rr^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(4 - r_rr - 6*r_rr^2 + 3*r_rr^3)) + (x[,"n_11"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(30 + 17*r_rr - 20*r_rr^2 + 12*r_rr^3)) + (x[,"n_21"] + x[,"n_24"])*log10(21 + 44*r_rr - 62*r_rr^2 + 36*r_rr^3 - 18*r_rr^4) + (x[,"n_13"] + x[,"n_32"])*log10(75 - 52*r_rr + 24*r_rr^2 - 12*r_rr^3 - 8*r_rr^4) + (x[,"n_12"] + x[,"n_33"])*log10(27 + 72*r_rr - 60*r_rr^2 + 44*r_rr^3 - 8*r_rr^4) + (x[,"n_10"] + x[,"n_35"])*log10(7*r_rr^2 - 4*r_rr^4) + (x[,"n_03"] + x[,"n_42"])*log10(9 + 14*r_rr - 35*r_rr^2 + 10*r_rr^3 + 2*r_rr^4) + (x[,"n_14"] + x[,"n_31"])*log10(39 - 52*r_rr + 29*r_rr^2 - 28*r_rr^3 + 12*r_rr^4) + (x[,"n_22"] + x[,"n_23"])*log10(111 - 52*r_rr + 64*r_rr^2 - 24*r_rr^3 + 12*r_rr^4)
logL_rr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_45"])*(3*log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 + 8*r_rr - 4*r_rr^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(12 + 2*r_rr - 3*r_rr^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(11 + 4*r_rr - 3*r_rr^2)) + (x[,"n_05"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 - 4*r_rr + r_rr^2)) + (x[,"n_02"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(18 + 7*r_rr - 18*r_rr^2 + 2*r_rr^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(4 - r_rr - 6*r_rr^2 + 3*r_rr^3)) + (x[,"n_11"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(30 + 17*r_rr - 20*r_rr^2 + 12*r_rr^3)) + (x[,"n_21"] + x[,"n_24"])*log(21 + 44*r_rr - 62*r_rr^2 + 36*r_rr^3 - 18*r_rr^4) + (x[,"n_13"] + x[,"n_32"])*log(75 - 52*r_rr + 24*r_rr^2 - 12*r_rr^3 - 8*r_rr^4) + (x[,"n_12"] + x[,"n_33"])*log(27 + 72*r_rr - 60*r_rr^2 + 44*r_rr^3 - 8*r_rr^4) + (x[,"n_10"] + x[,"n_35"])*log(7*r_rr^2 - 4*r_rr^4) + (x[,"n_03"] + x[,"n_42"])*log(9 + 14*r_rr - 35*r_rr^2 + 10*r_rr^3 + 2*r_rr^4) + (x[,"n_14"] + x[,"n_31"])*log(39 - 52*r_rr + 29*r_rr^2 - 28*r_rr^3 + 12*r_rr^4) + (x[,"n_22"] + x[,"n_23"])*log(111 - 52*r_rr + 64*r_rr^2 - 24*r_rr^3 + 12*r_rr^4)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_mc,r_mm,r_mr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_mc,LOD_mm,LOD_mr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_mc,logL_mm,logL_mr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","mixed coupling","mixed mixed","mixed repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.2_2.4 <- function(x){
logL_cc <- function(r,n01,n02,n04,n05,n11,n12,n13,n14,n15,n22,n24,n31,n32,n33,n34,n35,n41,n42,n44,n45,n03,n21,n23,n25,n43) {
L <- (-n01 - n02 - n04 - n05 - n11 - n12 + n13 - n14 - n15 + n22 + n24 - n31 - n32 + n33 - n34 - n35 - n41 - n42 - n44 - n45)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n05 + n41)*(3*log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n11 + n35 + n44)*(2*log(pmax(1e-6,1 - r)) + log(3 + 8*r - 4*r^2)) + (n01 + n45)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n03 + n21 + n25 + n43)*(log(pmax(1e-6,r)) + log(4 - r - 6*r^2 + 3*r^3)) + (n14 + n32)*(log(pmax(1e-6,r)) + log(18 + 3*r - 16*r^2 + 16*r^3)) + (n13 + n22 + n24 + n33)*log(3 + 5*r - 11*r^2 + 12*r^3 - 6*r^4) + (n04 + n15 + n31 + n42)*log(7*r^2 - 4*r^4) + (n12 + n34)*log(21 - 40*r + 51*r^2 - 48*r^3 + 16*r^4) + n23*log(21 - 28*r + 46*r^2 - 36*r^3 + 18*r^4)
return(L)}
interlogL_cc <- function(n01,n02,n04,n05,n11,n12,n13,n14,n15,n22,n24,n31,n32,n33,n34,n35,n41,n42,n44,n45,n03,n21,n23,n25,n43) {
optimize(logL_cc,c(0,0.5), n01,n02,n04,n05,n11,n12,n13,n14,n15,n22,n24,n31,n32,n33,n34,n35,n41,n42,n44,n45,n03,n21,n23,n25,n43, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_22"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_44"],x[,"n_45"],x[,"n_03"],x[,"n_21"],x[,"n_23"],x[,"n_25"],x[,"n_43"])
LOD_cc <- -((x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(-log10(2) + log10(3))) - (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(-log10(2) + log10(3)) - (x[,"n_05"] + x[,"n_41"])*(-4*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_45"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(5) + log10(7)) - x[,"n_23"]*(-3*log10(2) + 2*log10(11)) - (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(-4*log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(-3*log10(2) + log10(31)) + (x[,"n_05"] + x[,"n_41"])*(3*log10(pmax(1e-6,r_cc)) + log10(2 + r_cc)) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 8*r_cc - 4*r_cc^2)) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 - 4*r_cc + r_cc^2)) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(log10(pmax(1e-6,r_cc)) + log10(4 - r_cc - 6*r_cc^2 + 3*r_cc^3)) + (x[,"n_14"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(18 + 3*r_cc - 16*r_cc^2 + 16*r_cc^3)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log10(3 + 5*r_cc - 11*r_cc^2 + 12*r_cc^3 - 6*r_cc^4) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*log10(7*r_cc^2 - 4*r_cc^4) + (x[,"n_12"] + x[,"n_34"])*log10(21 - 40*r_cc + 51*r_cc^2 - 48*r_cc^3 + 16*r_cc^4) + x[,"n_23"]*log10(21 - 28*r_cc + 46*r_cc^2 - 36*r_cc^3 + 18*r_cc^4)
logL_cc <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] + x[,"n_13"] - x[,"n_14"] - x[,"n_15"] + x[,"n_22"] + x[,"n_24"] - x[,"n_31"] - x[,"n_32"] + x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_44"] - x[,"n_45"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_41"])*(3*log(pmax(1e-6,r_cc)) + log(2 + r_cc)) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 8*r_cc - 4*r_cc^2)) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 - 4*r_cc + r_cc^2)) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(log(pmax(1e-6,r_cc)) + log(4 - r_cc - 6*r_cc^2 + 3*r_cc^3)) + (x[,"n_14"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(18 + 3*r_cc - 16*r_cc^2 + 16*r_cc^3)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(3 + 5*r_cc - 11*r_cc^2 + 12*r_cc^3 - 6*r_cc^4) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*log(7*r_cc^2 - 4*r_cc^4) + (x[,"n_12"] + x[,"n_34"])*log(21 - 40*r_cc + 51*r_cc^2 - 48*r_cc^3 + 16*r_cc^4) + x[,"n_23"]*log(21 - 28*r_cc + 46*r_cc^2 - 36*r_cc^3 + 18*r_cc^4)
logL_cm <- function(r,n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n23,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n13,n22,n24,n33) {
L <- (-2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n11 - 2*n12 - 2*n14 - 2*n15 - 2*n21 - n23 - 2*n25 - 2*n31 - 2*n32 - 2*n34 - 2*n35 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(2) + (n13 + n22 + n24 + n33)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n01 + n45)*(2*log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n05 + n41)*(2*log(pmax(1e-6,r)) + log(1 - r^2)) + (n04 + n15 + n31 + n42)*(log(pmax(1e-6,r)) + log(2 + r - 4*r^2 + 4*r^3)) + n23*log(9 + 8*r - 26*r^2 + 36*r^3 - 18*r^4) + (n14 + n32)*log(3 + 8*r - 13*r^2 + 24*r^3 - 16*r^4) + (n12 + n34)*log(6 + 10*r - 37*r^2 + 40*r^3 - 16*r^4) + (n03 + n21 + n25 + n43)*log(1 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n13 + n22 + n24 + n33)*log(1 - r + 3*r^2 - 4*r^3 + 2*r^4) + (n02 + n11 + n35 + n44)*log(3 - 8*r + 13*r^2 - 12*r^3 + 4*r^4)
return(L)}
interlogL_cm <- function(n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n23,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n13,n22,n24,n33) {
optimize(logL_cm,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n23,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n13,n22,n24,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_23"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_13"],x[,"n_22"],x[,"n_24"],x[,"n_33"])
LOD_cm <- -((x[,"n_05"] + x[,"n_41"])*(-4*log10(2) + log10(3))) - (x[,"n_01"] + x[,"n_45"])*(-4*log10(2) + log10(3)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(3*log10(2) - log10(7)) - (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(-3*log10(2) + log10(13)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(23)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(23)) - x[,"n_23"]*(-3*log10(2) + log10(79)) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(2 - r_cm) + log10(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm)) + log10(1 - r_cm^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm)) + log10(2 + r_cm - 4*r_cm^2 + 4*r_cm^3)) + x[,"n_23"]*log10(9 + 8*r_cm - 26*r_cm^2 + 36*r_cm^3 - 18*r_cm^4) + (x[,"n_14"] + x[,"n_32"])*log10(3 + 8*r_cm - 13*r_cm^2 + 24*r_cm^3 - 16*r_cm^4) + (x[,"n_12"] + x[,"n_34"])*log10(6 + 10*r_cm - 37*r_cm^2 + 40*r_cm^3 - 16*r_cm^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log10(1 + 4*r_cm - 10*r_cm^2 + 12*r_cm^3 - 6*r_cm^4) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log10(1 - r_cm + 3*r_cm^2 - 4*r_cm^3 + 2*r_cm^4) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*log10(3 - 8*r_cm + 13*r_cm^2 - 12*r_cm^3 + 4*r_cm^4)
logL_cm <- (-2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - x[,"n_23"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm)) + log(2 - r_cm) + log(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm)) + log(1 - r_cm^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(log(pmax(1e-6,r_cm)) + log(2 + r_cm - 4*r_cm^2 + 4*r_cm^3)) + x[,"n_23"]*log(9 + 8*r_cm - 26*r_cm^2 + 36*r_cm^3 - 18*r_cm^4) + (x[,"n_14"] + x[,"n_32"])*log(3 + 8*r_cm - 13*r_cm^2 + 24*r_cm^3 - 16*r_cm^4) + (x[,"n_12"] + x[,"n_34"])*log(6 + 10*r_cm - 37*r_cm^2 + 40*r_cm^3 - 16*r_cm^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log(1 + 4*r_cm - 10*r_cm^2 + 12*r_cm^3 - 6*r_cm^4) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(1 - r_cm + 3*r_cm^2 - 4*r_cm^3 + 2*r_cm^4) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*log(3 - 8*r_cm + 13*r_cm^2 - 12*r_cm^3 + 4*r_cm^4)
logL_cr <- function(r,n02,n04,n11,n13,n15,n22,n24,n31,n33,n35,n42,n44,n23,n01,n03,n05,n12,n14,n21,n25,n32,n34,n41,n43,n45) {
L <- (n02 + n04 + n11 + 3*n13 + n15 + 3*n22 + 3*n24 + n31 + 3*n33 + n35 + n42 + n44)*log(2) + n23*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n01 + n05 + n41 + n45)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n02 + n04 + n11 + n15 + n31 + n35 + n42 + n44)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 - 2*r + 2*r^2)) + (n13 + n22 + n24 + n33)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n03 + n21 + n25 + n43)*log(1 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + n23*log(3 - 8*r + 20*r^2 - 24*r^3 + 12*r^4) + (n12 + n14 + n32 + n34)*log(3 - 10*r + 26*r^2 - 32*r^3 + 16*r^4)
return(L)}
interlogL_cr <- function(n02,n04,n11,n13,n15,n22,n24,n31,n33,n35,n42,n44,n23,n01,n03,n05,n12,n14,n21,n25,n32,n34,n41,n43,n45) {
optimize(logL_cr,c(0,0.5), n02,n04,n11,n13,n15,n22,n24,n31,n33,n35,n42,n44,n23,n01,n03,n05,n12,n14,n21,n25,n32,n34,n41,n43,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_02"],x[,"n_04"],x[,"n_11"],x[,"n_13"],x[,"n_15"],x[,"n_22"],x[,"n_24"],x[,"n_31"],x[,"n_33"],x[,"n_35"],x[,"n_42"],x[,"n_44"],x[,"n_23"],x[,"n_01"],x[,"n_03"],x[,"n_05"],x[,"n_12"],x[,"n_14"],x[,"n_21"],x[,"n_25"],x[,"n_32"],x[,"n_34"],x[,"n_41"],x[,"n_43"],x[,"n_45"])
LOD_cr <- 3*(x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*log10(2) + 4*(x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*log10(2) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(3*log10(2) - log10(3)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(-4*log10(2) + log10(5)) - x[,"n_23"]*(-2*log10(2) + log10(7)) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*(log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr)) + log10(1 - 2*r_cr + 2*r_cr^2)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 5*r_cr + 6*r_cr^2 - 3*r_cr^3)) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log10(1 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + x[,"n_23"]*log10(3 - 8*r_cr + 20*r_cr^2 - 24*r_cr^3 + 12*r_cr^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log10(3 - 10*r_cr + 26*r_cr^2 - 32*r_cr^3 + 16*r_cr^4)
logL_cr <- (x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + 3*x[,"n_13"] + x[,"n_15"] + 3*x[,"n_22"] + 3*x[,"n_24"] + x[,"n_31"] + 3*x[,"n_33"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*log(2) + x[,"n_23"]*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*(log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr)) + log(1 - 2*r_cr + 2*r_cr^2)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(log(pmax(1e-6,r_cr)) + log(2 - 5*r_cr + 6*r_cr^2 - 3*r_cr^3)) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log(1 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + x[,"n_23"]*log(3 - 8*r_cr + 20*r_cr^2 - 24*r_cr^3 + 12*r_cr^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log(3 - 10*r_cr + 26*r_cr^2 - 32*r_cr^3 + 16*r_cr^4)
logL_mc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-3*n01 - n02 - 3*n03 - n04 - 3*n05 - n11 - 2*n12 - 2*n13 - 2*n14 - n15 - 3*n21 - 2*n22 - 2*n23 - 2*n24 - 3*n25 - n31 - 2*n32 - 2*n33 - 2*n34 - n35 - 3*n41 - n42 - 3*n43 - n44 - 3*n45)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n05 + n41)*(log(2 - r) + 2*log(pmax(1e-6,r)) + log(2 + r)) + (n01 + n45)*(log(1 - r^2) + log(3 - 4*r + r^2)) + (n04 + n15 + n31 + n42)*(log(pmax(1e-6,r)) + log(3 - r^2 + r^3)) + n23*log(57 + 8*r - 26*r^2 + 36*r^3 - 18*r^4) + (n14 + n32)*log(9 + 18*r - 7*r^2 + 12*r^3 - 8*r^4) + (n12 + n34)*log(24 - 8*r - 19*r^2 + 20*r^3 - 8*r^4) + (n03 + n21 + n25 + n43)*log(9 + 8*r - 14*r^2 + 12*r^3 - 6*r^4) + (n02 + n11 + n35 + n44)*log(3 - 4*r + 3*r^2 - 3*r^3 + r^4) + (n13 + n22 + n24 + n33)*log(33 - 8*r + 20*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_mc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_mc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_mc <- -((x[,"n_05"] + x[,"n_41"])*(-4*log10(2) + log10(3) + log10(5))) - (x[,"n_01"] + x[,"n_45"])*(-4*log10(2) + log10(3) + log10(5)) - (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(-3*log10(2) + log10(5) + log10(17)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(23)) - (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(-4*log10(2) + log10(23)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(-2*log10(2) + log10(127)) - x[,"n_23"]*(-3*log10(2) + log10(463)) + (x[,"n_05"] + x[,"n_41"])*(log10(2 - r_mc) + 2*log10(pmax(1e-6,r_mc)) + log10(2 + r_mc)) + (x[,"n_01"] + x[,"n_45"])*(log10(1 - r_mc^2) + log10(3 - 4*r_mc + r_mc^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(log10(pmax(1e-6,r_mc)) + log10(3 - r_mc^2 + r_mc^3)) + x[,"n_23"]*log10(57 + 8*r_mc - 26*r_mc^2 + 36*r_mc^3 - 18*r_mc^4) + (x[,"n_14"] + x[,"n_32"])*log10(9 + 18*r_mc - 7*r_mc^2 + 12*r_mc^3 - 8*r_mc^4) + (x[,"n_12"] + x[,"n_34"])*log10(24 - 8*r_mc - 19*r_mc^2 + 20*r_mc^3 - 8*r_mc^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log10(9 + 8*r_mc - 14*r_mc^2 + 12*r_mc^3 - 6*r_mc^4) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*log10(3 - 4*r_mc + 3*r_mc^2 - 3*r_mc^3 + r_mc^4) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log10(33 - 8*r_mc + 20*r_mc^2 - 24*r_mc^3 + 12*r_mc^4)
logL_mc <- (-3*x[,"n_01"] - x[,"n_02"] - 3*x[,"n_03"] - x[,"n_04"] - 3*x[,"n_05"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 3*x[,"n_25"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - 3*x[,"n_41"] - x[,"n_42"] - 3*x[,"n_43"] - x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_41"])*(log(2 - r_mc) + 2*log(pmax(1e-6,r_mc)) + log(2 + r_mc)) + (x[,"n_01"] + x[,"n_45"])*(log(1 - r_mc^2) + log(3 - 4*r_mc + r_mc^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(log(pmax(1e-6,r_mc)) + log(3 - r_mc^2 + r_mc^3)) + x[,"n_23"]*log(57 + 8*r_mc - 26*r_mc^2 + 36*r_mc^3 - 18*r_mc^4) + (x[,"n_14"] + x[,"n_32"])*log(9 + 18*r_mc - 7*r_mc^2 + 12*r_mc^3 - 8*r_mc^4) + (x[,"n_12"] + x[,"n_34"])*log(24 - 8*r_mc - 19*r_mc^2 + 20*r_mc^3 - 8*r_mc^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log(9 + 8*r_mc - 14*r_mc^2 + 12*r_mc^3 - 6*r_mc^4) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*log(3 - 4*r_mc + 3*r_mc^2 - 3*r_mc^3 + r_mc^4) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(33 - 8*r_mc + 20*r_mc^2 - 24*r_mc^3 + 12*r_mc^4)
logL_mm <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n11 - n12 - 3*n13 - n14 - 4*n15 - 3*n21 - 3*n22 - 2*n23 - 3*n24 - 3*n25 - 4*n31 - n32 - 3*n33 - n34 - 4*n35 - 4*n41 - 4*n42 - 3*n43 - 4*n44 - 4*n45)*log(2) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n01 + n05 + n41 + n45)*(log(2 - r) + log(pmax(1e-6,r)) + log(1 - r^2)) + (n13 + n22 + n24 + n33)*log(21 + 4*r - 16*r^2 + 24*r^3 - 12*r^4) + (n02 + n04 + n11 + n15 + n31 + n35 + n42 + n44)*log(3 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n12 + n14 + n32 + n34)*log(3 - r + 3*r^2 - 4*r^3 + 2*r^4) + (n03 + n21 + n25 + n43)*log(5 - 6*r + 9*r^2 - 6*r^3 + 3*r^4) + n23*log(18 + 2*r + 7*r^2 - 18*r^3 + 9*r^4)
return(L)}
interlogL_mm <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_mm,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm <- mapply(interlogL_mm,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_mm <- -((x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(-4*log10(2) + 2*log10(3))) - (x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(-2*log10(2) + log10(5) + log10(17)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*(-3*log10(2) + log10(23)) - (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(-4*log10(2) + log10(59)) - x[,"n_23"]*(-4*log10(2) + log10(5) + log10(61)) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(log10(2 - r_mm) + log10(pmax(1e-6,r_mm)) + log10(1 - r_mm^2)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log10(21 + 4*r_mm - 16*r_mm^2 + 24*r_mm^3 - 12*r_mm^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*log10(3 + 4*r_mm - 8*r_mm^2 + 8*r_mm^3 - 4*r_mm^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log10(3 - r_mm + 3*r_mm^2 - 4*r_mm^3 + 2*r_mm^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log10(5 - 6*r_mm + 9*r_mm^2 - 6*r_mm^3 + 3*r_mm^4) + x[,"n_23"]*log10(18 + 2*r_mm + 7*r_mm^2 - 18*r_mm^3 + 9*r_mm^4)
logL_mm <- (-4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_11"] - x[,"n_12"] - 3*x[,"n_13"] - x[,"n_14"] - 4*x[,"n_15"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_31"] - x[,"n_32"] - 3*x[,"n_33"] - x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(log(2 - r_mm) + log(pmax(1e-6,r_mm)) + log(1 - r_mm^2)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(21 + 4*r_mm - 16*r_mm^2 + 24*r_mm^3 - 12*r_mm^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*log(3 + 4*r_mm - 8*r_mm^2 + 8*r_mm^3 - 4*r_mm^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log(3 - r_mm + 3*r_mm^2 - 4*r_mm^3 + 2*r_mm^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log(5 - 6*r_mm + 9*r_mm^2 - 6*r_mm^3 + 3*r_mm^4) + x[,"n_23"]*log(18 + 2*r_mm + 7*r_mm^2 - 18*r_mm^3 + 9*r_mm^4)
logL_mr <- function(r,n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n23,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n13,n22,n24,n33) {
L <- (-2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n11 - 2*n12 - 2*n14 - 2*n15 - 2*n21 - n23 - 2*n25 - 2*n31 - 2*n32 - 2*n34 - 2*n35 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(2) + (n13 + n22 + n24 + n33)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n41)*(2*log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n01 + n45)*(2*log(pmax(1e-6,r)) + log(1 - r^2)) + (n02 + n11 + n35 + n44)*(log(pmax(1e-6,r)) + log(2 + r - 4*r^2 + 4*r^3)) + n23*log(9 + 8*r - 26*r^2 + 36*r^3 - 18*r^4) + (n12 + n34)*log(3 + 8*r - 13*r^2 + 24*r^3 - 16*r^4) + (n14 + n32)*log(6 + 10*r - 37*r^2 + 40*r^3 - 16*r^4) + (n03 + n21 + n25 + n43)*log(1 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n13 + n22 + n24 + n33)*log(1 - r + 3*r^2 - 4*r^3 + 2*r^4) + (n04 + n15 + n31 + n42)*log(3 - 8*r + 13*r^2 - 12*r^3 + 4*r^4)
return(L)}
interlogL_mr <- function(n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n23,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n13,n22,n24,n33) {
optimize(logL_mr,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n14,n15,n21,n23,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n13,n22,n24,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mr <- mapply(interlogL_mr,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_23"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_13"],x[,"n_22"],x[,"n_24"],x[,"n_33"])
LOD_mr <- -((x[,"n_05"] + x[,"n_41"])*(-4*log10(2) + log10(3))) - (x[,"n_01"] + x[,"n_45"])*(-4*log10(2) + log10(3)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(3*log10(2) - log10(7)) - (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(-3*log10(2) + log10(13)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(23)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(23)) - x[,"n_23"]*(-3*log10(2) + log10(79)) + (x[,"n_05"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(2 - r_mr) + log10(pmax(1e-6,r_mr))) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_mr)) + log10(1 - r_mr^2)) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(log10(pmax(1e-6,r_mr)) + log10(2 + r_mr - 4*r_mr^2 + 4*r_mr^3)) + x[,"n_23"]*log10(9 + 8*r_mr - 26*r_mr^2 + 36*r_mr^3 - 18*r_mr^4) + (x[,"n_12"] + x[,"n_34"])*log10(3 + 8*r_mr - 13*r_mr^2 + 24*r_mr^3 - 16*r_mr^4) + (x[,"n_14"] + x[,"n_32"])*log10(6 + 10*r_mr - 37*r_mr^2 + 40*r_mr^3 - 16*r_mr^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log10(1 + 4*r_mr - 10*r_mr^2 + 12*r_mr^3 - 6*r_mr^4) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log10(1 - r_mr + 3*r_mr^2 - 4*r_mr^3 + 2*r_mr^4) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*log10(3 - 8*r_mr + 13*r_mr^2 - 12*r_mr^3 + 4*r_mr^4)
logL_mr <- (-2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - x[,"n_23"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_mr)) + log(2 - r_mr) + log(pmax(1e-6,r_mr))) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,r_mr)) + log(1 - r_mr^2)) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(log(pmax(1e-6,r_mr)) + log(2 + r_mr - 4*r_mr^2 + 4*r_mr^3)) + x[,"n_23"]*log(9 + 8*r_mr - 26*r_mr^2 + 36*r_mr^3 - 18*r_mr^4) + (x[,"n_12"] + x[,"n_34"])*log(3 + 8*r_mr - 13*r_mr^2 + 24*r_mr^3 - 16*r_mr^4) + (x[,"n_14"] + x[,"n_32"])*log(6 + 10*r_mr - 37*r_mr^2 + 40*r_mr^3 - 16*r_mr^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log(1 + 4*r_mr - 10*r_mr^2 + 12*r_mr^3 - 6*r_mr^4) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(1 - r_mr + 3*r_mr^2 - 4*r_mr^3 + 2*r_mr^4) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*log(3 - 8*r_mr + 13*r_mr^2 - 12*r_mr^3 + 4*r_mr^4)
logL_rc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n23) {
L <- (-2*n01 - 2*n02 - n03 - 2*n04 - 2*n05 - 2*n11 - n12 - n13 - n14 - 2*n15 - n21 - n22 - n24 - n25 - 2*n31 - n32 - n33 - n34 - 2*n35 - 2*n41 - 2*n42 - n43 - 2*n44 - 2*n45)*log(2) + (-2*n01 - 2*n02 - n03 - 2*n04 - 2*n05 - 2*n11 - 2*n12 - n13 - 2*n14 - 2*n15 - n21 - n22 - n23 - n24 - n25 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - 2*n41 - 2*n42 - n43 - 2*n44 - 2*n45)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*log(5) + (n01 + n05 + n41 + n45)*(log(pmax(1e-6,r)) + log(2 + r) + log(3 - 4*r + r^2)) + (n13 + n22 + n24 + n33)*log(15 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n02 + n04 + n11 + n15 + n31 + n35 + n42 + n44)*log(9 - 4*r^2 + 8*r^3 - 4*r^4) + (n03 + n21 + n25 + n43)*log(3 - 2*r + 3*r^2 - 2*r^3 + r^4) + n23*log(15 - 2*r + 5*r^2 - 6*r^3 + 3*r^4) + (n12 + n14 + n32 + n34)*log(27 - 6*r + 14*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n23) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_23"])
LOD_rc <- -((x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(-4*log10(2) + 2*log10(5))) - (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(-2*log10(2) + 2*log10(3) + log10(7)) - (x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*(log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(-4*log10(2) + log10(41)) - x[,"n_23"]*(-4*log10(2) + log10(5) + log10(47)) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(log10(pmax(1e-6,r_rc)) + log10(2 + r_rc) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log10(15 + 4*r_rc - 8*r_rc^2 + 8*r_rc^3 - 4*r_rc^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*log10(9 - 4*r_rc^2 + 8*r_rc^3 - 4*r_rc^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log10(3 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3 + r_rc^4) + x[,"n_23"]*log10(15 - 2*r_rc + 5*r_rc^2 - 6*r_rc^3 + 3*r_rc^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log10(27 - 6*r_rc + 14*r_rc^2 - 16*r_rc^3 + 8*r_rc^4)
logL_rc <- (-2*x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - 2*x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_24"] - x[,"n_25"] - 2*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (-2*x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(log(pmax(1e-6,r_rc)) + log(2 + r_rc) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(15 + 4*r_rc - 8*r_rc^2 + 8*r_rc^3 - 4*r_rc^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"] + x[,"n_42"] + x[,"n_44"])*log(9 - 4*r_rc^2 + 8*r_rc^3 - 4*r_rc^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log(3 - 2*r_rc + 3*r_rc^2 - 2*r_rc^3 + r_rc^4) + x[,"n_23"]*log(15 - 2*r_rc + 5*r_rc^2 - 6*r_rc^3 + 3*r_rc^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log(27 - 6*r_rc + 14*r_rc^2 - 16*r_rc^3 + 8*r_rc^4)
logL_rm <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
L <- (-3*n01 - n02 - 3*n03 - n04 - 3*n05 - n11 - 2*n12 - 2*n13 - 2*n14 - n15 - 3*n21 - 2*n22 - 2*n23 - 2*n24 - 3*n25 - n31 - 2*n32 - 2*n33 - 2*n34 - n35 - 3*n41 - n42 - 3*n43 - n44 - 3*n45)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n01 + n45)*(log(2 - r) + 2*log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n41)*(log(1 - r^2) + log(3 - 4*r + r^2)) + (n02 + n11 + n35 + n44)*(log(pmax(1e-6,r)) + log(3 - r^2 + r^3)) + n23*log(57 + 8*r - 26*r^2 + 36*r^3 - 18*r^4) + (n12 + n34)*log(9 + 18*r - 7*r^2 + 12*r^3 - 8*r^4) + (n14 + n32)*log(24 - 8*r - 19*r^2 + 20*r^3 - 8*r^4) + (n03 + n21 + n25 + n43)*log(9 + 8*r - 14*r^2 + 12*r^3 - 6*r^4) + (n04 + n15 + n31 + n42)*log(3 - 4*r + 3*r^2 - 3*r^3 + r^4) + (n13 + n22 + n24 + n33)*log(33 - 8*r + 20*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_rm <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45) {
optimize(logL_rm,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm <- -((x[,"n_05"] + x[,"n_41"])*(-4*log10(2) + log10(3) + log10(5))) - (x[,"n_01"] + x[,"n_45"])*(-4*log10(2) + log10(3) + log10(5)) - (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(-3*log10(2) + log10(5) + log10(17)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(23)) - (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(-4*log10(2) + log10(23)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(-2*log10(2) + log10(127)) - x[,"n_23"]*(-3*log10(2) + log10(463)) + (x[,"n_01"] + x[,"n_45"])*(log10(2 - r_rm) + 2*log10(pmax(1e-6,r_rm)) + log10(2 + r_rm)) + (x[,"n_05"] + x[,"n_41"])*(log10(1 - r_rm^2) + log10(3 - 4*r_rm + r_rm^2)) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm)) + log10(3 - r_rm^2 + r_rm^3)) + x[,"n_23"]*log10(57 + 8*r_rm - 26*r_rm^2 + 36*r_rm^3 - 18*r_rm^4) + (x[,"n_12"] + x[,"n_34"])*log10(9 + 18*r_rm - 7*r_rm^2 + 12*r_rm^3 - 8*r_rm^4) + (x[,"n_14"] + x[,"n_32"])*log10(24 - 8*r_rm - 19*r_rm^2 + 20*r_rm^3 - 8*r_rm^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log10(9 + 8*r_rm - 14*r_rm^2 + 12*r_rm^3 - 6*r_rm^4) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*log10(3 - 4*r_rm + 3*r_rm^2 - 3*r_rm^3 + r_rm^4) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log10(33 - 8*r_rm + 20*r_rm^2 - 24*r_rm^3 + 12*r_rm^4)
logL_rm <- (-3*x[,"n_01"] - x[,"n_02"] - 3*x[,"n_03"] - x[,"n_04"] - 3*x[,"n_05"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 3*x[,"n_25"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - 3*x[,"n_41"] - x[,"n_42"] - 3*x[,"n_43"] - x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_45"])*(log(2 - r_rm) + 2*log(pmax(1e-6,r_rm)) + log(2 + r_rm)) + (x[,"n_05"] + x[,"n_41"])*(log(1 - r_rm^2) + log(3 - 4*r_rm + r_rm^2)) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(log(pmax(1e-6,r_rm)) + log(3 - r_rm^2 + r_rm^3)) + x[,"n_23"]*log(57 + 8*r_rm - 26*r_rm^2 + 36*r_rm^3 - 18*r_rm^4) + (x[,"n_12"] + x[,"n_34"])*log(9 + 18*r_rm - 7*r_rm^2 + 12*r_rm^3 - 8*r_rm^4) + (x[,"n_14"] + x[,"n_32"])*log(24 - 8*r_rm - 19*r_rm^2 + 20*r_rm^3 - 8*r_rm^4) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*log(9 + 8*r_rm - 14*r_rm^2 + 12*r_rm^3 - 6*r_rm^4) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*log(3 - 4*r_rm + 3*r_rm^2 - 3*r_rm^3 + r_rm^4) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(33 - 8*r_rm + 20*r_rm^2 - 24*r_rm^3 + 12*r_rm^4)
logL_rr <- function(r,n01,n02,n04,n05,n11,n12,n13,n14,n15,n22,n24,n31,n32,n33,n34,n35,n41,n42,n44,n45,n03,n21,n23,n25,n43) {
L <- (-n01 - n02 - n04 - n05 - n11 - n12 + n13 - n14 - n15 + n22 + n24 - n31 - n32 + n33 - n34 - n35 - n41 - n42 - n44 - n45)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n01 + n45)*(3*log(pmax(1e-6,r)) + log(2 + r)) + (n04 + n15 + n31 + n42)*(2*log(pmax(1e-6,1 - r)) + log(3 + 8*r - 4*r^2)) + (n05 + n41)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n03 + n21 + n25 + n43)*(log(pmax(1e-6,r)) + log(4 - r - 6*r^2 + 3*r^3)) + (n12 + n34)*(log(pmax(1e-6,r)) + log(18 + 3*r - 16*r^2 + 16*r^3)) + (n13 + n22 + n24 + n33)*log(3 + 5*r - 11*r^2 + 12*r^3 - 6*r^4) + (n02 + n11 + n35 + n44)*log(7*r^2 - 4*r^4) + (n14 + n32)*log(21 - 40*r + 51*r^2 - 48*r^3 + 16*r^4) + n23*log(21 - 28*r + 46*r^2 - 36*r^3 + 18*r^4)
return(L)}
interlogL_rr <- function(n01,n02,n04,n05,n11,n12,n13,n14,n15,n22,n24,n31,n32,n33,n34,n35,n41,n42,n44,n45,n03,n21,n23,n25,n43) {
optimize(logL_rr,c(0,0.5), n01,n02,n04,n05,n11,n12,n13,n14,n15,n22,n24,n31,n32,n33,n34,n35,n41,n42,n44,n45,n03,n21,n23,n25,n43, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_01"],x[,"n_02"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_22"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_44"],x[,"n_45"],x[,"n_03"],x[,"n_21"],x[,"n_23"],x[,"n_25"],x[,"n_43"])
LOD_rr <- -((x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(-log10(2) + log10(3))) - (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*(-log10(2) + log10(3)) - (x[,"n_05"] + x[,"n_41"])*(-4*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_45"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(5) + log10(7)) - x[,"n_23"]*(-3*log10(2) + 2*log10(11)) - (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(-4*log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*(-3*log10(2) + log10(31)) + (x[,"n_01"] + x[,"n_45"])*(3*log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 + 8*r_rr - 4*r_rr^2)) + (x[,"n_05"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 - 4*r_rr + r_rr^2)) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(4 - r_rr - 6*r_rr^2 + 3*r_rr^3)) + (x[,"n_12"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(18 + 3*r_rr - 16*r_rr^2 + 16*r_rr^3)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log10(3 + 5*r_rr - 11*r_rr^2 + 12*r_rr^3 - 6*r_rr^4) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*log10(7*r_rr^2 - 4*r_rr^4) + (x[,"n_14"] + x[,"n_32"])*log10(21 - 40*r_rr + 51*r_rr^2 - 48*r_rr^3 + 16*r_rr^4) + x[,"n_23"]*log10(21 - 28*r_rr + 46*r_rr^2 - 36*r_rr^3 + 18*r_rr^4)
logL_rr <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] + x[,"n_13"] - x[,"n_14"] - x[,"n_15"] + x[,"n_22"] + x[,"n_24"] - x[,"n_31"] - x[,"n_32"] + x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_44"] - x[,"n_45"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_45"])*(3*log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_31"] + x[,"n_42"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 + 8*r_rr - 4*r_rr^2)) + (x[,"n_05"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 - 4*r_rr + r_rr^2)) + (x[,"n_03"] + x[,"n_21"] + x[,"n_25"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(4 - r_rr - 6*r_rr^2 + 3*r_rr^3)) + (x[,"n_12"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(18 + 3*r_rr - 16*r_rr^2 + 16*r_rr^3)) + (x[,"n_13"] + x[,"n_22"] + x[,"n_24"] + x[,"n_33"])*log(3 + 5*r_rr - 11*r_rr^2 + 12*r_rr^3 - 6*r_rr^4) + (x[,"n_02"] + x[,"n_11"] + x[,"n_35"] + x[,"n_44"])*log(7*r_rr^2 - 4*r_rr^4) + (x[,"n_14"] + x[,"n_32"])*log(21 - 40*r_rr + 51*r_rr^2 - 48*r_rr^3 + 16*r_rr^4) + x[,"n_23"]*log(21 - 28*r_rr + 46*r_rr^2 - 36*r_rr^3 + 18*r_rr^4)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_mc,r_mm,r_mr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_mc,LOD_mm,LOD_mr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_mc,logL_mm,logL_mr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","mixed coupling","mixed mixed","mixed repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.2_3.2 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35) {
L <- (-2*n00 - 2*n01 - n02 - n03 - 2*n04 - 2*n05 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - n20 - n21 - n22 - n23 - n24 - n25 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n40 - 2*n41 - n42 - n43 - 2*n44 - 2*n45)*log(2) + (n01 + n04 + n11 + n14 + n20 + n21 + n24 + n25 + n31 + n34 + n41 + n44)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + 4*(n00 + n45)*log(pmax(1e-6,1 - r)) + 4*(n05 + n40)*log(pmax(1e-6,r)) + (n10 + n35)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n20 + n25)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n15 + n30)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n04 + n41)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n01 + n44)*(3*log(pmax(1e-6,1 - r)) + log(1 + r)) + (n02 + n43)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(5 + r)) + (n03 + n42)*(2*log(pmax(1e-6,r)) + log(6 - 7*r + r^2)) + (n11 + n34)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r^2)) + (n14 + n31)*(2*log(pmax(1e-6,r)) + log(7 - 8*r + 4*r^2)) + (n13 + n32)*(log(pmax(1e-6,r)) + log(32 - 45*r + 36*r^2 - 8*r^3)) + (n21 + n24)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n12 + n33)*log(15 - 18*r + 15*r^2 - 4*r^3 - 8*r^4) + (n22 + n23)*log(11 - 12*r + 18*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n11,n12,n13,n14,n20,n21,n22,n23,n24,n25,n31,n32,n33,n34,n40,n41,n42,n43,n44,n45,n10,n15,n30,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_10"],x[,"n_15"],x[,"n_30"],x[,"n_35"])
LOD_cc <- 4*(x[,"n_20"] + x[,"n_25"])*log10(2) + 4*(x[,"n_15"] + x[,"n_30"])*log10(2) + 4*(x[,"n_10"] + x[,"n_35"])*log10(2) + 4*(x[,"n_05"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_33"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_03"] + x[,"n_42"])*(-4*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_43"])*(-4*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-4*log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_23"])*(-3*log10(2) + log10(67)) + 4*(x[,"n_00"] + x[,"n_45"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_35"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_25"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_41"])*(log10(2 - r_cc) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_44"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(1 + r_cc)) + (x[,"n_02"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc)) + log10(5 + r_cc)) + (x[,"n_03"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(6 - 7*r_cc + r_cc^2)) + (x[,"n_11"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 4*r_cc^2)) + (x[,"n_14"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(7 - 8*r_cc + 4*r_cc^2)) + (x[,"n_13"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(32 - 45*r_cc + 36*r_cc^2 - 8*r_cc^3)) + (x[,"n_21"] + x[,"n_24"])*(log10(pmax(1e-6,r_cc)) + log10(4 - 7*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_12"] + x[,"n_33"])*log10(15 - 18*r_cc + 15*r_cc^2 - 4*r_cc^3 - 8*r_cc^4) + (x[,"n_22"] + x[,"n_23"])*log10(11 - 12*r_cc + 18*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cc <- (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_40"] - 2*x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_11"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_34"] + x[,"n_41"] + x[,"n_44"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + 4*(x[,"n_00"] + x[,"n_45"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_35"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_25"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_41"])*(log(2 - r_cc) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_44"])*(3*log(pmax(1e-6,1 - r_cc)) + log(1 + r_cc)) + (x[,"n_02"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc)) + log(5 + r_cc)) + (x[,"n_03"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(6 - 7*r_cc + r_cc^2)) + (x[,"n_11"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 4*r_cc^2)) + (x[,"n_14"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(7 - 8*r_cc + 4*r_cc^2)) + (x[,"n_13"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(32 - 45*r_cc + 36*r_cc^2 - 8*r_cc^3)) + (x[,"n_21"] + x[,"n_24"])*(log(pmax(1e-6,r_cc)) + log(4 - 7*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_12"] + x[,"n_33"])*log(15 - 18*r_cc + 15*r_cc^2 - 4*r_cc^3 - 8*r_cc^4) + (x[,"n_22"] + x[,"n_23"])*log(11 - 12*r_cc + 18*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 3*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 3*n25 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n40)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n00 + n45)*(2*log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n04 + n41)*(2*log(pmax(1e-6,r)) + log(11 - 8*r + 3*r^2)) + (n01 + n44)*(2*log(pmax(1e-6,1 - r)) + log(6 + 2*r + 3*r^2)) + (n10 + n35)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r^2)) + (n15 + n30)*(2*log(pmax(1e-6,r)) + log(7 - 8*r + 4*r^2)) + (n14 + n31)*(log(pmax(1e-6,r)) + log(30 - 25*r + 28*r^2 - 12*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n03 + n42)*(log(pmax(1e-6,r)) + log(18 - 17*r + 10*r^2 - 2*r^3)) + (n11 + n34)*log(21 - 16*r - 13*r^2 + 20*r^3 - 12*r^4) + (n22 + n23)*log(67 - 4*r - 8*r^2 + 24*r^3 - 12*r^4) + (n02 + n43)*log(9 - 6*r + r^2 - 2*r^3 - 2*r^4) + (n13 + n32)*log(27 + 8*r + 30*r^2 - 28*r^3 + 8*r^4) + (n12 + n33)*log(45 - 16*r - 6*r^2 - 4*r^3 + 8*r^4) + (n21 + n24)*log(21 - 4*r + 22*r^2 - 36*r^3 + 18*r^4)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cm <- -((x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-4*log10(2) + log10(13)) - (x[,"n_21"] + x[,"n_24"])*(-3*log10(2) + 2*log10(13)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(23)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(23)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + 2*log10(3) + log10(29)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(31)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(31)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(71)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(71)) + (x[,"n_05"] + x[,"n_40"])*(log10(2 - r_cm) + 3*log10(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(1 - r_cm^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm)) + log10(11 - 8*r_cm + 3*r_cm^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(6 + 2*r_cm + 3*r_cm^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(3 + 4*r_cm^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_cm)) + log10(7 - 8*r_cm + 4*r_cm^2)) + (x[,"n_14"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(30 - 25*r_cm + 28*r_cm^2 - 12*r_cm^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_cm)) + log10(4 - 7*r_cm + 6*r_cm^2 - 3*r_cm^3)) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm)) + log10(18 - 17*r_cm + 10*r_cm^2 - 2*r_cm^3)) + (x[,"n_11"] + x[,"n_34"])*log10(21 - 16*r_cm - 13*r_cm^2 + 20*r_cm^3 - 12*r_cm^4) + (x[,"n_22"] + x[,"n_23"])*log10(67 - 4*r_cm - 8*r_cm^2 + 24*r_cm^3 - 12*r_cm^4) + (x[,"n_02"] + x[,"n_43"])*log10(9 - 6*r_cm + r_cm^2 - 2*r_cm^3 - 2*r_cm^4) + (x[,"n_13"] + x[,"n_32"])*log10(27 + 8*r_cm + 30*r_cm^2 - 28*r_cm^3 + 8*r_cm^4) + (x[,"n_12"] + x[,"n_33"])*log10(45 - 16*r_cm - 6*r_cm^2 - 4*r_cm^3 + 8*r_cm^4) + (x[,"n_21"] + x[,"n_24"])*log10(21 - 4*r_cm + 22*r_cm^2 - 36*r_cm^3 + 18*r_cm^4)
logL_cm <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 3*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_40"])*(log(2 - r_cm) + 3*log(pmax(1e-6,r_cm))) + (x[,"n_00"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm)) + log(1 - r_cm^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm)) + log(11 - 8*r_cm + 3*r_cm^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cm)) + log(6 + 2*r_cm + 3*r_cm^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cm)) + log(3 + 4*r_cm^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,r_cm)) + log(7 - 8*r_cm + 4*r_cm^2)) + (x[,"n_14"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(30 - 25*r_cm + 28*r_cm^2 - 12*r_cm^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_cm)) + log(4 - 7*r_cm + 6*r_cm^2 - 3*r_cm^3)) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,r_cm)) + log(18 - 17*r_cm + 10*r_cm^2 - 2*r_cm^3)) + (x[,"n_11"] + x[,"n_34"])*log(21 - 16*r_cm - 13*r_cm^2 + 20*r_cm^3 - 12*r_cm^4) + (x[,"n_22"] + x[,"n_23"])*log(67 - 4*r_cm - 8*r_cm^2 + 24*r_cm^3 - 12*r_cm^4) + (x[,"n_02"] + x[,"n_43"])*log(9 - 6*r_cm + r_cm^2 - 2*r_cm^3 - 2*r_cm^4) + (x[,"n_13"] + x[,"n_32"])*log(27 + 8*r_cm + 30*r_cm^2 - 28*r_cm^3 + 8*r_cm^4) + (x[,"n_12"] + x[,"n_33"])*log(45 - 16*r_cm - 6*r_cm^2 - 4*r_cm^3 + 8*r_cm^4) + (x[,"n_21"] + x[,"n_24"])*log(21 - 4*r_cm + 22*r_cm^2 - 36*r_cm^3 + 18*r_cm^4)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n20 - 2*n21 - 3*n22 - 3*n23 - 2*n24 - 3*n25 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n00 + n45)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n40)*(2*log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n15 + n30)*(log(pmax(1e-6,r)) + log(6 - 11*r + 12*r^2 - 4*r^3)) + (n04 + n41)*(log(pmax(1e-6,r)) + log(12 - 16*r + 10*r^2 - 3*r^3)) + (n21 + n24)*log(15 + 4*r - 13*r^2 + 18*r^3 - 9*r^4) + (n12 + n33)*log(57 - 4*r - 12*r^2 + 12*r^3 - 8*r^4) + (n13 + n32)*log(45 + 24*r - 24*r^2 + 20*r^3 - 8*r^4) + (n10 + n35)*log(3 - 4*r + r^2 + 4*r^3 - 4*r^4) + (n01 + n44)*log(3 + 2*r - 4*r^2 + 2*r^3 - 3*r^4) + (n03 + n42)*log(9 - 6*r + 13*r^2 - 6*r^3 + 2*r^4) + (n02 + n43)*log(12 - 10*r + 7*r^2 - 2*r^3 + 2*r^4) + (n20 + n25)*log(3 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + (n14 + n31)*log(18 - 12*r + 35*r^2 - 32*r^3 + 12*r^4) + (n22 + n23)*log(99 - 4*r + 16*r^2 - 24*r^3 + 12*r^4) + (n11 + n34)*log(21 - 10*r + 11*r^2 - 16*r^3 + 12*r^4)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_cr <- -((x[,"n_15"] + x[,"n_30"])*(-log10(2) + log10(3))) - (x[,"n_10"] + x[,"n_35"])*(-log10(2) + log10(3)) - (x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_20"] + x[,"n_25"])*(-3*log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_24"])*(-4*log10(2) + log10(13) + log10(19)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_13"] + x[,"n_32"])*log10(53) - (x[,"n_12"] + x[,"n_33"])*log10(53) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(5) + log10(79)) + (x[,"n_00"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr)) + log10(2 + r_cr)) + (x[,"n_05"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_cr)) + log10(3 - 4*r_cr + r_cr^2)) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,r_cr)) + log10(6 - 11*r_cr + 12*r_cr^2 - 4*r_cr^3)) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_cr)) + log10(12 - 16*r_cr + 10*r_cr^2 - 3*r_cr^3)) + (x[,"n_21"] + x[,"n_24"])*log10(15 + 4*r_cr - 13*r_cr^2 + 18*r_cr^3 - 9*r_cr^4) + (x[,"n_12"] + x[,"n_33"])*log10(57 - 4*r_cr - 12*r_cr^2 + 12*r_cr^3 - 8*r_cr^4) + (x[,"n_13"] + x[,"n_32"])*log10(45 + 24*r_cr - 24*r_cr^2 + 20*r_cr^3 - 8*r_cr^4) + (x[,"n_10"] + x[,"n_35"])*log10(3 - 4*r_cr + r_cr^2 + 4*r_cr^3 - 4*r_cr^4) + (x[,"n_01"] + x[,"n_44"])*log10(3 + 2*r_cr - 4*r_cr^2 + 2*r_cr^3 - 3*r_cr^4) + (x[,"n_03"] + x[,"n_42"])*log10(9 - 6*r_cr + 13*r_cr^2 - 6*r_cr^3 + 2*r_cr^4) + (x[,"n_02"] + x[,"n_43"])*log10(12 - 10*r_cr + 7*r_cr^2 - 2*r_cr^3 + 2*r_cr^4) + (x[,"n_20"] + x[,"n_25"])*log10(3 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + (x[,"n_14"] + x[,"n_31"])*log10(18 - 12*r_cr + 35*r_cr^2 - 32*r_cr^3 + 12*r_cr^4) + (x[,"n_22"] + x[,"n_23"])*log10(99 - 4*r_cr + 16*r_cr^2 - 24*r_cr^3 + 12*r_cr^4) + (x[,"n_11"] + x[,"n_34"])*log10(21 - 10*r_cr + 11*r_cr^2 - 16*r_cr^3 + 12*r_cr^4)
logL_cr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr)) + log(2 + r_cr)) + (x[,"n_05"] + x[,"n_40"])*(2*log(pmax(1e-6,r_cr)) + log(3 - 4*r_cr + r_cr^2)) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,r_cr)) + log(6 - 11*r_cr + 12*r_cr^2 - 4*r_cr^3)) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_cr)) + log(12 - 16*r_cr + 10*r_cr^2 - 3*r_cr^3)) + (x[,"n_21"] + x[,"n_24"])*log(15 + 4*r_cr - 13*r_cr^2 + 18*r_cr^3 - 9*r_cr^4) + (x[,"n_12"] + x[,"n_33"])*log(57 - 4*r_cr - 12*r_cr^2 + 12*r_cr^3 - 8*r_cr^4) + (x[,"n_13"] + x[,"n_32"])*log(45 + 24*r_cr - 24*r_cr^2 + 20*r_cr^3 - 8*r_cr^4) + (x[,"n_10"] + x[,"n_35"])*log(3 - 4*r_cr + r_cr^2 + 4*r_cr^3 - 4*r_cr^4) + (x[,"n_01"] + x[,"n_44"])*log(3 + 2*r_cr - 4*r_cr^2 + 2*r_cr^3 - 3*r_cr^4) + (x[,"n_03"] + x[,"n_42"])*log(9 - 6*r_cr + 13*r_cr^2 - 6*r_cr^3 + 2*r_cr^4) + (x[,"n_02"] + x[,"n_43"])*log(12 - 10*r_cr + 7*r_cr^2 - 2*r_cr^3 + 2*r_cr^4) + (x[,"n_20"] + x[,"n_25"])*log(3 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4) + (x[,"n_14"] + x[,"n_31"])*log(18 - 12*r_cr + 35*r_cr^2 - 32*r_cr^3 + 12*r_cr^4) + (x[,"n_22"] + x[,"n_23"])*log(99 - 4*r_cr + 16*r_cr^2 - 24*r_cr^3 + 12*r_cr^4) + (x[,"n_11"] + x[,"n_34"])*log(21 - 10*r_cr + 11*r_cr^2 - 16*r_cr^3 + 12*r_cr^4)
logL_mc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n21,n22,n23,n24,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n20,n25) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n00 + n45)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n40)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n04 + n41)*(2*log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n01 + n44)*(2*log(pmax(1e-6,1 - r)) + log(3 - r + 3*r^2)) + (n15 + n30)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n10 + n35)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n14 + n31)*(log(pmax(1e-6,r)) + log(24 - 29*r + 44*r^2 - 24*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n03 + n42)*(log(pmax(1e-6,r)) + log(7 - 8*r + 6*r^2 - 2*r^3)) + (n11 + n34)*log(15 - 2*r - 41*r^2 + 52*r^3 - 24*r^4) + (n22 + n23)*log(17 - 8*r^2 + 16*r^3 - 8*r^4) + (n02 + n43)*log(3 - r - 2*r^2 + 2*r^3 - 2*r^4) + (n13 + n32)*log(21 + 2*r + 34*r^2 - 40*r^3 + 16*r^4) + (n12 + n33)*log(33 - 14*r + 10*r^2 - 24*r^3 + 16*r^4) + (n21 + n24)*log(15 - 8*r + 44*r^2 - 72*r^3 + 36*r^4)
return(L)}
interlogL_mc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n21,n22,n23,n24,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n20,n25) {
optimize(logL_mc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n21,n22,n23,n24,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n20,n25, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_20"],x[,"n_25"])
LOD_mc <- 4*(x[,"n_05"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_45"])*log10(2) - (x[,"n_15"] + x[,"n_30"])*(-2*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_35"])*(-2*log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_31"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_11"] + x[,"n_34"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_22"] + x[,"n_23"])*(-log10(2) + log10(3) + log10(11)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(17)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(17)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(53)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(53)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(61)) + (x[,"n_00"] + x[,"n_45"])*(3*log10(pmax(1e-6,1 - r_mc)) + log10(pmax(1e-6,r_mc))) + (x[,"n_05"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_mc)) + 3*log10(pmax(1e-6,r_mc))) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_mc)) + log10(5 - 5*r_mc + 3*r_mc^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 - r_mc + 3*r_mc^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_mc)) + log10(7 - 12*r_mc + 8*r_mc^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 - 4*r_mc + 8*r_mc^2)) + (x[,"n_14"] + x[,"n_31"])*(log10(pmax(1e-6,r_mc)) + log10(24 - 29*r_mc + 44*r_mc^2 - 24*r_mc^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_mc)) + log10(2 - 5*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_03"] + x[,"n_42"])*(log10(pmax(1e-6,r_mc)) + log10(7 - 8*r_mc + 6*r_mc^2 - 2*r_mc^3)) + (x[,"n_11"] + x[,"n_34"])*log10(15 - 2*r_mc - 41*r_mc^2 + 52*r_mc^3 - 24*r_mc^4) + (x[,"n_22"] + x[,"n_23"])*log10(17 - 8*r_mc^2 + 16*r_mc^3 - 8*r_mc^4) + (x[,"n_02"] + x[,"n_43"])*log10(3 - r_mc - 2*r_mc^2 + 2*r_mc^3 - 2*r_mc^4) + (x[,"n_13"] + x[,"n_32"])*log10(21 + 2*r_mc + 34*r_mc^2 - 40*r_mc^3 + 16*r_mc^4) + (x[,"n_12"] + x[,"n_33"])*log10(33 - 14*r_mc + 10*r_mc^2 - 24*r_mc^3 + 16*r_mc^4) + (x[,"n_21"] + x[,"n_24"])*log10(15 - 8*r_mc + 44*r_mc^2 - 72*r_mc^3 + 36*r_mc^4)
logL_mc <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_00"] + x[,"n_45"])*(3*log(pmax(1e-6,1 - r_mc)) + log(pmax(1e-6,r_mc))) + (x[,"n_05"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_mc)) + 3*log(pmax(1e-6,r_mc))) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,r_mc)) + log(5 - 5*r_mc + 3*r_mc^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 - r_mc + 3*r_mc^2)) + (x[,"n_15"] + x[,"n_30"])*(2*log(pmax(1e-6,r_mc)) + log(7 - 12*r_mc + 8*r_mc^2)) + (x[,"n_10"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 - 4*r_mc + 8*r_mc^2)) + (x[,"n_14"] + x[,"n_31"])*(log(pmax(1e-6,r_mc)) + log(24 - 29*r_mc + 44*r_mc^2 - 24*r_mc^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_mc)) + log(2 - 5*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_03"] + x[,"n_42"])*(log(pmax(1e-6,r_mc)) + log(7 - 8*r_mc + 6*r_mc^2 - 2*r_mc^3)) + (x[,"n_11"] + x[,"n_34"])*log(15 - 2*r_mc - 41*r_mc^2 + 52*r_mc^3 - 24*r_mc^4) + (x[,"n_22"] + x[,"n_23"])*log(17 - 8*r_mc^2 + 16*r_mc^3 - 8*r_mc^4) + (x[,"n_02"] + x[,"n_43"])*log(3 - r_mc - 2*r_mc^2 + 2*r_mc^3 - 2*r_mc^4) + (x[,"n_13"] + x[,"n_32"])*log(21 + 2*r_mc + 34*r_mc^2 - 40*r_mc^3 + 16*r_mc^4) + (x[,"n_12"] + x[,"n_33"])*log(33 - 14*r_mc + 10*r_mc^2 - 24*r_mc^3 + 16*r_mc^4) + (x[,"n_21"] + x[,"n_24"])*log(15 - 8*r_mc + 44*r_mc^2 - 72*r_mc^3 + 36*r_mc^4)
logL_mm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 4*n10 - 4*n11 - 3*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n40)*(log(pmax(1e-6,1 - r)) + log(2 - r) + 2*log(pmax(1e-6,r))) + (n00 + n45)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 - r^2)) + (n15 + n30)*(log(pmax(1e-6,r)) + log(12 - 21*r + 20*r^2 - 8*r^3)) + (n04 + n41)*(log(pmax(1e-6,r)) + log(9 - 10*r + 7*r^2 - 3*r^3)) + (n21 + n24)*log(69 - 8*r - 28*r^2 + 72*r^3 - 36*r^4) + (n10 + n35)*log(3 + 2*r - 9*r^2 + 12*r^3 - 8*r^4) + (n13 + n32)*log(51 + 4*r - 2*r^2 + 12*r^3 - 8*r^4) + (n12 + n33)*log(57 - 4*r - 14*r^2 + 20*r^3 - 8*r^4) + (n01 + n44)*log(3 + 2*r - 7*r^2 + 5*r^3 - 3*r^4) + (n02 + n43)*log(12 - 10*r + 11*r^2 - 6*r^3 + 2*r^4) + (n03 + n42)*log(9 - 2*r + 5*r^2 - 2*r^3 + 2*r^4) + (n22 + n23)*log(31 + 4*r - 8*r^3 + 4*r^4) + (n20 + n25)*log(9 - 16*r + 28*r^2 - 24*r^3 + 12*r^4) + (n14 + n31)*log(27 + 6*r + 37*r^2 - 52*r^3 + 24*r^4) + (n11 + n34)*log(42 - 20*r + 25*r^2 - 44*r^3 + 24*r^4)
return(L)}
interlogL_mm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_mm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm <- mapply(interlogL_mm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_mm <- -((x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_30"])*(-2*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_35"])*(-2*log10(2) + log10(11)) - (x[,"n_20"] + x[,"n_25"])*(-2*log10(2) + log10(23)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(7) + log10(37)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(43)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(43)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(3) + log10(43)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(73)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(73)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(107)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(107)) - (x[,"n_14"] + x[,"n_31"])*(-2*log10(2) + log10(137)) - (x[,"n_11"] + x[,"n_34"])*(-2*log10(2) + log10(137)) + (x[,"n_05"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_mm)) + log10(2 - r_mm) + 2*log10(pmax(1e-6,r_mm))) + (x[,"n_00"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_mm)) + log10(pmax(1e-6,r_mm)) + log10(1 - r_mm^2)) + (x[,"n_15"] + x[,"n_30"])*(log10(pmax(1e-6,r_mm)) + log10(12 - 21*r_mm + 20*r_mm^2 - 8*r_mm^3)) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_mm)) + log10(9 - 10*r_mm + 7*r_mm^2 - 3*r_mm^3)) + (x[,"n_21"] + x[,"n_24"])*log10(69 - 8*r_mm - 28*r_mm^2 + 72*r_mm^3 - 36*r_mm^4) + (x[,"n_10"] + x[,"n_35"])*log10(3 + 2*r_mm - 9*r_mm^2 + 12*r_mm^3 - 8*r_mm^4) + (x[,"n_13"] + x[,"n_32"])*log10(51 + 4*r_mm - 2*r_mm^2 + 12*r_mm^3 - 8*r_mm^4) + (x[,"n_12"] + x[,"n_33"])*log10(57 - 4*r_mm - 14*r_mm^2 + 20*r_mm^3 - 8*r_mm^4) + (x[,"n_01"] + x[,"n_44"])*log10(3 + 2*r_mm - 7*r_mm^2 + 5*r_mm^3 - 3*r_mm^4) + (x[,"n_02"] + x[,"n_43"])*log10(12 - 10*r_mm + 11*r_mm^2 - 6*r_mm^3 + 2*r_mm^4) + (x[,"n_03"] + x[,"n_42"])*log10(9 - 2*r_mm + 5*r_mm^2 - 2*r_mm^3 + 2*r_mm^4) + (x[,"n_22"] + x[,"n_23"])*log10(31 + 4*r_mm - 8*r_mm^3 + 4*r_mm^4) + (x[,"n_20"] + x[,"n_25"])*log10(9 - 16*r_mm + 28*r_mm^2 - 24*r_mm^3 + 12*r_mm^4) + (x[,"n_14"] + x[,"n_31"])*log10(27 + 6*r_mm + 37*r_mm^2 - 52*r_mm^3 + 24*r_mm^4) + (x[,"n_11"] + x[,"n_34"])*log10(42 - 20*r_mm + 25*r_mm^2 - 44*r_mm^3 + 24*r_mm^4)
logL_mm <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_mm)) + log(2 - r_mm) + 2*log(pmax(1e-6,r_mm))) + (x[,"n_00"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_mm)) + log(pmax(1e-6,r_mm)) + log(1 - r_mm^2)) + (x[,"n_15"] + x[,"n_30"])*(log(pmax(1e-6,r_mm)) + log(12 - 21*r_mm + 20*r_mm^2 - 8*r_mm^3)) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_mm)) + log(9 - 10*r_mm + 7*r_mm^2 - 3*r_mm^3)) + (x[,"n_21"] + x[,"n_24"])*log(69 - 8*r_mm - 28*r_mm^2 + 72*r_mm^3 - 36*r_mm^4) + (x[,"n_10"] + x[,"n_35"])*log(3 + 2*r_mm - 9*r_mm^2 + 12*r_mm^3 - 8*r_mm^4) + (x[,"n_13"] + x[,"n_32"])*log(51 + 4*r_mm - 2*r_mm^2 + 12*r_mm^3 - 8*r_mm^4) + (x[,"n_12"] + x[,"n_33"])*log(57 - 4*r_mm - 14*r_mm^2 + 20*r_mm^3 - 8*r_mm^4) + (x[,"n_01"] + x[,"n_44"])*log(3 + 2*r_mm - 7*r_mm^2 + 5*r_mm^3 - 3*r_mm^4) + (x[,"n_02"] + x[,"n_43"])*log(12 - 10*r_mm + 11*r_mm^2 - 6*r_mm^3 + 2*r_mm^4) + (x[,"n_03"] + x[,"n_42"])*log(9 - 2*r_mm + 5*r_mm^2 - 2*r_mm^3 + 2*r_mm^4) + (x[,"n_22"] + x[,"n_23"])*log(31 + 4*r_mm - 8*r_mm^3 + 4*r_mm^4) + (x[,"n_20"] + x[,"n_25"])*log(9 - 16*r_mm + 28*r_mm^2 - 24*r_mm^3 + 12*r_mm^4) + (x[,"n_14"] + x[,"n_31"])*log(27 + 6*r_mm + 37*r_mm^2 - 52*r_mm^3 + 24*r_mm^4) + (x[,"n_11"] + x[,"n_34"])*log(42 - 20*r_mm + 25*r_mm^2 - 44*r_mm^3 + 24*r_mm^4)
logL_mr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 3*n10 - 3*n11 - n12 - n13 - 3*n14 - 3*n15 - 3*n20 - 3*n21 - 3*n24 - 3*n25 - 3*n30 - 3*n31 - n32 - n33 - 3*n34 - 3*n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(2) + (-2*n00 - n01 - 2*n02 - 2*n03 - n04 - 2*n05 - 2*n10 - n11 - 2*n12 - 2*n13 - n14 - 2*n15 - n20 - n21 - n22 - n23 - n24 - n25 - 2*n30 - n31 - 2*n32 - 2*n33 - n34 - 2*n35 - 2*n40 - n41 - 2*n42 - 2*n43 - n44 - 2*n45)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n00 + n45)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n40)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n01 + n44)*(log(pmax(1e-6,r)) + log(3 - r^2 + r^3)) + (n10 + n35)*(log(pmax(1e-6,r)) + log(12 - 7*r - 4*r^2 + 8*r^3)) + (n11 + n34)*log(9 + 18*r - 7*r^2 + 12*r^3 - 8*r^4) + (n14 + n31)*log(24 - 8*r - 19*r^2 + 20*r^3 - 8*r^4) + (n20 + n25)*log(3 - 4*r^2 + 8*r^3 - 4*r^4) + (n03 + n42)*log(18 - 12*r + 7*r^2 - 2*r^3 - 2*r^4) + (n02 + n43)*log(9 + 12*r - 11*r^2 + 10*r^3 - 2*r^4) + (n22 + n23)*log(12 + r - 2*r^2 + 2*r^3 - r^4) + (n04 + n41)*log(3 - 4*r + 3*r^2 - 3*r^3 + r^4) + (n12 + n33)*log(36 + 3*r + 16*r^2 - 14*r^3 + 4*r^4) + (n13 + n32)*log(45 - 9*r - 2*r^2 - 2*r^3 + 4*r^4) + (n15 + n30)*log(9 - 18*r + 29*r^2 - 28*r^3 + 8*r^4) + (n21 + n24)*log(33 - 8*r + 20*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_mr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23) {
optimize(logL_mr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mr <- mapply(interlogL_mr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_22"],x[,"n_23"])
LOD_mr <- -((x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_32"])*(3*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_33"])*(3*log10(2) + log10(5)) - (x[,"n_20"] + x[,"n_25"])*(-2*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_23"])*(-4*log10(2) + log10(3) + log10(5) + log10(13)) - (x[,"n_15"] + x[,"n_30"])*(-2*log10(2) + log10(17)) - (x[,"n_10"] + x[,"n_35"])*(-2*log10(2) + log10(17)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(23)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(23)) - (x[,"n_14"] + x[,"n_31"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_11"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(107)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(107)) - (x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(127)) + (x[,"n_00"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_mr)) + 2*log10(pmax(1e-6,r_mr)) + log10(2 + r_mr)) + (x[,"n_05"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_mr)) + log10(pmax(1e-6,r_mr)) + log10(3 - 4*r_mr + r_mr^2)) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_mr)) + log10(3 - r_mr^2 + r_mr^3)) + (x[,"n_10"] + x[,"n_35"])*(log10(pmax(1e-6,r_mr)) + log10(12 - 7*r_mr - 4*r_mr^2 + 8*r_mr^3)) + (x[,"n_11"] + x[,"n_34"])*log10(9 + 18*r_mr - 7*r_mr^2 + 12*r_mr^3 - 8*r_mr^4) + (x[,"n_14"] + x[,"n_31"])*log10(24 - 8*r_mr - 19*r_mr^2 + 20*r_mr^3 - 8*r_mr^4) + (x[,"n_20"] + x[,"n_25"])*log10(3 - 4*r_mr^2 + 8*r_mr^3 - 4*r_mr^4) + (x[,"n_03"] + x[,"n_42"])*log10(18 - 12*r_mr + 7*r_mr^2 - 2*r_mr^3 - 2*r_mr^4) + (x[,"n_02"] + x[,"n_43"])*log10(9 + 12*r_mr - 11*r_mr^2 + 10*r_mr^3 - 2*r_mr^4) + (x[,"n_22"] + x[,"n_23"])*log10(12 + r_mr - 2*r_mr^2 + 2*r_mr^3 - r_mr^4) + (x[,"n_04"] + x[,"n_41"])*log10(3 - 4*r_mr + 3*r_mr^2 - 3*r_mr^3 + r_mr^4) + (x[,"n_12"] + x[,"n_33"])*log10(36 + 3*r_mr + 16*r_mr^2 - 14*r_mr^3 + 4*r_mr^4) + (x[,"n_13"] + x[,"n_32"])*log10(45 - 9*r_mr - 2*r_mr^2 - 2*r_mr^3 + 4*r_mr^4) + (x[,"n_15"] + x[,"n_30"])*log10(9 - 18*r_mr + 29*r_mr^2 - 28*r_mr^3 + 8*r_mr^4) + (x[,"n_21"] + x[,"n_24"])*log10(33 - 8*r_mr + 20*r_mr^2 - 24*r_mr^3 + 12*r_mr^4)
logL_mr <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - 2*x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - 2*x[,"n_30"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_40"] - x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - x[,"n_44"] - 2*x[,"n_45"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_00"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_mr)) + 2*log(pmax(1e-6,r_mr)) + log(2 + r_mr)) + (x[,"n_05"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_mr)) + log(pmax(1e-6,r_mr)) + log(3 - 4*r_mr + r_mr^2)) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_mr)) + log(3 - r_mr^2 + r_mr^3)) + (x[,"n_10"] + x[,"n_35"])*(log(pmax(1e-6,r_mr)) + log(12 - 7*r_mr - 4*r_mr^2 + 8*r_mr^3)) + (x[,"n_11"] + x[,"n_34"])*log(9 + 18*r_mr - 7*r_mr^2 + 12*r_mr^3 - 8*r_mr^4) + (x[,"n_14"] + x[,"n_31"])*log(24 - 8*r_mr - 19*r_mr^2 + 20*r_mr^3 - 8*r_mr^4) + (x[,"n_20"] + x[,"n_25"])*log(3 - 4*r_mr^2 + 8*r_mr^3 - 4*r_mr^4) + (x[,"n_03"] + x[,"n_42"])*log(18 - 12*r_mr + 7*r_mr^2 - 2*r_mr^3 - 2*r_mr^4) + (x[,"n_02"] + x[,"n_43"])*log(9 + 12*r_mr - 11*r_mr^2 + 10*r_mr^3 - 2*r_mr^4) + (x[,"n_22"] + x[,"n_23"])*log(12 + r_mr - 2*r_mr^2 + 2*r_mr^3 - r_mr^4) + (x[,"n_04"] + x[,"n_41"])*log(3 - 4*r_mr + 3*r_mr^2 - 3*r_mr^3 + r_mr^4) + (x[,"n_12"] + x[,"n_33"])*log(36 + 3*r_mr + 16*r_mr^2 - 14*r_mr^3 + 4*r_mr^4) + (x[,"n_13"] + x[,"n_32"])*log(45 - 9*r_mr - 2*r_mr^2 - 2*r_mr^3 + 4*r_mr^4) + (x[,"n_15"] + x[,"n_30"])*log(9 - 18*r_mr + 29*r_mr^2 - 28*r_mr^3 + 8*r_mr^4) + (x[,"n_21"] + x[,"n_24"])*log(33 - 8*r_mr + 20*r_mr^2 - 24*r_mr^3 + 12*r_mr^4)
logL_rc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - n15 - 2*n20 - 2*n21 - n22 - n23 - 2*n24 - 2*n25 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45)*log(2) + (n21 + n24)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n00 + n05 + n40 + n45)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n41)*(log(pmax(1e-6,r)) + log(2 - 3*r + 4*r^2 - 3*r^3)) + (n01 + n44)*(log(pmax(1e-6,r)) + log(4 - 9*r + 8*r^2 - 3*r^3)) + (n13 + n32)*log(9 - 2*r + 3*r^2 + 4*r^3 - 8*r^4) + (n12 + n33)*log(6 + 16*r - 33*r^2 + 28*r^3 - 8*r^4) + (n21 + n24)*log(1 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n10 + n15 + n30 + n35)*log(r - 3*r^2 + 4*r^3 - 2*r^4) + (n02 + n43)*log(3 - 8*r + 14*r^2 - 10*r^3 + 2*r^4) + (n03 + n42)*log(1 + 2*r - 4*r^2 + 2*r^3 + 2*r^4) + (n20 + n25)*log(1 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + (n22 + n23)*log(9 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + (n11 + n34)*log(6 - 14*r + 27*r^2 - 28*r^3 + 12*r^4) + (n14 + n31)*log(3 - 4*r + 15*r^2 - 20*r^3 + 12*r^4)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rc <- 3*(x[,"n_10"] + x[,"n_15"] + x[,"n_30"] + x[,"n_35"])*log10(2) + 4*(x[,"n_00"] + x[,"n_05"] + x[,"n_40"] + x[,"n_45"])*log10(2) + (x[,"n_20"] + x[,"n_25"])*(3*log10(2) - log10(3)) - (x[,"n_14"] + x[,"n_31"])*log10(3) - (x[,"n_11"] + x[,"n_34"])*log10(3) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_13"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_33"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_24"])*(-3*log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_23"])*(-3*log10(2) + log10(67)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_40"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_rc)) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_04"] + x[,"n_41"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 3*r_rc + 4*r_rc^2 - 3*r_rc^3)) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_rc)) + log10(4 - 9*r_rc + 8*r_rc^2 - 3*r_rc^3)) + (x[,"n_13"] + x[,"n_32"])*log10(9 - 2*r_rc + 3*r_rc^2 + 4*r_rc^3 - 8*r_rc^4) + (x[,"n_12"] + x[,"n_33"])*log10(6 + 16*r_rc - 33*r_rc^2 + 28*r_rc^3 - 8*r_rc^4) + (x[,"n_21"] + x[,"n_24"])*log10(1 + 4*r_rc - 10*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_10"] + x[,"n_15"] + x[,"n_30"] + x[,"n_35"])*log10(r_rc - 3*r_rc^2 + 4*r_rc^3 - 2*r_rc^4) + (x[,"n_02"] + x[,"n_43"])*log10(3 - 8*r_rc + 14*r_rc^2 - 10*r_rc^3 + 2*r_rc^4) + (x[,"n_03"] + x[,"n_42"])*log10(1 + 2*r_rc - 4*r_rc^2 + 2*r_rc^3 + 2*r_rc^4) + (x[,"n_20"] + x[,"n_25"])*log10(1 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + (x[,"n_22"] + x[,"n_23"])*log10(9 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + (x[,"n_11"] + x[,"n_34"])*log10(6 - 14*r_rc + 27*r_rc^2 - 28*r_rc^3 + 12*r_rc^4) + (x[,"n_14"] + x[,"n_31"])*log10(3 - 4*r_rc + 15*r_rc^2 - 20*r_rc^3 + 12*r_rc^4)
logL_rc <- (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - 2*x[,"n_20"] - 2*x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"])*log(2) + (x[,"n_21"] + x[,"n_24"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_00"] + x[,"n_05"] + x[,"n_40"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_rc)) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_04"] + x[,"n_41"])*(log(pmax(1e-6,r_rc)) + log(2 - 3*r_rc + 4*r_rc^2 - 3*r_rc^3)) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_rc)) + log(4 - 9*r_rc + 8*r_rc^2 - 3*r_rc^3)) + (x[,"n_13"] + x[,"n_32"])*log(9 - 2*r_rc + 3*r_rc^2 + 4*r_rc^3 - 8*r_rc^4) + (x[,"n_12"] + x[,"n_33"])*log(6 + 16*r_rc - 33*r_rc^2 + 28*r_rc^3 - 8*r_rc^4) + (x[,"n_21"] + x[,"n_24"])*log(1 + 4*r_rc - 10*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_10"] + x[,"n_15"] + x[,"n_30"] + x[,"n_35"])*log(r_rc - 3*r_rc^2 + 4*r_rc^3 - 2*r_rc^4) + (x[,"n_02"] + x[,"n_43"])*log(3 - 8*r_rc + 14*r_rc^2 - 10*r_rc^3 + 2*r_rc^4) + (x[,"n_03"] + x[,"n_42"])*log(1 + 2*r_rc - 4*r_rc^2 + 2*r_rc^3 + 2*r_rc^4) + (x[,"n_20"] + x[,"n_25"])*log(1 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + (x[,"n_22"] + x[,"n_23"])*log(9 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + (x[,"n_11"] + x[,"n_34"])*log(6 - 14*r_rc + 27*r_rc^2 - 28*r_rc^3 + 12*r_rc^4) + (x[,"n_14"] + x[,"n_31"])*log(3 - 4*r_rc + 15*r_rc^2 - 20*r_rc^3 + 12*r_rc^4)
logL_rm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 3*n21 - 4*n22 - 4*n23 - 3*n24 - 4*n25 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45)*log(2) + (n22 + n23)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*log(5) + (n05 + n40)*(2*log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n00 + n45)*(2*log(pmax(1e-6,r)) + log(1 - r^2)) + (n01 + n44)*(log(pmax(1e-6,r)) + log(4 + 2*r - 6*r^2 + 3*r^3)) + (n10 + n35)*(log(pmax(1e-6,r)) + log(2 + r - 4*r^2 + 4*r^3)) + (n14 + n31)*log(15 - 2*r - 19*r^2 + 24*r^3 - 12*r^4) + (n11 + n34)*log(6 + 16*r - 19*r^2 + 24*r^3 - 12*r^4) + (n20 + n25)*log(1 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n22 + n23)*log(21 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n03 + n42)*log(10 - 14*r + 15*r^2 - 6*r^3 - 2*r^4) + (n02 + n43)*log(3 + 10*r - 15*r^2 + 14*r^3 - 2*r^4) + (n04 + n41)*log(3 - 2*r + 2*r^2 - 6*r^3 + 3*r^4) + (n15 + n30)*log(3 - 8*r + 13*r^2 - 12*r^3 + 4*r^4) + (n12 + n33)*log(33 - 8*r + 42*r^2 - 36*r^3 + 8*r^4) + (n13 + n32)*log(39 - 18*r^2 + 4*r^3 + 8*r^4) + (n21 + n24)*log(12 - 8*r + 17*r^2 - 18*r^3 + 9*r^4)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rm <- -((x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(3)) - (x[,"n_20"] + x[,"n_25"])*(-3*log10(2) + log10(13)) - (x[,"n_21"] + x[,"n_24"])*(-4*log10(2) + 2*log10(13)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(23)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(23)) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(3) + log10(29)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + log10(31)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + log10(31)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_32"])*(-log10(2) + log10(71)) - (x[,"n_12"] + x[,"n_33"])*(-log10(2) + log10(71)) + (x[,"n_05"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_rm)) + log10(2 - r_rm) + log10(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_rm)) + log10(1 - r_rm^2)) + (x[,"n_01"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm)) + log10(4 + 2*r_rm - 6*r_rm^2 + 3*r_rm^3)) + (x[,"n_10"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm)) + log10(2 + r_rm - 4*r_rm^2 + 4*r_rm^3)) + (x[,"n_14"] + x[,"n_31"])*log10(15 - 2*r_rm - 19*r_rm^2 + 24*r_rm^3 - 12*r_rm^4) + (x[,"n_11"] + x[,"n_34"])*log10(6 + 16*r_rm - 19*r_rm^2 + 24*r_rm^3 - 12*r_rm^4) + (x[,"n_20"] + x[,"n_25"])*log10(1 + 4*r_rm - 10*r_rm^2 + 12*r_rm^3 - 6*r_rm^4) + (x[,"n_22"] + x[,"n_23"])*log10(21 + 4*r_rm - 8*r_rm^2 + 8*r_rm^3 - 4*r_rm^4) + (x[,"n_03"] + x[,"n_42"])*log10(10 - 14*r_rm + 15*r_rm^2 - 6*r_rm^3 - 2*r_rm^4) + (x[,"n_02"] + x[,"n_43"])*log10(3 + 10*r_rm - 15*r_rm^2 + 14*r_rm^3 - 2*r_rm^4) + (x[,"n_04"] + x[,"n_41"])*log10(3 - 2*r_rm + 2*r_rm^2 - 6*r_rm^3 + 3*r_rm^4) + (x[,"n_15"] + x[,"n_30"])*log10(3 - 8*r_rm + 13*r_rm^2 - 12*r_rm^3 + 4*r_rm^4) + (x[,"n_12"] + x[,"n_33"])*log10(33 - 8*r_rm + 42*r_rm^2 - 36*r_rm^3 + 8*r_rm^4) + (x[,"n_13"] + x[,"n_32"])*log10(39 - 18*r_rm^2 + 4*r_rm^3 + 8*r_rm^4) + (x[,"n_21"] + x[,"n_24"])*log10(12 - 8*r_rm + 17*r_rm^2 - 18*r_rm^3 + 9*r_rm^4)
logL_rm <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 3*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"])*log(2) + (x[,"n_22"] + x[,"n_23"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*log(5) + (x[,"n_05"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_rm)) + log(2 - r_rm) + log(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_45"])*(2*log(pmax(1e-6,r_rm)) + log(1 - r_rm^2)) + (x[,"n_01"] + x[,"n_44"])*(log(pmax(1e-6,r_rm)) + log(4 + 2*r_rm - 6*r_rm^2 + 3*r_rm^3)) + (x[,"n_10"] + x[,"n_35"])*(log(pmax(1e-6,r_rm)) + log(2 + r_rm - 4*r_rm^2 + 4*r_rm^3)) + (x[,"n_14"] + x[,"n_31"])*log(15 - 2*r_rm - 19*r_rm^2 + 24*r_rm^3 - 12*r_rm^4) + (x[,"n_11"] + x[,"n_34"])*log(6 + 16*r_rm - 19*r_rm^2 + 24*r_rm^3 - 12*r_rm^4) + (x[,"n_20"] + x[,"n_25"])*log(1 + 4*r_rm - 10*r_rm^2 + 12*r_rm^3 - 6*r_rm^4) + (x[,"n_22"] + x[,"n_23"])*log(21 + 4*r_rm - 8*r_rm^2 + 8*r_rm^3 - 4*r_rm^4) + (x[,"n_03"] + x[,"n_42"])*log(10 - 14*r_rm + 15*r_rm^2 - 6*r_rm^3 - 2*r_rm^4) + (x[,"n_02"] + x[,"n_43"])*log(3 + 10*r_rm - 15*r_rm^2 + 14*r_rm^3 - 2*r_rm^4) + (x[,"n_04"] + x[,"n_41"])*log(3 - 2*r_rm + 2*r_rm^2 - 6*r_rm^3 + 3*r_rm^4) + (x[,"n_15"] + x[,"n_30"])*log(3 - 8*r_rm + 13*r_rm^2 - 12*r_rm^3 + 4*r_rm^4) + (x[,"n_12"] + x[,"n_33"])*log(33 - 8*r_rm + 42*r_rm^2 - 36*r_rm^3 + 8*r_rm^4) + (x[,"n_13"] + x[,"n_32"])*log(39 - 18*r_rm^2 + 4*r_rm^3 + 8*r_rm^4) + (x[,"n_21"] + x[,"n_24"])*log(12 - 8*r_rm + 17*r_rm^2 - 18*r_rm^3 + 9*r_rm^4)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 2*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 2*n25 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45)*(log(3) + 2*log(5)) + (n00 + n45)*(3*log(pmax(1e-6,r)) + log(2 + r)) + (n15 + n30)*(2*log(1 + r) + log(3 + 8*r - 4*r^2)) + (n04 + n41)*(2*log(pmax(1e-6,1 - r)) + log(12 + 2*r - 3*r^2)) + (n01 + n44)*(2*log(pmax(1e-6,r)) + log(11 + 4*r - 3*r^2)) + (n05 + n40)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n02 + n43)*(log(pmax(1e-6,r)) + log(18 + 7*r - 18*r^2 + 2*r^3)) + (n20 + n25)*(log(pmax(1e-6,r)) + log(4 - r - 6*r^2 + 3*r^3)) + (n11 + n34)*(log(pmax(1e-6,r)) + log(30 + 17*r - 20*r^2 + 12*r^3)) + (n21 + n24)*log(21 + 44*r - 62*r^2 + 36*r^3 - 18*r^4) + (n13 + n32)*log(75 - 52*r + 24*r^2 - 12*r^3 - 8*r^4) + (n12 + n33)*log(27 + 72*r - 60*r^2 + 44*r^3 - 8*r^4) + (n10 + n35)*log(7*r^2 - 4*r^4) + (n03 + n42)*log(9 + 14*r - 35*r^2 + 10*r^3 + 2*r^4) + (n14 + n31)*log(39 - 52*r + 29*r^2 - 28*r^3 + 12*r^4) + (n22 + n23)*log(111 - 52*r + 64*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"])
LOD_rr <- -((x[,"n_10"] + x[,"n_35"])*(-log10(2) + log10(3))) - (x[,"n_15"] + x[,"n_30"])*(log10(2) + log10(3) + 2*(-log10(2) + log10(3))) - (x[,"n_05"] + x[,"n_40"])*(-4*log10(2) + log10(5)) - (x[,"n_00"] + x[,"n_45"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_31"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_11"] + x[,"n_34"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_04"] + x[,"n_41"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_01"] + x[,"n_44"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_20"] + x[,"n_25"])*(-4*log10(2) + log10(19)) - (x[,"n_21"] + x[,"n_24"])*(-3*log10(2) + log10(13) + log10(19)) - (x[,"n_03"] + x[,"n_42"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_02"] + x[,"n_43"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_13"] + x[,"n_32"])*log10(53) - (x[,"n_12"] + x[,"n_33"])*log10(53) - (x[,"n_22"] + x[,"n_23"])*(-2*log10(2) + log10(5) + log10(79)) + (x[,"n_00"] + x[,"n_45"])*(3*log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_15"] + x[,"n_30"])*(2*log10(1 + r_rr) + log10(3 + 8*r_rr - 4*r_rr^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(12 + 2*r_rr - 3*r_rr^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(11 + 4*r_rr - 3*r_rr^2)) + (x[,"n_05"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 - 4*r_rr + r_rr^2)) + (x[,"n_02"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(18 + 7*r_rr - 18*r_rr^2 + 2*r_rr^3)) + (x[,"n_20"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(4 - r_rr - 6*r_rr^2 + 3*r_rr^3)) + (x[,"n_11"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(30 + 17*r_rr - 20*r_rr^2 + 12*r_rr^3)) + (x[,"n_21"] + x[,"n_24"])*log10(21 + 44*r_rr - 62*r_rr^2 + 36*r_rr^3 - 18*r_rr^4) + (x[,"n_13"] + x[,"n_32"])*log10(75 - 52*r_rr + 24*r_rr^2 - 12*r_rr^3 - 8*r_rr^4) + (x[,"n_12"] + x[,"n_33"])*log10(27 + 72*r_rr - 60*r_rr^2 + 44*r_rr^3 - 8*r_rr^4) + (x[,"n_10"] + x[,"n_35"])*log10(7*r_rr^2 - 4*r_rr^4) + (x[,"n_03"] + x[,"n_42"])*log10(9 + 14*r_rr - 35*r_rr^2 + 10*r_rr^3 + 2*r_rr^4) + (x[,"n_14"] + x[,"n_31"])*log10(39 - 52*r_rr + 29*r_rr^2 - 28*r_rr^3 + 12*r_rr^4) + (x[,"n_22"] + x[,"n_23"])*log10(111 - 52*r_rr + 64*r_rr^2 - 24*r_rr^3 + 12*r_rr^4)
logL_rr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 2*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_45"])*(3*log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_15"] + x[,"n_30"])*(2*log(1 + r_rr) + log(3 + 8*r_rr - 4*r_rr^2)) + (x[,"n_04"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(12 + 2*r_rr - 3*r_rr^2)) + (x[,"n_01"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(11 + 4*r_rr - 3*r_rr^2)) + (x[,"n_05"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 - 4*r_rr + r_rr^2)) + (x[,"n_02"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(18 + 7*r_rr - 18*r_rr^2 + 2*r_rr^3)) + (x[,"n_20"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(4 - r_rr - 6*r_rr^2 + 3*r_rr^3)) + (x[,"n_11"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(30 + 17*r_rr - 20*r_rr^2 + 12*r_rr^3)) + (x[,"n_21"] + x[,"n_24"])*log(21 + 44*r_rr - 62*r_rr^2 + 36*r_rr^3 - 18*r_rr^4) + (x[,"n_13"] + x[,"n_32"])*log(75 - 52*r_rr + 24*r_rr^2 - 12*r_rr^3 - 8*r_rr^4) + (x[,"n_12"] + x[,"n_33"])*log(27 + 72*r_rr - 60*r_rr^2 + 44*r_rr^3 - 8*r_rr^4) + (x[,"n_10"] + x[,"n_35"])*log(7*r_rr^2 - 4*r_rr^4) + (x[,"n_03"] + x[,"n_42"])*log(9 + 14*r_rr - 35*r_rr^2 + 10*r_rr^3 + 2*r_rr^4) + (x[,"n_14"] + x[,"n_31"])*log(39 - 52*r_rr + 29*r_rr^2 - 28*r_rr^3 + 12*r_rr^4) + (x[,"n_22"] + x[,"n_23"])*log(111 - 52*r_rr + 64*r_rr^2 - 24*r_rr^3 + 12*r_rr^4)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_mc,r_mm,r_mr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_mc,LOD_mm,LOD_mr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_mc,logL_mm,logL_mr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","mixed coupling","mixed mixed","mixed repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.2_3.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 4*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 3*n20 - 2*n21 - 3*n22 - 2*n23 - 3*n24 - 2*n25 - 3*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 4*n46)*log(2) + (n20 + n26)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*log(5) + 4*(n00 + n46)*log(pmax(1e-6,1 - r)) + 4*(n06 + n40)*log(pmax(1e-6,r)) + (n10 + n36)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n20 + n26)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n16 + n30)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n05 + n41)*(log(4 - r) + 3*log(pmax(1e-6,r))) + (n04 + n42)*(2*log(pmax(1e-6,r)) + log(22 - 12*r - r^2)) + (n01 + n45)*(2*log(pmax(1e-6,1 - r)) + log(3 - 2*r - r^2)) + (n02 + n44)*(2*log(pmax(1e-6,1 - r)) + log(9 + 14*r - r^2)) + (n15 + n31)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n11 + n35)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n21 + n25)*(log(pmax(1e-6,r)) + log(8 - 11*r + 6*r^2 - 3*r^3)) + (n03 + n43)*(log(pmax(1e-6,r)) + log(6 - 5*r - 2*r^2 + r^3)) + (n14 + n32)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n13 + n33)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n22 + n24)*log(21 + 4*r - 7*r^2 + 6*r^3 - 3*r^4) + (n12 + n34)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + n23*log(23 - 20*r + 26*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cc <- 4*(x[,"n_20"] + x[,"n_26"])*log10(2) + 4*(x[,"n_16"] + x[,"n_30"])*log10(2) + 4*(x[,"n_10"] + x[,"n_36"])*log10(2) + 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) - (x[,"n_15"] + x[,"n_31"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_43"])*(-4*log10(2) + 2*log10(5)) - (x[,"n_05"] + x[,"n_41"])*(-4*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_45"])*(-4*log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + 2*log10(3) + log10(7)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + 2*log10(3) + log10(7)) - x[,"n_23"]*(-3*log10(2) + log10(3) + 2*log10(7)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_14"] + x[,"n_32"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_12"] + x[,"n_34"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_21"] + x[,"n_25"])*(-4*log10(2) + log10(29)) - (x[,"n_22"] + x[,"n_24"])*(-4*log10(2) + log10(349)) + 4*(x[,"n_00"] + x[,"n_46"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_06"] + x[,"n_40"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_36"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_26"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_16"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_41"])*(log10(4 - r_cc) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(22 - 12*r_cc - r_cc^2)) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 - 2*r_cc - r_cc^2)) + (x[,"n_02"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(9 + 14*r_cc - r_cc^2)) + (x[,"n_15"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(7 - 6*r_cc + 2*r_cc^2)) + (x[,"n_11"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 2*r_cc + 2*r_cc^2)) + (x[,"n_21"] + x[,"n_25"])*(log10(pmax(1e-6,r_cc)) + log10(8 - 11*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_03"] + x[,"n_43"])*(log10(pmax(1e-6,r_cc)) + log10(6 - 5*r_cc - 2*r_cc^2 + r_cc^3)) + (x[,"n_14"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(15 - 9*r_cc + 5*r_cc^2 + r_cc^3)) + (x[,"n_13"] + x[,"n_33"])*log10(9 + 4*r_cc - 8*r_cc^2 + 8*r_cc^3 - 4*r_cc^4) + (x[,"n_22"] + x[,"n_24"])*log10(21 + 4*r_cc - 7*r_cc^2 + 6*r_cc^3 - 3*r_cc^4) + (x[,"n_12"] + x[,"n_34"])*log10(12 - 16*r_cc + 12*r_cc^2 - 9*r_cc^3 + r_cc^4) + x[,"n_23"]*log10(23 - 20*r_cc + 26*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cc <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 3*x[,"n_20"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 3*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (x[,"n_20"] + x[,"n_26"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*log(5) + 4*(x[,"n_00"] + x[,"n_46"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_06"] + x[,"n_40"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_36"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_26"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_16"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_41"])*(log(4 - r_cc) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(22 - 12*r_cc - r_cc^2)) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 - 2*r_cc - r_cc^2)) + (x[,"n_02"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(9 + 14*r_cc - r_cc^2)) + (x[,"n_15"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(7 - 6*r_cc + 2*r_cc^2)) + (x[,"n_11"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 2*r_cc + 2*r_cc^2)) + (x[,"n_21"] + x[,"n_25"])*(log(pmax(1e-6,r_cc)) + log(8 - 11*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_03"] + x[,"n_43"])*(log(pmax(1e-6,r_cc)) + log(6 - 5*r_cc - 2*r_cc^2 + r_cc^3)) + (x[,"n_14"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(15 - 9*r_cc + 5*r_cc^2 + r_cc^3)) + (x[,"n_13"] + x[,"n_33"])*log(9 + 4*r_cc - 8*r_cc^2 + 8*r_cc^3 - 4*r_cc^4) + (x[,"n_22"] + x[,"n_24"])*log(21 + 4*r_cc - 7*r_cc^2 + 6*r_cc^3 - 3*r_cc^4) + (x[,"n_12"] + x[,"n_34"])*log(12 - 16*r_cc + 12*r_cc^2 - 9*r_cc^3 + r_cc^4) + x[,"n_23"]*log(23 - 20*r_cc + 26*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cm <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n06 - 4*n10 - 2*n11 - 4*n12 - n13 - 4*n14 - 2*n15 - 4*n16 - 2*n20 - n21 - 2*n22 - n23 - 2*n24 - n25 - 2*n26 - 4*n30 - 2*n31 - 4*n32 - n33 - 4*n34 - 2*n35 - 4*n36 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 3*n46)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(log(3) + 2*log(5)) + (n00 + n46)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n06 + n40)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n05 + n41)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n01 + n45)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n16 + n30)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n10 + n36)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n15 + n31)*(log(pmax(1e-6,r)) + log(9 - 9*r + 10*r^2 - 4*r^3)) + (n20 + n26)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n04 + n42)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n14 + n32)*log(45 + 42*r + 29*r^2 - 12*r^3 - 8*r^4) + (n12 + n34)*log(96 - 32*r - 55*r^2 + 44*r^3 - 8*r^4) + n23*log(27 + 2*r - 8*r^2 + 12*r^3 - 6*r^4) + (n11 + n35)*log(6 - 5*r - 3*r^2 + 6*r^3 - 4*r^4) + (n03 + n43)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n02 + n44)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + (n21 + n25)*log(3 - r + 4*r^2 - 6*r^3 + 3*r^4) + (n22 + n24)*log(33 - 2*r + 5*r^2 - 6*r^3 + 3*r^4) + (n13 + n33)*log(15 - 2*r + 6*r^2 - 8*r^3 + 4*r^4)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cm <- 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) - (x[,"n_16"] + x[,"n_30"])*(-2*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_36"])*(-2*log10(2) + log10(3)) - (x[,"n_05"] + x[,"n_41"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_45"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_20"] + x[,"n_26"])*(-4*log10(2) + log10(5)) - (x[,"n_15"] + x[,"n_31"])*(-2*log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_35"])*(-2*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_43"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - x[,"n_23"]*(-3*log10(2) + log10(7) + log10(31)) - (x[,"n_21"] + x[,"n_25"])*(-4*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(59)) - (x[,"n_22"] + x[,"n_24"])*(-4*log10(2) + log10(523)) + (x[,"n_00"] + x[,"n_46"])*(3*log10(pmax(1e-6,1 - r_cm)) + log10(pmax(1e-6,r_cm))) + (x[,"n_06"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cm)) + 3*log10(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm)) + log10(7 - 6*r_cm + 2*r_cm^2)) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(3 + 2*r_cm + 2*r_cm^2)) + (x[,"n_16"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_cm)) + log10(7 - 12*r_cm + 8*r_cm^2)) + (x[,"n_10"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(3 - 4*r_cm + 8*r_cm^2)) + (x[,"n_15"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm)) + log10(9 - 9*r_cm + 10*r_cm^2 - 4*r_cm^3)) + (x[,"n_20"] + x[,"n_26"])*(log10(pmax(1e-6,r_cm)) + log10(2 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3)) + (x[,"n_04"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm)) + log10(15 - 9*r_cm + 5*r_cm^2 + r_cm^3)) + (x[,"n_14"] + x[,"n_32"])*log10(45 + 42*r_cm + 29*r_cm^2 - 12*r_cm^3 - 8*r_cm^4) + (x[,"n_12"] + x[,"n_34"])*log10(96 - 32*r_cm - 55*r_cm^2 + 44*r_cm^3 - 8*r_cm^4) + x[,"n_23"]*log10(27 + 2*r_cm - 8*r_cm^2 + 12*r_cm^3 - 6*r_cm^4) + (x[,"n_11"] + x[,"n_35"])*log10(6 - 5*r_cm - 3*r_cm^2 + 6*r_cm^3 - 4*r_cm^4) + (x[,"n_03"] + x[,"n_43"])*log10(9 + 4*r_cm - 8*r_cm^2 + 8*r_cm^3 - 4*r_cm^4) + (x[,"n_02"] + x[,"n_44"])*log10(12 - 16*r_cm + 12*r_cm^2 - 9*r_cm^3 + r_cm^4) + (x[,"n_21"] + x[,"n_25"])*log10(3 - r_cm + 4*r_cm^2 - 6*r_cm^3 + 3*r_cm^4) + (x[,"n_22"] + x[,"n_24"])*log10(33 - 2*r_cm + 5*r_cm^2 - 6*r_cm^3 + 3*r_cm^4) + (x[,"n_13"] + x[,"n_33"])*log10(15 - 2*r_cm + 6*r_cm^2 - 8*r_cm^3 + 4*r_cm^4)
logL_cm <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_06"] - 4*x[,"n_10"] - 2*x[,"n_11"] - 4*x[,"n_12"] - x[,"n_13"] - 4*x[,"n_14"] - 2*x[,"n_15"] - 4*x[,"n_16"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - 2*x[,"n_26"] - 4*x[,"n_30"] - 2*x[,"n_31"] - 4*x[,"n_32"] - x[,"n_33"] - 4*x[,"n_34"] - 2*x[,"n_35"] - 4*x[,"n_36"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_46"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_46"])*(3*log(pmax(1e-6,1 - r_cm)) + log(pmax(1e-6,r_cm))) + (x[,"n_06"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cm)) + 3*log(pmax(1e-6,r_cm))) + (x[,"n_05"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm)) + log(7 - 6*r_cm + 2*r_cm^2)) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm)) + log(3 + 2*r_cm + 2*r_cm^2)) + (x[,"n_16"] + x[,"n_30"])*(2*log(pmax(1e-6,r_cm)) + log(7 - 12*r_cm + 8*r_cm^2)) + (x[,"n_10"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_cm)) + log(3 - 4*r_cm + 8*r_cm^2)) + (x[,"n_15"] + x[,"n_31"])*(log(pmax(1e-6,r_cm)) + log(9 - 9*r_cm + 10*r_cm^2 - 4*r_cm^3)) + (x[,"n_20"] + x[,"n_26"])*(log(pmax(1e-6,r_cm)) + log(2 - 5*r_cm + 6*r_cm^2 - 3*r_cm^3)) + (x[,"n_04"] + x[,"n_42"])*(log(pmax(1e-6,r_cm)) + log(15 - 9*r_cm + 5*r_cm^2 + r_cm^3)) + (x[,"n_14"] + x[,"n_32"])*log(45 + 42*r_cm + 29*r_cm^2 - 12*r_cm^3 - 8*r_cm^4) + (x[,"n_12"] + x[,"n_34"])*log(96 - 32*r_cm - 55*r_cm^2 + 44*r_cm^3 - 8*r_cm^4) + x[,"n_23"]*log(27 + 2*r_cm - 8*r_cm^2 + 12*r_cm^3 - 6*r_cm^4) + (x[,"n_11"] + x[,"n_35"])*log(6 - 5*r_cm - 3*r_cm^2 + 6*r_cm^3 - 4*r_cm^4) + (x[,"n_03"] + x[,"n_43"])*log(9 + 4*r_cm - 8*r_cm^2 + 8*r_cm^3 - 4*r_cm^4) + (x[,"n_02"] + x[,"n_44"])*log(12 - 16*r_cm + 12*r_cm^2 - 9*r_cm^3 + r_cm^4) + (x[,"n_21"] + x[,"n_25"])*log(3 - r_cm + 4*r_cm^2 - 6*r_cm^3 + 3*r_cm^4) + (x[,"n_22"] + x[,"n_24"])*log(33 - 2*r_cm + 5*r_cm^2 - 6*r_cm^3 + 3*r_cm^4) + (x[,"n_13"] + x[,"n_33"])*log(15 - 2*r_cm + 6*r_cm^2 - 8*r_cm^3 + 4*r_cm^4)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 3*n21 - 4*n22 - n23 - 4*n24 - 3*n25 - 4*n26 - 3*n30 - 3*n31 - 3*n32 - 2*n33 - 3*n34 - 3*n35 - 3*n36 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44 - 3*n45 - 4*n46)*log(2) + n23*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*log(5) + (n00 + n06 + n40 + n46)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n05 + n41 + n45)*(log(pmax(1e-6,r)) + log(2 - 3*r + 2*r^2 - r^3)) + (n21 + n25)*log(3 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n22 + n24)*log(45 - 4*r - 2*r^2 + 12*r^3 - 6*r^4) + (n13 + n33)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n10 + n16 + n30 + n36)*log(r - 3*r^2 + 4*r^3 - 2*r^4) + (n02 + n04 + n42 + n44)*log(3 + 4*r - 5*r^2 + 2*r^3 - r^4) + n23*log(3 + r^2 - 2*r^3 + r^4) + (n12 + n14 + n32 + n34)*log(12 - r + 3*r^2 - 4*r^3 + 2*r^4) + (n03 + n43)*log(5 - 8*r + 10*r^2 - 4*r^3 + 2*r^4) + (n11 + n15 + n31 + n35)*log(3 - 4*r + 8*r^2 - 8*r^3 + 4*r^4) + (n20 + n26)*log(1 - 4*r + 10*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_cr <- 3*(x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*log10(2) + 4*(x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*log10(2) + (x[,"n_20"] + x[,"n_26"])*(3*log10(2) - log10(3)) - (x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"])*(-2*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_43"])*(-3*log10(2) + 2*log10(5)) - (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(-4*log10(2) + log10(7)) - (x[,"n_02"] + x[,"n_04"] + x[,"n_42"] + x[,"n_44"])*(-4*log10(2) + 2*log10(3) + log10(7)) - x[,"n_23"]*(-4*log10(2) + 2*log10(7)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*(-3*log10(2) + log10(5) + log10(19)) - (x[,"n_21"] + x[,"n_25"])*(-3*log10(2) + log10(29)) - (x[,"n_22"] + x[,"n_24"])*(-3*log10(2) + log10(349)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*(2*log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 3*r_cr + 2*r_cr^2 - r_cr^3)) + (x[,"n_21"] + x[,"n_25"])*log10(3 + 4*r_cr - 10*r_cr^2 + 12*r_cr^3 - 6*r_cr^4) + (x[,"n_22"] + x[,"n_24"])*log10(45 - 4*r_cr - 2*r_cr^2 + 12*r_cr^3 - 6*r_cr^4) + (x[,"n_13"] + x[,"n_33"])*log10(9 + 4*r_cr - 8*r_cr^2 + 8*r_cr^3 - 4*r_cr^4) + (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*log10(r_cr - 3*r_cr^2 + 4*r_cr^3 - 2*r_cr^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_42"] + x[,"n_44"])*log10(3 + 4*r_cr - 5*r_cr^2 + 2*r_cr^3 - r_cr^4) + x[,"n_23"]*log10(3 + r_cr^2 - 2*r_cr^3 + r_cr^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log10(12 - r_cr + 3*r_cr^2 - 4*r_cr^3 + 2*r_cr^4) + (x[,"n_03"] + x[,"n_43"])*log10(5 - 8*r_cr + 10*r_cr^2 - 4*r_cr^3 + 2*r_cr^4) + (x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"])*log10(3 - 4*r_cr + 8*r_cr^2 - 8*r_cr^3 + 4*r_cr^4) + (x[,"n_20"] + x[,"n_26"])*log10(1 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4)
logL_cr <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"])*log(2) + x[,"n_23"]*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*log(5) + (x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*(2*log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(log(pmax(1e-6,r_cr)) + log(2 - 3*r_cr + 2*r_cr^2 - r_cr^3)) + (x[,"n_21"] + x[,"n_25"])*log(3 + 4*r_cr - 10*r_cr^2 + 12*r_cr^3 - 6*r_cr^4) + (x[,"n_22"] + x[,"n_24"])*log(45 - 4*r_cr - 2*r_cr^2 + 12*r_cr^3 - 6*r_cr^4) + (x[,"n_13"] + x[,"n_33"])*log(9 + 4*r_cr - 8*r_cr^2 + 8*r_cr^3 - 4*r_cr^4) + (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*log(r_cr - 3*r_cr^2 + 4*r_cr^3 - 2*r_cr^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_42"] + x[,"n_44"])*log(3 + 4*r_cr - 5*r_cr^2 + 2*r_cr^3 - r_cr^4) + x[,"n_23"]*log(3 + r_cr^2 - 2*r_cr^3 + r_cr^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log(12 - r_cr + 3*r_cr^2 - 4*r_cr^3 + 2*r_cr^4) + (x[,"n_03"] + x[,"n_43"])*log(5 - 8*r_cr + 10*r_cr^2 - 4*r_cr^3 + 2*r_cr^4) + (x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"])*log(3 - 4*r_cr + 8*r_cr^2 - 8*r_cr^3 + 4*r_cr^4) + (x[,"n_20"] + x[,"n_26"])*log(1 - 4*r_cr + 10*r_cr^2 - 12*r_cr^3 + 6*r_cr^4)

logL_mm <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-2*n00 - n01 - 2*n02 - n03 - 2*n04 - n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - n13 - 2*n14 - 2*n15 - 2*n16 - 4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26 - 2*n30 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - 2*n36 - 2*n40 - n41 - 2*n42 - n43 - 2*n44 - n45 - 2*n46)*log(2) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n46)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*log(5) + (n00 + n06 + n40 + n46)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n05 + n41 + n45)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - r + r^2)) + (n10 + n16 + n30 + n36)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - 4*r + 4*r^2)) + (n21 + n25)*log(15 - 8*r + 16*r^3 - 8*r^4) + (n22 + n24)*log(129 + 8*r - 16*r^2 + 16*r^3 - 8*r^4) + (n13 + n33)*log(45 - 8*r^2 + 16*r^3 - 8*r^4) + (n02 + n04 + n42 + n44)*log(9 - r^2 + 2*r^3 - r^4) + (n03 + n43)*log(9 - 6*r + 8*r^2 - 4*r^3 + 2*r^4) + (n12 + n14 + n32 + n34)*log(54 - 3*r + 7*r^2 - 8*r^3 + 4*r^4) + n23*log(51 + 8*r - 16*r^3 + 8*r^4) + (n11 + n15 + n31 + n35)*log(9 + 8*r^2 - 16*r^3 + 8*r^4) + (n20 + n26)*log(3 - 8*r + 16*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_mm <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_mm,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm <- mapply(interlogL_mm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_mm <- (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*log10(2) + 4*(x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*log10(2) - (x[,"n_20"] + x[,"n_26"])*(-log10(2) + log10(3)) - (x[,"n_21"] + x[,"n_25"])*(-log10(2) + 2*log10(5)) - (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(-4*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_04"] + x[,"n_42"] + x[,"n_44"])*(-4*log10(2) + log10(11) + log10(13)) - (x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"])*(-log10(2) + log10(19)) - (x[,"n_22"] + x[,"n_24"])*(-log10(2) + 2*log10(3) + log10(29)) - (x[,"n_03"] + x[,"n_43"])*(-3*log10(2) + log10(61)) - (x[,"n_13"] + x[,"n_33"])*(-log10(2) + log10(89)) - x[,"n_23"]*(-log10(2) + log10(107)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*(-log10(2) + log10(107)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*(2*log10(pmax(1e-6,1 - r_mm)) + 2*log10(pmax(1e-6,r_mm))) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(log10(pmax(1e-6,1 - r_mm)) + log10(pmax(1e-6,r_mm)) + log10(3 - r_mm + r_mm^2)) + (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*(log10(pmax(1e-6,1 - r_mm)) + log10(pmax(1e-6,r_mm)) + log10(3 - 4*r_mm + 4*r_mm^2)) + (x[,"n_21"] + x[,"n_25"])*log10(15 - 8*r_mm + 16*r_mm^3 - 8*r_mm^4) + (x[,"n_22"] + x[,"n_24"])*log10(129 + 8*r_mm - 16*r_mm^2 + 16*r_mm^3 - 8*r_mm^4) + (x[,"n_13"] + x[,"n_33"])*log10(45 - 8*r_mm^2 + 16*r_mm^3 - 8*r_mm^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_42"] + x[,"n_44"])*log10(9 - r_mm^2 + 2*r_mm^3 - r_mm^4) + (x[,"n_03"] + x[,"n_43"])*log10(9 - 6*r_mm + 8*r_mm^2 - 4*r_mm^3 + 2*r_mm^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log10(54 - 3*r_mm + 7*r_mm^2 - 8*r_mm^3 + 4*r_mm^4) + x[,"n_23"]*log10(51 + 8*r_mm - 16*r_mm^3 + 8*r_mm^4) + (x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"])*log10(9 + 8*r_mm^2 - 16*r_mm^3 + 8*r_mm^4) + (x[,"n_20"] + x[,"n_26"])*log10(3 - 8*r_mm + 16*r_mm^2 - 16*r_mm^3 + 8*r_mm^4)
logL_mm <- (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - x[,"n_45"] - 2*x[,"n_46"])*log(2) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_46"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*log(5) + (x[,"n_00"] + x[,"n_06"] + x[,"n_40"] + x[,"n_46"])*(2*log(pmax(1e-6,1 - r_mm)) + 2*log(pmax(1e-6,r_mm))) + (x[,"n_01"] + x[,"n_05"] + x[,"n_41"] + x[,"n_45"])*(log(pmax(1e-6,1 - r_mm)) + log(pmax(1e-6,r_mm)) + log(3 - r_mm + r_mm^2)) + (x[,"n_10"] + x[,"n_16"] + x[,"n_30"] + x[,"n_36"])*(log(pmax(1e-6,1 - r_mm)) + log(pmax(1e-6,r_mm)) + log(3 - 4*r_mm + 4*r_mm^2)) + (x[,"n_21"] + x[,"n_25"])*log(15 - 8*r_mm + 16*r_mm^3 - 8*r_mm^4) + (x[,"n_22"] + x[,"n_24"])*log(129 + 8*r_mm - 16*r_mm^2 + 16*r_mm^3 - 8*r_mm^4) + (x[,"n_13"] + x[,"n_33"])*log(45 - 8*r_mm^2 + 16*r_mm^3 - 8*r_mm^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_42"] + x[,"n_44"])*log(9 - r_mm^2 + 2*r_mm^3 - r_mm^4) + (x[,"n_03"] + x[,"n_43"])*log(9 - 6*r_mm + 8*r_mm^2 - 4*r_mm^3 + 2*r_mm^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_32"] + x[,"n_34"])*log(54 - 3*r_mm + 7*r_mm^2 - 8*r_mm^3 + 4*r_mm^4) + x[,"n_23"]*log(51 + 8*r_mm - 16*r_mm^3 + 8*r_mm^4) + (x[,"n_11"] + x[,"n_15"] + x[,"n_31"] + x[,"n_35"])*log(9 + 8*r_mm^2 - 16*r_mm^3 + 8*r_mm^4) + (x[,"n_20"] + x[,"n_26"])*log(3 - 8*r_mm + 16*r_mm^2 - 16*r_mm^3 + 8*r_mm^4)
logL_mr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n06 - 4*n10 - 2*n11 - 4*n12 - n13 - 4*n14 - 2*n15 - 4*n16 - 2*n20 - n21 - 2*n22 - n23 - 2*n24 - n25 - 2*n26 - 4*n30 - 2*n31 - 4*n32 - n33 - 4*n34 - 2*n35 - 4*n36 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 3*n46)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*(log(3) + 2*log(5)) + (n06 + n40)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n46)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n45)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n05 + n41)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n10 + n36)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n16 + n30)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n11 + n35)*(log(pmax(1e-6,r)) + log(9 - 9*r + 10*r^2 - 4*r^3)) + (n20 + n26)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n02 + n44)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n12 + n34)*log(45 + 42*r + 29*r^2 - 12*r^3 - 8*r^4) + (n14 + n32)*log(96 - 32*r - 55*r^2 + 44*r^3 - 8*r^4) + n23*log(27 + 2*r - 8*r^2 + 12*r^3 - 6*r^4) + (n15 + n31)*log(6 - 5*r - 3*r^2 + 6*r^3 - 4*r^4) + (n03 + n43)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n04 + n42)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + (n21 + n25)*log(3 - r + 4*r^2 - 6*r^3 + 3*r^4) + (n22 + n24)*log(33 - 2*r + 5*r^2 - 6*r^3 + 3*r^4) + (n13 + n33)*log(15 - 2*r + 6*r^2 - 8*r^3 + 4*r^4)
return(L)}
interlogL_mr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_mr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mr <- mapply(interlogL_mr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_mr <- 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) - (x[,"n_16"] + x[,"n_30"])*(-2*log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_36"])*(-2*log10(2) + log10(3)) - (x[,"n_05"] + x[,"n_41"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_45"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_20"] + x[,"n_26"])*(-4*log10(2) + log10(5)) - (x[,"n_15"] + x[,"n_31"])*(-2*log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_35"])*(-2*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_43"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_14"] + x[,"n_32"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - (x[,"n_12"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - x[,"n_23"]*(-3*log10(2) + log10(7) + log10(31)) - (x[,"n_21"] + x[,"n_25"])*(-4*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(59)) - (x[,"n_22"] + x[,"n_24"])*(-4*log10(2) + log10(523)) + (x[,"n_06"] + x[,"n_40"])*(3*log10(pmax(1e-6,1 - r_mr)) + log10(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_46"])*(log10(pmax(1e-6,1 - r_mr)) + 3*log10(pmax(1e-6,r_mr))) + (x[,"n_01"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_mr)) + log10(7 - 6*r_mr + 2*r_mr^2)) + (x[,"n_05"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(3 + 2*r_mr + 2*r_mr^2)) + (x[,"n_10"] + x[,"n_36"])*(2*log10(pmax(1e-6,r_mr)) + log10(7 - 12*r_mr + 8*r_mr^2)) + (x[,"n_16"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(3 - 4*r_mr + 8*r_mr^2)) + (x[,"n_11"] + x[,"n_35"])*(log10(pmax(1e-6,r_mr)) + log10(9 - 9*r_mr + 10*r_mr^2 - 4*r_mr^3)) + (x[,"n_20"] + x[,"n_26"])*(log10(pmax(1e-6,r_mr)) + log10(2 - 5*r_mr + 6*r_mr^2 - 3*r_mr^3)) + (x[,"n_02"] + x[,"n_44"])*(log10(pmax(1e-6,r_mr)) + log10(15 - 9*r_mr + 5*r_mr^2 + r_mr^3)) + (x[,"n_12"] + x[,"n_34"])*log10(45 + 42*r_mr + 29*r_mr^2 - 12*r_mr^3 - 8*r_mr^4) + (x[,"n_14"] + x[,"n_32"])*log10(96 - 32*r_mr - 55*r_mr^2 + 44*r_mr^3 - 8*r_mr^4) + x[,"n_23"]*log10(27 + 2*r_mr - 8*r_mr^2 + 12*r_mr^3 - 6*r_mr^4) + (x[,"n_15"] + x[,"n_31"])*log10(6 - 5*r_mr - 3*r_mr^2 + 6*r_mr^3 - 4*r_mr^4) + (x[,"n_03"] + x[,"n_43"])*log10(9 + 4*r_mr - 8*r_mr^2 + 8*r_mr^3 - 4*r_mr^4) + (x[,"n_04"] + x[,"n_42"])*log10(12 - 16*r_mr + 12*r_mr^2 - 9*r_mr^3 + r_mr^4) + (x[,"n_21"] + x[,"n_25"])*log10(3 - r_mr + 4*r_mr^2 - 6*r_mr^3 + 3*r_mr^4) + (x[,"n_22"] + x[,"n_24"])*log10(33 - 2*r_mr + 5*r_mr^2 - 6*r_mr^3 + 3*r_mr^4) + (x[,"n_13"] + x[,"n_33"])*log10(15 - 2*r_mr + 6*r_mr^2 - 8*r_mr^3 + 4*r_mr^4)
logL_mr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_06"] - 4*x[,"n_10"] - 2*x[,"n_11"] - 4*x[,"n_12"] - x[,"n_13"] - 4*x[,"n_14"] - 2*x[,"n_15"] - 4*x[,"n_16"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - 2*x[,"n_26"] - 4*x[,"n_30"] - 2*x[,"n_31"] - 4*x[,"n_32"] - x[,"n_33"] - 4*x[,"n_34"] - 2*x[,"n_35"] - 4*x[,"n_36"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_46"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*(log(3) + 2*log(5)) + (x[,"n_06"] + x[,"n_40"])*(3*log(pmax(1e-6,1 - r_mr)) + log(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_46"])*(log(pmax(1e-6,1 - r_mr)) + 3*log(pmax(1e-6,r_mr))) + (x[,"n_01"] + x[,"n_45"])*(2*log(pmax(1e-6,r_mr)) + log(7 - 6*r_mr + 2*r_mr^2)) + (x[,"n_05"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_mr)) + log(3 + 2*r_mr + 2*r_mr^2)) + (x[,"n_10"] + x[,"n_36"])*(2*log(pmax(1e-6,r_mr)) + log(7 - 12*r_mr + 8*r_mr^2)) + (x[,"n_16"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_mr)) + log(3 - 4*r_mr + 8*r_mr^2)) + (x[,"n_11"] + x[,"n_35"])*(log(pmax(1e-6,r_mr)) + log(9 - 9*r_mr + 10*r_mr^2 - 4*r_mr^3)) + (x[,"n_20"] + x[,"n_26"])*(log(pmax(1e-6,r_mr)) + log(2 - 5*r_mr + 6*r_mr^2 - 3*r_mr^3)) + (x[,"n_02"] + x[,"n_44"])*(log(pmax(1e-6,r_mr)) + log(15 - 9*r_mr + 5*r_mr^2 + r_mr^3)) + (x[,"n_12"] + x[,"n_34"])*log(45 + 42*r_mr + 29*r_mr^2 - 12*r_mr^3 - 8*r_mr^4) + (x[,"n_14"] + x[,"n_32"])*log(96 - 32*r_mr - 55*r_mr^2 + 44*r_mr^3 - 8*r_mr^4) + x[,"n_23"]*log(27 + 2*r_mr - 8*r_mr^2 + 12*r_mr^3 - 6*r_mr^4) + (x[,"n_15"] + x[,"n_31"])*log(6 - 5*r_mr - 3*r_mr^2 + 6*r_mr^3 - 4*r_mr^4) + (x[,"n_03"] + x[,"n_43"])*log(9 + 4*r_mr - 8*r_mr^2 + 8*r_mr^3 - 4*r_mr^4) + (x[,"n_04"] + x[,"n_42"])*log(12 - 16*r_mr + 12*r_mr^2 - 9*r_mr^3 + r_mr^4) + (x[,"n_21"] + x[,"n_25"])*log(3 - r_mr + 4*r_mr^2 - 6*r_mr^3 + 3*r_mr^4) + (x[,"n_22"] + x[,"n_24"])*log(33 - 2*r_mr + 5*r_mr^2 - 6*r_mr^3 + 3*r_mr^4) + (x[,"n_13"] + x[,"n_33"])*log(15 - 2*r_mr + 6*r_mr^2 - 8*r_mr^3 + 4*r_mr^4)

logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
L <- (-4*n00 - 3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 4*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 3*n20 - 2*n21 - 3*n22 - 2*n23 - 3*n24 - 2*n25 - 3*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 4*n46)*log(2) + (n20 + n26)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46)*log(5) + 4*(n06 + n40)*log(pmax(1e-6,1 - r)) + 4*(n00 + n46)*log(pmax(1e-6,r)) + (n16 + n30)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n20 + n26)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n10 + n36)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n45)*(log(4 - r) + 3*log(pmax(1e-6,r))) + (n02 + n44)*(2*log(pmax(1e-6,r)) + log(22 - 12*r - r^2)) + (n05 + n41)*(2*log(pmax(1e-6,1 - r)) + log(3 - 2*r - r^2)) + (n04 + n42)*(2*log(pmax(1e-6,1 - r)) + log(9 + 14*r - r^2)) + (n11 + n35)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n15 + n31)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n21 + n25)*(log(pmax(1e-6,r)) + log(8 - 11*r + 6*r^2 - 3*r^3)) + (n03 + n43)*(log(pmax(1e-6,r)) + log(6 - 5*r - 2*r^2 + r^3)) + (n12 + n34)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n13 + n33)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n22 + n24)*log(21 + 4*r - 7*r^2 + 6*r^3 - 3*r^4) + (n14 + n32)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + n23*log(23 - 20*r + 26*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"])
LOD_rr <- 4*(x[,"n_20"] + x[,"n_26"])*log10(2) + 4*(x[,"n_16"] + x[,"n_30"])*log10(2) + 4*(x[,"n_10"] + x[,"n_36"])*log10(2) + 4*(x[,"n_06"] + x[,"n_40"])*log10(2) + 4*(x[,"n_00"] + x[,"n_46"])*log10(2) - (x[,"n_15"] + x[,"n_31"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_35"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_43"])*(-4*log10(2) + 2*log10(5)) - (x[,"n_05"] + x[,"n_41"])*(-4*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_45"])*(-4*log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_42"])*(-4*log10(2) + 2*log10(3) + log10(7)) - (x[,"n_02"] + x[,"n_44"])*(-4*log10(2) + 2*log10(3) + log10(7)) - x[,"n_23"]*(-3*log10(2) + log10(3) + 2*log10(7)) - (x[,"n_13"] + x[,"n_33"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_14"] + x[,"n_32"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_12"] + x[,"n_34"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_21"] + x[,"n_25"])*(-4*log10(2) + log10(29)) - (x[,"n_22"] + x[,"n_24"])*(-4*log10(2) + log10(349)) + 4*(x[,"n_06"] + x[,"n_40"])*log10(pmax(1e-6,1 - r_rr)) + 4*(x[,"n_00"] + x[,"n_46"])*log10(pmax(1e-6,r_rr)) + (x[,"n_16"] + x[,"n_30"])*(3*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr))) + (x[,"n_20"] + x[,"n_26"])*(2*log10(pmax(1e-6,1 - r_rr)) + 2*log10(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_36"])*(log10(pmax(1e-6,1 - r_rr)) + 3*log10(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_45"])*(log10(4 - r_rr) + 3*log10(pmax(1e-6,r_rr))) + (x[,"n_02"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(22 - 12*r_rr - r_rr^2)) + (x[,"n_05"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 - 2*r_rr - r_rr^2)) + (x[,"n_04"] + x[,"n_42"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(9 + 14*r_rr - r_rr^2)) + (x[,"n_11"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_rr)) + log10(7 - 6*r_rr + 2*r_rr^2)) + (x[,"n_15"] + x[,"n_31"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(3 + 2*r_rr + 2*r_rr^2)) + (x[,"n_21"] + x[,"n_25"])*(log10(pmax(1e-6,r_rr)) + log10(8 - 11*r_rr + 6*r_rr^2 - 3*r_rr^3)) + (x[,"n_03"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(6 - 5*r_rr - 2*r_rr^2 + r_rr^3)) + (x[,"n_12"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(15 - 9*r_rr + 5*r_rr^2 + r_rr^3)) + (x[,"n_13"] + x[,"n_33"])*log10(9 + 4*r_rr - 8*r_rr^2 + 8*r_rr^3 - 4*r_rr^4) + (x[,"n_22"] + x[,"n_24"])*log10(21 + 4*r_rr - 7*r_rr^2 + 6*r_rr^3 - 3*r_rr^4) + (x[,"n_14"] + x[,"n_32"])*log10(12 - 16*r_rr + 12*r_rr^2 - 9*r_rr^3 + r_rr^4) + x[,"n_23"]*log10(23 - 20*r_rr + 26*r_rr^2 - 12*r_rr^3 + 6*r_rr^4)
logL_rr <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 3*x[,"n_20"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 2*x[,"n_25"] - 3*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"])*log(2) + (x[,"n_20"] + x[,"n_26"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"])*log(5) + 4*(x[,"n_06"] + x[,"n_40"])*log(pmax(1e-6,1 - r_rr)) + 4*(x[,"n_00"] + x[,"n_46"])*log(pmax(1e-6,r_rr)) + (x[,"n_16"] + x[,"n_30"])*(3*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr))) + (x[,"n_20"] + x[,"n_26"])*(2*log(pmax(1e-6,1 - r_rr)) + 2*log(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_36"])*(log(pmax(1e-6,1 - r_rr)) + 3*log(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_45"])*(log(4 - r_rr) + 3*log(pmax(1e-6,r_rr))) + (x[,"n_02"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(22 - 12*r_rr - r_rr^2)) + (x[,"n_05"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 - 2*r_rr - r_rr^2)) + (x[,"n_04"] + x[,"n_42"])*(2*log(pmax(1e-6,1 - r_rr)) + log(9 + 14*r_rr - r_rr^2)) + (x[,"n_11"] + x[,"n_35"])*(2*log(pmax(1e-6,r_rr)) + log(7 - 6*r_rr + 2*r_rr^2)) + (x[,"n_15"] + x[,"n_31"])*(2*log(pmax(1e-6,1 - r_rr)) + log(3 + 2*r_rr + 2*r_rr^2)) + (x[,"n_21"] + x[,"n_25"])*(log(pmax(1e-6,r_rr)) + log(8 - 11*r_rr + 6*r_rr^2 - 3*r_rr^3)) + (x[,"n_03"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(6 - 5*r_rr - 2*r_rr^2 + r_rr^3)) + (x[,"n_12"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(15 - 9*r_rr + 5*r_rr^2 + r_rr^3)) + (x[,"n_13"] + x[,"n_33"])*log(9 + 4*r_rr - 8*r_rr^2 + 8*r_rr^3 - 4*r_rr^4) + (x[,"n_22"] + x[,"n_24"])*log(21 + 4*r_rr - 7*r_rr^2 + 6*r_rr^3 - 3*r_rr^4) + (x[,"n_14"] + x[,"n_32"])*log(12 - 16*r_rr + 12*r_rr^2 - 9*r_rr^3 + r_rr^4) + x[,"n_23"]*log(23 - 20*r_rr + 26*r_rr^2 - 12*r_rr^3 + 6*r_rr^4)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_mm,r_mr,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_mm,LOD_mr,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_mm,logL_mr,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","mixed mixed","mixed repulsion","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.3_2.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-2*n00 - 2*n01 - n02 - n03 - 2*n04 - 2*n05 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - n25 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n50 - 2*n51 - n52 - n53 - 2*n54 - 2*n55)*log(2) + (-2*n00 - n01 - n02 - n03 - n04 - 2*n05 - n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - n15 - n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - n25 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - n35 - n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - n45 - 2*n50 - n51 - n52 - n53 - n54 - 2*n55)*log(5) + 5*(n00 + n55)*log(pmax(1e-6,1 - r)) + 5*(n05 + n50)*log(pmax(1e-6,r)) + (n01 + n10 + n45 + n54)*(4*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n20 + n35 + n53)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n03 + n25 + n30 + n52)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n04 + n15 + n40 + n51)*(log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n14 + n41)*(3*log(pmax(1e-6,r)) + log(27 - 40*r + 25*r^2)) + (n11 + n44)*(3*log(pmax(1e-6,1 - r)) + log(12 - 10*r + 25*r^2)) + (n12 + n21 + n34 + n43)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(41 - 40*r + 50*r^2)) + (n13 + n24 + n31 + n42)*(2*log(pmax(1e-6,r)) + log(51 - 111*r + 110*r^2 - 50*r^3)) + (n23 + n32)*(log(pmax(1e-6,r)) + log(78 - 204*r + 303*r^2 - 240*r^3 + 100*r^4)) + (n22 + n33)*log(37 - 119*r + 265*r^2 - 343*r^3 + 260*r^4 - 100*r^5)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_cc <- 5*(x[,"n_05"] + x[,"n_50"])*log10(2) + 5*(x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*log10(2) + 5*(x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log10(2) + 5*(x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log10(2) + 5*(x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*log10(2) + 5*(x[,"n_00"] + x[,"n_55"])*log10(2) - (x[,"n_23"] + x[,"n_32"])*(log10(2) + log10(7)) - (x[,"n_22"] + x[,"n_33"])*(log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + log10(53)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + log10(53)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(67)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(67)) + 5*(x[,"n_00"] + x[,"n_55"])*log10(pmax(1e-6,1 - r_cc)) + 5*(x[,"n_05"] + x[,"n_50"])*log10(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(4*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(3*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(log10(pmax(1e-6,1 - r_cc)) + 4*log10(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_41"])*(3*log10(pmax(1e-6,r_cc)) + log10(27 - 40*r_cc + 25*r_cc^2)) + (x[,"n_11"] + x[,"n_44"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(12 - 10*r_cc + 25*r_cc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc)) + log10(41 - 40*r_cc + 50*r_cc^2)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(51 - 111*r_cc + 110*r_cc^2 - 50*r_cc^3)) + (x[,"n_23"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(78 - 204*r_cc + 303*r_cc^2 - 240*r_cc^3 + 100*r_cc^4)) + (x[,"n_22"] + x[,"n_33"])*log10(37 - 119*r_cc + 265*r_cc^2 - 343*r_cc^3 + 260*r_cc^4 - 100*r_cc^5)
logL_cc <- (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_50"] - 2*x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"])*log(2) + (-2*x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - x[,"n_45"] - 2*x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - 2*x[,"n_55"])*log(5) + 5*(x[,"n_00"] + x[,"n_55"])*log(pmax(1e-6,1 - r_cc)) + 5*(x[,"n_05"] + x[,"n_50"])*log(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(4*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(3*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(log(pmax(1e-6,1 - r_cc)) + 4*log(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_41"])*(3*log(pmax(1e-6,r_cc)) + log(27 - 40*r_cc + 25*r_cc^2)) + (x[,"n_11"] + x[,"n_44"])*(3*log(pmax(1e-6,1 - r_cc)) + log(12 - 10*r_cc + 25*r_cc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc)) + log(41 - 40*r_cc + 50*r_cc^2)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(51 - 111*r_cc + 110*r_cc^2 - 50*r_cc^3)) + (x[,"n_23"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(78 - 204*r_cc + 303*r_cc^2 - 240*r_cc^3 + 100*r_cc^4)) + (x[,"n_22"] + x[,"n_33"])*log(37 - 119*r_cc + 265*r_cc^2 - 343*r_cc^3 + 260*r_cc^4 - 100*r_cc^5)
logL_cm2c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-2*n00 - 2*n01 - n02 - n03 - 2*n04 - 2*n05 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - n25 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - n35 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n50 - 2*n51 - n52 - n53 - 2*n54 - 2*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n00 + n55)*(4*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n50)*(log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n04 + n15 + n40 + n51)*(3*log(pmax(1e-6,r)) + log(6 - 8*r + 5*r^2)) + (n02 + n20 + n35 + n53)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(5 - 4*r + 5*r^2)) + (n01 + n10 + n45 + n54)*(3*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 5*r^2)) + (n14 + n41)*(2*log(pmax(1e-6,r)) + log(31 - 43*r + 49*r^2 - 25*r^3)) + (n03 + n25 + n30 + n52)*(2*log(pmax(1e-6,r)) + log(6 - 12*r + 11*r^2 - 5*r^3)) + (n11 + n44)*(2*log(pmax(1e-6,1 - r)) + log(12 + 20*r - 26*r^2 + 25*r^3)) + (n13 + n24 + n31 + n42)*(log(pmax(1e-6,r)) + log(46 - 76*r + 117*r^2 - 116*r^3 + 50*r^4)) + (n23 + n32)*log(42 - 34*r + 139*r^2 - 243*r^3 + 244*r^4 - 100*r^5) + (n12 + n21 + n34 + n43)*log(21 - 31*r + 79*r^2 - 153*r^3 + 134*r^4 - 50*r^5) + (n22 + n33)*log(48 + 9*r - 126*r^2 + 267*r^3 - 256*r^4 + 100*r^5)
return(L)}
interlogL_cm2c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_cm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_cm2c <- 5*(x[,"n_05"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_55"])*log10(2) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-5*log10(2) + log10(17)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-5*log10(2) + log10(17)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + 2*log10(3) + log10(23)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + 2*log10(3) + log10(23)) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(83)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(83)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + log10(149)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + log10(149)) + (x[,"n_00"] + x[,"n_55"])*(4*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_50"])*(log10(pmax(1e-6,1 - r_cm2c)) + 4*log10(pmax(1e-6,r_cm2c))) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(3*log10(pmax(1e-6,r_cm2c)) + log10(6 - 8*r_cm2c + 5*r_cm2c^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c)) + log10(5 - 4*r_cm2c + 5*r_cm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 - 2*r_cm2c + 5*r_cm2c^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(31 - 43*r_cm2c + 49*r_cm2c^2 - 25*r_cm2c^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(6 - 12*r_cm2c + 11*r_cm2c^2 - 5*r_cm2c^3)) + (x[,"n_11"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(12 + 20*r_cm2c - 26*r_cm2c^2 + 25*r_cm2c^3)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm2c)) + log10(46 - 76*r_cm2c + 117*r_cm2c^2 - 116*r_cm2c^3 + 50*r_cm2c^4)) + (x[,"n_23"] + x[,"n_32"])*log10(42 - 34*r_cm2c + 139*r_cm2c^2 - 243*r_cm2c^3 + 244*r_cm2c^4 - 100*r_cm2c^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(21 - 31*r_cm2c + 79*r_cm2c^2 - 153*r_cm2c^3 + 134*r_cm2c^4 - 50*r_cm2c^5) + (x[,"n_22"] + x[,"n_33"])*log10(48 + 9*r_cm2c - 126*r_cm2c^2 + 267*r_cm2c^3 - 256*r_cm2c^4 + 100*r_cm2c^5)
logL_cm2c <- (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_50"] - 2*x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_55"])*(4*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_50"])*(log(pmax(1e-6,1 - r_cm2c)) + 4*log(pmax(1e-6,r_cm2c))) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(3*log(pmax(1e-6,r_cm2c)) + log(6 - 8*r_cm2c + 5*r_cm2c^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c)) + log(5 - 4*r_cm2c + 5*r_cm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(3 - 2*r_cm2c + 5*r_cm2c^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm2c)) + log(31 - 43*r_cm2c + 49*r_cm2c^2 - 25*r_cm2c^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log(pmax(1e-6,r_cm2c)) + log(6 - 12*r_cm2c + 11*r_cm2c^2 - 5*r_cm2c^3)) + (x[,"n_11"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(12 + 20*r_cm2c - 26*r_cm2c^2 + 25*r_cm2c^3)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log(pmax(1e-6,r_cm2c)) + log(46 - 76*r_cm2c + 117*r_cm2c^2 - 116*r_cm2c^3 + 50*r_cm2c^4)) + (x[,"n_23"] + x[,"n_32"])*log(42 - 34*r_cm2c + 139*r_cm2c^2 - 243*r_cm2c^3 + 244*r_cm2c^4 - 100*r_cm2c^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(21 - 31*r_cm2c + 79*r_cm2c^2 - 153*r_cm2c^3 + 134*r_cm2c^4 - 50*r_cm2c^5) + (x[,"n_22"] + x[,"n_33"])*log(48 + 9*r_cm2c - 126*r_cm2c^2 + 267*r_cm2c^3 - 256*r_cm2c^4 + 100*r_cm2c^5)
logL_cm1c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(2*log(2) + log(3) + 2*log(5)) + (n00 + n55)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n05 + n50)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n10 + n45 + n54)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(4 - 4*r + 5*r^2)) + (n04 + n15 + n40 + n51)*(2*log(pmax(1e-6,r)) + log(5 - 11*r + 11*r^2 - 5*r^3)) + (n03 + n25 + n30 + n52)*(log(pmax(1e-6,r)) + log(7 - 20*r + 30*r^2 - 24*r^3 + 10*r^4)) + (n14 + n41)*(log(pmax(1e-6,r)) + log(15 - 30*r + 60*r^2 - 58*r^3 + 25*r^4)) + (n22 + n33)*log(48 - 40*r + 131*r^2 - 255*r^3 + 252*r^4 - 100*r^5) + (n13 + n24 + n31 + n42)*log(12 - 14*r + 73*r^2 - 131*r^3 + 122*r^4 - 50*r^5) + (n11 + n44)*log(12 - 28*r + 52*r^2 - 78*r^3 + 67*r^4 - 25*r^5) + (n02 + n20 + n35 + n53)*log(3 - 11*r + 26*r^2 - 34*r^3 + 26*r^4 - 10*r^5) + (n12 + n21 + n34 + n43)*log(12 + 23*r - 88*r^2 + 143*r^3 - 128*r^4 + 50*r^5) + (n23 + n32)*log(36 + 35*r - 122*r^2 + 247*r^3 - 248*r^4 + 100*r^5)
return(L)}
interlogL_cm1c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_cm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_cm1c <- 5*(x[,"n_05"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_55"])*log10(2) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(17)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(17)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + 2*log10(3) + log10(23)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + 2*log10(3) + log10(23)) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(83)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(83)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + log10(149)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + log10(149)) + (x[,"n_00"] + x[,"n_55"])*(3*log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_05"] + x[,"n_50"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + 3*log10(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + log10(pmax(1e-6,r_cm1c)) + log10(4 - 4*r_cm1c + 5*r_cm1c^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_cm1c)) + log10(5 - 11*r_cm1c + 11*r_cm1c^2 - 5*r_cm1c^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(log10(pmax(1e-6,r_cm1c)) + log10(7 - 20*r_cm1c + 30*r_cm1c^2 - 24*r_cm1c^3 + 10*r_cm1c^4)) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm1c)) + log10(15 - 30*r_cm1c + 60*r_cm1c^2 - 58*r_cm1c^3 + 25*r_cm1c^4)) + (x[,"n_22"] + x[,"n_33"])*log10(48 - 40*r_cm1c + 131*r_cm1c^2 - 255*r_cm1c^3 + 252*r_cm1c^4 - 100*r_cm1c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(12 - 14*r_cm1c + 73*r_cm1c^2 - 131*r_cm1c^3 + 122*r_cm1c^4 - 50*r_cm1c^5) + (x[,"n_11"] + x[,"n_44"])*log10(12 - 28*r_cm1c + 52*r_cm1c^2 - 78*r_cm1c^3 + 67*r_cm1c^4 - 25*r_cm1c^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log10(3 - 11*r_cm1c + 26*r_cm1c^2 - 34*r_cm1c^3 + 26*r_cm1c^4 - 10*r_cm1c^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(12 + 23*r_cm1c - 88*r_cm1c^2 + 143*r_cm1c^3 - 128*r_cm1c^4 + 50*r_cm1c^5) + (x[,"n_23"] + x[,"n_32"])*log10(36 + 35*r_cm1c - 122*r_cm1c^2 + 247*r_cm1c^3 - 248*r_cm1c^4 + 100*r_cm1c^5)
logL_cm1c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(2*log(2) + log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_55"])*(3*log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_05"] + x[,"n_50"])*(2*log(pmax(1e-6,1 - r_cm1c)) + 3*log(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_cm1c)) + log(pmax(1e-6,r_cm1c)) + log(4 - 4*r_cm1c + 5*r_cm1c^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log(pmax(1e-6,r_cm1c)) + log(5 - 11*r_cm1c + 11*r_cm1c^2 - 5*r_cm1c^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(log(pmax(1e-6,r_cm1c)) + log(7 - 20*r_cm1c + 30*r_cm1c^2 - 24*r_cm1c^3 + 10*r_cm1c^4)) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,r_cm1c)) + log(15 - 30*r_cm1c + 60*r_cm1c^2 - 58*r_cm1c^3 + 25*r_cm1c^4)) + (x[,"n_22"] + x[,"n_33"])*log(48 - 40*r_cm1c + 131*r_cm1c^2 - 255*r_cm1c^3 + 252*r_cm1c^4 - 100*r_cm1c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(12 - 14*r_cm1c + 73*r_cm1c^2 - 131*r_cm1c^3 + 122*r_cm1c^4 - 50*r_cm1c^5) + (x[,"n_11"] + x[,"n_44"])*log(12 - 28*r_cm1c + 52*r_cm1c^2 - 78*r_cm1c^3 + 67*r_cm1c^4 - 25*r_cm1c^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log(3 - 11*r_cm1c + 26*r_cm1c^2 - 34*r_cm1c^3 + 26*r_cm1c^4 - 10*r_cm1c^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(12 + 23*r_cm1c - 88*r_cm1c^2 + 143*r_cm1c^3 - 128*r_cm1c^4 + 50*r_cm1c^5) + (x[,"n_23"] + x[,"n_32"])*log(36 + 35*r_cm1c - 122*r_cm1c^2 + 247*r_cm1c^3 - 248*r_cm1c^4 + 100*r_cm1c^5)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(2*log(2) + 2*log(5)) + (n05 + n50)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n00 + n55)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n10 + n45 + n54)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(3 - 6*r + 5*r^2)) + (n04 + n15 + n40 + n51)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 - 4*r + 5*r^2)) + (n02 + n20 + n35 + n53)*(log(pmax(1e-6,r)) + log(3 - 12*r + 24*r^2 - 24*r^3 + 10*r^4)) + (n11 + n44)*(log(pmax(1e-6,r)) + log(12 - 36*r + 60*r^2 - 58*r^3 + 25*r^4)) + (n13 + n24 + n31 + n42)*(log(pmax(1e-6,r)) + log(32 - 98*r + 153*r^2 - 128*r^3 + 50*r^4)) + (n22 + n33)*(log(pmax(1e-6,r)) + log(84 - 216*r + 307*r^2 - 248*r^3 + 100*r^4)) + (n23 + n32)*log(27 - 81*r + 217*r^2 - 315*r^3 + 252*r^4 - 100*r^5) + (n12 + n21 + n34 + n43)*log(9 - 33*r + 93*r^2 - 141*r^3 + 122*r^4 - 50*r^5) + (n14 + n41)*log(3 - 13*r + 46*r^2 - 78*r^3 + 67*r^4 - 25*r^5) + (n03 + n25 + n30 + n52)*log(1 - 5*r + 16*r^2 - 28*r^3 + 26*r^4 - 10*r^5)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_cr <- 5*(x[,"n_05"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_55"])*log10(2) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(4*log10(2) - log10(5)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_32"])*(log10(2) + log10(7)) - (x[,"n_22"] + x[,"n_33"])*(log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + log10(53)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + log10(53)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(67)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(67)) + (x[,"n_05"] + x[,"n_50"])*(3*log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_cr)) + 3*log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr)) + log10(3 - 6*r_cr + 5*r_cr^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 5*r_cr^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(log10(pmax(1e-6,r_cr)) + log10(3 - 12*r_cr + 24*r_cr^2 - 24*r_cr^3 + 10*r_cr^4)) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_cr)) + log10(12 - 36*r_cr + 60*r_cr^2 - 58*r_cr^3 + 25*r_cr^4)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log10(pmax(1e-6,r_cr)) + log10(32 - 98*r_cr + 153*r_cr^2 - 128*r_cr^3 + 50*r_cr^4)) + (x[,"n_22"] + x[,"n_33"])*(log10(pmax(1e-6,r_cr)) + log10(84 - 216*r_cr + 307*r_cr^2 - 248*r_cr^3 + 100*r_cr^4)) + (x[,"n_23"] + x[,"n_32"])*log10(27 - 81*r_cr + 217*r_cr^2 - 315*r_cr^3 + 252*r_cr^4 - 100*r_cr^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(9 - 33*r_cr + 93*r_cr^2 - 141*r_cr^3 + 122*r_cr^4 - 50*r_cr^5) + (x[,"n_14"] + x[,"n_41"])*log10(3 - 13*r_cr + 46*r_cr^2 - 78*r_cr^3 + 67*r_cr^4 - 25*r_cr^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log10(1 - 5*r_cr + 16*r_cr^2 - 28*r_cr^3 + 26*r_cr^4 - 10*r_cr^5)
logL_cr <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(2*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_50"])*(3*log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_cr)) + 3*log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr)) + log(3 - 6*r_cr + 5*r_cr^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 5*r_cr^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(log(pmax(1e-6,r_cr)) + log(3 - 12*r_cr + 24*r_cr^2 - 24*r_cr^3 + 10*r_cr^4)) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_cr)) + log(12 - 36*r_cr + 60*r_cr^2 - 58*r_cr^3 + 25*r_cr^4)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log(pmax(1e-6,r_cr)) + log(32 - 98*r_cr + 153*r_cr^2 - 128*r_cr^3 + 50*r_cr^4)) + (x[,"n_22"] + x[,"n_33"])*(log(pmax(1e-6,r_cr)) + log(84 - 216*r_cr + 307*r_cr^2 - 248*r_cr^3 + 100*r_cr^4)) + (x[,"n_23"] + x[,"n_32"])*log(27 - 81*r_cr + 217*r_cr^2 - 315*r_cr^3 + 252*r_cr^4 - 100*r_cr^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(9 - 33*r_cr + 93*r_cr^2 - 141*r_cr^3 + 122*r_cr^4 - 50*r_cr^5) + (x[,"n_14"] + x[,"n_41"])*log(3 - 13*r_cr + 46*r_cr^2 - 78*r_cr^3 + 67*r_cr^4 - 25*r_cr^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log(1 - 5*r_cr + 16*r_cr^2 - 28*r_cr^3 + 26*r_cr^4 - 10*r_cr^5)
logL_mc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 4*n55)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*log(5) + (n05 + n50)*(log(2 - r) + 4*log(pmax(1e-6,r))) + (n00 + n55)*(3*log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n04 + n15 + n40 + n51)*(3*log(pmax(1e-6,r)) + log(9 - 11*r + 5*r^2)) + (n01 + n10 + n45 + n54)*(3*log(pmax(1e-6,1 - r)) + log(3 + r + 5*r^2)) + (n02 + n20 + n35 + n53)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(11 - 6*r + 10*r^2)) + (n14 + n41)*(2*log(pmax(1e-6,r)) + log(42 - 60*r + 58*r^2 - 25*r^3)) + (n03 + n25 + n30 + n52)*(2*log(pmax(1e-6,r)) + log(15 - 29*r + 24*r^2 - 10*r^3)) + (n11 + n44)*(2*log(pmax(1e-6,1 - r)) + log(15 + 19*r - 17*r^2 + 25*r^3)) + (n13 + n24 + n31 + n42)*(log(pmax(1e-6,r)) + log(63 - 102*r + 141*r^2 - 122*r^3 + 50*r^4)) + (n23 + n32)*log(27 - 24*r + 96*r^2 - 141*r^3 + 124*r^4 - 50*r^5) + (n12 + n21 + n34 + n43)*log(30 - 44*r + 89*r^2 - 153*r^3 + 128*r^4 - 50*r^5) + (n22 + n33)*log(32 + 9*r - 83*r^2 + 145*r^3 - 126*r^4 + 50*r^5)
return(L)}
interlogL_mc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_mc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mc <- -((x[,"n_05"] + x[,"n_50"])*(-5*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_55"])*(-5*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(3) + log10(7)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(3) + log10(7)) - (x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + 2*log10(3) + 2*log10(7)) - (x[,"n_22"] + x[,"n_33"])*(-4*log10(2) + 2*log10(3) + 2*log10(7)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + log10(11) + log10(17)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + log10(11) + log10(17)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(19)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(281)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(281)) + (x[,"n_05"] + x[,"n_50"])*(log10(2 - r_mc) + 4*log10(pmax(1e-6,r_mc))) + (x[,"n_00"] + x[,"n_55"])*(3*log10(pmax(1e-6,1 - r_mc)) + log10(1 - r_mc^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(3*log10(pmax(1e-6,r_mc)) + log10(9 - 11*r_mc + 5*r_mc^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(3*log10(pmax(1e-6,1 - r_mc)) + log10(3 + r_mc + 5*r_mc^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(pmax(1e-6,r_mc)) + log10(11 - 6*r_mc + 10*r_mc^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_mc)) + log10(42 - 60*r_mc + 58*r_mc^2 - 25*r_mc^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_mc)) + log10(15 - 29*r_mc + 24*r_mc^2 - 10*r_mc^3)) + (x[,"n_11"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(15 + 19*r_mc - 17*r_mc^2 + 25*r_mc^3)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log10(pmax(1e-6,r_mc)) + log10(63 - 102*r_mc + 141*r_mc^2 - 122*r_mc^3 + 50*r_mc^4)) + (x[,"n_23"] + x[,"n_32"])*log10(27 - 24*r_mc + 96*r_mc^2 - 141*r_mc^3 + 124*r_mc^4 - 50*r_mc^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(30 - 44*r_mc + 89*r_mc^2 - 153*r_mc^3 + 128*r_mc^4 - 50*r_mc^5) + (x[,"n_22"] + x[,"n_33"])*log10(32 + 9*r_mc - 83*r_mc^2 + 145*r_mc^3 - 126*r_mc^4 + 50*r_mc^5)
logL_mc <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_05"] + x[,"n_50"])*(log(2 - r_mc) + 4*log(pmax(1e-6,r_mc))) + (x[,"n_00"] + x[,"n_55"])*(3*log(pmax(1e-6,1 - r_mc)) + log(1 - r_mc^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(3*log(pmax(1e-6,r_mc)) + log(9 - 11*r_mc + 5*r_mc^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(3*log(pmax(1e-6,1 - r_mc)) + log(3 + r_mc + 5*r_mc^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log(pmax(1e-6,1 - r_mc)) + log(pmax(1e-6,r_mc)) + log(11 - 6*r_mc + 10*r_mc^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log(pmax(1e-6,r_mc)) + log(42 - 60*r_mc + 58*r_mc^2 - 25*r_mc^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log(pmax(1e-6,r_mc)) + log(15 - 29*r_mc + 24*r_mc^2 - 10*r_mc^3)) + (x[,"n_11"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_mc)) + log(15 + 19*r_mc - 17*r_mc^2 + 25*r_mc^3)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log(pmax(1e-6,r_mc)) + log(63 - 102*r_mc + 141*r_mc^2 - 122*r_mc^3 + 50*r_mc^4)) + (x[,"n_23"] + x[,"n_32"])*log(27 - 24*r_mc + 96*r_mc^2 - 141*r_mc^3 + 124*r_mc^4 - 50*r_mc^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(30 - 44*r_mc + 89*r_mc^2 - 153*r_mc^3 + 128*r_mc^4 - 50*r_mc^5) + (x[,"n_22"] + x[,"n_33"])*log(32 + 9*r_mc - 83*r_mc^2 + 145*r_mc^3 - 126*r_mc^4 + 50*r_mc^5)
logL_mm2c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 4*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n05 + n50)*(log(pmax(1e-6,1 - r)) + log(2 - r) + 3*log(pmax(1e-6,r))) + (n00 + n55)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 - r^2)) + (n04 + n15 + n40 + n51)*(2*log(pmax(1e-6,r)) + log(11 - 17*r + 14*r^2 - 5*r^3)) + (n01 + n10 + n45 + n54)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r - r^2 + 5*r^3)) + (n03 + n25 + n30 + n52)*(log(pmax(1e-6,r)) + log(18 - 32*r + 39*r^2 - 26*r^3 + 10*r^4)) + (n14 + n41)*(log(pmax(1e-6,r)) + log(48 - 54*r + 78*r^2 - 67*r^3 + 25*r^4)) + (n22 + n33)*log(87 - 15*r + 40*r^2 - 111*r^3 + 124*r^4 - 50*r^5) + (n13 + n24 + n31 + n42)*log(45 - 5*r + 91*r^2 - 143*r^3 + 128*r^4 - 50*r^5) + (n11 + n44)*log(30 - 31*r + 28*r^2 - 60*r^3 + 58*r^4 - 25*r^5) + (n02 + n20 + n35 + n53)*log(9 - 17*r + 29*r^2 - 35*r^3 + 24*r^4 - 10*r^5) + (n23 + n32)*log(75 + 22*r - 49*r^2 + 115*r^3 - 126*r^4 + 50*r^5) + (n12 + n21 + n34 + n43)*log(66 - 10*r - 70*r^2 + 131*r^3 - 122*r^4 + 50*r^5)
return(L)}
interlogL_mm2c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_mm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm2c <- mapply(interlogL_mm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mm2c <- -((x[,"n_05"] + x[,"n_50"])*(-5*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_55"])*(-5*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + 2*log10(7) + log10(11)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + 2*log10(7) + log10(11)) - (x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + log10(7) + log10(11) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-4*log10(2) + log10(7) + log10(11) + log10(17)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(3) + log10(7) + log10(41)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(3) + log10(7) + log10(41)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(43)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(43)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(73)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(73)) + (x[,"n_05"] + x[,"n_50"])*(log10(pmax(1e-6,1 - r_mm2c)) + log10(2 - r_mm2c) + 3*log10(pmax(1e-6,r_mm2c))) + (x[,"n_00"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_mm2c)) + log10(pmax(1e-6,r_mm2c)) + log10(1 - r_mm2c^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_mm2c)) + log10(11 - 17*r_mm2c + 14*r_mm2c^2 - 5*r_mm2c^3)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_mm2c)) + log10(3 + 4*r_mm2c - r_mm2c^2 + 5*r_mm2c^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(log10(pmax(1e-6,r_mm2c)) + log10(18 - 32*r_mm2c + 39*r_mm2c^2 - 26*r_mm2c^3 + 10*r_mm2c^4)) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,r_mm2c)) + log10(48 - 54*r_mm2c + 78*r_mm2c^2 - 67*r_mm2c^3 + 25*r_mm2c^4)) + (x[,"n_22"] + x[,"n_33"])*log10(87 - 15*r_mm2c + 40*r_mm2c^2 - 111*r_mm2c^3 + 124*r_mm2c^4 - 50*r_mm2c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(45 - 5*r_mm2c + 91*r_mm2c^2 - 143*r_mm2c^3 + 128*r_mm2c^4 - 50*r_mm2c^5) + (x[,"n_11"] + x[,"n_44"])*log10(30 - 31*r_mm2c + 28*r_mm2c^2 - 60*r_mm2c^3 + 58*r_mm2c^4 - 25*r_mm2c^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log10(9 - 17*r_mm2c + 29*r_mm2c^2 - 35*r_mm2c^3 + 24*r_mm2c^4 - 10*r_mm2c^5) + (x[,"n_23"] + x[,"n_32"])*log10(75 + 22*r_mm2c - 49*r_mm2c^2 + 115*r_mm2c^3 - 126*r_mm2c^4 + 50*r_mm2c^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(66 - 10*r_mm2c - 70*r_mm2c^2 + 131*r_mm2c^3 - 122*r_mm2c^4 + 50*r_mm2c^5)
logL_mm2c <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_50"])*(log(pmax(1e-6,1 - r_mm2c)) + log(2 - r_mm2c) + 3*log(pmax(1e-6,r_mm2c))) + (x[,"n_00"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_mm2c)) + log(pmax(1e-6,r_mm2c)) + log(1 - r_mm2c^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log(pmax(1e-6,r_mm2c)) + log(11 - 17*r_mm2c + 14*r_mm2c^2 - 5*r_mm2c^3)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_mm2c)) + log(3 + 4*r_mm2c - r_mm2c^2 + 5*r_mm2c^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(log(pmax(1e-6,r_mm2c)) + log(18 - 32*r_mm2c + 39*r_mm2c^2 - 26*r_mm2c^3 + 10*r_mm2c^4)) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,r_mm2c)) + log(48 - 54*r_mm2c + 78*r_mm2c^2 - 67*r_mm2c^3 + 25*r_mm2c^4)) + (x[,"n_22"] + x[,"n_33"])*log(87 - 15*r_mm2c + 40*r_mm2c^2 - 111*r_mm2c^3 + 124*r_mm2c^4 - 50*r_mm2c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(45 - 5*r_mm2c + 91*r_mm2c^2 - 143*r_mm2c^3 + 128*r_mm2c^4 - 50*r_mm2c^5) + (x[,"n_11"] + x[,"n_44"])*log(30 - 31*r_mm2c + 28*r_mm2c^2 - 60*r_mm2c^3 + 58*r_mm2c^4 - 25*r_mm2c^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log(9 - 17*r_mm2c + 29*r_mm2c^2 - 35*r_mm2c^3 + 24*r_mm2c^4 - 10*r_mm2c^5) + (x[,"n_23"] + x[,"n_32"])*log(75 + 22*r_mm2c - 49*r_mm2c^2 + 115*r_mm2c^3 - 126*r_mm2c^4 + 50*r_mm2c^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(66 - 10*r_mm2c - 70*r_mm2c^2 + 131*r_mm2c^3 - 122*r_mm2c^4 + 50*r_mm2c^5)
logL_mm1c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 4*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n05 + n50)*(2*log(pmax(1e-6,1 - r)) + log(2 - r) + 2*log(pmax(1e-6,r))) + (n00 + n55)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(1 - r^2)) + (n01 + n10 + n45 + n54)*(log(pmax(1e-6,r)) + log(4 - r - 6*r^2 + 8*r^3 - 5*r^4)) + (n04 + n15 + n40 + n51)*(log(pmax(1e-6,r)) + log(9 - 21*r + 24*r^2 - 17*r^3 + 5*r^4)) + (n12 + n21 + n34 + n43)*log(57 - 28*r + 79*r^2 - 117*r^3 + 116*r^4 - 50*r^5) + (n23 + n32)*log(81 - 2*r + 42*r^2 - 121*r^3 + 128*r^4 - 50*r^5) + (n14 + n41)*log(18 - 12*r + 52*r^2 - 97*r^3 + 76*r^4 - 25*r^5) + (n03 + n25 + n30 + n52)*log(9 - 20*r + 37*r^2 - 41*r^3 + 28*r^4 - 10*r^5) + (n02 + n20 + n35 + n53)*log(3 + 7*r - 18*r^2 + 29*r^3 - 22*r^4 + 10*r^5) + (n11 + n44)*log(12 + 20*r - 33*r^2 + 43*r^3 - 49*r^4 + 25*r^5) + (n13 + n24 + n31 + n42)*log(57 + 7*r - 76*r^2 + 153*r^3 - 134*r^4 + 50*r^5) + (n22 + n33)*log(78 + 19*r - 53*r^2 + 109*r^3 - 122*r^4 + 50*r^5)
return(L)}
interlogL_mm1c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_mm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm1c <- mapply(interlogL_mm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mm1c <- -((x[,"n_05"] + x[,"n_50"])*(-5*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_55"])*(-5*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + 2*log10(7) + log10(11)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + 2*log10(7) + log10(11)) - (x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + log10(7) + log10(11) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-4*log10(2) + log10(7) + log10(11) + log10(17)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(3) + log10(7) + log10(41)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(3) + log10(7) + log10(41)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(43)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(43)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(73)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(73)) + (x[,"n_05"] + x[,"n_50"])*(2*log10(pmax(1e-6,1 - r_mm1c)) + log10(2 - r_mm1c) + 2*log10(pmax(1e-6,r_mm1c))) + (x[,"n_00"] + x[,"n_55"])*(log10(pmax(1e-6,1 - r_mm1c)) + 2*log10(pmax(1e-6,r_mm1c)) + log10(1 - r_mm1c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(log10(pmax(1e-6,r_mm1c)) + log10(4 - r_mm1c - 6*r_mm1c^2 + 8*r_mm1c^3 - 5*r_mm1c^4)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(log10(pmax(1e-6,r_mm1c)) + log10(9 - 21*r_mm1c + 24*r_mm1c^2 - 17*r_mm1c^3 + 5*r_mm1c^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(57 - 28*r_mm1c + 79*r_mm1c^2 - 117*r_mm1c^3 + 116*r_mm1c^4 - 50*r_mm1c^5) + (x[,"n_23"] + x[,"n_32"])*log10(81 - 2*r_mm1c + 42*r_mm1c^2 - 121*r_mm1c^3 + 128*r_mm1c^4 - 50*r_mm1c^5) + (x[,"n_14"] + x[,"n_41"])*log10(18 - 12*r_mm1c + 52*r_mm1c^2 - 97*r_mm1c^3 + 76*r_mm1c^4 - 25*r_mm1c^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log10(9 - 20*r_mm1c + 37*r_mm1c^2 - 41*r_mm1c^3 + 28*r_mm1c^4 - 10*r_mm1c^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log10(3 + 7*r_mm1c - 18*r_mm1c^2 + 29*r_mm1c^3 - 22*r_mm1c^4 + 10*r_mm1c^5) + (x[,"n_11"] + x[,"n_44"])*log10(12 + 20*r_mm1c - 33*r_mm1c^2 + 43*r_mm1c^3 - 49*r_mm1c^4 + 25*r_mm1c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(57 + 7*r_mm1c - 76*r_mm1c^2 + 153*r_mm1c^3 - 134*r_mm1c^4 + 50*r_mm1c^5) + (x[,"n_22"] + x[,"n_33"])*log10(78 + 19*r_mm1c - 53*r_mm1c^2 + 109*r_mm1c^3 - 122*r_mm1c^4 + 50*r_mm1c^5)
logL_mm1c <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_50"])*(2*log(pmax(1e-6,1 - r_mm1c)) + log(2 - r_mm1c) + 2*log(pmax(1e-6,r_mm1c))) + (x[,"n_00"] + x[,"n_55"])*(log(pmax(1e-6,1 - r_mm1c)) + 2*log(pmax(1e-6,r_mm1c)) + log(1 - r_mm1c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(log(pmax(1e-6,r_mm1c)) + log(4 - r_mm1c - 6*r_mm1c^2 + 8*r_mm1c^3 - 5*r_mm1c^4)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(log(pmax(1e-6,r_mm1c)) + log(9 - 21*r_mm1c + 24*r_mm1c^2 - 17*r_mm1c^3 + 5*r_mm1c^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(57 - 28*r_mm1c + 79*r_mm1c^2 - 117*r_mm1c^3 + 116*r_mm1c^4 - 50*r_mm1c^5) + (x[,"n_23"] + x[,"n_32"])*log(81 - 2*r_mm1c + 42*r_mm1c^2 - 121*r_mm1c^3 + 128*r_mm1c^4 - 50*r_mm1c^5) + (x[,"n_14"] + x[,"n_41"])*log(18 - 12*r_mm1c + 52*r_mm1c^2 - 97*r_mm1c^3 + 76*r_mm1c^4 - 25*r_mm1c^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log(9 - 20*r_mm1c + 37*r_mm1c^2 - 41*r_mm1c^3 + 28*r_mm1c^4 - 10*r_mm1c^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log(3 + 7*r_mm1c - 18*r_mm1c^2 + 29*r_mm1c^3 - 22*r_mm1c^4 + 10*r_mm1c^5) + (x[,"n_11"] + x[,"n_44"])*log(12 + 20*r_mm1c - 33*r_mm1c^2 + 43*r_mm1c^3 - 49*r_mm1c^4 + 25*r_mm1c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(57 + 7*r_mm1c - 76*r_mm1c^2 + 153*r_mm1c^3 - 134*r_mm1c^4 + 50*r_mm1c^5) + (x[,"n_22"] + x[,"n_33"])*log(78 + 19*r_mm1c - 53*r_mm1c^2 + 109*r_mm1c^3 - 122*r_mm1c^4 + 50*r_mm1c^5)
logL_mr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 4*n55)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*log(5) + (n05 + n50)*(3*log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n00 + n55)*(3*log(pmax(1e-6,r)) + log(1 - r^2)) + (n04 + n15 + n40 + n51)*(2*log(pmax(1e-6,1 - r)) + log(3 - 5*r + 10*r^2 - 5*r^3)) + (n01 + n10 + n45 + n54)*(2*log(pmax(1e-6,r)) + log(3 - 5*r^2 + 5*r^3)) + (n11 + n44)*(log(pmax(1e-6,r)) + log(12 + 6*r - 27*r^2 + 40*r^3 - 25*r^4)) + (n02 + n20 + n35 + n53)*(log(pmax(1e-6,r)) + log(3 + 3*r - 15*r^2 + 20*r^3 - 10*r^4)) + (n22 + n33)*log(27 - 21*r + 90*r^2 - 139*r^3 + 120*r^4 - 50*r^5) + (n13 + n24 + n31 + n42)*log(30 - 53*r + 113*r^2 - 171*r^3 + 140*r^4 - 50*r^5) + (n03 + n25 + n30 + n52)*log(1 + 6*r - 22*r^2 + 35*r^3 - 30*r^4 + 10*r^5) + (n14 + n41)*log(6 + 22*r - 85*r^2 + 117*r^3 - 85*r^4 + 25*r^5) + (n23 + n32)*log(27 + 28*r - 107*r^2 + 159*r^3 - 130*r^4 + 50*r^5) + (n12 + n21 + n34 + n43)*log(9 + 30*r - 60*r^2 + 111*r^3 - 110*r^4 + 50*r^5)
return(L)}
interlogL_mr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_mr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mr <- mapply(interlogL_mr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mr <- -((x[,"n_05"] + x[,"n_50"])*(-5*log10(2) + log10(3))) - (x[,"n_00"] + x[,"n_55"])*(-5*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(3) + log10(7)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(3) + log10(7)) - (x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + 2*log10(3) + 2*log10(7)) - (x[,"n_22"] + x[,"n_33"])*(-4*log10(2) + 2*log10(3) + 2*log10(7)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + log10(11) + log10(17)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + log10(11) + log10(17)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(19)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(281)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(281)) + (x[,"n_05"] + x[,"n_50"])*(3*log10(pmax(1e-6,1 - r_mr)) + log10(2 - r_mr) + log10(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_55"])*(3*log10(pmax(1e-6,r_mr)) + log10(1 - r_mr^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(3 - 5*r_mr + 10*r_mr^2 - 5*r_mr^3)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(2*log10(pmax(1e-6,r_mr)) + log10(3 - 5*r_mr^2 + 5*r_mr^3)) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_mr)) + log10(12 + 6*r_mr - 27*r_mr^2 + 40*r_mr^3 - 25*r_mr^4)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(log10(pmax(1e-6,r_mr)) + log10(3 + 3*r_mr - 15*r_mr^2 + 20*r_mr^3 - 10*r_mr^4)) + (x[,"n_22"] + x[,"n_33"])*log10(27 - 21*r_mr + 90*r_mr^2 - 139*r_mr^3 + 120*r_mr^4 - 50*r_mr^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(30 - 53*r_mr + 113*r_mr^2 - 171*r_mr^3 + 140*r_mr^4 - 50*r_mr^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log10(1 + 6*r_mr - 22*r_mr^2 + 35*r_mr^3 - 30*r_mr^4 + 10*r_mr^5) + (x[,"n_14"] + x[,"n_41"])*log10(6 + 22*r_mr - 85*r_mr^2 + 117*r_mr^3 - 85*r_mr^4 + 25*r_mr^5) + (x[,"n_23"] + x[,"n_32"])*log10(27 + 28*r_mr - 107*r_mr^2 + 159*r_mr^3 - 130*r_mr^4 + 50*r_mr^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(9 + 30*r_mr - 60*r_mr^2 + 111*r_mr^3 - 110*r_mr^4 + 50*r_mr^5)
logL_mr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_05"] + x[,"n_50"])*(3*log(pmax(1e-6,1 - r_mr)) + log(2 - r_mr) + log(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_55"])*(3*log(pmax(1e-6,r_mr)) + log(1 - r_mr^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_mr)) + log(3 - 5*r_mr + 10*r_mr^2 - 5*r_mr^3)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(2*log(pmax(1e-6,r_mr)) + log(3 - 5*r_mr^2 + 5*r_mr^3)) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_mr)) + log(12 + 6*r_mr - 27*r_mr^2 + 40*r_mr^3 - 25*r_mr^4)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(log(pmax(1e-6,r_mr)) + log(3 + 3*r_mr - 15*r_mr^2 + 20*r_mr^3 - 10*r_mr^4)) + (x[,"n_22"] + x[,"n_33"])*log(27 - 21*r_mr + 90*r_mr^2 - 139*r_mr^3 + 120*r_mr^4 - 50*r_mr^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(30 - 53*r_mr + 113*r_mr^2 - 171*r_mr^3 + 140*r_mr^4 - 50*r_mr^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log(1 + 6*r_mr - 22*r_mr^2 + 35*r_mr^3 - 30*r_mr^4 + 10*r_mr^5) + (x[,"n_14"] + x[,"n_41"])*log(6 + 22*r_mr - 85*r_mr^2 + 117*r_mr^3 - 85*r_mr^4 + 25*r_mr^5) + (x[,"n_23"] + x[,"n_32"])*log(27 + 28*r_mr - 107*r_mr^2 + 159*r_mr^3 - 130*r_mr^4 + 50*r_mr^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(9 + 30*r_mr - 60*r_mr^2 + 111*r_mr^3 - 110*r_mr^4 + 50*r_mr^5)
logL_rc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n20 - 3*n21 - 2*n22 - 2*n23 - 3*n24 - 3*n25 - 3*n30 - 3*n31 - 2*n32 - 2*n33 - 3*n34 - 3*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n00 + n55)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n50)*(3*log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n04 + n15 + n40 + n51)*(2*log(pmax(1e-6,r)) + log(9 - 18*r + 17*r^2 - 5*r^3)) + (n03 + n25 + n30 + n52)*(log(pmax(1e-6,r)) + log(9 - 21*r + 33*r^2 - 28*r^3 + 10*r^4)) + (n14 + n41)*(log(pmax(1e-6,r)) + log(36 - 66*r + 93*r^2 - 76*r^3 + 25*r^4)) + (n22 + n33)*log(54 - 53*r + 110*r^2 - 156*r^3 + 122*r^4 - 50*r^5) + (n13 + n24 + n31 + n42)*log(27 - 36*r + 126*r^2 - 171*r^3 + 134*r^4 - 50*r^5) + (n11 + n44)*log(12 - 4*r + 7*r^2 - 39*r^3 + 49*r^4 - 25*r^5) + (n02 + n20 + n35 + n53)*log(3 - 4*r + 10*r^2 - 21*r^3 + 22*r^4 - 10*r^5) + (n01 + n10 + n45 + n54)*log(3 - 7*r + 7*r^2 - 8*r^4 + 5*r^5) + (n23 + n32)*log(27 + 63*r - 126*r^2 + 168*r^3 - 128*r^4 + 50*r^5) + (n12 + n21 + n34 + n43)*log(30 + 11*r - 83*r^2 + 135*r^3 - 116*r^4 + 50*r^5)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rc <- -((x[,"n_05"] + x[,"n_50"])*(-5*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_55"])*(-5*log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + log10(5) + log10(7) + log10(19)) - (x[,"n_22"] + x[,"n_33"])*(-4*log10(2) + log10(5) + log10(7) + log10(19)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(5*log10(2) - log10(29)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(29)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(31)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(31)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(5) + log10(83)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(5) + log10(83)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + log10(293)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + log10(293)) + (x[,"n_00"] + x[,"n_55"])*(3*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc)) + log10(2 + r_rc)) + (x[,"n_05"] + x[,"n_50"])*(3*log10(pmax(1e-6,r_rc)) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_rc)) + log10(9 - 18*r_rc + 17*r_rc^2 - 5*r_rc^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(log10(pmax(1e-6,r_rc)) + log10(9 - 21*r_rc + 33*r_rc^2 - 28*r_rc^3 + 10*r_rc^4)) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,r_rc)) + log10(36 - 66*r_rc + 93*r_rc^2 - 76*r_rc^3 + 25*r_rc^4)) + (x[,"n_22"] + x[,"n_33"])*log10(54 - 53*r_rc + 110*r_rc^2 - 156*r_rc^3 + 122*r_rc^4 - 50*r_rc^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(27 - 36*r_rc + 126*r_rc^2 - 171*r_rc^3 + 134*r_rc^4 - 50*r_rc^5) + (x[,"n_11"] + x[,"n_44"])*log10(12 - 4*r_rc + 7*r_rc^2 - 39*r_rc^3 + 49*r_rc^4 - 25*r_rc^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log10(3 - 4*r_rc + 10*r_rc^2 - 21*r_rc^3 + 22*r_rc^4 - 10*r_rc^5) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*log10(3 - 7*r_rc + 7*r_rc^2 - 8*r_rc^4 + 5*r_rc^5) + (x[,"n_23"] + x[,"n_32"])*log10(27 + 63*r_rc - 126*r_rc^2 + 168*r_rc^3 - 128*r_rc^4 + 50*r_rc^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(30 + 11*r_rc - 83*r_rc^2 + 135*r_rc^3 - 116*r_rc^4 + 50*r_rc^5)
logL_rc <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_55"])*(3*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc)) + log(2 + r_rc)) + (x[,"n_05"] + x[,"n_50"])*(3*log(pmax(1e-6,r_rc)) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log(pmax(1e-6,r_rc)) + log(9 - 18*r_rc + 17*r_rc^2 - 5*r_rc^3)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(log(pmax(1e-6,r_rc)) + log(9 - 21*r_rc + 33*r_rc^2 - 28*r_rc^3 + 10*r_rc^4)) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,r_rc)) + log(36 - 66*r_rc + 93*r_rc^2 - 76*r_rc^3 + 25*r_rc^4)) + (x[,"n_22"] + x[,"n_33"])*log(54 - 53*r_rc + 110*r_rc^2 - 156*r_rc^3 + 122*r_rc^4 - 50*r_rc^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(27 - 36*r_rc + 126*r_rc^2 - 171*r_rc^3 + 134*r_rc^4 - 50*r_rc^5) + (x[,"n_11"] + x[,"n_44"])*log(12 - 4*r_rc + 7*r_rc^2 - 39*r_rc^3 + 49*r_rc^4 - 25*r_rc^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log(3 - 4*r_rc + 10*r_rc^2 - 21*r_rc^3 + 22*r_rc^4 - 10*r_rc^5) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*log(3 - 7*r_rc + 7*r_rc^2 - 8*r_rc^4 + 5*r_rc^5) + (x[,"n_23"] + x[,"n_32"])*log(27 + 63*r_rc - 126*r_rc^2 + 168*r_rc^3 - 128*r_rc^4 + 50*r_rc^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(30 + 11*r_rc - 83*r_rc^2 + 135*r_rc^3 - 116*r_rc^4 + 50*r_rc^5)
logL_rm2c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n20 - 3*n21 - 2*n22 - 2*n23 - 3*n24 - 3*n25 - 3*n30 - 3*n31 - 2*n32 - 2*n33 - 3*n34 - 3*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(2*log(3) + 2*log(5)) + (n00 + n55)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n50)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n01 + n10 + n45 + n54)*(log(pmax(1e-6,r)) + log(9 - 11*r + 2*r^2 + 5*r^3 - 5*r^4)) + (n04 + n15 + n40 + n51)*(log(pmax(1e-6,r)) + log(12 - 25*r + 28*r^2 - 20*r^3 + 5*r^4)) + (n12 + n21 + n34 + n43)*log(81 - 21*r + 70*r^2 - 109*r^3 + 110*r^4 - 50*r^5) + (n23 + n32)*log(126 - 21*r + 78*r^2 - 146*r^3 + 130*r^4 - 50*r^5) + (n14 + n41)*log(36 - 36*r + 71*r^2 - 113*r^3 + 85*r^4 - 25*r^5) + (n03 + n25 + n30 + n52)*log(9 - 15*r + 34*r^2 - 39*r^3 + 30*r^4 - 10*r^5) + (n02 + n20 + n35 + n53)*log(9 - 6*r - 3*r^2 + 19*r^3 - 20*r^4 + 10*r^5) + (n11 + n44)*log(18 + 18*r - 8*r^2 + 23*r^3 - 40*r^4 + 25*r^5) + (n13 + n24 + n31 + n42)*log(81 + 18*r - 97*r^2 + 169*r^3 - 140*r^4 + 50*r^5) + (n22 + n33)*log(117 + 33*r - 80*r^2 + 126*r^3 - 120*r^4 + 50*r^5)
return(L)}
interlogL_rm2c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_rm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rm2c <- -((x[,"n_05"] + x[,"n_50"])*(-5*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_55"])*(-5*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(3) + 2*log10(5) + log10(17)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(3) + 2*log10(5) + log10(17)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(3) + log10(23)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(3) + log10(23)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + 3*log10(3) + log10(31)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + 3*log10(3) + log10(31)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(107)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(107)) - (x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + log10(1973)) - (x[,"n_22"] + x[,"n_33"])*(-4*log10(2) + log10(1973)) + (x[,"n_00"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + 2*log10(pmax(1e-6,r_rm2c)) + log10(2 + r_rm2c)) + (x[,"n_05"] + x[,"n_50"])*(log10(pmax(1e-6,1 - r_rm2c)) + 2*log10(pmax(1e-6,r_rm2c)) + log10(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(log10(pmax(1e-6,r_rm2c)) + log10(9 - 11*r_rm2c + 2*r_rm2c^2 + 5*r_rm2c^3 - 5*r_rm2c^4)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(log10(pmax(1e-6,r_rm2c)) + log10(12 - 25*r_rm2c + 28*r_rm2c^2 - 20*r_rm2c^3 + 5*r_rm2c^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(81 - 21*r_rm2c + 70*r_rm2c^2 - 109*r_rm2c^3 + 110*r_rm2c^4 - 50*r_rm2c^5) + (x[,"n_23"] + x[,"n_32"])*log10(126 - 21*r_rm2c + 78*r_rm2c^2 - 146*r_rm2c^3 + 130*r_rm2c^4 - 50*r_rm2c^5) + (x[,"n_14"] + x[,"n_41"])*log10(36 - 36*r_rm2c + 71*r_rm2c^2 - 113*r_rm2c^3 + 85*r_rm2c^4 - 25*r_rm2c^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log10(9 - 15*r_rm2c + 34*r_rm2c^2 - 39*r_rm2c^3 + 30*r_rm2c^4 - 10*r_rm2c^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log10(9 - 6*r_rm2c - 3*r_rm2c^2 + 19*r_rm2c^3 - 20*r_rm2c^4 + 10*r_rm2c^5) + (x[,"n_11"] + x[,"n_44"])*log10(18 + 18*r_rm2c - 8*r_rm2c^2 + 23*r_rm2c^3 - 40*r_rm2c^4 + 25*r_rm2c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(81 + 18*r_rm2c - 97*r_rm2c^2 + 169*r_rm2c^3 - 140*r_rm2c^4 + 50*r_rm2c^5) + (x[,"n_22"] + x[,"n_33"])*log10(117 + 33*r_rm2c - 80*r_rm2c^2 + 126*r_rm2c^3 - 120*r_rm2c^4 + 50*r_rm2c^5)
logL_rm2c <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(2*log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_rm2c)) + 2*log(pmax(1e-6,r_rm2c)) + log(2 + r_rm2c)) + (x[,"n_05"] + x[,"n_50"])*(log(pmax(1e-6,1 - r_rm2c)) + 2*log(pmax(1e-6,r_rm2c)) + log(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(log(pmax(1e-6,r_rm2c)) + log(9 - 11*r_rm2c + 2*r_rm2c^2 + 5*r_rm2c^3 - 5*r_rm2c^4)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(log(pmax(1e-6,r_rm2c)) + log(12 - 25*r_rm2c + 28*r_rm2c^2 - 20*r_rm2c^3 + 5*r_rm2c^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(81 - 21*r_rm2c + 70*r_rm2c^2 - 109*r_rm2c^3 + 110*r_rm2c^4 - 50*r_rm2c^5) + (x[,"n_23"] + x[,"n_32"])*log(126 - 21*r_rm2c + 78*r_rm2c^2 - 146*r_rm2c^3 + 130*r_rm2c^4 - 50*r_rm2c^5) + (x[,"n_14"] + x[,"n_41"])*log(36 - 36*r_rm2c + 71*r_rm2c^2 - 113*r_rm2c^3 + 85*r_rm2c^4 - 25*r_rm2c^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log(9 - 15*r_rm2c + 34*r_rm2c^2 - 39*r_rm2c^3 + 30*r_rm2c^4 - 10*r_rm2c^5) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*log(9 - 6*r_rm2c - 3*r_rm2c^2 + 19*r_rm2c^3 - 20*r_rm2c^4 + 10*r_rm2c^5) + (x[,"n_11"] + x[,"n_44"])*log(18 + 18*r_rm2c - 8*r_rm2c^2 + 23*r_rm2c^3 - 40*r_rm2c^4 + 25*r_rm2c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(81 + 18*r_rm2c - 97*r_rm2c^2 + 169*r_rm2c^3 - 140*r_rm2c^4 + 50*r_rm2c^5) + (x[,"n_22"] + x[,"n_33"])*log(117 + 33*r_rm2c - 80*r_rm2c^2 + 126*r_rm2c^3 - 120*r_rm2c^4 + 50*r_rm2c^5)
logL_rm1c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n20 - 3*n21 - 2*n22 - 2*n23 - 3*n24 - 3*n25 - 3*n30 - 3*n31 - 2*n32 - 2*n33 - 3*n34 - 3*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(2*log(3) + 2*log(5)) + (n00 + n55)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n50)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n04 + n15 + n40 + n51)*(2*log(pmax(1e-6,1 - r)) + log(9 - 6*r + 13*r^2 - 5*r^3)) + (n01 + n10 + n45 + n54)*(2*log(pmax(1e-6,r)) + log(11 - 5*r - 2*r^2 + 5*r^3)) + (n11 + n44)*(log(pmax(1e-6,r)) + log(48 + 8*r - 8*r^2 + 31*r^3 - 25*r^4)) + (n02 + n20 + n35 + n53)*(log(pmax(1e-6,r)) + log(18 - 8*r - 9*r^2 + 18*r^3 - 10*r^4)) + (n22 + n33)*log(117 - 12*r + 90*r^2 - 128*r^3 + 118*r^4 - 50*r^5) + (n13 + n24 + n31 + n42)*log(108 - 72*r + 88*r^2 - 175*r^3 + 146*r^4 - 50*r^5) + (n03 + n25 + n30 + n52)*log(9 + 3*r - 27*r^2 + 37*r^3 - 32*r^4 + 10*r^5) + (n14 + n41)*log(54 - 39*r - 80*r^2 + 134*r^3 - 94*r^4 + 25*r^5) + (n23 + n32)*log(135 - 6*r - 86*r^2 + 156*r^3 - 132*r^4 + 50*r^5) + (n12 + n21 + n34 + n43)*log(45 + 87*r - 61*r^2 + 91*r^3 - 104*r^4 + 50*r^5)
return(L)}
interlogL_rm1c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_rm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rm1c <- -((x[,"n_05"] + x[,"n_50"])*(-5*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_55"])*(-5*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(3) + 2*log10(5) + log10(17)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(3) + 2*log10(5) + log10(17)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(3) + log10(23)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(3) + log10(23)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + 3*log10(3) + log10(31)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + 3*log10(3) + log10(31)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(107)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(107)) - (x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + log10(1973)) - (x[,"n_22"] + x[,"n_33"])*(-4*log10(2) + log10(1973)) + (x[,"n_00"] + x[,"n_55"])*(log10(pmax(1e-6,1 - r_rm1c)) + 3*log10(pmax(1e-6,r_rm1c)) + log10(2 + r_rm1c)) + (x[,"n_05"] + x[,"n_50"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(pmax(1e-6,r_rm1c)) + log10(3 - 4*r_rm1c + r_rm1c^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(9 - 6*r_rm1c + 13*r_rm1c^2 - 5*r_rm1c^3)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(11 - 5*r_rm1c - 2*r_rm1c^2 + 5*r_rm1c^3)) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm1c)) + log10(48 + 8*r_rm1c - 8*r_rm1c^2 + 31*r_rm1c^3 - 25*r_rm1c^4)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(log10(pmax(1e-6,r_rm1c)) + log10(18 - 8*r_rm1c - 9*r_rm1c^2 + 18*r_rm1c^3 - 10*r_rm1c^4)) + (x[,"n_22"] + x[,"n_33"])*log10(117 - 12*r_rm1c + 90*r_rm1c^2 - 128*r_rm1c^3 + 118*r_rm1c^4 - 50*r_rm1c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(108 - 72*r_rm1c + 88*r_rm1c^2 - 175*r_rm1c^3 + 146*r_rm1c^4 - 50*r_rm1c^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log10(9 + 3*r_rm1c - 27*r_rm1c^2 + 37*r_rm1c^3 - 32*r_rm1c^4 + 10*r_rm1c^5) + (x[,"n_14"] + x[,"n_41"])*log10(54 - 39*r_rm1c - 80*r_rm1c^2 + 134*r_rm1c^3 - 94*r_rm1c^4 + 25*r_rm1c^5) + (x[,"n_23"] + x[,"n_32"])*log10(135 - 6*r_rm1c - 86*r_rm1c^2 + 156*r_rm1c^3 - 132*r_rm1c^4 + 50*r_rm1c^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(45 + 87*r_rm1c - 61*r_rm1c^2 + 91*r_rm1c^3 - 104*r_rm1c^4 + 50*r_rm1c^5)
logL_rm1c <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(2*log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_55"])*(log(pmax(1e-6,1 - r_rm1c)) + 3*log(pmax(1e-6,r_rm1c)) + log(2 + r_rm1c)) + (x[,"n_05"] + x[,"n_50"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(pmax(1e-6,r_rm1c)) + log(3 - 4*r_rm1c + r_rm1c^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(9 - 6*r_rm1c + 13*r_rm1c^2 - 5*r_rm1c^3)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(2*log(pmax(1e-6,r_rm1c)) + log(11 - 5*r_rm1c - 2*r_rm1c^2 + 5*r_rm1c^3)) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_rm1c)) + log(48 + 8*r_rm1c - 8*r_rm1c^2 + 31*r_rm1c^3 - 25*r_rm1c^4)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(log(pmax(1e-6,r_rm1c)) + log(18 - 8*r_rm1c - 9*r_rm1c^2 + 18*r_rm1c^3 - 10*r_rm1c^4)) + (x[,"n_22"] + x[,"n_33"])*log(117 - 12*r_rm1c + 90*r_rm1c^2 - 128*r_rm1c^3 + 118*r_rm1c^4 - 50*r_rm1c^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(108 - 72*r_rm1c + 88*r_rm1c^2 - 175*r_rm1c^3 + 146*r_rm1c^4 - 50*r_rm1c^5) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*log(9 + 3*r_rm1c - 27*r_rm1c^2 + 37*r_rm1c^3 - 32*r_rm1c^4 + 10*r_rm1c^5) + (x[,"n_14"] + x[,"n_41"])*log(54 - 39*r_rm1c - 80*r_rm1c^2 + 134*r_rm1c^3 - 94*r_rm1c^4 + 25*r_rm1c^5) + (x[,"n_23"] + x[,"n_32"])*log(135 - 6*r_rm1c - 86*r_rm1c^2 + 156*r_rm1c^3 - 132*r_rm1c^4 + 50*r_rm1c^5) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(45 + 87*r_rm1c - 61*r_rm1c^2 + 91*r_rm1c^3 - 104*r_rm1c^4 + 50*r_rm1c^5)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n20 - 3*n21 - 2*n22 - 2*n23 - 3*n24 - 3*n25 - 3*n30 - 3*n31 - 2*n32 - 2*n33 - 3*n34 - 3*n35 - 3*n40 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n00 + n55)*(4*log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n25 + n30 + n52)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(11 + 14*r - 10*r^2)) + (n01 + n10 + n45 + n54)*(3*log(pmax(1e-6,r)) + log(9 - r - 5*r^2)) + (n04 + n15 + n40 + n51)*(3*log(pmax(1e-6,1 - r)) + log(3 + 11*r - 5*r^2)) + (n05 + n50)*(3*log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n14 + n41)*(2*log(pmax(1e-6,1 - r)) + log(39 - 25*r + 53*r^2 - 25*r^3)) + (n02 + n20 + n35 + n53)*(2*log(pmax(1e-6,r)) + log(15 - 9*r - 16*r^2 + 10*r^3)) + (n11 + n44)*(2*log(pmax(1e-6,r)) + log(42 - 6*r - 22*r^2 + 25*r^3)) + (n12 + n21 + n34 + n43)*log(63*r - 81*r^3 + 98*r^4 - 50*r^5) + (n23 + n32)*log(69 - 110*r + 182*r^2 - 198*r^3 + 134*r^4 - 50*r^5) + (n13 + n24 + n31 + n42)*log(30 + 38*r - 155*r^2 + 189*r^3 - 152*r^4 + 50*r^5) + (n22 + n33)*log(27 + 54*r - 108*r^2 + 162*r^3 - 116*r^4 + 50*r^5)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rr <- -((x[,"n_05"] + x[,"n_50"])*(-5*log10(2) + log10(5))) - (x[,"n_00"] + x[,"n_55"])*(-5*log10(2) + log10(5)) - (x[,"n_23"] + x[,"n_32"])*(-4*log10(2) + log10(5) + log10(7) + log10(19)) - (x[,"n_22"] + x[,"n_33"])*(-4*log10(2) + log10(5) + log10(7) + log10(19)) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-5*log10(2) + log10(29)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + log10(29)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(31)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(31)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(5) + log10(83)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(5) + log10(83)) - (x[,"n_14"] + x[,"n_41"])*(-5*log10(2) + log10(293)) - (x[,"n_11"] + x[,"n_44"])*(-5*log10(2) + log10(293)) + (x[,"n_00"] + x[,"n_55"])*(4*log10(pmax(1e-6,r_rr)) + log10(2 + r_rr)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr)) + log10(11 + 14*r_rr - 10*r_rr^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(3*log10(pmax(1e-6,r_rr)) + log10(9 - r_rr - 5*r_rr^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(3*log10(pmax(1e-6,1 - r_rr)) + log10(3 + 11*r_rr - 5*r_rr^2)) + (x[,"n_05"] + x[,"n_50"])*(3*log10(pmax(1e-6,1 - r_rr)) + log10(3 - 4*r_rr + r_rr^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(39 - 25*r_rr + 53*r_rr^2 - 25*r_rr^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log10(pmax(1e-6,r_rr)) + log10(15 - 9*r_rr - 16*r_rr^2 + 10*r_rr^3)) + (x[,"n_11"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(42 - 6*r_rr - 22*r_rr^2 + 25*r_rr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(63*r_rr - 81*r_rr^3 + 98*r_rr^4 - 50*r_rr^5) + (x[,"n_23"] + x[,"n_32"])*log10(69 - 110*r_rr + 182*r_rr^2 - 198*r_rr^3 + 134*r_rr^4 - 50*r_rr^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(30 + 38*r_rr - 155*r_rr^2 + 189*r_rr^3 - 152*r_rr^4 + 50*r_rr^5) + (x[,"n_22"] + x[,"n_33"])*log10(27 + 54*r_rr - 108*r_rr^2 + 162*r_rr^3 - 116*r_rr^4 + 50*r_rr^5)
logL_rr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_55"])*(4*log(pmax(1e-6,r_rr)) + log(2 + r_rr)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr)) + log(11 + 14*r_rr - 10*r_rr^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(3*log(pmax(1e-6,r_rr)) + log(9 - r_rr - 5*r_rr^2)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(3*log(pmax(1e-6,1 - r_rr)) + log(3 + 11*r_rr - 5*r_rr^2)) + (x[,"n_05"] + x[,"n_50"])*(3*log(pmax(1e-6,1 - r_rr)) + log(3 - 4*r_rr + r_rr^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(39 - 25*r_rr + 53*r_rr^2 - 25*r_rr^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log(pmax(1e-6,r_rr)) + log(15 - 9*r_rr - 16*r_rr^2 + 10*r_rr^3)) + (x[,"n_11"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(42 - 6*r_rr - 22*r_rr^2 + 25*r_rr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(63*r_rr - 81*r_rr^3 + 98*r_rr^4 - 50*r_rr^5) + (x[,"n_23"] + x[,"n_32"])*log(69 - 110*r_rr + 182*r_rr^2 - 198*r_rr^3 + 134*r_rr^4 - 50*r_rr^5) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(30 + 38*r_rr - 155*r_rr^2 + 189*r_rr^3 - 152*r_rr^4 + 50*r_rr^5) + (x[,"n_22"] + x[,"n_33"])*log(27 + 54*r_rr - 108*r_rr^2 + 162*r_rr^3 - 116*r_rr^4 + 50*r_rr^5)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_cr,r_mc,r_mm2c,r_mm1c,r_mr,r_rc,r_rm2c,r_rm1c,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_cr,LOD_mc,LOD_mm2c,LOD_mm1c,LOD_mr,LOD_rc,LOD_rm2c,LOD_rm1c,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_cr,logL_mc,logL_mm2c,logL_mm1c,logL_mr,logL_rc,logL_rm2c,logL_rm1c,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","coupling repulsion","mixed coupling","mixed mixed2c","mixed mixed1c","mixed repulsion","repulsion coupling","repulsion mixed2c","repulsion mixed1c","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.3_2.4 <- function(x){
logL_cc <- function(r,n01,n03,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n53,n55,n02,n04,n52,n54) {
L <- (-2*n01 - n03 - 2*n05 - 2*n11 - 2*n12 - n13 - 2*n14 - 2*n15 - n21 - 2*n22 - n23 - 2*n24 - n25 - n31 - 2*n32 - n33 - 2*n34 - n35 - 2*n41 - 2*n42 - n43 - 2*n44 - 2*n45 - 2*n51 - n53 - 2*n55)*log(2) + (n03 + n11 + n12 + n13 + n14 + n15 + n41 + n42 + n43 + n44 + n45 + n53)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + 4*(n01 + n55)*log(pmax(1e-6,1 - r)) + 4*(n05 + n51)*log(pmax(1e-6,r)) + (n02 + n54)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n53)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n52)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n15 + n41)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n11 + n45)*(3*log(pmax(1e-6,1 - r)) + log(1 + r)) + (n21 + n35)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(5 + r)) + (n25 + n31)*(2*log(pmax(1e-6,r)) + log(6 - 7*r + r^2)) + (n12 + n44)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r^2)) + (n14 + n42)*(2*log(pmax(1e-6,r)) + log(7 - 8*r + 4*r^2)) + (n24 + n32)*(log(pmax(1e-6,r)) + log(32 - 45*r + 36*r^2 - 8*r^3)) + (n13 + n43)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n22 + n34)*log(15 - 18*r + 15*r^2 - 4*r^3 - 8*r^4) + (n23 + n33)*log(11 - 12*r + 18*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_cc <- function(n01,n03,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n53,n55,n02,n04,n52,n54) {
optimize(logL_cc,c(0,0.5), n01,n03,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n53,n55,n02,n04,n52,n54, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_03"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_53"],x[,"n_55"],x[,"n_02"],x[,"n_04"],x[,"n_52"],x[,"n_54"])
LOD_cc <- 4*(x[,"n_05"] + x[,"n_51"])*log10(2) + 4*(x[,"n_04"] + x[,"n_52"])*log10(2) + 4*(x[,"n_03"] + x[,"n_53"])*log10(2) + 4*(x[,"n_02"] + x[,"n_54"])*log10(2) + 4*(x[,"n_01"] + x[,"n_55"])*log10(2) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(3)) - (x[,"n_24"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_22"] + x[,"n_34"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_25"] + x[,"n_31"])*(-4*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_35"])*(-4*log10(2) + log10(11)) - (x[,"n_13"] + x[,"n_43"])*(-4*log10(2) + log10(13)) - (x[,"n_23"] + x[,"n_33"])*(-3*log10(2) + log10(67)) + 4*(x[,"n_01"] + x[,"n_55"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_51"])*log10(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_54"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_53"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_52"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_41"])*(log10(2 - r_cc) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_45"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(1 + r_cc)) + (x[,"n_21"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc)) + log10(5 + r_cc)) + (x[,"n_25"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(6 - 7*r_cc + r_cc^2)) + (x[,"n_12"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 4*r_cc^2)) + (x[,"n_14"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(7 - 8*r_cc + 4*r_cc^2)) + (x[,"n_24"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(32 - 45*r_cc + 36*r_cc^2 - 8*r_cc^3)) + (x[,"n_13"] + x[,"n_43"])*(log10(pmax(1e-6,r_cc)) + log10(4 - 7*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_22"] + x[,"n_34"])*log10(15 - 18*r_cc + 15*r_cc^2 - 4*r_cc^3 - 8*r_cc^4) + (x[,"n_23"] + x[,"n_33"])*log10(11 - 12*r_cc + 18*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cc <- (-2*x[,"n_01"] - x[,"n_03"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_51"] - x[,"n_53"] - 2*x[,"n_55"])*log(2) + (x[,"n_03"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_15"] + x[,"n_41"] + x[,"n_42"] + x[,"n_43"] + x[,"n_44"] + x[,"n_45"] + x[,"n_53"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + 4*(x[,"n_01"] + x[,"n_55"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_51"])*log(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_54"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_53"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_52"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_41"])*(log(2 - r_cc) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_11"] + x[,"n_45"])*(3*log(pmax(1e-6,1 - r_cc)) + log(1 + r_cc)) + (x[,"n_21"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc)) + log(5 + r_cc)) + (x[,"n_25"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(6 - 7*r_cc + r_cc^2)) + (x[,"n_12"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 4*r_cc^2)) + (x[,"n_14"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(7 - 8*r_cc + 4*r_cc^2)) + (x[,"n_24"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(32 - 45*r_cc + 36*r_cc^2 - 8*r_cc^3)) + (x[,"n_13"] + x[,"n_43"])*(log(pmax(1e-6,r_cc)) + log(4 - 7*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_22"] + x[,"n_34"])*log(15 - 18*r_cc + 15*r_cc^2 - 4*r_cc^3 - 8*r_cc^4) + (x[,"n_23"] + x[,"n_33"])*log(11 - 12*r_cc + 18*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cm2c <- function(r,n01,n02,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n54,n55,n03,n53) {
L <- (-n01 - 2*n02 - 2*n04 - n05 - n11 - 2*n12 - 2*n13 - 2*n14 - n15 - n21 - 2*n22 - 2*n23 - 2*n24 - n25 - n31 - 2*n32 - 2*n33 - 2*n34 - n35 - n41 - 2*n42 - 2*n43 - 2*n44 - n45 - n51 - 2*n52 - 2*n54 - n55)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n24 - n25 - n31 - n32 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n01 + n55)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n51)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n15 + n41)*(2*log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n11 + n45)*(2*log(pmax(1e-6,1 - r)) + log(3 - r + 3*r^2)) + (n04 + n52)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n02 + n54)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n14 + n42)*(log(pmax(1e-6,r)) + log(24 - 29*r + 44*r^2 - 24*r^3)) + (n03 + n53)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n25 + n31)*(log(pmax(1e-6,r)) + log(7 - 8*r + 6*r^2 - 2*r^3)) + (n12 + n44)*log(15 - 2*r - 41*r^2 + 52*r^3 - 24*r^4) + (n23 + n33)*log(17 - 8*r^2 + 16*r^3 - 8*r^4) + (n21 + n35)*log(3 - r - 2*r^2 + 2*r^3 - 2*r^4) + (n24 + n32)*log(21 + 2*r + 34*r^2 - 40*r^3 + 16*r^4) + (n22 + n34)*log(33 - 14*r + 10*r^2 - 24*r^3 + 16*r^4) + (n13 + n43)*log(15 - 8*r + 44*r^2 - 72*r^3 + 36*r^4)
return(L)}
interlogL_cm2c <- function(n01,n02,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n54,n55,n03,n53) {
optimize(logL_cm2c,c(0,0.5), n01,n02,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n54,n55,n03,n53, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_01"],x[,"n_02"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_54"],x[,"n_55"],x[,"n_03"],x[,"n_53"])
LOD_cm2c <- 4*(x[,"n_05"] + x[,"n_51"])*log10(2) + 4*(x[,"n_01"] + x[,"n_55"])*log10(2) - (x[,"n_04"] + x[,"n_52"])*(-2*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_54"])*(-2*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_53"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_42"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_44"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_23"] + x[,"n_33"])*(-log10(2) + log10(3) + log10(11)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(13)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(13)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(17)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(17)) - (x[,"n_24"] + x[,"n_32"])*(-log10(2) + log10(53)) - (x[,"n_22"] + x[,"n_34"])*(-log10(2) + log10(53)) - (x[,"n_13"] + x[,"n_43"])*(-2*log10(2) + log10(61)) + (x[,"n_01"] + x[,"n_55"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_51"])*(log10(pmax(1e-6,1 - r_cm2c)) + 3*log10(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(5 - 5*r_cm2c + 3*r_cm2c^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 - r_cm2c + 3*r_cm2c^2)) + (x[,"n_04"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(7 - 12*r_cm2c + 8*r_cm2c^2)) + (x[,"n_02"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 - 4*r_cm2c + 8*r_cm2c^2)) + (x[,"n_14"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm2c)) + log10(24 - 29*r_cm2c + 44*r_cm2c^2 - 24*r_cm2c^3)) + (x[,"n_03"] + x[,"n_53"])*(log10(pmax(1e-6,r_cm2c)) + log10(2 - 5*r_cm2c + 6*r_cm2c^2 - 3*r_cm2c^3)) + (x[,"n_25"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm2c)) + log10(7 - 8*r_cm2c + 6*r_cm2c^2 - 2*r_cm2c^3)) + (x[,"n_12"] + x[,"n_44"])*log10(15 - 2*r_cm2c - 41*r_cm2c^2 + 52*r_cm2c^3 - 24*r_cm2c^4) + (x[,"n_23"] + x[,"n_33"])*log10(17 - 8*r_cm2c^2 + 16*r_cm2c^3 - 8*r_cm2c^4) + (x[,"n_21"] + x[,"n_35"])*log10(3 - r_cm2c - 2*r_cm2c^2 + 2*r_cm2c^3 - 2*r_cm2c^4) + (x[,"n_24"] + x[,"n_32"])*log10(21 + 2*r_cm2c + 34*r_cm2c^2 - 40*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_22"] + x[,"n_34"])*log10(33 - 14*r_cm2c + 10*r_cm2c^2 - 24*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_13"] + x[,"n_43"])*log10(15 - 8*r_cm2c + 44*r_cm2c^2 - 72*r_cm2c^3 + 36*r_cm2c^4)
logL_cm2c <- (-x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_54"] - x[,"n_55"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_01"] + x[,"n_55"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_51"])*(log(pmax(1e-6,1 - r_cm2c)) + 3*log(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm2c)) + log(5 - 5*r_cm2c + 3*r_cm2c^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(3 - r_cm2c + 3*r_cm2c^2)) + (x[,"n_04"] + x[,"n_52"])*(2*log(pmax(1e-6,r_cm2c)) + log(7 - 12*r_cm2c + 8*r_cm2c^2)) + (x[,"n_02"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(3 - 4*r_cm2c + 8*r_cm2c^2)) + (x[,"n_14"] + x[,"n_42"])*(log(pmax(1e-6,r_cm2c)) + log(24 - 29*r_cm2c + 44*r_cm2c^2 - 24*r_cm2c^3)) + (x[,"n_03"] + x[,"n_53"])*(log(pmax(1e-6,r_cm2c)) + log(2 - 5*r_cm2c + 6*r_cm2c^2 - 3*r_cm2c^3)) + (x[,"n_25"] + x[,"n_31"])*(log(pmax(1e-6,r_cm2c)) + log(7 - 8*r_cm2c + 6*r_cm2c^2 - 2*r_cm2c^3)) + (x[,"n_12"] + x[,"n_44"])*log(15 - 2*r_cm2c - 41*r_cm2c^2 + 52*r_cm2c^3 - 24*r_cm2c^4) + (x[,"n_23"] + x[,"n_33"])*log(17 - 8*r_cm2c^2 + 16*r_cm2c^3 - 8*r_cm2c^4) + (x[,"n_21"] + x[,"n_35"])*log(3 - r_cm2c - 2*r_cm2c^2 + 2*r_cm2c^3 - 2*r_cm2c^4) + (x[,"n_24"] + x[,"n_32"])*log(21 + 2*r_cm2c + 34*r_cm2c^2 - 40*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_22"] + x[,"n_34"])*log(33 - 14*r_cm2c + 10*r_cm2c^2 - 24*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_13"] + x[,"n_43"])*log(15 - 8*r_cm2c + 44*r_cm2c^2 - 72*r_cm2c^3 + 36*r_cm2c^4)
logL_cm1c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-2*n01 - n02 - 2*n03 - n04 - 2*n05 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - n23 - 2*n24 - 2*n25 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n51 - n52 - 2*n53 - n54 - 2*n55)*log(2) + (n13 + n43)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n01 + n05 + n51 + n55)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n15 + n41)*(log(pmax(1e-6,r)) + log(2 - 3*r + 4*r^2 - 3*r^3)) + (n11 + n45)*(log(pmax(1e-6,r)) + log(4 - 9*r + 8*r^2 - 3*r^3)) + (n24 + n32)*log(9 - 2*r + 3*r^2 + 4*r^3 - 8*r^4) + (n22 + n34)*log(6 + 16*r - 33*r^2 + 28*r^3 - 8*r^4) + (n13 + n43)*log(1 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n02 + n04 + n52 + n54)*log(r - 3*r^2 + 4*r^3 - 2*r^4) + (n21 + n35)*log(3 - 8*r + 14*r^2 - 10*r^3 + 2*r^4) + (n25 + n31)*log(1 + 2*r - 4*r^2 + 2*r^3 + 2*r^4) + (n03 + n53)*log(1 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + (n23 + n33)*log(9 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + (n12 + n44)*log(6 - 14*r + 27*r^2 - 28*r^3 + 12*r^4) + (n14 + n42)*log(3 - 4*r + 15*r^2 - 20*r^3 + 12*r^4)
return(L)}
interlogL_cm1c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_cm1c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_cm1c <- 3*(x[,"n_02"] + x[,"n_04"] + x[,"n_52"] + x[,"n_54"])*log10(2) + 4*(x[,"n_01"] + x[,"n_05"] + x[,"n_51"] + x[,"n_55"])*log10(2) + (x[,"n_03"] + x[,"n_53"])*(3*log10(2) - log10(3)) - (x[,"n_14"] + x[,"n_42"])*log10(3) - (x[,"n_12"] + x[,"n_44"])*log10(3) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_24"] + x[,"n_32"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_22"] + x[,"n_34"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(11)) - (x[,"n_13"] + x[,"n_43"])*(-3*log10(2) + log10(13)) - (x[,"n_23"] + x[,"n_33"])*(-3*log10(2) + log10(67)) + (x[,"n_01"] + x[,"n_05"] + x[,"n_51"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_15"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm1c)) + log10(2 - 3*r_cm1c + 4*r_cm1c^2 - 3*r_cm1c^3)) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,r_cm1c)) + log10(4 - 9*r_cm1c + 8*r_cm1c^2 - 3*r_cm1c^3)) + (x[,"n_24"] + x[,"n_32"])*log10(9 - 2*r_cm1c + 3*r_cm1c^2 + 4*r_cm1c^3 - 8*r_cm1c^4) + (x[,"n_22"] + x[,"n_34"])*log10(6 + 16*r_cm1c - 33*r_cm1c^2 + 28*r_cm1c^3 - 8*r_cm1c^4) + (x[,"n_13"] + x[,"n_43"])*log10(1 + 4*r_cm1c - 10*r_cm1c^2 + 12*r_cm1c^3 - 6*r_cm1c^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_52"] + x[,"n_54"])*log10(r_cm1c - 3*r_cm1c^2 + 4*r_cm1c^3 - 2*r_cm1c^4) + (x[,"n_21"] + x[,"n_35"])*log10(3 - 8*r_cm1c + 14*r_cm1c^2 - 10*r_cm1c^3 + 2*r_cm1c^4) + (x[,"n_25"] + x[,"n_31"])*log10(1 + 2*r_cm1c - 4*r_cm1c^2 + 2*r_cm1c^3 + 2*r_cm1c^4) + (x[,"n_03"] + x[,"n_53"])*log10(1 - 4*r_cm1c + 10*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_23"] + x[,"n_33"])*log10(9 - 4*r_cm1c + 10*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_12"] + x[,"n_44"])*log10(6 - 14*r_cm1c + 27*r_cm1c^2 - 28*r_cm1c^3 + 12*r_cm1c^4) + (x[,"n_14"] + x[,"n_42"])*log10(3 - 4*r_cm1c + 15*r_cm1c^2 - 20*r_cm1c^3 + 12*r_cm1c^4)
logL_cm1c <- (-2*x[,"n_01"] - x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_51"] - x[,"n_52"] - 2*x[,"n_53"] - x[,"n_54"] - 2*x[,"n_55"])*log(2) + (x[,"n_13"] + x[,"n_43"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_01"] + x[,"n_05"] + x[,"n_51"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_15"] + x[,"n_41"])*(log(pmax(1e-6,r_cm1c)) + log(2 - 3*r_cm1c + 4*r_cm1c^2 - 3*r_cm1c^3)) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,r_cm1c)) + log(4 - 9*r_cm1c + 8*r_cm1c^2 - 3*r_cm1c^3)) + (x[,"n_24"] + x[,"n_32"])*log(9 - 2*r_cm1c + 3*r_cm1c^2 + 4*r_cm1c^3 - 8*r_cm1c^4) + (x[,"n_22"] + x[,"n_34"])*log(6 + 16*r_cm1c - 33*r_cm1c^2 + 28*r_cm1c^3 - 8*r_cm1c^4) + (x[,"n_13"] + x[,"n_43"])*log(1 + 4*r_cm1c - 10*r_cm1c^2 + 12*r_cm1c^3 - 6*r_cm1c^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_52"] + x[,"n_54"])*log(r_cm1c - 3*r_cm1c^2 + 4*r_cm1c^3 - 2*r_cm1c^4) + (x[,"n_21"] + x[,"n_35"])*log(3 - 8*r_cm1c + 14*r_cm1c^2 - 10*r_cm1c^3 + 2*r_cm1c^4) + (x[,"n_25"] + x[,"n_31"])*log(1 + 2*r_cm1c - 4*r_cm1c^2 + 2*r_cm1c^3 + 2*r_cm1c^4) + (x[,"n_03"] + x[,"n_53"])*log(1 - 4*r_cm1c + 10*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_23"] + x[,"n_33"])*log(9 - 4*r_cm1c + 10*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4) + (x[,"n_12"] + x[,"n_44"])*log(6 - 14*r_cm1c + 27*r_cm1c^2 - 28*r_cm1c^3 + 12*r_cm1c^4) + (x[,"n_14"] + x[,"n_42"])*log(3 - 4*r_cm1c + 15*r_cm1c^2 - 20*r_cm1c^3 + 12*r_cm1c^4)
logL_mc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 4*n25 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n51 - 4*n52 - 3*n53 - 4*n54 - 4*n55)*log(2) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n05 + n51)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n01 + n55)*(2*log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n15 + n41)*(2*log(pmax(1e-6,r)) + log(11 - 8*r + 3*r^2)) + (n11 + n45)*(2*log(pmax(1e-6,1 - r)) + log(6 + 2*r + 3*r^2)) + (n02 + n54)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r^2)) + (n04 + n52)*(2*log(pmax(1e-6,r)) + log(7 - 8*r + 4*r^2)) + (n14 + n42)*(log(pmax(1e-6,r)) + log(30 - 25*r + 28*r^2 - 12*r^3)) + (n03 + n53)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n25 + n31)*(log(pmax(1e-6,r)) + log(18 - 17*r + 10*r^2 - 2*r^3)) + (n12 + n44)*log(21 - 16*r - 13*r^2 + 20*r^3 - 12*r^4) + (n23 + n33)*log(67 - 4*r - 8*r^2 + 24*r^3 - 12*r^4) + (n21 + n35)*log(9 - 6*r + r^2 - 2*r^3 - 2*r^4) + (n24 + n32)*log(27 + 8*r + 30*r^2 - 28*r^3 + 8*r^4) + (n22 + n34)*log(45 - 16*r - 6*r^2 - 4*r^3 + 8*r^4) + (n13 + n43)*log(21 - 4*r + 22*r^2 - 36*r^3 + 18*r^4)
return(L)}
interlogL_mc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_mc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mc <- -((x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(3))) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_53"])*(-4*log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_43"])*(-3*log10(2) + 2*log10(13)) - (x[,"n_14"] + x[,"n_42"])*(-log10(2) + log10(23)) - (x[,"n_12"] + x[,"n_44"])*(-log10(2) + log10(23)) - (x[,"n_23"] + x[,"n_33"])*(-2*log10(2) + 2*log10(3) + log10(29)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(31)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(31)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(47)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(47)) - (x[,"n_24"] + x[,"n_32"])*(-log10(2) + log10(71)) - (x[,"n_22"] + x[,"n_34"])*(-log10(2) + log10(71)) + (x[,"n_05"] + x[,"n_51"])*(log10(2 - r_mc) + 3*log10(pmax(1e-6,r_mc))) + (x[,"n_01"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(1 - r_mc^2)) + (x[,"n_15"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_mc)) + log10(11 - 8*r_mc + 3*r_mc^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(6 + 2*r_mc + 3*r_mc^2)) + (x[,"n_02"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 + 4*r_mc^2)) + (x[,"n_04"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_mc)) + log10(7 - 8*r_mc + 4*r_mc^2)) + (x[,"n_14"] + x[,"n_42"])*(log10(pmax(1e-6,r_mc)) + log10(30 - 25*r_mc + 28*r_mc^2 - 12*r_mc^3)) + (x[,"n_03"] + x[,"n_53"])*(log10(pmax(1e-6,r_mc)) + log10(4 - 7*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_25"] + x[,"n_31"])*(log10(pmax(1e-6,r_mc)) + log10(18 - 17*r_mc + 10*r_mc^2 - 2*r_mc^3)) + (x[,"n_12"] + x[,"n_44"])*log10(21 - 16*r_mc - 13*r_mc^2 + 20*r_mc^3 - 12*r_mc^4) + (x[,"n_23"] + x[,"n_33"])*log10(67 - 4*r_mc - 8*r_mc^2 + 24*r_mc^3 - 12*r_mc^4) + (x[,"n_21"] + x[,"n_35"])*log10(9 - 6*r_mc + r_mc^2 - 2*r_mc^3 - 2*r_mc^4) + (x[,"n_24"] + x[,"n_32"])*log10(27 + 8*r_mc + 30*r_mc^2 - 28*r_mc^3 + 8*r_mc^4) + (x[,"n_22"] + x[,"n_34"])*log10(45 - 16*r_mc - 6*r_mc^2 - 4*r_mc^3 + 8*r_mc^4) + (x[,"n_13"] + x[,"n_43"])*log10(21 - 4*r_mc + 22*r_mc^2 - 36*r_mc^3 + 18*r_mc^4)
logL_mc <- (-4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 3*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_05"] + x[,"n_51"])*(log(2 - r_mc) + 3*log(pmax(1e-6,r_mc))) + (x[,"n_01"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_mc)) + log(1 - r_mc^2)) + (x[,"n_15"] + x[,"n_41"])*(2*log(pmax(1e-6,r_mc)) + log(11 - 8*r_mc + 3*r_mc^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_mc)) + log(6 + 2*r_mc + 3*r_mc^2)) + (x[,"n_02"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 + 4*r_mc^2)) + (x[,"n_04"] + x[,"n_52"])*(2*log(pmax(1e-6,r_mc)) + log(7 - 8*r_mc + 4*r_mc^2)) + (x[,"n_14"] + x[,"n_42"])*(log(pmax(1e-6,r_mc)) + log(30 - 25*r_mc + 28*r_mc^2 - 12*r_mc^3)) + (x[,"n_03"] + x[,"n_53"])*(log(pmax(1e-6,r_mc)) + log(4 - 7*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_25"] + x[,"n_31"])*(log(pmax(1e-6,r_mc)) + log(18 - 17*r_mc + 10*r_mc^2 - 2*r_mc^3)) + (x[,"n_12"] + x[,"n_44"])*log(21 - 16*r_mc - 13*r_mc^2 + 20*r_mc^3 - 12*r_mc^4) + (x[,"n_23"] + x[,"n_33"])*log(67 - 4*r_mc - 8*r_mc^2 + 24*r_mc^3 - 12*r_mc^4) + (x[,"n_21"] + x[,"n_35"])*log(9 - 6*r_mc + r_mc^2 - 2*r_mc^3 - 2*r_mc^4) + (x[,"n_24"] + x[,"n_32"])*log(27 + 8*r_mc + 30*r_mc^2 - 28*r_mc^3 + 8*r_mc^4) + (x[,"n_22"] + x[,"n_34"])*log(45 - 16*r_mc - 6*r_mc^2 - 4*r_mc^3 + 8*r_mc^4) + (x[,"n_13"] + x[,"n_43"])*log(21 - 4*r_mc + 22*r_mc^2 - 36*r_mc^3 + 18*r_mc^4)
logL_mm2c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-3*n01 - 4*n02 - 4*n03 - 4*n04 - 3*n05 - 3*n11 - 4*n12 - 4*n13 - 4*n14 - 3*n15 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 3*n25 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n41 - 4*n42 - 4*n43 - 4*n44 - 3*n45 - 3*n51 - 4*n52 - 4*n53 - 4*n54 - 3*n55)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n24 - n25 - n31 - n32 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n05 + n51)*(log(pmax(1e-6,1 - r)) + log(2 - r) + 2*log(pmax(1e-6,r))) + (n01 + n55)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(1 - r^2)) + (n04 + n52)*(log(pmax(1e-6,r)) + log(12 - 21*r + 20*r^2 - 8*r^3)) + (n15 + n41)*(log(pmax(1e-6,r)) + log(9 - 10*r + 7*r^2 - 3*r^3)) + (n13 + n43)*log(69 - 8*r - 28*r^2 + 72*r^3 - 36*r^4) + (n02 + n54)*log(3 + 2*r - 9*r^2 + 12*r^3 - 8*r^4) + (n24 + n32)*log(51 + 4*r - 2*r^2 + 12*r^3 - 8*r^4) + (n22 + n34)*log(57 - 4*r - 14*r^2 + 20*r^3 - 8*r^4) + (n11 + n45)*log(3 + 2*r - 7*r^2 + 5*r^3 - 3*r^4) + (n21 + n35)*log(12 - 10*r + 11*r^2 - 6*r^3 + 2*r^4) + (n25 + n31)*log(9 - 2*r + 5*r^2 - 2*r^3 + 2*r^4) + (n23 + n33)*log(31 + 4*r - 8*r^3 + 4*r^4) + (n03 + n53)*log(9 - 16*r + 28*r^2 - 24*r^3 + 12*r^4) + (n14 + n42)*log(27 + 6*r + 37*r^2 - 52*r^3 + 24*r^4) + (n12 + n44)*log(42 - 20*r + 25*r^2 - 44*r^3 + 24*r^4)
return(L)}
interlogL_mm2c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_mm2c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm2c <- mapply(interlogL_mm2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mm2c <- -((x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(3))) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(3)) - (x[,"n_04"] + x[,"n_52"])*(-2*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_54"])*(-2*log10(2) + log10(11)) - (x[,"n_03"] + x[,"n_53"])*(-2*log10(2) + log10(23)) - (x[,"n_13"] + x[,"n_43"])*(-2*log10(2) + log10(7) + log10(37)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(43)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(43)) - (x[,"n_23"] + x[,"n_33"])*(-2*log10(2) + log10(3) + log10(43)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(73)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(73)) - (x[,"n_24"] + x[,"n_32"])*(-log10(2) + log10(107)) - (x[,"n_22"] + x[,"n_34"])*(-log10(2) + log10(107)) - (x[,"n_14"] + x[,"n_42"])*(-2*log10(2) + log10(137)) - (x[,"n_12"] + x[,"n_44"])*(-2*log10(2) + log10(137)) + (x[,"n_05"] + x[,"n_51"])*(log10(pmax(1e-6,1 - r_mm2c)) + log10(2 - r_mm2c) + 2*log10(pmax(1e-6,r_mm2c))) + (x[,"n_01"] + x[,"n_55"])*(log10(pmax(1e-6,1 - r_mm2c)) + log10(pmax(1e-6,r_mm2c)) + log10(1 - r_mm2c^2)) + (x[,"n_04"] + x[,"n_52"])*(log10(pmax(1e-6,r_mm2c)) + log10(12 - 21*r_mm2c + 20*r_mm2c^2 - 8*r_mm2c^3)) + (x[,"n_15"] + x[,"n_41"])*(log10(pmax(1e-6,r_mm2c)) + log10(9 - 10*r_mm2c + 7*r_mm2c^2 - 3*r_mm2c^3)) + (x[,"n_13"] + x[,"n_43"])*log10(69 - 8*r_mm2c - 28*r_mm2c^2 + 72*r_mm2c^3 - 36*r_mm2c^4) + (x[,"n_02"] + x[,"n_54"])*log10(3 + 2*r_mm2c - 9*r_mm2c^2 + 12*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_24"] + x[,"n_32"])*log10(51 + 4*r_mm2c - 2*r_mm2c^2 + 12*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_22"] + x[,"n_34"])*log10(57 - 4*r_mm2c - 14*r_mm2c^2 + 20*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_11"] + x[,"n_45"])*log10(3 + 2*r_mm2c - 7*r_mm2c^2 + 5*r_mm2c^3 - 3*r_mm2c^4) + (x[,"n_21"] + x[,"n_35"])*log10(12 - 10*r_mm2c + 11*r_mm2c^2 - 6*r_mm2c^3 + 2*r_mm2c^4) + (x[,"n_25"] + x[,"n_31"])*log10(9 - 2*r_mm2c + 5*r_mm2c^2 - 2*r_mm2c^3 + 2*r_mm2c^4) + (x[,"n_23"] + x[,"n_33"])*log10(31 + 4*r_mm2c - 8*r_mm2c^3 + 4*r_mm2c^4) + (x[,"n_03"] + x[,"n_53"])*log10(9 - 16*r_mm2c + 28*r_mm2c^2 - 24*r_mm2c^3 + 12*r_mm2c^4) + (x[,"n_14"] + x[,"n_42"])*log10(27 + 6*r_mm2c + 37*r_mm2c^2 - 52*r_mm2c^3 + 24*r_mm2c^4) + (x[,"n_12"] + x[,"n_44"])*log10(42 - 20*r_mm2c + 25*r_mm2c^2 - 44*r_mm2c^3 + 24*r_mm2c^4)
logL_mm2c <- (-3*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_05"] + x[,"n_51"])*(log(pmax(1e-6,1 - r_mm2c)) + log(2 - r_mm2c) + 2*log(pmax(1e-6,r_mm2c))) + (x[,"n_01"] + x[,"n_55"])*(log(pmax(1e-6,1 - r_mm2c)) + log(pmax(1e-6,r_mm2c)) + log(1 - r_mm2c^2)) + (x[,"n_04"] + x[,"n_52"])*(log(pmax(1e-6,r_mm2c)) + log(12 - 21*r_mm2c + 20*r_mm2c^2 - 8*r_mm2c^3)) + (x[,"n_15"] + x[,"n_41"])*(log(pmax(1e-6,r_mm2c)) + log(9 - 10*r_mm2c + 7*r_mm2c^2 - 3*r_mm2c^3)) + (x[,"n_13"] + x[,"n_43"])*log(69 - 8*r_mm2c - 28*r_mm2c^2 + 72*r_mm2c^3 - 36*r_mm2c^4) + (x[,"n_02"] + x[,"n_54"])*log(3 + 2*r_mm2c - 9*r_mm2c^2 + 12*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_24"] + x[,"n_32"])*log(51 + 4*r_mm2c - 2*r_mm2c^2 + 12*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_22"] + x[,"n_34"])*log(57 - 4*r_mm2c - 14*r_mm2c^2 + 20*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_11"] + x[,"n_45"])*log(3 + 2*r_mm2c - 7*r_mm2c^2 + 5*r_mm2c^3 - 3*r_mm2c^4) + (x[,"n_21"] + x[,"n_35"])*log(12 - 10*r_mm2c + 11*r_mm2c^2 - 6*r_mm2c^3 + 2*r_mm2c^4) + (x[,"n_25"] + x[,"n_31"])*log(9 - 2*r_mm2c + 5*r_mm2c^2 - 2*r_mm2c^3 + 2*r_mm2c^4) + (x[,"n_23"] + x[,"n_33"])*log(31 + 4*r_mm2c - 8*r_mm2c^3 + 4*r_mm2c^4) + (x[,"n_03"] + x[,"n_53"])*log(9 - 16*r_mm2c + 28*r_mm2c^2 - 24*r_mm2c^3 + 12*r_mm2c^4) + (x[,"n_14"] + x[,"n_42"])*log(27 + 6*r_mm2c + 37*r_mm2c^2 - 52*r_mm2c^3 + 24*r_mm2c^4) + (x[,"n_12"] + x[,"n_44"])*log(42 - 20*r_mm2c + 25*r_mm2c^2 - 44*r_mm2c^3 + 24*r_mm2c^4)
logL_mm1c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n11 - 4*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 4*n25 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n41 - 4*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 4*n55)*log(2) + (n23 + n33)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n05 + n51)*(2*log(pmax(1e-6,1 - r)) + log(2 - r) + log(pmax(1e-6,r))) + (n01 + n55)*(2*log(pmax(1e-6,r)) + log(1 - r^2)) + (n11 + n45)*(log(pmax(1e-6,r)) + log(4 + 2*r - 6*r^2 + 3*r^3)) + (n02 + n54)*(log(pmax(1e-6,r)) + log(2 + r - 4*r^2 + 4*r^3)) + (n14 + n42)*log(15 - 2*r - 19*r^2 + 24*r^3 - 12*r^4) + (n12 + n44)*log(6 + 16*r - 19*r^2 + 24*r^3 - 12*r^4) + (n03 + n53)*log(1 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n23 + n33)*log(21 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n25 + n31)*log(10 - 14*r + 15*r^2 - 6*r^3 - 2*r^4) + (n21 + n35)*log(3 + 10*r - 15*r^2 + 14*r^3 - 2*r^4) + (n15 + n41)*log(3 - 2*r + 2*r^2 - 6*r^3 + 3*r^4) + (n04 + n52)*log(3 - 8*r + 13*r^2 - 12*r^3 + 4*r^4) + (n22 + n34)*log(33 - 8*r + 42*r^2 - 36*r^3 + 8*r^4) + (n24 + n32)*log(39 - 18*r^2 + 4*r^3 + 8*r^4) + (n13 + n43)*log(12 - 8*r + 17*r^2 - 18*r^3 + 9*r^4)
return(L)}
interlogL_mm1c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_mm1c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm1c <- mapply(interlogL_mm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mm1c <- -((x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(3))) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_53"])*(-3*log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_43"])*(-4*log10(2) + 2*log10(13)) - (x[,"n_14"] + x[,"n_42"])*(-log10(2) + log10(23)) - (x[,"n_12"] + x[,"n_44"])*(-log10(2) + log10(23)) - (x[,"n_23"] + x[,"n_33"])*(-2*log10(2) + log10(3) + log10(29)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(31)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(31)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(47)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(47)) - (x[,"n_24"] + x[,"n_32"])*(-log10(2) + log10(71)) - (x[,"n_22"] + x[,"n_34"])*(-log10(2) + log10(71)) + (x[,"n_05"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_mm1c)) + log10(2 - r_mm1c) + log10(pmax(1e-6,r_mm1c))) + (x[,"n_01"] + x[,"n_55"])*(2*log10(pmax(1e-6,r_mm1c)) + log10(1 - r_mm1c^2)) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,r_mm1c)) + log10(4 + 2*r_mm1c - 6*r_mm1c^2 + 3*r_mm1c^3)) + (x[,"n_02"] + x[,"n_54"])*(log10(pmax(1e-6,r_mm1c)) + log10(2 + r_mm1c - 4*r_mm1c^2 + 4*r_mm1c^3)) + (x[,"n_14"] + x[,"n_42"])*log10(15 - 2*r_mm1c - 19*r_mm1c^2 + 24*r_mm1c^3 - 12*r_mm1c^4) + (x[,"n_12"] + x[,"n_44"])*log10(6 + 16*r_mm1c - 19*r_mm1c^2 + 24*r_mm1c^3 - 12*r_mm1c^4) + (x[,"n_03"] + x[,"n_53"])*log10(1 + 4*r_mm1c - 10*r_mm1c^2 + 12*r_mm1c^3 - 6*r_mm1c^4) + (x[,"n_23"] + x[,"n_33"])*log10(21 + 4*r_mm1c - 8*r_mm1c^2 + 8*r_mm1c^3 - 4*r_mm1c^4) + (x[,"n_25"] + x[,"n_31"])*log10(10 - 14*r_mm1c + 15*r_mm1c^2 - 6*r_mm1c^3 - 2*r_mm1c^4) + (x[,"n_21"] + x[,"n_35"])*log10(3 + 10*r_mm1c - 15*r_mm1c^2 + 14*r_mm1c^3 - 2*r_mm1c^4) + (x[,"n_15"] + x[,"n_41"])*log10(3 - 2*r_mm1c + 2*r_mm1c^2 - 6*r_mm1c^3 + 3*r_mm1c^4) + (x[,"n_04"] + x[,"n_52"])*log10(3 - 8*r_mm1c + 13*r_mm1c^2 - 12*r_mm1c^3 + 4*r_mm1c^4) + (x[,"n_22"] + x[,"n_34"])*log10(33 - 8*r_mm1c + 42*r_mm1c^2 - 36*r_mm1c^3 + 8*r_mm1c^4) + (x[,"n_24"] + x[,"n_32"])*log10(39 - 18*r_mm1c^2 + 4*r_mm1c^3 + 8*r_mm1c^4) + (x[,"n_13"] + x[,"n_43"])*log10(12 - 8*r_mm1c + 17*r_mm1c^2 - 18*r_mm1c^3 + 9*r_mm1c^4)
logL_mm1c <- (-4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + (x[,"n_23"] + x[,"n_33"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_05"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_mm1c)) + log(2 - r_mm1c) + log(pmax(1e-6,r_mm1c))) + (x[,"n_01"] + x[,"n_55"])*(2*log(pmax(1e-6,r_mm1c)) + log(1 - r_mm1c^2)) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,r_mm1c)) + log(4 + 2*r_mm1c - 6*r_mm1c^2 + 3*r_mm1c^3)) + (x[,"n_02"] + x[,"n_54"])*(log(pmax(1e-6,r_mm1c)) + log(2 + r_mm1c - 4*r_mm1c^2 + 4*r_mm1c^3)) + (x[,"n_14"] + x[,"n_42"])*log(15 - 2*r_mm1c - 19*r_mm1c^2 + 24*r_mm1c^3 - 12*r_mm1c^4) + (x[,"n_12"] + x[,"n_44"])*log(6 + 16*r_mm1c - 19*r_mm1c^2 + 24*r_mm1c^3 - 12*r_mm1c^4) + (x[,"n_03"] + x[,"n_53"])*log(1 + 4*r_mm1c - 10*r_mm1c^2 + 12*r_mm1c^3 - 6*r_mm1c^4) + (x[,"n_23"] + x[,"n_33"])*log(21 + 4*r_mm1c - 8*r_mm1c^2 + 8*r_mm1c^3 - 4*r_mm1c^4) + (x[,"n_25"] + x[,"n_31"])*log(10 - 14*r_mm1c + 15*r_mm1c^2 - 6*r_mm1c^3 - 2*r_mm1c^4) + (x[,"n_21"] + x[,"n_35"])*log(3 + 10*r_mm1c - 15*r_mm1c^2 + 14*r_mm1c^3 - 2*r_mm1c^4) + (x[,"n_15"] + x[,"n_41"])*log(3 - 2*r_mm1c + 2*r_mm1c^2 - 6*r_mm1c^3 + 3*r_mm1c^4) + (x[,"n_04"] + x[,"n_52"])*log(3 - 8*r_mm1c + 13*r_mm1c^2 - 12*r_mm1c^3 + 4*r_mm1c^4) + (x[,"n_22"] + x[,"n_34"])*log(33 - 8*r_mm1c + 42*r_mm1c^2 - 36*r_mm1c^3 + 8*r_mm1c^4) + (x[,"n_24"] + x[,"n_32"])*log(39 - 18*r_mm1c^2 + 4*r_mm1c^3 + 8*r_mm1c^4) + (x[,"n_13"] + x[,"n_43"])*log(12 - 8*r_mm1c + 17*r_mm1c^2 - 18*r_mm1c^3 + 9*r_mm1c^4)
logL_rc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 3*n25 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n41 - 3*n42 - 2*n43 - 3*n44 - 3*n45 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n01 + n55)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n51)*(2*log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n04 + n52)*(log(pmax(1e-6,r)) + log(6 - 11*r + 12*r^2 - 4*r^3)) + (n15 + n41)*(log(pmax(1e-6,r)) + log(12 - 16*r + 10*r^2 - 3*r^3)) + (n13 + n43)*log(15 + 4*r - 13*r^2 + 18*r^3 - 9*r^4) + (n22 + n34)*log(57 - 4*r - 12*r^2 + 12*r^3 - 8*r^4) + (n24 + n32)*log(45 + 24*r - 24*r^2 + 20*r^3 - 8*r^4) + (n02 + n54)*log(3 - 4*r + r^2 + 4*r^3 - 4*r^4) + (n11 + n45)*log(3 + 2*r - 4*r^2 + 2*r^3 - 3*r^4) + (n25 + n31)*log(9 - 6*r + 13*r^2 - 6*r^3 + 2*r^4) + (n21 + n35)*log(12 - 10*r + 7*r^2 - 2*r^3 + 2*r^4) + (n03 + n53)*log(3 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + (n14 + n42)*log(18 - 12*r + 35*r^2 - 32*r^3 + 12*r^4) + (n23 + n33)*log(99 - 4*r + 16*r^2 - 24*r^3 + 12*r^4) + (n12 + n44)*log(21 - 10*r + 11*r^2 - 16*r^3 + 12*r^4)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rc <- -((x[,"n_04"] + x[,"n_52"])*(-log10(2) + log10(3))) - (x[,"n_02"] + x[,"n_54"])*(-log10(2) + log10(3)) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_42"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_44"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_03"] + x[,"n_53"])*(-3*log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_43"])*(-4*log10(2) + log10(13) + log10(19)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_24"] + x[,"n_32"])*log10(53) - (x[,"n_22"] + x[,"n_34"])*log10(53) - (x[,"n_23"] + x[,"n_33"])*(-2*log10(2) + log10(5) + log10(79)) + (x[,"n_01"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc)) + log10(2 + r_rc)) + (x[,"n_05"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_rc)) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_04"] + x[,"n_52"])*(log10(pmax(1e-6,r_rc)) + log10(6 - 11*r_rc + 12*r_rc^2 - 4*r_rc^3)) + (x[,"n_15"] + x[,"n_41"])*(log10(pmax(1e-6,r_rc)) + log10(12 - 16*r_rc + 10*r_rc^2 - 3*r_rc^3)) + (x[,"n_13"] + x[,"n_43"])*log10(15 + 4*r_rc - 13*r_rc^2 + 18*r_rc^3 - 9*r_rc^4) + (x[,"n_22"] + x[,"n_34"])*log10(57 - 4*r_rc - 12*r_rc^2 + 12*r_rc^3 - 8*r_rc^4) + (x[,"n_24"] + x[,"n_32"])*log10(45 + 24*r_rc - 24*r_rc^2 + 20*r_rc^3 - 8*r_rc^4) + (x[,"n_02"] + x[,"n_54"])*log10(3 - 4*r_rc + r_rc^2 + 4*r_rc^3 - 4*r_rc^4) + (x[,"n_11"] + x[,"n_45"])*log10(3 + 2*r_rc - 4*r_rc^2 + 2*r_rc^3 - 3*r_rc^4) + (x[,"n_25"] + x[,"n_31"])*log10(9 - 6*r_rc + 13*r_rc^2 - 6*r_rc^3 + 2*r_rc^4) + (x[,"n_21"] + x[,"n_35"])*log10(12 - 10*r_rc + 7*r_rc^2 - 2*r_rc^3 + 2*r_rc^4) + (x[,"n_03"] + x[,"n_53"])*log10(3 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + (x[,"n_14"] + x[,"n_42"])*log10(18 - 12*r_rc + 35*r_rc^2 - 32*r_rc^3 + 12*r_rc^4) + (x[,"n_23"] + x[,"n_33"])*log10(99 - 4*r_rc + 16*r_rc^2 - 24*r_rc^3 + 12*r_rc^4) + (x[,"n_12"] + x[,"n_44"])*log10(21 - 10*r_rc + 11*r_rc^2 - 16*r_rc^3 + 12*r_rc^4)
logL_rc <- (-3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 2*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc)) + log(2 + r_rc)) + (x[,"n_05"] + x[,"n_51"])*(2*log(pmax(1e-6,r_rc)) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_04"] + x[,"n_52"])*(log(pmax(1e-6,r_rc)) + log(6 - 11*r_rc + 12*r_rc^2 - 4*r_rc^3)) + (x[,"n_15"] + x[,"n_41"])*(log(pmax(1e-6,r_rc)) + log(12 - 16*r_rc + 10*r_rc^2 - 3*r_rc^3)) + (x[,"n_13"] + x[,"n_43"])*log(15 + 4*r_rc - 13*r_rc^2 + 18*r_rc^3 - 9*r_rc^4) + (x[,"n_22"] + x[,"n_34"])*log(57 - 4*r_rc - 12*r_rc^2 + 12*r_rc^3 - 8*r_rc^4) + (x[,"n_24"] + x[,"n_32"])*log(45 + 24*r_rc - 24*r_rc^2 + 20*r_rc^3 - 8*r_rc^4) + (x[,"n_02"] + x[,"n_54"])*log(3 - 4*r_rc + r_rc^2 + 4*r_rc^3 - 4*r_rc^4) + (x[,"n_11"] + x[,"n_45"])*log(3 + 2*r_rc - 4*r_rc^2 + 2*r_rc^3 - 3*r_rc^4) + (x[,"n_25"] + x[,"n_31"])*log(9 - 6*r_rc + 13*r_rc^2 - 6*r_rc^3 + 2*r_rc^4) + (x[,"n_21"] + x[,"n_35"])*log(12 - 10*r_rc + 7*r_rc^2 - 2*r_rc^3 + 2*r_rc^4) + (x[,"n_03"] + x[,"n_53"])*log(3 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + (x[,"n_14"] + x[,"n_42"])*log(18 - 12*r_rc + 35*r_rc^2 - 32*r_rc^3 + 12*r_rc^4) + (x[,"n_23"] + x[,"n_33"])*log(99 - 4*r_rc + 16*r_rc^2 - 24*r_rc^3 + 12*r_rc^4) + (x[,"n_12"] + x[,"n_44"])*log(21 - 10*r_rc + 11*r_rc^2 - 16*r_rc^3 + 12*r_rc^4)
logL_rm2c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n23,n33) {
L <- (-2*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 2*n11 - 3*n12 - 3*n13 - 3*n14 - 2*n15 - 2*n21 - n22 - n24 - 2*n25 - 2*n31 - n32 - n34 - 2*n35 - 2*n41 - 3*n42 - 3*n43 - 3*n44 - 2*n45 - 2*n51 - 3*n52 - 3*n53 - 3*n54 - 2*n55)*log(2) + (-2*n01 - 2*n02 - n03 - 2*n04 - 2*n05 - n11 - n12 - n13 - n14 - n15 - 2*n21 - 2*n22 - n23 - 2*n24 - 2*n25 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - n41 - n42 - n43 - n44 - n45 - 2*n51 - 2*n52 - n53 - 2*n54 - 2*n55)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n01 + n55)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(2 + r)) + (n05 + n51)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n11 + n45)*(log(pmax(1e-6,r)) + log(3 - r^2 + r^3)) + (n02 + n54)*(log(pmax(1e-6,r)) + log(12 - 7*r - 4*r^2 + 8*r^3)) + (n12 + n44)*log(9 + 18*r - 7*r^2 + 12*r^3 - 8*r^4) + (n14 + n42)*log(24 - 8*r - 19*r^2 + 20*r^3 - 8*r^4) + (n03 + n53)*log(3 - 4*r^2 + 8*r^3 - 4*r^4) + (n25 + n31)*log(18 - 12*r + 7*r^2 - 2*r^3 - 2*r^4) + (n21 + n35)*log(9 + 12*r - 11*r^2 + 10*r^3 - 2*r^4) + (n23 + n33)*log(12 + r - 2*r^2 + 2*r^3 - r^4) + (n15 + n41)*log(3 - 4*r + 3*r^2 - 3*r^3 + r^4) + (n22 + n34)*log(36 + 3*r + 16*r^2 - 14*r^3 + 4*r^4) + (n24 + n32)*log(45 - 9*r - 2*r^2 - 2*r^3 + 4*r^4) + (n04 + n52)*log(9 - 18*r + 29*r^2 - 28*r^3 + 8*r^4) + (n13 + n43)*log(33 - 8*r + 20*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_rm2c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n23,n33) {
optimize(logL_rm2c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n23,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_23"],x[,"n_33"])
LOD_rm2c <- -((x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(5))) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(5)) - (x[,"n_24"] + x[,"n_32"])*(3*log10(2) + log10(5)) - (x[,"n_22"] + x[,"n_34"])*(3*log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_53"])*(-2*log10(2) + log10(11)) - (x[,"n_23"] + x[,"n_33"])*(-4*log10(2) + log10(3) + log10(5) + log10(13)) - (x[,"n_04"] + x[,"n_52"])*(-2*log10(2) + log10(17)) - (x[,"n_02"] + x[,"n_54"])*(-2*log10(2) + log10(17)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(23)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(23)) - (x[,"n_14"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_12"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(107)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(107)) - (x[,"n_13"] + x[,"n_43"])*(-2*log10(2) + log10(127)) + (x[,"n_01"] + x[,"n_55"])*(log10(pmax(1e-6,1 - r_rm2c)) + 2*log10(pmax(1e-6,r_rm2c)) + log10(2 + r_rm2c)) + (x[,"n_05"] + x[,"n_51"])*(log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c)) + log10(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,r_rm2c)) + log10(3 - r_rm2c^2 + r_rm2c^3)) + (x[,"n_02"] + x[,"n_54"])*(log10(pmax(1e-6,r_rm2c)) + log10(12 - 7*r_rm2c - 4*r_rm2c^2 + 8*r_rm2c^3)) + (x[,"n_12"] + x[,"n_44"])*log10(9 + 18*r_rm2c - 7*r_rm2c^2 + 12*r_rm2c^3 - 8*r_rm2c^4) + (x[,"n_14"] + x[,"n_42"])*log10(24 - 8*r_rm2c - 19*r_rm2c^2 + 20*r_rm2c^3 - 8*r_rm2c^4) + (x[,"n_03"] + x[,"n_53"])*log10(3 - 4*r_rm2c^2 + 8*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_25"] + x[,"n_31"])*log10(18 - 12*r_rm2c + 7*r_rm2c^2 - 2*r_rm2c^3 - 2*r_rm2c^4) + (x[,"n_21"] + x[,"n_35"])*log10(9 + 12*r_rm2c - 11*r_rm2c^2 + 10*r_rm2c^3 - 2*r_rm2c^4) + (x[,"n_23"] + x[,"n_33"])*log10(12 + r_rm2c - 2*r_rm2c^2 + 2*r_rm2c^3 - r_rm2c^4) + (x[,"n_15"] + x[,"n_41"])*log10(3 - 4*r_rm2c + 3*r_rm2c^2 - 3*r_rm2c^3 + r_rm2c^4) + (x[,"n_22"] + x[,"n_34"])*log10(36 + 3*r_rm2c + 16*r_rm2c^2 - 14*r_rm2c^3 + 4*r_rm2c^4) + (x[,"n_24"] + x[,"n_32"])*log10(45 - 9*r_rm2c - 2*r_rm2c^2 - 2*r_rm2c^3 + 4*r_rm2c^4) + (x[,"n_04"] + x[,"n_52"])*log10(9 - 18*r_rm2c + 29*r_rm2c^2 - 28*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_13"] + x[,"n_43"])*log10(33 - 8*r_rm2c + 20*r_rm2c^2 - 24*r_rm2c^3 + 12*r_rm2c^4)
logL_rm2c <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - x[,"n_22"] - x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - x[,"n_32"] - x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 2*x[,"n_55"])*log(2) + (-2*x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - 2*x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_01"] + x[,"n_55"])*(log(pmax(1e-6,1 - r_rm2c)) + 2*log(pmax(1e-6,r_rm2c)) + log(2 + r_rm2c)) + (x[,"n_05"] + x[,"n_51"])*(log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c)) + log(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,r_rm2c)) + log(3 - r_rm2c^2 + r_rm2c^3)) + (x[,"n_02"] + x[,"n_54"])*(log(pmax(1e-6,r_rm2c)) + log(12 - 7*r_rm2c - 4*r_rm2c^2 + 8*r_rm2c^3)) + (x[,"n_12"] + x[,"n_44"])*log(9 + 18*r_rm2c - 7*r_rm2c^2 + 12*r_rm2c^3 - 8*r_rm2c^4) + (x[,"n_14"] + x[,"n_42"])*log(24 - 8*r_rm2c - 19*r_rm2c^2 + 20*r_rm2c^3 - 8*r_rm2c^4) + (x[,"n_03"] + x[,"n_53"])*log(3 - 4*r_rm2c^2 + 8*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_25"] + x[,"n_31"])*log(18 - 12*r_rm2c + 7*r_rm2c^2 - 2*r_rm2c^3 - 2*r_rm2c^4) + (x[,"n_21"] + x[,"n_35"])*log(9 + 12*r_rm2c - 11*r_rm2c^2 + 10*r_rm2c^3 - 2*r_rm2c^4) + (x[,"n_23"] + x[,"n_33"])*log(12 + r_rm2c - 2*r_rm2c^2 + 2*r_rm2c^3 - r_rm2c^4) + (x[,"n_15"] + x[,"n_41"])*log(3 - 4*r_rm2c + 3*r_rm2c^2 - 3*r_rm2c^3 + r_rm2c^4) + (x[,"n_22"] + x[,"n_34"])*log(36 + 3*r_rm2c + 16*r_rm2c^2 - 14*r_rm2c^3 + 4*r_rm2c^4) + (x[,"n_24"] + x[,"n_32"])*log(45 - 9*r_rm2c - 2*r_rm2c^2 - 2*r_rm2c^3 + 4*r_rm2c^4) + (x[,"n_04"] + x[,"n_52"])*log(9 - 18*r_rm2c + 29*r_rm2c^2 - 28*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_13"] + x[,"n_43"])*log(33 - 8*r_rm2c + 20*r_rm2c^2 - 24*r_rm2c^3 + 12*r_rm2c^4)
logL_rm1c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-3*n01 - 3*n02 - 2*n03 - 3*n04 - 3*n05 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 3*n25 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 3*n51 - 3*n52 - 2*n53 - 3*n54 - 3*n55)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n01 + n55)*(3*log(pmax(1e-6,r)) + log(2 + r)) + (n04 + n52)*(2*log(pmax(1e-6,1 - r)) + log(3 + 8*r - 4*r^2)) + (n15 + n41)*(2*log(pmax(1e-6,1 - r)) + log(12 + 2*r - 3*r^2)) + (n11 + n45)*(2*log(pmax(1e-6,r)) + log(11 + 4*r - 3*r^2)) + (n05 + n51)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + r^2)) + (n21 + n35)*(log(pmax(1e-6,r)) + log(18 + 7*r - 18*r^2 + 2*r^3)) + (n03 + n53)*(log(pmax(1e-6,r)) + log(4 - r - 6*r^2 + 3*r^3)) + (n12 + n44)*(log(pmax(1e-6,r)) + log(30 + 17*r - 20*r^2 + 12*r^3)) + (n13 + n43)*log(21 + 44*r - 62*r^2 + 36*r^3 - 18*r^4) + (n24 + n32)*log(75 - 52*r + 24*r^2 - 12*r^3 - 8*r^4) + (n22 + n34)*log(27 + 72*r - 60*r^2 + 44*r^3 - 8*r^4) + (n02 + n54)*log(7*r^2 - 4*r^4) + (n25 + n31)*log(9 + 14*r - 35*r^2 + 10*r^3 + 2*r^4) + (n14 + n42)*log(39 - 52*r + 29*r^2 - 28*r^3 + 12*r^4) + (n23 + n33)*log(111 - 52*r + 64*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_rm1c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_rm1c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rm1c <- -((x[,"n_04"] + x[,"n_52"])*(-log10(2) + log10(3))) - (x[,"n_02"] + x[,"n_54"])*(-log10(2) + log10(3)) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_42"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_44"])*(-log10(2) + log10(5) + log10(7)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_03"] + x[,"n_53"])*(-4*log10(2) + log10(19)) - (x[,"n_13"] + x[,"n_43"])*(-3*log10(2) + log10(13) + log10(19)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(3) + log10(23)) - (x[,"n_24"] + x[,"n_32"])*log10(53) - (x[,"n_22"] + x[,"n_34"])*log10(53) - (x[,"n_23"] + x[,"n_33"])*(-2*log10(2) + log10(5) + log10(79)) + (x[,"n_01"] + x[,"n_55"])*(3*log10(pmax(1e-6,r_rm1c)) + log10(2 + r_rm1c)) + (x[,"n_04"] + x[,"n_52"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 + 8*r_rm1c - 4*r_rm1c^2)) + (x[,"n_15"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(12 + 2*r_rm1c - 3*r_rm1c^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(11 + 4*r_rm1c - 3*r_rm1c^2)) + (x[,"n_05"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 - 4*r_rm1c + r_rm1c^2)) + (x[,"n_21"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm1c)) + log10(18 + 7*r_rm1c - 18*r_rm1c^2 + 2*r_rm1c^3)) + (x[,"n_03"] + x[,"n_53"])*(log10(pmax(1e-6,r_rm1c)) + log10(4 - r_rm1c - 6*r_rm1c^2 + 3*r_rm1c^3)) + (x[,"n_12"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm1c)) + log10(30 + 17*r_rm1c - 20*r_rm1c^2 + 12*r_rm1c^3)) + (x[,"n_13"] + x[,"n_43"])*log10(21 + 44*r_rm1c - 62*r_rm1c^2 + 36*r_rm1c^3 - 18*r_rm1c^4) + (x[,"n_24"] + x[,"n_32"])*log10(75 - 52*r_rm1c + 24*r_rm1c^2 - 12*r_rm1c^3 - 8*r_rm1c^4) + (x[,"n_22"] + x[,"n_34"])*log10(27 + 72*r_rm1c - 60*r_rm1c^2 + 44*r_rm1c^3 - 8*r_rm1c^4) + (x[,"n_02"] + x[,"n_54"])*log10(7*r_rm1c^2 - 4*r_rm1c^4) + (x[,"n_25"] + x[,"n_31"])*log10(9 + 14*r_rm1c - 35*r_rm1c^2 + 10*r_rm1c^3 + 2*r_rm1c^4) + (x[,"n_14"] + x[,"n_42"])*log10(39 - 52*r_rm1c + 29*r_rm1c^2 - 28*r_rm1c^3 + 12*r_rm1c^4) + (x[,"n_23"] + x[,"n_33"])*log10(111 - 52*r_rm1c + 64*r_rm1c^2 - 24*r_rm1c^3 + 12*r_rm1c^4)
logL_rm1c <- (-3*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 2*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_55"])*(3*log(pmax(1e-6,r_rm1c)) + log(2 + r_rm1c)) + (x[,"n_04"] + x[,"n_52"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(3 + 8*r_rm1c - 4*r_rm1c^2)) + (x[,"n_15"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(12 + 2*r_rm1c - 3*r_rm1c^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log(pmax(1e-6,r_rm1c)) + log(11 + 4*r_rm1c - 3*r_rm1c^2)) + (x[,"n_05"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(3 - 4*r_rm1c + r_rm1c^2)) + (x[,"n_21"] + x[,"n_35"])*(log(pmax(1e-6,r_rm1c)) + log(18 + 7*r_rm1c - 18*r_rm1c^2 + 2*r_rm1c^3)) + (x[,"n_03"] + x[,"n_53"])*(log(pmax(1e-6,r_rm1c)) + log(4 - r_rm1c - 6*r_rm1c^2 + 3*r_rm1c^3)) + (x[,"n_12"] + x[,"n_44"])*(log(pmax(1e-6,r_rm1c)) + log(30 + 17*r_rm1c - 20*r_rm1c^2 + 12*r_rm1c^3)) + (x[,"n_13"] + x[,"n_43"])*log(21 + 44*r_rm1c - 62*r_rm1c^2 + 36*r_rm1c^3 - 18*r_rm1c^4) + (x[,"n_24"] + x[,"n_32"])*log(75 - 52*r_rm1c + 24*r_rm1c^2 - 12*r_rm1c^3 - 8*r_rm1c^4) + (x[,"n_22"] + x[,"n_34"])*log(27 + 72*r_rm1c - 60*r_rm1c^2 + 44*r_rm1c^3 - 8*r_rm1c^4) + (x[,"n_02"] + x[,"n_54"])*log(7*r_rm1c^2 - 4*r_rm1c^4) + (x[,"n_25"] + x[,"n_31"])*log(9 + 14*r_rm1c - 35*r_rm1c^2 + 10*r_rm1c^3 + 2*r_rm1c^4) + (x[,"n_14"] + x[,"n_42"])*log(39 - 52*r_rm1c + 29*r_rm1c^2 - 28*r_rm1c^3 + 12*r_rm1c^4) + (x[,"n_23"] + x[,"n_33"])*log(111 - 52*r_rm1c + 64*r_rm1c^2 - 24*r_rm1c^3 + 12*r_rm1c^4)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_mc,r_mm2c,r_mm1c,r_rc,r_rm2c,r_rm1c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_mc,LOD_mm2c,LOD_mm1c,LOD_rc,LOD_rm2c,LOD_rm1c,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_mc,logL_mm2c,logL_mm1c,logL_rc,logL_rm2c,logL_rm1c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","mixed coupling","mixed mixed2c","mixed mixed1c","repulsion coupling","repulsion mixed2c","repulsion mixed1c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.3_3.2 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-4*n00 - 4*n01 - 3*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 3*n12 - 3*n13 - 4*n14 - 4*n15 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 3*n25 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 4*n40 - 4*n41 - 3*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n50 - 4*n51 - 3*n52 - 3*n53 - 4*n54 - 4*n55)*log(2) + (n01 + n04 + n10 + n15 + n40 + n45 + n51 + n54)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*log(5) + 4*(n00 + n55)*log(pmax(1e-6,1 - r)) + 4*(n05 + n50)*log(pmax(1e-6,r)) + (n04 + n15 + n40 + n51)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n01 + n10 + n45 + n54)*(3*log(pmax(1e-6,1 - r)) + log(1 + r)) + (n02 + n20 + n35 + n53)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(5 + r)) + (n03 + n25 + n30 + n52)*(2*log(pmax(1e-6,r)) + log(6 - 7*r + r^2)) + (n14 + n41)*(2*log(pmax(1e-6,r)) + log(32 - 20*r + 9*r^2)) + (n11 + n44)*(2*log(pmax(1e-6,1 - r)) + log(21 + 2*r + 9*r^2)) + (n13 + n24 + n31 + n42)*(log(pmax(1e-6,r)) + log(25 - 19*r + 9*r^2 - 3*r^3)) + (n12 + n21 + n34 + n43)*log(12 - 2*r - 10*r^2 + 3*r^3 - 3*r^4) + (n22 + n33)*log(39 - 38*r + 40*r^2 - 20*r^3 + 2*r^4) + (n23 + n32)*log(23 + 10*r - 8*r^2 + 12*r^3 + 2*r^4)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_cc <- 4*(x[,"n_05"] + x[,"n_50"])*log10(2) + 4*(x[,"n_00"] + x[,"n_55"])*log10(2) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-4*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-4*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(11)) - (x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(13) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-3*log10(2) + log10(13) + log10(17)) - (x[,"n_14"] + x[,"n_41"])*(-4*log10(2) + log10(97)) - (x[,"n_11"] + x[,"n_44"])*(-4*log10(2) + log10(97)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(139)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(139)) + 4*(x[,"n_00"] + x[,"n_55"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_50"])*log10(pmax(1e-6,r_cc)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(log10(2 - r_cc) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(1 + r_cc)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc)) + log10(5 + r_cc)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_cc)) + log10(6 - 7*r_cc + r_cc^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cc)) + log10(32 - 20*r_cc + 9*r_cc^2)) + (x[,"n_11"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(21 + 2*r_cc + 9*r_cc^2)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log10(pmax(1e-6,r_cc)) + log10(25 - 19*r_cc + 9*r_cc^2 - 3*r_cc^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log10(12 - 2*r_cc - 10*r_cc^2 + 3*r_cc^3 - 3*r_cc^4) + (x[,"n_22"] + x[,"n_33"])*log10(39 - 38*r_cc + 40*r_cc^2 - 20*r_cc^3 + 2*r_cc^4) + (x[,"n_23"] + x[,"n_32"])*log10(23 + 10*r_cc - 8*r_cc^2 + 12*r_cc^3 + 2*r_cc^4)
logL_cc <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_10"] + x[,"n_15"] + x[,"n_40"] + x[,"n_45"] + x[,"n_51"] + x[,"n_54"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + 4*(x[,"n_00"] + x[,"n_55"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_50"])*log(pmax(1e-6,r_cc)) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(log(2 - r_cc) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(3*log(pmax(1e-6,1 - r_cc)) + log(1 + r_cc)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc)) + log(5 + r_cc)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log(pmax(1e-6,r_cc)) + log(6 - 7*r_cc + r_cc^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cc)) + log(32 - 20*r_cc + 9*r_cc^2)) + (x[,"n_11"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(21 + 2*r_cc + 9*r_cc^2)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(log(pmax(1e-6,r_cc)) + log(25 - 19*r_cc + 9*r_cc^2 - 3*r_cc^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*log(12 - 2*r_cc - 10*r_cc^2 + 3*r_cc^3 - 3*r_cc^4) + (x[,"n_22"] + x[,"n_33"])*log(39 - 38*r_cc + 40*r_cc^2 - 20*r_cc^3 + 2*r_cc^4) + (x[,"n_23"] + x[,"n_32"])*log(23 + 10*r_cc - 8*r_cc^2 + 12*r_cc^3 + 2*r_cc^4)
logL_cm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-3*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 3*n05 - 3*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 3*n15 - 3*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 3*n25 - 3*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 3*n35 - 3*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 3*n45 - 3*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 3*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + 4*(n00 + n55)*log(pmax(1e-6,r)) + (n05 + n50)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n15 + n40)*(2*log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n10 + n45)*(2*log(pmax(1e-6,1 - r)) + log(3 - r + 3*r^2)) + (n04 + n51)*(2*log(pmax(1e-6,r)) + log(11 - 14*r + 6*r^2)) + (n01 + n54)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 6*r^2)) + (n14 + n41)*(log(pmax(1e-6,r)) + log(42 - 29*r + 38*r^2 - 18*r^3)) + (n03 + n52)*(log(pmax(1e-6,r)) + log(18 - 29*r + 24*r^2 - 4*r^3)) + (n25 + n30)*(log(pmax(1e-6,r)) + log(7 - 8*r + 6*r^2 - 2*r^3)) + (n11 + n44)*log(33 - 26*r - 23*r^2 + 34*r^3 - 18*r^4) + (n23 + n32)*log(153 + 8*r + 40*r^2 - 16*r^3 - 8*r^4) + (n22 + n33)*log(177 - 8*r - 56*r^2 + 48*r^3 - 8*r^4) + (n02 + n53)*log(9 - 16*r + 19*r^2 - 8*r^3 - 4*r^4) + (n20 + n35)*log(3 - r - 2*r^2 + 2*r^3 - 2*r^4) + (n13 + n42)*log(36 + 24*r + 37*r^2 - 40*r^3 + 12*r^4) + (n21 + n34)*log(69 - 38*r + 24*r^2 - 28*r^3 + 12*r^4) + (n24 + n31)*log(39 + 26*r + 12*r^2 - 20*r^3 + 12*r^4) + (n12 + n43)*log(69 - 26*r - 11*r^2 - 8*r^3 + 12*r^4)
return(L)}
interlogL_cm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_cm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm <- mapply(interlogL_cm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_cm <- 4*(x[,"n_05"] + x[,"n_50"])*log10(2) + 4*(x[,"n_00"] + x[,"n_55"])*log10(2) - (x[,"n_03"] + x[,"n_52"])*(-log10(2) + 2*log10(3)) - (x[,"n_02"] + x[,"n_53"])*(-log10(2) + 2*log10(3)) - (x[,"n_04"] + x[,"n_51"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_54"])*(-3*log10(2) + log10(11)) - (x[,"n_15"] + x[,"n_40"])*(-4*log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_45"])*(-4*log10(2) + log10(13)) - (x[,"n_25"] + x[,"n_30"])*(-3*log10(2) + log10(17)) - (x[,"n_20"] + x[,"n_35"])*(-3*log10(2) + log10(17)) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(7) + log10(47)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(7) + log10(47)) - (x[,"n_13"] + x[,"n_42"])*log10(53) - (x[,"n_12"] + x[,"n_43"])*log10(53) - (x[,"n_24"] + x[,"n_31"])*(-2*log10(2) + log10(3) + log10(71)) - (x[,"n_21"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(71)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(139)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(139)) + 4*(x[,"n_00"] + x[,"n_55"])*log10(pmax(1e-6,r_cm)) + (x[,"n_05"] + x[,"n_50"])*(log10(pmax(1e-6,1 - r_cm)) + 3*log10(pmax(1e-6,r_cm))) + (x[,"n_15"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_cm)) + log10(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_10"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(3 - r_cm + 3*r_cm^2)) + (x[,"n_04"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_cm)) + log10(11 - 14*r_cm + 6*r_cm^2)) + (x[,"n_01"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_cm)) + log10(3 + 2*r_cm + 6*r_cm^2)) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm)) + log10(42 - 29*r_cm + 38*r_cm^2 - 18*r_cm^3)) + (x[,"n_03"] + x[,"n_52"])*(log10(pmax(1e-6,r_cm)) + log10(18 - 29*r_cm + 24*r_cm^2 - 4*r_cm^3)) + (x[,"n_25"] + x[,"n_30"])*(log10(pmax(1e-6,r_cm)) + log10(7 - 8*r_cm + 6*r_cm^2 - 2*r_cm^3)) + (x[,"n_11"] + x[,"n_44"])*log10(33 - 26*r_cm - 23*r_cm^2 + 34*r_cm^3 - 18*r_cm^4) + (x[,"n_23"] + x[,"n_32"])*log10(153 + 8*r_cm + 40*r_cm^2 - 16*r_cm^3 - 8*r_cm^4) + (x[,"n_22"] + x[,"n_33"])*log10(177 - 8*r_cm - 56*r_cm^2 + 48*r_cm^3 - 8*r_cm^4) + (x[,"n_02"] + x[,"n_53"])*log10(9 - 16*r_cm + 19*r_cm^2 - 8*r_cm^3 - 4*r_cm^4) + (x[,"n_20"] + x[,"n_35"])*log10(3 - r_cm - 2*r_cm^2 + 2*r_cm^3 - 2*r_cm^4) + (x[,"n_13"] + x[,"n_42"])*log10(36 + 24*r_cm + 37*r_cm^2 - 40*r_cm^3 + 12*r_cm^4) + (x[,"n_21"] + x[,"n_34"])*log10(69 - 38*r_cm + 24*r_cm^2 - 28*r_cm^3 + 12*r_cm^4) + (x[,"n_24"] + x[,"n_31"])*log10(39 + 26*r_cm + 12*r_cm^2 - 20*r_cm^3 + 12*r_cm^4) + (x[,"n_12"] + x[,"n_43"])*log10(69 - 26*r_cm - 11*r_cm^2 - 8*r_cm^3 + 12*r_cm^4)
logL_cm <- (-3*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + 4*(x[,"n_00"] + x[,"n_55"])*log(pmax(1e-6,r_cm)) + (x[,"n_05"] + x[,"n_50"])*(log(pmax(1e-6,1 - r_cm)) + 3*log(pmax(1e-6,r_cm))) + (x[,"n_15"] + x[,"n_40"])*(2*log(pmax(1e-6,r_cm)) + log(5 - 5*r_cm + 3*r_cm^2)) + (x[,"n_10"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm)) + log(3 - r_cm + 3*r_cm^2)) + (x[,"n_04"] + x[,"n_51"])*(2*log(pmax(1e-6,r_cm)) + log(11 - 14*r_cm + 6*r_cm^2)) + (x[,"n_01"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_cm)) + log(3 + 2*r_cm + 6*r_cm^2)) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,r_cm)) + log(42 - 29*r_cm + 38*r_cm^2 - 18*r_cm^3)) + (x[,"n_03"] + x[,"n_52"])*(log(pmax(1e-6,r_cm)) + log(18 - 29*r_cm + 24*r_cm^2 - 4*r_cm^3)) + (x[,"n_25"] + x[,"n_30"])*(log(pmax(1e-6,r_cm)) + log(7 - 8*r_cm + 6*r_cm^2 - 2*r_cm^3)) + (x[,"n_11"] + x[,"n_44"])*log(33 - 26*r_cm - 23*r_cm^2 + 34*r_cm^3 - 18*r_cm^4) + (x[,"n_23"] + x[,"n_32"])*log(153 + 8*r_cm + 40*r_cm^2 - 16*r_cm^3 - 8*r_cm^4) + (x[,"n_22"] + x[,"n_33"])*log(177 - 8*r_cm - 56*r_cm^2 + 48*r_cm^3 - 8*r_cm^4) + (x[,"n_02"] + x[,"n_53"])*log(9 - 16*r_cm + 19*r_cm^2 - 8*r_cm^3 - 4*r_cm^4) + (x[,"n_20"] + x[,"n_35"])*log(3 - r_cm - 2*r_cm^2 + 2*r_cm^3 - 2*r_cm^4) + (x[,"n_13"] + x[,"n_42"])*log(36 + 24*r_cm + 37*r_cm^2 - 40*r_cm^3 + 12*r_cm^4) + (x[,"n_21"] + x[,"n_34"])*log(69 - 38*r_cm + 24*r_cm^2 - 28*r_cm^3 + 12*r_cm^4) + (x[,"n_24"] + x[,"n_31"])*log(39 + 26*r_cm + 12*r_cm^2 - 20*r_cm^3 + 12*r_cm^4) + (x[,"n_12"] + x[,"n_43"])*log(69 - 26*r_cm - 11*r_cm^2 - 8*r_cm^3 + 12*r_cm^4)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 4*n55)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*log(5) + (n00 + n05 + n50 + n55)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n15 + n40 + n54)*(log(pmax(1e-6,r)) + log(2 - 3*r + 4*r^2 - 3*r^3)) + (n04 + n10 + n45 + n51)*(log(pmax(1e-6,r)) + log(4 - 9*r + 8*r^2 - 3*r^3)) + (n12 + n24 + n31 + n43)*log(21 - 12*r + 10*r^2 + 2*r^3 - 6*r^4) + (n13 + n21 + n34 + n42)*log(15 + 10*r - 20*r^2 + 22*r^3 - 6*r^4) + (n03 + n20 + n35 + n52)*log(3 - 8*r + 14*r^2 - 10*r^3 + 2*r^4) + (n22 + n23 + n32 + n33)*log(27 + 2*r - 4*r^3 + 2*r^4) + (n02 + n25 + n30 + n53)*log(1 + 2*r - 4*r^2 + 2*r^3 + 2*r^4) + (n11 + n14 + n41 + n44)*log(6 - 2*r + 11*r^2 - 18*r^3 + 9*r^4)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_cr <- 4*(x[,"n_00"] + x[,"n_05"] + x[,"n_50"] + x[,"n_55"])*log10(2) - (x[,"n_04"] + x[,"n_10"] + x[,"n_45"] + x[,"n_51"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_15"] + x[,"n_40"] + x[,"n_54"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_20"] + x[,"n_35"] + x[,"n_52"])*(-3*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_25"] + x[,"n_30"] + x[,"n_53"])*(-3*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*(-3*log10(2) + log10(13) + log10(17)) - (x[,"n_11"] + x[,"n_14"] + x[,"n_41"] + x[,"n_44"])*(-4*log10(2) + log10(97)) - (x[,"n_13"] + x[,"n_21"] + x[,"n_34"] + x[,"n_42"])*(-3*log10(2) + log10(139)) - (x[,"n_12"] + x[,"n_24"] + x[,"n_31"] + x[,"n_43"])*(-3*log10(2) + log10(139)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_50"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_15"] + x[,"n_40"] + x[,"n_54"])*(log10(pmax(1e-6,r_cr)) + log10(2 - 3*r_cr + 4*r_cr^2 - 3*r_cr^3)) + (x[,"n_04"] + x[,"n_10"] + x[,"n_45"] + x[,"n_51"])*(log10(pmax(1e-6,r_cr)) + log10(4 - 9*r_cr + 8*r_cr^2 - 3*r_cr^3)) + (x[,"n_12"] + x[,"n_24"] + x[,"n_31"] + x[,"n_43"])*log10(21 - 12*r_cr + 10*r_cr^2 + 2*r_cr^3 - 6*r_cr^4) + (x[,"n_13"] + x[,"n_21"] + x[,"n_34"] + x[,"n_42"])*log10(15 + 10*r_cr - 20*r_cr^2 + 22*r_cr^3 - 6*r_cr^4) + (x[,"n_03"] + x[,"n_20"] + x[,"n_35"] + x[,"n_52"])*log10(3 - 8*r_cr + 14*r_cr^2 - 10*r_cr^3 + 2*r_cr^4) + (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*log10(27 + 2*r_cr - 4*r_cr^3 + 2*r_cr^4) + (x[,"n_02"] + x[,"n_25"] + x[,"n_30"] + x[,"n_53"])*log10(1 + 2*r_cr - 4*r_cr^2 + 2*r_cr^3 + 2*r_cr^4) + (x[,"n_11"] + x[,"n_14"] + x[,"n_41"] + x[,"n_44"])*log10(6 - 2*r_cr + 11*r_cr^2 - 18*r_cr^3 + 9*r_cr^4)
logL_cr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_00"] + x[,"n_05"] + x[,"n_50"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_15"] + x[,"n_40"] + x[,"n_54"])*(log(pmax(1e-6,r_cr)) + log(2 - 3*r_cr + 4*r_cr^2 - 3*r_cr^3)) + (x[,"n_04"] + x[,"n_10"] + x[,"n_45"] + x[,"n_51"])*(log(pmax(1e-6,r_cr)) + log(4 - 9*r_cr + 8*r_cr^2 - 3*r_cr^3)) + (x[,"n_12"] + x[,"n_24"] + x[,"n_31"] + x[,"n_43"])*log(21 - 12*r_cr + 10*r_cr^2 + 2*r_cr^3 - 6*r_cr^4) + (x[,"n_13"] + x[,"n_21"] + x[,"n_34"] + x[,"n_42"])*log(15 + 10*r_cr - 20*r_cr^2 + 22*r_cr^3 - 6*r_cr^4) + (x[,"n_03"] + x[,"n_20"] + x[,"n_35"] + x[,"n_52"])*log(3 - 8*r_cr + 14*r_cr^2 - 10*r_cr^3 + 2*r_cr^4) + (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*log(27 + 2*r_cr - 4*r_cr^3 + 2*r_cr^4) + (x[,"n_02"] + x[,"n_25"] + x[,"n_30"] + x[,"n_53"])*log(1 + 2*r_cr - 4*r_cr^2 + 2*r_cr^3 + 2*r_cr^4) + (x[,"n_11"] + x[,"n_14"] + x[,"n_41"] + x[,"n_44"])*log(6 - 2*r_cr + 11*r_cr^2 - 18*r_cr^3 + 9*r_cr^4)
logL_mc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n00 + n55)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n50)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n04 + n51)*(2*log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n01 + n54)*(2*log(pmax(1e-6,1 - r)) + log(3 - r + 3*r^2)) + (n15 + n40)*(2*log(pmax(1e-6,r)) + log(11 - 14*r + 6*r^2)) + (n10 + n45)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 6*r^2)) + (n14 + n41)*(log(pmax(1e-6,r)) + log(42 - 29*r + 38*r^2 - 18*r^3)) + (n25 + n30)*(log(pmax(1e-6,r)) + log(18 - 29*r + 24*r^2 - 4*r^3)) + (n03 + n52)*(log(pmax(1e-6,r)) + log(7 - 8*r + 6*r^2 - 2*r^3)) + (n11 + n44)*log(33 - 26*r - 23*r^2 + 34*r^3 - 18*r^4) + (n23 + n32)*log(153 + 8*r + 40*r^2 - 16*r^3 - 8*r^4) + (n22 + n33)*log(177 - 8*r - 56*r^2 + 48*r^3 - 8*r^4) + (n20 + n35)*log(9 - 16*r + 19*r^2 - 8*r^3 - 4*r^4) + (n02 + n53)*log(3 - r - 2*r^2 + 2*r^3 - 2*r^4) + (n24 + n31)*log(36 + 24*r + 37*r^2 - 40*r^3 + 12*r^4) + (n12 + n43)*log(69 - 38*r + 24*r^2 - 28*r^3 + 12*r^4) + (n13 + n42)*log(39 + 26*r + 12*r^2 - 20*r^3 + 12*r^4) + (n21 + n34)*log(69 - 26*r - 11*r^2 - 8*r^3 + 12*r^4)
return(L)}
interlogL_mc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_mc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mc <- 4*(x[,"n_05"] + x[,"n_50"])*log10(2) + 4*(x[,"n_00"] + x[,"n_55"])*log10(2) - (x[,"n_25"] + x[,"n_30"])*(-log10(2) + 2*log10(3)) - (x[,"n_20"] + x[,"n_35"])*(-log10(2) + 2*log10(3)) - (x[,"n_15"] + x[,"n_40"])*(-3*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_45"])*(-3*log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_51"])*(-4*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_54"])*(-4*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_52"])*(-3*log10(2) + log10(17)) - (x[,"n_02"] + x[,"n_53"])*(-3*log10(2) + log10(17)) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(7) + log10(47)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(7) + log10(47)) - (x[,"n_24"] + x[,"n_31"])*log10(53) - (x[,"n_21"] + x[,"n_34"])*log10(53) - (x[,"n_13"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(71)) - (x[,"n_12"] + x[,"n_43"])*(-2*log10(2) + log10(3) + log10(71)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(139)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(139)) + (x[,"n_00"] + x[,"n_55"])*(3*log10(pmax(1e-6,1 - r_mc)) + log10(pmax(1e-6,r_mc))) + (x[,"n_05"] + x[,"n_50"])*(log10(pmax(1e-6,1 - r_mc)) + 3*log10(pmax(1e-6,r_mc))) + (x[,"n_04"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_mc)) + log10(5 - 5*r_mc + 3*r_mc^2)) + (x[,"n_01"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 - r_mc + 3*r_mc^2)) + (x[,"n_15"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_mc)) + log10(11 - 14*r_mc + 6*r_mc^2)) + (x[,"n_10"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 + 2*r_mc + 6*r_mc^2)) + (x[,"n_14"] + x[,"n_41"])*(log10(pmax(1e-6,r_mc)) + log10(42 - 29*r_mc + 38*r_mc^2 - 18*r_mc^3)) + (x[,"n_25"] + x[,"n_30"])*(log10(pmax(1e-6,r_mc)) + log10(18 - 29*r_mc + 24*r_mc^2 - 4*r_mc^3)) + (x[,"n_03"] + x[,"n_52"])*(log10(pmax(1e-6,r_mc)) + log10(7 - 8*r_mc + 6*r_mc^2 - 2*r_mc^3)) + (x[,"n_11"] + x[,"n_44"])*log10(33 - 26*r_mc - 23*r_mc^2 + 34*r_mc^3 - 18*r_mc^4) + (x[,"n_23"] + x[,"n_32"])*log10(153 + 8*r_mc + 40*r_mc^2 - 16*r_mc^3 - 8*r_mc^4) + (x[,"n_22"] + x[,"n_33"])*log10(177 - 8*r_mc - 56*r_mc^2 + 48*r_mc^3 - 8*r_mc^4) + (x[,"n_20"] + x[,"n_35"])*log10(9 - 16*r_mc + 19*r_mc^2 - 8*r_mc^3 - 4*r_mc^4) + (x[,"n_02"] + x[,"n_53"])*log10(3 - r_mc - 2*r_mc^2 + 2*r_mc^3 - 2*r_mc^4) + (x[,"n_24"] + x[,"n_31"])*log10(36 + 24*r_mc + 37*r_mc^2 - 40*r_mc^3 + 12*r_mc^4) + (x[,"n_12"] + x[,"n_43"])*log10(69 - 38*r_mc + 24*r_mc^2 - 28*r_mc^3 + 12*r_mc^4) + (x[,"n_13"] + x[,"n_42"])*log10(39 + 26*r_mc + 12*r_mc^2 - 20*r_mc^3 + 12*r_mc^4) + (x[,"n_21"] + x[,"n_34"])*log10(69 - 26*r_mc - 11*r_mc^2 - 8*r_mc^3 + 12*r_mc^4)
logL_mc <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_55"])*(3*log(pmax(1e-6,1 - r_mc)) + log(pmax(1e-6,r_mc))) + (x[,"n_05"] + x[,"n_50"])*(log(pmax(1e-6,1 - r_mc)) + 3*log(pmax(1e-6,r_mc))) + (x[,"n_04"] + x[,"n_51"])*(2*log(pmax(1e-6,r_mc)) + log(5 - 5*r_mc + 3*r_mc^2)) + (x[,"n_01"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 - r_mc + 3*r_mc^2)) + (x[,"n_15"] + x[,"n_40"])*(2*log(pmax(1e-6,r_mc)) + log(11 - 14*r_mc + 6*r_mc^2)) + (x[,"n_10"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 + 2*r_mc + 6*r_mc^2)) + (x[,"n_14"] + x[,"n_41"])*(log(pmax(1e-6,r_mc)) + log(42 - 29*r_mc + 38*r_mc^2 - 18*r_mc^3)) + (x[,"n_25"] + x[,"n_30"])*(log(pmax(1e-6,r_mc)) + log(18 - 29*r_mc + 24*r_mc^2 - 4*r_mc^3)) + (x[,"n_03"] + x[,"n_52"])*(log(pmax(1e-6,r_mc)) + log(7 - 8*r_mc + 6*r_mc^2 - 2*r_mc^3)) + (x[,"n_11"] + x[,"n_44"])*log(33 - 26*r_mc - 23*r_mc^2 + 34*r_mc^3 - 18*r_mc^4) + (x[,"n_23"] + x[,"n_32"])*log(153 + 8*r_mc + 40*r_mc^2 - 16*r_mc^3 - 8*r_mc^4) + (x[,"n_22"] + x[,"n_33"])*log(177 - 8*r_mc - 56*r_mc^2 + 48*r_mc^3 - 8*r_mc^4) + (x[,"n_20"] + x[,"n_35"])*log(9 - 16*r_mc + 19*r_mc^2 - 8*r_mc^3 - 4*r_mc^4) + (x[,"n_02"] + x[,"n_53"])*log(3 - r_mc - 2*r_mc^2 + 2*r_mc^3 - 2*r_mc^4) + (x[,"n_24"] + x[,"n_31"])*log(36 + 24*r_mc + 37*r_mc^2 - 40*r_mc^3 + 12*r_mc^4) + (x[,"n_12"] + x[,"n_43"])*log(69 - 38*r_mc + 24*r_mc^2 - 28*r_mc^3 + 12*r_mc^4) + (x[,"n_13"] + x[,"n_42"])*log(39 + 26*r_mc + 12*r_mc^2 - 20*r_mc^3 + 12*r_mc^4) + (x[,"n_21"] + x[,"n_34"])*log(69 - 26*r_mc - 11*r_mc^2 - 8*r_mc^3 + 12*r_mc^4)
logL_mm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-2*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 2*n05 - 3*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 3*n15 - 3*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 3*n25 - 3*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 3*n35 - 3*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 3*n45 - 2*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 2*n55)*log(2) + (-2*n00 - n01 - 2*n02 - 2*n03 - n04 - 2*n05 - n10 - n11 - n12 - n13 - n14 - n15 - 2*n20 - n21 - 2*n22 - 2*n23 - n24 - 2*n25 - 2*n30 - n31 - 2*n32 - 2*n33 - n34 - 2*n35 - n40 - n41 - n42 - n43 - n44 - n45 - 2*n50 - n51 - 2*n52 - 2*n53 - n54 - 2*n55)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*log(5) + (n00 + n05 + n50 + n55)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n04 + n10 + n15 + n40 + n45 + n51 + n54)*(log(pmax(1e-6,r)) + log(3 - 5*r + 4*r^2 - 2*r^3)) + (n12 + n13 + n21 + n24 + n31 + n34 + n42 + n43)*log(57 - 10*r + 2*r^2 + 16*r^3 - 8*r^4) + (n02 + n03 + n20 + n25 + n30 + n35 + n52 + n53)*log(9 - 9*r + 13*r^2 - 8*r^3 + 4*r^4) + (n11 + n14 + n41 + n44)*log(15 + 4*r + 8*r^2 - 24*r^3 + 12*r^4) + (n22 + n23 + n32 + n33)*log(477 + 48*r - 32*r^2 - 32*r^3 + 16*r^4)
return(L)}
interlogL_mm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_mm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm <- mapply(interlogL_mm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mm <- 4*(x[,"n_00"] + x[,"n_05"] + x[,"n_50"] + x[,"n_55"])*log10(2) - (x[,"n_12"] + x[,"n_13"] + x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_42"] + x[,"n_43"])*(log10(2) + 3*log10(3)) - (x[,"n_01"] + x[,"n_04"] + x[,"n_10"] + x[,"n_15"] + x[,"n_40"] + x[,"n_45"] + x[,"n_51"] + x[,"n_54"])*(-3*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_03"] + x[,"n_20"] + x[,"n_25"] + x[,"n_30"] + x[,"n_35"] + x[,"n_52"] + x[,"n_53"])*log10(7) - (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*(log10(2) + log10(5) + 2*log10(7)) - (x[,"n_11"] + x[,"n_14"] + x[,"n_41"] + x[,"n_44"])*(-2*log10(2) + log10(67)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_50"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_mm)) + 2*log10(pmax(1e-6,r_mm))) + (x[,"n_01"] + x[,"n_04"] + x[,"n_10"] + x[,"n_15"] + x[,"n_40"] + x[,"n_45"] + x[,"n_51"] + x[,"n_54"])*(log10(pmax(1e-6,r_mm)) + log10(3 - 5*r_mm + 4*r_mm^2 - 2*r_mm^3)) + (x[,"n_12"] + x[,"n_13"] + x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_42"] + x[,"n_43"])*log10(57 - 10*r_mm + 2*r_mm^2 + 16*r_mm^3 - 8*r_mm^4) + (x[,"n_02"] + x[,"n_03"] + x[,"n_20"] + x[,"n_25"] + x[,"n_30"] + x[,"n_35"] + x[,"n_52"] + x[,"n_53"])*log10(9 - 9*r_mm + 13*r_mm^2 - 8*r_mm^3 + 4*r_mm^4) + (x[,"n_11"] + x[,"n_14"] + x[,"n_41"] + x[,"n_44"])*log10(15 + 4*r_mm + 8*r_mm^2 - 24*r_mm^3 + 12*r_mm^4) + (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*log10(477 + 48*r_mm - 32*r_mm^2 - 32*r_mm^3 + 16*r_mm^4)
logL_mm <- (-2*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 2*x[,"n_05"] - 3*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 2*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 2*x[,"n_55"])*log(2) + (-2*x[,"n_00"] - x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - 2*x[,"n_20"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_30"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"] - 2*x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - 2*x[,"n_50"] - x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - x[,"n_54"] - 2*x[,"n_55"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_00"] + x[,"n_05"] + x[,"n_50"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_mm)) + 2*log(pmax(1e-6,r_mm))) + (x[,"n_01"] + x[,"n_04"] + x[,"n_10"] + x[,"n_15"] + x[,"n_40"] + x[,"n_45"] + x[,"n_51"] + x[,"n_54"])*(log(pmax(1e-6,r_mm)) + log(3 - 5*r_mm + 4*r_mm^2 - 2*r_mm^3)) + (x[,"n_12"] + x[,"n_13"] + x[,"n_21"] + x[,"n_24"] + x[,"n_31"] + x[,"n_34"] + x[,"n_42"] + x[,"n_43"])*log(57 - 10*r_mm + 2*r_mm^2 + 16*r_mm^3 - 8*r_mm^4) + (x[,"n_02"] + x[,"n_03"] + x[,"n_20"] + x[,"n_25"] + x[,"n_30"] + x[,"n_35"] + x[,"n_52"] + x[,"n_53"])*log(9 - 9*r_mm + 13*r_mm^2 - 8*r_mm^3 + 4*r_mm^4) + (x[,"n_11"] + x[,"n_14"] + x[,"n_41"] + x[,"n_44"])*log(15 + 4*r_mm + 8*r_mm^2 - 24*r_mm^3 + 12*r_mm^4) + (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*log(477 + 48*r_mm - 32*r_mm^2 - 32*r_mm^3 + 16*r_mm^4)
logL_mr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n05 + n50)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n55)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n54)*(2*log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n04 + n51)*(2*log(pmax(1e-6,1 - r)) + log(3 - r + 3*r^2)) + (n10 + n45)*(2*log(pmax(1e-6,r)) + log(11 - 14*r + 6*r^2)) + (n15 + n40)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 6*r^2)) + (n11 + n44)*(log(pmax(1e-6,r)) + log(42 - 29*r + 38*r^2 - 18*r^3)) + (n20 + n35)*(log(pmax(1e-6,r)) + log(18 - 29*r + 24*r^2 - 4*r^3)) + (n02 + n53)*(log(pmax(1e-6,r)) + log(7 - 8*r + 6*r^2 - 2*r^3)) + (n14 + n41)*log(33 - 26*r - 23*r^2 + 34*r^3 - 18*r^4) + (n22 + n33)*log(153 + 8*r + 40*r^2 - 16*r^3 - 8*r^4) + (n23 + n32)*log(177 - 8*r - 56*r^2 + 48*r^3 - 8*r^4) + (n25 + n30)*log(9 - 16*r + 19*r^2 - 8*r^3 - 4*r^4) + (n03 + n52)*log(3 - r - 2*r^2 + 2*r^3 - 2*r^4) + (n21 + n34)*log(36 + 24*r + 37*r^2 - 40*r^3 + 12*r^4) + (n13 + n42)*log(69 - 38*r + 24*r^2 - 28*r^3 + 12*r^4) + (n12 + n43)*log(39 + 26*r + 12*r^2 - 20*r^3 + 12*r^4) + (n24 + n31)*log(69 - 26*r - 11*r^2 - 8*r^3 + 12*r^4)
return(L)}
interlogL_mr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_mr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mr <- mapply(interlogL_mr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mr <- 4*(x[,"n_05"] + x[,"n_50"])*log10(2) + 4*(x[,"n_00"] + x[,"n_55"])*log10(2) - (x[,"n_25"] + x[,"n_30"])*(-log10(2) + 2*log10(3)) - (x[,"n_20"] + x[,"n_35"])*(-log10(2) + 2*log10(3)) - (x[,"n_15"] + x[,"n_40"])*(-3*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_45"])*(-3*log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_51"])*(-4*log10(2) + log10(13)) - (x[,"n_01"] + x[,"n_54"])*(-4*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_52"])*(-3*log10(2) + log10(17)) - (x[,"n_02"] + x[,"n_53"])*(-3*log10(2) + log10(17)) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(7) + log10(47)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(7) + log10(47)) - (x[,"n_24"] + x[,"n_31"])*log10(53) - (x[,"n_21"] + x[,"n_34"])*log10(53) - (x[,"n_13"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(71)) - (x[,"n_12"] + x[,"n_43"])*(-2*log10(2) + log10(3) + log10(71)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(139)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(139)) + (x[,"n_05"] + x[,"n_50"])*(3*log10(pmax(1e-6,1 - r_mr)) + log10(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_55"])*(log10(pmax(1e-6,1 - r_mr)) + 3*log10(pmax(1e-6,r_mr))) + (x[,"n_01"] + x[,"n_54"])*(2*log10(pmax(1e-6,r_mr)) + log10(5 - 5*r_mr + 3*r_mr^2)) + (x[,"n_04"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(3 - r_mr + 3*r_mr^2)) + (x[,"n_10"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_mr)) + log10(11 - 14*r_mr + 6*r_mr^2)) + (x[,"n_15"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(3 + 2*r_mr + 6*r_mr^2)) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_mr)) + log10(42 - 29*r_mr + 38*r_mr^2 - 18*r_mr^3)) + (x[,"n_20"] + x[,"n_35"])*(log10(pmax(1e-6,r_mr)) + log10(18 - 29*r_mr + 24*r_mr^2 - 4*r_mr^3)) + (x[,"n_02"] + x[,"n_53"])*(log10(pmax(1e-6,r_mr)) + log10(7 - 8*r_mr + 6*r_mr^2 - 2*r_mr^3)) + (x[,"n_14"] + x[,"n_41"])*log10(33 - 26*r_mr - 23*r_mr^2 + 34*r_mr^3 - 18*r_mr^4) + (x[,"n_22"] + x[,"n_33"])*log10(153 + 8*r_mr + 40*r_mr^2 - 16*r_mr^3 - 8*r_mr^4) + (x[,"n_23"] + x[,"n_32"])*log10(177 - 8*r_mr - 56*r_mr^2 + 48*r_mr^3 - 8*r_mr^4) + (x[,"n_25"] + x[,"n_30"])*log10(9 - 16*r_mr + 19*r_mr^2 - 8*r_mr^3 - 4*r_mr^4) + (x[,"n_03"] + x[,"n_52"])*log10(3 - r_mr - 2*r_mr^2 + 2*r_mr^3 - 2*r_mr^4) + (x[,"n_21"] + x[,"n_34"])*log10(36 + 24*r_mr + 37*r_mr^2 - 40*r_mr^3 + 12*r_mr^4) + (x[,"n_13"] + x[,"n_42"])*log10(69 - 38*r_mr + 24*r_mr^2 - 28*r_mr^3 + 12*r_mr^4) + (x[,"n_12"] + x[,"n_43"])*log10(39 + 26*r_mr + 12*r_mr^2 - 20*r_mr^3 + 12*r_mr^4) + (x[,"n_24"] + x[,"n_31"])*log10(69 - 26*r_mr - 11*r_mr^2 - 8*r_mr^3 + 12*r_mr^4)
logL_mr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_50"])*(3*log(pmax(1e-6,1 - r_mr)) + log(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_55"])*(log(pmax(1e-6,1 - r_mr)) + 3*log(pmax(1e-6,r_mr))) + (x[,"n_01"] + x[,"n_54"])*(2*log(pmax(1e-6,r_mr)) + log(5 - 5*r_mr + 3*r_mr^2)) + (x[,"n_04"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_mr)) + log(3 - r_mr + 3*r_mr^2)) + (x[,"n_10"] + x[,"n_45"])*(2*log(pmax(1e-6,r_mr)) + log(11 - 14*r_mr + 6*r_mr^2)) + (x[,"n_15"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_mr)) + log(3 + 2*r_mr + 6*r_mr^2)) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_mr)) + log(42 - 29*r_mr + 38*r_mr^2 - 18*r_mr^3)) + (x[,"n_20"] + x[,"n_35"])*(log(pmax(1e-6,r_mr)) + log(18 - 29*r_mr + 24*r_mr^2 - 4*r_mr^3)) + (x[,"n_02"] + x[,"n_53"])*(log(pmax(1e-6,r_mr)) + log(7 - 8*r_mr + 6*r_mr^2 - 2*r_mr^3)) + (x[,"n_14"] + x[,"n_41"])*log(33 - 26*r_mr - 23*r_mr^2 + 34*r_mr^3 - 18*r_mr^4) + (x[,"n_22"] + x[,"n_33"])*log(153 + 8*r_mr + 40*r_mr^2 - 16*r_mr^3 - 8*r_mr^4) + (x[,"n_23"] + x[,"n_32"])*log(177 - 8*r_mr - 56*r_mr^2 + 48*r_mr^3 - 8*r_mr^4) + (x[,"n_25"] + x[,"n_30"])*log(9 - 16*r_mr + 19*r_mr^2 - 8*r_mr^3 - 4*r_mr^4) + (x[,"n_03"] + x[,"n_52"])*log(3 - r_mr - 2*r_mr^2 + 2*r_mr^3 - 2*r_mr^4) + (x[,"n_21"] + x[,"n_34"])*log(36 + 24*r_mr + 37*r_mr^2 - 40*r_mr^3 + 12*r_mr^4) + (x[,"n_13"] + x[,"n_42"])*log(69 - 38*r_mr + 24*r_mr^2 - 28*r_mr^3 + 12*r_mr^4) + (x[,"n_12"] + x[,"n_43"])*log(39 + 26*r_mr + 12*r_mr^2 - 20*r_mr^3 + 12*r_mr^4) + (x[,"n_24"] + x[,"n_31"])*log(69 - 26*r_mr - 11*r_mr^2 - 8*r_mr^3 + 12*r_mr^4)
logL_rc <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n20 - 4*n21 - 3*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n30 - 4*n31 - 3*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 4*n55)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*log(5) + (n00 + n05 + n50 + n55)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n10 + n45 + n51)*(log(pmax(1e-6,r)) + log(2 - 3*r + 4*r^2 - 3*r^3)) + (n01 + n15 + n40 + n54)*(log(pmax(1e-6,r)) + log(4 - 9*r + 8*r^2 - 3*r^3)) + (n13 + n21 + n34 + n42)*log(21 - 12*r + 10*r^2 + 2*r^3 - 6*r^4) + (n12 + n24 + n31 + n43)*log(15 + 10*r - 20*r^2 + 22*r^3 - 6*r^4) + (n02 + n25 + n30 + n53)*log(3 - 8*r + 14*r^2 - 10*r^3 + 2*r^4) + (n22 + n23 + n32 + n33)*log(27 + 2*r - 4*r^3 + 2*r^4) + (n03 + n20 + n35 + n52)*log(1 + 2*r - 4*r^2 + 2*r^3 + 2*r^4) + (n11 + n14 + n41 + n44)*log(6 - 2*r + 11*r^2 - 18*r^3 + 9*r^4)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rc <- 4*(x[,"n_00"] + x[,"n_05"] + x[,"n_50"] + x[,"n_55"])*log10(2) - (x[,"n_04"] + x[,"n_10"] + x[,"n_45"] + x[,"n_51"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_15"] + x[,"n_40"] + x[,"n_54"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_20"] + x[,"n_35"] + x[,"n_52"])*(-3*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_25"] + x[,"n_30"] + x[,"n_53"])*(-3*log10(2) + log10(11)) - (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*(-3*log10(2) + log10(13) + log10(17)) - (x[,"n_11"] + x[,"n_14"] + x[,"n_41"] + x[,"n_44"])*(-4*log10(2) + log10(97)) - (x[,"n_13"] + x[,"n_21"] + x[,"n_34"] + x[,"n_42"])*(-3*log10(2) + log10(139)) - (x[,"n_12"] + x[,"n_24"] + x[,"n_31"] + x[,"n_43"])*(-3*log10(2) + log10(139)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_50"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_rc)) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_04"] + x[,"n_10"] + x[,"n_45"] + x[,"n_51"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 3*r_rc + 4*r_rc^2 - 3*r_rc^3)) + (x[,"n_01"] + x[,"n_15"] + x[,"n_40"] + x[,"n_54"])*(log10(pmax(1e-6,r_rc)) + log10(4 - 9*r_rc + 8*r_rc^2 - 3*r_rc^3)) + (x[,"n_13"] + x[,"n_21"] + x[,"n_34"] + x[,"n_42"])*log10(21 - 12*r_rc + 10*r_rc^2 + 2*r_rc^3 - 6*r_rc^4) + (x[,"n_12"] + x[,"n_24"] + x[,"n_31"] + x[,"n_43"])*log10(15 + 10*r_rc - 20*r_rc^2 + 22*r_rc^3 - 6*r_rc^4) + (x[,"n_02"] + x[,"n_25"] + x[,"n_30"] + x[,"n_53"])*log10(3 - 8*r_rc + 14*r_rc^2 - 10*r_rc^3 + 2*r_rc^4) + (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*log10(27 + 2*r_rc - 4*r_rc^3 + 2*r_rc^4) + (x[,"n_03"] + x[,"n_20"] + x[,"n_35"] + x[,"n_52"])*log10(1 + 2*r_rc - 4*r_rc^2 + 2*r_rc^3 + 2*r_rc^4) + (x[,"n_11"] + x[,"n_14"] + x[,"n_41"] + x[,"n_44"])*log10(6 - 2*r_rc + 11*r_rc^2 - 18*r_rc^3 + 9*r_rc^4)
logL_rc <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_00"] + x[,"n_05"] + x[,"n_50"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_rc)) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_04"] + x[,"n_10"] + x[,"n_45"] + x[,"n_51"])*(log(pmax(1e-6,r_rc)) + log(2 - 3*r_rc + 4*r_rc^2 - 3*r_rc^3)) + (x[,"n_01"] + x[,"n_15"] + x[,"n_40"] + x[,"n_54"])*(log(pmax(1e-6,r_rc)) + log(4 - 9*r_rc + 8*r_rc^2 - 3*r_rc^3)) + (x[,"n_13"] + x[,"n_21"] + x[,"n_34"] + x[,"n_42"])*log(21 - 12*r_rc + 10*r_rc^2 + 2*r_rc^3 - 6*r_rc^4) + (x[,"n_12"] + x[,"n_24"] + x[,"n_31"] + x[,"n_43"])*log(15 + 10*r_rc - 20*r_rc^2 + 22*r_rc^3 - 6*r_rc^4) + (x[,"n_02"] + x[,"n_25"] + x[,"n_30"] + x[,"n_53"])*log(3 - 8*r_rc + 14*r_rc^2 - 10*r_rc^3 + 2*r_rc^4) + (x[,"n_22"] + x[,"n_23"] + x[,"n_32"] + x[,"n_33"])*log(27 + 2*r_rc - 4*r_rc^3 + 2*r_rc^4) + (x[,"n_03"] + x[,"n_20"] + x[,"n_35"] + x[,"n_52"])*log(1 + 2*r_rc - 4*r_rc^2 + 2*r_rc^3 + 2*r_rc^4) + (x[,"n_11"] + x[,"n_14"] + x[,"n_41"] + x[,"n_44"])*log(6 - 2*r_rc + 11*r_rc^2 - 18*r_rc^3 + 9*r_rc^4)
logL_rm <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-3*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 3*n05 - 3*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 3*n15 - 3*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 3*n25 - 3*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 3*n35 - 3*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 3*n45 - 3*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 3*n55)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n05 + n50)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n55)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n10 + n45)*(2*log(pmax(1e-6,r)) + log(5 - 5*r + 3*r^2)) + (n15 + n40)*(2*log(pmax(1e-6,1 - r)) + log(3 - r + 3*r^2)) + (n01 + n54)*(2*log(pmax(1e-6,r)) + log(11 - 14*r + 6*r^2)) + (n04 + n51)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 6*r^2)) + (n11 + n44)*(log(pmax(1e-6,r)) + log(42 - 29*r + 38*r^2 - 18*r^3)) + (n02 + n53)*(log(pmax(1e-6,r)) + log(18 - 29*r + 24*r^2 - 4*r^3)) + (n20 + n35)*(log(pmax(1e-6,r)) + log(7 - 8*r + 6*r^2 - 2*r^3)) + (n14 + n41)*log(33 - 26*r - 23*r^2 + 34*r^3 - 18*r^4) + (n22 + n33)*log(153 + 8*r + 40*r^2 - 16*r^3 - 8*r^4) + (n23 + n32)*log(177 - 8*r - 56*r^2 + 48*r^3 - 8*r^4) + (n03 + n52)*log(9 - 16*r + 19*r^2 - 8*r^3 - 4*r^4) + (n25 + n30)*log(3 - r - 2*r^2 + 2*r^3 - 2*r^4) + (n12 + n43)*log(36 + 24*r + 37*r^2 - 40*r^3 + 12*r^4) + (n24 + n31)*log(69 - 38*r + 24*r^2 - 28*r^3 + 12*r^4) + (n21 + n34)*log(39 + 26*r + 12*r^2 - 20*r^3 + 12*r^4) + (n13 + n42)*log(69 - 26*r - 11*r^2 - 8*r^3 + 12*r^4)
return(L)}
interlogL_rm <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_rm,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm <- mapply(interlogL_rm,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rm <- 4*(x[,"n_05"] + x[,"n_50"])*log10(2) + 4*(x[,"n_00"] + x[,"n_55"])*log10(2) - (x[,"n_03"] + x[,"n_52"])*(-log10(2) + 2*log10(3)) - (x[,"n_02"] + x[,"n_53"])*(-log10(2) + 2*log10(3)) - (x[,"n_04"] + x[,"n_51"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_54"])*(-3*log10(2) + log10(11)) - (x[,"n_15"] + x[,"n_40"])*(-4*log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_45"])*(-4*log10(2) + log10(13)) - (x[,"n_25"] + x[,"n_30"])*(-3*log10(2) + log10(17)) - (x[,"n_20"] + x[,"n_35"])*(-3*log10(2) + log10(17)) - (x[,"n_23"] + x[,"n_32"])*(-log10(2) + log10(7) + log10(47)) - (x[,"n_22"] + x[,"n_33"])*(-log10(2) + log10(7) + log10(47)) - (x[,"n_13"] + x[,"n_42"])*log10(53) - (x[,"n_12"] + x[,"n_43"])*log10(53) - (x[,"n_24"] + x[,"n_31"])*(-2*log10(2) + log10(3) + log10(71)) - (x[,"n_21"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(71)) - (x[,"n_14"] + x[,"n_41"])*(-3*log10(2) + log10(139)) - (x[,"n_11"] + x[,"n_44"])*(-3*log10(2) + log10(139)) + (x[,"n_05"] + x[,"n_50"])*(3*log10(pmax(1e-6,1 - r_rm)) + log10(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_55"])*(log10(pmax(1e-6,1 - r_rm)) + 3*log10(pmax(1e-6,r_rm))) + (x[,"n_10"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_rm)) + log10(5 - 5*r_rm + 3*r_rm^2)) + (x[,"n_15"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_rm)) + log10(3 - r_rm + 3*r_rm^2)) + (x[,"n_01"] + x[,"n_54"])*(2*log10(pmax(1e-6,r_rm)) + log10(11 - 14*r_rm + 6*r_rm^2)) + (x[,"n_04"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_rm)) + log10(3 + 2*r_rm + 6*r_rm^2)) + (x[,"n_11"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm)) + log10(42 - 29*r_rm + 38*r_rm^2 - 18*r_rm^3)) + (x[,"n_02"] + x[,"n_53"])*(log10(pmax(1e-6,r_rm)) + log10(18 - 29*r_rm + 24*r_rm^2 - 4*r_rm^3)) + (x[,"n_20"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm)) + log10(7 - 8*r_rm + 6*r_rm^2 - 2*r_rm^3)) + (x[,"n_14"] + x[,"n_41"])*log10(33 - 26*r_rm - 23*r_rm^2 + 34*r_rm^3 - 18*r_rm^4) + (x[,"n_22"] + x[,"n_33"])*log10(153 + 8*r_rm + 40*r_rm^2 - 16*r_rm^3 - 8*r_rm^4) + (x[,"n_23"] + x[,"n_32"])*log10(177 - 8*r_rm - 56*r_rm^2 + 48*r_rm^3 - 8*r_rm^4) + (x[,"n_03"] + x[,"n_52"])*log10(9 - 16*r_rm + 19*r_rm^2 - 8*r_rm^3 - 4*r_rm^4) + (x[,"n_25"] + x[,"n_30"])*log10(3 - r_rm - 2*r_rm^2 + 2*r_rm^3 - 2*r_rm^4) + (x[,"n_12"] + x[,"n_43"])*log10(36 + 24*r_rm + 37*r_rm^2 - 40*r_rm^3 + 12*r_rm^4) + (x[,"n_24"] + x[,"n_31"])*log10(69 - 38*r_rm + 24*r_rm^2 - 28*r_rm^3 + 12*r_rm^4) + (x[,"n_21"] + x[,"n_34"])*log10(39 + 26*r_rm + 12*r_rm^2 - 20*r_rm^3 + 12*r_rm^4) + (x[,"n_13"] + x[,"n_42"])*log10(69 - 26*r_rm - 11*r_rm^2 - 8*r_rm^3 + 12*r_rm^4)
logL_rm <- (-3*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_50"])*(3*log(pmax(1e-6,1 - r_rm)) + log(pmax(1e-6,r_rm))) + (x[,"n_00"] + x[,"n_55"])*(log(pmax(1e-6,1 - r_rm)) + 3*log(pmax(1e-6,r_rm))) + (x[,"n_10"] + x[,"n_45"])*(2*log(pmax(1e-6,r_rm)) + log(5 - 5*r_rm + 3*r_rm^2)) + (x[,"n_15"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_rm)) + log(3 - r_rm + 3*r_rm^2)) + (x[,"n_01"] + x[,"n_54"])*(2*log(pmax(1e-6,r_rm)) + log(11 - 14*r_rm + 6*r_rm^2)) + (x[,"n_04"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_rm)) + log(3 + 2*r_rm + 6*r_rm^2)) + (x[,"n_11"] + x[,"n_44"])*(log(pmax(1e-6,r_rm)) + log(42 - 29*r_rm + 38*r_rm^2 - 18*r_rm^3)) + (x[,"n_02"] + x[,"n_53"])*(log(pmax(1e-6,r_rm)) + log(18 - 29*r_rm + 24*r_rm^2 - 4*r_rm^3)) + (x[,"n_20"] + x[,"n_35"])*(log(pmax(1e-6,r_rm)) + log(7 - 8*r_rm + 6*r_rm^2 - 2*r_rm^3)) + (x[,"n_14"] + x[,"n_41"])*log(33 - 26*r_rm - 23*r_rm^2 + 34*r_rm^3 - 18*r_rm^4) + (x[,"n_22"] + x[,"n_33"])*log(153 + 8*r_rm + 40*r_rm^2 - 16*r_rm^3 - 8*r_rm^4) + (x[,"n_23"] + x[,"n_32"])*log(177 - 8*r_rm - 56*r_rm^2 + 48*r_rm^3 - 8*r_rm^4) + (x[,"n_03"] + x[,"n_52"])*log(9 - 16*r_rm + 19*r_rm^2 - 8*r_rm^3 - 4*r_rm^4) + (x[,"n_25"] + x[,"n_30"])*log(3 - r_rm - 2*r_rm^2 + 2*r_rm^3 - 2*r_rm^4) + (x[,"n_12"] + x[,"n_43"])*log(36 + 24*r_rm + 37*r_rm^2 - 40*r_rm^3 + 12*r_rm^4) + (x[,"n_24"] + x[,"n_31"])*log(69 - 38*r_rm + 24*r_rm^2 - 28*r_rm^3 + 12*r_rm^4) + (x[,"n_21"] + x[,"n_34"])*log(39 + 26*r_rm + 12*r_rm^2 - 20*r_rm^3 + 12*r_rm^4) + (x[,"n_13"] + x[,"n_42"])*log(69 - 26*r_rm - 11*r_rm^2 - 8*r_rm^3 + 12*r_rm^4)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
L <- (-4*n00 - 4*n01 - 3*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n10 - 4*n11 - 3*n12 - 3*n13 - 4*n14 - 4*n15 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 3*n25 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 4*n40 - 4*n41 - 3*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n50 - 4*n51 - 3*n52 - 3*n53 - 4*n54 - 4*n55)*log(2) + (n01 + n04 + n10 + n15 + n40 + n45 + n51 + n54)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35 - n40 - n41 - n42 - n43 - n44 - n45 - n50 - n51 - n52 - n53 - n54 - n55)*log(5) + 4*(n05 + n50)*log(pmax(1e-6,1 - r)) + 4*(n00 + n55)*log(pmax(1e-6,r)) + (n01 + n10 + n45 + n54)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n04 + n15 + n40 + n51)*(3*log(pmax(1e-6,1 - r)) + log(1 + r)) + (n03 + n25 + n30 + n52)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(5 + r)) + (n02 + n20 + n35 + n53)*(2*log(pmax(1e-6,r)) + log(6 - 7*r + r^2)) + (n11 + n44)*(2*log(pmax(1e-6,r)) + log(32 - 20*r + 9*r^2)) + (n14 + n41)*(2*log(pmax(1e-6,1 - r)) + log(21 + 2*r + 9*r^2)) + (n12 + n21 + n34 + n43)*(log(pmax(1e-6,r)) + log(25 - 19*r + 9*r^2 - 3*r^3)) + (n13 + n24 + n31 + n42)*log(12 - 2*r - 10*r^2 + 3*r^3 - 3*r^4) + (n23 + n32)*log(39 - 38*r + 40*r^2 - 20*r^3 + 2*r^4) + (n22 + n33)*log(23 + 10*r - 8*r^2 + 12*r^3 + 2*r^4)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35,n40,n41,n42,n43,n44,n45,n50,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rr <- 4*(x[,"n_05"] + x[,"n_50"])*log10(2) + 4*(x[,"n_00"] + x[,"n_55"])*log10(2) - (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(-4*log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(-4*log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(-4*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(11)) - (x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(13) + log10(17)) - (x[,"n_22"] + x[,"n_33"])*(-3*log10(2) + log10(13) + log10(17)) - (x[,"n_14"] + x[,"n_41"])*(-4*log10(2) + log10(97)) - (x[,"n_11"] + x[,"n_44"])*(-4*log10(2) + log10(97)) - (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*(-4*log10(2) + log10(139)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(139)) + 4*(x[,"n_05"] + x[,"n_50"])*log10(pmax(1e-6,1 - r_rr)) + 4*(x[,"n_00"] + x[,"n_55"])*log10(pmax(1e-6,r_rr)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(log10(2 - r_rr) + 3*log10(pmax(1e-6,r_rr))) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(3*log10(pmax(1e-6,1 - r_rr)) + log10(1 + r_rr)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr)) + log10(5 + r_rr)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log10(pmax(1e-6,r_rr)) + log10(6 - 7*r_rr + r_rr^2)) + (x[,"n_11"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(32 - 20*r_rr + 9*r_rr^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(21 + 2*r_rr + 9*r_rr^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(25 - 19*r_rr + 9*r_rr^2 - 3*r_rr^3)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log10(12 - 2*r_rr - 10*r_rr^2 + 3*r_rr^3 - 3*r_rr^4) + (x[,"n_23"] + x[,"n_32"])*log10(39 - 38*r_rr + 40*r_rr^2 - 20*r_rr^3 + 2*r_rr^4) + (x[,"n_22"] + x[,"n_33"])*log10(23 + 10*r_rr - 8*r_rr^2 + 12*r_rr^3 + 2*r_rr^4)
logL_rr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"])*log(2) + (x[,"n_01"] + x[,"n_04"] + x[,"n_10"] + x[,"n_15"] + x[,"n_40"] + x[,"n_45"] + x[,"n_51"] + x[,"n_54"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + 4*(x[,"n_05"] + x[,"n_50"])*log(pmax(1e-6,1 - r_rr)) + 4*(x[,"n_00"] + x[,"n_55"])*log(pmax(1e-6,r_rr)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_45"] + x[,"n_54"])*(log(2 - r_rr) + 3*log(pmax(1e-6,r_rr))) + (x[,"n_04"] + x[,"n_15"] + x[,"n_40"] + x[,"n_51"])*(3*log(pmax(1e-6,1 - r_rr)) + log(1 + r_rr)) + (x[,"n_03"] + x[,"n_25"] + x[,"n_30"] + x[,"n_52"])*(2*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr)) + log(5 + r_rr)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_35"] + x[,"n_53"])*(2*log(pmax(1e-6,r_rr)) + log(6 - 7*r_rr + r_rr^2)) + (x[,"n_11"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(32 - 20*r_rr + 9*r_rr^2)) + (x[,"n_14"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rr)) + log(21 + 2*r_rr + 9*r_rr^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_34"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(25 - 19*r_rr + 9*r_rr^2 - 3*r_rr^3)) + (x[,"n_13"] + x[,"n_24"] + x[,"n_31"] + x[,"n_42"])*log(12 - 2*r_rr - 10*r_rr^2 + 3*r_rr^3 - 3*r_rr^4) + (x[,"n_23"] + x[,"n_32"])*log(39 - 38*r_rr + 40*r_rr^2 - 20*r_rr^3 + 2*r_rr^4) + (x[,"n_22"] + x[,"n_33"])*log(23 + 10*r_rr - 8*r_rr^2 + 12*r_rr^3 + 2*r_rr^4)

return(list(
r_mat = cbind(r_cc,r_cm,r_cr,r_mc,r_mm,r_mr,r_rc,r_rm,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm,LOD_cr,LOD_mc,LOD_mm,LOD_mr,LOD_rc,LOD_rm,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm,logL_cr,logL_mc,logL_mm,logL_mr,logL_rc,logL_rm,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed","coupling repulsion","mixed coupling","mixed mixed","mixed repulsion","repulsion coupling","repulsion mixed","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.3_3.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-4*n00 - 4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 4*n11 - 4*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n16 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 3*n25 - 3*n26 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n36 - 4*n40 - 4*n41 - 4*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n46 - 4*n50 - 4*n51 - 4*n52 - 3*n53 - 4*n54 - 4*n55 - 4*n56)*log(2) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - n16 - n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - n26 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - n36 - n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - n46 - 2*n50 - 2*n51 - 2*n52 - 2*n53 - 2*n54 - 2*n55 - 2*n56)*log(5) + (n03 + n53)*log(11) + 5*(n00 + n56)*log(pmax(1e-6,1 - r)) + 5*(n06 + n50)*log(pmax(1e-6,r)) + (n10 + n46)*(4*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n53)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n20 + n36)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n26 + n30)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n05 + n51)*(log(7 - 4*r) + 4*log(pmax(1e-6,r))) + (n16 + n40)*(log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n01 + n55)*(4*log(pmax(1e-6,1 - r)) + log(3 + 4*r)) + (n02 + n54)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(13 + 5*r)) + (n04 + n52)*(3*log(pmax(1e-6,r)) + log(18 - 23*r + 5*r^2)) + (n15 + n41)*(3*log(pmax(1e-6,r)) + log(36 - 41*r + 20*r^2)) + (n21 + n35)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(26 - 13*r + 20*r^2)) + (n11 + n45)*(3*log(pmax(1e-6,1 - r)) + log(15 + r + 20*r^2)) + (n13 + n43)*(log(pmax(1e-6,r)) + log(46 - 79*r + 66*r^2 - 33*r^3)) + (n14 + n42)*(2*log(pmax(1e-6,r)) + log(90 - 126*r + 94*r^2 - 25*r^3)) + (n25 + n31)*(2*log(pmax(1e-6,r)) + log(33 - 60*r + 47*r^2 - 20*r^3)) + (n12 + n44)*(2*log(pmax(1e-6,1 - r)) + log(33 + 13*r + 19*r^2 + 25*r^3)) + (n23 + n33)*log(41 - 46*r + 68*r^2 - 44*r^3 + 22*r^4) + (n24 + n32)*(log(pmax(1e-6,r)) + log(69 - 111*r + 123*r^2 - 73*r^3 + 25*r^4)) + (n22 + n34)*log(33 - 49*r + 70*r^2 - 81*r^3 + 52*r^4 - 25*r^5)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_cc <- 5*(x[,"n_26"] + x[,"n_30"])*log10(2) + 5*(x[,"n_20"] + x[,"n_36"])*log10(2) + 5*(x[,"n_16"] + x[,"n_40"])*log10(2) + 5*(x[,"n_10"] + x[,"n_46"])*log10(2) + 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 4*(x[,"n_03"] + x[,"n_53"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(5)) - (x[,"n_25"] + x[,"n_31"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_21"] + x[,"n_35"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_23"] + x[,"n_33"])*(-3*log10(2) + log10(13) + log10(19)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(31)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(31)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(41)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(41)) - (x[,"n_13"] + x[,"n_43"])*(-4*log10(2) + log10(151)) - (x[,"n_14"] + x[,"n_42"])*(-5*log10(2) + log10(379)) - (x[,"n_12"] + x[,"n_44"])*(-5*log10(2) + log10(379)) - (x[,"n_24"] + x[,"n_32"])*(-5*log10(2) + log10(587)) - (x[,"n_22"] + x[,"n_34"])*(-5*log10(2) + log10(587)) + 5*(x[,"n_00"] + x[,"n_56"])*log10(pmax(1e-6,1 - r_cc)) + 5*(x[,"n_06"] + x[,"n_50"])*log10(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_46"])*(4*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_53"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_36"])*(3*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_26"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_51"])*(log10(7 - 4*r_cc) + 4*log10(pmax(1e-6,r_cc))) + (x[,"n_16"] + x[,"n_40"])*(log10(pmax(1e-6,1 - r_cc)) + 4*log10(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_55"])*(4*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 4*r_cc)) + (x[,"n_02"] + x[,"n_54"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc)) + log10(13 + 5*r_cc)) + (x[,"n_04"] + x[,"n_52"])*(3*log10(pmax(1e-6,r_cc)) + log10(18 - 23*r_cc + 5*r_cc^2)) + (x[,"n_15"] + x[,"n_41"])*(3*log10(pmax(1e-6,r_cc)) + log10(36 - 41*r_cc + 20*r_cc^2)) + (x[,"n_21"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc)) + log10(26 - 13*r_cc + 20*r_cc^2)) + (x[,"n_11"] + x[,"n_45"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(15 + r_cc + 20*r_cc^2)) + (x[,"n_13"] + x[,"n_43"])*(log10(pmax(1e-6,r_cc)) + log10(46 - 79*r_cc + 66*r_cc^2 - 33*r_cc^3)) + (x[,"n_14"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(90 - 126*r_cc + 94*r_cc^2 - 25*r_cc^3)) + (x[,"n_25"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_cc)) + log10(33 - 60*r_cc + 47*r_cc^2 - 20*r_cc^3)) + (x[,"n_12"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(33 + 13*r_cc + 19*r_cc^2 + 25*r_cc^3)) + (x[,"n_23"] + x[,"n_33"])*log10(41 - 46*r_cc + 68*r_cc^2 - 44*r_cc^3 + 22*r_cc^4) + (x[,"n_24"] + x[,"n_32"])*(log10(pmax(1e-6,r_cc)) + log10(69 - 111*r_cc + 123*r_cc^2 - 73*r_cc^3 + 25*r_cc^4)) + (x[,"n_22"] + x[,"n_34"])*log10(33 - 49*r_cc + 70*r_cc^2 - 81*r_cc^3 + 52*r_cc^4 - 25*r_cc^5)
logL_cc <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 3*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"] - 4*x[,"n_56"])*log(2) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - x[,"n_46"] - 2*x[,"n_50"] - 2*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"] - 2*x[,"n_56"])*log(5) + (x[,"n_03"] + x[,"n_53"])*log(11) + 5*(x[,"n_00"] + x[,"n_56"])*log(pmax(1e-6,1 - r_cc)) + 5*(x[,"n_06"] + x[,"n_50"])*log(pmax(1e-6,r_cc)) + (x[,"n_10"] + x[,"n_46"])*(4*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_53"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_20"] + x[,"n_36"])*(3*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_26"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_51"])*(log(7 - 4*r_cc) + 4*log(pmax(1e-6,r_cc))) + (x[,"n_16"] + x[,"n_40"])*(log(pmax(1e-6,1 - r_cc)) + 4*log(pmax(1e-6,r_cc))) + (x[,"n_01"] + x[,"n_55"])*(4*log(pmax(1e-6,1 - r_cc)) + log(3 + 4*r_cc)) + (x[,"n_02"] + x[,"n_54"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc)) + log(13 + 5*r_cc)) + (x[,"n_04"] + x[,"n_52"])*(3*log(pmax(1e-6,r_cc)) + log(18 - 23*r_cc + 5*r_cc^2)) + (x[,"n_15"] + x[,"n_41"])*(3*log(pmax(1e-6,r_cc)) + log(36 - 41*r_cc + 20*r_cc^2)) + (x[,"n_21"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc)) + log(26 - 13*r_cc + 20*r_cc^2)) + (x[,"n_11"] + x[,"n_45"])*(3*log(pmax(1e-6,1 - r_cc)) + log(15 + r_cc + 20*r_cc^2)) + (x[,"n_13"] + x[,"n_43"])*(log(pmax(1e-6,r_cc)) + log(46 - 79*r_cc + 66*r_cc^2 - 33*r_cc^3)) + (x[,"n_14"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(90 - 126*r_cc + 94*r_cc^2 - 25*r_cc^3)) + (x[,"n_25"] + x[,"n_31"])*(2*log(pmax(1e-6,r_cc)) + log(33 - 60*r_cc + 47*r_cc^2 - 20*r_cc^3)) + (x[,"n_12"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(33 + 13*r_cc + 19*r_cc^2 + 25*r_cc^3)) + (x[,"n_23"] + x[,"n_33"])*log(41 - 46*r_cc + 68*r_cc^2 - 44*r_cc^3 + 22*r_cc^4) + (x[,"n_24"] + x[,"n_32"])*(log(pmax(1e-6,r_cc)) + log(69 - 111*r_cc + 123*r_cc^2 - 73*r_cc^3 + 25*r_cc^4)) + (x[,"n_22"] + x[,"n_34"])*log(33 - 49*r_cc + 70*r_cc^2 - 81*r_cc^3 + 52*r_cc^4 - 25*r_cc^5)
logL_cm2c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-4*n00 - 4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 4*n11 - 4*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n16 - 3*n20 - 3*n21 - 3*n22 - 2*n23 - 3*n24 - 3*n25 - 3*n26 - 3*n30 - 3*n31 - 3*n32 - 2*n33 - 3*n34 - 3*n35 - 3*n36 - 4*n40 - 4*n41 - 4*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n46 - 4*n50 - 4*n51 - 4*n52 - 3*n53 - 4*n54 - 4*n55 - 4*n56)*log(2) + (-n00 - n01 - n02 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n54 - n55 - n56)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*log(5) + (n00 + n56)*(4*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n06 + n50)*(log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n05 + n51)*(3*log(pmax(1e-6,r)) + log(8 - 9*r + 4*r^2)) + (n01 + n55)*(3*log(pmax(1e-6,1 - r)) + log(3 + r + 4*r^2)) + (n16 + n40)*(3*log(pmax(1e-6,r)) + log(6 - 8*r + 5*r^2)) + (n20 + n36)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(5 - 4*r + 5*r^2)) + (n10 + n46)*(3*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 5*r^2)) + (n15 + n41)*(2*log(pmax(1e-6,r)) + log(42 - 48*r + 47*r^2 - 20*r^3)) + (n26 + n30)*(2*log(pmax(1e-6,r)) + log(6 - 12*r + 11*r^2 - 5*r^3)) + (n04 + n52)*(2*log(pmax(1e-6,r)) + log(22 - 30*r + 22*r^2 - 5*r^3)) + (n03 + n53)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n02 + n54)*(2*log(pmax(1e-6,1 - r)) + log(9 + r + 7*r^2 + 5*r^3)) + (n11 + n45)*(2*log(pmax(1e-6,1 - r)) + log(21 + 14*r - 13*r^2 + 20*r^3)) + (n23 + n33)*log(48 - 7*r - 2*r^2 + 18*r^3 - 9*r^4) + (n25 + n31)*(log(pmax(1e-6,r)) + log(32 - 42*r + 54*r^2 - 49*r^3 + 20*r^4)) + (n14 + n42)*(log(pmax(1e-6,r)) + log(90 - 78*r + 120*r^2 - 91*r^3 + 25*r^4)) + (n13 + n43)*log(27 + 2*r + 25*r^2 - 54*r^3 + 27*r^4) + (n12 + n44)*log(66 - 55*r - 14*r^2 - 6*r^3 + 34*r^4 - 25*r^5) + (n24 + n32)*log(42 + 14*r + 52*r^2 - 86*r^3 + 72*r^4 - 25*r^5) + (n21 + n35)*log(15 - 14*r + 26*r^2 - 58*r^3 + 51*r^4 - 20*r^5) + (n22 + n34)*log(69 - 23*r - 24*r^2 + 48*r^3 - 53*r^4 + 25*r^5)
return(L)}
interlogL_cm2c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_cm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_cm2c <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_16"] + x[,"n_40"])*(-5*log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_46"])*(-5*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_53"])*(-4*log10(2) + log10(13)) - (x[,"n_26"] + x[,"n_30"])*(-5*log10(2) + log10(17)) - (x[,"n_20"] + x[,"n_36"])*(-5*log10(2) + log10(17)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(5) + log10(19)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(5) + log10(19)) - (x[,"n_23"] + x[,"n_33"])*(-4*log10(2) + log10(17) + log10(43)) - (x[,"n_14"] + x[,"n_42"])*(-5*log10(2) + log10(17) + log10(67)) - (x[,"n_12"] + x[,"n_44"])*(-5*log10(2) + log10(17) + log10(67)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(109)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(109)) - (x[,"n_25"] + x[,"n_31"])*(-4*log10(2) + log10(157)) - (x[,"n_21"] + x[,"n_35"])*(-4*log10(2) + log10(157)) - (x[,"n_13"] + x[,"n_43"])*(-4*log10(2) + log10(467)) - (x[,"n_24"] + x[,"n_32"])*(-5*log10(2) + log10(1759)) - (x[,"n_22"] + x[,"n_34"])*(-5*log10(2) + log10(1759)) + (x[,"n_00"] + x[,"n_56"])*(4*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_06"] + x[,"n_50"])*(log10(pmax(1e-6,1 - r_cm2c)) + 4*log10(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_51"])*(3*log10(pmax(1e-6,r_cm2c)) + log10(8 - 9*r_cm2c + 4*r_cm2c^2)) + (x[,"n_01"] + x[,"n_55"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 + r_cm2c + 4*r_cm2c^2)) + (x[,"n_16"] + x[,"n_40"])*(3*log10(pmax(1e-6,r_cm2c)) + log10(6 - 8*r_cm2c + 5*r_cm2c^2)) + (x[,"n_20"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c)) + log10(5 - 4*r_cm2c + 5*r_cm2c^2)) + (x[,"n_10"] + x[,"n_46"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 - 2*r_cm2c + 5*r_cm2c^2)) + (x[,"n_15"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(42 - 48*r_cm2c + 47*r_cm2c^2 - 20*r_cm2c^3)) + (x[,"n_26"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(6 - 12*r_cm2c + 11*r_cm2c^2 - 5*r_cm2c^3)) + (x[,"n_04"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(22 - 30*r_cm2c + 22*r_cm2c^2 - 5*r_cm2c^3)) + (x[,"n_03"] + x[,"n_53"])*(log10(pmax(1e-6,r_cm2c)) + log10(4 - 7*r_cm2c + 6*r_cm2c^2 - 3*r_cm2c^3)) + (x[,"n_02"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(9 + r_cm2c + 7*r_cm2c^2 + 5*r_cm2c^3)) + (x[,"n_11"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(21 + 14*r_cm2c - 13*r_cm2c^2 + 20*r_cm2c^3)) + (x[,"n_23"] + x[,"n_33"])*log10(48 - 7*r_cm2c - 2*r_cm2c^2 + 18*r_cm2c^3 - 9*r_cm2c^4) + (x[,"n_25"] + x[,"n_31"])*(log10(pmax(1e-6,r_cm2c)) + log10(32 - 42*r_cm2c + 54*r_cm2c^2 - 49*r_cm2c^3 + 20*r_cm2c^4)) + (x[,"n_14"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm2c)) + log10(90 - 78*r_cm2c + 120*r_cm2c^2 - 91*r_cm2c^3 + 25*r_cm2c^4)) + (x[,"n_13"] + x[,"n_43"])*log10(27 + 2*r_cm2c + 25*r_cm2c^2 - 54*r_cm2c^3 + 27*r_cm2c^4) + (x[,"n_12"] + x[,"n_44"])*log10(66 - 55*r_cm2c - 14*r_cm2c^2 - 6*r_cm2c^3 + 34*r_cm2c^4 - 25*r_cm2c^5) + (x[,"n_24"] + x[,"n_32"])*log10(42 + 14*r_cm2c + 52*r_cm2c^2 - 86*r_cm2c^3 + 72*r_cm2c^4 - 25*r_cm2c^5) + (x[,"n_21"] + x[,"n_35"])*log10(15 - 14*r_cm2c + 26*r_cm2c^2 - 58*r_cm2c^3 + 51*r_cm2c^4 - 20*r_cm2c^5) + (x[,"n_22"] + x[,"n_34"])*log10(69 - 23*r_cm2c - 24*r_cm2c^2 + 48*r_cm2c^3 - 53*r_cm2c^4 + 25*r_cm2c^5)
logL_cm2c <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 3*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"] - 4*x[,"n_56"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*log(5) + (x[,"n_00"] + x[,"n_56"])*(4*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_06"] + x[,"n_50"])*(log(pmax(1e-6,1 - r_cm2c)) + 4*log(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_51"])*(3*log(pmax(1e-6,r_cm2c)) + log(8 - 9*r_cm2c + 4*r_cm2c^2)) + (x[,"n_01"] + x[,"n_55"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(3 + r_cm2c + 4*r_cm2c^2)) + (x[,"n_16"] + x[,"n_40"])*(3*log(pmax(1e-6,r_cm2c)) + log(6 - 8*r_cm2c + 5*r_cm2c^2)) + (x[,"n_20"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c)) + log(5 - 4*r_cm2c + 5*r_cm2c^2)) + (x[,"n_10"] + x[,"n_46"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(3 - 2*r_cm2c + 5*r_cm2c^2)) + (x[,"n_15"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cm2c)) + log(42 - 48*r_cm2c + 47*r_cm2c^2 - 20*r_cm2c^3)) + (x[,"n_26"] + x[,"n_30"])*(2*log(pmax(1e-6,r_cm2c)) + log(6 - 12*r_cm2c + 11*r_cm2c^2 - 5*r_cm2c^3)) + (x[,"n_04"] + x[,"n_52"])*(2*log(pmax(1e-6,r_cm2c)) + log(22 - 30*r_cm2c + 22*r_cm2c^2 - 5*r_cm2c^3)) + (x[,"n_03"] + x[,"n_53"])*(log(pmax(1e-6,r_cm2c)) + log(4 - 7*r_cm2c + 6*r_cm2c^2 - 3*r_cm2c^3)) + (x[,"n_02"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(9 + r_cm2c + 7*r_cm2c^2 + 5*r_cm2c^3)) + (x[,"n_11"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(21 + 14*r_cm2c - 13*r_cm2c^2 + 20*r_cm2c^3)) + (x[,"n_23"] + x[,"n_33"])*log(48 - 7*r_cm2c - 2*r_cm2c^2 + 18*r_cm2c^3 - 9*r_cm2c^4) + (x[,"n_25"] + x[,"n_31"])*(log(pmax(1e-6,r_cm2c)) + log(32 - 42*r_cm2c + 54*r_cm2c^2 - 49*r_cm2c^3 + 20*r_cm2c^4)) + (x[,"n_14"] + x[,"n_42"])*(log(pmax(1e-6,r_cm2c)) + log(90 - 78*r_cm2c + 120*r_cm2c^2 - 91*r_cm2c^3 + 25*r_cm2c^4)) + (x[,"n_13"] + x[,"n_43"])*log(27 + 2*r_cm2c + 25*r_cm2c^2 - 54*r_cm2c^3 + 27*r_cm2c^4) + (x[,"n_12"] + x[,"n_44"])*log(66 - 55*r_cm2c - 14*r_cm2c^2 - 6*r_cm2c^3 + 34*r_cm2c^4 - 25*r_cm2c^5) + (x[,"n_24"] + x[,"n_32"])*log(42 + 14*r_cm2c + 52*r_cm2c^2 - 86*r_cm2c^3 + 72*r_cm2c^4 - 25*r_cm2c^5) + (x[,"n_21"] + x[,"n_35"])*log(15 - 14*r_cm2c + 26*r_cm2c^2 - 58*r_cm2c^3 + 51*r_cm2c^4 - 20*r_cm2c^5) + (x[,"n_22"] + x[,"n_34"])*log(69 - 23*r_cm2c - 24*r_cm2c^2 + 48*r_cm2c^3 - 53*r_cm2c^4 + 25*r_cm2c^5)
logL_cm1c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- 2*(-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 2*n20 - 2*n21 - 2*n22 - n23 - 2*n24 - 2*n25 - 2*n26 - 2*n30 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - 2*n36 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n46 - 2*n50 - 2*n51 - 2*n52 - 2*n53 - 2*n54 - 2*n55 - 2*n56)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*(log(3) + 2*log(5)) + (n00 + n56)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n06 + n50)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n55)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(4 - r + 4*r^2)) + (n10 + n46)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(4 - 4*r + 5*r^2)) + (n16 + n40)*(2*log(pmax(1e-6,r)) + log(5 - 11*r + 11*r^2 - 5*r^3)) + (n05 + n51)*(2*log(pmax(1e-6,r)) + log(7 - 14*r + 11*r^2 - 4*r^3)) + (n13 + n43)*log(57 + 16*r - 58*r^2 + 84*r^3 - 42*r^4) + (n04 + n52)*(log(pmax(1e-6,r)) + log(15 - 31*r + 35*r^2 - 21*r^3 + 5*r^4)) + (n23 + n33)*log(45 + r + 6*r^2 - 14*r^3 + 7*r^4) + (n26 + n30)*(log(pmax(1e-6,r)) + log(7 - 20*r + 30*r^2 - 24*r^3 + 10*r^4)) + (n03 + n53)*log(9 - 20*r + 34*r^2 - 28*r^3 + 14*r^4) + (n15 + n41)*(log(pmax(1e-6,r)) + log(24 - 41*r + 62*r^2 - 53*r^3 + 20*r^4)) + (n22 + n34)*log(120 - 25*r + 32*r^2 - 86*r^3 + 108*r^4 - 50*r^5) + (n25 + n31)*log(21 - 20*r + 70*r^2 - 112*r^3 + 102*r^4 - 40*r^5) + (n14 + n42)*log(27 - 3*r + 82*r^2 - 121*r^3 + 88*r^4 - 25*r^5) + (n11 + n45)*log(12 - 16*r + 27*r^2 - 50*r^3 + 47*r^4 - 20*r^5) + (n20 + n36)*log(3 - 11*r + 26*r^2 - 34*r^3 + 26*r^4 - 10*r^5) + (n02 + n54)*log(3 + r - 2*r^2 - r^3 + 4*r^4 - 5*r^5) + (n12 + n44)*log(48 - 25*r - 3*r^2 + 19*r^3 - 37*r^4 + 25*r^5) + (n21 + n35)*log(21 + 8*r - 54*r^2 + 104*r^3 - 98*r^4 + 40*r^5) + (n24 + n32)*log(99 + 37*r - 78*r^2 + 154*r^3 - 142*r^4 + 50*r^5)
return(L)}
interlogL_cm1c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_cm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_cm1c <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_16"] + x[,"n_40"])*(-5*log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_46"])*(-5*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_53"])*(-3*log10(2) + log10(3) + log10(13)) - (x[,"n_26"] + x[,"n_30"])*(-4*log10(2) + log10(17)) - (x[,"n_20"] + x[,"n_36"])*(-4*log10(2) + log10(17)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(5) + log10(19)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(5) + log10(19)) - (x[,"n_23"] + x[,"n_33"])*(-4*log10(2) + log10(17) + log10(43)) - (x[,"n_14"] + x[,"n_42"])*(-5*log10(2) + log10(17) + log10(67)) - (x[,"n_12"] + x[,"n_44"])*(-5*log10(2) + log10(17) + log10(67)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(109)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(109)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(157)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(157)) - (x[,"n_13"] + x[,"n_43"])*(-3*log10(2) + log10(467)) - (x[,"n_24"] + x[,"n_32"])*(-4*log10(2) + log10(1759)) - (x[,"n_22"] + x[,"n_34"])*(-4*log10(2) + log10(1759)) + (x[,"n_00"] + x[,"n_56"])*(3*log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_06"] + x[,"n_50"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + 3*log10(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + log10(pmax(1e-6,r_cm1c)) + log10(4 - r_cm1c + 4*r_cm1c^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + log10(pmax(1e-6,r_cm1c)) + log10(4 - 4*r_cm1c + 5*r_cm1c^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_cm1c)) + log10(5 - 11*r_cm1c + 11*r_cm1c^2 - 5*r_cm1c^3)) + (x[,"n_05"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_cm1c)) + log10(7 - 14*r_cm1c + 11*r_cm1c^2 - 4*r_cm1c^3)) + (x[,"n_13"] + x[,"n_43"])*log10(57 + 16*r_cm1c - 58*r_cm1c^2 + 84*r_cm1c^3 - 42*r_cm1c^4) + (x[,"n_04"] + x[,"n_52"])*(log10(pmax(1e-6,r_cm1c)) + log10(15 - 31*r_cm1c + 35*r_cm1c^2 - 21*r_cm1c^3 + 5*r_cm1c^4)) + (x[,"n_23"] + x[,"n_33"])*log10(45 + r_cm1c + 6*r_cm1c^2 - 14*r_cm1c^3 + 7*r_cm1c^4) + (x[,"n_26"] + x[,"n_30"])*(log10(pmax(1e-6,r_cm1c)) + log10(7 - 20*r_cm1c + 30*r_cm1c^2 - 24*r_cm1c^3 + 10*r_cm1c^4)) + (x[,"n_03"] + x[,"n_53"])*log10(9 - 20*r_cm1c + 34*r_cm1c^2 - 28*r_cm1c^3 + 14*r_cm1c^4) + (x[,"n_15"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm1c)) + log10(24 - 41*r_cm1c + 62*r_cm1c^2 - 53*r_cm1c^3 + 20*r_cm1c^4)) + (x[,"n_22"] + x[,"n_34"])*log10(120 - 25*r_cm1c + 32*r_cm1c^2 - 86*r_cm1c^3 + 108*r_cm1c^4 - 50*r_cm1c^5) + (x[,"n_25"] + x[,"n_31"])*log10(21 - 20*r_cm1c + 70*r_cm1c^2 - 112*r_cm1c^3 + 102*r_cm1c^4 - 40*r_cm1c^5) + (x[,"n_14"] + x[,"n_42"])*log10(27 - 3*r_cm1c + 82*r_cm1c^2 - 121*r_cm1c^3 + 88*r_cm1c^4 - 25*r_cm1c^5) + (x[,"n_11"] + x[,"n_45"])*log10(12 - 16*r_cm1c + 27*r_cm1c^2 - 50*r_cm1c^3 + 47*r_cm1c^4 - 20*r_cm1c^5) + (x[,"n_20"] + x[,"n_36"])*log10(3 - 11*r_cm1c + 26*r_cm1c^2 - 34*r_cm1c^3 + 26*r_cm1c^4 - 10*r_cm1c^5) + (x[,"n_02"] + x[,"n_54"])*log10(3 + r_cm1c - 2*r_cm1c^2 - r_cm1c^3 + 4*r_cm1c^4 - 5*r_cm1c^5) + (x[,"n_12"] + x[,"n_44"])*log10(48 - 25*r_cm1c - 3*r_cm1c^2 + 19*r_cm1c^3 - 37*r_cm1c^4 + 25*r_cm1c^5) + (x[,"n_21"] + x[,"n_35"])*log10(21 + 8*r_cm1c - 54*r_cm1c^2 + 104*r_cm1c^3 - 98*r_cm1c^4 + 40*r_cm1c^5) + (x[,"n_24"] + x[,"n_32"])*log10(99 + 37*r_cm1c - 78*r_cm1c^2 + 154*r_cm1c^3 - 142*r_cm1c^4 + 50*r_cm1c^5)
logL_cm1c <- 2*(-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_46"] - 2*x[,"n_50"] - 2*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"] - 2*x[,"n_56"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_56"])*(3*log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_06"] + x[,"n_50"])*(2*log(pmax(1e-6,1 - r_cm1c)) + 3*log(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_cm1c)) + log(pmax(1e-6,r_cm1c)) + log(4 - r_cm1c + 4*r_cm1c^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log(pmax(1e-6,1 - r_cm1c)) + log(pmax(1e-6,r_cm1c)) + log(4 - 4*r_cm1c + 5*r_cm1c^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log(pmax(1e-6,r_cm1c)) + log(5 - 11*r_cm1c + 11*r_cm1c^2 - 5*r_cm1c^3)) + (x[,"n_05"] + x[,"n_51"])*(2*log(pmax(1e-6,r_cm1c)) + log(7 - 14*r_cm1c + 11*r_cm1c^2 - 4*r_cm1c^3)) + (x[,"n_13"] + x[,"n_43"])*log(57 + 16*r_cm1c - 58*r_cm1c^2 + 84*r_cm1c^3 - 42*r_cm1c^4) + (x[,"n_04"] + x[,"n_52"])*(log(pmax(1e-6,r_cm1c)) + log(15 - 31*r_cm1c + 35*r_cm1c^2 - 21*r_cm1c^3 + 5*r_cm1c^4)) + (x[,"n_23"] + x[,"n_33"])*log(45 + r_cm1c + 6*r_cm1c^2 - 14*r_cm1c^3 + 7*r_cm1c^4) + (x[,"n_26"] + x[,"n_30"])*(log(pmax(1e-6,r_cm1c)) + log(7 - 20*r_cm1c + 30*r_cm1c^2 - 24*r_cm1c^3 + 10*r_cm1c^4)) + (x[,"n_03"] + x[,"n_53"])*log(9 - 20*r_cm1c + 34*r_cm1c^2 - 28*r_cm1c^3 + 14*r_cm1c^4) + (x[,"n_15"] + x[,"n_41"])*(log(pmax(1e-6,r_cm1c)) + log(24 - 41*r_cm1c + 62*r_cm1c^2 - 53*r_cm1c^3 + 20*r_cm1c^4)) + (x[,"n_22"] + x[,"n_34"])*log(120 - 25*r_cm1c + 32*r_cm1c^2 - 86*r_cm1c^3 + 108*r_cm1c^4 - 50*r_cm1c^5) + (x[,"n_25"] + x[,"n_31"])*log(21 - 20*r_cm1c + 70*r_cm1c^2 - 112*r_cm1c^3 + 102*r_cm1c^4 - 40*r_cm1c^5) + (x[,"n_14"] + x[,"n_42"])*log(27 - 3*r_cm1c + 82*r_cm1c^2 - 121*r_cm1c^3 + 88*r_cm1c^4 - 25*r_cm1c^5) + (x[,"n_11"] + x[,"n_45"])*log(12 - 16*r_cm1c + 27*r_cm1c^2 - 50*r_cm1c^3 + 47*r_cm1c^4 - 20*r_cm1c^5) + (x[,"n_20"] + x[,"n_36"])*log(3 - 11*r_cm1c + 26*r_cm1c^2 - 34*r_cm1c^3 + 26*r_cm1c^4 - 10*r_cm1c^5) + (x[,"n_02"] + x[,"n_54"])*log(3 + r_cm1c - 2*r_cm1c^2 - r_cm1c^3 + 4*r_cm1c^4 - 5*r_cm1c^5) + (x[,"n_12"] + x[,"n_44"])*log(48 - 25*r_cm1c - 3*r_cm1c^2 + 19*r_cm1c^3 - 37*r_cm1c^4 + 25*r_cm1c^5) + (x[,"n_21"] + x[,"n_35"])*log(21 + 8*r_cm1c - 54*r_cm1c^2 + 104*r_cm1c^3 - 98*r_cm1c^4 + 40*r_cm1c^5) + (x[,"n_24"] + x[,"n_32"])*log(99 + 37*r_cm1c - 78*r_cm1c^2 + 154*r_cm1c^3 - 142*r_cm1c^4 + 50*r_cm1c^5)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 4*n31 - 4*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46 - 4*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 4*n55 - 4*n56)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*log(5) + (n06 + n50)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n00 + n56)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n05 + n51)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(4 - 5*r + 4*r^2)) + (n16 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 - 4*r + 5*r^2)) + (n10 + n46)*(2*log(pmax(1e-6,r)) + log(3 - 9*r + 11*r^2 - 5*r^3)) + (n01 + n55)*(2*log(pmax(1e-6,r)) + log(3 - 6*r + 7*r^2 - 4*r^3)) + (n23 + n33)*log(27 + 18*r - 28*r^2 + 20*r^3 - 10*r^4) + (n03 + n53)*log(1 + 4*r - 14*r^2 + 20*r^3 - 10*r^4) + (n02 + n54)*(log(pmax(1e-6,r)) + log(3 - 3*r + 3*r^2 - 5*r^3 + 5*r^4)) + (n20 + n36)*(log(pmax(1e-6,r)) + log(3 - 12*r + 24*r^2 - 24*r^3 + 10*r^4)) + (n11 + n45)*(log(pmax(1e-6,r)) + log(12 - 27*r + 42*r^2 - 41*r^3 + 20*r^4)) + (n13 + n43)*log(27 - 40*r + 70*r^2 - 60*r^3 + 30*r^4) + (n24 + n32)*log(54 - 77*r + 152*r^2 - 192*r^3 + 140*r^4 - 50*r^5) + (n21 + n35)*log(9 - 24*r + 72*r^2 - 108*r^3 + 94*r^4 - 40*r^5) + (n12 + n44)*log(9 + 3*r + 18*r^2 - 39*r^3 + 40*r^4 - 25*r^5) + (n15 + n41)*log(6 - 20*r + 53*r^2 - 78*r^3 + 59*r^4 - 20*r^5) + (n26 + n30)*log(1 - 5*r + 16*r^2 - 28*r^3 + 26*r^4 - 10*r^5) + (n04 + n52)*log(3 - 11*r + 26*r^2 - 33*r^3 + 20*r^4 - 5*r^5) + (n14 + n42)*log(6 + 43*r - 109*r^2 + 129*r^3 - 85*r^4 + 25*r^5) + (n25 + n31)*log(3 + 28*r - 88*r^2 + 132*r^3 - 106*r^4 + 40*r^5) + (n22 + n34)*log(27 + 39*r - 84*r^2 + 132*r^3 - 110*r^4 + 50*r^5)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_cr <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) + (x[,"n_26"] + x[,"n_30"])*(4*log10(2) - log10(5)) - (x[,"n_16"] + x[,"n_40"])*(-5*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_46"])*(-5*log10(2) + log10(5)) - (x[,"n_20"] + x[,"n_36"])*(-4*log10(2) + log10(5)) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(5)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + 2*log10(7)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + 2*log10(7)) - (x[,"n_03"] + x[,"n_53"])*(-3*log10(2) + log10(11)) - (x[,"n_23"] + x[,"n_33"])*(-3*log10(2) + log10(13) + log10(19)) + (x[,"n_04"] + x[,"n_52"])*(5*log10(2) - log10(31)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(31)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(41)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(41)) - (x[,"n_13"] + x[,"n_43"])*(-3*log10(2) + log10(151)) - (x[,"n_14"] + x[,"n_42"])*(-5*log10(2) + log10(379)) - (x[,"n_12"] + x[,"n_44"])*(-5*log10(2) + log10(379)) - (x[,"n_24"] + x[,"n_32"])*(-4*log10(2) + log10(587)) - (x[,"n_22"] + x[,"n_34"])*(-4*log10(2) + log10(587)) + (x[,"n_06"] + x[,"n_50"])*(3*log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_56"])*(2*log10(pmax(1e-6,1 - r_cr)) + 3*log10(pmax(1e-6,r_cr))) + (x[,"n_05"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr)) + log10(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_cr)) + log10(pmax(1e-6,r_cr)) + log10(2 - 4*r_cr + 5*r_cr^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log10(pmax(1e-6,r_cr)) + log10(3 - 9*r_cr + 11*r_cr^2 - 5*r_cr^3)) + (x[,"n_01"] + x[,"n_55"])*(2*log10(pmax(1e-6,r_cr)) + log10(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3)) + (x[,"n_23"] + x[,"n_33"])*log10(27 + 18*r_cr - 28*r_cr^2 + 20*r_cr^3 - 10*r_cr^4) + (x[,"n_03"] + x[,"n_53"])*log10(1 + 4*r_cr - 14*r_cr^2 + 20*r_cr^3 - 10*r_cr^4) + (x[,"n_02"] + x[,"n_54"])*(log10(pmax(1e-6,r_cr)) + log10(3 - 3*r_cr + 3*r_cr^2 - 5*r_cr^3 + 5*r_cr^4)) + (x[,"n_20"] + x[,"n_36"])*(log10(pmax(1e-6,r_cr)) + log10(3 - 12*r_cr + 24*r_cr^2 - 24*r_cr^3 + 10*r_cr^4)) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,r_cr)) + log10(12 - 27*r_cr + 42*r_cr^2 - 41*r_cr^3 + 20*r_cr^4)) + (x[,"n_13"] + x[,"n_43"])*log10(27 - 40*r_cr + 70*r_cr^2 - 60*r_cr^3 + 30*r_cr^4) + (x[,"n_24"] + x[,"n_32"])*log10(54 - 77*r_cr + 152*r_cr^2 - 192*r_cr^3 + 140*r_cr^4 - 50*r_cr^5) + (x[,"n_21"] + x[,"n_35"])*log10(9 - 24*r_cr + 72*r_cr^2 - 108*r_cr^3 + 94*r_cr^4 - 40*r_cr^5) + (x[,"n_12"] + x[,"n_44"])*log10(9 + 3*r_cr + 18*r_cr^2 - 39*r_cr^3 + 40*r_cr^4 - 25*r_cr^5) + (x[,"n_15"] + x[,"n_41"])*log10(6 - 20*r_cr + 53*r_cr^2 - 78*r_cr^3 + 59*r_cr^4 - 20*r_cr^5) + (x[,"n_26"] + x[,"n_30"])*log10(1 - 5*r_cr + 16*r_cr^2 - 28*r_cr^3 + 26*r_cr^4 - 10*r_cr^5) + (x[,"n_04"] + x[,"n_52"])*log10(3 - 11*r_cr + 26*r_cr^2 - 33*r_cr^3 + 20*r_cr^4 - 5*r_cr^5) + (x[,"n_14"] + x[,"n_42"])*log10(6 + 43*r_cr - 109*r_cr^2 + 129*r_cr^3 - 85*r_cr^4 + 25*r_cr^5) + (x[,"n_25"] + x[,"n_31"])*log10(3 + 28*r_cr - 88*r_cr^2 + 132*r_cr^3 - 106*r_cr^4 + 40*r_cr^5) + (x[,"n_22"] + x[,"n_34"])*log10(27 + 39*r_cr - 84*r_cr^2 + 132*r_cr^3 - 110*r_cr^4 + 50*r_cr^5)
logL_cr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"] - 4*x[,"n_56"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*log(5) + (x[,"n_06"] + x[,"n_50"])*(3*log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr))) + (x[,"n_00"] + x[,"n_56"])*(2*log(pmax(1e-6,1 - r_cr)) + 3*log(pmax(1e-6,r_cr))) + (x[,"n_05"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr)) + log(4 - 5*r_cr + 4*r_cr^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_cr)) + log(pmax(1e-6,r_cr)) + log(2 - 4*r_cr + 5*r_cr^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log(pmax(1e-6,r_cr)) + log(3 - 9*r_cr + 11*r_cr^2 - 5*r_cr^3)) + (x[,"n_01"] + x[,"n_55"])*(2*log(pmax(1e-6,r_cr)) + log(3 - 6*r_cr + 7*r_cr^2 - 4*r_cr^3)) + (x[,"n_23"] + x[,"n_33"])*log(27 + 18*r_cr - 28*r_cr^2 + 20*r_cr^3 - 10*r_cr^4) + (x[,"n_03"] + x[,"n_53"])*log(1 + 4*r_cr - 14*r_cr^2 + 20*r_cr^3 - 10*r_cr^4) + (x[,"n_02"] + x[,"n_54"])*(log(pmax(1e-6,r_cr)) + log(3 - 3*r_cr + 3*r_cr^2 - 5*r_cr^3 + 5*r_cr^4)) + (x[,"n_20"] + x[,"n_36"])*(log(pmax(1e-6,r_cr)) + log(3 - 12*r_cr + 24*r_cr^2 - 24*r_cr^3 + 10*r_cr^4)) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,r_cr)) + log(12 - 27*r_cr + 42*r_cr^2 - 41*r_cr^3 + 20*r_cr^4)) + (x[,"n_13"] + x[,"n_43"])*log(27 - 40*r_cr + 70*r_cr^2 - 60*r_cr^3 + 30*r_cr^4) + (x[,"n_24"] + x[,"n_32"])*log(54 - 77*r_cr + 152*r_cr^2 - 192*r_cr^3 + 140*r_cr^4 - 50*r_cr^5) + (x[,"n_21"] + x[,"n_35"])*log(9 - 24*r_cr + 72*r_cr^2 - 108*r_cr^3 + 94*r_cr^4 - 40*r_cr^5) + (x[,"n_12"] + x[,"n_44"])*log(9 + 3*r_cr + 18*r_cr^2 - 39*r_cr^3 + 40*r_cr^4 - 25*r_cr^5) + (x[,"n_15"] + x[,"n_41"])*log(6 - 20*r_cr + 53*r_cr^2 - 78*r_cr^3 + 59*r_cr^4 - 20*r_cr^5) + (x[,"n_26"] + x[,"n_30"])*log(1 - 5*r_cr + 16*r_cr^2 - 28*r_cr^3 + 26*r_cr^4 - 10*r_cr^5) + (x[,"n_04"] + x[,"n_52"])*log(3 - 11*r_cr + 26*r_cr^2 - 33*r_cr^3 + 20*r_cr^4 - 5*r_cr^5) + (x[,"n_14"] + x[,"n_42"])*log(6 + 43*r_cr - 109*r_cr^2 + 129*r_cr^3 - 85*r_cr^4 + 25*r_cr^5) + (x[,"n_25"] + x[,"n_31"])*log(3 + 28*r_cr - 88*r_cr^2 + 132*r_cr^3 - 106*r_cr^4 + 40*r_cr^5) + (x[,"n_22"] + x[,"n_34"])*log(27 + 39*r_cr - 84*r_cr^2 + 132*r_cr^3 - 110*r_cr^4 + 50*r_cr^5)
logL_mc <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-3*n00 - 3*n01 - 3*n02 - 2*n03 - 3*n04 - 3*n05 - 3*n06 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 2*n53 - 3*n54 - 3*n55 - 3*n56)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*(log(3) + 2*log(5)) + (n00 + n56)*(4*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n06 + n50)*(log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n05 + n51)*(3*log(pmax(1e-6,r)) + log(6 - 7*r + 4*r^2)) + (n01 + n55)*(3*log(pmax(1e-6,1 - r)) + log(3 - r + 4*r^2)) + (n16 + n40)*(3*log(pmax(1e-6,r)) + log(9 - 16*r + 10*r^2)) + (n10 + n46)*(3*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 10*r^2)) + (n20 + n36)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(11 - 16*r + 20*r^2)) + (n15 + n41)*(2*log(pmax(1e-6,r)) + log(51 - 78*r + 82*r^2 - 40*r^3)) + (n26 + n30)*(2*log(pmax(1e-6,r)) + log(15 - 39*r + 44*r^2 - 20*r^3)) + (n04 + n52)*(2*log(pmax(1e-6,r)) + log(12 - 18*r + 14*r^2 - 5*r^3)) + (n03 + n53)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2 - 3*r^3)) + (n02 + n54)*(2*log(pmax(1e-6,1 - r)) + log(3 + 5*r - r^2 + 5*r^3)) + (n11 + n45)*(2*log(pmax(1e-6,1 - r)) + log(15 + 34*r - 38*r^2 + 40*r^3)) + (n23 + n33)*log(189 - 16*r - 8*r^2 + 48*r^3 - 24*r^4) + (n13 + n43)*log(57 - 4*r + 40*r^2 - 72*r^3 + 36*r^4) + (n14 + n42)*(log(pmax(1e-6,r)) + log(99 - 111*r + 165*r^2 - 134*r^3 + 50*r^4)) + (n25 + n31)*(log(pmax(1e-6,r)) + log(72 - 129*r + 198*r^2 - 188*r^3 + 80*r^4)) + (n24 + n32)*log(81 + 9*r + 198*r^2 - 303*r^3 + 256*r^4 - 100*r^5) + (n21 + n35)*log(33 - 56*r + 137*r^2 - 246*r^3 + 212*r^4 - 80*r^5) + (n12 + n44)*log(69 - 86*r + 80*r^2 - 129*r^3 + 116*r^4 - 50*r^5) + (n22 + n34)*log(141 - 20*r - 175*r^2 + 279*r^3 - 244*r^4 + 100*r^5)
return(L)}
interlogL_mc <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_mc,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_mc <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_16"] + x[,"n_40"])*(-4*log10(2) + log10(7)) - (x[,"n_10"] + x[,"n_46"])*(-4*log10(2) + log10(7)) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_41"])*(-3*log10(2) + log10(5) + log10(11)) - (x[,"n_11"] + x[,"n_45"])*(-3*log10(2) + log10(5) + log10(11)) - (x[,"n_24"] + x[,"n_32"])*(log10(2) + log10(5) + log10(11)) - (x[,"n_22"] + x[,"n_34"])*(log10(2) + log10(5) + log10(11)) - (x[,"n_25"] + x[,"n_31"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_21"] + x[,"n_35"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_03"] + x[,"n_53"])*(-4*log10(2) + log10(17)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_43"])*(-2*log10(2) + log10(233)) - (x[,"n_23"] + x[,"n_33"])*(-log10(2) + log10(367)) - (x[,"n_14"] + x[,"n_42"])*(-4*log10(2) + log10(569)) - (x[,"n_12"] + x[,"n_44"])*(-4*log10(2) + log10(569)) + (x[,"n_00"] + x[,"n_56"])*(4*log10(pmax(1e-6,1 - r_mc)) + log10(pmax(1e-6,r_mc))) + (x[,"n_06"] + x[,"n_50"])*(log10(pmax(1e-6,1 - r_mc)) + 4*log10(pmax(1e-6,r_mc))) + (x[,"n_05"] + x[,"n_51"])*(3*log10(pmax(1e-6,r_mc)) + log10(6 - 7*r_mc + 4*r_mc^2)) + (x[,"n_01"] + x[,"n_55"])*(3*log10(pmax(1e-6,1 - r_mc)) + log10(3 - r_mc + 4*r_mc^2)) + (x[,"n_16"] + x[,"n_40"])*(3*log10(pmax(1e-6,r_mc)) + log10(9 - 16*r_mc + 10*r_mc^2)) + (x[,"n_10"] + x[,"n_46"])*(3*log10(pmax(1e-6,1 - r_mc)) + log10(3 - 4*r_mc + 10*r_mc^2)) + (x[,"n_20"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(pmax(1e-6,r_mc)) + log10(11 - 16*r_mc + 20*r_mc^2)) + (x[,"n_15"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_mc)) + log10(51 - 78*r_mc + 82*r_mc^2 - 40*r_mc^3)) + (x[,"n_26"] + x[,"n_30"])*(2*log10(pmax(1e-6,r_mc)) + log10(15 - 39*r_mc + 44*r_mc^2 - 20*r_mc^3)) + (x[,"n_04"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_mc)) + log10(12 - 18*r_mc + 14*r_mc^2 - 5*r_mc^3)) + (x[,"n_03"] + x[,"n_53"])*(log10(pmax(1e-6,r_mc)) + log10(5 - 8*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_02"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 + 5*r_mc - r_mc^2 + 5*r_mc^3)) + (x[,"n_11"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(15 + 34*r_mc - 38*r_mc^2 + 40*r_mc^3)) + (x[,"n_23"] + x[,"n_33"])*log10(189 - 16*r_mc - 8*r_mc^2 + 48*r_mc^3 - 24*r_mc^4) + (x[,"n_13"] + x[,"n_43"])*log10(57 - 4*r_mc + 40*r_mc^2 - 72*r_mc^3 + 36*r_mc^4) + (x[,"n_14"] + x[,"n_42"])*(log10(pmax(1e-6,r_mc)) + log10(99 - 111*r_mc + 165*r_mc^2 - 134*r_mc^3 + 50*r_mc^4)) + (x[,"n_25"] + x[,"n_31"])*(log10(pmax(1e-6,r_mc)) + log10(72 - 129*r_mc + 198*r_mc^2 - 188*r_mc^3 + 80*r_mc^4)) + (x[,"n_24"] + x[,"n_32"])*log10(81 + 9*r_mc + 198*r_mc^2 - 303*r_mc^3 + 256*r_mc^4 - 100*r_mc^5) + (x[,"n_21"] + x[,"n_35"])*log10(33 - 56*r_mc + 137*r_mc^2 - 246*r_mc^3 + 212*r_mc^4 - 80*r_mc^5) + (x[,"n_12"] + x[,"n_44"])*log10(69 - 86*r_mc + 80*r_mc^2 - 129*r_mc^3 + 116*r_mc^4 - 50*r_mc^5) + (x[,"n_22"] + x[,"n_34"])*log10(141 - 20*r_mc - 175*r_mc^2 + 279*r_mc^3 - 244*r_mc^4 + 100*r_mc^5)
logL_mc <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 2*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_56"])*(4*log(pmax(1e-6,1 - r_mc)) + log(pmax(1e-6,r_mc))) + (x[,"n_06"] + x[,"n_50"])*(log(pmax(1e-6,1 - r_mc)) + 4*log(pmax(1e-6,r_mc))) + (x[,"n_05"] + x[,"n_51"])*(3*log(pmax(1e-6,r_mc)) + log(6 - 7*r_mc + 4*r_mc^2)) + (x[,"n_01"] + x[,"n_55"])*(3*log(pmax(1e-6,1 - r_mc)) + log(3 - r_mc + 4*r_mc^2)) + (x[,"n_16"] + x[,"n_40"])*(3*log(pmax(1e-6,r_mc)) + log(9 - 16*r_mc + 10*r_mc^2)) + (x[,"n_10"] + x[,"n_46"])*(3*log(pmax(1e-6,1 - r_mc)) + log(3 - 4*r_mc + 10*r_mc^2)) + (x[,"n_20"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_mc)) + log(pmax(1e-6,r_mc)) + log(11 - 16*r_mc + 20*r_mc^2)) + (x[,"n_15"] + x[,"n_41"])*(2*log(pmax(1e-6,r_mc)) + log(51 - 78*r_mc + 82*r_mc^2 - 40*r_mc^3)) + (x[,"n_26"] + x[,"n_30"])*(2*log(pmax(1e-6,r_mc)) + log(15 - 39*r_mc + 44*r_mc^2 - 20*r_mc^3)) + (x[,"n_04"] + x[,"n_52"])*(2*log(pmax(1e-6,r_mc)) + log(12 - 18*r_mc + 14*r_mc^2 - 5*r_mc^3)) + (x[,"n_03"] + x[,"n_53"])*(log(pmax(1e-6,r_mc)) + log(5 - 8*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_02"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 + 5*r_mc - r_mc^2 + 5*r_mc^3)) + (x[,"n_11"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_mc)) + log(15 + 34*r_mc - 38*r_mc^2 + 40*r_mc^3)) + (x[,"n_23"] + x[,"n_33"])*log(189 - 16*r_mc - 8*r_mc^2 + 48*r_mc^3 - 24*r_mc^4) + (x[,"n_13"] + x[,"n_43"])*log(57 - 4*r_mc + 40*r_mc^2 - 72*r_mc^3 + 36*r_mc^4) + (x[,"n_14"] + x[,"n_42"])*(log(pmax(1e-6,r_mc)) + log(99 - 111*r_mc + 165*r_mc^2 - 134*r_mc^3 + 50*r_mc^4)) + (x[,"n_25"] + x[,"n_31"])*(log(pmax(1e-6,r_mc)) + log(72 - 129*r_mc + 198*r_mc^2 - 188*r_mc^3 + 80*r_mc^4)) + (x[,"n_24"] + x[,"n_32"])*log(81 + 9*r_mc + 198*r_mc^2 - 303*r_mc^3 + 256*r_mc^4 - 100*r_mc^5) + (x[,"n_21"] + x[,"n_35"])*log(33 - 56*r_mc + 137*r_mc^2 - 246*r_mc^3 + 212*r_mc^4 - 80*r_mc^5) + (x[,"n_12"] + x[,"n_44"])*log(69 - 86*r_mc + 80*r_mc^2 - 129*r_mc^3 + 116*r_mc^4 - 50*r_mc^5) + (x[,"n_22"] + x[,"n_34"])*log(141 - 20*r_mc - 175*r_mc^2 + 279*r_mc^3 - 244*r_mc^4 + 100*r_mc^5)
logL_mm2c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n06 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55 - 3*n56)*log(2) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - n13 - 2*n14 - 2*n15 - 2*n16 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - 2*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 2*n40 - 2*n41 - 2*n42 - n43 - 2*n44 - 2*n45 - 2*n46 - 2*n50 - 2*n51 - 2*n52 - 2*n53 - 2*n54 - 2*n55 - 2*n56)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*log(5) + (n00 + n56)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n06 + n50)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n55)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(6 - 3*r + 4*r^2)) + (n10 + n46)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(9 - 8*r + 10*r^2)) + (n16 + n40)*(2*log(pmax(1e-6,r)) + log(11 - 23*r + 22*r^2 - 10*r^3)) + (n05 + n51)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 9*r^2 - 4*r^3)) + (n13 + n43)*log(63 - 12*r + 8*r^2 + 8*r^3 - 4*r^4) + (n03 + n53)*log(9 - 6*r + 8*r^2 - 4*r^3 + 2*r^4) + (n04 + n52)*(log(pmax(1e-6,r)) + log(15 - 21*r + 23*r^2 - 13*r^3 + 5*r^4)) + (n23 + n33)*log(531 + 48*r - 40*r^2 - 16*r^3 + 8*r^4) + (n26 + n30)*(log(pmax(1e-6,r)) + log(18 - 44*r + 63*r^2 - 48*r^3 + 20*r^4)) + (n15 + n41)*(log(pmax(1e-6,r)) + log(60 - 77*r + 98*r^2 - 94*r^3 + 40*r^4)) + (n22 + n34)*log(351 - 30*r + 12*r^2 - 175*r^3 + 248*r^4 - 100*r^5) + (n25 + n31)*log(54 - 12*r + 105*r^2 - 182*r^3 + 196*r^4 - 80*r^5) + (n14 + n42)*log(81 + 33*r + 82*r^2 - 139*r^3 + 128*r^4 - 50*r^5) + (n11 + n45)*log(27 - 24*r + 53*r^2 - 122*r^3 + 106*r^4 - 40*r^5) + (n20 + n36)*log(9 - 27*r + 57*r^2 - 71*r^3 + 52*r^4 - 20*r^5) + (n02 + n54)*log(9 - 15*r + 20*r^2 - 21*r^3 + 12*r^4 - 5*r^5) + (n12 + n44)*log(135 - 42*r - 67*r^2 + 127*r^3 - 122*r^4 + 50*r^5) + (n21 + n35)*log(81 - 36*r - 65*r^2 + 198*r^3 - 204*r^4 + 80*r^5) + (n24 + n32)*log(306 + 39*r - 25*r^2 + 183*r^3 - 252*r^4 + 100*r^5)
return(L)}
interlogL_mm2c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_mm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm2c <- mapply(interlogL_mm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_mm2c <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_16"] + x[,"n_40"])*(-4*log10(2) + log10(3) + log10(5)) - (x[,"n_10"] + x[,"n_46"])*(-4*log10(2) + log10(3) + log10(5)) - (x[,"n_26"] + x[,"n_30"])*(-log10(2) + log10(7)) - (x[,"n_20"] + x[,"n_36"])*(-log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_41"])*(-3*log10(2) + log10(3) + 2*log10(7)) - (x[,"n_11"] + x[,"n_45"])*(-3*log10(2) + log10(3) + 2*log10(7)) - (x[,"n_25"] + x[,"n_31"])*(-2*log10(2) + log10(5) + 2*log10(7)) - (x[,"n_21"] + x[,"n_35"])*(-2*log10(2) + log10(5) + 2*log10(7)) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(11) + log10(13)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(11) + log10(13)) - (x[,"n_03"] + x[,"n_53"])*(-3*log10(2) + log10(61)) - (x[,"n_13"] + x[,"n_43"])*(-2*log10(2) + log10(239)) - (x[,"n_14"] + x[,"n_42"])*(-4*log10(2) + log10(3) + log10(571)) - (x[,"n_12"] + x[,"n_44"])*(-4*log10(2) + log10(3) + log10(571)) - (x[,"n_24"] + x[,"n_32"])*(-log10(2) + log10(659)) - (x[,"n_22"] + x[,"n_34"])*(-log10(2) + log10(659)) - (x[,"n_23"] + x[,"n_33"])*(-log10(2) + log10(1087)) + (x[,"n_00"] + x[,"n_56"])*(3*log10(pmax(1e-6,1 - r_mm2c)) + 2*log10(pmax(1e-6,r_mm2c))) + (x[,"n_06"] + x[,"n_50"])*(2*log10(pmax(1e-6,1 - r_mm2c)) + 3*log10(pmax(1e-6,r_mm2c))) + (x[,"n_01"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_mm2c)) + log10(pmax(1e-6,r_mm2c)) + log10(6 - 3*r_mm2c + 4*r_mm2c^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log10(pmax(1e-6,1 - r_mm2c)) + log10(pmax(1e-6,r_mm2c)) + log10(9 - 8*r_mm2c + 10*r_mm2c^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_mm2c)) + log10(11 - 23*r_mm2c + 22*r_mm2c^2 - 10*r_mm2c^3)) + (x[,"n_05"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_mm2c)) + log10(7 - 12*r_mm2c + 9*r_mm2c^2 - 4*r_mm2c^3)) + (x[,"n_13"] + x[,"n_43"])*log10(63 - 12*r_mm2c + 8*r_mm2c^2 + 8*r_mm2c^3 - 4*r_mm2c^4) + (x[,"n_03"] + x[,"n_53"])*log10(9 - 6*r_mm2c + 8*r_mm2c^2 - 4*r_mm2c^3 + 2*r_mm2c^4) + (x[,"n_04"] + x[,"n_52"])*(log10(pmax(1e-6,r_mm2c)) + log10(15 - 21*r_mm2c + 23*r_mm2c^2 - 13*r_mm2c^3 + 5*r_mm2c^4)) + (x[,"n_23"] + x[,"n_33"])*log10(531 + 48*r_mm2c - 40*r_mm2c^2 - 16*r_mm2c^3 + 8*r_mm2c^4) + (x[,"n_26"] + x[,"n_30"])*(log10(pmax(1e-6,r_mm2c)) + log10(18 - 44*r_mm2c + 63*r_mm2c^2 - 48*r_mm2c^3 + 20*r_mm2c^4)) + (x[,"n_15"] + x[,"n_41"])*(log10(pmax(1e-6,r_mm2c)) + log10(60 - 77*r_mm2c + 98*r_mm2c^2 - 94*r_mm2c^3 + 40*r_mm2c^4)) + (x[,"n_22"] + x[,"n_34"])*log10(351 - 30*r_mm2c + 12*r_mm2c^2 - 175*r_mm2c^3 + 248*r_mm2c^4 - 100*r_mm2c^5) + (x[,"n_25"] + x[,"n_31"])*log10(54 - 12*r_mm2c + 105*r_mm2c^2 - 182*r_mm2c^3 + 196*r_mm2c^4 - 80*r_mm2c^5) + (x[,"n_14"] + x[,"n_42"])*log10(81 + 33*r_mm2c + 82*r_mm2c^2 - 139*r_mm2c^3 + 128*r_mm2c^4 - 50*r_mm2c^5) + (x[,"n_11"] + x[,"n_45"])*log10(27 - 24*r_mm2c + 53*r_mm2c^2 - 122*r_mm2c^3 + 106*r_mm2c^4 - 40*r_mm2c^5) + (x[,"n_20"] + x[,"n_36"])*log10(9 - 27*r_mm2c + 57*r_mm2c^2 - 71*r_mm2c^3 + 52*r_mm2c^4 - 20*r_mm2c^5) + (x[,"n_02"] + x[,"n_54"])*log10(9 - 15*r_mm2c + 20*r_mm2c^2 - 21*r_mm2c^3 + 12*r_mm2c^4 - 5*r_mm2c^5) + (x[,"n_12"] + x[,"n_44"])*log10(135 - 42*r_mm2c - 67*r_mm2c^2 + 127*r_mm2c^3 - 122*r_mm2c^4 + 50*r_mm2c^5) + (x[,"n_21"] + x[,"n_35"])*log10(81 - 36*r_mm2c - 65*r_mm2c^2 + 198*r_mm2c^3 - 204*r_mm2c^4 + 80*r_mm2c^5) + (x[,"n_24"] + x[,"n_32"])*log10(306 + 39*r_mm2c - 25*r_mm2c^2 + 183*r_mm2c^3 - 252*r_mm2c^4 + 100*r_mm2c^5)
logL_mm2c <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"])*log(2) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_46"] - 2*x[,"n_50"] - 2*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"] - 2*x[,"n_56"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*log(5) + (x[,"n_00"] + x[,"n_56"])*(3*log(pmax(1e-6,1 - r_mm2c)) + 2*log(pmax(1e-6,r_mm2c))) + (x[,"n_06"] + x[,"n_50"])*(2*log(pmax(1e-6,1 - r_mm2c)) + 3*log(pmax(1e-6,r_mm2c))) + (x[,"n_01"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_mm2c)) + log(pmax(1e-6,r_mm2c)) + log(6 - 3*r_mm2c + 4*r_mm2c^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log(pmax(1e-6,1 - r_mm2c)) + log(pmax(1e-6,r_mm2c)) + log(9 - 8*r_mm2c + 10*r_mm2c^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log(pmax(1e-6,r_mm2c)) + log(11 - 23*r_mm2c + 22*r_mm2c^2 - 10*r_mm2c^3)) + (x[,"n_05"] + x[,"n_51"])*(2*log(pmax(1e-6,r_mm2c)) + log(7 - 12*r_mm2c + 9*r_mm2c^2 - 4*r_mm2c^3)) + (x[,"n_13"] + x[,"n_43"])*log(63 - 12*r_mm2c + 8*r_mm2c^2 + 8*r_mm2c^3 - 4*r_mm2c^4) + (x[,"n_03"] + x[,"n_53"])*log(9 - 6*r_mm2c + 8*r_mm2c^2 - 4*r_mm2c^3 + 2*r_mm2c^4) + (x[,"n_04"] + x[,"n_52"])*(log(pmax(1e-6,r_mm2c)) + log(15 - 21*r_mm2c + 23*r_mm2c^2 - 13*r_mm2c^3 + 5*r_mm2c^4)) + (x[,"n_23"] + x[,"n_33"])*log(531 + 48*r_mm2c - 40*r_mm2c^2 - 16*r_mm2c^3 + 8*r_mm2c^4) + (x[,"n_26"] + x[,"n_30"])*(log(pmax(1e-6,r_mm2c)) + log(18 - 44*r_mm2c + 63*r_mm2c^2 - 48*r_mm2c^3 + 20*r_mm2c^4)) + (x[,"n_15"] + x[,"n_41"])*(log(pmax(1e-6,r_mm2c)) + log(60 - 77*r_mm2c + 98*r_mm2c^2 - 94*r_mm2c^3 + 40*r_mm2c^4)) + (x[,"n_22"] + x[,"n_34"])*log(351 - 30*r_mm2c + 12*r_mm2c^2 - 175*r_mm2c^3 + 248*r_mm2c^4 - 100*r_mm2c^5) + (x[,"n_25"] + x[,"n_31"])*log(54 - 12*r_mm2c + 105*r_mm2c^2 - 182*r_mm2c^3 + 196*r_mm2c^4 - 80*r_mm2c^5) + (x[,"n_14"] + x[,"n_42"])*log(81 + 33*r_mm2c + 82*r_mm2c^2 - 139*r_mm2c^3 + 128*r_mm2c^4 - 50*r_mm2c^5) + (x[,"n_11"] + x[,"n_45"])*log(27 - 24*r_mm2c + 53*r_mm2c^2 - 122*r_mm2c^3 + 106*r_mm2c^4 - 40*r_mm2c^5) + (x[,"n_20"] + x[,"n_36"])*log(9 - 27*r_mm2c + 57*r_mm2c^2 - 71*r_mm2c^3 + 52*r_mm2c^4 - 20*r_mm2c^5) + (x[,"n_02"] + x[,"n_54"])*log(9 - 15*r_mm2c + 20*r_mm2c^2 - 21*r_mm2c^3 + 12*r_mm2c^4 - 5*r_mm2c^5) + (x[,"n_12"] + x[,"n_44"])*log(135 - 42*r_mm2c - 67*r_mm2c^2 + 127*r_mm2c^3 - 122*r_mm2c^4 + 50*r_mm2c^5) + (x[,"n_21"] + x[,"n_35"])*log(81 - 36*r_mm2c - 65*r_mm2c^2 + 198*r_mm2c^3 - 204*r_mm2c^4 + 80*r_mm2c^5) + (x[,"n_24"] + x[,"n_32"])*log(306 + 39*r_mm2c - 25*r_mm2c^2 + 183*r_mm2c^3 - 252*r_mm2c^4 + 100*r_mm2c^5)
logL_mm1c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-3*n00 - 3*n01 - 3*n02 - 3*n03 - 3*n04 - 3*n05 - 3*n06 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55 - 3*n56)*log(2) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - n13 - 2*n14 - 2*n15 - 2*n16 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - 2*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 2*n40 - 2*n41 - 2*n42 - n43 - 2*n44 - 2*n45 - 2*n46 - 2*n50 - 2*n51 - 2*n52 - 2*n53 - 2*n54 - 2*n55 - 2*n56)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*log(5) + (n06 + n50)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n00 + n56)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n05 + n51)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(6 - 3*r + 4*r^2)) + (n16 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(9 - 8*r + 10*r^2)) + (n10 + n46)*(2*log(pmax(1e-6,r)) + log(11 - 23*r + 22*r^2 - 10*r^3)) + (n01 + n55)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 9*r^2 - 4*r^3)) + (n13 + n43)*log(63 - 12*r + 8*r^2 + 8*r^3 - 4*r^4) + (n03 + n53)*log(9 - 6*r + 8*r^2 - 4*r^3 + 2*r^4) + (n02 + n54)*(log(pmax(1e-6,r)) + log(15 - 21*r + 23*r^2 - 13*r^3 + 5*r^4)) + (n23 + n33)*log(531 + 48*r - 40*r^2 - 16*r^3 + 8*r^4) + (n20 + n36)*(log(pmax(1e-6,r)) + log(18 - 44*r + 63*r^2 - 48*r^3 + 20*r^4)) + (n11 + n45)*(log(pmax(1e-6,r)) + log(60 - 77*r + 98*r^2 - 94*r^3 + 40*r^4)) + (n24 + n32)*log(351 - 30*r + 12*r^2 - 175*r^3 + 248*r^4 - 100*r^5) + (n21 + n35)*log(54 - 12*r + 105*r^2 - 182*r^3 + 196*r^4 - 80*r^5) + (n12 + n44)*log(81 + 33*r + 82*r^2 - 139*r^3 + 128*r^4 - 50*r^5) + (n15 + n41)*log(27 - 24*r + 53*r^2 - 122*r^3 + 106*r^4 - 40*r^5) + (n26 + n30)*log(9 - 27*r + 57*r^2 - 71*r^3 + 52*r^4 - 20*r^5) + (n04 + n52)*log(9 - 15*r + 20*r^2 - 21*r^3 + 12*r^4 - 5*r^5) + (n14 + n42)*log(135 - 42*r - 67*r^2 + 127*r^3 - 122*r^4 + 50*r^5) + (n25 + n31)*log(81 - 36*r - 65*r^2 + 198*r^3 - 204*r^4 + 80*r^5) + (n22 + n34)*log(306 + 39*r - 25*r^2 + 183*r^3 - 252*r^4 + 100*r^5)
return(L)}
interlogL_mm1c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_mm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm1c <- mapply(interlogL_mm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_mm1c <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_16"] + x[,"n_40"])*(-4*log10(2) + log10(3) + log10(5)) - (x[,"n_10"] + x[,"n_46"])*(-4*log10(2) + log10(3) + log10(5)) - (x[,"n_26"] + x[,"n_30"])*(-log10(2) + log10(7)) - (x[,"n_20"] + x[,"n_36"])*(-log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_41"])*(-3*log10(2) + log10(3) + 2*log10(7)) - (x[,"n_11"] + x[,"n_45"])*(-3*log10(2) + log10(3) + 2*log10(7)) - (x[,"n_25"] + x[,"n_31"])*(-2*log10(2) + log10(5) + 2*log10(7)) - (x[,"n_21"] + x[,"n_35"])*(-2*log10(2) + log10(5) + 2*log10(7)) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(11) + log10(13)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(11) + log10(13)) - (x[,"n_03"] + x[,"n_53"])*(-3*log10(2) + log10(61)) - (x[,"n_13"] + x[,"n_43"])*(-2*log10(2) + log10(239)) - (x[,"n_14"] + x[,"n_42"])*(-4*log10(2) + log10(3) + log10(571)) - (x[,"n_12"] + x[,"n_44"])*(-4*log10(2) + log10(3) + log10(571)) - (x[,"n_24"] + x[,"n_32"])*(-log10(2) + log10(659)) - (x[,"n_22"] + x[,"n_34"])*(-log10(2) + log10(659)) - (x[,"n_23"] + x[,"n_33"])*(-log10(2) + log10(1087)) + (x[,"n_06"] + x[,"n_50"])*(3*log10(pmax(1e-6,1 - r_mm1c)) + 2*log10(pmax(1e-6,r_mm1c))) + (x[,"n_00"] + x[,"n_56"])*(2*log10(pmax(1e-6,1 - r_mm1c)) + 3*log10(pmax(1e-6,r_mm1c))) + (x[,"n_05"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_mm1c)) + log10(pmax(1e-6,r_mm1c)) + log10(6 - 3*r_mm1c + 4*r_mm1c^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_mm1c)) + log10(pmax(1e-6,r_mm1c)) + log10(9 - 8*r_mm1c + 10*r_mm1c^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log10(pmax(1e-6,r_mm1c)) + log10(11 - 23*r_mm1c + 22*r_mm1c^2 - 10*r_mm1c^3)) + (x[,"n_01"] + x[,"n_55"])*(2*log10(pmax(1e-6,r_mm1c)) + log10(7 - 12*r_mm1c + 9*r_mm1c^2 - 4*r_mm1c^3)) + (x[,"n_13"] + x[,"n_43"])*log10(63 - 12*r_mm1c + 8*r_mm1c^2 + 8*r_mm1c^3 - 4*r_mm1c^4) + (x[,"n_03"] + x[,"n_53"])*log10(9 - 6*r_mm1c + 8*r_mm1c^2 - 4*r_mm1c^3 + 2*r_mm1c^4) + (x[,"n_02"] + x[,"n_54"])*(log10(pmax(1e-6,r_mm1c)) + log10(15 - 21*r_mm1c + 23*r_mm1c^2 - 13*r_mm1c^3 + 5*r_mm1c^4)) + (x[,"n_23"] + x[,"n_33"])*log10(531 + 48*r_mm1c - 40*r_mm1c^2 - 16*r_mm1c^3 + 8*r_mm1c^4) + (x[,"n_20"] + x[,"n_36"])*(log10(pmax(1e-6,r_mm1c)) + log10(18 - 44*r_mm1c + 63*r_mm1c^2 - 48*r_mm1c^3 + 20*r_mm1c^4)) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,r_mm1c)) + log10(60 - 77*r_mm1c + 98*r_mm1c^2 - 94*r_mm1c^3 + 40*r_mm1c^4)) + (x[,"n_24"] + x[,"n_32"])*log10(351 - 30*r_mm1c + 12*r_mm1c^2 - 175*r_mm1c^3 + 248*r_mm1c^4 - 100*r_mm1c^5) + (x[,"n_21"] + x[,"n_35"])*log10(54 - 12*r_mm1c + 105*r_mm1c^2 - 182*r_mm1c^3 + 196*r_mm1c^4 - 80*r_mm1c^5) + (x[,"n_12"] + x[,"n_44"])*log10(81 + 33*r_mm1c + 82*r_mm1c^2 - 139*r_mm1c^3 + 128*r_mm1c^4 - 50*r_mm1c^5) + (x[,"n_15"] + x[,"n_41"])*log10(27 - 24*r_mm1c + 53*r_mm1c^2 - 122*r_mm1c^3 + 106*r_mm1c^4 - 40*r_mm1c^5) + (x[,"n_26"] + x[,"n_30"])*log10(9 - 27*r_mm1c + 57*r_mm1c^2 - 71*r_mm1c^3 + 52*r_mm1c^4 - 20*r_mm1c^5) + (x[,"n_04"] + x[,"n_52"])*log10(9 - 15*r_mm1c + 20*r_mm1c^2 - 21*r_mm1c^3 + 12*r_mm1c^4 - 5*r_mm1c^5) + (x[,"n_14"] + x[,"n_42"])*log10(135 - 42*r_mm1c - 67*r_mm1c^2 + 127*r_mm1c^3 - 122*r_mm1c^4 + 50*r_mm1c^5) + (x[,"n_25"] + x[,"n_31"])*log10(81 - 36*r_mm1c - 65*r_mm1c^2 + 198*r_mm1c^3 - 204*r_mm1c^4 + 80*r_mm1c^5) + (x[,"n_22"] + x[,"n_34"])*log10(306 + 39*r_mm1c - 25*r_mm1c^2 + 183*r_mm1c^3 - 252*r_mm1c^4 + 100*r_mm1c^5)
logL_mm1c <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"])*log(2) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_46"] - 2*x[,"n_50"] - 2*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"] - 2*x[,"n_56"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*log(5) + (x[,"n_06"] + x[,"n_50"])*(3*log(pmax(1e-6,1 - r_mm1c)) + 2*log(pmax(1e-6,r_mm1c))) + (x[,"n_00"] + x[,"n_56"])*(2*log(pmax(1e-6,1 - r_mm1c)) + 3*log(pmax(1e-6,r_mm1c))) + (x[,"n_05"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_mm1c)) + log(pmax(1e-6,r_mm1c)) + log(6 - 3*r_mm1c + 4*r_mm1c^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_mm1c)) + log(pmax(1e-6,r_mm1c)) + log(9 - 8*r_mm1c + 10*r_mm1c^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log(pmax(1e-6,r_mm1c)) + log(11 - 23*r_mm1c + 22*r_mm1c^2 - 10*r_mm1c^3)) + (x[,"n_01"] + x[,"n_55"])*(2*log(pmax(1e-6,r_mm1c)) + log(7 - 12*r_mm1c + 9*r_mm1c^2 - 4*r_mm1c^3)) + (x[,"n_13"] + x[,"n_43"])*log(63 - 12*r_mm1c + 8*r_mm1c^2 + 8*r_mm1c^3 - 4*r_mm1c^4) + (x[,"n_03"] + x[,"n_53"])*log(9 - 6*r_mm1c + 8*r_mm1c^2 - 4*r_mm1c^3 + 2*r_mm1c^4) + (x[,"n_02"] + x[,"n_54"])*(log(pmax(1e-6,r_mm1c)) + log(15 - 21*r_mm1c + 23*r_mm1c^2 - 13*r_mm1c^3 + 5*r_mm1c^4)) + (x[,"n_23"] + x[,"n_33"])*log(531 + 48*r_mm1c - 40*r_mm1c^2 - 16*r_mm1c^3 + 8*r_mm1c^4) + (x[,"n_20"] + x[,"n_36"])*(log(pmax(1e-6,r_mm1c)) + log(18 - 44*r_mm1c + 63*r_mm1c^2 - 48*r_mm1c^3 + 20*r_mm1c^4)) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,r_mm1c)) + log(60 - 77*r_mm1c + 98*r_mm1c^2 - 94*r_mm1c^3 + 40*r_mm1c^4)) + (x[,"n_24"] + x[,"n_32"])*log(351 - 30*r_mm1c + 12*r_mm1c^2 - 175*r_mm1c^3 + 248*r_mm1c^4 - 100*r_mm1c^5) + (x[,"n_21"] + x[,"n_35"])*log(54 - 12*r_mm1c + 105*r_mm1c^2 - 182*r_mm1c^3 + 196*r_mm1c^4 - 80*r_mm1c^5) + (x[,"n_12"] + x[,"n_44"])*log(81 + 33*r_mm1c + 82*r_mm1c^2 - 139*r_mm1c^3 + 128*r_mm1c^4 - 50*r_mm1c^5) + (x[,"n_15"] + x[,"n_41"])*log(27 - 24*r_mm1c + 53*r_mm1c^2 - 122*r_mm1c^3 + 106*r_mm1c^4 - 40*r_mm1c^5) + (x[,"n_26"] + x[,"n_30"])*log(9 - 27*r_mm1c + 57*r_mm1c^2 - 71*r_mm1c^3 + 52*r_mm1c^4 - 20*r_mm1c^5) + (x[,"n_04"] + x[,"n_52"])*log(9 - 15*r_mm1c + 20*r_mm1c^2 - 21*r_mm1c^3 + 12*r_mm1c^4 - 5*r_mm1c^5) + (x[,"n_14"] + x[,"n_42"])*log(135 - 42*r_mm1c - 67*r_mm1c^2 + 127*r_mm1c^3 - 122*r_mm1c^4 + 50*r_mm1c^5) + (x[,"n_25"] + x[,"n_31"])*log(81 - 36*r_mm1c - 65*r_mm1c^2 + 198*r_mm1c^3 - 204*r_mm1c^4 + 80*r_mm1c^5) + (x[,"n_22"] + x[,"n_34"])*log(306 + 39*r_mm1c - 25*r_mm1c^2 + 183*r_mm1c^3 - 252*r_mm1c^4 + 100*r_mm1c^5)
logL_mr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-3*n00 - 3*n01 - 3*n02 - 2*n03 - 3*n04 - 3*n05 - 3*n06 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 4*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 4*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 2*n53 - 3*n54 - 3*n55 - 3*n56)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*(log(3) + 2*log(5)) + (n06 + n50)*(4*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n56)*(log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n01 + n55)*(3*log(pmax(1e-6,r)) + log(6 - 7*r + 4*r^2)) + (n05 + n51)*(3*log(pmax(1e-6,1 - r)) + log(3 - r + 4*r^2)) + (n10 + n46)*(3*log(pmax(1e-6,r)) + log(9 - 16*r + 10*r^2)) + (n16 + n40)*(3*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 10*r^2)) + (n26 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(11 - 16*r + 20*r^2)) + (n11 + n45)*(2*log(pmax(1e-6,r)) + log(51 - 78*r + 82*r^2 - 40*r^3)) + (n20 + n36)*(2*log(pmax(1e-6,r)) + log(15 - 39*r + 44*r^2 - 20*r^3)) + (n02 + n54)*(2*log(pmax(1e-6,r)) + log(12 - 18*r + 14*r^2 - 5*r^3)) + (n03 + n53)*(log(pmax(1e-6,r)) + log(5 - 8*r + 6*r^2 - 3*r^3)) + (n04 + n52)*(2*log(pmax(1e-6,1 - r)) + log(3 + 5*r - r^2 + 5*r^3)) + (n15 + n41)*(2*log(pmax(1e-6,1 - r)) + log(15 + 34*r - 38*r^2 + 40*r^3)) + (n23 + n33)*log(189 - 16*r - 8*r^2 + 48*r^3 - 24*r^4) + (n13 + n43)*log(57 - 4*r + 40*r^2 - 72*r^3 + 36*r^4) + (n12 + n44)*(log(pmax(1e-6,r)) + log(99 - 111*r + 165*r^2 - 134*r^3 + 50*r^4)) + (n21 + n35)*(log(pmax(1e-6,r)) + log(72 - 129*r + 198*r^2 - 188*r^3 + 80*r^4)) + (n22 + n34)*log(81 + 9*r + 198*r^2 - 303*r^3 + 256*r^4 - 100*r^5) + (n25 + n31)*log(33 - 56*r + 137*r^2 - 246*r^3 + 212*r^4 - 80*r^5) + (n14 + n42)*log(69 - 86*r + 80*r^2 - 129*r^3 + 116*r^4 - 50*r^5) + (n24 + n32)*log(141 - 20*r - 175*r^2 + 279*r^3 - 244*r^4 + 100*r^5)
return(L)}
interlogL_mr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_mr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mr <- mapply(interlogL_mr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_mr <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_16"] + x[,"n_40"])*(-4*log10(2) + log10(7)) - (x[,"n_10"] + x[,"n_46"])*(-4*log10(2) + log10(7)) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_41"])*(-3*log10(2) + log10(5) + log10(11)) - (x[,"n_11"] + x[,"n_45"])*(-3*log10(2) + log10(5) + log10(11)) - (x[,"n_24"] + x[,"n_32"])*(log10(2) + log10(5) + log10(11)) - (x[,"n_22"] + x[,"n_34"])*(log10(2) + log10(5) + log10(11)) - (x[,"n_25"] + x[,"n_31"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_21"] + x[,"n_35"])*(-2*log10(2) + log10(7) + log10(11)) - (x[,"n_03"] + x[,"n_53"])*(-4*log10(2) + log10(17)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_43"])*(-2*log10(2) + log10(233)) - (x[,"n_23"] + x[,"n_33"])*(-log10(2) + log10(367)) - (x[,"n_14"] + x[,"n_42"])*(-4*log10(2) + log10(569)) - (x[,"n_12"] + x[,"n_44"])*(-4*log10(2) + log10(569)) + (x[,"n_06"] + x[,"n_50"])*(4*log10(pmax(1e-6,1 - r_mr)) + log10(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_56"])*(log10(pmax(1e-6,1 - r_mr)) + 4*log10(pmax(1e-6,r_mr))) + (x[,"n_01"] + x[,"n_55"])*(3*log10(pmax(1e-6,r_mr)) + log10(6 - 7*r_mr + 4*r_mr^2)) + (x[,"n_05"] + x[,"n_51"])*(3*log10(pmax(1e-6,1 - r_mr)) + log10(3 - r_mr + 4*r_mr^2)) + (x[,"n_10"] + x[,"n_46"])*(3*log10(pmax(1e-6,r_mr)) + log10(9 - 16*r_mr + 10*r_mr^2)) + (x[,"n_16"] + x[,"n_40"])*(3*log10(pmax(1e-6,1 - r_mr)) + log10(3 - 4*r_mr + 10*r_mr^2)) + (x[,"n_26"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(pmax(1e-6,r_mr)) + log10(11 - 16*r_mr + 20*r_mr^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_mr)) + log10(51 - 78*r_mr + 82*r_mr^2 - 40*r_mr^3)) + (x[,"n_20"] + x[,"n_36"])*(2*log10(pmax(1e-6,r_mr)) + log10(15 - 39*r_mr + 44*r_mr^2 - 20*r_mr^3)) + (x[,"n_02"] + x[,"n_54"])*(2*log10(pmax(1e-6,r_mr)) + log10(12 - 18*r_mr + 14*r_mr^2 - 5*r_mr^3)) + (x[,"n_03"] + x[,"n_53"])*(log10(pmax(1e-6,r_mr)) + log10(5 - 8*r_mr + 6*r_mr^2 - 3*r_mr^3)) + (x[,"n_04"] + x[,"n_52"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(3 + 5*r_mr - r_mr^2 + 5*r_mr^3)) + (x[,"n_15"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_mr)) + log10(15 + 34*r_mr - 38*r_mr^2 + 40*r_mr^3)) + (x[,"n_23"] + x[,"n_33"])*log10(189 - 16*r_mr - 8*r_mr^2 + 48*r_mr^3 - 24*r_mr^4) + (x[,"n_13"] + x[,"n_43"])*log10(57 - 4*r_mr + 40*r_mr^2 - 72*r_mr^3 + 36*r_mr^4) + (x[,"n_12"] + x[,"n_44"])*(log10(pmax(1e-6,r_mr)) + log10(99 - 111*r_mr + 165*r_mr^2 - 134*r_mr^3 + 50*r_mr^4)) + (x[,"n_21"] + x[,"n_35"])*(log10(pmax(1e-6,r_mr)) + log10(72 - 129*r_mr + 198*r_mr^2 - 188*r_mr^3 + 80*r_mr^4)) + (x[,"n_22"] + x[,"n_34"])*log10(81 + 9*r_mr + 198*r_mr^2 - 303*r_mr^3 + 256*r_mr^4 - 100*r_mr^5) + (x[,"n_25"] + x[,"n_31"])*log10(33 - 56*r_mr + 137*r_mr^2 - 246*r_mr^3 + 212*r_mr^4 - 80*r_mr^5) + (x[,"n_14"] + x[,"n_42"])*log10(69 - 86*r_mr + 80*r_mr^2 - 129*r_mr^3 + 116*r_mr^4 - 50*r_mr^5) + (x[,"n_24"] + x[,"n_32"])*log10(141 - 20*r_mr - 175*r_mr^2 + 279*r_mr^3 - 244*r_mr^4 + 100*r_mr^5)
logL_mr <- (-3*x[,"n_00"] - 3*x[,"n_01"] - 3*x[,"n_02"] - 2*x[,"n_03"] - 3*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 4*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 2*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*(log(3) + 2*log(5)) + (x[,"n_06"] + x[,"n_50"])*(4*log(pmax(1e-6,1 - r_mr)) + log(pmax(1e-6,r_mr))) + (x[,"n_00"] + x[,"n_56"])*(log(pmax(1e-6,1 - r_mr)) + 4*log(pmax(1e-6,r_mr))) + (x[,"n_01"] + x[,"n_55"])*(3*log(pmax(1e-6,r_mr)) + log(6 - 7*r_mr + 4*r_mr^2)) + (x[,"n_05"] + x[,"n_51"])*(3*log(pmax(1e-6,1 - r_mr)) + log(3 - r_mr + 4*r_mr^2)) + (x[,"n_10"] + x[,"n_46"])*(3*log(pmax(1e-6,r_mr)) + log(9 - 16*r_mr + 10*r_mr^2)) + (x[,"n_16"] + x[,"n_40"])*(3*log(pmax(1e-6,1 - r_mr)) + log(3 - 4*r_mr + 10*r_mr^2)) + (x[,"n_26"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_mr)) + log(pmax(1e-6,r_mr)) + log(11 - 16*r_mr + 20*r_mr^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log(pmax(1e-6,r_mr)) + log(51 - 78*r_mr + 82*r_mr^2 - 40*r_mr^3)) + (x[,"n_20"] + x[,"n_36"])*(2*log(pmax(1e-6,r_mr)) + log(15 - 39*r_mr + 44*r_mr^2 - 20*r_mr^3)) + (x[,"n_02"] + x[,"n_54"])*(2*log(pmax(1e-6,r_mr)) + log(12 - 18*r_mr + 14*r_mr^2 - 5*r_mr^3)) + (x[,"n_03"] + x[,"n_53"])*(log(pmax(1e-6,r_mr)) + log(5 - 8*r_mr + 6*r_mr^2 - 3*r_mr^3)) + (x[,"n_04"] + x[,"n_52"])*(2*log(pmax(1e-6,1 - r_mr)) + log(3 + 5*r_mr - r_mr^2 + 5*r_mr^3)) + (x[,"n_15"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_mr)) + log(15 + 34*r_mr - 38*r_mr^2 + 40*r_mr^3)) + (x[,"n_23"] + x[,"n_33"])*log(189 - 16*r_mr - 8*r_mr^2 + 48*r_mr^3 - 24*r_mr^4) + (x[,"n_13"] + x[,"n_43"])*log(57 - 4*r_mr + 40*r_mr^2 - 72*r_mr^3 + 36*r_mr^4) + (x[,"n_12"] + x[,"n_44"])*(log(pmax(1e-6,r_mr)) + log(99 - 111*r_mr + 165*r_mr^2 - 134*r_mr^3 + 50*r_mr^4)) + (x[,"n_21"] + x[,"n_35"])*(log(pmax(1e-6,r_mr)) + log(72 - 129*r_mr + 198*r_mr^2 - 188*r_mr^3 + 80*r_mr^4)) + (x[,"n_22"] + x[,"n_34"])*log(81 + 9*r_mr + 198*r_mr^2 - 303*r_mr^3 + 256*r_mr^4 - 100*r_mr^5) + (x[,"n_25"] + x[,"n_31"])*log(33 - 56*r_mr + 137*r_mr^2 - 246*r_mr^3 + 212*r_mr^4 - 80*r_mr^5) + (x[,"n_14"] + x[,"n_42"])*log(69 - 86*r_mr + 80*r_mr^2 - 129*r_mr^3 + 116*r_mr^4 - 50*r_mr^5) + (x[,"n_24"] + x[,"n_32"])*log(141 - 20*r_mr - 175*r_mr^2 + 279*r_mr^3 - 244*r_mr^4 + 100*r_mr^5)
logL_rc <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-4*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 4*n11 - 4*n12 - 4*n13 - 4*n14 - 4*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26 - 4*n30 - 4*n31 - 4*n32 - 3*n33 - 4*n34 - 4*n35 - 4*n36 - 4*n40 - 4*n41 - 4*n42 - 4*n43 - 4*n44 - 4*n45 - 4*n46 - 4*n50 - 4*n51 - 4*n52 - 4*n53 - 4*n54 - 4*n55 - 4*n56)*log(2) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*log(5) + (n00 + n56)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n06 + n50)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n55)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(4 - 5*r + 4*r^2)) + (n10 + n46)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 - 4*r + 5*r^2)) + (n16 + n40)*(2*log(pmax(1e-6,r)) + log(3 - 9*r + 11*r^2 - 5*r^3)) + (n05 + n51)*(2*log(pmax(1e-6,r)) + log(3 - 6*r + 7*r^2 - 4*r^3)) + (n23 + n33)*log(27 + 18*r - 28*r^2 + 20*r^3 - 10*r^4) + (n03 + n53)*log(1 + 4*r - 14*r^2 + 20*r^3 - 10*r^4) + (n04 + n52)*(log(pmax(1e-6,r)) + log(3 - 3*r + 3*r^2 - 5*r^3 + 5*r^4)) + (n26 + n30)*(log(pmax(1e-6,r)) + log(3 - 12*r + 24*r^2 - 24*r^3 + 10*r^4)) + (n15 + n41)*(log(pmax(1e-6,r)) + log(12 - 27*r + 42*r^2 - 41*r^3 + 20*r^4)) + (n13 + n43)*log(27 - 40*r + 70*r^2 - 60*r^3 + 30*r^4) + (n22 + n34)*log(54 - 77*r + 152*r^2 - 192*r^3 + 140*r^4 - 50*r^5) + (n25 + n31)*log(9 - 24*r + 72*r^2 - 108*r^3 + 94*r^4 - 40*r^5) + (n14 + n42)*log(9 + 3*r + 18*r^2 - 39*r^3 + 40*r^4 - 25*r^5) + (n11 + n45)*log(6 - 20*r + 53*r^2 - 78*r^3 + 59*r^4 - 20*r^5) + (n20 + n36)*log(1 - 5*r + 16*r^2 - 28*r^3 + 26*r^4 - 10*r^5) + (n02 + n54)*log(3 - 11*r + 26*r^2 - 33*r^3 + 20*r^4 - 5*r^5) + (n12 + n44)*log(6 + 43*r - 109*r^2 + 129*r^3 - 85*r^4 + 25*r^5) + (n21 + n35)*log(3 + 28*r - 88*r^2 + 132*r^3 - 106*r^4 + 40*r^5) + (n24 + n32)*log(27 + 39*r - 84*r^2 + 132*r^3 - 110*r^4 + 50*r^5)
return(L)}
interlogL_rc <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_rc,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_rc <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) + (x[,"n_20"] + x[,"n_36"])*(4*log10(2) - log10(5)) - (x[,"n_16"] + x[,"n_40"])*(-5*log10(2) + log10(5)) - (x[,"n_10"] + x[,"n_46"])*(-5*log10(2) + log10(5)) - (x[,"n_26"] + x[,"n_30"])*(-4*log10(2) + log10(5)) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(5)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + 2*log10(7)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + 2*log10(7)) - (x[,"n_03"] + x[,"n_53"])*(-3*log10(2) + log10(11)) - (x[,"n_23"] + x[,"n_33"])*(-3*log10(2) + log10(13) + log10(19)) + (x[,"n_02"] + x[,"n_54"])*(5*log10(2) - log10(31)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(31)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(41)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(41)) - (x[,"n_13"] + x[,"n_43"])*(-3*log10(2) + log10(151)) - (x[,"n_14"] + x[,"n_42"])*(-5*log10(2) + log10(379)) - (x[,"n_12"] + x[,"n_44"])*(-5*log10(2) + log10(379)) - (x[,"n_24"] + x[,"n_32"])*(-4*log10(2) + log10(587)) - (x[,"n_22"] + x[,"n_34"])*(-4*log10(2) + log10(587)) + (x[,"n_00"] + x[,"n_56"])*(3*log10(pmax(1e-6,1 - r_rc)) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_06"] + x[,"n_50"])*(2*log10(pmax(1e-6,1 - r_rc)) + 3*log10(pmax(1e-6,r_rc))) + (x[,"n_01"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc)) + log10(4 - 5*r_rc + 4*r_rc^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log10(pmax(1e-6,1 - r_rc)) + log10(pmax(1e-6,r_rc)) + log10(2 - 4*r_rc + 5*r_rc^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log10(pmax(1e-6,r_rc)) + log10(3 - 9*r_rc + 11*r_rc^2 - 5*r_rc^3)) + (x[,"n_05"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_rc)) + log10(3 - 6*r_rc + 7*r_rc^2 - 4*r_rc^3)) + (x[,"n_23"] + x[,"n_33"])*log10(27 + 18*r_rc - 28*r_rc^2 + 20*r_rc^3 - 10*r_rc^4) + (x[,"n_03"] + x[,"n_53"])*log10(1 + 4*r_rc - 14*r_rc^2 + 20*r_rc^3 - 10*r_rc^4) + (x[,"n_04"] + x[,"n_52"])*(log10(pmax(1e-6,r_rc)) + log10(3 - 3*r_rc + 3*r_rc^2 - 5*r_rc^3 + 5*r_rc^4)) + (x[,"n_26"] + x[,"n_30"])*(log10(pmax(1e-6,r_rc)) + log10(3 - 12*r_rc + 24*r_rc^2 - 24*r_rc^3 + 10*r_rc^4)) + (x[,"n_15"] + x[,"n_41"])*(log10(pmax(1e-6,r_rc)) + log10(12 - 27*r_rc + 42*r_rc^2 - 41*r_rc^3 + 20*r_rc^4)) + (x[,"n_13"] + x[,"n_43"])*log10(27 - 40*r_rc + 70*r_rc^2 - 60*r_rc^3 + 30*r_rc^4) + (x[,"n_22"] + x[,"n_34"])*log10(54 - 77*r_rc + 152*r_rc^2 - 192*r_rc^3 + 140*r_rc^4 - 50*r_rc^5) + (x[,"n_25"] + x[,"n_31"])*log10(9 - 24*r_rc + 72*r_rc^2 - 108*r_rc^3 + 94*r_rc^4 - 40*r_rc^5) + (x[,"n_14"] + x[,"n_42"])*log10(9 + 3*r_rc + 18*r_rc^2 - 39*r_rc^3 + 40*r_rc^4 - 25*r_rc^5) + (x[,"n_11"] + x[,"n_45"])*log10(6 - 20*r_rc + 53*r_rc^2 - 78*r_rc^3 + 59*r_rc^4 - 20*r_rc^5) + (x[,"n_20"] + x[,"n_36"])*log10(1 - 5*r_rc + 16*r_rc^2 - 28*r_rc^3 + 26*r_rc^4 - 10*r_rc^5) + (x[,"n_02"] + x[,"n_54"])*log10(3 - 11*r_rc + 26*r_rc^2 - 33*r_rc^3 + 20*r_rc^4 - 5*r_rc^5) + (x[,"n_12"] + x[,"n_44"])*log10(6 + 43*r_rc - 109*r_rc^2 + 129*r_rc^3 - 85*r_rc^4 + 25*r_rc^5) + (x[,"n_21"] + x[,"n_35"])*log10(3 + 28*r_rc - 88*r_rc^2 + 132*r_rc^3 - 106*r_rc^4 + 40*r_rc^5) + (x[,"n_24"] + x[,"n_32"])*log10(27 + 39*r_rc - 84*r_rc^2 + 132*r_rc^3 - 110*r_rc^4 + 50*r_rc^5)
logL_rc <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 4*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 4*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 3*x[,"n_33"] - 4*x[,"n_34"] - 4*x[,"n_35"] - 4*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 4*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 4*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"] - 4*x[,"n_56"])*log(2) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*log(5) + (x[,"n_00"] + x[,"n_56"])*(3*log(pmax(1e-6,1 - r_rc)) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_06"] + x[,"n_50"])*(2*log(pmax(1e-6,1 - r_rc)) + 3*log(pmax(1e-6,r_rc))) + (x[,"n_01"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc)) + log(4 - 5*r_rc + 4*r_rc^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log(pmax(1e-6,1 - r_rc)) + log(pmax(1e-6,r_rc)) + log(2 - 4*r_rc + 5*r_rc^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log(pmax(1e-6,r_rc)) + log(3 - 9*r_rc + 11*r_rc^2 - 5*r_rc^3)) + (x[,"n_05"] + x[,"n_51"])*(2*log(pmax(1e-6,r_rc)) + log(3 - 6*r_rc + 7*r_rc^2 - 4*r_rc^3)) + (x[,"n_23"] + x[,"n_33"])*log(27 + 18*r_rc - 28*r_rc^2 + 20*r_rc^3 - 10*r_rc^4) + (x[,"n_03"] + x[,"n_53"])*log(1 + 4*r_rc - 14*r_rc^2 + 20*r_rc^3 - 10*r_rc^4) + (x[,"n_04"] + x[,"n_52"])*(log(pmax(1e-6,r_rc)) + log(3 - 3*r_rc + 3*r_rc^2 - 5*r_rc^3 + 5*r_rc^4)) + (x[,"n_26"] + x[,"n_30"])*(log(pmax(1e-6,r_rc)) + log(3 - 12*r_rc + 24*r_rc^2 - 24*r_rc^3 + 10*r_rc^4)) + (x[,"n_15"] + x[,"n_41"])*(log(pmax(1e-6,r_rc)) + log(12 - 27*r_rc + 42*r_rc^2 - 41*r_rc^3 + 20*r_rc^4)) + (x[,"n_13"] + x[,"n_43"])*log(27 - 40*r_rc + 70*r_rc^2 - 60*r_rc^3 + 30*r_rc^4) + (x[,"n_22"] + x[,"n_34"])*log(54 - 77*r_rc + 152*r_rc^2 - 192*r_rc^3 + 140*r_rc^4 - 50*r_rc^5) + (x[,"n_25"] + x[,"n_31"])*log(9 - 24*r_rc + 72*r_rc^2 - 108*r_rc^3 + 94*r_rc^4 - 40*r_rc^5) + (x[,"n_14"] + x[,"n_42"])*log(9 + 3*r_rc + 18*r_rc^2 - 39*r_rc^3 + 40*r_rc^4 - 25*r_rc^5) + (x[,"n_11"] + x[,"n_45"])*log(6 - 20*r_rc + 53*r_rc^2 - 78*r_rc^3 + 59*r_rc^4 - 20*r_rc^5) + (x[,"n_20"] + x[,"n_36"])*log(1 - 5*r_rc + 16*r_rc^2 - 28*r_rc^3 + 26*r_rc^4 - 10*r_rc^5) + (x[,"n_02"] + x[,"n_54"])*log(3 - 11*r_rc + 26*r_rc^2 - 33*r_rc^3 + 20*r_rc^4 - 5*r_rc^5) + (x[,"n_12"] + x[,"n_44"])*log(6 + 43*r_rc - 109*r_rc^2 + 129*r_rc^3 - 85*r_rc^4 + 25*r_rc^5) + (x[,"n_21"] + x[,"n_35"])*log(3 + 28*r_rc - 88*r_rc^2 + 132*r_rc^3 - 106*r_rc^4 + 40*r_rc^5) + (x[,"n_24"] + x[,"n_32"])*log(27 + 39*r_rc - 84*r_rc^2 + 132*r_rc^3 - 110*r_rc^4 + 50*r_rc^5)
logL_rm2c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- 2*(-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - 2*n20 - 2*n21 - 2*n22 - n23 - 2*n24 - 2*n25 - 2*n26 - 2*n30 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - 2*n36 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n46 - 2*n50 - 2*n51 - 2*n52 - 2*n53 - 2*n54 - 2*n55 - 2*n56)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*(log(3) + 2*log(5)) + (n06 + n50)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n00 + n56)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n05 + n51)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(4 - r + 4*r^2)) + (n16 + n40)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(4 - 4*r + 5*r^2)) + (n10 + n46)*(2*log(pmax(1e-6,r)) + log(5 - 11*r + 11*r^2 - 5*r^3)) + (n01 + n55)*(2*log(pmax(1e-6,r)) + log(7 - 14*r + 11*r^2 - 4*r^3)) + (n13 + n43)*log(57 + 16*r - 58*r^2 + 84*r^3 - 42*r^4) + (n02 + n54)*(log(pmax(1e-6,r)) + log(15 - 31*r + 35*r^2 - 21*r^3 + 5*r^4)) + (n23 + n33)*log(45 + r + 6*r^2 - 14*r^3 + 7*r^4) + (n20 + n36)*(log(pmax(1e-6,r)) + log(7 - 20*r + 30*r^2 - 24*r^3 + 10*r^4)) + (n03 + n53)*log(9 - 20*r + 34*r^2 - 28*r^3 + 14*r^4) + (n11 + n45)*(log(pmax(1e-6,r)) + log(24 - 41*r + 62*r^2 - 53*r^3 + 20*r^4)) + (n24 + n32)*log(120 - 25*r + 32*r^2 - 86*r^3 + 108*r^4 - 50*r^5) + (n21 + n35)*log(21 - 20*r + 70*r^2 - 112*r^3 + 102*r^4 - 40*r^5) + (n12 + n44)*log(27 - 3*r + 82*r^2 - 121*r^3 + 88*r^4 - 25*r^5) + (n15 + n41)*log(12 - 16*r + 27*r^2 - 50*r^3 + 47*r^4 - 20*r^5) + (n26 + n30)*log(3 - 11*r + 26*r^2 - 34*r^3 + 26*r^4 - 10*r^5) + (n04 + n52)*log(3 + r - 2*r^2 - r^3 + 4*r^4 - 5*r^5) + (n14 + n42)*log(48 - 25*r - 3*r^2 + 19*r^3 - 37*r^4 + 25*r^5) + (n25 + n31)*log(21 + 8*r - 54*r^2 + 104*r^3 - 98*r^4 + 40*r^5) + (n22 + n34)*log(99 + 37*r - 78*r^2 + 154*r^3 - 142*r^4 + 50*r^5)
return(L)}
interlogL_rm2c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_rm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_rm2c <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_16"] + x[,"n_40"])*(-5*log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_46"])*(-5*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_53"])*(-3*log10(2) + log10(3) + log10(13)) - (x[,"n_26"] + x[,"n_30"])*(-4*log10(2) + log10(17)) - (x[,"n_20"] + x[,"n_36"])*(-4*log10(2) + log10(17)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(5) + log10(19)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(5) + log10(19)) - (x[,"n_23"] + x[,"n_33"])*(-4*log10(2) + log10(17) + log10(43)) - (x[,"n_14"] + x[,"n_42"])*(-5*log10(2) + log10(17) + log10(67)) - (x[,"n_12"] + x[,"n_44"])*(-5*log10(2) + log10(17) + log10(67)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(109)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(109)) - (x[,"n_25"] + x[,"n_31"])*(-3*log10(2) + log10(157)) - (x[,"n_21"] + x[,"n_35"])*(-3*log10(2) + log10(157)) - (x[,"n_13"] + x[,"n_43"])*(-3*log10(2) + log10(467)) - (x[,"n_24"] + x[,"n_32"])*(-4*log10(2) + log10(1759)) - (x[,"n_22"] + x[,"n_34"])*(-4*log10(2) + log10(1759)) + (x[,"n_06"] + x[,"n_50"])*(3*log10(pmax(1e-6,1 - r_rm2c)) + 2*log10(pmax(1e-6,r_rm2c))) + (x[,"n_00"] + x[,"n_56"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + 3*log10(pmax(1e-6,r_rm2c))) + (x[,"n_05"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c)) + log10(4 - r_rm2c + 4*r_rm2c^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c)) + log10(4 - 4*r_rm2c + 5*r_rm2c^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log10(pmax(1e-6,r_rm2c)) + log10(5 - 11*r_rm2c + 11*r_rm2c^2 - 5*r_rm2c^3)) + (x[,"n_01"] + x[,"n_55"])*(2*log10(pmax(1e-6,r_rm2c)) + log10(7 - 14*r_rm2c + 11*r_rm2c^2 - 4*r_rm2c^3)) + (x[,"n_13"] + x[,"n_43"])*log10(57 + 16*r_rm2c - 58*r_rm2c^2 + 84*r_rm2c^3 - 42*r_rm2c^4) + (x[,"n_02"] + x[,"n_54"])*(log10(pmax(1e-6,r_rm2c)) + log10(15 - 31*r_rm2c + 35*r_rm2c^2 - 21*r_rm2c^3 + 5*r_rm2c^4)) + (x[,"n_23"] + x[,"n_33"])*log10(45 + r_rm2c + 6*r_rm2c^2 - 14*r_rm2c^3 + 7*r_rm2c^4) + (x[,"n_20"] + x[,"n_36"])*(log10(pmax(1e-6,r_rm2c)) + log10(7 - 20*r_rm2c + 30*r_rm2c^2 - 24*r_rm2c^3 + 10*r_rm2c^4)) + (x[,"n_03"] + x[,"n_53"])*log10(9 - 20*r_rm2c + 34*r_rm2c^2 - 28*r_rm2c^3 + 14*r_rm2c^4) + (x[,"n_11"] + x[,"n_45"])*(log10(pmax(1e-6,r_rm2c)) + log10(24 - 41*r_rm2c + 62*r_rm2c^2 - 53*r_rm2c^3 + 20*r_rm2c^4)) + (x[,"n_24"] + x[,"n_32"])*log10(120 - 25*r_rm2c + 32*r_rm2c^2 - 86*r_rm2c^3 + 108*r_rm2c^4 - 50*r_rm2c^5) + (x[,"n_21"] + x[,"n_35"])*log10(21 - 20*r_rm2c + 70*r_rm2c^2 - 112*r_rm2c^3 + 102*r_rm2c^4 - 40*r_rm2c^5) + (x[,"n_12"] + x[,"n_44"])*log10(27 - 3*r_rm2c + 82*r_rm2c^2 - 121*r_rm2c^3 + 88*r_rm2c^4 - 25*r_rm2c^5) + (x[,"n_15"] + x[,"n_41"])*log10(12 - 16*r_rm2c + 27*r_rm2c^2 - 50*r_rm2c^3 + 47*r_rm2c^4 - 20*r_rm2c^5) + (x[,"n_26"] + x[,"n_30"])*log10(3 - 11*r_rm2c + 26*r_rm2c^2 - 34*r_rm2c^3 + 26*r_rm2c^4 - 10*r_rm2c^5) + (x[,"n_04"] + x[,"n_52"])*log10(3 + r_rm2c - 2*r_rm2c^2 - r_rm2c^3 + 4*r_rm2c^4 - 5*r_rm2c^5) + (x[,"n_14"] + x[,"n_42"])*log10(48 - 25*r_rm2c - 3*r_rm2c^2 + 19*r_rm2c^3 - 37*r_rm2c^4 + 25*r_rm2c^5) + (x[,"n_25"] + x[,"n_31"])*log10(21 + 8*r_rm2c - 54*r_rm2c^2 + 104*r_rm2c^3 - 98*r_rm2c^4 + 40*r_rm2c^5) + (x[,"n_22"] + x[,"n_34"])*log10(99 + 37*r_rm2c - 78*r_rm2c^2 + 154*r_rm2c^3 - 142*r_rm2c^4 + 50*r_rm2c^5)
logL_rm2c <- 2*(-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_46"] - 2*x[,"n_50"] - 2*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"] - 2*x[,"n_56"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*(log(3) + 2*log(5)) + (x[,"n_06"] + x[,"n_50"])*(3*log(pmax(1e-6,1 - r_rm2c)) + 2*log(pmax(1e-6,r_rm2c))) + (x[,"n_00"] + x[,"n_56"])*(2*log(pmax(1e-6,1 - r_rm2c)) + 3*log(pmax(1e-6,r_rm2c))) + (x[,"n_05"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c)) + log(4 - r_rm2c + 4*r_rm2c^2)) + (x[,"n_16"] + x[,"n_40"])*(2*log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c)) + log(4 - 4*r_rm2c + 5*r_rm2c^2)) + (x[,"n_10"] + x[,"n_46"])*(2*log(pmax(1e-6,r_rm2c)) + log(5 - 11*r_rm2c + 11*r_rm2c^2 - 5*r_rm2c^3)) + (x[,"n_01"] + x[,"n_55"])*(2*log(pmax(1e-6,r_rm2c)) + log(7 - 14*r_rm2c + 11*r_rm2c^2 - 4*r_rm2c^3)) + (x[,"n_13"] + x[,"n_43"])*log(57 + 16*r_rm2c - 58*r_rm2c^2 + 84*r_rm2c^3 - 42*r_rm2c^4) + (x[,"n_02"] + x[,"n_54"])*(log(pmax(1e-6,r_rm2c)) + log(15 - 31*r_rm2c + 35*r_rm2c^2 - 21*r_rm2c^3 + 5*r_rm2c^4)) + (x[,"n_23"] + x[,"n_33"])*log(45 + r_rm2c + 6*r_rm2c^2 - 14*r_rm2c^3 + 7*r_rm2c^4) + (x[,"n_20"] + x[,"n_36"])*(log(pmax(1e-6,r_rm2c)) + log(7 - 20*r_rm2c + 30*r_rm2c^2 - 24*r_rm2c^3 + 10*r_rm2c^4)) + (x[,"n_03"] + x[,"n_53"])*log(9 - 20*r_rm2c + 34*r_rm2c^2 - 28*r_rm2c^3 + 14*r_rm2c^4) + (x[,"n_11"] + x[,"n_45"])*(log(pmax(1e-6,r_rm2c)) + log(24 - 41*r_rm2c + 62*r_rm2c^2 - 53*r_rm2c^3 + 20*r_rm2c^4)) + (x[,"n_24"] + x[,"n_32"])*log(120 - 25*r_rm2c + 32*r_rm2c^2 - 86*r_rm2c^3 + 108*r_rm2c^4 - 50*r_rm2c^5) + (x[,"n_21"] + x[,"n_35"])*log(21 - 20*r_rm2c + 70*r_rm2c^2 - 112*r_rm2c^3 + 102*r_rm2c^4 - 40*r_rm2c^5) + (x[,"n_12"] + x[,"n_44"])*log(27 - 3*r_rm2c + 82*r_rm2c^2 - 121*r_rm2c^3 + 88*r_rm2c^4 - 25*r_rm2c^5) + (x[,"n_15"] + x[,"n_41"])*log(12 - 16*r_rm2c + 27*r_rm2c^2 - 50*r_rm2c^3 + 47*r_rm2c^4 - 20*r_rm2c^5) + (x[,"n_26"] + x[,"n_30"])*log(3 - 11*r_rm2c + 26*r_rm2c^2 - 34*r_rm2c^3 + 26*r_rm2c^4 - 10*r_rm2c^5) + (x[,"n_04"] + x[,"n_52"])*log(3 + r_rm2c - 2*r_rm2c^2 - r_rm2c^3 + 4*r_rm2c^4 - 5*r_rm2c^5) + (x[,"n_14"] + x[,"n_42"])*log(48 - 25*r_rm2c - 3*r_rm2c^2 + 19*r_rm2c^3 - 37*r_rm2c^4 + 25*r_rm2c^5) + (x[,"n_25"] + x[,"n_31"])*log(21 + 8*r_rm2c - 54*r_rm2c^2 + 104*r_rm2c^3 - 98*r_rm2c^4 + 40*r_rm2c^5) + (x[,"n_22"] + x[,"n_34"])*log(99 + 37*r_rm2c - 78*r_rm2c^2 + 154*r_rm2c^3 - 142*r_rm2c^4 + 50*r_rm2c^5)
logL_rm1c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-4*n00 - 4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 4*n11 - 4*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n16 - 3*n20 - 3*n21 - 3*n22 - 2*n23 - 3*n24 - 3*n25 - 3*n26 - 3*n30 - 3*n31 - 3*n32 - 2*n33 - 3*n34 - 3*n35 - 3*n36 - 4*n40 - 4*n41 - 4*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n46 - 4*n50 - 4*n51 - 4*n52 - 3*n53 - 4*n54 - 4*n55 - 4*n56)*log(2) + (-n00 - n01 - n02 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n54 - n55 - n56)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56)*log(5) + (n06 + n50)*(4*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n56)*(log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n01 + n55)*(3*log(pmax(1e-6,r)) + log(8 - 9*r + 4*r^2)) + (n05 + n51)*(3*log(pmax(1e-6,1 - r)) + log(3 + r + 4*r^2)) + (n10 + n46)*(3*log(pmax(1e-6,r)) + log(6 - 8*r + 5*r^2)) + (n26 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(5 - 4*r + 5*r^2)) + (n16 + n40)*(3*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 5*r^2)) + (n11 + n45)*(2*log(pmax(1e-6,r)) + log(42 - 48*r + 47*r^2 - 20*r^3)) + (n20 + n36)*(2*log(pmax(1e-6,r)) + log(6 - 12*r + 11*r^2 - 5*r^3)) + (n02 + n54)*(2*log(pmax(1e-6,r)) + log(22 - 30*r + 22*r^2 - 5*r^3)) + (n03 + n53)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n04 + n52)*(2*log(pmax(1e-6,1 - r)) + log(9 + r + 7*r^2 + 5*r^3)) + (n15 + n41)*(2*log(pmax(1e-6,1 - r)) + log(21 + 14*r - 13*r^2 + 20*r^3)) + (n23 + n33)*log(48 - 7*r - 2*r^2 + 18*r^3 - 9*r^4) + (n21 + n35)*(log(pmax(1e-6,r)) + log(32 - 42*r + 54*r^2 - 49*r^3 + 20*r^4)) + (n12 + n44)*(log(pmax(1e-6,r)) + log(90 - 78*r + 120*r^2 - 91*r^3 + 25*r^4)) + (n13 + n43)*log(27 + 2*r + 25*r^2 - 54*r^3 + 27*r^4) + (n14 + n42)*log(66 - 55*r - 14*r^2 - 6*r^3 + 34*r^4 - 25*r^5) + (n22 + n34)*log(42 + 14*r + 52*r^2 - 86*r^3 + 72*r^4 - 25*r^5) + (n25 + n31)*log(15 - 14*r + 26*r^2 - 58*r^3 + 51*r^4 - 20*r^5) + (n24 + n32)*log(69 - 23*r - 24*r^2 + 48*r^3 - 53*r^4 + 25*r^5)
return(L)}
interlogL_rm1c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_rm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_rm1c <- 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + 2*log10(3)) - (x[,"n_16"] + x[,"n_40"])*(-5*log10(2) + log10(13)) - (x[,"n_10"] + x[,"n_46"])*(-5*log10(2) + log10(13)) - (x[,"n_03"] + x[,"n_53"])*(-4*log10(2) + log10(13)) - (x[,"n_26"] + x[,"n_30"])*(-5*log10(2) + log10(17)) - (x[,"n_20"] + x[,"n_36"])*(-5*log10(2) + log10(17)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(5) + log10(19)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(5) + log10(19)) - (x[,"n_23"] + x[,"n_33"])*(-4*log10(2) + log10(17) + log10(43)) - (x[,"n_14"] + x[,"n_42"])*(-5*log10(2) + log10(17) + log10(67)) - (x[,"n_12"] + x[,"n_44"])*(-5*log10(2) + log10(17) + log10(67)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(109)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(109)) - (x[,"n_25"] + x[,"n_31"])*(-4*log10(2) + log10(157)) - (x[,"n_21"] + x[,"n_35"])*(-4*log10(2) + log10(157)) - (x[,"n_13"] + x[,"n_43"])*(-4*log10(2) + log10(467)) - (x[,"n_24"] + x[,"n_32"])*(-5*log10(2) + log10(1759)) - (x[,"n_22"] + x[,"n_34"])*(-5*log10(2) + log10(1759)) + (x[,"n_06"] + x[,"n_50"])*(4*log10(pmax(1e-6,1 - r_rm1c)) + log10(pmax(1e-6,r_rm1c))) + (x[,"n_00"] + x[,"n_56"])*(log10(pmax(1e-6,1 - r_rm1c)) + 4*log10(pmax(1e-6,r_rm1c))) + (x[,"n_01"] + x[,"n_55"])*(3*log10(pmax(1e-6,r_rm1c)) + log10(8 - 9*r_rm1c + 4*r_rm1c^2)) + (x[,"n_05"] + x[,"n_51"])*(3*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 + r_rm1c + 4*r_rm1c^2)) + (x[,"n_10"] + x[,"n_46"])*(3*log10(pmax(1e-6,r_rm1c)) + log10(6 - 8*r_rm1c + 5*r_rm1c^2)) + (x[,"n_26"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(pmax(1e-6,r_rm1c)) + log10(5 - 4*r_rm1c + 5*r_rm1c^2)) + (x[,"n_16"] + x[,"n_40"])*(3*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 - 2*r_rm1c + 5*r_rm1c^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(42 - 48*r_rm1c + 47*r_rm1c^2 - 20*r_rm1c^3)) + (x[,"n_20"] + x[,"n_36"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(6 - 12*r_rm1c + 11*r_rm1c^2 - 5*r_rm1c^3)) + (x[,"n_02"] + x[,"n_54"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(22 - 30*r_rm1c + 22*r_rm1c^2 - 5*r_rm1c^3)) + (x[,"n_03"] + x[,"n_53"])*(log10(pmax(1e-6,r_rm1c)) + log10(4 - 7*r_rm1c + 6*r_rm1c^2 - 3*r_rm1c^3)) + (x[,"n_04"] + x[,"n_52"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(9 + r_rm1c + 7*r_rm1c^2 + 5*r_rm1c^3)) + (x[,"n_15"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(21 + 14*r_rm1c - 13*r_rm1c^2 + 20*r_rm1c^3)) + (x[,"n_23"] + x[,"n_33"])*log10(48 - 7*r_rm1c - 2*r_rm1c^2 + 18*r_rm1c^3 - 9*r_rm1c^4) + (x[,"n_21"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm1c)) + log10(32 - 42*r_rm1c + 54*r_rm1c^2 - 49*r_rm1c^3 + 20*r_rm1c^4)) + (x[,"n_12"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm1c)) + log10(90 - 78*r_rm1c + 120*r_rm1c^2 - 91*r_rm1c^3 + 25*r_rm1c^4)) + (x[,"n_13"] + x[,"n_43"])*log10(27 + 2*r_rm1c + 25*r_rm1c^2 - 54*r_rm1c^3 + 27*r_rm1c^4) + (x[,"n_14"] + x[,"n_42"])*log10(66 - 55*r_rm1c - 14*r_rm1c^2 - 6*r_rm1c^3 + 34*r_rm1c^4 - 25*r_rm1c^5) + (x[,"n_22"] + x[,"n_34"])*log10(42 + 14*r_rm1c + 52*r_rm1c^2 - 86*r_rm1c^3 + 72*r_rm1c^4 - 25*r_rm1c^5) + (x[,"n_25"] + x[,"n_31"])*log10(15 - 14*r_rm1c + 26*r_rm1c^2 - 58*r_rm1c^3 + 51*r_rm1c^4 - 20*r_rm1c^5) + (x[,"n_24"] + x[,"n_32"])*log10(69 - 23*r_rm1c - 24*r_rm1c^2 + 48*r_rm1c^3 - 53*r_rm1c^4 + 25*r_rm1c^5)
logL_rm1c <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 2*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 2*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 3*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"] - 4*x[,"n_56"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"])*log(5) + (x[,"n_06"] + x[,"n_50"])*(4*log(pmax(1e-6,1 - r_rm1c)) + log(pmax(1e-6,r_rm1c))) + (x[,"n_00"] + x[,"n_56"])*(log(pmax(1e-6,1 - r_rm1c)) + 4*log(pmax(1e-6,r_rm1c))) + (x[,"n_01"] + x[,"n_55"])*(3*log(pmax(1e-6,r_rm1c)) + log(8 - 9*r_rm1c + 4*r_rm1c^2)) + (x[,"n_05"] + x[,"n_51"])*(3*log(pmax(1e-6,1 - r_rm1c)) + log(3 + r_rm1c + 4*r_rm1c^2)) + (x[,"n_10"] + x[,"n_46"])*(3*log(pmax(1e-6,r_rm1c)) + log(6 - 8*r_rm1c + 5*r_rm1c^2)) + (x[,"n_26"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(pmax(1e-6,r_rm1c)) + log(5 - 4*r_rm1c + 5*r_rm1c^2)) + (x[,"n_16"] + x[,"n_40"])*(3*log(pmax(1e-6,1 - r_rm1c)) + log(3 - 2*r_rm1c + 5*r_rm1c^2)) + (x[,"n_11"] + x[,"n_45"])*(2*log(pmax(1e-6,r_rm1c)) + log(42 - 48*r_rm1c + 47*r_rm1c^2 - 20*r_rm1c^3)) + (x[,"n_20"] + x[,"n_36"])*(2*log(pmax(1e-6,r_rm1c)) + log(6 - 12*r_rm1c + 11*r_rm1c^2 - 5*r_rm1c^3)) + (x[,"n_02"] + x[,"n_54"])*(2*log(pmax(1e-6,r_rm1c)) + log(22 - 30*r_rm1c + 22*r_rm1c^2 - 5*r_rm1c^3)) + (x[,"n_03"] + x[,"n_53"])*(log(pmax(1e-6,r_rm1c)) + log(4 - 7*r_rm1c + 6*r_rm1c^2 - 3*r_rm1c^3)) + (x[,"n_04"] + x[,"n_52"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(9 + r_rm1c + 7*r_rm1c^2 + 5*r_rm1c^3)) + (x[,"n_15"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(21 + 14*r_rm1c - 13*r_rm1c^2 + 20*r_rm1c^3)) + (x[,"n_23"] + x[,"n_33"])*log(48 - 7*r_rm1c - 2*r_rm1c^2 + 18*r_rm1c^3 - 9*r_rm1c^4) + (x[,"n_21"] + x[,"n_35"])*(log(pmax(1e-6,r_rm1c)) + log(32 - 42*r_rm1c + 54*r_rm1c^2 - 49*r_rm1c^3 + 20*r_rm1c^4)) + (x[,"n_12"] + x[,"n_44"])*(log(pmax(1e-6,r_rm1c)) + log(90 - 78*r_rm1c + 120*r_rm1c^2 - 91*r_rm1c^3 + 25*r_rm1c^4)) + (x[,"n_13"] + x[,"n_43"])*log(27 + 2*r_rm1c + 25*r_rm1c^2 - 54*r_rm1c^3 + 27*r_rm1c^4) + (x[,"n_14"] + x[,"n_42"])*log(66 - 55*r_rm1c - 14*r_rm1c^2 - 6*r_rm1c^3 + 34*r_rm1c^4 - 25*r_rm1c^5) + (x[,"n_22"] + x[,"n_34"])*log(42 + 14*r_rm1c + 52*r_rm1c^2 - 86*r_rm1c^3 + 72*r_rm1c^4 - 25*r_rm1c^5) + (x[,"n_25"] + x[,"n_31"])*log(15 - 14*r_rm1c + 26*r_rm1c^2 - 58*r_rm1c^3 + 51*r_rm1c^4 - 20*r_rm1c^5) + (x[,"n_24"] + x[,"n_32"])*log(69 - 23*r_rm1c - 24*r_rm1c^2 + 48*r_rm1c^3 - 53*r_rm1c^4 + 25*r_rm1c^5)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
L <- (-4*n00 - 4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 4*n11 - 4*n12 - 3*n13 - 4*n14 - 4*n15 - 4*n16 - 3*n20 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 3*n25 - 3*n26 - 3*n30 - 3*n31 - 3*n32 - 3*n33 - 3*n34 - 3*n35 - 3*n36 - 4*n40 - 4*n41 - 4*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n46 - 4*n50 - 4*n51 - 4*n52 - 3*n53 - 4*n54 - 4*n55 - 4*n56)*log(2) + (-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - n16 - n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - n26 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - n36 - n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - n46 - 2*n50 - 2*n51 - 2*n52 - 2*n53 - 2*n54 - 2*n55 - 2*n56)*log(5) + (n03 + n53)*log(11) + 5*(n06 + n50)*log(pmax(1e-6,1 - r)) + 5*(n00 + n56)*log(pmax(1e-6,r)) + (n16 + n40)*(4*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n53)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n26 + n30)*(3*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n20 + n36)*(2*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n55)*(log(7 - 4*r) + 4*log(pmax(1e-6,r))) + (n10 + n46)*(log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n05 + n51)*(4*log(pmax(1e-6,1 - r)) + log(3 + 4*r)) + (n04 + n52)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(13 + 5*r)) + (n02 + n54)*(3*log(pmax(1e-6,r)) + log(18 - 23*r + 5*r^2)) + (n11 + n45)*(3*log(pmax(1e-6,r)) + log(36 - 41*r + 20*r^2)) + (n25 + n31)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(26 - 13*r + 20*r^2)) + (n15 + n41)*(3*log(pmax(1e-6,1 - r)) + log(15 + r + 20*r^2)) + (n13 + n43)*(log(pmax(1e-6,r)) + log(46 - 79*r + 66*r^2 - 33*r^3)) + (n12 + n44)*(2*log(pmax(1e-6,r)) + log(90 - 126*r + 94*r^2 - 25*r^3)) + (n21 + n35)*(2*log(pmax(1e-6,r)) + log(33 - 60*r + 47*r^2 - 20*r^3)) + (n14 + n42)*(2*log(pmax(1e-6,1 - r)) + log(33 + 13*r + 19*r^2 + 25*r^3)) + (n23 + n33)*log(41 - 46*r + 68*r^2 - 44*r^3 + 22*r^4) + (n22 + n34)*(log(pmax(1e-6,r)) + log(69 - 111*r + 123*r^2 - 73*r^3 + 25*r^4)) + (n24 + n32)*log(33 - 49*r + 70*r^2 - 81*r^3 + 52*r^4 - 25*r^5)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"])
LOD_rr <- 5*(x[,"n_26"] + x[,"n_30"])*log10(2) + 5*(x[,"n_20"] + x[,"n_36"])*log10(2) + 5*(x[,"n_16"] + x[,"n_40"])*log10(2) + 5*(x[,"n_10"] + x[,"n_46"])*log10(2) + 5*(x[,"n_06"] + x[,"n_50"])*log10(2) + 4*(x[,"n_03"] + x[,"n_53"])*log10(2) + 5*(x[,"n_00"] + x[,"n_56"])*log10(2) - (x[,"n_05"] + x[,"n_51"])*(-4*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_55"])*(-4*log10(2) + log10(5)) - (x[,"n_25"] + x[,"n_31"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_21"] + x[,"n_35"])*(-4*log10(2) + 2*log10(7)) - (x[,"n_23"] + x[,"n_33"])*(-3*log10(2) + log10(13) + log10(19)) - (x[,"n_04"] + x[,"n_52"])*(-5*log10(2) + log10(31)) - (x[,"n_02"] + x[,"n_54"])*(-5*log10(2) + log10(31)) - (x[,"n_15"] + x[,"n_41"])*(-4*log10(2) + log10(41)) - (x[,"n_11"] + x[,"n_45"])*(-4*log10(2) + log10(41)) - (x[,"n_13"] + x[,"n_43"])*(-4*log10(2) + log10(151)) - (x[,"n_14"] + x[,"n_42"])*(-5*log10(2) + log10(379)) - (x[,"n_12"] + x[,"n_44"])*(-5*log10(2) + log10(379)) - (x[,"n_24"] + x[,"n_32"])*(-5*log10(2) + log10(587)) - (x[,"n_22"] + x[,"n_34"])*(-5*log10(2) + log10(587)) + 5*(x[,"n_06"] + x[,"n_50"])*log10(pmax(1e-6,1 - r_rr)) + 5*(x[,"n_00"] + x[,"n_56"])*log10(pmax(1e-6,r_rr)) + (x[,"n_16"] + x[,"n_40"])*(4*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr))) + (x[,"n_03"] + x[,"n_53"])*(2*log10(pmax(1e-6,1 - r_rr)) + 2*log10(pmax(1e-6,r_rr))) + (x[,"n_26"] + x[,"n_30"])*(3*log10(pmax(1e-6,1 - r_rr)) + 2*log10(pmax(1e-6,r_rr))) + (x[,"n_20"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_rr)) + 3*log10(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_55"])*(log10(7 - 4*r_rr) + 4*log10(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_46"])*(log10(pmax(1e-6,1 - r_rr)) + 4*log10(pmax(1e-6,r_rr))) + (x[,"n_05"] + x[,"n_51"])*(4*log10(pmax(1e-6,1 - r_rr)) + log10(3 + 4*r_rr)) + (x[,"n_04"] + x[,"n_52"])*(3*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr)) + log10(13 + 5*r_rr)) + (x[,"n_02"] + x[,"n_54"])*(3*log10(pmax(1e-6,r_rr)) + log10(18 - 23*r_rr + 5*r_rr^2)) + (x[,"n_11"] + x[,"n_45"])*(3*log10(pmax(1e-6,r_rr)) + log10(36 - 41*r_rr + 20*r_rr^2)) + (x[,"n_25"] + x[,"n_31"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr)) + log10(26 - 13*r_rr + 20*r_rr^2)) + (x[,"n_15"] + x[,"n_41"])*(3*log10(pmax(1e-6,1 - r_rr)) + log10(15 + r_rr + 20*r_rr^2)) + (x[,"n_13"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(46 - 79*r_rr + 66*r_rr^2 - 33*r_rr^3)) + (x[,"n_12"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(90 - 126*r_rr + 94*r_rr^2 - 25*r_rr^3)) + (x[,"n_21"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_rr)) + log10(33 - 60*r_rr + 47*r_rr^2 - 20*r_rr^3)) + (x[,"n_14"] + x[,"n_42"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(33 + 13*r_rr + 19*r_rr^2 + 25*r_rr^3)) + (x[,"n_23"] + x[,"n_33"])*log10(41 - 46*r_rr + 68*r_rr^2 - 44*r_rr^3 + 22*r_rr^4) + (x[,"n_22"] + x[,"n_34"])*(log10(pmax(1e-6,r_rr)) + log10(69 - 111*r_rr + 123*r_rr^2 - 73*r_rr^3 + 25*r_rr^4)) + (x[,"n_24"] + x[,"n_32"])*log10(33 - 49*r_rr + 70*r_rr^2 - 81*r_rr^3 + 52*r_rr^4 - 25*r_rr^5)
logL_rr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_12"] - 3*x[,"n_13"] - 4*x[,"n_14"] - 4*x[,"n_15"] - 4*x[,"n_16"] - 3*x[,"n_20"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 4*x[,"n_50"] - 4*x[,"n_51"] - 4*x[,"n_52"] - 3*x[,"n_53"] - 4*x[,"n_54"] - 4*x[,"n_55"] - 4*x[,"n_56"])*log(2) + (-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - x[,"n_46"] - 2*x[,"n_50"] - 2*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"] - 2*x[,"n_56"])*log(5) + (x[,"n_03"] + x[,"n_53"])*log(11) + 5*(x[,"n_06"] + x[,"n_50"])*log(pmax(1e-6,1 - r_rr)) + 5*(x[,"n_00"] + x[,"n_56"])*log(pmax(1e-6,r_rr)) + (x[,"n_16"] + x[,"n_40"])*(4*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr))) + (x[,"n_03"] + x[,"n_53"])*(2*log(pmax(1e-6,1 - r_rr)) + 2*log(pmax(1e-6,r_rr))) + (x[,"n_26"] + x[,"n_30"])*(3*log(pmax(1e-6,1 - r_rr)) + 2*log(pmax(1e-6,r_rr))) + (x[,"n_20"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_rr)) + 3*log(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_55"])*(log(7 - 4*r_rr) + 4*log(pmax(1e-6,r_rr))) + (x[,"n_10"] + x[,"n_46"])*(log(pmax(1e-6,1 - r_rr)) + 4*log(pmax(1e-6,r_rr))) + (x[,"n_05"] + x[,"n_51"])*(4*log(pmax(1e-6,1 - r_rr)) + log(3 + 4*r_rr)) + (x[,"n_04"] + x[,"n_52"])*(3*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr)) + log(13 + 5*r_rr)) + (x[,"n_02"] + x[,"n_54"])*(3*log(pmax(1e-6,r_rr)) + log(18 - 23*r_rr + 5*r_rr^2)) + (x[,"n_11"] + x[,"n_45"])*(3*log(pmax(1e-6,r_rr)) + log(36 - 41*r_rr + 20*r_rr^2)) + (x[,"n_25"] + x[,"n_31"])*(2*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr)) + log(26 - 13*r_rr + 20*r_rr^2)) + (x[,"n_15"] + x[,"n_41"])*(3*log(pmax(1e-6,1 - r_rr)) + log(15 + r_rr + 20*r_rr^2)) + (x[,"n_13"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(46 - 79*r_rr + 66*r_rr^2 - 33*r_rr^3)) + (x[,"n_12"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(90 - 126*r_rr + 94*r_rr^2 - 25*r_rr^3)) + (x[,"n_21"] + x[,"n_35"])*(2*log(pmax(1e-6,r_rr)) + log(33 - 60*r_rr + 47*r_rr^2 - 20*r_rr^3)) + (x[,"n_14"] + x[,"n_42"])*(2*log(pmax(1e-6,1 - r_rr)) + log(33 + 13*r_rr + 19*r_rr^2 + 25*r_rr^3)) + (x[,"n_23"] + x[,"n_33"])*log(41 - 46*r_rr + 68*r_rr^2 - 44*r_rr^3 + 22*r_rr^4) + (x[,"n_22"] + x[,"n_34"])*(log(pmax(1e-6,r_rr)) + log(69 - 111*r_rr + 123*r_rr^2 - 73*r_rr^3 + 25*r_rr^4)) + (x[,"n_24"] + x[,"n_32"])*log(33 - 49*r_rr + 70*r_rr^2 - 81*r_rr^3 + 52*r_rr^4 - 25*r_rr^5)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_cr,r_mc,r_mm2c,r_mm1c,r_mr,r_rc,r_rm2c,r_rm1c,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_cr,LOD_mc,LOD_mm2c,LOD_mm1c,LOD_mr,LOD_rc,LOD_rm2c,LOD_rm1c,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_cr,logL_mc,logL_mm2c,logL_mm1c,logL_mr,logL_rc,logL_rm2c,logL_rm1c,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","coupling repulsion","mixed coupling","mixed mixed2c","mixed mixed1c","mixed repulsion","repulsion coupling","repulsion mixed2c","repulsion mixed1c","repulsion repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_2.4_2.4 <- function(x){
logL_cc <- function(r,n12,n13,n14,n21,n22,n23,n24,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n52,n53,n54,n11,n15,n33,n51,n55) {
L <- (2*n12 + n13 + 2*n14 + 2*n21 + n22 + 3*n23 + n24 + 2*n25 + n31 + 3*n32 + 3*n34 + n35 + 2*n41 + n42 + 3*n43 + n44 + 2*n45 + 2*n52 + n53 + 2*n54)*log(2) + (n13 + n31 + n35 + n53)*log(3) + 2*(-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + 4*(n11 + n55)*log(pmax(1e-6,1 - r)) + 4*(n15 + n51)*log(pmax(1e-6,r)) + (n12 + n21 + n45 + n54)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n13 + n31 + n35 + n53)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n14 + n25 + n41 + n52)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n24 + n42)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n22 + n44)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n23 + n32 + n34 + n43)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + n33*log(11 - 32*r + 68*r^2 - 72*r^3 + 36*r^4)
return(L)}
interlogL_cc <- function(n12,n13,n14,n21,n22,n23,n24,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n52,n53,n54,n11,n15,n33,n51,n55) {
optimize(logL_cc,c(0,0.5), n12,n13,n14,n21,n22,n23,n24,n25,n31,n32,n34,n35,n41,n42,n43,n44,n45,n52,n53,n54,n11,n15,n33,n51,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_11"],x[,"n_15"],x[,"n_33"],x[,"n_51"],x[,"n_55"])
LOD_cc <- 4*(x[,"n_15"] + x[,"n_51"])*log10(2) + 4*(x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*log10(2) + 4*(x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(2) + 4*(x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log10(2) + 4*(x[,"n_11"] + x[,"n_55"])*log10(2) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(3)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(3)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(5)) - x[,"n_33"]*(-2*log10(2) + log10(3) + log10(7)) + 4*(x[,"n_11"] + x[,"n_55"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_15"] + x[,"n_51"])*log10(pmax(1e-6,r_cc)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_24"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(7 - 12*r_cc + 8*r_cc^2)) + (x[,"n_22"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 - 4*r_cc + 8*r_cc^2)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log10(pmax(1e-6,r_cc)) + log10(2 - 5*r_cc + 6*r_cc^2 - 3*r_cc^3)) + x[,"n_33"]*log10(11 - 32*r_cc + 68*r_cc^2 - 72*r_cc^3 + 36*r_cc^4)
logL_cc <- (2*x[,"n_12"] + x[,"n_13"] + 2*x[,"n_14"] + 2*x[,"n_21"] + x[,"n_22"] + 3*x[,"n_23"] + x[,"n_24"] + 2*x[,"n_25"] + x[,"n_31"] + 3*x[,"n_32"] + 3*x[,"n_34"] + x[,"n_35"] + 2*x[,"n_41"] + x[,"n_42"] + 3*x[,"n_43"] + x[,"n_44"] + 2*x[,"n_45"] + 2*x[,"n_52"] + x[,"n_53"] + 2*x[,"n_54"])*log(2) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(3) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + 4*(x[,"n_11"] + x[,"n_55"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_15"] + x[,"n_51"])*log(pmax(1e-6,r_cc)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_24"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(7 - 12*r_cc + 8*r_cc^2)) + (x[,"n_22"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 - 4*r_cc + 8*r_cc^2)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log(pmax(1e-6,r_cc)) + log(2 - 5*r_cc + 6*r_cc^2 - 3*r_cc^3)) + x[,"n_33"]*log(11 - 32*r_cc + 68*r_cc^2 - 72*r_cc^3 + 36*r_cc^4)
logL_cm3c <- function(r,n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n35,n41,n42,n44,n45,n51,n52,n53,n54,n55,n23,n32,n34,n43,n33) {
L <- (-2*n11 - 2*n12 - n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - 2*n24 - 2*n25 - n31 - n35 - 2*n41 - 2*n42 - 2*n44 - 2*n45 - 2*n51 - 2*n52 - n53 - 2*n54 - 2*n55)*log(2) + (n23 + n32 + n34 + n43)*log(3) + 2*(-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n15 + n51)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n11 + n55)*(2*log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n12 + n21 + n45 + n54)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r^2)) + (n14 + n25 + n41 + n52)*(2*log(pmax(1e-6,r)) + log(7 - 8*r + 4*r^2)) + (n24 + n42)*(log(pmax(1e-6,r)) + log(18 - 25*r + 32*r^2 - 16*r^3)) + (n13 + n31 + n35 + n53)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n22 + n44)*log(9 - 25*r^2 + 32*r^3 - 16*r^4) + n33*log(5 + 2*r - 11*r^2 + 18*r^3 - 9*r^4) + (n23 + n32 + n34 + n43)*log(1 - r + 3*r^2 - 4*r^3 + 2*r^4)
return(L)}
interlogL_cm3c <- function(n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n35,n41,n42,n44,n45,n51,n52,n53,n54,n55,n23,n32,n34,n43,n33) {
optimize(logL_cm3c,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n35,n41,n42,n44,n45,n51,n52,n53,n54,n55,n23,n32,n34,n43,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm3c <- mapply(interlogL_cm3c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_23"],x[,"n_32"],x[,"n_34"],x[,"n_43"],x[,"n_33"])
LOD_cm3c <- -((x[,"n_15"] + x[,"n_51"])*(-4*log10(2) + log10(3))) - (x[,"n_11"] + x[,"n_55"])*(-4*log10(2) + log10(3)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(3*log10(2) - log10(7)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(13)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(23)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(23)) - x[,"n_33"]*(-4*log10(2) + log10(79)) + (x[,"n_15"] + x[,"n_51"])*(log10(2 - r_cm3c) + 3*log10(pmax(1e-6,r_cm3c))) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_cm3c)) + log10(1 - r_cm3c^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_cm3c)) + log10(3 + 4*r_cm3c^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_cm3c)) + log10(7 - 8*r_cm3c + 4*r_cm3c^2)) + (x[,"n_24"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm3c)) + log10(18 - 25*r_cm3c + 32*r_cm3c^2 - 16*r_cm3c^3)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log10(pmax(1e-6,r_cm3c)) + log10(4 - 7*r_cm3c + 6*r_cm3c^2 - 3*r_cm3c^3)) + (x[,"n_22"] + x[,"n_44"])*log10(9 - 25*r_cm3c^2 + 32*r_cm3c^3 - 16*r_cm3c^4) + x[,"n_33"]*log10(5 + 2*r_cm3c - 11*r_cm3c^2 + 18*r_cm3c^3 - 9*r_cm3c^4) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(1 - r_cm3c + 3*r_cm3c^2 - 4*r_cm3c^3 + 2*r_cm3c^4)
logL_cm3c <- (-2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_31"] - x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"])*log(2) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(3) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_15"] + x[,"n_51"])*(log(2 - r_cm3c) + 3*log(pmax(1e-6,r_cm3c))) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_cm3c)) + log(1 - r_cm3c^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_cm3c)) + log(3 + 4*r_cm3c^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(2*log(pmax(1e-6,r_cm3c)) + log(7 - 8*r_cm3c + 4*r_cm3c^2)) + (x[,"n_24"] + x[,"n_42"])*(log(pmax(1e-6,r_cm3c)) + log(18 - 25*r_cm3c + 32*r_cm3c^2 - 16*r_cm3c^3)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log(pmax(1e-6,r_cm3c)) + log(4 - 7*r_cm3c + 6*r_cm3c^2 - 3*r_cm3c^3)) + (x[,"n_22"] + x[,"n_44"])*log(9 - 25*r_cm3c^2 + 32*r_cm3c^3 - 16*r_cm3c^4) + x[,"n_33"]*log(5 + 2*r_cm3c - 11*r_cm3c^2 + 18*r_cm3c^3 - 9*r_cm3c^4) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(1 - r_cm3c + 3*r_cm3c^2 - 4*r_cm3c^3 + 2*r_cm3c^4)
logL_cm2c <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-n11 - n12 - n13 - n14 - n15 - n21 - n22 + n23 - n24 - n25 - n31 + n32 + n33 + n34 - n35 - n41 - n42 + n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(2) + (-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n11 + n55)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 + r)) + (n15 + n51)*(2*log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n14 + n25 + n41 + n52)*(log(pmax(1e-6,r)) + log(6 - 11*r + 12*r^2 - 4*r^3)) + (n23 + n32 + n34 + n43)*log(3 + 5*r - 11*r^2 + 12*r^3 - 6*r^4) + (n12 + n21 + n45 + n54)*log(3 - 4*r + r^2 + 4*r^3 - 4*r^4) + (n13 + n31 + n35 + n53)*log(3 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + n33*log(9 - 8*r + 17*r^2 - 18*r^3 + 9*r^4) + (n24 + n42)*log(9 - 12*r + 39*r^2 - 40*r^3 + 16*r^4) + (n22 + n44)*log(12 - 10*r + 15*r^2 - 24*r^3 + 16*r^4)
return(L)}
interlogL_cm2c <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_cm2c,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_cm2c <- -((x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-log10(2) + log10(3))) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_51"])*(-4*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_55"])*(-4*log10(2) + log10(5)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(5) + log10(7)) - x[,"n_33"]*(-4*log10(2) + 2*log10(11)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-3*log10(2) + log10(19)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-3*log10(2) + log10(31)) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c)) + log10(2 + r_cm2c)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(3 - 4*r_cm2c + r_cm2c^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log10(pmax(1e-6,r_cm2c)) + log10(6 - 11*r_cm2c + 12*r_cm2c^2 - 4*r_cm2c^3)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(3 + 5*r_cm2c - 11*r_cm2c^2 + 12*r_cm2c^3 - 6*r_cm2c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log10(3 - 4*r_cm2c + r_cm2c^2 + 4*r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(3 - 4*r_cm2c + 10*r_cm2c^2 - 12*r_cm2c^3 + 6*r_cm2c^4) + x[,"n_33"]*log10(9 - 8*r_cm2c + 17*r_cm2c^2 - 18*r_cm2c^3 + 9*r_cm2c^4) + (x[,"n_24"] + x[,"n_42"])*log10(9 - 12*r_cm2c + 39*r_cm2c^2 - 40*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_22"] + x[,"n_44"])*log10(12 - 10*r_cm2c + 15*r_cm2c^2 - 24*r_cm2c^3 + 16*r_cm2c^4)
logL_cm2c <- (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] + x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] + x[,"n_32"] + x[,"n_33"] + x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] + x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(2) + (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c)) + log(2 + r_cm2c)) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,r_cm2c)) + log(3 - 4*r_cm2c + r_cm2c^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log(pmax(1e-6,r_cm2c)) + log(6 - 11*r_cm2c + 12*r_cm2c^2 - 4*r_cm2c^3)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(3 + 5*r_cm2c - 11*r_cm2c^2 + 12*r_cm2c^3 - 6*r_cm2c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log(3 - 4*r_cm2c + r_cm2c^2 + 4*r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(3 - 4*r_cm2c + 10*r_cm2c^2 - 12*r_cm2c^3 + 6*r_cm2c^4) + x[,"n_33"]*log(9 - 8*r_cm2c + 17*r_cm2c^2 - 18*r_cm2c^3 + 9*r_cm2c^4) + (x[,"n_24"] + x[,"n_42"])*log(9 - 12*r_cm2c + 39*r_cm2c^2 - 40*r_cm2c^3 + 16*r_cm2c^4) + (x[,"n_22"] + x[,"n_44"])*log(12 - 10*r_cm2c + 15*r_cm2c^2 - 24*r_cm2c^3 + 16*r_cm2c^4)
logL_mc <- function(r,n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n35,n41,n42,n44,n45,n51,n52,n53,n54,n55,n23,n32,n34,n43,n33) {
L <- (-2*n11 - 2*n12 - n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - 2*n24 - 2*n25 - n31 - n35 - 2*n41 - 2*n42 - 2*n44 - 2*n45 - 2*n51 - 2*n52 - n53 - 2*n54 - 2*n55)*log(2) + (n23 + n32 + n34 + n43)*log(3) + 2*(-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n15 + n51)*(log(2 - r) + 3*log(pmax(1e-6,r))) + (n11 + n55)*(2*log(pmax(1e-6,1 - r)) + log(1 - r^2)) + (n12 + n21 + n45 + n54)*(2*log(pmax(1e-6,1 - r)) + log(3 + 4*r^2)) + (n14 + n25 + n41 + n52)*(2*log(pmax(1e-6,r)) + log(7 - 8*r + 4*r^2)) + (n24 + n42)*(log(pmax(1e-6,r)) + log(18 - 25*r + 32*r^2 - 16*r^3)) + (n13 + n31 + n35 + n53)*(log(pmax(1e-6,r)) + log(4 - 7*r + 6*r^2 - 3*r^3)) + (n22 + n44)*log(9 - 25*r^2 + 32*r^3 - 16*r^4) + n33*log(5 + 2*r - 11*r^2 + 18*r^3 - 9*r^4) + (n23 + n32 + n34 + n43)*log(1 - r + 3*r^2 - 4*r^3 + 2*r^4)
return(L)}
interlogL_mc <- function(n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n35,n41,n42,n44,n45,n51,n52,n53,n54,n55,n23,n32,n34,n43,n33) {
optimize(logL_mc,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n24,n25,n31,n35,n41,n42,n44,n45,n51,n52,n53,n54,n55,n23,n32,n34,n43,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_23"],x[,"n_32"],x[,"n_34"],x[,"n_43"],x[,"n_33"])
LOD_mc <- -((x[,"n_15"] + x[,"n_51"])*(-4*log10(2) + log10(3))) - (x[,"n_11"] + x[,"n_55"])*(-4*log10(2) + log10(3)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(3*log10(2) - log10(7)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(13)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(23)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(23)) - x[,"n_33"]*(-4*log10(2) + log10(79)) + (x[,"n_15"] + x[,"n_51"])*(log10(2 - r_mc) + 3*log10(pmax(1e-6,r_mc))) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(1 - r_mc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 + 4*r_mc^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_mc)) + log10(7 - 8*r_mc + 4*r_mc^2)) + (x[,"n_24"] + x[,"n_42"])*(log10(pmax(1e-6,r_mc)) + log10(18 - 25*r_mc + 32*r_mc^2 - 16*r_mc^3)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log10(pmax(1e-6,r_mc)) + log10(4 - 7*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_22"] + x[,"n_44"])*log10(9 - 25*r_mc^2 + 32*r_mc^3 - 16*r_mc^4) + x[,"n_33"]*log10(5 + 2*r_mc - 11*r_mc^2 + 18*r_mc^3 - 9*r_mc^4) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(1 - r_mc + 3*r_mc^2 - 4*r_mc^3 + 2*r_mc^4)
logL_mc <- (-2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_31"] - x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"])*log(2) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(3) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_15"] + x[,"n_51"])*(log(2 - r_mc) + 3*log(pmax(1e-6,r_mc))) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_mc)) + log(1 - r_mc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 + 4*r_mc^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(2*log(pmax(1e-6,r_mc)) + log(7 - 8*r_mc + 4*r_mc^2)) + (x[,"n_24"] + x[,"n_42"])*(log(pmax(1e-6,r_mc)) + log(18 - 25*r_mc + 32*r_mc^2 - 16*r_mc^3)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log(pmax(1e-6,r_mc)) + log(4 - 7*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_22"] + x[,"n_44"])*log(9 - 25*r_mc^2 + 32*r_mc^3 - 16*r_mc^4) + x[,"n_33"]*log(5 + 2*r_mc - 11*r_mc^2 + 18*r_mc^3 - 9*r_mc^4) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(1 - r_mc + 3*r_mc^2 - 4*r_mc^3 + 2*r_mc^4)
logL_mm3c <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-4*n11 - 3*n12 - 4*n13 - 3*n14 - 4*n15 - 3*n21 - 3*n22 - 3*n23 - 3*n24 - 3*n25 - 4*n31 - 3*n32 - 3*n33 - 3*n34 - 4*n35 - 3*n41 - 3*n42 - 3*n43 - 3*n44 - 3*n45 - 4*n51 - 3*n52 - 4*n53 - 3*n54 - 4*n55)*log(2) + 2*(-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n15 + n51)*(2*log(2 - r) + 2*log(pmax(1e-6,r))) + 2*(n11 + n55)*log(1 - r^2) + (n14 + n25 + n41 + n52)*(log(2 - r) + log(pmax(1e-6,r)) + log(3 - 2*r + 2*r^2)) + (n12 + n21 + n45 + n54)*(log(1 - r^2) + log(3 - 2*r + 2*r^2)) + (n23 + n32 + n34 + n43)*log(21 + 4*r - 16*r^2 + 24*r^3 - 12*r^4) + (n13 + n31 + n35 + n53)*log(9 - 8*r + 14*r^2 - 12*r^3 + 6*r^4) + (n24 + n42)*log(9 + 18*r^2 - 20*r^3 + 8*r^4) + (n22 + n44)*log(15 - 8*r + 6*r^2 - 12*r^3 + 8*r^4) + n33*log(37 + 18*r^2 - 36*r^3 + 18*r^4)
return(L)}
interlogL_mm3c <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_mm3c,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm3c <- mapply(interlogL_mm3c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mm3c <- 2*(x[,"n_11"] + x[,"n_55"])*(2*log10(2) - log10(3)) - (x[,"n_15"] + x[,"n_51"])*(-2*log10(2) + 2*(-log10(2) + log10(3))) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-3*log10(2) + log10(3) + log10(5)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + log10(5) + log10(17)) - (x[,"n_24"] + x[,"n_42"])*(-log10(2) + log10(23)) - (x[,"n_22"] + x[,"n_44"])*(-log10(2) + log10(23)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-3*log10(2) + log10(59)) - x[,"n_33"]*(-3*log10(2) + log10(5) + log10(61)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(2 - r_mm3c) + 2*log10(pmax(1e-6,r_mm3c))) + 2*(x[,"n_11"] + x[,"n_55"])*log10(1 - r_mm3c^2) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log10(2 - r_mm3c) + log10(pmax(1e-6,r_mm3c)) + log10(3 - 2*r_mm3c + 2*r_mm3c^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(log10(1 - r_mm3c^2) + log10(3 - 2*r_mm3c + 2*r_mm3c^2)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(21 + 4*r_mm3c - 16*r_mm3c^2 + 24*r_mm3c^3 - 12*r_mm3c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(9 - 8*r_mm3c + 14*r_mm3c^2 - 12*r_mm3c^3 + 6*r_mm3c^4) + (x[,"n_24"] + x[,"n_42"])*log10(9 + 18*r_mm3c^2 - 20*r_mm3c^3 + 8*r_mm3c^4) + (x[,"n_22"] + x[,"n_44"])*log10(15 - 8*r_mm3c + 6*r_mm3c^2 - 12*r_mm3c^3 + 8*r_mm3c^4) + x[,"n_33"]*log10(37 + 18*r_mm3c^2 - 36*r_mm3c^3 + 18*r_mm3c^4)
logL_mm3c <- (-4*x[,"n_11"] - 3*x[,"n_12"] - 4*x[,"n_13"] - 3*x[,"n_14"] - 4*x[,"n_15"] - 3*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 3*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 3*x[,"n_34"] - 4*x[,"n_35"] - 3*x[,"n_41"] - 3*x[,"n_42"] - 3*x[,"n_43"] - 3*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_51"] - 3*x[,"n_52"] - 4*x[,"n_53"] - 3*x[,"n_54"] - 4*x[,"n_55"])*log(2) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_15"] + x[,"n_51"])*(2*log(2 - r_mm3c) + 2*log(pmax(1e-6,r_mm3c))) + 2*(x[,"n_11"] + x[,"n_55"])*log(1 - r_mm3c^2) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log(2 - r_mm3c) + log(pmax(1e-6,r_mm3c)) + log(3 - 2*r_mm3c + 2*r_mm3c^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(log(1 - r_mm3c^2) + log(3 - 2*r_mm3c + 2*r_mm3c^2)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(21 + 4*r_mm3c - 16*r_mm3c^2 + 24*r_mm3c^3 - 12*r_mm3c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(9 - 8*r_mm3c + 14*r_mm3c^2 - 12*r_mm3c^3 + 6*r_mm3c^4) + (x[,"n_24"] + x[,"n_42"])*log(9 + 18*r_mm3c^2 - 20*r_mm3c^3 + 8*r_mm3c^4) + (x[,"n_22"] + x[,"n_44"])*log(15 - 8*r_mm3c + 6*r_mm3c^2 - 12*r_mm3c^3 + 8*r_mm3c^4) + x[,"n_33"]*log(37 + 18*r_mm3c^2 - 36*r_mm3c^3 + 18*r_mm3c^4)
logL_mm2c <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n21 - 2*n22 - 2*n23 - 2*n24 - 3*n25 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - 3*n41 - 2*n42 - 2*n43 - 2*n44 - 3*n45 - 3*n51 - 3*n52 - 2*n53 - 3*n54 - 3*n55)*log(2) + (-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n11 + n55)*(log(pmax(1e-6,r)) + log(2 + r) + log(1 - r^2)) + (n15 + n51)*(log(2 - r) + log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + n33*log(27 + 10*r - 19*r^2 + 18*r^3 - 9*r^4) + (n22 + n44)*log(15 + 4*r - r^2 + 8*r^3 - 8*r^4) + (n24 + n42)*log(18 + 6*r - 25*r^2 + 24*r^3 - 8*r^4) + (n13 + n31 + n35 + n53)*log(6 - 2*r - r^2 + 6*r^3 - 3*r^4) + (n12 + n21 + n45 + n54)*log(3 + 8*r - 6*r^2 + 4*r^4) + (n14 + n25 + n41 + n52)*log(9 - 12*r + 18*r^2 - 16*r^3 + 4*r^4) + (n23 + n32 + n34 + n43)*log(33 - 8*r + 20*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_mm2c <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_mm2c,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm2c <- mapply(interlogL_mm2c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_mm2c <- -((x[,"n_15"] + x[,"n_51"])*(-4*log10(2) + log10(3) + log10(5))) - (x[,"n_11"] + x[,"n_55"])*(-4*log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(5) + log10(17)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-2*log10(2) + log10(23)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-2*log10(2) + log10(23)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + log10(127)) - x[,"n_33"]*(-4*log10(2) + log10(463)) + (x[,"n_11"] + x[,"n_55"])*(log10(pmax(1e-6,r_mm2c)) + log10(2 + r_mm2c) + log10(1 - r_mm2c^2)) + (x[,"n_15"] + x[,"n_51"])*(log10(2 - r_mm2c) + log10(pmax(1e-6,r_mm2c)) + log10(3 - 4*r_mm2c + r_mm2c^2)) + x[,"n_33"]*log10(27 + 10*r_mm2c - 19*r_mm2c^2 + 18*r_mm2c^3 - 9*r_mm2c^4) + (x[,"n_22"] + x[,"n_44"])*log10(15 + 4*r_mm2c - r_mm2c^2 + 8*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_24"] + x[,"n_42"])*log10(18 + 6*r_mm2c - 25*r_mm2c^2 + 24*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(6 - 2*r_mm2c - r_mm2c^2 + 6*r_mm2c^3 - 3*r_mm2c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log10(3 + 8*r_mm2c - 6*r_mm2c^2 + 4*r_mm2c^4) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*log10(9 - 12*r_mm2c + 18*r_mm2c^2 - 16*r_mm2c^3 + 4*r_mm2c^4) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(33 - 8*r_mm2c + 20*r_mm2c^2 - 24*r_mm2c^3 + 12*r_mm2c^4)
logL_mm2c <- (-3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 3*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 3*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 2*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_11"] + x[,"n_55"])*(log(pmax(1e-6,r_mm2c)) + log(2 + r_mm2c) + log(1 - r_mm2c^2)) + (x[,"n_15"] + x[,"n_51"])*(log(2 - r_mm2c) + log(pmax(1e-6,r_mm2c)) + log(3 - 4*r_mm2c + r_mm2c^2)) + x[,"n_33"]*log(27 + 10*r_mm2c - 19*r_mm2c^2 + 18*r_mm2c^3 - 9*r_mm2c^4) + (x[,"n_22"] + x[,"n_44"])*log(15 + 4*r_mm2c - r_mm2c^2 + 8*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_24"] + x[,"n_42"])*log(18 + 6*r_mm2c - 25*r_mm2c^2 + 24*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(6 - 2*r_mm2c - r_mm2c^2 + 6*r_mm2c^3 - 3*r_mm2c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log(3 + 8*r_mm2c - 6*r_mm2c^2 + 4*r_mm2c^4) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*log(9 - 12*r_mm2c + 18*r_mm2c^2 - 16*r_mm2c^3 + 4*r_mm2c^4) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(33 - 8*r_mm2c + 20*r_mm2c^2 - 24*r_mm2c^3 + 12*r_mm2c^4)
logL_rc <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-n11 - n12 - n13 - n14 - n15 - n21 - n22 + n23 - n24 - n25 - n31 + n32 + n33 + n34 - n35 - n41 - n42 + n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(2) + (-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n11 + n55)*(log(pmax(1e-6,r)) + 2*log(1 + r) + log(2 + r)) + (n15 + n51)*(2*log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + (n14 + n25 + n41 + n52)*(log(pmax(1e-6,r)) + log(6 - 11*r + 12*r^2 - 4*r^3)) + (n23 + n32 + n34 + n43)*log(3 + 5*r - 11*r^2 + 12*r^3 - 6*r^4) + (n12 + n21 + n45 + n54)*log(3 - 4*r + r^2 + 4*r^3 - 4*r^4) + (n13 + n31 + n35 + n53)*log(3 - 4*r + 10*r^2 - 12*r^3 + 6*r^4) + n33*log(9 - 8*r + 17*r^2 - 18*r^3 + 9*r^4) + (n24 + n42)*log(9 - 12*r + 39*r^2 - 40*r^3 + 16*r^4) + (n22 + n44)*log(12 - 10*r + 15*r^2 - 24*r^3 + 16*r^4)
return(L)}
interlogL_rc <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_rc,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rc <- -((x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-log10(2) + log10(3))) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_51"])*(-4*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_55"])*(-2*log10(2) + 2*(-log10(2) + log10(3)) + log10(5)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(5) + log10(7)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(5) + log10(7)) - x[,"n_33"]*(-4*log10(2) + 2*log10(11)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-3*log10(2) + log10(19)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-3*log10(2) + log10(31)) + (x[,"n_11"] + x[,"n_55"])*(log10(pmax(1e-6,r_rc)) + 2*log10(1 + r_rc) + log10(2 + r_rc)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_rc)) + log10(3 - 4*r_rc + r_rc^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log10(pmax(1e-6,r_rc)) + log10(6 - 11*r_rc + 12*r_rc^2 - 4*r_rc^3)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(3 + 5*r_rc - 11*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log10(3 - 4*r_rc + r_rc^2 + 4*r_rc^3 - 4*r_rc^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(3 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + x[,"n_33"]*log10(9 - 8*r_rc + 17*r_rc^2 - 18*r_rc^3 + 9*r_rc^4) + (x[,"n_24"] + x[,"n_42"])*log10(9 - 12*r_rc + 39*r_rc^2 - 40*r_rc^3 + 16*r_rc^4) + (x[,"n_22"] + x[,"n_44"])*log10(12 - 10*r_rc + 15*r_rc^2 - 24*r_rc^3 + 16*r_rc^4)
logL_rc <- (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] + x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] + x[,"n_32"] + x[,"n_33"] + x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] + x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(2) + (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_11"] + x[,"n_55"])*(log(pmax(1e-6,r_rc)) + 2*log(1 + r_rc) + log(2 + r_rc)) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,r_rc)) + log(3 - 4*r_rc + r_rc^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log(pmax(1e-6,r_rc)) + log(6 - 11*r_rc + 12*r_rc^2 - 4*r_rc^3)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(3 + 5*r_rc - 11*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log(3 - 4*r_rc + r_rc^2 + 4*r_rc^3 - 4*r_rc^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(3 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4) + x[,"n_33"]*log(9 - 8*r_rc + 17*r_rc^2 - 18*r_rc^3 + 9*r_rc^4) + (x[,"n_24"] + x[,"n_42"])*log(9 - 12*r_rc + 39*r_rc^2 - 40*r_rc^3 + 16*r_rc^4) + (x[,"n_22"] + x[,"n_44"])*log(12 - 10*r_rc + 15*r_rc^2 - 24*r_rc^3 + 16*r_rc^4)
logL_rm3c <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n21 - 2*n22 - 2*n23 - 2*n24 - 3*n25 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - 3*n41 - 2*n42 - 2*n43 - 2*n44 - 3*n45 - 3*n51 - 3*n52 - 2*n53 - 3*n54 - 3*n55)*log(2) + (-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*(log(3) + 2*log(5)) + (n11 + n55)*(log(pmax(1e-6,r)) + log(2 + r) + log(1 - r^2)) + (n15 + n51)*(log(2 - r) + log(pmax(1e-6,r)) + log(3 - 4*r + r^2)) + n33*log(27 + 10*r - 19*r^2 + 18*r^3 - 9*r^4) + (n22 + n44)*log(15 + 4*r - r^2 + 8*r^3 - 8*r^4) + (n24 + n42)*log(18 + 6*r - 25*r^2 + 24*r^3 - 8*r^4) + (n13 + n31 + n35 + n53)*log(6 - 2*r - r^2 + 6*r^3 - 3*r^4) + (n12 + n21 + n45 + n54)*log(3 + 8*r - 6*r^2 + 4*r^4) + (n14 + n25 + n41 + n52)*log(9 - 12*r + 18*r^2 - 16*r^3 + 4*r^4) + (n23 + n32 + n34 + n43)*log(33 - 8*r + 20*r^2 - 24*r^3 + 12*r^4)
return(L)}
interlogL_rm3c <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_rm3c,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm3c <- mapply(interlogL_rm3c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rm3c <- -((x[,"n_15"] + x[,"n_51"])*(-4*log10(2) + log10(3) + log10(5))) - (x[,"n_11"] + x[,"n_55"])*(-4*log10(2) + log10(3) + log10(5)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(5) + log10(17)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-2*log10(2) + log10(23)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-2*log10(2) + log10(23)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(23)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + log10(127)) - x[,"n_33"]*(-4*log10(2) + log10(463)) + (x[,"n_11"] + x[,"n_55"])*(log10(pmax(1e-6,r_rm3c)) + log10(2 + r_rm3c) + log10(1 - r_rm3c^2)) + (x[,"n_15"] + x[,"n_51"])*(log10(2 - r_rm3c) + log10(pmax(1e-6,r_rm3c)) + log10(3 - 4*r_rm3c + r_rm3c^2)) + x[,"n_33"]*log10(27 + 10*r_rm3c - 19*r_rm3c^2 + 18*r_rm3c^3 - 9*r_rm3c^4) + (x[,"n_22"] + x[,"n_44"])*log10(15 + 4*r_rm3c - r_rm3c^2 + 8*r_rm3c^3 - 8*r_rm3c^4) + (x[,"n_24"] + x[,"n_42"])*log10(18 + 6*r_rm3c - 25*r_rm3c^2 + 24*r_rm3c^3 - 8*r_rm3c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(6 - 2*r_rm3c - r_rm3c^2 + 6*r_rm3c^3 - 3*r_rm3c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log10(3 + 8*r_rm3c - 6*r_rm3c^2 + 4*r_rm3c^4) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*log10(9 - 12*r_rm3c + 18*r_rm3c^2 - 16*r_rm3c^3 + 4*r_rm3c^4) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(33 - 8*r_rm3c + 20*r_rm3c^2 - 24*r_rm3c^3 + 12*r_rm3c^4)
logL_rm3c <- (-3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 3*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 3*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 2*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"])*log(2) + (-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*(log(3) + 2*log(5)) + (x[,"n_11"] + x[,"n_55"])*(log(pmax(1e-6,r_rm3c)) + log(2 + r_rm3c) + log(1 - r_rm3c^2)) + (x[,"n_15"] + x[,"n_51"])*(log(2 - r_rm3c) + log(pmax(1e-6,r_rm3c)) + log(3 - 4*r_rm3c + r_rm3c^2)) + x[,"n_33"]*log(27 + 10*r_rm3c - 19*r_rm3c^2 + 18*r_rm3c^3 - 9*r_rm3c^4) + (x[,"n_22"] + x[,"n_44"])*log(15 + 4*r_rm3c - r_rm3c^2 + 8*r_rm3c^3 - 8*r_rm3c^4) + (x[,"n_24"] + x[,"n_42"])*log(18 + 6*r_rm3c - 25*r_rm3c^2 + 24*r_rm3c^3 - 8*r_rm3c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(6 - 2*r_rm3c - r_rm3c^2 + 6*r_rm3c^3 - 3*r_rm3c^4) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log(3 + 8*r_rm3c - 6*r_rm3c^2 + 4*r_rm3c^4) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*log(9 - 12*r_rm3c + 18*r_rm3c^2 - 16*r_rm3c^3 + 4*r_rm3c^4) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(33 - 8*r_rm3c + 20*r_rm3c^2 - 24*r_rm3c^3 + 12*r_rm3c^4)
logL_rm2c <- function(r,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
L <- (-2*n11 - n12 - 2*n13 - n14 - 2*n15 - n21 - n22 - n23 - n24 - n25 - 2*n31 - n32 - n33 - n34 - 2*n35 - n41 - n42 - n43 - n44 - n45 - 2*n51 - n52 - 2*n53 - n54 - 2*n55)*log(2) + (-2*n11 - 2*n12 - n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - n23 - 2*n24 - 2*n25 - n31 - n32 - n33 - n34 - n35 - 2*n41 - 2*n42 - n43 - 2*n44 - 2*n45 - 2*n51 - 2*n52 - n53 - 2*n54 - 2*n55)*log(3) + 2*(-n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55)*log(5) + (n11 + n55)*(2*log(pmax(1e-6,r)) + 2*log(2 + r)) + (n12 + n21 + n45 + n54)*(log(pmax(1e-6,r)) + log(2 + r) + log(3 + 2*r - 2*r^2)) + 2*(n15 + n51)*log(3 - 4*r + r^2) + (n14 + n25 + n41 + n52)*(log(3 + 2*r - 2*r^2) + log(3 - 4*r + r^2)) + (n23 + n32 + n34 + n43)*log(15 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n13 + n31 + n35 + n53)*log(3 + 8*r - 6*r^2 - 4*r^3 + 2*r^4) + n33*log(33 - 16*r + 22*r^2 - 12*r^3 + 6*r^4) + (n24 + n42)*log(45 - 48*r + 32*r^2 - 28*r^3 + 8*r^4) + (n22 + n44)*log(9 + 36*r - 4*r^2 - 4*r^3 + 8*r^4)
return(L)}
interlogL_rm2c <- function(n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55) {
optimize(logL_rm2c,c(0,0.5), n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"])
LOD_rm2c <- -2*(x[,"n_15"] + x[,"n_51"])*(-2*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_55"])*(-2*log10(2) + 2*(-log10(2) + log10(5))) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-2*log10(2) + 2*log10(3) + log10(7)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-3*log10(2) + log10(5) + log10(7)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-3*log10(2) + log10(5) + log10(7)) - (x[,"n_24"] + x[,"n_42"])*(log10(2) + log10(13)) - (x[,"n_22"] + x[,"n_44"])*(log10(2) + log10(13)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-3*log10(2) + log10(41)) - x[,"n_33"]*(-3*log10(2) + log10(5) + log10(47)) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,r_rm2c)) + 2*log10(2 + r_rm2c)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(log10(pmax(1e-6,r_rm2c)) + log10(2 + r_rm2c) + log10(3 + 2*r_rm2c - 2*r_rm2c^2)) + 2*(x[,"n_15"] + x[,"n_51"])*log10(3 - 4*r_rm2c + r_rm2c^2) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log10(3 + 2*r_rm2c - 2*r_rm2c^2) + log10(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(15 + 4*r_rm2c - 8*r_rm2c^2 + 8*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(3 + 8*r_rm2c - 6*r_rm2c^2 - 4*r_rm2c^3 + 2*r_rm2c^4) + x[,"n_33"]*log10(33 - 16*r_rm2c + 22*r_rm2c^2 - 12*r_rm2c^3 + 6*r_rm2c^4) + (x[,"n_24"] + x[,"n_42"])*log10(45 - 48*r_rm2c + 32*r_rm2c^2 - 28*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_22"] + x[,"n_44"])*log10(9 + 36*r_rm2c - 4*r_rm2c^2 - 4*r_rm2c^3 + 8*r_rm2c^4)
logL_rm2c <- (-2*x[,"n_11"] - x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - 2*x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - 2*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - 2*x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - 2*x[,"n_51"] - x[,"n_52"] - 2*x[,"n_53"] - x[,"n_54"] - 2*x[,"n_55"])*log(2) + (-2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"])*log(3) + 2*(-x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"])*log(5) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,r_rm2c)) + 2*log(2 + r_rm2c)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(log(pmax(1e-6,r_rm2c)) + log(2 + r_rm2c) + log(3 + 2*r_rm2c - 2*r_rm2c^2)) + 2*(x[,"n_15"] + x[,"n_51"])*log(3 - 4*r_rm2c + r_rm2c^2) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log(3 + 2*r_rm2c - 2*r_rm2c^2) + log(3 - 4*r_rm2c + r_rm2c^2)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(15 + 4*r_rm2c - 8*r_rm2c^2 + 8*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(3 + 8*r_rm2c - 6*r_rm2c^2 - 4*r_rm2c^3 + 2*r_rm2c^4) + x[,"n_33"]*log(33 - 16*r_rm2c + 22*r_rm2c^2 - 12*r_rm2c^3 + 6*r_rm2c^4) + (x[,"n_24"] + x[,"n_42"])*log(45 - 48*r_rm2c + 32*r_rm2c^2 - 28*r_rm2c^3 + 8*r_rm2c^4) + (x[,"n_22"] + x[,"n_44"])*log(9 + 36*r_rm2c - 4*r_rm2c^2 - 4*r_rm2c^3 + 8*r_rm2c^4)

return(list(
r_mat = cbind(r_cc,r_cm3c,r_cm2c,r_mc,r_mm3c,r_mm2c,r_rc,r_rm3c,r_rm2c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm3c,LOD_cm2c,LOD_mc,LOD_mm3c,LOD_mm2c,LOD_rc,LOD_rm3c,LOD_rm2c,0),
logL_mat = cbind(logL_cc,logL_cm3c,logL_cm2c,logL_mc,logL_mm3c,logL_mm2c,logL_rc,logL_rm3c,logL_rm2c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed3c","coupling mixed2c","mixed coupling","mixed mixed3c","mixed mixed2c","repulsion coupling","repulsion mixed3c","repulsion mixed2c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_1.3 <- function(x){
logL_c <- function(r,n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34) {
L <- (2*n02 + n10 + n11 + 4*n12 + n13 + n14 + n20 + n21 + 4*n22 + n23 + n24 + 2*n32)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(4*log(2) + 2*log(5)) + (n01 + n33)*log(9 - 8*r) + (n11 + n23)*log(19 - 8*r) + (n00 + n34)*log(pmax(1e-6,1 - r)) + (n10 + n24)*log(2 - r) + (n04 + n30)*log(pmax(1e-6,r)) + (n14 + n20)*log(1 + r) + (n03 + n31)*log(1 + 8*r) + (n13 + n21)*log(11 + 8*r)
return(L)}
interlogL_c <- function(n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34) {
optimize(logL_c,c(0,0.5), n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"])
LOD_c <- (x[,"n_04"] + x[,"n_30"])*log10(2) + (x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_14"] + x[,"n_20"])*(-log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_24"])*(-log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_31"])*log10(5) - (x[,"n_01"] + x[,"n_33"])*log10(5) - (x[,"n_13"] + x[,"n_21"])*(log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_23"])*(log10(3) + log10(5)) + (x[,"n_01"] + x[,"n_33"])*log10(9 - 8*r_c) + (x[,"n_11"] + x[,"n_23"])*log10(19 - 8*r_c) + (x[,"n_00"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_10"] + x[,"n_24"])*log10(2 - r_c) + (x[,"n_04"] + x[,"n_30"])*log10(pmax(1e-6,r_c)) + (x[,"n_14"] + x[,"n_20"])*log10(1 + r_c) + (x[,"n_03"] + x[,"n_31"])*log10(1 + 8*r_c) + (x[,"n_13"] + x[,"n_21"])*log10(11 + 8*r_c)
logL_c <- (2*x[,"n_02"] + x[,"n_10"] + x[,"n_11"] + 4*x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + 4*x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + 2*x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(4*log(2) + 2*log(5)) + (x[,"n_01"] + x[,"n_33"])*log(9 - 8*r_c) + (x[,"n_11"] + x[,"n_23"])*log(19 - 8*r_c) + (x[,"n_00"] + x[,"n_34"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_10"] + x[,"n_24"])*log(2 - r_c) + (x[,"n_04"] + x[,"n_30"])*log(pmax(1e-6,r_c)) + (x[,"n_14"] + x[,"n_20"])*log(1 + r_c) + (x[,"n_03"] + x[,"n_31"])*log(1 + 8*r_c) + (x[,"n_13"] + x[,"n_21"])*log(11 + 8*r_c)
logL_r <- function(r,n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34) {
L <- (2*n02 + n10 + n11 + 4*n12 + n13 + n14 + n20 + n21 + 4*n22 + n23 + n24 + 2*n32)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(4*log(2) + 2*log(5)) + (n03 + n31)*log(9 - 8*r) + (n13 + n21)*log(19 - 8*r) + (n04 + n30)*log(pmax(1e-6,1 - r)) + (n14 + n20)*log(2 - r) + (n00 + n34)*log(pmax(1e-6,r)) + (n10 + n24)*log(1 + r) + (n01 + n33)*log(1 + 8*r) + (n11 + n23)*log(11 + 8*r)
return(L)}
interlogL_r <- function(n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34) {
optimize(logL_r,c(0,0.5), n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"])
LOD_r <- (x[,"n_04"] + x[,"n_30"])*log10(2) + (x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_14"] + x[,"n_20"])*(-log10(2) + log10(3)) - (x[,"n_10"] + x[,"n_24"])*(-log10(2) + log10(3)) - (x[,"n_03"] + x[,"n_31"])*log10(5) - (x[,"n_01"] + x[,"n_33"])*log10(5) - (x[,"n_13"] + x[,"n_21"])*(log10(3) + log10(5)) - (x[,"n_11"] + x[,"n_23"])*(log10(3) + log10(5)) + (x[,"n_03"] + x[,"n_31"])*log10(9 - 8*r_r) + (x[,"n_13"] + x[,"n_21"])*log10(19 - 8*r_r) + (x[,"n_04"] + x[,"n_30"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_14"] + x[,"n_20"])*log10(2 - r_r) + (x[,"n_00"] + x[,"n_34"])*log10(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_24"])*log10(1 + r_r) + (x[,"n_01"] + x[,"n_33"])*log10(1 + 8*r_r) + (x[,"n_11"] + x[,"n_23"])*log10(11 + 8*r_r)
logL_r <- (2*x[,"n_02"] + x[,"n_10"] + x[,"n_11"] + 4*x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + 4*x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + 2*x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(4*log(2) + 2*log(5)) + (x[,"n_03"] + x[,"n_31"])*log(9 - 8*r_r) + (x[,"n_13"] + x[,"n_21"])*log(19 - 8*r_r) + (x[,"n_04"] + x[,"n_30"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_14"] + x[,"n_20"])*log(2 - r_r) + (x[,"n_00"] + x[,"n_34"])*log(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_24"])*log(1 + r_r) + (x[,"n_01"] + x[,"n_33"])*log(1 + 8*r_r) + (x[,"n_11"] + x[,"n_23"])*log(11 + 8*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_1.4 <- function(x){
logL_c <- function(r,n11,n12,n13,n14,n21,n22,n23,n24,n01,n02,n03,n04,n31,n32,n33,n34) {
L <- (n11 + n12 + n13 + n14 + n21 + n22 + n23 + n24)*log(3) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(2*log(2) + 2*log(5)) + (n02 + n33)*log(3 - 2*r) + (n12 + n23)*log(7 - 2*r) + (n01 + n34)*log(pmax(1e-6,1 - r)) + (n11 + n24)*log(2 - r) + (n04 + n31)*log(pmax(1e-6,r)) + (n14 + n21)*log(1 + r) + (n03 + n32)*log(1 + 2*r) + (n13 + n22)*log(5 + 2*r)
return(L)}
interlogL_c <- function(n11,n12,n13,n14,n21,n22,n23,n24,n01,n02,n03,n04,n31,n32,n33,n34) {
optimize(logL_c,c(0,0.5), n11,n12,n13,n14,n21,n22,n23,n24,n01,n02,n03,n04,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_c <- (x[,"n_04"] + x[,"n_31"])*log10(2) - (x[,"n_03"] + x[,"n_32"])*log10(2) - (x[,"n_02"] + x[,"n_33"])*log10(2) + (x[,"n_01"] + x[,"n_34"])*log10(2) - (x[,"n_14"] + x[,"n_21"])*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_24"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_22"])*(log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_23"])*(log10(2) + log10(3)) + (x[,"n_02"] + x[,"n_33"])*log10(3 - 2*r_c) + (x[,"n_12"] + x[,"n_23"])*log10(7 - 2*r_c) + (x[,"n_01"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_11"] + x[,"n_24"])*log10(2 - r_c) + (x[,"n_04"] + x[,"n_31"])*log10(pmax(1e-6,r_c)) + (x[,"n_14"] + x[,"n_21"])*log10(1 + r_c) + (x[,"n_03"] + x[,"n_32"])*log10(1 + 2*r_c) + (x[,"n_13"] + x[,"n_22"])*log10(5 + 2*r_c)
logL_c <- (x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(2*log(2) + 2*log(5)) + (x[,"n_02"] + x[,"n_33"])*log(3 - 2*r_c) + (x[,"n_12"] + x[,"n_23"])*log(7 - 2*r_c) + (x[,"n_01"] + x[,"n_34"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_11"] + x[,"n_24"])*log(2 - r_c) + (x[,"n_04"] + x[,"n_31"])*log(pmax(1e-6,r_c)) + (x[,"n_14"] + x[,"n_21"])*log(1 + r_c) + (x[,"n_03"] + x[,"n_32"])*log(1 + 2*r_c) + (x[,"n_13"] + x[,"n_22"])*log(5 + 2*r_c)
logL_r <- function(r,n11,n12,n13,n14,n21,n22,n23,n24,n01,n02,n03,n04,n31,n32,n33,n34) {
L <- (n11 + n12 + n13 + n14 + n21 + n22 + n23 + n24)*log(3) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(2*log(2) + 2*log(5)) + (n03 + n32)*log(3 - 2*r) + (n13 + n22)*log(7 - 2*r) + (n04 + n31)*log(pmax(1e-6,1 - r)) + (n14 + n21)*log(2 - r) + (n01 + n34)*log(pmax(1e-6,r)) + (n11 + n24)*log(1 + r) + (n02 + n33)*log(1 + 2*r) + (n12 + n23)*log(5 + 2*r)
return(L)}
interlogL_r <- function(n11,n12,n13,n14,n21,n22,n23,n24,n01,n02,n03,n04,n31,n32,n33,n34) {
optimize(logL_r,c(0,0.5), n11,n12,n13,n14,n21,n22,n23,n24,n01,n02,n03,n04,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_r <- (x[,"n_04"] + x[,"n_31"])*log10(2) - (x[,"n_03"] + x[,"n_32"])*log10(2) - (x[,"n_02"] + x[,"n_33"])*log10(2) + (x[,"n_01"] + x[,"n_34"])*log10(2) - (x[,"n_14"] + x[,"n_21"])*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_24"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_22"])*(log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_23"])*(log10(2) + log10(3)) + (x[,"n_03"] + x[,"n_32"])*log10(3 - 2*r_r) + (x[,"n_13"] + x[,"n_22"])*log10(7 - 2*r_r) + (x[,"n_04"] + x[,"n_31"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_14"] + x[,"n_21"])*log10(2 - r_r) + (x[,"n_01"] + x[,"n_34"])*log10(pmax(1e-6,r_r)) + (x[,"n_11"] + x[,"n_24"])*log10(1 + r_r) + (x[,"n_02"] + x[,"n_33"])*log10(1 + 2*r_r) + (x[,"n_12"] + x[,"n_23"])*log10(5 + 2*r_r)
logL_r <- (x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"])*log(3) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(2*log(2) + 2*log(5)) + (x[,"n_03"] + x[,"n_32"])*log(3 - 2*r_r) + (x[,"n_13"] + x[,"n_22"])*log(7 - 2*r_r) + (x[,"n_04"] + x[,"n_31"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_14"] + x[,"n_21"])*log(2 - r_r) + (x[,"n_01"] + x[,"n_34"])*log(pmax(1e-6,r_r)) + (x[,"n_11"] + x[,"n_24"])*log(1 + r_r) + (x[,"n_02"] + x[,"n_33"])*log(1 + 2*r_r) + (x[,"n_12"] + x[,"n_23"])*log(5 + 2*r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_1.5 <- function(x){
logL_c <- function(r,n12,n13,n14,n22,n23,n24,n02,n03,n04,n32,n33,n34) {
L <- (n12 + 2*n13 + n14 + n22 + 2*n23 + n24)*log(3) + (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*(3*log(2) + log(5)) + (n02 + n34)*log(pmax(1e-6,1 - r)) + (n12 + n24)*log(2 - r) + (n04 + n32)*log(pmax(1e-6,r)) + (n14 + n22)*log(1 + r)
return(L)}
interlogL_c <- function(n12,n13,n14,n22,n23,n24,n02,n03,n04,n32,n33,n34) {
optimize(logL_c,c(0,0.5), n12,n13,n14,n22,n23,n24,n02,n03,n04,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_c <- (x[,"n_04"] + x[,"n_32"])*log10(2) + (x[,"n_02"] + x[,"n_34"])*log10(2) - (x[,"n_14"] + x[,"n_22"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_24"])*(-log10(2) + log10(3)) + (x[,"n_02"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_12"] + x[,"n_24"])*log10(2 - r_c) + (x[,"n_04"] + x[,"n_32"])*log10(pmax(1e-6,r_c)) + (x[,"n_14"] + x[,"n_22"])*log10(1 + r_c)
logL_c <- (x[,"n_12"] + 2*x[,"n_13"] + x[,"n_14"] + x[,"n_22"] + 2*x[,"n_23"] + x[,"n_24"])*log(3) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + (x[,"n_02"] + x[,"n_34"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_12"] + x[,"n_24"])*log(2 - r_c) + (x[,"n_04"] + x[,"n_32"])*log(pmax(1e-6,r_c)) + (x[,"n_14"] + x[,"n_22"])*log(1 + r_c)
logL_r <- function(r,n12,n13,n14,n22,n23,n24,n02,n03,n04,n32,n33,n34) {
L <- (n12 + 2*n13 + n14 + n22 + 2*n23 + n24)*log(3) + (-n02 - n03 - n04 - n12 - n13 - n14 - n22 - n23 - n24 - n32 - n33 - n34)*(3*log(2) + log(5)) + (n04 + n32)*log(pmax(1e-6,1 - r)) + (n14 + n22)*log(2 - r) + (n02 + n34)*log(pmax(1e-6,r)) + (n12 + n24)*log(1 + r)
return(L)}
interlogL_r <- function(n12,n13,n14,n22,n23,n24,n02,n03,n04,n32,n33,n34) {
optimize(logL_r,c(0,0.5), n12,n13,n14,n22,n23,n24,n02,n03,n04,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_r <- (x[,"n_04"] + x[,"n_32"])*log10(2) + (x[,"n_02"] + x[,"n_34"])*log10(2) - (x[,"n_14"] + x[,"n_22"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_24"])*(-log10(2) + log10(3)) + (x[,"n_04"] + x[,"n_32"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_14"] + x[,"n_22"])*log10(2 - r_r) + (x[,"n_02"] + x[,"n_34"])*log10(pmax(1e-6,r_r)) + (x[,"n_12"] + x[,"n_24"])*log10(1 + r_r)
logL_r <- (x[,"n_12"] + 2*x[,"n_13"] + x[,"n_14"] + x[,"n_22"] + 2*x[,"n_23"] + x[,"n_24"])*log(3) + (-x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + (x[,"n_04"] + x[,"n_32"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_14"] + x[,"n_22"])*log(2 - r_r) + (x[,"n_02"] + x[,"n_34"])*log(pmax(1e-6,r_r)) + (x[,"n_12"] + x[,"n_24"])*log(1 + r_r)

return(list(
r_mat = cbind(r_c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_r,0),
logL_mat = cbind(logL_c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_2.2 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(2*log(2) + 2*log(5)) + 2*(n00 + n34)*log(pmax(1e-6,1 - r)) + 2*(n04 + n30)*log(pmax(1e-6,r)) + (n14 + n20)*(log(4 - r) + log(pmax(1e-6,r))) + (n03 + n31)*(log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n32)*log(1 + 4*r - 4*r^2) + (n11 + n23)*log(15 - 8*r - r^2) + (n10 + n24)*log(3 - 2*r - r^2) + (n13 + n21)*log(6 + 10*r - r^2) + (n01 + n33)*log(3 - 4*r + r^2) + (n12 + n22)*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_c <- 2*(x[,"n_04"] + x[,"n_30"])*log10(2) - (x[,"n_02"] + x[,"n_32"])*log10(2) + 2*(x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_03"] + x[,"n_31"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_33"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_22"])*(2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_20"])*(-2*log10(2) + log10(7)) - (x[,"n_10"] + x[,"n_24"])*(-2*log10(2) + log10(7)) - (x[,"n_13"] + x[,"n_21"])*(-2*log10(2) + log10(43)) - (x[,"n_11"] + x[,"n_23"])*(-2*log10(2) + log10(43)) + 2*(x[,"n_00"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_04"] + x[,"n_30"])*log10(pmax(1e-6,r_c)) + (x[,"n_14"] + x[,"n_20"])*(log10(4 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_03"] + x[,"n_31"])*(log10(pmax(1e-6,r_c)) + log10(2 + r_c)) + (x[,"n_02"] + x[,"n_32"])*log10(1 + 4*r_c - 4*r_c^2) + (x[,"n_11"] + x[,"n_23"])*log10(15 - 8*r_c - r_c^2) + (x[,"n_10"] + x[,"n_24"])*log10(3 - 2*r_c - r_c^2) + (x[,"n_13"] + x[,"n_21"])*log10(6 + 10*r_c - r_c^2) + (x[,"n_01"] + x[,"n_33"])*log10(3 - 4*r_c + r_c^2) + (x[,"n_12"] + x[,"n_22"])*log10(21 - 4*r_c + 4*r_c^2)
logL_c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(2*log(2) + 2*log(5)) + 2*(x[,"n_00"] + x[,"n_34"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_04"] + x[,"n_30"])*log(pmax(1e-6,r_c)) + (x[,"n_14"] + x[,"n_20"])*(log(4 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_03"] + x[,"n_31"])*(log(pmax(1e-6,r_c)) + log(2 + r_c)) + (x[,"n_02"] + x[,"n_32"])*log(1 + 4*r_c - 4*r_c^2) + (x[,"n_11"] + x[,"n_23"])*log(15 - 8*r_c - r_c^2) + (x[,"n_10"] + x[,"n_24"])*log(3 - 2*r_c - r_c^2) + (x[,"n_13"] + x[,"n_21"])*log(6 + 10*r_c - r_c^2) + (x[,"n_01"] + x[,"n_33"])*log(3 - 4*r_c + r_c^2) + (x[,"n_12"] + x[,"n_22"])*log(21 - 4*r_c + 4*r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-n00 - 2*n01 - 2*n02 - 2*n03 - n04 - n10 - 2*n11 - 2*n12 - 2*n13 - n14 - n20 - 2*n21 - 2*n22 - 2*n23 - n24 - n30 - 2*n31 - 2*n32 - 2*n33 - n34)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(log(3) + 2*log(5)) + (n00 + n04 + n30 + n34)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n12 + n22)*log(57 + 8*r - 8*r^2) + (n01 + n03 + n31 + n33)*log(3 + 2*r - 2*r^2) + (n10 + n14 + n20 + n24)*log(3 - r + r^2) + (n11 + n13 + n21 + n23)*log(33 - 2*r + 2*r^2) + (n02 + n32)*log(9 - 8*r + 8*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_m <- 2*(x[,"n_00"] + x[,"n_04"] + x[,"n_30"] + x[,"n_34"])*log10(2) - (x[,"n_02"] + x[,"n_32"])*log10(7) - (x[,"n_01"] + x[,"n_03"] + x[,"n_31"] + x[,"n_33"])*(-log10(2) + log10(7)) - (x[,"n_10"] + x[,"n_14"] + x[,"n_20"] + x[,"n_24"])*(-2*log10(2) + log10(11)) - (x[,"n_11"] + x[,"n_13"] + x[,"n_21"] + x[,"n_23"])*(-log10(2) + log10(5) + log10(13)) - (x[,"n_12"] + x[,"n_22"])*log10(59) + (x[,"n_00"] + x[,"n_04"] + x[,"n_30"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_m)) + log10(pmax(1e-6,r_m))) + (x[,"n_12"] + x[,"n_22"])*log10(57 + 8*r_m - 8*r_m^2) + (x[,"n_01"] + x[,"n_03"] + x[,"n_31"] + x[,"n_33"])*log10(3 + 2*r_m - 2*r_m^2) + (x[,"n_10"] + x[,"n_14"] + x[,"n_20"] + x[,"n_24"])*log10(3 - r_m + r_m^2) + (x[,"n_11"] + x[,"n_13"] + x[,"n_21"] + x[,"n_23"])*log10(33 - 2*r_m + 2*r_m^2) + (x[,"n_02"] + x[,"n_32"])*log10(9 - 8*r_m + 8*r_m^2)
logL_m <- (-x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_04"] + x[,"n_30"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_m)) + log(pmax(1e-6,r_m))) + (x[,"n_12"] + x[,"n_22"])*log(57 + 8*r_m - 8*r_m^2) + (x[,"n_01"] + x[,"n_03"] + x[,"n_31"] + x[,"n_33"])*log(3 + 2*r_m - 2*r_m^2) + (x[,"n_10"] + x[,"n_14"] + x[,"n_20"] + x[,"n_24"])*log(3 - r_m + r_m^2) + (x[,"n_11"] + x[,"n_13"] + x[,"n_21"] + x[,"n_23"])*log(33 - 2*r_m + 2*r_m^2) + (x[,"n_02"] + x[,"n_32"])*log(9 - 8*r_m + 8*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
L <- (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(2*log(2) + 2*log(5)) + 2*(n04 + n30)*log(pmax(1e-6,1 - r)) + 2*(n00 + n34)*log(pmax(1e-6,r)) + (n10 + n24)*(log(4 - r) + log(pmax(1e-6,r))) + (n01 + n33)*(log(pmax(1e-6,r)) + log(2 + r)) + (n02 + n32)*log(1 + 4*r - 4*r^2) + (n13 + n21)*log(15 - 8*r - r^2) + (n14 + n20)*log(3 - 2*r - r^2) + (n11 + n23)*log(6 + 10*r - r^2) + (n03 + n31)*log(3 - 4*r + r^2) + (n12 + n22)*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n30,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_r <- 2*(x[,"n_04"] + x[,"n_30"])*log10(2) - (x[,"n_02"] + x[,"n_32"])*log10(2) + 2*(x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_03"] + x[,"n_31"])*(-2*log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_33"])*(-2*log10(2) + log10(5)) - (x[,"n_12"] + x[,"n_22"])*(2*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_20"])*(-2*log10(2) + log10(7)) - (x[,"n_10"] + x[,"n_24"])*(-2*log10(2) + log10(7)) - (x[,"n_13"] + x[,"n_21"])*(-2*log10(2) + log10(43)) - (x[,"n_11"] + x[,"n_23"])*(-2*log10(2) + log10(43)) + 2*(x[,"n_04"] + x[,"n_30"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_34"])*log10(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_24"])*(log10(4 - r_r) + log10(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_33"])*(log10(pmax(1e-6,r_r)) + log10(2 + r_r)) + (x[,"n_02"] + x[,"n_32"])*log10(1 + 4*r_r - 4*r_r^2) + (x[,"n_13"] + x[,"n_21"])*log10(15 - 8*r_r - r_r^2) + (x[,"n_14"] + x[,"n_20"])*log10(3 - 2*r_r - r_r^2) + (x[,"n_11"] + x[,"n_23"])*log10(6 + 10*r_r - r_r^2) + (x[,"n_03"] + x[,"n_31"])*log10(3 - 4*r_r + r_r^2) + (x[,"n_12"] + x[,"n_22"])*log10(21 - 4*r_r + 4*r_r^2)
logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(2*log(2) + 2*log(5)) + 2*(x[,"n_04"] + x[,"n_30"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_34"])*log(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_24"])*(log(4 - r_r) + log(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_33"])*(log(pmax(1e-6,r_r)) + log(2 + r_r)) + (x[,"n_02"] + x[,"n_32"])*log(1 + 4*r_r - 4*r_r^2) + (x[,"n_13"] + x[,"n_21"])*log(15 - 8*r_r - r_r^2) + (x[,"n_14"] + x[,"n_20"])*log(3 - 2*r_r - r_r^2) + (x[,"n_11"] + x[,"n_23"])*log(6 + 10*r_r - r_r^2) + (x[,"n_03"] + x[,"n_31"])*log(3 - 4*r_r + r_r^2) + (x[,"n_12"] + x[,"n_22"])*log(21 - 4*r_r + 4*r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_2.3 <- function(x){
logL_c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(4*log(2) + 2*log(5)) + 2*(n00 + n35)*log(pmax(1e-6,1 - r)) + 2*(n05 + n30)*log(pmax(1e-6,r)) + (n15 + n20)*(log(4 - r) + log(pmax(1e-6,r))) + (n04 + n31)*(log(pmax(1e-6,r)) + log(2 + 7*r)) + (n02 + n33)*log(9 - 8*r^2) + (n03 + n32)*log(1 + 16*r - 8*r^2) + (n11 + n24)*log(33 - 20*r - 7*r^2) + (n14 + n21)*log(6 + 34*r - 7*r^2) + (n10 + n25)*log(3 - 2*r - r^2) + (n01 + n34)*log(9 - 16*r + 7*r^2) + (n12 + n23)*log(81 - 32*r + 8*r^2) + (n13 + n22)*log(57 + 16*r + 8*r^2)
return(L)}
interlogL_c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_c <- 2*(x[,"n_05"] + x[,"n_30"])*log10(2) + 2*(x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_03"] + x[,"n_32"])*log10(7) - (x[,"n_02"] + x[,"n_33"])*log10(7) - (x[,"n_15"] + x[,"n_20"])*(-2*log10(2) + log10(7)) - (x[,"n_10"] + x[,"n_25"])*(-2*log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_31"])*(-2*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_34"])*(-2*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_21"])*(-2*log10(2) + log10(5) + log10(17)) - (x[,"n_11"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(17)) - (x[,"n_13"] + x[,"n_22"])*log10(67) - (x[,"n_12"] + x[,"n_23"])*log10(67) + 2*(x[,"n_00"] + x[,"n_35"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_30"])*log10(pmax(1e-6,r_c)) + (x[,"n_15"] + x[,"n_20"])*(log10(4 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_31"])*(log10(pmax(1e-6,r_c)) + log10(2 + 7*r_c)) + (x[,"n_02"] + x[,"n_33"])*log10(9 - 8*r_c^2) + (x[,"n_03"] + x[,"n_32"])*log10(1 + 16*r_c - 8*r_c^2) + (x[,"n_11"] + x[,"n_24"])*log10(33 - 20*r_c - 7*r_c^2) + (x[,"n_14"] + x[,"n_21"])*log10(6 + 34*r_c - 7*r_c^2) + (x[,"n_10"] + x[,"n_25"])*log10(3 - 2*r_c - r_c^2) + (x[,"n_01"] + x[,"n_34"])*log10(9 - 16*r_c + 7*r_c^2) + (x[,"n_12"] + x[,"n_23"])*log10(81 - 32*r_c + 8*r_c^2) + (x[,"n_13"] + x[,"n_22"])*log10(57 + 16*r_c + 8*r_c^2)
logL_c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(4*log(2) + 2*log(5)) + 2*(x[,"n_00"] + x[,"n_35"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_30"])*log(pmax(1e-6,r_c)) + (x[,"n_15"] + x[,"n_20"])*(log(4 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_31"])*(log(pmax(1e-6,r_c)) + log(2 + 7*r_c)) + (x[,"n_02"] + x[,"n_33"])*log(9 - 8*r_c^2) + (x[,"n_03"] + x[,"n_32"])*log(1 + 16*r_c - 8*r_c^2) + (x[,"n_11"] + x[,"n_24"])*log(33 - 20*r_c - 7*r_c^2) + (x[,"n_14"] + x[,"n_21"])*log(6 + 34*r_c - 7*r_c^2) + (x[,"n_10"] + x[,"n_25"])*log(3 - 2*r_c - r_c^2) + (x[,"n_01"] + x[,"n_34"])*log(9 - 16*r_c + 7*r_c^2) + (x[,"n_12"] + x[,"n_23"])*log(81 - 32*r_c + 8*r_c^2) + (x[,"n_13"] + x[,"n_22"])*log(57 + 16*r_c + 8*r_c^2)
logL_m <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n12,n13,n22,n23) {
L <- (-3*n00 - 4*n01 - 4*n02 - 4*n03 - 4*n04 - 3*n05 - 3*n10 - 4*n11 - 4*n14 - 3*n15 - 3*n20 - 4*n21 - 4*n24 - 3*n25 - 3*n30 - 4*n31 - 4*n32 - 4*n33 - 4*n34 - 3*n35)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(log(3) + 2*log(5)) + (n00 + n05 + n30 + n35)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n04 + n31 + n34)*log(3 + 14*r - 14*r^2) + (n12 + n13 + n22 + n23)*log(13/80 + r/75 - r^2/75) + (n10 + n15 + n20 + n25)*log(3 - r + r^2) + (n11 + n14 + n21 + n24)*log(69 - 14*r + 14*r^2) + (n02 + n03 + n32 + n33)*log(27 - 16*r + 16*r^2)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n04,n05,n10,n11,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n12,n13,n22,n23) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n12,n13,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_12"],x[,"n_13"],x[,"n_22"],x[,"n_23"])
LOD_m <- 2*(x[,"n_00"] + x[,"n_05"] + x[,"n_30"] + x[,"n_35"])*log10(2) - (x[,"n_10"] + x[,"n_15"] + x[,"n_20"] + x[,"n_25"])*(-2*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*(-log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_03"] + x[,"n_32"] + x[,"n_33"])*log10(23) - (x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*(-log10(2) + log10(131)) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*(4*log10(2) + log10(3) + 2*log10(5) - log10(199)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_30"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_m)) + log10(pmax(1e-6,r_m))) + (x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*log10(3 + 14*r_m - 14*r_m^2) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log10(13/80 + r_m/75 - r_m^2/75) + (x[,"n_10"] + x[,"n_15"] + x[,"n_20"] + x[,"n_25"])*log10(3 - r_m + r_m^2) + (x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*log10(69 - 14*r_m + 14*r_m^2) + (x[,"n_02"] + x[,"n_03"] + x[,"n_32"] + x[,"n_33"])*log10(27 - 16*r_m + 16*r_m^2)
logL_m <- (-3*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 4*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_10"] - 4*x[,"n_11"] - 4*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_30"] - 4*x[,"n_31"] - 4*x[,"n_32"] - 4*x[,"n_33"] - 4*x[,"n_34"] - 3*x[,"n_35"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_05"] + x[,"n_30"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_m)) + log(pmax(1e-6,r_m))) + (x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*log(3 + 14*r_m - 14*r_m^2) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log(13/80 + r_m/75 - r_m^2/75) + (x[,"n_10"] + x[,"n_15"] + x[,"n_20"] + x[,"n_25"])*log(3 - r_m + r_m^2) + (x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*log(69 - 14*r_m + 14*r_m^2) + (x[,"n_02"] + x[,"n_03"] + x[,"n_32"] + x[,"n_33"])*log(27 - 16*r_m + 16*r_m^2)
logL_r <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(4*log(2) + 2*log(5)) + 2*(n05 + n30)*log(pmax(1e-6,1 - r)) + 2*(n00 + n35)*log(pmax(1e-6,r)) + (n10 + n25)*(log(4 - r) + log(pmax(1e-6,r))) + (n01 + n34)*(log(pmax(1e-6,r)) + log(2 + 7*r)) + (n03 + n32)*log(9 - 8*r^2) + (n02 + n33)*log(1 + 16*r - 8*r^2) + (n14 + n21)*log(33 - 20*r - 7*r^2) + (n11 + n24)*log(6 + 34*r - 7*r^2) + (n15 + n20)*log(3 - 2*r - r^2) + (n04 + n31)*log(9 - 16*r + 7*r^2) + (n13 + n22)*log(81 - 32*r + 8*r^2) + (n12 + n23)*log(57 + 16*r + 8*r^2)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n30,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_r <- 2*(x[,"n_05"] + x[,"n_30"])*log10(2) + 2*(x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_03"] + x[,"n_32"])*log10(7) - (x[,"n_02"] + x[,"n_33"])*log10(7) - (x[,"n_15"] + x[,"n_20"])*(-2*log10(2) + log10(7)) - (x[,"n_10"] + x[,"n_25"])*(-2*log10(2) + log10(7)) - (x[,"n_04"] + x[,"n_31"])*(-2*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_34"])*(-2*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_21"])*(-2*log10(2) + log10(5) + log10(17)) - (x[,"n_11"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(17)) - (x[,"n_13"] + x[,"n_22"])*log10(67) - (x[,"n_12"] + x[,"n_23"])*log10(67) + 2*(x[,"n_05"] + x[,"n_30"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_35"])*log10(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_25"])*(log10(4 - r_r) + log10(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_34"])*(log10(pmax(1e-6,r_r)) + log10(2 + 7*r_r)) + (x[,"n_03"] + x[,"n_32"])*log10(9 - 8*r_r^2) + (x[,"n_02"] + x[,"n_33"])*log10(1 + 16*r_r - 8*r_r^2) + (x[,"n_14"] + x[,"n_21"])*log10(33 - 20*r_r - 7*r_r^2) + (x[,"n_11"] + x[,"n_24"])*log10(6 + 34*r_r - 7*r_r^2) + (x[,"n_15"] + x[,"n_20"])*log10(3 - 2*r_r - r_r^2) + (x[,"n_04"] + x[,"n_31"])*log10(9 - 16*r_r + 7*r_r^2) + (x[,"n_13"] + x[,"n_22"])*log10(81 - 32*r_r + 8*r_r^2) + (x[,"n_12"] + x[,"n_23"])*log10(57 + 16*r_r + 8*r_r^2)
logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(4*log(2) + 2*log(5)) + 2*(x[,"n_05"] + x[,"n_30"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_00"] + x[,"n_35"])*log(pmax(1e-6,r_r)) + (x[,"n_10"] + x[,"n_25"])*(log(4 - r_r) + log(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_34"])*(log(pmax(1e-6,r_r)) + log(2 + 7*r_r)) + (x[,"n_03"] + x[,"n_32"])*log(9 - 8*r_r^2) + (x[,"n_02"] + x[,"n_33"])*log(1 + 16*r_r - 8*r_r^2) + (x[,"n_14"] + x[,"n_21"])*log(33 - 20*r_r - 7*r_r^2) + (x[,"n_11"] + x[,"n_24"])*log(6 + 34*r_r - 7*r_r^2) + (x[,"n_15"] + x[,"n_20"])*log(3 - 2*r_r - r_r^2) + (x[,"n_04"] + x[,"n_31"])*log(9 - 16*r_r + 7*r_r^2) + (x[,"n_13"] + x[,"n_22"])*log(81 - 32*r_r + 8*r_r^2) + (x[,"n_12"] + x[,"n_23"])*log(57 + 16*r_r + 8*r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_2.4 <- function(x){
logL_c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*(2*log(2) + 2*log(5)) + 2*(n01 + n35)*log(pmax(1e-6,1 - r)) + 2*(n05 + n31)*log(pmax(1e-6,r)) + (n15 + n21)*(log(4 - r) + log(pmax(1e-6,r))) + (n04 + n32)*(log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n33)*log(1 + 4*r - 4*r^2) + (n12 + n24)*log(15 - 8*r - r^2) + (n11 + n25)*log(3 - 2*r - r^2) + (n14 + n22)*log(6 + 10*r - r^2) + (n02 + n34)*log(3 - 4*r + r^2) + (n13 + n23)*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_c <- 2*(x[,"n_05"] + x[,"n_31"])*log10(2) - (x[,"n_03"] + x[,"n_33"])*log10(2) + 2*(x[,"n_01"] + x[,"n_35"])*log10(2) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_23"])*(2*log10(2) + log10(5)) - (x[,"n_15"] + x[,"n_21"])*(-2*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_25"])*(-2*log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + log10(43)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + log10(43)) + 2*(x[,"n_01"] + x[,"n_35"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_31"])*log10(pmax(1e-6,r_c)) + (x[,"n_15"] + x[,"n_21"])*(log10(4 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_32"])*(log10(pmax(1e-6,r_c)) + log10(2 + r_c)) + (x[,"n_03"] + x[,"n_33"])*log10(1 + 4*r_c - 4*r_c^2) + (x[,"n_12"] + x[,"n_24"])*log10(15 - 8*r_c - r_c^2) + (x[,"n_11"] + x[,"n_25"])*log10(3 - 2*r_c - r_c^2) + (x[,"n_14"] + x[,"n_22"])*log10(6 + 10*r_c - r_c^2) + (x[,"n_02"] + x[,"n_34"])*log10(3 - 4*r_c + r_c^2) + (x[,"n_13"] + x[,"n_23"])*log10(21 - 4*r_c + 4*r_c^2)
logL_c <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(2*log(2) + 2*log(5)) + 2*(x[,"n_01"] + x[,"n_35"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_05"] + x[,"n_31"])*log(pmax(1e-6,r_c)) + (x[,"n_15"] + x[,"n_21"])*(log(4 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_04"] + x[,"n_32"])*(log(pmax(1e-6,r_c)) + log(2 + r_c)) + (x[,"n_03"] + x[,"n_33"])*log(1 + 4*r_c - 4*r_c^2) + (x[,"n_12"] + x[,"n_24"])*log(15 - 8*r_c - r_c^2) + (x[,"n_11"] + x[,"n_25"])*log(3 - 2*r_c - r_c^2) + (x[,"n_14"] + x[,"n_22"])*log(6 + 10*r_c - r_c^2) + (x[,"n_02"] + x[,"n_34"])*log(3 - 4*r_c + r_c^2) + (x[,"n_13"] + x[,"n_23"])*log(21 - 4*r_c + 4*r_c^2)
logL_m <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-n01 - 2*n02 - 2*n03 - 2*n04 - n05 - n11 - 2*n12 - 2*n13 - 2*n14 - n15 - n21 - 2*n22 - 2*n23 - 2*n24 - n25 - n31 - 2*n32 - 2*n33 - 2*n34 - n35)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*(log(3) + 2*log(5)) + (n01 + n05 + n31 + n35)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n13 + n23)*log(57 + 8*r - 8*r^2) + (n02 + n04 + n32 + n34)*log(3 + 2*r - 2*r^2) + (n11 + n15 + n21 + n25)*log(3 - r + r^2) + (n12 + n14 + n22 + n24)*log(33 - 2*r + 2*r^2) + (n03 + n33)*log(9 - 8*r + 8*r^2)
return(L)}
interlogL_m <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_m,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_m <- 2*(x[,"n_01"] + x[,"n_05"] + x[,"n_31"] + x[,"n_35"])*log10(2) - (x[,"n_03"] + x[,"n_33"])*log10(7) - (x[,"n_02"] + x[,"n_04"] + x[,"n_32"] + x[,"n_34"])*(-log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_15"] + x[,"n_21"] + x[,"n_25"])*(-2*log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_22"] + x[,"n_24"])*(-log10(2) + log10(5) + log10(13)) - (x[,"n_13"] + x[,"n_23"])*log10(59) + (x[,"n_01"] + x[,"n_05"] + x[,"n_31"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_m)) + log10(pmax(1e-6,r_m))) + (x[,"n_13"] + x[,"n_23"])*log10(57 + 8*r_m - 8*r_m^2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_32"] + x[,"n_34"])*log10(3 + 2*r_m - 2*r_m^2) + (x[,"n_11"] + x[,"n_15"] + x[,"n_21"] + x[,"n_25"])*log10(3 - r_m + r_m^2) + (x[,"n_12"] + x[,"n_14"] + x[,"n_22"] + x[,"n_24"])*log10(33 - 2*r_m + 2*r_m^2) + (x[,"n_03"] + x[,"n_33"])*log10(9 - 8*r_m + 8*r_m^2)
logL_m <- (-x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - x[,"n_35"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_05"] + x[,"n_31"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_m)) + log(pmax(1e-6,r_m))) + (x[,"n_13"] + x[,"n_23"])*log(57 + 8*r_m - 8*r_m^2) + (x[,"n_02"] + x[,"n_04"] + x[,"n_32"] + x[,"n_34"])*log(3 + 2*r_m - 2*r_m^2) + (x[,"n_11"] + x[,"n_15"] + x[,"n_21"] + x[,"n_25"])*log(3 - r_m + r_m^2) + (x[,"n_12"] + x[,"n_14"] + x[,"n_22"] + x[,"n_24"])*log(33 - 2*r_m + 2*r_m^2) + (x[,"n_03"] + x[,"n_33"])*log(9 - 8*r_m + 8*r_m^2)
logL_r <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
L <- (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35)*(2*log(2) + 2*log(5)) + 2*(n05 + n31)*log(pmax(1e-6,1 - r)) + 2*(n01 + n35)*log(pmax(1e-6,r)) + (n11 + n25)*(log(4 - r) + log(pmax(1e-6,r))) + (n02 + n34)*(log(pmax(1e-6,r)) + log(2 + r)) + (n03 + n33)*log(1 + 4*r - 4*r^2) + (n14 + n22)*log(15 - 8*r - r^2) + (n15 + n21)*log(3 - 2*r - r^2) + (n12 + n24)*log(6 + 10*r - r^2) + (n04 + n32)*log(3 - 4*r + r^2) + (n13 + n23)*log(21 - 4*r + 4*r^2)
return(L)}
interlogL_r <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35) {
optimize(logL_r,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"])
LOD_r <- 2*(x[,"n_05"] + x[,"n_31"])*log10(2) - (x[,"n_03"] + x[,"n_33"])*log10(2) + 2*(x[,"n_01"] + x[,"n_35"])*log10(2) - (x[,"n_04"] + x[,"n_32"])*(-2*log10(2) + log10(5)) - (x[,"n_02"] + x[,"n_34"])*(-2*log10(2) + log10(5)) - (x[,"n_13"] + x[,"n_23"])*(2*log10(2) + log10(5)) - (x[,"n_15"] + x[,"n_21"])*(-2*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_25"])*(-2*log10(2) + log10(7)) - (x[,"n_14"] + x[,"n_22"])*(-2*log10(2) + log10(43)) - (x[,"n_12"] + x[,"n_24"])*(-2*log10(2) + log10(43)) + 2*(x[,"n_05"] + x[,"n_31"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_01"] + x[,"n_35"])*log10(pmax(1e-6,r_r)) + (x[,"n_11"] + x[,"n_25"])*(log10(4 - r_r) + log10(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_34"])*(log10(pmax(1e-6,r_r)) + log10(2 + r_r)) + (x[,"n_03"] + x[,"n_33"])*log10(1 + 4*r_r - 4*r_r^2) + (x[,"n_14"] + x[,"n_22"])*log10(15 - 8*r_r - r_r^2) + (x[,"n_15"] + x[,"n_21"])*log10(3 - 2*r_r - r_r^2) + (x[,"n_12"] + x[,"n_24"])*log10(6 + 10*r_r - r_r^2) + (x[,"n_04"] + x[,"n_32"])*log10(3 - 4*r_r + r_r^2) + (x[,"n_13"] + x[,"n_23"])*log10(21 - 4*r_r + 4*r_r^2)
logL_r <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(2*log(2) + 2*log(5)) + 2*(x[,"n_05"] + x[,"n_31"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_01"] + x[,"n_35"])*log(pmax(1e-6,r_r)) + (x[,"n_11"] + x[,"n_25"])*(log(4 - r_r) + log(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_34"])*(log(pmax(1e-6,r_r)) + log(2 + r_r)) + (x[,"n_03"] + x[,"n_33"])*log(1 + 4*r_r - 4*r_r^2) + (x[,"n_14"] + x[,"n_22"])*log(15 - 8*r_r - r_r^2) + (x[,"n_15"] + x[,"n_21"])*log(3 - 2*r_r - r_r^2) + (x[,"n_12"] + x[,"n_24"])*log(6 + 10*r_r - r_r^2) + (x[,"n_04"] + x[,"n_32"])*log(3 - 4*r_r + r_r^2) + (x[,"n_13"] + x[,"n_23"])*log(21 - 4*r_r + 4*r_r^2)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_3.0 <- function(x){
logL_c <- function(r,n01,n02,n10,n11,n12,n13,n20,n21,n22,n23,n31,n32,n00,n03,n30,n33) {
L <- (n01 + n02 + n10 + n11 + n12 + n13 + n20 + n21 + n22 + n23 + n31 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(2*log(2) + log(5)) + 3*(n00 + n33)*log(pmax(1e-6,1 - r)) + 3*(n03 + n30)*log(pmax(1e-6,r)) + (n01 + n10 + n23 + n32)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n13 + n20 + n31)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n12 + n21)*(log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n11 + n22)*log(3 - 5*r + 5*r^2 - 3*r^3)
return(L)}
interlogL_c <- function(n01,n02,n10,n11,n12,n13,n20,n21,n22,n23,n31,n32,n00,n03,n30,n33) {
optimize(logL_c,c(0,0.5), n01,n02,n10,n11,n12,n13,n20,n21,n22,n23,n31,n32,n00,n03,n30,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_31"],x[,"n_32"],x[,"n_00"],x[,"n_03"],x[,"n_30"],x[,"n_33"])
LOD_c <- 3*(x[,"n_03"] + x[,"n_30"])*log10(2) + 3*(x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*log10(2) + 3*(x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(2) + 3*(x[,"n_00"] + x[,"n_33"])*log10(2) - (x[,"n_12"] + x[,"n_21"])*(-3*log10(2) + log10(11)) - (x[,"n_11"] + x[,"n_22"])*(-3*log10(2) + log10(11)) + 3*(x[,"n_00"] + x[,"n_33"])*log10(pmax(1e-6,1 - r_c)) + 3*(x[,"n_03"] + x[,"n_30"])*log10(pmax(1e-6,r_c)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(2*log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log10(pmax(1e-6,1 - r_c)) + 2*log10(pmax(1e-6,r_c))) + (x[,"n_12"] + x[,"n_21"])*(log10(pmax(1e-6,r_c)) + log10(4 - 4*r_c + 3*r_c^2)) + (x[,"n_11"] + x[,"n_22"])*log10(3 - 5*r_c + 5*r_c^2 - 3*r_c^3)
logL_c <- (x[,"n_01"] + x[,"n_02"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_31"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(2*log(2) + log(5)) + 3*(x[,"n_00"] + x[,"n_33"])*log(pmax(1e-6,1 - r_c)) + 3*(x[,"n_03"] + x[,"n_30"])*log(pmax(1e-6,r_c)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(2*log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log(pmax(1e-6,1 - r_c)) + 2*log(pmax(1e-6,r_c))) + (x[,"n_12"] + x[,"n_21"])*(log(pmax(1e-6,r_c)) + log(4 - 4*r_c + 3*r_c^2)) + (x[,"n_11"] + x[,"n_22"])*log(3 - 5*r_c + 5*r_c^2 - 3*r_c^3)
logL_m <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(2*log(2) + log(3) + log(5)) + (n00 + n33)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n02 + n13 + n20 + n31)*(log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n12 + n21)*log(12 - 4*r + 13*r^2 - 9*r^3) + (n01 + n10 + n23 + n32)*log(3 - 5*r + 5*r^2 - 3*r^3) + (n11 + n22)*log(12 + 5*r - 14*r^2 + 9*r^3)
return(L)}
interlogL_m <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_m,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m <- mapply(interlogL_m,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_m <- 3*(x[,"n_03"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_33"])*log10(2) - (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_21"])*(-3*log10(2) + log10(97)) - (x[,"n_11"] + x[,"n_22"])*(-3*log10(2) + log10(97)) + (x[,"n_00"] + x[,"n_33"])*(2*log10(pmax(1e-6,1 - r_m)) + log10(pmax(1e-6,r_m))) + (x[,"n_03"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_m)) + 2*log10(pmax(1e-6,r_m))) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log10(pmax(1e-6,r_m)) + log10(4 - 4*r_m + 3*r_m^2)) + (x[,"n_12"] + x[,"n_21"])*log10(12 - 4*r_m + 13*r_m^2 - 9*r_m^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(3 - 5*r_m + 5*r_m^2 - 3*r_m^3) + (x[,"n_11"] + x[,"n_22"])*log10(12 + 5*r_m - 14*r_m^2 + 9*r_m^3)
logL_m <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(2*log(2) + log(3) + log(5)) + (x[,"n_00"] + x[,"n_33"])*(2*log(pmax(1e-6,1 - r_m)) + log(pmax(1e-6,r_m))) + (x[,"n_03"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_m)) + 2*log(pmax(1e-6,r_m))) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log(pmax(1e-6,r_m)) + log(4 - 4*r_m + 3*r_m^2)) + (x[,"n_12"] + x[,"n_21"])*log(12 - 4*r_m + 13*r_m^2 - 9*r_m^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log(3 - 5*r_m + 5*r_m^2 - 3*r_m^3) + (x[,"n_11"] + x[,"n_22"])*log(12 + 5*r_m - 14*r_m^2 + 9*r_m^3)
logL_r <- function(r,n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
L <- (-n00 - n01 - n02 - n03 - n10 - n11 - n12 - n13 - n20 - n21 - n22 - n23 - n30 - n31 - n32 - n33)*(2*log(2) + log(3) + log(5)) + (n00 + n33)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n02 + n13 + n20 + n31)*(log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n12 + n21)*log(12 - 4*r + 13*r^2 - 9*r^3) + (n01 + n10 + n23 + n32)*log(3 - 5*r + 5*r^2 - 3*r^3) + (n11 + n22)*log(12 + 5*r - 14*r^2 + 9*r^3)
return(L)}
interlogL_r <- function(n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33) {
optimize(logL_r,c(0,0.5), n00,n01,n02,n03,n10,n11,n12,n13,n20,n21,n22,n23,n30,n31,n32,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"])
LOD_r <- 3*(x[,"n_03"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_33"])*log10(2) - (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(-3*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*(-3*log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_21"])*(-3*log10(2) + log10(97)) - (x[,"n_11"] + x[,"n_22"])*(-3*log10(2) + log10(97)) + (x[,"n_00"] + x[,"n_33"])*(2*log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_03"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_r)) + 2*log10(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log10(pmax(1e-6,r_r)) + log10(4 - 4*r_r + 3*r_r^2)) + (x[,"n_12"] + x[,"n_21"])*log10(12 - 4*r_r + 13*r_r^2 - 9*r_r^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log10(3 - 5*r_r + 5*r_r^2 - 3*r_r^3) + (x[,"n_11"] + x[,"n_22"])*log10(12 + 5*r_r - 14*r_r^2 + 9*r_r^3)
logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"])*(2*log(2) + log(3) + log(5)) + (x[,"n_00"] + x[,"n_33"])*(2*log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_03"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_r)) + 2*log(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_13"] + x[,"n_20"] + x[,"n_31"])*(log(pmax(1e-6,r_r)) + log(4 - 4*r_r + 3*r_r^2)) + (x[,"n_12"] + x[,"n_21"])*log(12 - 4*r_r + 13*r_r^2 - 9*r_r^3) + (x[,"n_01"] + x[,"n_10"] + x[,"n_23"] + x[,"n_32"])*log(3 - 5*r_r + 5*r_r^2 - 3*r_r^3) + (x[,"n_11"] + x[,"n_22"])*log(12 + 5*r_r - 14*r_r^2 + 9*r_r^3)

return(list(
r_mat = cbind(r_c,r_m,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m,LOD_r,0),
logL_mat = cbind(logL_c,logL_m,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_3.1 <- function(x){
logL_c <- function(r,n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34) {
L <- (n02 + n10 + n11 + n12 + n13 + n14 + n20 + n21 + n22 + n23 + n24 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(3*log(2) + log(5)) + 3*(n00 + n34)*log(pmax(1e-6,1 - r)) + 3*(n04 + n30)*log(pmax(1e-6,r)) + (n02 + n32)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n24)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n31)*(log(3 - 2*r) + 2*log(pmax(1e-6,r))) + (n14 + n20)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n33)*(2*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n12 + n22)*log(3 - r + r^2) + (n13 + n21)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n11 + n23)*log(3 - 4*r + 3*r^2 - 2*r^3)
return(L)}
interlogL_c <- function(n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34) {
optimize(logL_c,c(0,0.5), n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"])
LOD_c <- 3*(x[,"n_14"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_24"])*log10(2) + 3*(x[,"n_04"] + x[,"n_30"])*log10(2) + (x[,"n_03"] + x[,"n_31"])*log10(2) + 2*(x[,"n_02"] + x[,"n_32"])*log10(2) + (x[,"n_01"] + x[,"n_33"])*log10(2) + 3*(x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_13"] + x[,"n_21"])*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_23"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_22"])*(-2*log10(2) + log10(11)) + 3*(x[,"n_00"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_c)) + 3*(x[,"n_04"] + x[,"n_30"])*log10(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_32"])*(log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_10"] + x[,"n_24"])*(2*log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_03"] + x[,"n_31"])*(log10(3 - 2*r_c) + 2*log10(pmax(1e-6,r_c))) + (x[,"n_14"] + x[,"n_20"])*(log10(pmax(1e-6,1 - r_c)) + 2*log10(pmax(1e-6,r_c))) + (x[,"n_01"] + x[,"n_33"])*(2*log10(pmax(1e-6,1 - r_c)) + log10(1 + 2*r_c)) + (x[,"n_12"] + x[,"n_22"])*log10(3 - r_c + r_c^2) + (x[,"n_13"] + x[,"n_21"])*(log10(pmax(1e-6,r_c)) + log10(4 - 3*r_c + 2*r_c^2)) + (x[,"n_11"] + x[,"n_23"])*log10(3 - 4*r_c + 3*r_c^2 - 2*r_c^3)
logL_c <- (x[,"n_02"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + 3*(x[,"n_00"] + x[,"n_34"])*log(pmax(1e-6,1 - r_c)) + 3*(x[,"n_04"] + x[,"n_30"])*log(pmax(1e-6,r_c)) + (x[,"n_02"] + x[,"n_32"])*(log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_10"] + x[,"n_24"])*(2*log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_03"] + x[,"n_31"])*(log(3 - 2*r_c) + 2*log(pmax(1e-6,r_c))) + (x[,"n_14"] + x[,"n_20"])*(log(pmax(1e-6,1 - r_c)) + 2*log(pmax(1e-6,r_c))) + (x[,"n_01"] + x[,"n_33"])*(2*log(pmax(1e-6,1 - r_c)) + log(1 + 2*r_c)) + (x[,"n_12"] + x[,"n_22"])*log(3 - r_c + r_c^2) + (x[,"n_13"] + x[,"n_21"])*(log(pmax(1e-6,r_c)) + log(4 - 3*r_c + 2*r_c^2)) + (x[,"n_11"] + x[,"n_23"])*log(3 - 4*r_c + 3*r_c^2 - 2*r_c^3)
logL_m1c <- function(r,n00,n01,n02,n03,n04,n10,n12,n14,n20,n22,n24,n30,n31,n32,n33,n34,n11,n13,n21,n23) {
L <- (-n00 - n01 - n02 - n03 - n04 - n10 - n12 - n14 - n20 - n22 - n24 - n30 - n31 - n32 - n33 - n34)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(3*log(2) + log(5)) + (n04 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n34)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n12 + n22)*log(24 + r - r^2) + (n02 + n32)*log(3 - r + r^2) + (n01 + n33)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n10 + n24)*(log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n14 + n20)*log(3 - 5*r + 5*r^2 - 3*r^3) + (n11 + n23)*log(4 + 3*r^2 - 2*r^3) + (n03 + n31)*log(3 - 4*r + 3*r^2 - 2*r^3) + (n13 + n21)*log(5 - 3*r^2 + 2*r^3)
return(L)}
interlogL_m1c <- function(n00,n01,n02,n03,n04,n10,n12,n14,n20,n22,n24,n30,n31,n32,n33,n34,n11,n13,n21,n23) {
optimize(logL_m1c,c(0,0.5), n00,n01,n02,n03,n04,n10,n12,n14,n20,n22,n24,n30,n31,n32,n33,n34,n11,n13,n21,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m1c <- mapply(interlogL_m1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_12"],x[,"n_14"],x[,"n_20"],x[,"n_22"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_11"],x[,"n_13"],x[,"n_21"],x[,"n_23"])
LOD_m1c <- 3*(x[,"n_04"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_03"] + x[,"n_31"])*(-log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_33"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_21"])*(-log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_23"])*(-log10(2) + 2*log10(3)) - (x[,"n_14"] + x[,"n_20"])*(-3*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_24"])*(-3*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_32"])*(-2*log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_22"])*(-2*log10(2) + log10(97)) + (x[,"n_04"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_m1c)) + log10(pmax(1e-6,r_m1c))) + (x[,"n_00"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_m1c)) + 2*log10(pmax(1e-6,r_m1c))) + (x[,"n_12"] + x[,"n_22"])*log10(24 + r_m1c - r_m1c^2) + (x[,"n_02"] + x[,"n_32"])*log10(3 - r_m1c + r_m1c^2) + (x[,"n_01"] + x[,"n_33"])*(log10(pmax(1e-6,r_m1c)) + log10(4 - 3*r_m1c + 2*r_m1c^2)) + (x[,"n_10"] + x[,"n_24"])*(log10(pmax(1e-6,r_m1c)) + log10(4 - 4*r_m1c + 3*r_m1c^2)) + (x[,"n_14"] + x[,"n_20"])*log10(3 - 5*r_m1c + 5*r_m1c^2 - 3*r_m1c^3) + (x[,"n_11"] + x[,"n_23"])*log10(4 + 3*r_m1c^2 - 2*r_m1c^3) + (x[,"n_03"] + x[,"n_31"])*log10(3 - 4*r_m1c + 3*r_m1c^2 - 2*r_m1c^3) + (x[,"n_13"] + x[,"n_21"])*log10(5 - 3*r_m1c^2 + 2*r_m1c^3)
logL_m1c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_12"] - x[,"n_14"] - x[,"n_20"] - x[,"n_22"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + (x[,"n_04"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_m1c)) + log(pmax(1e-6,r_m1c))) + (x[,"n_00"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_m1c)) + 2*log(pmax(1e-6,r_m1c))) + (x[,"n_12"] + x[,"n_22"])*log(24 + r_m1c - r_m1c^2) + (x[,"n_02"] + x[,"n_32"])*log(3 - r_m1c + r_m1c^2) + (x[,"n_01"] + x[,"n_33"])*(log(pmax(1e-6,r_m1c)) + log(4 - 3*r_m1c + 2*r_m1c^2)) + (x[,"n_10"] + x[,"n_24"])*(log(pmax(1e-6,r_m1c)) + log(4 - 4*r_m1c + 3*r_m1c^2)) + (x[,"n_14"] + x[,"n_20"])*log(3 - 5*r_m1c + 5*r_m1c^2 - 3*r_m1c^3) + (x[,"n_11"] + x[,"n_23"])*log(4 + 3*r_m1c^2 - 2*r_m1c^3) + (x[,"n_03"] + x[,"n_31"])*log(3 - 4*r_m1c + 3*r_m1c^2 - 2*r_m1c^3) + (x[,"n_13"] + x[,"n_21"])*log(5 - 3*r_m1c^2 + 2*r_m1c^3)
logL_m2c <- function(r,n00,n01,n02,n03,n04,n10,n12,n14,n20,n22,n24,n30,n31,n32,n33,n34,n11,n13,n21,n23) {
L <- (-n00 - n01 - n02 - n03 - n04 - n10 - n12 - n14 - n20 - n22 - n24 - n30 - n31 - n32 - n33 - n34)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(3*log(2) + log(5)) + (n00 + n34)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n12 + n22)*log(24 + r - r^2) + (n02 + n32)*log(3 - r + r^2) + (n03 + n31)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n14 + n20)*(log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n10 + n24)*log(3 - 5*r + 5*r^2 - 3*r^3) + (n13 + n21)*log(4 + 3*r^2 - 2*r^3) + (n01 + n33)*log(3 - 4*r + 3*r^2 - 2*r^3) + (n11 + n23)*log(5 - 3*r^2 + 2*r^3)
return(L)}
interlogL_m2c <- function(n00,n01,n02,n03,n04,n10,n12,n14,n20,n22,n24,n30,n31,n32,n33,n34,n11,n13,n21,n23) {
optimize(logL_m2c,c(0,0.5), n00,n01,n02,n03,n04,n10,n12,n14,n20,n22,n24,n30,n31,n32,n33,n34,n11,n13,n21,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m2c <- mapply(interlogL_m2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_10"],x[,"n_12"],x[,"n_14"],x[,"n_20"],x[,"n_22"],x[,"n_24"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_11"],x[,"n_13"],x[,"n_21"],x[,"n_23"])
LOD_m2c <- 3*(x[,"n_04"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_03"] + x[,"n_31"])*(-log10(2) + log10(3)) - (x[,"n_01"] + x[,"n_33"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_21"])*(-log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_23"])*(-log10(2) + 2*log10(3)) - (x[,"n_14"] + x[,"n_20"])*(-3*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_24"])*(-3*log10(2) + log10(11)) - (x[,"n_02"] + x[,"n_32"])*(-2*log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_22"])*(-2*log10(2) + log10(97)) + (x[,"n_00"] + x[,"n_34"])*(2*log10(pmax(1e-6,1 - r_m2c)) + log10(pmax(1e-6,r_m2c))) + (x[,"n_04"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_m2c)) + 2*log10(pmax(1e-6,r_m2c))) + (x[,"n_12"] + x[,"n_22"])*log10(24 + r_m2c - r_m2c^2) + (x[,"n_02"] + x[,"n_32"])*log10(3 - r_m2c + r_m2c^2) + (x[,"n_03"] + x[,"n_31"])*(log10(pmax(1e-6,r_m2c)) + log10(4 - 3*r_m2c + 2*r_m2c^2)) + (x[,"n_14"] + x[,"n_20"])*(log10(pmax(1e-6,r_m2c)) + log10(4 - 4*r_m2c + 3*r_m2c^2)) + (x[,"n_10"] + x[,"n_24"])*log10(3 - 5*r_m2c + 5*r_m2c^2 - 3*r_m2c^3) + (x[,"n_13"] + x[,"n_21"])*log10(4 + 3*r_m2c^2 - 2*r_m2c^3) + (x[,"n_01"] + x[,"n_33"])*log10(3 - 4*r_m2c + 3*r_m2c^2 - 2*r_m2c^3) + (x[,"n_11"] + x[,"n_23"])*log10(5 - 3*r_m2c^2 + 2*r_m2c^3)
logL_m2c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_12"] - x[,"n_14"] - x[,"n_20"] - x[,"n_22"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + (x[,"n_00"] + x[,"n_34"])*(2*log(pmax(1e-6,1 - r_m2c)) + log(pmax(1e-6,r_m2c))) + (x[,"n_04"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_m2c)) + 2*log(pmax(1e-6,r_m2c))) + (x[,"n_12"] + x[,"n_22"])*log(24 + r_m2c - r_m2c^2) + (x[,"n_02"] + x[,"n_32"])*log(3 - r_m2c + r_m2c^2) + (x[,"n_03"] + x[,"n_31"])*(log(pmax(1e-6,r_m2c)) + log(4 - 3*r_m2c + 2*r_m2c^2)) + (x[,"n_14"] + x[,"n_20"])*(log(pmax(1e-6,r_m2c)) + log(4 - 4*r_m2c + 3*r_m2c^2)) + (x[,"n_10"] + x[,"n_24"])*log(3 - 5*r_m2c + 5*r_m2c^2 - 3*r_m2c^3) + (x[,"n_13"] + x[,"n_21"])*log(4 + 3*r_m2c^2 - 2*r_m2c^3) + (x[,"n_01"] + x[,"n_33"])*log(3 - 4*r_m2c + 3*r_m2c^2 - 2*r_m2c^3) + (x[,"n_11"] + x[,"n_23"])*log(5 - 3*r_m2c^2 + 2*r_m2c^3)
logL_r <- function(r,n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34) {
L <- (n02 + n10 + n11 + n12 + n13 + n14 + n20 + n21 + n22 + n23 + n24 + n32)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n10 - n11 - n12 - n13 - n14 - n20 - n21 - n22 - n23 - n24 - n30 - n31 - n32 - n33 - n34)*(3*log(2) + log(5)) + 3*(n04 + n30)*log(pmax(1e-6,1 - r)) + 3*(n00 + n34)*log(pmax(1e-6,r)) + (n02 + n32)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n14 + n20)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n33)*(log(3 - 2*r) + 2*log(pmax(1e-6,r))) + (n10 + n24)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n03 + n31)*(2*log(pmax(1e-6,1 - r)) + log(1 + 2*r)) + (n12 + n22)*log(3 - r + r^2) + (n11 + n23)*(log(pmax(1e-6,r)) + log(4 - 3*r + 2*r^2)) + (n13 + n21)*log(3 - 4*r + 3*r^2 - 2*r^3)
return(L)}
interlogL_r <- function(n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34) {
optimize(logL_r,c(0,0.5), n02,n10,n11,n12,n13,n14,n20,n21,n22,n23,n24,n32,n00,n01,n03,n04,n30,n31,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_02"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_32"],x[,"n_00"],x[,"n_01"],x[,"n_03"],x[,"n_04"],x[,"n_30"],x[,"n_31"],x[,"n_33"],x[,"n_34"])
LOD_r <- 3*(x[,"n_14"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_24"])*log10(2) + 3*(x[,"n_04"] + x[,"n_30"])*log10(2) + (x[,"n_03"] + x[,"n_31"])*log10(2) + 2*(x[,"n_02"] + x[,"n_32"])*log10(2) + (x[,"n_01"] + x[,"n_33"])*log10(2) + 3*(x[,"n_00"] + x[,"n_34"])*log10(2) - (x[,"n_13"] + x[,"n_21"])*(-log10(2) + log10(3)) - (x[,"n_11"] + x[,"n_23"])*(-log10(2) + log10(3)) - (x[,"n_12"] + x[,"n_22"])*(-2*log10(2) + log10(11)) + 3*(x[,"n_04"] + x[,"n_30"])*log10(pmax(1e-6,1 - r_r)) + 3*(x[,"n_00"] + x[,"n_34"])*log10(pmax(1e-6,r_r)) + (x[,"n_02"] + x[,"n_32"])*(log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_14"] + x[,"n_20"])*(2*log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_33"])*(log10(3 - 2*r_r) + 2*log10(pmax(1e-6,r_r))) + (x[,"n_10"] + x[,"n_24"])*(log10(pmax(1e-6,1 - r_r)) + 2*log10(pmax(1e-6,r_r))) + (x[,"n_03"] + x[,"n_31"])*(2*log10(pmax(1e-6,1 - r_r)) + log10(1 + 2*r_r)) + (x[,"n_12"] + x[,"n_22"])*log10(3 - r_r + r_r^2) + (x[,"n_11"] + x[,"n_23"])*(log10(pmax(1e-6,r_r)) + log10(4 - 3*r_r + 2*r_r^2)) + (x[,"n_13"] + x[,"n_21"])*log10(3 - 4*r_r + 3*r_r^2 - 2*r_r^3)
logL_r <- (x[,"n_02"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_32"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + 3*(x[,"n_04"] + x[,"n_30"])*log(pmax(1e-6,1 - r_r)) + 3*(x[,"n_00"] + x[,"n_34"])*log(pmax(1e-6,r_r)) + (x[,"n_02"] + x[,"n_32"])*(log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_14"] + x[,"n_20"])*(2*log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_33"])*(log(3 - 2*r_r) + 2*log(pmax(1e-6,r_r))) + (x[,"n_10"] + x[,"n_24"])*(log(pmax(1e-6,1 - r_r)) + 2*log(pmax(1e-6,r_r))) + (x[,"n_03"] + x[,"n_31"])*(2*log(pmax(1e-6,1 - r_r)) + log(1 + 2*r_r)) + (x[,"n_12"] + x[,"n_22"])*log(3 - r_r + r_r^2) + (x[,"n_11"] + x[,"n_23"])*(log(pmax(1e-6,r_r)) + log(4 - 3*r_r + 2*r_r^2)) + (x[,"n_13"] + x[,"n_21"])*log(3 - 4*r_r + 3*r_r^2 - 2*r_r^3)

return(list(
r_mat = cbind(r_c,r_m1c,r_m2c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m1c,LOD_m2c,LOD_r,0),
logL_mat = cbind(logL_c,logL_m1c,logL_m2c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed1c","mixed2c","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_3.2 <- function(x){
logL_c <- function(r,n01,n04,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n31,n34,n00,n02,n03,n05,n30,n32,n33,n35) {
L <- (n01 + n04 + n10 + n11 + n12 + n13 + n14 + n15 + n20 + n21 + n22 + n23 + n24 + n25 + n31 + n34)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(2*log(2) + 2*log(5)) + 2*(n01 + n34)*log(pmax(1e-6,1 - r)) + 3*(n00 + n35)*log(pmax(1e-6,1 - r)) + 3*(n05 + n30)*log(pmax(1e-6,r)) + 2*(n04 + n31)*log(pmax(1e-6,r)) + (n10 + n25)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n14 + n21)*(log(4 - r) + log(pmax(1e-6,r))) + (n15 + n20)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n03 + n32)*(log(pmax(1e-6,r)) + log(3 + 3*r - 5*r^2)) + (n11 + n24)*log(3 - 2*r - r^2) + (n12 + n23)*log(9 - 10*r + 9*r^2 - 5*r^3) + (n02 + n33)*log(1 + 6*r - 12*r^2 + 5*r^3) + (n13 + n22)*log(3 + 7*r - 6*r^2 + 5*r^3)
return(L)}
interlogL_c <- function(n01,n04,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n31,n34,n00,n02,n03,n05,n30,n32,n33,n35) {
optimize(logL_c,c(0,0.5), n01,n04,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n31,n34,n00,n02,n03,n05,n30,n32,n33,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_34"],x[,"n_00"],x[,"n_02"],x[,"n_03"],x[,"n_05"],x[,"n_30"],x[,"n_32"],x[,"n_33"],x[,"n_35"])
LOD_c <- 3*(x[,"n_15"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_25"])*log10(2) + 3*(x[,"n_05"] + x[,"n_30"])*log10(2) + 2*(x[,"n_04"] + x[,"n_31"])*log10(2) + 2*(x[,"n_01"] + x[,"n_34"])*log10(2) + 3*(x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_14"] + x[,"n_21"])*(-2*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_24"])*(-2*log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_32"])*(-3*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_33"])*(-3*log10(2) + log10(13)) + 2*(x[,"n_01"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_c)) + 3*(x[,"n_00"] + x[,"n_35"])*log10(pmax(1e-6,1 - r_c)) + 3*(x[,"n_05"] + x[,"n_30"])*log10(pmax(1e-6,r_c)) + 2*(x[,"n_04"] + x[,"n_31"])*log10(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_25"])*(2*log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_14"] + x[,"n_21"])*(log10(4 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_15"] + x[,"n_20"])*(log10(pmax(1e-6,1 - r_c)) + 2*log10(pmax(1e-6,r_c))) + (x[,"n_03"] + x[,"n_32"])*(log10(pmax(1e-6,r_c)) + log10(3 + 3*r_c - 5*r_c^2)) + (x[,"n_11"] + x[,"n_24"])*log10(3 - 2*r_c - r_c^2) + (x[,"n_12"] + x[,"n_23"])*log10(9 - 10*r_c + 9*r_c^2 - 5*r_c^3) + (x[,"n_02"] + x[,"n_33"])*log10(1 + 6*r_c - 12*r_c^2 + 5*r_c^3) + (x[,"n_13"] + x[,"n_22"])*log10(3 + 7*r_c - 6*r_c^2 + 5*r_c^3)
logL_c <- (x[,"n_01"] + x[,"n_04"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_15"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_34"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(2*log(2) + 2*log(5)) + 2*(x[,"n_01"] + x[,"n_34"])*log(pmax(1e-6,1 - r_c)) + 3*(x[,"n_00"] + x[,"n_35"])*log(pmax(1e-6,1 - r_c)) + 3*(x[,"n_05"] + x[,"n_30"])*log(pmax(1e-6,r_c)) + 2*(x[,"n_04"] + x[,"n_31"])*log(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_25"])*(2*log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_14"] + x[,"n_21"])*(log(4 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_15"] + x[,"n_20"])*(log(pmax(1e-6,1 - r_c)) + 2*log(pmax(1e-6,r_c))) + (x[,"n_03"] + x[,"n_32"])*(log(pmax(1e-6,r_c)) + log(3 + 3*r_c - 5*r_c^2)) + (x[,"n_11"] + x[,"n_24"])*log(3 - 2*r_c - r_c^2) + (x[,"n_12"] + x[,"n_23"])*log(9 - 10*r_c + 9*r_c^2 - 5*r_c^3) + (x[,"n_02"] + x[,"n_33"])*log(1 + 6*r_c - 12*r_c^2 + 5*r_c^3) + (x[,"n_13"] + x[,"n_22"])*log(3 + 7*r_c - 6*r_c^2 + 5*r_c^3)
logL_m1c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n12,n13,n22,n23) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(2*log(2) + 2*log(5)) + (n05 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n34)*(log(4 - r) + log(pmax(1e-6,r))) + (n00 + n35)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n31)*log(3 - 2*r - r^2) + (n14 + n21)*log(21 - 10*r + r^2) + (n11 + n24)*log(12 + 8*r + r^2) + (n10 + n25)*(log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n12 + n23)*log(16 - r + 7*r^2 - 5*r^3) + (n03 + n32)*log(9 - 10*r + 9*r^2 - 5*r^3) + (n15 + n20)*log(3 - 5*r + 5*r^2 - 3*r^3) + (n13 + n22)*log(17 + 2*r - 8*r^2 + 5*r^3) + (n02 + n33)*log(3 + 7*r - 6*r^2 + 5*r^3)
return(L)}
interlogL_m1c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n12,n13,n22,n23) {
optimize(logL_m1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n12,n13,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m1c <- mapply(interlogL_m1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_12"],x[,"n_13"],x[,"n_22"],x[,"n_23"])
LOD_m1c <- 3*(x[,"n_05"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_03"] + x[,"n_32"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_33"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_04"] + x[,"n_31"])*(-2*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_34"])*(-2*log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_20"])*(-3*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_25"])*(-3*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_21"])*(-2*log10(2) + log10(5) + log10(13)) - (x[,"n_11"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(13)) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + log10(7) + log10(19)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + log10(7) + log10(19)) + (x[,"n_05"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_m1c)) + log10(pmax(1e-6,r_m1c))) + (x[,"n_01"] + x[,"n_34"])*(log10(4 - r_m1c) + log10(pmax(1e-6,r_m1c))) + (x[,"n_00"] + x[,"n_35"])*(log10(pmax(1e-6,1 - r_m1c)) + 2*log10(pmax(1e-6,r_m1c))) + (x[,"n_04"] + x[,"n_31"])*log10(3 - 2*r_m1c - r_m1c^2) + (x[,"n_14"] + x[,"n_21"])*log10(21 - 10*r_m1c + r_m1c^2) + (x[,"n_11"] + x[,"n_24"])*log10(12 + 8*r_m1c + r_m1c^2) + (x[,"n_10"] + x[,"n_25"])*(log10(pmax(1e-6,r_m1c)) + log10(4 - 4*r_m1c + 3*r_m1c^2)) + (x[,"n_12"] + x[,"n_23"])*log10(16 - r_m1c + 7*r_m1c^2 - 5*r_m1c^3) + (x[,"n_03"] + x[,"n_32"])*log10(9 - 10*r_m1c + 9*r_m1c^2 - 5*r_m1c^3) + (x[,"n_15"] + x[,"n_20"])*log10(3 - 5*r_m1c + 5*r_m1c^2 - 3*r_m1c^3) + (x[,"n_13"] + x[,"n_22"])*log10(17 + 2*r_m1c - 8*r_m1c^2 + 5*r_m1c^3) + (x[,"n_02"] + x[,"n_33"])*log10(3 + 7*r_m1c - 6*r_m1c^2 + 5*r_m1c^3)
logL_m1c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(2*log(2) + 2*log(5)) + (x[,"n_05"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_m1c)) + log(pmax(1e-6,r_m1c))) + (x[,"n_01"] + x[,"n_34"])*(log(4 - r_m1c) + log(pmax(1e-6,r_m1c))) + (x[,"n_00"] + x[,"n_35"])*(log(pmax(1e-6,1 - r_m1c)) + 2*log(pmax(1e-6,r_m1c))) + (x[,"n_04"] + x[,"n_31"])*log(3 - 2*r_m1c - r_m1c^2) + (x[,"n_14"] + x[,"n_21"])*log(21 - 10*r_m1c + r_m1c^2) + (x[,"n_11"] + x[,"n_24"])*log(12 + 8*r_m1c + r_m1c^2) + (x[,"n_10"] + x[,"n_25"])*(log(pmax(1e-6,r_m1c)) + log(4 - 4*r_m1c + 3*r_m1c^2)) + (x[,"n_12"] + x[,"n_23"])*log(16 - r_m1c + 7*r_m1c^2 - 5*r_m1c^3) + (x[,"n_03"] + x[,"n_32"])*log(9 - 10*r_m1c + 9*r_m1c^2 - 5*r_m1c^3) + (x[,"n_15"] + x[,"n_20"])*log(3 - 5*r_m1c + 5*r_m1c^2 - 3*r_m1c^3) + (x[,"n_13"] + x[,"n_22"])*log(17 + 2*r_m1c - 8*r_m1c^2 + 5*r_m1c^3) + (x[,"n_02"] + x[,"n_33"])*log(3 + 7*r_m1c - 6*r_m1c^2 + 5*r_m1c^3)
logL_m2c <- function(r,n00,n01,n02,n03,n04,n05,n10,n11,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n12,n13,n22,n23) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n14 - n15 - n20 - n21 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(2*log(2) + 2*log(5)) + (n00 + n35)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n31)*(log(4 - r) + log(pmax(1e-6,r))) + (n05 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n34)*log(3 - 2*r - r^2) + (n11 + n24)*log(21 - 10*r + r^2) + (n14 + n21)*log(12 + 8*r + r^2) + (n15 + n20)*(log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n13 + n22)*log(16 - r + 7*r^2 - 5*r^3) + (n02 + n33)*log(9 - 10*r + 9*r^2 - 5*r^3) + (n10 + n25)*log(3 - 5*r + 5*r^2 - 3*r^3) + (n12 + n23)*log(17 + 2*r - 8*r^2 + 5*r^3) + (n03 + n32)*log(3 + 7*r - 6*r^2 + 5*r^3)
return(L)}
interlogL_m2c <- function(n00,n01,n02,n03,n04,n05,n10,n11,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n12,n13,n22,n23) {
optimize(logL_m2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n10,n11,n14,n15,n20,n21,n24,n25,n30,n31,n32,n33,n34,n35,n12,n13,n22,n23, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m2c <- mapply(interlogL_m2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_24"],x[,"n_25"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_12"],x[,"n_13"],x[,"n_22"],x[,"n_23"])
LOD_m2c <- 3*(x[,"n_05"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_03"] + x[,"n_32"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_02"] + x[,"n_33"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_04"] + x[,"n_31"])*(-2*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_34"])*(-2*log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_20"])*(-3*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_25"])*(-3*log10(2) + log10(11)) - (x[,"n_14"] + x[,"n_21"])*(-2*log10(2) + log10(5) + log10(13)) - (x[,"n_11"] + x[,"n_24"])*(-2*log10(2) + log10(5) + log10(13)) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + log10(7) + log10(19)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + log10(7) + log10(19)) + (x[,"n_00"] + x[,"n_35"])*(2*log10(pmax(1e-6,1 - r_m2c)) + log10(pmax(1e-6,r_m2c))) + (x[,"n_04"] + x[,"n_31"])*(log10(4 - r_m2c) + log10(pmax(1e-6,r_m2c))) + (x[,"n_05"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_m2c)) + 2*log10(pmax(1e-6,r_m2c))) + (x[,"n_01"] + x[,"n_34"])*log10(3 - 2*r_m2c - r_m2c^2) + (x[,"n_11"] + x[,"n_24"])*log10(21 - 10*r_m2c + r_m2c^2) + (x[,"n_14"] + x[,"n_21"])*log10(12 + 8*r_m2c + r_m2c^2) + (x[,"n_15"] + x[,"n_20"])*(log10(pmax(1e-6,r_m2c)) + log10(4 - 4*r_m2c + 3*r_m2c^2)) + (x[,"n_13"] + x[,"n_22"])*log10(16 - r_m2c + 7*r_m2c^2 - 5*r_m2c^3) + (x[,"n_02"] + x[,"n_33"])*log10(9 - 10*r_m2c + 9*r_m2c^2 - 5*r_m2c^3) + (x[,"n_10"] + x[,"n_25"])*log10(3 - 5*r_m2c + 5*r_m2c^2 - 3*r_m2c^3) + (x[,"n_12"] + x[,"n_23"])*log10(17 + 2*r_m2c - 8*r_m2c^2 + 5*r_m2c^3) + (x[,"n_03"] + x[,"n_32"])*log10(3 + 7*r_m2c - 6*r_m2c^2 + 5*r_m2c^3)
logL_m2c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(2*log(2) + 2*log(5)) + (x[,"n_00"] + x[,"n_35"])*(2*log(pmax(1e-6,1 - r_m2c)) + log(pmax(1e-6,r_m2c))) + (x[,"n_04"] + x[,"n_31"])*(log(4 - r_m2c) + log(pmax(1e-6,r_m2c))) + (x[,"n_05"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_m2c)) + 2*log(pmax(1e-6,r_m2c))) + (x[,"n_01"] + x[,"n_34"])*log(3 - 2*r_m2c - r_m2c^2) + (x[,"n_11"] + x[,"n_24"])*log(21 - 10*r_m2c + r_m2c^2) + (x[,"n_14"] + x[,"n_21"])*log(12 + 8*r_m2c + r_m2c^2) + (x[,"n_15"] + x[,"n_20"])*(log(pmax(1e-6,r_m2c)) + log(4 - 4*r_m2c + 3*r_m2c^2)) + (x[,"n_13"] + x[,"n_22"])*log(16 - r_m2c + 7*r_m2c^2 - 5*r_m2c^3) + (x[,"n_02"] + x[,"n_33"])*log(9 - 10*r_m2c + 9*r_m2c^2 - 5*r_m2c^3) + (x[,"n_10"] + x[,"n_25"])*log(3 - 5*r_m2c + 5*r_m2c^2 - 3*r_m2c^3) + (x[,"n_12"] + x[,"n_23"])*log(17 + 2*r_m2c - 8*r_m2c^2 + 5*r_m2c^3) + (x[,"n_03"] + x[,"n_32"])*log(3 + 7*r_m2c - 6*r_m2c^2 + 5*r_m2c^3)
logL_r <- function(r,n01,n04,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n31,n34,n00,n02,n03,n05,n30,n32,n33,n35) {
L <- (n01 + n04 + n10 + n11 + n12 + n13 + n14 + n15 + n20 + n21 + n22 + n23 + n24 + n25 + n31 + n34)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n10 - n11 - n12 - n13 - n14 - n15 - n20 - n21 - n22 - n23 - n24 - n25 - n30 - n31 - n32 - n33 - n34 - n35)*(2*log(2) + 2*log(5)) + 3*(n05 + n30)*log(pmax(1e-6,1 - r)) + 2*(n04 + n31)*log(pmax(1e-6,1 - r)) + 2*(n01 + n34)*log(pmax(1e-6,r)) + 3*(n00 + n35)*log(pmax(1e-6,r)) + (n15 + n20)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n11 + n24)*(log(4 - r) + log(pmax(1e-6,r))) + (n10 + n25)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n02 + n33)*(log(pmax(1e-6,r)) + log(3 + 3*r - 5*r^2)) + (n14 + n21)*log(3 - 2*r - r^2) + (n13 + n22)*log(9 - 10*r + 9*r^2 - 5*r^3) + (n03 + n32)*log(1 + 6*r - 12*r^2 + 5*r^3) + (n12 + n23)*log(3 + 7*r - 6*r^2 + 5*r^3)
return(L)}
interlogL_r <- function(n01,n04,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n31,n34,n00,n02,n03,n05,n30,n32,n33,n35) {
optimize(logL_r,c(0,0.5), n01,n04,n10,n11,n12,n13,n14,n15,n20,n21,n22,n23,n24,n25,n31,n34,n00,n02,n03,n05,n30,n32,n33,n35, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_04"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_34"],x[,"n_00"],x[,"n_02"],x[,"n_03"],x[,"n_05"],x[,"n_30"],x[,"n_32"],x[,"n_33"],x[,"n_35"])
LOD_r <- 3*(x[,"n_15"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_25"])*log10(2) + 3*(x[,"n_05"] + x[,"n_30"])*log10(2) + 2*(x[,"n_04"] + x[,"n_31"])*log10(2) + 2*(x[,"n_01"] + x[,"n_34"])*log10(2) + 3*(x[,"n_00"] + x[,"n_35"])*log10(2) - (x[,"n_13"] + x[,"n_22"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_12"] + x[,"n_23"])*(-3*log10(2) + 2*log10(3) + log10(5)) - (x[,"n_14"] + x[,"n_21"])*(-2*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_24"])*(-2*log10(2) + log10(7)) - (x[,"n_03"] + x[,"n_32"])*(-3*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_33"])*(-3*log10(2) + log10(13)) + 3*(x[,"n_05"] + x[,"n_30"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_04"] + x[,"n_31"])*log10(pmax(1e-6,1 - r_r)) + 2*(x[,"n_01"] + x[,"n_34"])*log10(pmax(1e-6,r_r)) + 3*(x[,"n_00"] + x[,"n_35"])*log10(pmax(1e-6,r_r)) + (x[,"n_15"] + x[,"n_20"])*(2*log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_11"] + x[,"n_24"])*(log10(4 - r_r) + log10(pmax(1e-6,r_r))) + (x[,"n_10"] + x[,"n_25"])*(log10(pmax(1e-6,1 - r_r)) + 2*log10(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_33"])*(log10(pmax(1e-6,r_r)) + log10(3 + 3*r_r - 5*r_r^2)) + (x[,"n_14"] + x[,"n_21"])*log10(3 - 2*r_r - r_r^2) + (x[,"n_13"] + x[,"n_22"])*log10(9 - 10*r_r + 9*r_r^2 - 5*r_r^3) + (x[,"n_03"] + x[,"n_32"])*log10(1 + 6*r_r - 12*r_r^2 + 5*r_r^3) + (x[,"n_12"] + x[,"n_23"])*log10(3 + 7*r_r - 6*r_r^2 + 5*r_r^3)
logL_r <- (x[,"n_01"] + x[,"n_04"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_15"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_31"] + x[,"n_34"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"])*(2*log(2) + 2*log(5)) + 3*(x[,"n_05"] + x[,"n_30"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_04"] + x[,"n_31"])*log(pmax(1e-6,1 - r_r)) + 2*(x[,"n_01"] + x[,"n_34"])*log(pmax(1e-6,r_r)) + 3*(x[,"n_00"] + x[,"n_35"])*log(pmax(1e-6,r_r)) + (x[,"n_15"] + x[,"n_20"])*(2*log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_11"] + x[,"n_24"])*(log(4 - r_r) + log(pmax(1e-6,r_r))) + (x[,"n_10"] + x[,"n_25"])*(log(pmax(1e-6,1 - r_r)) + 2*log(pmax(1e-6,r_r))) + (x[,"n_02"] + x[,"n_33"])*(log(pmax(1e-6,r_r)) + log(3 + 3*r_r - 5*r_r^2)) + (x[,"n_14"] + x[,"n_21"])*log(3 - 2*r_r - r_r^2) + (x[,"n_13"] + x[,"n_22"])*log(9 - 10*r_r + 9*r_r^2 - 5*r_r^3) + (x[,"n_03"] + x[,"n_32"])*log(1 + 6*r_r - 12*r_r^2 + 5*r_r^3) + (x[,"n_12"] + x[,"n_23"])*log(3 + 7*r_r - 6*r_r^2 + 5*r_r^3)

return(list(
r_mat = cbind(r_c,r_m1c,r_m2c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m1c,LOD_m2c,LOD_r,0),
logL_mat = cbind(logL_c,logL_m1c,logL_m2c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed1c","mixed2c","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_3.3 <- function(x){
logL_c <- function(r,n01,n02,n04,n05,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n31,n32,n34,n35,n00,n03,n06,n30,n33,n36) {
L <- (n01 + n02 + n04 + n05 + n10 + n11 + n12 + n13 + n14 + n15 + n16 + n20 + n21 + n22 + n23 + n24 + n25 + n26 + n31 + n32 + n34 + n35)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(4*log(2) + 2*log(5)) + 3*(n00 + n36)*log(pmax(1e-6,1 - r)) + (n01 + n35)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r))) + 3*(n06 + n30)*log(pmax(1e-6,r)) + (n10 + n26)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n16 + n20)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n05 + n31)*(2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n03 + n33)*log(1 + 24*r - 24*r^2) + (n15 + n21)*(log(pmax(1e-6,r)) + log(4 + 5*r - 6*r^2)) + (n04 + n32)*(log(pmax(1e-6,r)) + log(1 + 7*r - 5*r^2)) + (n13 + n23)*log(27 - 8*r + 8*r^2) + (n12 + n24)*log(27 - 32*r + 23*r^2 - 15*r^3) + (n02 + n34)*log(3 - 8*r^2 + 5*r^3) + (n11 + n25)*log(3 + 4*r - 13*r^2 + 6*r^3) + (n14 + n22)*log(3 + 31*r - 22*r^2 + 15*r^3)
return(L)}
interlogL_c <- function(n01,n02,n04,n05,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n31,n32,n34,n35,n00,n03,n06,n30,n33,n36) {
optimize(logL_c,c(0,0.5), n01,n02,n04,n05,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n31,n32,n34,n35,n00,n03,n06,n30,n33,n36, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_00"],x[,"n_03"],x[,"n_06"],x[,"n_30"],x[,"n_33"],x[,"n_36"])
LOD_c <- 3*(x[,"n_16"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_26"])*log10(2) + 3*(x[,"n_06"] + x[,"n_30"])*log10(2) + (x[,"n_05"] + x[,"n_31"])*log10(2) + (x[,"n_01"] + x[,"n_35"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) - 2*(x[,"n_13"] + x[,"n_23"])*log10(5) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_33"])*log10(7) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(13)) - (x[,"n_14"] + x[,"n_22"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_12"] + x[,"n_24"])*(-3*log10(2) + log10(7) + log10(17)) + 3*(x[,"n_00"] + x[,"n_36"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_35"])*(log10(3 - 2*r_c) + 2*log10(pmax(1e-6,1 - r_c))) + 3*(x[,"n_06"] + x[,"n_30"])*log10(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_26"])*(2*log10(pmax(1e-6,1 - r_c)) + log10(pmax(1e-6,r_c))) + (x[,"n_16"] + x[,"n_20"])*(log10(pmax(1e-6,1 - r_c)) + 2*log10(pmax(1e-6,r_c))) + (x[,"n_05"] + x[,"n_31"])*(2*log10(pmax(1e-6,r_c)) + log10(1 + 2*r_c)) + (x[,"n_03"] + x[,"n_33"])*log10(1 + 24*r_c - 24*r_c^2) + (x[,"n_15"] + x[,"n_21"])*(log10(pmax(1e-6,r_c)) + log10(4 + 5*r_c - 6*r_c^2)) + (x[,"n_04"] + x[,"n_32"])*(log10(pmax(1e-6,r_c)) + log10(1 + 7*r_c - 5*r_c^2)) + (x[,"n_13"] + x[,"n_23"])*log10(27 - 8*r_c + 8*r_c^2) + (x[,"n_12"] + x[,"n_24"])*log10(27 - 32*r_c + 23*r_c^2 - 15*r_c^3) + (x[,"n_02"] + x[,"n_34"])*log10(3 - 8*r_c^2 + 5*r_c^3) + (x[,"n_11"] + x[,"n_25"])*log10(3 + 4*r_c - 13*r_c^2 + 6*r_c^3) + (x[,"n_14"] + x[,"n_22"])*log10(3 + 31*r_c - 22*r_c^2 + 15*r_c^3)
logL_c <- (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_15"] + x[,"n_16"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_26"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(4*log(2) + 2*log(5)) + 3*(x[,"n_00"] + x[,"n_36"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_35"])*(log(3 - 2*r_c) + 2*log(pmax(1e-6,1 - r_c))) + 3*(x[,"n_06"] + x[,"n_30"])*log(pmax(1e-6,r_c)) + (x[,"n_10"] + x[,"n_26"])*(2*log(pmax(1e-6,1 - r_c)) + log(pmax(1e-6,r_c))) + (x[,"n_16"] + x[,"n_20"])*(log(pmax(1e-6,1 - r_c)) + 2*log(pmax(1e-6,r_c))) + (x[,"n_05"] + x[,"n_31"])*(2*log(pmax(1e-6,r_c)) + log(1 + 2*r_c)) + (x[,"n_03"] + x[,"n_33"])*log(1 + 24*r_c - 24*r_c^2) + (x[,"n_15"] + x[,"n_21"])*(log(pmax(1e-6,r_c)) + log(4 + 5*r_c - 6*r_c^2)) + (x[,"n_04"] + x[,"n_32"])*(log(pmax(1e-6,r_c)) + log(1 + 7*r_c - 5*r_c^2)) + (x[,"n_13"] + x[,"n_23"])*log(27 - 8*r_c + 8*r_c^2) + (x[,"n_12"] + x[,"n_24"])*log(27 - 32*r_c + 23*r_c^2 - 15*r_c^3) + (x[,"n_02"] + x[,"n_34"])*log(3 - 8*r_c^2 + 5*r_c^3) + (x[,"n_11"] + x[,"n_25"])*log(3 + 4*r_c - 13*r_c^2 + 6*r_c^3) + (x[,"n_14"] + x[,"n_22"])*log(3 + 31*r_c - 22*r_c^2 + 15*r_c^3)
logL_m1c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(4*log(2) + log(3) + 2*log(5)) + (n06 + n30)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n36)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n13 + n23)*log(219 + 8*r - 8*r^2) + (n01 + n35)*(log(pmax(1e-6,r)) + log(4 + 5*r - 6*r^2)) + (n10 + n26)*(log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n03 + n33)*log(27 - 8*r + 8*r^2) + (n12 + n24)*log(120 + 5*r + 67*r^2 - 45*r^3) + (n15 + n21)*log(39 - 40*r + 31*r^2 - 18*r^3) + (n04 + n32)*log(27 - 32*r + 23*r^2 - 15*r^3) + (n16 + n20)*log(3 - 5*r + 5*r^2 - 3*r^3) + (n05 + n31)*log(3 + 4*r - 13*r^2 + 6*r^3) + (n02 + n34)*log(3 + 31*r - 22*r^2 + 15*r^3) + (n11 + n25)*log(12 + 32*r - 23*r^2 + 18*r^3) + (n14 + n22)*log(147 - 4*r - 68*r^2 + 45*r^3)
return(L)}
interlogL_m1c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36) {
optimize(logL_m1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m1c <- mapply(interlogL_m1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"])
LOD_m1c <- 3*(x[,"n_06"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) - 2*(x[,"n_03"] + x[,"n_33"])*log10(5) - (x[,"n_05"] + x[,"n_31"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_35"])*(-log10(2) + log10(5)) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + 2*log10(7)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + 2*log10(7)) - (x[,"n_16"] + x[,"n_20"])*(-3*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_26"])*(-3*log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_13"] + x[,"n_23"])*(log10(13) + log10(17)) - (x[,"n_14"] + x[,"n_22"])*(-3*log10(2) + log10(1069)) - (x[,"n_12"] + x[,"n_24"])*(-3*log10(2) + log10(1069)) + (x[,"n_06"] + x[,"n_30"])*(2*log10(pmax(1e-6,1 - r_m1c)) + log10(pmax(1e-6,r_m1c))) + (x[,"n_00"] + x[,"n_36"])*(log10(pmax(1e-6,1 - r_m1c)) + 2*log10(pmax(1e-6,r_m1c))) + (x[,"n_13"] + x[,"n_23"])*log10(219 + 8*r_m1c - 8*r_m1c^2) + (x[,"n_01"] + x[,"n_35"])*(log10(pmax(1e-6,r_m1c)) + log10(4 + 5*r_m1c - 6*r_m1c^2)) + (x[,"n_10"] + x[,"n_26"])*(log10(pmax(1e-6,r_m1c)) + log10(4 - 4*r_m1c + 3*r_m1c^2)) + (x[,"n_03"] + x[,"n_33"])*log10(27 - 8*r_m1c + 8*r_m1c^2) + (x[,"n_12"] + x[,"n_24"])*log10(120 + 5*r_m1c + 67*r_m1c^2 - 45*r_m1c^3) + (x[,"n_15"] + x[,"n_21"])*log10(39 - 40*r_m1c + 31*r_m1c^2 - 18*r_m1c^3) + (x[,"n_04"] + x[,"n_32"])*log10(27 - 32*r_m1c + 23*r_m1c^2 - 15*r_m1c^3) + (x[,"n_16"] + x[,"n_20"])*log10(3 - 5*r_m1c + 5*r_m1c^2 - 3*r_m1c^3) + (x[,"n_05"] + x[,"n_31"])*log10(3 + 4*r_m1c - 13*r_m1c^2 + 6*r_m1c^3) + (x[,"n_02"] + x[,"n_34"])*log10(3 + 31*r_m1c - 22*r_m1c^2 + 15*r_m1c^3) + (x[,"n_11"] + x[,"n_25"])*log10(12 + 32*r_m1c - 23*r_m1c^2 + 18*r_m1c^3) + (x[,"n_14"] + x[,"n_22"])*log10(147 - 4*r_m1c - 68*r_m1c^2 + 45*r_m1c^3)
logL_m1c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(4*log(2) + log(3) + 2*log(5)) + (x[,"n_06"] + x[,"n_30"])*(2*log(pmax(1e-6,1 - r_m1c)) + log(pmax(1e-6,r_m1c))) + (x[,"n_00"] + x[,"n_36"])*(log(pmax(1e-6,1 - r_m1c)) + 2*log(pmax(1e-6,r_m1c))) + (x[,"n_13"] + x[,"n_23"])*log(219 + 8*r_m1c - 8*r_m1c^2) + (x[,"n_01"] + x[,"n_35"])*(log(pmax(1e-6,r_m1c)) + log(4 + 5*r_m1c - 6*r_m1c^2)) + (x[,"n_10"] + x[,"n_26"])*(log(pmax(1e-6,r_m1c)) + log(4 - 4*r_m1c + 3*r_m1c^2)) + (x[,"n_03"] + x[,"n_33"])*log(27 - 8*r_m1c + 8*r_m1c^2) + (x[,"n_12"] + x[,"n_24"])*log(120 + 5*r_m1c + 67*r_m1c^2 - 45*r_m1c^3) + (x[,"n_15"] + x[,"n_21"])*log(39 - 40*r_m1c + 31*r_m1c^2 - 18*r_m1c^3) + (x[,"n_04"] + x[,"n_32"])*log(27 - 32*r_m1c + 23*r_m1c^2 - 15*r_m1c^3) + (x[,"n_16"] + x[,"n_20"])*log(3 - 5*r_m1c + 5*r_m1c^2 - 3*r_m1c^3) + (x[,"n_05"] + x[,"n_31"])*log(3 + 4*r_m1c - 13*r_m1c^2 + 6*r_m1c^3) + (x[,"n_02"] + x[,"n_34"])*log(3 + 31*r_m1c - 22*r_m1c^2 + 15*r_m1c^3) + (x[,"n_11"] + x[,"n_25"])*log(12 + 32*r_m1c - 23*r_m1c^2 + 18*r_m1c^3) + (x[,"n_14"] + x[,"n_22"])*log(147 - 4*r_m1c - 68*r_m1c^2 + 45*r_m1c^3)
logL_m2c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36) {
L <- (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(4*log(2) + log(3) + 2*log(5)) + (n00 + n36)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n06 + n30)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n13 + n23)*log(219 + 8*r - 8*r^2) + (n05 + n31)*(log(pmax(1e-6,r)) + log(4 + 5*r - 6*r^2)) + (n16 + n20)*(log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n03 + n33)*log(27 - 8*r + 8*r^2) + (n14 + n22)*log(120 + 5*r + 67*r^2 - 45*r^3) + (n11 + n25)*log(39 - 40*r + 31*r^2 - 18*r^3) + (n02 + n34)*log(27 - 32*r + 23*r^2 - 15*r^3) + (n10 + n26)*log(3 - 5*r + 5*r^2 - 3*r^3) + (n01 + n35)*log(3 + 4*r - 13*r^2 + 6*r^3) + (n04 + n32)*log(3 + 31*r - 22*r^2 + 15*r^3) + (n15 + n21)*log(12 + 32*r - 23*r^2 + 18*r^3) + (n12 + n24)*log(147 - 4*r - 68*r^2 + 45*r^3)
return(L)}
interlogL_m2c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36) {
optimize(logL_m2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m2c <- mapply(interlogL_m2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"])
LOD_m2c <- 3*(x[,"n_06"] + x[,"n_30"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) - 2*(x[,"n_03"] + x[,"n_33"])*log10(5) - (x[,"n_05"] + x[,"n_31"])*(-log10(2) + log10(5)) - (x[,"n_01"] + x[,"n_35"])*(-log10(2) + log10(5)) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + 2*log10(7)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + 2*log10(7)) - (x[,"n_16"] + x[,"n_20"])*(-3*log10(2) + log10(11)) - (x[,"n_10"] + x[,"n_26"])*(-3*log10(2) + log10(11)) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_13"] + x[,"n_23"])*(log10(13) + log10(17)) - (x[,"n_14"] + x[,"n_22"])*(-3*log10(2) + log10(1069)) - (x[,"n_12"] + x[,"n_24"])*(-3*log10(2) + log10(1069)) + (x[,"n_00"] + x[,"n_36"])*(2*log10(pmax(1e-6,1 - r_m2c)) + log10(pmax(1e-6,r_m2c))) + (x[,"n_06"] + x[,"n_30"])*(log10(pmax(1e-6,1 - r_m2c)) + 2*log10(pmax(1e-6,r_m2c))) + (x[,"n_13"] + x[,"n_23"])*log10(219 + 8*r_m2c - 8*r_m2c^2) + (x[,"n_05"] + x[,"n_31"])*(log10(pmax(1e-6,r_m2c)) + log10(4 + 5*r_m2c - 6*r_m2c^2)) + (x[,"n_16"] + x[,"n_20"])*(log10(pmax(1e-6,r_m2c)) + log10(4 - 4*r_m2c + 3*r_m2c^2)) + (x[,"n_03"] + x[,"n_33"])*log10(27 - 8*r_m2c + 8*r_m2c^2) + (x[,"n_14"] + x[,"n_22"])*log10(120 + 5*r_m2c + 67*r_m2c^2 - 45*r_m2c^3) + (x[,"n_11"] + x[,"n_25"])*log10(39 - 40*r_m2c + 31*r_m2c^2 - 18*r_m2c^3) + (x[,"n_02"] + x[,"n_34"])*log10(27 - 32*r_m2c + 23*r_m2c^2 - 15*r_m2c^3) + (x[,"n_10"] + x[,"n_26"])*log10(3 - 5*r_m2c + 5*r_m2c^2 - 3*r_m2c^3) + (x[,"n_01"] + x[,"n_35"])*log10(3 + 4*r_m2c - 13*r_m2c^2 + 6*r_m2c^3) + (x[,"n_04"] + x[,"n_32"])*log10(3 + 31*r_m2c - 22*r_m2c^2 + 15*r_m2c^3) + (x[,"n_15"] + x[,"n_21"])*log10(12 + 32*r_m2c - 23*r_m2c^2 + 18*r_m2c^3) + (x[,"n_12"] + x[,"n_24"])*log10(147 - 4*r_m2c - 68*r_m2c^2 + 45*r_m2c^3)
logL_m2c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(4*log(2) + log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_36"])*(2*log(pmax(1e-6,1 - r_m2c)) + log(pmax(1e-6,r_m2c))) + (x[,"n_06"] + x[,"n_30"])*(log(pmax(1e-6,1 - r_m2c)) + 2*log(pmax(1e-6,r_m2c))) + (x[,"n_13"] + x[,"n_23"])*log(219 + 8*r_m2c - 8*r_m2c^2) + (x[,"n_05"] + x[,"n_31"])*(log(pmax(1e-6,r_m2c)) + log(4 + 5*r_m2c - 6*r_m2c^2)) + (x[,"n_16"] + x[,"n_20"])*(log(pmax(1e-6,r_m2c)) + log(4 - 4*r_m2c + 3*r_m2c^2)) + (x[,"n_03"] + x[,"n_33"])*log(27 - 8*r_m2c + 8*r_m2c^2) + (x[,"n_14"] + x[,"n_22"])*log(120 + 5*r_m2c + 67*r_m2c^2 - 45*r_m2c^3) + (x[,"n_11"] + x[,"n_25"])*log(39 - 40*r_m2c + 31*r_m2c^2 - 18*r_m2c^3) + (x[,"n_02"] + x[,"n_34"])*log(27 - 32*r_m2c + 23*r_m2c^2 - 15*r_m2c^3) + (x[,"n_10"] + x[,"n_26"])*log(3 - 5*r_m2c + 5*r_m2c^2 - 3*r_m2c^3) + (x[,"n_01"] + x[,"n_35"])*log(3 + 4*r_m2c - 13*r_m2c^2 + 6*r_m2c^3) + (x[,"n_04"] + x[,"n_32"])*log(3 + 31*r_m2c - 22*r_m2c^2 + 15*r_m2c^3) + (x[,"n_15"] + x[,"n_21"])*log(12 + 32*r_m2c - 23*r_m2c^2 + 18*r_m2c^3) + (x[,"n_12"] + x[,"n_24"])*log(147 - 4*r_m2c - 68*r_m2c^2 + 45*r_m2c^3)
logL_r <- function(r,n01,n02,n04,n05,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n31,n32,n34,n35,n00,n03,n06,n30,n33,n36) {
L <- (n01 + n02 + n04 + n05 + n10 + n11 + n12 + n13 + n14 + n15 + n16 + n20 + n21 + n22 + n23 + n24 + n25 + n26 + n31 + n32 + n34 + n35)*log(3) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36)*(4*log(2) + 2*log(5)) + 3*(n06 + n30)*log(pmax(1e-6,1 - r)) + (n05 + n31)*(log(3 - 2*r) + 2*log(pmax(1e-6,1 - r))) + 3*(n00 + n36)*log(pmax(1e-6,r)) + (n16 + n20)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n10 + n26)*(log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n01 + n35)*(2*log(pmax(1e-6,r)) + log(1 + 2*r)) + (n03 + n33)*log(1 + 24*r - 24*r^2) + (n11 + n25)*(log(pmax(1e-6,r)) + log(4 + 5*r - 6*r^2)) + (n02 + n34)*(log(pmax(1e-6,r)) + log(1 + 7*r - 5*r^2)) + (n13 + n23)*log(27 - 8*r + 8*r^2) + (n14 + n22)*log(27 - 32*r + 23*r^2 - 15*r^3) + (n04 + n32)*log(3 - 8*r^2 + 5*r^3) + (n15 + n21)*log(3 + 4*r - 13*r^2 + 6*r^3) + (n12 + n24)*log(3 + 31*r - 22*r^2 + 15*r^3)
return(L)}
interlogL_r <- function(n01,n02,n04,n05,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n31,n32,n34,n35,n00,n03,n06,n30,n33,n36) {
optimize(logL_r,c(0,0.5), n01,n02,n04,n05,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n31,n32,n34,n35,n00,n03,n06,n30,n33,n36, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_r <- mapply(interlogL_r,x[,"n_01"],x[,"n_02"],x[,"n_04"],x[,"n_05"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_00"],x[,"n_03"],x[,"n_06"],x[,"n_30"],x[,"n_33"],x[,"n_36"])
LOD_r <- 3*(x[,"n_16"] + x[,"n_20"])*log10(2) + 3*(x[,"n_10"] + x[,"n_26"])*log10(2) + 3*(x[,"n_06"] + x[,"n_30"])*log10(2) + (x[,"n_05"] + x[,"n_31"])*log10(2) + (x[,"n_01"] + x[,"n_35"])*log10(2) + 3*(x[,"n_00"] + x[,"n_36"])*log10(2) - 2*(x[,"n_13"] + x[,"n_23"])*log10(5) - (x[,"n_15"] + x[,"n_21"])*(-log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_25"])*(-log10(2) + log10(5)) - (x[,"n_03"] + x[,"n_33"])*log10(7) - (x[,"n_04"] + x[,"n_32"])*(-3*log10(2) + log10(13)) - (x[,"n_02"] + x[,"n_34"])*(-3*log10(2) + log10(13)) - (x[,"n_14"] + x[,"n_22"])*(-3*log10(2) + log10(7) + log10(17)) - (x[,"n_12"] + x[,"n_24"])*(-3*log10(2) + log10(7) + log10(17)) + 3*(x[,"n_06"] + x[,"n_30"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_05"] + x[,"n_31"])*(log10(3 - 2*r_r) + 2*log10(pmax(1e-6,1 - r_r))) + 3*(x[,"n_00"] + x[,"n_36"])*log10(pmax(1e-6,r_r)) + (x[,"n_16"] + x[,"n_20"])*(2*log10(pmax(1e-6,1 - r_r)) + log10(pmax(1e-6,r_r))) + (x[,"n_10"] + x[,"n_26"])*(log10(pmax(1e-6,1 - r_r)) + 2*log10(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_35"])*(2*log10(pmax(1e-6,r_r)) + log10(1 + 2*r_r)) + (x[,"n_03"] + x[,"n_33"])*log10(1 + 24*r_r - 24*r_r^2) + (x[,"n_11"] + x[,"n_25"])*(log10(pmax(1e-6,r_r)) + log10(4 + 5*r_r - 6*r_r^2)) + (x[,"n_02"] + x[,"n_34"])*(log10(pmax(1e-6,r_r)) + log10(1 + 7*r_r - 5*r_r^2)) + (x[,"n_13"] + x[,"n_23"])*log10(27 - 8*r_r + 8*r_r^2) + (x[,"n_14"] + x[,"n_22"])*log10(27 - 32*r_r + 23*r_r^2 - 15*r_r^3) + (x[,"n_04"] + x[,"n_32"])*log10(3 - 8*r_r^2 + 5*r_r^3) + (x[,"n_15"] + x[,"n_21"])*log10(3 + 4*r_r - 13*r_r^2 + 6*r_r^3) + (x[,"n_12"] + x[,"n_24"])*log10(3 + 31*r_r - 22*r_r^2 + 15*r_r^3)
logL_r <- (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_15"] + x[,"n_16"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_26"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"])*log(3) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"])*(4*log(2) + 2*log(5)) + 3*(x[,"n_06"] + x[,"n_30"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_05"] + x[,"n_31"])*(log(3 - 2*r_r) + 2*log(pmax(1e-6,1 - r_r))) + 3*(x[,"n_00"] + x[,"n_36"])*log(pmax(1e-6,r_r)) + (x[,"n_16"] + x[,"n_20"])*(2*log(pmax(1e-6,1 - r_r)) + log(pmax(1e-6,r_r))) + (x[,"n_10"] + x[,"n_26"])*(log(pmax(1e-6,1 - r_r)) + 2*log(pmax(1e-6,r_r))) + (x[,"n_01"] + x[,"n_35"])*(2*log(pmax(1e-6,r_r)) + log(1 + 2*r_r)) + (x[,"n_03"] + x[,"n_33"])*log(1 + 24*r_r - 24*r_r^2) + (x[,"n_11"] + x[,"n_25"])*(log(pmax(1e-6,r_r)) + log(4 + 5*r_r - 6*r_r^2)) + (x[,"n_02"] + x[,"n_34"])*(log(pmax(1e-6,r_r)) + log(1 + 7*r_r - 5*r_r^2)) + (x[,"n_13"] + x[,"n_23"])*log(27 - 8*r_r + 8*r_r^2) + (x[,"n_14"] + x[,"n_22"])*log(27 - 32*r_r + 23*r_r^2 - 15*r_r^3) + (x[,"n_04"] + x[,"n_32"])*log(3 - 8*r_r^2 + 5*r_r^3) + (x[,"n_15"] + x[,"n_21"])*log(3 + 4*r_r - 13*r_r^2 + 6*r_r^3) + (x[,"n_12"] + x[,"n_24"])*log(3 + 31*r_r - 22*r_r^2 + 15*r_r^3)

return(list(
r_mat = cbind(r_c,r_m1c,r_m2c,r_r,0.499),
LOD_mat = cbind(LOD_c,LOD_m1c,LOD_m2c,LOD_r,0),
logL_mat = cbind(logL_c,logL_m1c,logL_m2c,logL_r,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed1c","mixed2c","repulsion","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.0_4.1 <- function(x){
logL_c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(3*log(2) + log(5)) + 2*(n01 + n34)*log(pmax(1e-6,1 - r)) + 2*(n04 + n31)*log(pmax(1e-6,r)) + (n03 + n32)*(log(2 - r) + log(pmax(1e-6,r))) + (n14 + n21)*(log(4 - r) + log(pmax(1e-6,r))) + (n02 + n33)*log(1 - r^2) + (n11 + n24)*log(3 - 2*r - r^2) + (n12 + n23)*log(9 - 4*r + r^2) + (n13 + n22)*log(6 + 2*r + r^2)
return(L)}
interlogL_c <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_c <- mapply(interlogL_c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_c <- 2*(x[,"n_04"] + x[,"n_31"])*log10(2) + 2*(x[,"n_01"] + x[,"n_34"])*log10(2) + (x[,"n_02"] + x[,"n_33"])*(2*log10(2) - log10(3)) - (x[,"n_03"] + x[,"n_32"])*(-2*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_21"])*(-2*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_24"])*(-2*log10(2) + log10(7)) - (x[,"n_13"] + x[,"n_22"])*(-2*log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_23"])*(-2*log10(2) + log10(29)) + 2*(x[,"n_01"] + x[,"n_34"])*log10(pmax(1e-6,1 - r_c)) + 2*(x[,"n_04"] + x[,"n_31"])*log10(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_32"])*(log10(2 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_14"] + x[,"n_21"])*(log10(4 - r_c) + log10(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_33"])*log10(1 - r_c^2) + (x[,"n_11"] + x[,"n_24"])*log10(3 - 2*r_c - r_c^2) + (x[,"n_12"] + x[,"n_23"])*log10(9 - 4*r_c + r_c^2) + (x[,"n_13"] + x[,"n_22"])*log10(6 + 2*r_c + r_c^2)
logL_c <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + 2*(x[,"n_01"] + x[,"n_34"])*log(pmax(1e-6,1 - r_c)) + 2*(x[,"n_04"] + x[,"n_31"])*log(pmax(1e-6,r_c)) + (x[,"n_03"] + x[,"n_32"])*(log(2 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_14"] + x[,"n_21"])*(log(4 - r_c) + log(pmax(1e-6,r_c))) + (x[,"n_02"] + x[,"n_33"])*log(1 - r_c^2) + (x[,"n_11"] + x[,"n_24"])*log(3 - 2*r_c - r_c^2) + (x[,"n_12"] + x[,"n_23"])*log(9 - 4*r_c + r_c^2) + (x[,"n_13"] + x[,"n_22"])*log(6 + 2*r_c + r_c^2)
logL_m1c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(3*log(2) + log(5)) + 2*(n04 + n31)*log(pmax(1e-6,1 - r)) + 2*(n01 + n34)*log(pmax(1e-6,r)) + (n02 + n33)*(log(2 - r) + log(pmax(1e-6,r))) + (n11 + n24)*(log(4 - r) + log(pmax(1e-6,r))) + (n03 + n32)*log(1 - r^2) + (n14 + n21)*log(3 - 2*r - r^2) + (n13 + n22)*log(9 - 4*r + r^2) + (n12 + n23)*log(6 + 2*r + r^2)
return(L)}
interlogL_m1c <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_m1c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m1c <- mapply(interlogL_m1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_m1c <- 2*(x[,"n_04"] + x[,"n_31"])*log10(2) + 2*(x[,"n_01"] + x[,"n_34"])*log10(2) + (x[,"n_03"] + x[,"n_32"])*(2*log10(2) - log10(3)) - (x[,"n_02"] + x[,"n_33"])*(-2*log10(2) + log10(3)) - (x[,"n_14"] + x[,"n_21"])*(-2*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_24"])*(-2*log10(2) + log10(7)) - (x[,"n_13"] + x[,"n_22"])*(-2*log10(2) + log10(29)) - (x[,"n_12"] + x[,"n_23"])*(-2*log10(2) + log10(29)) + 2*(x[,"n_04"] + x[,"n_31"])*log10(pmax(1e-6,1 - r_m1c)) + 2*(x[,"n_01"] + x[,"n_34"])*log10(pmax(1e-6,r_m1c)) + (x[,"n_02"] + x[,"n_33"])*(log10(2 - r_m1c) + log10(pmax(1e-6,r_m1c))) + (x[,"n_11"] + x[,"n_24"])*(log10(4 - r_m1c) + log10(pmax(1e-6,r_m1c))) + (x[,"n_03"] + x[,"n_32"])*log10(1 - r_m1c^2) + (x[,"n_14"] + x[,"n_21"])*log10(3 - 2*r_m1c - r_m1c^2) + (x[,"n_13"] + x[,"n_22"])*log10(9 - 4*r_m1c + r_m1c^2) + (x[,"n_12"] + x[,"n_23"])*log10(6 + 2*r_m1c + r_m1c^2)
logL_m1c <- (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(3*log(2) + log(5)) + 2*(x[,"n_04"] + x[,"n_31"])*log(pmax(1e-6,1 - r_m1c)) + 2*(x[,"n_01"] + x[,"n_34"])*log(pmax(1e-6,r_m1c)) + (x[,"n_02"] + x[,"n_33"])*(log(2 - r_m1c) + log(pmax(1e-6,r_m1c))) + (x[,"n_11"] + x[,"n_24"])*(log(4 - r_m1c) + log(pmax(1e-6,r_m1c))) + (x[,"n_03"] + x[,"n_32"])*log(1 - r_m1c^2) + (x[,"n_14"] + x[,"n_21"])*log(3 - 2*r_m1c - r_m1c^2) + (x[,"n_13"] + x[,"n_22"])*log(9 - 4*r_m1c + r_m1c^2) + (x[,"n_12"] + x[,"n_23"])*log(6 + 2*r_m1c + r_m1c^2)
logL_m2c <- function(r,n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
L <- (-2*n01 - 3*n02 - 3*n03 - 2*n04 - 2*n11 - 3*n12 - 3*n13 - 2*n14 - 2*n21 - 3*n22 - 3*n23 - 2*n24 - 2*n31 - 3*n32 - 3*n33 - 2*n34)*log(2) + (-n01 - n02 - n03 - n04 - n11 - n12 - n13 - n14 - n21 - n22 - n23 - n24 - n31 - n32 - n33 - n34)*(log(3) + log(5)) + (n01 + n04 + n31 + n34)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n12 + n13 + n22 + n23)*log(21 + 2*r - 2*r^2) + (n11 + n14 + n21 + n24)*log(3 - r + r^2) + (n02 + n03 + n32 + n33)*log(3 - 2*r + 2*r^2)
return(L)}
interlogL_m2c <- function(n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34) {
optimize(logL_m2c,c(0,0.5), n01,n02,n03,n04,n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m2c <- mapply(interlogL_m2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"])
LOD_m2c <- 2*(x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*log10(2) - (x[,"n_02"] + x[,"n_03"] + x[,"n_32"] + x[,"n_33"])*(-log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*(-2*log10(2) + log10(11)) - (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*(-log10(2) + log10(43)) + (x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*(log10(pmax(1e-6,1 - r_m2c)) + log10(pmax(1e-6,r_m2c))) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log10(21 + 2*r_m2c - 2*r_m2c^2) + (x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*log10(3 - r_m2c + r_m2c^2) + (x[,"n_02"] + x[,"n_03"] + x[,"n_32"] + x[,"n_33"])*log10(3 - 2*r_m2c + 2*r_m2c^2)
logL_m2c <- (-2*x[,"n_01"] - 3*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_21"] - 3*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_31"] - 3*x[,"n_32"] - 3*x[,"n_33"] - 2*x[,"n_34"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"])*(log(3) + log(5)) + (x[,"n_01"] + x[,"n_04"] + x[,"n_31"] + x[,"n_34"])*(log(pmax(1e-6,1 - r_m2c)) + log(pmax(1e-6,r_m2c))) + (x[,"n_12"] + x[,"n_13"] + x[,"n_22"] + x[,"n_23"])*log(21 + 2*r_m2c - 2*r_m2c^2) + (x[,"n_11"] + x[,"n_14"] + x[,"n_21"] + x[,"n_24"])*log(3 - r_m2c + r_m2c^2) + (x[,"n_02"] + x[,"n_03"] + x[,"n_32"] + x[,"n_33"])*log(3 - 2*r_m2c + 2*r_m2c^2)

return(list(
r_mat = cbind(r_c,r_m1c,r_m2c,0.499),
LOD_mat = cbind(LOD_c,LOD_m1c,LOD_m2c,0),
logL_mat = cbind(logL_c,logL_m1c,logL_m2c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling","mixed1c","mixed2c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.3_2.4 <- function(x){
logL_cc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
L <- (-4*n01 - 2*n02 - 3*n03 - 2*n04 - 4*n05 - 3*n11 - 2*n12 - 2*n13 - 2*n14 - 3*n15 - 4*n21 - 2*n22 - 3*n23 - 2*n24 - 4*n25 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 4*n41 - 2*n42 - 3*n43 - 2*n44 - 4*n45 - 3*n51 - 2*n52 - 2*n53 - 2*n54 - 3*n55 - 4*n61 - 2*n62 - 3*n63 - 2*n64 - 4*n65)*log(2) + (n03 + n63)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55 - n61 - n62 - n63 - n64 - n65)*log(5) + 4*(n01 + n65)*log(pmax(1e-6,1 - r)) + 4*(n05 + n61)*log(pmax(1e-6,r)) + (n02 + n64)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n63)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n04 + n62)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n15 + n51)*(log(4 - r) + 3*log(pmax(1e-6,r))) + (n25 + n41)*(2*log(pmax(1e-6,r)) + log(22 - 12*r - r^2)) + (n11 + n55)*(2*log(pmax(1e-6,1 - r)) + log(3 - 2*r - r^2)) + (n21 + n45)*(2*log(pmax(1e-6,1 - r)) + log(9 + 14*r - r^2)) + (n14 + n52)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n12 + n54)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n13 + n53)*(log(pmax(1e-6,r)) + log(8 - 11*r + 6*r^2 - 3*r^3)) + (n31 + n35)*(log(pmax(1e-6,r)) + log(6 - 5*r - 2*r^2 + r^3)) + (n24 + n42)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n32 + n34)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n23 + n43)*log(21 + 4*r - 7*r^2 + 6*r^3 - 3*r^4) + (n22 + n44)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + n33*log(23 - 20*r + 26*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_cc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
optimize(logL_cc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"])
LOD_cc <- 4*(x[,"n_05"] + x[,"n_61"])*log10(2) + 4*(x[,"n_04"] + x[,"n_62"])*log10(2) + 4*(x[,"n_03"] + x[,"n_63"])*log10(2) + 4*(x[,"n_02"] + x[,"n_64"])*log10(2) + 4*(x[,"n_01"] + x[,"n_65"])*log10(2) - (x[,"n_14"] + x[,"n_52"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_12"] + x[,"n_54"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_31"] + x[,"n_35"])*(-4*log10(2) + 2*log10(5)) - (x[,"n_15"] + x[,"n_51"])*(-4*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_55"])*(-4*log10(2) + log10(7)) - (x[,"n_25"] + x[,"n_41"])*(-4*log10(2) + 2*log10(3) + log10(7)) - (x[,"n_21"] + x[,"n_45"])*(-4*log10(2) + 2*log10(3) + log10(7)) - x[,"n_33"]*(-3*log10(2) + log10(3) + 2*log10(7)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_24"] + x[,"n_42"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_22"] + x[,"n_44"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_13"] + x[,"n_53"])*(-4*log10(2) + log10(29)) - (x[,"n_23"] + x[,"n_43"])*(-4*log10(2) + log10(349)) + 4*(x[,"n_01"] + x[,"n_65"])*log10(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_61"])*log10(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_64"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_63"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_62"])*(log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_51"])*(log10(4 - r_cc) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_25"] + x[,"n_41"])*(2*log10(pmax(1e-6,r_cc)) + log10(22 - 12*r_cc - r_cc^2)) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 - 2*r_cc - r_cc^2)) + (x[,"n_21"] + x[,"n_45"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(9 + 14*r_cc - r_cc^2)) + (x[,"n_14"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_cc)) + log10(7 - 6*r_cc + 2*r_cc^2)) + (x[,"n_12"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(3 + 2*r_cc + 2*r_cc^2)) + (x[,"n_13"] + x[,"n_53"])*(log10(pmax(1e-6,r_cc)) + log10(8 - 11*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_31"] + x[,"n_35"])*(log10(pmax(1e-6,r_cc)) + log10(6 - 5*r_cc - 2*r_cc^2 + r_cc^3)) + (x[,"n_24"] + x[,"n_42"])*(log10(pmax(1e-6,r_cc)) + log10(15 - 9*r_cc + 5*r_cc^2 + r_cc^3)) + (x[,"n_32"] + x[,"n_34"])*log10(9 + 4*r_cc - 8*r_cc^2 + 8*r_cc^3 - 4*r_cc^4) + (x[,"n_23"] + x[,"n_43"])*log10(21 + 4*r_cc - 7*r_cc^2 + 6*r_cc^3 - 3*r_cc^4) + (x[,"n_22"] + x[,"n_44"])*log10(12 - 16*r_cc + 12*r_cc^2 - 9*r_cc^3 + r_cc^4) + x[,"n_33"]*log10(23 - 20*r_cc + 26*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cc <- (-4*x[,"n_01"] - 2*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 4*x[,"n_05"] - 3*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_21"] - 2*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 4*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 4*x[,"n_41"] - 2*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"] - 4*x[,"n_45"] - 3*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 3*x[,"n_55"] - 4*x[,"n_61"] - 2*x[,"n_62"] - 3*x[,"n_63"] - 2*x[,"n_64"] - 4*x[,"n_65"])*log(2) + (x[,"n_03"] + x[,"n_63"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"])*log(5) + 4*(x[,"n_01"] + x[,"n_65"])*log(pmax(1e-6,1 - r_cc)) + 4*(x[,"n_05"] + x[,"n_61"])*log(pmax(1e-6,r_cc)) + (x[,"n_02"] + x[,"n_64"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_63"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_62"])*(log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_51"])*(log(4 - r_cc) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_25"] + x[,"n_41"])*(2*log(pmax(1e-6,r_cc)) + log(22 - 12*r_cc - r_cc^2)) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 - 2*r_cc - r_cc^2)) + (x[,"n_21"] + x[,"n_45"])*(2*log(pmax(1e-6,1 - r_cc)) + log(9 + 14*r_cc - r_cc^2)) + (x[,"n_14"] + x[,"n_52"])*(2*log(pmax(1e-6,r_cc)) + log(7 - 6*r_cc + 2*r_cc^2)) + (x[,"n_12"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_cc)) + log(3 + 2*r_cc + 2*r_cc^2)) + (x[,"n_13"] + x[,"n_53"])*(log(pmax(1e-6,r_cc)) + log(8 - 11*r_cc + 6*r_cc^2 - 3*r_cc^3)) + (x[,"n_31"] + x[,"n_35"])*(log(pmax(1e-6,r_cc)) + log(6 - 5*r_cc - 2*r_cc^2 + r_cc^3)) + (x[,"n_24"] + x[,"n_42"])*(log(pmax(1e-6,r_cc)) + log(15 - 9*r_cc + 5*r_cc^2 + r_cc^3)) + (x[,"n_32"] + x[,"n_34"])*log(9 + 4*r_cc - 8*r_cc^2 + 8*r_cc^3 - 4*r_cc^4) + (x[,"n_23"] + x[,"n_43"])*log(21 + 4*r_cc - 7*r_cc^2 + 6*r_cc^3 - 3*r_cc^4) + (x[,"n_22"] + x[,"n_44"])*log(12 - 16*r_cc + 12*r_cc^2 - 9*r_cc^3 + r_cc^4) + x[,"n_33"]*log(23 - 20*r_cc + 26*r_cc^2 - 12*r_cc^3 + 6*r_cc^4)
logL_cm2c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
L <- (-3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 3*n11 - 2*n12 - n13 - 2*n14 - 3*n15 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 3*n31 - n32 - n33 - n34 - 3*n35 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 3*n51 - 2*n52 - n53 - 2*n54 - 3*n55 - 3*n61 - 4*n62 - 2*n63 - 4*n64 - 3*n65)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55 - n61 - n62 - n63 - n64 - n65)*(log(3) + 2*log(5)) + (n01 + n65)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n61)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n15 + n51)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n11 + n55)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n04 + n62)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n02 + n64)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n14 + n52)*(log(pmax(1e-6,r)) + log(9 - 9*r + 10*r^2 - 4*r^3)) + (n03 + n63)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n25 + n41)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n24 + n42)*log(45 + 42*r + 29*r^2 - 12*r^3 - 8*r^4) + (n22 + n44)*log(96 - 32*r - 55*r^2 + 44*r^3 - 8*r^4) + n33*log(27 + 2*r - 8*r^2 + 12*r^3 - 6*r^4) + (n12 + n54)*log(6 - 5*r - 3*r^2 + 6*r^3 - 4*r^4) + (n31 + n35)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n21 + n45)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + (n13 + n53)*log(3 - r + 4*r^2 - 6*r^3 + 3*r^4) + (n23 + n43)*log(33 - 2*r + 5*r^2 - 6*r^3 + 3*r^4) + (n32 + n34)*log(15 - 2*r + 6*r^2 - 8*r^3 + 4*r^4)
return(L)}
interlogL_cm2c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
optimize(logL_cm2c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"])
LOD_cm2c <- 4*(x[,"n_05"] + x[,"n_61"])*log10(2) + 4*(x[,"n_01"] + x[,"n_65"])*log10(2) - (x[,"n_04"] + x[,"n_62"])*(-2*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_64"])*(-2*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_51"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_55"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_63"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_52"])*(-2*log10(2) + log10(13)) - (x[,"n_12"] + x[,"n_54"])*(-2*log10(2) + log10(13)) - (x[,"n_31"] + x[,"n_35"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_25"] + x[,"n_41"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_21"] + x[,"n_45"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - x[,"n_33"]*(-3*log10(2) + log10(7) + log10(31)) - (x[,"n_13"] + x[,"n_53"])*(-4*log10(2) + log10(47)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(59)) - (x[,"n_23"] + x[,"n_43"])*(-4*log10(2) + log10(523)) + (x[,"n_01"] + x[,"n_65"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_61"])*(log10(pmax(1e-6,1 - r_cm2c)) + 3*log10(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(7 - 6*r_cm2c + 2*r_cm2c^2)) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 + 2*r_cm2c + 2*r_cm2c^2)) + (x[,"n_04"] + x[,"n_62"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(7 - 12*r_cm2c + 8*r_cm2c^2)) + (x[,"n_02"] + x[,"n_64"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 - 4*r_cm2c + 8*r_cm2c^2)) + (x[,"n_14"] + x[,"n_52"])*(log10(pmax(1e-6,r_cm2c)) + log10(9 - 9*r_cm2c + 10*r_cm2c^2 - 4*r_cm2c^3)) + (x[,"n_03"] + x[,"n_63"])*(log10(pmax(1e-6,r_cm2c)) + log10(2 - 5*r_cm2c + 6*r_cm2c^2 - 3*r_cm2c^3)) + (x[,"n_25"] + x[,"n_41"])*(log10(pmax(1e-6,r_cm2c)) + log10(15 - 9*r_cm2c + 5*r_cm2c^2 + r_cm2c^3)) + (x[,"n_24"] + x[,"n_42"])*log10(45 + 42*r_cm2c + 29*r_cm2c^2 - 12*r_cm2c^3 - 8*r_cm2c^4) + (x[,"n_22"] + x[,"n_44"])*log10(96 - 32*r_cm2c - 55*r_cm2c^2 + 44*r_cm2c^3 - 8*r_cm2c^4) + x[,"n_33"]*log10(27 + 2*r_cm2c - 8*r_cm2c^2 + 12*r_cm2c^3 - 6*r_cm2c^4) + (x[,"n_12"] + x[,"n_54"])*log10(6 - 5*r_cm2c - 3*r_cm2c^2 + 6*r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_31"] + x[,"n_35"])*log10(9 + 4*r_cm2c - 8*r_cm2c^2 + 8*r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_21"] + x[,"n_45"])*log10(12 - 16*r_cm2c + 12*r_cm2c^2 - 9*r_cm2c^3 + r_cm2c^4) + (x[,"n_13"] + x[,"n_53"])*log10(3 - r_cm2c + 4*r_cm2c^2 - 6*r_cm2c^3 + 3*r_cm2c^4) + (x[,"n_23"] + x[,"n_43"])*log10(33 - 2*r_cm2c + 5*r_cm2c^2 - 6*r_cm2c^3 + 3*r_cm2c^4) + (x[,"n_32"] + x[,"n_34"])*log10(15 - 2*r_cm2c + 6*r_cm2c^2 - 8*r_cm2c^3 + 4*r_cm2c^4)
logL_cm2c <- (-3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 2*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"])*(log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_65"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_61"])*(log(pmax(1e-6,1 - r_cm2c)) + 3*log(pmax(1e-6,r_cm2c))) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,r_cm2c)) + log(7 - 6*r_cm2c + 2*r_cm2c^2)) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(3 + 2*r_cm2c + 2*r_cm2c^2)) + (x[,"n_04"] + x[,"n_62"])*(2*log(pmax(1e-6,r_cm2c)) + log(7 - 12*r_cm2c + 8*r_cm2c^2)) + (x[,"n_02"] + x[,"n_64"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(3 - 4*r_cm2c + 8*r_cm2c^2)) + (x[,"n_14"] + x[,"n_52"])*(log(pmax(1e-6,r_cm2c)) + log(9 - 9*r_cm2c + 10*r_cm2c^2 - 4*r_cm2c^3)) + (x[,"n_03"] + x[,"n_63"])*(log(pmax(1e-6,r_cm2c)) + log(2 - 5*r_cm2c + 6*r_cm2c^2 - 3*r_cm2c^3)) + (x[,"n_25"] + x[,"n_41"])*(log(pmax(1e-6,r_cm2c)) + log(15 - 9*r_cm2c + 5*r_cm2c^2 + r_cm2c^3)) + (x[,"n_24"] + x[,"n_42"])*log(45 + 42*r_cm2c + 29*r_cm2c^2 - 12*r_cm2c^3 - 8*r_cm2c^4) + (x[,"n_22"] + x[,"n_44"])*log(96 - 32*r_cm2c - 55*r_cm2c^2 + 44*r_cm2c^3 - 8*r_cm2c^4) + x[,"n_33"]*log(27 + 2*r_cm2c - 8*r_cm2c^2 + 12*r_cm2c^3 - 6*r_cm2c^4) + (x[,"n_12"] + x[,"n_54"])*log(6 - 5*r_cm2c - 3*r_cm2c^2 + 6*r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_31"] + x[,"n_35"])*log(9 + 4*r_cm2c - 8*r_cm2c^2 + 8*r_cm2c^3 - 4*r_cm2c^4) + (x[,"n_21"] + x[,"n_45"])*log(12 - 16*r_cm2c + 12*r_cm2c^2 - 9*r_cm2c^3 + r_cm2c^4) + (x[,"n_13"] + x[,"n_53"])*log(3 - r_cm2c + 4*r_cm2c^2 - 6*r_cm2c^3 + 3*r_cm2c^4) + (x[,"n_23"] + x[,"n_43"])*log(33 - 2*r_cm2c + 5*r_cm2c^2 - 6*r_cm2c^3 + 3*r_cm2c^4) + (x[,"n_32"] + x[,"n_34"])*log(15 - 2*r_cm2c + 6*r_cm2c^2 - 8*r_cm2c^3 + 4*r_cm2c^4)
logL_cm1c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
L <- (-4*n01 - 3*n02 - 4*n03 - 3*n04 - 4*n05 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 4*n21 - 3*n22 - 4*n23 - 3*n24 - 4*n25 - 3*n31 - 2*n32 - n33 - 2*n34 - 3*n35 - 4*n41 - 3*n42 - 4*n43 - 3*n44 - 4*n45 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55 - 4*n61 - 3*n62 - 4*n63 - 3*n64 - 4*n65)*log(2) + n33*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55 - n61 - n62 - n63 - n64 - n65)*log(5) + (n01 + n05 + n61 + n65)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n15 + n51 + n55)*(log(pmax(1e-6,r)) + log(2 - 3*r + 2*r^2 - r^3)) + (n13 + n53)*log(3 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n23 + n43)*log(45 - 4*r - 2*r^2 + 12*r^3 - 6*r^4) + (n32 + n34)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n02 + n04 + n62 + n64)*log(r - 3*r^2 + 4*r^3 - 2*r^4) + (n21 + n25 + n41 + n45)*log(3 + 4*r - 5*r^2 + 2*r^3 - r^4) + n33*log(3 + r^2 - 2*r^3 + r^4) + (n22 + n24 + n42 + n44)*log(12 - r + 3*r^2 - 4*r^3 + 2*r^4) + (n31 + n35)*log(5 - 8*r + 10*r^2 - 4*r^3 + 2*r^4) + (n12 + n14 + n52 + n54)*log(3 - 4*r + 8*r^2 - 8*r^3 + 4*r^4) + (n03 + n63)*log(1 - 4*r + 10*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_cm1c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
optimize(logL_cm1c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"])
LOD_cm1c <- 3*(x[,"n_02"] + x[,"n_04"] + x[,"n_62"] + x[,"n_64"])*log10(2) + 4*(x[,"n_01"] + x[,"n_05"] + x[,"n_61"] + x[,"n_65"])*log10(2) + (x[,"n_03"] + x[,"n_63"])*(3*log10(2) - log10(3)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_52"] + x[,"n_54"])*(-2*log10(2) + 2*log10(3)) - (x[,"n_31"] + x[,"n_35"])*(-3*log10(2) + 2*log10(5)) - (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(-4*log10(2) + log10(7)) - (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*(-4*log10(2) + 2*log10(3) + log10(7)) - x[,"n_33"]*(-4*log10(2) + 2*log10(7)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*(-3*log10(2) + log10(5) + log10(19)) - (x[,"n_13"] + x[,"n_53"])*(-3*log10(2) + log10(29)) - (x[,"n_23"] + x[,"n_43"])*(-3*log10(2) + log10(349)) + (x[,"n_01"] + x[,"n_05"] + x[,"n_61"] + x[,"n_65"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log10(pmax(1e-6,r_cm1c)) + log10(2 - 3*r_cm1c + 2*r_cm1c^2 - r_cm1c^3)) + (x[,"n_13"] + x[,"n_53"])*log10(3 + 4*r_cm1c - 10*r_cm1c^2 + 12*r_cm1c^3 - 6*r_cm1c^4) + (x[,"n_23"] + x[,"n_43"])*log10(45 - 4*r_cm1c - 2*r_cm1c^2 + 12*r_cm1c^3 - 6*r_cm1c^4) + (x[,"n_32"] + x[,"n_34"])*log10(9 + 4*r_cm1c - 8*r_cm1c^2 + 8*r_cm1c^3 - 4*r_cm1c^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_62"] + x[,"n_64"])*log10(r_cm1c - 3*r_cm1c^2 + 4*r_cm1c^3 - 2*r_cm1c^4) + (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*log10(3 + 4*r_cm1c - 5*r_cm1c^2 + 2*r_cm1c^3 - r_cm1c^4) + x[,"n_33"]*log10(3 + r_cm1c^2 - 2*r_cm1c^3 + r_cm1c^4) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log10(12 - r_cm1c + 3*r_cm1c^2 - 4*r_cm1c^3 + 2*r_cm1c^4) + (x[,"n_31"] + x[,"n_35"])*log10(5 - 8*r_cm1c + 10*r_cm1c^2 - 4*r_cm1c^3 + 2*r_cm1c^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_52"] + x[,"n_54"])*log10(3 - 4*r_cm1c + 8*r_cm1c^2 - 8*r_cm1c^3 + 4*r_cm1c^4) + (x[,"n_03"] + x[,"n_63"])*log10(1 - 4*r_cm1c + 10*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4)
logL_cm1c <- (-4*x[,"n_01"] - 3*x[,"n_02"] - 4*x[,"n_03"] - 3*x[,"n_04"] - 4*x[,"n_05"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 4*x[,"n_23"] - 3*x[,"n_24"] - 4*x[,"n_25"] - 3*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 4*x[,"n_43"] - 3*x[,"n_44"] - 4*x[,"n_45"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 4*x[,"n_61"] - 3*x[,"n_62"] - 4*x[,"n_63"] - 3*x[,"n_64"] - 4*x[,"n_65"])*log(2) + x[,"n_33"]*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"])*log(5) + (x[,"n_01"] + x[,"n_05"] + x[,"n_61"] + x[,"n_65"])*(2*log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log(pmax(1e-6,r_cm1c)) + log(2 - 3*r_cm1c + 2*r_cm1c^2 - r_cm1c^3)) + (x[,"n_13"] + x[,"n_53"])*log(3 + 4*r_cm1c - 10*r_cm1c^2 + 12*r_cm1c^3 - 6*r_cm1c^4) + (x[,"n_23"] + x[,"n_43"])*log(45 - 4*r_cm1c - 2*r_cm1c^2 + 12*r_cm1c^3 - 6*r_cm1c^4) + (x[,"n_32"] + x[,"n_34"])*log(9 + 4*r_cm1c - 8*r_cm1c^2 + 8*r_cm1c^3 - 4*r_cm1c^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_62"] + x[,"n_64"])*log(r_cm1c - 3*r_cm1c^2 + 4*r_cm1c^3 - 2*r_cm1c^4) + (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*log(3 + 4*r_cm1c - 5*r_cm1c^2 + 2*r_cm1c^3 - r_cm1c^4) + x[,"n_33"]*log(3 + r_cm1c^2 - 2*r_cm1c^3 + r_cm1c^4) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log(12 - r_cm1c + 3*r_cm1c^2 - 4*r_cm1c^3 + 2*r_cm1c^4) + (x[,"n_31"] + x[,"n_35"])*log(5 - 8*r_cm1c + 10*r_cm1c^2 - 4*r_cm1c^3 + 2*r_cm1c^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_52"] + x[,"n_54"])*log(3 - 4*r_cm1c + 8*r_cm1c^2 - 8*r_cm1c^3 + 4*r_cm1c^4) + (x[,"n_03"] + x[,"n_63"])*log(1 - 4*r_cm1c + 10*r_cm1c^2 - 12*r_cm1c^3 + 6*r_cm1c^4)
logL_mc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
L <- (-3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 3*n11 - 2*n12 - n13 - 2*n14 - 3*n15 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 3*n31 - n32 - n33 - n34 - 3*n35 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 3*n51 - 2*n52 - n53 - 2*n54 - 3*n55 - 3*n61 - 4*n62 - 2*n63 - 4*n64 - 3*n65)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55 - n61 - n62 - n63 - n64 - n65)*(log(3) + 2*log(5)) + (n01 + n65)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n05 + n61)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n15 + n51)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n11 + n55)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n04 + n62)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n02 + n64)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n14 + n52)*(log(pmax(1e-6,r)) + log(9 - 9*r + 10*r^2 - 4*r^3)) + (n03 + n63)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n25 + n41)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n24 + n42)*log(45 + 42*r + 29*r^2 - 12*r^3 - 8*r^4) + (n22 + n44)*log(96 - 32*r - 55*r^2 + 44*r^3 - 8*r^4) + n33*log(27 + 2*r - 8*r^2 + 12*r^3 - 6*r^4) + (n12 + n54)*log(6 - 5*r - 3*r^2 + 6*r^3 - 4*r^4) + (n31 + n35)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n21 + n45)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + (n13 + n53)*log(3 - r + 4*r^2 - 6*r^3 + 3*r^4) + (n23 + n43)*log(33 - 2*r + 5*r^2 - 6*r^3 + 3*r^4) + (n32 + n34)*log(15 - 2*r + 6*r^2 - 8*r^3 + 4*r^4)
return(L)}
interlogL_mc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
optimize(logL_mc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mc <- mapply(interlogL_mc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"])
LOD_mc <- 4*(x[,"n_05"] + x[,"n_61"])*log10(2) + 4*(x[,"n_01"] + x[,"n_65"])*log10(2) - (x[,"n_04"] + x[,"n_62"])*(-2*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_64"])*(-2*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_51"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_55"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_63"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_52"])*(-2*log10(2) + log10(13)) - (x[,"n_12"] + x[,"n_54"])*(-2*log10(2) + log10(13)) - (x[,"n_31"] + x[,"n_35"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_25"] + x[,"n_41"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_21"] + x[,"n_45"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - x[,"n_33"]*(-3*log10(2) + log10(7) + log10(31)) - (x[,"n_13"] + x[,"n_53"])*(-4*log10(2) + log10(47)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(59)) - (x[,"n_23"] + x[,"n_43"])*(-4*log10(2) + log10(523)) + (x[,"n_01"] + x[,"n_65"])*(3*log10(pmax(1e-6,1 - r_mc)) + log10(pmax(1e-6,r_mc))) + (x[,"n_05"] + x[,"n_61"])*(log10(pmax(1e-6,1 - r_mc)) + 3*log10(pmax(1e-6,r_mc))) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_mc)) + log10(7 - 6*r_mc + 2*r_mc^2)) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 + 2*r_mc + 2*r_mc^2)) + (x[,"n_04"] + x[,"n_62"])*(2*log10(pmax(1e-6,r_mc)) + log10(7 - 12*r_mc + 8*r_mc^2)) + (x[,"n_02"] + x[,"n_64"])*(2*log10(pmax(1e-6,1 - r_mc)) + log10(3 - 4*r_mc + 8*r_mc^2)) + (x[,"n_14"] + x[,"n_52"])*(log10(pmax(1e-6,r_mc)) + log10(9 - 9*r_mc + 10*r_mc^2 - 4*r_mc^3)) + (x[,"n_03"] + x[,"n_63"])*(log10(pmax(1e-6,r_mc)) + log10(2 - 5*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_25"] + x[,"n_41"])*(log10(pmax(1e-6,r_mc)) + log10(15 - 9*r_mc + 5*r_mc^2 + r_mc^3)) + (x[,"n_24"] + x[,"n_42"])*log10(45 + 42*r_mc + 29*r_mc^2 - 12*r_mc^3 - 8*r_mc^4) + (x[,"n_22"] + x[,"n_44"])*log10(96 - 32*r_mc - 55*r_mc^2 + 44*r_mc^3 - 8*r_mc^4) + x[,"n_33"]*log10(27 + 2*r_mc - 8*r_mc^2 + 12*r_mc^3 - 6*r_mc^4) + (x[,"n_12"] + x[,"n_54"])*log10(6 - 5*r_mc - 3*r_mc^2 + 6*r_mc^3 - 4*r_mc^4) + (x[,"n_31"] + x[,"n_35"])*log10(9 + 4*r_mc - 8*r_mc^2 + 8*r_mc^3 - 4*r_mc^4) + (x[,"n_21"] + x[,"n_45"])*log10(12 - 16*r_mc + 12*r_mc^2 - 9*r_mc^3 + r_mc^4) + (x[,"n_13"] + x[,"n_53"])*log10(3 - r_mc + 4*r_mc^2 - 6*r_mc^3 + 3*r_mc^4) + (x[,"n_23"] + x[,"n_43"])*log10(33 - 2*r_mc + 5*r_mc^2 - 6*r_mc^3 + 3*r_mc^4) + (x[,"n_32"] + x[,"n_34"])*log10(15 - 2*r_mc + 6*r_mc^2 - 8*r_mc^3 + 4*r_mc^4)
logL_mc <- (-3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 2*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"])*(log(3) + 2*log(5)) + (x[,"n_01"] + x[,"n_65"])*(3*log(pmax(1e-6,1 - r_mc)) + log(pmax(1e-6,r_mc))) + (x[,"n_05"] + x[,"n_61"])*(log(pmax(1e-6,1 - r_mc)) + 3*log(pmax(1e-6,r_mc))) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,r_mc)) + log(7 - 6*r_mc + 2*r_mc^2)) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 + 2*r_mc + 2*r_mc^2)) + (x[,"n_04"] + x[,"n_62"])*(2*log(pmax(1e-6,r_mc)) + log(7 - 12*r_mc + 8*r_mc^2)) + (x[,"n_02"] + x[,"n_64"])*(2*log(pmax(1e-6,1 - r_mc)) + log(3 - 4*r_mc + 8*r_mc^2)) + (x[,"n_14"] + x[,"n_52"])*(log(pmax(1e-6,r_mc)) + log(9 - 9*r_mc + 10*r_mc^2 - 4*r_mc^3)) + (x[,"n_03"] + x[,"n_63"])*(log(pmax(1e-6,r_mc)) + log(2 - 5*r_mc + 6*r_mc^2 - 3*r_mc^3)) + (x[,"n_25"] + x[,"n_41"])*(log(pmax(1e-6,r_mc)) + log(15 - 9*r_mc + 5*r_mc^2 + r_mc^3)) + (x[,"n_24"] + x[,"n_42"])*log(45 + 42*r_mc + 29*r_mc^2 - 12*r_mc^3 - 8*r_mc^4) + (x[,"n_22"] + x[,"n_44"])*log(96 - 32*r_mc - 55*r_mc^2 + 44*r_mc^3 - 8*r_mc^4) + x[,"n_33"]*log(27 + 2*r_mc - 8*r_mc^2 + 12*r_mc^3 - 6*r_mc^4) + (x[,"n_12"] + x[,"n_54"])*log(6 - 5*r_mc - 3*r_mc^2 + 6*r_mc^3 - 4*r_mc^4) + (x[,"n_31"] + x[,"n_35"])*log(9 + 4*r_mc - 8*r_mc^2 + 8*r_mc^3 - 4*r_mc^4) + (x[,"n_21"] + x[,"n_45"])*log(12 - 16*r_mc + 12*r_mc^2 - 9*r_mc^3 + r_mc^4) + (x[,"n_13"] + x[,"n_53"])*log(3 - r_mc + 4*r_mc^2 - 6*r_mc^3 + 3*r_mc^4) + (x[,"n_23"] + x[,"n_43"])*log(33 - 2*r_mc + 5*r_mc^2 - 6*r_mc^3 + 3*r_mc^4) + (x[,"n_32"] + x[,"n_34"])*log(15 - 2*r_mc + 6*r_mc^2 - 8*r_mc^3 + 4*r_mc^4)
logL_mm2c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
L <- (-2*n01 - 2*n02 - 4*n03 - 2*n04 - 2*n05 - n11 - 2*n12 - 3*n13 - 2*n14 - n15 - 2*n21 - 2*n22 - 4*n23 - 2*n24 - 2*n25 - n31 - n32 - 2*n33 - n34 - n35 - 2*n41 - 2*n42 - 4*n43 - 2*n44 - 2*n45 - n51 - 2*n52 - 3*n53 - 2*n54 - n55 - 2*n61 - 2*n62 - 4*n63 - 2*n64 - 2*n65)*log(2) + (-2*n01 - 2*n02 - n03 - 2*n04 - 2*n05 - 2*n11 - 2*n12 - n13 - 2*n14 - 2*n15 - 2*n21 - 2*n22 - n23 - 2*n24 - 2*n25 - 2*n31 - 2*n32 - n33 - 2*n34 - 2*n35 - 2*n41 - 2*n42 - n43 - 2*n44 - 2*n45 - 2*n51 - 2*n52 - n53 - 2*n54 - 2*n55 - 2*n61 - 2*n62 - n63 - 2*n64 - 2*n65)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55 - n61 - n62 - n63 - n64 - n65)*log(5) + (n01 + n05 + n61 + n65)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n15 + n51 + n55)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - r + r^2)) + (n02 + n04 + n62 + n64)*(log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - 4*r + 4*r^2)) + (n13 + n53)*log(15 - 8*r + 16*r^3 - 8*r^4) + (n23 + n43)*log(129 + 8*r - 16*r^2 + 16*r^3 - 8*r^4) + (n32 + n34)*log(45 - 8*r^2 + 16*r^3 - 8*r^4) + (n21 + n25 + n41 + n45)*log(9 - r^2 + 2*r^3 - r^4) + (n31 + n35)*log(9 - 6*r + 8*r^2 - 4*r^3 + 2*r^4) + (n22 + n24 + n42 + n44)*log(54 - 3*r + 7*r^2 - 8*r^3 + 4*r^4) + n33*log(51 + 8*r - 16*r^3 + 8*r^4) + (n12 + n14 + n52 + n54)*log(9 + 8*r^2 - 16*r^3 + 8*r^4) + (n03 + n63)*log(3 - 8*r + 16*r^2 - 16*r^3 + 8*r^4)
return(L)}
interlogL_mm2c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
optimize(logL_mm2c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm2c <- mapply(interlogL_mm2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"])
LOD_mm2c <- (x[,"n_02"] + x[,"n_04"] + x[,"n_62"] + x[,"n_64"])*log10(2) + 4*(x[,"n_01"] + x[,"n_05"] + x[,"n_61"] + x[,"n_65"])*log10(2) - (x[,"n_03"] + x[,"n_63"])*(-log10(2) + log10(3)) - (x[,"n_13"] + x[,"n_53"])*(-log10(2) + 2*log10(5)) - (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(-4*log10(2) + log10(11)) - (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*(-4*log10(2) + log10(11) + log10(13)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_52"] + x[,"n_54"])*(-log10(2) + log10(19)) - (x[,"n_23"] + x[,"n_43"])*(-log10(2) + 2*log10(3) + log10(29)) - (x[,"n_31"] + x[,"n_35"])*(-3*log10(2) + log10(61)) - (x[,"n_32"] + x[,"n_34"])*(-log10(2) + log10(89)) - x[,"n_33"]*(-log10(2) + log10(107)) - (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*(-log10(2) + log10(107)) + (x[,"n_01"] + x[,"n_05"] + x[,"n_61"] + x[,"n_65"])*(2*log10(pmax(1e-6,1 - r_mm2c)) + 2*log10(pmax(1e-6,r_mm2c))) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log10(pmax(1e-6,1 - r_mm2c)) + log10(pmax(1e-6,r_mm2c)) + log10(3 - r_mm2c + r_mm2c^2)) + (x[,"n_02"] + x[,"n_04"] + x[,"n_62"] + x[,"n_64"])*(log10(pmax(1e-6,1 - r_mm2c)) + log10(pmax(1e-6,r_mm2c)) + log10(3 - 4*r_mm2c + 4*r_mm2c^2)) + (x[,"n_13"] + x[,"n_53"])*log10(15 - 8*r_mm2c + 16*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_23"] + x[,"n_43"])*log10(129 + 8*r_mm2c - 16*r_mm2c^2 + 16*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_32"] + x[,"n_34"])*log10(45 - 8*r_mm2c^2 + 16*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*log10(9 - r_mm2c^2 + 2*r_mm2c^3 - r_mm2c^4) + (x[,"n_31"] + x[,"n_35"])*log10(9 - 6*r_mm2c + 8*r_mm2c^2 - 4*r_mm2c^3 + 2*r_mm2c^4) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log10(54 - 3*r_mm2c + 7*r_mm2c^2 - 8*r_mm2c^3 + 4*r_mm2c^4) + x[,"n_33"]*log10(51 + 8*r_mm2c - 16*r_mm2c^3 + 8*r_mm2c^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_52"] + x[,"n_54"])*log10(9 + 8*r_mm2c^2 - 16*r_mm2c^3 + 8*r_mm2c^4) + (x[,"n_03"] + x[,"n_63"])*log10(3 - 8*r_mm2c + 16*r_mm2c^2 - 16*r_mm2c^3 + 8*r_mm2c^4)
logL_mm2c <- (-2*x[,"n_01"] - 2*x[,"n_02"] - 4*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - x[,"n_11"] - 2*x[,"n_12"] - 3*x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 4*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - 2*x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 4*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - x[,"n_51"] - 2*x[,"n_52"] - 3*x[,"n_53"] - 2*x[,"n_54"] - x[,"n_55"] - 2*x[,"n_61"] - 2*x[,"n_62"] - 4*x[,"n_63"] - 2*x[,"n_64"] - 2*x[,"n_65"])*log(2) + (-2*x[,"n_01"] - 2*x[,"n_02"] - x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_21"] - 2*x[,"n_22"] - x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_41"] - 2*x[,"n_42"] - x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"] - 2*x[,"n_61"] - 2*x[,"n_62"] - x[,"n_63"] - 2*x[,"n_64"] - 2*x[,"n_65"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"])*log(5) + (x[,"n_01"] + x[,"n_05"] + x[,"n_61"] + x[,"n_65"])*(2*log(pmax(1e-6,1 - r_mm2c)) + 2*log(pmax(1e-6,r_mm2c))) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log(pmax(1e-6,1 - r_mm2c)) + log(pmax(1e-6,r_mm2c)) + log(3 - r_mm2c + r_mm2c^2)) + (x[,"n_02"] + x[,"n_04"] + x[,"n_62"] + x[,"n_64"])*(log(pmax(1e-6,1 - r_mm2c)) + log(pmax(1e-6,r_mm2c)) + log(3 - 4*r_mm2c + 4*r_mm2c^2)) + (x[,"n_13"] + x[,"n_53"])*log(15 - 8*r_mm2c + 16*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_23"] + x[,"n_43"])*log(129 + 8*r_mm2c - 16*r_mm2c^2 + 16*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_32"] + x[,"n_34"])*log(45 - 8*r_mm2c^2 + 16*r_mm2c^3 - 8*r_mm2c^4) + (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*log(9 - r_mm2c^2 + 2*r_mm2c^3 - r_mm2c^4) + (x[,"n_31"] + x[,"n_35"])*log(9 - 6*r_mm2c + 8*r_mm2c^2 - 4*r_mm2c^3 + 2*r_mm2c^4) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log(54 - 3*r_mm2c + 7*r_mm2c^2 - 8*r_mm2c^3 + 4*r_mm2c^4) + x[,"n_33"]*log(51 + 8*r_mm2c - 16*r_mm2c^3 + 8*r_mm2c^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_52"] + x[,"n_54"])*log(9 + 8*r_mm2c^2 - 16*r_mm2c^3 + 8*r_mm2c^4) + (x[,"n_03"] + x[,"n_63"])*log(3 - 8*r_mm2c + 16*r_mm2c^2 - 16*r_mm2c^3 + 8*r_mm2c^4)
logL_mm1c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
L <- (-3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 3*n11 - 2*n12 - n13 - 2*n14 - 3*n15 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 3*n31 - n32 - n33 - n34 - 3*n35 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 3*n51 - 2*n52 - n53 - 2*n54 - 3*n55 - 3*n61 - 4*n62 - 2*n63 - 4*n64 - 3*n65)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55 - n61 - n62 - n63 - n64 - n65)*(log(3) + 2*log(5)) + (n05 + n61)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n65)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n11 + n55)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n15 + n51)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n02 + n64)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n04 + n62)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n12 + n54)*(log(pmax(1e-6,r)) + log(9 - 9*r + 10*r^2 - 4*r^3)) + (n03 + n63)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n21 + n45)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n22 + n44)*log(45 + 42*r + 29*r^2 - 12*r^3 - 8*r^4) + (n24 + n42)*log(96 - 32*r - 55*r^2 + 44*r^3 - 8*r^4) + n33*log(27 + 2*r - 8*r^2 + 12*r^3 - 6*r^4) + (n14 + n52)*log(6 - 5*r - 3*r^2 + 6*r^3 - 4*r^4) + (n31 + n35)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n25 + n41)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + (n13 + n53)*log(3 - r + 4*r^2 - 6*r^3 + 3*r^4) + (n23 + n43)*log(33 - 2*r + 5*r^2 - 6*r^3 + 3*r^4) + (n32 + n34)*log(15 - 2*r + 6*r^2 - 8*r^3 + 4*r^4)
return(L)}
interlogL_mm1c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
optimize(logL_mm1c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_mm1c <- mapply(interlogL_mm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"])
LOD_mm1c <- 4*(x[,"n_05"] + x[,"n_61"])*log10(2) + 4*(x[,"n_01"] + x[,"n_65"])*log10(2) - (x[,"n_04"] + x[,"n_62"])*(-2*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_64"])*(-2*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_51"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_55"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_63"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_52"])*(-2*log10(2) + log10(13)) - (x[,"n_12"] + x[,"n_54"])*(-2*log10(2) + log10(13)) - (x[,"n_31"] + x[,"n_35"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_25"] + x[,"n_41"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_21"] + x[,"n_45"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - x[,"n_33"]*(-3*log10(2) + log10(7) + log10(31)) - (x[,"n_13"] + x[,"n_53"])*(-4*log10(2) + log10(47)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(59)) - (x[,"n_23"] + x[,"n_43"])*(-4*log10(2) + log10(523)) + (x[,"n_05"] + x[,"n_61"])*(3*log10(pmax(1e-6,1 - r_mm1c)) + log10(pmax(1e-6,r_mm1c))) + (x[,"n_01"] + x[,"n_65"])*(log10(pmax(1e-6,1 - r_mm1c)) + 3*log10(pmax(1e-6,r_mm1c))) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,r_mm1c)) + log10(7 - 6*r_mm1c + 2*r_mm1c^2)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_mm1c)) + log10(3 + 2*r_mm1c + 2*r_mm1c^2)) + (x[,"n_02"] + x[,"n_64"])*(2*log10(pmax(1e-6,r_mm1c)) + log10(7 - 12*r_mm1c + 8*r_mm1c^2)) + (x[,"n_04"] + x[,"n_62"])*(2*log10(pmax(1e-6,1 - r_mm1c)) + log10(3 - 4*r_mm1c + 8*r_mm1c^2)) + (x[,"n_12"] + x[,"n_54"])*(log10(pmax(1e-6,r_mm1c)) + log10(9 - 9*r_mm1c + 10*r_mm1c^2 - 4*r_mm1c^3)) + (x[,"n_03"] + x[,"n_63"])*(log10(pmax(1e-6,r_mm1c)) + log10(2 - 5*r_mm1c + 6*r_mm1c^2 - 3*r_mm1c^3)) + (x[,"n_21"] + x[,"n_45"])*(log10(pmax(1e-6,r_mm1c)) + log10(15 - 9*r_mm1c + 5*r_mm1c^2 + r_mm1c^3)) + (x[,"n_22"] + x[,"n_44"])*log10(45 + 42*r_mm1c + 29*r_mm1c^2 - 12*r_mm1c^3 - 8*r_mm1c^4) + (x[,"n_24"] + x[,"n_42"])*log10(96 - 32*r_mm1c - 55*r_mm1c^2 + 44*r_mm1c^3 - 8*r_mm1c^4) + x[,"n_33"]*log10(27 + 2*r_mm1c - 8*r_mm1c^2 + 12*r_mm1c^3 - 6*r_mm1c^4) + (x[,"n_14"] + x[,"n_52"])*log10(6 - 5*r_mm1c - 3*r_mm1c^2 + 6*r_mm1c^3 - 4*r_mm1c^4) + (x[,"n_31"] + x[,"n_35"])*log10(9 + 4*r_mm1c - 8*r_mm1c^2 + 8*r_mm1c^3 - 4*r_mm1c^4) + (x[,"n_25"] + x[,"n_41"])*log10(12 - 16*r_mm1c + 12*r_mm1c^2 - 9*r_mm1c^3 + r_mm1c^4) + (x[,"n_13"] + x[,"n_53"])*log10(3 - r_mm1c + 4*r_mm1c^2 - 6*r_mm1c^3 + 3*r_mm1c^4) + (x[,"n_23"] + x[,"n_43"])*log10(33 - 2*r_mm1c + 5*r_mm1c^2 - 6*r_mm1c^3 + 3*r_mm1c^4) + (x[,"n_32"] + x[,"n_34"])*log10(15 - 2*r_mm1c + 6*r_mm1c^2 - 8*r_mm1c^3 + 4*r_mm1c^4)
logL_mm1c <- (-3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 2*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_61"])*(3*log(pmax(1e-6,1 - r_mm1c)) + log(pmax(1e-6,r_mm1c))) + (x[,"n_01"] + x[,"n_65"])*(log(pmax(1e-6,1 - r_mm1c)) + 3*log(pmax(1e-6,r_mm1c))) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,r_mm1c)) + log(7 - 6*r_mm1c + 2*r_mm1c^2)) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_mm1c)) + log(3 + 2*r_mm1c + 2*r_mm1c^2)) + (x[,"n_02"] + x[,"n_64"])*(2*log(pmax(1e-6,r_mm1c)) + log(7 - 12*r_mm1c + 8*r_mm1c^2)) + (x[,"n_04"] + x[,"n_62"])*(2*log(pmax(1e-6,1 - r_mm1c)) + log(3 - 4*r_mm1c + 8*r_mm1c^2)) + (x[,"n_12"] + x[,"n_54"])*(log(pmax(1e-6,r_mm1c)) + log(9 - 9*r_mm1c + 10*r_mm1c^2 - 4*r_mm1c^3)) + (x[,"n_03"] + x[,"n_63"])*(log(pmax(1e-6,r_mm1c)) + log(2 - 5*r_mm1c + 6*r_mm1c^2 - 3*r_mm1c^3)) + (x[,"n_21"] + x[,"n_45"])*(log(pmax(1e-6,r_mm1c)) + log(15 - 9*r_mm1c + 5*r_mm1c^2 + r_mm1c^3)) + (x[,"n_22"] + x[,"n_44"])*log(45 + 42*r_mm1c + 29*r_mm1c^2 - 12*r_mm1c^3 - 8*r_mm1c^4) + (x[,"n_24"] + x[,"n_42"])*log(96 - 32*r_mm1c - 55*r_mm1c^2 + 44*r_mm1c^3 - 8*r_mm1c^4) + x[,"n_33"]*log(27 + 2*r_mm1c - 8*r_mm1c^2 + 12*r_mm1c^3 - 6*r_mm1c^4) + (x[,"n_14"] + x[,"n_52"])*log(6 - 5*r_mm1c - 3*r_mm1c^2 + 6*r_mm1c^3 - 4*r_mm1c^4) + (x[,"n_31"] + x[,"n_35"])*log(9 + 4*r_mm1c - 8*r_mm1c^2 + 8*r_mm1c^3 - 4*r_mm1c^4) + (x[,"n_25"] + x[,"n_41"])*log(12 - 16*r_mm1c + 12*r_mm1c^2 - 9*r_mm1c^3 + r_mm1c^4) + (x[,"n_13"] + x[,"n_53"])*log(3 - r_mm1c + 4*r_mm1c^2 - 6*r_mm1c^3 + 3*r_mm1c^4) + (x[,"n_23"] + x[,"n_43"])*log(33 - 2*r_mm1c + 5*r_mm1c^2 - 6*r_mm1c^3 + 3*r_mm1c^4) + (x[,"n_32"] + x[,"n_34"])*log(15 - 2*r_mm1c + 6*r_mm1c^2 - 8*r_mm1c^3 + 4*r_mm1c^4)
logL_rc <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
L <- (-4*n01 - 3*n02 - 4*n03 - 3*n04 - 4*n05 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 4*n21 - 3*n22 - 4*n23 - 3*n24 - 4*n25 - 3*n31 - 2*n32 - n33 - 2*n34 - 3*n35 - 4*n41 - 3*n42 - 4*n43 - 3*n44 - 4*n45 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55 - 4*n61 - 3*n62 - 4*n63 - 3*n64 - 4*n65)*log(2) + n33*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55 - n61 - n62 - n63 - n64 - n65)*log(5) + (n01 + n05 + n61 + n65)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n11 + n15 + n51 + n55)*(log(pmax(1e-6,r)) + log(2 - 3*r + 2*r^2 - r^3)) + (n13 + n53)*log(3 + 4*r - 10*r^2 + 12*r^3 - 6*r^4) + (n23 + n43)*log(45 - 4*r - 2*r^2 + 12*r^3 - 6*r^4) + (n32 + n34)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n02 + n04 + n62 + n64)*log(r - 3*r^2 + 4*r^3 - 2*r^4) + (n21 + n25 + n41 + n45)*log(3 + 4*r - 5*r^2 + 2*r^3 - r^4) + n33*log(3 + r^2 - 2*r^3 + r^4) + (n22 + n24 + n42 + n44)*log(12 - r + 3*r^2 - 4*r^3 + 2*r^4) + (n31 + n35)*log(5 - 8*r + 10*r^2 - 4*r^3 + 2*r^4) + (n12 + n14 + n52 + n54)*log(3 - 4*r + 8*r^2 - 8*r^3 + 4*r^4) + (n03 + n63)*log(1 - 4*r + 10*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_rc <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
optimize(logL_rc,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rc <- mapply(interlogL_rc,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"])
LOD_rc <- 3*(x[,"n_02"] + x[,"n_04"] + x[,"n_62"] + x[,"n_64"])*log10(2) + 4*(x[,"n_01"] + x[,"n_05"] + x[,"n_61"] + x[,"n_65"])*log10(2) + (x[,"n_03"] + x[,"n_63"])*(3*log10(2) - log10(3)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_52"] + x[,"n_54"])*(-2*log10(2) + 2*log10(3)) - (x[,"n_31"] + x[,"n_35"])*(-3*log10(2) + 2*log10(5)) - (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(-4*log10(2) + log10(7)) - (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*(-4*log10(2) + 2*log10(3) + log10(7)) - x[,"n_33"]*(-4*log10(2) + 2*log10(7)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*(-3*log10(2) + log10(5) + log10(19)) - (x[,"n_13"] + x[,"n_53"])*(-3*log10(2) + log10(29)) - (x[,"n_23"] + x[,"n_43"])*(-3*log10(2) + log10(349)) + (x[,"n_01"] + x[,"n_05"] + x[,"n_61"] + x[,"n_65"])*(2*log10(pmax(1e-6,1 - r_rc)) + 2*log10(pmax(1e-6,r_rc))) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log10(pmax(1e-6,r_rc)) + log10(2 - 3*r_rc + 2*r_rc^2 - r_rc^3)) + (x[,"n_13"] + x[,"n_53"])*log10(3 + 4*r_rc - 10*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_23"] + x[,"n_43"])*log10(45 - 4*r_rc - 2*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_32"] + x[,"n_34"])*log10(9 + 4*r_rc - 8*r_rc^2 + 8*r_rc^3 - 4*r_rc^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_62"] + x[,"n_64"])*log10(r_rc - 3*r_rc^2 + 4*r_rc^3 - 2*r_rc^4) + (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*log10(3 + 4*r_rc - 5*r_rc^2 + 2*r_rc^3 - r_rc^4) + x[,"n_33"]*log10(3 + r_rc^2 - 2*r_rc^3 + r_rc^4) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log10(12 - r_rc + 3*r_rc^2 - 4*r_rc^3 + 2*r_rc^4) + (x[,"n_31"] + x[,"n_35"])*log10(5 - 8*r_rc + 10*r_rc^2 - 4*r_rc^3 + 2*r_rc^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_52"] + x[,"n_54"])*log10(3 - 4*r_rc + 8*r_rc^2 - 8*r_rc^3 + 4*r_rc^4) + (x[,"n_03"] + x[,"n_63"])*log10(1 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4)
logL_rc <- (-4*x[,"n_01"] - 3*x[,"n_02"] - 4*x[,"n_03"] - 3*x[,"n_04"] - 4*x[,"n_05"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_21"] - 3*x[,"n_22"] - 4*x[,"n_23"] - 3*x[,"n_24"] - 4*x[,"n_25"] - 3*x[,"n_31"] - 2*x[,"n_32"] - x[,"n_33"] - 2*x[,"n_34"] - 3*x[,"n_35"] - 4*x[,"n_41"] - 3*x[,"n_42"] - 4*x[,"n_43"] - 3*x[,"n_44"] - 4*x[,"n_45"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 4*x[,"n_61"] - 3*x[,"n_62"] - 4*x[,"n_63"] - 3*x[,"n_64"] - 4*x[,"n_65"])*log(2) + x[,"n_33"]*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"])*log(5) + (x[,"n_01"] + x[,"n_05"] + x[,"n_61"] + x[,"n_65"])*(2*log(pmax(1e-6,1 - r_rc)) + 2*log(pmax(1e-6,r_rc))) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log(pmax(1e-6,r_rc)) + log(2 - 3*r_rc + 2*r_rc^2 - r_rc^3)) + (x[,"n_13"] + x[,"n_53"])*log(3 + 4*r_rc - 10*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_23"] + x[,"n_43"])*log(45 - 4*r_rc - 2*r_rc^2 + 12*r_rc^3 - 6*r_rc^4) + (x[,"n_32"] + x[,"n_34"])*log(9 + 4*r_rc - 8*r_rc^2 + 8*r_rc^3 - 4*r_rc^4) + (x[,"n_02"] + x[,"n_04"] + x[,"n_62"] + x[,"n_64"])*log(r_rc - 3*r_rc^2 + 4*r_rc^3 - 2*r_rc^4) + (x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"])*log(3 + 4*r_rc - 5*r_rc^2 + 2*r_rc^3 - r_rc^4) + x[,"n_33"]*log(3 + r_rc^2 - 2*r_rc^3 + r_rc^4) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log(12 - r_rc + 3*r_rc^2 - 4*r_rc^3 + 2*r_rc^4) + (x[,"n_31"] + x[,"n_35"])*log(5 - 8*r_rc + 10*r_rc^2 - 4*r_rc^3 + 2*r_rc^4) + (x[,"n_12"] + x[,"n_14"] + x[,"n_52"] + x[,"n_54"])*log(3 - 4*r_rc + 8*r_rc^2 - 8*r_rc^3 + 4*r_rc^4) + (x[,"n_03"] + x[,"n_63"])*log(1 - 4*r_rc + 10*r_rc^2 - 12*r_rc^3 + 6*r_rc^4)
logL_rm2c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
L <- (-3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 3*n11 - 2*n12 - n13 - 2*n14 - 3*n15 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 3*n31 - n32 - n33 - n34 - 3*n35 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 3*n51 - 2*n52 - n53 - 2*n54 - 3*n55 - 3*n61 - 4*n62 - 2*n63 - 4*n64 - 3*n65)*log(2) + (-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55 - n61 - n62 - n63 - n64 - n65)*(log(3) + 2*log(5)) + (n05 + n61)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n01 + n65)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n11 + n55)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n15 + n51)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n02 + n64)*(2*log(pmax(1e-6,r)) + log(7 - 12*r + 8*r^2)) + (n04 + n62)*(2*log(pmax(1e-6,1 - r)) + log(3 - 4*r + 8*r^2)) + (n12 + n54)*(log(pmax(1e-6,r)) + log(9 - 9*r + 10*r^2 - 4*r^3)) + (n03 + n63)*(log(pmax(1e-6,r)) + log(2 - 5*r + 6*r^2 - 3*r^3)) + (n21 + n45)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n22 + n44)*log(45 + 42*r + 29*r^2 - 12*r^3 - 8*r^4) + (n24 + n42)*log(96 - 32*r - 55*r^2 + 44*r^3 - 8*r^4) + n33*log(27 + 2*r - 8*r^2 + 12*r^3 - 6*r^4) + (n14 + n52)*log(6 - 5*r - 3*r^2 + 6*r^3 - 4*r^4) + (n31 + n35)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n25 + n41)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + (n13 + n53)*log(3 - r + 4*r^2 - 6*r^3 + 3*r^4) + (n23 + n43)*log(33 - 2*r + 5*r^2 - 6*r^3 + 3*r^4) + (n32 + n34)*log(15 - 2*r + 6*r^2 - 8*r^3 + 4*r^4)
return(L)}
interlogL_rm2c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
optimize(logL_rm2c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm2c <- mapply(interlogL_rm2c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"])
LOD_rm2c <- 4*(x[,"n_05"] + x[,"n_61"])*log10(2) + 4*(x[,"n_01"] + x[,"n_65"])*log10(2) - (x[,"n_04"] + x[,"n_62"])*(-2*log10(2) + log10(3)) - (x[,"n_02"] + x[,"n_64"])*(-2*log10(2) + log10(3)) - (x[,"n_15"] + x[,"n_51"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_11"] + x[,"n_55"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_03"] + x[,"n_63"])*(-4*log10(2) + log10(5)) - (x[,"n_14"] + x[,"n_52"])*(-2*log10(2) + log10(13)) - (x[,"n_12"] + x[,"n_54"])*(-2*log10(2) + log10(13)) - (x[,"n_31"] + x[,"n_35"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_25"] + x[,"n_41"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_21"] + x[,"n_45"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_24"] + x[,"n_42"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - (x[,"n_22"] + x[,"n_44"])*(-2*log10(2) + log10(3) + log10(5) + log10(19)) - x[,"n_33"]*(-3*log10(2) + log10(7) + log10(31)) - (x[,"n_13"] + x[,"n_53"])*(-4*log10(2) + log10(47)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(59)) - (x[,"n_23"] + x[,"n_43"])*(-4*log10(2) + log10(523)) + (x[,"n_05"] + x[,"n_61"])*(3*log10(pmax(1e-6,1 - r_rm2c)) + log10(pmax(1e-6,r_rm2c))) + (x[,"n_01"] + x[,"n_65"])*(log10(pmax(1e-6,1 - r_rm2c)) + 3*log10(pmax(1e-6,r_rm2c))) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,r_rm2c)) + log10(7 - 6*r_rm2c + 2*r_rm2c^2)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + log10(3 + 2*r_rm2c + 2*r_rm2c^2)) + (x[,"n_02"] + x[,"n_64"])*(2*log10(pmax(1e-6,r_rm2c)) + log10(7 - 12*r_rm2c + 8*r_rm2c^2)) + (x[,"n_04"] + x[,"n_62"])*(2*log10(pmax(1e-6,1 - r_rm2c)) + log10(3 - 4*r_rm2c + 8*r_rm2c^2)) + (x[,"n_12"] + x[,"n_54"])*(log10(pmax(1e-6,r_rm2c)) + log10(9 - 9*r_rm2c + 10*r_rm2c^2 - 4*r_rm2c^3)) + (x[,"n_03"] + x[,"n_63"])*(log10(pmax(1e-6,r_rm2c)) + log10(2 - 5*r_rm2c + 6*r_rm2c^2 - 3*r_rm2c^3)) + (x[,"n_21"] + x[,"n_45"])*(log10(pmax(1e-6,r_rm2c)) + log10(15 - 9*r_rm2c + 5*r_rm2c^2 + r_rm2c^3)) + (x[,"n_22"] + x[,"n_44"])*log10(45 + 42*r_rm2c + 29*r_rm2c^2 - 12*r_rm2c^3 - 8*r_rm2c^4) + (x[,"n_24"] + x[,"n_42"])*log10(96 - 32*r_rm2c - 55*r_rm2c^2 + 44*r_rm2c^3 - 8*r_rm2c^4) + x[,"n_33"]*log10(27 + 2*r_rm2c - 8*r_rm2c^2 + 12*r_rm2c^3 - 6*r_rm2c^4) + (x[,"n_14"] + x[,"n_52"])*log10(6 - 5*r_rm2c - 3*r_rm2c^2 + 6*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_31"] + x[,"n_35"])*log10(9 + 4*r_rm2c - 8*r_rm2c^2 + 8*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_25"] + x[,"n_41"])*log10(12 - 16*r_rm2c + 12*r_rm2c^2 - 9*r_rm2c^3 + r_rm2c^4) + (x[,"n_13"] + x[,"n_53"])*log10(3 - r_rm2c + 4*r_rm2c^2 - 6*r_rm2c^3 + 3*r_rm2c^4) + (x[,"n_23"] + x[,"n_43"])*log10(33 - 2*r_rm2c + 5*r_rm2c^2 - 6*r_rm2c^3 + 3*r_rm2c^4) + (x[,"n_32"] + x[,"n_34"])*log10(15 - 2*r_rm2c + 6*r_rm2c^2 - 8*r_rm2c^3 + 4*r_rm2c^4)
logL_rm2c <- (-3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 3*x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 3*x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 3*x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 2*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"])*log(2) + (-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"])*(log(3) + 2*log(5)) + (x[,"n_05"] + x[,"n_61"])*(3*log(pmax(1e-6,1 - r_rm2c)) + log(pmax(1e-6,r_rm2c))) + (x[,"n_01"] + x[,"n_65"])*(log(pmax(1e-6,1 - r_rm2c)) + 3*log(pmax(1e-6,r_rm2c))) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,r_rm2c)) + log(7 - 6*r_rm2c + 2*r_rm2c^2)) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_rm2c)) + log(3 + 2*r_rm2c + 2*r_rm2c^2)) + (x[,"n_02"] + x[,"n_64"])*(2*log(pmax(1e-6,r_rm2c)) + log(7 - 12*r_rm2c + 8*r_rm2c^2)) + (x[,"n_04"] + x[,"n_62"])*(2*log(pmax(1e-6,1 - r_rm2c)) + log(3 - 4*r_rm2c + 8*r_rm2c^2)) + (x[,"n_12"] + x[,"n_54"])*(log(pmax(1e-6,r_rm2c)) + log(9 - 9*r_rm2c + 10*r_rm2c^2 - 4*r_rm2c^3)) + (x[,"n_03"] + x[,"n_63"])*(log(pmax(1e-6,r_rm2c)) + log(2 - 5*r_rm2c + 6*r_rm2c^2 - 3*r_rm2c^3)) + (x[,"n_21"] + x[,"n_45"])*(log(pmax(1e-6,r_rm2c)) + log(15 - 9*r_rm2c + 5*r_rm2c^2 + r_rm2c^3)) + (x[,"n_22"] + x[,"n_44"])*log(45 + 42*r_rm2c + 29*r_rm2c^2 - 12*r_rm2c^3 - 8*r_rm2c^4) + (x[,"n_24"] + x[,"n_42"])*log(96 - 32*r_rm2c - 55*r_rm2c^2 + 44*r_rm2c^3 - 8*r_rm2c^4) + x[,"n_33"]*log(27 + 2*r_rm2c - 8*r_rm2c^2 + 12*r_rm2c^3 - 6*r_rm2c^4) + (x[,"n_14"] + x[,"n_52"])*log(6 - 5*r_rm2c - 3*r_rm2c^2 + 6*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_31"] + x[,"n_35"])*log(9 + 4*r_rm2c - 8*r_rm2c^2 + 8*r_rm2c^3 - 4*r_rm2c^4) + (x[,"n_25"] + x[,"n_41"])*log(12 - 16*r_rm2c + 12*r_rm2c^2 - 9*r_rm2c^3 + r_rm2c^4) + (x[,"n_13"] + x[,"n_53"])*log(3 - r_rm2c + 4*r_rm2c^2 - 6*r_rm2c^3 + 3*r_rm2c^4) + (x[,"n_23"] + x[,"n_43"])*log(33 - 2*r_rm2c + 5*r_rm2c^2 - 6*r_rm2c^3 + 3*r_rm2c^4) + (x[,"n_32"] + x[,"n_34"])*log(15 - 2*r_rm2c + 6*r_rm2c^2 - 8*r_rm2c^3 + 4*r_rm2c^4)
logL_rm1c <- function(r,n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
L <- (-4*n01 - 2*n02 - 3*n03 - 2*n04 - 4*n05 - 3*n11 - 2*n12 - 2*n13 - 2*n14 - 3*n15 - 4*n21 - 2*n22 - 3*n23 - 2*n24 - 4*n25 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 4*n41 - 2*n42 - 3*n43 - 2*n44 - 4*n45 - 3*n51 - 2*n52 - 2*n53 - 2*n54 - 3*n55 - 4*n61 - 2*n62 - 3*n63 - 2*n64 - 4*n65)*log(2) + (n03 + n63)*log(3) + 2*(-n01 - n02 - n03 - n04 - n05 - n11 - n12 - n13 - n14 - n15 - n21 - n22 - n23 - n24 - n25 - n31 - n32 - n33 - n34 - n35 - n41 - n42 - n43 - n44 - n45 - n51 - n52 - n53 - n54 - n55 - n61 - n62 - n63 - n64 - n65)*log(5) + 4*(n05 + n61)*log(pmax(1e-6,1 - r)) + 4*(n01 + n65)*log(pmax(1e-6,r)) + (n04 + n62)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n03 + n63)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n02 + n64)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n11 + n55)*(log(4 - r) + 3*log(pmax(1e-6,r))) + (n21 + n45)*(2*log(pmax(1e-6,r)) + log(22 - 12*r - r^2)) + (n15 + n51)*(2*log(pmax(1e-6,1 - r)) + log(3 - 2*r - r^2)) + (n25 + n41)*(2*log(pmax(1e-6,1 - r)) + log(9 + 14*r - r^2)) + (n12 + n54)*(2*log(pmax(1e-6,r)) + log(7 - 6*r + 2*r^2)) + (n14 + n52)*(2*log(pmax(1e-6,1 - r)) + log(3 + 2*r + 2*r^2)) + (n13 + n53)*(log(pmax(1e-6,r)) + log(8 - 11*r + 6*r^2 - 3*r^3)) + (n31 + n35)*(log(pmax(1e-6,r)) + log(6 - 5*r - 2*r^2 + r^3)) + (n22 + n44)*(log(pmax(1e-6,r)) + log(15 - 9*r + 5*r^2 + r^3)) + (n32 + n34)*log(9 + 4*r - 8*r^2 + 8*r^3 - 4*r^4) + (n23 + n43)*log(21 + 4*r - 7*r^2 + 6*r^3 - 3*r^4) + (n24 + n42)*log(12 - 16*r + 12*r^2 - 9*r^3 + r^4) + n33*log(23 - 20*r + 26*r^2 - 12*r^3 + 6*r^4)
return(L)}
interlogL_rm1c <- function(n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65) {
optimize(logL_rm1c,c(0,0.5), n01,n02,n03,n04,n05,n11,n12,n13,n14,n15,n21,n22,n23,n24,n25,n31,n32,n33,n34,n35,n41,n42,n43,n44,n45,n51,n52,n53,n54,n55,n61,n62,n63,n64,n65, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rm1c <- mapply(interlogL_rm1c,x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"])
LOD_rm1c <- 4*(x[,"n_05"] + x[,"n_61"])*log10(2) + 4*(x[,"n_04"] + x[,"n_62"])*log10(2) + 4*(x[,"n_03"] + x[,"n_63"])*log10(2) + 4*(x[,"n_02"] + x[,"n_64"])*log10(2) + 4*(x[,"n_01"] + x[,"n_65"])*log10(2) - (x[,"n_14"] + x[,"n_52"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_12"] + x[,"n_54"])*(-3*log10(2) + 2*log10(3)) - (x[,"n_31"] + x[,"n_35"])*(-4*log10(2) + 2*log10(5)) - (x[,"n_15"] + x[,"n_51"])*(-4*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_55"])*(-4*log10(2) + log10(7)) - (x[,"n_25"] + x[,"n_41"])*(-4*log10(2) + 2*log10(3) + log10(7)) - (x[,"n_21"] + x[,"n_45"])*(-4*log10(2) + 2*log10(3) + log10(7)) - x[,"n_33"]*(-3*log10(2) + log10(3) + 2*log10(7)) - (x[,"n_32"] + x[,"n_34"])*(-2*log10(2) + log10(3) + log10(13)) - (x[,"n_24"] + x[,"n_42"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_22"] + x[,"n_44"])*(-4*log10(2) + log10(5) + log10(19)) - (x[,"n_13"] + x[,"n_53"])*(-4*log10(2) + log10(29)) - (x[,"n_23"] + x[,"n_43"])*(-4*log10(2) + log10(349)) + 4*(x[,"n_05"] + x[,"n_61"])*log10(pmax(1e-6,1 - r_rm1c)) + 4*(x[,"n_01"] + x[,"n_65"])*log10(pmax(1e-6,r_rm1c)) + (x[,"n_04"] + x[,"n_62"])*(3*log10(pmax(1e-6,1 - r_rm1c)) + log10(pmax(1e-6,r_rm1c))) + (x[,"n_03"] + x[,"n_63"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + 2*log10(pmax(1e-6,r_rm1c))) + (x[,"n_02"] + x[,"n_64"])*(log10(pmax(1e-6,1 - r_rm1c)) + 3*log10(pmax(1e-6,r_rm1c))) + (x[,"n_11"] + x[,"n_55"])*(log10(4 - r_rm1c) + 3*log10(pmax(1e-6,r_rm1c))) + (x[,"n_21"] + x[,"n_45"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(22 - 12*r_rm1c - r_rm1c^2)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 - 2*r_rm1c - r_rm1c^2)) + (x[,"n_25"] + x[,"n_41"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(9 + 14*r_rm1c - r_rm1c^2)) + (x[,"n_12"] + x[,"n_54"])*(2*log10(pmax(1e-6,r_rm1c)) + log10(7 - 6*r_rm1c + 2*r_rm1c^2)) + (x[,"n_14"] + x[,"n_52"])*(2*log10(pmax(1e-6,1 - r_rm1c)) + log10(3 + 2*r_rm1c + 2*r_rm1c^2)) + (x[,"n_13"] + x[,"n_53"])*(log10(pmax(1e-6,r_rm1c)) + log10(8 - 11*r_rm1c + 6*r_rm1c^2 - 3*r_rm1c^3)) + (x[,"n_31"] + x[,"n_35"])*(log10(pmax(1e-6,r_rm1c)) + log10(6 - 5*r_rm1c - 2*r_rm1c^2 + r_rm1c^3)) + (x[,"n_22"] + x[,"n_44"])*(log10(pmax(1e-6,r_rm1c)) + log10(15 - 9*r_rm1c + 5*r_rm1c^2 + r_rm1c^3)) + (x[,"n_32"] + x[,"n_34"])*log10(9 + 4*r_rm1c - 8*r_rm1c^2 + 8*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_23"] + x[,"n_43"])*log10(21 + 4*r_rm1c - 7*r_rm1c^2 + 6*r_rm1c^3 - 3*r_rm1c^4) + (x[,"n_24"] + x[,"n_42"])*log10(12 - 16*r_rm1c + 12*r_rm1c^2 - 9*r_rm1c^3 + r_rm1c^4) + x[,"n_33"]*log10(23 - 20*r_rm1c + 26*r_rm1c^2 - 12*r_rm1c^3 + 6*r_rm1c^4)
logL_rm1c <- (-4*x[,"n_01"] - 2*x[,"n_02"] - 3*x[,"n_03"] - 2*x[,"n_04"] - 4*x[,"n_05"] - 3*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 3*x[,"n_15"] - 4*x[,"n_21"] - 2*x[,"n_22"] - 3*x[,"n_23"] - 2*x[,"n_24"] - 4*x[,"n_25"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 4*x[,"n_41"] - 2*x[,"n_42"] - 3*x[,"n_43"] - 2*x[,"n_44"] - 4*x[,"n_45"] - 3*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 3*x[,"n_55"] - 4*x[,"n_61"] - 2*x[,"n_62"] - 3*x[,"n_63"] - 2*x[,"n_64"] - 4*x[,"n_65"])*log(2) + (x[,"n_03"] + x[,"n_63"])*log(3) + 2*(-x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"])*log(5) + 4*(x[,"n_05"] + x[,"n_61"])*log(pmax(1e-6,1 - r_rm1c)) + 4*(x[,"n_01"] + x[,"n_65"])*log(pmax(1e-6,r_rm1c)) + (x[,"n_04"] + x[,"n_62"])*(3*log(pmax(1e-6,1 - r_rm1c)) + log(pmax(1e-6,r_rm1c))) + (x[,"n_03"] + x[,"n_63"])*(2*log(pmax(1e-6,1 - r_rm1c)) + 2*log(pmax(1e-6,r_rm1c))) + (x[,"n_02"] + x[,"n_64"])*(log(pmax(1e-6,1 - r_rm1c)) + 3*log(pmax(1e-6,r_rm1c))) + (x[,"n_11"] + x[,"n_55"])*(log(4 - r_rm1c) + 3*log(pmax(1e-6,r_rm1c))) + (x[,"n_21"] + x[,"n_45"])*(2*log(pmax(1e-6,r_rm1c)) + log(22 - 12*r_rm1c - r_rm1c^2)) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(3 - 2*r_rm1c - r_rm1c^2)) + (x[,"n_25"] + x[,"n_41"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(9 + 14*r_rm1c - r_rm1c^2)) + (x[,"n_12"] + x[,"n_54"])*(2*log(pmax(1e-6,r_rm1c)) + log(7 - 6*r_rm1c + 2*r_rm1c^2)) + (x[,"n_14"] + x[,"n_52"])*(2*log(pmax(1e-6,1 - r_rm1c)) + log(3 + 2*r_rm1c + 2*r_rm1c^2)) + (x[,"n_13"] + x[,"n_53"])*(log(pmax(1e-6,r_rm1c)) + log(8 - 11*r_rm1c + 6*r_rm1c^2 - 3*r_rm1c^3)) + (x[,"n_31"] + x[,"n_35"])*(log(pmax(1e-6,r_rm1c)) + log(6 - 5*r_rm1c - 2*r_rm1c^2 + r_rm1c^3)) + (x[,"n_22"] + x[,"n_44"])*(log(pmax(1e-6,r_rm1c)) + log(15 - 9*r_rm1c + 5*r_rm1c^2 + r_rm1c^3)) + (x[,"n_32"] + x[,"n_34"])*log(9 + 4*r_rm1c - 8*r_rm1c^2 + 8*r_rm1c^3 - 4*r_rm1c^4) + (x[,"n_23"] + x[,"n_43"])*log(21 + 4*r_rm1c - 7*r_rm1c^2 + 6*r_rm1c^3 - 3*r_rm1c^4) + (x[,"n_24"] + x[,"n_42"])*log(12 - 16*r_rm1c + 12*r_rm1c^2 - 9*r_rm1c^3 + r_rm1c^4) + x[,"n_33"]*log(23 - 20*r_rm1c + 26*r_rm1c^2 - 12*r_rm1c^3 + 6*r_rm1c^4)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_mc,r_mm2c,r_mm1c,r_rc,r_rm2c,r_rm1c,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_mc,LOD_mm2c,LOD_mm1c,LOD_rc,LOD_rm2c,LOD_rm1c,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_mc,logL_mm2c,logL_mm1c,logL_rc,logL_rm2c,logL_rm1c,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","mixed coupling","mixed mixed2c","mixed mixed1c","repulsion coupling","repulsion mixed2c","repulsion mixed1c","unknown")
)
)
}

#' @rdname r6_functions 
#' @noRd
r6_3.3_3.3 <- function(x){
logL_cc <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
L <- (-4*n00 - 3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 2*n53 - 3*n54 - 3*n55 - 3*n56 - 4*n60 - 3*n61 - 4*n62 - 2*n63 - 4*n64 - 3*n65 - 4*n66)*log(2) + (n01 + n02 + n04 + n05 + n10 + n11 + n12 + n13 + n14 + n15 + n16 + n20 + n21 + n22 + n23 + n24 + n25 + n26 + n31 + n32 + n34 + n35 + n40 + n41 + n42 + n43 + n44 + n45 + n46 + n50 + n51 + n52 + n53 + n54 + n55 + n56 + n61 + n62 + n64 + n65)*log(3) + (-2*n00 - 2*n01 - n02 - n03 - n04 - 2*n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - n26 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - n36 - n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - n46 - 2*n50 - 2*n51 - 2*n52 - 2*n53 - 2*n54 - 2*n55 - 2*n56 - 2*n60 - 2*n61 - n62 - n63 - n64 - 2*n65 - 2*n66)*log(5) + 6*(n00 + n66)*log(pmax(1e-6,1 - r)) + 6*(n06 + n60)*log(pmax(1e-6,r)) + (n01 + n10 + n56 + n65)*(5*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n02 + n20 + n46 + n64)*(4*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n03 + n30 + n36 + n63)*(3*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n04 + n26 + n40 + n62)*(2*log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n05 + n16 + n50 + n61)*(log(pmax(1e-6,1 - r)) + 5*log(pmax(1e-6,r))) + (n15 + n51)*(4*log(pmax(1e-6,r)) + log(7 - 10*r + 6*r^2)) + (n11 + n55)*(4*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 6*r^2)) + (n13 + n31 + n35 + n53)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(11 - 10*r + 10*r^2)) + (n12 + n21 + n45 + n54)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(13 - 10*r + 15*r^2)) + (n14 + n25 + n41 + n52)*(3*log(pmax(1e-6,r)) + log(18 - 38*r + 35*r^2 - 15*r^3)) + (n24 + n42)*(2*log(pmax(1e-6,r)) + log(90 - 216*r + 284*r^2 - 200*r^3 + 75*r^4)) + (n22 + n44)*(2*log(pmax(1e-6,1 - r)) + log(33 - 52*r + 134*r^2 - 100*r^3 + 75*r^4)) + (n23 + n32 + n34 + n43)*(log(pmax(1e-6,r)) + log(23 - 67*r + 113*r^2 - 119*r^3 + 75*r^4 - 25*r^5)) + n33*log(41 - 138*r + 336*r^2 - 496*r^3 + 498*r^4 - 300*r^5 + 100*r^6)
return(L)}
interlogL_cc <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
optimize(logL_cc,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cc <- mapply(interlogL_cc,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"])
LOD_cc <- 6*(x[,"n_06"] + x[,"n_60"])*log10(2) + 6*(x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*log10(2) + 6*(x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*log10(2) + 6*(x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*log10(2) + 6*(x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*log10(2) + 6*(x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*log10(2) + 6*(x[,"n_00"] + x[,"n_66"])*log10(2) - (x[,"n_15"] + x[,"n_51"])*(-5*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_55"])*(-5*log10(2) + log10(7)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-5*log10(2) + log10(17)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-6*log10(2) + log10(7) + log10(31)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-6*log10(2) + log10(47)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-6*log10(2) + log10(47)) - x[,"n_33"]*(-4*log10(2) + log10(277)) - (x[,"n_24"] + x[,"n_42"])*(-6*log10(2) + log10(523)) - (x[,"n_22"] + x[,"n_44"])*(-6*log10(2) + log10(523)) + 6*(x[,"n_00"] + x[,"n_66"])*log10(pmax(1e-6,1 - r_cc)) + 6*(x[,"n_06"] + x[,"n_60"])*log10(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(5*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(4*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(3*log10(pmax(1e-6,1 - r_cc)) + 3*log10(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log10(pmax(1e-6,1 - r_cc)) + 4*log10(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(log10(pmax(1e-6,1 - r_cc)) + 5*log10(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_51"])*(4*log10(pmax(1e-6,r_cc)) + log10(7 - 10*r_cc + 6*r_cc^2)) + (x[,"n_11"] + x[,"n_55"])*(4*log10(pmax(1e-6,1 - r_cc)) + log10(3 - 2*r_cc + 6*r_cc^2)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(2*log10(pmax(1e-6,1 - r_cc)) + 2*log10(pmax(1e-6,r_cc)) + log10(11 - 10*r_cc + 10*r_cc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(3*log10(pmax(1e-6,1 - r_cc)) + log10(pmax(1e-6,r_cc)) + log10(13 - 10*r_cc + 15*r_cc^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(3*log10(pmax(1e-6,r_cc)) + log10(18 - 38*r_cc + 35*r_cc^2 - 15*r_cc^3)) + (x[,"n_24"] + x[,"n_42"])*(2*log10(pmax(1e-6,r_cc)) + log10(90 - 216*r_cc + 284*r_cc^2 - 200*r_cc^3 + 75*r_cc^4)) + (x[,"n_22"] + x[,"n_44"])*(2*log10(pmax(1e-6,1 - r_cc)) + log10(33 - 52*r_cc + 134*r_cc^2 - 100*r_cc^3 + 75*r_cc^4)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log10(pmax(1e-6,r_cc)) + log10(23 - 67*r_cc + 113*r_cc^2 - 119*r_cc^3 + 75*r_cc^4 - 25*r_cc^5)) + x[,"n_33"]*log10(41 - 138*r_cc + 336*r_cc^2 - 496*r_cc^3 + 498*r_cc^4 - 300*r_cc^5 + 100*r_cc^6)
logL_cc <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 2*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"] - 4*x[,"n_60"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 2*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"] - 4*x[,"n_66"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_15"] + x[,"n_16"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_26"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_40"] + x[,"n_41"] + x[,"n_42"] + x[,"n_43"] + x[,"n_44"] + x[,"n_45"] + x[,"n_46"] + x[,"n_50"] + x[,"n_51"] + x[,"n_52"] + x[,"n_53"] + x[,"n_54"] + x[,"n_55"] + x[,"n_56"] + x[,"n_61"] + x[,"n_62"] + x[,"n_64"] + x[,"n_65"])*log(3) + (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - x[,"n_46"] - 2*x[,"n_50"] - 2*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"] - 2*x[,"n_56"] - 2*x[,"n_60"] - 2*x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - 2*x[,"n_65"] - 2*x[,"n_66"])*log(5) + 6*(x[,"n_00"] + x[,"n_66"])*log(pmax(1e-6,1 - r_cc)) + 6*(x[,"n_06"] + x[,"n_60"])*log(pmax(1e-6,r_cc)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(5*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(4*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc))) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(3*log(pmax(1e-6,1 - r_cc)) + 3*log(pmax(1e-6,r_cc))) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log(pmax(1e-6,1 - r_cc)) + 4*log(pmax(1e-6,r_cc))) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(log(pmax(1e-6,1 - r_cc)) + 5*log(pmax(1e-6,r_cc))) + (x[,"n_15"] + x[,"n_51"])*(4*log(pmax(1e-6,r_cc)) + log(7 - 10*r_cc + 6*r_cc^2)) + (x[,"n_11"] + x[,"n_55"])*(4*log(pmax(1e-6,1 - r_cc)) + log(3 - 2*r_cc + 6*r_cc^2)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(2*log(pmax(1e-6,1 - r_cc)) + 2*log(pmax(1e-6,r_cc)) + log(11 - 10*r_cc + 10*r_cc^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(3*log(pmax(1e-6,1 - r_cc)) + log(pmax(1e-6,r_cc)) + log(13 - 10*r_cc + 15*r_cc^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(3*log(pmax(1e-6,r_cc)) + log(18 - 38*r_cc + 35*r_cc^2 - 15*r_cc^3)) + (x[,"n_24"] + x[,"n_42"])*(2*log(pmax(1e-6,r_cc)) + log(90 - 216*r_cc + 284*r_cc^2 - 200*r_cc^3 + 75*r_cc^4)) + (x[,"n_22"] + x[,"n_44"])*(2*log(pmax(1e-6,1 - r_cc)) + log(33 - 52*r_cc + 134*r_cc^2 - 100*r_cc^3 + 75*r_cc^4)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log(pmax(1e-6,r_cc)) + log(23 - 67*r_cc + 113*r_cc^2 - 119*r_cc^3 + 75*r_cc^4 - 25*r_cc^5)) + x[,"n_33"]*log(41 - 138*r_cc + 336*r_cc^2 - 496*r_cc^3 + 498*r_cc^4 - 300*r_cc^5 + 100*r_cc^6)
logL_cm2c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
L <- (-4*n00 - 4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 2*n11 - 4*n12 - n13 - 4*n14 - 2*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26 - 3*n30 - n31 - 3*n32 - n33 - 3*n34 - n35 - 3*n36 - 4*n40 - 4*n41 - 4*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n46 - 4*n50 - 2*n51 - 4*n52 - n53 - 4*n54 - 2*n55 - 4*n56 - 4*n60 - 4*n61 - 4*n62 - 3*n63 - 4*n64 - 4*n65 - 4*n66)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56 - n60 - n61 - n62 - n63 - n64 - n65 - n66)*(log(3) + 2*log(5)) + (n00 + n66)*(5*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n06 + n60)*(log(pmax(1e-6,1 - r)) + 5*log(pmax(1e-6,r))) + (n05 + n16 + n50 + n61)*(4*log(pmax(1e-6,r)) + log(7 - 10*r + 6*r^2)) + (n01 + n10 + n56 + n65)*(4*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 6*r^2)) + (n03 + n30 + n36 + n63)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(11 - 10*r + 10*r^2)) + (n02 + n20 + n46 + n64)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(13 - 10*r + 15*r^2)) + (n04 + n26 + n40 + n62)*(3*log(pmax(1e-6,r)) + log(18 - 38*r + 35*r^2 - 15*r^3)) + (n15 + n51)*(3*log(pmax(1e-6,r)) + log(12 - 19*r + 19*r^2 - 9*r^3)) + (n11 + n55)*(3*log(pmax(1e-6,1 - r)) + log(3 + 8*r - 8*r^2 + 9*r^3)) + (n14 + n25 + n41 + n52)*(2*log(pmax(1e-6,r)) + log(114 - 216*r + 281*r^2 - 230*r^3 + 90*r^4)) + (n12 + n21 + n45 + n54)*(2*log(pmax(1e-6,1 - r)) + log(39 - 16*r + 131*r^2 - 130*r^3 + 90*r^4)) + (n24 + n42)*(log(pmax(1e-6,r)) + log(234 - 378*r + 696*r^2 - 832*r^3 + 625*r^4 - 225*r^5)) + (n13 + n31 + n35 + n53)*(log(pmax(1e-6,r)) + log(14 - 32*r + 51*r^2 - 63*r^3 + 45*r^4 - 15*r^5)) + (n22 + n44)*log(120 - 13*r - 407*r^2 + 882*r^3 - 1082*r^4 + 725*r^5 - 225*r^6) + n33*log(27 + 7*r - 71*r^2 + 178*r^3 - 214*r^4 + 150*r^5 - 50*r^6) + (n23 + n32 + n34 + n43)*log(69 - 70*r + 259*r^2 - 528*r^3 + 639*r^4 - 450*r^5 + 150*r^6)
return(L)}
interlogL_cm2c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
optimize(logL_cm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm2c <- mapply(interlogL_cm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"])
LOD_cm2c <- 6*(x[,"n_06"] + x[,"n_60"])*log10(2) + 6*(x[,"n_00"] + x[,"n_66"])*log10(2) - (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(-5*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(-5*log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_51"])*(-6*log10(2) + 2*log10(7)) - (x[,"n_11"] + x[,"n_55"])*(-6*log10(2) + 2*log10(7)) - (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(-5*log10(2) + log10(17)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-5*log10(2) + 2*log10(5) + log10(17)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + 2*log10(5) + log10(17)) - x[,"n_33"]*(-5*log10(2) + log10(19) + log10(43)) - (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(-6*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(-6*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-6*log10(2) + log10(167)) - (x[,"n_24"] + x[,"n_42"])*(-6*log10(2) + log10(5) + log10(941)) - (x[,"n_22"] + x[,"n_44"])*(-6*log10(2) + log10(5) + log10(941)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-5*log10(2) + log10(1951)) + (x[,"n_00"] + x[,"n_66"])*(5*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c))) + (x[,"n_06"] + x[,"n_60"])*(log10(pmax(1e-6,1 - r_cm2c)) + 5*log10(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(4*log10(pmax(1e-6,r_cm2c)) + log10(7 - 10*r_cm2c + 6*r_cm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(4*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 - 2*r_cm2c + 6*r_cm2c^2)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + 2*log10(pmax(1e-6,r_cm2c)) + log10(11 - 10*r_cm2c + 10*r_cm2c^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(pmax(1e-6,r_cm2c)) + log10(13 - 10*r_cm2c + 15*r_cm2c^2)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(3*log10(pmax(1e-6,r_cm2c)) + log10(18 - 38*r_cm2c + 35*r_cm2c^2 - 15*r_cm2c^3)) + (x[,"n_15"] + x[,"n_51"])*(3*log10(pmax(1e-6,r_cm2c)) + log10(12 - 19*r_cm2c + 19*r_cm2c^2 - 9*r_cm2c^3)) + (x[,"n_11"] + x[,"n_55"])*(3*log10(pmax(1e-6,1 - r_cm2c)) + log10(3 + 8*r_cm2c - 8*r_cm2c^2 + 9*r_cm2c^3)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(2*log10(pmax(1e-6,r_cm2c)) + log10(114 - 216*r_cm2c + 281*r_cm2c^2 - 230*r_cm2c^3 + 90*r_cm2c^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(2*log10(pmax(1e-6,1 - r_cm2c)) + log10(39 - 16*r_cm2c + 131*r_cm2c^2 - 130*r_cm2c^3 + 90*r_cm2c^4)) + (x[,"n_24"] + x[,"n_42"])*(log10(pmax(1e-6,r_cm2c)) + log10(234 - 378*r_cm2c + 696*r_cm2c^2 - 832*r_cm2c^3 + 625*r_cm2c^4 - 225*r_cm2c^5)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log10(pmax(1e-6,r_cm2c)) + log10(14 - 32*r_cm2c + 51*r_cm2c^2 - 63*r_cm2c^3 + 45*r_cm2c^4 - 15*r_cm2c^5)) + (x[,"n_22"] + x[,"n_44"])*log10(120 - 13*r_cm2c - 407*r_cm2c^2 + 882*r_cm2c^3 - 1082*r_cm2c^4 + 725*r_cm2c^5 - 225*r_cm2c^6) + x[,"n_33"]*log10(27 + 7*r_cm2c - 71*r_cm2c^2 + 178*r_cm2c^3 - 214*r_cm2c^4 + 150*r_cm2c^5 - 50*r_cm2c^6) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(69 - 70*r_cm2c + 259*r_cm2c^2 - 528*r_cm2c^3 + 639*r_cm2c^4 - 450*r_cm2c^5 + 150*r_cm2c^6)
logL_cm2c <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 2*x[,"n_11"] - 4*x[,"n_12"] - x[,"n_13"] - 4*x[,"n_14"] - 2*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 3*x[,"n_30"] - x[,"n_31"] - 3*x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 4*x[,"n_50"] - 2*x[,"n_51"] - 4*x[,"n_52"] - x[,"n_53"] - 4*x[,"n_54"] - 2*x[,"n_55"] - 4*x[,"n_56"] - 4*x[,"n_60"] - 4*x[,"n_61"] - 4*x[,"n_62"] - 3*x[,"n_63"] - 4*x[,"n_64"] - 4*x[,"n_65"] - 4*x[,"n_66"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"] - x[,"n_60"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"] - x[,"n_66"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_66"])*(5*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c))) + (x[,"n_06"] + x[,"n_60"])*(log(pmax(1e-6,1 - r_cm2c)) + 5*log(pmax(1e-6,r_cm2c))) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(4*log(pmax(1e-6,r_cm2c)) + log(7 - 10*r_cm2c + 6*r_cm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(4*log(pmax(1e-6,1 - r_cm2c)) + log(3 - 2*r_cm2c + 6*r_cm2c^2)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(2*log(pmax(1e-6,1 - r_cm2c)) + 2*log(pmax(1e-6,r_cm2c)) + log(11 - 10*r_cm2c + 10*r_cm2c^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(pmax(1e-6,r_cm2c)) + log(13 - 10*r_cm2c + 15*r_cm2c^2)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(3*log(pmax(1e-6,r_cm2c)) + log(18 - 38*r_cm2c + 35*r_cm2c^2 - 15*r_cm2c^3)) + (x[,"n_15"] + x[,"n_51"])*(3*log(pmax(1e-6,r_cm2c)) + log(12 - 19*r_cm2c + 19*r_cm2c^2 - 9*r_cm2c^3)) + (x[,"n_11"] + x[,"n_55"])*(3*log(pmax(1e-6,1 - r_cm2c)) + log(3 + 8*r_cm2c - 8*r_cm2c^2 + 9*r_cm2c^3)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(2*log(pmax(1e-6,r_cm2c)) + log(114 - 216*r_cm2c + 281*r_cm2c^2 - 230*r_cm2c^3 + 90*r_cm2c^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(2*log(pmax(1e-6,1 - r_cm2c)) + log(39 - 16*r_cm2c + 131*r_cm2c^2 - 130*r_cm2c^3 + 90*r_cm2c^4)) + (x[,"n_24"] + x[,"n_42"])*(log(pmax(1e-6,r_cm2c)) + log(234 - 378*r_cm2c + 696*r_cm2c^2 - 832*r_cm2c^3 + 625*r_cm2c^4 - 225*r_cm2c^5)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log(pmax(1e-6,r_cm2c)) + log(14 - 32*r_cm2c + 51*r_cm2c^2 - 63*r_cm2c^3 + 45*r_cm2c^4 - 15*r_cm2c^5)) + (x[,"n_22"] + x[,"n_44"])*log(120 - 13*r_cm2c - 407*r_cm2c^2 + 882*r_cm2c^3 - 1082*r_cm2c^4 + 725*r_cm2c^5 - 225*r_cm2c^6) + x[,"n_33"]*log(27 + 7*r_cm2c - 71*r_cm2c^2 + 178*r_cm2c^3 - 214*r_cm2c^4 + 150*r_cm2c^5 - 50*r_cm2c^6) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(69 - 70*r_cm2c + 259*r_cm2c^2 - 528*r_cm2c^3 + 639*r_cm2c^4 - 450*r_cm2c^5 + 150*r_cm2c^6)
logL_cm1c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 3*n21 - 4*n22 - 3*n23 - 4*n24 - 3*n25 - 4*n26 - 3*n30 - 3*n31 - 3*n32 - n33 - 3*n34 - 3*n35 - 3*n36 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44 - 3*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55 - 3*n56 - 4*n60 - 3*n61 - 4*n62 - 3*n63 - 4*n64 - 3*n65 - 4*n66)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56 - n60 - n61 - n62 - n63 - n64 - n65 - n66)*(log(3) + 2*log(5)) + (n00 + n66)*(4*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n06 + n60)*(2*log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n01 + n10 + n56 + n65)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 - 2*r + 3*r^2)) + (n05 + n16 + n50 + n61)*(3*log(pmax(1e-6,r)) + log(3 - 7*r + 7*r^2 - 3*r^3)) + (n04 + n26 + n40 + n62)*(2*log(pmax(1e-6,r)) + log(12 - 36*r + 52*r^2 - 40*r^3 + 15*r^4)) + (n02 + n20 + n46 + n64)*(2*log(pmax(1e-6,1 - r)) + log(3 - 8*r + 22*r^2 - 20*r^3 + 15*r^4)) + (n15 + n51)*(2*log(pmax(1e-6,r)) + log(15 - 36*r + 55*r^2 - 46*r^3 + 18*r^4)) + (n11 + n55)*(2*log(pmax(1e-6,1 - r)) + log(6 - 8*r + 25*r^2 - 26*r^3 + 18*r^4)) + (n14 + n25 + n41 + n52)*(log(pmax(1e-6,r)) + log(27 - 63*r + 129*r^2 - 167*r^3 + 125*r^4 - 45*r^5)) + (n03 + n30 + n36 + n63)*(log(pmax(1e-6,r)) + log(5 - 19*r + 38*r^2 - 44*r^3 + 30*r^4 - 10*r^5)) + (n23 + n32 + n34 + n43)*log(54 + 65*r - 251*r^2 + 522*r^3 - 636*r^4 + 450*r^5 - 150*r^6) + (n12 + n21 + n45 + n54)*log(6 + 25*r - 103*r^2 + 189*r^3 - 217*r^4 + 145*r^5 - 45*r^6) + n33*log(27 - 18*r + 79*r^2 - 172*r^3 + 211*r^4 - 150*r^5 + 50*r^6) + (n13 + n31 + n35 + n53)*log(15 - 34*r + 112*r^2 - 216*r^3 + 258*r^4 - 180*r^5 + 60*r^6) + (n22 + n44)*log(120 - 184*r + 427*r^2 - 852*r^3 + 1018*r^4 - 700*r^5 + 225*r^6) + (n24 + n42)*log(54 - 36*r + 354*r^2 - 720*r^3 + 893*r^4 - 650*r^5 + 225*r^6)
return(L)}
interlogL_cm1c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
optimize(logL_cm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cm1c <- mapply(interlogL_cm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"])
LOD_cm1c <- 6*(x[,"n_06"] + x[,"n_60"])*log10(2) + 6*(x[,"n_00"] + x[,"n_66"])*log10(2) - (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(-6*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(-6*log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_51"])*(-5*log10(2) + 2*log10(7)) - (x[,"n_11"] + x[,"n_55"])*(-5*log10(2) + 2*log10(7)) - (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(-5*log10(2) + log10(17)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-6*log10(2) + 2*log10(5) + log10(17)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-6*log10(2) + 2*log10(5) + log10(17)) - x[,"n_33"]*(-5*log10(2) + log10(19) + log10(43)) - (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(-6*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(-6*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(167)) - (x[,"n_24"] + x[,"n_42"])*(-6*log10(2) + log10(5) + log10(941)) - (x[,"n_22"] + x[,"n_44"])*(-6*log10(2) + log10(5) + log10(941)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-5*log10(2) + log10(1951)) + (x[,"n_00"] + x[,"n_66"])*(4*log10(pmax(1e-6,1 - r_cm1c)) + 2*log10(pmax(1e-6,r_cm1c))) + (x[,"n_06"] + x[,"n_60"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + 4*log10(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(3*log10(pmax(1e-6,1 - r_cm1c)) + log10(pmax(1e-6,r_cm1c)) + log10(2 - 2*r_cm1c + 3*r_cm1c^2)) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(3*log10(pmax(1e-6,r_cm1c)) + log10(3 - 7*r_cm1c + 7*r_cm1c^2 - 3*r_cm1c^3)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log10(pmax(1e-6,r_cm1c)) + log10(12 - 36*r_cm1c + 52*r_cm1c^2 - 40*r_cm1c^3 + 15*r_cm1c^4)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + log10(3 - 8*r_cm1c + 22*r_cm1c^2 - 20*r_cm1c^3 + 15*r_cm1c^4)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_cm1c)) + log10(15 - 36*r_cm1c + 55*r_cm1c^2 - 46*r_cm1c^3 + 18*r_cm1c^4)) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_cm1c)) + log10(6 - 8*r_cm1c + 25*r_cm1c^2 - 26*r_cm1c^3 + 18*r_cm1c^4)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log10(pmax(1e-6,r_cm1c)) + log10(27 - 63*r_cm1c + 129*r_cm1c^2 - 167*r_cm1c^3 + 125*r_cm1c^4 - 45*r_cm1c^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(log10(pmax(1e-6,r_cm1c)) + log10(5 - 19*r_cm1c + 38*r_cm1c^2 - 44*r_cm1c^3 + 30*r_cm1c^4 - 10*r_cm1c^5)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(54 + 65*r_cm1c - 251*r_cm1c^2 + 522*r_cm1c^3 - 636*r_cm1c^4 + 450*r_cm1c^5 - 150*r_cm1c^6) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log10(6 + 25*r_cm1c - 103*r_cm1c^2 + 189*r_cm1c^3 - 217*r_cm1c^4 + 145*r_cm1c^5 - 45*r_cm1c^6) + x[,"n_33"]*log10(27 - 18*r_cm1c + 79*r_cm1c^2 - 172*r_cm1c^3 + 211*r_cm1c^4 - 150*r_cm1c^5 + 50*r_cm1c^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(15 - 34*r_cm1c + 112*r_cm1c^2 - 216*r_cm1c^3 + 258*r_cm1c^4 - 180*r_cm1c^5 + 60*r_cm1c^6) + (x[,"n_22"] + x[,"n_44"])*log10(120 - 184*r_cm1c + 427*r_cm1c^2 - 852*r_cm1c^3 + 1018*r_cm1c^4 - 700*r_cm1c^5 + 225*r_cm1c^6) + (x[,"n_24"] + x[,"n_42"])*log10(54 - 36*r_cm1c + 354*r_cm1c^2 - 720*r_cm1c^3 + 893*r_cm1c^4 - 650*r_cm1c^5 + 225*r_cm1c^6)
logL_cm1c <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"] - 4*x[,"n_60"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 3*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"] - 4*x[,"n_66"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"] - x[,"n_60"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"] - x[,"n_66"])*(log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_66"])*(4*log(pmax(1e-6,1 - r_cm1c)) + 2*log(pmax(1e-6,r_cm1c))) + (x[,"n_06"] + x[,"n_60"])*(2*log(pmax(1e-6,1 - r_cm1c)) + 4*log(pmax(1e-6,r_cm1c))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(3*log(pmax(1e-6,1 - r_cm1c)) + log(pmax(1e-6,r_cm1c)) + log(2 - 2*r_cm1c + 3*r_cm1c^2)) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(3*log(pmax(1e-6,r_cm1c)) + log(3 - 7*r_cm1c + 7*r_cm1c^2 - 3*r_cm1c^3)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log(pmax(1e-6,r_cm1c)) + log(12 - 36*r_cm1c + 52*r_cm1c^2 - 40*r_cm1c^3 + 15*r_cm1c^4)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log(pmax(1e-6,1 - r_cm1c)) + log(3 - 8*r_cm1c + 22*r_cm1c^2 - 20*r_cm1c^3 + 15*r_cm1c^4)) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,r_cm1c)) + log(15 - 36*r_cm1c + 55*r_cm1c^2 - 46*r_cm1c^3 + 18*r_cm1c^4)) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_cm1c)) + log(6 - 8*r_cm1c + 25*r_cm1c^2 - 26*r_cm1c^3 + 18*r_cm1c^4)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log(pmax(1e-6,r_cm1c)) + log(27 - 63*r_cm1c + 129*r_cm1c^2 - 167*r_cm1c^3 + 125*r_cm1c^4 - 45*r_cm1c^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(log(pmax(1e-6,r_cm1c)) + log(5 - 19*r_cm1c + 38*r_cm1c^2 - 44*r_cm1c^3 + 30*r_cm1c^4 - 10*r_cm1c^5)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(54 + 65*r_cm1c - 251*r_cm1c^2 + 522*r_cm1c^3 - 636*r_cm1c^4 + 450*r_cm1c^5 - 150*r_cm1c^6) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log(6 + 25*r_cm1c - 103*r_cm1c^2 + 189*r_cm1c^3 - 217*r_cm1c^4 + 145*r_cm1c^5 - 45*r_cm1c^6) + x[,"n_33"]*log(27 - 18*r_cm1c + 79*r_cm1c^2 - 172*r_cm1c^3 + 211*r_cm1c^4 - 150*r_cm1c^5 + 50*r_cm1c^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(15 - 34*r_cm1c + 112*r_cm1c^2 - 216*r_cm1c^3 + 258*r_cm1c^4 - 180*r_cm1c^5 + 60*r_cm1c^6) + (x[,"n_22"] + x[,"n_44"])*log(120 - 184*r_cm1c + 427*r_cm1c^2 - 852*r_cm1c^3 + 1018*r_cm1c^4 - 700*r_cm1c^5 + 225*r_cm1c^6) + (x[,"n_24"] + x[,"n_42"])*log(54 - 36*r_cm1c + 354*r_cm1c^2 - 720*r_cm1c^3 + 893*r_cm1c^4 - 650*r_cm1c^5 + 225*r_cm1c^6)
logL_cr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66,n33) {
L <- 2*(-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - 2*n10 - n11 - 2*n12 - n13 - 2*n14 - n15 - 2*n16 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - 2*n26 - 2*n30 - n31 - 2*n32 - 2*n34 - n35 - 2*n36 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n46 - 2*n50 - n51 - 2*n52 - n53 - 2*n54 - n55 - 2*n56 - 2*n60 - 2*n61 - 2*n62 - 2*n63 - 2*n64 - 2*n65 - 2*n66)*log(2) + (n01 + n02 + n04 + n05 + n10 + n11 + n12 + n13 + n14 + n15 + n16 + n20 + n21 + n22 + n23 + n24 + n25 + n26 + n31 + n32 + n34 + n35 + n40 + n41 + n42 + n43 + n44 + n45 + n46 + n50 + n51 + n52 + n53 + n54 + n55 + n56 + n61 + n62 + n64 + n65)*log(3) + 2*(-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56 - n60 - n61 - n62 - n63 - n64 - n65 - n66)*log(5) + (n00 + n06 + n60 + n66)*(3*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n05 + n10 + n16 + n50 + n56 + n61 + n65)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log((-1 + r)^2 + r^2)) + (n22 + n24 + n42 + n44)*(log(pmax(1e-6,r)) + log(58 - 178*r + 315*r^2 - 345*r^3 + 225*r^4 - 75*r^5)) + n33*(log(pmax(1e-6,r)) + log(27 - 72*r + 115*r^2 - 120*r^3 + 75*r^4 - 25*r^5)) + (n13 + n31 + n35 + n53)*(log(pmax(1e-6,r)) + log(5 - 19*r + 38*r^2 - 44*r^3 + 30*r^4 - 10*r^5)) + (n02 + n04 + n20 + n26 + n40 + n46 + n62 + n64)*(log(pmax(1e-6,r)) + log(1 - 5*r + 13*r^2 - 19*r^3 + 15*r^4 - 5*r^5)) + (n11 + n15 + n51 + n55)*(log(pmax(1e-6,r)) + log(1 - 4*r + 9*r^2 - 12*r^3 + 9*r^4 - 3*r^5)) + (n03 + n30 + n36 + n63)*log(1 - 6*r + 24*r^2 - 56*r^3 + 78*r^4 - 60*r^5 + 20*r^6) + (n12 + n14 + n21 + n25 + n41 + n45 + n52 + n54)*log(3 - 14*r + 53*r^2 - 108*r^3 + 129*r^4 - 90*r^5 + 30*r^6) + (n23 + n32 + n34 + n43)*log(27 - 90*r + 260*r^2 - 440*r^3 + 470*r^4 - 300*r^5 + 100*r^6)
return(L)}
interlogL_cr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66,n33) {
optimize(logL_cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_cr <- mapply(interlogL_cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"],x[,"n_33"])
LOD_cr <- 5*(x[,"n_01"] + x[,"n_05"] + x[,"n_10"] + x[,"n_16"] + x[,"n_50"] + x[,"n_56"] + x[,"n_61"] + x[,"n_65"])*log10(2) + 6*(x[,"n_00"] + x[,"n_06"] + x[,"n_60"] + x[,"n_66"])*log10(2) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(4*log10(2) - log10(5)) - (x[,"n_02"] + x[,"n_04"] + x[,"n_20"] + x[,"n_26"] + x[,"n_40"] + x[,"n_46"] + x[,"n_62"] + x[,"n_64"])*(-6*log10(2) + log10(5)) - (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(-6*log10(2) + log10(7)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-5*log10(2) + log10(17)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(7) + log10(31)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"] + x[,"n_52"] + x[,"n_54"])*(-5*log10(2) + log10(47)) - x[,"n_33"]*(-6*log10(2) + log10(277)) - (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*(-6*log10(2) + log10(523)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_60"] + x[,"n_66"])*(3*log10(pmax(1e-6,1 - r_cr)) + 3*log10(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_05"] + x[,"n_10"] + x[,"n_16"] + x[,"n_50"] + x[,"n_56"] + x[,"n_61"] + x[,"n_65"])*(2*log10(pmax(1e-6,1 - r_cr)) + 2*log10(pmax(1e-6,r_cr)) + log10((-1 + r_cr)^2 + r_cr^2)) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*(log10(pmax(1e-6,r_cr)) + log10(58 - 178*r_cr + 315*r_cr^2 - 345*r_cr^3 + 225*r_cr^4 - 75*r_cr^5)) + x[,"n_33"]*(log10(pmax(1e-6,r_cr)) + log10(27 - 72*r_cr + 115*r_cr^2 - 120*r_cr^3 + 75*r_cr^4 - 25*r_cr^5)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log10(pmax(1e-6,r_cr)) + log10(5 - 19*r_cr + 38*r_cr^2 - 44*r_cr^3 + 30*r_cr^4 - 10*r_cr^5)) + (x[,"n_02"] + x[,"n_04"] + x[,"n_20"] + x[,"n_26"] + x[,"n_40"] + x[,"n_46"] + x[,"n_62"] + x[,"n_64"])*(log10(pmax(1e-6,r_cr)) + log10(1 - 5*r_cr + 13*r_cr^2 - 19*r_cr^3 + 15*r_cr^4 - 5*r_cr^5)) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log10(pmax(1e-6,r_cr)) + log10(1 - 4*r_cr + 9*r_cr^2 - 12*r_cr^3 + 9*r_cr^4 - 3*r_cr^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*log10(1 - 6*r_cr + 24*r_cr^2 - 56*r_cr^3 + 78*r_cr^4 - 60*r_cr^5 + 20*r_cr^6) + (x[,"n_12"] + x[,"n_14"] + x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"] + x[,"n_52"] + x[,"n_54"])*log10(3 - 14*r_cr + 53*r_cr^2 - 108*r_cr^3 + 129*r_cr^4 - 90*r_cr^5 + 30*r_cr^6) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(27 - 90*r_cr + 260*r_cr^2 - 440*r_cr^3 + 470*r_cr^4 - 300*r_cr^5 + 100*r_cr^6)
logL_cr <- 2*(-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - 2*x[,"n_16"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_26"] - 2*x[,"n_30"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_34"] - x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_46"] - 2*x[,"n_50"] - x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - x[,"n_55"] - 2*x[,"n_56"] - 2*x[,"n_60"] - 2*x[,"n_61"] - 2*x[,"n_62"] - 2*x[,"n_63"] - 2*x[,"n_64"] - 2*x[,"n_65"] - 2*x[,"n_66"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_15"] + x[,"n_16"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_26"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_40"] + x[,"n_41"] + x[,"n_42"] + x[,"n_43"] + x[,"n_44"] + x[,"n_45"] + x[,"n_46"] + x[,"n_50"] + x[,"n_51"] + x[,"n_52"] + x[,"n_53"] + x[,"n_54"] + x[,"n_55"] + x[,"n_56"] + x[,"n_61"] + x[,"n_62"] + x[,"n_64"] + x[,"n_65"])*log(3) + 2*(-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"] - x[,"n_60"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"] - x[,"n_66"])*log(5) + (x[,"n_00"] + x[,"n_06"] + x[,"n_60"] + x[,"n_66"])*(3*log(pmax(1e-6,1 - r_cr)) + 3*log(pmax(1e-6,r_cr))) + (x[,"n_01"] + x[,"n_05"] + x[,"n_10"] + x[,"n_16"] + x[,"n_50"] + x[,"n_56"] + x[,"n_61"] + x[,"n_65"])*(2*log(pmax(1e-6,1 - r_cr)) + 2*log(pmax(1e-6,r_cr)) + log((-1 + r_cr)^2 + r_cr^2)) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*(log(pmax(1e-6,r_cr)) + log(58 - 178*r_cr + 315*r_cr^2 - 345*r_cr^3 + 225*r_cr^4 - 75*r_cr^5)) + x[,"n_33"]*(log(pmax(1e-6,r_cr)) + log(27 - 72*r_cr + 115*r_cr^2 - 120*r_cr^3 + 75*r_cr^4 - 25*r_cr^5)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log(pmax(1e-6,r_cr)) + log(5 - 19*r_cr + 38*r_cr^2 - 44*r_cr^3 + 30*r_cr^4 - 10*r_cr^5)) + (x[,"n_02"] + x[,"n_04"] + x[,"n_20"] + x[,"n_26"] + x[,"n_40"] + x[,"n_46"] + x[,"n_62"] + x[,"n_64"])*(log(pmax(1e-6,r_cr)) + log(1 - 5*r_cr + 13*r_cr^2 - 19*r_cr^3 + 15*r_cr^4 - 5*r_cr^5)) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log(pmax(1e-6,r_cr)) + log(1 - 4*r_cr + 9*r_cr^2 - 12*r_cr^3 + 9*r_cr^4 - 3*r_cr^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*log(1 - 6*r_cr + 24*r_cr^2 - 56*r_cr^3 + 78*r_cr^4 - 60*r_cr^5 + 20*r_cr^6) + (x[,"n_12"] + x[,"n_14"] + x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"] + x[,"n_52"] + x[,"n_54"])*log(3 - 14*r_cr + 53*r_cr^2 - 108*r_cr^3 + 129*r_cr^4 - 90*r_cr^5 + 30*r_cr^6) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(27 - 90*r_cr + 260*r_cr^2 - 440*r_cr^3 + 470*r_cr^4 - 300*r_cr^5 + 100*r_cr^6)
logL_m2cm2c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
L <- (-4*n00 - 3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 2*n53 - 3*n54 - 3*n55 - 3*n56 - 4*n60 - 3*n61 - 4*n62 - 2*n63 - 4*n64 - 3*n65 - 4*n66)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56 - n60 - n61 - n62 - n63 - n64 - n65 - n66)*(2*log(3) + 2*log(5)) + (n00 + n66)*(4*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n06 + n60)*(2*log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n05 + n16 + n50 + n61)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n01 + n10 + n56 + n65)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - 5*r + 5*r^2 - 3*r^3)) + (n04 + n26 + n40 + n62)*(2*log(pmax(1e-6,r)) + log(22 - 48*r + 60*r^2 - 40*r^3 + 15*r^4)) + (n02 + n20 + n46 + n64)*(2*log(pmax(1e-6,1 - r)) + log(9 - 12*r + 30*r^2 - 20*r^3 + 15*r^4)) + (n15 + n51)*(2*log(pmax(1e-6,r)) + log(28 - 48*r + 57*r^2 - 46*r^3 + 18*r^4)) + (n11 + n55)*(2*log(pmax(1e-6,1 - r)) + log(9 + 27*r^2 - 26*r^3 + 18*r^4)) + (n14 + n25 + n41 + n52)*(log(pmax(1e-6,r)) + log(60 - 84*r + 146*r^2 - 166*r^3 + 125*r^4 - 45*r^5)) + (n03 + n30 + n36 + n63)*(log(pmax(1e-6,r)) + log(6 - 16*r + 25*r^2 - 25*r^3 + 15*r^4 - 5*r^5)) + (n23 + n32 + n34 + n43)*log(90 + 15*r - 73*r^2 + 191*r^3 - 283*r^4 + 225*r^5 - 75*r^6) + (n12 + n21 + n45 + n54)*log(36 - 21*r - 67*r^2 + 168*r^3 - 216*r^4 + 145*r^5 - 45*r^6) + (n13 + n31 + n35 + n53)*log(18 - 12*r + 49*r^2 - 104*r^3 + 127*r^4 - 90*r^5 + 30*r^6) + n33*log(153 - 18*r + 80*r^2 - 224*r^3 + 362*r^4 - 300*r^5 + 100*r^6) + (n22 + n44)*log(288 - 144*r + 159*r^2 - 604*r^3 + 938*r^4 - 700*r^5 + 225*r^6) + (n24 + n42)*log(162 + 36*r + 350*r^2 - 648*r^3 + 813*r^4 - 650*r^5 + 225*r^6)
return(L)}
interlogL_m2cm2c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
optimize(logL_m2cm2c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m2cm2c <- mapply(interlogL_m2cm2c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"])
LOD_m2cm2c <- 6*(x[,"n_06"] + x[,"n_60"])*log10(2) + 6*(x[,"n_00"] + x[,"n_66"])*log10(2) - (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(-6*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(-6*log10(2) + log10(11)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-5*log10(2) + 2*log10(7) + log10(11)) - (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(-6*log10(2) + log10(11) + log10(13)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(-6*log10(2) + log10(11) + log10(13)) - (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(-6*log10(2) + log10(61)) - (x[,"n_15"] + x[,"n_51"])*(-5*log10(2) + log10(109)) - (x[,"n_11"] + x[,"n_55"])*(-5*log10(2) + log10(109)) - x[,"n_33"]*(-4*log10(2) + log10(19) + log10(127)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-6*log10(2) + log10(5) + log10(257)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-6*log10(2) + log10(5) + log10(257)) - (x[,"n_24"] + x[,"n_42"])*(-6*log10(2) + log10(11) + log10(1283)) - (x[,"n_22"] + x[,"n_44"])*(-6*log10(2) + log10(11) + log10(1283)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-6*log10(2) + log10(5843)) + (x[,"n_00"] + x[,"n_66"])*(4*log10(pmax(1e-6,1 - r_m2cm2c)) + 2*log10(pmax(1e-6,r_m2cm2c))) + (x[,"n_06"] + x[,"n_60"])*(2*log10(pmax(1e-6,1 - r_m2cm2c)) + 4*log10(pmax(1e-6,r_m2cm2c))) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(log10(pmax(1e-6,1 - r_m2cm2c)) + 3*log10(pmax(1e-6,r_m2cm2c)) + log10(4 - 4*r_m2cm2c + 3*r_m2cm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(2*log10(pmax(1e-6,1 - r_m2cm2c)) + log10(pmax(1e-6,r_m2cm2c)) + log10(3 - 5*r_m2cm2c + 5*r_m2cm2c^2 - 3*r_m2cm2c^3)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log10(pmax(1e-6,r_m2cm2c)) + log10(22 - 48*r_m2cm2c + 60*r_m2cm2c^2 - 40*r_m2cm2c^3 + 15*r_m2cm2c^4)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log10(pmax(1e-6,1 - r_m2cm2c)) + log10(9 - 12*r_m2cm2c + 30*r_m2cm2c^2 - 20*r_m2cm2c^3 + 15*r_m2cm2c^4)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,r_m2cm2c)) + log10(28 - 48*r_m2cm2c + 57*r_m2cm2c^2 - 46*r_m2cm2c^3 + 18*r_m2cm2c^4)) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,1 - r_m2cm2c)) + log10(9 + 27*r_m2cm2c^2 - 26*r_m2cm2c^3 + 18*r_m2cm2c^4)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log10(pmax(1e-6,r_m2cm2c)) + log10(60 - 84*r_m2cm2c + 146*r_m2cm2c^2 - 166*r_m2cm2c^3 + 125*r_m2cm2c^4 - 45*r_m2cm2c^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(log10(pmax(1e-6,r_m2cm2c)) + log10(6 - 16*r_m2cm2c + 25*r_m2cm2c^2 - 25*r_m2cm2c^3 + 15*r_m2cm2c^4 - 5*r_m2cm2c^5)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(90 + 15*r_m2cm2c - 73*r_m2cm2c^2 + 191*r_m2cm2c^3 - 283*r_m2cm2c^4 + 225*r_m2cm2c^5 - 75*r_m2cm2c^6) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log10(36 - 21*r_m2cm2c - 67*r_m2cm2c^2 + 168*r_m2cm2c^3 - 216*r_m2cm2c^4 + 145*r_m2cm2c^5 - 45*r_m2cm2c^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(18 - 12*r_m2cm2c + 49*r_m2cm2c^2 - 104*r_m2cm2c^3 + 127*r_m2cm2c^4 - 90*r_m2cm2c^5 + 30*r_m2cm2c^6) + x[,"n_33"]*log10(153 - 18*r_m2cm2c + 80*r_m2cm2c^2 - 224*r_m2cm2c^3 + 362*r_m2cm2c^4 - 300*r_m2cm2c^5 + 100*r_m2cm2c^6) + (x[,"n_22"] + x[,"n_44"])*log10(288 - 144*r_m2cm2c + 159*r_m2cm2c^2 - 604*r_m2cm2c^3 + 938*r_m2cm2c^4 - 700*r_m2cm2c^5 + 225*r_m2cm2c^6) + (x[,"n_24"] + x[,"n_42"])*log10(162 + 36*r_m2cm2c + 350*r_m2cm2c^2 - 648*r_m2cm2c^3 + 813*r_m2cm2c^4 - 650*r_m2cm2c^5 + 225*r_m2cm2c^6)
logL_m2cm2c <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 2*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"] - 4*x[,"n_60"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 2*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"] - 4*x[,"n_66"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"] - x[,"n_60"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"] - x[,"n_66"])*(2*log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_66"])*(4*log(pmax(1e-6,1 - r_m2cm2c)) + 2*log(pmax(1e-6,r_m2cm2c))) + (x[,"n_06"] + x[,"n_60"])*(2*log(pmax(1e-6,1 - r_m2cm2c)) + 4*log(pmax(1e-6,r_m2cm2c))) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(log(pmax(1e-6,1 - r_m2cm2c)) + 3*log(pmax(1e-6,r_m2cm2c)) + log(4 - 4*r_m2cm2c + 3*r_m2cm2c^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(2*log(pmax(1e-6,1 - r_m2cm2c)) + log(pmax(1e-6,r_m2cm2c)) + log(3 - 5*r_m2cm2c + 5*r_m2cm2c^2 - 3*r_m2cm2c^3)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log(pmax(1e-6,r_m2cm2c)) + log(22 - 48*r_m2cm2c + 60*r_m2cm2c^2 - 40*r_m2cm2c^3 + 15*r_m2cm2c^4)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log(pmax(1e-6,1 - r_m2cm2c)) + log(9 - 12*r_m2cm2c + 30*r_m2cm2c^2 - 20*r_m2cm2c^3 + 15*r_m2cm2c^4)) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,r_m2cm2c)) + log(28 - 48*r_m2cm2c + 57*r_m2cm2c^2 - 46*r_m2cm2c^3 + 18*r_m2cm2c^4)) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,1 - r_m2cm2c)) + log(9 + 27*r_m2cm2c^2 - 26*r_m2cm2c^3 + 18*r_m2cm2c^4)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(log(pmax(1e-6,r_m2cm2c)) + log(60 - 84*r_m2cm2c + 146*r_m2cm2c^2 - 166*r_m2cm2c^3 + 125*r_m2cm2c^4 - 45*r_m2cm2c^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(log(pmax(1e-6,r_m2cm2c)) + log(6 - 16*r_m2cm2c + 25*r_m2cm2c^2 - 25*r_m2cm2c^3 + 15*r_m2cm2c^4 - 5*r_m2cm2c^5)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(90 + 15*r_m2cm2c - 73*r_m2cm2c^2 + 191*r_m2cm2c^3 - 283*r_m2cm2c^4 + 225*r_m2cm2c^5 - 75*r_m2cm2c^6) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*log(36 - 21*r_m2cm2c - 67*r_m2cm2c^2 + 168*r_m2cm2c^3 - 216*r_m2cm2c^4 + 145*r_m2cm2c^5 - 45*r_m2cm2c^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(18 - 12*r_m2cm2c + 49*r_m2cm2c^2 - 104*r_m2cm2c^3 + 127*r_m2cm2c^4 - 90*r_m2cm2c^5 + 30*r_m2cm2c^6) + x[,"n_33"]*log(153 - 18*r_m2cm2c + 80*r_m2cm2c^2 - 224*r_m2cm2c^3 + 362*r_m2cm2c^4 - 300*r_m2cm2c^5 + 100*r_m2cm2c^6) + (x[,"n_22"] + x[,"n_44"])*log(288 - 144*r_m2cm2c + 159*r_m2cm2c^2 - 604*r_m2cm2c^3 + 938*r_m2cm2c^4 - 700*r_m2cm2c^5 + 225*r_m2cm2c^6) + (x[,"n_24"] + x[,"n_42"])*log(162 + 36*r_m2cm2c + 350*r_m2cm2c^2 - 648*r_m2cm2c^3 + 813*r_m2cm2c^4 - 650*r_m2cm2c^5 + 225*r_m2cm2c^6)
logL_m2cm1c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66,n33) {
L <- 2*(-2*n00 - 2*n01 - 2*n02 - 2*n03 - 2*n04 - 2*n05 - 2*n06 - 2*n10 - n11 - 2*n12 - n13 - 2*n14 - n15 - 2*n16 - 2*n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - 2*n26 - 2*n30 - n31 - 2*n32 - 2*n34 - n35 - 2*n36 - 2*n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - 2*n46 - 2*n50 - n51 - 2*n52 - n53 - 2*n54 - n55 - 2*n56 - 2*n60 - 2*n61 - 2*n62 - 2*n63 - 2*n64 - 2*n65 - 2*n66)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56 - n60 - n61 - n62 - n63 - n64 - n65 - n66)*(2*log(3) + 2*log(5)) + (n00 + n06 + n60 + n66)*(3*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n01 + n05 + n10 + n16 + n50 + n56 + n61 + n65)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(7 - 6*r + 6*r^2)) + (n02 + n04 + n20 + n26 + n40 + n46 + n62 + n64)*(log(pmax(1e-6,r)) + log(15 - 43*r + 71*r^2 - 73*r^3 + 45*r^4 - 15*r^5)) + (n11 + n15 + n51 + n55)*(log(pmax(1e-6,r)) + log(9 - 20*r + 31*r^2 - 38*r^3 + 27*r^4 - 9*r^5)) + (n22 + n24 + n42 + n44)*log(216 + 54*r - 254*r^2 + 625*r^3 - 875*r^4 + 675*r^5 - 225*r^6) + (n13 + n31 + n35 + n53)*log(18 + 3*r - 41*r^2 + 106*r^3 - 128*r^4 + 90*r^5 - 30*r^6) + n33*log(36 + 9*r - 24*r^2 + 55*r^3 - 90*r^4 + 75*r^5 - 25*r^6) + (n03 + n30 + n36 + n63)*log(9 - 30*r + 72*r^2 - 104*r^3 + 102*r^4 - 60*r^5 + 20*r^6) + (n12 + n14 + n21 + n25 + n41 + n45 + n52 + n54)*log(45 - 42*r + 155*r^2 - 316*r^3 + 383*r^4 - 270*r^5 + 90*r^6) + (n23 + n32 + n34 + n43)*log(369 - 54*r + 284*r^2 - 760*r^3 + 1130*r^4 - 900*r^5 + 300*r^6)
return(L)}
interlogL_m2cm1c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66,n33) {
optimize(logL_m2cm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66,n33, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m2cm1c <- mapply(interlogL_m2cm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"],x[,"n_33"])
LOD_m2cm1c <- 6*(x[,"n_00"] + x[,"n_06"] + x[,"n_60"] + x[,"n_66"])*log10(2) - (x[,"n_01"] + x[,"n_05"] + x[,"n_10"] + x[,"n_16"] + x[,"n_50"] + x[,"n_56"] + x[,"n_61"] + x[,"n_65"])*(-5*log10(2) + log10(11)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-5*log10(2) + 2*log10(7) + log10(11)) - (x[,"n_02"] + x[,"n_04"] + x[,"n_20"] + x[,"n_26"] + x[,"n_40"] + x[,"n_46"] + x[,"n_62"] + x[,"n_64"])*(-6*log10(2) + log10(11) + log10(13)) - (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(-4*log10(2) + log10(61)) - (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(-6*log10(2) + log10(109)) - x[,"n_33"]*(-6*log10(2) + log10(19) + log10(127)) - (x[,"n_12"] + x[,"n_14"] + x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"] + x[,"n_52"] + x[,"n_54"])*(-5*log10(2) + log10(5) + log10(257)) - (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*(-6*log10(2) + log10(11) + log10(1283)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-4*log10(2) + log10(5843)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_60"] + x[,"n_66"])*(3*log10(pmax(1e-6,1 - r_m2cm1c)) + 3*log10(pmax(1e-6,r_m2cm1c))) + (x[,"n_01"] + x[,"n_05"] + x[,"n_10"] + x[,"n_16"] + x[,"n_50"] + x[,"n_56"] + x[,"n_61"] + x[,"n_65"])*(2*log10(pmax(1e-6,1 - r_m2cm1c)) + 2*log10(pmax(1e-6,r_m2cm1c)) + log10(7 - 6*r_m2cm1c + 6*r_m2cm1c^2)) + (x[,"n_02"] + x[,"n_04"] + x[,"n_20"] + x[,"n_26"] + x[,"n_40"] + x[,"n_46"] + x[,"n_62"] + x[,"n_64"])*(log10(pmax(1e-6,r_m2cm1c)) + log10(15 - 43*r_m2cm1c + 71*r_m2cm1c^2 - 73*r_m2cm1c^3 + 45*r_m2cm1c^4 - 15*r_m2cm1c^5)) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log10(pmax(1e-6,r_m2cm1c)) + log10(9 - 20*r_m2cm1c + 31*r_m2cm1c^2 - 38*r_m2cm1c^3 + 27*r_m2cm1c^4 - 9*r_m2cm1c^5)) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log10(216 + 54*r_m2cm1c - 254*r_m2cm1c^2 + 625*r_m2cm1c^3 - 875*r_m2cm1c^4 + 675*r_m2cm1c^5 - 225*r_m2cm1c^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(18 + 3*r_m2cm1c - 41*r_m2cm1c^2 + 106*r_m2cm1c^3 - 128*r_m2cm1c^4 + 90*r_m2cm1c^5 - 30*r_m2cm1c^6) + x[,"n_33"]*log10(36 + 9*r_m2cm1c - 24*r_m2cm1c^2 + 55*r_m2cm1c^3 - 90*r_m2cm1c^4 + 75*r_m2cm1c^5 - 25*r_m2cm1c^6) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*log10(9 - 30*r_m2cm1c + 72*r_m2cm1c^2 - 104*r_m2cm1c^3 + 102*r_m2cm1c^4 - 60*r_m2cm1c^5 + 20*r_m2cm1c^6) + (x[,"n_12"] + x[,"n_14"] + x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"] + x[,"n_52"] + x[,"n_54"])*log10(45 - 42*r_m2cm1c + 155*r_m2cm1c^2 - 316*r_m2cm1c^3 + 383*r_m2cm1c^4 - 270*r_m2cm1c^5 + 90*r_m2cm1c^6) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(369 - 54*r_m2cm1c + 284*r_m2cm1c^2 - 760*r_m2cm1c^3 + 1130*r_m2cm1c^4 - 900*r_m2cm1c^5 + 300*r_m2cm1c^6)
logL_m2cm1c <- 2*(-2*x[,"n_00"] - 2*x[,"n_01"] - 2*x[,"n_02"] - 2*x[,"n_03"] - 2*x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - x[,"n_11"] - 2*x[,"n_12"] - x[,"n_13"] - 2*x[,"n_14"] - x[,"n_15"] - 2*x[,"n_16"] - 2*x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - 2*x[,"n_26"] - 2*x[,"n_30"] - x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_34"] - x[,"n_35"] - 2*x[,"n_36"] - 2*x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - 2*x[,"n_46"] - 2*x[,"n_50"] - x[,"n_51"] - 2*x[,"n_52"] - x[,"n_53"] - 2*x[,"n_54"] - x[,"n_55"] - 2*x[,"n_56"] - 2*x[,"n_60"] - 2*x[,"n_61"] - 2*x[,"n_62"] - 2*x[,"n_63"] - 2*x[,"n_64"] - 2*x[,"n_65"] - 2*x[,"n_66"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"] - x[,"n_60"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"] - x[,"n_66"])*(2*log(3) + 2*log(5)) + (x[,"n_00"] + x[,"n_06"] + x[,"n_60"] + x[,"n_66"])*(3*log(pmax(1e-6,1 - r_m2cm1c)) + 3*log(pmax(1e-6,r_m2cm1c))) + (x[,"n_01"] + x[,"n_05"] + x[,"n_10"] + x[,"n_16"] + x[,"n_50"] + x[,"n_56"] + x[,"n_61"] + x[,"n_65"])*(2*log(pmax(1e-6,1 - r_m2cm1c)) + 2*log(pmax(1e-6,r_m2cm1c)) + log(7 - 6*r_m2cm1c + 6*r_m2cm1c^2)) + (x[,"n_02"] + x[,"n_04"] + x[,"n_20"] + x[,"n_26"] + x[,"n_40"] + x[,"n_46"] + x[,"n_62"] + x[,"n_64"])*(log(pmax(1e-6,r_m2cm1c)) + log(15 - 43*r_m2cm1c + 71*r_m2cm1c^2 - 73*r_m2cm1c^3 + 45*r_m2cm1c^4 - 15*r_m2cm1c^5)) + (x[,"n_11"] + x[,"n_15"] + x[,"n_51"] + x[,"n_55"])*(log(pmax(1e-6,r_m2cm1c)) + log(9 - 20*r_m2cm1c + 31*r_m2cm1c^2 - 38*r_m2cm1c^3 + 27*r_m2cm1c^4 - 9*r_m2cm1c^5)) + (x[,"n_22"] + x[,"n_24"] + x[,"n_42"] + x[,"n_44"])*log(216 + 54*r_m2cm1c - 254*r_m2cm1c^2 + 625*r_m2cm1c^3 - 875*r_m2cm1c^4 + 675*r_m2cm1c^5 - 225*r_m2cm1c^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(18 + 3*r_m2cm1c - 41*r_m2cm1c^2 + 106*r_m2cm1c^3 - 128*r_m2cm1c^4 + 90*r_m2cm1c^5 - 30*r_m2cm1c^6) + x[,"n_33"]*log(36 + 9*r_m2cm1c - 24*r_m2cm1c^2 + 55*r_m2cm1c^3 - 90*r_m2cm1c^4 + 75*r_m2cm1c^5 - 25*r_m2cm1c^6) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*log(9 - 30*r_m2cm1c + 72*r_m2cm1c^2 - 104*r_m2cm1c^3 + 102*r_m2cm1c^4 - 60*r_m2cm1c^5 + 20*r_m2cm1c^6) + (x[,"n_12"] + x[,"n_14"] + x[,"n_21"] + x[,"n_25"] + x[,"n_41"] + x[,"n_45"] + x[,"n_52"] + x[,"n_54"])*log(45 - 42*r_m2cm1c + 155*r_m2cm1c^2 - 316*r_m2cm1c^3 + 383*r_m2cm1c^4 - 270*r_m2cm1c^5 + 90*r_m2cm1c^6) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(369 - 54*r_m2cm1c + 284*r_m2cm1c^2 - 760*r_m2cm1c^3 + 1130*r_m2cm1c^4 - 900*r_m2cm1c^5 + 300*r_m2cm1c^6)
logL_m2cr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
L <- (-4*n00 - 3*n01 - 4*n02 - 3*n03 - 4*n04 - 3*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 3*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 3*n21 - 4*n22 - 3*n23 - 4*n24 - 3*n25 - 4*n26 - 3*n30 - 3*n31 - 3*n32 - n33 - 3*n34 - 3*n35 - 3*n36 - 4*n40 - 3*n41 - 4*n42 - 3*n43 - 4*n44 - 3*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 3*n53 - 3*n54 - 3*n55 - 3*n56 - 4*n60 - 3*n61 - 4*n62 - 3*n63 - 4*n64 - 3*n65 - 4*n66)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56 - n60 - n61 - n62 - n63 - n64 - n65 - n66)*(log(3) + 2*log(5)) + (n06 + n60)*(4*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n00 + n66)*(2*log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n05 + n16 + n50 + n61)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(2 - 2*r + 3*r^2)) + (n01 + n10 + n56 + n65)*(3*log(pmax(1e-6,r)) + log(3 - 7*r + 7*r^2 - 3*r^3)) + (n02 + n20 + n46 + n64)*(2*log(pmax(1e-6,r)) + log(12 - 36*r + 52*r^2 - 40*r^3 + 15*r^4)) + (n04 + n26 + n40 + n62)*(2*log(pmax(1e-6,1 - r)) + log(3 - 8*r + 22*r^2 - 20*r^3 + 15*r^4)) + (n11 + n55)*(2*log(pmax(1e-6,r)) + log(15 - 36*r + 55*r^2 - 46*r^3 + 18*r^4)) + (n15 + n51)*(2*log(pmax(1e-6,1 - r)) + log(6 - 8*r + 25*r^2 - 26*r^3 + 18*r^4)) + (n12 + n21 + n45 + n54)*(log(pmax(1e-6,r)) + log(27 - 63*r + 129*r^2 - 167*r^3 + 125*r^4 - 45*r^5)) + (n03 + n30 + n36 + n63)*(log(pmax(1e-6,r)) + log(5 - 19*r + 38*r^2 - 44*r^3 + 30*r^4 - 10*r^5)) + (n23 + n32 + n34 + n43)*log(54 + 65*r - 251*r^2 + 522*r^3 - 636*r^4 + 450*r^5 - 150*r^6) + (n14 + n25 + n41 + n52)*log(6 + 25*r - 103*r^2 + 189*r^3 - 217*r^4 + 145*r^5 - 45*r^6) + n33*log(27 - 18*r + 79*r^2 - 172*r^3 + 211*r^4 - 150*r^5 + 50*r^6) + (n13 + n31 + n35 + n53)*log(15 - 34*r + 112*r^2 - 216*r^3 + 258*r^4 - 180*r^5 + 60*r^6) + (n24 + n42)*log(120 - 184*r + 427*r^2 - 852*r^3 + 1018*r^4 - 700*r^5 + 225*r^6) + (n22 + n44)*log(54 - 36*r + 354*r^2 - 720*r^3 + 893*r^4 - 650*r^5 + 225*r^6)
return(L)}
interlogL_m2cr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
optimize(logL_m2cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m2cr <- mapply(interlogL_m2cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"])
LOD_m2cr <- 6*(x[,"n_06"] + x[,"n_60"])*log10(2) + 6*(x[,"n_00"] + x[,"n_66"])*log10(2) - (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(-6*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(-6*log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_51"])*(-5*log10(2) + 2*log10(7)) - (x[,"n_11"] + x[,"n_55"])*(-5*log10(2) + 2*log10(7)) - (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(-5*log10(2) + log10(17)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-6*log10(2) + 2*log10(5) + log10(17)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-6*log10(2) + 2*log10(5) + log10(17)) - x[,"n_33"]*(-5*log10(2) + log10(19) + log10(43)) - (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(-6*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(-6*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-4*log10(2) + log10(167)) - (x[,"n_24"] + x[,"n_42"])*(-6*log10(2) + log10(5) + log10(941)) - (x[,"n_22"] + x[,"n_44"])*(-6*log10(2) + log10(5) + log10(941)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-5*log10(2) + log10(1951)) + (x[,"n_06"] + x[,"n_60"])*(4*log10(pmax(1e-6,1 - r_m2cr)) + 2*log10(pmax(1e-6,r_m2cr))) + (x[,"n_00"] + x[,"n_66"])*(2*log10(pmax(1e-6,1 - r_m2cr)) + 4*log10(pmax(1e-6,r_m2cr))) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(3*log10(pmax(1e-6,1 - r_m2cr)) + log10(pmax(1e-6,r_m2cr)) + log10(2 - 2*r_m2cr + 3*r_m2cr^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(3*log10(pmax(1e-6,r_m2cr)) + log10(3 - 7*r_m2cr + 7*r_m2cr^2 - 3*r_m2cr^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log10(pmax(1e-6,r_m2cr)) + log10(12 - 36*r_m2cr + 52*r_m2cr^2 - 40*r_m2cr^3 + 15*r_m2cr^4)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log10(pmax(1e-6,1 - r_m2cr)) + log10(3 - 8*r_m2cr + 22*r_m2cr^2 - 20*r_m2cr^3 + 15*r_m2cr^4)) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,r_m2cr)) + log10(15 - 36*r_m2cr + 55*r_m2cr^2 - 46*r_m2cr^3 + 18*r_m2cr^4)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_m2cr)) + log10(6 - 8*r_m2cr + 25*r_m2cr^2 - 26*r_m2cr^3 + 18*r_m2cr^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(log10(pmax(1e-6,r_m2cr)) + log10(27 - 63*r_m2cr + 129*r_m2cr^2 - 167*r_m2cr^3 + 125*r_m2cr^4 - 45*r_m2cr^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(log10(pmax(1e-6,r_m2cr)) + log10(5 - 19*r_m2cr + 38*r_m2cr^2 - 44*r_m2cr^3 + 30*r_m2cr^4 - 10*r_m2cr^5)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(54 + 65*r_m2cr - 251*r_m2cr^2 + 522*r_m2cr^3 - 636*r_m2cr^4 + 450*r_m2cr^5 - 150*r_m2cr^6) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*log10(6 + 25*r_m2cr - 103*r_m2cr^2 + 189*r_m2cr^3 - 217*r_m2cr^4 + 145*r_m2cr^5 - 45*r_m2cr^6) + x[,"n_33"]*log10(27 - 18*r_m2cr + 79*r_m2cr^2 - 172*r_m2cr^3 + 211*r_m2cr^4 - 150*r_m2cr^5 + 50*r_m2cr^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(15 - 34*r_m2cr + 112*r_m2cr^2 - 216*r_m2cr^3 + 258*r_m2cr^4 - 180*r_m2cr^5 + 60*r_m2cr^6) + (x[,"n_24"] + x[,"n_42"])*log10(120 - 184*r_m2cr + 427*r_m2cr^2 - 852*r_m2cr^3 + 1018*r_m2cr^4 - 700*r_m2cr^5 + 225*r_m2cr^6) + (x[,"n_22"] + x[,"n_44"])*log10(54 - 36*r_m2cr + 354*r_m2cr^2 - 720*r_m2cr^3 + 893*r_m2cr^4 - 650*r_m2cr^5 + 225*r_m2cr^6)
logL_m2cr <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 3*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 3*x[,"n_30"] - 3*x[,"n_31"] - 3*x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - 3*x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 3*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"] - 4*x[,"n_60"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 3*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"] - 4*x[,"n_66"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"] - x[,"n_60"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"] - x[,"n_66"])*(log(3) + 2*log(5)) + (x[,"n_06"] + x[,"n_60"])*(4*log(pmax(1e-6,1 - r_m2cr)) + 2*log(pmax(1e-6,r_m2cr))) + (x[,"n_00"] + x[,"n_66"])*(2*log(pmax(1e-6,1 - r_m2cr)) + 4*log(pmax(1e-6,r_m2cr))) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(3*log(pmax(1e-6,1 - r_m2cr)) + log(pmax(1e-6,r_m2cr)) + log(2 - 2*r_m2cr + 3*r_m2cr^2)) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(3*log(pmax(1e-6,r_m2cr)) + log(3 - 7*r_m2cr + 7*r_m2cr^2 - 3*r_m2cr^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log(pmax(1e-6,r_m2cr)) + log(12 - 36*r_m2cr + 52*r_m2cr^2 - 40*r_m2cr^3 + 15*r_m2cr^4)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log(pmax(1e-6,1 - r_m2cr)) + log(3 - 8*r_m2cr + 22*r_m2cr^2 - 20*r_m2cr^3 + 15*r_m2cr^4)) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,r_m2cr)) + log(15 - 36*r_m2cr + 55*r_m2cr^2 - 46*r_m2cr^3 + 18*r_m2cr^4)) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_m2cr)) + log(6 - 8*r_m2cr + 25*r_m2cr^2 - 26*r_m2cr^3 + 18*r_m2cr^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(log(pmax(1e-6,r_m2cr)) + log(27 - 63*r_m2cr + 129*r_m2cr^2 - 167*r_m2cr^3 + 125*r_m2cr^4 - 45*r_m2cr^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(log(pmax(1e-6,r_m2cr)) + log(5 - 19*r_m2cr + 38*r_m2cr^2 - 44*r_m2cr^3 + 30*r_m2cr^4 - 10*r_m2cr^5)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(54 + 65*r_m2cr - 251*r_m2cr^2 + 522*r_m2cr^3 - 636*r_m2cr^4 + 450*r_m2cr^5 - 150*r_m2cr^6) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*log(6 + 25*r_m2cr - 103*r_m2cr^2 + 189*r_m2cr^3 - 217*r_m2cr^4 + 145*r_m2cr^5 - 45*r_m2cr^6) + x[,"n_33"]*log(27 - 18*r_m2cr + 79*r_m2cr^2 - 172*r_m2cr^3 + 211*r_m2cr^4 - 150*r_m2cr^5 + 50*r_m2cr^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(15 - 34*r_m2cr + 112*r_m2cr^2 - 216*r_m2cr^3 + 258*r_m2cr^4 - 180*r_m2cr^5 + 60*r_m2cr^6) + (x[,"n_24"] + x[,"n_42"])*log(120 - 184*r_m2cr + 427*r_m2cr^2 - 852*r_m2cr^3 + 1018*r_m2cr^4 - 700*r_m2cr^5 + 225*r_m2cr^6) + (x[,"n_22"] + x[,"n_44"])*log(54 - 36*r_m2cr + 354*r_m2cr^2 - 720*r_m2cr^3 + 893*r_m2cr^4 - 650*r_m2cr^5 + 225*r_m2cr^6)
logL_m1cm1c <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
L <- (-4*n00 - 3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 2*n53 - 3*n54 - 3*n55 - 3*n56 - 4*n60 - 3*n61 - 4*n62 - 2*n63 - 4*n64 - 3*n65 - 4*n66)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56 - n60 - n61 - n62 - n63 - n64 - n65 - n66)*(2*log(3) + 2*log(5)) + (n06 + n60)*(4*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n00 + n66)*(2*log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n01 + n10 + n56 + n65)*(log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r)) + log(4 - 4*r + 3*r^2)) + (n05 + n16 + n50 + n61)*(2*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(3 - 5*r + 5*r^2 - 3*r^3)) + (n02 + n20 + n46 + n64)*(2*log(pmax(1e-6,r)) + log(22 - 48*r + 60*r^2 - 40*r^3 + 15*r^4)) + (n04 + n26 + n40 + n62)*(2*log(pmax(1e-6,1 - r)) + log(9 - 12*r + 30*r^2 - 20*r^3 + 15*r^4)) + (n11 + n55)*(2*log(pmax(1e-6,r)) + log(28 - 48*r + 57*r^2 - 46*r^3 + 18*r^4)) + (n15 + n51)*(2*log(pmax(1e-6,1 - r)) + log(9 + 27*r^2 - 26*r^3 + 18*r^4)) + (n12 + n21 + n45 + n54)*(log(pmax(1e-6,r)) + log(60 - 84*r + 146*r^2 - 166*r^3 + 125*r^4 - 45*r^5)) + (n03 + n30 + n36 + n63)*(log(pmax(1e-6,r)) + log(6 - 16*r + 25*r^2 - 25*r^3 + 15*r^4 - 5*r^5)) + (n23 + n32 + n34 + n43)*log(90 + 15*r - 73*r^2 + 191*r^3 - 283*r^4 + 225*r^5 - 75*r^6) + (n14 + n25 + n41 + n52)*log(36 - 21*r - 67*r^2 + 168*r^3 - 216*r^4 + 145*r^5 - 45*r^6) + (n13 + n31 + n35 + n53)*log(18 - 12*r + 49*r^2 - 104*r^3 + 127*r^4 - 90*r^5 + 30*r^6) + n33*log(153 - 18*r + 80*r^2 - 224*r^3 + 362*r^4 - 300*r^5 + 100*r^6) + (n24 + n42)*log(288 - 144*r + 159*r^2 - 604*r^3 + 938*r^4 - 700*r^5 + 225*r^6) + (n22 + n44)*log(162 + 36*r + 350*r^2 - 648*r^3 + 813*r^4 - 650*r^5 + 225*r^6)
return(L)}
interlogL_m1cm1c <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
optimize(logL_m1cm1c,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m1cm1c <- mapply(interlogL_m1cm1c,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"])
LOD_m1cm1c <- 6*(x[,"n_06"] + x[,"n_60"])*log10(2) + 6*(x[,"n_00"] + x[,"n_66"])*log10(2) - (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(-6*log10(2) + log10(11)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(-6*log10(2) + log10(11)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-5*log10(2) + 2*log10(7) + log10(11)) - (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(-6*log10(2) + log10(11) + log10(13)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(-6*log10(2) + log10(11) + log10(13)) - (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(-6*log10(2) + log10(61)) - (x[,"n_15"] + x[,"n_51"])*(-5*log10(2) + log10(109)) - (x[,"n_11"] + x[,"n_55"])*(-5*log10(2) + log10(109)) - x[,"n_33"]*(-4*log10(2) + log10(19) + log10(127)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-6*log10(2) + log10(5) + log10(257)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-6*log10(2) + log10(5) + log10(257)) - (x[,"n_24"] + x[,"n_42"])*(-6*log10(2) + log10(11) + log10(1283)) - (x[,"n_22"] + x[,"n_44"])*(-6*log10(2) + log10(11) + log10(1283)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-6*log10(2) + log10(5843)) + (x[,"n_06"] + x[,"n_60"])*(4*log10(pmax(1e-6,1 - r_m1cm1c)) + 2*log10(pmax(1e-6,r_m1cm1c))) + (x[,"n_00"] + x[,"n_66"])*(2*log10(pmax(1e-6,1 - r_m1cm1c)) + 4*log10(pmax(1e-6,r_m1cm1c))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(log10(pmax(1e-6,1 - r_m1cm1c)) + 3*log10(pmax(1e-6,r_m1cm1c)) + log10(4 - 4*r_m1cm1c + 3*r_m1cm1c^2)) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(2*log10(pmax(1e-6,1 - r_m1cm1c)) + log10(pmax(1e-6,r_m1cm1c)) + log10(3 - 5*r_m1cm1c + 5*r_m1cm1c^2 - 3*r_m1cm1c^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log10(pmax(1e-6,r_m1cm1c)) + log10(22 - 48*r_m1cm1c + 60*r_m1cm1c^2 - 40*r_m1cm1c^3 + 15*r_m1cm1c^4)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log10(pmax(1e-6,1 - r_m1cm1c)) + log10(9 - 12*r_m1cm1c + 30*r_m1cm1c^2 - 20*r_m1cm1c^3 + 15*r_m1cm1c^4)) + (x[,"n_11"] + x[,"n_55"])*(2*log10(pmax(1e-6,r_m1cm1c)) + log10(28 - 48*r_m1cm1c + 57*r_m1cm1c^2 - 46*r_m1cm1c^3 + 18*r_m1cm1c^4)) + (x[,"n_15"] + x[,"n_51"])*(2*log10(pmax(1e-6,1 - r_m1cm1c)) + log10(9 + 27*r_m1cm1c^2 - 26*r_m1cm1c^3 + 18*r_m1cm1c^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(log10(pmax(1e-6,r_m1cm1c)) + log10(60 - 84*r_m1cm1c + 146*r_m1cm1c^2 - 166*r_m1cm1c^3 + 125*r_m1cm1c^4 - 45*r_m1cm1c^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(log10(pmax(1e-6,r_m1cm1c)) + log10(6 - 16*r_m1cm1c + 25*r_m1cm1c^2 - 25*r_m1cm1c^3 + 15*r_m1cm1c^4 - 5*r_m1cm1c^5)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(90 + 15*r_m1cm1c - 73*r_m1cm1c^2 + 191*r_m1cm1c^3 - 283*r_m1cm1c^4 + 225*r_m1cm1c^5 - 75*r_m1cm1c^6) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*log10(36 - 21*r_m1cm1c - 67*r_m1cm1c^2 + 168*r_m1cm1c^3 - 216*r_m1cm1c^4 + 145*r_m1cm1c^5 - 45*r_m1cm1c^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log10(18 - 12*r_m1cm1c + 49*r_m1cm1c^2 - 104*r_m1cm1c^3 + 127*r_m1cm1c^4 - 90*r_m1cm1c^5 + 30*r_m1cm1c^6) + x[,"n_33"]*log10(153 - 18*r_m1cm1c + 80*r_m1cm1c^2 - 224*r_m1cm1c^3 + 362*r_m1cm1c^4 - 300*r_m1cm1c^5 + 100*r_m1cm1c^6) + (x[,"n_24"] + x[,"n_42"])*log10(288 - 144*r_m1cm1c + 159*r_m1cm1c^2 - 604*r_m1cm1c^3 + 938*r_m1cm1c^4 - 700*r_m1cm1c^5 + 225*r_m1cm1c^6) + (x[,"n_22"] + x[,"n_44"])*log10(162 + 36*r_m1cm1c + 350*r_m1cm1c^2 - 648*r_m1cm1c^3 + 813*r_m1cm1c^4 - 650*r_m1cm1c^5 + 225*r_m1cm1c^6)
logL_m1cm1c <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 2*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"] - 4*x[,"n_60"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 2*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"] - 4*x[,"n_66"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"] - x[,"n_60"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"] - x[,"n_66"])*(2*log(3) + 2*log(5)) + (x[,"n_06"] + x[,"n_60"])*(4*log(pmax(1e-6,1 - r_m1cm1c)) + 2*log(pmax(1e-6,r_m1cm1c))) + (x[,"n_00"] + x[,"n_66"])*(2*log(pmax(1e-6,1 - r_m1cm1c)) + 4*log(pmax(1e-6,r_m1cm1c))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(log(pmax(1e-6,1 - r_m1cm1c)) + 3*log(pmax(1e-6,r_m1cm1c)) + log(4 - 4*r_m1cm1c + 3*r_m1cm1c^2)) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(2*log(pmax(1e-6,1 - r_m1cm1c)) + log(pmax(1e-6,r_m1cm1c)) + log(3 - 5*r_m1cm1c + 5*r_m1cm1c^2 - 3*r_m1cm1c^3)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log(pmax(1e-6,r_m1cm1c)) + log(22 - 48*r_m1cm1c + 60*r_m1cm1c^2 - 40*r_m1cm1c^3 + 15*r_m1cm1c^4)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(2*log(pmax(1e-6,1 - r_m1cm1c)) + log(9 - 12*r_m1cm1c + 30*r_m1cm1c^2 - 20*r_m1cm1c^3 + 15*r_m1cm1c^4)) + (x[,"n_11"] + x[,"n_55"])*(2*log(pmax(1e-6,r_m1cm1c)) + log(28 - 48*r_m1cm1c + 57*r_m1cm1c^2 - 46*r_m1cm1c^3 + 18*r_m1cm1c^4)) + (x[,"n_15"] + x[,"n_51"])*(2*log(pmax(1e-6,1 - r_m1cm1c)) + log(9 + 27*r_m1cm1c^2 - 26*r_m1cm1c^3 + 18*r_m1cm1c^4)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(log(pmax(1e-6,r_m1cm1c)) + log(60 - 84*r_m1cm1c + 146*r_m1cm1c^2 - 166*r_m1cm1c^3 + 125*r_m1cm1c^4 - 45*r_m1cm1c^5)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(log(pmax(1e-6,r_m1cm1c)) + log(6 - 16*r_m1cm1c + 25*r_m1cm1c^2 - 25*r_m1cm1c^3 + 15*r_m1cm1c^4 - 5*r_m1cm1c^5)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(90 + 15*r_m1cm1c - 73*r_m1cm1c^2 + 191*r_m1cm1c^3 - 283*r_m1cm1c^4 + 225*r_m1cm1c^5 - 75*r_m1cm1c^6) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*log(36 - 21*r_m1cm1c - 67*r_m1cm1c^2 + 168*r_m1cm1c^3 - 216*r_m1cm1c^4 + 145*r_m1cm1c^5 - 45*r_m1cm1c^6) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*log(18 - 12*r_m1cm1c + 49*r_m1cm1c^2 - 104*r_m1cm1c^3 + 127*r_m1cm1c^4 - 90*r_m1cm1c^5 + 30*r_m1cm1c^6) + x[,"n_33"]*log(153 - 18*r_m1cm1c + 80*r_m1cm1c^2 - 224*r_m1cm1c^3 + 362*r_m1cm1c^4 - 300*r_m1cm1c^5 + 100*r_m1cm1c^6) + (x[,"n_24"] + x[,"n_42"])*log(288 - 144*r_m1cm1c + 159*r_m1cm1c^2 - 604*r_m1cm1c^3 + 938*r_m1cm1c^4 - 700*r_m1cm1c^5 + 225*r_m1cm1c^6) + (x[,"n_22"] + x[,"n_44"])*log(162 + 36*r_m1cm1c + 350*r_m1cm1c^2 - 648*r_m1cm1c^3 + 813*r_m1cm1c^4 - 650*r_m1cm1c^5 + 225*r_m1cm1c^6)
logL_m1cr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
L <- (-4*n00 - 4*n01 - 4*n02 - 3*n03 - 4*n04 - 4*n05 - 4*n06 - 4*n10 - 2*n11 - 4*n12 - n13 - 4*n14 - 2*n15 - 4*n16 - 4*n20 - 4*n21 - 4*n22 - 3*n23 - 4*n24 - 4*n25 - 4*n26 - 3*n30 - n31 - 3*n32 - n33 - 3*n34 - n35 - 3*n36 - 4*n40 - 4*n41 - 4*n42 - 3*n43 - 4*n44 - 4*n45 - 4*n46 - 4*n50 - 2*n51 - 4*n52 - n53 - 4*n54 - 2*n55 - 4*n56 - 4*n60 - 4*n61 - 4*n62 - 3*n63 - 4*n64 - 4*n65 - 4*n66)*log(2) + (-n00 - n01 - n02 - n03 - n04 - n05 - n06 - n10 - n11 - n12 - n13 - n14 - n15 - n16 - n20 - n21 - n22 - n23 - n24 - n25 - n26 - n30 - n31 - n32 - n33 - n34 - n35 - n36 - n40 - n41 - n42 - n43 - n44 - n45 - n46 - n50 - n51 - n52 - n53 - n54 - n55 - n56 - n60 - n61 - n62 - n63 - n64 - n65 - n66)*(log(3) + 2*log(5)) + (n06 + n60)*(5*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n00 + n66)*(log(pmax(1e-6,1 - r)) + 5*log(pmax(1e-6,r))) + (n01 + n10 + n56 + n65)*(4*log(pmax(1e-6,r)) + log(7 - 10*r + 6*r^2)) + (n05 + n16 + n50 + n61)*(4*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 6*r^2)) + (n03 + n30 + n36 + n63)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(11 - 10*r + 10*r^2)) + (n04 + n26 + n40 + n62)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(13 - 10*r + 15*r^2)) + (n02 + n20 + n46 + n64)*(3*log(pmax(1e-6,r)) + log(18 - 38*r + 35*r^2 - 15*r^3)) + (n11 + n55)*(3*log(pmax(1e-6,r)) + log(12 - 19*r + 19*r^2 - 9*r^3)) + (n15 + n51)*(3*log(pmax(1e-6,1 - r)) + log(3 + 8*r - 8*r^2 + 9*r^3)) + (n12 + n21 + n45 + n54)*(2*log(pmax(1e-6,r)) + log(114 - 216*r + 281*r^2 - 230*r^3 + 90*r^4)) + (n14 + n25 + n41 + n52)*(2*log(pmax(1e-6,1 - r)) + log(39 - 16*r + 131*r^2 - 130*r^3 + 90*r^4)) + (n22 + n44)*(log(pmax(1e-6,r)) + log(234 - 378*r + 696*r^2 - 832*r^3 + 625*r^4 - 225*r^5)) + (n13 + n31 + n35 + n53)*(log(pmax(1e-6,r)) + log(14 - 32*r + 51*r^2 - 63*r^3 + 45*r^4 - 15*r^5)) + (n24 + n42)*log(120 - 13*r - 407*r^2 + 882*r^3 - 1082*r^4 + 725*r^5 - 225*r^6) + n33*log(27 + 7*r - 71*r^2 + 178*r^3 - 214*r^4 + 150*r^5 - 50*r^6) + (n23 + n32 + n34 + n43)*log(69 - 70*r + 259*r^2 - 528*r^3 + 639*r^4 - 450*r^5 + 150*r^6)
return(L)}
interlogL_m1cr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
optimize(logL_m1cr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_m1cr <- mapply(interlogL_m1cr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"])
LOD_m1cr <- 6*(x[,"n_06"] + x[,"n_60"])*log10(2) + 6*(x[,"n_00"] + x[,"n_66"])*log10(2) - (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(-5*log10(2) + log10(7)) - (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(-5*log10(2) + log10(7)) - (x[,"n_15"] + x[,"n_51"])*(-6*log10(2) + 2*log10(7)) - (x[,"n_11"] + x[,"n_55"])*(-6*log10(2) + 2*log10(7)) - (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(-5*log10(2) + log10(17)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-5*log10(2) + 2*log10(5) + log10(17)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-5*log10(2) + 2*log10(5) + log10(17)) - x[,"n_33"]*(-5*log10(2) + log10(19) + log10(43)) - (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(-6*log10(2) + log10(47)) - (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(-6*log10(2) + log10(47)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-6*log10(2) + log10(167)) - (x[,"n_24"] + x[,"n_42"])*(-6*log10(2) + log10(5) + log10(941)) - (x[,"n_22"] + x[,"n_44"])*(-6*log10(2) + log10(5) + log10(941)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-5*log10(2) + log10(1951)) + (x[,"n_06"] + x[,"n_60"])*(5*log10(pmax(1e-6,1 - r_m1cr)) + log10(pmax(1e-6,r_m1cr))) + (x[,"n_00"] + x[,"n_66"])*(log10(pmax(1e-6,1 - r_m1cr)) + 5*log10(pmax(1e-6,r_m1cr))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(4*log10(pmax(1e-6,r_m1cr)) + log10(7 - 10*r_m1cr + 6*r_m1cr^2)) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(4*log10(pmax(1e-6,1 - r_m1cr)) + log10(3 - 2*r_m1cr + 6*r_m1cr^2)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(2*log10(pmax(1e-6,1 - r_m1cr)) + 2*log10(pmax(1e-6,r_m1cr)) + log10(11 - 10*r_m1cr + 10*r_m1cr^2)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(3*log10(pmax(1e-6,1 - r_m1cr)) + log10(pmax(1e-6,r_m1cr)) + log10(13 - 10*r_m1cr + 15*r_m1cr^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(3*log10(pmax(1e-6,r_m1cr)) + log10(18 - 38*r_m1cr + 35*r_m1cr^2 - 15*r_m1cr^3)) + (x[,"n_11"] + x[,"n_55"])*(3*log10(pmax(1e-6,r_m1cr)) + log10(12 - 19*r_m1cr + 19*r_m1cr^2 - 9*r_m1cr^3)) + (x[,"n_15"] + x[,"n_51"])*(3*log10(pmax(1e-6,1 - r_m1cr)) + log10(3 + 8*r_m1cr - 8*r_m1cr^2 + 9*r_m1cr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(2*log10(pmax(1e-6,r_m1cr)) + log10(114 - 216*r_m1cr + 281*r_m1cr^2 - 230*r_m1cr^3 + 90*r_m1cr^4)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(2*log10(pmax(1e-6,1 - r_m1cr)) + log10(39 - 16*r_m1cr + 131*r_m1cr^2 - 130*r_m1cr^3 + 90*r_m1cr^4)) + (x[,"n_22"] + x[,"n_44"])*(log10(pmax(1e-6,r_m1cr)) + log10(234 - 378*r_m1cr + 696*r_m1cr^2 - 832*r_m1cr^3 + 625*r_m1cr^4 - 225*r_m1cr^5)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log10(pmax(1e-6,r_m1cr)) + log10(14 - 32*r_m1cr + 51*r_m1cr^2 - 63*r_m1cr^3 + 45*r_m1cr^4 - 15*r_m1cr^5)) + (x[,"n_24"] + x[,"n_42"])*log10(120 - 13*r_m1cr - 407*r_m1cr^2 + 882*r_m1cr^3 - 1082*r_m1cr^4 + 725*r_m1cr^5 - 225*r_m1cr^6) + x[,"n_33"]*log10(27 + 7*r_m1cr - 71*r_m1cr^2 + 178*r_m1cr^3 - 214*r_m1cr^4 + 150*r_m1cr^5 - 50*r_m1cr^6) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log10(69 - 70*r_m1cr + 259*r_m1cr^2 - 528*r_m1cr^3 + 639*r_m1cr^4 - 450*r_m1cr^5 + 150*r_m1cr^6)
logL_m1cr <- (-4*x[,"n_00"] - 4*x[,"n_01"] - 4*x[,"n_02"] - 3*x[,"n_03"] - 4*x[,"n_04"] - 4*x[,"n_05"] - 4*x[,"n_06"] - 4*x[,"n_10"] - 2*x[,"n_11"] - 4*x[,"n_12"] - x[,"n_13"] - 4*x[,"n_14"] - 2*x[,"n_15"] - 4*x[,"n_16"] - 4*x[,"n_20"] - 4*x[,"n_21"] - 4*x[,"n_22"] - 3*x[,"n_23"] - 4*x[,"n_24"] - 4*x[,"n_25"] - 4*x[,"n_26"] - 3*x[,"n_30"] - x[,"n_31"] - 3*x[,"n_32"] - x[,"n_33"] - 3*x[,"n_34"] - x[,"n_35"] - 3*x[,"n_36"] - 4*x[,"n_40"] - 4*x[,"n_41"] - 4*x[,"n_42"] - 3*x[,"n_43"] - 4*x[,"n_44"] - 4*x[,"n_45"] - 4*x[,"n_46"] - 4*x[,"n_50"] - 2*x[,"n_51"] - 4*x[,"n_52"] - x[,"n_53"] - 4*x[,"n_54"] - 2*x[,"n_55"] - 4*x[,"n_56"] - 4*x[,"n_60"] - 4*x[,"n_61"] - 4*x[,"n_62"] - 3*x[,"n_63"] - 4*x[,"n_64"] - 4*x[,"n_65"] - 4*x[,"n_66"])*log(2) + (-x[,"n_00"] - x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - x[,"n_05"] - x[,"n_06"] - x[,"n_10"] - x[,"n_11"] - x[,"n_12"] - x[,"n_13"] - x[,"n_14"] - x[,"n_15"] - x[,"n_16"] - x[,"n_20"] - x[,"n_21"] - x[,"n_22"] - x[,"n_23"] - x[,"n_24"] - x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - x[,"n_31"] - x[,"n_32"] - x[,"n_33"] - x[,"n_34"] - x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - x[,"n_41"] - x[,"n_42"] - x[,"n_43"] - x[,"n_44"] - x[,"n_45"] - x[,"n_46"] - x[,"n_50"] - x[,"n_51"] - x[,"n_52"] - x[,"n_53"] - x[,"n_54"] - x[,"n_55"] - x[,"n_56"] - x[,"n_60"] - x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - x[,"n_65"] - x[,"n_66"])*(log(3) + 2*log(5)) + (x[,"n_06"] + x[,"n_60"])*(5*log(pmax(1e-6,1 - r_m1cr)) + log(pmax(1e-6,r_m1cr))) + (x[,"n_00"] + x[,"n_66"])*(log(pmax(1e-6,1 - r_m1cr)) + 5*log(pmax(1e-6,r_m1cr))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(4*log(pmax(1e-6,r_m1cr)) + log(7 - 10*r_m1cr + 6*r_m1cr^2)) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(4*log(pmax(1e-6,1 - r_m1cr)) + log(3 - 2*r_m1cr + 6*r_m1cr^2)) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(2*log(pmax(1e-6,1 - r_m1cr)) + 2*log(pmax(1e-6,r_m1cr)) + log(11 - 10*r_m1cr + 10*r_m1cr^2)) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(3*log(pmax(1e-6,1 - r_m1cr)) + log(pmax(1e-6,r_m1cr)) + log(13 - 10*r_m1cr + 15*r_m1cr^2)) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(3*log(pmax(1e-6,r_m1cr)) + log(18 - 38*r_m1cr + 35*r_m1cr^2 - 15*r_m1cr^3)) + (x[,"n_11"] + x[,"n_55"])*(3*log(pmax(1e-6,r_m1cr)) + log(12 - 19*r_m1cr + 19*r_m1cr^2 - 9*r_m1cr^3)) + (x[,"n_15"] + x[,"n_51"])*(3*log(pmax(1e-6,1 - r_m1cr)) + log(3 + 8*r_m1cr - 8*r_m1cr^2 + 9*r_m1cr^3)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(2*log(pmax(1e-6,r_m1cr)) + log(114 - 216*r_m1cr + 281*r_m1cr^2 - 230*r_m1cr^3 + 90*r_m1cr^4)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(2*log(pmax(1e-6,1 - r_m1cr)) + log(39 - 16*r_m1cr + 131*r_m1cr^2 - 130*r_m1cr^3 + 90*r_m1cr^4)) + (x[,"n_22"] + x[,"n_44"])*(log(pmax(1e-6,r_m1cr)) + log(234 - 378*r_m1cr + 696*r_m1cr^2 - 832*r_m1cr^3 + 625*r_m1cr^4 - 225*r_m1cr^5)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(log(pmax(1e-6,r_m1cr)) + log(14 - 32*r_m1cr + 51*r_m1cr^2 - 63*r_m1cr^3 + 45*r_m1cr^4 - 15*r_m1cr^5)) + (x[,"n_24"] + x[,"n_42"])*log(120 - 13*r_m1cr - 407*r_m1cr^2 + 882*r_m1cr^3 - 1082*r_m1cr^4 + 725*r_m1cr^5 - 225*r_m1cr^6) + x[,"n_33"]*log(27 + 7*r_m1cr - 71*r_m1cr^2 + 178*r_m1cr^3 - 214*r_m1cr^4 + 150*r_m1cr^5 - 50*r_m1cr^6) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*log(69 - 70*r_m1cr + 259*r_m1cr^2 - 528*r_m1cr^3 + 639*r_m1cr^4 - 450*r_m1cr^5 + 150*r_m1cr^6)
logL_rr <- function(r,n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
L <- (-4*n00 - 3*n01 - 4*n02 - 2*n03 - 4*n04 - 3*n05 - 4*n06 - 3*n10 - 3*n11 - 3*n12 - 2*n13 - 3*n14 - 3*n15 - 3*n16 - 4*n20 - 3*n21 - 4*n22 - 2*n23 - 4*n24 - 3*n25 - 4*n26 - 2*n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - 2*n36 - 4*n40 - 3*n41 - 4*n42 - 2*n43 - 4*n44 - 3*n45 - 4*n46 - 3*n50 - 3*n51 - 3*n52 - 2*n53 - 3*n54 - 3*n55 - 3*n56 - 4*n60 - 3*n61 - 4*n62 - 2*n63 - 4*n64 - 3*n65 - 4*n66)*log(2) + (n01 + n02 + n04 + n05 + n10 + n11 + n12 + n13 + n14 + n15 + n16 + n20 + n21 + n22 + n23 + n24 + n25 + n26 + n31 + n32 + n34 + n35 + n40 + n41 + n42 + n43 + n44 + n45 + n46 + n50 + n51 + n52 + n53 + n54 + n55 + n56 + n61 + n62 + n64 + n65)*log(3) + (-2*n00 - 2*n01 - n02 - n03 - n04 - 2*n05 - 2*n06 - 2*n10 - 2*n11 - 2*n12 - 2*n13 - 2*n14 - 2*n15 - 2*n16 - n20 - 2*n21 - 2*n22 - 2*n23 - 2*n24 - 2*n25 - n26 - n30 - 2*n31 - 2*n32 - 2*n33 - 2*n34 - 2*n35 - n36 - n40 - 2*n41 - 2*n42 - 2*n43 - 2*n44 - 2*n45 - n46 - 2*n50 - 2*n51 - 2*n52 - 2*n53 - 2*n54 - 2*n55 - 2*n56 - 2*n60 - 2*n61 - n62 - n63 - n64 - 2*n65 - 2*n66)*log(5) + 6*(n06 + n60)*log(pmax(1e-6,1 - r)) + 6*(n00 + n66)*log(pmax(1e-6,r)) + (n05 + n16 + n50 + n61)*(5*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r))) + (n04 + n26 + n40 + n62)*(4*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r))) + (n03 + n30 + n36 + n63)*(3*log(pmax(1e-6,1 - r)) + 3*log(pmax(1e-6,r))) + (n02 + n20 + n46 + n64)*(2*log(pmax(1e-6,1 - r)) + 4*log(pmax(1e-6,r))) + (n01 + n10 + n56 + n65)*(log(pmax(1e-6,1 - r)) + 5*log(pmax(1e-6,r))) + (n11 + n55)*(4*log(pmax(1e-6,r)) + log(7 - 10*r + 6*r^2)) + (n15 + n51)*(4*log(pmax(1e-6,1 - r)) + log(3 - 2*r + 6*r^2)) + (n13 + n31 + n35 + n53)*(2*log(pmax(1e-6,1 - r)) + 2*log(pmax(1e-6,r)) + log(11 - 10*r + 10*r^2)) + (n14 + n25 + n41 + n52)*(3*log(pmax(1e-6,1 - r)) + log(pmax(1e-6,r)) + log(13 - 10*r + 15*r^2)) + (n12 + n21 + n45 + n54)*(3*log(pmax(1e-6,r)) + log(18 - 38*r + 35*r^2 - 15*r^3)) + (n22 + n44)*(2*log(pmax(1e-6,r)) + log(90 - 216*r + 284*r^2 - 200*r^3 + 75*r^4)) + (n24 + n42)*(2*log(pmax(1e-6,1 - r)) + log(33 - 52*r + 134*r^2 - 100*r^3 + 75*r^4)) + (n23 + n32 + n34 + n43)*(log(pmax(1e-6,r)) + log(23 - 67*r + 113*r^2 - 119*r^3 + 75*r^4 - 25*r^5)) + n33*log(41 - 138*r + 336*r^2 - 496*r^3 + 498*r^4 - 300*r^5 + 100*r^6)
return(L)}
interlogL_rr <- function(n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66) {
optimize(logL_rr,c(0,0.5), n00,n01,n02,n03,n04,n05,n06,n10,n11,n12,n13,n14,n15,n16,n20,n21,n22,n23,n24,n25,n26,n30,n31,n32,n33,n34,n35,n36,n40,n41,n42,n43,n44,n45,n46,n50,n51,n52,n53,n54,n55,n56,n60,n61,n62,n63,n64,n65,n66, maximum=TRUE, lower=0, upper=0.5)$maximum}
r_rr <- mapply(interlogL_rr,x[,"n_00"],x[,"n_01"],x[,"n_02"],x[,"n_03"],x[,"n_04"],x[,"n_05"],x[,"n_06"],x[,"n_10"],x[,"n_11"],x[,"n_12"],x[,"n_13"],x[,"n_14"],x[,"n_15"],x[,"n_16"],x[,"n_20"],x[,"n_21"],x[,"n_22"],x[,"n_23"],x[,"n_24"],x[,"n_25"],x[,"n_26"],x[,"n_30"],x[,"n_31"],x[,"n_32"],x[,"n_33"],x[,"n_34"],x[,"n_35"],x[,"n_36"],x[,"n_40"],x[,"n_41"],x[,"n_42"],x[,"n_43"],x[,"n_44"],x[,"n_45"],x[,"n_46"],x[,"n_50"],x[,"n_51"],x[,"n_52"],x[,"n_53"],x[,"n_54"],x[,"n_55"],x[,"n_56"],x[,"n_60"],x[,"n_61"],x[,"n_62"],x[,"n_63"],x[,"n_64"],x[,"n_65"],x[,"n_66"])
LOD_rr <- 6*(x[,"n_06"] + x[,"n_60"])*log10(2) + 6*(x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*log10(2) + 6*(x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*log10(2) + 6*(x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*log10(2) + 6*(x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*log10(2) + 6*(x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*log10(2) + 6*(x[,"n_00"] + x[,"n_66"])*log10(2) - (x[,"n_15"] + x[,"n_51"])*(-5*log10(2) + log10(7)) - (x[,"n_11"] + x[,"n_55"])*(-5*log10(2) + log10(7)) - (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(-5*log10(2) + log10(17)) - (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(-6*log10(2) + log10(7) + log10(31)) - (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(-6*log10(2) + log10(47)) - (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(-6*log10(2) + log10(47)) - x[,"n_33"]*(-4*log10(2) + log10(277)) - (x[,"n_24"] + x[,"n_42"])*(-6*log10(2) + log10(523)) - (x[,"n_22"] + x[,"n_44"])*(-6*log10(2) + log10(523)) + 6*(x[,"n_06"] + x[,"n_60"])*log10(pmax(1e-6,1 - r_rr)) + 6*(x[,"n_00"] + x[,"n_66"])*log10(pmax(1e-6,r_rr)) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(5*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr))) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(4*log10(pmax(1e-6,1 - r_rr)) + 2*log10(pmax(1e-6,r_rr))) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(3*log10(pmax(1e-6,1 - r_rr)) + 3*log10(pmax(1e-6,r_rr))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log10(pmax(1e-6,1 - r_rr)) + 4*log10(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(log10(pmax(1e-6,1 - r_rr)) + 5*log10(pmax(1e-6,r_rr))) + (x[,"n_11"] + x[,"n_55"])*(4*log10(pmax(1e-6,r_rr)) + log10(7 - 10*r_rr + 6*r_rr^2)) + (x[,"n_15"] + x[,"n_51"])*(4*log10(pmax(1e-6,1 - r_rr)) + log10(3 - 2*r_rr + 6*r_rr^2)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(2*log10(pmax(1e-6,1 - r_rr)) + 2*log10(pmax(1e-6,r_rr)) + log10(11 - 10*r_rr + 10*r_rr^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(3*log10(pmax(1e-6,1 - r_rr)) + log10(pmax(1e-6,r_rr)) + log10(13 - 10*r_rr + 15*r_rr^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(3*log10(pmax(1e-6,r_rr)) + log10(18 - 38*r_rr + 35*r_rr^2 - 15*r_rr^3)) + (x[,"n_22"] + x[,"n_44"])*(2*log10(pmax(1e-6,r_rr)) + log10(90 - 216*r_rr + 284*r_rr^2 - 200*r_rr^3 + 75*r_rr^4)) + (x[,"n_24"] + x[,"n_42"])*(2*log10(pmax(1e-6,1 - r_rr)) + log10(33 - 52*r_rr + 134*r_rr^2 - 100*r_rr^3 + 75*r_rr^4)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log10(pmax(1e-6,r_rr)) + log10(23 - 67*r_rr + 113*r_rr^2 - 119*r_rr^3 + 75*r_rr^4 - 25*r_rr^5)) + x[,"n_33"]*log10(41 - 138*r_rr + 336*r_rr^2 - 496*r_rr^3 + 498*r_rr^4 - 300*r_rr^5 + 100*r_rr^6)
logL_rr <- (-4*x[,"n_00"] - 3*x[,"n_01"] - 4*x[,"n_02"] - 2*x[,"n_03"] - 4*x[,"n_04"] - 3*x[,"n_05"] - 4*x[,"n_06"] - 3*x[,"n_10"] - 3*x[,"n_11"] - 3*x[,"n_12"] - 2*x[,"n_13"] - 3*x[,"n_14"] - 3*x[,"n_15"] - 3*x[,"n_16"] - 4*x[,"n_20"] - 3*x[,"n_21"] - 4*x[,"n_22"] - 2*x[,"n_23"] - 4*x[,"n_24"] - 3*x[,"n_25"] - 4*x[,"n_26"] - 2*x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - 2*x[,"n_36"] - 4*x[,"n_40"] - 3*x[,"n_41"] - 4*x[,"n_42"] - 2*x[,"n_43"] - 4*x[,"n_44"] - 3*x[,"n_45"] - 4*x[,"n_46"] - 3*x[,"n_50"] - 3*x[,"n_51"] - 3*x[,"n_52"] - 2*x[,"n_53"] - 3*x[,"n_54"] - 3*x[,"n_55"] - 3*x[,"n_56"] - 4*x[,"n_60"] - 3*x[,"n_61"] - 4*x[,"n_62"] - 2*x[,"n_63"] - 4*x[,"n_64"] - 3*x[,"n_65"] - 4*x[,"n_66"])*log(2) + (x[,"n_01"] + x[,"n_02"] + x[,"n_04"] + x[,"n_05"] + x[,"n_10"] + x[,"n_11"] + x[,"n_12"] + x[,"n_13"] + x[,"n_14"] + x[,"n_15"] + x[,"n_16"] + x[,"n_20"] + x[,"n_21"] + x[,"n_22"] + x[,"n_23"] + x[,"n_24"] + x[,"n_25"] + x[,"n_26"] + x[,"n_31"] + x[,"n_32"] + x[,"n_34"] + x[,"n_35"] + x[,"n_40"] + x[,"n_41"] + x[,"n_42"] + x[,"n_43"] + x[,"n_44"] + x[,"n_45"] + x[,"n_46"] + x[,"n_50"] + x[,"n_51"] + x[,"n_52"] + x[,"n_53"] + x[,"n_54"] + x[,"n_55"] + x[,"n_56"] + x[,"n_61"] + x[,"n_62"] + x[,"n_64"] + x[,"n_65"])*log(3) + (-2*x[,"n_00"] - 2*x[,"n_01"] - x[,"n_02"] - x[,"n_03"] - x[,"n_04"] - 2*x[,"n_05"] - 2*x[,"n_06"] - 2*x[,"n_10"] - 2*x[,"n_11"] - 2*x[,"n_12"] - 2*x[,"n_13"] - 2*x[,"n_14"] - 2*x[,"n_15"] - 2*x[,"n_16"] - x[,"n_20"] - 2*x[,"n_21"] - 2*x[,"n_22"] - 2*x[,"n_23"] - 2*x[,"n_24"] - 2*x[,"n_25"] - x[,"n_26"] - x[,"n_30"] - 2*x[,"n_31"] - 2*x[,"n_32"] - 2*x[,"n_33"] - 2*x[,"n_34"] - 2*x[,"n_35"] - x[,"n_36"] - x[,"n_40"] - 2*x[,"n_41"] - 2*x[,"n_42"] - 2*x[,"n_43"] - 2*x[,"n_44"] - 2*x[,"n_45"] - x[,"n_46"] - 2*x[,"n_50"] - 2*x[,"n_51"] - 2*x[,"n_52"] - 2*x[,"n_53"] - 2*x[,"n_54"] - 2*x[,"n_55"] - 2*x[,"n_56"] - 2*x[,"n_60"] - 2*x[,"n_61"] - x[,"n_62"] - x[,"n_63"] - x[,"n_64"] - 2*x[,"n_65"] - 2*x[,"n_66"])*log(5) + 6*(x[,"n_06"] + x[,"n_60"])*log(pmax(1e-6,1 - r_rr)) + 6*(x[,"n_00"] + x[,"n_66"])*log(pmax(1e-6,r_rr)) + (x[,"n_05"] + x[,"n_16"] + x[,"n_50"] + x[,"n_61"])*(5*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr))) + (x[,"n_04"] + x[,"n_26"] + x[,"n_40"] + x[,"n_62"])*(4*log(pmax(1e-6,1 - r_rr)) + 2*log(pmax(1e-6,r_rr))) + (x[,"n_03"] + x[,"n_30"] + x[,"n_36"] + x[,"n_63"])*(3*log(pmax(1e-6,1 - r_rr)) + 3*log(pmax(1e-6,r_rr))) + (x[,"n_02"] + x[,"n_20"] + x[,"n_46"] + x[,"n_64"])*(2*log(pmax(1e-6,1 - r_rr)) + 4*log(pmax(1e-6,r_rr))) + (x[,"n_01"] + x[,"n_10"] + x[,"n_56"] + x[,"n_65"])*(log(pmax(1e-6,1 - r_rr)) + 5*log(pmax(1e-6,r_rr))) + (x[,"n_11"] + x[,"n_55"])*(4*log(pmax(1e-6,r_rr)) + log(7 - 10*r_rr + 6*r_rr^2)) + (x[,"n_15"] + x[,"n_51"])*(4*log(pmax(1e-6,1 - r_rr)) + log(3 - 2*r_rr + 6*r_rr^2)) + (x[,"n_13"] + x[,"n_31"] + x[,"n_35"] + x[,"n_53"])*(2*log(pmax(1e-6,1 - r_rr)) + 2*log(pmax(1e-6,r_rr)) + log(11 - 10*r_rr + 10*r_rr^2)) + (x[,"n_14"] + x[,"n_25"] + x[,"n_41"] + x[,"n_52"])*(3*log(pmax(1e-6,1 - r_rr)) + log(pmax(1e-6,r_rr)) + log(13 - 10*r_rr + 15*r_rr^2)) + (x[,"n_12"] + x[,"n_21"] + x[,"n_45"] + x[,"n_54"])*(3*log(pmax(1e-6,r_rr)) + log(18 - 38*r_rr + 35*r_rr^2 - 15*r_rr^3)) + (x[,"n_22"] + x[,"n_44"])*(2*log(pmax(1e-6,r_rr)) + log(90 - 216*r_rr + 284*r_rr^2 - 200*r_rr^3 + 75*r_rr^4)) + (x[,"n_24"] + x[,"n_42"])*(2*log(pmax(1e-6,1 - r_rr)) + log(33 - 52*r_rr + 134*r_rr^2 - 100*r_rr^3 + 75*r_rr^4)) + (x[,"n_23"] + x[,"n_32"] + x[,"n_34"] + x[,"n_43"])*(log(pmax(1e-6,r_rr)) + log(23 - 67*r_rr + 113*r_rr^2 - 119*r_rr^3 + 75*r_rr^4 - 25*r_rr^5)) + x[,"n_33"]*log(41 - 138*r_rr + 336*r_rr^2 - 496*r_rr^3 + 498*r_rr^4 - 300*r_rr^5 + 100*r_rr^6)

return(list(
r_mat = cbind(r_cc,r_cm2c,r_cm1c,r_cr,r_m2cm2c,r_m2cm1c,r_m2cr,r_m1cm1c,r_m1cr,r_rr,0.499),
LOD_mat = cbind(LOD_cc,LOD_cm2c,LOD_cm1c,LOD_cr,LOD_m2cm2c,LOD_m2cm1c,LOD_m2cr,LOD_m1cm1c,LOD_m1cr,LOD_rr,0),
logL_mat = cbind(logL_cc,logL_cm2c,logL_cm1c,logL_cr,logL_m2cm2c,logL_m2cm1c,logL_m2cr,logL_m1cm1c,logL_m1cr,logL_rr,-1e6),
phasing_strategy = "MLL",
possible_phases = c("coupling coupling","coupling mixed2c","coupling mixed1c","coupling repulsion","mixed2c mixed2c","mixed2c mixed1c","mixed2c repulsion","mixed1c mixed1c","mixed1c repulsion","repulsion repulsion","unknown")
)
)
}


#' @rdname r6_functions 
#' @noRd
p6_1.0_1.0 <- function(x){
  r_c <- (x[,"n_01"] + x[,"n_10"])/(x[,"n_00"] + x[,"n_01"] + x[,"n_10"] + x[,"n_11"])
  logL_c <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_10"] - x[,"n_11"])*log(2) + (x[,"n_00"] + x[,"n_11"])*log(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_10"])*log(pmax(1e-6,r_c))
  LOD_c <- (x[,"n_01"] + x[,"n_10"])*log10(2) + (x[,"n_00"] + x[,"n_11"])*log10(2) + (x[,"n_00"] + x[,"n_11"])*log10(pmax(1e-6,1 - r_c)) + (x[,"n_01"] + x[,"n_10"])*log10(pmax(1e-6,r_c))
  r_r <- (x[,"n_11"] + x[,"n_00"])/(x[,"n_00"] + x[,"n_01"] + x[,"n_10"] + x[,"n_11"])
  logL_r <- (-x[,"n_00"] - x[,"n_01"] - x[,"n_10"] - x[,"n_11"])*log(2) + (x[,"n_01"] + x[,"n_10"])*log(pmax(1e-6,1 - r_r)) + (x[,"n_11"] + x[,"n_00"])*log(pmax(1e-6,r_r))
  LOD_r <- (x[,"n_11"] + x[,"n_00"])*log10(2) + (x[,"n_01"] + x[,"n_10"])*log10(2) + (x[,"n_01"] + x[,"n_10"])*log10(pmax(1e-6,1 - r_r)) + (x[,"n_00"] + x[,"n_11"])*log10(pmax(1e-6,r_r))
  
  return(list(
    r_mat = cbind(r_c,r_r,0.499),
    LOD_mat = cbind(LOD_c,LOD_r,0),
    logL_mat = cbind(logL_c,logL_r,-1e6),
    phasing_strategy = "MLL",
    possible_phases = c("coupling","repulsion","unknown")
  )
  )
}

